/**************************************************************************

                    =====  JUNO codes  =====

   Copyright (C) 2019 California Institute of Technology (Caltech)
   All rights reserved.

   Written by Fabiano Oyafuso (fabiano@jpl.nasa.gov)

**************************************************************************/

#ifndef HDR_ATMOSPHERE_H2
#define HDR_ATMOSPHERE_H2

#include <algorithm>

struct CIA_Orton {
  static const double wn[];
  static const double logT[10];
  static const double H2H2_e[][10];
  static const double H2H2_n[][10];
  static const double H2He_e[][10];
  static const double H2He_n[][10];
  static const double H2CH4_e[][10];
  static const double H2CH4_n[][10];
  /*
  static const double N2N2[][10];
  static const double N2H2[][10];
  static const double N2CH4[][10];
  static const double CH4CH4[][10];
  */

  static double CIA_H2_H2_e(int It0, int Inu0, double nu, double logT0) {
    return interp(It0, Inu0, nu, logT0, H2H2_e);
  }
  static double CIA_H2_H2_n(int It0, int Inu0, double nu, double logT0) {
    return interp(It0, Inu0, nu, logT0, H2H2_n);
  }
  static double CIA_H2_He_e(int It0, int Inu0, double nu, double logT0) {
    return interp(It0, Inu0, nu, logT0, H2He_e);
  }
  static double CIA_H2_He_n(int It0, int Inu0, double nu, double logT0) {
    return interp(It0, Inu0, nu, logT0, H2He_n);
  }
  static double CIA_H2_CH4_e(int It0, int Inu0, double nu, double logT0) {
    return interp(It0, Inu0, nu, logT0, H2CH4_e);
  }
  static double CIA_H2_CH4_n(int It0, int Inu0, double nu, double logT0) {
    return interp(It0, Inu0, nu, logT0, H2CH4_n);
  }

  static int find_index_wavenumber(double nu) {
    return (std::lower_bound(wn, wn + 2428 - 1, nu) - wn - 1);
  }
  static int find_index_temp(double logT0) {
    if (logT0 < logT[9])
      return (std::lower_bound(logT, logT + 9, logT0) - logT - 1);
    else
      return 8;
  }

 private:
  static double interp(int It0, int Inu0, double nu, double logT0,
                       double const f[][10]);
};

inline double CIA_Orton::interp(int It0, int Inu0, double nu, double logT0,
                                double const f[][10]) {
  int const It1 = It0 + 1;
  int const Inu1 = Inu0 + 1;

  double df_dT0 = (f[Inu0][It1] - f[Inu0][It0]) / (logT[It1] - logT[It0]);
  double g0 = df_dT0 * logT0 + (f[Inu0][It1] - df_dT0 * logT[It1]);
  double df_dT1 = (f[Inu1][It1] - f[Inu1][It0]) / (logT[It1] - logT[It0]);
  double g1 = df_dT1 * logT0 + (f[Inu1][It1] - df_dT1 * logT[It1]);
  double m = (g1 - g0) / (log(wn[Inu1] / wn[Inu0]));

  return m * log(nu) + (g1 - m * log(wn[Inu1]));
};

// temperature grid, linear in log T from 40K to 400K
const double CIA_Orton::logT[10] = {
    3.688879454113936e+00, 3.944722242224386e+00, 4.200565030334835e+00,
    4.456407818445284e+00, 4.712250606555735e+00, 4.968093394666184e+00,
    5.223936182776633e+00, 5.479778970887082e+00, 5.735621758997532e+00,
    5.991464547107982e+00};

const double CIA_Orton::wn[] = {
    0.0200,    0.0500,    0.1000,    0.2000,    0.3000,    0.4000,    0.5000,
    0.6000,    0.7000,    0.8000,    0.9000,    1.0000,    1.1000,    1.2000,
    1.3000,    1.4000,    1.5000,    1.6000,    1.7000,    1.8000,    1.9000,
    2.0000,    2.5000,    3.0000,    3.5000,    4.0000,    4.5000,    5.0000,
    5.5000,    6.0000,    6.5000,    7.0000,    7.5000,    8.0000,    8.5000,
    9.0000,    9.5000,    10.0000,   11.0000,   12.0000,   13.0000,   14.0000,
    15.0000,   16.0000,   17.0000,   18.0000,   19.0000,   20.0000,   21.0000,
    22.0000,   23.0000,   24.0000,   25.0000,   26.0000,   27.0000,   28.0000,
    29.0000,   30.0000,   31.0000,   32.0000,   33.0000,   34.0000,   35.0000,
    36.0000,   37.0000,   38.0000,   39.0000,   40.0000,   41.0000,   42.0000,
    43.0000,   44.0000,   45.0000,   46.0000,   47.0000,   48.0000,   49.0000,
    50.0000,   51.0000,   52.0000,   53.0000,   54.0000,   55.0000,   56.0000,
    57.0000,   58.0000,   59.0000,   60.0000,   61.0000,   62.0000,   63.0000,
    64.0000,   65.0000,   66.0000,   67.0000,   68.0000,   69.0000,   70.0000,
    71.0000,   72.0000,   73.0000,   74.0000,   75.0000,   76.0000,   77.0000,
    78.0000,   79.0000,   80.0000,   81.0000,   82.0000,   83.0000,   84.0000,
    85.0000,   86.0000,   87.0000,   88.0000,   89.0000,   90.0000,   91.0000,
    92.0000,   93.0000,   94.0000,   95.0000,   96.0000,   97.0000,   98.0000,
    99.0000,   100.0000,  101.0000,  102.0000,  103.0000,  104.0000,  105.0000,
    106.0000,  107.0000,  108.0000,  109.0000,  110.0000,  111.0000,  112.0000,
    113.0000,  114.0000,  115.0000,  116.0000,  117.0000,  118.0000,  119.0000,
    120.0000,  121.0000,  122.0000,  123.0000,  124.0000,  125.0000,  126.0000,
    127.0000,  128.0000,  129.0000,  130.0000,  131.0000,  132.0000,  133.0000,
    134.0000,  135.0000,  136.0000,  137.0000,  138.0000,  139.0000,  140.0000,
    141.0000,  142.0000,  143.0000,  144.0000,  145.0000,  146.0000,  147.0000,
    148.0000,  149.0000,  150.0000,  151.0000,  152.0000,  153.0000,  154.0000,
    155.0000,  156.0000,  157.0000,  158.0000,  159.0000,  160.0000,  161.0000,
    162.0000,  163.0000,  164.0000,  165.0000,  166.0000,  167.0000,  168.0000,
    169.0000,  170.0000,  171.0000,  172.0000,  173.0000,  174.0000,  175.0000,
    176.0000,  177.0000,  178.0000,  179.0000,  180.0000,  181.0000,  182.0000,
    183.0000,  184.0000,  185.0000,  186.0000,  187.0000,  188.0000,  189.0000,
    190.0000,  191.0000,  192.0000,  193.0000,  194.0000,  195.0000,  196.0000,
    197.0000,  198.0000,  199.0000,  200.0000,  201.0000,  202.0000,  203.0000,
    204.0000,  205.0000,  206.0000,  207.0000,  208.0000,  209.0000,  210.0000,
    211.0000,  212.0000,  213.0000,  214.0000,  215.0000,  216.0000,  217.0000,
    218.0000,  219.0000,  220.0000,  221.0000,  222.0000,  223.0000,  224.0000,
    225.0000,  226.0000,  227.0000,  228.0000,  229.0000,  230.0000,  231.0000,
    232.0000,  233.0000,  234.0000,  235.0000,  236.0000,  237.0000,  238.0000,
    239.0000,  240.0000,  241.0000,  242.0000,  243.0000,  244.0000,  245.0000,
    246.0000,  247.0000,  248.0000,  249.0000,  250.0000,  251.0000,  252.0000,
    253.0000,  254.0000,  255.0000,  256.0000,  257.0000,  258.0000,  259.0000,
    260.0000,  261.0000,  262.0000,  263.0000,  264.0000,  265.0000,  266.0000,
    267.0000,  268.0000,  269.0000,  270.0000,  271.0000,  272.0000,  273.0000,
    274.0000,  275.0000,  276.0000,  277.0000,  278.0000,  279.0000,  280.0000,
    281.0000,  282.0000,  283.0000,  284.0000,  285.0000,  286.0000,  287.0000,
    288.0000,  289.0000,  290.0000,  291.0000,  292.0000,  293.0000,  294.0000,
    295.0000,  296.0000,  297.0000,  298.0000,  299.0000,  300.0000,  301.0000,
    302.0000,  303.0000,  304.0000,  305.0000,  306.0000,  307.0000,  308.0000,
    309.0000,  310.0000,  311.0000,  312.0000,  313.0000,  314.0000,  315.0000,
    316.0000,  317.0000,  318.0000,  319.0000,  320.0000,  321.0000,  322.0000,
    323.0000,  324.0000,  325.0000,  326.0000,  327.0000,  328.0000,  329.0000,
    330.0000,  331.0000,  332.0000,  333.0000,  334.0000,  335.0000,  336.0000,
    337.0000,  338.0000,  339.0000,  340.0000,  341.0000,  342.0000,  343.0000,
    344.0000,  345.0000,  346.0000,  347.0000,  348.0000,  349.0000,  350.0000,
    351.0000,  352.0000,  353.0000,  354.0000,  355.0000,  356.0000,  357.0000,
    358.0000,  359.0000,  360.0000,  361.0000,  362.0000,  363.0000,  364.0000,
    365.0000,  366.0000,  367.0000,  368.0000,  369.0000,  370.0000,  371.0000,
    372.0000,  373.0000,  374.0000,  375.0000,  376.0000,  377.0000,  378.0000,
    379.0000,  380.0000,  381.0000,  382.0000,  383.0000,  384.0000,  385.0000,
    386.0000,  387.0000,  388.0000,  389.0000,  390.0000,  391.0000,  392.0000,
    393.0000,  394.0000,  395.0000,  396.0000,  397.0000,  398.0000,  399.0000,
    400.0000,  401.0000,  402.0000,  403.0000,  404.0000,  405.0000,  406.0000,
    407.0000,  408.0000,  409.0000,  410.0000,  411.0000,  412.0000,  413.0000,
    414.0000,  415.0000,  416.0000,  417.0000,  418.0000,  419.0000,  420.0000,
    421.0000,  422.0000,  423.0000,  424.0000,  425.0000,  426.0000,  427.0000,
    428.0000,  429.0000,  430.0000,  431.0000,  432.0000,  433.0000,  434.0000,
    435.0000,  436.0000,  437.0000,  438.0000,  439.0000,  440.0000,  441.0000,
    442.0000,  443.0000,  444.0000,  445.0000,  446.0000,  447.0000,  448.0000,
    449.0000,  450.0000,  451.0000,  452.0000,  453.0000,  454.0000,  455.0000,
    456.0000,  457.0000,  458.0000,  459.0000,  460.0000,  461.0000,  462.0000,
    463.0000,  464.0000,  465.0000,  466.0000,  467.0000,  468.0000,  469.0000,
    470.0000,  471.0000,  472.0000,  473.0000,  474.0000,  475.0000,  476.0000,
    477.0000,  478.0000,  479.0000,  480.0000,  481.0000,  482.0000,  483.0000,
    484.0000,  485.0000,  486.0000,  487.0000,  488.0000,  489.0000,  490.0000,
    491.0000,  492.0000,  493.0000,  494.0000,  495.0000,  496.0000,  497.0000,
    498.0000,  499.0000,  500.0000,  501.0000,  502.0000,  503.0000,  504.0000,
    505.0000,  506.0000,  507.0000,  508.0000,  509.0000,  510.0000,  511.0000,
    512.0000,  513.0000,  514.0000,  515.0000,  516.0000,  517.0000,  518.0000,
    519.0000,  520.0000,  521.0000,  522.0000,  523.0000,  524.0000,  525.0000,
    526.0000,  527.0000,  528.0000,  529.0000,  530.0000,  531.0000,  532.0000,
    533.0000,  534.0000,  535.0000,  536.0000,  537.0000,  538.0000,  539.0000,
    540.0000,  541.0000,  542.0000,  543.0000,  544.0000,  545.0000,  546.0000,
    547.0000,  548.0000,  549.0000,  550.0000,  551.0000,  552.0000,  553.0000,
    554.0000,  555.0000,  556.0000,  557.0000,  558.0000,  559.0000,  560.0000,
    561.0000,  562.0000,  563.0000,  564.0000,  565.0000,  566.0000,  567.0000,
    568.0000,  569.0000,  570.0000,  571.0000,  572.0000,  573.0000,  574.0000,
    575.0000,  576.0000,  577.0000,  578.0000,  579.0000,  580.0000,  581.0000,
    582.0000,  583.0000,  584.0000,  585.0000,  586.0000,  587.0000,  588.0000,
    589.0000,  590.0000,  591.0000,  592.0000,  593.0000,  594.0000,  595.0000,
    596.0000,  597.0000,  598.0000,  599.0000,  600.0000,  601.0000,  602.0000,
    603.0000,  604.0000,  605.0000,  606.0000,  607.0000,  608.0000,  609.0000,
    610.0000,  611.0000,  612.0000,  613.0000,  614.0000,  615.0000,  616.0000,
    617.0000,  618.0000,  619.0000,  620.0000,  621.0000,  622.0000,  623.0000,
    624.0000,  625.0000,  626.0000,  627.0000,  628.0000,  629.0000,  630.0000,
    631.0000,  632.0000,  633.0000,  634.0000,  635.0000,  636.0000,  637.0000,
    638.0000,  639.0000,  640.0000,  641.0000,  642.0000,  643.0000,  644.0000,
    645.0000,  646.0000,  647.0000,  648.0000,  649.0000,  650.0000,  651.0000,
    652.0000,  653.0000,  654.0000,  655.0000,  656.0000,  657.0000,  658.0000,
    659.0000,  660.0000,  661.0000,  662.0000,  663.0000,  664.0000,  665.0000,
    666.0000,  667.0000,  668.0000,  669.0000,  670.0000,  671.0000,  672.0000,
    673.0000,  674.0000,  675.0000,  676.0000,  677.0000,  678.0000,  679.0000,
    680.0000,  681.0000,  682.0000,  683.0000,  684.0000,  685.0000,  686.0000,
    687.0000,  688.0000,  689.0000,  690.0000,  691.0000,  692.0000,  693.0000,
    694.0000,  695.0000,  696.0000,  697.0000,  698.0000,  699.0000,  700.0000,
    701.0000,  702.0000,  703.0000,  704.0000,  705.0000,  706.0000,  707.0000,
    708.0000,  709.0000,  710.0000,  711.0000,  712.0000,  713.0000,  714.0000,
    715.0000,  716.0000,  717.0000,  718.0000,  719.0000,  720.0000,  721.0000,
    722.0000,  723.0000,  724.0000,  725.0000,  726.0000,  727.0000,  728.0000,
    729.0000,  730.0000,  731.0000,  732.0000,  733.0000,  734.0000,  735.0000,
    736.0000,  737.0000,  738.0000,  739.0000,  740.0000,  741.0000,  742.0000,
    743.0000,  744.0000,  745.0000,  746.0000,  747.0000,  748.0000,  749.0000,
    750.0000,  751.0000,  752.0000,  753.0000,  754.0000,  755.0000,  756.0000,
    757.0000,  758.0000,  759.0000,  760.0000,  761.0000,  762.0000,  763.0000,
    764.0000,  765.0000,  766.0000,  767.0000,  768.0000,  769.0000,  770.0000,
    771.0000,  772.0000,  773.0000,  774.0000,  775.0000,  776.0000,  777.0000,
    778.0000,  779.0000,  780.0000,  781.0000,  782.0000,  783.0000,  784.0000,
    785.0000,  786.0000,  787.0000,  788.0000,  789.0000,  790.0000,  791.0000,
    792.0000,  793.0000,  794.0000,  795.0000,  796.0000,  797.0000,  798.0000,
    799.0000,  800.0000,  801.0000,  802.0000,  803.0000,  804.0000,  805.0000,
    806.0000,  807.0000,  808.0000,  809.0000,  810.0000,  811.0000,  812.0000,
    813.0000,  814.0000,  815.0000,  816.0000,  817.0000,  818.0000,  819.0000,
    820.0000,  821.0000,  822.0000,  823.0000,  824.0000,  825.0000,  826.0000,
    827.0000,  828.0000,  829.0000,  830.0000,  831.0000,  832.0000,  833.0000,
    834.0000,  835.0000,  836.0000,  837.0000,  838.0000,  839.0000,  840.0000,
    841.0000,  842.0000,  843.0000,  844.0000,  845.0000,  846.0000,  847.0000,
    848.0000,  849.0000,  850.0000,  851.0000,  852.0000,  853.0000,  854.0000,
    855.0000,  856.0000,  857.0000,  858.0000,  859.0000,  860.0000,  861.0000,
    862.0000,  863.0000,  864.0000,  865.0000,  866.0000,  867.0000,  868.0000,
    869.0000,  870.0000,  871.0000,  872.0000,  873.0000,  874.0000,  875.0000,
    876.0000,  877.0000,  878.0000,  879.0000,  880.0000,  881.0000,  882.0000,
    883.0000,  884.0000,  885.0000,  886.0000,  887.0000,  888.0000,  889.0000,
    890.0000,  891.0000,  892.0000,  893.0000,  894.0000,  895.0000,  896.0000,
    897.0000,  898.0000,  899.0000,  900.0000,  901.0000,  902.0000,  903.0000,
    904.0000,  905.0000,  906.0000,  907.0000,  908.0000,  909.0000,  910.0000,
    911.0000,  912.0000,  913.0000,  914.0000,  915.0000,  916.0000,  917.0000,
    918.0000,  919.0000,  920.0000,  921.0000,  922.0000,  923.0000,  924.0000,
    925.0000,  926.0000,  927.0000,  928.0000,  929.0000,  930.0000,  931.0000,
    932.0000,  933.0000,  934.0000,  935.0000,  936.0000,  937.0000,  938.0000,
    939.0000,  940.0000,  941.0000,  942.0000,  943.0000,  944.0000,  945.0000,
    946.0000,  947.0000,  948.0000,  949.0000,  950.0000,  951.0000,  952.0000,
    953.0000,  954.0000,  955.0000,  956.0000,  957.0000,  958.0000,  959.0000,
    960.0000,  961.0000,  962.0000,  963.0000,  964.0000,  965.0000,  966.0000,
    967.0000,  968.0000,  969.0000,  970.0000,  971.0000,  972.0000,  973.0000,
    974.0000,  975.0000,  976.0000,  977.0000,  978.0000,  979.0000,  980.0000,
    981.0000,  982.0000,  983.0000,  984.0000,  985.0000,  986.0000,  987.0000,
    988.0000,  989.0000,  990.0000,  991.0000,  992.0000,  993.0000,  994.0000,
    995.0000,  996.0000,  997.0000,  998.0000,  999.0000,  1000.0000, 1001.0000,
    1002.0000, 1003.0000, 1004.0000, 1005.0000, 1006.0000, 1007.0000, 1008.0000,
    1009.0000, 1010.0000, 1011.0000, 1012.0000, 1013.0000, 1014.0000, 1015.0000,
    1016.0000, 1017.0000, 1018.0000, 1019.0000, 1020.0000, 1021.0000, 1022.0000,
    1023.0000, 1024.0000, 1025.0000, 1026.0000, 1027.0000, 1028.0000, 1029.0000,
    1030.0000, 1031.0000, 1032.0000, 1033.0000, 1034.0000, 1035.0000, 1036.0000,
    1037.0000, 1038.0000, 1039.0000, 1040.0000, 1041.0000, 1042.0000, 1043.0000,
    1044.0000, 1045.0000, 1046.0000, 1047.0000, 1048.0000, 1049.0000, 1050.0000,
    1051.0000, 1052.0000, 1053.0000, 1054.0000, 1055.0000, 1056.0000, 1057.0000,
    1058.0000, 1059.0000, 1060.0000, 1061.0000, 1062.0000, 1063.0000, 1064.0000,
    1065.0000, 1066.0000, 1067.0000, 1068.0000, 1069.0000, 1070.0000, 1071.0000,
    1072.0000, 1073.0000, 1074.0000, 1075.0000, 1076.0000, 1077.0000, 1078.0000,
    1079.0000, 1080.0000, 1081.0000, 1082.0000, 1083.0000, 1084.0000, 1085.0000,
    1086.0000, 1087.0000, 1088.0000, 1089.0000, 1090.0000, 1091.0000, 1092.0000,
    1093.0000, 1094.0000, 1095.0000, 1096.0000, 1097.0000, 1098.0000, 1099.0000,
    1100.0000, 1101.0000, 1102.0000, 1103.0000, 1104.0000, 1105.0000, 1106.0000,
    1107.0000, 1108.0000, 1109.0000, 1110.0000, 1111.0000, 1112.0000, 1113.0000,
    1114.0000, 1115.0000, 1116.0000, 1117.0000, 1118.0000, 1119.0000, 1120.0000,
    1121.0000, 1122.0000, 1123.0000, 1124.0000, 1125.0000, 1126.0000, 1127.0000,
    1128.0000, 1129.0000, 1130.0000, 1131.0000, 1132.0000, 1133.0000, 1134.0000,
    1135.0000, 1136.0000, 1137.0000, 1138.0000, 1139.0000, 1140.0000, 1141.0000,
    1142.0000, 1143.0000, 1144.0000, 1145.0000, 1146.0000, 1147.0000, 1148.0000,
    1149.0000, 1150.0000, 1151.0000, 1152.0000, 1153.0000, 1154.0000, 1155.0000,
    1156.0000, 1157.0000, 1158.0000, 1159.0000, 1160.0000, 1161.0000, 1162.0000,
    1163.0000, 1164.0000, 1165.0000, 1166.0000, 1167.0000, 1168.0000, 1169.0000,
    1170.0000, 1171.0000, 1172.0000, 1173.0000, 1174.0000, 1175.0000, 1176.0000,
    1177.0000, 1178.0000, 1179.0000, 1180.0000, 1181.0000, 1182.0000, 1183.0000,
    1184.0000, 1185.0000, 1186.0000, 1187.0000, 1188.0000, 1189.0000, 1190.0000,
    1191.0000, 1192.0000, 1193.0000, 1194.0000, 1195.0000, 1196.0000, 1197.0000,
    1198.0000, 1199.0000, 1200.0000, 1201.0000, 1202.0000, 1203.0000, 1204.0000,
    1205.0000, 1206.0000, 1207.0000, 1208.0000, 1209.0000, 1210.0000, 1211.0000,
    1212.0000, 1213.0000, 1214.0000, 1215.0000, 1216.0000, 1217.0000, 1218.0000,
    1219.0000, 1220.0000, 1221.0000, 1222.0000, 1223.0000, 1224.0000, 1225.0000,
    1226.0000, 1227.0000, 1228.0000, 1229.0000, 1230.0000, 1231.0000, 1232.0000,
    1233.0000, 1234.0000, 1235.0000, 1236.0000, 1237.0000, 1238.0000, 1239.0000,
    1240.0000, 1241.0000, 1242.0000, 1243.0000, 1244.0000, 1245.0000, 1246.0000,
    1247.0000, 1248.0000, 1249.0000, 1250.0000, 1251.0000, 1252.0000, 1253.0000,
    1254.0000, 1255.0000, 1256.0000, 1257.0000, 1258.0000, 1259.0000, 1260.0000,
    1261.0000, 1262.0000, 1263.0000, 1264.0000, 1265.0000, 1266.0000, 1267.0000,
    1268.0000, 1269.0000, 1270.0000, 1271.0000, 1272.0000, 1273.0000, 1274.0000,
    1275.0000, 1276.0000, 1277.0000, 1278.0000, 1279.0000, 1280.0000, 1281.0000,
    1282.0000, 1283.0000, 1284.0000, 1285.0000, 1286.0000, 1287.0000, 1288.0000,
    1289.0000, 1290.0000, 1291.0000, 1292.0000, 1293.0000, 1294.0000, 1295.0000,
    1296.0000, 1297.0000, 1298.0000, 1299.0000, 1300.0000, 1301.0000, 1302.0000,
    1303.0000, 1304.0000, 1305.0000, 1306.0000, 1307.0000, 1308.0000, 1309.0000,
    1310.0000, 1311.0000, 1312.0000, 1313.0000, 1314.0000, 1315.0000, 1316.0000,
    1317.0000, 1318.0000, 1319.0000, 1320.0000, 1321.0000, 1322.0000, 1323.0000,
    1324.0000, 1325.0000, 1326.0000, 1327.0000, 1328.0000, 1329.0000, 1330.0000,
    1331.0000, 1332.0000, 1333.0000, 1334.0000, 1335.0000, 1336.0000, 1337.0000,
    1338.0000, 1339.0000, 1340.0000, 1341.0000, 1342.0000, 1343.0000, 1344.0000,
    1345.0000, 1346.0000, 1347.0000, 1348.0000, 1349.0000, 1350.0000, 1351.0000,
    1352.0000, 1353.0000, 1354.0000, 1355.0000, 1356.0000, 1357.0000, 1358.0000,
    1359.0000, 1360.0000, 1361.0000, 1362.0000, 1363.0000, 1364.0000, 1365.0000,
    1366.0000, 1367.0000, 1368.0000, 1369.0000, 1370.0000, 1371.0000, 1372.0000,
    1373.0000, 1374.0000, 1375.0000, 1376.0000, 1377.0000, 1378.0000, 1379.0000,
    1380.0000, 1381.0000, 1382.0000, 1383.0000, 1384.0000, 1385.0000, 1386.0000,
    1387.0000, 1388.0000, 1389.0000, 1390.0000, 1391.0000, 1392.0000, 1393.0000,
    1394.0000, 1395.0000, 1396.0000, 1397.0000, 1398.0000, 1399.0000, 1400.0000,
    1401.0000, 1402.0000, 1403.0000, 1404.0000, 1405.0000, 1406.0000, 1407.0000,
    1408.0000, 1409.0000, 1410.0000, 1411.0000, 1412.0000, 1413.0000, 1414.0000,
    1415.0000, 1416.0000, 1417.0000, 1418.0000, 1419.0000, 1420.0000, 1421.0000,
    1422.0000, 1423.0000, 1424.0000, 1425.0000, 1426.0000, 1427.0000, 1428.0000,
    1429.0000, 1430.0000, 1431.0000, 1432.0000, 1433.0000, 1434.0000, 1435.0000,
    1436.0000, 1437.0000, 1438.0000, 1439.0000, 1440.0000, 1441.0000, 1442.0000,
    1443.0000, 1444.0000, 1445.0000, 1446.0000, 1447.0000, 1448.0000, 1449.0000,
    1450.0000, 1451.0000, 1452.0000, 1453.0000, 1454.0000, 1455.0000, 1456.0000,
    1457.0000, 1458.0000, 1459.0000, 1460.0000, 1461.0000, 1462.0000, 1463.0000,
    1464.0000, 1465.0000, 1466.0000, 1467.0000, 1468.0000, 1469.0000, 1470.0000,
    1471.0000, 1472.0000, 1473.0000, 1474.0000, 1475.0000, 1476.0000, 1477.0000,
    1478.0000, 1479.0000, 1480.0000, 1481.0000, 1482.0000, 1483.0000, 1484.0000,
    1485.0000, 1486.0000, 1487.0000, 1488.0000, 1489.0000, 1490.0000, 1491.0000,
    1492.0000, 1493.0000, 1494.0000, 1495.0000, 1496.0000, 1497.0000, 1498.0000,
    1499.0000, 1500.0000, 1501.0000, 1502.0000, 1503.0000, 1504.0000, 1505.0000,
    1506.0000, 1507.0000, 1508.0000, 1509.0000, 1510.0000, 1511.0000, 1512.0000,
    1513.0000, 1514.0000, 1515.0000, 1516.0000, 1517.0000, 1518.0000, 1519.0000,
    1520.0000, 1521.0000, 1522.0000, 1523.0000, 1524.0000, 1525.0000, 1526.0000,
    1527.0000, 1528.0000, 1529.0000, 1530.0000, 1531.0000, 1532.0000, 1533.0000,
    1534.0000, 1535.0000, 1536.0000, 1537.0000, 1538.0000, 1539.0000, 1540.0000,
    1541.0000, 1542.0000, 1543.0000, 1544.0000, 1545.0000, 1546.0000, 1547.0000,
    1548.0000, 1549.0000, 1550.0000, 1551.0000, 1552.0000, 1553.0000, 1554.0000,
    1555.0000, 1556.0000, 1557.0000, 1558.0000, 1559.0000, 1560.0000, 1561.0000,
    1562.0000, 1563.0000, 1564.0000, 1565.0000, 1566.0000, 1567.0000, 1568.0000,
    1569.0000, 1570.0000, 1571.0000, 1572.0000, 1573.0000, 1574.0000, 1575.0000,
    1576.0000, 1577.0000, 1578.0000, 1579.0000, 1580.0000, 1581.0000, 1582.0000,
    1583.0000, 1584.0000, 1585.0000, 1586.0000, 1587.0000, 1588.0000, 1589.0000,
    1590.0000, 1591.0000, 1592.0000, 1593.0000, 1594.0000, 1595.0000, 1596.0000,
    1597.0000, 1598.0000, 1599.0000, 1600.0000, 1601.0000, 1602.0000, 1603.0000,
    1604.0000, 1605.0000, 1606.0000, 1607.0000, 1608.0000, 1609.0000, 1610.0000,
    1611.0000, 1612.0000, 1613.0000, 1614.0000, 1615.0000, 1616.0000, 1617.0000,
    1618.0000, 1619.0000, 1620.0000, 1621.0000, 1622.0000, 1623.0000, 1624.0000,
    1625.0000, 1626.0000, 1627.0000, 1628.0000, 1629.0000, 1630.0000, 1631.0000,
    1632.0000, 1633.0000, 1634.0000, 1635.0000, 1636.0000, 1637.0000, 1638.0000,
    1639.0000, 1640.0000, 1641.0000, 1642.0000, 1643.0000, 1644.0000, 1645.0000,
    1646.0000, 1647.0000, 1648.0000, 1649.0000, 1650.0000, 1651.0000, 1652.0000,
    1653.0000, 1654.0000, 1655.0000, 1656.0000, 1657.0000, 1658.0000, 1659.0000,
    1660.0000, 1661.0000, 1662.0000, 1663.0000, 1664.0000, 1665.0000, 1666.0000,
    1667.0000, 1668.0000, 1669.0000, 1670.0000, 1671.0000, 1672.0000, 1673.0000,
    1674.0000, 1675.0000, 1676.0000, 1677.0000, 1678.0000, 1679.0000, 1680.0000,
    1681.0000, 1682.0000, 1683.0000, 1684.0000, 1685.0000, 1686.0000, 1687.0000,
    1688.0000, 1689.0000, 1690.0000, 1691.0000, 1692.0000, 1693.0000, 1694.0000,
    1695.0000, 1696.0000, 1697.0000, 1698.0000, 1699.0000, 1700.0000, 1701.0000,
    1702.0000, 1703.0000, 1704.0000, 1705.0000, 1706.0000, 1707.0000, 1708.0000,
    1709.0000, 1710.0000, 1711.0000, 1712.0000, 1713.0000, 1714.0000, 1715.0000,
    1716.0000, 1717.0000, 1718.0000, 1719.0000, 1720.0000, 1721.0000, 1722.0000,
    1723.0000, 1724.0000, 1725.0000, 1726.0000, 1727.0000, 1728.0000, 1729.0000,
    1730.0000, 1731.0000, 1732.0000, 1733.0000, 1734.0000, 1735.0000, 1736.0000,
    1737.0000, 1738.0000, 1739.0000, 1740.0000, 1741.0000, 1742.0000, 1743.0000,
    1744.0000, 1745.0000, 1746.0000, 1747.0000, 1748.0000, 1749.0000, 1750.0000,
    1751.0000, 1752.0000, 1753.0000, 1754.0000, 1755.0000, 1756.0000, 1757.0000,
    1758.0000, 1759.0000, 1760.0000, 1761.0000, 1762.0000, 1763.0000, 1764.0000,
    1765.0000, 1766.0000, 1767.0000, 1768.0000, 1769.0000, 1770.0000, 1771.0000,
    1772.0000, 1773.0000, 1774.0000, 1775.0000, 1776.0000, 1777.0000, 1778.0000,
    1779.0000, 1780.0000, 1781.0000, 1782.0000, 1783.0000, 1784.0000, 1785.0000,
    1786.0000, 1787.0000, 1788.0000, 1789.0000, 1790.0000, 1791.0000, 1792.0000,
    1793.0000, 1794.0000, 1795.0000, 1796.0000, 1797.0000, 1798.0000, 1799.0000,
    1800.0000, 1801.0000, 1802.0000, 1803.0000, 1804.0000, 1805.0000, 1806.0000,
    1807.0000, 1808.0000, 1809.0000, 1810.0000, 1811.0000, 1812.0000, 1813.0000,
    1814.0000, 1815.0000, 1816.0000, 1817.0000, 1818.0000, 1819.0000, 1820.0000,
    1821.0000, 1822.0000, 1823.0000, 1824.0000, 1825.0000, 1826.0000, 1827.0000,
    1828.0000, 1829.0000, 1830.0000, 1831.0000, 1832.0000, 1833.0000, 1834.0000,
    1835.0000, 1836.0000, 1837.0000, 1838.0000, 1839.0000, 1840.0000, 1841.0000,
    1842.0000, 1843.0000, 1844.0000, 1845.0000, 1846.0000, 1847.0000, 1848.0000,
    1849.0000, 1850.0000, 1851.0000, 1852.0000, 1853.0000, 1854.0000, 1855.0000,
    1856.0000, 1857.0000, 1858.0000, 1859.0000, 1860.0000, 1861.0000, 1862.0000,
    1863.0000, 1864.0000, 1865.0000, 1866.0000, 1867.0000, 1868.0000, 1869.0000,
    1870.0000, 1871.0000, 1872.0000, 1873.0000, 1874.0000, 1875.0000, 1876.0000,
    1877.0000, 1878.0000, 1879.0000, 1880.0000, 1881.0000, 1882.0000, 1883.0000,
    1884.0000, 1885.0000, 1886.0000, 1887.0000, 1888.0000, 1889.0000, 1890.0000,
    1891.0000, 1892.0000, 1893.0000, 1894.0000, 1895.0000, 1896.0000, 1897.0000,
    1898.0000, 1899.0000, 1900.0000, 1901.0000, 1902.0000, 1903.0000, 1904.0000,
    1905.0000, 1906.0000, 1907.0000, 1908.0000, 1909.0000, 1910.0000, 1911.0000,
    1912.0000, 1913.0000, 1914.0000, 1915.0000, 1916.0000, 1917.0000, 1918.0000,
    1919.0000, 1920.0000, 1921.0000, 1922.0000, 1923.0000, 1924.0000, 1925.0000,
    1926.0000, 1927.0000, 1928.0000, 1929.0000, 1930.0000, 1931.0000, 1932.0000,
    1933.0000, 1934.0000, 1935.0000, 1936.0000, 1937.0000, 1938.0000, 1939.0000,
    1940.0000, 1941.0000, 1942.0000, 1943.0000, 1944.0000, 1945.0000, 1946.0000,
    1947.0000, 1948.0000, 1949.0000, 1950.0000, 1951.0000, 1952.0000, 1953.0000,
    1954.0000, 1955.0000, 1956.0000, 1957.0000, 1958.0000, 1959.0000, 1960.0000,
    1961.0000, 1962.0000, 1963.0000, 1964.0000, 1965.0000, 1966.0000, 1967.0000,
    1968.0000, 1969.0000, 1970.0000, 1971.0000, 1972.0000, 1973.0000, 1974.0000,
    1975.0000, 1976.0000, 1977.0000, 1978.0000, 1979.0000, 1980.0000, 1981.0000,
    1982.0000, 1983.0000, 1984.0000, 1985.0000, 1986.0000, 1987.0000, 1988.0000,
    1989.0000, 1990.0000, 1991.0000, 1992.0000, 1993.0000, 1994.0000, 1995.0000,
    1996.0000, 1997.0000, 1998.0000, 1999.0000, 2000.0000, 2001.0000, 2002.0000,
    2003.0000, 2004.0000, 2005.0000, 2006.0000, 2007.0000, 2008.0000, 2009.0000,
    2010.0000, 2011.0000, 2012.0000, 2013.0000, 2014.0000, 2015.0000, 2016.0000,
    2017.0000, 2018.0000, 2019.0000, 2020.0000, 2021.0000, 2022.0000, 2023.0000,
    2024.0000, 2025.0000, 2026.0000, 2027.0000, 2028.0000, 2029.0000, 2030.0000,
    2031.0000, 2032.0000, 2033.0000, 2034.0000, 2035.0000, 2036.0000, 2037.0000,
    2038.0000, 2039.0000, 2040.0000, 2041.0000, 2042.0000, 2043.0000, 2044.0000,
    2045.0000, 2046.0000, 2047.0000, 2048.0000, 2049.0000, 2050.0000, 2051.0000,
    2052.0000, 2053.0000, 2054.0000, 2055.0000, 2056.0000, 2057.0000, 2058.0000,
    2059.0000, 2060.0000, 2061.0000, 2062.0000, 2063.0000, 2064.0000, 2065.0000,
    2066.0000, 2067.0000, 2068.0000, 2069.0000, 2070.0000, 2071.0000, 2072.0000,
    2073.0000, 2074.0000, 2075.0000, 2076.0000, 2077.0000, 2078.0000, 2079.0000,
    2080.0000, 2081.0000, 2082.0000, 2083.0000, 2084.0000, 2085.0000, 2086.0000,
    2087.0000, 2088.0000, 2089.0000, 2090.0000, 2091.0000, 2092.0000, 2093.0000,
    2094.0000, 2095.0000, 2096.0000, 2097.0000, 2098.0000, 2099.0000, 2100.0000,
    2101.0000, 2102.0000, 2103.0000, 2104.0000, 2105.0000, 2106.0000, 2107.0000,
    2108.0000, 2109.0000, 2110.0000, 2111.0000, 2112.0000, 2113.0000, 2114.0000,
    2115.0000, 2116.0000, 2117.0000, 2118.0000, 2119.0000, 2120.0000, 2121.0000,
    2122.0000, 2123.0000, 2124.0000, 2125.0000, 2126.0000, 2127.0000, 2128.0000,
    2129.0000, 2130.0000, 2131.0000, 2132.0000, 2133.0000, 2134.0000, 2135.0000,
    2136.0000, 2137.0000, 2138.0000, 2139.0000, 2140.0000, 2141.0000, 2142.0000,
    2143.0000, 2144.0000, 2145.0000, 2146.0000, 2147.0000, 2148.0000, 2149.0000,
    2150.0000, 2151.0000, 2152.0000, 2153.0000, 2154.0000, 2155.0000, 2156.0000,
    2157.0000, 2158.0000, 2159.0000, 2160.0000, 2161.0000, 2162.0000, 2163.0000,
    2164.0000, 2165.0000, 2166.0000, 2167.0000, 2168.0000, 2169.0000, 2170.0000,
    2171.0000, 2172.0000, 2173.0000, 2174.0000, 2175.0000, 2176.0000, 2177.0000,
    2178.0000, 2179.0000, 2180.0000, 2181.0000, 2182.0000, 2183.0000, 2184.0000,
    2185.0000, 2186.0000, 2187.0000, 2188.0000, 2189.0000, 2190.0000, 2191.0000,
    2192.0000, 2193.0000, 2194.0000, 2195.0000, 2196.0000, 2197.0000, 2198.0000,
    2199.0000, 2200.0000, 2201.0000, 2202.0000, 2203.0000, 2204.0000, 2205.0000,
    2206.0000, 2207.0000, 2208.0000, 2209.0000, 2210.0000, 2211.0000, 2212.0000,
    2213.0000, 2214.0000, 2215.0000, 2216.0000, 2217.0000, 2218.0000, 2219.0000,
    2220.0000, 2221.0000, 2222.0000, 2223.0000, 2224.0000, 2225.0000, 2226.0000,
    2227.0000, 2228.0000, 2229.0000, 2230.0000, 2231.0000, 2232.0000, 2233.0000,
    2234.0000, 2235.0000, 2236.0000, 2237.0000, 2238.0000, 2239.0000, 2240.0000,
    2241.0000, 2242.0000, 2243.0000, 2244.0000, 2245.0000, 2246.0000, 2247.0000,
    2248.0000, 2249.0000, 2250.0000, 2251.0000, 2252.0000, 2253.0000, 2254.0000,
    2255.0000, 2256.0000, 2257.0000, 2258.0000, 2259.0000, 2260.0000, 2261.0000,
    2262.0000, 2263.0000, 2264.0000, 2265.0000, 2266.0000, 2267.0000, 2268.0000,
    2269.0000, 2270.0000, 2271.0000, 2272.0000, 2273.0000, 2274.0000, 2275.0000,
    2276.0000, 2277.0000, 2278.0000, 2279.0000, 2280.0000, 2281.0000, 2282.0000,
    2283.0000, 2284.0000, 2285.0000, 2286.0000, 2287.0000, 2288.0000, 2289.0000,
    2290.0000, 2291.0000, 2292.0000, 2293.0000, 2294.0000, 2295.0000, 2296.0000,
    2297.0000, 2298.0000, 2299.0000, 2300.0000, 2301.0000, 2302.0000, 2303.0000,
    2304.0000, 2305.0000, 2306.0000, 2307.0000, 2308.0000, 2309.0000, 2310.0000,
    2311.0000, 2312.0000, 2313.0000, 2314.0000, 2315.0000, 2316.0000, 2317.0000,
    2318.0000, 2319.0000, 2320.0000, 2321.0000, 2322.0000, 2323.0000, 2324.0000,
    2325.0000, 2326.0000, 2327.0000, 2328.0000, 2329.0000, 2330.0000, 2331.0000,
    2332.0000, 2333.0000, 2334.0000, 2335.0000, 2336.0000, 2337.0000, 2338.0000,
    2339.0000, 2340.0000, 2341.0000, 2342.0000, 2343.0000, 2344.0000, 2345.0000,
    2346.0000, 2347.0000, 2348.0000, 2349.0000, 2350.0000, 2351.0000, 2352.0000,
    2353.0000, 2354.0000, 2355.0000, 2356.0000, 2357.0000, 2358.0000, 2359.0000,
    2360.0000, 2361.0000, 2362.0000, 2363.0000, 2364.0000, 2365.0000, 2366.0000,
    2367.0000, 2368.0000, 2369.0000, 2370.0000, 2371.0000, 2372.0000, 2373.0000,
    2374.0000, 2375.0000, 2376.0000, 2377.0000, 2378.0000, 2379.0000, 2380.0000,
    2381.0000, 2382.0000, 2383.0000, 2384.0000, 2385.0000, 2386.0000, 2387.0000,
    2388.0000, 2389.0000, 2390.0000, 2391.0000, 2392.0000, 2393.0000, 2394.0000,
    2395.0000, 2396.0000, 2397.0000, 2398.0000, 2399.0000, 2400.0000};

const double CIA_Orton::H2H2_e[][10] = {
    -31.421, -30.965, -30.795, -30.810, -30.951, -31.174, -31.425, -31.691,
    -31.969, -32.246, -29.588, -29.132, -28.962, -28.977, -29.118, -29.341,
    -29.592, -29.858, -30.136, -30.413, -28.202, -27.746, -27.576, -27.591,
    -27.732, -27.955, -28.206, -28.472, -28.750, -29.027, -26.817, -26.360,
    -26.190, -26.205, -26.346, -26.569, -26.820, -27.086, -27.363, -27.640,
    -26.007, -25.550, -25.379, -25.393, -25.536, -25.758, -26.009, -26.275,
    -26.553, -26.829, -25.433, -24.975, -24.803, -24.818, -24.961, -25.183,
    -25.434, -25.700, -25.977, -26.254, -24.987, -24.529, -24.357, -24.371,
    -24.515, -24.737, -24.988, -25.254, -25.531, -25.808, -24.623, -24.164,
    -23.991, -24.007, -24.151, -24.373, -24.623, -24.890, -25.167, -25.443,
    -24.315, -23.856, -23.682, -23.698, -23.843, -24.065, -24.315, -24.581,
    -24.859, -25.135, -24.048, -23.589, -23.415, -23.431, -23.576, -23.798,
    -24.048, -24.315, -24.592, -24.868, -23.812, -23.353, -23.178, -23.196,
    -23.341, -23.563, -23.813, -24.079, -24.356, -24.633, -23.601, -23.141,
    -22.967, -22.985, -23.131, -23.352, -23.602, -23.869, -24.145, -24.422,
    -23.409, -22.950, -22.775, -22.795, -22.941, -23.162, -23.412, -23.678,
    -23.955, -24.232, -23.235, -22.775, -22.600, -22.621, -22.767, -22.988,
    -23.238, -23.504, -23.781, -24.058, -23.074, -22.614, -22.440, -22.461,
    -22.608, -22.828, -23.078, -23.344, -23.621, -23.898, -22.924, -22.464,
    -22.291, -22.313, -22.460, -22.680, -22.930, -23.196, -23.473, -23.750,
    -22.785, -22.325, -22.152, -22.175, -22.322, -22.542, -22.792, -23.058,
    -23.335, -23.612, -22.654, -22.194, -22.022, -22.047, -22.193, -22.413,
    -22.663, -22.930, -23.206, -23.483, -22.531, -22.072, -21.901, -21.926,
    -22.073, -22.292, -22.542, -22.808, -23.085, -23.361, -22.414, -21.956,
    -21.786, -21.811, -21.959, -22.177, -22.427, -22.693, -22.970, -23.246,
    -22.304, -21.847, -21.677, -21.703, -21.850, -22.069, -22.319, -22.585,
    -22.862, -23.138, -22.200, -21.743, -21.575, -21.602, -21.748, -21.967,
    -22.217, -22.484, -22.760, -23.037, -21.743, -21.292, -21.130, -21.158,
    -21.303, -21.521, -21.771, -22.038, -22.315, -22.591, -21.372, -20.927,
    -20.768, -20.795, -20.939, -21.157, -21.407, -21.674, -21.950, -22.227,
    -21.063, -20.621, -20.463, -20.489, -20.632, -20.849, -21.100, -21.367,
    -21.643, -21.919, -20.801, -20.360, -20.200, -20.224, -20.365, -20.582,
    -20.834, -21.101, -21.376, -21.652, -20.575, -20.131, -19.969, -19.991,
    -20.130, -20.348, -20.599, -20.866, -21.141, -21.417, -20.375, -19.928,
    -19.763, -19.782, -19.920, -20.138, -20.390, -20.656, -20.931, -21.207,
    -20.196, -19.744, -19.577, -19.593, -19.730, -19.949, -20.201, -20.466,
    -20.741, -21.017, -20.033, -19.578, -19.408, -19.422, -19.557, -19.776,
    -20.028, -20.293, -20.567, -20.843, -19.884, -19.425, -19.253, -19.264,
    -19.399, -19.618, -19.869, -20.133, -20.408, -20.683, -19.747, -19.284,
    -19.109, -19.119, -19.253, -19.472, -19.723, -19.986, -20.260, -20.536,
    -19.620, -19.153, -18.976, -18.984, -19.118, -19.337, -19.586, -19.849,
    -20.123, -20.398, -19.501, -19.032, -18.852, -18.858, -18.992, -19.210,
    -19.459, -19.721, -19.995, -20.270, -19.390, -18.918, -18.736, -18.741,
    -18.874, -19.091, -19.339, -19.601, -19.874, -20.149, -19.286, -18.811,
    -18.627, -18.630, -18.763, -18.979, -19.226, -19.487, -19.761, -20.035,
    -19.188, -18.710, -18.524, -18.527, -18.658, -18.874, -19.119, -19.380,
    -19.653, -19.927, -19.095, -18.615, -18.427, -18.428, -18.559, -18.773,
    -19.018, -19.279, -19.552, -19.825, -18.923, -18.438, -18.247, -18.247,
    -18.376, -18.588, -18.831, -19.090, -19.363, -19.636, -18.767, -18.279,
    -18.084, -18.082, -18.208, -18.418, -18.660, -18.919, -19.190, -19.463,
    -18.624, -18.132, -17.934, -17.930, -18.054, -18.262, -18.503, -18.761,
    -19.031, -19.304, -18.493, -17.997, -17.797, -17.790, -17.913, -18.119,
    -18.359, -18.615, -18.886, -19.157, -18.371, -17.872, -17.669, -17.660,
    -17.781, -17.985, -18.224, -18.479, -18.749, -19.019, -18.258, -17.757,
    -17.551, -17.540, -17.659, -17.862, -18.099, -18.354, -18.622, -18.892,
    -18.153, -17.649, -17.440, -17.427, -17.544, -17.745, -17.981, -18.235,
    -18.502, -18.772, -18.056, -17.548, -17.337, -17.322, -17.437, -17.636,
    -17.871, -18.124, -18.391, -18.660, -17.965, -17.454, -17.240, -17.223,
    -17.336, -17.533, -17.766, -18.018, -18.285, -18.553, -17.882, -17.367,
    -17.151, -17.130, -17.241, -17.436, -17.668, -17.919, -18.185, -18.452,
    -17.803, -17.285, -17.066, -17.042, -17.150, -17.343, -17.574, -17.824,
    -18.089, -18.355, -17.731, -17.208, -16.986, -16.959, -17.065, -17.256,
    -17.485, -17.734, -17.998, -18.263, -17.663, -17.136, -16.911, -16.881,
    -16.984, -17.173, -17.401, -17.648, -17.911, -18.176, -17.602, -17.070,
    -16.841, -16.808, -16.908, -17.096, -17.321, -17.568, -17.829, -18.093,
    -17.546, -17.010, -16.776, -16.740, -16.837, -17.023, -17.247, -17.492,
    -17.752, -18.015, -17.491, -16.951, -16.713, -16.673, -16.767, -16.951,
    -17.173, -17.417, -17.676, -17.938, -17.441, -16.895, -16.653, -16.610,
    -16.701, -16.883, -17.104, -17.345, -17.603, -17.864, -17.394, -16.844,
    -16.598, -16.551, -16.639, -16.818, -17.037, -17.277, -17.533, -17.794,
    -17.352, -16.796, -16.546, -16.496, -16.581, -16.757, -16.974, -17.212,
    -17.468, -17.727, -17.314, -16.753, -16.498, -16.444, -16.526, -16.700,
    -16.915, -17.151, -17.405, -17.663, -17.276, -16.710, -16.451, -16.393,
    -16.472, -16.643, -16.856, -17.090, -17.343, -17.600, -17.241, -16.670,
    -16.406, -16.345, -16.421, -16.589, -16.800, -17.032, -17.283, -17.539,
    -17.209, -16.633, -16.364, -16.299, -16.372, -16.537, -16.746, -16.976,
    -17.226, -17.480, -17.180, -16.599, -16.325, -16.256, -16.326, -16.488,
    -16.694, -16.923, -17.171, -17.424, -17.154, -16.567, -16.289, -16.216,
    -16.283, -16.442, -16.645, -16.872, -17.118, -17.370, -17.128, -16.536,
    -16.253, -16.177, -16.240, -16.396, -16.597, -16.822, -17.066, -17.317,
    -17.104, -16.507, -16.220, -16.139, -16.200, -16.353, -16.550, -16.773,
    -17.016, -17.265, -17.082, -16.480, -16.188, -16.104, -16.161, -16.311,
    -16.506, -16.727, -16.967, -17.215, -17.063, -16.456, -16.159, -16.070,
    -16.124, -16.271, -16.464, -16.682, -16.921, -17.167, -17.045, -16.433,
    -16.131, -16.039, -16.089, -16.233, -16.423, -16.639, -16.876, -17.121,
    -17.028, -16.411, -16.104, -16.008, -16.055, -16.196, -16.383, -16.597,
    -16.832, -17.075, -17.012, -16.390, -16.078, -15.979, -16.022, -16.159,
    -16.344, -16.556, -16.789, -17.031, -16.997, -16.370, -16.054, -15.951,
    -15.990, -16.125, -16.307, -16.516, -16.748, -16.988, -16.984, -16.352,
    -16.032, -15.925, -15.960, -16.092, -16.271, -16.478, -16.708, -16.947,
    -16.973, -16.336, -16.011, -15.900, -15.932, -16.060, -16.236, -16.441,
    -16.670, -16.907, -16.962, -16.320, -15.991, -15.876, -15.904, -16.029,
    -16.202, -16.405, -16.632, -16.867, -16.951, -16.305, -15.971, -15.853,
    -15.878, -15.999, -16.170, -16.370, -16.595, -16.829, -16.942, -16.291,
    -15.953, -15.831, -15.852, -15.971, -16.138, -16.336, -16.559, -16.791,
    -16.934, -16.279, -15.937, -15.810, -15.828, -15.943, -16.108, -16.303,
    -16.525, -16.755, -16.927, -16.267, -15.921, -15.790, -15.805, -15.917,
    -16.079, -16.272, -16.491, -16.719, -16.920, -16.256, -15.906, -15.771,
    -15.782, -15.891, -16.050, -16.241, -16.458, -16.684, -16.914, -16.246,
    -15.891, -15.753, -15.761, -15.866, -16.022, -16.211, -16.426, -16.650,
    -16.909, -16.236, -15.878, -15.736, -15.740, -15.842, -15.995, -16.182,
    -16.394, -16.617, -16.905, -16.228, -15.865, -15.719, -15.720, -15.818,
    -15.969, -16.153, -16.364, -16.585, -16.901, -16.220, -15.853, -15.704,
    -15.701, -15.796, -15.944, -16.126, -16.334, -16.554, -16.898, -16.213,
    -15.842, -15.689, -15.682, -15.774, -15.919, -16.099, -16.305, -16.523,
    -16.895, -16.206, -15.831, -15.675, -15.665, -15.753, -15.896, -16.073,
    -16.277, -16.493, -16.893, -16.200, -15.821, -15.661, -15.648, -15.733,
    -15.873, -16.048, -16.249, -16.464, -16.892, -16.195, -15.812, -15.648,
    -15.631, -15.714, -15.850, -16.023, -16.222, -16.436, -16.891, -16.190,
    -15.803, -15.636, -15.616, -15.695, -15.829, -15.999, -16.196, -16.408,
    -16.890, -16.186, -15.795, -15.624, -15.601, -15.677, -15.808, -15.975,
    -16.170, -16.381, -16.890, -16.182, -15.787, -15.613, -15.586, -15.659,
    -15.787, -15.952, -16.145, -16.354, -16.890, -16.178, -15.780, -15.602,
    -15.573, -15.642, -15.768, -15.930, -16.121, -16.328, -16.891, -16.175,
    -15.774, -15.592, -15.560, -15.626, -15.749, -15.908, -16.097, -16.302,
    -16.892, -16.173, -15.768, -15.583, -15.547, -15.610, -15.730, -15.887,
    -16.074, -16.277, -16.893, -16.171, -15.762, -15.574, -15.535, -15.594,
    -15.712, -15.866, -16.051, -16.252, -16.895, -16.169, -15.757, -15.565,
    -15.523, -15.579, -15.694, -15.846, -16.029, -16.228, -16.897, -16.168,
    -15.752, -15.557, -15.511, -15.565, -15.676, -15.827, -16.007, -16.205,
    -16.900, -16.167, -15.748, -15.549, -15.500, -15.551, -15.659, -15.807,
    -15.986, -16.182, -16.903, -16.167, -15.744, -15.542, -15.490, -15.537,
    -15.643, -15.789, -15.965, -16.159, -16.906, -16.167, -15.740, -15.535,
    -15.480, -15.524, -15.627, -15.771, -15.945, -16.137, -16.909, -16.167,
    -15.737, -15.529, -15.470, -15.511, -15.612, -15.753, -15.924, -16.115,
    -16.913, -16.167, -15.734, -15.523, -15.461, -15.499, -15.597, -15.735,
    -15.905, -16.093, -16.917, -16.168, -15.732, -15.517, -15.452, -15.487,
    -15.583, -15.719, -15.886, -16.073, -16.921, -16.169, -15.730, -15.512,
    -15.444, -15.476, -15.569, -15.702, -15.867, -16.052, -16.925, -16.170,
    -15.728, -15.507, -15.436, -15.465, -15.555, -15.686, -15.849, -16.032,
    -16.930, -16.172, -15.727, -15.503, -15.428, -15.454, -15.542, -15.670,
    -15.831, -16.012, -16.934, -16.174, -15.725, -15.498, -15.421, -15.444,
    -15.529, -15.655, -15.813, -15.993, -16.939, -16.176, -15.725, -15.494,
    -15.414, -15.434, -15.516, -15.640, -15.797, -15.974, -16.944, -16.178,
    -15.724, -15.491, -15.408, -15.425, -15.504, -15.625, -15.780, -15.955,
    -16.949, -16.181, -15.724, -15.488, -15.402, -15.416, -15.492, -15.611,
    -15.763, -15.937, -16.955, -16.183, -15.724, -15.485, -15.396, -15.407,
    -15.480, -15.596, -15.747, -15.918, -16.960, -16.186, -15.724, -15.482,
    -15.390, -15.398, -15.469, -15.583, -15.731, -15.901, -16.966, -16.190,
    -15.724, -15.479, -15.384, -15.390, -15.458, -15.569, -15.715, -15.883,
    -16.972, -16.193, -15.725, -15.477, -15.379, -15.382, -15.447, -15.556,
    -15.700, -15.866, -16.978, -16.197, -15.726, -15.475, -15.374, -15.374,
    -15.437, -15.543, -15.685, -15.849, -16.984, -16.200, -15.727, -15.473,
    -15.369, -15.367, -15.426, -15.531, -15.670, -15.833, -16.990, -16.204,
    -15.728, -15.471, -15.365, -15.360, -15.417, -15.518, -15.656, -15.816,
    -16.997, -16.209, -15.729, -15.470, -15.361, -15.353, -15.407, -15.507,
    -15.642, -15.801, -17.003, -16.213, -15.731, -15.469, -15.357, -15.346,
    -15.398, -15.495, -15.628, -15.785, -17.009, -16.217, -15.733, -15.468,
    -15.353, -15.340, -15.389, -15.484, -15.614, -15.770, -17.016, -16.222,
    -15.735, -15.467, -15.350, -15.333, -15.380, -15.473, -15.601, -15.755,
    -17.023, -16.226, -15.737, -15.467, -15.347, -15.327, -15.372, -15.462,
    -15.588, -15.740, -17.029, -16.231, -15.739, -15.467, -15.344, -15.322,
    -15.364, -15.451, -15.575, -15.725, -17.036, -16.236, -15.742, -15.467,
    -15.341, -15.316, -15.356, -15.441, -15.563, -15.711, -17.043, -16.241,
    -15.745, -15.467, -15.338, -15.311, -15.348, -15.431, -15.551, -15.697,
    -17.050, -16.246, -15.747, -15.467, -15.336, -15.305, -15.340, -15.421,
    -15.539, -15.683, -17.057, -16.251, -15.750, -15.467, -15.333, -15.300,
    -15.333, -15.411, -15.527, -15.669, -17.065, -16.257, -15.753, -15.468,
    -15.331, -15.296, -15.325, -15.401, -15.515, -15.655, -17.072, -16.262,
    -15.756, -15.468, -15.329, -15.291, -15.318, -15.392, -15.504, -15.642,
    -17.079, -16.268, -15.759, -15.469, -15.327, -15.287, -15.311, -15.383,
    -15.493, -15.629, -17.087, -16.273, -15.763, -15.470, -15.326, -15.282,
    -15.304, -15.374, -15.482, -15.616, -17.094, -16.279, -15.766, -15.471,
    -15.324, -15.278, -15.298, -15.365, -15.471, -15.603, -17.102, -16.285,
    -15.770, -15.472, -15.323, -15.275, -15.291, -15.356, -15.460, -15.591,
    -17.109, -16.291, -15.774, -15.473, -15.322, -15.271, -15.285, -15.348,
    -15.450, -15.578, -17.117, -16.297, -15.778, -15.475, -15.321, -15.268,
    -15.279, -15.340, -15.439, -15.566, -17.125, -16.303, -15.782, -15.477,
    -15.320, -15.264, -15.274, -15.332, -15.429, -15.554, -17.132, -16.309,
    -15.786, -15.479, -15.320, -15.261, -15.268, -15.324, -15.419, -15.543,
    -17.140, -16.316, -15.790, -15.481, -15.319, -15.259, -15.263, -15.316,
    -15.410, -15.531, -17.148, -16.322, -15.795, -15.483, -15.319, -15.256,
    -15.258, -15.309, -15.400, -15.520, -17.156, -16.328, -15.799, -15.485,
    -15.319, -15.253, -15.253, -15.302, -15.390, -15.509, -17.164, -16.335,
    -15.804, -15.487, -15.319, -15.251, -15.248, -15.294, -15.381, -15.498,
    -17.172, -16.341, -15.808, -15.490, -15.319, -15.248, -15.243, -15.287,
    -15.372, -15.487, -17.180, -16.347, -15.813, -15.492, -15.319, -15.246,
    -15.238, -15.280, -15.363, -15.476, -17.188, -16.354, -15.818, -15.495,
    -15.319, -15.243, -15.233, -15.273, -15.354, -15.465, -17.196, -16.360,
    -15.822, -15.497, -15.319, -15.241, -15.229, -15.266, -15.345, -15.455,
    -17.204, -16.367, -15.827, -15.500, -15.320, -15.239, -15.224, -15.260,
    -15.337, -15.445, -17.212, -16.374, -15.832, -15.503, -15.320, -15.237,
    -15.220, -15.253, -15.328, -15.435, -17.220, -16.380, -15.837, -15.506,
    -15.321, -15.236, -15.216, -15.247, -15.320, -15.425, -17.228, -16.387,
    -15.842, -15.509, -15.322, -15.234, -15.212, -15.241, -15.312, -15.415,
    -17.236, -16.394, -15.847, -15.512, -15.323, -15.232, -15.208, -15.235,
    -15.304, -15.405, -17.244, -16.401, -15.852, -15.515, -15.323, -15.231,
    -15.204, -15.229, -15.296, -15.396, -17.252, -16.407, -15.857, -15.518,
    -15.324, -15.229, -15.201, -15.223, -15.288, -15.386, -17.260, -16.414,
    -15.862, -15.521, -15.325, -15.228, -15.197, -15.217, -15.280, -15.377,
    -17.268, -16.421, -15.868, -15.525, -15.326, -15.227, -15.193, -15.212,
    -15.273, -15.367, -17.276, -16.428, -15.873, -15.528, -15.327, -15.226,
    -15.190, -15.206, -15.265, -15.358, -17.285, -16.435, -15.878, -15.531,
    -15.328, -15.225, -15.187, -15.201, -15.258, -15.349, -17.293, -16.443,
    -15.884, -15.535, -15.329, -15.224, -15.184, -15.196, -15.251, -15.340,
    -17.301, -16.450, -15.889, -15.538, -15.331, -15.223, -15.181, -15.191,
    -15.244, -15.332, -17.309, -16.457, -15.895, -15.542, -15.332, -15.222,
    -15.178, -15.186, -15.237, -15.323, -17.317, -16.464, -15.901, -15.546,
    -15.333, -15.221, -15.175, -15.181, -15.230, -15.314, -17.325, -16.471,
    -15.906, -15.549, -15.335, -15.220, -15.172, -15.176, -15.223, -15.306,
    -17.333, -16.478, -15.912, -15.553, -15.336, -15.220, -15.169, -15.171,
    -15.216, -15.297, -17.341, -16.485, -15.917, -15.557, -15.338, -15.219,
    -15.167, -15.166, -15.210, -15.289, -17.350, -16.492, -15.923, -15.560,
    -15.340, -15.219, -15.164, -15.161, -15.203, -15.281, -17.358, -16.499,
    -15.929, -15.564, -15.341, -15.218, -15.161, -15.157, -15.197, -15.273,
    -17.366, -16.506, -15.934, -15.568, -15.343, -15.218, -15.159, -15.152,
    -15.190, -15.265, -17.374, -16.513, -15.940, -15.572, -15.345, -15.217,
    -15.157, -15.148, -15.184, -15.257, -17.382, -16.520, -15.946, -15.576,
    -15.347, -15.217, -15.154, -15.143, -15.178, -15.249, -17.390, -16.527,
    -15.952, -15.580, -15.349, -15.217, -15.152, -15.139, -15.172, -15.241,
    -17.398, -16.534, -15.958, -15.584, -15.351, -15.217, -15.150, -15.135,
    -15.166, -15.233, -17.406, -16.541, -15.963, -15.588, -15.353, -15.216,
    -15.147, -15.130, -15.160, -15.226, -17.414, -16.548, -15.969, -15.592,
    -15.355, -15.216, -15.145, -15.126, -15.154, -15.218, -17.422, -16.555,
    -15.975, -15.595, -15.357, -15.216, -15.143, -15.122, -15.148, -15.210,
    -17.430, -16.562, -15.980, -15.599, -15.359, -15.216, -15.141, -15.118,
    -15.142, -15.203, -17.438, -16.569, -15.986, -15.603, -15.361, -15.216,
    -15.138, -15.114, -15.136, -15.196, -17.446, -16.577, -15.991, -15.607,
    -15.363, -15.216, -15.136, -15.110, -15.131, -15.188, -17.454, -16.584,
    -15.997, -15.611, -15.365, -15.216, -15.134, -15.106, -15.125, -15.181,
    -17.462, -16.591, -16.002, -15.615, -15.368, -15.216, -15.132, -15.102,
    -15.119, -15.174, -17.470, -16.598, -16.008, -15.619, -15.370, -15.216,
    -15.130, -15.098, -15.114, -15.167, -17.478, -16.605, -16.013, -15.623,
    -15.372, -15.216, -15.128, -15.094, -15.108, -15.160, -17.486, -16.612,
    -16.019, -15.627, -15.374, -15.216, -15.126, -15.090, -15.103, -15.153,
    -17.493, -16.619, -16.024, -15.630, -15.376, -15.216, -15.124, -15.087,
    -15.098, -15.146, -17.501, -16.625, -16.030, -15.634, -15.378, -15.216,
    -15.122, -15.083, -15.092, -15.139, -17.509, -16.632, -16.035, -15.638,
    -15.379, -15.215, -15.120, -15.079, -15.087, -15.133, -17.517, -16.639,
    -16.040, -15.641, -15.381, -15.215, -15.118, -15.076, -15.082, -15.126,
    -17.524, -16.645, -16.045, -15.645, -15.383, -15.215, -15.116, -15.072,
    -15.076, -15.119, -17.532, -16.652, -16.051, -15.649, -15.385, -15.215,
    -15.115, -15.069, -15.071, -15.113, -17.539, -16.658, -16.056, -15.652,
    -15.386, -15.215, -15.113, -15.065, -15.066, -15.106, -17.547, -16.665,
    -16.061, -15.656, -15.388, -15.215, -15.111, -15.062, -15.061, -15.100,
    -17.554, -16.671, -16.066, -15.659, -15.390, -15.215, -15.109, -15.058,
    -15.056, -15.094, -17.562, -16.678, -16.071, -15.663, -15.391, -15.215,
    -15.107, -15.055, -15.051, -15.087, -17.569, -16.684, -16.077, -15.666,
    -15.393, -15.215, -15.105, -15.051, -15.046, -15.081, -17.576, -16.690,
    -16.082, -15.670, -15.394, -15.214, -15.103, -15.048, -15.041, -15.074,
    -17.583, -16.696, -16.087, -15.673, -15.396, -15.214, -15.101, -15.044,
    -15.036, -15.068, -17.590, -16.702, -16.092, -15.676, -15.397, -15.214,
    -15.099, -15.040, -15.031, -15.062, -17.597, -16.708, -16.097, -15.680,
    -15.399, -15.214, -15.097, -15.037, -15.026, -15.055, -17.604, -16.714,
    -16.102, -15.683, -15.400, -15.213, -15.095, -15.033, -15.021, -15.049,
    -17.611, -16.720, -16.107, -15.686, -15.402, -15.213, -15.093, -15.030,
    -15.016, -15.043, -17.618, -16.726, -16.112, -15.689, -15.403, -15.213,
    -15.091, -15.026, -15.011, -15.037, -17.625, -16.731, -16.116, -15.692,
    -15.404, -15.212, -15.089, -15.022, -15.006, -15.031, -17.631, -16.737,
    -16.120, -15.694, -15.405, -15.212, -15.087, -15.019, -15.001, -15.025,
    -17.638, -16.742, -16.124, -15.697, -15.406, -15.211, -15.085, -15.015,
    -14.996, -15.019, -17.644, -16.747, -16.128, -15.699, -15.407, -15.211,
    -15.082, -15.012, -14.992, -15.013, -17.651, -16.752, -16.131, -15.701,
    -15.408, -15.210, -15.080, -15.008, -14.987, -15.007, -17.657, -16.757,
    -16.135, -15.703, -15.409, -15.209, -15.078, -15.004, -14.982, -15.001,
    -17.663, -16.762, -16.138, -15.705, -15.409, -15.208, -15.076, -15.001,
    -14.977, -14.995, -17.669, -16.766, -16.141, -15.707, -15.410, -15.207,
    -15.074, -14.997, -14.972, -14.989, -17.675, -16.771, -16.144, -15.708,
    -15.411, -15.206, -15.071, -14.994, -14.968, -14.984, -17.681, -16.775,
    -16.147, -15.710, -15.411, -15.205, -15.069, -14.990, -14.963, -14.978,
    -17.686, -16.779, -16.150, -15.711, -15.411, -15.204, -15.066, -14.986,
    -14.958, -14.972, -17.692, -16.783, -16.152, -15.713, -15.411, -15.202,
    -15.064, -14.982, -14.953, -14.966, -17.697, -16.788, -16.155, -15.714,
    -15.410, -15.201, -15.061, -14.978, -14.949, -14.960, -17.702, -16.792,
    -16.157, -15.715, -15.410, -15.199, -15.059, -14.974, -14.944, -14.955,
    -17.707, -16.795, -16.159, -15.716, -15.409, -15.198, -15.056, -14.970,
    -14.939, -14.949, -17.711, -16.799, -16.161, -15.717, -15.408, -15.196,
    -15.053, -14.966, -14.934, -14.943, -17.716, -16.803, -16.163, -15.718,
    -15.407, -15.194, -15.050, -14.962, -14.929, -14.937, -17.720, -16.806,
    -16.164, -15.719, -15.406, -15.192, -15.047, -14.958, -14.925, -14.932,
    -17.724, -16.810, -16.166, -15.719, -15.405, -15.190, -15.044, -14.954,
    -14.920, -14.926, -17.728, -16.813, -16.167, -15.720, -15.403, -15.188,
    -15.041, -14.950, -14.915, -14.920, -17.731, -16.814, -16.168, -15.720,
    -15.402, -15.186, -15.038, -14.946, -14.910, -14.914, -17.734, -16.816,
    -16.169, -15.719, -15.400, -15.183, -15.034, -14.942, -14.905, -14.909,
    -17.737, -16.816, -16.169, -15.718, -15.399, -15.181, -15.031, -14.938,
    -14.900, -14.903, -17.739, -16.816, -16.170, -15.717, -15.397, -15.178,
    -15.027, -14.934, -14.895, -14.897, -17.741, -16.816, -16.170, -15.716,
    -15.395, -15.175, -15.024, -14.930, -14.890, -14.891, -17.743, -16.816,
    -16.170, -15.714, -15.393, -15.173, -15.020, -14.926, -14.885, -14.886,
    -17.744, -16.815, -16.170, -15.712, -15.391, -15.170, -15.016, -14.921,
    -14.880, -14.880, -17.745, -16.813, -16.170, -15.710, -15.389, -15.166,
    -15.012, -14.917, -14.874, -14.874, -17.746, -16.811, -16.170, -15.708,
    -15.387, -15.163, -15.008, -14.913, -14.869, -14.869, -17.747, -16.809,
    -16.169, -15.705, -15.385, -15.160, -15.004, -14.909, -14.864, -14.863,
    -17.746, -16.808, -16.166, -15.702, -15.382, -15.156, -15.000, -14.904,
    -14.859, -14.857, -17.745, -16.806, -16.163, -15.699, -15.378, -15.152,
    -14.996, -14.900, -14.854, -14.851, -17.743, -16.805, -16.160, -15.696,
    -15.375, -15.148, -14.991, -14.895, -14.849, -14.846, -17.741, -16.803,
    -16.156, -15.692, -15.371, -15.144, -14.987, -14.890, -14.843, -14.840,
    -17.738, -16.801, -16.151, -15.688, -15.366, -15.139, -14.982, -14.885,
    -14.838, -14.834, -17.735, -16.799, -16.146, -15.684, -15.362, -15.135,
    -14.977, -14.880, -14.833, -14.828, -17.731, -16.797, -16.140, -15.680,
    -15.357, -15.130, -14.973, -14.875, -14.828, -14.823, -17.727, -16.794,
    -16.134, -15.675, -15.352, -15.125, -14.968, -14.870, -14.822, -14.817,
    -17.722, -16.792, -16.127, -15.670, -15.347, -15.119, -14.963, -14.864,
    -14.817, -14.811, -17.717, -16.789, -16.120, -15.665, -15.341, -15.114,
    -14.958, -14.859, -14.812, -14.805, -17.710, -16.783, -16.113, -15.659,
    -15.335, -15.109, -14.953, -14.854, -14.806, -14.799, -17.702, -16.775,
    -16.106, -15.652, -15.329, -15.103, -14.947, -14.848, -14.801, -14.793,
    -17.693, -16.767, -16.099, -15.645, -15.323, -15.097, -14.941, -14.842,
    -14.795, -14.787, -17.683, -16.758, -16.091, -15.638, -15.317, -15.091,
    -14.936, -14.837, -14.790, -14.782, -17.673, -16.749, -16.083, -15.630,
    -15.310, -15.085, -14.930, -14.831, -14.784, -14.776, -17.662, -16.738,
    -16.075, -15.622, -15.303, -15.079, -14.924, -14.825, -14.778, -14.770,
    -17.650, -16.727, -16.067, -15.613, -15.296, -15.072, -14.917, -14.819,
    -14.772, -14.764, -17.637, -16.715, -16.058, -15.604, -15.289, -15.066,
    -14.911, -14.813, -14.766, -14.758, -17.623, -16.702, -16.049, -15.595,
    -15.281, -15.059, -14.905, -14.807, -14.760, -14.752, -17.609, -16.688,
    -16.039, -15.585, -15.273, -15.052, -14.898, -14.801, -14.754, -14.746,
    -17.593, -16.673, -16.028, -15.575, -15.264, -15.044, -14.891, -14.795,
    -14.748, -14.740, -17.576, -16.658, -16.017, -15.564, -15.256, -15.037,
    -14.884, -14.789, -14.742, -14.734, -17.558, -16.641, -16.005, -15.553,
    -15.246, -15.029, -14.877, -14.782, -14.735, -14.728, -17.539, -16.624,
    -15.993, -15.542, -15.237, -15.020, -14.870, -14.776, -14.729, -14.722,
    -17.519, -16.606, -15.980, -15.530, -15.227, -15.012, -14.863, -14.769,
    -14.723, -14.716, -17.498, -16.587, -15.967, -15.517, -15.217, -15.003,
    -14.855, -14.763, -14.716, -14.710, -17.476, -16.567, -15.953, -15.505,
    -15.206, -14.994, -14.848, -14.756, -14.710, -14.704, -17.453, -16.547,
    -15.939, -15.491, -15.196, -14.985, -14.840, -14.750, -14.703, -14.698,
    -17.430, -16.525, -15.924, -15.478, -15.184, -14.976, -14.832, -14.743,
    -14.697, -14.692, -17.405, -16.503, -15.909, -15.464, -15.173, -14.966,
    -14.824, -14.736, -14.690, -14.686, -17.379, -16.481, -15.891, -15.449,
    -15.161, -14.956, -14.816, -14.729, -14.683, -14.680, -17.353, -16.459,
    -15.873, -15.435, -15.149, -14.946, -14.807, -14.722, -14.677, -14.673,
    -17.325, -16.436, -15.853, -15.419, -15.137, -14.936, -14.799, -14.714,
    -14.670, -14.667, -17.297, -16.412, -15.833, -15.403, -15.124, -14.925,
    -14.790, -14.707, -14.664, -14.661, -17.267, -16.388, -15.811, -15.387,
    -15.111, -14.914, -14.781, -14.699, -14.657, -14.654, -17.237, -16.364,
    -15.789, -15.370, -15.098, -14.903, -14.772, -14.691, -14.651, -14.648,
    -17.205, -16.338, -15.766, -15.353, -15.084, -14.892, -14.763, -14.683,
    -14.644, -14.641, -17.173, -16.313, -15.742, -15.336, -15.070, -14.881,
    -14.754, -14.675, -14.637, -14.634, -17.139, -16.286, -15.718, -15.317,
    -15.056, -14.869, -14.745, -14.667, -14.631, -14.628, -17.105, -16.259,
    -15.692, -15.299, -15.041, -14.857, -14.735, -14.659, -14.624, -14.621,
    -17.070, -16.232, -15.668, -15.280, -15.026, -14.845, -14.725, -14.651,
    -14.617, -14.614, -17.035, -16.205, -15.644, -15.261, -15.011, -14.833,
    -14.715, -14.642, -14.610, -14.608, -16.998, -16.178, -15.620, -15.241,
    -14.995, -14.820, -14.705, -14.634, -14.603, -14.601, -16.961, -16.150,
    -15.595, -15.221, -14.980, -14.808, -14.695, -14.625, -14.595, -14.595,
    -16.923, -16.122, -15.570, -15.200, -14.963, -14.795, -14.684, -14.616,
    -14.588, -14.588, -16.884, -16.093, -15.545, -15.179, -14.947, -14.782,
    -14.673, -14.608, -14.581, -14.582, -16.844, -16.064, -15.519, -15.158,
    -14.930, -14.768, -14.663, -14.599, -14.573, -14.575, -16.804, -16.034,
    -15.493, -15.136, -14.913, -14.755, -14.652, -14.589, -14.565, -14.568,
    -16.762, -16.004, -15.467, -15.114, -14.896, -14.741, -14.640, -14.580,
    -14.558, -14.562, -16.720, -15.974, -15.440, -15.091, -14.878, -14.727,
    -14.629, -14.571, -14.550, -14.555, -16.678, -15.940, -15.412, -15.068,
    -14.859, -14.713, -14.618, -14.562, -14.542, -14.548, -16.635, -15.905,
    -15.384, -15.045, -14.839, -14.698, -14.606, -14.552, -14.535, -14.541,
    -16.591, -15.868, -15.355, -15.021, -14.819, -14.683, -14.594, -14.543,
    -14.527, -14.534, -16.547, -15.831, -15.325, -14.997, -14.799, -14.668,
    -14.582, -14.533, -14.519, -14.527, -16.502, -15.792, -15.295, -14.972,
    -14.777, -14.653, -14.570, -14.524, -14.511, -14.521, -16.456, -15.753,
    -15.265, -14.947, -14.756, -14.637, -14.558, -14.514, -14.503, -14.513,
    -16.409, -15.713, -15.234, -14.922, -14.734, -14.621, -14.545, -14.504,
    -14.495, -14.506, -16.362, -15.671, -15.202, -14.896, -14.711, -14.605,
    -14.533, -14.494, -14.487, -14.499, -16.315, -15.629, -15.170, -14.870,
    -14.688, -14.589, -14.520, -14.484, -14.479, -14.492, -16.266, -15.585,
    -15.138, -14.843, -14.664, -14.572, -14.507, -14.474, -14.471, -14.485,
    -16.218, -15.544, -15.105, -14.816, -14.643, -14.556, -14.494, -14.464,
    -14.463, -14.478, -16.169, -15.502, -15.072, -14.789, -14.623, -14.539,
    -14.480, -14.454, -14.454, -14.471, -16.120, -15.460, -15.039, -14.762,
    -14.603, -14.523, -14.467, -14.444, -14.446, -14.464, -16.071, -15.417,
    -15.005, -14.735, -14.583, -14.506, -14.453, -14.434, -14.437, -14.456,
    -16.021, -15.374, -14.970, -14.707, -14.563, -14.489, -14.439, -14.423,
    -14.428, -14.449, -15.971, -15.331, -14.935, -14.678, -14.543, -14.472,
    -14.425, -14.413, -14.419, -14.442, -15.920, -15.287, -14.900, -14.650,
    -14.524, -14.455, -14.411, -14.403, -14.410, -14.435, -15.868, -15.243,
    -14.864, -14.621, -14.504, -14.438, -14.396, -14.392, -14.401, -14.428,
    -15.816, -15.198, -14.828, -14.592, -14.485, -14.421, -14.382, -14.382,
    -14.392, -14.420, -15.764, -15.153, -14.791, -14.562, -14.466, -14.403,
    -14.367, -14.371, -14.383, -14.413, -15.710, -15.109, -14.755, -14.533,
    -14.444, -14.385, -14.353, -14.360, -14.374, -14.406, -15.656, -15.064,
    -14.719, -14.503, -14.421, -14.367, -14.338, -14.349, -14.366, -14.398,
    -15.602, -15.019, -14.683, -14.473, -14.398, -14.349, -14.324, -14.338,
    -14.357, -14.391, -15.546, -14.974, -14.646, -14.443, -14.374, -14.330,
    -14.310, -14.326, -14.348, -14.384, -15.490, -14.928, -14.609, -14.413,
    -14.350, -14.312, -14.295, -14.315, -14.339, -14.376, -15.434, -14.883,
    -14.572, -14.382, -14.325, -14.293, -14.281, -14.303, -14.331, -14.369,
    -15.376, -14.837, -14.535, -14.351, -14.300, -14.274, -14.266, -14.291,
    -14.322, -14.361, -15.318, -14.791, -14.498, -14.320, -14.274, -14.255,
    -14.251, -14.279, -14.313, -14.354, -15.260, -14.745, -14.460, -14.289,
    -14.248, -14.236, -14.237, -14.267, -14.305, -14.346, -15.201, -14.698,
    -14.422, -14.257, -14.222, -14.216, -14.222, -14.255, -14.296, -14.339,
    -15.142, -14.652, -14.384, -14.226, -14.198, -14.197, -14.208, -14.243,
    -14.287, -14.331, -15.083, -14.605, -14.346, -14.194, -14.173, -14.177,
    -14.193, -14.231, -14.277, -14.323, -15.024, -14.559, -14.308, -14.162,
    -14.149, -14.158, -14.179, -14.219, -14.268, -14.315, -14.964, -14.512,
    -14.270, -14.130, -14.124, -14.138, -14.165, -14.207, -14.258, -14.307,
    -14.904, -14.465, -14.231, -14.098, -14.100, -14.118, -14.151, -14.195,
    -14.248, -14.299, -14.844, -14.418, -14.192, -14.066, -14.076, -14.098,
    -14.136, -14.183, -14.239, -14.291, -14.783, -14.371, -14.154, -14.033,
    -14.051, -14.078, -14.122, -14.171, -14.229, -14.283, -14.721, -14.323,
    -14.115, -14.001, -14.027, -14.058, -14.108, -14.159, -14.218, -14.275,
    -14.659, -14.276, -14.075, -13.968, -14.002, -14.037, -14.094, -14.146,
    -14.208, -14.266, -14.597, -14.228, -14.036, -13.935, -13.978, -14.017,
    -14.080, -14.134, -14.198, -14.258, -14.537, -14.180, -13.996, -13.903,
    -13.953, -13.997, -14.065, -14.121, -14.189, -14.250, -14.477, -14.131,
    -13.956, -13.870, -13.928, -13.977, -14.050, -14.109, -14.179, -14.242,
    -14.417, -14.082, -13.915, -13.837, -13.903, -13.957, -14.034, -14.096,
    -14.170, -14.235, -14.357, -14.032, -13.875, -13.805, -13.878, -13.937,
    -14.019, -14.083, -14.161, -14.227, -14.297, -13.983, -13.834, -13.772,
    -13.853, -13.917, -14.003, -14.070, -14.152, -14.219, -14.237, -13.933,
    -13.792, -13.739, -13.827, -13.898, -13.988, -14.057, -14.143, -14.212,
    -14.177, -13.883, -13.751, -13.707, -13.802, -13.878, -13.972, -14.044,
    -14.134, -14.204, -14.117, -13.833, -13.709, -13.674, -13.776, -13.858,
    -13.956, -14.031, -14.125, -14.197, -14.058, -13.782, -13.667, -13.641,
    -13.751, -13.838, -13.939, -14.017, -14.117, -14.189, -13.998, -13.731,
    -13.625, -13.608, -13.725, -13.818, -13.923, -14.004, -14.108, -14.182,
    -13.939, -13.682, -13.586, -13.576, -13.700, -13.798, -13.907, -13.992,
    -14.098, -14.174, -13.880, -13.633, -13.547, -13.544, -13.674, -13.777,
    -13.892, -13.981, -14.089, -14.166, -13.821, -13.585, -13.509, -13.512,
    -13.649, -13.757, -13.876, -13.969, -14.079, -14.158, -13.763, -13.536,
    -13.472, -13.480, -13.624, -13.736, -13.860, -13.958, -14.069, -14.150,
    -13.704, -13.488, -13.434, -13.448, -13.598, -13.715, -13.845, -13.947,
    -14.059, -14.142, -13.646, -13.440, -13.397, -13.416, -13.573, -13.695,
    -13.829, -13.937, -14.049, -14.134, -13.588, -13.391, -13.361, -13.384,
    -13.548, -13.674, -13.814, -13.926, -14.038, -14.126, -13.530, -13.343,
    -13.324, -13.353, -13.523, -13.653, -13.798, -13.916, -14.028, -14.118,
    -13.472, -13.296, -13.288, -13.321, -13.497, -13.632, -13.783, -13.906,
    -14.018, -14.109, -13.414, -13.248, -13.253, -13.290, -13.472, -13.611,
    -13.767, -13.896, -14.007, -14.101, -13.357, -13.201, -13.212, -13.259,
    -13.448, -13.592, -13.752, -13.884, -13.996, -14.093, -13.300, -13.155,
    -13.170, -13.229, -13.424, -13.573, -13.736, -13.872, -13.985, -14.086,
    -13.243, -13.109, -13.128, -13.199, -13.400, -13.554, -13.721, -13.860,
    -13.975, -14.078, -13.186, -13.063, -13.085, -13.169, -13.376, -13.535,
    -13.706, -13.847, -13.964, -14.070, -13.130, -13.017, -13.042, -13.139,
    -13.352, -13.517, -13.691, -13.835, -13.952, -14.063, -13.074, -12.972,
    -12.997, -13.109, -13.329, -13.499, -13.676, -13.822, -13.941, -14.055,
    -13.018, -12.927, -12.952, -13.080, -13.306, -13.481, -13.661, -13.808,
    -13.930, -14.048, -12.962, -12.882, -12.906, -13.051, -13.283, -13.463,
    -13.646, -13.795, -13.918, -14.041, -12.906, -12.837, -12.860, -13.022,
    -13.260, -13.446, -13.631, -13.782, -13.907, -14.033, -12.851, -12.793,
    -12.813, -12.993, -13.237, -13.429, -13.616, -13.768, -13.895, -14.026,
    -12.798, -12.751, -12.779, -12.966, -13.209, -13.412, -13.602, -13.757,
    -13.886, -14.019, -12.746, -12.709, -12.747, -12.939, -13.181, -13.395,
    -13.588, -13.746, -13.878, -14.012, -12.694, -12.668, -12.717, -12.912,
    -13.152, -13.378, -13.574, -13.735, -13.871, -14.006, -12.643, -12.628,
    -12.688, -12.886, -13.122, -13.361, -13.560, -13.724, -13.864, -13.999,
    -12.592, -12.588, -12.662, -12.860, -13.091, -13.344, -13.546, -13.714,
    -13.857, -13.993, -12.542, -12.549, -12.637, -12.835, -13.060, -13.327,
    -13.533, -13.704, -13.851, -13.986, -12.493, -12.511, -12.614, -12.810,
    -13.028, -13.311, -13.520, -13.694, -13.845, -13.980, -12.445, -12.473,
    -12.593, -12.786, -12.995, -13.295, -13.507, -13.685, -13.840, -13.974,
    -12.396, -12.436, -12.573, -12.762, -12.961, -13.279, -13.494, -13.676,
    -13.835, -13.968, -12.349, -12.399, -12.556, -12.738, -12.927, -13.263,
    -13.481, -13.667, -13.831, -13.962, -12.303, -12.360, -12.526, -12.716,
    -12.909, -13.247, -13.469, -13.658, -13.823, -13.955, -12.258, -12.322,
    -12.495, -12.694, -12.893, -13.232, -13.457, -13.650, -13.815, -13.947,
    -12.214, -12.284, -12.464, -12.673, -12.880, -13.217, -13.445, -13.642,
    -13.807, -13.939, -12.171, -12.246, -12.432, -12.652, -12.869, -13.202,
    -13.434, -13.635, -13.799, -13.931, -12.128, -12.208, -12.399, -12.632,
    -12.861, -13.187, -13.423, -13.627, -13.790, -13.922, -12.086, -12.170,
    -12.365, -12.612, -12.855, -13.173, -13.412, -13.620, -13.781, -13.914,
    -12.045, -12.133, -12.331, -12.593, -12.851, -13.158, -13.401, -13.614,
    -13.772, -13.905, -12.005, -12.095, -12.296, -12.575, -12.850, -13.144,
    -13.391, -13.607, -13.763, -13.896, -11.966, -12.058, -12.260, -12.557,
    -12.851, -13.130, -13.381, -13.601, -13.753, -13.887, -11.927, -12.021,
    -12.223, -12.540, -12.854, -13.116, -13.371, -13.595, -13.743, -13.878,
    -11.886, -11.987, -12.199, -12.521, -12.840, -13.103, -13.360, -13.586,
    -13.736, -13.871, -11.846, -11.955, -12.176, -12.502, -12.824, -13.091,
    -13.349, -13.577, -13.728, -13.864, -11.806, -11.923, -12.155, -12.483,
    -12.807, -13.079, -13.338, -13.568, -13.721, -13.858, -11.766, -11.892,
    -12.135, -12.465, -12.788, -13.067, -13.326, -13.558, -13.715, -13.852,
    -11.726, -11.861, -12.117, -12.446, -12.769, -13.055, -13.315, -13.549,
    -13.708, -13.846, -11.687, -11.832, -12.101, -12.428, -12.748, -13.044,
    -13.304, -13.538, -13.702, -13.840, -11.648, -11.803, -12.086, -12.410,
    -12.726, -13.033, -13.293, -13.528, -13.696, -13.835, -11.609, -11.775,
    -12.073, -12.392, -12.702, -13.023, -13.282, -13.517, -13.691, -13.830,
    -11.570, -11.747, -12.062, -12.374, -12.677, -13.013, -13.270, -13.506,
    -13.685, -13.825, -11.532, -11.721, -12.052, -12.356, -12.651, -13.003,
    -13.259, -13.495, -13.680, -13.820, -11.501, -11.700, -12.038, -12.344,
    -12.641, -12.995, -13.252, -13.488, -13.674, -13.814, -11.471, -11.681,
    -12.024, -12.333, -12.632, -12.987, -13.246, -13.481, -13.668, -13.808,
    -11.443, -11.663, -12.010, -12.323, -12.625, -12.980, -13.240, -13.475,
    -13.662, -13.803, -11.417, -11.647, -11.997, -12.314, -12.620, -12.974,
    -13.235, -13.469, -13.657, -13.797, -11.392, -11.633, -11.985, -12.306,
    -12.617, -12.968, -13.231, -13.464, -13.651, -13.791, -11.368, -11.620,
    -11.972, -12.300, -12.615, -12.963, -13.228, -13.459, -13.646, -13.785,
    -11.347, -11.609, -11.961, -12.294, -12.615, -12.958, -13.225, -13.455,
    -13.640, -13.779, -11.326, -11.600, -11.950, -12.290, -12.617, -12.954,
    -13.223, -13.451, -13.635, -13.774, -11.307, -11.592, -11.939, -12.287,
    -12.621, -12.950, -13.222, -13.448, -13.629, -13.768, -11.290, -11.586,
    -11.929, -12.285, -12.627, -12.947, -13.222, -13.445, -13.624, -13.762,
    -11.288, -11.583, -11.924, -12.280, -12.624, -12.943, -13.218, -13.440,
    -13.619, -13.756, -11.290, -11.582, -11.921, -12.276, -12.620, -12.939,
    -13.213, -13.435, -13.613, -13.750, -11.296, -11.582, -11.918, -12.273,
    -12.617, -12.935, -13.209, -13.429, -13.608, -13.745, -11.307, -11.585,
    -11.918, -12.270, -12.613, -12.932, -13.204, -13.424, -13.602, -13.739,
    -11.321, -11.589, -11.918, -12.267, -12.609, -12.929, -13.199, -13.419,
    -13.597, -13.733, -11.320, -11.586, -11.915, -12.264, -12.606, -12.925,
    -13.195, -13.414, -13.592, -13.727, -11.316, -11.583, -11.911, -12.261,
    -12.603, -12.922, -13.191, -13.409, -13.586, -13.721, -11.310, -11.578,
    -11.908, -12.258, -12.600, -12.918, -13.187, -13.404, -13.581, -13.716,
    -11.301, -11.572, -11.903, -12.255, -12.597, -12.915, -13.183, -13.400,
    -13.575, -13.710, -11.290, -11.565, -11.899, -12.252, -12.594, -12.911,
    -13.179, -13.395, -13.570, -13.704, -11.287, -11.563, -11.897, -12.250,
    -12.592, -12.908, -13.175, -13.390, -13.565, -13.698, -11.285, -11.561,
    -11.896, -12.249, -12.590, -12.905, -13.172, -13.386, -13.559, -13.692,
    -11.284, -11.561, -11.895, -12.248, -12.588, -12.903, -13.169, -13.382,
    -13.554, -13.687, -11.285, -11.561, -11.895, -12.247, -12.587, -12.900,
    -13.165, -13.377, -13.549, -13.681, -11.286, -11.562, -11.895, -12.247,
    -12.586, -12.898, -13.162, -13.373, -13.544, -13.675, -11.288, -11.563,
    -11.896, -12.247, -12.585, -12.896, -13.159, -13.369, -13.538, -13.669,
    -11.292, -11.566, -11.898, -12.247, -12.584, -12.895, -13.157, -13.365,
    -13.533, -13.663, -11.297, -11.570, -11.900, -12.248, -12.584, -12.893,
    -13.154, -13.361, -13.528, -13.658, -11.302, -11.574, -11.903, -12.249,
    -12.584, -12.892, -13.151, -13.358, -13.523, -13.652, -11.309, -11.579,
    -11.906, -12.251, -12.584, -12.891, -13.149, -13.354, -13.518, -13.646,
    -11.317, -11.585, -11.910, -12.254, -12.585, -12.891, -13.147, -13.351,
    -13.513, -13.640, -11.325, -11.591, -11.915, -12.257, -12.587, -12.891,
    -13.146, -13.347, -13.508, -13.635, -11.335, -11.599, -11.920, -12.260,
    -12.589, -12.891, -13.144, -13.344, -13.504, -13.629, -11.345, -11.607,
    -11.926, -12.265, -12.591, -12.892, -13.143, -13.341, -13.499, -13.624,
    -11.357, -11.616, -11.933, -12.269, -12.594, -12.893, -13.142, -13.338,
    -13.494, -13.618, -11.370, -11.625, -11.940, -12.274, -12.598, -12.895,
    -13.142, -13.335, -13.490, -13.613, -11.384, -11.636, -11.948, -12.280,
    -12.601, -12.897, -13.141, -13.332, -13.485, -13.608, -11.398, -11.647,
    -11.956, -12.286, -12.606, -12.899, -13.141, -13.330, -13.481, -13.602,
    -11.414, -11.659, -11.965, -12.293, -12.610, -12.901, -13.141, -13.327,
    -13.476, -13.597, -11.431, -11.672, -11.975, -12.300, -12.615, -12.904,
    -13.141, -13.325, -13.472, -13.592, -11.445, -11.683, -11.984, -12.307,
    -12.620, -12.907, -13.141, -13.323, -13.468, -13.587, -11.460, -11.695,
    -11.993, -12.314, -12.625, -12.909, -13.141, -13.321, -13.464, -13.581,
    -11.475, -11.708, -12.003, -12.322, -12.630, -12.912, -13.142, -13.318,
    -13.460, -13.576, -11.490, -11.720, -12.013, -12.330, -12.636, -12.915,
    -13.142, -13.316, -13.456, -13.571, -11.506, -11.733, -12.024, -12.338,
    -12.642, -12.919, -13.143, -13.315, -13.452, -13.565, -11.522, -11.747,
    -12.035, -12.347, -12.648, -12.922, -13.144, -13.313, -13.449, -13.560,
    -11.538, -11.760, -12.046, -12.356, -12.655, -12.926, -13.145, -13.311,
    -13.445, -13.555, -11.555, -11.774, -12.058, -12.365, -12.662, -12.930,
    -13.146, -13.310, -13.442, -13.550, -11.572, -11.789, -12.070, -12.375,
    -12.669, -12.935, -13.147, -13.308, -13.438, -13.544, -11.590, -11.804,
    -12.082, -12.385, -12.677, -12.939, -13.149, -13.307, -13.435, -13.539,
    -11.607, -11.818, -12.094, -12.395, -12.684, -12.943, -13.150, -13.305,
    -13.431, -13.534, -11.624, -11.833, -12.106, -12.404, -12.691, -12.948,
    -13.152, -13.304, -13.428, -13.529, -11.641, -11.847, -12.118, -12.414,
    -12.699, -12.952, -13.153, -13.302, -13.424, -13.524, -11.659, -11.862,
    -12.131, -12.425, -12.707, -12.957, -13.154, -13.301, -13.420, -13.519,
    -11.676, -11.878, -12.144, -12.435, -12.714, -12.962, -13.156, -13.299,
    -13.417, -13.514, -11.694, -11.893, -12.157, -12.446, -12.722, -12.967,
    -13.158, -13.298, -13.413, -13.509, -11.712, -11.908, -12.170, -12.456,
    -12.731, -12.972, -13.159, -13.297, -13.409, -13.504, -11.730, -11.924,
    -12.183, -12.467, -12.739, -12.978, -13.161, -13.295, -13.405, -13.499,
    -11.748, -11.940, -12.197, -12.479, -12.747, -12.983, -13.163, -13.294,
    -13.402, -13.494, -11.767, -11.956, -12.211, -12.490, -12.756, -12.989,
    -13.165, -13.293, -13.398, -13.489, -11.785, -11.972, -12.224, -12.501,
    -12.764, -12.994, -13.167, -13.292, -13.394, -13.484, -11.803, -11.987,
    -12.238, -12.512, -12.773, -12.999, -13.169, -13.290, -13.391, -13.479,
    -11.821, -12.003, -12.251, -12.523, -12.781, -13.005, -13.170, -13.289,
    -13.387, -13.474, -11.839, -12.019, -12.265, -12.535, -12.790, -13.010,
    -13.172, -13.288, -13.383, -13.468, -11.857, -12.035, -12.279, -12.546,
    -12.799, -13.016, -13.174, -13.286, -13.380, -13.463, -11.875, -12.052,
    -12.293, -12.558, -12.808, -13.021, -13.176, -13.285, -13.376, -13.458,
    -11.893, -12.068, -12.307, -12.569, -12.817, -13.026, -13.178, -13.284,
    -13.372, -13.453, -11.911, -12.084, -12.321, -12.581, -12.826, -13.032,
    -13.180, -13.283, -13.368, -13.448, -11.930, -12.100, -12.335, -12.593,
    -12.835, -13.037, -13.182, -13.281, -13.365, -13.442, -11.948, -12.117,
    -12.349, -12.605, -12.844, -13.043, -13.184, -13.280, -13.361, -13.437,
    -11.966, -12.133, -12.363, -12.617, -12.853, -13.048, -13.186, -13.278,
    -13.357, -13.432, -11.984, -12.149, -12.377, -12.628, -12.862, -13.053,
    -13.187, -13.277, -13.353, -13.426, -12.002, -12.165, -12.391, -12.640,
    -12.871, -13.058, -13.189, -13.275, -13.349, -13.421, -12.020, -12.182,
    -12.405, -12.652, -12.879, -13.063, -13.190, -13.273, -13.345, -13.416,
    -12.038, -12.198, -12.420, -12.664, -12.888, -13.069, -13.192, -13.272,
    -13.341, -13.411, -12.056, -12.214, -12.434, -12.676, -12.897, -13.073,
    -13.193, -13.270, -13.336, -13.405, -12.074, -12.230, -12.448, -12.688,
    -12.906, -13.078, -13.194, -13.268, -13.332, -13.400, -12.092, -12.247,
    -12.462, -12.700, -12.915, -13.083, -13.196, -13.266, -13.328, -13.395,
    -12.110, -12.263, -12.477, -12.712, -12.924, -13.088, -13.197, -13.264,
    -13.323, -13.389, -12.128, -12.279, -12.491, -12.724, -12.933, -13.093,
    -13.198, -13.262, -13.319, -13.384, -12.146, -12.295, -12.505, -12.736,
    -12.942, -13.098, -13.199, -13.260, -13.315, -13.379, -12.163, -12.311,
    -12.519, -12.747, -12.950, -13.103, -13.200, -13.257, -13.310, -13.373,
    -12.181, -12.327, -12.533, -12.759, -12.959, -13.107, -13.201, -13.255,
    -13.306, -13.368, -12.198, -12.343, -12.547, -12.771, -12.967, -13.112,
    -13.201, -13.252, -13.301, -13.362, -12.216, -12.359, -12.561, -12.782,
    -12.976, -13.117, -13.202, -13.250, -13.297, -13.357, -12.233, -12.375,
    -12.575, -12.794, -12.984, -13.122, -13.203, -13.247, -13.292, -13.351,
    -12.250, -12.391, -12.589, -12.806, -12.993, -13.126, -13.203, -13.244,
    -13.287, -13.346, -12.268, -12.406, -12.603, -12.817, -13.001, -13.131,
    -13.203, -13.241, -13.283, -13.340, -12.285, -12.422, -12.617, -12.829,
    -13.010, -13.135, -13.204, -13.238, -13.278, -13.335, -12.302, -12.438,
    -12.631, -12.840, -13.018, -13.140, -13.204, -13.235, -13.273, -13.329,
    -12.319, -12.454, -12.645, -12.851, -13.026, -13.144, -13.204, -13.231,
    -13.268, -13.323, -12.336, -12.469, -12.659, -12.863, -13.034, -13.148,
    -13.204, -13.228, -13.263, -13.317, -12.353, -12.485, -12.673, -12.874,
    -13.042, -13.152, -13.203, -13.224, -13.257, -13.311, -12.370, -12.500,
    -12.686, -12.885, -13.049, -13.155, -13.203, -13.220, -13.252, -13.305,
    -12.387, -12.516, -12.700, -12.897, -13.057, -13.159, -13.203, -13.216,
    -13.247, -13.299, -12.404, -12.531, -12.714, -12.908, -13.065, -13.162,
    -13.202, -13.212, -13.241, -13.293, -12.420, -12.547, -12.728, -12.919,
    -13.072, -13.165, -13.201, -13.207, -13.235, -13.286, -12.437, -12.562,
    -12.742, -12.930, -13.080, -13.168, -13.200, -13.203, -13.230, -13.280,
    -12.453, -12.578, -12.755, -12.941, -13.087, -13.171, -13.199, -13.198,
    -13.224, -13.274, -12.470, -12.593, -12.769, -12.952, -13.094, -13.174,
    -13.198, -13.193, -13.218, -13.267, -12.486, -12.608, -12.782, -12.963,
    -13.101, -13.176, -13.196, -13.189, -13.212, -13.261, -12.503, -12.623,
    -12.796, -12.973, -13.108, -13.179, -13.195, -13.185, -13.207, -13.254,
    -12.519, -12.638, -12.809, -12.984, -13.114, -13.181, -13.193, -13.181,
    -13.201, -13.248, -12.535, -12.653, -12.822, -12.994, -13.120, -13.183,
    -13.191, -13.177, -13.195, -13.242, -12.551, -12.668, -12.835, -13.004,
    -13.127, -13.185, -13.189, -13.173, -13.189, -13.235, -12.567, -12.683,
    -12.848, -13.014, -13.133, -13.186, -13.186, -13.169, -13.183, -13.229,
    -12.583, -12.698, -12.861, -13.024, -13.139, -13.188, -13.184, -13.165,
    -13.177, -13.222, -12.599, -12.713, -12.874, -13.034, -13.145, -13.189,
    -13.181, -13.161, -13.171, -13.216, -12.614, -12.727, -12.887, -13.044,
    -13.150, -13.190, -13.178, -13.157, -13.165, -13.209, -12.630, -12.742,
    -12.900, -13.054, -13.156, -13.191, -13.175, -13.153, -13.159, -13.203,
    -12.646, -12.757, -12.913, -13.063, -13.161, -13.192, -13.171, -13.148,
    -13.153, -13.196, -12.661, -12.771, -12.925, -13.073, -13.166, -13.192,
    -13.167, -13.142, -13.146, -13.189, -12.677, -12.786, -12.938, -13.082,
    -13.171, -13.192, -13.163, -13.137, -13.139, -13.183, -12.692, -12.800,
    -12.950, -13.091, -13.175, -13.192, -13.159, -13.131, -13.133, -13.176,
    -12.708, -12.814, -12.963, -13.100, -13.180, -13.192, -13.154, -13.125,
    -13.126, -13.169, -12.723, -12.829, -12.975, -13.109, -13.184, -13.192,
    -13.149, -13.119, -13.119, -13.162, -12.738, -12.843, -12.987, -13.118,
    -13.188, -13.191, -13.144, -13.113, -13.112, -13.155, -12.754, -12.857,
    -12.999, -13.126, -13.192, -13.190, -13.139, -13.106, -13.105, -13.147,
    -12.769, -12.871, -13.011, -13.135, -13.196, -13.189, -13.133, -13.099,
    -13.097, -13.140, -12.784, -12.885, -13.023, -13.143, -13.199, -13.188,
    -13.127, -13.092, -13.090, -13.133, -12.799, -12.899, -13.035, -13.151,
    -13.201, -13.186, -13.122, -13.085, -13.082, -13.126, -12.814, -12.913,
    -13.046, -13.158, -13.203, -13.183, -13.118, -13.078, -13.075, -13.119,
    -12.829, -12.926, -13.057, -13.166, -13.205, -13.180, -13.113, -13.072,
    -13.067, -13.111, -12.844, -12.940, -13.069, -13.173, -13.206, -13.177,
    -13.108, -13.065, -13.059, -13.104, -12.858, -12.954, -13.080, -13.180,
    -13.207, -13.173, -13.104, -13.057, -13.051, -13.097, -12.873, -12.967,
    -13.091, -13.187, -13.207, -13.169, -13.099, -13.050, -13.043, -13.090,
    -12.888, -12.980, -13.101, -13.193, -13.207, -13.164, -13.095, -13.043,
    -13.035, -13.082, -12.902, -12.994, -13.112, -13.200, -13.207, -13.160,
    -13.090, -13.035, -13.027, -13.075, -12.917, -13.007, -13.123, -13.206,
    -13.206, -13.155, -13.086, -13.028, -13.018, -13.067, -12.931, -13.020,
    -13.133, -13.212, -13.205, -13.149, -13.081, -13.020, -13.010, -13.060,
    -12.945, -13.033, -13.143, -13.217, -13.205, -13.145, -13.075, -13.012,
    -13.002, -13.053, -12.959, -13.046, -13.153, -13.222, -13.205, -13.140,
    -13.068, -13.005, -12.994, -13.045, -12.974, -13.059, -13.162, -13.227,
    -13.205, -13.136, -13.061, -12.997, -12.986, -13.038, -12.988, -13.071,
    -13.171, -13.232, -13.205, -13.131, -13.054, -12.989, -12.979, -13.031,
    -13.002, -13.084, -13.181, -13.236, -13.205, -13.127, -13.047, -12.981,
    -12.971, -13.024, -13.015, -13.096, -13.189, -13.240, -13.205, -13.122,
    -13.039, -12.972, -12.963, -13.016, -13.029, -13.109, -13.198, -13.244,
    -13.204, -13.118, -13.031, -12.964, -12.955, -13.009, -13.043, -13.121,
    -13.207, -13.247, -13.204, -13.113, -13.023, -12.956, -12.947, -13.002,
    -13.056, -13.133, -13.215, -13.250, -13.204, -13.108, -13.014, -12.948,
    -12.940, -12.994, -13.070, -13.145, -13.223, -13.253, -13.203, -13.103,
    -13.005, -12.939, -12.932, -12.987, -13.083, -13.157, -13.231, -13.255,
    -13.200, -13.096, -12.996, -12.930, -12.924, -12.979, -13.097, -13.168,
    -13.238, -13.257, -13.196, -13.089, -12.988, -12.921, -12.915, -12.971,
    -13.110, -13.179, -13.246, -13.258, -13.192, -13.082, -12.978, -12.912,
    -12.907, -12.963, -13.123, -13.191, -13.253, -13.259, -13.188, -13.075,
    -12.969, -12.903, -12.898, -12.955, -13.136, -13.201, -13.260, -13.260,
    -13.182, -13.066, -12.960, -12.893, -12.889, -12.947, -13.150, -13.212,
    -13.267, -13.260, -13.177, -13.058, -12.950, -12.884, -12.880, -12.939,
    -13.163, -13.223, -13.274, -13.260, -13.171, -13.049, -12.940, -12.874,
    -12.871, -12.931, -13.175, -13.233, -13.280, -13.260, -13.164, -13.040,
    -12.931, -12.864, -12.862, -12.922, -13.188, -13.244, -13.286, -13.259,
    -13.157, -13.031, -12.920, -12.854, -12.853, -12.914, -13.201, -13.254,
    -13.292, -13.258, -13.149, -13.021, -12.910, -12.844, -12.844, -12.905,
    -13.214, -13.264, -13.297, -13.256, -13.142, -13.011, -12.900, -12.834,
    -12.835, -12.897, -13.226, -13.274, -13.301, -13.254, -13.135, -13.001,
    -12.889, -12.824, -12.826, -12.889, -13.238, -13.283, -13.304, -13.251,
    -13.127, -12.991, -12.879, -12.813, -12.817, -12.881, -13.251, -13.293,
    -13.308, -13.248, -13.119, -12.981, -12.868, -12.803, -12.808, -12.873,
    -13.263, -13.302, -13.310, -13.244, -13.111, -12.970, -12.857, -12.792,
    -12.799, -12.866, -13.275, -13.311, -13.313, -13.240, -13.103, -12.959,
    -12.846, -12.781, -12.790, -12.858, -13.287, -13.320, -13.315, -13.236,
    -13.094, -12.948, -12.835, -12.770, -12.781, -12.850, -13.299, -13.328,
    -13.316, -13.231, -13.085, -12.937, -12.823, -12.759, -12.772, -12.842,
    -13.310, -13.337, -13.317, -13.226, -13.076, -12.925, -12.812, -12.748,
    -12.763, -12.835, -13.322, -13.345, -13.318, -13.220, -13.066, -12.913,
    -12.800, -12.737, -12.754, -12.827, -13.333, -13.352, -13.319, -13.214,
    -13.055, -12.901, -12.788, -12.726, -12.745, -12.819, -13.345, -13.359,
    -13.319, -13.207, -13.045, -12.889, -12.776, -12.715, -12.735, -12.810,
    -13.356, -13.366, -13.318, -13.199, -13.033, -12.877, -12.764, -12.704,
    -12.725, -12.802, -13.367, -13.372, -13.318, -13.191, -13.022, -12.865,
    -12.752, -12.693, -12.716, -12.794, -13.378, -13.378, -13.317, -13.183,
    -13.010, -12.852, -12.740, -12.682, -12.706, -12.785, -13.388, -13.384,
    -13.315, -13.174, -12.997, -12.840, -12.728, -12.671, -12.696, -12.776,
    -13.399, -13.390, -13.314, -13.164, -12.984, -12.827, -12.715, -12.660,
    -12.686, -12.768, -13.409, -13.395, -13.311, -13.154, -12.971, -12.814,
    -12.703, -12.649, -12.676, -12.759, -13.420, -13.400, -13.309, -13.144,
    -12.957, -12.800, -12.690, -12.637, -12.666, -12.750, -13.430, -13.404,
    -13.306, -13.133, -12.943, -12.787, -12.677, -12.626, -12.656, -12.741,
    -13.439, -13.406, -13.301, -13.122, -12.929, -12.771, -12.664, -12.615,
    -12.647, -12.733, -13.448, -13.408, -13.295, -13.110, -12.915, -12.755,
    -12.651, -12.605, -12.637, -12.725, -13.457, -13.409, -13.288, -13.098,
    -12.901, -12.738, -12.638, -12.594, -12.628, -12.717, -13.466, -13.409,
    -13.280, -13.085, -12.886, -12.720, -12.625, -12.584, -12.619, -12.709,
    -13.474, -13.409, -13.272, -13.072, -12.871, -12.701, -12.612, -12.574,
    -12.610, -12.702, -13.482, -13.408, -13.263, -13.058, -12.856, -12.683,
    -12.599, -12.563, -12.602, -12.694, -13.490, -13.407, -13.254, -13.044,
    -12.841, -12.663, -12.586, -12.553, -12.593, -12.687, -13.498, -13.404,
    -13.243, -13.030, -12.825, -12.643, -12.572, -12.543, -12.584, -12.679,
    -13.505, -13.402, -13.232, -13.015, -12.809, -12.622, -12.559, -12.533,
    -12.575, -12.672, -13.512, -13.398, -13.220, -12.999, -12.793, -12.601,
    -12.545, -12.523, -12.567, -12.665, -13.518, -13.396, -13.209, -12.983,
    -12.777, -12.586, -12.532, -12.512, -12.557, -12.655, -13.523, -13.393,
    -13.198, -12.967, -12.760, -12.572, -12.518, -12.501, -12.548, -12.645,
    -13.528, -13.390, -13.186, -12.951, -12.744, -12.558, -12.505, -12.490,
    -12.538, -12.635, -13.532, -13.386, -13.174, -12.934, -12.727, -12.545,
    -12.491, -12.478, -12.528, -12.624, -13.536, -13.383, -13.161, -12.917,
    -12.710, -12.533, -12.478, -12.467, -12.518, -12.613, -13.539, -13.379,
    -13.148, -12.899, -12.692, -12.521, -12.464, -12.455, -12.508, -12.602,
    -13.542, -13.374, -13.135, -12.881, -12.675, -12.510, -12.451, -12.444,
    -12.498, -12.590, -13.544, -13.369, -13.121, -12.863, -12.657, -12.500,
    -12.437, -12.432, -12.487, -12.578, -13.546, -13.364, -13.107, -12.844,
    -12.639, -12.490, -12.424, -12.420, -12.477, -12.565, -13.548, -13.359,
    -13.092, -12.825, -12.621, -12.481, -12.410, -12.408, -12.466, -12.552,
    -13.549, -13.351, -13.075, -12.806, -12.598, -12.467, -12.397, -12.397,
    -12.458, -12.545, -13.550, -13.342, -13.058, -12.787, -12.574, -12.453,
    -12.383, -12.386, -12.449, -12.539, -13.550, -13.332, -13.039, -12.768,
    -12.550, -12.438, -12.370, -12.374, -12.442, -12.533, -13.550, -13.322,
    -13.020, -12.749, -12.524, -12.423, -12.357, -12.363, -12.434, -12.528,
    -13.550, -13.311, -13.001, -12.729, -12.497, -12.407, -12.343, -12.352,
    -12.427, -12.524, -13.549, -13.299, -12.980, -12.709, -12.469, -12.392,
    -12.330, -12.341, -12.420, -12.521, -13.548, -13.286, -12.959, -12.689,
    -12.440, -12.376, -12.317, -12.330, -12.413, -12.519, -13.546, -13.273,
    -12.937, -12.669, -12.410, -12.359, -12.303, -12.320, -12.407, -12.517,
    -13.544, -13.259, -12.914, -12.648, -12.379, -12.342, -12.290, -12.309,
    -12.401, -12.516, -13.541, -13.244, -12.891, -12.627, -12.347, -12.325,
    -12.277, -12.298, -12.395, -12.516, -13.537, -13.229, -12.871, -12.607,
    -12.330, -12.310, -12.265, -12.288, -12.387, -12.512, -13.532, -13.213,
    -12.850, -12.587, -12.316, -12.295, -12.252, -12.279, -12.379, -12.507,
    -13.526, -13.197, -12.830, -12.567, -12.303, -12.281, -12.240, -12.270,
    -12.371, -12.503, -13.520, -13.180, -12.810, -12.547, -12.292, -12.267,
    -12.228, -12.261, -12.362, -12.498, -13.513, -13.162, -12.789, -12.527,
    -12.283, -12.252, -12.217, -12.252, -12.354, -12.493, -13.506, -13.144,
    -12.769, -12.506, -12.276, -12.238, -12.205, -12.244, -12.345, -12.488,
    -13.497, -13.125, -12.749, -12.486, -12.271, -12.224, -12.193, -12.236,
    -12.336, -12.483, -13.489, -13.105, -12.728, -12.466, -12.268, -12.210,
    -12.182, -12.228, -12.327, -12.478, -13.479, -13.085, -12.708, -12.445,
    -12.267, -12.197, -12.171, -12.220, -12.317, -12.472, -13.469, -13.064,
    -12.687, -12.425, -12.268, -12.183, -12.160, -12.213, -12.308, -12.467,
    -13.459, -13.040, -12.667, -12.406, -12.255, -12.171, -12.150, -12.205,
    -12.297, -12.460, -13.449, -13.015, -12.646, -12.388, -12.240, -12.160,
    -12.140, -12.198, -12.287, -12.453, -13.439, -12.990, -12.626, -12.370,
    -12.225, -12.149, -12.130, -12.191, -12.275, -12.446, -13.428, -12.963,
    -12.606, -12.352, -12.209, -12.138, -12.121, -12.183, -12.264, -12.439,
    -13.416, -12.935, -12.585, -12.335, -12.192, -12.128, -12.112, -12.177,
    -12.252, -12.431, -13.404, -12.906, -12.565, -12.317, -12.174, -12.118,
    -12.103, -12.170, -12.239, -12.423, -13.392, -12.876, -12.545, -12.300,
    -12.154, -12.109, -12.094, -12.163, -12.227, -12.414, -13.379, -12.844,
    -12.525, -12.284, -12.134, -12.100, -12.086, -12.157, -12.214, -12.405,
    -13.366, -12.812, -12.504, -12.267, -12.113, -12.091, -12.078, -12.150,
    -12.201, -12.396, -13.352, -12.779, -12.484, -12.251, -12.091, -12.083,
    -12.070, -12.144, -12.187, -12.387, -13.338, -12.759, -12.462, -12.235,
    -12.078, -12.063, -12.061, -12.136, -12.182, -12.382, -13.323, -12.740,
    -12.439, -12.218, -12.066, -12.042, -12.053, -12.128, -12.178, -12.377,
    -13.308, -12.723, -12.416, -12.203, -12.055, -12.019, -12.044, -12.119,
    -12.175, -12.373, -13.293, -12.708, -12.393, -12.187, -12.045, -11.994,
    -12.036, -12.110, -12.174, -12.370, -13.277, -12.693, -12.369, -12.171,
    -12.035, -11.967, -12.028, -12.101, -12.174, -12.367, -13.261, -12.681,
    -12.345, -12.156, -12.027, -11.939, -12.019, -12.092, -12.175, -12.364,
    -13.244, -12.670, -12.321, -12.141, -12.019, -11.908, -12.011, -12.083,
    -12.178, -12.363, -13.227, -12.660, -12.296, -12.125, -12.012, -11.876,
    -12.003, -12.073, -12.182, -12.361, -13.209, -12.652, -12.271, -12.111,
    -12.006, -11.843, -11.995, -12.063, -12.187, -12.360, -13.191, -12.646,
    -12.246, -12.096, -12.001, -11.807, -11.987, -12.053, -12.193, -12.360,
    -13.175, -12.633, -12.232, -12.084, -11.992, -11.800, -11.981, -12.048,
    -12.192, -12.357, -13.160, -12.620, -12.220, -12.072, -11.984, -11.798,
    -11.976, -12.044, -12.190, -12.355, -13.144, -12.607, -12.209, -12.061,
    -11.976, -11.799, -11.972, -12.041, -12.189, -12.352, -13.128, -12.595,
    -12.201, -12.051, -11.967, -11.804, -11.968, -12.038, -12.187, -12.350,
    -13.113, -12.582, -12.194, -12.041, -11.959, -11.812, -11.964, -12.036,
    -12.184, -12.347, -13.097, -12.570, -12.189, -12.031, -11.951, -11.825,
    -11.961, -12.035, -12.182, -12.345, -13.082, -12.557, -12.185, -12.023,
    -11.943, -11.841, -11.959, -12.035, -12.179, -12.342, -13.067, -12.545,
    -12.184, -12.014, -11.935, -11.861, -11.957, -12.035, -12.175, -12.340,
    -13.051, -12.532, -12.184, -12.007, -11.927, -11.885, -11.955, -12.036,
    -12.172, -12.337, -13.036, -12.520, -12.186, -12.000, -11.919, -11.913,
    -11.954, -12.038, -12.168, -12.335, -13.030, -12.515, -12.183, -11.994,
    -11.914, -11.917, -11.951, -12.036, -12.166, -12.333, -13.027, -12.511,
    -12.180, -11.990, -11.909, -11.919, -11.949, -12.034, -12.163, -12.330,
    -13.025, -12.508, -12.178, -11.986, -11.906, -11.920, -11.946, -12.032,
    -12.161, -12.327, -13.025, -12.506, -12.176, -11.982, -11.902, -11.920,
    -11.944, -12.030, -12.159, -12.325, -13.026, -12.506, -12.174, -11.980,
    -11.899, -11.918, -11.942, -12.028, -12.156, -12.323, -13.030, -12.507,
    -12.174, -11.978, -11.897, -11.916, -11.940, -12.025, -12.154, -12.320,
    -13.035, -12.510, -12.173, -11.977, -11.896, -11.912, -11.938, -12.023,
    -12.152, -12.318, -13.042, -12.514, -12.173, -11.976, -11.894, -11.906,
    -11.936, -12.020, -12.149, -12.315, -13.051, -12.519, -12.174, -11.977,
    -11.894, -11.900, -11.935, -12.018, -12.147, -12.313, -13.062, -12.525,
    -12.175, -11.978, -11.894, -11.892, -11.933, -12.015, -12.145, -12.310,
    -13.065, -12.526, -12.175, -11.977, -11.893, -11.891, -11.932, -12.013,
    -12.143, -12.308, -13.068, -12.527, -12.175, -11.977, -11.892, -11.889,
    -11.930, -12.012, -12.141, -12.305, -13.070, -12.529, -12.176, -11.976,
    -11.891, -11.888, -11.929, -12.010, -12.139, -12.303, -13.073, -12.530,
    -12.176, -11.976, -11.890, -11.888, -11.928, -12.009, -12.137, -12.301,
    -13.076, -12.532, -12.177, -11.977, -11.890, -11.887, -11.927, -12.007,
    -12.136, -12.299, -13.079, -12.533, -12.178, -11.977, -11.890, -11.886,
    -11.926, -12.006, -12.134, -12.297, -13.082, -12.535, -12.180, -11.978,
    -11.890, -11.886, -11.925, -12.005, -12.132, -12.295, -13.085, -12.537,
    -12.181, -11.979, -11.891, -11.885, -11.925, -12.004, -12.131, -12.293,
    -13.088, -12.539, -12.183, -11.980, -11.891, -11.885, -11.924, -12.003,
    -12.129, -12.291, -13.091, -12.541, -12.185, -11.982, -11.892, -11.885,
    -11.924, -12.002, -12.128, -12.289, -13.098, -12.546, -12.189, -11.985,
    -11.894, -11.886, -11.924, -12.002, -12.127, -12.287, -13.105, -12.552,
    -12.193, -11.988, -11.896, -11.887, -11.925, -12.002, -12.126, -12.286,
    -13.113, -12.559, -12.199, -11.992, -11.899, -11.889, -11.926, -12.002,
    -12.126, -12.285, -13.122, -12.566, -12.204, -11.996, -11.902, -11.891,
    -11.927, -12.002, -12.125, -12.283, -13.132, -12.574, -12.210, -12.000,
    -11.905, -11.893, -11.928, -12.002, -12.125, -12.282, -13.142, -12.582,
    -12.217, -12.006, -11.909, -11.896, -11.930, -12.003, -12.125, -12.281,
    -13.153, -12.591, -12.224, -12.011, -11.913, -11.899, -11.931, -12.004,
    -12.124, -12.280, -13.165, -12.601, -12.232, -12.017, -11.918, -11.902,
    -11.933, -12.005, -12.124, -12.279, -13.177, -12.612, -12.240, -12.023,
    -11.923, -11.905, -11.936, -12.006, -12.125, -12.279, -13.190, -12.623,
    -12.249, -12.030, -11.928, -11.909, -11.938, -12.007, -12.125, -12.278,
    -13.202, -12.634, -12.258, -12.037, -11.934, -11.913, -11.941, -12.009,
    -12.125, -12.278, -13.215, -12.645, -12.267, -12.045, -11.940, -11.917,
    -11.944, -12.010, -12.126, -12.277, -13.228, -12.656, -12.277, -12.053,
    -11.946, -11.922, -11.947, -12.012, -12.127, -12.277, -13.241, -12.668,
    -12.287, -12.061, -11.952, -11.927, -11.950, -12.014, -12.128, -12.277,
    -13.255, -12.681, -12.297, -12.070, -11.960, -11.932, -11.954, -12.017,
    -12.129, -12.277, -13.269, -12.694, -12.309, -12.079, -11.967, -11.938,
    -11.958, -12.019, -12.130, -12.277, -13.284, -12.707, -12.320, -12.089,
    -11.975, -11.944, -11.962, -12.022, -12.131, -12.277, -13.299, -12.721,
    -12.332, -12.099, -11.983, -11.950, -11.966, -12.025, -12.133, -12.277,
    -13.314, -12.735, -12.344, -12.109, -11.991, -11.956, -11.971, -12.028,
    -12.134, -12.278, -13.330, -12.750, -12.357, -12.120, -12.000, -11.963,
    -11.976, -12.031, -12.136, -12.278, -13.345, -12.764, -12.369, -12.130,
    -12.008, -11.969, -11.981, -12.034, -12.138, -12.278, -13.360, -12.778,
    -12.381, -12.141, -12.017, -11.976, -11.986, -12.037, -12.140, -12.279,
    -13.375, -12.792, -12.393, -12.151, -12.026, -11.983, -11.991, -12.041,
    -12.142, -12.280, -13.390, -12.806, -12.406, -12.162, -12.035, -11.990,
    -11.996, -12.045, -12.144, -12.280, -13.405, -12.821, -12.419, -12.173,
    -12.044, -11.997, -12.001, -12.048, -12.146, -12.281, -13.421, -12.835,
    -12.432, -12.185, -12.053, -12.005, -12.006, -12.052, -12.148, -12.282,
    -13.436, -12.850, -12.445, -12.196, -12.063, -12.012, -12.012, -12.056,
    -12.151, -12.283, -13.452, -12.865, -12.459, -12.208, -12.073, -12.020,
    -12.018, -12.060, -12.153, -12.284, -13.468, -12.881, -12.472, -12.220,
    -12.083, -12.028, -12.024, -12.065, -12.156, -12.285, -13.484, -12.896,
    -12.486, -12.232, -12.093, -12.036, -12.030, -12.069, -12.159, -12.286,
    -13.499, -12.911, -12.500, -12.244, -12.103, -12.044, -12.036, -12.073,
    -12.162, -12.287, -13.515, -12.926, -12.513, -12.256, -12.113, -12.052,
    -12.042, -12.078, -12.165, -12.289, -13.530, -12.941, -12.527, -12.268,
    -12.123, -12.060, -12.049, -12.082, -12.167, -12.290, -13.545, -12.956,
    -12.541, -12.280, -12.133, -12.068, -12.055, -12.087, -12.170, -12.291,
    -13.561, -12.972, -12.555, -12.292, -12.144, -12.077, -12.062, -12.092,
    -12.174, -12.293, -13.576, -12.987, -12.569, -12.304, -12.154, -12.085,
    -12.068, -12.097, -12.177, -12.295, -13.591, -13.002, -12.583, -12.317,
    -12.165, -12.094, -12.075, -12.102, -12.180, -12.296, -13.606, -13.018,
    -12.597, -12.330, -12.175, -12.102, -12.082, -12.107, -12.183, -12.298,
    -13.622, -13.033, -12.612, -12.342, -12.186, -12.111, -12.089, -12.112,
    -12.187, -12.300, -13.637, -13.049, -12.626, -12.355, -12.197, -12.120,
    -12.096, -12.117, -12.190, -12.302, -13.652, -13.064, -12.640, -12.368,
    -12.208, -12.129, -12.103, -12.122, -12.193, -12.304, -13.667, -13.079,
    -12.654, -12.380, -12.219, -12.138, -12.110, -12.127, -12.197, -12.306,
    -13.681, -13.094, -12.668, -12.393, -12.229, -12.147, -12.117, -12.133,
    -12.200, -12.308, -13.696, -13.109, -12.683, -12.405, -12.240, -12.156,
    -12.124, -12.138, -12.204, -12.310, -13.710, -13.124, -12.697, -12.418,
    -12.251, -12.165, -12.131, -12.144, -12.208, -12.312, -13.724, -13.139,
    -12.711, -12.431, -12.262, -12.174, -12.138, -12.149, -12.211, -12.314,
    -13.739, -13.154, -12.725, -12.443, -12.273, -12.184, -12.146, -12.155,
    -12.215, -12.316, -13.753, -13.169, -12.739, -12.456, -12.285, -12.193,
    -12.153, -12.160, -12.219, -12.318, -13.767, -13.184, -12.754, -12.469,
    -12.296, -12.202, -12.161, -12.166, -12.223, -12.321, -13.781, -13.199,
    -12.768, -12.482, -12.307, -12.212, -12.168, -12.172, -12.227, -12.323,
    -13.794, -13.214, -12.782, -12.495, -12.318, -12.221, -12.175, -12.178,
    -12.231, -12.325, -13.808, -13.228, -12.796, -12.508, -12.329, -12.231,
    -12.183, -12.183, -12.235, -12.327, -13.821, -13.243, -12.810, -12.520,
    -12.341, -12.240, -12.191, -12.189, -12.239, -12.330, -13.834, -13.257,
    -12.824, -12.533, -12.352, -12.250, -12.198, -12.195, -12.243, -12.332,
    -13.847, -13.271, -12.838, -12.546, -12.363, -12.259, -12.206, -12.201,
    -12.247, -12.334, -13.859, -13.286, -12.852, -12.559, -12.375, -12.269,
    -12.214, -12.206, -12.251, -12.337, -13.872, -13.300, -12.866, -12.572,
    -12.386, -12.278, -12.221, -12.212, -12.255, -12.339, -13.884, -13.314,
    -12.880, -12.585, -12.398, -12.288, -12.229, -12.218, -12.259, -12.341,
    -13.896, -13.328, -12.894, -12.598, -12.409, -12.297, -12.237, -12.224,
    -12.263, -12.344, -13.908, -13.342, -12.908, -12.611, -12.421, -12.307,
    -12.245, -12.230, -12.267, -12.346, -13.919, -13.356, -12.922, -12.624,
    -12.432, -12.317, -12.253, -12.236, -12.271, -12.348, -13.930, -13.369,
    -12.935, -12.637, -12.444, -12.326, -12.261, -12.242, -12.275, -12.351,
    -13.941, -13.383, -12.949, -12.650, -12.455, -12.336, -12.269, -12.248,
    -12.279, -12.353, -13.952, -13.396, -12.963, -12.662, -12.467, -12.346,
    -12.277, -12.254, -12.284, -12.356, -13.962, -13.409, -12.976, -12.675,
    -12.478, -12.356, -12.285, -12.260, -12.288, -12.359, -13.972, -13.422,
    -12.990, -12.688, -12.489, -12.365, -12.293, -12.266, -12.292, -12.361,
    -13.981, -13.435, -13.003, -12.701, -12.501, -12.375, -12.301, -12.272,
    -12.296, -12.364, -13.991, -13.448, -13.016, -12.713, -12.512, -12.385,
    -12.309, -12.279, -12.300, -12.367, -14.000, -13.460, -13.030, -12.726,
    -12.524, -12.395, -12.317, -12.285, -12.305, -12.369, -14.009, -13.473,
    -13.043, -12.739, -12.535, -12.405, -12.325, -12.291, -12.309, -12.372,
    -14.018, -13.485, -13.056, -12.752, -12.546, -12.415, -12.333, -12.297,
    -12.313, -12.375, -14.026, -13.498, -13.069, -12.764, -12.558, -12.425,
    -12.341, -12.303, -12.318, -12.377, -14.034, -13.510, -13.083, -12.777,
    -12.569, -12.434, -12.349, -12.309, -12.322, -12.380, -14.041, -13.522,
    -13.096, -12.790, -12.581, -12.444, -12.357, -12.315, -12.326, -12.382,
    -14.049, -13.534, -13.109, -12.802, -12.592, -12.454, -12.365, -12.322,
    -12.330, -12.385, -14.056, -13.545, -13.122, -12.815, -12.603, -12.464,
    -12.373, -12.328, -12.335, -12.388, -14.062, -13.557, -13.135, -12.827,
    -12.615, -12.474, -12.381, -12.334, -12.339, -12.390, -14.069, -13.568,
    -13.147, -12.840, -12.626, -12.483, -12.389, -12.340, -12.343, -12.393,
    -14.075, -13.580, -13.160, -12.852, -12.638, -12.493, -12.397, -12.346,
    -12.348, -12.395, -14.081, -13.591, -13.173, -12.865, -12.649, -12.503,
    -12.405, -12.352, -12.352, -12.398, -14.086, -13.602, -13.186, -12.877,
    -12.660, -12.513, -12.413, -12.358, -12.356, -12.401, -14.091, -13.613,
    -13.198, -12.890, -12.672, -12.522, -12.421, -12.364, -12.361, -12.403,
    -14.095, -13.623, -13.211, -12.902, -12.683, -12.532, -12.429, -12.370,
    -12.365, -12.406, -14.099, -13.633, -13.223, -12.914, -12.694, -12.542,
    -12.437, -12.377, -12.369, -12.408, -14.102, -13.644, -13.236, -12.927,
    -12.706, -12.552, -12.445, -12.383, -12.374, -12.411, -14.105, -13.654,
    -13.248, -12.939, -12.717, -12.561, -12.453, -12.389, -12.378, -12.413,
    -14.108, -13.663, -13.260, -12.951, -12.728, -12.571, -12.461, -12.395,
    -12.382, -12.416, -14.111, -13.673, -13.273, -12.963, -12.739, -12.581,
    -12.469, -12.402, -12.386, -12.418, -14.112, -13.683, -13.285, -12.975,
    -12.751, -12.590, -12.477, -12.408, -12.391, -12.421, -14.114, -13.692,
    -13.297, -12.987, -12.762, -12.600, -12.485, -12.414, -12.395, -12.423,
    -14.117, -13.702, -13.309, -12.999, -12.773, -12.610, -12.493, -12.420,
    -12.399, -12.426, -14.119, -13.711, -13.321, -13.011, -12.784, -12.619,
    -12.501, -12.426, -12.403, -12.428, -14.121, -13.721, -13.333, -13.023,
    -12.795, -12.629, -12.509, -12.432, -12.408, -12.431, -14.123, -13.730,
    -13.345, -13.035, -12.806, -12.638, -12.516, -12.438, -12.412, -12.433,
    -14.125, -13.739, -13.357, -13.047, -12.817, -12.648, -12.524, -12.444,
    -12.416, -12.436, -14.127, -13.748, -13.369, -13.059, -12.828, -12.657,
    -12.532, -12.450, -12.420, -12.438, -14.129, -13.757, -13.380, -13.071,
    -12.839, -12.667, -12.540, -12.456, -12.425, -12.441, -14.130, -13.766,
    -13.392, -13.083, -12.850, -12.676, -12.548, -12.461, -12.429, -12.444,
    -14.132, -13.775, -13.403, -13.095, -12.861, -12.686, -12.555, -12.467,
    -12.433, -12.446, -14.133, -13.784, -13.415, -13.107, -12.872, -12.695,
    -12.563, -12.473, -12.437, -12.449, -14.138, -13.794, -13.427, -13.119,
    -12.883, -12.704, -12.571, -12.479, -12.441, -12.451, -14.143, -13.804,
    -13.439, -13.131, -12.894, -12.714, -12.578, -12.485, -12.445, -12.454,
    -14.149, -13.814, -13.451, -13.143, -12.904, -12.723, -12.586, -12.490,
    -12.449, -12.456, -14.155, -13.825, -13.462, -13.154, -12.915, -12.732,
    -12.593, -12.496, -12.453, -12.459, -14.162, -13.835, -13.474, -13.166,
    -12.926, -12.742, -12.601, -12.502, -12.457, -12.461, -14.169, -13.846,
    -13.486, -13.178, -12.937, -12.751, -12.608, -12.507, -12.461, -12.463,
    -14.177, -13.857, -13.498, -13.190, -12.947, -12.760, -12.616, -12.513,
    -12.465, -12.465, -14.186, -13.868, -13.510, -13.202, -12.958, -12.770,
    -12.623, -12.519, -12.468, -12.468, -14.195, -13.879, -13.522, -13.213,
    -12.968, -12.779, -12.631, -12.524, -12.472, -12.470, -14.204, -13.890,
    -13.534, -13.225, -12.979, -12.788, -12.638, -12.530, -12.476, -12.472,
    -14.211, -13.900, -13.546, -13.237, -12.990, -12.797, -12.645, -12.536,
    -12.480, -12.474, -14.217, -13.910, -13.557, -13.248, -13.000, -12.806,
    -12.652, -12.541, -12.483, -12.476, -14.224, -13.921, -13.569, -13.259,
    -13.011, -12.815, -12.660, -12.547, -12.487, -12.479, -14.230, -13.931,
    -13.580, -13.271, -13.021, -12.824, -12.667, -12.552, -12.491, -12.481,
    -14.237, -13.941, -13.592, -13.282, -13.032, -12.833, -12.674, -12.557,
    -12.494, -12.483, -14.243, -13.951, -13.603, -13.293, -13.042, -12.842,
    -12.681, -12.563, -12.498, -12.485, -14.250, -13.961, -13.615, -13.305,
    -13.053, -12.851, -12.688, -12.568, -12.502, -12.487, -14.256, -13.970,
    -13.626, -13.316, -13.063, -12.860, -12.694, -12.573, -12.505, -12.489,
    -14.262, -13.980, -13.638, -13.327, -13.074, -12.869, -12.701, -12.579,
    -12.509, -12.491, -14.268, -13.990, -13.649, -13.338, -13.084, -12.878,
    -12.708, -12.584, -12.512, -12.493, -14.277, -14.001, -13.661, -13.349,
    -13.094, -12.887, -12.715, -12.589, -12.515, -12.495, -14.286, -14.012,
    -13.672, -13.360, -13.104, -12.895, -12.722, -12.594, -12.519, -12.497,
    -14.296, -14.023, -13.684, -13.372, -13.115, -12.904, -12.728, -12.599,
    -12.522, -12.499, -14.307, -14.034, -13.695, -13.383, -13.125, -12.912,
    -12.735, -12.604, -12.526, -12.501, -14.318, -14.045, -13.707, -13.394,
    -13.135, -12.921, -12.742, -12.609, -12.529, -12.503, -14.329, -14.057,
    -13.718, -13.405, -13.145, -12.929, -12.749, -12.614, -12.533, -12.505,
    -14.341, -14.068, -13.730, -13.417, -13.155, -12.937, -12.755, -12.618,
    -12.536, -12.507, -14.353, -14.080, -13.742, -13.428, -13.164, -12.946,
    -12.762, -12.623, -12.539, -12.509, -14.366, -14.092, -13.753, -13.439,
    -13.174, -12.954, -12.768, -12.627, -12.543, -12.511, -14.379, -14.104,
    -13.765, -13.450, -13.184, -12.962, -12.775, -12.632, -12.546, -12.513,
    -14.391, -14.116, -13.776, -13.461, -13.194, -12.970, -12.781, -12.637,
    -12.549, -12.515, -14.403, -14.127, -13.788, -13.472, -13.204, -12.978,
    -12.788, -12.641, -12.552, -12.516, -14.416, -14.139, -13.799, -13.483,
    -13.213, -12.986, -12.794, -12.646, -12.555, -12.518, -14.428, -14.151,
    -13.811, -13.493, -13.223, -12.994, -12.800, -12.650, -12.558, -12.520,
    -14.441, -14.163, -13.822, -13.504, -13.233, -13.002, -12.807, -12.655,
    -12.561, -12.521, -14.454, -14.175, -13.834, -13.515, -13.242, -13.010,
    -12.813, -12.659, -12.564, -12.523, -14.467, -14.187, -13.845, -13.525,
    -13.252, -13.018, -12.819, -12.663, -12.566, -12.524, -14.481, -14.199,
    -13.856, -13.536, -13.261, -13.026, -12.825, -12.668, -12.569, -12.525,
    -14.494, -14.211, -13.868, -13.547, -13.271, -13.033, -12.831, -12.672,
    -12.571, -12.527, -14.508, -14.223, -13.879, -13.557, -13.280, -13.041,
    -12.837, -12.676, -12.574, -12.528, -14.521, -14.235, -13.890, -13.568,
    -13.289, -13.049, -12.843, -12.680, -12.577, -12.530, -14.534, -14.247,
    -13.902, -13.578, -13.298, -13.056, -12.848, -12.684, -12.579, -12.531,
    -14.548, -14.259, -13.913, -13.589, -13.308, -13.063, -12.854, -12.687,
    -12.582, -12.532, -14.561, -14.271, -13.925, -13.599, -13.317, -13.071,
    -12.860, -12.691, -12.585, -12.534, -14.574, -14.283, -13.936, -13.609,
    -13.326, -13.078, -12.865, -12.695, -12.588, -12.535, -14.588, -14.295,
    -13.947, -13.620, -13.335, -13.085, -12.870, -12.698, -12.590, -12.537,
    -14.601, -14.307, -13.959, -13.630, -13.344, -13.092, -12.876, -12.701,
    -12.593, -12.538, -14.615, -14.320, -13.970, -13.640, -13.352, -13.099,
    -12.881, -12.705, -12.596, -12.540, -14.628, -14.332, -13.982, -13.651,
    -13.361, -13.106, -12.886, -12.708, -12.598, -12.542, -14.642, -14.344,
    -13.993, -13.661, -13.370, -13.113, -12.891, -12.711, -12.601, -12.543,
    -14.655, -14.356, -14.004, -13.671, -13.379, -13.120, -12.896, -12.715,
    -12.603, -12.544, -14.669, -14.368, -14.015, -13.681, -13.387, -13.126,
    -12.901, -12.718, -12.606, -12.546, -14.682, -14.380, -14.026, -13.691,
    -13.396, -13.133, -12.906, -12.722, -12.608, -12.547, -14.695, -14.392,
    -14.037, -13.701, -13.404, -13.139, -12.911, -12.726, -12.610, -12.548,
    -14.709, -14.404, -14.049, -13.711, -13.412, -13.146, -12.915, -12.729,
    -12.612, -12.549, -14.722, -14.416, -14.059, -13.721, -13.421, -13.152,
    -12.920, -12.733, -12.614, -12.550, -14.735, -14.428, -14.070, -13.730,
    -13.429, -13.158, -12.925, -12.737, -12.616, -12.551, -14.748, -14.439,
    -14.081, -13.740, -13.437, -13.164, -12.929, -12.740, -12.618, -12.552,
    -14.762, -14.451, -14.092, -13.749, -13.445, -13.170, -12.934, -12.744,
    -12.619, -12.553, -14.775, -14.463, -14.103, -13.759, -13.453, -13.176,
    -12.938, -12.748, -12.621, -12.554, -14.788, -14.475, -14.114, -13.768,
    -13.461, -13.182, -12.942, -12.751, -12.622, -12.555, -14.801, -14.487,
    -14.124, -13.778, -13.469, -13.188, -12.947, -12.754, -12.624, -12.556,
    -14.814, -14.498, -14.135, -13.787, -13.476, -13.194, -12.951, -12.757,
    -12.625, -12.557, -14.827, -14.510, -14.145, -13.797, -13.484, -13.200,
    -12.955, -12.760, -12.626, -12.558, -14.840, -14.522, -14.156, -13.806,
    -13.491, -13.206, -12.959, -12.762, -12.627, -12.559, -14.852, -14.533,
    -14.166, -13.815, -13.499, -13.211, -12.963, -12.765, -12.628, -12.561,
    -14.865, -14.545, -14.176, -13.824, -13.506, -13.217, -12.967, -12.767,
    -12.628, -12.562, -14.878, -14.557, -14.187, -13.833, -13.514, -13.223,
    -12.971, -12.770, -12.629, -12.563, -14.890, -14.568, -14.197, -13.842,
    -13.521, -13.228, -12.975, -12.772, -12.630, -12.564, -14.903, -14.580,
    -14.207, -13.851, -13.528, -13.234, -12.979, -12.774, -12.630, -12.565,
    -14.916, -14.591, -14.217, -13.860, -13.534, -13.239, -12.983, -12.777,
    -12.632, -12.566, -14.928, -14.603, -14.228, -13.869, -13.540, -13.245,
    -12.987, -12.779, -12.634, -12.568, -14.940, -14.614, -14.238, -13.878,
    -13.546, -13.250, -12.990, -12.782, -12.637, -12.569, -14.953, -14.625,
    -14.248, -13.886, -13.552, -13.255, -12.994, -12.784, -12.639, -12.571,
    -14.965, -14.637, -14.258, -13.895, -13.557, -13.260, -12.998, -12.787,
    -12.642, -12.573, -14.977, -14.648, -14.268, -13.904, -13.562, -13.265,
    -13.001, -12.790, -12.645, -12.574, -14.990, -14.659, -14.279, -13.912,
    -13.567, -13.270, -13.005, -12.792, -12.649, -12.576, -15.002, -14.670,
    -14.289, -13.921, -13.571, -13.275, -13.008, -12.795, -12.652, -12.578,
    -15.014, -14.681, -14.299, -13.930, -13.576, -13.279, -13.012, -12.797,
    -12.656, -12.580, -15.026, -14.692, -14.309, -13.938, -13.580, -13.284,
    -13.015, -12.800, -12.660, -12.582, -15.038, -14.703, -14.319, -13.946,
    -13.587, -13.289, -13.018, -12.803, -12.662, -12.583, -15.050, -14.714,
    -14.328, -13.955, -13.594, -13.293, -13.022, -12.806, -12.664, -12.584,
    -15.062, -14.725, -14.338, -13.963, -13.601, -13.298, -13.025, -12.809,
    -12.666, -12.584, -15.074, -14.736, -14.348, -13.971, -13.609, -13.302,
    -13.029, -12.813, -12.668, -12.585, -15.085, -14.747, -14.357, -13.980,
    -13.617, -13.306, -13.032, -12.816, -12.669, -12.585, -15.097, -14.758,
    -14.366, -13.988, -13.625, -13.311, -13.036, -12.819, -12.670, -12.585,
    -15.109, -14.769, -14.376, -13.996, -13.633, -13.315, -13.039, -12.823,
    -12.671, -12.585, -15.120, -14.779, -14.385, -14.004, -13.642, -13.319,
    -13.042, -12.827, -12.672, -12.585, -15.132, -14.790, -14.394, -14.012,
    -13.651, -13.323, -13.046, -12.830, -12.673, -12.584, -15.143, -14.801,
    -14.403, -14.020, -13.660, -13.327, -13.049, -12.834, -12.674, -12.584,
    -15.154, -14.812, -14.412, -14.028, -13.666, -13.331, -13.052, -12.837,
    -12.676, -12.585, -15.166, -14.822, -14.422, -14.036, -13.672, -13.336,
    -13.055, -12.839, -12.678, -12.586, -15.177, -14.833, -14.431, -14.043,
    -13.677, -13.340, -13.058, -12.841, -12.680, -12.588, -15.188, -14.843,
    -14.441, -14.051, -13.682, -13.345, -13.061, -12.843, -12.682, -12.589,
    -15.199, -14.854, -14.450, -14.059, -13.686, -13.349, -13.064, -12.845,
    -12.684, -12.591, -15.210, -14.864, -14.460, -14.066, -13.690, -13.354,
    -13.067, -12.847, -12.687, -12.592, -15.221, -14.875, -14.469, -14.074,
    -13.694, -13.358, -13.069, -12.849, -12.689, -12.594, -15.232, -14.885,
    -14.479, -14.081, -13.697, -13.363, -13.072, -12.850, -12.692, -12.596,
    -15.243, -14.896, -14.488, -14.089, -13.700, -13.367, -13.074, -12.852,
    -12.694, -12.599, -15.254, -14.906, -14.498, -14.096, -13.703, -13.372,
    -13.077, -12.853, -12.697, -12.601, -15.265, -14.916, -14.507, -14.104,
    -13.709, -13.377, -13.081, -12.856, -12.699, -12.603, -15.276, -14.927,
    -14.516, -14.112, -13.716, -13.382, -13.085, -12.859, -12.702, -12.604,
    -15.286, -14.937, -14.526, -14.119, -13.722, -13.387, -13.089, -12.862,
    -12.704, -12.606, -15.297, -14.947, -14.535, -14.127, -13.729, -13.392,
    -13.094, -12.865, -12.707, -12.607, -15.308, -14.957, -14.544, -14.135,
    -13.737, -13.398, -13.099, -12.868, -12.710, -12.609, -15.318, -14.968,
    -14.553, -14.143, -13.745, -13.403, -13.103, -12.872, -12.712, -12.610,
    -15.329, -14.978, -14.562, -14.151, -13.753, -13.408, -13.108, -12.876,
    -12.715, -12.612, -15.339, -14.988, -14.570, -14.159, -13.761, -13.414,
    -13.114, -12.880, -12.718, -12.613, -15.350, -14.998, -14.579, -14.167,
    -13.770, -13.419, -13.119, -12.884, -12.720, -12.615, -15.360, -15.008,
    -14.588, -14.175, -13.779, -13.425, -13.125, -12.888, -12.723, -12.616,
    -15.370, -15.018, -14.597, -14.183, -13.786, -13.430, -13.130, -12.891,
    -12.726, -12.618, -15.381, -15.028, -14.606, -14.191, -13.793, -13.436,
    -13.134, -12.895, -12.728, -12.619, -15.391, -15.038, -14.615, -14.199,
    -13.800, -13.441, -13.139, -12.898, -12.731, -12.620, -15.401, -15.048,
    -14.624, -14.207, -13.807, -13.447, -13.143, -12.902, -12.733, -12.622,
    -15.412, -15.058, -14.633, -14.215, -13.814, -13.452, -13.148, -12.905,
    -12.736, -12.623, -15.422, -15.067, -14.642, -14.223, -13.821, -13.458,
    -13.152, -12.909, -12.738, -12.625, -15.432, -15.077, -14.652, -14.231,
    -13.827, -13.463, -13.156, -12.912, -12.741, -12.626, -15.442, -15.087,
    -14.661, -14.239, -13.834, -13.469, -13.161, -12.915, -12.743, -12.628,
    -15.452, -15.096, -14.670, -14.247, -13.841, -13.474, -13.165, -12.919,
    -12.746, -12.629, -15.462, -15.106, -14.679, -14.255, -13.847, -13.480,
    -13.169, -12.922, -12.748, -12.631, -15.472, -15.116, -14.688, -14.263,
    -13.854, -13.486, -13.174, -12.926, -12.750, -12.632, -15.482, -15.125,
    -14.697, -14.271, -13.861, -13.491, -13.178, -12.929, -12.753, -12.634,
    -15.492, -15.135, -14.706, -14.279, -13.867, -13.497, -13.183, -12.933,
    -12.756, -12.635, -15.501, -15.144, -14.714, -14.287, -13.874, -13.503,
    -13.187, -12.936, -12.758, -12.637, -15.511, -15.153, -14.723, -14.295,
    -13.881, -13.508, -13.192, -12.940, -12.760, -12.638, -15.521, -15.163,
    -14.732, -14.303, -13.888, -13.514, -13.196, -12.944, -12.763, -12.640,
    -15.530, -15.172, -14.740, -14.310, -13.895, -13.520, -13.201, -12.948,
    -12.765, -12.641, -15.540, -15.182, -14.749, -14.318, -13.901, -13.525,
    -13.206, -12.951, -12.768, -12.642, -15.550, -15.191, -14.757, -14.326,
    -13.908, -13.531, -13.210, -12.955, -12.771, -12.644, -15.559, -15.200,
    -14.766, -14.334, -13.915, -13.537, -13.215, -12.959, -12.773, -12.645,
    -15.568, -15.209, -14.775, -14.342, -13.922, -13.543, -13.220, -12.963,
    -12.776, -12.647, -15.578, -15.218, -14.784, -14.350, -13.929, -13.549,
    -13.225, -12.967, -12.779, -12.648, -15.587, -15.227, -14.792, -14.358,
    -13.937, -13.556, -13.230, -12.971, -12.781, -12.650, -15.597, -15.236,
    -14.801, -14.367, -13.944, -13.562, -13.236, -12.975, -12.784, -12.651,
    -15.606, -15.245, -14.810, -14.375, -13.952, -13.569, -13.241, -12.979,
    -12.787, -12.653, -15.615, -15.254, -14.819, -14.383, -13.959, -13.575,
    -13.246, -12.984, -12.790, -12.654, -15.625, -15.263, -14.827, -14.392,
    -13.967, -13.582, -13.252, -12.988, -12.793, -12.656, -15.634, -15.272,
    -14.836, -14.400, -13.974, -13.588, -13.258, -12.992, -12.797, -12.658,
    -15.643, -15.281, -14.845, -14.409, -13.982, -13.595, -13.263, -12.997,
    -12.800, -12.659, -15.652, -15.290, -14.854, -14.417, -13.990, -13.602,
    -13.269, -13.001, -12.803, -12.661, -15.661, -15.299, -14.863, -14.425,
    -13.998, -13.609, -13.275, -13.006, -12.806, -12.663, -15.670, -15.307,
    -14.871, -14.434, -14.006, -13.616, -13.281, -13.010, -12.809, -12.665,
    -15.679, -15.316, -14.880, -14.442, -14.013, -13.623, -13.287, -13.015,
    -12.813, -12.666, -15.688, -15.325, -14.888, -14.450, -14.021, -13.630,
    -13.293, -13.019, -12.816, -12.668, -15.697, -15.333, -14.896, -14.459,
    -14.029, -13.637, -13.299, -13.024, -12.819, -12.670, -15.706, -15.342,
    -14.905, -14.467, -14.037, -13.645, -13.305, -13.029, -12.822, -12.672,
    -15.715, -15.350, -14.913, -14.476, -14.045, -13.652, -13.312, -13.034,
    -12.826, -12.674, -15.724, -15.358, -14.922, -14.484, -14.054, -13.660,
    -13.318, -13.039, -12.829, -12.676, -15.733, -15.367, -14.930, -14.493,
    -14.062, -13.667, -13.324, -13.044, -12.833, -12.678, -15.742, -15.375,
    -14.938, -14.501, -14.070, -13.675, -13.331, -13.049, -12.836, -12.680,
    -15.751, -15.383, -14.946, -14.509, -14.078, -13.682, -13.337, -13.054,
    -12.840, -12.682, -15.759, -15.391, -14.954, -14.518, -14.086, -13.690,
    -13.344, -13.059, -12.843, -12.684, -15.768, -15.399, -14.962, -14.526,
    -14.094, -13.697, -13.350, -13.064, -12.847, -12.686, -15.776, -15.407,
    -14.970, -14.534, -14.103, -13.705, -13.357, -13.069, -12.850, -12.688,
    -15.785, -15.415, -14.978, -14.542, -14.111, -13.713, -13.363, -13.075,
    -12.854, -12.690, -15.793, -15.423, -14.986, -14.550, -14.119, -13.720,
    -13.370, -13.080, -12.858, -12.692, -15.801, -15.431, -14.994, -14.559,
    -14.127, -13.728, -13.376, -13.085, -12.861, -12.694, -15.810, -15.439,
    -15.002, -14.567, -14.135, -13.736, -13.383, -13.090, -12.865, -12.696,
    -15.818, -15.446, -15.010, -14.575, -14.144, -13.743, -13.389, -13.096,
    -12.869, -12.698, -15.826, -15.454, -15.018, -14.583, -14.152, -13.751,
    -13.396, -13.101, -12.873, -12.700, -15.834, -15.461, -15.025, -14.591,
    -14.160, -13.759, -13.403, -13.106, -12.877, -12.702, -15.842, -15.469,
    -15.033, -14.599, -14.168, -13.766, -13.409, -13.112, -12.880, -12.704,
    -15.850, -15.476, -15.040, -14.606, -14.176, -13.774, -13.416, -13.117,
    -12.884, -12.706, -15.857, -15.483, -15.047, -14.614, -14.184, -13.781,
    -13.423, -13.123, -12.888, -12.708, -15.865, -15.491, -15.054, -14.622,
    -14.192, -13.789, -13.430, -13.128, -12.892, -12.710, -15.873, -15.498,
    -15.061, -14.629, -14.200, -13.797, -13.436, -13.134, -12.895, -12.712,
    -15.880, -15.505, -15.068, -14.637, -14.208, -13.804, -13.443, -13.139,
    -12.899, -12.714, -15.887, -15.511, -15.075, -14.644, -14.216, -13.812,
    -13.450, -13.145, -12.903, -12.716, -15.895, -15.518, -15.081, -14.652,
    -14.224, -13.819, -13.457, -13.150, -12.906, -12.718, -15.902, -15.525,
    -15.088, -14.659, -14.232, -13.827, -13.464, -13.156, -12.910, -12.720,
    -15.909, -15.531, -15.095, -14.666, -14.240, -13.835, -13.471, -13.162,
    -12.914, -12.722, -15.916, -15.537, -15.101, -14.673, -14.248, -13.842,
    -13.478, -13.167, -12.918, -12.724, -15.923, -15.543, -15.107, -14.680,
    -14.255, -13.850, -13.485, -13.173, -12.921, -12.726, -15.930, -15.549,
    -15.113, -14.687, -14.263, -13.858, -13.491, -13.178, -12.925, -12.728,
    -15.937, -15.555, -15.120, -14.694, -14.271, -13.865, -13.498, -13.184,
    -12.929, -12.730, -15.943, -15.560, -15.126, -14.701, -14.279, -13.873,
    -13.505, -13.189, -12.933, -12.732, -15.950, -15.565, -15.132, -14.708,
    -14.286, -13.881, -13.512, -13.194, -12.936, -12.734, -15.956, -15.570,
    -15.138, -14.715, -14.294, -13.889, -13.519, -13.200, -12.940, -12.736,
    -15.963, -15.575, -15.143, -14.721, -14.301, -13.896, -13.526, -13.206,
    -12.944, -12.738, -15.969, -15.580, -15.149, -14.728, -14.309, -13.904,
    -13.533, -13.211, -12.948, -12.740, -15.975, -15.584, -15.154, -14.734,
    -14.316, -13.912, -13.540, -13.216, -12.952, -12.742, -15.980, -15.588,
    -15.159, -14.741, -14.324, -13.919, -13.547, -13.222, -12.955, -12.744,
    -15.986, -15.592, -15.163, -14.747, -14.331, -13.927, -13.553, -13.227,
    -12.959, -12.745, -15.991, -15.595, -15.167, -14.753, -14.338, -13.934,
    -13.560, -13.233, -12.962, -12.747, -15.997, -15.599, -15.171, -14.759,
    -14.345, -13.942, -13.567, -13.238, -12.966, -12.749, -16.002, -15.601,
    -15.175, -14.765, -14.352, -13.949, -13.573, -13.244, -12.969, -12.751,
    -16.006, -15.604, -15.179, -14.770, -14.359, -13.957, -13.580, -13.249,
    -12.973, -12.752, -16.011, -15.607, -15.182, -14.776, -14.366, -13.964,
    -13.587, -13.254, -12.976, -12.754, -16.016, -15.609, -15.185, -14.782,
    -14.373, -13.972, -13.593, -13.260, -12.980, -12.755, -16.020, -15.611,
    -15.188, -14.787, -14.380, -13.979, -13.600, -13.265, -12.983, -12.757,
    -16.024, -15.613, -15.191, -14.792, -14.387, -13.986, -13.607, -13.270,
    -12.986, -12.759, -16.027, -15.615, -15.195, -14.797, -14.394, -13.993,
    -13.613, -13.276, -12.990, -12.760, -16.030, -15.617, -15.198, -14.803,
    -14.400, -14.001, -13.620, -13.281, -12.993, -12.762, -16.033, -15.618,
    -15.201, -14.808, -14.407, -14.008, -13.627, -13.286, -12.996, -12.764,
    -16.036, -15.620, -15.204, -14.812, -14.414, -14.015, -13.634, -13.291,
    -12.999, -12.765, -16.039, -15.621, -15.207, -14.817, -14.420, -14.022,
    -13.640, -13.296, -13.003, -12.767, -16.041, -15.622, -15.209, -14.822,
    -14.427, -14.029, -13.647, -13.302, -13.006, -12.768, -16.043, -15.623,
    -15.212, -14.826, -14.433, -14.035, -13.654, -13.307, -13.009, -12.770,
    -16.045, -15.624, -15.215, -14.831, -14.440, -14.042, -13.660, -13.312,
    -13.012, -12.771, -16.046, -15.624, -15.217, -14.835, -14.446, -14.049,
    -13.667, -13.317, -13.015, -12.773, -16.047, -15.624, -15.218, -14.839,
    -14.452, -14.056, -13.674, -13.322, -13.018, -12.774, -16.048, -15.624,
    -15.219, -14.844, -14.459, -14.063, -13.680, -13.327, -13.021, -12.776,
    -16.048, -15.624, -15.220, -14.848, -14.465, -14.070, -13.687, -13.332,
    -13.024, -12.777, -16.048, -15.623, -15.220, -14.852, -14.471, -14.077,
    -13.693, -13.337, -13.028, -12.778, -16.048, -15.622, -15.221, -14.856,
    -14.477, -14.084, -13.700, -13.342, -13.031, -12.779, -16.048, -15.621,
    -15.220, -14.859, -14.483, -14.091, -13.706, -13.347, -13.034, -12.780,
    -16.047, -15.620, -15.220, -14.863, -14.489, -14.099, -13.713, -13.352,
    -13.037, -12.781, -16.046, -15.619, -15.219, -14.867, -14.495, -14.106,
    -13.719, -13.356, -13.040, -12.782, -16.045, -15.617, -15.218, -14.871,
    -14.501, -14.113, -13.726, -13.361, -13.043, -12.783, -16.043, -15.615,
    -15.217, -14.874, -14.507, -14.120, -13.732, -13.366, -13.046, -12.784,
    -16.041, -15.612, -15.217, -14.878, -14.512, -14.127, -13.738, -13.371,
    -13.049, -12.785, -16.039, -15.609, -15.218, -14.881, -14.517, -14.134,
    -13.745, -13.375, -13.051, -12.786, -16.037, -15.605, -15.219, -14.885,
    -14.522, -14.141, -13.751, -13.380, -13.054, -12.787, -16.035, -15.601,
    -15.219, -14.888, -14.526, -14.147, -13.757, -13.384, -13.057, -12.788,
    -16.032, -15.596, -15.220, -14.892, -14.531, -14.154, -13.763, -13.389,
    -13.059, -12.788, -16.029, -15.591, -15.221, -14.895, -14.535, -14.161,
    -13.770, -13.393, -13.062, -12.789, -16.025, -15.586, -15.222, -14.899,
    -14.539, -14.167, -13.776, -13.398, -13.064, -12.790, -16.022, -15.580,
    -15.223, -14.902, -14.543, -14.174, -13.782, -13.402, -13.066, -12.791,
    -16.018, -15.574, -15.224, -14.906, -14.547, -14.180, -13.788, -13.407,
    -13.069, -12.791, -16.014, -15.568, -15.225, -14.909, -14.550, -14.187,
    -13.794, -13.411, -13.071, -12.792, -16.009, -15.563, -15.225, -14.912,
    -14.556, -14.194, -13.800, -13.415, -13.073, -12.793, -16.003, -15.559,
    -15.224, -14.915, -14.562, -14.200, -13.806, -13.420, -13.076, -12.793,
    -15.997, -15.554, -15.223, -14.918, -14.568, -14.207, -13.812, -13.424,
    -13.078, -12.793, -15.991, -15.549, -15.222, -14.920, -14.574, -14.213,
    -13.818, -13.428, -13.080, -12.794, -15.984, -15.545, -15.221, -14.923,
    -14.581, -14.220, -13.823, -13.432, -13.082, -12.794, -15.977, -15.540,
    -15.220, -14.926, -14.588, -14.226, -13.829, -13.436, -13.085, -12.794,
    -15.969, -15.535, -15.218, -14.928, -14.596, -14.233, -13.835, -13.440,
    -13.087, -12.795, -15.961, -15.530, -15.216, -14.931, -14.603, -14.239,
    -13.841, -13.444, -13.089, -12.795, -15.953, -15.526, -15.214, -14.933,
    -14.611, -14.246, -13.846, -13.448, -13.091, -12.795, -15.944, -15.521,
    -15.212, -14.935, -14.619, -14.252, -13.852, -13.452, -13.093, -12.795,
    -15.942, -15.521, -15.213, -14.939, -14.626, -14.259, -13.858, -13.456,
    -13.095, -12.795, -15.942, -15.521, -15.215, -14.943, -14.632, -14.265,
    -13.863, -13.460, -13.096, -12.795, -15.942, -15.522, -15.217, -14.947,
    -14.638, -14.272, -13.868, -13.464, -13.098, -12.796, -15.943, -15.524,
    -15.220, -14.951, -14.645, -14.278, -13.874, -13.467, -13.100, -12.796,
    -15.945, -15.526, -15.223, -14.956, -14.651, -14.285, -13.879, -13.471,
    -13.101, -12.796, -15.949, -15.530, -15.226, -14.961, -14.657, -14.291,
    -13.884, -13.475, -13.102, -12.796, -15.953, -15.534, -15.230, -14.966,
    -14.663, -14.298, -13.890, -13.478, -13.104, -12.797, -15.958, -15.539,
    -15.235, -14.971, -14.669, -14.305, -13.895, -13.482, -13.105, -12.797,
    -15.964, -15.545, -15.240, -14.976, -14.675, -14.311, -13.900, -13.485,
    -13.106, -12.797, -15.971, -15.551, -15.245, -14.982, -14.681, -14.318,
    -13.905, -13.489, -13.107, -12.797, -15.975, -15.555, -15.249, -14.987,
    -14.687, -14.324, -13.910, -13.492, -13.108, -12.797, -15.979, -15.558,
    -15.253, -14.992, -14.694, -14.331, -13.915, -13.495, -13.110, -12.797,
    -15.983, -15.562, -15.257, -14.998, -14.700, -14.337, -13.921, -13.499,
    -13.111, -12.797, -15.987, -15.566, -15.262, -15.003, -14.706, -14.344,
    -13.926, -13.502, -13.112, -12.796, -15.991, -15.570, -15.266, -15.009,
    -14.713, -14.350, -13.931, -13.505, -13.113, -12.796, -15.996, -15.574,
    -15.271, -15.015, -14.719, -14.356, -13.936, -13.507, -13.114, -12.796,
    -16.001, -15.579, -15.276, -15.020, -14.726, -14.362, -13.942, -13.510,
    -13.116, -12.795, -16.006, -15.583, -15.281, -15.026, -14.732, -14.369,
    -13.947, -13.513, -13.117, -12.795, -16.011, -15.587, -15.286, -15.032,
    -14.739, -14.375, -13.952, -13.515, -13.118, -12.795, -16.016, -15.592,
    -15.291, -15.038, -14.745, -14.381, -13.957, -13.518, -13.119, -12.794,
    -16.022, -15.597, -15.297, -15.044, -14.752, -14.387, -13.962, -13.521,
    -13.120, -12.793, -16.028, -15.603, -15.302, -15.050, -14.758, -14.393,
    -13.967, -13.523, -13.120, -12.793, -16.035, -15.609, -15.308, -15.057,
    -14.765, -14.400, -13.971, -13.525, -13.121, -12.792, -16.042, -15.615,
    -15.314, -15.063, -14.771, -14.406, -13.976, -13.528, -13.121, -12.791,
    -16.049, -15.622, -15.320, -15.069, -14.778, -14.412, -13.981, -13.530,
    -13.121, -12.790, -16.056, -15.629, -15.327, -15.076, -14.784, -14.418,
    -13.985, -13.532, -13.121, -12.789, -16.064, -15.636, -15.333, -15.083,
    -14.791, -14.424, -13.990, -13.534, -13.121, -12.788, -16.071, -15.643,
    -15.340, -15.089, -14.798, -14.430, -13.994, -13.536, -13.121, -12.787,
    -16.080, -15.650, -15.347, -15.096, -14.804, -14.436, -13.999, -13.538,
    -13.121, -12.785, -16.088, -15.658, -15.354, -15.103, -14.811, -14.442,
    -14.003, -13.540, -13.121, -12.784, -16.097, -15.666, -15.362, -15.110,
    -14.818, -14.448, -14.007, -13.542, -13.121, -12.783, -16.106, -15.675,
    -15.370, -15.118, -14.825, -14.454, -14.011, -13.544, -13.122, -12.782,
    -16.116, -15.684, -15.378, -15.125, -14.832, -14.460, -14.016, -13.546,
    -13.122, -12.782, -16.126, -15.694, -15.386, -15.133, -14.839, -14.467,
    -14.020, -13.548, -13.122, -12.781, -16.137, -15.704, -15.395, -15.141,
    -14.847, -14.473, -14.024, -13.549, -13.123, -12.780, -16.148, -15.714,
    -15.404, -15.149, -14.854, -14.479, -14.028, -13.551, -13.123, -12.779,
    -16.159, -15.724, -15.413, -15.157, -14.861, -14.485, -14.032, -13.553,
    -13.124, -12.779, -16.170, -15.735, -15.422, -15.165, -14.869, -14.492,
    -14.035, -13.555, -13.124, -12.778, -16.182, -15.746, -15.432, -15.173,
    -14.876, -14.498, -14.039, -13.556, -13.125, -12.778, -16.194, -15.758,
    -15.442, -15.182, -14.884, -14.504, -14.043, -13.558, -13.125, -12.777,
    -16.205, -15.769, -15.452, -15.190, -14.891, -14.510, -14.047, -13.559,
    -13.125, -12.776, -16.217, -15.780, -15.461, -15.199, -14.899, -14.516,
    -14.050, -13.561, -13.124, -12.775, -16.228, -15.791, -15.471, -15.208,
    -14.906, -14.522, -14.054, -13.562, -13.124, -12.774, -16.240, -15.802,
    -15.481, -15.216, -14.914, -14.528, -14.058, -13.564, -13.123, -12.772,
    -16.252, -15.814, -15.491, -15.225, -14.922, -14.534, -14.061, -13.565,
    -13.122, -12.771, -16.263, -15.825, -15.501, -15.234, -14.929, -14.539,
    -14.065, -13.566, -13.122, -12.769, -16.275, -15.837, -15.511, -15.243,
    -14.937, -14.545, -14.068, -13.567, -13.121, -12.768, -16.287, -15.849,
    -15.522, -15.252, -14.945, -14.551, -14.071, -13.568, -13.120, -12.766,
    -16.300, -15.861, -15.532, -15.262, -14.952, -14.556, -14.075, -13.569,
    -13.118, -12.765, -16.312, -15.873, -15.543, -15.271, -14.960, -14.562,
    -14.078, -13.570, -13.117, -12.763, -16.324, -15.885, -15.553, -15.280,
    -14.968, -14.568, -14.081, -13.571, -13.115, -12.762, -16.336, -15.896,
    -15.564, -15.289, -14.975, -14.573, -14.084, -13.571, -13.112, -12.761,
    -16.348, -15.908, -15.574, -15.298, -14.983, -14.579, -14.088, -13.572,
    -13.110, -12.760, -16.359, -15.920, -15.585, -15.307, -14.991, -14.585,
    -14.091, -13.573, -13.107, -12.759, -16.371, -15.932, -15.595, -15.317,
    -14.998, -14.591, -14.094, -13.573, -13.104, -12.757, -16.383, -15.943,
    -15.606, -15.326, -15.006, -14.596, -14.097, -13.573, -13.100, -12.757,
    -16.395, -15.955, -15.617, -15.335, -15.014, -14.602, -14.100, -13.574,
    -13.097, -12.756, -16.407, -15.967, -15.627, -15.344, -15.022, -14.608,
    -14.102, -13.574, -13.093, -12.755, -16.419, -15.979, -15.638, -15.354,
    -15.029, -14.613, -14.105, -13.574, -13.088, -12.754, -16.431, -15.991,
    -15.649, -15.363, -15.037, -14.619, -14.108, -13.574, -13.084, -12.753,
    -16.443, -16.003, -15.660, -15.372, -15.045, -14.625, -14.111, -13.575,
    -13.084, -12.753, -16.454, -16.014, -15.670, -15.382, -15.052, -14.630,
    -14.114, -13.575, -13.084, -12.752, -16.466, -16.026, -15.681, -15.391,
    -15.060, -14.636, -14.116, -13.576, -13.086, -12.752, -16.477, -16.038,
    -15.691, -15.400, -15.067, -14.641, -14.119, -13.576, -13.087, -12.752,
    -16.489, -16.050, -15.702, -15.409, -15.075, -14.646, -14.121, -13.577,
    -13.090, -12.752, -16.500, -16.061, -15.713, -15.419, -15.082, -14.652,
    -14.124, -13.577, -13.093, -12.752, -16.511, -16.073, -15.723, -15.428,
    -15.090, -14.657, -14.127, -13.578, -13.096, -12.753, -16.523, -16.085,
    -15.734, -15.437, -15.097, -14.662, -14.129, -13.579, -13.100, -12.753,
    -16.534, -16.096, -15.744, -15.447, -15.105, -14.668, -14.132, -13.579,
    -13.105, -12.753, -16.545, -16.108, -15.755, -15.456, -15.112, -14.673,
    -14.134, -13.580, -13.110, -12.754, -16.556, -16.119, -15.765, -15.465,
    -15.120, -14.678, -14.137, -13.581, -13.111, -12.753, -16.567, -16.131,
    -15.776, -15.474, -15.127, -14.683, -14.139, -13.583, -13.111, -12.752,
    -16.578, -16.142, -15.786, -15.483, -15.134, -14.688, -14.142, -13.585,
    -13.111, -12.750, -16.589, -16.153, -15.797, -15.492, -15.142, -14.693,
    -14.145, -13.587, -13.111, -12.749, -16.600, -16.165, -15.807, -15.501,
    -15.149, -14.697, -14.147, -13.589, -13.110, -12.747, -16.611, -16.176,
    -15.817, -15.510, -15.157, -14.702, -14.150, -13.591, -13.109, -12.745,
    -16.622, -16.187, -15.827, -15.519, -15.165, -14.707, -14.153, -13.593,
    -13.108, -12.742, -16.633, -16.198, -15.838, -15.528, -15.172, -14.711,
    -14.156, -13.596, -13.106, -12.740, -16.643, -16.209, -15.848, -15.537,
    -15.180, -14.716, -14.158, -13.598, -13.104, -12.737, -16.654, -16.220,
    -15.858, -15.546, -15.187, -14.720, -14.161, -13.601, -13.102, -12.734,
    -16.665, -16.231, -15.868, -15.555, -15.194, -14.725, -14.163, -13.602,
    -13.102, -12.733, -16.675, -16.242, -15.878, -15.564, -15.201, -14.730,
    -14.166, -13.603, -13.102, -12.733, -16.685, -16.253, -15.888, -15.573,
    -15.208, -14.735, -14.168, -13.603, -13.103, -12.733, -16.696, -16.263,
    -15.898, -15.582, -15.214, -14.740, -14.170, -13.604, -13.104, -12.734,
    -16.706, -16.274, -15.908, -15.591, -15.221, -14.745, -14.172, -13.604,
    -13.104, -12.734, -16.716, -16.285, -15.918, -15.600, -15.228, -14.750,
    -14.174, -13.604, -13.105, -12.735, -16.726, -16.295, -15.928, -15.608,
    -15.234, -14.755, -14.176, -13.604, -13.107, -12.736, -16.736, -16.306,
    -15.938, -15.617, -15.241, -14.760, -14.178, -13.603, -13.108, -12.737,
    -16.746, -16.317, -15.948, -15.626, -15.247, -14.766, -14.180, -13.603,
    -13.109, -12.739, -16.756, -16.327, -15.958, -15.635, -15.253, -14.771,
    -14.182, -13.602, -13.111, -12.741, -16.766, -16.337, -15.968, -15.644,
    -15.260, -14.776, -14.185, -13.603, -13.112, -12.742, -16.776, -16.348,
    -15.978, -15.652, -15.267, -14.782, -14.189, -13.605, -13.114, -12.743,
    -16.785, -16.358, -15.987, -15.661, -15.275, -14.787, -14.192, -13.607,
    -13.115, -12.743, -16.795, -16.368, -15.997, -15.670, -15.282, -14.792,
    -14.196, -13.609, -13.116, -12.744, -16.804, -16.379, -16.007, -15.678,
    -15.290, -14.798, -14.200, -13.611, -13.118, -12.745, -16.814, -16.389,
    -16.016, -15.687, -15.297, -14.803, -14.204, -13.614, -13.119, -12.746,
    -16.823, -16.399, -16.026, -15.696, -15.305, -14.808, -14.209, -13.617,
    -13.121, -12.747, -16.833, -16.409, -16.035, -15.704, -15.313, -14.814,
    -14.213, -13.620, -13.123, -12.748, -16.842, -16.419, -16.045, -15.713,
    -15.321, -14.819, -14.218, -13.623, -13.124, -12.749, -16.851, -16.429,
    -16.054, -15.721, -15.329, -14.825, -14.223, -13.627, -13.126, -12.750,
    -16.860, -16.439, -16.063, -15.729, -15.336, -14.830, -14.227, -13.629,
    -13.127, -12.751, -16.869, -16.449, -16.073, -15.738, -15.343, -14.836,
    -14.231, -13.632, -13.129, -12.752, -16.879, -16.458, -16.082, -15.746,
    -15.350, -14.841, -14.235, -13.634, -13.131, -12.753, -16.888, -16.468,
    -16.091, -15.754, -15.357, -14.847, -14.239, -13.637, -13.132, -12.753,
    -16.897, -16.477, -16.100, -15.762, -15.364, -14.852, -14.242, -13.639,
    -13.134, -12.754, -16.906, -16.487, -16.109, -15.771, -15.371, -14.858,
    -14.246, -13.642, -13.135, -12.755, -16.915, -16.496, -16.118, -15.779,
    -15.378, -14.863, -14.250, -13.644, -13.137, -12.756, -16.923, -16.506,
    -16.127, -15.787, -15.385, -14.869, -14.254, -13.646, -13.138, -12.756,
    -16.932, -16.515, -16.136, -15.795, -15.391, -14.874, -14.257, -13.649,
    -13.140, -12.757, -16.941, -16.524, -16.145, -15.803, -15.398, -14.880,
    -14.261, -13.651, -13.141, -12.758, -16.950, -16.533, -16.154, -15.811,
    -15.405, -14.885, -14.265, -13.654, -13.143, -12.759, -16.958, -16.542,
    -16.162, -15.819, -15.412, -14.891, -14.269, -13.656, -13.144, -12.760,
    -16.967, -16.552, -16.171, -15.827, -15.419, -14.896, -14.273, -13.659,
    -13.146, -12.761, -16.975, -16.561, -16.180, -15.835, -15.425, -14.902,
    -14.277, -13.662, -13.148, -12.762, -16.984, -16.570, -16.188, -15.843,
    -15.432, -14.907, -14.281, -13.665, -13.150, -12.763, -16.992, -16.579,
    -16.197, -15.850, -15.439, -14.913, -14.285, -13.667, -13.152, -12.764,
    -17.000, -16.588, -16.205, -15.858, -15.446, -14.918, -14.289, -13.670,
    -13.154, -12.765, -17.009, -16.596, -16.213, -15.866, -15.452, -14.923,
    -14.293, -13.673, -13.156, -12.767, -17.017, -16.605, -16.222, -15.873,
    -15.459, -14.929, -14.297, -13.676, -13.158, -12.768, -17.025, -16.614,
    -16.230, -15.881, -15.466, -14.934, -14.301, -13.679, -13.160, -12.769,
    -17.033, -16.623, -16.238, -15.889, -15.473, -14.940, -14.305, -13.682,
    -13.162, -12.770, -17.041, -16.631, -16.246, -15.896, -15.479, -14.945,
    -14.310, -13.686, -13.165, -12.772, -17.049, -16.639, -16.255, -15.903,
    -15.486, -14.951, -14.315, -13.689, -13.167, -12.774, -17.057, -16.648,
    -16.263, -15.911, -15.492, -14.957, -14.319, -13.693, -13.170, -12.775,
    -17.064, -16.656, -16.271, -15.918, -15.499, -14.962, -14.324, -13.696,
    -13.172, -12.777, -17.072, -16.664, -16.279, -15.925, -15.505, -14.968,
    -14.329, -13.700, -13.175, -12.779, -17.080, -16.672, -16.287, -15.933,
    -15.512, -14.974, -14.334, -13.704, -13.177, -12.781, -17.087, -16.680,
    -16.294, -15.940, -15.518, -14.980, -14.339, -13.708, -13.180, -12.783,
    -17.095, -16.688, -16.302, -15.947, -15.525, -14.986, -14.344, -13.712,
    -13.183, -12.785, -17.102, -16.696, -16.310, -15.954, -15.531, -14.992,
    -14.349, -13.716, -13.186, -12.787, -17.109, -16.704, -16.318, -15.961,
    -15.537, -14.998, -14.354, -13.720, -13.189, -12.789, -17.117, -16.711,
    -16.325, -15.968, -15.544, -15.004, -14.359, -13.724, -13.192, -12.791,
    -17.124, -16.719, -16.333, -15.975, -15.550, -15.010, -14.365, -13.729,
    -13.196, -12.794, -17.131, -16.726, -16.340, -15.981, -15.556, -15.016,
    -14.370, -13.733, -13.199, -12.796, -17.138, -16.734, -16.347, -15.988,
    -15.562, -15.022, -14.375, -13.738, -13.202, -12.799, -17.145, -16.741,
    -16.355, -15.994, -15.568, -15.028, -14.381, -13.743, -13.206, -12.801,
    -17.152, -16.748, -16.362, -16.001, -15.574, -15.034, -14.386, -13.747,
    -13.210, -12.804, -17.159, -16.755, -16.369, -16.007, -15.580, -15.040,
    -14.392, -13.752, -13.213, -12.806, -17.165, -16.762, -16.376, -16.014,
    -15.586, -15.046, -14.397, -13.757, -13.217, -12.809, -17.172, -16.769,
    -16.383, -16.020, -15.592, -15.052, -14.403, -13.762, -13.221, -12.812,
    -17.178, -16.776, -16.390, -16.026, -15.598, -15.058, -14.409, -13.767,
    -13.225, -12.815, -17.185, -16.783, -16.397, -16.032, -15.604, -15.064,
    -14.414, -13.772, -13.229, -12.818, -17.191, -16.790, -16.404, -16.038,
    -15.610, -15.070, -14.420, -13.777, -13.233, -12.821, -17.197, -16.797,
    -16.410, -16.044, -15.616, -15.076, -14.426, -13.782, -13.237, -12.824,
    -17.203, -16.804, -16.417, -16.050, -15.621, -15.082, -14.432, -13.787,
    -13.241, -12.827, -17.209, -16.810, -16.423, -16.056, -15.627, -15.088,
    -14.437, -13.792, -13.245, -12.830, -17.215, -16.817, -16.430, -16.061,
    -15.633, -15.093, -14.443, -13.798, -13.250, -12.833, -17.221, -16.824,
    -16.436, -16.067, -15.639, -15.099, -14.449, -13.803, -13.254, -12.836,
    -17.226, -16.830, -16.443, -16.073, -15.644, -15.105, -14.455, -13.809,
    -13.258, -12.840, -17.232, -16.837, -16.449, -16.078, -15.650, -15.111,
    -14.461, -13.814, -13.263, -12.843, -17.237, -16.843, -16.455, -16.083,
    -15.655, -15.117, -14.467, -13.819, -13.267, -12.846, -17.243, -16.849,
    -16.461, -16.089, -15.660, -15.122, -14.473, -13.825, -13.272, -12.850,
    -17.248, -16.855, -16.466, -16.094, -15.665, -15.128, -14.479, -13.830,
    -13.277, -12.853, -17.253, -16.861, -16.472, -16.099, -15.670, -15.133,
    -14.484, -13.836, -13.281, -12.857, -17.258, -16.867, -16.477, -16.104,
    -15.675, -15.139, -14.490, -13.842, -13.286, -12.860, -17.263, -16.873,
    -16.482, -16.108, -15.680, -15.144, -14.496, -13.847, -13.291, -12.864,
    -17.267, -16.879, -16.487, -16.113, -15.685, -15.149, -14.502, -13.853,
    -13.295, -12.867, -17.272, -16.884, -16.492, -16.118, -15.689, -15.155,
    -14.508, -13.858, -13.300, -12.871, -17.277, -16.890, -16.497, -16.123,
    -15.694, -15.160, -14.514, -13.864, -13.305, -12.874, -17.281, -16.895,
    -16.502, -16.127, -15.698, -15.165, -14.520, -13.870, -13.310, -12.878,
    -17.285, -16.900, -16.507, -16.131, -15.702, -15.170, -14.526, -13.876,
    -13.315, -12.882, -17.290, -16.905, -16.512, -16.135, -15.707, -15.176,
    -14.532, -13.882, -13.320, -12.885, -17.294, -16.910, -16.516, -16.139,
    -15.711, -15.181, -14.538, -13.887, -13.325, -12.889, -17.298, -16.914,
    -16.521, -16.143, -15.716, -15.186, -14.544, -13.893, -13.330, -12.893,
    -17.302, -16.919, -16.525, -16.146, -15.720, -15.191, -14.550, -13.899,
    -13.335, -12.897, -17.306, -16.923, -16.529, -16.150, -15.724, -15.196,
    -14.556, -13.905, -13.340, -12.901, -17.310, -16.927, -16.534, -16.153,
    -15.728, -15.202, -14.562, -13.911, -13.345, -12.905, -17.314, -16.931,
    -16.538, -16.156, -15.732, -15.207, -14.568, -13.917, -13.350, -12.909,
    -17.317, -16.935, -16.542, -16.159, -15.736, -15.212, -14.574, -13.923,
    -13.355, -12.913, -17.321, -16.939, -16.546, -16.162, -15.740, -15.217,
    -14.580, -13.929, -13.360, -12.917, -17.324, -16.942, -16.549, -16.165,
    -15.743, -15.222, -14.586, -13.935, -13.365, -12.921, -17.326, -16.946,
    -16.552, -16.167, -15.747, -15.226, -14.592, -13.941, -13.370, -12.925,
    -17.329, -16.948, -16.555, -16.169, -15.750, -15.231, -14.598, -13.946,
    -13.375, -12.929, -17.331, -16.951, -16.557, -16.172, -15.753, -15.235,
    -14.603, -13.952, -13.380, -12.933, -17.333, -16.954, -16.559, -16.174,
    -15.756, -15.239, -14.609, -13.958, -13.385, -12.938, -17.335, -16.956,
    -16.561, -16.175, -15.759, -15.243, -14.615, -13.963, -13.390, -12.942,
    -17.337, -16.958, -16.563, -16.177, -15.762, -15.247, -14.621, -13.969,
    -13.396, -12.946, -17.338, -16.961, -16.564, -16.178, -15.764, -15.251,
    -14.627, -13.975, -13.401, -12.950, -17.339, -16.962, -16.565, -16.180,
    -15.767, -15.255, -14.632, -13.980, -13.406, -12.955, -17.340, -16.964,
    -16.566, -16.181, -15.769, -15.259, -14.638, -13.986, -13.411, -12.959,
    -17.341, -16.965, -16.568, -16.182, -15.771, -15.263, -14.644, -13.992,
    -13.416, -12.963, -17.341, -16.966, -16.569, -16.182, -15.773, -15.267,
    -14.649, -13.998, -13.421, -12.967, -17.341, -16.967, -16.571, -16.182,
    -15.775, -15.271, -14.655, -14.004, -13.427, -12.972, -17.341, -16.968,
    -16.572, -16.182, -15.777, -15.275, -14.660, -14.010, -13.432, -12.976,
    -17.341, -16.969, -16.573, -16.182, -15.779, -15.278, -14.665, -14.016,
    -13.437, -12.980, -17.340, -16.969, -16.575, -16.182, -15.780, -15.282,
    -14.671, -14.023, -13.443, -12.984, -17.340, -16.969, -16.576, -16.181,
    -15.782, -15.286, -14.676, -14.029, -13.448, -12.988, -17.339, -16.969,
    -16.577, -16.180, -15.783, -15.290, -14.682, -14.035, -13.453, -12.993,
    -17.337, -16.969, -16.578, -16.179, -15.784, -15.293, -14.687, -14.042,
    -13.459, -12.997, -17.336, -16.968, -16.579, -16.178, -15.785, -15.297,
    -14.692, -14.048, -13.464, -13.001, -17.334, -16.968, -16.578, -16.176,
    -15.786, -15.300, -14.697, -14.054, -13.469, -13.005, -17.332, -16.967,
    -16.577, -16.174, -15.787, -15.304, -14.703, -14.060, -13.475, -13.010,
    -17.330, -16.966, -16.575, -16.172, -15.788, -15.307, -14.708, -14.066,
    -13.480, -13.014, -17.327, -16.965, -16.572, -16.170, -15.788, -15.311,
    -14.713, -14.072, -13.485, -13.018, -17.324, -16.964, -16.570, -16.167,
    -15.789, -15.314, -14.718, -14.078, -13.491, -13.023, -17.321, -16.963,
    -16.567, -16.164, -15.789, -15.317, -14.723, -14.084, -13.496, -13.027,
    -17.317, -16.962, -16.563, -16.161, -15.790, -15.320, -14.729, -14.090,
    -13.501, -13.032, -17.313, -16.960, -16.559, -16.158, -15.790, -15.323,
    -14.734, -14.095, -13.506, -13.036, -17.309, -16.959, -16.555, -16.154,
    -15.790, -15.326, -14.739, -14.101, -13.512, -13.041, -17.305, -16.957,
    -16.550, -16.150, -15.790, -15.329, -14.744, -14.107, -13.517, -13.045,
    -17.300, -16.955, -16.548, -16.146, -15.788, -15.331, -14.749, -14.113,
    -13.522, -13.049, -17.295, -16.952, -16.545, -16.142, -15.786, -15.333,
    -14.754, -14.119, -13.528, -13.054, -17.289, -16.949, -16.543, -16.138,
    -15.783, -15.334, -14.759, -14.125, -13.533, -13.058, -17.283, -16.946,
    -16.541, -16.133, -15.780, -15.335, -14.764, -14.131, -13.539, -13.063,
    -17.277, -16.943, -16.539, -16.128, -15.776, -15.336, -14.769, -14.137,
    -13.544, -13.067, -17.271, -16.939, -16.537, -16.123, -15.772, -15.337,
    -14.774, -14.143, -13.550, -13.072, -17.264, -16.935, -16.535, -16.118,
    -15.768, -15.338, -14.779, -14.149, -13.555, -13.076, -17.256, -16.931,
    -16.533, -16.112, -15.763, -15.338, -14.784, -14.155, -13.561, -13.081,
    -17.249, -16.927, -16.532, -16.107, -15.758, -15.338, -14.789, -14.161,
    -13.566, -13.085, -17.241, -16.922, -16.530, -16.101, -15.752, -15.338,
    -14.794, -14.167, -13.572, -13.090, -17.232, -16.916, -16.526, -16.098,
    -15.752, -15.341, -14.799, -14.173, -13.577, -13.094, -17.224, -16.909,
    -16.522, -16.095, -15.753, -15.343, -14.804, -14.179, -13.583, -13.099,
    -17.214, -16.903, -16.517, -16.092, -15.755, -15.347, -14.809, -14.185,
    -13.588, -13.103, -17.204, -16.895, -16.512, -16.089, -15.757, -15.350,
    -14.814, -14.191, -13.594, -13.108, -17.194, -16.887, -16.506, -16.087,
    -15.760, -15.354, -14.819, -14.197, -13.599, -13.112, -17.184, -16.879,
    -16.500, -16.085, -15.764, -15.358, -14.824, -14.203, -13.605, -13.117,
    -17.173, -16.870, -16.494, -16.083, -15.768, -15.362, -14.828, -14.210,
    -13.610, -13.121, -17.162, -16.861, -16.487, -16.082, -15.773, -15.367,
    -14.833, -14.216, -13.615, -13.125, -17.150, -16.851, -16.480, -16.081,
    -15.779, -15.372, -14.838, -14.222, -13.621, -13.130, -17.138, -16.841,
    -16.473, -16.080, -15.786, -15.377, -14.843, -14.228, -13.626, -13.134,
    -17.128, -16.834, -16.469, -16.076, -15.786, -15.380, -14.848, -14.234,
    -13.631, -13.139, -17.118, -16.828, -16.465, -16.073, -15.785, -15.383,
    -14.853, -14.240, -13.637, -13.143, -17.109, -16.822, -16.462, -16.068,
    -15.784, -15.386, -14.857, -14.246, -13.642, -13.148, -17.099, -16.816,
    -16.458, -16.064, -15.783, -15.388, -14.862, -14.252, -13.648, -13.152,
    -17.090, -16.811, -16.455, -16.059, -15.780, -15.391, -14.867, -14.258,
    -13.653, -13.157, -17.080, -16.805, -16.453, -16.054, -15.777, -15.393,
    -14.871, -14.264, -13.659, -13.161, -17.071, -16.800, -16.450, -16.049,
    -15.774, -15.395, -14.876, -14.270, -13.664, -13.166, -17.061, -16.796,
    -16.448, -16.044, -15.769, -15.397, -14.881, -14.275, -13.670, -13.171,
    -17.052, -16.791, -16.446, -16.038, -15.764, -15.399, -14.885, -14.281,
    -13.675, -13.175, -17.043, -16.787, -16.445, -16.032, -15.759, -15.401,
    -14.890, -14.287, -13.681, -13.180, -17.037, -16.784, -16.443, -16.034,
    -15.760, -15.404, -14.895, -14.293, -13.686, -13.184, -17.032, -16.782,
    -16.442, -16.037, -15.761, -15.407, -14.900, -14.299, -13.692, -13.189,
    -17.028, -16.781, -16.441, -16.041, -15.763, -15.411, -14.906, -14.305,
    -13.697, -13.193, -17.024, -16.780, -16.439, -16.046, -15.765, -15.415,
    -14.911, -14.311, -13.702, -13.198, -17.021, -16.779, -16.438, -16.052,
    -15.768, -15.418, -14.917, -14.317, -13.708, -13.202, -17.018, -16.779,
    -16.438, -16.060, -15.772, -15.422, -14.923, -14.323, -13.713, -13.207,
    -17.016, -16.780, -16.437, -16.069, -15.776, -15.426, -14.928, -14.329,
    -13.718, -13.211, -17.015, -16.781, -16.437, -16.079, -15.781, -15.430,
    -14.934, -14.336, -13.724, -13.215, -17.015, -16.782, -16.437, -16.090,
    -15.787, -15.435, -14.940, -14.342, -13.729, -13.220, -17.015, -16.784,
    -16.437, -16.102, -15.793, -15.439, -14.946, -14.348, -13.734, -13.224,
    -17.013, -16.784, -16.438, -16.107, -15.796, -15.443, -14.951, -14.354,
    -13.739, -13.228, -17.010, -16.784, -16.439, -16.111, -15.800, -15.447,
    -14.957, -14.360, -13.745, -13.233, -17.008, -16.785, -16.440, -16.114,
    -15.803, -15.451, -14.962, -14.366, -13.750, -13.237, -17.006, -16.785,
    -16.441, -16.117, -15.806, -15.455, -14.968, -14.372, -13.755, -13.242,
    -17.004, -16.786, -16.443, -16.120, -15.809, -15.459, -14.973, -14.378,
    -13.761, -13.246, -17.002, -16.787, -16.445, -16.122, -15.812, -15.464,
    -14.979, -14.385, -13.766, -13.250, -17.000, -16.787, -16.447, -16.124,
    -15.816, -15.468, -14.984, -14.391, -13.772, -13.255, -16.998, -16.788,
    -16.450, -16.126, -15.819, -15.472, -14.989, -14.397, -13.777, -13.259,
    -16.997, -16.789, -16.453, -16.127, -15.822, -15.477, -14.995, -14.403,
    -13.783, -13.264, -16.995, -16.790, -16.456, -16.127, -15.825, -15.481,
    -15.000, -14.409, -13.788, -13.268, -16.994, -16.791, -16.459, -16.130,
    -15.829, -15.485, -15.005, -14.415, -13.793, -13.272, -16.993, -16.793,
    -16.461, -16.133, -15.832, -15.490, -15.011, -14.421, -13.799, -13.277,
    -16.993, -16.794, -16.464, -16.136, -15.836, -15.494, -15.016, -14.427,
    -13.804, -13.281, -16.993, -16.796, -16.467, -16.140, -15.839, -15.499,
    -15.022, -14.433, -13.809, -13.285, -16.993, -16.798, -16.470, -16.143,
    -15.843, -15.503, -15.027, -14.440, -13.814, -13.289, -16.993, -16.800,
    -16.473, -16.147, -15.847, -15.508, -15.033, -14.446, -13.819, -13.293,
    -16.993, -16.802, -16.476, -16.150, -15.851, -15.513, -15.038, -14.452,
    -13.825, -13.298, -16.993, -16.804, -16.480, -16.154, -15.855, -15.517,
    -15.044, -14.458, -13.830, -13.302, -16.994, -16.807, -16.483, -16.158,
    -15.859, -15.522, -15.049, -14.464, -13.835, -13.306, -16.995, -16.809,
    -16.487, -16.162, -15.863, -15.527, -15.055, -14.470, -13.840, -13.310,
    -16.995, -16.812, -16.491, -16.167, -15.868, -15.532, -15.061, -14.476,
    -13.845, -13.314, -16.996, -16.814, -16.496, -16.172, -15.873, -15.537,
    -15.067, -14.482, -13.850, -13.319, -16.996, -16.817, -16.501, -16.177,
    -15.878, -15.543, -15.072, -14.489, -13.856, -13.323, -16.997, -16.821,
    -16.506, -16.183, -15.884, -15.548, -15.078, -14.495, -13.861, -13.327,
    -16.997, -16.824, -16.511, -16.188, -15.889, -15.554, -15.084, -14.501,
    -13.866, -13.332, -16.998, -16.827, -16.516, -16.195, -15.895, -15.559,
    -15.090, -14.508, -13.871, -13.336, -16.999, -16.831, -16.522, -16.201,
    -15.901, -15.565, -15.096, -14.514, -13.876, -13.341, -16.999, -16.834,
    -16.528, -16.207, -15.907, -15.571, -15.103, -14.520, -13.882, -13.345,
    -17.000, -16.838, -16.534, -16.214, -15.913, -15.577, -15.109, -14.527,
    -13.887, -13.350, -17.001, -16.842, -16.540, -16.221, -15.920, -15.583,
    -15.115, -14.533, -13.892, -13.354, -17.002, -16.846, -16.546, -16.228,
    -15.926, -15.589, -15.121, -14.539, -13.897, -13.358, -17.003, -16.850,
    -16.552, -16.235, -15.933, -15.595, -15.128, -14.546, -13.902, -13.363,
    -17.004, -16.854, -16.559, -16.242, -15.940, -15.602, -15.134, -14.552,
    -13.908, -13.367, -17.005, -16.858, -16.565, -16.249, -15.947, -15.608,
    -15.140, -14.558, -13.913, -13.372, -17.006, -16.862, -16.572, -16.256,
    -15.954, -15.615, -15.147, -14.565, -13.918, -13.376, -17.007, -16.866,
    -16.579, -16.264, -15.961, -15.621, -15.153, -14.571, -13.923, -13.381,
    -17.008, -16.870, -16.585, -16.271, -15.968, -15.628, -15.160, -14.578,
    -13.928, -13.385, -17.009, -16.875, -16.592, -16.279, -15.976, -15.635,
    -15.167, -14.584, -13.934, -13.390, -17.011, -16.879, -16.600, -16.287,
    -15.983, -15.642, -15.173, -14.591, -13.939, -13.394, -17.012, -16.884,
    -16.607, -16.295, -15.991, -15.649, -15.180, -14.597, -13.944, -13.399,
    -17.014, -16.889, -16.614, -16.303, -15.998, -15.656, -15.187, -14.603,
    -13.949, -13.403, -17.016, -16.893, -16.621, -16.311, -16.006, -15.663,
    -15.193, -14.610, -13.954, -13.407, -17.018, -16.898, -16.628, -16.319,
    -16.014, -15.670, -15.200, -14.616, -13.959, -13.412, -17.020, -16.903,
    -16.635, -16.327, -16.021, -15.677, -15.207, -14.623, -13.964, -13.416,
    -17.022, -16.908, -16.643, -16.335, -16.029, -15.684, -15.213, -14.629,
    -13.969, -13.420, -17.025, -16.913, -16.650, -16.343, -16.037, -15.691,
    -15.220, -14.635, -13.974, -13.424, -17.027, -16.918, -16.657, -16.351,
    -16.045, -15.698, -15.227, -14.642, -13.979, -13.428, -17.030, -16.923,
    -16.665, -16.359, -16.053, -15.705, -15.233, -14.648, -13.983, -13.431,
    -17.033, -16.928, -16.672, -16.368, -16.061, -15.713, -15.240, -14.655,
    -13.988, -13.435, -17.036, -16.933, -16.680, -16.376, -16.069, -15.720,
    -15.247, -14.661, -13.993, -13.439, -17.039, -16.938, -16.687, -16.384,
    -16.077, -15.727, -15.254, -14.667, -13.998, -13.443, -17.042, -16.943,
    -16.695, -16.393, -16.085, -15.734, -15.261, -14.674, -14.004, -13.448,
    -17.045, -16.948, -16.702, -16.401, -16.093, -15.742, -15.267, -14.680,
    -14.009, -13.452, -17.048, -16.953, -16.709, -16.409, -16.101, -15.749,
    -15.274, -14.687, -14.014, -13.457, -17.051, -16.959, -16.717, -16.418,
    -16.109, -15.756, -15.281, -14.693, -14.020, -13.461, -17.055, -16.964,
    -16.724, -16.426, -16.117, -15.764, -15.288, -14.700, -14.025, -13.466,
    -17.058, -16.969, -16.732, -16.434, -16.125, -15.771, -15.294, -14.706,
    -14.031, -13.471, -17.062, -16.974, -16.739, -16.443, -16.134, -15.778,
    -15.301, -14.713, -14.037, -13.475, -17.065, -16.980, -16.747, -16.451,
    -16.142, -15.786, -15.308, -14.719, -14.042, -13.480, -17.069, -16.985,
    -16.754, -16.460, -16.150, -15.793, -15.315, -14.726, -14.048, -13.485,
    -17.073, -16.990, -16.761, -16.468, -16.158, -15.800, -15.322, -14.732,
    -14.053, -13.490, -17.077, -16.996, -16.769, -16.477, -16.166, -15.808,
    -15.329, -14.739, -14.059, -13.494, -17.080, -17.001, -16.776, -16.485,
    -16.174, -15.815, -15.336, -14.746, -14.064, -13.499, -17.084, -17.006,
    -16.784, -16.494, -16.183, -15.823, -15.342, -14.752, -14.070, -13.503,
    -17.089, -17.012, -16.791, -16.502, -16.191, -15.830, -15.349, -14.758,
    -14.075, -13.508, -17.093, -17.017, -16.799, -16.510, -16.199, -15.838,
    -15.356, -14.765, -14.081, -13.513, -17.097, -17.023, -16.806, -16.519,
    -16.207, -15.845, -15.363, -14.771, -14.086, -13.517, -17.101, -17.028,
    -16.814, -16.527, -16.216, -15.853, -15.370, -14.778, -14.091, -13.522,
    -17.106, -17.034, -16.821, -16.536, -16.224, -15.860, -15.377, -14.785,
    -14.097, -13.526, -17.110, -17.039, -16.829, -16.544, -16.232, -15.868,
    -15.384, -14.791, -14.102, -13.531, -17.114, -17.045, -16.836, -16.552,
    -16.240, -15.876, -15.391, -14.798, -14.107, -13.535, -17.119, -17.050,
    -16.844, -16.561, -16.248, -15.883, -15.398, -14.804, -14.113, -13.540,
    -17.123, -17.056, -16.851, -16.569, -16.257, -15.891, -15.405, -14.811,
    -14.118, -13.544, -17.128, -17.061, -16.858, -16.577, -16.265, -15.898,
    -15.411, -14.817, -14.123, -13.549, -17.133, -17.067, -16.866, -16.586,
    -16.273, -15.906, -15.418, -14.823, -14.129, -13.553, -17.137, -17.073,
    -16.873, -16.594, -16.281, -15.913, -15.425, -14.830, -14.134, -13.558,
    -17.142, -17.079, -16.880, -16.602, -16.289, -15.921, -15.432, -14.837,
    -14.139, -13.562, -17.147, -17.084, -16.888, -16.610, -16.298, -15.928,
    -15.438, -14.843, -14.144, -13.566, -17.152, -17.090, -16.895, -16.619,
    -16.306, -15.935, -15.445, -14.849, -14.150, -13.571, -17.157, -17.096,
    -16.902, -16.627, -16.314, -15.943, -15.452, -14.856, -14.155, -13.575,
    -17.162, -17.102, -16.909, -16.635, -16.322, -15.950, -15.459, -14.862,
    -14.160, -13.580, -17.167, -17.108, -16.917, -16.643, -16.330, -15.958,
    -15.466, -14.869, -14.166, -13.584, -17.172, -17.113, -16.924, -16.652,
    -16.338, -15.965, -15.472, -14.875, -14.171, -13.589, -17.177, -17.119,
    -16.931, -16.660, -16.346, -15.973, -15.479, -14.881, -14.176, -13.593,
    -17.182, -17.125, -16.938, -16.668, -16.354, -15.980, -15.486, -14.888,
    -14.182, -13.598, -17.188, -17.131, -16.946, -16.676, -16.362, -15.988,
    -15.493, -14.894, -14.187, -13.602, -17.193, -17.136, -16.953, -16.684,
    -16.370, -15.995, -15.500, -14.900, -14.192, -13.607, -17.198, -17.142,
    -16.960, -16.692, -16.378, -16.002, -15.506, -14.907, -14.197, -13.612,
    -17.204, -17.148, -16.968, -16.700, -16.386, -16.010, -15.513, -14.913,
    -14.203, -13.616, -17.209, -17.154, -16.975, -16.708, -16.394, -16.017,
    -15.520, -14.919, -14.208, -13.621, -17.214, -17.160, -16.982, -16.716,
    -16.402, -16.024, -15.527, -14.925, -14.213, -13.626, -17.220, -17.166,
    -16.989, -16.724, -16.410, -16.032, -15.533, -14.932, -14.219, -13.630,
    -17.225, -17.172, -16.997, -16.732, -16.418, -16.039, -15.540, -14.938,
    -14.224, -13.635, -17.231, -17.178, -17.004, -16.740, -16.426, -16.046,
    -15.547, -14.944, -14.229, -13.639, -17.236, -17.184, -17.011, -16.748,
    -16.434, -16.054, -15.554, -14.950, -14.235, -13.644, -17.242, -17.190,
    -17.018, -16.756, -16.442, -16.061, -15.560, -14.957, -14.240, -13.648,
    -17.247, -17.196, -17.025, -16.764, -16.449, -16.068, -15.567, -14.963,
    -14.246, -13.653, -17.253, -17.202, -17.032, -16.771, -16.457, -16.075,
    -15.574, -14.969, -14.251, -13.657, -17.258, -17.208, -17.039, -16.779,
    -16.465, -16.083, -15.580, -14.976, -14.257, -13.662, -17.264, -17.214,
    -17.046, -16.787, -16.473, -16.090, -15.587, -14.982, -14.262, -13.666,
    -17.270, -17.220, -17.053, -16.795, -16.481, -16.097, -15.594, -14.988,
    -14.267, -13.671, -17.275, -17.226, -17.060, -16.803, -16.488, -16.104,
    -15.600, -14.994, -14.273, -13.675, -17.281, -17.232, -17.067, -16.810,
    -16.496, -16.112, -15.607, -15.001, -14.278, -13.680, -17.287, -17.238,
    -17.074, -16.818, -16.504, -16.119, -15.614, -15.007, -14.284, -13.685,
    -17.293, -17.244, -17.081, -16.826, -16.511, -16.126, -15.620, -15.013,
    -14.289, -13.689, -17.298, -17.250, -17.088, -16.833, -16.519, -16.133,
    -15.627, -15.019, -14.294, -13.694, -17.304, -17.257, -17.095, -16.841,
    -16.527, -16.140, -15.633, -15.026, -14.300, -13.699, -17.310, -17.263,
    -17.102, -16.849, -16.534, -16.147, -15.640, -15.032, -14.305, -13.703,
    -17.316, -17.269, -17.109, -16.856, -16.542, -16.154, -15.646, -15.038,
    -14.311, -13.708, -17.322, -17.275, -17.116, -16.864, -16.549, -16.161,
    -15.653, -15.044, -14.316, -13.713, -17.328, -17.281, -17.123, -16.872,
    -16.557, -16.168, -15.659, -15.050, -14.321, -13.718, -17.334, -17.287,
    -17.130, -16.879, -16.564, -16.175, -15.666, -15.056, -14.327, -13.723,
    -17.340, -17.293, -17.137, -16.887, -16.572, -16.182, -15.672, -15.062,
    -14.332, -13.727, -17.346, -17.300, -17.144, -16.894, -16.579, -16.189,
    -15.679, -15.068, -14.338, -13.732, -17.352, -17.306, -17.151, -16.902,
    -16.587, -16.196, -15.685, -15.074, -14.343, -13.737, -17.358, -17.312,
    -17.158, -16.909, -16.594, -16.203, -15.692, -15.080, -14.349, -13.742,
    -17.364, -17.318, -17.164, -16.917, -16.602, -16.210, -15.698, -15.086,
    -14.354, -13.747, -17.371, -17.325, -17.171, -16.924, -16.609, -16.216,
    -15.704, -15.092, -14.360, -13.752, -17.377, -17.331, -17.178, -16.932,
    -16.617, -16.223, -15.711, -15.098, -14.365, -13.757, -17.383, -17.337,
    -17.185, -16.939, -16.624, -16.230, -15.717, -15.104, -14.371, -13.762,
    -17.389, -17.343, -17.192, -16.946, -16.631, -16.237, -15.723, -15.110,
    -14.376, -13.767, -17.395, -17.349, -17.199, -16.954, -16.639, -16.244,
    -15.730, -15.116, -14.382, -13.772, -17.401, -17.356, -17.206, -16.961,
    -16.646, -16.250, -15.736, -15.122, -14.387, -13.777, -17.408, -17.362,
    -17.212, -16.968, -16.653, -16.257, -15.742, -15.128, -14.393, -13.782,
    -17.414, -17.368, -17.219, -16.976, -16.661, -16.264, -15.748, -15.133,
    -14.398, -13.787, -17.420, -17.374, -17.226, -16.983, -16.668, -16.270,
    -15.754, -15.139, -14.404, -13.791, -17.426, -17.381, -17.233, -16.990,
    -16.675, -16.277, -15.760, -15.145, -14.409, -13.796, -17.433, -17.387,
    -17.240, -16.998, -16.682, -16.283, -15.767, -15.151, -14.415, -13.801,
    -17.439, -17.393, -17.246, -17.005, -16.689, -16.290, -15.773, -15.157,
    -14.420, -13.806, -17.445, -17.399, -17.253, -17.012, -16.696, -16.297,
    -15.779, -15.163, -14.426, -13.811, -17.451, -17.406, -17.260, -17.019,
    -16.703, -16.303, -15.785, -15.168, -14.431, -13.816, -17.458, -17.412,
    -17.267, -17.026, -16.711, -16.310, -15.791, -15.174, -14.437, -13.821,
    -17.464, -17.418, -17.274, -17.034, -16.718, -16.316, -15.797, -15.180,
    -14.442, -13.826, -17.470, -17.425, -17.280, -17.041, -16.725, -16.323,
    -15.803, -15.186, -14.448, -13.831, -17.476, -17.431, -17.287, -17.048,
    -16.732, -16.329, -15.809, -15.191, -14.453, -13.836, -17.483, -17.437,
    -17.294, -17.055, -16.739, -16.335, -15.815, -15.197, -14.459, -13.841,
    -17.489, -17.443, -17.300, -17.062, -16.745, -16.342, -15.821, -15.203,
    -14.464, -13.846, -17.495, -17.450, -17.307, -17.069, -16.752, -16.348,
    -15.826, -15.209, -14.470, -13.850, -17.502, -17.456, -17.314, -17.076,
    -16.759, -16.354, -15.832, -15.214, -14.475, -13.855, -17.508, -17.462,
    -17.321, -17.083, -16.766, -16.361, -15.838, -15.220, -14.481, -13.860,
    -17.514, -17.469, -17.327, -17.090, -16.773, -16.367, -15.844, -15.226,
    -14.486, -13.865, -17.521, -17.475, -17.334, -17.097, -16.780, -16.373,
    -15.850, -15.231, -14.492, -13.870, -17.527, -17.481, -17.341, -17.104,
    -16.787, -16.380, -15.856, -15.237, -14.497, -13.875, -17.534, -17.488,
    -17.347, -17.111, -16.793, -16.386, -15.861, -15.242, -14.503, -13.880,
    -17.540, -17.494, -17.354, -17.118, -16.800, -16.392, -15.867, -15.248,
    -14.508, -13.885, -17.546, -17.500, -17.361, -17.125, -16.807, -16.398,
    -15.873, -15.254, -14.513, -13.890, -17.553, -17.507, -17.367, -17.132,
    -16.813, -16.404, -15.878, -15.259, -14.519, -13.895, -17.559, -17.513,
    -17.374, -17.139, -16.820, -16.410, -15.884, -15.264, -14.524, -13.900,
    -17.566, -17.519, -17.380, -17.145, -16.826, -16.416, -15.889, -15.270,
    -14.530, -13.905, -17.572, -17.526, -17.387, -17.152, -16.833, -16.422,
    -15.895, -15.275, -14.535, -13.910, -17.579, -17.532, -17.394, -17.159,
    -16.839, -16.428, -15.900, -15.281, -14.540, -13.915, -17.585, -17.538,
    -17.400, -17.166, -16.846, -16.434, -15.906, -15.286, -14.546, -13.920,
    -17.592, -17.545, -17.407, -17.173, -16.852, -16.440, -15.911, -15.292,
    -14.551, -13.925, -17.598, -17.551, -17.413, -17.179, -16.859, -16.445,
    -15.917, -15.297, -14.557, -13.930, -17.604, -17.557, -17.420, -17.186,
    -16.865, -16.451, -15.922, -15.302, -14.562, -13.934, -17.611, -17.564,
    -17.426, -17.193, -16.871, -16.457, -15.927, -15.308, -14.567, -13.939,
    -17.617, -17.570, -17.433, -17.199, -16.878, -16.463, -15.933, -15.313,
    -14.573, -13.944, -17.624, -17.576, -17.440, -17.206, -16.884, -16.468,
    -15.938, -15.318, -14.578, -13.949, -17.631, -17.583, -17.446, -17.213,
    -16.890, -16.474, -15.943, -15.323, -14.583, -13.954, -17.637, -17.589,
    -17.453, -17.219, -16.896, -16.480, -15.948, -15.329, -14.589, -13.959,
    -17.644, -17.596, -17.459, -17.226, -16.903, -16.485, -15.954, -15.334,
    -14.594, -13.964, -17.650, -17.602, -17.465, -17.232, -16.909, -16.491,
    -15.959, -15.339, -14.599, -13.969, -17.657, -17.608, -17.472, -17.239,
    -16.915, -16.496, -15.964, -15.344, -14.605, -13.974, -17.663, -17.615,
    -17.478, -17.245, -16.921, -16.502, -15.969, -15.349, -14.610, -13.979,
    -17.670, -17.621, -17.485, -17.252, -16.927, -16.507, -15.974, -15.355,
    -14.615, -13.984, -17.677, -17.627, -17.491, -17.258, -16.933, -16.513,
    -15.979, -15.360, -14.621, -13.989, -17.683, -17.634, -17.498, -17.265,
    -16.939, -16.518, -15.984, -15.365, -14.626, -13.994, -17.690, -17.640,
    -17.504, -17.271, -16.945, -16.523, -15.989, -15.370, -14.631, -13.999,
    -17.696, -17.647, -17.511, -17.278, -16.951, -16.529, -15.994, -15.375,
    -14.637, -14.004, -17.703, -17.653, -17.517, -17.284, -16.957, -16.534,
    -15.999, -15.380, -14.642, -14.009, -17.710, -17.659, -17.523, -17.290,
    -16.963, -16.539, -16.004, -15.385, -14.647, -14.014, -17.716, -17.666,
    -17.530, -17.297, -16.969, -16.544, -16.009, -15.390, -14.653, -14.019,
    -17.723, -17.672, -17.536, -17.303, -16.974, -16.549, -16.014, -15.395,
    -14.658, -14.024, -17.729, -17.678, -17.543, -17.309, -16.980, -16.555,
    -16.019, -15.399, -14.663, -14.029, -17.736, -17.685, -17.549, -17.316,
    -16.986, -16.560, -16.023, -15.404, -14.668, -14.034, -17.742, -17.691,
    -17.555, -17.322, -16.991, -16.565, -16.028, -15.409, -14.673, -14.039,
    -17.749, -17.698, -17.562, -17.328, -16.997, -16.570, -16.033, -15.414,
    -14.679, -14.044, -17.755, -17.704, -17.568, -17.334, -17.003, -16.575,
    -16.038, -15.418, -14.684, -14.049, -17.762, -17.710, -17.574, -17.340,
    -17.008, -16.579, -16.043, -15.423, -14.689, -14.054, -17.768, -17.717,
    -17.581, -17.346, -17.014, -16.584, -16.047, -15.428, -14.694, -14.059,
    -17.775, -17.723, -17.587, -17.353, -17.019, -16.589, -16.052, -15.432,
    -14.699, -14.064, -17.782, -17.729, -17.593, -17.359, -17.025, -16.594,
    -16.057, -15.437, -14.705, -14.069, -17.788, -17.736, -17.599, -17.365,
    -17.030, -16.599, -16.061, -15.442, -14.710, -14.073, -17.795, -17.742,
    -17.606, -17.371, -17.035, -16.603, -16.066, -15.446, -14.715, -14.078,
    -17.801, -17.748, -17.612, -17.377, -17.041, -16.608, -16.071, -15.451,
    -14.720, -14.083, -17.808, -17.755, -17.618, -17.383, -17.046, -16.613,
    -16.075, -15.455, -14.725, -14.088, -17.814, -17.761, -17.624, -17.389,
    -17.051, -16.617, -16.080, -15.459, -14.730, -14.093, -17.821, -17.767,
    -17.631, -17.395, -17.056, -16.622, -16.084, -15.464, -14.735, -14.098,
    -17.828, -17.774, -17.637, -17.401, -17.062, -16.627, -16.089, -15.468,
    -14.740, -14.103, -17.834, -17.780, -17.643, -17.407, -17.067, -16.631,
    -16.093, -15.473, -14.746, -14.108, -17.841, -17.787, -17.649, -17.413,
    -17.072, -16.636, -16.098, -15.477, -14.751, -14.113, -17.847, -17.793,
    -17.655, -17.419, -17.077, -16.640, -16.102, -15.481, -14.756, -14.118,
    -17.854, -17.799, -17.662, -17.424, -17.082, -16.644, -16.106, -15.486,
    -14.761, -14.123, -17.860, -17.806, -17.668, -17.430, -17.087, -16.649,
    -16.111, -15.490, -14.766, -14.128, -17.867, -17.812, -17.674, -17.436,
    -17.092, -16.653, -16.115, -15.494, -14.771, -14.133, -17.874, -17.818,
    -17.680, -17.442, -17.097, -16.658, -16.119, -15.499, -14.776, -14.138,
    -17.880, -17.825, -17.686, -17.448, -17.102, -16.662, -16.124, -15.504,
    -14.781, -14.143, -17.887, -17.831, -17.693, -17.453, -17.107, -16.667,
    -16.128, -15.508, -14.786, -14.148, -17.893, -17.837, -17.699, -17.459,
    -17.112, -16.671, -16.133, -15.513, -14.791, -14.153, -17.900, -17.844,
    -17.705, -17.465, -17.117, -16.675, -16.137, -15.518, -14.797, -14.158,
    -17.907, -17.850, -17.711, -17.471, -17.122, -16.680, -16.142, -15.523,
    -14.802, -14.162, -17.913, -17.856, -17.717, -17.476, -17.127, -16.684,
    -16.146, -15.528, -14.807, -14.167, -17.920, -17.863, -17.723, -17.482,
    -17.132, -16.689, -16.151, -15.533, -14.812, -14.172, -17.926, -17.869,
    -17.729, -17.488, -17.137, -16.693, -16.155, -15.537, -14.817, -14.177,
    -17.933, -17.875, -17.735, -17.493, -17.142, -16.698, -16.159, -15.542,
    -14.822, -14.182, -17.939, -17.882, -17.741, -17.499, -17.147, -16.702,
    -16.164, -15.547, -14.827, -14.187, -17.946, -17.888, -17.748, -17.504,
    -17.151, -16.706, -16.168, -15.552, -14.832, -14.192, -17.953, -17.894,
    -17.754, -17.510, -17.156, -16.711, -16.173, -15.557, -14.837, -14.197,
    -17.959, -17.901, -17.760, -17.515, -17.161, -16.715, -16.177, -15.562,
    -14.842, -14.202, -17.966, -17.907, -17.766, -17.521, -17.166, -16.719,
    -16.181, -15.567, -14.847, -14.206, -17.972, -17.913, -17.772, -17.526,
    -17.171, -16.723, -16.186, -15.572, -14.853, -14.211, -17.979, -17.920,
    -17.778, -17.532, -17.175, -16.728, -16.190, -15.577, -14.858, -14.216,
    -17.986, -17.926, -17.784, -17.537, -17.180, -16.732, -16.194, -15.582,
    -14.863, -14.221, -17.992, -17.932, -17.790, -17.543, -17.185, -16.736,
    -16.198, -15.587, -14.868, -14.226, -17.999, -17.938, -17.796, -17.548,
    -17.189, -16.740, -16.203, -15.592, -14.873, -14.231, -18.006, -17.945,
    -17.802, -17.554, -17.194, -16.744, -16.207, -15.597, -14.878, -14.236,
    -18.012, -17.951, -17.808, -17.559, -17.198, -16.749, -16.211, -15.603,
    -14.883, -14.240, -18.019, -17.957, -17.814, -17.564, -17.203, -16.753,
    -16.216, -15.608, -14.888, -14.245, -18.025, -17.964, -17.820, -17.570,
    -17.207, -16.757, -16.220, -15.613, -14.893, -14.250, -18.032, -17.970,
    -17.826, -17.575, -17.212, -16.761, -16.224, -15.618, -14.898, -14.255,
    -18.039, -17.976, -17.832, -17.581, -17.217, -16.766, -16.228, -15.623,
    -14.903, -14.260, -18.045, -17.983, -17.838, -17.586, -17.222, -16.770,
    -16.233, -15.628, -14.908, -14.265, -18.052, -17.989, -17.844, -17.592,
    -17.227, -16.775, -16.237, -15.632, -14.913, -14.270, -18.058, -17.995,
    -17.850, -17.598, -17.232, -16.779, -16.242, -15.637, -14.918, -14.274,
    -18.065, -18.002, -17.856, -17.603, -17.237, -16.784, -16.247, -15.642,
    -14.923, -14.279, -18.071, -18.008, -17.862, -17.609, -17.242, -16.788,
    -16.251, -15.647, -14.928, -14.284, -18.078, -18.014, -17.868, -17.614,
    -17.247, -16.793, -16.256, -15.652, -14.933, -14.289, -18.085, -18.021,
    -17.874, -17.620, -17.252, -16.798, -16.260, -15.656, -14.938, -14.294,
    -18.091, -18.027, -17.880, -17.626, -17.257, -16.802, -16.265, -15.661,
    -14.943, -14.299, -18.098, -18.033, -17.887, -17.631, -17.262, -16.807,
    -16.269, -15.666, -14.948, -14.304, -18.104, -18.040, -17.893, -17.637,
    -17.267, -16.812, -16.274, -15.671, -14.953, -14.308, -18.111, -18.046,
    -17.899, -17.643, -17.272, -16.816, -16.278, -15.675, -14.958, -14.313,
    -18.117, -18.052, -17.905, -17.648, -17.278, -16.821, -16.283, -15.680,
    -14.963, -14.318, -18.124, -18.058, -17.911, -17.654, -17.283, -16.826,
    -16.288, -15.685, -14.968, -14.323, -18.131, -18.065, -17.917, -17.660,
    -17.288, -16.831, -16.292, -15.690, -14.973, -14.328, -18.137, -18.071,
    -17.923, -17.666, -17.293, -16.835, -16.297, -15.694, -14.978, -14.333,
    -18.144, -18.077, -17.929, -17.671, -17.298, -16.840, -16.302, -15.699,
    -14.983, -14.337, -18.150, -18.084, -17.935, -17.677, -17.303, -16.845,
    -16.306, -15.704, -14.988, -14.342, -18.157, -18.090, -17.941, -17.683,
    -17.309, -16.850, -16.311, -15.708, -14.993, -14.347, -18.163, -18.096,
    -17.947, -17.688, -17.314, -16.855, -16.316, -15.713, -14.998, -14.352,
    -18.170, -18.103, -17.953, -17.694, -17.319, -16.859, -16.320, -15.718,
    -15.003, -14.357, -18.176, -18.109, -17.959, -17.700, -17.324, -16.864,
    -16.325, -15.722, -15.008, -14.362, -18.183, -18.115, -17.965, -17.706,
    -17.329, -16.869, -16.330, -15.727, -15.013, -14.366, -18.189, -18.122,
    -17.971, -17.711, -17.335, -16.874, -16.334, -15.731, -15.018, -14.371,
    -18.196, -18.128, -17.977, -17.717, -17.340, -16.879, -16.339, -15.736,
    -15.023, -14.376, -18.203, -18.134, -17.983, -17.723, -17.345, -16.884,
    -16.344, -15.741, -15.028, -14.381, -18.209, -18.141, -17.989, -17.728,
    -17.351, -16.889, -16.349, -15.746, -15.033, -14.386, -18.216, -18.147,
    -17.995, -17.734, -17.356, -16.894, -16.353, -15.750, -15.038, -14.390,
    -18.222, -18.153, -18.001, -17.740, -17.361, -16.899, -16.358, -15.755,
    -15.043, -14.395, -18.229, -18.159, -18.007, -17.745, -17.367, -16.904,
    -16.363, -15.760, -15.048, -14.400, -18.235, -18.165, -18.013, -17.751,
    -17.372, -16.909, -16.368, -15.765, -15.053, -14.405, -18.242, -18.172,
    -18.019, -17.757, -17.378, -16.914, -16.373, -15.770, -15.058, -14.410,
    -18.248, -18.178, -18.025, -17.762, -17.383, -16.919, -16.378, -15.775,
    -15.063, -14.415, -18.255, -18.184, -18.031, -17.768, -17.389, -16.924,
    -16.383, -15.779, -15.068, -14.419, -18.261, -18.190, -18.037, -17.774,
    -17.394, -16.930, -16.388, -15.784, -15.073, -14.424, -18.268, -18.197,
    -18.043, -17.779, -17.399, -16.935, -16.393, -15.789, -15.078, -14.429,
    -18.274, -18.203, -18.049, -17.785, -17.405, -16.940, -16.398, -15.794,
    -15.083, -14.434, -18.280, -18.209, -18.055, -17.791, -17.411, -16.945,
    -16.403, -15.799, -15.088, -14.439, -18.287, -18.215, -18.061, -17.797,
    -17.416, -16.950, -16.408, -15.804, -15.093, -14.443, -18.293, -18.221,
    -18.067, -17.802, -17.422, -16.956, -16.413, -15.809, -15.098, -14.448,
    -18.300, -18.228, -18.073, -17.808, -17.427, -16.961, -16.418, -15.814,
    -15.102, -14.453, -18.306, -18.234, -18.079, -17.814, -17.433, -16.966,
    -16.423, -15.819, -15.107, -14.458, -18.313, -18.240, -18.085, -17.819,
    -17.438, -16.971, -16.428, -15.823, -15.112, -14.463, -18.319, -18.246,
    -18.091, -17.825, -17.444, -16.977, -16.433, -15.828, -15.117, -14.467,
    -18.326, -18.252, -18.097, -17.831, -17.450, -16.982, -16.438, -15.833,
    -15.122, -14.472, -18.332, -18.258, -18.103, -17.836, -17.455, -16.987,
    -16.443, -15.838, -15.127, -14.477, -18.339, -18.265, -18.109, -17.842,
    -17.461, -16.993, -16.448, -15.843, -15.132, -14.482, -18.345, -18.271,
    -18.115, -17.848, -17.467, -16.998, -16.454, -15.848, -15.137, -14.486,
    -18.352, -18.277, -18.121, -17.853, -17.472, -17.004, -16.459, -15.853,
    -15.142, -14.491, -18.358, -18.283, -18.127, -17.859, -17.478, -17.009,
    -16.464, -15.858, -15.147, -14.496, -18.364, -18.289, -18.133, -17.865,
    -17.483, -17.014, -16.469, -15.863, -15.152, -14.501, -18.371, -18.295,
    -18.139, -17.870, -17.489, -17.019, -16.474, -15.868, -15.157, -14.506,
    -18.377, -18.301, -18.144, -17.876, -17.494, -17.025, -16.479, -15.873,
    -15.162, -14.510, -18.384, -18.307, -18.150, -17.881, -17.499, -17.030,
    -16.484, -15.878, -15.167, -14.515, -18.390, -18.314, -18.156, -17.887,
    -17.505, -17.035, -16.489, -15.883, -15.172, -14.520, -18.397, -18.320,
    -18.162, -17.892, -17.510, -17.040, -16.494, -15.887, -15.177, -14.525,
    -18.403, -18.326, -18.167, -17.898, -17.515, -17.045, -16.499, -15.892,
    -15.182, -14.529, -18.409, -18.332, -18.173, -17.903, -17.520, -17.050,
    -16.504, -15.897, -15.187, -14.534, -18.416, -18.338, -18.179, -17.908,
    -17.526, -17.056, -16.509, -15.902, -15.192, -14.539, -18.422, -18.344,
    -18.184, -17.914, -17.531, -17.061, -16.514, -15.907, -15.197, -14.543,
    -18.429, -18.350, -18.190, -17.919, -17.536, -17.066, -16.519, -15.912,
    -15.202, -14.548, -18.435, -18.356, -18.195, -17.925, -17.541, -17.071,
    -16.524, -15.917, -15.207, -14.553, -18.441, -18.362, -18.201, -17.930,
    -17.547, -17.076, -16.529, -15.922, -15.211, -14.558, -18.448, -18.368,
    -18.207, -17.935, -17.552, -17.081, -16.534, -15.927, -15.216, -14.562,
    -18.454, -18.374, -18.212, -17.941, -17.557, -17.087, -16.539, -15.932,
    -15.221, -14.567, -18.461, -18.380, -18.218, -17.946, -17.562, -17.092,
    -16.545, -15.937, -15.226, -14.572, -18.467, -18.386, -18.223, -17.952,
    -17.567, -17.097, -16.550, -15.942, -15.231, -14.576, -18.473, -18.392,
    -18.229, -17.957, -17.572, -17.102, -16.555, -15.946, -15.236, -14.581,
    -18.480, -18.398, -18.234, -17.962, -17.578, -17.107, -16.560, -15.951,
    -15.241, -14.586, -18.486, -18.404, -18.240, -17.968, -17.583, -17.112,
    -16.565, -15.956, -15.246, -14.591, -18.493, -18.410, -18.245, -17.973,
    -17.588, -17.117, -16.570, -15.961, -15.251, -14.595, -18.499, -18.416,
    -18.251, -17.978, -17.593, -17.123, -16.575, -15.966, -15.256, -14.600,
    -18.505, -18.422, -18.256, -17.983, -17.598, -17.128, -16.580, -15.971,
    -15.261, -14.605, -18.512, -18.428, -18.262, -17.989, -17.603, -17.133,
    -16.585, -15.976, -15.266, -14.609, -18.518, -18.434, -18.267, -17.994,
    -17.608, -17.138, -16.590, -15.981, -15.271, -14.614, -18.524, -18.440,
    -18.272, -17.999, -17.613, -17.143, -16.595, -15.986, -15.276, -14.619,
    -18.531, -18.446, -18.277, -18.004, -17.617, -17.147, -16.600, -15.990,
    -15.281, -14.623, -18.537, -18.452, -18.283, -18.008, -17.622, -17.152,
    -16.604, -15.995, -15.285, -14.628, -18.543, -18.457, -18.288, -18.013,
    -17.627, -17.157, -16.609, -16.000, -15.290, -14.633, -18.550, -18.463,
    -18.293, -18.018, -17.631, -17.162, -16.614, -16.005, -15.295, -14.637,
    -18.556, -18.469, -18.298, -18.022, -17.636, -17.166, -16.619, -16.009,
    -15.300, -14.642, -18.562, -18.475, -18.303, -18.027, -17.640, -17.171,
    -16.623, -16.014, -15.304, -14.646, -18.569, -18.481, -18.308, -18.031,
    -17.645, -17.175, -16.628, -16.019, -15.309, -14.651, -18.575, -18.486,
    -18.313, -18.036, -17.649, -17.180, -16.633, -16.023, -15.314, -14.656,
    -18.581, -18.492, -18.318, -18.040, -17.653, -17.184, -16.637, -16.028,
    -15.319, -14.660, -18.587, -18.498, -18.323, -18.045, -17.658, -17.189,
    -16.642, -16.033, -15.323, -14.665, -18.594, -18.504, -18.328, -18.049,
    -17.662, -17.193, -16.647, -16.037, -15.328, -14.669, -18.600, -18.509,
    -18.333, -18.054, -17.666, -17.198, -16.651, -16.042, -15.333, -14.674,
    -18.606, -18.515, -18.338, -18.058, -17.670, -17.202, -16.656, -16.046,
    -15.337, -14.678, -18.613, -18.521, -18.343, -18.062, -17.674, -17.206,
    -16.660, -16.051, -15.342, -14.683, -18.619, -18.526, -18.348, -18.066,
    -17.679, -17.211, -16.665, -16.056, -15.347, -14.687, -18.625, -18.532,
    -18.352, -18.070, -17.683, -17.215, -16.669, -16.060, -15.351, -14.692,
    -18.631, -18.538, -18.357, -18.074, -17.687, -17.219, -16.674, -16.065,
    -15.356, -14.696, -18.638, -18.543, -18.362, -18.079, -17.691, -17.223,
    -16.678, -16.069, -15.361, -14.701, -18.644, -18.549, -18.367, -18.083,
    -17.695, -17.228, -16.683, -16.074, -15.365, -14.706, -18.650, -18.555,
    -18.371, -18.087, -17.699, -17.232, -16.687, -16.078, -15.370, -14.710,
    -18.656, -18.560, -18.376, -18.090, -17.703, -17.236, -16.692, -16.083,
    -15.374, -14.715, -18.663, -18.566, -18.381, -18.094, -17.706, -17.240,
    -16.696, -16.087, -15.379, -14.719, -18.669, -18.571, -18.385, -18.098,
    -17.710, -17.244, -16.701, -16.092, -15.383, -14.724, -18.675, -18.577,
    -18.390, -18.102, -17.714, -17.248, -16.705, -16.096, -15.388, -14.728,
    -18.681, -18.582, -18.394, -18.106, -17.717, -17.252, -16.709, -16.100,
    -15.393, -14.732, -18.687, -18.588, -18.398, -18.109, -17.721, -17.256,
    -16.713, -16.105, -15.397, -14.737, -18.693, -18.593, -18.403, -18.112,
    -17.724, -17.259, -16.718, -16.109, -15.402, -14.741, -18.699, -18.598,
    -18.407, -18.116, -17.728, -17.263, -16.722, -16.113, -15.406, -14.746,
    -18.705, -18.604, -18.411, -18.119, -17.731, -17.267, -16.726, -16.118,
    -15.411, -14.750, -18.711, -18.609, -18.415, -18.122, -17.734, -17.270,
    -16.730, -16.122, -15.415, -14.755, -18.718, -18.614, -18.419, -18.125,
    -17.737, -17.274, -16.734, -16.126, -15.420, -14.759, -18.724, -18.619,
    -18.423, -18.129, -17.740, -17.277, -16.738, -16.130, -15.424, -14.763,
    -18.730, -18.625, -18.426, -18.132, -17.743, -17.281, -16.742, -16.135,
    -15.428, -14.768, -18.736, -18.630, -18.430, -18.135, -17.746, -17.284,
    -16.746, -16.139, -15.433, -14.772, -18.742, -18.635, -18.434, -18.137,
    -17.749, -17.288, -16.750, -16.143, -15.437, -14.777, -18.748, -18.640,
    -18.438, -18.140, -17.752, -17.291, -16.754, -16.147, -15.442, -14.781,
    -18.753, -18.645, -18.441, -18.143, -17.755, -17.295, -16.758, -16.151,
    -15.446, -14.785, -18.759, -18.650, -18.445, -18.146, -17.758, -17.298,
    -16.762, -16.156, -15.450, -14.790, -18.765, -18.655, -18.448, -18.149,
    -17.760, -17.301, -16.765, -16.160, -15.455, -14.794, -18.771, -18.660,
    -18.452, -18.151, -17.763, -17.304, -16.769, -16.164, -15.459, -14.798,
    -18.777, -18.665, -18.455, -18.154, -17.766, -17.307, -16.773, -16.168,
    -15.464, -14.803, -18.783, -18.670, -18.459, -18.156, -17.768, -17.310,
    -16.777, -16.172, -15.468, -14.807, -18.789, -18.675, -18.462, -18.159,
    -17.770, -17.314, -16.780, -16.176, -15.472, -14.811, -18.795, -18.680,
    -18.465, -18.161, -17.773, -17.317, -16.784, -16.180, -15.477, -14.816,
    -18.801, -18.685, -18.469, -18.163, -17.775, -17.320, -16.788, -16.184,
    -15.481, -14.820, -18.807, -18.690, -18.472, -18.166, -17.778, -17.322,
    -16.791, -16.188, -15.485, -14.824, -18.812, -18.694, -18.475, -18.168,
    -17.780, -17.325, -16.795, -16.192, -15.489, -14.828, -18.818, -18.699,
    -18.478, -18.170, -17.782, -17.328, -16.798, -16.196, -15.494, -14.833,
    -18.824, -18.704, -18.481, -18.172, -17.784, -17.331, -16.802, -16.200,
    -15.498, -14.837, -18.830, -18.708, -18.484, -18.174, -17.786, -17.334,
    -16.805, -16.204, -15.502, -14.841, -18.835, -18.712, -18.486, -18.175,
    -17.788, -17.336, -16.809, -16.208, -15.506, -14.845, -18.841, -18.716,
    -18.489, -18.177, -17.790, -17.339, -16.812, -16.211, -15.510, -14.850,
    -18.846, -18.720, -18.491, -18.178, -17.791, -17.341, -16.815, -16.215,
    -15.515, -14.854, -18.852, -18.724, -18.493, -18.180, -17.793, -17.344,
    -16.818, -16.219, -15.519, -14.858, -18.858, -18.728, -18.495, -18.181,
    -17.794, -17.346, -16.822, -16.222, -15.523, -14.862, -18.863, -18.732,
    -18.498, -18.182, -17.796, -17.349, -16.825, -16.226, -15.527, -14.866,
    -18.869, -18.736, -18.500, -18.183, -17.797, -17.351, -16.828, -16.230,
    -15.531, -14.870, -18.874, -18.739, -18.502, -18.184, -17.799, -17.353,
    -16.831, -16.233, -15.535, -14.875, -18.879, -18.743, -18.503, -18.185,
    -17.800, -17.355, -16.834, -16.237, -15.539, -14.879, -18.885, -18.746,
    -18.505, -18.186, -17.801, -17.357, -16.837, -16.240, -15.543, -14.883,
    -18.890, -18.750, -18.507, -18.186, -17.802, -17.359, -16.840, -16.244,
    -15.547, -14.887, -18.896, -18.753, -18.509, -18.187, -17.803, -17.362,
    -16.843, -16.247, -15.551, -14.891, -18.901, -18.756, -18.510, -18.188,
    -17.804, -17.364, -16.845, -16.251, -15.555, -14.895, -18.906, -18.760,
    -18.512, -18.188, -17.805, -17.365, -16.848, -16.254, -15.559, -14.899,
    -18.911, -18.763, -18.513, -18.188, -17.806, -17.367, -16.851, -16.258,
    -15.563, -14.903, -18.917, -18.766, -18.514, -18.189, -17.807, -17.369,
    -16.854, -16.261, -15.567, -14.907, -18.922, -18.769, -18.516, -18.189,
    -17.808, -17.371, -16.856, -16.264, -15.570, -14.911, -18.927, -18.772,
    -18.517, -18.189, -17.808, -17.373, -16.859, -16.268, -15.574, -14.915,
    -18.932, -18.775, -18.518, -18.189, -17.809, -17.374, -16.862, -16.271,
    -15.578, -14.919, -18.938, -18.777, -18.519, -18.189, -17.810, -17.376,
    -16.864, -16.274, -15.582, -14.923, -18.943, -18.780, -18.520, -18.189,
    -17.810, -17.378, -16.867, -16.277, -15.586, -14.927, -18.948, -18.783,
    -18.521, -18.189, -17.810, -17.379, -16.869, -16.281, -15.589, -14.931,
    -18.953, -18.785, -18.521, -18.188, -17.811, -17.381, -16.872, -16.284,
    -15.593, -14.935, -18.958, -18.788, -18.522, -18.188, -17.811, -17.382,
    -16.874, -16.287, -15.597, -14.939, -18.962, -18.790, -18.522, -18.187,
    -17.811, -17.383, -16.876, -16.290, -15.601, -14.943, -18.967, -18.792,
    -18.522, -18.187, -17.811, -17.385, -16.879, -16.293, -15.604, -14.947,
    -18.971, -18.794, -18.522, -18.186, -17.811, -17.386, -16.881, -16.296,
    -15.608, -14.951, -18.975, -18.796, -18.521, -18.185, -17.811, -17.387,
    -16.883, -16.299, -15.612, -14.955, -18.979, -18.797, -18.521, -18.184,
    -17.811, -17.388, -16.886, -16.302, -15.615, -14.959, -18.983, -18.799,
    -18.520, -18.184, -17.811, -17.390, -16.888, -16.305, -15.619, -14.962,
    -18.987, -18.800, -18.519, -18.182, -17.811, -17.391, -16.890, -16.308,
    -15.623, -14.966, -18.991, -18.802, -18.518, -18.181, -17.811, -17.392,
    -16.892, -16.311, -15.626, -14.970, -18.995, -18.803, -18.517, -18.180,
    -17.810, -17.393, -16.894, -16.314, -15.630, -14.974, -18.999, -18.804,
    -18.516, -18.179, -17.810, -17.394, -16.896, -16.317, -15.633, -14.978,
    -19.003, -18.805, -18.515, -18.177, -17.810, -17.395, -16.898, -16.320,
    -15.637, -14.982, -19.006, -18.806, -18.514, -18.176, -17.809, -17.395,
    -16.900, -16.322, -15.640, -14.985, -19.010, -18.807, -18.512, -18.174,
    -17.808, -17.396, -16.902, -16.325, -15.644, -14.989, -19.013, -18.808,
    -18.510, -18.173, -17.808, -17.397, -16.904, -16.328, -15.647, -14.993,
    -19.017, -18.808, -18.509, -18.171, -17.807, -17.398, -16.906, -16.331,
    -15.651, -14.997, -19.020, -18.809, -18.507, -18.169, -17.806, -17.398,
    -16.908, -16.333, -15.654, -15.001, -19.023, -18.809, -18.505, -18.167,
    -17.805, -17.399, -16.910, -16.336, -15.657, -15.004, -19.027, -18.809,
    -18.503, -18.165, -17.804, -17.400, -16.912, -16.338, -15.661, -15.008,
    -19.030, -18.810, -18.500, -18.163, -17.803, -17.400, -16.913, -16.341,
    -15.664, -15.012, -19.033, -18.810, -18.498, -18.160, -17.802, -17.400,
    -16.915, -16.344, -15.668, -15.016, -19.036, -18.810, -18.495, -18.158,
    -17.801, -17.401, -16.917, -16.346, -15.671, -15.019, -19.039, -18.810,
    -18.493, -18.156, -17.800, -17.401, -16.918, -16.349, -15.674, -15.023,
    -19.042, -18.810, -18.490, -18.153, -17.799, -17.402, -16.920, -16.351,
    -15.677, -15.027, -19.044, -18.809, -18.487, -18.151, -17.797, -17.402,
    -16.921, -16.354, -15.681, -15.030, -19.047, -18.809, -18.484, -18.148,
    -17.796, -17.402, -16.923, -16.356, -15.684, -15.034, -19.049, -18.808,
    -18.481, -18.145, -17.795, -17.402, -16.925, -16.358, -15.687, -15.038,
    -19.051, -18.807, -18.478, -18.143, -17.793, -17.403, -16.926, -16.361,
    -15.691, -15.041, -19.053, -18.805, -18.474, -18.140, -17.792, -17.403,
    -16.928, -16.363, -15.694, -15.045, -19.055, -18.804, -18.470, -18.137,
    -17.790, -17.403, -16.930, -16.366, -15.697, -15.049, -19.057, -18.802,
    -18.467, -18.134, -17.789, -17.403, -16.931, -16.368, -15.701, -15.053,
    -19.059, -18.800, -18.463, -18.131, -17.787, -17.403, -16.933, -16.370,
    -15.704, -15.056, -19.060, -18.798, -18.459, -18.128, -17.786, -17.403,
    -16.934, -16.373, -15.707, -15.060, -19.062, -18.796, -18.455, -18.124,
    -17.784, -17.403, -16.936, -16.375, -15.711, -15.064, -19.063, -18.794,
    -18.451, -18.121, -17.782, -17.403, -16.937, -16.377, -15.714, -15.067,
    -19.065, -18.792, -18.446, -18.118, -17.780, -17.403, -16.939, -16.380,
    -15.717, -15.071, -19.066, -18.789, -18.442, -18.114, -17.779, -17.403,
    -16.940, -16.382, -15.721, -15.075, -19.067, -18.786, -18.437, -18.111,
    -17.777, -17.403, -16.942, -16.384, -15.724, -15.079, -19.068, -18.784,
    -18.433, -18.107, -17.775, -17.403, -16.943, -16.386, -15.727, -15.082,
    -19.069, -18.781, -18.428, -18.103, -17.773, -17.402, -16.944, -16.389,
    -15.731, -15.086, -19.070, -18.777, -18.423, -18.100, -17.771, -17.402,
    -16.946, -16.391, -15.734, -15.090, -19.070, -18.774, -18.418, -18.096,
    -17.769, -17.402, -16.947, -16.393, -15.737, -15.093, -19.071, -18.771,
    -18.412, -18.092, -17.766, -17.401, -16.948, -16.395, -15.740, -15.097,
    -19.071, -18.767, -18.407, -18.088, -17.764, -17.401, -16.950, -16.397,
    -15.744, -15.101, -19.072, -18.763, -18.402, -18.083, -17.762, -17.400,
    -16.951, -16.399, -15.747, -15.104, -19.072, -18.760, -18.396, -18.079,
    -17.760, -17.400, -16.952, -16.401, -15.750, -15.108, -19.072, -18.756,
    -18.390, -18.075, -17.757, -17.399, -16.953, -16.403, -15.753, -15.111,
    -19.072, -18.751, -18.384, -18.071, -17.755, -17.399, -16.955, -16.405,
    -15.756, -15.115, -19.072, -18.747, -18.378, -18.066, -17.752, -17.398,
    -16.956, -16.407, -15.760, -15.119, -19.072, -18.743, -18.372, -18.062,
    -17.750, -17.398, -16.957, -16.409, -15.763, -15.122, -19.072, -18.738,
    -18.366, -18.057, -17.747, -17.397, -16.958, -16.411, -15.766, -15.126,
    -19.071, -18.733, -18.361, -18.053, -17.745, -17.397, -16.959, -16.413,
    -15.769, -15.130, -19.069, -18.729, -18.356, -18.049, -17.743, -17.396,
    -16.960, -16.416, -15.772, -15.133, -19.067, -18.724, -18.351, -18.045,
    -17.740, -17.396, -16.962, -16.418, -15.776, -15.137, -19.065, -18.719,
    -18.346, -18.041, -17.738, -17.396, -16.963, -16.420, -15.779, -15.141,
    -19.063, -18.714, -18.340, -18.037, -17.736, -17.395, -16.964, -16.422,
    -15.782, -15.145, -19.061, -18.708, -18.335, -18.032, -17.733, -17.395,
    -16.965, -16.425, -15.785, -15.148, -19.059, -18.703, -18.330, -18.028,
    -17.731, -17.395, -16.966, -16.427, -15.789, -15.152, -19.056, -18.698,
    -18.325, -18.024, -17.729, -17.394, -16.967, -16.429, -15.792, -15.156,
    -19.053, -18.692, -18.320, -18.020, -17.726, -17.394, -16.968, -16.432,
    -15.795, -15.160, -19.050, -18.686, -18.314, -18.015, -17.724, -17.394,
    -16.969, -16.434, -15.798, -15.163, -19.047, -18.680, -18.309, -18.011,
    -17.722, -17.393, -16.970, -16.436, -15.801, -15.167, -19.044, -18.674,
    -18.304, -18.007, -17.719, -17.393, -16.971, -16.439, -15.805, -15.171,
    -19.040, -18.668, -18.298, -18.002, -17.717, -17.392, -16.972, -16.441,
    -15.808, -15.175, -19.036, -18.662, -18.293, -17.998, -17.714, -17.392,
    -16.973, -16.444, -15.811, -15.178, -19.032, -18.655, -18.288, -17.993,
    -17.712, -17.392, -16.974, -16.446, -15.814, -15.182, -19.028, -18.648,
    -18.282, -17.989, -17.709, -17.391, -16.975, -16.448, -15.817, -15.186,
    -19.024, -18.642, -18.277, -17.984, -17.707, -17.391, -16.976, -16.451,
    -15.821, -15.190, -19.019, -18.635, -18.271, -17.980, -17.704, -17.390,
    -16.977, -16.453, -15.824, -15.193, -19.015, -18.628, -18.266, -17.975,
    -17.702, -17.390, -16.978, -16.455, -15.827, -15.197, -19.010, -18.621,
    -18.260, -17.971, -17.699, -17.389, -16.979, -16.458, -15.830, -15.201,
    -19.005, -18.613, -18.255, -17.966, -17.697, -17.389, -16.980, -16.460,
    -15.833, -15.205, -19.000, -18.606, -18.249, -17.961, -17.694, -17.388,
    -16.981, -16.463, -15.837, -15.209, -18.994, -18.598, -18.243, -17.956,
    -17.691, -17.388, -16.981, -16.465, -15.840, -15.212, -18.989, -18.591,
    -18.238, -17.952, -17.689, -17.387, -16.982, -16.468, -15.843, -15.216,
    -18.983, -18.583, -18.232, -17.947, -17.686, -17.387, -16.983, -16.470,
    -15.846, -15.220, -18.982, -18.580, -18.229, -17.944, -17.685, -17.387,
    -16.984, -16.473, -15.849, -15.224, -18.981, -18.577, -18.226, -17.942,
    -17.684, -17.388, -16.986, -16.475, -15.853, -15.227, -18.981, -18.574,
    -18.223, -17.940, -17.683, -17.388, -16.988, -16.478, -15.856, -15.231,
    -18.980, -18.572, -18.221, -17.938, -17.682, -17.388, -16.989, -16.480,
    -15.859, -15.235, -18.980, -18.570, -18.218, -17.936, -17.681, -17.389,
    -16.991, -16.483, -15.863, -15.239, -18.980, -18.568, -18.216, -17.934,
    -17.680, -17.389, -16.992, -16.485, -15.866, -15.242, -18.980, -18.566,
    -18.214, -17.932, -17.679, -17.390, -16.994, -16.488, -15.869, -15.246,
    -18.980, -18.564, -18.212, -17.931, -17.679, -17.390, -16.996, -16.491,
    -15.873, -15.250, -18.981, -18.563, -18.211, -17.929, -17.678, -17.391,
    -16.998, -16.493, -15.876, -15.254, -18.982, -18.562, -18.209, -17.928,
    -17.678, -17.392, -16.999, -16.496, -15.879, -15.257, -18.982, -18.561,
    -18.208, -17.927, -17.677, -17.392, -17.001, -16.499, -15.883, -15.261,
    -18.984, -18.560, -18.207, -17.926, -17.677, -17.393, -17.003, -16.502,
    -15.886, -15.265, -18.985, -18.560, -18.206, -17.925, -17.677, -17.394,
    -17.005, -16.504, -15.890, -15.268, -18.986, -18.560, -18.205, -17.924,
    -17.677, -17.395, -17.007, -16.507, -15.893, -15.272, -18.988, -18.560,
    -18.204, -17.924, -17.677, -17.396, -17.009, -16.510, -15.896, -15.276,
    -18.990, -18.560, -18.204, -17.923, -17.677, -17.397, -17.011, -16.512,
    -15.900, -15.280, -18.992, -18.561, -18.204, -17.923, -17.677, -17.398,
    -17.013, -16.515, -15.903, -15.283, -18.994, -18.562, -18.204, -17.923,
    -17.677, -17.399, -17.015, -16.518, -15.907, -15.287, -18.997, -18.563,
    -18.204, -17.923, -17.678, -17.400, -17.017, -16.521, -15.910, -15.291,
    -19.000, -18.564, -18.204, -17.923, -17.678, -17.401, -17.020, -16.524,
    -15.914, -15.294, -19.003, -18.566, -18.204, -17.923, -17.679, -17.403,
    -17.022, -16.527, -15.917, -15.298, -19.006, -18.567, -18.205, -17.924,
    -17.679, -17.404, -17.024, -16.529, -15.921, -15.302, -19.009, -18.569,
    -18.206, -17.924, -17.680, -17.405, -17.026, -16.532, -15.924, -15.306,
    -19.012, -18.572, -18.207, -17.925, -17.681, -17.407, -17.029, -16.535,
    -15.928, -15.309, -19.016, -18.574, -18.208, -17.926, -17.682, -17.408,
    -17.031, -16.538, -15.931, -15.313, -19.018, -18.574, -18.208, -17.926,
    -17.682, -17.409, -17.033, -16.541, -15.935, -15.317, -19.019, -18.574,
    -18.207, -17.925, -17.682, -17.410, -17.035, -16.544, -15.938, -15.321,
    -19.021, -18.574, -18.207, -17.925, -17.683, -17.411, -17.038, -16.547,
    -15.942, -15.324, -19.023, -18.575, -18.207, -17.925, -17.683, -17.413,
    -17.040, -16.550, -15.945, -15.328, -19.024, -18.575, -18.206, -17.925,
    -17.683, -17.414, -17.042, -16.553, -15.949, -15.332, -19.026, -18.575,
    -18.206, -17.925, -17.684, -17.415, -17.044, -16.556, -15.953, -15.336,
    -19.028, -18.576, -18.206, -17.924, -17.684, -17.416, -17.047, -16.559,
    -15.956, -15.340, -19.030, -18.576, -18.206, -17.924, -17.685, -17.418,
    -17.049, -16.563, -15.960, -15.343, -19.031, -18.576, -18.206, -17.924,
    -17.685, -17.419, -17.051, -16.566, -15.964, -15.347, -19.033, -18.577,
    -18.206, -17.924, -17.685, -17.420, -17.054, -16.569, -15.967, -15.351,
    -19.035, -18.577, -18.206, -17.924, -17.686, -17.422, -17.056, -16.572,
    -15.971, -15.355, -19.037, -18.578, -18.206, -17.924, -17.686, -17.423,
    -17.058, -16.575, -15.975, -15.359, -19.039, -18.578, -18.206, -17.925,
    -17.687, -17.424, -17.061, -16.578, -15.978, -15.363, -19.041, -18.579,
    -18.206, -17.925, -17.688, -17.426, -17.063, -16.582, -15.982, -15.366,
    -19.043, -18.580, -18.206, -17.925, -17.688, -17.427, -17.066, -16.585,
    -15.986, -15.370, -19.044, -18.580, -18.206, -17.925, -17.689, -17.429,
    -17.068, -16.588, -15.990, -15.374, -19.046, -18.581, -18.207, -17.925,
    -17.690, -17.430, -17.070, -16.592, -15.993, -15.378, -19.048, -18.582,
    -18.207, -17.926, -17.690, -17.432, -17.073, -16.595, -15.997, -15.382,
    -19.050, -18.583, -18.207, -17.926, -17.691, -17.433, -17.076, -16.598,
    -16.001, -15.386, -19.052, -18.583, -18.208, -17.926, -17.692, -17.435,
    -17.078, -16.602, -16.005, -15.390, -19.055, -18.584, -18.208, -17.927,
    -17.693, -17.436, -17.081, -16.605, -16.009, -15.393, -19.057, -18.585,
    -18.208, -17.927, -17.693, -17.438, -17.083, -16.609, -16.012, -15.397,
    -19.059, -18.586, -18.209, -17.928, -17.694, -17.440, -17.086, -16.612,
    -16.016, -15.401, -19.061, -18.587, -18.209, -17.928, -17.695, -17.441,
    -17.088, -16.616, -16.020, -15.405, -19.063, -18.588, -18.210, -17.929,
    -17.696, -17.443, -17.091, -16.619, -16.024, -15.409, -19.066, -18.590,
    -18.211, -17.930, -17.697, -17.445, -17.094, -16.622, -16.028, -15.413,
    -19.068, -18.591, -18.212, -17.931, -17.698, -17.446, -17.096, -16.626,
    -16.032, -15.417, -19.071, -18.593, -18.213, -17.932, -17.699, -17.448,
    -17.099, -16.629, -16.036, -15.421, -19.074, -18.595, -18.214, -17.933,
    -17.701, -17.450, -17.102, -16.633, -16.039, -15.425, -19.077, -18.596,
    -18.216, -17.934, -17.702, -17.452, -17.104, -16.636, -16.043, -15.429,
    -19.080, -18.598, -18.217, -17.935, -17.703, -17.454, -17.107, -16.640,
    -16.047, -15.432, -19.083, -18.600, -18.218, -17.936, -17.704, -17.456,
    -17.110, -16.643, -16.051, -15.436, -19.086, -18.602, -18.220, -17.937,
    -17.706, -17.458, -17.113, -16.647, -16.055, -15.440, -19.089, -18.605,
    -18.221, -17.939, -17.707, -17.460, -17.116, -16.650, -16.059, -15.444,
    -19.093, -18.607, -18.223, -17.940, -17.709, -17.461, -17.118, -16.654,
    -16.063, -15.448, -19.096, -18.609, -18.225, -17.941, -17.710, -17.463,
    -17.121, -16.657, -16.067, -15.452, -19.099, -18.611, -18.227, -17.943,
    -17.712, -17.465, -17.124, -16.661, -16.071, -15.456, -19.103, -18.614,
    -18.228, -17.944, -17.713, -17.468, -17.127, -16.665, -16.075, -15.460,
    -19.106, -18.616, -18.230, -17.946, -17.715, -17.470, -17.130, -16.668,
    -16.079, -15.464, -19.110, -18.619, -18.232, -17.948, -17.716, -17.472,
    -17.133, -16.672, -16.083, -15.468, -19.114, -18.622, -18.234, -17.949,
    -17.718, -17.474, -17.136, -16.676, -16.087, -15.472, -19.117, -18.624,
    -18.236, -17.951, -17.720, -17.476, -17.139, -16.679, -16.091, -15.476,
    -19.121, -18.627, -18.239, -17.953, -17.722, -17.478, -17.142, -16.683,
    -16.095, -15.480, -19.125, -18.630, -18.241, -17.955, -17.723, -17.480,
    -17.145, -16.687, -16.099, -15.484, -19.129, -18.633, -18.243, -17.957,
    -17.725, -17.483, -17.148, -16.690, -16.103, -15.488, -19.133, -18.636,
    -18.246, -17.959, -17.727, -17.485, -17.151, -16.694, -16.107, -15.492,
    -19.137, -18.639, -18.248, -17.961, -17.729, -17.487, -17.154, -16.698,
    -16.111, -15.496, -19.141, -18.642, -18.251, -17.963, -17.731, -17.489,
    -17.157, -16.701, -16.115, -15.500, -19.145, -18.646, -18.253, -17.965,
    -17.733, -17.492, -17.160, -16.705, -16.119, -15.504, -19.149, -18.649,
    -18.256, -17.967, -17.735, -17.494, -17.163, -16.709, -16.123, -15.508,
    -19.153, -18.652, -18.259, -17.969, -17.737, -17.496, -17.166, -16.713,
    -16.127, -15.512, -19.157, -18.656, -18.261, -17.972, -17.739, -17.499,
    -17.169, -16.716, -16.131, -15.516, -19.162, -18.659, -18.264, -17.974,
    -17.742, -17.501, -17.172, -16.720, -16.135, -15.520, -19.166, -18.663,
    -18.267, -17.976, -17.744, -17.504, -17.175, -16.724, -16.139, -15.524,
    -19.170, -18.666, -18.270, -17.979, -17.746, -17.506, -17.178, -16.727,
    -16.143, -15.528, -19.175, -18.670, -18.273, -17.981, -17.749, -17.509,
    -17.182, -16.731, -16.147, -15.532, -19.179, -18.674, -18.276, -17.984,
    -17.751, -17.512, -17.185, -16.735, -16.151, -15.536, -19.184, -18.677,
    -18.279, -17.987, -17.753, -17.514, -17.188, -16.739, -16.155, -15.540,
    -19.189, -18.681, -18.282, -17.989, -17.756, -17.517, -17.191, -16.742,
    -16.159, -15.544, -19.193, -18.685, -18.286, -17.992, -17.758, -17.519,
    -17.194, -16.746, -16.163, -15.548, -19.198, -18.689, -18.289, -17.995,
    -17.761, -17.522, -17.198, -16.750, -16.167, -15.552, -19.203, -18.693,
    -18.292, -17.998, -17.764, -17.525, -17.201, -16.754, -16.171, -15.556,
    -19.208, -18.697, -18.296, -18.001, -17.766, -17.528, -17.204, -16.757,
    -16.175, -15.560, -19.212, -18.702, -18.299, -18.004, -17.769, -17.530,
    -17.207, -16.761, -16.179, -15.564, -19.217, -18.706, -18.303, -18.007,
    -17.772, -17.533, -17.211, -16.765, -16.183, -15.568, -19.222, -18.710,
    -18.306, -18.010, -17.775, -17.536, -17.214, -16.769, -16.187, -15.572,
    -19.227, -18.714, -18.310, -18.013, -17.778, -17.539, -17.217, -16.772,
    -16.191, -15.576, -19.233, -18.719, -18.314, -18.017, -17.780, -17.542,
    -17.220, -16.776, -16.196, -15.580, -19.238, -18.723, -18.318, -18.020,
    -17.783, -17.545, -17.224, -16.780, -16.200, -15.584, -19.243, -18.728,
    -18.322, -18.023, -17.786, -17.548, -17.227, -16.784, -16.204, -15.588,
    -19.248, -18.733, -18.326, -18.027, -17.789, -17.551, -17.230, -16.788,
    -16.208, -15.592, -19.254, -18.737, -18.330, -18.030, -17.792, -17.554,
    -17.234, -16.792, -16.212, -15.596, -19.259, -18.742, -18.334, -18.034,
    -17.796, -17.557, -17.237, -16.795, -16.216, -15.600, -19.264, -18.747,
    -18.338, -18.037, -17.799, -17.560, -17.241, -16.799, -16.220, -15.604,
    -19.270, -18.752, -18.342, -18.041, -17.802, -17.563, -17.244, -16.803,
    -16.224, -15.608, -19.275, -18.757, -18.346, -18.044, -17.805, -17.566,
    -17.247, -16.807, -16.228, -15.612, -19.281, -18.761, -18.350, -18.048,
    -17.808, -17.569, -17.251, -16.811, -16.232, -15.616, -19.286, -18.766,
    -18.354, -18.051, -17.811, -17.572, -17.254, -16.814, -16.236, -15.620,
    -19.291, -18.771, -18.358, -18.055, -17.815, -17.575, -17.258, -16.818,
    -16.240, -15.624, -19.297, -18.776, -18.362, -18.059, -17.818, -17.578,
    -17.261, -16.822, -16.244, -15.628, -19.302, -18.780, -18.367, -18.062,
    -17.821, -17.582, -17.264, -16.826, -16.248, -15.632, -19.307, -18.785,
    -18.371, -18.066, -17.824, -17.585, -17.268, -16.830, -16.252, -15.636,
    -19.313, -18.790, -18.375, -18.070, -17.828, -17.588, -17.271, -16.834,
    -16.256, -15.640, -19.318, -18.795, -18.380, -18.074, -17.831, -17.591,
    -17.275, -16.837, -16.260, -15.644, -19.324, -18.800, -18.384, -18.077,
    -17.834, -17.595, -17.278, -16.841, -16.264, -15.647, -19.330, -18.805,
    -18.388, -18.081, -17.838, -17.598, -17.282, -16.845, -16.268, -15.651,
    -19.335, -18.810, -18.393, -18.085, -17.841, -17.601, -17.285, -16.849,
    -16.272, -15.655, -19.341, -18.815, -18.397, -18.089, -17.845, -17.604,
    -17.288, -16.853, -16.276, -15.659, -19.346, -18.821, -18.402, -18.093,
    -17.848, -17.608, -17.292, -16.857, -16.280, -15.663, -19.352, -18.826,
    -18.407, -18.097, -17.852, -17.611, -17.295, -16.860, -16.284, -15.667,
    -19.358, -18.831, -18.411, -18.101, -17.855, -17.615, -17.299, -16.864,
    -16.288, -15.671, -19.364, -18.836, -18.416, -18.105, -17.859, -17.618,
    -17.303, -16.868, -16.292, -15.675, -19.370, -18.842, -18.421, -18.109,
    -17.863, -17.621, -17.306, -16.872, -16.295, -15.679, -19.375, -18.847,
    -18.425, -18.113, -17.866, -17.625, -17.310, -16.876, -16.299, -15.683,
    -19.381, -18.852, -18.430, -18.117, -17.870, -17.628, -17.313, -16.880,
    -16.303, -15.687, -19.387, -18.858, -18.435, -18.122, -17.874, -17.632,
    -17.317, -16.884, -16.307, -15.690, -19.393, -18.863, -18.440, -18.126,
    -17.878, -17.635, -17.320, -16.887, -16.311, -15.694, -19.399, -18.869,
    -18.445, -18.130, -17.881, -17.639, -17.324, -16.891, -16.315, -15.698,
    -19.405, -18.874, -18.450, -18.135, -17.885, -17.642, -17.327, -16.895,
    -16.319, -15.702, -19.411, -18.880, -18.455, -18.139, -17.889, -17.646,
    -17.331, -16.899, -16.323, -15.706, -19.417, -18.885, -18.460, -18.143,
    -17.893, -17.650, -17.335, -16.903, -16.327, -15.710, -19.423, -18.891,
    -18.465, -18.148, -17.897, -17.653, -17.338, -16.907, -16.331, -15.714,
    -19.429, -18.896, -18.470, -18.152, -17.900, -17.657, -17.342, -16.910,
    -16.335, -15.717, -19.434, -18.902, -18.475, -18.156, -17.904, -17.660,
    -17.345, -16.914, -16.339, -15.721, -19.440, -18.907, -18.480, -18.161,
    -17.908, -17.664, -17.349, -16.918, -16.343, -15.725, -19.446, -18.913,
    -18.484, -18.165, -17.912, -17.667, -17.353, -16.922, -16.346, -15.729,
    -19.452, -18.919, -18.490, -18.169, -17.916, -17.671, -17.356, -16.926,
    -16.350, -15.733, -19.458, -18.924, -18.495, -18.174, -17.920, -17.675,
    -17.360, -16.930, -16.354, -15.737, -19.464, -18.930, -18.500, -18.178,
    -17.924, -17.678, -17.364, -16.933, -16.358, -15.740, -19.470, -18.935,
    -18.505, -18.183, -17.928, -17.682, -17.367, -16.937, -16.362, -15.744,
    -19.476, -18.941, -18.510, -18.188, -17.932, -17.686, -17.371, -16.941,
    -16.366, -15.748, -19.482, -18.947, -18.515, -18.192, -17.936, -17.689,
    -17.375, -16.945, -16.370, -15.752, -19.488, -18.952, -18.520, -18.197,
    -17.940, -17.693, -17.378, -16.949, -16.374, -15.755, -19.494, -18.958,
    -18.525, -18.201, -17.944, -17.697, -17.382, -16.952, -16.378, -15.759,
    -19.500, -18.964, -18.530, -18.206, -17.948, -17.701, -17.386, -16.956,
    -16.381, -15.763, -19.506, -18.970, -18.536, -18.211, -17.953, -17.704,
    -17.389, -16.960, -16.385, -15.767, -19.512, -18.976, -18.541, -18.215,
    -17.957, -17.708, -17.393, -16.964, -16.389, -15.770, -19.518, -18.981,
    -18.546, -18.220, -17.961, -17.712, -17.397, -16.968, -16.393, -15.774,
    -19.524, -18.987, -18.552, -18.225, -17.965, -17.716, -17.400, -16.971,
    -16.397, -15.778, -19.530, -18.993, -18.557, -18.230, -17.969, -17.720,
    -17.404, -16.975, -16.401, -15.782, -19.536, -18.999, -18.562, -18.234,
    -17.974, -17.723, -17.408, -16.979, -16.405, -15.785, -19.543, -19.005,
    -18.568, -18.239, -17.978, -17.727, -17.412, -16.983, -16.408, -15.789,
    -19.549, -19.011, -18.573, -18.244, -17.982, -17.731, -17.415, -16.986,
    -16.412, -15.793, -19.555, -19.017, -18.579, -18.249, -17.987, -17.735,
    -17.419, -16.990, -16.416, -15.796, -19.561, -19.023, -18.584, -18.254,
    -17.991, -17.739, -17.423, -16.994, -16.420, -15.800, -19.567, -19.029,
    -18.589, -18.259, -17.995, -17.743, -17.427, -16.998, -16.424, -15.804,
    -19.573, -19.035, -18.595, -18.264, -18.000, -17.747, -17.431, -17.002,
    -16.428, -15.807, -19.579, -19.041, -18.600, -18.269, -18.004, -17.751,
    -17.434, -17.005, -16.431, -15.811, -19.585, -19.047, -18.606, -18.273,
    -18.008, -17.755, -17.438, -17.009, -16.435, -15.815, -19.592, -19.052,
    -18.611, -18.278, -18.013, -17.759, -17.442, -17.013, -16.439, -15.819,
    -19.598, -19.058, -18.616, -18.283, -18.017, -17.763, -17.446, -17.017,
    -16.443, -15.822, -19.604, -19.064, -18.622, -18.288, -18.021, -17.766,
    -17.449, -17.020, -16.446, -15.826, -19.610, -19.070, -18.627, -18.293,
    -18.026, -17.770, -17.453, -17.024, -16.450, -15.830, -19.616, -19.076,
    -18.633, -18.298, -18.030, -17.774, -17.457, -17.028, -16.454, -15.833,
    -19.622, -19.082, -18.638, -18.303, -18.035, -17.778, -17.461, -17.032,
    -16.457, -15.837, -19.629, -19.088, -18.644, -18.308, -18.039, -17.782,
    -17.465, -17.035, -16.461, -15.841, -19.635, -19.094, -18.649, -18.313,
    -18.044, -17.786, -17.469, -17.039, -16.465, -15.844, -19.641, -19.100,
    -18.655, -18.318, -18.048, -17.791, -17.472, -17.043, -16.468, -15.848,
    -19.647, -19.106, -18.661, -18.323, -18.052, -17.795, -17.476, -17.047,
    -16.472, -15.852, -19.653, -19.112, -18.666, -18.328, -18.057, -17.799,
    -17.480, -17.051, -16.476, -15.855, -19.660, -19.118, -18.672, -18.333,
    -18.062, -17.803, -17.484, -17.054, -16.479, -15.859, -19.666, -19.124,
    -18.677, -18.338, -18.066, -17.807, -17.488, -17.058, -16.483, -15.863,
    -19.672, -19.130, -18.683, -18.344, -18.071, -17.811, -17.492, -17.062,
    -16.487, -15.866, -19.678, -19.136, -18.689, -18.349, -18.075, -17.815,
    -17.496, -17.066, -16.490, -15.870, -19.685, -19.142, -18.694, -18.354,
    -18.080, -17.819, -17.499, -17.069, -16.494, -15.874, -19.691, -19.149,
    -18.700, -18.359, -18.084, -17.823, -17.503, -17.073, -16.498, -15.877,
    -19.697, -19.155, -18.706, -18.364, -18.089, -17.827, -17.507, -17.077,
    -16.501, -15.881, -19.703, -19.161, -18.712, -18.370, -18.094, -17.832,
    -17.511, -17.081, -16.505, -15.885, -19.710, -19.167, -18.717, -18.375,
    -18.098, -17.836, -17.515, -17.084, -16.508, -15.888, -19.716, -19.173,
    -18.723, -18.380, -18.103, -17.840, -17.519, -17.088, -16.512, -15.892,
    -19.722, -19.179, -18.729, -18.385, -18.108, -17.844, -17.523, -17.092,
    -16.516, -15.896, -19.728, -19.185, -18.734, -18.390, -18.112, -17.848,
    -17.527, -17.095, -16.519, -15.899, -19.735, -19.191, -18.740, -18.396,
    -18.117, -17.852, -17.531, -17.099, -16.523, -15.903, -19.741, -19.197,
    -18.746, -18.401, -18.122, -17.857, -17.534, -17.103, -16.527, -15.906,
    -19.747, -19.203, -18.752, -18.406, -18.126, -17.861, -17.538, -17.107,
    -16.530, -15.910, -19.753, -19.210, -18.757, -18.411, -18.131, -17.865,
    -17.542, -17.110, -16.534, -15.914, -19.759, -19.216, -18.763, -18.417,
    -18.136, -17.869, -17.546, -17.114, -16.537, -15.917, -19.766, -19.222,
    -18.769, -18.422, -18.141, -17.873, -17.550, -17.118, -16.541, -15.921,
    -19.772, -19.228, -18.775, -18.427, -18.145, -17.878, -17.554, -17.122,
    -16.545, -15.924, -19.778, -19.234, -18.781, -18.432, -18.150, -17.882,
    -17.558, -17.125, -16.548, -15.928, -19.784, -19.240, -18.786, -18.438,
    -18.155, -17.886, -17.562, -17.129, -16.552, -15.932, -19.790, -19.246,
    -18.792, -18.443, -18.160, -17.890, -17.565, -17.133, -16.555, -15.935,
    -19.797, -19.253, -18.798, -18.448, -18.164, -17.894, -17.569, -17.136,
    -16.559, -15.939, -19.803, -19.259, -18.804, -18.454, -18.169, -17.899,
    -17.573, -17.140, -16.563, -15.942, -19.809, -19.265, -18.810, -18.459,
    -18.174, -17.903, -17.577, -17.144, -16.566, -15.946, -19.815, -19.271,
    -18.816, -18.464, -18.179, -17.907, -17.581, -17.148, -16.570, -15.949,
    -19.821, -19.277, -18.822, -18.470, -18.184, -17.912, -17.585, -17.151,
    -16.573, -15.953, -19.828, -19.283, -18.827, -18.475, -18.189, -17.916,
    -17.589, -17.155, -16.577, -15.956, -19.834, -19.290, -18.833, -18.480,
    -18.193, -17.920, -17.593, -17.159, -16.581, -15.960, -19.840, -19.296,
    -18.839, -18.486, -18.198, -17.924, -17.597, -17.162, -16.584, -15.963,
    -19.846, -19.302, -18.845, -18.491, -18.203, -17.929, -17.600, -17.166,
    -16.588, -15.967, -19.852, -19.308, -18.851, -18.497, -18.208, -17.933,
    -17.604, -17.170, -16.591, -15.970, -19.859, -19.315, -18.857, -18.502,
    -18.213, -17.937, -17.608, -17.174, -16.595, -15.974, -19.865, -19.321,
    -18.863, -18.508, -18.218, -17.942, -17.612, -17.177, -16.598, -15.977,
    -19.871, -19.327, -18.869, -18.513, -18.223, -17.946, -17.616, -17.181,
    -16.602, -15.981, -19.877, -19.333, -18.875, -18.518, -18.228, -17.950,
    -17.620, -17.185, -16.606, -15.985, -19.883, -19.339, -18.881, -18.524,
    -18.233, -17.955, -17.624, -17.188, -16.609, -15.988, -19.890, -19.346,
    -18.887, -18.529, -18.238, -17.959, -17.628, -17.192, -16.613, -15.992,
    -19.896, -19.352, -18.893, -18.535, -18.243, -17.963, -17.632, -17.196,
    -16.616, -15.995, -19.902, -19.358, -18.898, -18.540, -18.247, -17.968,
    -17.636, -17.200, -16.620, -15.999, -19.908, -19.364, -18.904, -18.546,
    -18.252, -17.972, -17.640, -17.203, -16.624, -16.002, -19.914, -19.370,
    -18.910, -18.551, -18.257, -17.977, -17.644, -17.207, -16.627, -16.006,
    -19.921, -19.377, -18.916, -18.556, -18.262, -17.981, -17.648, -17.211,
    -16.631, -16.009, -19.927, -19.383, -18.922, -18.562, -18.267, -17.985,
    -17.652, -17.214, -16.635, -16.013, -19.933, -19.389, -18.928, -18.567,
    -18.272, -17.990, -17.656, -17.218, -16.638, -16.017, -19.939, -19.395,
    -18.934, -18.573, -18.277, -17.994, -17.660, -17.222, -16.642, -16.020,
    -19.945, -19.402, -18.940, -18.578, -18.282, -17.999, -17.664, -17.226,
    -16.646, -16.024, -19.952, -19.408, -18.946, -18.584, -18.287, -18.003,
    -17.668, -17.229, -16.649, -16.027, -19.958, -19.414, -18.952, -18.589,
    -18.292, -18.008, -17.672, -17.233, -16.653, -16.031, -19.964, -19.420,
    -18.958, -18.595, -18.297, -18.012, -17.676, -17.237, -16.656, -16.034,
    -19.970, -19.427, -18.963, -18.600, -18.302, -18.016, -17.680, -17.241,
    -16.660, -16.038, -19.976, -19.433, -18.969, -18.606, -18.307, -18.021,
    -17.684, -17.244, -16.664, -16.042, -19.983, -19.439, -18.975, -18.611,
    -18.312, -18.025, -17.688, -17.248, -16.667, -16.045, -19.989, -19.445,
    -18.981, -18.617, -18.317, -18.030, -17.692, -17.252, -16.671, -16.049,
    -19.995, -19.452, -18.987, -18.622, -18.322, -18.034, -17.696, -17.255,
    -16.675, -16.052, -20.001, -19.458, -18.993, -18.628, -18.327, -18.039,
    -17.700, -17.259, -16.678, -16.056, -20.007, -19.464, -18.999, -18.633,
    -18.332, -18.043, -17.704, -17.263, -16.682, -16.059, -20.014, -19.470,
    -19.005, -18.639, -18.337, -18.048, -17.708, -17.267, -16.686, -16.063,
    -20.020, -19.477, -19.011, -18.644, -18.342, -18.052, -17.712, -17.270,
    -16.689, -16.066, -20.026, -19.483, -19.017, -18.650, -18.347, -18.057,
    -17.716, -17.274, -16.693, -16.070, -20.032, -19.489, -19.023, -18.656,
    -18.352, -18.062, -17.720, -17.278, -16.697, -16.074, -20.038, -19.495,
    -19.029, -18.661, -18.357, -18.066, -17.724, -17.282, -16.700, -16.077,
    -20.045, -19.502, -19.035, -18.667, -18.362, -18.071, -17.728, -17.285,
    -16.704, -16.081, -20.051, -19.508, -19.041, -18.672, -18.367, -18.075,
    -17.732, -17.289, -16.708, -16.084, -20.057, -19.514, -19.047, -18.678,
    -18.372, -18.080, -17.736, -17.293, -16.711, -16.088, -20.063, -19.520,
    -19.053, -18.683, -18.377, -18.084, -17.740, -17.297, -16.715, -16.092,
    -20.069, -19.527, -19.059, -18.689, -18.382, -18.089, -17.744, -17.300,
    -16.719, -16.095, -20.075, -19.533, -19.065, -18.695, -18.387, -18.093,
    -17.748, -17.304, -16.722, -16.099, -20.082, -19.539, -19.071, -18.700,
    -18.392, -18.098, -17.752, -17.308, -16.726, -16.102, -20.088, -19.545,
    -19.077, -18.706, -18.397, -18.102, -17.757, -17.312, -16.730, -16.106,
    -20.094, -19.552, -19.083, -18.712, -18.403, -18.107, -17.761, -17.316,
    -16.733, -16.109, -20.100, -19.558, -19.089, -18.717, -18.408, -18.111,
    -17.765, -17.319, -16.737, -16.113, -20.106, -19.564, -19.095, -18.723,
    -18.413, -18.116, -17.769, -17.323, -16.741, -16.117, -20.112, -19.570,
    -19.101, -18.728, -18.418, -18.120, -17.773, -17.327, -16.744, -16.120,
    -20.119, -19.577, -19.107, -18.734, -18.423, -18.125, -17.777, -17.331,
    -16.748, -16.124, -20.125, -19.583, -19.113, -18.740, -18.428, -18.130,
    -17.781, -17.335, -16.752, -16.127, -20.131, -19.589, -19.119, -18.745,
    -18.433, -18.134, -17.785, -17.338, -16.755, -16.131, -20.137, -19.595,
    -19.125, -18.751, -18.438, -18.139, -17.789, -17.342, -16.759, -16.135,
    -20.143, -19.602, -19.131, -18.757, -18.443, -18.143, -17.793, -17.346,
    -16.763, -16.138, -20.149, -19.608, -19.138, -18.762, -18.448, -18.148,
    -17.797, -17.350, -16.767, -16.142, -20.155, -19.614, -19.144, -18.768,
    -18.453, -18.153, -17.802, -17.354, -16.770, -16.146, -20.162, -19.620,
    -19.150, -18.774, -18.459, -18.157, -17.806, -17.357, -16.774, -16.149,
    -20.168, -19.627, -19.156, -18.780, -18.464, -18.162, -17.810, -17.361,
    -16.778, -16.153, -20.174, -19.633, -19.162, -18.785, -18.469, -18.166,
    -17.814, -17.365, -16.781, -16.156, -20.180, -19.639, -19.168, -18.791,
    -18.474, -18.171, -17.818, -17.369, -16.785, -16.160, -20.186, -19.645,
    -19.174, -18.797, -18.479, -18.176, -17.822, -17.373, -16.789, -16.164,
    -20.192, -19.651, -19.180, -18.802, -18.484, -18.180, -17.826, -17.377,
    -16.792, -16.167, -20.198, -19.658, -19.186, -18.808, -18.489, -18.185,
    -17.830, -17.381, -16.796, -16.171, -20.204, -19.664, -19.192, -18.814,
    -18.495, -18.189, -17.834, -17.384, -16.800, -16.174, -20.211, -19.670,
    -19.198, -18.819, -18.500, -18.194, -17.839, -17.388, -16.804, -16.178,
    -20.217, -19.676, -19.204, -18.825, -18.505, -18.199, -17.843, -17.392,
    -16.807, -16.182, -20.223, -19.683, -19.210, -18.831, -18.510, -18.203,
    -17.847, -17.396, -16.811, -16.185, -20.229, -19.689, -19.216, -18.836,
    -18.515, -18.208, -17.851, -17.400, -16.815, -16.189, -20.235, -19.695,
    -19.222, -18.842, -18.520, -18.212, -17.855, -17.404, -16.819, -16.193,
    -20.241, -19.701, -19.228, -18.848, -18.526, -18.217, -17.859, -17.408,
    -16.822, -16.196, -20.247, -19.708, -19.234, -18.853, -18.531, -18.221,
    -17.863, -17.411, -16.826, -16.200, -20.253, -19.714, -19.240, -18.859,
    -18.536, -18.226, -17.867, -17.415, -16.830, -16.204, -20.259, -19.720,
    -19.247, -18.865, -18.541, -18.231, -17.871, -17.419, -16.834, -16.207,
    -20.265, -19.726, -19.253, -18.870, -18.546, -18.235, -17.876, -17.423,
    -16.837, -16.211, -20.271, -19.733, -19.259, -18.876, -18.552, -18.240,
    -17.880, -17.427, -16.841, -16.214, -20.278, -19.739, -19.265, -18.882,
    -18.557, -18.244, -17.884, -17.431, -16.845, -16.218, -20.284, -19.745,
    -19.271, -18.887, -18.562, -18.249, -17.888, -17.435, -16.849, -16.222,
    -20.290, -19.751, -19.277, -18.893, -18.567, -18.254, -17.892, -17.439,
    -16.852, -16.225, -20.296, -19.758, -19.283, -18.899, -18.573, -18.258,
    -17.896, -17.443, -16.856, -16.229, -20.302, -19.764, -19.289, -18.905,
    -18.578, -18.263, -17.900, -17.446, -16.860, -16.233, -20.308, -19.770,
    -19.295, -18.910, -18.583, -18.268, -17.904, -17.450, -16.864, -16.236,
    -20.314, -19.776, -19.301, -18.916, -18.588, -18.272, -17.909, -17.454,
    -16.868, -16.240, -20.320, -19.783, -19.307, -18.922, -18.593, -18.277,
    -17.913, -17.458, -16.871, -16.244, -20.326, -19.789, -19.313, -18.927,
    -18.599, -18.281, -17.917, -17.462, -16.875, -16.247, -20.332, -19.795,
    -19.319, -18.933, -18.604, -18.286, -17.921, -17.466, -16.879, -16.251,
    -20.338, -19.801, -19.325, -18.939, -18.609, -18.291, -17.925, -17.470,
    -16.883, -16.255, -20.344, -19.807, -19.331, -18.944, -18.614, -18.295,
    -17.929, -17.474, -16.887, -16.258, -20.350, -19.814, -19.337, -18.950,
    -18.620, -18.300, -17.933, -17.478, -16.890, -16.262, -20.356, -19.820,
    -19.343, -18.956, -18.625, -18.304, -17.938, -17.482, -16.894, -16.266,
    -20.362, -19.826, -19.349, -18.961, -18.630, -18.309, -17.942, -17.485,
    -16.898, -16.269, -20.368, -19.832, -19.355, -18.967, -18.635, -18.314,
    -17.946, -17.489, -16.902, -16.273, -20.374, -19.838, -19.361, -18.973,
    -18.640, -18.318, -17.950, -17.493, -16.906, -16.277, -20.380, -19.844,
    -19.367, -18.978, -18.645, -18.323, -17.954, -17.497, -16.910, -16.280,
    -20.386, -19.851, -19.373, -18.984, -18.651, -18.327, -17.958, -17.501,
    -16.913, -16.284, -20.392, -19.857, -19.379, -18.990, -18.656, -18.332,
    -17.962, -17.505, -16.917, -16.288, -20.398, -19.863, -19.385, -18.995,
    -18.661, -18.337, -17.967, -17.509, -16.921, -16.291, -20.404, -19.869,
    -19.391, -19.001, -18.666, -18.341, -17.971, -17.513, -16.925, -16.295,
    -20.410, -19.875, -19.398, -19.007, -18.671, -18.346, -17.975, -17.516,
    -16.929, -16.299, -20.416, -19.881, -19.404, -19.013, -18.676, -18.350,
    -17.979, -17.520, -16.933, -16.302, -20.422, -19.888, -19.410, -19.018,
    -18.682, -18.355, -17.983, -17.524, -16.936, -16.306, -20.428, -19.894,
    -19.416, -19.024, -18.687, -18.360, -17.987, -17.528, -16.940, -16.310,
    -20.434, -19.900, -19.422, -19.030, -18.692, -18.364, -17.992, -17.532,
    -16.944, -16.313, -20.440, -19.906, -19.428, -19.035, -18.697, -18.369,
    -17.996, -17.536, -16.948, -16.317, -20.446, -19.912, -19.434, -19.041,
    -18.702, -18.373, -18.000, -17.540, -16.952, -16.321, -20.452, -19.918,
    -19.440, -19.047, -18.707, -18.378, -18.004, -17.544, -16.956, -16.325,
    -20.458, -19.924, -19.446, -19.052, -18.712, -18.383, -18.008, -17.547,
    -16.960, -16.328, -20.464, -19.931, -19.452, -19.058, -18.718, -18.387,
    -18.012, -17.551, -16.963, -16.332, -20.470, -19.937, -19.458, -19.064,
    -18.723, -18.392, -18.017, -17.555, -16.967, -16.336, -20.476, -19.943,
    -19.464, -19.069, -18.728, -18.396, -18.021, -17.559, -16.971, -16.339,
    -20.482, -19.949, -19.470, -19.075, -18.733, -18.401, -18.025, -17.563,
    -16.975, -16.343, -20.488, -19.955, -19.476, -19.081, -18.738, -18.406,
    -18.029, -17.567, -16.979, -16.347, -20.494, -19.961, -19.482, -19.086,
    -18.743, -18.410, -18.033, -17.571, -16.983, -16.350, -20.500, -19.967,
    -19.488, -19.092, -18.748, -18.415, -18.037, -17.575, -16.986, -16.354,
    -20.506, -19.974, -19.494, -19.098, -18.754, -18.419, -18.041, -17.578,
    -16.990, -16.358, -20.512, -19.980, -19.500, -19.103, -18.759, -18.424,
    -18.045, -17.582, -16.994, -16.361, -20.518, -19.986, -19.506, -19.109,
    -18.764, -18.428, -18.050, -17.586, -16.998, -16.365, -20.523, -19.992,
    -19.512, -19.115, -18.769, -18.433, -18.054, -17.590, -17.002, -16.368,
    -20.529, -19.998, -19.518, -19.120, -18.774, -18.437, -18.058, -17.594,
    -17.005, -16.372, -20.535, -20.004, -19.524, -19.126, -18.779, -18.442,
    -18.062, -17.597, -17.009, -16.376, -20.541, -20.010, -19.530, -19.132,
    -18.785, -18.446, -18.066, -17.601, -17.013, -16.379, -20.547, -20.016,
    -19.536, -19.137, -18.790, -18.451, -18.070, -17.605, -17.017, -16.383,
    -20.553, -20.023, -19.542, -19.143, -18.795, -18.455, -18.074, -17.609,
    -17.020, -16.386, -20.559, -20.029, -19.548, -19.148, -18.800, -18.460,
    -18.078, -17.613, -17.024, -16.390, -20.565, -20.035, -19.554, -19.154,
    -18.805, -18.465, -18.082, -17.616, -17.028, -16.394, -20.571, -20.041,
    -19.560, -19.160, -18.810, -18.469, -18.086, -17.620, -17.032, -16.397,
    -20.576, -20.047, -19.566, -19.165, -18.816, -18.474, -18.090, -17.624,
    -17.035, -16.401, -20.582, -20.053, -19.572, -19.171, -18.821, -18.478,
    -18.094, -17.628, -17.039, -16.404, -20.588, -20.059, -19.578, -19.177,
    -18.826, -18.483, -18.098, -17.632, -17.043, -16.408, -20.594, -20.065,
    -19.584, -19.182, -18.831, -18.487, -18.102, -17.635, -17.047, -16.412,
    -20.600, -20.071, -19.590, -19.188, -18.836, -18.492, -18.106, -17.639,
    -17.050, -16.415, -20.606, -20.078, -19.596, -19.194, -18.841, -18.496,
    -18.110, -17.643, -17.054, -16.419, -20.612, -20.084, -19.602, -19.199,
    -18.847, -18.501, -18.114, -17.647, -17.058, -16.422, -20.617, -20.090,
    -19.608, -19.205, -18.852, -18.505, -18.118, -17.650, -17.062, -16.426,
    -20.623, -20.096, -19.614, -19.210, -18.857, -18.510, -18.122, -17.654,
    -17.065, -16.429, -20.629, -20.102, -19.620, -19.216, -18.862, -18.514,
    -18.126, -17.658, -17.069, -16.433, -20.635, -20.108, -19.626, -19.222,
    -18.867, -18.518, -18.130, -17.662, -17.073, -16.437, -20.641, -20.114,
    -19.632, -19.227, -18.872, -18.523, -18.134, -17.665, -17.076, -16.440,
    -20.647, -20.120, -19.638, -19.233, -18.877, -18.527, -18.138, -17.669,
    -17.080, -16.444, -20.652, -20.126, -19.644, -19.238, -18.882, -18.532,
    -18.142, -17.673, -17.084, -16.447, -20.658, -20.132, -19.650, -19.244,
    -18.887, -18.536, -18.146, -17.677, -17.087, -16.451, -20.664, -20.138,
    -19.656, -19.250, -18.892, -18.541, -18.150, -17.680, -17.091, -16.454,
    -20.670, -20.144, -19.662, -19.255, -18.897, -18.545, -18.154, -17.684,
    -17.094, -16.457, -20.676, -20.151, -19.668, -19.261, -18.902, -18.550,
    -18.157, -17.688, -17.098, -16.461, -20.681, -20.157, -19.674, -19.267,
    -18.907, -18.554, -18.161, -17.691, -17.102, -16.464, -20.687, -20.163,
    -19.680, -19.272, -18.912, -18.558, -18.165, -17.695, -17.105, -16.468,
    -20.693, -20.169, -19.686, -19.278, -18.917, -18.563, -18.169, -17.699,
    -17.109, -16.471, -20.699, -20.175, -19.692, -19.283, -18.922, -18.567,
    -18.173, -17.702, -17.112, -16.475, -20.705, -20.181, -19.698, -19.289,
    -18.927, -18.572, -18.177, -17.706, -17.116, -16.478, -20.710, -20.187,
    -19.704, -19.295, -18.932, -18.576, -18.181, -17.709, -17.119, -16.482,
    -20.716, -20.193, -19.710, -19.300, -18.937, -18.580, -18.185, -17.713,
    -17.123, -16.485, -20.722, -20.199, -19.715, -19.306, -18.942, -18.585,
    -18.189, -17.717, -17.127, -16.489, -20.728, -20.205, -19.721, -19.311,
    -18.947, -18.589, -18.192, -17.720, -17.130, -16.492, -20.734, -20.211,
    -19.727, -19.317, -18.952, -18.594, -18.196, -17.724, -17.134, -16.495,
    -20.739, -20.217, -19.733, -19.322, -18.957, -18.598, -18.200, -17.728,
    -17.137, -16.499, -20.745, -20.223, -19.739, -19.328, -18.962, -18.602,
    -18.204, -17.731, -17.141, -16.502, -20.751, -20.229, -19.745, -19.334,
    -18.967, -18.607, -18.208, -17.735, -17.144, -16.506, -20.757, -20.235,
    -19.751, -19.339, -18.972, -18.611, -18.212, -17.738, -17.148, -16.509,
    -20.762, -20.241, -19.757, -19.345, -18.977, -18.615, -18.215, -17.742,
    -17.151, -16.512, -20.768, -20.247, -19.763, -19.350, -18.982, -18.620,
    -18.219, -17.745, -17.155, -16.516, -20.774, -20.253, -19.769, -19.356,
    -18.987, -18.624, -18.223, -17.749, -17.158, -16.519, -20.780, -20.259,
    -19.775, -19.362, -18.992, -18.628, -18.227, -17.753, -17.161, -16.522,
    -20.786, -20.265, -19.781, -19.367, -18.997, -18.633, -18.230, -17.756,
    -17.165, -16.526, -20.791, -20.271, -19.787, -19.373, -19.002, -18.637,
    -18.234, -17.760, -17.168, -16.529, -20.797, -20.277, -19.793, -19.378,
    -19.007, -18.641, -18.238, -17.763, -17.172, -16.532, -20.803, -20.283,
    -19.798, -19.384, -19.012, -18.645, -18.242, -17.766, -17.175, -16.536,
    -20.809, -20.289, -19.804, -19.389, -19.016, -18.650, -18.245, -17.770,
    -17.179, -16.539, -20.814, -20.295, -19.810, -19.395, -19.021, -18.654,
    -18.249, -17.773, -17.182, -16.542, -20.820, -20.301, -19.816, -19.400,
    -19.026, -18.658, -18.253, -17.777, -17.185, -16.546, -20.826, -20.307,
    -19.822, -19.406, -19.031, -18.662, -18.256, -17.780, -17.189, -16.549,
    -20.832, -20.313, -19.828, -19.411, -19.036, -18.666, -18.260, -17.784,
    -17.192, -16.552, -20.837, -20.319, -19.834, -19.417, -19.041, -18.670,
    -18.264, -17.787, -17.196, -16.556, -20.843, -20.325, -19.839, -19.422,
    -19.046, -18.675, -18.267, -17.791, -17.199, -16.559, -20.849, -20.331,
    -19.845, -19.428, -19.050, -18.679, -18.271, -17.794, -17.202, -16.562,
    -20.855, -20.337, -19.851, -19.433, -19.055, -18.683, -18.275, -17.797,
    -17.206, -16.565, -20.861, -20.343, -19.857, -19.439, -19.060, -18.687,
    -18.278, -17.801, -17.209, -16.569, -20.866, -20.349, -19.863, -19.444,
    -19.065, -18.691, -18.282, -17.804, -17.212, -16.572, -20.872, -20.355,
    -19.869, -19.449, -19.070, -18.695, -18.286, -17.807, -17.216, -16.575,
    -20.878, -20.360, -19.874, -19.455, -19.075, -18.699, -18.289, -17.811,
    -17.219, -16.578, -20.884, -20.366, -19.880, -19.460, -19.079, -18.704,
    -18.293, -17.814, -17.222, -16.582, -20.889, -20.372, -19.886, -19.466,
    -19.084, -18.708, -18.296, -17.817, -17.226, -16.585, -20.895, -20.378,
    -19.892, -19.471, -19.089, -18.712, -18.300, -17.821, -17.229, -16.588,
    -20.901, -20.384, -19.898, -19.477, -19.094, -18.716, -18.303, -17.824,
    -17.232, -16.591, -20.907, -20.390, -19.903, -19.482, -19.098, -18.720,
    -18.307, -17.827, -17.235, -16.595, -20.912, -20.396, -19.909, -19.488,
    -19.103, -18.724, -18.310, -17.831, -17.239, -16.598, -20.918, -20.402,
    -19.915, -19.493, -19.108, -18.728, -18.314, -17.834, -17.242, -16.601,
    -20.924, -20.408, -19.921, -19.498, -19.113, -18.732, -18.318, -17.837,
    -17.245, -16.604, -20.929, -20.414, -19.927, -19.504, -19.117, -18.736,
    -18.321, -17.841, -17.249, -16.607, -20.935, -20.420, -19.932, -19.509,
    -19.122, -18.740, -18.325, -17.844, -17.252, -16.611, -20.941, -20.426,
    -19.938, -19.514, -19.127, -18.744, -18.328, -17.847, -17.255, -16.614,
    -20.946, -20.432, -19.944, -19.520, -19.132, -18.748, -18.332, -17.850,
    -17.258, -16.617, -20.952, -20.437, -19.950, -19.525, -19.136, -18.752,
    -18.335, -17.854, -17.262, -16.620, -20.958, -20.443, -19.956, -19.530,
    -19.141, -18.756, -18.339, -17.857, -17.265, -16.623, -20.963, -20.449,
    -19.961, -19.536, -19.146, -18.760, -18.342, -17.860, -17.268, -16.626,
    -20.969, -20.455, -19.967, -19.541, -19.150, -18.764, -18.346, -17.863,
    -17.271, -16.630, -20.974, -20.461, -19.973, -19.546, -19.155, -18.768,
    -18.349, -17.867, -17.274, -16.633, -20.980, -20.467, -19.979, -19.552,
    -19.160, -18.772, -18.352, -17.870, -17.278, -16.636, -20.986, -20.473,
    -19.984, -19.557, -19.164, -18.776, -18.356, -17.873, -17.281, -16.639,
    -20.991, -20.479, -19.990, -19.562, -19.169, -18.780, -18.359, -17.876,
    -17.284, -16.642, -20.997, -20.485, -19.996, -19.568, -19.174, -18.784,
    -18.363, -17.879, -17.287, -16.645, -21.002, -20.490, -20.002, -19.573,
    -19.178, -18.788, -18.366, -17.883, -17.290, -16.648, -21.008, -20.496,
    -20.007, -19.578, -19.183, -18.791, -18.370, -17.886, -17.294, -16.651,
    -21.014, -20.502, -20.013, -19.583, -19.187, -18.795, -18.373, -17.889,
    -17.297, -16.654, -21.019, -20.508, -20.019, -19.589, -19.192, -18.799,
    -18.376, -17.892, -17.300, -16.658, -21.025, -20.514, -20.025, -19.594,
    -19.197, -18.803, -18.380, -17.895, -17.303, -16.661, -21.030, -20.520,
    -20.030, -19.599, -19.201, -18.807, -18.383, -17.898, -17.306, -16.664,
    -21.036, -20.526, -20.036, -19.604, -19.206, -18.811, -18.387, -17.902,
    -17.309, -16.667, -21.041, -20.531, -20.042, -19.609, -19.210, -18.815,
    -18.390, -17.905, -17.313, -16.670, -21.047, -20.537, -20.048, -19.615,
    -19.215, -18.818, -18.393, -17.908, -17.316, -16.673, -21.052, -20.543,
    -20.053, -19.620, -19.219, -18.822, -18.397, -17.911, -17.319, -16.676,
    -21.058, -20.549, -20.059, -19.625, -19.224, -18.826, -18.400, -17.914,
    -17.322, -16.679, -21.064, -20.555, -20.065, -19.630, -19.228, -18.830,
    -18.403, -17.917, -17.325, -16.682, -21.069, -20.561, -20.070, -19.635,
    -19.233, -18.834, -18.407, -17.920, -17.328, -16.685, -21.075, -20.566,
    -20.076, -19.641, -19.237, -18.837, -18.410, -17.923, -17.331, -16.688,
    -21.080, -20.572, -20.081, -19.646, -19.242, -18.841, -18.413, -17.927,
    -17.334, -16.691, -21.086, -20.578, -20.087, -19.651, -19.246, -18.845,
    -18.417, -17.930, -17.337, -16.694, -21.091, -20.584, -20.093, -19.656,
    -19.251, -18.849, -18.420, -17.933, -17.340, -16.697, -21.097, -20.590,
    -20.098, -19.661, -19.255, -18.852, -18.424, -17.936, -17.344, -16.700,
    -21.102, -20.595, -20.104, -19.666, -19.259, -18.856, -18.427, -17.939,
    -17.347, -16.703, -21.108, -20.601, -20.109, -19.671, -19.264, -18.860,
    -18.430, -17.942, -17.350, -16.706, -21.113, -20.607, -20.115, -19.676,
    -19.268, -18.864, -18.434, -17.945, -17.353, -16.709, -21.119, -20.613,
    -20.120, -19.681, -19.273, -18.867, -18.437, -17.948, -17.356, -16.712,
    -21.124, -20.618, -20.126, -19.687, -19.277, -18.871, -18.440, -17.951,
    -17.359, -16.715, -21.130, -20.624, -20.131, -19.692, -19.281, -18.875,
    -18.444, -17.954, -17.362, -16.718, -21.136, -20.630, -20.137, -19.697,
    -19.285, -18.878, -18.447, -17.957, -17.365, -16.721, -21.141, -20.636,
    -20.142, -19.702, -19.290, -18.882, -18.450, -17.961, -17.368, -16.724,
    -21.147, -20.642, -20.147, -19.707, -19.294, -18.886, -18.453, -17.964,
    -17.371, -16.727, -21.152, -20.647, -20.153, -19.712, -19.298, -18.889,
    -18.457, -17.967, -17.374, -16.730, -21.158, -20.653, -20.158, -19.717,
    -19.303, -18.893, -18.460, -17.970, -17.377, -16.733, -21.163, -20.659,
    -20.164, -19.722, -19.307, -18.896, -18.463, -17.973, -17.379, -16.736,
    -21.169, -20.664, -20.169, -19.727, -19.311, -18.900, -18.467, -17.976,
    -17.382, -16.739, -21.174, -20.670, -20.175, -19.732, -19.315, -18.904,
    -18.470, -17.979, -17.385, -16.742, -21.180, -20.676, -20.180, -19.737,
    -19.319, -18.907, -18.473, -17.982, -17.388, -16.745, -21.185, -20.682,
    -20.185, -19.742, -19.324, -18.911, -18.476, -17.985, -17.391, -16.748,
    -21.191, -20.687, -20.191, -19.747, -19.328, -18.914, -18.480, -17.988,
    -17.394, -16.751, -21.196, -20.693, -20.196, -19.752, -19.332, -18.918,
    -18.483, -17.991, -17.397, -16.754, -21.201, -20.699, -20.202, -19.757,
    -19.336, -18.922, -18.486, -17.994, -17.400, -16.757, -21.207, -20.704,
    -20.207, -19.762, -19.341, -18.925, -18.490, -17.997, -17.403, -16.760,
    -21.212, -20.710, -20.213, -19.767, -19.345, -18.929, -18.493, -18.000,
    -17.406, -16.763, -21.218, -20.716, -20.218, -19.772, -19.349, -18.933,
    -18.496, -18.003, -17.409, -16.766, -21.223, -20.722, -20.224, -19.777,
    -19.354, -18.937, -18.499, -18.007, -17.412, -16.769, -21.229, -20.727,
    -20.229, -19.782, -19.358, -18.940, -18.503, -18.010, -17.415, -16.773,
    -21.234, -20.733, -20.235, -19.787, -19.362, -18.944, -18.506, -18.013,
    -17.419, -16.776, -21.240, -20.739, -20.240, -19.792, -19.367, -18.948,
    -18.509, -18.016, -17.422, -16.779, -21.245, -20.745, -20.246, -19.797,
    -19.371, -18.951, -18.512, -18.019, -17.425, -16.782, -21.251, -20.750,
    -20.251, -19.802, -19.375, -18.955, -18.516, -18.022, -17.428, -16.785,
    -21.256, -20.756, -20.257, -19.807, -19.380, -18.959, -18.519, -18.025,
    -17.431, -16.788, -21.262, -20.762, -20.262, -19.812, -19.384, -18.963,
    -18.522, -18.028, -17.434, -16.791, -21.267, -20.767, -20.268, -19.817,
    -19.389, -18.966, -18.525, -18.031, -17.437, -16.795, -21.272, -20.773,
    -20.274, -19.822, -19.393, -18.970, -18.529, -18.035, -17.440, -16.798,
    -21.278, -20.779, -20.279, -19.827, -19.397, -18.974, -18.532, -18.038,
    -17.443, -16.801, -21.283, -20.785, -20.285, -19.832, -19.402, -18.977,
    -18.535, -18.041, -17.447, -16.804, -21.289, -20.790, -20.290, -19.837,
    -19.406, -18.981, -18.538, -18.044, -17.450, -16.807, -21.294, -20.796,
    -20.296, -19.842, -19.410, -18.985, -18.542, -18.047, -17.453, -16.810,
    -21.300, -20.802, -20.301, -19.847, -19.415, -18.989, -18.545, -18.050,
    -17.456, -16.814, -21.305, -20.807, -20.307, -19.851, -19.419, -18.992,
    -18.548, -18.053, -17.459, -16.817, -21.310, -20.813, -20.313, -19.856,
    -19.423, -18.996, -18.551, -18.056, -17.462, -16.820, -21.316, -20.819,
    -20.318, -19.861, -19.428, -19.000, -18.554, -18.060, -17.465, -16.823,
    -21.321, -20.825, -20.324, -19.866, -19.432, -19.004, -18.558, -18.063,
    -17.469, -16.827, -21.327, -20.830, -20.329, -19.871, -19.437, -19.007,
    -18.561, -18.066, -17.472, -16.830, -21.332, -20.836, -20.335, -19.876,
    -19.441, -19.011, -18.564, -18.069, -17.475, -16.833, -21.337, -20.842,
    -20.341, -19.881, -19.445, -19.015, -18.567, -18.072, -17.478, -16.836,
    -21.343, -20.847, -20.346, -19.886, -19.450, -19.019, -18.571, -18.075,
    -17.481, -16.839, -21.348, -20.853, -20.352, -19.891, -19.454, -19.022,
    -18.574, -18.078, -17.484, -16.842, -21.354, -20.859, -20.357, -19.896,
    -19.459, -19.026, -18.577, -18.082, -17.488, -16.845, -21.359, -20.864,
    -20.363, -19.901, -19.463, -19.030, -18.581, -18.085, -17.491, -16.849,
    -21.364, -20.870, -20.368, -19.906, -19.467, -19.034, -18.584, -18.088,
    -17.494, -16.852, -21.370, -20.876, -20.374, -19.911, -19.472, -19.038,
    -18.587, -18.091, -17.497, -16.855, -21.375, -20.881, -20.379, -19.916,
    -19.476, -19.041, -18.591, -18.094, -17.500, -16.858, -21.380, -20.887,
    -20.385, -19.921, -19.480, -19.045, -18.594, -18.098, -17.503, -16.861,
    -21.386, -20.893, -20.390, -19.926, -19.485, -19.049, -18.598, -18.101,
    -17.506, -16.864, -21.391, -20.898, -20.396, -19.931, -19.489, -19.053,
    -18.601, -18.104, -17.510, -16.867, -21.397, -20.904, -20.401, -19.936,
    -19.494, -19.057, -18.604, -18.107, -17.513, -16.870, -21.402, -20.910,
    -20.406, -19.941, -19.498, -19.061, -18.608, -18.110, -17.516, -16.873,
    -21.407, -20.915, -20.412, -19.946, -19.502, -19.064, -18.611, -18.114,
    -17.519, -16.876, -21.413, -20.921, -20.417, -19.951, -19.507, -19.068,
    -18.615, -18.117, -17.522, -16.879, -21.418, -20.926, -20.423, -19.956,
    -19.511, -19.072, -18.618, -18.120, -17.525, -16.883, -21.423, -20.932,
    -20.428, -19.961, -19.516, -19.076, -18.621, -18.123, -17.529, -16.886,
    -21.429, -20.938, -20.434, -19.967, -19.520, -19.080, -18.625, -18.126,
    -17.532, -16.889, -21.434, -20.943, -20.439, -19.972, -19.525, -19.084,
    -18.628, -18.130, -17.535, -16.892, -21.439, -20.949, -20.445, -19.977,
    -19.529, -19.088, -18.632, -18.133, -17.538, -16.895, -21.445, -20.955,
    -20.450, -19.982, -19.533, -19.091, -18.635, -18.136, -17.541, -16.898,
    -21.450, -20.960, -20.456, -19.987, -19.538, -19.095, -18.639, -18.139,
    -17.544, -16.901, -21.455, -20.966, -20.461, -19.992, -19.542, -19.099,
    -18.642, -18.143, -17.548, -16.904, -21.461, -20.971, -20.467, -19.997,
    -19.547, -19.103, -18.646, -18.146, -17.551, -16.907, -21.466, -20.977,
    -20.472, -20.002, -19.551, -19.107, -18.649, -18.149, -17.554, -16.910,
    -21.471, -20.983, -20.477, -20.007, -19.555, -19.111, -18.652, -18.152,
    -17.557, -16.913, -21.477, -20.988, -20.483, -20.012, -19.560, -19.115,
    -18.656, -18.155, -17.560, -16.916, -21.482, -20.994, -20.488, -20.017,
    -19.564, -19.119, -18.659, -18.159, -17.564, -16.920, -21.487, -20.999,
    -20.494, -20.022, -19.569, -19.122, -18.663, -18.162, -17.567, -16.923,
    -21.493, -21.005, -20.499, -20.027, -19.573, -19.126, -18.666, -18.165,
    -17.570, -16.926, -21.498, -21.011, -20.505, -20.032, -19.578, -19.130,
    -18.670, -18.168, -17.573, -16.929, -21.503, -21.016, -20.510, -20.037,
    -19.582, -19.134, -18.673, -18.172, -17.577, -16.932, -21.509, -21.022,
    -20.516, -20.043, -19.587, -19.138, -18.677, -18.175, -17.580, -16.935,
    -21.514, -21.027, -20.521, -20.048, -19.591, -19.142, -18.680, -18.178,
    -17.583, -16.939, -21.519, -21.033, -20.527, -20.053, -19.596, -19.146,
    -18.684, -18.182, -17.586, -16.942, -21.525, -21.038, -20.532, -20.058,
    -19.600, -19.150, -18.687, -18.185, -17.590, -16.945, -21.530, -21.044,
    -20.537, -20.063, -19.605, -19.153, -18.691, -18.188, -17.593, -16.948,
    -21.535, -21.050, -20.543, -20.068, -19.609, -19.157, -18.694, -18.191,
    -17.596, -16.952, -21.541, -21.055, -20.548, -20.073, -19.613, -19.161,
    -18.697, -18.195, -17.600, -16.955, -21.546, -21.061, -20.554, -20.078,
    -19.618, -19.165, -18.701, -18.198, -17.603, -16.958, -21.551, -21.066,
    -20.559, -20.083, -19.622, -19.169, -18.704, -18.201, -17.606, -16.961,
    -21.557, -21.072, -20.565, -20.088, -19.627, -19.173, -18.708, -18.205,
    -17.609, -16.965, -21.562, -21.077, -20.570, -20.093, -19.631, -19.177,
    -18.711, -18.208, -17.613, -16.968, -21.567, -21.083, -20.575, -20.098,
    -19.636, -19.181, -18.715, -18.211, -17.616, -16.971, -21.572, -21.088,
    -20.581, -20.104, -19.640, -19.185, -18.718, -18.214, -17.619, -16.974,
    -21.578, -21.094, -20.586, -20.109, -19.645, -19.188, -18.722, -18.218,
    -17.623, -16.978, -21.583, -21.099, -20.592, -20.114, -19.649, -19.192,
    -18.725, -18.221, -17.626, -16.981, -21.588, -21.105, -20.597, -20.119,
    -19.654, -19.196, -18.729, -18.224, -17.629, -16.984, -21.594, -21.110,
    -20.603, -20.124, -19.658, -19.200, -18.732, -18.228, -17.633, -16.988,
    -21.599, -21.116, -20.608, -20.129, -19.663, -19.204, -18.736, -18.231,
    -17.636, -16.991, -21.604, -21.121, -20.613, -20.134, -19.668, -19.208,
    -18.740, -18.234, -17.639, -16.994, -21.609, -21.127, -20.619, -20.139,
    -19.672, -19.212, -18.743, -18.238, -17.643, -16.998, -21.615, -21.132,
    -20.624, -20.144, -19.677, -19.216, -18.747, -18.241, -17.646, -17.001,
    -21.620, -21.138, -20.630, -20.149, -19.681, -19.220, -18.750, -18.244,
    -17.649, -17.004, -21.625, -21.143, -20.635, -20.154, -19.686, -19.224,
    -18.754, -18.248, -17.653, -17.007, -21.630, -21.149, -20.641, -20.159,
    -19.690, -19.228, -18.757, -18.251, -17.656, -17.011, -21.636, -21.154,
    -20.646, -20.165, -19.695, -19.232, -18.761, -18.254, -17.659, -17.014,
    -21.641, -21.160, -20.652, -20.170, -19.699, -19.236, -18.764, -18.258,
    -17.663, -17.017, -21.646, -21.165, -20.657, -20.175, -19.704, -19.240,
    -18.768, -18.261, -17.666, -17.020, -21.651, -21.171, -20.662, -20.180,
    -19.709, -19.244, -18.772, -18.265, -17.669, -17.024, -21.656, -21.176,
    -20.668, -20.185, -19.713, -19.248, -18.775, -18.268, -17.673, -17.027,
    -21.661, -21.182, -20.673, -20.190, -19.718, -19.252, -18.779, -18.271,
    -17.676, -17.030, -21.667, -21.187, -20.679, -20.195, -19.722, -19.256,
    -18.782, -18.275, -17.679, -17.034, -21.672, -21.193, -20.684, -20.200,
    -19.727, -19.260, -18.786, -18.278, -17.683, -17.037, -21.677, -21.198,
    -20.690, -20.205, -19.732, -19.264, -18.790, -18.282, -17.686, -17.040,
    -21.682, -21.203, -20.695, -20.210, -19.736, -19.268, -18.793, -18.285,
    -17.690, -17.043, -21.687, -21.209, -20.700, -20.215, -19.741, -19.272,
    -18.797, -18.288, -17.693, -17.047, -21.692, -21.214, -20.706, -20.220,
    -19.745, -19.276, -18.800, -18.292, -17.696, -17.050, -21.697, -21.220,
    -20.711, -20.226, -19.750, -19.280, -18.804, -18.295, -17.700, -17.053,
    -21.703, -21.225, -20.717, -20.231, -19.755, -19.284, -18.808, -18.299,
    -17.703, -17.057, -21.708, -21.230, -20.722, -20.236, -19.759, -19.289,
    -18.811, -18.302, -17.706, -17.060, -21.713, -21.236, -20.728, -20.241,
    -19.764, -19.293, -18.815, -18.306, -17.710, -17.063, -21.718, -21.241,
    -20.733, -20.246, -19.769, -19.297, -18.819, -18.309, -17.713, -17.066,
    -21.723, -21.247, -20.739, -20.251, -19.773, -19.301, -18.822, -18.313,
    -17.717, -17.070, -21.728, -21.252, -20.744, -20.256, -19.778, -19.305,
    -18.826, -18.316, -17.720, -17.073, -21.733, -21.257, -20.749, -20.261,
    -19.783, -19.309, -18.830, -18.319, -17.723, -17.076, -21.738, -21.263,
    -20.755, -20.266, -19.787, -19.313, -18.833, -18.323, -17.727, -17.080,
    -21.743, -21.268, -20.760, -20.271, -19.792, -19.317, -18.837, -18.326,
    -17.730, -17.083, -21.748, -21.274, -20.766, -20.277, -19.797, -19.322,
    -18.841, -18.330, -17.734, -17.086, -21.754, -21.279, -20.771, -20.282,
    -19.801, -19.326, -18.845, -18.333, -17.737, -17.090, -21.759, -21.284,
    -20.776, -20.287, -19.806, -19.330, -18.848, -18.337, -17.740, -17.093,
    -21.764, -21.290, -20.782, -20.292, -19.811, -19.334, -18.852, -18.340,
    -17.744, -17.096, -21.769, -21.295, -20.787, -20.297, -19.815, -19.338,
    -18.856, -18.344, -17.747, -17.099, -21.774, -21.301, -20.793, -20.302,
    -19.820, -19.343, -18.860, -18.347, -17.750, -17.103, -21.779, -21.306,
    -20.798, -20.308, -19.825, -19.347, -18.863, -18.351, -17.754, -17.106,
    -21.784, -21.312, -20.804, -20.313, -19.830, -19.351, -18.867, -18.354,
    -17.757, -17.109, -21.789, -21.317, -20.809, -20.318, -19.834, -19.355,
    -18.871, -18.358, -17.761, -17.113, -21.794, -21.322, -20.814, -20.323,
    -19.839, -19.359, -18.875, -18.361, -17.764, -17.116, -21.799, -21.328,
    -20.820, -20.328, -19.844, -19.364, -18.879, -18.364, -17.767, -17.119,
    -21.804, -21.333, -20.825, -20.334, -19.848, -19.368, -18.882, -18.368,
    -17.771, -17.123, -21.810, -21.339, -20.831, -20.339, -19.853, -19.372,
    -18.886, -18.371, -17.774, -17.126, -21.815, -21.344, -20.836, -20.344,
    -19.858, -19.377, -18.890, -18.375, -17.778, -17.129, -21.820, -21.349,
    -20.841, -20.349, -19.863, -19.381, -18.894, -18.378, -17.781, -17.133,
    -21.825, -21.355, -20.847, -20.355, -19.867, -19.385, -18.898, -18.382,
    -17.785, -17.136, -21.830, -21.360, -20.852, -20.360, -19.872, -19.389,
    -18.902, -18.385, -17.788, -17.139, -21.835, -21.365, -20.857, -20.365,
    -19.877, -19.394, -18.905, -18.389, -17.791, -17.143, -21.840, -21.371,
    -20.863, -20.370, -19.882, -19.398, -18.909, -18.392, -17.795, -17.146,
    -21.845, -21.376, -20.868, -20.375, -19.886, -19.402, -18.913, -18.396,
    -17.798, -17.149, -21.850, -21.382, -20.874, -20.381, -19.891, -19.407,
    -18.917, -18.399, -17.802, -17.153, -21.855, -21.387, -20.879, -20.386,
    -19.896, -19.411, -18.921, -18.403, -17.805, -17.156, -21.860, -21.392,
    -20.884, -20.391, -19.901, -19.415, -18.925, -18.406, -17.808, -17.159,
    -21.865, -21.398, -20.890, -20.396, -19.906, -19.420, -18.929, -18.410,
    -17.812, -17.163, -21.870, -21.403, -20.895, -20.402, -19.910, -19.424,
    -18.933, -18.414, -17.815, -17.166, -21.875, -21.408, -20.901, -20.407,
    -19.915, -19.428, -18.936, -18.417, -17.819, -17.169, -21.880, -21.414,
    -20.906, -20.412, -19.920, -19.433, -18.940, -18.421, -17.822, -17.172,
    -21.885, -21.419, -20.911, -20.417, -19.925, -19.437, -18.944, -18.424,
    -17.825, -17.175, -21.890, -21.424, -20.917, -20.422, -19.930, -19.441,
    -18.948, -18.428, -17.829, -17.179, -21.895, -21.430, -20.922, -20.428,
    -19.935, -19.446, -18.952, -18.431, -17.832, -17.182, -21.900, -21.435,
    -20.927, -20.433, -19.939, -19.450, -18.956, -18.435, -17.835, -17.185,
    -21.906, -21.440, -20.933, -20.438, -19.944, -19.454, -18.960, -18.438,
    -17.839, -17.188, -21.911, -21.446, -20.938, -20.443, -19.949, -19.459,
    -18.963, -18.442, -17.842, -17.192, -21.916, -21.451, -20.943, -20.448,
    -19.954, -19.463, -18.967, -18.445, -17.845, -17.195, -21.921, -21.456,
    -20.949, -20.454, -19.959, -19.467, -18.971, -18.449, -17.849, -17.198,
    -21.926, -21.462, -20.954, -20.459, -19.964, -19.472, -18.975, -18.452,
    -17.852, -17.201, -21.931, -21.467, -20.960, -20.464, -19.969, -19.476,
    -18.979, -18.456, -17.856, -17.204, -21.936, -21.472, -20.965, -20.469,
    -19.974, -19.480, -18.983, -18.459, -17.859, -17.208, -21.941, -21.478,
    -20.970, -20.474, -19.979, -19.485, -18.987, -18.463, -17.862, -17.211,
    -21.946, -21.483, -20.976, -20.480, -19.983, -19.489, -18.991, -18.466,
    -17.866, -17.214, -21.951, -21.488, -20.981, -20.485, -19.988, -19.494,
    -18.995, -18.470, -17.869, -17.217, -21.956, -21.494, -20.986, -20.490,
    -19.993, -19.498, -18.998, -18.473, -17.872, -17.220, -21.961, -21.499,
    -20.992, -20.495, -19.998, -19.502, -19.002, -18.477, -17.876, -17.223,
    -21.966, -21.504, -20.997, -20.500, -20.003, -19.507, -19.006, -18.480,
    -17.879, -17.227, -21.971, -21.509, -21.002, -20.506, -20.008, -19.511,
    -19.010, -18.484, -17.882, -17.230, -21.976, -21.515, -21.008, -20.511,
    -20.013, -19.516, -19.014, -18.487, -17.886, -17.233, -21.981, -21.520,
    -21.013, -20.516, -20.018, -19.520, -19.018, -18.491, -17.889, -17.236,
    -21.986, -21.525, -21.018, -20.521, -20.023, -19.524, -19.022, -18.495,
    -17.892, -17.239, -21.991, -21.531, -21.024, -20.526, -20.028, -19.529,
    -19.026, -18.498, -17.896, -17.242, -21.996, -21.536, -21.029, -20.532,
    -20.033, -19.533, -19.030, -18.502, -17.899, -17.245, -22.001, -21.541,
    -21.034, -20.537, -20.038, -19.538, -19.034, -18.505, -17.902, -17.249,
    -22.006, -21.546, -21.040, -20.542, -20.043, -19.542, -19.038, -18.509,
    -17.906, -17.252, -22.011, -21.552, -21.045, -20.547, -20.047, -19.547,
    -19.042, -18.512, -17.909, -17.255, -22.016, -21.557, -21.051, -20.553,
    -20.052, -19.551, -19.046, -18.516, -17.912, -17.258, -22.021, -21.562,
    -21.056, -20.558, -20.057, -19.556, -19.050, -18.519, -17.915, -17.261,
    -22.026, -21.567, -21.061, -20.563, -20.062, -19.560, -19.054, -18.523,
    -17.919, -17.264, -22.031, -21.573, -21.067, -20.568, -20.067, -19.564,
    -19.058, -18.526, -17.922, -17.267, -22.036, -21.578, -21.072, -20.574,
    -20.072, -19.569, -19.062, -18.530, -17.925, -17.271, -22.041, -21.583,
    -21.077, -20.579, -20.077, -19.573, -19.066, -18.534, -17.929, -17.274,
    -22.046, -21.588, -21.083, -20.584, -20.082, -19.578, -19.070, -18.537,
    -17.932, -17.277, -22.051, -21.594, -21.088, -20.589, -20.087, -19.582,
    -19.074, -18.541, -17.935, -17.280, -22.056, -21.599, -21.093, -20.594,
    -20.092, -19.587, -19.078, -18.544, -17.938, -17.283, -22.061, -21.604,
    -21.099, -20.600, -20.097, -19.591, -19.082, -18.548, -17.942, -17.286,
    -22.066, -21.609, -21.104, -20.605, -20.102, -19.596, -19.086, -18.551,
    -17.945, -17.289, -22.071, -21.615, -21.109, -20.610, -20.106, -19.600,
    -19.090, -18.555, -17.948, -17.292, -22.076, -21.620, -21.115, -20.615,
    -20.111, -19.605, -19.094, -18.559, -17.952, -17.295, -22.081, -21.625,
    -21.120, -20.621, -20.116, -19.609, -19.098, -18.562, -17.955, -17.299,
    -22.086, -21.630, -21.126, -20.626, -20.121, -19.614, -19.102, -18.566,
    -17.958, -17.302, -22.091, -21.635, -21.131, -20.631, -20.126, -19.618,
    -19.106, -18.569, -17.961, -17.305, -22.096, -21.641, -21.136, -20.636,
    -20.131, -19.623, -19.110, -18.573, -17.965, -17.308, -22.101, -21.646,
    -21.142, -20.642, -20.136, -19.627, -19.114, -18.576, -17.968, -17.311,
    -22.106, -21.651, -21.147, -20.647, -20.141, -19.632, -19.118, -18.580,
    -17.971, -17.314, -22.111, -21.656, -21.152, -20.652, -20.146, -19.637,
    -19.122, -18.584, -17.974, -17.317, -22.116, -21.661, -21.158, -20.657,
    -20.151, -19.641, -19.126, -18.587, -17.977, -17.320, -22.121, -21.667,
    -21.163, -20.663, -20.156, -19.646, -19.130, -18.591, -17.981, -17.323,
    -22.126, -21.672, -21.168, -20.668, -20.161, -19.650, -19.134, -18.594,
    -17.984, -17.326, -22.131, -21.677, -21.174, -20.673, -20.166, -19.655,
    -19.138, -18.598, -17.987, -17.329, -22.136, -21.682, -21.179, -20.678,
    -20.171, -19.659, -19.142, -18.601, -17.990, -17.332, -22.140, -21.687,
    -21.184, -20.684, -20.176, -19.664, -19.146, -18.605, -17.993, -17.335,
    -22.145, -21.692, -21.189, -20.689, -20.181, -19.668, -19.150, -18.608,
    -17.997, -17.338, -22.150, -21.697, -21.195, -20.694, -20.186, -19.673,
    -19.154, -18.612, -18.000, -17.341, -22.155, -21.703, -21.200, -20.699,
    -20.190, -19.677, -19.158, -18.616, -18.003, -17.344, -22.160, -21.708,
    -21.205, -20.704, -20.195, -19.682, -19.162, -18.619, -18.006, -17.347,
    -22.165, -21.713, -21.211, -20.710, -20.200, -19.686, -19.166, -18.623,
    -18.009, -17.350, -22.170, -21.718, -21.216, -20.715, -20.205, -19.691,
    -19.170, -18.626, -18.012, -17.353, -22.175, -21.723, -21.221, -20.720,
    -20.210, -19.696, -19.174, -18.630, -18.016, -17.356, -22.180, -21.728,
    -21.226, -20.725, -20.215, -19.700, -19.178, -18.633, -18.019, -17.359,
    -22.184, -21.733, -21.232, -20.730, -20.220, -19.705, -19.183, -18.637,
    -18.022, -17.362, -22.189, -21.738, -21.237, -20.736, -20.225, -19.709,
    -19.187, -18.641, -18.025, -17.365, -22.194, -21.743, -21.242, -20.741,
    -20.230, -19.714, -19.191, -18.644, -18.028, -17.368, -22.199, -21.749,
    -21.248, -20.746, -20.235, -19.718, -19.195, -18.648, -18.031, -17.371,
    -22.204, -21.754, -21.253, -20.751, -20.240, -19.723, -19.199, -18.651,
    -18.034, -17.374, -22.209, -21.759, -21.258, -20.756, -20.245, -19.728,
    -19.203, -18.655, -18.038, -17.377, -22.214, -21.764, -21.263, -20.761,
    -20.250, -19.732, -19.207, -18.658, -18.041, -17.380, -22.218, -21.769,
    -21.269, -20.767, -20.255, -19.737, -19.211, -18.662, -18.044, -17.383,
    -22.223, -21.774, -21.274, -20.772, -20.260, -19.741, -19.215, -18.665,
    -18.047, -17.386, -22.228, -21.779, -21.279, -20.777, -20.265, -19.746,
    -19.219, -18.669, -18.050, -17.389, -22.233, -21.784, -21.284, -20.782,
    -20.270, -19.751, -19.223, -18.673, -18.053, -17.392, -22.238, -21.789,
    -21.289, -20.787, -20.275, -19.755, -19.227, -18.676, -18.056, -17.395,
    -22.242, -21.794, -21.295, -20.793, -20.280, -19.760, -19.231, -18.680,
    -18.059, -17.398, -22.247, -21.799, -21.300, -20.798, -20.285, -19.764,
    -19.235, -18.683, -18.062, -17.401, -22.252, -21.804, -21.305, -20.803,
    -20.290, -19.769, -19.239, -18.687, -18.066, -17.404, -22.257, -21.810,
    -21.310, -20.808, -20.295, -19.773, -19.243, -18.690, -18.069, -17.407,
    -22.262, -21.815, -21.316, -20.813, -20.300, -19.778, -19.247, -18.694,
    -18.072, -17.409, -22.266, -21.820, -21.321, -20.819, -20.305, -19.782,
    -19.251, -18.697, -18.075, -17.412, -22.271, -21.825, -21.326, -20.824,
    -20.310, -19.787, -19.255, -18.701, -18.078, -17.415, -22.276, -21.830,
    -21.331, -20.829, -20.315, -19.792, -19.259, -18.704, -18.081, -17.418,
    -22.281, -21.835, -21.336, -20.834, -20.320, -19.796, -19.263, -18.708,
    -18.084, -17.421, -22.285, -21.840, -21.342, -20.839, -20.325, -19.801,
    -19.267, -18.711, -18.087, -17.424, -22.290, -21.845, -21.347, -20.845,
    -20.330, -19.805, -19.272, -18.714, -18.090, -17.427, -22.295, -21.850,
    -21.352, -20.850, -20.335, -19.810, -19.276, -18.718, -18.093, -17.430,
    -22.300, -21.855, -21.357, -20.855, -20.340, -19.814, -19.280, -18.721,
    -18.096, -17.433, -22.305, -21.860, -21.362, -20.860, -20.345, -19.819,
    -19.284, -18.725, -18.100, -17.435, -22.309, -21.865, -21.368, -20.865,
    -20.350, -19.824, -19.288, -18.728, -18.103, -17.438, -22.314, -21.871,
    -21.373, -20.871, -20.355, -19.828, -19.292, -18.732, -18.106, -17.441,
    -22.319, -21.876, -21.378, -20.876, -20.360, -19.833, -19.296, -18.735,
    -18.109, -17.444, -22.323, -21.881, -21.383, -20.881, -20.365, -19.837,
    -19.300, -18.739, -18.112, -17.447, -22.328, -21.886, -21.388, -20.886,
    -20.370, -19.842, -19.304, -18.742, -18.115, -17.450, -22.333, -21.891,
    -21.394, -20.891, -20.375, -19.846, -19.308, -18.746, -18.118, -17.453,
    -22.338, -21.896, -21.399, -20.897, -20.380, -19.851, -19.312, -18.749,
    -18.121, -17.455, -22.342, -21.901, -21.404, -20.902, -20.385, -19.855,
    -19.316, -18.753, -18.124, -17.458, -22.347, -21.906, -21.409, -20.907,
    -20.390, -19.860, -19.320, -18.756, -18.127, -17.461, -22.352, -21.911,
    -21.414, -20.912, -20.395, -19.865, -19.324, -18.759, -18.130, -17.464,
    -22.357, -21.916, -21.419, -20.917, -20.400, -19.869, -19.328, -18.763,
    -18.133, -17.467, -22.361, -21.921, -21.425, -20.923, -20.405, -19.874,
    -19.332, -18.766, -18.136, -17.469, -22.366, -21.926, -21.430, -20.928,
    -20.410, -19.878, -19.336, -18.770, -18.139, -17.472, -22.371, -21.931,
    -21.435, -20.933, -20.415, -19.883, -19.340, -18.773, -18.142, -17.475,
    -22.376, -21.936, -21.440, -20.938, -20.420, -19.888, -19.344, -18.777,
    -18.145, -17.478, -22.380, -21.941, -21.445, -20.943, -20.425, -19.892,
    -19.348, -18.780, -18.148, -17.481, -22.385, -21.946, -21.451, -20.948,
    -20.430, -19.897, -19.352, -18.783, -18.151, -17.483, -22.390, -21.951,
    -21.456, -20.954, -20.435, -19.901, -19.356, -18.787, -18.154, -17.486,
    -22.395, -21.956, -21.461, -20.959, -20.440, -19.906, -19.360, -18.790,
    -18.157, -17.489, -22.399, -21.961, -21.466, -20.964, -20.445, -19.910,
    -19.364, -18.794, -18.160, -17.492, -22.404, -21.966, -21.471, -20.969,
    -20.450, -19.915, -19.368, -18.797, -18.163, -17.494, -22.409, -21.971,
    -21.477, -20.974, -20.455, -19.920, -19.372, -18.800, -18.166, -17.497,
    -22.414, -21.976, -21.482, -20.979, -20.460, -19.924, -19.376, -18.804,
    -18.169, -17.500, -22.418, -21.981, -21.487, -20.985, -20.465, -19.929,
    -19.380, -18.807, -18.172, -17.503, -22.423, -21.986, -21.492, -20.990,
    -20.470, -19.933, -19.384, -18.811, -18.175, -17.506, -22.428, -21.991,
    -21.497, -20.995, -20.475, -19.938, -19.388, -18.814, -18.178, -17.508,
    -22.433, -21.996, -21.503, -21.000, -20.480, -19.943, -19.392, -18.817,
    -18.181, -17.511, -22.437, -22.001, -21.508, -21.005, -20.485, -19.947,
    -19.396, -18.821, -18.184, -17.514, -22.442, -22.006, -21.513, -21.010,
    -20.490, -19.952, -19.400, -18.824, -18.186, -17.516, -22.447, -22.011,
    -21.518, -21.015, -20.495, -19.957, -19.404, -18.828, -18.189, -17.519,
    -22.452, -22.016, -21.523, -21.021, -20.500, -19.961, -19.408, -18.831,
    -18.192, -17.522, -22.456, -22.021, -21.529, -21.026, -20.505, -19.966,
    -19.412, -18.834, -18.195, -17.525, -22.461, -22.026, -21.534, -21.031,
    -20.510, -19.970, -19.416, -18.838, -18.198, -17.527, -22.466, -22.031,
    -21.539, -21.036, -20.515, -19.975, -19.420, -18.841, -18.201, -17.530,
    -22.471, -22.036, -21.544, -21.041, -20.520, -19.980, -19.424, -18.844,
    -18.204, -17.533, -22.475, -22.041, -21.549, -21.046, -20.525, -19.984,
    -19.428, -18.848, -18.207, -17.535, -22.480, -22.046, -21.554, -21.051,
    -20.530, -19.989, -19.432, -18.851, -18.210, -17.538, -22.485, -22.051,
    -21.560, -21.057, -20.535, -19.993, -19.436, -18.854, -18.212, -17.541,
    -22.490, -22.056, -21.565, -21.062, -20.540, -19.998, -19.440, -18.857,
    -18.215, -17.543, -22.494, -22.061, -21.570, -21.067, -20.545, -20.002,
    -19.444, -18.861, -18.218, -17.546, -22.499, -22.066, -21.575, -21.072,
    -20.550, -20.007, -19.448, -18.864, -18.221, -17.549, -22.504, -22.071,
    -21.580, -21.077, -20.555, -20.012, -19.452, -18.867, -18.224, -17.551,
    -22.509, -22.076, -21.585, -21.082, -20.560, -20.016, -19.456, -18.870,
    -18.226, -17.554, -22.513, -22.081, -21.590, -21.087, -20.565, -20.021,
    -19.459, -18.874, -18.229, -17.556, -22.518, -22.086, -21.595, -21.092,
    -20.570, -20.025, -19.463, -18.877, -18.232, -17.559, -22.523, -22.091,
    -21.600, -21.098, -20.575, -20.030, -19.467, -18.880, -18.235, -17.562,
    -22.527, -22.096, -21.605, -21.103, -20.580, -20.034, -19.471, -18.883,
    -18.238, -17.564, -22.532, -22.101, -21.610, -21.108, -20.585, -20.039,
    -19.475, -18.886, -18.240, -17.567, -22.537, -22.106, -21.616, -21.113,
    -20.590, -20.044, -19.479, -18.890, -18.243, -17.569, -22.542, -22.110,
    -21.621, -21.118, -20.594, -20.048, -19.483, -18.893, -18.246, -17.572,
    -22.546, -22.115, -21.626, -21.123, -20.599, -20.053, -19.487, -18.896,
    -18.248, -17.575, -22.551, -22.120, -21.631, -21.128, -20.604, -20.057,
    -19.491, -18.899, -18.251, -17.577, -22.556, -22.125, -21.636, -21.133,
    -20.609, -20.062, -19.495, -18.902, -18.254, -17.580, -22.560, -22.130,
    -21.641, -21.138, -20.614, -20.066, -19.499, -18.905, -18.257, -17.582,
    -22.565, -22.135, -21.646, -21.144, -20.619, -20.071, -19.502, -18.909,
    -18.259, -17.585, -22.570, -22.140, -21.651, -21.149, -20.624, -20.075,
    -19.506, -18.912, -18.262, -17.587, -22.575, -22.145, -21.656, -21.154,
    -20.629, -20.080, -19.510, -18.915, -18.265, -17.590, -22.579, -22.150,
    -21.661, -21.159, -20.634, -20.084, -19.514, -18.918, -18.267, -17.592,
    -22.584, -22.155, -21.666, -21.164, -20.639, -20.089, -19.518, -18.921,
    -18.270, -17.595, -22.589, -22.160, -21.671, -21.169, -20.644, -20.094,
    -19.522, -18.924, -18.273, -17.598, -22.593, -22.165, -21.676, -21.174,
    -20.649, -20.098, -19.526, -18.927, -18.275, -17.600, -22.598, -22.170,
    -21.681, -21.179, -20.654, -20.103, -19.530, -18.930, -18.278, -17.603,
    -22.603, -22.175, -21.686, -21.184, -20.659, -20.107, -19.533, -18.933,
    -18.281, -17.605, -22.607, -22.180, -21.691, -21.189, -20.664, -20.112,
    -19.537, -18.936, -18.283, -17.608, -22.612, -22.185, -21.696, -21.195,
    -20.669, -20.116, -19.541, -18.939, -18.286, -17.610, -22.617, -22.190,
    -21.701, -21.200, -20.674, -20.121, -19.545, -18.943, -18.289, -17.613,
    -22.622, -22.194, -21.706, -21.205, -20.678, -20.125, -19.549, -18.946,
    -18.291, -17.615, -22.626, -22.199, -21.711, -21.210, -20.683, -20.130,
    -19.553, -18.949, -18.294, -17.618, -22.631, -22.204, -21.717, -21.215,
    -20.688, -20.134, -19.556, -18.952, -18.296, -17.620, -22.636, -22.209,
    -21.722, -21.220, -20.693, -20.139, -19.560, -18.955, -18.299, -17.623,
    -22.640, -22.214, -21.727, -21.225, -20.698, -20.143, -19.564, -18.958,
    -18.302, -17.625, -22.645, -22.219, -21.732, -21.230, -20.703, -20.148,
    -19.568, -18.961, -18.304, -17.628, -22.650, -22.224, -21.737, -21.235,
    -20.708, -20.152, -19.572, -18.964, -18.307, -17.631, -22.654, -22.229,
    -21.742, -21.240, -20.713, -20.157, -19.575, -18.967, -18.309, -17.633,
    -22.659, -22.234, -21.747, -21.245, -20.718, -20.161, -19.579, -18.970,
    -18.312, -17.636, -22.664, -22.239, -21.752, -21.250, -20.723, -20.166,
    -19.583, -18.973, -18.315, -17.638, -22.668, -22.244, -21.757, -21.256,
    -20.728, -20.170, -19.587, -18.976, -18.317, -17.641, -22.673, -22.249,
    -21.762, -21.261, -20.733, -20.175, -19.590, -18.978, -18.320, -17.643,
    -22.678, -22.253, -21.767, -21.266, -20.737, -20.179, -19.594, -18.981,
    -18.322, -17.646, -22.682, -22.258, -21.772, -21.271, -20.742, -20.184,
    -19.598, -18.984, -18.325, -17.648, -22.687, -22.263, -21.777, -21.276,
    -20.747, -20.188, -19.602, -18.987, -18.327, -17.651, -22.692, -22.268,
    -21.782, -21.281, -20.752, -20.193, -19.606, -18.990, -18.330, -17.653,
    -22.696, -22.273, -21.787, -21.286, -20.757, -20.198, -19.609, -18.993,
    -18.332, -17.656, -22.701, -22.278, -21.792, -21.291, -20.762, -20.202,
    -19.613, -18.996, -18.335, -17.658, -22.706, -22.283, -21.797, -21.296,
    -20.767, -20.206, -19.617, -18.999, -18.337, -17.660, -22.710, -22.288,
    -21.802, -21.301, -20.772, -20.211, -19.620, -19.002, -18.340, -17.663,
    -22.715, -22.293, -21.807, -21.306, -20.777, -20.215, -19.624, -19.005,
    -18.342, -17.665, -22.719, -22.297, -21.812, -21.311, -20.781, -20.220,
    -19.628, -19.007, -18.345, -17.668, -22.724, -22.302, -21.817, -21.316,
    -20.786, -20.224, -19.631, -19.010, -18.347, -17.670, -22.728, -22.307,
    -21.822, -21.321, -20.791, -20.229, -19.635, -19.013, -18.350, -17.673,
    -22.733, -22.312, -21.827, -21.326, -20.796, -20.233, -19.639, -19.016,
    -18.352, -17.675, -22.738, -22.317, -21.832, -21.331, -20.801, -20.237,
    -19.642, -19.019, -18.355, -17.677, -22.742, -22.322, -21.837, -21.336,
    -20.806, -20.242, -19.646, -19.021, -18.357, -17.680, -22.747, -22.326,
    -21.842, -21.342, -20.811, -20.246, -19.650, -19.024, -18.359, -17.682,
    -22.751, -22.331, -21.847, -21.347, -20.815, -20.250, -19.653, -19.027,
    -18.362, -17.684, -22.756, -22.336, -21.852, -21.352, -20.820, -20.255,
    -19.657, -19.030, -18.364, -17.687, -22.760, -22.341, -21.857, -21.357,
    -20.825, -20.259, -19.660, -19.032, -18.366, -17.689, -22.765, -22.346,
    -21.862, -21.362, -20.830, -20.264, -19.664, -19.035, -18.369, -17.692,
    -22.769, -22.350, -21.867, -21.367, -20.835, -20.268, -19.668, -19.038,
    -18.371, -17.694, -22.774, -22.355, -21.872, -21.372, -20.840, -20.272,
    -19.671, -19.040, -18.373, -17.696, -22.778, -22.360, -21.877, -21.377,
    -20.844, -20.277, -19.675, -19.043, -18.376, -17.699, -22.783, -22.365,
    -21.882, -21.382, -20.849, -20.281, -19.678, -19.046, -18.378, -17.701,
    -22.787, -22.369, -21.887, -21.387, -20.854, -20.285, -19.682, -19.048,
    -18.380, -17.703, -22.792, -22.374, -21.892, -21.392, -20.859, -20.290,
    -19.685, -19.051, -18.383, -17.706, -22.796, -22.379, -21.897, -21.397,
    -20.864, -20.294, -19.689, -19.054, -18.385, -17.708, -22.801, -22.384,
    -21.902, -21.402, -20.869, -20.298, -19.692, -19.056, -18.387, -17.710,
    -22.805, -22.389, -21.907, -21.407, -20.873, -20.302, -19.696, -19.059,
    -18.389, -17.712, -22.810, -22.393, -21.912, -21.412, -20.878, -20.307,
    -19.699, -19.061, -18.392, -17.715, -22.814, -22.398, -21.917, -21.417,
    -20.883, -20.311, -19.703, -19.064, -18.394, -17.717, -22.818, -22.403,
    -21.922, -21.422, -20.888, -20.315, -19.706, -19.067, -18.396, -17.719,
    -22.823, -22.408, -21.927, -21.427, -20.893, -20.320, -19.710, -19.069,
    -18.398, -17.722, -22.827, -22.412, -21.932, -21.432, -20.897, -20.324,
    -19.713, -19.072, -18.401, -17.724, -22.832, -22.417, -21.937, -21.437,
    -20.902, -20.328, -19.717, -19.074, -18.403, -17.726, -22.836, -22.422,
    -21.942, -21.442, -20.907, -20.332, -19.720, -19.077, -18.405, -17.728,
    -22.841, -22.427, -21.947, -21.447, -20.912, -20.336, -19.723, -19.080,
    -18.407, -17.731, -22.845, -22.431, -21.952, -21.452, -20.917, -20.341,
    -19.727, -19.082, -18.409, -17.733, -22.850, -22.436, -21.957, -21.457,
    -20.922, -20.345, -19.730, -19.085, -18.411, -17.735, -22.854, -22.441,
    -21.962, -21.462, -20.926, -20.349, -19.733, -19.087, -18.414, -17.737,
    -22.859, -22.445, -21.966, -21.467, -20.931, -20.353, -19.737, -19.090,
    -18.416, -17.739, -22.863, -22.450, -21.971, -21.472, -20.936, -20.357,
    -19.740, -19.092, -18.418, -17.742, -22.868, -22.455, -21.976, -21.477,
    -20.941, -20.362, -19.743, -19.095, -18.420, -17.744, -22.872, -22.460,
    -21.981, -21.482, -20.946, -20.366, -19.747, -19.097, -18.422, -17.746,
    -22.877, -22.464, -21.986, -21.487, -20.950, -20.370, -19.750, -19.100,
    -18.424, -17.748, -22.881, -22.469, -21.991, -21.492, -20.955, -20.374,
    -19.753, -19.102, -18.426, -17.750, -22.886, -22.474, -21.996, -21.497,
    -20.960, -20.378, -19.756, -19.105, -18.428, -17.753, -22.890, -22.478,
    -22.001, -21.502, -20.965, -20.382, -19.760, -19.107, -18.430, -17.755,
    -22.895, -22.483, -22.006, -21.506, -20.970, -20.386, -19.763, -19.110,
    -18.432, -17.757, -22.899, -22.488, -22.011, -21.511, -20.974, -20.390,
    -19.766, -19.112, -18.434, -17.759, -22.904, -22.493, -22.016, -21.516,
    -20.979, -20.395, -19.769, -19.115, -18.436, -17.761, -22.908, -22.497,
    -22.020, -21.521, -20.984, -20.399, -19.772, -19.117, -18.438, -17.763,
    -22.913, -22.502, -22.025, -21.526, -20.989, -20.403, -19.776, -19.120,
    -18.440, -17.766, -22.917, -22.507, -22.030, -21.531, -20.993, -20.407,
    -19.779, -19.122, -18.442, -17.768, -22.922, -22.511, -22.035, -21.536,
    -20.998, -20.411, -19.782, -19.125, -18.444, -17.770, -22.926, -22.516,
    -22.040, -21.541, -21.003, -20.415, -19.785, -19.127, -18.446, -17.772};

const double CIA_Orton::H2H2_n[][10] = {
    -29.519, -29.859, -30.193, -30.505, -30.814, -31.122, -31.411, -31.689,
    -31.968, -32.245, -27.686, -28.026, -28.360, -28.672, -28.981, -29.289,
    -29.578, -29.856, -30.135, -30.412, -26.300, -26.640, -26.974, -27.286,
    -27.595, -27.903, -28.192, -28.470, -28.749, -29.026, -24.915, -25.254,
    -25.588, -25.900, -26.209, -26.518, -26.806, -27.084, -27.363, -27.640,
    -24.105, -24.444, -24.776, -25.089, -25.398, -25.707, -25.995, -26.273,
    -26.552, -26.829, -23.531, -23.869, -24.201, -24.513, -24.823, -25.132,
    -25.420, -25.698, -25.977, -26.253, -23.085, -23.423, -23.754, -24.067,
    -24.378, -24.686, -24.974, -25.252, -25.531, -25.807, -22.721, -23.058,
    -23.389, -23.702, -24.014, -24.322, -24.609, -24.887, -25.166, -25.443,
    -22.413, -22.750, -23.080, -23.394, -23.706, -24.014, -24.301, -24.579,
    -24.858, -25.134, -22.146, -22.483, -22.812, -23.126, -23.439, -23.747,
    -24.034, -24.312, -24.591, -24.867, -21.910, -22.247, -22.576, -22.891,
    -23.204, -23.512, -23.798, -24.077, -24.356, -24.632, -21.699, -22.035,
    -22.364, -22.680, -22.994, -23.301, -23.588, -23.866, -24.145, -24.421,
    -21.507, -21.844, -22.173, -22.490, -22.804, -23.111, -23.397, -23.676,
    -23.954, -24.231, -21.333, -21.669, -21.998, -22.316, -22.630, -22.937,
    -23.223, -23.502, -23.781, -24.057, -21.172, -21.508, -21.837, -22.156,
    -22.470, -22.777, -23.063, -23.342, -23.621, -23.897, -21.022, -21.358,
    -21.688, -22.008, -22.323, -22.629, -22.915, -23.194, -23.472, -23.749,
    -20.883, -21.219, -21.549, -21.870, -22.185, -22.491, -22.777, -23.056,
    -23.335, -23.611, -20.752, -21.089, -21.420, -21.742, -22.056, -22.362,
    -22.648, -22.927, -23.206, -23.482, -20.629, -20.966, -21.298, -21.621,
    -21.935, -22.241, -22.527, -22.806, -23.084, -23.361, -20.512, -20.850,
    -21.183, -21.507, -21.820, -22.126, -22.413, -22.691, -22.969, -23.246,
    -20.402, -20.740, -21.074, -21.399, -21.712, -22.018, -22.305, -22.583,
    -22.861, -23.138, -20.298, -20.637, -20.972, -21.297, -21.611, -21.916,
    -22.203, -22.482, -22.760, -23.036, -19.841, -20.186, -20.527, -20.853,
    -21.166, -21.470, -21.757, -22.036, -22.314, -22.590, -19.470, -19.821,
    -20.165, -20.491, -20.802, -21.106, -21.393, -21.672, -21.950, -22.226,
    -19.161, -19.515, -19.861, -20.184, -20.494, -20.798, -21.086, -21.365,
    -21.642, -21.918, -18.899, -19.254, -19.598, -19.919, -20.228, -20.531,
    -20.819, -21.098, -21.375, -21.651, -18.673, -19.025, -19.367, -19.686,
    -19.992, -20.296, -20.585, -20.864, -21.140, -21.416, -18.473, -18.822,
    -19.161, -19.477, -19.782, -20.087, -20.376, -20.654, -20.930, -21.206,
    -18.294, -18.638, -18.975, -19.289, -19.593, -19.898, -20.186, -20.464,
    -20.740, -21.016, -18.131, -18.472, -18.805, -19.117, -19.420, -19.725,
    -20.014, -20.291, -20.567, -20.842, -17.982, -18.319, -18.650, -18.959,
    -19.262, -19.567, -19.855, -20.131, -20.407, -20.683, -17.845, -18.178,
    -18.507, -18.814, -19.116, -19.421, -19.708, -19.984, -20.260, -20.535,
    -17.718, -18.047, -18.373, -18.679, -18.981, -19.285, -19.572, -19.847,
    -20.123, -20.397, -17.599, -17.926, -18.249, -18.553, -18.855, -19.159,
    -19.444, -19.719, -19.994, -20.269, -17.488, -17.812, -18.133, -18.436,
    -18.737, -19.040, -19.324, -19.598, -19.874, -20.148, -17.384, -17.705,
    -18.024, -18.326, -18.626, -18.928, -19.212, -19.485, -19.760, -20.034,
    -17.286, -17.604, -17.921, -18.222, -18.521, -18.822, -19.105, -19.378,
    -19.653, -19.927, -17.193, -17.509, -17.824, -18.124, -18.422, -18.722,
    -19.004, -19.277, -19.551, -19.825, -17.021, -17.332, -17.645, -17.942,
    -18.238, -18.537, -18.817, -19.088, -19.362, -19.635, -16.865, -17.173,
    -17.481, -17.777, -18.071, -18.367, -18.646, -18.917, -19.190, -19.462,
    -16.722, -17.026, -17.331, -17.626, -17.917, -18.211, -18.489, -18.759,
    -19.031, -19.302, -16.591, -16.891, -17.194, -17.486, -17.776, -18.068,
    -18.344, -18.613, -18.885, -19.156, -16.469, -16.766, -17.066, -17.355,
    -17.644, -17.934, -18.209, -18.477, -18.748, -19.019, -16.356, -16.651,
    -16.948, -17.235, -17.522, -17.811, -18.084, -18.352, -18.622, -18.892,
    -16.251, -16.543, -16.837, -17.122, -17.407, -17.694, -17.966, -18.233,
    -18.502, -18.772, -16.154, -16.442, -16.735, -17.017, -17.300, -17.585,
    -17.856, -18.122, -18.390, -18.659, -16.063, -16.348, -16.638, -16.918,
    -17.198, -17.482, -17.752, -18.016, -18.284, -18.552, -15.980, -16.261,
    -16.548, -16.826, -17.103, -17.385, -17.654, -17.917, -18.184, -18.451,
    -15.901, -16.179, -16.462, -16.738, -17.012, -17.292, -17.560, -17.822,
    -18.088, -18.354, -15.829, -16.102, -16.382, -16.655, -16.926, -17.205,
    -17.471, -17.732, -17.997, -18.262, -15.761, -16.030, -16.307, -16.577,
    -16.846, -17.122, -17.387, -17.646, -17.910, -18.175, -15.700, -15.964,
    -16.237, -16.503, -16.770, -17.045, -17.308, -17.566, -17.828, -18.093,
    -15.644, -15.904, -16.173, -16.435, -16.700, -16.972, -17.233, -17.490,
    -17.751, -18.015, -15.589, -15.845, -16.110, -16.368, -16.630, -16.900,
    -17.159, -17.415, -17.675, -17.938, -15.539, -15.789, -16.050, -16.305,
    -16.565, -16.832, -17.089, -17.343, -17.602, -17.864, -15.492, -15.738,
    -15.995, -16.246, -16.502, -16.767, -17.023, -17.275, -17.532, -17.793,
    -15.450, -15.690, -15.943, -16.191, -16.444, -16.706, -16.960, -17.210,
    -17.467, -17.726, -15.412, -15.647, -15.895, -16.139, -16.389, -16.649,
    -16.900, -17.149, -17.404, -17.662, -15.374, -15.604, -15.848, -16.088,
    -16.335, -16.592, -16.841, -17.088, -17.342, -17.599, -15.339, -15.564,
    -15.803, -16.040, -16.283, -16.538, -16.785, -17.030, -17.282, -17.538,
    -15.307, -15.527, -15.761, -15.994, -16.235, -16.486, -16.731, -16.974,
    -17.225, -17.479, -15.278, -15.493, -15.722, -15.951, -16.188, -16.437,
    -16.680, -16.921, -17.170, -17.423, -15.252, -15.462, -15.686, -15.911,
    -16.145, -16.391, -16.631, -16.870, -17.117, -17.369, -15.226, -15.431,
    -15.650, -15.872, -16.102, -16.345, -16.583, -16.820, -17.065, -17.316,
    -15.202, -15.402, -15.617, -15.834, -16.062, -16.302, -16.536, -16.771,
    -17.015, -17.264, -15.180, -15.375, -15.586, -15.799, -16.023, -16.260,
    -16.492, -16.725, -16.967, -17.215, -15.161, -15.350, -15.556, -15.765,
    -15.987, -16.220, -16.449, -16.680, -16.921, -17.167, -15.143, -15.327,
    -15.529, -15.734, -15.952, -16.182, -16.408, -16.637, -16.876, -17.121,
    -15.126, -15.305, -15.502, -15.703, -15.918, -16.145, -16.368, -16.595,
    -16.832, -17.075, -15.110, -15.284, -15.477, -15.674, -15.885, -16.108,
    -16.329, -16.554, -16.789, -17.031, -15.095, -15.264, -15.453, -15.646,
    -15.853, -16.074, -16.292, -16.514, -16.748, -16.988, -15.082, -15.246,
    -15.430, -15.620, -15.824, -16.041, -16.256, -16.476, -16.708, -16.946,
    -15.071, -15.230, -15.409, -15.595, -15.795, -16.009, -16.222, -16.439,
    -16.669, -16.906, -15.060, -15.214, -15.388, -15.571, -15.767, -15.978,
    -16.188, -16.403, -16.631, -16.866, -15.049, -15.199, -15.369, -15.548,
    -15.741, -15.948, -16.155, -16.368, -16.594, -16.828, -15.040, -15.185,
    -15.351, -15.526, -15.715, -15.920, -16.124, -16.334, -16.558, -16.790,
    -15.032, -15.173, -15.334, -15.505, -15.691, -15.892, -16.093, -16.301,
    -16.524, -16.754, -15.025, -15.161, -15.318, -15.486, -15.668, -15.866,
    -16.064, -16.270, -16.490, -16.719, -15.018, -15.150, -15.303, -15.467,
    -15.645, -15.840, -16.035, -16.239, -16.457, -16.684, -15.012, -15.140,
    -15.288, -15.449, -15.624, -15.815, -16.008, -16.209, -16.425, -16.650,
    -15.007, -15.130, -15.275, -15.431, -15.603, -15.791, -15.981, -16.180,
    -16.394, -16.617, -15.003, -15.122, -15.262, -15.415, -15.583, -15.767,
    -15.955, -16.151, -16.363, -16.585, -14.999, -15.114, -15.251, -15.399,
    -15.564, -15.745, -15.930, -16.124, -16.334, -16.554, -14.996, -15.107,
    -15.240, -15.384, -15.545, -15.723, -15.905, -16.097, -16.305, -16.523,
    -14.993, -15.100, -15.229, -15.370, -15.528, -15.702, -15.882, -16.071,
    -16.277, -16.493, -14.991, -15.094, -15.219, -15.356, -15.511, -15.682,
    -15.859, -16.046, -16.249, -16.464, -14.990, -15.089, -15.210, -15.343,
    -15.494, -15.663, -15.836, -16.021, -16.222, -16.435, -14.989, -15.084,
    -15.201, -15.331, -15.479, -15.644, -15.815, -15.997, -16.196, -16.407,
    -14.988, -15.080, -15.193, -15.319, -15.464, -15.626, -15.794, -15.973,
    -16.170, -16.380, -14.988, -15.076, -15.185, -15.308, -15.449, -15.608,
    -15.773, -15.950, -16.145, -16.353, -14.988, -15.072, -15.178, -15.297,
    -15.436, -15.591, -15.753, -15.928, -16.121, -16.326, -14.989, -15.069,
    -15.171, -15.287, -15.423, -15.575, -15.734, -15.906, -16.097, -16.301,
    -14.990, -15.067, -15.165, -15.278, -15.410, -15.559, -15.715, -15.885,
    -16.074, -16.276, -14.991, -15.065, -15.159, -15.269, -15.398, -15.544,
    -15.697, -15.864, -16.051, -16.251, -14.993, -15.063, -15.154, -15.260,
    -15.386, -15.529, -15.679, -15.844, -16.029, -16.227, -14.995, -15.062,
    -15.150, -15.252, -15.375, -15.514, -15.662, -15.825, -16.007, -16.204,
    -14.998, -15.061, -15.146, -15.245, -15.364, -15.500, -15.645, -15.805,
    -15.986, -16.181, -15.001, -15.061, -15.142, -15.238, -15.354, -15.487,
    -15.629, -15.787, -15.965, -16.158, -15.004, -15.061, -15.138, -15.231,
    -15.344, -15.474, -15.613, -15.769, -15.945, -16.136, -15.007, -15.061,
    -15.135, -15.225, -15.334, -15.461, -15.598, -15.751, -15.924, -16.114,
    -15.011, -15.061, -15.133, -15.219, -15.325, -15.449, -15.583, -15.733,
    -15.905, -16.092, -15.015, -15.062, -15.130, -15.213, -15.316, -15.437,
    -15.569, -15.717, -15.886, -16.072, -15.019, -15.063, -15.128, -15.208,
    -15.308, -15.426, -15.555, -15.700, -15.867, -16.051, -15.023, -15.064,
    -15.126, -15.203, -15.300, -15.415, -15.541, -15.684, -15.849, -16.031,
    -15.028, -15.066, -15.125, -15.199, -15.292, -15.404, -15.528, -15.668,
    -15.831, -16.011, -15.032, -15.068, -15.123, -15.194, -15.285, -15.394,
    -15.515, -15.653, -15.813, -15.992, -15.037, -15.070, -15.123, -15.190,
    -15.278, -15.384, -15.502, -15.638, -15.796, -15.973, -15.042, -15.073,
    -15.122, -15.187, -15.272, -15.375, -15.490, -15.623, -15.779, -15.954,
    -15.047, -15.076, -15.122, -15.184, -15.266, -15.366, -15.478, -15.609,
    -15.762, -15.936, -15.053, -15.078, -15.122, -15.181, -15.260, -15.357,
    -15.467, -15.594, -15.746, -15.918, -15.058, -15.081, -15.122, -15.178,
    -15.254, -15.348, -15.455, -15.581, -15.731, -15.900, -15.064, -15.085,
    -15.122, -15.175, -15.248, -15.340, -15.444, -15.567, -15.715, -15.883,
    -15.070, -15.088, -15.123, -15.173, -15.243, -15.332, -15.434, -15.554,
    -15.700, -15.866, -15.076, -15.092, -15.124, -15.171, -15.238, -15.324,
    -15.424, -15.541, -15.685, -15.849, -15.082, -15.095, -15.125, -15.169,
    -15.234, -15.317, -15.414, -15.529, -15.670, -15.832, -15.088, -15.099,
    -15.126, -15.168, -15.229, -15.310, -15.404, -15.516, -15.656, -15.816,
    -15.095, -15.104, -15.127, -15.167, -15.226, -15.303, -15.394, -15.505,
    -15.642, -15.800, -15.101, -15.108, -15.129, -15.166, -15.222, -15.296,
    -15.385, -15.493, -15.628, -15.784, -15.108, -15.112, -15.131, -15.165,
    -15.218, -15.290, -15.376, -15.482, -15.614, -15.769, -15.114, -15.117,
    -15.133, -15.165, -15.215, -15.284, -15.367, -15.471, -15.601, -15.753,
    -15.121, -15.121, -15.135, -15.164, -15.212, -15.278, -15.358, -15.460,
    -15.588, -15.739, -15.128, -15.126, -15.138, -15.164, -15.209, -15.272,
    -15.350, -15.449, -15.575, -15.724, -15.135, -15.131, -15.141, -15.164,
    -15.206, -15.267, -15.342, -15.439, -15.563, -15.710, -15.142, -15.136,
    -15.144, -15.164, -15.203, -15.262, -15.334, -15.429, -15.551, -15.696,
    -15.149, -15.141, -15.147, -15.164, -15.201, -15.257, -15.326, -15.419,
    -15.539, -15.682, -15.156, -15.146, -15.150, -15.165, -15.199, -15.252,
    -15.319, -15.409, -15.527, -15.668, -15.164, -15.152, -15.153, -15.165,
    -15.197, -15.247, -15.312, -15.399, -15.515, -15.655, -15.171, -15.157,
    -15.156, -15.166, -15.195, -15.242, -15.305, -15.390, -15.504, -15.642,
    -15.178, -15.163, -15.159, -15.167, -15.193, -15.238, -15.298, -15.381,
    -15.493, -15.629, -15.186, -15.168, -15.163, -15.168, -15.192, -15.233,
    -15.292, -15.372, -15.482, -15.616, -15.193, -15.174, -15.166, -15.169,
    -15.190, -15.230, -15.285, -15.363, -15.471, -15.603, -15.201, -15.180,
    -15.170, -15.170, -15.189, -15.226, -15.279, -15.354, -15.460, -15.591,
    -15.208, -15.186, -15.174, -15.172, -15.188, -15.222, -15.273, -15.346,
    -15.450, -15.579, -15.216, -15.192, -15.178, -15.174, -15.188, -15.219,
    -15.267, -15.338, -15.439, -15.567, -15.224, -15.198, -15.182, -15.176,
    -15.187, -15.216, -15.261, -15.330, -15.429, -15.555, -15.231, -15.204,
    -15.186, -15.178, -15.186, -15.213, -15.256, -15.322, -15.419, -15.543,
    -15.239, -15.211, -15.191, -15.180, -15.186, -15.210, -15.250, -15.314,
    -15.410, -15.531, -15.247, -15.217, -15.195, -15.182, -15.186, -15.208,
    -15.245, -15.307, -15.400, -15.520, -15.255, -15.223, -15.199, -15.184,
    -15.186, -15.205, -15.240, -15.300, -15.390, -15.509, -15.263, -15.230,
    -15.204, -15.187, -15.186, -15.203, -15.235, -15.292, -15.381, -15.498,
    -15.271, -15.236, -15.209, -15.189, -15.186, -15.200, -15.230, -15.285,
    -15.372, -15.487, -15.279, -15.243, -15.213, -15.192, -15.186, -15.198,
    -15.225, -15.278, -15.363, -15.476, -15.287, -15.250, -15.218, -15.194,
    -15.186, -15.196, -15.221, -15.272, -15.354, -15.465, -15.295, -15.256,
    -15.223, -15.197, -15.187, -15.194, -15.216, -15.265, -15.345, -15.454,
    -15.303, -15.263, -15.228, -15.200, -15.187, -15.192, -15.212, -15.258,
    -15.337, -15.444, -15.311, -15.270, -15.233, -15.203, -15.188, -15.190,
    -15.208, -15.252, -15.328, -15.434, -15.319, -15.277, -15.239, -15.207,
    -15.189, -15.189, -15.204, -15.246, -15.320, -15.424, -15.327, -15.284,
    -15.244, -15.210, -15.190, -15.187, -15.200, -15.240, -15.312, -15.414,
    -15.335, -15.291, -15.249, -15.213, -15.191, -15.185, -15.196, -15.234,
    -15.304, -15.404, -15.343, -15.298, -15.255, -15.217, -15.192, -15.184,
    -15.192, -15.228, -15.296, -15.394, -15.351, -15.305, -15.260, -15.220,
    -15.193, -15.183, -15.189, -15.222, -15.288, -15.385, -15.360, -15.312,
    -15.265, -15.223, -15.194, -15.181, -15.185, -15.216, -15.280, -15.376,
    -15.368, -15.319, -15.271, -15.227, -15.195, -15.180, -15.182, -15.211,
    -15.273, -15.366, -15.376, -15.326, -15.276, -15.231, -15.197, -15.179,
    -15.178, -15.205, -15.265, -15.357, -15.384, -15.333, -15.282, -15.234,
    -15.198, -15.178, -15.175, -15.200, -15.258, -15.348, -15.392, -15.340,
    -15.288, -15.238, -15.200, -15.177, -15.172, -15.194, -15.251, -15.339,
    -15.401, -15.348, -15.293, -15.242, -15.201, -15.177, -15.169, -15.189,
    -15.244, -15.331, -15.409, -15.355, -15.299, -15.246, -15.203, -15.176,
    -15.166, -15.184, -15.237, -15.322, -15.417, -15.362, -15.305, -15.250,
    -15.205, -15.175, -15.163, -15.179, -15.230, -15.313, -15.425, -15.369,
    -15.311, -15.254, -15.207, -15.175, -15.160, -15.174, -15.223, -15.305,
    -15.434, -15.377, -15.316, -15.258, -15.208, -15.174, -15.158, -15.169,
    -15.216, -15.296, -15.442, -15.384, -15.322, -15.262, -15.210, -15.174,
    -15.155, -15.164, -15.210, -15.288, -15.450, -15.391, -15.328, -15.266,
    -15.212, -15.174, -15.152, -15.159, -15.203, -15.280, -15.458, -15.399,
    -15.334, -15.270, -15.215, -15.174, -15.150, -15.155, -15.197, -15.272,
    -15.466, -15.406, -15.340, -15.275, -15.217, -15.174, -15.148, -15.150,
    -15.190, -15.264, -15.475, -15.413, -15.347, -15.279, -15.219, -15.174,
    -15.145, -15.146, -15.184, -15.256, -15.483, -15.421, -15.353, -15.284,
    -15.222, -15.174, -15.143, -15.141, -15.178, -15.248, -15.491, -15.428,
    -15.359, -15.288, -15.224, -15.174, -15.141, -15.137, -15.172, -15.240,
    -15.499, -15.435, -15.365, -15.292, -15.226, -15.174, -15.139, -15.133,
    -15.166, -15.232, -15.508, -15.443, -15.371, -15.297, -15.229, -15.174,
    -15.136, -15.128, -15.160, -15.225, -15.516, -15.450, -15.377, -15.301,
    -15.231, -15.174, -15.134, -15.124, -15.154, -15.217, -15.524, -15.458,
    -15.384, -15.306, -15.234, -15.174, -15.132, -15.120, -15.148, -15.210,
    -15.532, -15.465, -15.390, -15.310, -15.236, -15.174, -15.130, -15.116,
    -15.142, -15.202, -15.541, -15.473, -15.396, -15.315, -15.239, -15.174,
    -15.128, -15.112, -15.136, -15.195, -15.549, -15.480, -15.402, -15.319,
    -15.241, -15.174, -15.126, -15.108, -15.131, -15.188, -15.557, -15.488,
    -15.408, -15.324, -15.244, -15.175, -15.124, -15.104, -15.125, -15.181,
    -15.566, -15.495, -15.415, -15.328, -15.246, -15.175, -15.122, -15.101,
    -15.119, -15.174, -15.574, -15.503, -15.421, -15.333, -15.249, -15.175,
    -15.120, -15.097, -15.114, -15.167, -15.582, -15.510, -15.427, -15.338,
    -15.252, -15.175, -15.118, -15.093, -15.108, -15.160, -15.590, -15.518,
    -15.434, -15.342, -15.254, -15.176, -15.116, -15.089, -15.103, -15.153,
    -15.599, -15.526, -15.440, -15.347, -15.257, -15.176, -15.114, -15.086,
    -15.098, -15.146, -15.607, -15.533, -15.446, -15.352, -15.259, -15.176,
    -15.112, -15.082, -15.092, -15.139, -15.615, -15.540, -15.453, -15.356,
    -15.262, -15.177, -15.111, -15.078, -15.087, -15.133, -15.623, -15.548,
    -15.459, -15.361, -15.264, -15.177, -15.109, -15.075, -15.082, -15.126,
    -15.632, -15.555, -15.466, -15.366, -15.267, -15.177, -15.107, -15.071,
    -15.076, -15.119, -15.640, -15.563, -15.472, -15.370, -15.270, -15.178,
    -15.105, -15.068, -15.071, -15.113, -15.648, -15.570, -15.479, -15.375,
    -15.272, -15.178, -15.104, -15.064, -15.066, -15.106, -15.656, -15.578,
    -15.485, -15.380, -15.275, -15.179, -15.102, -15.061, -15.061, -15.100,
    -15.664, -15.586, -15.491, -15.385, -15.278, -15.179, -15.100, -15.057,
    -15.056, -15.094, -15.672, -15.593, -15.498, -15.389, -15.280, -15.180,
    -15.099, -15.054, -15.051, -15.087, -15.680, -15.601, -15.504, -15.394,
    -15.283, -15.180, -15.097, -15.050, -15.046, -15.081, -15.688, -15.608,
    -15.511, -15.399, -15.285, -15.180, -15.095, -15.047, -15.041, -15.074,
    -15.696, -15.615, -15.517, -15.403, -15.288, -15.180, -15.093, -15.043,
    -15.036, -15.068, -15.704, -15.623, -15.523, -15.408, -15.290, -15.180,
    -15.092, -15.039, -15.031, -15.062, -15.712, -15.630, -15.530, -15.413,
    -15.293, -15.181, -15.090, -15.036, -15.026, -15.055, -15.720, -15.638,
    -15.536, -15.418, -15.295, -15.181, -15.088, -15.032, -15.021, -15.049,
    -15.728, -15.645, -15.543, -15.422, -15.298, -15.181, -15.087, -15.029,
    -15.016, -15.043, -15.736, -15.653, -15.549, -15.427, -15.300, -15.181,
    -15.085, -15.025, -15.011, -15.037, -15.744, -15.660, -15.555, -15.431,
    -15.302, -15.181, -15.083, -15.021, -15.006, -15.031, -15.752, -15.668,
    -15.561, -15.436, -15.305, -15.181, -15.081, -15.018, -15.001, -15.025,
    -15.760, -15.675, -15.568, -15.440, -15.307, -15.181, -15.079, -15.014,
    -14.996, -15.019, -15.768, -15.682, -15.574, -15.444, -15.309, -15.181,
    -15.077, -15.011, -14.992, -15.013, -15.776, -15.690, -15.580, -15.449,
    -15.311, -15.181, -15.075, -15.007, -14.987, -15.007, -15.784, -15.697,
    -15.586, -15.453, -15.314, -15.181, -15.073, -15.003, -14.982, -15.001,
    -15.792, -15.704, -15.592, -15.457, -15.316, -15.181, -15.071, -15.000,
    -14.977, -14.995, -15.800, -15.712, -15.598, -15.461, -15.318, -15.180,
    -15.069, -14.996, -14.972, -14.989, -15.808, -15.719, -15.604, -15.465,
    -15.320, -15.180, -15.066, -14.993, -14.968, -14.984, -15.816, -15.726,
    -15.610, -15.469, -15.322, -15.180, -15.064, -14.989, -14.963, -14.978,
    -15.824, -15.733, -15.616, -15.473, -15.324, -15.180, -15.062, -14.985,
    -14.958, -14.972, -15.832, -15.740, -15.622, -15.477, -15.325, -15.179,
    -15.059, -14.981, -14.953, -14.966, -15.840, -15.748, -15.628, -15.481,
    -15.326, -15.178, -15.057, -14.978, -14.949, -14.960, -15.847, -15.755,
    -15.634, -15.485, -15.328, -15.178, -15.054, -14.974, -14.944, -14.955,
    -15.855, -15.762, -15.639, -15.489, -15.329, -15.177, -15.052, -14.970,
    -14.939, -14.949, -15.863, -15.769, -15.645, -15.493, -15.330, -15.176,
    -15.049, -14.966, -14.934, -14.943, -15.871, -15.777, -15.651, -15.497,
    -15.331, -15.175, -15.047, -14.962, -14.929, -14.937, -15.879, -15.784,
    -15.657, -15.500, -15.332, -15.174, -15.044, -14.958, -14.925, -14.932,
    -15.886, -15.791, -15.662, -15.504, -15.332, -15.173, -15.042, -14.954,
    -14.920, -14.926, -15.894, -15.798, -15.668, -15.508, -15.333, -15.172,
    -15.039, -14.950, -14.915, -14.920, -15.902, -15.805, -15.673, -15.511,
    -15.334, -15.171, -15.036, -14.946, -14.910, -14.914, -15.909, -15.811,
    -15.679, -15.514, -15.334, -15.169, -15.033, -14.942, -14.905, -14.909,
    -15.917, -15.818, -15.684, -15.517, -15.335, -15.167, -15.030, -14.938,
    -14.900, -14.903, -15.925, -15.824, -15.690, -15.520, -15.336, -15.166,
    -15.027, -14.934, -14.895, -14.897, -15.932, -15.830, -15.695, -15.523,
    -15.336, -15.164, -15.023, -14.930, -14.890, -14.891, -15.940, -15.836,
    -15.700, -15.525, -15.337, -15.162, -15.020, -14.926, -14.885, -14.886,
    -15.947, -15.842, -15.705, -15.528, -15.337, -15.160, -15.017, -14.922,
    -14.880, -14.880, -15.955, -15.848, -15.710, -15.530, -15.337, -15.158,
    -15.013, -14.917, -14.874, -14.874, -15.962, -15.854, -15.715, -15.532,
    -15.338, -15.155, -15.010, -14.913, -14.869, -14.869, -15.970, -15.860,
    -15.720, -15.534, -15.338, -15.153, -15.006, -14.909, -14.864, -14.863,
    -15.977, -15.866, -15.724, -15.536, -15.337, -15.150, -15.002, -14.904,
    -14.859, -14.857, -15.985, -15.873, -15.729, -15.538, -15.337, -15.148,
    -14.998, -14.900, -14.854, -14.852, -15.992, -15.879, -15.733, -15.539,
    -15.336, -15.145, -14.994, -14.895, -14.849, -14.846, -15.999, -15.886,
    -15.737, -15.541, -15.334, -15.142, -14.990, -14.890, -14.843, -14.840,
    -16.007, -15.892, -15.740, -15.542, -15.333, -15.139, -14.985, -14.886,
    -14.838, -14.835, -16.014, -15.899, -15.744, -15.543, -15.331, -15.135,
    -14.981, -14.881, -14.833, -14.829, -16.021, -15.905, -15.747, -15.544,
    -15.330, -15.132, -14.976, -14.876, -14.828, -14.823, -16.028, -15.912,
    -15.751, -15.545, -15.328, -15.128, -14.972, -14.870, -14.822, -14.817,
    -16.035, -15.918, -15.754, -15.546, -15.325, -15.125, -14.967, -14.865,
    -14.817, -14.812, -16.042, -15.925, -15.757, -15.547, -15.323, -15.121,
    -14.962, -14.860, -14.812, -14.806, -16.049, -15.931, -15.760, -15.547,
    -15.321, -15.117, -14.957, -14.855, -14.807, -14.800, -16.056, -15.936,
    -15.762, -15.547, -15.318, -15.113, -14.952, -14.849, -14.801, -14.794,
    -16.062, -15.941, -15.765, -15.546, -15.315, -15.108, -14.946, -14.844,
    -14.795, -14.789, -16.069, -15.946, -15.767, -15.546, -15.312, -15.103,
    -14.941, -14.838, -14.790, -14.783, -16.076, -15.951, -15.769, -15.545,
    -15.309, -15.099, -14.935, -14.833, -14.784, -14.777, -16.082, -15.956,
    -15.772, -15.544, -15.305, -15.094, -14.930, -14.827, -14.778, -14.771,
    -16.089, -15.960, -15.773, -15.542, -15.302, -15.088, -14.924, -14.821,
    -14.773, -14.765, -16.095, -15.965, -15.775, -15.541, -15.298, -15.083,
    -14.918, -14.815, -14.767, -14.759, -16.102, -15.969, -15.777, -15.539,
    -15.294, -15.078, -14.912, -14.809, -14.761, -14.753, -16.108, -15.973,
    -15.778, -15.537, -15.290, -15.072, -14.906, -14.803, -14.755, -14.747,
    -16.114, -15.977, -15.779, -15.534, -15.285, -15.066, -14.900, -14.797,
    -14.749, -14.741, -16.120, -15.980, -15.780, -15.532, -15.280, -15.060,
    -14.893, -14.791, -14.743, -14.735, -16.126, -15.983, -15.780, -15.528,
    -15.275, -15.053, -14.887, -14.784, -14.737, -14.729, -16.131, -15.987,
    -15.781, -15.525, -15.269, -15.046, -14.880, -14.778, -14.730, -14.723,
    -16.137, -15.989, -15.781, -15.521, -15.263, -15.039, -14.873, -14.772,
    -14.724, -14.717, -16.142, -15.992, -15.781, -15.517, -15.257, -15.032,
    -14.866, -14.765, -14.717, -14.711, -16.148, -15.995, -15.780, -15.513,
    -15.251, -15.025, -14.859, -14.758, -14.711, -14.705, -16.153, -15.997,
    -15.780, -15.508, -15.244, -15.017, -14.852, -14.752, -14.704, -14.698,
    -16.158, -15.999, -15.779, -15.503, -15.237, -15.009, -14.844, -14.745,
    -14.698, -14.692, -16.163, -16.001, -15.778, -15.498, -15.230, -15.001,
    -14.837, -14.738, -14.691, -14.686, -16.168, -16.002, -15.776, -15.492,
    -15.222, -14.993, -14.829, -14.731, -14.684, -14.680, -16.172, -16.003,
    -15.773, -15.486, -15.215, -14.984, -14.821, -14.724, -14.678, -14.673,
    -16.176, -16.004, -15.769, -15.479, -15.206, -14.975, -14.813, -14.716,
    -14.671, -14.667, -16.181, -16.004, -15.766, -15.472, -15.198, -14.966,
    -14.805, -14.709, -14.665, -14.661, -16.185, -16.004, -15.762, -15.465,
    -15.189, -14.957, -14.796, -14.701, -14.658, -14.654, -16.188, -16.004,
    -15.757, -15.457, -15.180, -14.948, -14.788, -14.694, -14.652, -14.648,
    -16.192, -16.003, -15.752, -15.449, -15.171, -14.938, -14.779, -14.686,
    -14.645, -14.642, -16.195, -16.003, -15.746, -15.441, -15.161, -14.928,
    -14.770, -14.678, -14.638, -14.635, -16.199, -16.001, -15.740, -15.432,
    -15.151, -14.918, -14.761, -14.670, -14.632, -14.629, -16.202, -16.000,
    -15.734, -15.423, -15.141, -14.908, -14.752, -14.662, -14.625, -14.622,
    -16.204, -15.998, -15.727, -15.413, -15.130, -14.898, -14.743, -14.654,
    -14.618, -14.615, -16.206, -15.995, -15.720, -15.403, -15.119, -14.887,
    -14.733, -14.645, -14.611, -14.609, -16.208, -15.992, -15.712, -15.392,
    -15.108, -14.876, -14.723, -14.637, -14.604, -14.602, -16.209, -15.988,
    -15.704, -15.381, -15.096, -14.865, -14.714, -14.628, -14.596, -14.596,
    -16.211, -15.984, -15.695, -15.369, -15.084, -14.854, -14.704, -14.619,
    -14.589, -14.589, -16.211, -15.980, -15.686, -15.357, -15.072, -14.843,
    -14.693, -14.611, -14.582, -14.583, -16.212, -15.975, -15.676, -15.345,
    -15.060, -14.831, -14.683, -14.602, -14.574, -14.576, -16.212, -15.970,
    -15.666, -15.332, -15.047, -14.819, -14.672, -14.593, -14.566, -14.569,
    -16.212, -15.965, -15.655, -15.319, -15.034, -14.807, -14.662, -14.583,
    -14.559, -14.563, -16.212, -15.959, -15.644, -15.305, -15.020, -14.795,
    -14.651, -14.574, -14.551, -14.556, -16.210, -15.951, -15.632, -15.291,
    -15.005, -14.782, -14.640, -14.565, -14.543, -14.549, -16.208, -15.942,
    -15.620, -15.276, -14.989, -14.769, -14.629, -14.555, -14.536, -14.542,
    -16.206, -15.932, -15.607, -15.261, -14.973, -14.756, -14.617, -14.546,
    -14.528, -14.535, -16.203, -15.921, -15.594, -15.245, -14.957, -14.742,
    -14.606, -14.537, -14.520, -14.528, -16.199, -15.910, -15.580, -15.229,
    -14.940, -14.728, -14.594, -14.527, -14.512, -14.521, -16.195, -15.897,
    -15.566, -15.213, -14.922, -14.714, -14.582, -14.517, -14.504, -14.514,
    -16.190, -15.884, -15.551, -15.196, -14.904, -14.700, -14.570, -14.508,
    -14.496, -14.507, -16.185, -15.871, -15.535, -15.178, -14.885, -14.685,
    -14.558, -14.498, -14.488, -14.500, -16.180, -15.856, -15.519, -15.160,
    -14.866, -14.670, -14.546, -14.488, -14.480, -14.493, -16.174, -15.841,
    -15.503, -15.142, -14.846, -14.655, -14.533, -14.478, -14.472, -14.486,
    -16.166, -15.826, -15.485, -15.123, -14.829, -14.640, -14.520, -14.468,
    -14.464, -14.479, -16.158, -15.811, -15.467, -15.104, -14.813, -14.625,
    -14.507, -14.458, -14.455, -14.472, -16.149, -15.795, -15.448, -15.084,
    -14.796, -14.610, -14.494, -14.448, -14.447, -14.465, -16.139, -15.779,
    -15.428, -15.064, -14.779, -14.595, -14.481, -14.438, -14.438, -14.457,
    -16.128, -15.762, -15.407, -15.044, -14.763, -14.579, -14.468, -14.428,
    -14.429, -14.450, -16.117, -15.745, -15.386, -15.023, -14.747, -14.563,
    -14.454, -14.418, -14.420, -14.443, -16.105, -15.727, -15.364, -15.001,
    -14.731, -14.547, -14.440, -14.407, -14.411, -14.436, -16.092, -15.708,
    -15.342, -14.979, -14.715, -14.531, -14.426, -14.397, -14.402, -14.429,
    -16.079, -15.689, -15.319, -14.957, -14.699, -14.515, -14.412, -14.386,
    -14.393, -14.421, -16.065, -15.670, -15.295, -14.934, -14.683, -14.499,
    -14.398, -14.376, -14.384, -14.414, -16.048, -15.648, -15.271, -14.911,
    -14.664, -14.482, -14.384, -14.365, -14.375, -14.407, -16.030, -15.626,
    -15.247, -14.887, -14.644, -14.466, -14.370, -14.354, -14.367, -14.399,
    -16.011, -15.602, -15.223, -14.864, -14.624, -14.448, -14.356, -14.343,
    -14.358, -14.392, -15.990, -15.578, -15.198, -14.839, -14.603, -14.431,
    -14.342, -14.332, -14.349, -14.385, -15.969, -15.553, -15.172, -14.815,
    -14.582, -14.414, -14.328, -14.320, -14.341, -14.378, -15.946, -15.527,
    -15.146, -14.790, -14.560, -14.396, -14.314, -14.309, -14.332, -14.370,
    -15.923, -15.500, -15.120, -14.764, -14.537, -14.378, -14.299, -14.297,
    -14.324, -14.363, -15.898, -15.473, -15.093, -14.738, -14.514, -14.361,
    -14.285, -14.285, -14.315, -14.356, -15.872, -15.444, -15.065, -14.712,
    -14.491, -14.342, -14.271, -14.273, -14.307, -14.348, -15.845, -15.415,
    -15.037, -14.685, -14.467, -14.324, -14.256, -14.261, -14.298, -14.341,
    -15.816, -15.386, -15.009, -14.658, -14.445, -14.306, -14.242, -14.249,
    -14.289, -14.333, -15.785, -15.355, -14.980, -14.631, -14.423, -14.287,
    -14.228, -14.237, -14.279, -14.325, -15.752, -15.324, -14.950, -14.604,
    -14.401, -14.269, -14.214, -14.225, -14.270, -14.318, -15.719, -15.293,
    -14.920, -14.577, -14.379, -14.250, -14.200, -14.213, -14.260, -14.310,
    -15.684, -15.260, -14.890, -14.549, -14.357, -14.231, -14.186, -14.201,
    -14.250, -14.301, -15.648, -15.227, -14.859, -14.521, -14.335, -14.212,
    -14.172, -14.189, -14.241, -14.293, -15.610, -15.194, -14.827, -14.492,
    -14.313, -14.193, -14.158, -14.177, -14.231, -14.285, -15.571, -15.159,
    -14.795, -14.463, -14.291, -14.173, -14.145, -14.165, -14.221, -14.277,
    -15.531, -15.124, -14.763, -14.434, -14.268, -14.154, -14.131, -14.152,
    -14.210, -14.268, -15.489, -15.088, -14.730, -14.405, -14.246, -14.134,
    -14.117, -14.140, -14.200, -14.260, -15.446, -15.050, -14.696, -14.376,
    -14.223, -14.115, -14.102, -14.127, -14.191, -14.252, -15.402, -15.011,
    -14.662, -14.347, -14.200, -14.096, -14.087, -14.115, -14.181, -14.244,
    -15.357, -14.972, -14.627, -14.318, -14.177, -14.077, -14.072, -14.102,
    -14.172, -14.237, -15.310, -14.931, -14.592, -14.288, -14.153, -14.058,
    -14.057, -14.089, -14.163, -14.229, -15.262, -14.890, -14.556, -14.259,
    -14.130, -14.039, -14.042, -14.076, -14.154, -14.221, -15.213, -14.847,
    -14.519, -14.229, -14.106, -14.020, -14.026, -14.063, -14.145, -14.214,
    -15.162, -14.804, -14.482, -14.199, -14.082, -14.001, -14.011, -14.050,
    -14.136, -14.206, -15.111, -14.759, -14.445, -14.169, -14.058, -13.982,
    -13.995, -14.037, -14.127, -14.199, -15.057, -14.714, -14.407, -14.138,
    -14.034, -13.963, -13.979, -14.023, -14.119, -14.191, -15.003, -14.668,
    -14.368, -14.108, -14.010, -13.944, -13.963, -14.010, -14.110, -14.184,
    -14.955, -14.626, -14.334, -14.078, -13.986, -13.924, -13.948, -13.998,
    -14.100, -14.176, -14.907, -14.583, -14.300, -14.049, -13.962, -13.905,
    -13.932, -13.987, -14.091, -14.168, -14.859, -14.541, -14.266, -14.019,
    -13.939, -13.885, -13.917, -13.975, -14.081, -14.160, -14.811, -14.498,
    -14.233, -13.989, -13.915, -13.865, -13.901, -13.964, -14.071, -14.152,
    -14.763, -14.455, -14.200, -13.960, -13.891, -13.845, -13.886, -13.953,
    -14.061, -14.144, -14.715, -14.413, -14.167, -13.930, -13.867, -13.825,
    -13.870, -13.943, -14.050, -14.136, -14.668, -14.370, -14.135, -13.901,
    -13.843, -13.805, -13.855, -13.932, -14.040, -14.128, -14.620, -14.327,
    -14.103, -13.871, -13.819, -13.785, -13.839, -13.922, -14.029, -14.120,
    -14.572, -14.284, -14.071, -13.842, -13.795, -13.764, -13.824, -13.912,
    -14.019, -14.112, -14.525, -14.241, -14.040, -13.812, -13.771, -13.744,
    -13.808, -13.902, -14.008, -14.104, -14.476, -14.199, -14.000, -13.783,
    -13.748, -13.725, -13.793, -13.890, -13.997, -14.096, -14.427, -14.158,
    -13.959, -13.755, -13.725, -13.707, -13.778, -13.878, -13.987, -14.088,
    -14.378, -14.117, -13.917, -13.726, -13.702, -13.688, -13.763, -13.866,
    -13.976, -14.081, -14.329, -14.076, -13.874, -13.698, -13.679, -13.670,
    -13.748, -13.853, -13.965, -14.073, -14.279, -14.035, -13.829, -13.670,
    -13.656, -13.652, -13.733, -13.841, -13.954, -14.066, -14.230, -13.995,
    -13.783, -13.642, -13.634, -13.635, -13.718, -13.828, -13.942, -14.058,
    -14.180, -13.955, -13.735, -13.614, -13.611, -13.617, -13.704, -13.815,
    -13.931, -14.050, -14.130, -13.915, -13.687, -13.586, -13.589, -13.600,
    -13.689, -13.802, -13.920, -14.043, -14.079, -13.875, -13.637, -13.558,
    -13.567, -13.583, -13.674, -13.789, -13.909, -14.035, -14.029, -13.835,
    -13.586, -13.531, -13.545, -13.566, -13.660, -13.775, -13.897, -14.028,
    -13.981, -13.796, -13.554, -13.505, -13.518, -13.549, -13.646, -13.764,
    -13.888, -14.021, -13.933, -13.758, -13.525, -13.479, -13.489, -13.532,
    -13.632, -13.753, -13.880, -14.014, -13.885, -13.719, -13.498, -13.454,
    -13.460, -13.516, -13.618, -13.742, -13.873, -14.008, -13.837, -13.682,
    -13.473, -13.429, -13.430, -13.499, -13.605, -13.731, -13.866, -14.001,
    -13.790, -13.644, -13.451, -13.405, -13.399, -13.483, -13.591, -13.721,
    -13.859, -13.994, -13.743, -13.607, -13.432, -13.381, -13.367, -13.466,
    -13.578, -13.711, -13.853, -13.988, -13.696, -13.570, -13.415, -13.357,
    -13.334, -13.450, -13.565, -13.702, -13.847, -13.982, -13.649, -13.533,
    -13.400, -13.334, -13.301, -13.434, -13.552, -13.692, -13.842, -13.976,
    -13.603, -13.497, -13.388, -13.311, -13.266, -13.419, -13.539, -13.683,
    -13.837, -13.970, -13.557, -13.461, -13.378, -13.288, -13.231, -13.403,
    -13.526, -13.674, -13.833, -13.964, -13.514, -13.422, -13.349, -13.267,
    -13.214, -13.388, -13.514, -13.665, -13.825, -13.957, -13.472, -13.382,
    -13.319, -13.246, -13.199, -13.372, -13.502, -13.657, -13.817, -13.949,
    -13.430, -13.342, -13.287, -13.226, -13.187, -13.357, -13.490, -13.649,
    -13.809, -13.941, -13.389, -13.301, -13.254, -13.206, -13.177, -13.343,
    -13.479, -13.642, -13.801, -13.933, -13.349, -13.260, -13.219, -13.188,
    -13.170, -13.328, -13.468, -13.634, -13.792, -13.925, -13.310, -13.219,
    -13.183, -13.169, -13.165, -13.313, -13.457, -13.627, -13.783, -13.916,
    -13.272, -13.177, -13.146, -13.151, -13.163, -13.299, -13.446, -13.621,
    -13.774, -13.908, -13.234, -13.135, -13.106, -13.134, -13.163, -13.285,
    -13.436, -13.614, -13.765, -13.899, -13.197, -13.093, -13.066, -13.117,
    -13.166, -13.271, -13.426, -13.608, -13.755, -13.890, -13.161, -13.050,
    -13.024, -13.101, -13.172, -13.257, -13.416, -13.602, -13.745, -13.881,
    -13.121, -13.015, -13.000, -13.083, -13.158, -13.244, -13.405, -13.593,
    -13.738, -13.874, -13.082, -12.981, -12.979, -13.064, -13.142, -13.232,
    -13.394, -13.584, -13.730, -13.867, -13.042, -12.948, -12.960, -13.046,
    -13.125, -13.220, -13.383, -13.575, -13.723, -13.861, -13.003, -12.916,
    -12.943, -13.028, -13.107, -13.208, -13.372, -13.565, -13.717, -13.855,
    -12.963, -12.885, -12.928, -13.010, -13.087, -13.197, -13.361, -13.556,
    -13.710, -13.849, -12.923, -12.855, -12.915, -12.992, -13.066, -13.186,
    -13.350, -13.545, -13.704, -13.843, -12.884, -12.827, -12.905, -12.974,
    -13.043, -13.176, -13.338, -13.535, -13.698, -13.838, -12.844, -12.799,
    -12.896, -12.956, -13.018, -13.165, -13.327, -13.524, -13.693, -13.833,
    -12.805, -12.772, -12.890, -12.938, -12.992, -13.156, -13.316, -13.513,
    -13.687, -13.828, -12.765, -12.747, -12.886, -12.920, -12.965, -13.146,
    -13.305, -13.502, -13.682, -13.823, -12.736, -12.729, -12.873, -12.908,
    -12.955, -13.138, -13.298, -13.495, -13.676, -13.817, -12.708, -12.713,
    -12.861, -12.897, -12.946, -13.130, -13.292, -13.488, -13.670, -13.811,
    -12.682, -12.700, -12.848, -12.887, -12.940, -13.123, -13.286, -13.482,
    -13.664, -13.806, -12.659, -12.689, -12.836, -12.878, -12.935, -13.117,
    -13.281, -13.476, -13.659, -13.800, -12.637, -12.680, -12.824, -12.871,
    -12.932, -13.111, -13.277, -13.471, -13.653, -13.794, -12.617, -12.673,
    -12.812, -12.864, -12.931, -13.106, -13.274, -13.466, -13.648, -13.788,
    -12.598, -12.668, -12.800, -12.859, -12.932, -13.101, -13.271, -13.462,
    -13.642, -13.782, -12.582, -12.666, -12.789, -12.854, -12.935, -13.097,
    -13.269, -13.458, -13.637, -13.777, -12.568, -12.666, -12.777, -12.851,
    -12.939, -13.093, -13.268, -13.455, -13.631, -13.771, -12.555, -12.668,
    -12.766, -12.849, -12.946, -13.090, -13.268, -13.452, -13.626, -13.765,
    -12.551, -12.665, -12.761, -12.844, -12.943, -13.086, -13.264, -13.447,
    -13.621, -13.759, -12.550, -12.664, -12.757, -12.840, -12.939, -13.082,
    -13.259, -13.442, -13.615, -13.753, -12.553, -12.663, -12.755, -12.837,
    -12.936, -13.078, -13.255, -13.436, -13.610, -13.748, -12.558, -12.664,
    -12.754, -12.834, -12.932, -13.075, -13.250, -13.431, -13.604, -13.742,
    -12.567, -12.665, -12.754, -12.831, -12.928, -13.072, -13.245, -13.426,
    -13.599, -13.736, -12.566, -12.663, -12.751, -12.828, -12.925, -13.068,
    -13.241, -13.421, -13.594, -13.730, -12.564, -12.660, -12.748, -12.825,
    -12.922, -13.065, -13.236, -13.416, -13.588, -13.724, -12.561, -12.657,
    -12.745, -12.823, -12.919, -13.061, -13.232, -13.411, -13.583, -13.718,
    -12.556, -12.654, -12.742, -12.820, -12.916, -13.058, -13.228, -13.407,
    -13.577, -13.712, -12.551, -12.650, -12.739, -12.818, -12.914, -13.054,
    -13.224, -13.402, -13.572, -13.706, -12.550, -12.649, -12.737, -12.816,
    -12.912, -13.051, -13.220, -13.397, -13.567, -13.700, -12.549, -12.648,
    -12.736, -12.815, -12.910, -13.048, -13.217, -13.393, -13.561, -13.694,
    -12.549, -12.648, -12.736, -12.814, -12.908, -13.046, -13.213, -13.389,
    -13.556, -13.688, -12.550, -12.648, -12.736, -12.813, -12.907, -13.043,
    -13.210, -13.384, -13.551, -13.682, -12.552, -12.649, -12.736, -12.813,
    -12.906, -13.041, -13.207, -13.380, -13.546, -13.677, -12.554, -12.650,
    -12.737, -12.813, -12.905, -13.039, -13.204, -13.376, -13.540, -13.671,
    -12.557, -12.653, -12.738, -12.813, -12.904, -13.037, -13.201, -13.372,
    -13.535, -13.665, -12.560, -12.655, -12.740, -12.814, -12.903, -13.036,
    -13.199, -13.368, -13.530, -13.659, -12.564, -12.658, -12.742, -12.815,
    -12.903, -13.034, -13.196, -13.365, -13.525, -13.654, -12.569, -12.662,
    -12.745, -12.816, -12.903, -13.033, -13.194, -13.361, -13.520, -13.648,
    -12.576, -12.667, -12.749, -12.819, -12.904, -13.033, -13.192, -13.358,
    -13.515, -13.642, -12.584, -12.674, -12.754, -12.822, -12.906, -13.033,
    -13.191, -13.354, -13.510, -13.637, -12.593, -12.681, -12.759, -12.825,
    -12.907, -13.033, -13.189, -13.351, -13.506, -13.631, -12.603, -12.689,
    -12.764, -12.829, -12.910, -13.034, -13.188, -13.348, -13.501, -13.626,
    -12.614, -12.697, -12.771, -12.834, -12.912, -13.034, -13.187, -13.345,
    -13.496, -13.621, -12.626, -12.707, -12.778, -12.839, -12.916, -13.036,
    -13.186, -13.342, -13.492, -13.615, -12.639, -12.717, -12.785, -12.844,
    -12.919, -13.037, -13.185, -13.339, -13.487, -13.610, -12.653, -12.728,
    -12.794, -12.850, -12.923, -13.039, -13.185, -13.337, -13.483, -13.604,
    -12.668, -12.739, -12.803, -12.857, -12.927, -13.041, -13.185, -13.334,
    -13.478, -13.599, -12.684, -12.752, -12.812, -12.864, -12.932, -13.044,
    -13.185, -13.332, -13.474, -13.594, -12.698, -12.763, -12.821, -12.871,
    -12.937, -13.046, -13.185, -13.330, -13.470, -13.589, -12.712, -12.775,
    -12.830, -12.878, -12.942, -13.049, -13.185, -13.327, -13.466, -13.583,
    -12.727, -12.787, -12.840, -12.885, -12.947, -13.052, -13.185, -13.325,
    -13.462, -13.578, -12.742, -12.799, -12.850, -12.893, -12.953, -13.055,
    -13.186, -13.323, -13.458, -13.573, -12.758, -12.812, -12.860, -12.901,
    -12.958, -13.058, -13.186, -13.321, -13.454, -13.567, -12.774, -12.826,
    -12.871, -12.910, -12.965, -13.061, -13.187, -13.319, -13.450, -13.562,
    -12.790, -12.839, -12.883, -12.918, -12.971, -13.065, -13.188, -13.318,
    -13.447, -13.557, -12.807, -12.854, -12.894, -12.928, -12.978, -13.069,
    -13.189, -13.316, -13.443, -13.552, -12.824, -12.868, -12.906, -12.937,
    -12.985, -13.073, -13.190, -13.314, -13.439, -13.546, -12.842, -12.883,
    -12.919, -12.947, -12.993, -13.077, -13.191, -13.313, -13.436, -13.541,
    -12.859, -12.897, -12.931, -12.956, -13.000, -13.081, -13.192, -13.311,
    -13.432, -13.536, -12.876, -12.912, -12.943, -12.966, -13.007, -13.085,
    -13.193, -13.310, -13.429, -13.531, -12.893, -12.926, -12.955, -12.976,
    -13.014, -13.089, -13.195, -13.308, -13.425, -13.526, -12.910, -12.941,
    -12.967, -12.986, -13.021, -13.094, -13.196, -13.307, -13.421, -13.521,
    -12.927, -12.956, -12.980, -12.996, -13.028, -13.098, -13.197, -13.306,
    -13.417, -13.516, -12.945, -12.972, -12.992, -13.006, -13.036, -13.103,
    -13.199, -13.304, -13.414, -13.511, -12.963, -12.987, -13.005, -13.017,
    -13.043, -13.108, -13.201, -13.303, -13.410, -13.506, -12.981, -13.003,
    -13.018, -13.028, -13.051, -13.113, -13.202, -13.302, -13.406, -13.501,
    -12.999, -13.019, -13.032, -13.039, -13.059, -13.118, -13.204, -13.301,
    -13.403, -13.496, -13.017, -13.035, -13.045, -13.050, -13.067, -13.123,
    -13.206, -13.300, -13.399, -13.491, -13.035, -13.051, -13.058, -13.061,
    -13.075, -13.128, -13.208, -13.299, -13.395, -13.486, -13.053, -13.066,
    -13.071, -13.071, -13.083, -13.133, -13.209, -13.297, -13.392, -13.481,
    -13.070, -13.082, -13.084, -13.082, -13.091, -13.137, -13.211, -13.296,
    -13.388, -13.476, -13.088, -13.097, -13.098, -13.093, -13.099, -13.142,
    -13.212, -13.295, -13.384, -13.470, -13.106, -13.113, -13.111, -13.103,
    -13.107, -13.147, -13.214, -13.293, -13.381, -13.465, -13.124, -13.129,
    -13.124, -13.114, -13.115, -13.152, -13.215, -13.292, -13.377, -13.460,
    -13.142, -13.145, -13.138, -13.125, -13.124, -13.157, -13.217, -13.290,
    -13.373, -13.455, -13.161, -13.161, -13.152, -13.136, -13.132, -13.162,
    -13.219, -13.289, -13.369, -13.450, -13.179, -13.177, -13.165, -13.147,
    -13.140, -13.167, -13.220, -13.287, -13.366, -13.444, -13.197, -13.193,
    -13.179, -13.158, -13.149, -13.172, -13.222, -13.286, -13.362, -13.439,
    -13.215, -13.209, -13.193, -13.169, -13.157, -13.177, -13.223, -13.284,
    -13.358, -13.434, -13.233, -13.225, -13.206, -13.180, -13.165, -13.181,
    -13.225, -13.283, -13.354, -13.428, -13.250, -13.240, -13.220, -13.191,
    -13.173, -13.186, -13.226, -13.281, -13.350, -13.423, -13.268, -13.256,
    -13.233, -13.202, -13.181, -13.190, -13.227, -13.279, -13.346, -13.418,
    -13.286, -13.272, -13.247, -13.213, -13.189, -13.194, -13.228, -13.277,
    -13.342, -13.413, -13.304, -13.288, -13.260, -13.224, -13.197, -13.199,
    -13.229, -13.275, -13.337, -13.407, -13.321, -13.304, -13.274, -13.235,
    -13.205, -13.203, -13.230, -13.273, -13.333, -13.402, -13.339, -13.319,
    -13.288, -13.247, -13.212, -13.207, -13.231, -13.271, -13.329, -13.397,
    -13.356, -13.335, -13.301, -13.258, -13.220, -13.211, -13.232, -13.269,
    -13.324, -13.391, -13.374, -13.351, -13.315, -13.269, -13.228, -13.215,
    -13.233, -13.267, -13.320, -13.386, -13.391, -13.367, -13.328, -13.280,
    -13.236, -13.219, -13.234, -13.265, -13.316, -13.381, -13.409, -13.382,
    -13.341, -13.290, -13.243, -13.223, -13.234, -13.262, -13.311, -13.375,
    -13.426, -13.398, -13.355, -13.301, -13.250, -13.227, -13.235, -13.260,
    -13.307, -13.370, -13.443, -13.413, -13.368, -13.311, -13.258, -13.231,
    -13.235, -13.257, -13.302, -13.364, -13.460, -13.429, -13.381, -13.322,
    -13.265, -13.235, -13.236, -13.255, -13.297, -13.358, -13.477, -13.444,
    -13.394, -13.332, -13.272, -13.239, -13.236, -13.252, -13.293, -13.353,
    -13.494, -13.459, -13.407, -13.342, -13.279, -13.243, -13.236, -13.249,
    -13.288, -13.347, -13.511, -13.475, -13.419, -13.352, -13.286, -13.247,
    -13.236, -13.246, -13.284, -13.341, -13.528, -13.490, -13.432, -13.362,
    -13.293, -13.250, -13.236, -13.243, -13.279, -13.336, -13.545, -13.505,
    -13.445, -13.372, -13.300, -13.254, -13.236, -13.240, -13.274, -13.330,
    -13.562, -13.520, -13.458, -13.382, -13.306, -13.257, -13.236, -13.236,
    -13.269, -13.324, -13.578, -13.535, -13.470, -13.391, -13.313, -13.260,
    -13.235, -13.233, -13.264, -13.318, -13.595, -13.549, -13.482, -13.401,
    -13.319, -13.263, -13.234, -13.229, -13.258, -13.312, -13.611, -13.564,
    -13.494, -13.410, -13.325, -13.265, -13.233, -13.225, -13.253, -13.306,
    -13.627, -13.579, -13.507, -13.419, -13.331, -13.268, -13.232, -13.220,
    -13.248, -13.299, -13.644, -13.593, -13.519, -13.428, -13.336, -13.270,
    -13.231, -13.216, -13.242, -13.293, -13.660, -13.607, -13.531, -13.437,
    -13.342, -13.272, -13.230, -13.212, -13.236, -13.287, -13.676, -13.622,
    -13.543, -13.446, -13.348, -13.274, -13.228, -13.207, -13.231, -13.280,
    -13.692, -13.636, -13.554, -13.454, -13.353, -13.275, -13.227, -13.202,
    -13.225, -13.274, -13.708, -13.650, -13.566, -13.463, -13.358, -13.277,
    -13.225, -13.197, -13.219, -13.267, -13.724, -13.664, -13.577, -13.471,
    -13.363, -13.278, -13.223, -13.193, -13.213, -13.261, -13.739, -13.678,
    -13.589, -13.479, -13.367, -13.279, -13.221, -13.189, -13.208, -13.254,
    -13.755, -13.691, -13.600, -13.487, -13.372, -13.280, -13.219, -13.184,
    -13.202, -13.248, -13.770, -13.705, -13.611, -13.495, -13.376, -13.281,
    -13.216, -13.180, -13.196, -13.242, -13.786, -13.718, -13.622, -13.502,
    -13.380, -13.281, -13.213, -13.176, -13.190, -13.235, -13.801, -13.731,
    -13.632, -13.509, -13.384, -13.282, -13.211, -13.172, -13.184, -13.229,
    -13.816, -13.744, -13.643, -13.517, -13.387, -13.282, -13.208, -13.168,
    -13.178, -13.222, -13.831, -13.757, -13.653, -13.524, -13.391, -13.282,
    -13.205, -13.164, -13.172, -13.216, -13.846, -13.770, -13.664, -13.530,
    -13.394, -13.281, -13.201, -13.160, -13.166, -13.209, -13.861, -13.783,
    -13.674, -13.537, -13.397, -13.281, -13.198, -13.156, -13.160, -13.203,
    -13.876, -13.795, -13.683, -13.543, -13.399, -13.280, -13.194, -13.151,
    -13.153, -13.196, -13.890, -13.807, -13.693, -13.549, -13.402, -13.279,
    -13.189, -13.145, -13.147, -13.189, -13.904, -13.819, -13.702, -13.554,
    -13.404, -13.278, -13.185, -13.140, -13.140, -13.183, -13.919, -13.831,
    -13.710, -13.559, -13.405, -13.276, -13.180, -13.134, -13.133, -13.176,
    -13.933, -13.842, -13.719, -13.564, -13.407, -13.275, -13.175, -13.128,
    -13.126, -13.169, -13.947, -13.854, -13.727, -13.568, -13.408, -13.273,
    -13.169, -13.122, -13.119, -13.162, -13.960, -13.865, -13.735, -13.573,
    -13.409, -13.271, -13.164, -13.115, -13.112, -13.155, -13.974, -13.876,
    -13.742, -13.577, -13.409, -13.268, -13.158, -13.109, -13.105, -13.147,
    -13.988, -13.886, -13.750, -13.581, -13.410, -13.266, -13.151, -13.102,
    -13.097, -13.140, -14.001, -13.897, -13.757, -13.584, -13.410, -13.263,
    -13.145, -13.095, -13.090, -13.133, -14.014, -13.907, -13.763, -13.587,
    -13.409, -13.259, -13.140, -13.088, -13.082, -13.126, -14.027, -13.918,
    -13.770, -13.589, -13.407, -13.255, -13.135, -13.081, -13.075, -13.119,
    -14.039, -13.927, -13.775, -13.591, -13.405, -13.250, -13.129, -13.074,
    -13.067, -13.112, -14.051, -13.937, -13.781, -13.593, -13.402, -13.245,
    -13.124, -13.067, -13.059, -13.104, -14.063, -13.947, -13.786, -13.594,
    -13.399, -13.239, -13.119, -13.060, -13.051, -13.097, -14.075, -13.956,
    -13.791, -13.595, -13.395, -13.233, -13.114, -13.053, -13.043, -13.090,
    -14.087, -13.965, -13.795, -13.596, -13.391, -13.227, -13.109, -13.045,
    -13.035, -13.083, -14.099, -13.975, -13.800, -13.596, -13.386, -13.221,
    -13.104, -13.038, -13.027, -13.076, -14.110, -13.983, -13.803, -13.596,
    -13.381, -13.213, -13.099, -13.030, -13.018, -13.068, -14.121, -13.992,
    -13.807, -13.596, -13.375, -13.206, -13.094, -13.022, -13.010, -13.061,
    -14.131, -13.999, -13.809, -13.595, -13.371, -13.200, -13.087, -13.014,
    -13.002, -13.054, -14.141, -14.005, -13.811, -13.593, -13.367, -13.194,
    -13.080, -13.006, -12.994, -13.046, -14.151, -14.011, -13.811, -13.591,
    -13.363, -13.188, -13.073, -12.998, -12.987, -13.039, -14.161, -14.017,
    -13.812, -13.589, -13.359, -13.182, -13.065, -12.990, -12.979, -13.032,
    -14.170, -14.022, -13.812, -13.586, -13.355, -13.176, -13.058, -12.982,
    -12.971, -13.024, -14.179, -14.026, -13.811, -13.583, -13.351, -13.170,
    -13.049, -12.974, -12.963, -13.017, -14.187, -14.031, -13.810, -13.579,
    -13.347, -13.164, -13.041, -12.966, -12.956, -13.009, -14.195, -14.034,
    -13.808, -13.575, -13.343, -13.158, -13.032, -12.957, -12.948, -13.002,
    -14.203, -14.037, -13.806, -13.570, -13.338, -13.151, -13.023, -12.949,
    -12.941, -12.994, -14.211, -14.040, -13.803, -13.565, -13.334, -13.145,
    -13.014, -12.940, -12.933, -12.987, -14.217, -14.041, -13.800, -13.559,
    -13.327, -13.137, -13.005, -12.931, -12.925, -12.979, -14.223, -14.041,
    -13.797, -13.553, -13.319, -13.128, -12.996, -12.922, -12.916, -12.971,
    -14.229, -14.041, -13.793, -13.546, -13.310, -13.119, -12.986, -12.913,
    -12.908, -12.963, -14.234, -14.040, -13.789, -13.538, -13.301, -13.110,
    -12.977, -12.903, -12.899, -12.955, -14.238, -14.038, -13.785, -13.530,
    -13.291, -13.100, -12.967, -12.894, -12.890, -12.947, -14.242, -14.035,
    -13.780, -13.522, -13.281, -13.090, -12.957, -12.884, -12.881, -12.939,
    -14.245, -14.032, -13.775, -13.513, -13.270, -13.079, -12.947, -12.874,
    -12.872, -12.930, -14.248, -14.028, -13.769, -13.503, -13.259, -13.068,
    -12.936, -12.864, -12.863, -12.922, -14.251, -14.023, -13.763, -13.493,
    -13.247, -13.057, -12.926, -12.854, -12.853, -12.914, -14.253, -14.018,
    -13.756, -13.483, -13.234, -13.045, -12.915, -12.844, -12.844, -12.905,
    -14.253, -14.012, -13.746, -13.472, -13.223, -13.034, -12.904, -12.834,
    -12.835, -12.897, -14.252, -14.006, -13.736, -13.460, -13.211, -13.022,
    -12.893, -12.823, -12.826, -12.889, -14.251, -13.999, -13.725, -13.448,
    -13.199, -13.010, -12.882, -12.813, -12.817, -12.881, -14.249, -13.991,
    -13.712, -13.436, -13.187, -12.998, -12.871, -12.803, -12.808, -12.873,
    -14.246, -13.983, -13.699, -13.423, -13.174, -12.986, -12.859, -12.792,
    -12.799, -12.866, -14.242, -13.974, -13.685, -13.409, -13.162, -12.973,
    -12.848, -12.781, -12.790, -12.858, -14.238, -13.965, -13.670, -13.395,
    -13.149, -12.961, -12.836, -12.770, -12.781, -12.850, -14.233, -13.954,
    -13.654, -13.381, -13.135, -12.948, -12.824, -12.759, -12.771, -12.842,
    -14.227, -13.943, -13.638, -13.366, -13.122, -12.934, -12.812, -12.748,
    -12.762, -12.835, -14.220, -13.932, -13.620, -13.350, -13.108, -12.921,
    -12.800, -12.737, -12.753, -12.827, -14.211, -13.918, -13.604, -13.334,
    -13.093, -12.908, -12.788, -12.726, -12.744, -12.819, -14.200, -13.903,
    -13.588, -13.318, -13.078, -12.894, -12.776, -12.715, -12.734, -12.810,
    -14.189, -13.887, -13.571, -13.301, -13.063, -12.881, -12.763, -12.704,
    -12.724, -12.802, -14.176, -13.870, -13.554, -13.283, -13.047, -12.867,
    -12.751, -12.693, -12.715, -12.794, -14.163, -13.852, -13.537, -13.265,
    -13.031, -12.853, -12.738, -12.682, -12.705, -12.785, -14.148, -13.833,
    -13.519, -13.247, -13.014, -12.839, -12.726, -12.670, -12.695, -12.776,
    -14.132, -13.813, -13.500, -13.228, -12.997, -12.825, -12.713, -12.659,
    -12.685, -12.768, -14.115, -13.793, -13.482, -13.209, -12.980, -12.811,
    -12.700, -12.648, -12.675, -12.759, -14.097, -13.771, -13.463, -13.189,
    -12.962, -12.797, -12.687, -12.636, -12.665, -12.750, -14.078, -13.748,
    -13.443, -13.169, -12.944, -12.782, -12.674, -12.625, -12.655, -12.741,
    -14.056, -13.722, -13.420, -13.149, -12.926, -12.765, -12.661, -12.614,
    -12.646, -12.733, -14.032, -13.694, -13.397, -13.128, -12.909, -12.747,
    -12.648, -12.604, -12.637, -12.725, -14.006, -13.665, -13.372, -13.107,
    -12.891, -12.729, -12.634, -12.593, -12.628, -12.717, -13.979, -13.634,
    -13.347, -13.086, -12.873, -12.710, -12.621, -12.583, -12.619, -12.709,
    -13.951, -13.602, -13.320, -13.064, -12.854, -12.690, -12.608, -12.572,
    -12.610, -12.702, -13.921, -13.568, -13.293, -13.042, -12.836, -12.670,
    -12.594, -12.562, -12.601, -12.694, -13.889, -13.534, -13.265, -13.019,
    -12.817, -12.649, -12.580, -12.552, -12.592, -12.687, -13.856, -13.497,
    -13.236, -12.996, -12.798, -12.628, -12.567, -12.542, -12.584, -12.679,
    -13.822, -13.459, -13.206, -12.973, -12.778, -12.606, -12.553, -12.532,
    -12.575, -12.672, -13.786, -13.420, -13.175, -12.949, -12.759, -12.583,
    -12.539, -12.522, -12.567, -12.665, -13.749, -13.389, -13.148, -12.926,
    -12.740, -12.567, -12.525, -12.511, -12.557, -12.655, -13.710, -13.358,
    -13.120, -12.902, -12.720, -12.552, -12.512, -12.500, -12.548, -12.645,
    -13.670, -13.327, -13.093, -12.878, -12.701, -12.537, -12.498, -12.489,
    -12.538, -12.635, -13.628, -13.297, -13.066, -12.854, -12.681, -12.523,
    -12.484, -12.477, -12.528, -12.624, -13.585, -13.266, -13.038, -12.830,
    -12.661, -12.510, -12.470, -12.466, -12.518, -12.613, -13.540, -13.237,
    -13.010, -12.805, -12.641, -12.498, -12.457, -12.454, -12.508, -12.601,
    -13.494, -13.207, -12.983, -12.781, -12.621, -12.486, -12.443, -12.442,
    -12.498, -12.589, -13.446, -13.178, -12.955, -12.756, -12.601, -12.475,
    -12.429, -12.430, -12.487, -12.577, -13.397, -13.149, -12.927, -12.730,
    -12.581, -12.465, -12.415, -12.418, -12.477, -12.564, -13.346, -13.120,
    -12.899, -12.705, -12.560, -12.455, -12.401, -12.406, -12.466, -12.551,
    -13.303, -13.087, -12.868, -12.680, -12.535, -12.440, -12.387, -12.395,
    -12.458, -12.544, -13.261, -13.053, -12.837, -12.655, -12.508, -12.425,
    -12.374, -12.384, -12.449, -12.538, -13.218, -13.018, -12.806, -12.630,
    -12.481, -12.410, -12.360, -12.373, -12.442, -12.532, -13.176, -12.983,
    -12.774, -12.605, -12.452, -12.394, -12.347, -12.362, -12.434, -12.527,
    -13.135, -12.947, -12.741, -12.580, -12.423, -12.378, -12.333, -12.351,
    -12.427, -12.523, -13.093, -12.910, -12.708, -12.555, -12.392, -12.361,
    -12.320, -12.340, -12.420, -12.520, -13.052, -12.873, -12.674, -12.529,
    -12.361, -12.345, -12.306, -12.329, -12.413, -12.518, -13.011, -12.835,
    -12.640, -12.504, -12.328, -12.327, -12.293, -12.318, -12.407, -12.516,
    -12.970, -12.797, -12.605, -12.479, -12.295, -12.310, -12.279, -12.308,
    -12.401, -12.515, -12.929, -12.758, -12.570, -12.453, -12.260, -12.292,
    -12.266, -12.297, -12.395, -12.515, -12.886, -12.721, -12.539, -12.429,
    -12.242, -12.277, -12.253, -12.287, -12.387, -12.511, -12.843, -12.684,
    -12.509, -12.404, -12.225, -12.261, -12.241, -12.278, -12.379, -12.506,
    -12.800, -12.647, -12.479, -12.380, -12.211, -12.246, -12.229, -12.269,
    -12.370, -12.502, -12.757, -12.609, -12.450, -12.356, -12.199, -12.231,
    -12.217, -12.260, -12.362, -12.497, -12.713, -12.572, -12.421, -12.332,
    -12.189, -12.216, -12.205, -12.251, -12.353, -12.492, -12.669, -12.534,
    -12.392, -12.309, -12.181, -12.202, -12.193, -12.243, -12.344, -12.487,
    -12.625, -12.496, -12.364, -12.285, -12.176, -12.187, -12.182, -12.234,
    -12.335, -12.482, -12.581, -12.458, -12.336, -12.261, -12.172, -12.173,
    -12.170, -12.226, -12.326, -12.477, -12.537, -12.420, -12.309, -12.238,
    -12.170, -12.159, -12.159, -12.219, -12.317, -12.471, -12.492, -12.382,
    -12.282, -12.215, -12.171, -12.145, -12.148, -12.211, -12.307, -12.466,
    -12.451, -12.341, -12.255, -12.193, -12.157, -12.133, -12.138, -12.203,
    -12.296, -12.459, -12.411, -12.299, -12.228, -12.172, -12.141, -12.121,
    -12.128, -12.196, -12.285, -12.452, -12.371, -12.256, -12.201, -12.151,
    -12.125, -12.109, -12.118, -12.189, -12.274, -12.445, -12.332, -12.213,
    -12.175, -12.130, -12.108, -12.098, -12.108, -12.181, -12.263, -12.438,
    -12.294, -12.169, -12.148, -12.110, -12.089, -12.088, -12.099, -12.174,
    -12.251, -12.430, -12.256, -12.124, -12.123, -12.090, -12.070, -12.077,
    -12.090, -12.168, -12.238, -12.422, -12.218, -12.079, -12.097, -12.071,
    -12.050, -12.068, -12.081, -12.161, -12.226, -12.413, -12.181, -12.033,
    -12.072, -12.052, -12.029, -12.058, -12.073, -12.154, -12.213, -12.404,
    -12.145, -11.986, -12.047, -12.033, -12.006, -12.049, -12.065, -12.148,
    -12.200, -12.395, -12.109, -11.939, -12.022, -12.015, -11.983, -12.041,
    -12.057, -12.142, -12.186, -12.386, -12.072, -11.908, -11.995, -11.997,
    -11.969, -12.021, -12.048, -12.134, -12.181, -12.381, -12.035, -11.880,
    -11.967, -11.978, -11.956, -11.999, -12.040, -12.126, -12.177, -12.376,
    -11.999, -11.854, -11.940, -11.961, -11.945, -11.975, -12.031, -12.117,
    -12.174, -12.372, -11.963, -11.831, -11.912, -11.943, -11.934, -11.950,
    -12.023, -12.108, -12.173, -12.369, -11.927, -11.810, -11.884, -11.926,
    -11.924, -11.923, -12.014, -12.099, -12.173, -12.366, -11.891, -11.792,
    -11.856, -11.909, -11.914, -11.894, -12.006, -12.090, -12.174, -12.363,
    -11.855, -11.776, -11.828, -11.892, -11.906, -11.863, -11.997, -12.081,
    -12.177, -12.362, -11.820, -11.762, -11.799, -11.876, -11.899, -11.830,
    -11.989, -12.071, -12.181, -12.360, -11.785, -11.751, -11.770, -11.860,
    -11.892, -11.796, -11.981, -12.061, -12.186, -12.359, -11.750, -11.743,
    -11.741, -11.844, -11.887, -11.760, -11.973, -12.051, -12.192, -12.359,
    -11.719, -11.725, -11.724, -11.830, -11.878, -11.753, -11.967, -12.046,
    -12.191, -12.356, -11.688, -11.707, -11.709, -11.817, -11.869, -11.750,
    -11.962, -12.042, -12.189, -12.354, -11.658, -11.689, -11.697, -11.805,
    -11.860, -11.751, -11.958, -12.038, -12.188, -12.351, -11.629, -11.672,
    -11.686, -11.794, -11.851, -11.756, -11.954, -12.036, -12.186, -12.348,
    -11.601, -11.655, -11.677, -11.783, -11.842, -11.765, -11.950, -12.034,
    -12.183, -12.346, -11.573, -11.638, -11.671, -11.772, -11.834, -11.777,
    -11.947, -12.032, -12.181, -12.343, -11.546, -11.621, -11.666, -11.762,
    -11.825, -11.794, -11.945, -12.032, -12.178, -12.341, -11.520, -11.605,
    -11.664, -11.753, -11.817, -11.814, -11.943, -12.032, -12.174, -12.338,
    -11.495, -11.589, -11.663, -11.745, -11.808, -11.838, -11.941, -12.033,
    -12.171, -12.336, -11.471, -11.573, -11.665, -11.737, -11.800, -11.866,
    -11.940, -12.035, -12.167, -12.333, -11.458, -11.565, -11.661, -11.731,
    -11.795, -11.870, -11.937, -12.033, -12.165, -12.330, -11.448, -11.558,
    -11.657, -11.725, -11.790, -11.872, -11.935, -12.031, -12.162, -12.328,
    -11.441, -11.553, -11.653, -11.720, -11.785, -11.873, -11.932, -12.029,
    -12.160, -12.325, -11.437, -11.550, -11.651, -11.717, -11.782, -11.873,
    -11.930, -12.027, -12.158, -12.323, -11.435, -11.548, -11.648, -11.714,
    -11.779, -11.871, -11.928, -12.024, -12.155, -12.320, -11.437, -11.547,
    -11.647, -11.711, -11.776, -11.868, -11.925, -12.022, -12.153, -12.318,
    -11.441, -11.549, -11.646, -11.710, -11.774, -11.864, -11.923, -12.020,
    -12.151, -12.315, -11.447, -11.551, -11.645, -11.709, -11.773, -11.859,
    -11.922, -12.017, -12.148, -12.313, -11.457, -11.556, -11.645, -11.709,
    -11.772, -11.852, -11.920, -12.015, -12.146, -12.310, -11.469, -11.562,
    -11.646, -11.710, -11.772, -11.844, -11.918, -12.012, -12.144, -12.308,
    -11.469, -11.562, -11.645, -11.709, -11.771, -11.843, -11.916, -12.010,
    -12.142, -12.306, -11.470, -11.562, -11.645, -11.708, -11.769, -11.841,
    -11.915, -12.009, -12.140, -12.304, -11.470, -11.562, -11.645, -11.707,
    -11.768, -11.840, -11.914, -12.007, -12.138, -12.301, -11.470, -11.562,
    -11.645, -11.707, -11.768, -11.839, -11.913, -12.006, -12.137, -12.299,
    -11.471, -11.562, -11.645, -11.707, -11.767, -11.839, -11.912, -12.004,
    -12.135, -12.297, -11.472, -11.563, -11.645, -11.707, -11.767, -11.838,
    -11.911, -12.003, -12.134, -12.295, -11.472, -11.564, -11.646, -11.708,
    -11.767, -11.838, -11.910, -12.002, -12.132, -12.293, -11.473, -11.565,
    -11.647, -11.709, -11.767, -11.837, -11.909, -12.001, -12.131, -12.292,
    -11.474, -11.566, -11.648, -11.710, -11.767, -11.837, -11.908, -12.000,
    -12.129, -12.290, -11.475, -11.567, -11.649, -11.711, -11.768, -11.837,
    -11.908, -11.999, -12.128, -12.288, -11.481, -11.572, -11.653, -11.714,
    -11.770, -11.838, -11.908, -11.999, -12.127, -12.286, -11.488, -11.577,
    -11.657, -11.717, -11.772, -11.839, -11.909, -11.999, -12.126, -12.285,
    -11.495, -11.583, -11.662, -11.720, -11.774, -11.841, -11.910, -11.999,
    -12.126, -12.284, -11.504, -11.590, -11.667, -11.724, -11.777, -11.843,
    -11.911, -11.999, -12.125, -12.283, -11.514, -11.598, -11.673, -11.729,
    -11.781, -11.845, -11.912, -11.999, -12.124, -12.281, -11.524, -11.606,
    -11.680, -11.734, -11.784, -11.847, -11.914, -12.000, -12.124, -12.280,
    -11.536, -11.616, -11.687, -11.739, -11.788, -11.850, -11.915, -12.001,
    -12.124, -12.279, -11.548, -11.626, -11.695, -11.745, -11.793, -11.853,
    -11.917, -12.002, -12.124, -12.278, -11.562, -11.636, -11.703, -11.751,
    -11.798, -11.856, -11.920, -12.003, -12.124, -12.278, -11.576, -11.648,
    -11.712, -11.758, -11.803, -11.860, -11.922, -12.004, -12.124, -12.277,
    -11.589, -11.659, -11.721, -11.765, -11.809, -11.864, -11.925, -12.006,
    -12.124, -12.276, -11.603, -11.670, -11.730, -11.773, -11.814, -11.868,
    -11.928, -12.007, -12.125, -12.276, -11.618, -11.682, -11.740, -11.780,
    -11.821, -11.873, -11.931, -12.009, -12.126, -12.276, -11.633, -11.695,
    -11.750, -11.789, -11.827, -11.878, -11.934, -12.011, -12.127, -12.275,
    -11.648, -11.708, -11.761, -11.798, -11.834, -11.883, -11.938, -12.014,
    -12.128, -12.275, -11.664, -11.722, -11.772, -11.807, -11.842, -11.889,
    -11.942, -12.016, -12.129, -12.275, -11.681, -11.736, -11.783, -11.816,
    -11.850, -11.895, -11.946, -12.019, -12.130, -12.275, -11.699, -11.750,
    -11.796, -11.827, -11.858, -11.901, -11.950, -12.022, -12.132, -12.275,
    -11.716, -11.765, -11.808, -11.837, -11.866, -11.907, -11.955, -12.025,
    -12.133, -12.276, -11.735, -11.781, -11.821, -11.848, -11.875, -11.914,
    -11.960, -12.028, -12.135, -12.276, -11.752, -11.796, -11.833, -11.858,
    -11.883, -11.920, -11.965, -12.031, -12.137, -12.276, -11.769, -11.810,
    -11.846, -11.869, -11.892, -11.927, -11.970, -12.034, -12.139, -12.277,
    -11.787, -11.826, -11.859, -11.879, -11.901, -11.934, -11.975, -12.038,
    -12.141, -12.278, -11.805, -11.841, -11.872, -11.890, -11.910, -11.941,
    -11.980, -12.042, -12.143, -12.278, -11.823, -11.856, -11.885, -11.902,
    -11.919, -11.948, -11.985, -12.045, -12.145, -12.279, -11.841, -11.872,
    -11.898, -11.913, -11.928, -11.956, -11.991, -12.049, -12.147, -12.280,
    -11.859, -11.888, -11.912, -11.925, -11.938, -11.963, -11.997, -12.053,
    -12.150, -12.281, -11.878, -11.905, -11.926, -11.937, -11.948, -11.971,
    -12.003, -12.057, -12.152, -12.282, -11.897, -11.921, -11.940, -11.949,
    -11.958, -11.979, -12.009, -12.062, -12.155, -12.284, -11.916, -11.938,
    -11.955, -11.961, -11.968, -11.987, -12.015, -12.066, -12.158, -12.285,
    -11.934, -11.954, -11.969, -11.973, -11.978, -11.995, -12.021, -12.070,
    -12.161, -12.286, -11.953, -11.971, -11.983, -11.985, -11.988, -12.003,
    -12.027, -12.075, -12.164, -12.288, -11.972, -11.987, -11.997, -11.997,
    -11.998, -12.011, -12.034, -12.080, -12.166, -12.289, -11.990, -12.004,
    -12.012, -12.010, -12.009, -12.019, -12.040, -12.084, -12.169, -12.291,
    -12.009, -12.020, -12.026, -12.022, -12.019, -12.028, -12.047, -12.089,
    -12.172, -12.292, -12.028, -12.037, -12.041, -12.035, -12.030, -12.036,
    -12.053, -12.094, -12.176, -12.294, -12.047, -12.054, -12.055, -12.047,
    -12.040, -12.045, -12.060, -12.099, -12.179, -12.295, -12.066, -12.071,
    -12.070, -12.060, -12.051, -12.053, -12.067, -12.104, -12.182, -12.297,
    -12.085, -12.088, -12.085, -12.073, -12.062, -12.062, -12.074, -12.110,
    -12.186, -12.299, -12.104, -12.105, -12.100, -12.086, -12.073, -12.071,
    -12.081, -12.115, -12.189, -12.301, -12.123, -12.122, -12.115, -12.099,
    -12.084, -12.080, -12.088, -12.120, -12.192, -12.303, -12.141, -12.139,
    -12.130, -12.112, -12.095, -12.089, -12.095, -12.125, -12.196, -12.305,
    -12.160, -12.155, -12.144, -12.124, -12.106, -12.098, -12.102, -12.131,
    -12.199, -12.307, -12.179, -12.172, -12.159, -12.137, -12.117, -12.107,
    -12.109, -12.136, -12.203, -12.309, -12.197, -12.189, -12.174, -12.150,
    -12.128, -12.116, -12.116, -12.141, -12.207, -12.311, -12.216, -12.206,
    -12.189, -12.163, -12.139, -12.125, -12.123, -12.147, -12.210, -12.313,
    -12.234, -12.222, -12.204, -12.176, -12.151, -12.135, -12.131, -12.152,
    -12.214, -12.315, -12.253, -12.239, -12.220, -12.190, -12.162, -12.144,
    -12.138, -12.158, -12.218, -12.317, -12.271, -12.256, -12.235, -12.203,
    -12.173, -12.153, -12.146, -12.163, -12.222, -12.320, -12.290, -12.273,
    -12.250, -12.216, -12.185, -12.163, -12.153, -12.169, -12.226, -12.322,
    -12.308, -12.290, -12.265, -12.229, -12.196, -12.172, -12.161, -12.175,
    -12.230, -12.324, -12.327, -12.306, -12.280, -12.242, -12.208, -12.182,
    -12.168, -12.180, -12.234, -12.326, -12.345, -12.323, -12.295, -12.256,
    -12.219, -12.191, -12.176, -12.186, -12.238, -12.329, -12.363, -12.340,
    -12.310, -12.269, -12.231, -12.201, -12.183, -12.192, -12.242, -12.331,
    -12.381, -12.356, -12.325, -12.282, -12.242, -12.211, -12.191, -12.198,
    -12.246, -12.333, -12.399, -12.373, -12.340, -12.295, -12.254, -12.220,
    -12.199, -12.204, -12.250, -12.336, -12.417, -12.389, -12.354, -12.309,
    -12.265, -12.230, -12.207, -12.210, -12.254, -12.338, -12.435, -12.406,
    -12.369, -12.322, -12.277, -12.239, -12.215, -12.216, -12.258, -12.340,
    -12.453, -12.422, -12.384, -12.336, -12.288, -12.249, -12.223, -12.222,
    -12.263, -12.343, -12.471, -12.439, -12.399, -12.349, -12.300, -12.259,
    -12.231, -12.228, -12.267, -12.345, -12.489, -12.455, -12.414, -12.362,
    -12.312, -12.269, -12.239, -12.234, -12.271, -12.347, -12.506, -12.472,
    -12.429, -12.376, -12.323, -12.279, -12.247, -12.240, -12.275, -12.350,
    -12.524, -12.488, -12.444, -12.389, -12.335, -12.288, -12.255, -12.246,
    -12.279, -12.352, -12.541, -12.504, -12.458, -12.402, -12.346, -12.298,
    -12.262, -12.252, -12.284, -12.355, -12.559, -12.521, -12.473, -12.416,
    -12.358, -12.308, -12.270, -12.258, -12.288, -12.358, -12.576, -12.537,
    -12.488, -12.429, -12.369, -12.318, -12.278, -12.264, -12.292, -12.360,
    -12.594, -12.553, -12.503, -12.442, -12.381, -12.328, -12.286, -12.270,
    -12.296, -12.363, -12.611, -12.569, -12.517, -12.455, -12.393, -12.338,
    -12.294, -12.276, -12.301, -12.366, -12.628, -12.585, -12.532, -12.469,
    -12.404, -12.348, -12.302, -12.282, -12.305, -12.368, -12.645, -12.601,
    -12.547, -12.482, -12.416, -12.358, -12.310, -12.288, -12.309, -12.371,
    -12.662, -12.617, -12.562, -12.495, -12.428, -12.368, -12.318, -12.294,
    -12.313, -12.374, -12.679, -12.633, -12.576, -12.508, -12.439, -12.378,
    -12.326, -12.300, -12.318, -12.376, -12.696, -12.649, -12.591, -12.522,
    -12.451, -12.388, -12.334, -12.306, -12.322, -12.379, -12.712, -12.664,
    -12.605, -12.535, -12.462, -12.398, -12.342, -12.313, -12.326, -12.381,
    -12.729, -12.680, -12.620, -12.548, -12.474, -12.407, -12.350, -12.319,
    -12.330, -12.384, -12.746, -12.696, -12.634, -12.561, -12.486, -12.417,
    -12.358, -12.325, -12.335, -12.387, -12.762, -12.711, -12.649, -12.574,
    -12.497, -12.427, -12.366, -12.331, -12.339, -12.389, -12.778, -12.727,
    -12.663, -12.587, -12.509, -12.437, -12.375, -12.337, -12.343, -12.392,
    -12.795, -12.743, -12.678, -12.600, -12.520, -12.447, -12.383, -12.344,
    -12.348, -12.394, -12.811, -12.758, -12.692, -12.613, -12.532, -12.457,
    -12.391, -12.350, -12.352, -12.397, -12.827, -12.773, -12.706, -12.626,
    -12.544, -12.467, -12.399, -12.356, -12.356, -12.400, -12.843, -12.789,
    -12.720, -12.639, -12.555, -12.477, -12.407, -12.362, -12.361, -12.402,
    -12.859, -12.804, -12.735, -12.652, -12.567, -12.487, -12.415, -12.368,
    -12.365, -12.405, -12.875, -12.819, -12.749, -12.665, -12.578, -12.496,
    -12.423, -12.375, -12.369, -12.407, -12.891, -12.834, -12.763, -12.677,
    -12.590, -12.506, -12.431, -12.381, -12.374, -12.410, -12.907, -12.849,
    -12.777, -12.690, -12.601, -12.516, -12.439, -12.387, -12.378, -12.413,
    -12.923, -12.864, -12.791, -12.703, -12.613, -12.526, -12.447, -12.393,
    -12.382, -12.415, -12.939, -12.879, -12.804, -12.716, -12.624, -12.535,
    -12.455, -12.399, -12.386, -12.418, -12.954, -12.893, -12.818, -12.728,
    -12.636, -12.545, -12.463, -12.405, -12.391, -12.420, -12.970, -12.908,
    -12.832, -12.741, -12.647, -12.555, -12.471, -12.411, -12.395, -12.423,
    -12.985, -12.923, -12.846, -12.754, -12.658, -12.565, -12.479, -12.417,
    -12.399, -12.426, -13.001, -12.937, -12.860, -12.766, -12.670, -12.574,
    -12.487, -12.423, -12.403, -12.428, -13.016, -12.952, -12.873, -12.779,
    -12.681, -12.584, -12.495, -12.429, -12.408, -12.431, -13.031, -12.967,
    -12.887, -12.792, -12.692, -12.594, -12.503, -12.435, -12.412, -12.433,
    -13.046, -12.981, -12.900, -12.804, -12.704, -12.604, -12.511, -12.441,
    -12.416, -12.436, -13.061, -12.996, -12.914, -12.817, -12.715, -12.613,
    -12.519, -12.447, -12.420, -12.438, -13.076, -13.010, -12.928, -12.829,
    -12.726, -12.623, -12.526, -12.453, -12.424, -12.441, -13.091, -13.024,
    -12.941, -12.842, -12.738, -12.633, -12.534, -12.459, -12.428, -12.443,
    -13.105, -13.039, -12.955, -12.854, -12.749, -12.642, -12.542, -12.465,
    -12.432, -12.446, -13.120, -13.053, -12.968, -12.867, -12.760, -12.652,
    -12.550, -12.471, -12.436, -12.448, -13.135, -13.067, -12.981, -12.879,
    -12.771, -12.662, -12.558, -12.477, -12.440, -12.450, -13.150, -13.081,
    -12.995, -12.892, -12.782, -12.671, -12.565, -12.483, -12.444, -12.453,
    -13.164, -13.096, -13.008, -12.904, -12.793, -12.681, -12.573, -12.488,
    -12.448, -12.455, -13.179, -13.110, -13.022, -12.917, -12.804, -12.690,
    -12.581, -12.494, -12.452, -12.457, -13.193, -13.124, -13.035, -12.929,
    -12.815, -12.700, -12.588, -12.500, -12.456, -12.460, -13.208, -13.138,
    -13.048, -12.941, -12.826, -12.709, -12.596, -12.506, -12.460, -12.462,
    -13.223, -13.152, -13.061, -12.953, -12.837, -12.719, -12.604, -12.511,
    -12.464, -12.464, -13.237, -13.165, -13.075, -12.966, -12.848, -12.728,
    -12.611, -12.517, -12.468, -12.467, -13.252, -13.179, -13.088, -12.978,
    -12.859, -12.738, -12.619, -12.522, -12.472, -12.469, -13.266, -13.193,
    -13.101, -12.990, -12.870, -12.747, -12.626, -12.528, -12.476, -12.471,
    -13.280, -13.207, -13.114, -13.002, -12.881, -12.756, -12.633, -12.534,
    -12.480, -12.473, -13.294, -13.221, -13.127, -13.014, -12.892, -12.766,
    -12.641, -12.539, -12.484, -12.475, -13.308, -13.234, -13.140, -13.026,
    -12.903, -12.775, -12.648, -12.545, -12.487, -12.478, -13.322, -13.248,
    -13.153, -13.038, -12.913, -12.784, -12.655, -12.550, -12.491, -12.480,
    -13.336, -13.262, -13.166, -13.050, -12.924, -12.793, -12.663, -12.555,
    -12.495, -12.482, -13.350, -13.275, -13.179, -13.062, -12.935, -12.802,
    -12.670, -12.561, -12.498, -12.484, -13.364, -13.289, -13.192, -13.074,
    -12.946, -12.811, -12.677, -12.566, -12.502, -12.487, -13.378, -13.302,
    -13.205, -13.086, -12.957, -12.820, -12.684, -12.571, -12.505, -12.489,
    -13.391, -13.316, -13.217, -13.098, -12.967, -12.829, -12.691, -12.577,
    -12.509, -12.491, -13.405, -13.329, -13.230, -13.110, -12.978, -12.838,
    -12.698, -12.582, -12.512, -12.493, -13.419, -13.342, -13.243, -13.122,
    -12.989, -12.847, -12.705, -12.587, -12.515, -12.495, -13.432, -13.356,
    -13.255, -13.134, -12.999, -12.856, -12.712, -12.592, -12.519, -12.497,
    -13.446, -13.369, -13.268, -13.145, -13.010, -12.865, -12.719, -12.597,
    -12.522, -12.499, -13.460, -13.382, -13.280, -13.157, -13.020, -12.874,
    -12.726, -12.602, -12.526, -12.501, -13.473, -13.395, -13.293, -13.169,
    -13.030, -12.882, -12.732, -12.607, -12.529, -12.503, -13.487, -13.408,
    -13.305, -13.181, -13.041, -12.891, -12.739, -12.612, -12.533, -12.505,
    -13.501, -13.421, -13.318, -13.192, -13.051, -12.900, -12.746, -12.617,
    -12.536, -12.507, -13.514, -13.434, -13.330, -13.204, -13.062, -12.909,
    -12.753, -12.622, -12.539, -12.509, -13.528, -13.447, -13.343, -13.215,
    -13.072, -12.917, -12.759, -12.626, -12.543, -12.511, -13.541, -13.460,
    -13.355, -13.227, -13.082, -12.926, -12.766, -12.631, -12.546, -12.513,
    -13.554, -13.473, -13.367, -13.238, -13.092, -12.934, -12.773, -12.636,
    -12.549, -12.515, -13.567, -13.486, -13.380, -13.250, -13.102, -12.943,
    -12.779, -12.640, -12.552, -12.516, -13.581, -13.499, -13.392, -13.261,
    -13.113, -12.951, -12.786, -12.645, -12.555, -12.518, -13.594, -13.511,
    -13.404, -13.273, -13.123, -12.960, -12.792, -12.649, -12.558, -12.520,
    -13.607, -13.524, -13.416, -13.284, -13.133, -12.968, -12.799, -12.654,
    -12.561, -12.521, -13.620, -13.537, -13.429, -13.295, -13.143, -12.976,
    -12.805, -12.658, -12.564, -12.523, -13.633, -13.549, -13.441, -13.307,
    -13.153, -12.984, -12.811, -12.662, -12.566, -12.524, -13.645, -13.562,
    -13.453, -13.318, -13.163, -12.992, -12.818, -12.667, -12.569, -12.525,
    -13.658, -13.574, -13.465, -13.329, -13.173, -13.000, -12.824, -12.671,
    -12.571, -12.527, -13.671, -13.587, -13.477, -13.340, -13.183, -13.008,
    -12.830, -12.675, -12.574, -12.528, -13.684, -13.599, -13.489, -13.351,
    -13.193, -13.016, -12.836, -12.679, -12.577, -12.530, -13.697, -13.612,
    -13.501, -13.362, -13.203, -13.024, -12.842, -12.683, -12.579, -12.531,
    -13.709, -13.624, -13.513, -13.373, -13.212, -13.031, -12.848, -12.687,
    -12.582, -12.533, -13.722, -13.637, -13.525, -13.384, -13.222, -13.039,
    -12.853, -12.690, -12.585, -12.534, -13.735, -13.649, -13.537, -13.395,
    -13.232, -13.047, -12.859, -12.694, -12.588, -12.536, -13.748, -13.661,
    -13.548, -13.406, -13.241, -13.054, -12.865, -12.698, -12.590, -12.537,
    -13.760, -13.674, -13.560, -13.417, -13.251, -13.062, -12.870, -12.701,
    -12.593, -12.539, -13.773, -13.686, -13.572, -13.428, -13.260, -13.069,
    -12.875, -12.704, -12.596, -12.541, -13.785, -13.698, -13.583, -13.439,
    -13.270, -13.077, -12.881, -12.708, -12.598, -12.542, -13.798, -13.710,
    -13.595, -13.450, -13.279, -13.084, -12.886, -12.711, -12.601, -12.544,
    -13.810, -13.722, -13.607, -13.461, -13.288, -13.091, -12.891, -12.715,
    -12.603, -12.545, -13.823, -13.734, -13.618, -13.471, -13.297, -13.098,
    -12.896, -12.718, -12.606, -12.547, -13.835, -13.746, -13.630, -13.482,
    -13.307, -13.105, -12.901, -12.722, -12.608, -12.548, -13.847, -13.758,
    -13.641, -13.493, -13.316, -13.112, -12.906, -12.726, -12.610, -12.549,
    -13.860, -13.770, -13.653, -13.503, -13.325, -13.119, -12.911, -12.729,
    -12.612, -12.550, -13.872, -13.782, -13.664, -13.514, -13.334, -13.126,
    -12.916, -12.733, -12.614, -12.551, -13.884, -13.794, -13.676, -13.524,
    -13.343, -13.133, -12.921, -12.737, -12.616, -12.552, -13.896, -13.806,
    -13.687, -13.534, -13.351, -13.140, -12.926, -12.741, -12.618, -12.553,
    -13.908, -13.818, -13.699, -13.545, -13.360, -13.146, -12.930, -12.744,
    -12.619, -12.554, -13.920, -13.830, -13.710, -13.555, -13.369, -13.153,
    -12.935, -12.748, -12.621, -12.555, -13.932, -13.842, -13.721, -13.565,
    -13.378, -13.160, -12.940, -12.751, -12.622, -12.556, -13.944, -13.853,
    -13.732, -13.576, -13.386, -13.166, -12.944, -12.754, -12.624, -12.557,
    -13.956, -13.865, -13.744, -13.586, -13.395, -13.173, -12.949, -12.757,
    -12.625, -12.558, -13.968, -13.877, -13.755, -13.596, -13.404, -13.179,
    -12.953, -12.760, -12.626, -12.559, -13.979, -13.888, -13.766, -13.606,
    -13.412, -13.186, -12.957, -12.762, -12.627, -12.561, -13.991, -13.900,
    -13.777, -13.616, -13.421, -13.192, -12.962, -12.765, -12.628, -12.562,
    -14.003, -13.911, -13.788, -13.626, -13.429, -13.198, -12.966, -12.767,
    -12.628, -12.563, -14.015, -13.922, -13.798, -13.636, -13.437, -13.205,
    -12.970, -12.770, -12.629, -12.564, -14.026, -13.934, -13.809, -13.646,
    -13.446, -13.211, -12.974, -12.772, -12.630, -12.565, -14.038, -13.945,
    -13.820, -13.656, -13.454, -13.217, -12.978, -12.774, -12.630, -12.566,
    -14.050, -13.956, -13.831, -13.666, -13.462, -13.223, -12.982, -12.777,
    -12.632, -12.567, -14.061, -13.968, -13.842, -13.676, -13.469, -13.229,
    -12.986, -12.779, -12.634, -12.569, -14.073, -13.979, -13.853, -13.686,
    -13.476, -13.235, -12.990, -12.782, -12.637, -12.570, -14.084, -13.990,
    -13.863, -13.695, -13.483, -13.240, -12.994, -12.785, -12.639, -12.572,
    -14.096, -14.002, -13.874, -13.705, -13.490, -13.246, -12.998, -12.787,
    -12.642, -12.574, -14.107, -14.013, -13.885, -13.715, -13.497, -13.252,
    -13.001, -12.790, -12.645, -12.575, -14.119, -14.024, -13.896, -13.724,
    -13.503, -13.257, -13.005, -12.793, -12.649, -12.577, -14.130, -14.036,
    -13.907, -13.734, -13.510, -13.262, -13.009, -12.796, -12.652, -12.579,
    -14.142, -14.047, -13.917, -13.743, -13.516, -13.268, -13.012, -12.798,
    -12.656, -12.581, -14.153, -14.058, -13.928, -13.753, -13.522, -13.273,
    -13.016, -12.801, -12.660, -12.583, -14.164, -14.069, -13.939, -13.762,
    -13.530, -13.278, -13.020, -12.804, -12.662, -12.584, -14.176, -14.080,
    -13.949, -13.772, -13.538, -13.284, -13.023, -12.807, -12.664, -12.585,
    -14.187, -14.091, -13.959, -13.781, -13.546, -13.289, -13.027, -12.810,
    -12.666, -12.585, -14.198, -14.102, -13.970, -13.791, -13.554, -13.294,
    -13.031, -12.814, -12.668, -12.586, -14.209, -14.113, -13.980, -13.800,
    -13.562, -13.299, -13.035, -12.817, -12.669, -12.586, -14.220, -14.124,
    -13.990, -13.809, -13.571, -13.304, -13.038, -12.821, -12.670, -12.586,
    -14.231, -14.135, -14.001, -13.819, -13.580, -13.309, -13.042, -12.824,
    -12.671, -12.586, -14.242, -14.146, -14.011, -13.828, -13.589, -13.314,
    -13.046, -12.828, -12.672, -12.586, -14.253, -14.156, -14.021, -13.837,
    -13.598, -13.319, -13.049, -12.831, -12.673, -12.585, -14.264, -14.167,
    -14.031, -13.846, -13.607, -13.324, -13.053, -12.835, -12.674, -12.585,
    -14.275, -14.178, -14.041, -13.855, -13.614, -13.329, -13.056, -12.838,
    -12.676, -12.586, -14.286, -14.188, -14.051, -13.864, -13.621, -13.334,
    -13.060, -12.840, -12.678, -12.587, -14.297, -14.199, -14.062, -13.873,
    -13.628, -13.339, -13.063, -12.842, -12.680, -12.589, -14.307, -14.210,
    -14.072, -13.882, -13.634, -13.344, -13.066, -12.845, -12.682, -12.590,
    -14.318, -14.220, -14.082, -13.891, -13.640, -13.350, -13.069, -12.847,
    -12.685, -12.592, -14.329, -14.231, -14.092, -13.900, -13.646, -13.355,
    -13.072, -12.849, -12.687, -12.593, -14.339, -14.242, -14.102, -13.909,
    -13.652, -13.360, -13.075, -12.850, -12.690, -12.595, -14.350, -14.252,
    -14.112, -13.918, -13.657, -13.365, -13.078, -12.852, -12.692, -12.597,
    -14.360, -14.263, -14.122, -13.926, -13.662, -13.370, -13.081, -12.854,
    -12.695, -12.600, -14.371, -14.273, -14.132, -13.935, -13.667, -13.375,
    -13.084, -12.855, -12.698, -12.602, -14.382, -14.283, -14.142, -13.944,
    -13.674, -13.380, -13.088, -12.858, -12.701, -12.604, -14.392, -14.294,
    -14.152, -13.953, -13.682, -13.386, -13.092, -12.861, -12.703, -12.605,
    -14.403, -14.304, -14.162, -13.962, -13.689, -13.392, -13.097, -12.864,
    -12.706, -12.607, -14.413, -14.315, -14.172, -13.971, -13.697, -13.397,
    -13.101, -12.867, -12.708, -12.609, -14.424, -14.325, -14.181, -13.979,
    -13.705, -13.403, -13.106, -12.871, -12.711, -12.610, -14.434, -14.335,
    -14.191, -13.988, -13.713, -13.409, -13.111, -12.874, -12.713, -12.612,
    -14.445, -14.345, -14.201, -13.997, -13.721, -13.415, -13.116, -12.878,
    -12.716, -12.613, -14.455, -14.356, -14.211, -14.006, -13.730, -13.421,
    -13.122, -12.882, -12.719, -12.615, -14.466, -14.366, -14.220, -14.015,
    -13.739, -13.427, -13.127, -12.886, -12.721, -12.616, -14.476, -14.376,
    -14.230, -14.024, -13.748, -13.433, -13.133, -12.890, -12.724, -12.618,
    -14.486, -14.386, -14.240, -14.033, -13.756, -13.439, -13.138, -12.894,
    -12.726, -12.619, -14.497, -14.396, -14.249, -14.042, -13.764, -13.445,
    -13.142, -12.897, -12.729, -12.621, -14.507, -14.406, -14.259, -14.051,
    -13.771, -13.451, -13.147, -12.901, -12.731, -12.622, -14.517, -14.416,
    -14.269, -14.060, -13.779, -13.456, -13.151, -12.904, -12.734, -12.624,
    -14.527, -14.426, -14.278, -14.069, -13.787, -13.462, -13.156, -12.908,
    -12.736, -12.625, -14.538, -14.436, -14.288, -14.077, -13.794, -13.468,
    -13.160, -12.911, -12.739, -12.627, -14.548, -14.446, -14.297, -14.086,
    -13.802, -13.474, -13.165, -12.915, -12.741, -12.628, -14.558, -14.456,
    -14.307, -14.095, -13.809, -13.480, -13.169, -12.918, -12.743, -12.629,
    -14.568, -14.466, -14.316, -14.104, -13.817, -13.486, -13.174, -12.922,
    -12.746, -12.631, -14.578, -14.476, -14.326, -14.113, -13.824, -13.492,
    -13.178, -12.925, -12.748, -12.632, -14.588, -14.486, -14.335, -14.122,
    -13.831, -13.498, -13.183, -12.929, -12.751, -12.633, -14.598, -14.496,
    -14.345, -14.130, -13.839, -13.504, -13.187, -12.932, -12.753, -12.635,
    -14.608, -14.505, -14.354, -14.139, -13.846, -13.510, -13.192, -12.936,
    -12.756, -12.636, -14.618, -14.515, -14.364, -14.148, -13.854, -13.516,
    -13.197, -12.939, -12.758, -12.638, -14.627, -14.525, -14.373, -14.156,
    -13.861, -13.522, -13.202, -12.943, -12.761, -12.639, -14.637, -14.535,
    -14.383, -14.165, -13.869, -13.528, -13.206, -12.946, -12.763, -12.640,
    -14.647, -14.544, -14.392, -14.174, -13.876, -13.534, -13.211, -12.950,
    -12.766, -12.642, -14.657, -14.554, -14.401, -14.182, -13.883, -13.541,
    -13.216, -12.954, -12.769, -12.643, -14.666, -14.563, -14.411, -14.191,
    -13.891, -13.547, -13.221, -12.957, -12.771, -12.645, -14.676, -14.573,
    -14.420, -14.199, -13.898, -13.553, -13.226, -12.961, -12.774, -12.646,
    -14.686, -14.583, -14.429, -14.208, -13.906, -13.559, -13.231, -12.965,
    -12.777, -12.648, -14.695, -14.592, -14.439, -14.216, -13.913, -13.566,
    -13.236, -12.969, -12.780, -12.649, -14.705, -14.602, -14.448, -14.225,
    -13.921, -13.572, -13.242, -12.973, -12.782, -12.651, -14.715, -14.611,
    -14.457, -14.234, -13.929, -13.579, -13.247, -12.977, -12.785, -12.652,
    -14.725, -14.621, -14.466, -14.242, -13.937, -13.586, -13.253, -12.981,
    -12.788, -12.654, -14.734, -14.630, -14.475, -14.251, -13.945, -13.592,
    -13.258, -12.985, -12.791, -12.655, -14.744, -14.640, -14.485, -14.260,
    -13.953, -13.599, -13.264, -12.990, -12.794, -12.657, -14.754, -14.649,
    -14.494, -14.269, -13.961, -13.606, -13.269, -12.994, -12.797, -12.659,
    -14.763, -14.659, -14.503, -14.277, -13.969, -13.613, -13.275, -12.999,
    -12.800, -12.660, -14.773, -14.668, -14.512, -14.286, -13.977, -13.620,
    -13.281, -13.003, -12.803, -12.662, -14.782, -14.677, -14.521, -14.295,
    -13.985, -13.627, -13.287, -13.008, -12.806, -12.664, -14.792, -14.687,
    -14.530, -14.303, -13.993, -13.634, -13.293, -13.012, -12.809, -12.666,
    -14.801, -14.696, -14.539, -14.312, -14.001, -13.641, -13.299, -13.017,
    -12.813, -12.667, -14.811, -14.705, -14.548, -14.321, -14.009, -13.649,
    -13.305, -13.022, -12.816, -12.669, -14.820, -14.714, -14.556, -14.330,
    -14.018, -13.656, -13.311, -13.027, -12.819, -12.671, -14.829, -14.723,
    -14.565, -14.338, -14.026, -13.664, -13.317, -13.032, -12.823, -12.673,
    -14.839, -14.732, -14.574, -14.347, -14.034, -13.671, -13.324, -13.037,
    -12.826, -12.675, -14.848, -14.741, -14.583, -14.356, -14.042, -13.679,
    -13.330, -13.042, -12.830, -12.677, -14.857, -14.750, -14.591, -14.364,
    -14.051, -13.686, -13.336, -13.047, -12.833, -12.679, -14.866, -14.759,
    -14.600, -14.373, -14.059, -13.694, -13.343, -13.052, -12.837, -12.681,
    -14.875, -14.768, -14.609, -14.381, -14.067, -13.702, -13.349, -13.057,
    -12.841, -12.683, -14.884, -14.777, -14.617, -14.390, -14.075, -13.709,
    -13.356, -13.062, -12.844, -12.685, -14.893, -14.786, -14.626, -14.398,
    -14.084, -13.717, -13.362, -13.067, -12.848, -12.687, -14.902, -14.794,
    -14.635, -14.407, -14.092, -13.724, -13.369, -13.072, -12.851, -12.689,
    -14.911, -14.803, -14.643, -14.415, -14.100, -13.732, -13.376, -13.078,
    -12.855, -12.691, -14.920, -14.812, -14.652, -14.423, -14.108, -13.739,
    -13.382, -13.083, -12.859, -12.693, -14.929, -14.820, -14.660, -14.432,
    -14.116, -13.747, -13.389, -13.088, -12.863, -12.695, -14.938, -14.829,
    -14.669, -14.440, -14.125, -13.755, -13.395, -13.093, -12.866, -12.697,
    -14.947, -14.837, -14.677, -14.448, -14.133, -13.762, -13.402, -13.099,
    -12.870, -12.699, -14.956, -14.846, -14.686, -14.456, -14.141, -13.770,
    -13.409, -13.104, -12.874, -12.701, -14.965, -14.854, -14.694, -14.464,
    -14.149, -13.778, -13.416, -13.109, -12.878, -12.703, -14.973, -14.863,
    -14.702, -14.472, -14.157, -13.785, -13.422, -13.115, -12.881, -12.705,
    -14.982, -14.871, -14.710, -14.480, -14.165, -13.793, -13.429, -13.120,
    -12.885, -12.707, -14.990, -14.879, -14.718, -14.488, -14.173, -13.801,
    -13.436, -13.126, -12.889, -12.709, -14.998, -14.887, -14.726, -14.496,
    -14.181, -13.808, -13.443, -13.131, -12.892, -12.711, -15.007, -14.895,
    -14.734, -14.504, -14.189, -13.816, -13.450, -13.137, -12.896, -12.713,
    -15.015, -14.903, -14.741, -14.512, -14.197, -13.824, -13.456, -13.142,
    -12.900, -12.715, -15.023, -14.911, -14.749, -14.520, -14.205, -13.832,
    -13.463, -13.148, -12.904, -12.717, -15.031, -14.919, -14.757, -14.528,
    -14.213, -13.839, -13.470, -13.153, -12.907, -12.719, -15.039, -14.927,
    -14.764, -14.536, -14.221, -13.847, -13.477, -13.159, -12.911, -12.721,
    -15.047, -14.935, -14.772, -14.544, -14.229, -13.855, -13.484, -13.164,
    -12.915, -12.723, -15.055, -14.942, -14.779, -14.551, -14.237, -13.862,
    -13.491, -13.170, -12.919, -12.725, -15.063, -14.950, -14.786, -14.559,
    -14.244, -13.870, -13.497, -13.175, -12.922, -12.727, -15.070, -14.957,
    -14.794, -14.566, -14.252, -13.878, -13.504, -13.181, -12.926, -12.729,
    -15.078, -14.964, -14.801, -14.574, -14.260, -13.885, -13.511, -13.186,
    -12.930, -12.731, -15.085, -14.971, -14.808, -14.581, -14.268, -13.893,
    -13.518, -13.192, -12.934, -12.733, -15.093, -14.978, -14.816, -14.588,
    -14.275, -13.900, -13.525, -13.197, -12.937, -12.735, -15.100, -14.985,
    -14.823, -14.596, -14.283, -13.908, -13.531, -13.202, -12.941, -12.737,
    -15.108, -14.992, -14.830, -14.603, -14.290, -13.915, -13.538, -13.208,
    -12.945, -12.739, -15.115, -14.999, -14.837, -14.610, -14.298, -13.923,
    -13.545, -13.213, -12.949, -12.741, -15.122, -15.005, -14.844, -14.617,
    -14.305, -13.930, -13.552, -13.218, -12.953, -12.743, -15.129, -15.012,
    -14.850, -14.624, -14.313, -13.938, -13.558, -13.224, -12.956, -12.745,
    -15.135, -15.018, -14.856, -14.631, -14.320, -13.945, -13.565, -13.229,
    -12.960, -12.746, -15.141, -15.024, -14.863, -14.637, -14.328, -13.953,
    -13.572, -13.235, -12.963, -12.748, -15.148, -15.029, -14.869, -14.644,
    -14.335, -13.960, -13.579, -13.240, -12.967, -12.750, -15.154, -15.035,
    -14.875, -14.651, -14.342, -13.968, -13.585, -13.245, -12.970, -12.752,
    -15.160, -15.040, -14.880, -14.657, -14.350, -13.975, -13.592, -13.251,
    -12.974, -12.753, -15.166, -15.046, -14.886, -14.663, -14.357, -13.982,
    -13.599, -13.256, -12.977, -12.755, -15.171, -15.051, -14.892, -14.670,
    -14.364, -13.990, -13.605, -13.262, -12.981, -12.756, -15.177, -15.056,
    -14.897, -14.676, -14.371, -13.997, -13.612, -13.267, -12.984, -12.758,
    -15.182, -15.061, -14.903, -14.682, -14.378, -14.004, -13.619, -13.272,
    -12.987, -12.760, -15.187, -15.066, -14.908, -14.688, -14.385, -14.011,
    -13.625, -13.278, -12.991, -12.761, -15.192, -15.071, -14.913, -14.694,
    -14.392, -14.019, -13.632, -13.283, -12.994, -12.763, -15.196, -15.076,
    -14.919, -14.700, -14.399, -14.026, -13.639, -13.288, -12.997, -12.765,
    -15.201, -15.080, -14.924, -14.706, -14.405, -14.033, -13.645, -13.293,
    -13.000, -12.766, -15.205, -15.085, -14.929, -14.712, -14.412, -14.040,
    -13.652, -13.298, -13.003, -12.768, -15.208, -15.089, -14.934, -14.718,
    -14.419, -14.047, -13.658, -13.304, -13.006, -12.769, -15.212, -15.094,
    -14.939, -14.724, -14.425, -14.054, -13.665, -13.309, -13.009, -12.771,
    -15.216, -15.098, -14.944, -14.729, -14.432, -14.061, -13.671, -13.314,
    -13.012, -12.772, -15.219, -15.102, -14.949, -14.735, -14.438, -14.068,
    -13.678, -13.319, -13.015, -12.774, -15.222, -15.106, -14.953, -14.741,
    -14.445, -14.075, -13.685, -13.324, -13.018, -12.775, -15.224, -15.109,
    -14.957, -14.746, -14.451, -14.082, -13.691, -13.329, -13.021, -12.777,
    -15.227, -15.112, -14.961, -14.751, -14.458, -14.089, -13.698, -13.334,
    -13.024, -12.778, -15.229, -15.116, -14.964, -14.757, -14.464, -14.096,
    -13.704, -13.339, -13.028, -12.779, -15.231, -15.119, -14.968, -14.762,
    -14.470, -14.103, -13.711, -13.344, -13.031, -12.780, -15.232, -15.121,
    -14.971, -14.767, -14.477, -14.110, -13.718, -13.349, -13.034, -12.781,
    -15.234, -15.124, -14.974, -14.772, -14.483, -14.117, -13.724, -13.354,
    -13.037, -12.782, -15.235, -15.127, -14.977, -14.777, -14.489, -14.124,
    -13.731, -13.358, -13.041, -12.783, -15.235, -15.129, -14.980, -14.782,
    -14.496, -14.131, -13.737, -13.363, -13.044, -12.784, -15.236, -15.131,
    -14.982, -14.787, -14.502, -14.138, -13.744, -13.368, -13.047, -12.785,
    -15.236, -15.132, -14.986, -14.792, -14.508, -14.145, -13.750, -13.373,
    -13.050, -12.786, -15.236, -15.134, -14.989, -14.797, -14.513, -14.152,
    -13.757, -13.377, -13.053, -12.787, -15.236, -15.134, -14.993, -14.802,
    -14.518, -14.158, -13.763, -13.382, -13.055, -12.788, -15.236, -15.135,
    -14.996, -14.807, -14.523, -14.165, -13.769, -13.386, -13.058, -12.789,
    -15.235, -15.135, -15.000, -14.811, -14.528, -14.172, -13.775, -13.391,
    -13.060, -12.789, -15.234, -15.135, -15.004, -14.816, -14.533, -14.179,
    -13.781, -13.395, -13.063, -12.790, -15.233, -15.135, -15.007, -14.821,
    -14.537, -14.185, -13.787, -13.400, -13.065, -12.791, -15.232, -15.135,
    -15.011, -14.826, -14.542, -14.192, -13.793, -13.404, -13.068, -12.792,
    -15.230, -15.134, -15.015, -14.830, -14.546, -14.198, -13.799, -13.409,
    -13.070, -12.792, -15.228, -15.133, -15.019, -14.835, -14.550, -14.205,
    -13.805, -13.413, -13.072, -12.793, -15.225, -15.133, -15.022, -14.839,
    -14.556, -14.212, -13.811, -13.417, -13.074, -12.794, -15.222, -15.133,
    -15.024, -14.844, -14.562, -14.218, -13.817, -13.422, -13.077, -12.794,
    -15.218, -15.133, -15.027, -14.848, -14.569, -14.225, -13.823, -13.426,
    -13.079, -12.794, -15.214, -15.133, -15.029, -14.852, -14.575, -14.231,
    -13.828, -13.430, -13.081, -12.795, -15.210, -15.133, -15.031, -14.856,
    -14.582, -14.237, -13.834, -13.434, -13.083, -12.795, -15.205, -15.132,
    -15.034, -14.860, -14.589, -14.244, -13.840, -13.438, -13.085, -12.795,
    -15.200, -15.132, -15.035, -14.864, -14.597, -14.250, -13.845, -13.442,
    -13.087, -12.796, -15.194, -15.132, -15.037, -14.868, -14.604, -14.256,
    -13.851, -13.446, -13.089, -12.796, -15.188, -15.131, -15.039, -14.872,
    -14.612, -14.263, -13.856, -13.450, -13.091, -12.796, -15.182, -15.131,
    -15.040, -14.876, -14.620, -14.269, -13.862, -13.454, -13.093, -12.796,
    -15.182, -15.133, -15.044, -14.881, -14.627, -14.275, -13.867, -13.458,
    -13.095, -12.796, -15.183, -15.136, -15.048, -14.886, -14.633, -14.282,
    -13.873, -13.462, -13.096, -12.796, -15.185, -15.140, -15.052, -14.891,
    -14.640, -14.288, -13.878, -13.465, -13.098, -12.797, -15.187, -15.144,
    -15.056, -14.896, -14.646, -14.295, -13.884, -13.469, -13.100, -12.797,
    -15.191, -15.148, -15.061, -14.902, -14.652, -14.301, -13.889, -13.473,
    -13.101, -12.797, -15.195, -15.153, -15.066, -14.907, -14.659, -14.308,
    -13.894, -13.476, -13.103, -12.797, -15.200, -15.159, -15.071, -14.913,
    -14.665, -14.314, -13.900, -13.480, -13.104, -12.797, -15.206, -15.165,
    -15.077, -14.919, -14.671, -14.321, -13.905, -13.483, -13.105, -12.797,
    -15.212, -15.171, -15.083, -14.925, -14.677, -14.327, -13.910, -13.487,
    -13.107, -12.797, -15.220, -15.178, -15.089, -14.931, -14.683, -14.334,
    -13.915, -13.490, -13.108, -12.797, -15.225, -15.183, -15.094, -14.937,
    -14.689, -14.340, -13.920, -13.493, -13.109, -12.797, -15.229, -15.188,
    -15.100, -14.943, -14.696, -14.347, -13.925, -13.496, -13.111, -12.797,
    -15.234, -15.194, -15.105, -14.949, -14.702, -14.353, -13.931, -13.499,
    -13.112, -12.797, -15.239, -15.199, -15.111, -14.955, -14.709, -14.359,
    -13.936, -13.502, -13.113, -12.797, -15.244, -15.205, -15.117, -14.961,
    -14.715, -14.366, -13.941, -13.505, -13.114, -12.796, -15.249, -15.210,
    -15.123, -14.967, -14.722, -14.372, -13.946, -13.508, -13.115, -12.796,
    -15.254, -15.216, -15.129, -14.974, -14.728, -14.378, -13.951, -13.511,
    -13.116, -12.796, -15.260, -15.221, -15.135, -14.980, -14.735, -14.385,
    -13.956, -13.514, -13.117, -12.796, -15.265, -15.227, -15.141, -14.986,
    -14.741, -14.391, -13.961, -13.516, -13.118, -12.795, -15.271, -15.233,
    -15.147, -14.993, -14.748, -14.397, -13.966, -13.519, -13.119, -12.795,
    -15.278, -15.239, -15.153, -14.999, -14.755, -14.403, -13.971, -13.522,
    -13.120, -12.794, -15.284, -15.246, -15.160, -15.006, -14.761, -14.409,
    -13.975, -13.524, -13.120, -12.794, -15.291, -15.253, -15.166, -15.013,
    -14.768, -14.415, -13.980, -13.527, -13.121, -12.793, -15.299, -15.259,
    -15.173, -15.019, -14.774, -14.421, -13.985, -13.529, -13.121, -12.792,
    -15.306, -15.266, -15.180, -15.026, -14.781, -14.427, -13.989, -13.531,
    -13.122, -12.791, -15.314, -15.274, -15.187, -15.033, -14.787, -14.433,
    -13.994, -13.534, -13.122, -12.789, -15.322, -15.281, -15.194, -15.039,
    -14.794, -14.439, -13.998, -13.536, -13.122, -12.788, -15.331, -15.289,
    -15.201, -15.046, -14.801, -14.445, -14.002, -13.538, -13.122, -12.787,
    -15.339, -15.296, -15.208, -15.053, -14.807, -14.451, -14.007, -13.540,
    -13.122, -12.786, -15.348, -15.304, -15.215, -15.060, -14.814, -14.457,
    -14.011, -13.542, -13.122, -12.784, -15.357, -15.312, -15.223, -15.067,
    -14.821, -14.463, -14.015, -13.544, -13.122, -12.783, -15.367, -15.321,
    -15.231, -15.075, -14.828, -14.469, -14.019, -13.546, -13.123, -12.782,
    -15.377, -15.330, -15.239, -15.082, -14.835, -14.475, -14.023, -13.548,
    -13.123, -12.781, -15.387, -15.339, -15.247, -15.090, -14.842, -14.481,
    -14.027, -13.550, -13.123, -12.781, -15.398, -15.349, -15.255, -15.098,
    -14.849, -14.487, -14.031, -13.551, -13.123, -12.780, -15.409, -15.358,
    -15.264, -15.106, -14.856, -14.493, -14.035, -13.553, -13.124, -12.779,
    -15.420, -15.368, -15.273, -15.114, -14.863, -14.498, -14.039, -13.555,
    -13.124, -12.779, -15.431, -15.378, -15.282, -15.122, -14.870, -14.504,
    -14.043, -13.556, -13.124, -12.778, -15.443, -15.388, -15.291, -15.131,
    -14.878, -14.510, -14.046, -13.558, -13.125, -12.778, -15.455, -15.399,
    -15.300, -15.139, -14.885, -14.516, -14.050, -13.559, -13.125, -12.777,
    -15.466, -15.409, -15.309, -15.147, -14.892, -14.522, -14.054, -13.560,
    -13.125, -12.776, -15.478, -15.419, -15.318, -15.155, -14.900, -14.528,
    -14.057, -13.562, -13.124, -12.775, -15.489, -15.429, -15.327, -15.164,
    -14.907, -14.533, -14.061, -13.563, -13.124, -12.774, -15.501, -15.440,
    -15.336, -15.172, -14.914, -14.539, -14.064, -13.564, -13.123, -12.772,
    -15.512, -15.450, -15.346, -15.180, -14.922, -14.545, -14.068, -13.566,
    -13.123, -12.771, -15.524, -15.460, -15.355, -15.189, -14.929, -14.550,
    -14.071, -13.567, -13.122, -12.769, -15.536, -15.471, -15.364, -15.197,
    -14.936, -14.556, -14.074, -13.568, -13.121, -12.768, -15.548, -15.481,
    -15.374, -15.206, -14.944, -14.562, -14.078, -13.569, -13.120, -12.766,
    -15.560, -15.492, -15.383, -15.214, -14.951, -14.567, -14.081, -13.570,
    -13.119, -12.765, -15.572, -15.503, -15.393, -15.223, -14.959, -14.573,
    -14.084, -13.571, -13.118, -12.763, -15.584, -15.514, -15.402, -15.232,
    -14.966, -14.579, -14.087, -13.572, -13.116, -12.762, -15.596, -15.524,
    -15.412, -15.240, -14.974, -14.584, -14.090, -13.572, -13.113, -12.761,
    -15.607, -15.535, -15.421, -15.249, -14.981, -14.590, -14.093, -13.573,
    -13.111, -12.760, -15.619, -15.545, -15.431, -15.257, -14.988, -14.595,
    -14.096, -13.573, -13.108, -12.759, -15.631, -15.556, -15.440, -15.266,
    -14.996, -14.600, -14.099, -13.574, -13.104, -12.757, -15.643, -15.567,
    -15.450, -15.274, -15.003, -14.606, -14.102, -13.574, -13.101, -12.757,
    -15.654, -15.578, -15.459, -15.283, -15.010, -14.611, -14.104, -13.574,
    -13.097, -12.756, -15.666, -15.588, -15.469, -15.292, -15.017, -14.616,
    -14.107, -13.574, -13.093, -12.755, -15.678, -15.599, -15.478, -15.300,
    -15.025, -14.622, -14.109, -13.574, -13.089, -12.754, -15.690, -15.610,
    -15.488, -15.309, -15.032, -14.627, -14.112, -13.574, -13.084, -12.753,
    -15.702, -15.621, -15.498, -15.318, -15.039, -14.632, -14.115, -13.575,
    -13.084, -12.753, -15.713, -15.631, -15.507, -15.326, -15.046, -14.638,
    -14.117, -13.575, -13.084, -12.752, -15.725, -15.642, -15.517, -15.335,
    -15.054, -14.643, -14.120, -13.576, -13.085, -12.752, -15.736, -15.652,
    -15.526, -15.344, -15.061, -14.648, -14.122, -13.576, -13.087, -12.752,
    -15.748, -15.663, -15.536, -15.352, -15.068, -14.653, -14.125, -13.577,
    -13.090, -12.752, -15.759, -15.673, -15.545, -15.361, -15.075, -14.658,
    -14.127, -13.578, -13.092, -12.752, -15.771, -15.684, -15.554, -15.370,
    -15.082, -14.664, -14.130, -13.578, -13.096, -12.753, -15.782, -15.694,
    -15.564, -15.379, -15.089, -14.669, -14.132, -13.579, -13.100, -12.753,
    -15.794, -15.705, -15.573, -15.387, -15.096, -14.674, -14.135, -13.580,
    -13.105, -12.753, -15.805, -15.715, -15.583, -15.396, -15.103, -14.679,
    -14.137, -13.581, -13.110, -12.754, -15.816, -15.725, -15.592, -15.405,
    -15.110, -14.684, -14.139, -13.582, -13.111, -12.753, -15.827, -15.736,
    -15.602, -15.413, -15.117, -14.688, -14.142, -13.584, -13.111, -12.752,
    -15.838, -15.746, -15.611, -15.422, -15.124, -14.693, -14.145, -13.586,
    -13.111, -12.750, -15.849, -15.756, -15.621, -15.430, -15.131, -14.697,
    -14.147, -13.587, -13.111, -12.749, -15.860, -15.766, -15.630, -15.439,
    -15.139, -14.702, -14.149, -13.589, -13.110, -12.747, -15.871, -15.776,
    -15.640, -15.447, -15.146, -14.706, -14.152, -13.591, -13.109, -12.745,
    -15.882, -15.786, -15.649, -15.456, -15.153, -14.710, -14.154, -13.594,
    -13.108, -12.742, -15.893, -15.796, -15.658, -15.464, -15.160, -14.714,
    -14.157, -13.596, -13.106, -12.740, -15.903, -15.806, -15.668, -15.473,
    -15.167, -14.718, -14.160, -13.598, -13.104, -12.737, -15.914, -15.816,
    -15.677, -15.481, -15.174, -14.722, -14.162, -13.601, -13.102, -12.734,
    -15.925, -15.826, -15.686, -15.489, -15.181, -14.727, -14.164, -13.602,
    -13.102, -12.733, -15.935, -15.836, -15.695, -15.498, -15.187, -14.731,
    -14.166, -13.603, -13.102, -12.733, -15.946, -15.846, -15.704, -15.506,
    -15.193, -14.736, -14.169, -13.603, -13.103, -12.733, -15.956, -15.855,
    -15.713, -15.514, -15.200, -14.741, -14.171, -13.603, -13.104, -12.734,
    -15.966, -15.865, -15.723, -15.523, -15.206, -14.746, -14.173, -13.604,
    -13.104, -12.734, -15.977, -15.875, -15.731, -15.531, -15.212, -14.751,
    -14.175, -13.603, -13.105, -12.735, -15.987, -15.884, -15.740, -15.539,
    -15.218, -14.756, -14.177, -13.603, -13.107, -12.736, -15.997, -15.894,
    -15.749, -15.548, -15.224, -14.761, -14.178, -13.603, -13.108, -12.737,
    -16.007, -15.903, -15.758, -15.556, -15.229, -14.766, -14.180, -13.602,
    -13.109, -12.739, -16.017, -15.913, -15.767, -15.564, -15.235, -14.771,
    -14.182, -13.601, -13.111, -12.741, -16.027, -15.922, -15.776, -15.572,
    -15.242, -14.776, -14.185, -13.602, -13.112, -12.742, -16.037, -15.932,
    -15.785, -15.580, -15.249, -14.781, -14.188, -13.604, -13.114, -12.743,
    -16.047, -15.941, -15.793, -15.588, -15.256, -14.786, -14.192, -13.606,
    -13.115, -12.743, -16.057, -15.951, -15.802, -15.596, -15.263, -14.791,
    -14.196, -13.608, -13.116, -12.744, -16.067, -15.960, -15.811, -15.604,
    -15.270, -14.797, -14.200, -13.610, -13.118, -12.745, -16.076, -15.969,
    -15.819, -15.612, -15.277, -14.802, -14.204, -13.613, -13.119, -12.746,
    -16.086, -15.979, -15.828, -15.620, -15.285, -14.807, -14.208, -13.616,
    -13.121, -12.746, -16.096, -15.988, -15.836, -15.628, -15.292, -14.812,
    -14.212, -13.619, -13.123, -12.747, -16.105, -15.997, -15.845, -15.636,
    -15.300, -14.818, -14.217, -13.622, -13.124, -12.748, -16.115, -16.006,
    -15.853, -15.644, -15.308, -14.823, -14.222, -13.626, -13.126, -12.749,
    -16.125, -16.015, -15.861, -15.652, -15.315, -14.828, -14.226, -13.628,
    -13.127, -12.750, -16.134, -16.024, -15.870, -15.660, -15.322, -14.834,
    -14.230, -13.631, -13.129, -12.751, -16.143, -16.033, -15.878, -15.667,
    -15.329, -14.839, -14.234, -13.633, -13.131, -12.752, -16.153, -16.042,
    -15.887, -15.675, -15.336, -14.844, -14.237, -13.636, -13.132, -12.752,
    -16.162, -16.051, -15.895, -15.683, -15.342, -14.849, -14.241, -13.638,
    -13.134, -12.753, -16.171, -16.060, -15.903, -15.690, -15.349, -14.855,
    -14.245, -13.641, -13.135, -12.754, -16.181, -16.069, -15.911, -15.698,
    -15.356, -14.860, -14.249, -13.643, -13.137, -12.755, -16.190, -16.078,
    -15.920, -15.705, -15.362, -14.865, -14.253, -13.645, -13.138, -12.756,
    -16.199, -16.086, -15.928, -15.713, -15.369, -14.871, -14.256, -13.648,
    -13.140, -12.757, -16.208, -16.095, -15.936, -15.720, -15.375, -14.876,
    -14.260, -13.650, -13.141, -12.758, -16.217, -16.104, -15.944, -15.727,
    -15.381, -14.881, -14.264, -13.653, -13.143, -12.759, -16.226, -16.112,
    -15.952, -15.735, -15.388, -14.887, -14.268, -13.655, -13.144, -12.760,
    -16.235, -16.121, -15.960, -15.742, -15.394, -14.892, -14.272, -13.658,
    -13.146, -12.761, -16.244, -16.129, -15.968, -15.749, -15.401, -14.897,
    -14.275, -13.661, -13.148, -12.762, -16.253, -16.138, -15.976, -15.757,
    -15.407, -14.903, -14.279, -13.664, -13.150, -12.763, -16.262, -16.146,
    -15.983, -15.764, -15.413, -14.908, -14.283, -13.666, -13.152, -12.764,
    -16.271, -16.154, -15.991, -15.771, -15.420, -14.913, -14.287, -13.669,
    -13.154, -12.765, -16.280, -16.163, -15.999, -15.778, -15.426, -14.918,
    -14.291, -13.672, -13.156, -12.767, -16.288, -16.171, -16.006, -15.785,
    -15.432, -14.924, -14.295, -13.675, -13.158, -12.768, -16.297, -16.179,
    -16.014, -15.792, -15.438, -14.929, -14.299, -13.678, -13.160, -12.769,
    -16.306, -16.187, -16.022, -15.799, -15.444, -14.934, -14.303, -13.681,
    -13.162, -12.770, -16.314, -16.195, -16.029, -15.806, -15.450, -14.940,
    -14.308, -13.685, -13.165, -12.772, -16.323, -16.203, -16.036, -15.813,
    -15.457, -14.945, -14.312, -13.688, -13.167, -12.773, -16.331, -16.211,
    -16.044, -15.819, -15.463, -14.951, -14.317, -13.692, -13.170, -12.775,
    -16.340, -16.219, -16.051, -15.826, -15.469, -14.957, -14.322, -13.695,
    -13.172, -12.777, -16.348, -16.227, -16.058, -15.833, -15.475, -14.962,
    -14.327, -13.699, -13.175, -12.778, -16.356, -16.234, -16.066, -15.839,
    -15.482, -14.968, -14.332, -13.703, -13.177, -12.780, -16.365, -16.242,
    -16.073, -15.846, -15.488, -14.974, -14.337, -13.707, -13.180, -12.782,
    -16.373, -16.249, -16.080, -15.853, -15.494, -14.979, -14.342, -13.711,
    -13.183, -12.784, -16.381, -16.257, -16.087, -15.859, -15.500, -14.985,
    -14.347, -13.715, -13.186, -12.786, -16.389, -16.264, -16.094, -15.865,
    -15.506, -14.991, -14.352, -13.719, -13.189, -12.788, -16.397, -16.271,
    -16.101, -15.871, -15.512, -14.996, -14.357, -13.723, -13.192, -12.790,
    -16.405, -16.278, -16.107, -15.878, -15.518, -15.002, -14.363, -13.728,
    -13.196, -12.793, -16.413, -16.285, -16.114, -15.884, -15.524, -15.008,
    -14.368, -13.732, -13.199, -12.795, -16.421, -16.292, -16.120, -15.890,
    -15.529, -15.014, -14.373, -13.737, -13.202, -12.798, -16.429, -16.299,
    -16.127, -15.895, -15.535, -15.020, -14.379, -13.742, -13.206, -12.800,
    -16.436, -16.306, -16.133, -15.901, -15.541, -15.025, -14.384, -13.746,
    -13.210, -12.803, -16.444, -16.312, -16.140, -15.907, -15.547, -15.031,
    -14.390, -13.751, -13.213, -12.805, -16.451, -16.319, -16.146, -15.912,
    -15.552, -15.037, -14.395, -13.756, -13.217, -12.808, -16.459, -16.325,
    -16.152, -15.918, -15.558, -15.043, -14.401, -13.761, -13.221, -12.811,
    -16.466, -16.332, -16.158, -15.923, -15.564, -15.049, -14.407, -13.766,
    -13.225, -12.814, -16.474, -16.338, -16.164, -15.929, -15.569, -15.055,
    -14.412, -13.771, -13.229, -12.817, -16.481, -16.345, -16.170, -15.934,
    -15.575, -15.060, -14.418, -13.776, -13.233, -12.820, -16.488, -16.352,
    -16.175, -15.939, -15.580, -15.066, -14.423, -13.781, -13.237, -12.823,
    -16.495, -16.358, -16.181, -15.944, -15.586, -15.072, -14.429, -13.786,
    -13.241, -12.826, -16.502, -16.365, -16.186, -15.949, -15.591, -15.078,
    -14.435, -13.791, -13.245, -12.829, -16.509, -16.371, -16.192, -15.954,
    -15.596, -15.084, -14.441, -13.797, -13.250, -12.832, -16.516, -16.378,
    -16.197, -15.959, -15.602, -15.089, -14.446, -13.802, -13.254, -12.835,
    -16.522, -16.384, -16.202, -15.963, -15.607, -15.095, -14.452, -13.808,
    -13.258, -12.839, -16.529, -16.391, -16.207, -15.968, -15.612, -15.101,
    -14.458, -13.813, -13.263, -12.842, -16.535, -16.397, -16.212, -15.972,
    -15.617, -15.107, -14.464, -13.818, -13.267, -12.845, -16.541, -16.402,
    -16.216, -15.977, -15.621, -15.112, -14.470, -13.824, -13.272, -12.849,
    -16.547, -16.407, -16.220, -15.981, -15.626, -15.117, -14.475, -13.829,
    -13.277, -12.852, -16.553, -16.413, -16.225, -15.985, -15.631, -15.123,
    -14.481, -13.835, -13.281, -12.856, -16.559, -16.418, -16.229, -15.989,
    -15.635, -15.128, -14.487, -13.840, -13.286, -12.859, -16.564, -16.422,
    -16.232, -15.992, -15.639, -15.133, -14.493, -13.846, -13.291, -12.863,
    -16.570, -16.427, -16.236, -15.996, -15.643, -15.139, -14.498, -13.852,
    -13.295, -12.867, -16.575, -16.432, -16.239, -15.999, -15.647, -15.144,
    -14.504, -13.857, -13.300, -12.870, -16.580, -16.436, -16.243, -16.003,
    -15.651, -15.149, -14.510, -13.863, -13.305, -12.874, -16.585, -16.440,
    -16.246, -16.006, -15.655, -15.154, -14.516, -13.869, -13.310, -12.878,
    -16.589, -16.444, -16.249, -16.009, -15.659, -15.159, -14.522, -13.875,
    -13.315, -12.882, -16.594, -16.447, -16.252, -16.012, -15.663, -15.164,
    -14.528, -13.881, -13.320, -12.886, -16.598, -16.451, -16.255, -16.014,
    -15.667, -15.169, -14.534, -13.886, -13.324, -12.889, -16.602, -16.454,
    -16.258, -16.016, -15.670, -15.174, -14.540, -13.892, -13.329, -12.893,
    -16.605, -16.457, -16.261, -16.019, -15.674, -15.179, -14.546, -13.898,
    -13.334, -12.897, -16.609, -16.460, -16.263, -16.021, -15.677, -15.184,
    -14.552, -13.904, -13.339, -12.901, -16.612, -16.462, -16.266, -16.022,
    -15.681, -15.189, -14.558, -13.910, -13.344, -12.905, -16.615, -16.465,
    -16.268, -16.024, -15.684, -15.194, -14.565, -13.916, -13.349, -12.909,
    -16.617, -16.467, -16.270, -16.026, -15.688, -15.199, -14.571, -13.922,
    -13.354, -12.913, -16.620, -16.469, -16.272, -16.027, -15.691, -15.204,
    -14.577, -13.928, -13.359, -12.917, -16.622, -16.469, -16.272, -16.028,
    -15.694, -15.209, -14.583, -13.934, -13.364, -12.921, -16.624, -16.469,
    -16.272, -16.028, -15.696, -15.213, -14.589, -13.940, -13.369, -12.925,
    -16.625, -16.469, -16.271, -16.029, -15.699, -15.217, -14.595, -13.946,
    -13.374, -12.929, -16.626, -16.468, -16.270, -16.029, -15.701, -15.221,
    -14.600, -13.951, -13.379, -12.933, -16.627, -16.466, -16.268, -16.029,
    -15.703, -15.226, -14.606, -13.957, -13.385, -12.937, -16.627, -16.464,
    -16.266, -16.029, -15.705, -15.230, -14.612, -13.963, -13.390, -12.941,
    -16.627, -16.462, -16.263, -16.028, -15.707, -15.234, -14.617, -13.969,
    -13.395, -12.945, -16.627, -16.459, -16.260, -16.027, -15.709, -15.238,
    -14.623, -13.974, -13.400, -12.950, -16.627, -16.456, -16.256, -16.026,
    -15.711, -15.241, -14.628, -13.980, -13.406, -12.954, -16.626, -16.452,
    -16.252, -16.025, -15.712, -15.245, -14.634, -13.986, -13.411, -12.958,
    -16.624, -16.449, -16.250, -16.023, -15.713, -15.249, -14.640, -13.992,
    -13.416, -12.962, -16.622, -16.446, -16.248, -16.021, -15.715, -15.252,
    -14.645, -13.998, -13.422, -12.966, -16.619, -16.443, -16.246, -16.019,
    -15.716, -15.256, -14.651, -14.004, -13.427, -12.971, -16.616, -16.439,
    -16.245, -16.017, -15.716, -15.259, -14.656, -14.010, -13.432, -12.975,
    -16.612, -16.436, -16.243, -16.014, -15.717, -15.263, -14.661, -14.016,
    -13.437, -12.979, -16.608, -16.432, -16.241, -16.011, -15.718, -15.266,
    -14.667, -14.022, -13.443, -12.984, -16.603, -16.428, -16.239, -16.008,
    -15.718, -15.270, -14.672, -14.028, -13.448, -12.988, -16.598, -16.423,
    -16.237, -16.004, -15.719, -15.273, -14.677, -14.035, -13.453, -12.992,
    -16.592, -16.419, -16.236, -16.000, -15.719, -15.276, -14.683, -14.041,
    -13.459, -12.997, -16.586, -16.414, -16.234, -15.996, -15.719, -15.279,
    -14.688, -14.047, -13.464, -13.001, -16.578, -16.408, -16.228, -15.991,
    -15.719, -15.282, -14.693, -14.053, -13.469, -13.005, -16.570, -16.401,
    -16.221, -15.986, -15.719, -15.285, -14.698, -14.059, -13.475, -13.010,
    -16.561, -16.393, -16.213, -15.981, -15.719, -15.288, -14.704, -14.065,
    -13.480, -13.014, -16.552, -16.386, -16.204, -15.975, -15.718, -15.291,
    -14.709, -14.071, -13.485, -13.018, -16.541, -16.377, -16.194, -15.969,
    -15.718, -15.294, -14.714, -14.077, -13.491, -13.023, -16.530, -16.369,
    -16.184, -15.963, -15.717, -15.297, -14.719, -14.083, -13.496, -13.027,
    -16.519, -16.359, -16.173, -15.956, -15.716, -15.300, -14.724, -14.089,
    -13.501, -13.032, -16.506, -16.350, -16.161, -15.949, -15.715, -15.303,
    -14.729, -14.094, -13.506, -13.036, -16.494, -16.340, -16.148, -15.942,
    -15.714, -15.305, -14.734, -14.100, -13.512, -13.041, -16.480, -16.329,
    -16.134, -15.934, -15.713, -15.308, -14.739, -14.106, -13.517, -13.045,
    -16.466, -16.318, -16.126, -15.927, -15.710, -15.310, -14.744, -14.112,
    -13.522, -13.049, -16.452, -16.307, -16.119, -15.919, -15.706, -15.311,
    -14.749, -14.118, -13.528, -13.054, -16.437, -16.296, -16.113, -15.911,
    -15.701, -15.312, -14.754, -14.124, -13.533, -13.058, -16.422, -16.284,
    -16.107, -15.903, -15.696, -15.313, -14.759, -14.130, -13.539, -13.063,
    -16.406, -16.271, -16.101, -15.894, -15.691, -15.313, -14.764, -14.136,
    -13.544, -13.067, -16.390, -16.258, -16.096, -15.886, -15.685, -15.314,
    -14.769, -14.142, -13.550, -13.072, -16.372, -16.245, -16.091, -15.877,
    -15.678, -15.314, -14.774, -14.148, -13.555, -13.076, -16.355, -16.231,
    -16.086, -15.868, -15.671, -15.314, -14.779, -14.154, -13.561, -13.081,
    -16.337, -16.217, -16.082, -15.859, -15.664, -15.314, -14.784, -14.160,
    -13.566, -13.085, -16.318, -16.203, -16.079, -15.849, -15.656, -15.313,
    -14.789, -14.166, -13.572, -13.090, -16.300, -16.185, -16.070, -15.843,
    -15.655, -15.315, -14.794, -14.172, -13.577, -13.094, -16.282, -16.166,
    -16.059, -15.838, -15.655, -15.318, -14.799, -14.178, -13.583, -13.099,
    -16.264, -16.146, -16.049, -15.833, -15.656, -15.321, -14.804, -14.184,
    -13.588, -13.103, -16.245, -16.126, -16.037, -15.828, -15.658, -15.324,
    -14.809, -14.190, -13.594, -13.108, -16.226, -16.104, -16.025, -15.824,
    -15.661, -15.328, -14.814, -14.196, -13.599, -13.112, -16.206, -16.081,
    -16.012, -15.821, -15.665, -15.332, -14.819, -14.202, -13.605, -13.117,
    -16.187, -16.058, -15.998, -15.817, -15.669, -15.336, -14.823, -14.209,
    -13.610, -13.121, -16.167, -16.033, -15.984, -15.815, -15.674, -15.341,
    -14.828, -14.215, -13.615, -13.125, -16.147, -16.008, -15.969, -15.813,
    -15.681, -15.346, -14.833, -14.221, -13.621, -13.130, -16.126, -15.981,
    -15.953, -15.811, -15.688, -15.351, -14.838, -14.227, -13.626, -13.134,
    -16.104, -15.965, -15.944, -15.805, -15.687, -15.354, -14.843, -14.233,
    -13.631, -13.138, -16.081, -15.950, -15.936, -15.798, -15.686, -15.357,
    -14.847, -14.239, -13.637, -13.143, -16.058, -15.937, -15.929, -15.791,
    -15.684, -15.359, -14.852, -14.245, -13.642, -13.147, -16.034, -15.924,
    -15.922, -15.783, -15.681, -15.361, -14.857, -14.251, -13.648, -13.152,
    -16.010, -15.913, -15.916, -15.775, -15.677, -15.364, -14.861, -14.257,
    -13.653, -13.156, -15.985, -15.903, -15.911, -15.767, -15.673, -15.366,
    -14.866, -14.263, -13.658, -13.161, -15.959, -15.893, -15.906, -15.758,
    -15.668, -15.367, -14.871, -14.269, -13.664, -13.165, -15.934, -15.885,
    -15.903, -15.749, -15.662, -15.369, -14.875, -14.274, -13.669, -13.170,
    -15.907, -15.879, -15.899, -15.739, -15.655, -15.371, -14.880, -14.280,
    -13.675, -13.174, -15.880, -15.873, -15.897, -15.729, -15.648, -15.372,
    -14.884, -14.286, -13.680, -13.179, -15.864, -15.867, -15.892, -15.730,
    -15.648, -15.375, -14.889, -14.292, -13.685, -13.183, -15.851, -15.861,
    -15.888, -15.732, -15.649, -15.378, -14.894, -14.298, -13.691, -13.188,
    -15.838, -15.857, -15.884, -15.736, -15.650, -15.381, -14.900, -14.304,
    -13.696, -13.192, -15.828, -15.853, -15.879, -15.742, -15.652, -15.385,
    -14.905, -14.310, -13.702, -13.197, -15.819, -15.850, -15.875, -15.750,
    -15.655, -15.388, -14.910, -14.316, -13.707, -13.201, -15.812, -15.848,
    -15.872, -15.759, -15.659, -15.392, -14.916, -14.322, -13.712, -13.206,
    -15.806, -15.846, -15.868, -15.769, -15.663, -15.396, -14.922, -14.328,
    -13.718, -13.210, -15.802, -15.846, -15.864, -15.781, -15.668, -15.400,
    -14.927, -14.335, -13.723, -13.214, -15.800, -15.846, -15.861, -15.795,
    -15.674, -15.404, -14.933, -14.341, -13.729, -13.219, -15.799, -15.847,
    -15.858, -15.811, -15.680, -15.408, -14.939, -14.347, -13.734, -13.223,
    -15.797, -15.847, -15.857, -15.816, -15.683, -15.412, -14.944, -14.353,
    -13.739, -13.227, -15.795, -15.847, -15.857, -15.820, -15.686, -15.416,
    -14.950, -14.359, -13.745, -13.232, -15.795, -15.848, -15.857, -15.824,
    -15.689, -15.420, -14.955, -14.365, -13.750, -13.236, -15.796, -15.849,
    -15.858, -15.827, -15.692, -15.424, -14.961, -14.371, -13.755, -13.241,
    -15.798, -15.851, -15.859, -15.829, -15.695, -15.428, -14.966, -14.378,
    -13.761, -13.245, -15.800, -15.853, -15.861, -15.831, -15.698, -15.432,
    -14.972, -14.384, -13.766, -13.249, -15.804, -15.855, -15.864, -15.832,
    -15.701, -15.437, -14.977, -14.390, -13.771, -13.254, -15.809, -15.858,
    -15.867, -15.833, -15.703, -15.441, -14.982, -14.396, -13.776, -13.258,
    -15.815, -15.861, -15.871, -15.833, -15.706, -15.446, -14.988, -14.402,
    -13.782, -13.263, -15.822, -15.865, -15.875, -15.832, -15.709, -15.450,
    -14.993, -14.408, -13.787, -13.267, -15.825, -15.868, -15.878, -15.835,
    -15.712, -15.454, -14.999, -14.414, -13.792, -13.271, -15.829, -15.871,
    -15.881, -15.838, -15.716, -15.459, -15.004, -14.420, -13.797, -13.276,
    -15.833, -15.875, -15.884, -15.841, -15.719, -15.463, -15.010, -14.426,
    -13.803, -13.280, -15.837, -15.878, -15.887, -15.844, -15.723, -15.467,
    -15.015, -14.432, -13.808, -13.284, -15.841, -15.882, -15.891, -15.848,
    -15.726, -15.472, -15.021, -14.438, -13.813, -13.289, -15.846, -15.886,
    -15.895, -15.851, -15.730, -15.476, -15.026, -14.445, -13.818, -13.293,
    -15.851, -15.891, -15.899, -15.855, -15.734, -15.481, -15.032, -14.451,
    -13.824, -13.297, -15.856, -15.895, -15.903, -15.859, -15.738, -15.485,
    -15.038, -14.457, -13.829, -13.301, -15.861, -15.900, -15.907, -15.863,
    -15.742, -15.490, -15.043, -14.463, -13.834, -13.306, -15.866, -15.905,
    -15.912, -15.867, -15.746, -15.494, -15.049, -14.469, -13.839, -13.310,
    -15.874, -15.912, -15.918, -15.872, -15.751, -15.499, -15.055, -14.475,
    -13.844, -13.314, -15.883, -15.920, -15.924, -15.878, -15.756, -15.504,
    -15.061, -14.481, -13.850, -13.319, -15.893, -15.928, -15.931, -15.884,
    -15.761, -15.510, -15.067, -14.488, -13.855, -13.323, -15.904, -15.936,
    -15.939, -15.890, -15.767, -15.515, -15.073, -14.494, -13.860, -13.327,
    -15.915, -15.946, -15.946, -15.896, -15.773, -15.521, -15.079, -14.500,
    -13.865, -13.332, -15.927, -15.955, -15.954, -15.903, -15.779, -15.527,
    -15.085, -14.507, -13.871, -13.336, -15.940, -15.966, -15.963, -15.911,
    -15.785, -15.533, -15.091, -14.513, -13.876, -13.341, -15.953, -15.977,
    -15.972, -15.918, -15.791, -15.539, -15.097, -14.519, -13.881, -13.345,
    -15.967, -15.988, -15.981, -15.926, -15.798, -15.545, -15.103, -14.526,
    -13.887, -13.350, -15.982, -16.000, -15.991, -15.934, -15.805, -15.551,
    -15.109, -14.532, -13.892, -13.354, -15.995, -16.011, -16.000, -15.942,
    -15.812, -15.557, -15.115, -14.538, -13.897, -13.358, -16.009, -16.023,
    -16.010, -15.950, -15.819, -15.563, -15.121, -14.545, -13.902, -13.363,
    -16.023, -16.035, -16.020, -15.958, -15.826, -15.570, -15.128, -14.551,
    -13.908, -13.367, -16.037, -16.047, -16.030, -15.967, -15.833, -15.576,
    -15.134, -14.557, -13.913, -13.372, -16.052, -16.059, -16.040, -15.975,
    -15.840, -15.583, -15.141, -14.564, -13.918, -13.376, -16.067, -16.072,
    -16.051, -15.984, -15.848, -15.590, -15.147, -14.570, -13.923, -13.381,
    -16.082, -16.085, -16.062, -15.993, -15.856, -15.596, -15.153, -14.577,
    -13.929, -13.385, -16.098, -16.098, -16.073, -16.003, -15.864, -15.603,
    -15.160, -14.583, -13.934, -13.390, -16.114, -16.112, -16.084, -16.012,
    -15.872, -15.610, -15.166, -14.590, -13.939, -13.394, -16.130, -16.126,
    -16.096, -16.022, -15.880, -15.617, -15.173, -14.596, -13.944, -13.399,
    -16.145, -16.139, -16.107, -16.031, -15.888, -15.624, -15.180, -14.602,
    -13.949, -13.403, -16.161, -16.152, -16.118, -16.041, -15.896, -15.631,
    -15.186, -14.609, -13.954, -13.407, -16.176, -16.166, -16.130, -16.051,
    -15.904, -15.638, -15.193, -14.615, -13.959, -13.412, -16.191, -16.179,
    -16.141, -16.060, -15.913, -15.645, -15.199, -14.622, -13.964, -13.416,
    -16.207, -16.193, -16.153, -16.070, -15.921, -15.653, -15.206, -14.628,
    -13.969, -13.420, -16.223, -16.206, -16.164, -16.080, -15.929, -15.660,
    -15.213, -14.634, -13.974, -13.424, -16.238, -16.220, -16.176, -16.090,
    -15.938, -15.667, -15.220, -14.641, -13.979, -13.428, -16.254, -16.234,
    -16.188, -16.100, -15.946, -15.675, -15.226, -14.647, -13.983, -13.431,
    -16.270, -16.248, -16.200, -16.111, -15.955, -15.682, -15.233, -14.654,
    -13.988, -13.435, -16.286, -16.262, -16.212, -16.121, -15.964, -15.690,
    -15.240, -14.660, -13.993, -13.439, -16.301, -16.276, -16.224, -16.131,
    -15.973, -15.697, -15.247, -14.666, -13.998, -13.443, -16.317, -16.289,
    -16.236, -16.141, -15.981, -15.705, -15.254, -14.673, -14.004, -13.448,
    -16.332, -16.303, -16.247, -16.151, -15.990, -15.712, -15.260, -14.679,
    -14.009, -13.452, -16.348, -16.317, -16.259, -16.162, -15.998, -15.720,
    -15.267, -14.686, -14.014, -13.457, -16.363, -16.330, -16.271, -16.172,
    -16.007, -15.727, -15.274, -14.692, -14.020, -13.461, -16.378, -16.344,
    -16.283, -16.182, -16.016, -15.735, -15.281, -14.699, -14.025, -13.466,
    -16.393, -16.358, -16.295, -16.193, -16.025, -15.742, -15.288, -14.705,
    -14.031, -13.471, -16.409, -16.372, -16.307, -16.203, -16.033, -15.750,
    -15.295, -14.712, -14.037, -13.475, -16.424, -16.385, -16.319, -16.214,
    -16.042, -15.757, -15.302, -14.718, -14.042, -13.480, -16.439, -16.399,
    -16.331, -16.224, -16.051, -15.765, -15.309, -14.725, -14.048, -13.485,
    -16.454, -16.412, -16.343, -16.234, -16.060, -15.773, -15.316, -14.731,
    -14.053, -13.490, -16.469, -16.426, -16.355, -16.245, -16.069, -15.780,
    -15.323, -14.738, -14.059, -13.494, -16.484, -16.439, -16.367, -16.255,
    -16.077, -15.788, -15.330, -14.744, -14.064, -13.499, -16.499, -16.453,
    -16.378, -16.265, -16.086, -15.795, -15.337, -14.751, -14.070, -13.503,
    -16.514, -16.466, -16.390, -16.276, -16.095, -15.803, -15.344, -14.757,
    -14.075, -13.508, -16.528, -16.479, -16.402, -16.286, -16.104, -15.810,
    -15.350, -14.764, -14.081, -13.513, -16.543, -16.493, -16.414, -16.296,
    -16.113, -15.818, -15.357, -14.771, -14.086, -13.517, -16.558, -16.506,
    -16.425, -16.306, -16.122, -15.826, -15.364, -14.777, -14.091, -13.522,
    -16.572, -16.519, -16.437, -16.317, -16.131, -15.833, -15.371, -14.784,
    -14.097, -13.526, -16.587, -16.532, -16.449, -16.327, -16.140, -15.841,
    -15.378, -14.790, -14.102, -13.531, -16.601, -16.545, -16.461, -16.337,
    -16.149, -15.849, -15.385, -14.796, -14.107, -13.536, -16.616, -16.558,
    -16.472, -16.347, -16.158, -15.856, -15.392, -14.803, -14.113, -13.540,
    -16.630, -16.571, -16.484, -16.358, -16.167, -15.864, -15.399, -14.809,
    -14.118, -13.545, -16.644, -16.584, -16.496, -16.368, -16.175, -15.872,
    -15.406, -14.816, -14.123, -13.549, -16.658, -16.597, -16.507, -16.378,
    -16.184, -15.879, -15.412, -14.823, -14.129, -13.554, -16.671, -16.609,
    -16.519, -16.388, -16.193, -15.887, -15.419, -14.829, -14.134, -13.558,
    -16.685, -16.622, -16.530, -16.399, -16.202, -15.895, -15.426, -14.835,
    -14.139, -13.563, -16.699, -16.635, -16.542, -16.409, -16.210, -15.903,
    -15.433, -14.842, -14.144, -13.567, -16.713, -16.647, -16.553, -16.419,
    -16.219, -15.910, -15.440, -14.849, -14.150, -13.572, -16.726, -16.660,
    -16.565, -16.429, -16.228, -15.918, -15.447, -14.855, -14.155, -13.576,
    -16.739, -16.673, -16.576, -16.439, -16.237, -15.926, -15.454, -14.861,
    -14.160, -13.580, -16.753, -16.685, -16.588, -16.449, -16.245, -15.933,
    -15.461, -14.868, -14.166, -13.585, -16.766, -16.697, -16.599, -16.459,
    -16.254, -15.941, -15.468, -14.874, -14.171, -13.590, -16.779, -16.710,
    -16.610, -16.469, -16.263, -15.949, -15.474, -14.881, -14.176, -13.594,
    -16.792, -16.722, -16.621, -16.479, -16.272, -15.956, -15.481, -14.887,
    -14.182, -13.599, -16.806, -16.734, -16.632, -16.489, -16.280, -15.964,
    -15.488, -14.894, -14.187, -13.603, -16.819, -16.747, -16.643, -16.499,
    -16.289, -15.971, -15.495, -14.900, -14.192, -13.608, -16.831, -16.759,
    -16.654, -16.508, -16.298, -15.979, -15.501, -14.906, -14.197, -13.612,
    -16.844, -16.771, -16.665, -16.518, -16.306, -15.986, -15.508, -14.913,
    -14.203, -13.617, -16.857, -16.783, -16.676, -16.528, -16.315, -15.994,
    -15.515, -14.919, -14.208, -13.621, -16.870, -16.795, -16.687, -16.538,
    -16.324, -16.001, -15.522, -14.925, -14.213, -13.626, -16.882, -16.807,
    -16.698, -16.548, -16.332, -16.009, -15.529, -14.932, -14.219, -13.630,
    -16.895, -16.819, -16.709, -16.557, -16.341, -16.016, -15.535, -14.938,
    -14.224, -13.635, -16.907, -16.830, -16.719, -16.567, -16.349, -16.024,
    -15.542, -14.944, -14.229, -13.639, -16.920, -16.842, -16.730, -16.577,
    -16.358, -16.031, -15.549, -14.950, -14.235, -13.644, -16.932, -16.853,
    -16.741, -16.586, -16.366, -16.039, -15.556, -14.956, -14.240, -13.648,
    -16.945, -16.865, -16.751, -16.596, -16.375, -16.046, -15.563, -14.963,
    -14.246, -13.653, -16.957, -16.876, -16.762, -16.606, -16.383, -16.053,
    -15.569, -14.969, -14.251, -13.658, -16.969, -16.888, -16.772, -16.615,
    -16.392, -16.061, -15.576, -14.975, -14.257, -13.662, -16.981, -16.899,
    -16.783, -16.625, -16.400, -16.068, -15.583, -14.981, -14.262, -13.667,
    -16.993, -16.910, -16.793, -16.634, -16.408, -16.075, -15.590, -14.987,
    -14.267, -13.672, -17.005, -16.921, -16.804, -16.644, -16.417, -16.083,
    -15.596, -14.993, -14.273, -13.676, -17.017, -16.933, -16.814, -16.653,
    -16.425, -16.090, -15.603, -15.000, -14.278, -13.681, -17.028, -16.944,
    -16.824, -16.663, -16.434, -16.097, -15.610, -15.006, -14.284, -13.686,
    -17.040, -16.955, -16.835, -16.672, -16.442, -16.105, -15.616, -15.012,
    -14.289, -13.690, -17.052, -16.966, -16.845, -16.681, -16.451, -16.112,
    -15.623, -15.018, -14.294, -13.695, -17.063, -16.977, -16.855, -16.691,
    -16.459, -16.119, -15.629, -15.024, -14.300, -13.700, -17.074, -16.987,
    -16.865, -16.700, -16.467, -16.126, -15.636, -15.031, -14.305, -13.705,
    -17.086, -16.998, -16.875, -16.709, -16.476, -16.134, -15.642, -15.037,
    -14.311, -13.709, -17.097, -17.009, -16.885, -16.718, -16.484, -16.141,
    -15.649, -15.043, -14.316, -13.714, -17.108, -17.020, -16.895, -16.727,
    -16.492, -16.148, -15.655, -15.049, -14.321, -13.719, -17.119, -17.030,
    -16.905, -16.736, -16.500, -16.155, -15.662, -15.055, -14.327, -13.724,
    -17.130, -17.041, -16.915, -16.745, -16.508, -16.163, -15.668, -15.061,
    -14.332, -13.728, -17.141, -17.051, -16.925, -16.754, -16.516, -16.170,
    -15.675, -15.068, -14.338, -13.733, -17.152, -17.062, -16.934, -16.763,
    -16.524, -16.177, -15.681, -15.074, -14.343, -13.738, -17.163, -17.072,
    -16.944, -16.772, -16.532, -16.184, -15.688, -15.080, -14.349, -13.743,
    -17.174, -17.082, -16.954, -16.781, -16.540, -16.192, -15.694, -15.086,
    -14.354, -13.747, -17.185, -17.093, -16.964, -16.790, -16.548, -16.199,
    -15.700, -15.092, -14.360, -13.752, -17.195, -17.103, -16.973, -16.799,
    -16.556, -16.206, -15.707, -15.098, -14.365, -13.757, -17.206, -17.113,
    -16.983, -16.808, -16.564, -16.213, -15.713, -15.104, -14.371, -13.762,
    -17.217, -17.123, -16.993, -16.817, -16.572, -16.220, -15.719, -15.110,
    -14.376, -13.767, -17.227, -17.133, -17.002, -16.826, -16.580, -16.227,
    -15.726, -15.116, -14.382, -13.772, -17.238, -17.143, -17.012, -16.834,
    -16.588, -16.234, -15.732, -15.122, -14.387, -13.777, -17.248, -17.153,
    -17.021, -16.843, -16.595, -16.241, -15.738, -15.128, -14.393, -13.781,
    -17.259, -17.163, -17.031, -16.852, -16.603, -16.248, -15.744, -15.134,
    -14.398, -13.786, -17.269, -17.173, -17.040, -16.860, -16.611, -16.255,
    -15.751, -15.140, -14.404, -13.791, -17.279, -17.183, -17.049, -16.869,
    -16.619, -16.262, -15.757, -15.146, -14.409, -13.796, -17.289, -17.193,
    -17.059, -16.878, -16.626, -16.268, -15.763, -15.151, -14.415, -13.801,
    -17.300, -17.203, -17.068, -16.886, -16.634, -16.275, -15.769, -15.157,
    -14.420, -13.806, -17.310, -17.213, -17.077, -16.895, -16.642, -16.282,
    -15.775, -15.163, -14.426, -13.811, -17.320, -17.223, -17.087, -16.903,
    -16.649, -16.289, -15.781, -15.169, -14.431, -13.816, -17.330, -17.232,
    -17.096, -16.912, -16.657, -16.295, -15.787, -15.175, -14.437, -13.821,
    -17.340, -17.242, -17.105, -16.920, -16.665, -16.302, -15.793, -15.180,
    -14.442, -13.825, -17.350, -17.252, -17.114, -16.929, -16.672, -16.309,
    -15.800, -15.186, -14.448, -13.830, -17.360, -17.261, -17.123, -16.937,
    -16.680, -16.315, -15.806, -15.192, -14.453, -13.835, -17.370, -17.271,
    -17.132, -16.945, -16.687, -16.322, -15.812, -15.198, -14.459, -13.840,
    -17.379, -17.280, -17.141, -16.954, -16.695, -16.329, -15.818, -15.203,
    -14.464, -13.845, -17.389, -17.290, -17.150, -16.962, -16.702, -16.335,
    -15.824, -15.209, -14.470, -13.850, -17.399, -17.299, -17.159, -16.970,
    -16.710, -16.342, -15.830, -15.215, -14.475, -13.855, -17.409, -17.309,
    -17.168, -16.978, -16.717, -16.348, -15.836, -15.220, -14.481, -13.860,
    -17.418, -17.318, -17.177, -16.987, -16.725, -16.355, -15.841, -15.226,
    -14.486, -13.865, -17.428, -17.327, -17.186, -16.995, -16.732, -16.361,
    -15.847, -15.231, -14.492, -13.870, -17.437, -17.337, -17.195, -17.003,
    -16.739, -16.367, -15.853, -15.237, -14.497, -13.875, -17.447, -17.346,
    -17.203, -17.011, -16.747, -16.374, -15.859, -15.242, -14.503, -13.880,
    -17.456, -17.355, -17.212, -17.019, -16.754, -16.380, -15.865, -15.248,
    -14.508, -13.885, -17.465, -17.364, -17.221, -17.027, -16.761, -16.386,
    -15.871, -15.254, -14.513, -13.890, -17.475, -17.373, -17.229, -17.035,
    -16.768, -16.392, -15.876, -15.259, -14.519, -13.895, -17.484, -17.382,
    -17.238, -17.043, -16.776, -16.399, -15.882, -15.264, -14.524, -13.900,
    -17.494, -17.391, -17.247, -17.051, -16.783, -16.405, -15.888, -15.270,
    -14.530, -13.905, -17.503, -17.400, -17.255, -17.059, -16.790, -16.411,
    -15.893, -15.275, -14.535, -13.910, -17.512, -17.409, -17.264, -17.067,
    -16.797, -16.417, -15.899, -15.281, -14.540, -13.915, -17.522, -17.418,
    -17.272, -17.075, -16.804, -16.423, -15.904, -15.286, -14.546, -13.920,
    -17.531, -17.427, -17.281, -17.083, -16.811, -16.429, -15.910, -15.292,
    -14.551, -13.925, -17.540, -17.436, -17.289, -17.091, -16.818, -16.435,
    -15.915, -15.297, -14.557, -13.930, -17.549, -17.445, -17.297, -17.098,
    -16.825, -16.441, -15.921, -15.302, -14.562, -13.934, -17.558, -17.454,
    -17.306, -17.106, -16.832, -16.447, -15.926, -15.308, -14.567, -13.939,
    -17.567, -17.462, -17.314, -17.114, -16.838, -16.453, -15.932, -15.313,
    -14.573, -13.944, -17.576, -17.471, -17.322, -17.122, -16.845, -16.459,
    -15.937, -15.318, -14.578, -13.949, -17.585, -17.480, -17.331, -17.129,
    -16.852, -16.465, -15.942, -15.323, -14.583, -13.954, -17.594, -17.488,
    -17.339, -17.137, -16.859, -16.471, -15.948, -15.329, -14.589, -13.959,
    -17.603, -17.497, -17.347, -17.145, -16.866, -16.476, -15.953, -15.334,
    -14.594, -13.964, -17.612, -17.505, -17.355, -17.152, -16.872, -16.482,
    -15.958, -15.339, -14.599, -13.969, -17.621, -17.514, -17.363, -17.160,
    -16.879, -16.488, -15.963, -15.344, -14.605, -13.974, -17.629, -17.522,
    -17.371, -17.167, -16.886, -16.494, -15.968, -15.349, -14.610, -13.979,
    -17.638, -17.531, -17.379, -17.175, -16.892, -16.499, -15.974, -15.355,
    -14.615, -13.984, -17.647, -17.539, -17.387, -17.182, -16.899, -16.505,
    -15.979, -15.360, -14.621, -13.989, -17.655, -17.547, -17.395, -17.190,
    -16.905, -16.510, -15.984, -15.365, -14.626, -13.994, -17.664, -17.556,
    -17.403, -17.197, -16.912, -16.516, -15.989, -15.370, -14.631, -13.999,
    -17.672, -17.564, -17.411, -17.205, -16.919, -16.522, -15.994, -15.375,
    -14.637, -14.004, -17.681, -17.572, -17.419, -17.212, -16.925, -16.527,
    -15.999, -15.380, -14.642, -14.009, -17.690, -17.580, -17.427, -17.219,
    -16.931, -16.532, -16.004, -15.385, -14.647, -14.014, -17.698, -17.589,
    -17.435, -17.227, -16.938, -16.538, -16.009, -15.390, -14.653, -14.019,
    -17.707, -17.597, -17.443, -17.234, -16.944, -16.543, -16.014, -15.395,
    -14.658, -14.024, -17.715, -17.605, -17.451, -17.241, -16.950, -16.549,
    -16.019, -15.399, -14.663, -14.029, -17.724, -17.613, -17.458, -17.248,
    -16.957, -16.554, -16.024, -15.404, -14.668, -14.034, -17.732, -17.622,
    -17.466, -17.255, -16.963, -16.559, -16.029, -15.409, -14.673, -14.039,
    -17.740, -17.630, -17.474, -17.263, -16.969, -16.564, -16.033, -15.414,
    -14.679, -14.044, -17.749, -17.638, -17.482, -17.270, -16.975, -16.570,
    -16.038, -15.418, -14.684, -14.049, -17.757, -17.646, -17.490, -17.277,
    -16.981, -16.575, -16.043, -15.423, -14.689, -14.054, -17.765, -17.654,
    -17.497, -17.284, -16.987, -16.580, -16.048, -15.428, -14.694, -14.059,
    -17.774, -17.662, -17.505, -17.291, -16.993, -16.585, -16.053, -15.432,
    -14.699, -14.064, -17.782, -17.670, -17.513, -17.298, -16.999, -16.590,
    -16.057, -15.437, -14.705, -14.069, -17.790, -17.678, -17.520, -17.305,
    -17.005, -16.595, -16.062, -15.442, -14.710, -14.073, -17.798, -17.686,
    -17.528, -17.312, -17.011, -16.600, -16.067, -15.446, -14.715, -14.078,
    -17.806, -17.694, -17.536, -17.319, -17.017, -16.605, -16.071, -15.451,
    -14.720, -14.083, -17.815, -17.702, -17.543, -17.326, -17.023, -16.610,
    -16.076, -15.455, -14.725, -14.088, -17.823, -17.710, -17.551, -17.332,
    -17.029, -16.615, -16.080, -15.459, -14.730, -14.093, -17.831, -17.718,
    -17.558, -17.339, -17.034, -16.620, -16.085, -15.464, -14.735, -14.098,
    -17.839, -17.726, -17.566, -17.346, -17.040, -16.625, -16.089, -15.468,
    -14.740, -14.103, -17.847, -17.733, -17.573, -17.353, -17.046, -16.629,
    -16.094, -15.473, -14.746, -14.108, -17.855, -17.741, -17.581, -17.359,
    -17.052, -16.634, -16.098, -15.477, -14.751, -14.113, -17.862, -17.749,
    -17.588, -17.366, -17.057, -16.639, -16.103, -15.481, -14.756, -14.118,
    -17.870, -17.757, -17.595, -17.373, -17.063, -16.644, -16.107, -15.486,
    -14.761, -14.123, -17.878, -17.764, -17.603, -17.379, -17.068, -16.648,
    -16.112, -15.490, -14.766, -14.128, -17.886, -17.772, -17.610, -17.386,
    -17.074, -16.653, -16.116, -15.494, -14.771, -14.133, -17.894, -17.780,
    -17.617, -17.393, -17.080, -16.658, -16.121, -15.499, -14.776, -14.138,
    -17.902, -17.787, -17.625, -17.399, -17.085, -16.663, -16.125, -15.504,
    -14.781, -14.143, -17.910, -17.795, -17.632, -17.406, -17.091, -16.667,
    -16.130, -15.508, -14.786, -14.148, -17.917, -17.803, -17.639, -17.412,
    -17.096, -16.672, -16.134, -15.513, -14.791, -14.153, -17.925, -17.810,
    -17.646, -17.419, -17.102, -16.677, -16.139, -15.518, -14.797, -14.158,
    -17.933, -17.818, -17.654, -17.425, -17.107, -16.681, -16.143, -15.523,
    -14.802, -14.163, -17.941, -17.826, -17.661, -17.432, -17.113, -16.686,
    -16.148, -15.528, -14.807, -14.168, -17.949, -17.833, -17.668, -17.438,
    -17.118, -16.691, -16.152, -15.533, -14.812, -14.172, -17.956, -17.841,
    -17.675, -17.445, -17.123, -16.695, -16.157, -15.538, -14.817, -14.177,
    -17.964, -17.848, -17.682, -17.451, -17.129, -16.700, -16.161, -15.543,
    -14.822, -14.182, -17.972, -17.856, -17.689, -17.457, -17.134, -16.704,
    -16.166, -15.548, -14.827, -14.187, -17.979, -17.863, -17.696, -17.464,
    -17.139, -16.709, -16.170, -15.553, -14.832, -14.192, -17.987, -17.871,
    -17.703, -17.470, -17.145, -16.713, -16.174, -15.558, -14.837, -14.197,
    -17.995, -17.878, -17.710, -17.476, -17.150, -16.718, -16.179, -15.563,
    -14.842, -14.202, -18.002, -17.886, -17.717, -17.483, -17.155, -16.722,
    -16.183, -15.568, -14.847, -14.207, -18.010, -17.893, -17.724, -17.489,
    -17.160, -16.727, -16.188, -15.573, -14.853, -14.212, -18.017, -17.901,
    -17.731, -17.495, -17.166, -16.731, -16.192, -15.578, -14.858, -14.217,
    -18.025, -17.908, -17.738, -17.501, -17.171, -16.736, -16.197, -15.583,
    -14.863, -14.222, -18.032, -17.915, -17.745, -17.508, -17.176, -16.740,
    -16.201, -15.588, -14.868, -14.227, -18.040, -17.923, -17.752, -17.514,
    -17.181, -16.744, -16.205, -15.593, -14.873, -14.231, -18.047, -17.930,
    -17.759, -17.520, -17.186, -16.749, -16.210, -15.598, -14.878, -14.236,
    -18.055, -17.937, -17.766, -17.526, -17.191, -16.753, -16.214, -15.603,
    -14.883, -14.241, -18.062, -17.945, -17.772, -17.532, -17.196, -16.757,
    -16.218, -15.609, -14.888, -14.246, -18.070, -17.952, -17.779, -17.538,
    -17.201, -16.762, -16.223, -15.614, -14.893, -14.251, -18.077, -17.959,
    -17.786, -17.544, -17.206, -16.766, -16.227, -15.619, -14.898, -14.256,
    -18.084, -17.966, -17.793, -17.550, -17.211, -16.771, -16.232, -15.624,
    -14.903, -14.261, -18.092, -17.973, -17.800, -17.556, -17.217, -16.775,
    -16.236, -15.629, -14.908, -14.266, -18.099, -17.981, -17.807, -17.563,
    -17.222, -16.780, -16.241, -15.633, -14.913, -14.271, -18.107, -17.988,
    -17.813, -17.569, -17.227, -16.785, -16.245, -15.638, -14.918, -14.275,
    -18.114, -17.995, -17.820, -17.575, -17.233, -16.789, -16.250, -15.643,
    -14.923, -14.280, -18.121, -18.002, -17.827, -17.581, -17.238, -16.794,
    -16.254, -15.648, -14.928, -14.285, -18.129, -18.009, -17.834, -17.587,
    -17.243, -16.799, -16.259, -15.653, -14.933, -14.290, -18.136, -18.017,
    -17.840, -17.594, -17.249, -16.804, -16.264, -15.657, -14.938, -14.295,
    -18.143, -18.024, -17.847, -17.600, -17.254, -16.808, -16.268, -15.662,
    -14.943, -14.300, -18.150, -18.031, -17.854, -17.606, -17.259, -16.813,
    -16.273, -15.667, -14.948, -14.305, -18.158, -18.038, -17.861, -17.612,
    -17.265, -16.818, -16.277, -15.671, -14.953, -14.309, -18.165, -18.045,
    -17.867, -17.618, -17.270, -16.823, -16.282, -15.676, -14.958, -14.314,
    -18.172, -18.052, -17.874, -17.624, -17.275, -16.828, -16.287, -15.681,
    -14.963, -14.319, -18.179, -18.059, -17.881, -17.630, -17.281, -16.832,
    -16.291, -15.686, -14.968, -14.324, -18.186, -18.066, -17.887, -17.636,
    -17.286, -16.837, -16.296, -15.690, -14.973, -14.329, -18.193, -18.073,
    -17.894, -17.642, -17.292, -16.842, -16.301, -15.695, -14.978, -14.334,
    -18.201, -18.080, -17.901, -17.649, -17.297, -16.847, -16.305, -15.699,
    -14.983, -14.338, -18.208, -18.087, -17.907, -17.655, -17.302, -16.852,
    -16.310, -15.704, -14.988, -14.343, -18.215, -18.094, -17.914, -17.661,
    -17.308, -16.857, -16.315, -15.709, -14.993, -14.348, -18.222, -18.101,
    -17.920, -17.667, -17.313, -16.861, -16.319, -15.713, -14.998, -14.353,
    -18.229, -18.108, -17.927, -17.673, -17.318, -16.866, -16.324, -15.718,
    -15.003, -14.358, -18.236, -18.114, -17.933, -17.679, -17.324, -16.871,
    -16.329, -15.722, -15.008, -14.363, -18.243, -18.121, -17.940, -17.685,
    -17.329, -16.876, -16.334, -15.727, -15.013, -14.367, -18.250, -18.128,
    -17.946, -17.691, -17.335, -16.881, -16.338, -15.731, -15.018, -14.372,
    -18.257, -18.135, -17.953, -17.697, -17.340, -16.886, -16.343, -15.736,
    -15.023, -14.377, -18.264, -18.142, -17.959, -17.703, -17.345, -16.891,
    -16.348, -15.741, -15.028, -14.382, -18.271, -18.149, -17.966, -17.709,
    -17.351, -16.896, -16.353, -15.746, -15.033, -14.387, -18.278, -18.155,
    -17.972, -17.715, -17.356, -16.901, -16.357, -15.750, -15.038, -14.391,
    -18.285, -18.162, -17.979, -17.721, -17.362, -16.906, -16.362, -15.755,
    -15.043, -14.396, -18.292, -18.169, -17.985, -17.727, -17.367, -16.911,
    -16.367, -15.760, -15.048, -14.401, -18.299, -18.176, -17.992, -17.733,
    -17.373, -16.916, -16.372, -15.765, -15.053, -14.406, -18.306, -18.182,
    -17.998, -17.739, -17.378, -16.921, -16.377, -15.770, -15.058, -14.411,
    -18.313, -18.189, -18.005, -17.744, -17.384, -16.927, -16.382, -15.774,
    -15.063, -14.415, -18.320, -18.196, -18.011, -17.750, -17.389, -16.932,
    -16.387, -15.779, -15.068, -14.420, -18.326, -18.202, -18.017, -17.756,
    -17.395, -16.937, -16.392, -15.784, -15.073, -14.425, -18.333, -18.209,
    -18.024, -17.762, -17.401, -16.942, -16.397, -15.789, -15.078, -14.430,
    -18.340, -18.216, -18.030, -17.768, -17.406, -16.947, -16.402, -15.794,
    -15.083, -14.434, -18.347, -18.222, -18.036, -17.774, -17.412, -16.952,
    -16.406, -15.799, -15.088, -14.439, -18.354, -18.229, -18.043, -17.780,
    -17.417, -16.958, -16.411, -15.804, -15.093, -14.444, -18.361, -18.235,
    -18.049, -17.785, -17.423, -16.963, -16.416, -15.809, -15.098, -14.449,
    -18.367, -18.242, -18.055, -17.791, -17.428, -16.968, -16.421, -15.814,
    -15.103, -14.453, -18.374, -18.248, -18.061, -17.797, -17.434, -16.973,
    -16.426, -15.818, -15.108, -14.458, -18.381, -18.255, -18.068, -17.803,
    -17.440, -16.979, -16.431, -15.823, -15.113, -14.463, -18.388, -18.261,
    -18.074, -17.809, -17.445, -16.984, -16.437, -15.828, -15.118, -14.468,
    -18.395, -18.268, -18.080, -17.814, -17.451, -16.989, -16.442, -15.833,
    -15.123, -14.472, -18.401, -18.274, -18.086, -17.820, -17.456, -16.995,
    -16.447, -15.838, -15.128, -14.477, -18.408, -18.281, -18.093, -17.826,
    -17.462, -17.000, -16.452, -15.843, -15.133, -14.482, -18.415, -18.287,
    -18.099, -17.832, -17.468, -17.005, -16.457, -15.848, -15.138, -14.487,
    -18.421, -18.294, -18.105, -17.837, -17.473, -17.011, -16.462, -15.853,
    -15.143, -14.491, -18.428, -18.300, -18.111, -17.843, -17.479, -17.016,
    -16.467, -15.858, -15.148, -14.496, -18.435, -18.306, -18.117, -17.849,
    -17.484, -17.021, -16.472, -15.863, -15.153, -14.501, -18.441, -18.313,
    -18.123, -17.854, -17.490, -17.026, -16.477, -15.868, -15.158, -14.505,
    -18.448, -18.319, -18.129, -17.860, -17.495, -17.031, -16.482, -15.873,
    -15.163, -14.510, -18.454, -18.325, -18.134, -17.865, -17.500, -17.037,
    -16.487, -15.878, -15.168, -14.515, -18.461, -18.331, -18.140, -17.870,
    -17.505, -17.042, -16.492, -15.883, -15.173, -14.520, -18.467, -18.338,
    -18.146, -17.876, -17.510, -17.047, -16.497, -15.887, -15.178, -14.524,
    -18.474, -18.344, -18.152, -17.881, -17.516, -17.052, -16.502, -15.892,
    -15.183, -14.529, -18.480, -18.350, -18.157, -17.887, -17.521, -17.057,
    -16.507, -15.897, -15.188, -14.534, -18.487, -18.356, -18.163, -17.892,
    -17.526, -17.062, -16.512, -15.902, -15.193, -14.539, -18.493, -18.362,
    -18.169, -17.897, -17.531, -17.067, -16.517, -15.907, -15.197, -14.543,
    -18.500, -18.368, -18.174, -17.902, -17.536, -17.072, -16.522, -15.912,
    -15.202, -14.548, -18.506, -18.375, -18.180, -17.908, -17.541, -17.077,
    -16.527, -15.917, -15.207, -14.553, -18.513, -18.381, -18.185, -17.913,
    -17.546, -17.082, -16.532, -15.922, -15.212, -14.557, -18.519, -18.387,
    -18.191, -17.918, -17.551, -17.087, -16.537, -15.927, -15.217, -14.562,
    -18.525, -18.393, -18.196, -17.923, -17.556, -17.093, -16.542, -15.932,
    -15.222, -14.567, -18.532, -18.399, -18.202, -17.929, -17.561, -17.098,
    -16.547, -15.937, -15.227, -14.572, -18.538, -18.405, -18.207, -17.934,
    -17.566, -17.103, -16.552, -15.942, -15.232, -14.576, -18.544, -18.411,
    -18.212, -17.939, -17.571, -17.108, -16.558, -15.946, -15.237, -14.581,
    -18.551, -18.417, -18.218, -17.944, -17.576, -17.113, -16.563, -15.951,
    -15.242, -14.586, -18.557, -18.423, -18.223, -17.949, -17.581, -17.118,
    -16.568, -15.956, -15.247, -14.590, -18.563, -18.429, -18.228, -17.954,
    -17.586, -17.123, -16.573, -15.961, -15.251, -14.595, -18.569, -18.434,
    -18.233, -17.959, -17.591, -17.128, -16.578, -15.966, -15.256, -14.600,
    -18.576, -18.440, -18.239, -17.964, -17.595, -17.133, -16.583, -15.971,
    -15.261, -14.605, -18.582, -18.446, -18.244, -17.969, -17.600, -17.138,
    -16.588, -15.976, -15.266, -14.609, -18.588, -18.452, -18.249, -17.974,
    -17.605, -17.143, -16.593, -15.981, -15.271, -14.614, -18.594, -18.458,
    -18.254, -17.979, -17.609, -17.148, -16.598, -15.986, -15.276, -14.619,
    -18.600, -18.463, -18.259, -17.983, -17.614, -17.152, -16.603, -15.990,
    -15.281, -14.623, -18.606, -18.469, -18.263, -17.987, -17.618, -17.157,
    -16.607, -15.995, -15.285, -14.628, -18.612, -18.474, -18.268, -17.992,
    -17.623, -17.161, -16.612, -16.000, -15.290, -14.633, -18.619, -18.480,
    -18.273, -17.996, -17.627, -17.166, -16.617, -16.005, -15.295, -14.637,
    -18.625, -18.485, -18.278, -18.000, -17.631, -17.171, -16.621, -16.009,
    -15.300, -14.642, -18.631, -18.490, -18.282, -18.005, -17.635, -17.175,
    -16.626, -16.014, -15.304, -14.646, -18.637, -18.496, -18.287, -18.009,
    -17.639, -17.179, -16.631, -16.019, -15.309, -14.651, -18.643, -18.501,
    -18.291, -18.013, -17.644, -17.184, -16.635, -16.023, -15.314, -14.656,
    -18.649, -18.506, -18.296, -18.017, -17.648, -17.188, -16.640, -16.028,
    -15.319, -14.660, -18.655, -18.512, -18.300, -18.021, -17.652, -17.193,
    -16.645, -16.033, -15.323, -14.665, -18.661, -18.517, -18.304, -18.025,
    -17.656, -17.197, -16.649, -16.037, -15.328, -14.669, -18.666, -18.522,
    -18.309, -18.029, -17.659, -17.201, -16.654, -16.042, -15.333, -14.674,
    -18.672, -18.527, -18.313, -18.033, -17.663, -17.205, -16.658, -16.046,
    -15.337, -14.678, -18.678, -18.532, -18.317, -18.036, -17.667, -17.210,
    -16.663, -16.051, -15.342, -14.683, -18.684, -18.537, -18.321, -18.040,
    -17.671, -17.214, -16.667, -16.056, -15.347, -14.687, -18.690, -18.542,
    -18.325, -18.044, -17.675, -17.218, -16.672, -16.060, -15.351, -14.692,
    -18.696, -18.547, -18.329, -18.047, -17.678, -17.222, -16.676, -16.065,
    -15.356, -14.696, -18.702, -18.552, -18.333, -18.051, -17.682, -17.226,
    -16.681, -16.069, -15.360, -14.701, -18.707, -18.557, -18.337, -18.054,
    -17.686, -17.230, -16.685, -16.074, -15.365, -14.706, -18.713, -18.562,
    -18.341, -18.058, -17.689, -17.234, -16.690, -16.078, -15.370, -14.710,
    -18.719, -18.567, -18.345, -18.061, -17.693, -17.238, -16.694, -16.083,
    -15.374, -14.715, -18.725, -18.572, -18.349, -18.064, -17.696, -17.242,
    -16.698, -16.087, -15.379, -14.719, -18.730, -18.576, -18.352, -18.068,
    -17.700, -17.246, -16.703, -16.092, -15.383, -14.724, -18.736, -18.581,
    -18.356, -18.071, -17.703, -17.250, -16.707, -16.096, -15.388, -14.728,
    -18.741, -18.585, -18.359, -18.074, -17.706, -17.254, -16.711, -16.100,
    -15.393, -14.732, -18.746, -18.589, -18.362, -18.077, -17.709, -17.257,
    -16.715, -16.105, -15.397, -14.737, -18.751, -18.593, -18.365, -18.079,
    -17.712, -17.261, -16.719, -16.109, -15.402, -14.741, -18.756, -18.597,
    -18.368, -18.082, -17.715, -17.264, -16.724, -16.113, -15.406, -14.746,
    -18.761, -18.601, -18.371, -18.084, -17.718, -17.268, -16.728, -16.118,
    -15.411, -14.750, -18.765, -18.605, -18.374, -18.087, -17.721, -17.271,
    -16.732, -16.122, -15.415, -14.755, -18.770, -18.609, -18.376, -18.089,
    -17.723, -17.275, -16.736, -16.126, -15.420, -14.759, -18.775, -18.612,
    -18.379, -18.091, -17.726, -17.278, -16.740, -16.130, -15.424, -14.763,
    -18.779, -18.616, -18.381, -18.094, -17.729, -17.281, -16.744, -16.135,
    -15.428, -14.768, -18.784, -18.619, -18.384, -18.096, -17.731, -17.285,
    -16.748, -16.139, -15.433, -14.772, -18.788, -18.623, -18.386, -18.098,
    -17.734, -17.288, -16.751, -16.143, -15.437, -14.777, -18.792, -18.626,
    -18.388, -18.100, -17.736, -17.291, -16.755, -16.147, -15.442, -14.781,
    -18.797, -18.629, -18.390, -18.102, -17.739, -17.294, -16.759, -16.151,
    -15.446, -14.785, -18.801, -18.632, -18.392, -18.103, -17.741, -17.297,
    -16.763, -16.156, -15.450, -14.790, -18.805, -18.635, -18.394, -18.105,
    -17.743, -17.300, -16.767, -16.160, -15.455, -14.794, -18.809, -18.638,
    -18.396, -18.107, -17.745, -17.303, -16.771, -16.164, -15.459, -14.798,
    -18.813, -18.641, -18.398, -18.108, -17.747, -17.306, -16.774, -16.168,
    -15.464, -14.803, -18.817, -18.644, -18.400, -18.110, -17.749, -17.309,
    -16.778, -16.172, -15.468, -14.807, -18.821, -18.647, -18.401, -18.111,
    -17.751, -17.312, -16.782, -16.176, -15.472, -14.811, -18.825, -18.650,
    -18.403, -18.113, -17.753, -17.315, -16.785, -16.180, -15.477, -14.816,
    -18.829, -18.652, -18.404, -18.114, -17.755, -17.317, -16.789, -16.184,
    -15.481, -14.820, -18.832, -18.655, -18.405, -18.115, -17.757, -17.320,
    -16.792, -16.188, -15.485, -14.824, -18.836, -18.657, -18.407, -18.116,
    -17.759, -17.323, -16.796, -16.192, -15.489, -14.828, -18.840, -18.660,
    -18.408, -18.117, -17.760, -17.325, -16.800, -16.196, -15.494, -14.833,
    -18.843, -18.662, -18.409, -18.118, -17.762, -17.328, -16.803, -16.200,
    -15.498, -14.837, -18.845, -18.663, -18.410, -18.119, -17.763, -17.330,
    -16.806, -16.204, -15.502, -14.841, -18.848, -18.664, -18.410, -18.119,
    -17.765, -17.333, -16.810, -16.208, -15.506, -14.845, -18.850, -18.664,
    -18.410, -18.119, -17.766, -17.335, -16.813, -16.211, -15.510, -14.850,
    -18.852, -18.665, -18.410, -18.120, -17.767, -17.337, -16.816, -16.215,
    -15.515, -14.854, -18.854, -18.665, -18.410, -18.120, -17.768, -17.340,
    -16.819, -16.219, -15.519, -14.858, -18.856, -18.665, -18.410, -18.120,
    -17.769, -17.342, -16.822, -16.222, -15.523, -14.862, -18.857, -18.665,
    -18.410, -18.120, -17.770, -17.344, -16.825, -16.226, -15.527, -14.866,
    -18.859, -18.665, -18.410, -18.120, -17.771, -17.346, -16.828, -16.230,
    -15.531, -14.871, -18.860, -18.664, -18.409, -18.119, -17.772, -17.348,
    -16.831, -16.233, -15.535, -14.875, -18.861, -18.664, -18.409, -18.119,
    -17.772, -17.350, -16.834, -16.237, -15.539, -14.879, -18.863, -18.663,
    -18.408, -18.119, -17.773, -17.352, -16.837, -16.240, -15.543, -14.883,
    -18.863, -18.662, -18.407, -18.118, -17.774, -17.354, -16.840, -16.244,
    -15.547, -14.887, -18.864, -18.661, -18.406, -18.118, -17.774, -17.356,
    -16.843, -16.247, -15.551, -14.891, -18.865, -18.659, -18.405, -18.117,
    -17.775, -17.357, -16.845, -16.251, -15.555, -14.895, -18.865, -18.658,
    -18.404, -18.116, -17.775, -17.359, -16.848, -16.254, -15.559, -14.900,
    -18.866, -18.656, -18.403, -18.115, -17.775, -17.361, -16.851, -16.258,
    -15.563, -14.904, -18.866, -18.654, -18.401, -18.114, -17.775, -17.362,
    -16.853, -16.261, -15.567, -14.908, -18.866, -18.652, -18.399, -18.113,
    -17.775, -17.364, -16.856, -16.264, -15.570, -14.912, -18.866, -18.650,
    -18.398, -18.112, -17.775, -17.365, -16.859, -16.268, -15.574, -14.916,
    -18.866, -18.648, -18.396, -18.110, -17.775, -17.367, -16.861, -16.271,
    -15.578, -14.920, -18.866, -18.645, -18.394, -18.109, -17.775, -17.368,
    -16.864, -16.274, -15.582, -14.924, -18.865, -18.642, -18.392, -18.107,
    -17.775, -17.369, -16.866, -16.277, -15.586, -14.928, -18.865, -18.639,
    -18.390, -18.106, -17.775, -17.371, -16.868, -16.281, -15.589, -14.932,
    -18.864, -18.636, -18.387, -18.104, -17.774, -17.372, -16.871, -16.284,
    -15.593, -14.936, -18.863, -18.633, -18.385, -18.102, -17.774, -17.373,
    -16.873, -16.287, -15.597, -14.940, -18.860, -18.629, -18.382, -18.100,
    -17.774, -17.374, -16.875, -16.290, -15.601, -14.944, -18.856, -18.625,
    -18.378, -18.098, -17.773, -17.375, -16.878, -16.293, -15.604, -14.948,
    -18.852, -18.621, -18.375, -18.096, -17.772, -17.376, -16.880, -16.296,
    -15.608, -14.952, -18.848, -18.617, -18.371, -18.093, -17.772, -17.377,
    -16.882, -16.299, -15.612, -14.956, -18.843, -18.613, -18.367, -18.091,
    -17.771, -17.378, -16.884, -16.302, -15.615, -14.960, -18.838, -18.608,
    -18.363, -18.088, -17.770, -17.379, -16.886, -16.305, -15.619, -14.963,
    -18.833, -18.603, -18.359, -18.086, -17.769, -17.380, -16.888, -16.308,
    -15.623, -14.967, -18.827, -18.598, -18.355, -18.083, -17.769, -17.381,
    -16.891, -16.311, -15.626, -14.971, -18.821, -18.593, -18.350, -18.080,
    -17.768, -17.381, -16.893, -16.314, -15.630, -14.975, -18.815, -18.587,
    -18.345, -18.077, -17.766, -17.382, -16.895, -16.316, -15.633, -14.979,
    -18.808, -18.582, -18.340, -18.074, -17.765, -17.383, -16.897, -16.319,
    -15.637, -14.982, -18.802, -18.576, -18.335, -18.071, -17.764, -17.383,
    -16.898, -16.322, -15.640, -14.986, -18.794, -18.570, -18.330, -18.068,
    -17.763, -17.384, -16.900, -16.325, -15.644, -14.990, -18.787, -18.563,
    -18.324, -18.065, -17.762, -17.384, -16.902, -16.327, -15.647, -14.994,
    -18.779, -18.557, -18.319, -18.061, -17.760, -17.385, -16.904, -16.330,
    -15.651, -14.997, -18.771, -18.550, -18.313, -18.058, -17.759, -17.385,
    -16.906, -16.333, -15.654, -15.001, -18.762, -18.544, -18.307, -18.054,
    -17.757, -17.385, -16.907, -16.335, -15.657, -15.005, -18.754, -18.537,
    -18.301, -18.050, -17.755, -17.386, -16.909, -16.338, -15.661, -15.009,
    -18.744, -18.529, -18.295, -18.046, -17.754, -17.386, -16.911, -16.340,
    -15.664, -15.012, -18.735, -18.522, -18.288, -18.042, -17.752, -17.386,
    -16.912, -16.343, -15.668, -15.016, -18.725, -18.514, -18.281, -18.038,
    -17.750, -17.386, -16.914, -16.345, -15.671, -15.020, -18.715, -18.507,
    -18.275, -18.034, -17.748, -17.386, -16.916, -16.348, -15.674, -15.023,
    -18.705, -18.499, -18.268, -18.030, -17.746, -17.386, -16.917, -16.350,
    -15.677, -15.027, -18.694, -18.490, -18.260, -18.025, -17.744, -17.386,
    -16.919, -16.353, -15.681, -15.030, -18.683, -18.482, -18.253, -18.021,
    -17.742, -17.386, -16.920, -16.355, -15.684, -15.034, -18.673, -18.473,
    -18.246, -18.017, -17.740, -17.386, -16.922, -16.357, -15.687, -15.038,
    -18.662, -18.464, -18.239, -18.012, -17.738, -17.386, -16.923, -16.360,
    -15.691, -15.041, -18.651, -18.455, -18.232, -18.008, -17.736, -17.386,
    -16.925, -16.362, -15.694, -15.045, -18.640, -18.446, -18.224, -18.003,
    -17.734, -17.386, -16.926, -16.365, -15.697, -15.049, -18.628, -18.437,
    -18.217, -17.999, -17.732, -17.386, -16.928, -16.367, -15.701, -15.053,
    -18.617, -18.427, -18.209, -17.994, -17.729, -17.386, -16.929, -16.369,
    -15.704, -15.056, -18.605, -18.417, -18.201, -17.990, -17.727, -17.385,
    -16.931, -16.372, -15.707, -15.060, -18.592, -18.407, -18.193, -17.985,
    -17.725, -17.385, -16.932, -16.374, -15.711, -15.064, -18.580, -18.397,
    -18.185, -17.980, -17.722, -17.385, -16.934, -16.376, -15.714, -15.067,
    -18.567, -18.386, -18.177, -17.975, -17.720, -17.384, -16.935, -16.379,
    -15.717, -15.071, -18.554, -18.376, -18.168, -17.970, -17.717, -17.384,
    -16.937, -16.381, -15.721, -15.075, -18.541, -18.365, -18.160, -17.965,
    -17.715, -17.384, -16.938, -16.383, -15.724, -15.078, -18.528, -18.354,
    -18.151, -17.960, -17.712, -17.383, -16.939, -16.385, -15.727, -15.082,
    -18.514, -18.342, -18.142, -17.954, -17.710, -17.383, -16.941, -16.387,
    -15.731, -15.086, -18.500, -18.331, -18.133, -17.949, -17.707, -17.382,
    -16.942, -16.390, -15.734, -15.089, -18.486, -18.319, -18.124, -17.943,
    -17.704, -17.382, -16.943, -16.392, -15.737, -15.093, -18.471, -18.307,
    -18.115, -17.938, -17.701, -17.381, -16.945, -16.394, -15.740, -15.097,
    -18.457, -18.295, -18.106, -17.932, -17.699, -17.380, -16.946, -16.396,
    -15.744, -15.100, -18.442, -18.283, -18.097, -17.927, -17.696, -17.380,
    -16.947, -16.398, -15.747, -15.104, -18.427, -18.270, -18.087, -17.921,
    -17.693, -17.379, -16.948, -16.400, -15.750, -15.108, -18.411, -18.258,
    -18.077, -17.915, -17.690, -17.378, -16.949, -16.402, -15.753, -15.111,
    -18.396, -18.245, -18.067, -17.909, -17.687, -17.378, -16.951, -16.404,
    -15.756, -15.115, -18.380, -18.232, -18.057, -17.903, -17.683, -17.377,
    -16.952, -16.406, -15.760, -15.119, -18.363, -18.219, -18.047, -17.897,
    -17.680, -17.376, -16.953, -16.408, -15.763, -15.122, -18.347, -18.205,
    -18.037, -17.891, -17.677, -17.375, -16.954, -16.410, -15.766, -15.126,
    -18.331, -18.193, -18.029, -17.886, -17.674, -17.374, -16.955, -16.412,
    -15.769, -15.130, -18.314, -18.180, -18.020, -17.880, -17.671, -17.374,
    -16.956, -16.415, -15.772, -15.133, -18.297, -18.168, -18.012, -17.875,
    -17.669, -17.373, -16.957, -16.417, -15.776, -15.137, -18.280, -18.155,
    -18.003, -17.869, -17.666, -17.373, -16.958, -16.419, -15.779, -15.141,
    -18.263, -18.142, -17.995, -17.863, -17.663, -17.372, -16.959, -16.421,
    -15.782, -15.145, -18.246, -18.129, -17.986, -17.858, -17.660, -17.372,
    -16.961, -16.424, -15.785, -15.148, -18.228, -18.117, -17.978, -17.852,
    -17.657, -17.371, -16.962, -16.426, -15.789, -15.152, -18.210, -18.103,
    -17.970, -17.847, -17.654, -17.371, -16.963, -16.428, -15.792, -15.156,
    -18.192, -18.090, -17.961, -17.841, -17.651, -17.370, -16.964, -16.431,
    -15.795, -15.160, -18.174, -18.077, -17.953, -17.835, -17.648, -17.369,
    -16.965, -16.433, -15.798, -15.163, -18.155, -18.064, -17.945, -17.830,
    -17.645, -17.369, -16.965, -16.435, -15.801, -15.167, -18.136, -18.050,
    -17.937, -17.824, -17.642, -17.368, -16.966, -16.438, -15.805, -15.171,
    -18.117, -18.037, -17.928, -17.818, -17.639, -17.368, -16.967, -16.440,
    -15.808, -15.175, -18.098, -18.023, -17.920, -17.813, -17.636, -17.367,
    -16.968, -16.443, -15.811, -15.178, -18.079, -18.010, -17.912, -17.807,
    -17.633, -17.366, -16.969, -16.445, -15.814, -15.182, -18.059, -17.996,
    -17.903, -17.801, -17.630, -17.366, -16.970, -16.447, -15.817, -15.186,
    -18.039, -17.982, -17.895, -17.795, -17.627, -17.365, -16.971, -16.450,
    -15.821, -15.190, -18.019, -17.968, -17.887, -17.789, -17.624, -17.365,
    -16.972, -16.452, -15.824, -15.193, -17.999, -17.954, -17.879, -17.784,
    -17.621, -17.364, -16.972, -16.454, -15.827, -15.197, -17.978, -17.939,
    -17.871, -17.778, -17.618, -17.363, -16.973, -16.457, -15.830, -15.201,
    -17.957, -17.925, -17.863, -17.772, -17.615, -17.363, -16.974, -16.459,
    -15.833, -15.205, -17.936, -17.911, -17.854, -17.766, -17.612, -17.362,
    -16.975, -16.462, -15.837, -15.209, -17.915, -17.896, -17.846, -17.760,
    -17.608, -17.361, -16.976, -16.464, -15.840, -15.212, -17.894, -17.882,
    -17.838, -17.754, -17.605, -17.361, -16.976, -16.467, -15.843, -15.216,
    -17.872, -17.867, -17.830, -17.748, -17.602, -17.360, -16.977, -16.469,
    -15.846, -15.220, -17.861, -17.859, -17.825, -17.744, -17.600, -17.360,
    -16.978, -16.472, -15.849, -15.224, -17.851, -17.851, -17.820, -17.741,
    -17.599, -17.360, -16.980, -16.474, -15.853, -15.227, -17.841, -17.844,
    -17.815, -17.738, -17.597, -17.360, -16.981, -16.477, -15.856, -15.231,
    -17.832, -17.838, -17.811, -17.735, -17.596, -17.361, -16.983, -16.479,
    -15.859, -15.235, -17.823, -17.831, -17.807, -17.732, -17.595, -17.361,
    -16.984, -16.482, -15.863, -15.239, -17.816, -17.825, -17.803, -17.729,
    -17.593, -17.361, -16.986, -16.485, -15.866, -15.242, -17.808, -17.820,
    -17.799, -17.727, -17.592, -17.362, -16.988, -16.487, -15.869, -15.246,
    -17.802, -17.815, -17.795, -17.724, -17.591, -17.362, -16.989, -16.490,
    -15.873, -15.250, -17.796, -17.810, -17.792, -17.722, -17.590, -17.363,
    -16.991, -16.492, -15.876, -15.254, -17.790, -17.806, -17.789, -17.720,
    -17.590, -17.363, -16.993, -16.495, -15.879, -15.257, -17.786, -17.802,
    -17.786, -17.718, -17.589, -17.364, -16.994, -16.498, -15.883, -15.261,
    -17.781, -17.799, -17.784, -17.717, -17.588, -17.365, -16.996, -16.501,
    -15.886, -15.265, -17.778, -17.796, -17.782, -17.715, -17.588, -17.365,
    -16.998, -16.503, -15.890, -15.268, -17.775, -17.793, -17.780, -17.714,
    -17.588, -17.366, -17.000, -16.506, -15.893, -15.272, -17.773, -17.791,
    -17.778, -17.713, -17.587, -17.367, -17.002, -16.509, -15.896, -15.276,
    -17.771, -17.790, -17.776, -17.712, -17.587, -17.368, -17.004, -16.511,
    -15.900, -15.280, -17.770, -17.789, -17.775, -17.711, -17.587, -17.369,
    -17.006, -16.514, -15.903, -15.283, -17.770, -17.788, -17.774, -17.710,
    -17.587, -17.370, -17.008, -16.517, -15.907, -15.287, -17.770, -17.787,
    -17.773, -17.710, -17.587, -17.371, -17.010, -16.520, -15.910, -15.291,
    -17.771, -17.787, -17.772, -17.709, -17.587, -17.372, -17.012, -16.523,
    -15.914, -15.294, -17.772, -17.788, -17.772, -17.709, -17.588, -17.373,
    -17.014, -16.526, -15.917, -15.298, -17.774, -17.789, -17.772, -17.709,
    -17.588, -17.374, -17.016, -16.528, -15.921, -15.302, -17.777, -17.790,
    -17.772, -17.709, -17.589, -17.375, -17.018, -16.531, -15.924, -15.306,
    -17.780, -17.792, -17.772, -17.710, -17.589, -17.377, -17.021, -16.534,
    -15.928, -15.309, -17.784, -17.794, -17.773, -17.710, -17.590, -17.378,
    -17.023, -16.537, -15.931, -15.313, -17.781, -17.792, -17.771, -17.709,
    -17.590, -17.379, -17.025, -16.540, -15.935, -15.317, -17.779, -17.790,
    -17.770, -17.708, -17.590, -17.380, -17.027, -16.543, -15.938, -15.321,
    -17.777, -17.788, -17.769, -17.707, -17.590, -17.381, -17.029, -16.546,
    -15.942, -15.324, -17.774, -17.786, -17.767, -17.707, -17.590, -17.382,
    -17.032, -16.549, -15.945, -15.328, -17.772, -17.784, -17.766, -17.706,
    -17.590, -17.383, -17.034, -16.552, -15.949, -15.332, -17.770, -17.783,
    -17.765, -17.705, -17.590, -17.384, -17.036, -16.555, -15.953, -15.336,
    -17.768, -17.781, -17.763, -17.705, -17.590, -17.386, -17.038, -16.558,
    -15.956, -15.340, -17.766, -17.780, -17.762, -17.704, -17.590, -17.387,
    -17.041, -16.562, -15.960, -15.343, -17.764, -17.778, -17.761, -17.703,
    -17.591, -17.388, -17.043, -16.565, -15.964, -15.347, -17.763, -17.777,
    -17.760, -17.703, -17.591, -17.389, -17.045, -16.568, -15.967, -15.351,
    -17.761, -17.775, -17.759, -17.702, -17.591, -17.390, -17.048, -16.571,
    -15.971, -15.355, -17.759, -17.774, -17.758, -17.702, -17.591, -17.392,
    -17.050, -16.574, -15.975, -15.359, -17.758, -17.773, -17.758, -17.702,
    -17.592, -17.393, -17.052, -16.577, -15.978, -15.363, -17.756, -17.771,
    -17.757, -17.701, -17.592, -17.394, -17.055, -16.581, -15.982, -15.366,
    -17.755, -17.770, -17.756, -17.701, -17.592, -17.396, -17.057, -16.584,
    -15.986, -15.370, -17.753, -17.769, -17.755, -17.701, -17.593, -17.397,
    -17.060, -16.587, -15.990, -15.374, -17.752, -17.768, -17.755, -17.701,
    -17.593, -17.398, -17.062, -16.591, -15.993, -15.378, -17.751, -17.767,
    -17.754, -17.701, -17.594, -17.400, -17.065, -16.594, -15.997, -15.382,
    -17.750, -17.767, -17.754, -17.701, -17.594, -17.401, -17.067, -16.597,
    -16.001, -15.386, -17.749, -17.766, -17.753, -17.700, -17.595, -17.402,
    -17.070, -16.601, -16.005, -15.390, -17.748, -17.765, -17.753, -17.700,
    -17.595, -17.404, -17.072, -16.604, -16.009, -15.393, -17.747, -17.765,
    -17.753, -17.701, -17.596, -17.405, -17.075, -16.608, -16.012, -15.397,
    -17.746, -17.764, -17.752, -17.701, -17.597, -17.407, -17.078, -16.611,
    -16.016, -15.401, -17.746, -17.763, -17.752, -17.701, -17.597, -17.408,
    -17.080, -16.615, -16.020, -15.405, -17.745, -17.763, -17.752, -17.701,
    -17.598, -17.410, -17.083, -16.618, -16.024, -15.409, -17.745, -17.763,
    -17.752, -17.701, -17.599, -17.412, -17.086, -16.621, -16.028, -15.413,
    -17.745, -17.763, -17.753, -17.702, -17.600, -17.413, -17.088, -16.625,
    -16.032, -15.417, -17.746, -17.764, -17.753, -17.703, -17.601, -17.415,
    -17.091, -16.628, -16.036, -15.421, -17.746, -17.764, -17.754, -17.703,
    -17.602, -17.417, -17.094, -16.632, -16.039, -15.425, -17.747, -17.765,
    -17.754, -17.704, -17.603, -17.419, -17.096, -16.635, -16.043, -15.429,
    -17.748, -17.766, -17.755, -17.705, -17.604, -17.420, -17.099, -16.639,
    -16.047, -15.432, -17.749, -17.766, -17.756, -17.706, -17.605, -17.422,
    -17.102, -16.642, -16.051, -15.436, -17.750, -17.767, -17.756, -17.707,
    -17.606, -17.424, -17.105, -16.646, -16.055, -15.440, -17.751, -17.768,
    -17.757, -17.708, -17.607, -17.426, -17.108, -16.649, -16.059, -15.444,
    -17.752, -17.769, -17.758, -17.709, -17.609, -17.428, -17.110, -16.653,
    -16.063, -15.448, -17.754, -17.771, -17.760, -17.710, -17.610, -17.430,
    -17.113, -16.656, -16.067, -15.452, -17.755, -17.772, -17.761, -17.711,
    -17.611, -17.432, -17.116, -16.660, -16.071, -15.456, -17.757, -17.773,
    -17.762, -17.712, -17.613, -17.434, -17.119, -16.664, -16.075, -15.460,
    -17.759, -17.775, -17.763, -17.713, -17.614, -17.436, -17.122, -16.667,
    -16.079, -15.464, -17.761, -17.777, -17.765, -17.715, -17.616, -17.438,
    -17.125, -16.671, -16.083, -15.468, -17.763, -17.778, -17.766, -17.716,
    -17.617, -17.440, -17.128, -16.675, -16.087, -15.472, -17.765, -17.780,
    -17.768, -17.718, -17.619, -17.442, -17.131, -16.678, -16.091, -15.476,
    -17.768, -17.782, -17.770, -17.719, -17.620, -17.444, -17.134, -16.682,
    -16.095, -15.480, -17.770, -17.784, -17.771, -17.721, -17.622, -17.446,
    -17.137, -16.686, -16.099, -15.484, -17.773, -17.786, -17.773, -17.723,
    -17.624, -17.449, -17.140, -16.689, -16.103, -15.488, -17.776, -17.789,
    -17.775, -17.724, -17.626, -17.451, -17.143, -16.693, -16.107, -15.492,
    -17.779, -17.791, -17.777, -17.726, -17.627, -17.453, -17.146, -16.697,
    -16.111, -15.496, -17.782, -17.794, -17.779, -17.728, -17.629, -17.455,
    -17.149, -16.700, -16.115, -15.500, -17.785, -17.796, -17.782, -17.730,
    -17.631, -17.458, -17.152, -16.704, -16.119, -15.504, -17.788, -17.799,
    -17.784, -17.732, -17.633, -17.460, -17.155, -16.708, -16.123, -15.508,
    -17.791, -17.802, -17.786, -17.734, -17.635, -17.462, -17.158, -16.712,
    -16.127, -15.512, -17.794, -17.804, -17.789, -17.736, -17.637, -17.465,
    -17.161, -16.715, -16.131, -15.516, -17.798, -17.807, -17.791, -17.738,
    -17.639, -17.467, -17.164, -16.719, -16.135, -15.520, -17.801, -17.810,
    -17.794, -17.740, -17.641, -17.470, -17.167, -16.723, -16.139, -15.524,
    -17.805, -17.813, -17.796, -17.743, -17.643, -17.472, -17.170, -16.726,
    -16.143, -15.528, -17.808, -17.816, -17.799, -17.745, -17.646, -17.475,
    -17.173, -16.730, -16.147, -15.532, -17.812, -17.819, -17.802, -17.747,
    -17.648, -17.477, -17.177, -16.734, -16.151, -15.536, -17.816, -17.823,
    -17.804, -17.750, -17.650, -17.480, -17.180, -16.738, -16.155, -15.540,
    -17.820, -17.826, -17.807, -17.752, -17.653, -17.482, -17.183, -16.741,
    -16.159, -15.544, -17.824, -17.829, -17.810, -17.755, -17.655, -17.485,
    -17.186, -16.745, -16.163, -15.548, -17.828, -17.833, -17.813, -17.758,
    -17.658, -17.487, -17.189, -16.749, -16.167, -15.552, -17.832, -17.837,
    -17.816, -17.760, -17.660, -17.490, -17.192, -16.753, -16.171, -15.556,
    -17.836, -17.840, -17.820, -17.763, -17.663, -17.493, -17.196, -16.756,
    -16.175, -15.560, -17.841, -17.844, -17.823, -17.766, -17.665, -17.495,
    -17.199, -16.760, -16.179, -15.564, -17.845, -17.848, -17.826, -17.769,
    -17.668, -17.498, -17.202, -16.764, -16.183, -15.568, -17.850, -17.852,
    -17.830, -17.772, -17.671, -17.501, -17.205, -16.768, -16.187, -15.572,
    -17.855, -17.856, -17.833, -17.775, -17.673, -17.504, -17.209, -16.771,
    -16.191, -15.576, -17.860, -17.860, -17.837, -17.778, -17.676, -17.507,
    -17.212, -16.775, -16.196, -15.580, -17.865, -17.864, -17.840, -17.781,
    -17.679, -17.509, -17.215, -16.779, -16.200, -15.584, -17.870, -17.869,
    -17.844, -17.784, -17.682, -17.512, -17.218, -16.783, -16.204, -15.588,
    -17.875, -17.873, -17.848, -17.787, -17.685, -17.515, -17.222, -16.787,
    -16.208, -15.592, -17.880, -17.878, -17.852, -17.791, -17.688, -17.518,
    -17.225, -16.791, -16.212, -15.596, -17.886, -17.882, -17.856, -17.794,
    -17.691, -17.521, -17.228, -16.794, -16.216, -15.600, -17.891, -17.887,
    -17.860, -17.798, -17.694, -17.524, -17.232, -16.798, -16.220, -15.604,
    -17.897, -17.892, -17.864, -17.801, -17.697, -17.527, -17.235, -16.802,
    -16.224, -15.608, -17.902, -17.896, -17.868, -17.804, -17.700, -17.530,
    -17.238, -16.806, -16.228, -15.612, -17.907, -17.901, -17.872, -17.808,
    -17.703, -17.533, -17.242, -16.810, -16.232, -15.616, -17.912, -17.905,
    -17.876, -17.811, -17.706, -17.536, -17.245, -16.813, -16.236, -15.620,
    -17.918, -17.910, -17.880, -17.815, -17.709, -17.539, -17.248, -16.817,
    -16.240, -15.624, -17.923, -17.915, -17.884, -17.818, -17.712, -17.542,
    -17.252, -16.821, -16.244, -15.628, -17.928, -17.919, -17.888, -17.822,
    -17.716, -17.545, -17.255, -16.825, -16.248, -15.632, -17.933, -17.924,
    -17.892, -17.825, -17.719, -17.548, -17.259, -16.829, -16.252, -15.636,
    -17.939, -17.929, -17.896, -17.829, -17.722, -17.552, -17.262, -16.833,
    -16.256, -15.640, -17.944, -17.934, -17.900, -17.832, -17.725, -17.555,
    -17.266, -16.836, -16.260, -15.644, -17.950, -17.939, -17.904, -17.836,
    -17.729, -17.558, -17.269, -16.840, -16.264, -15.647, -17.955, -17.944,
    -17.909, -17.840, -17.732, -17.561, -17.272, -16.844, -16.268, -15.651,
    -17.961, -17.949, -17.913, -17.844, -17.736, -17.564, -17.276, -16.848,
    -16.272, -15.655, -17.967, -17.954, -17.917, -17.848, -17.739, -17.568,
    -17.279, -16.852, -16.276, -15.659, -17.972, -17.959, -17.922, -17.851,
    -17.742, -17.571, -17.283, -16.856, -16.280, -15.663, -17.978, -17.964,
    -17.926, -17.855, -17.746, -17.574, -17.286, -16.859, -16.284, -15.667,
    -17.984, -17.969, -17.931, -17.859, -17.750, -17.578, -17.290, -16.863,
    -16.288, -15.671, -17.990, -17.975, -17.936, -17.863, -17.753, -17.581,
    -17.293, -16.867, -16.292, -15.675, -17.996, -17.980, -17.940, -17.867,
    -17.757, -17.585, -17.297, -16.871, -16.295, -15.679, -18.002, -17.985,
    -17.945, -17.871, -17.760, -17.588, -17.300, -16.875, -16.299, -15.683,
    -18.008, -17.991, -17.950, -17.876, -17.764, -17.591, -17.304, -16.879,
    -16.303, -15.687, -18.014, -17.996, -17.954, -17.880, -17.768, -17.595,
    -17.308, -16.883, -16.307, -15.690, -18.020, -18.002, -17.959, -17.884,
    -17.772, -17.598, -17.311, -16.886, -16.311, -15.694, -18.026, -18.008,
    -17.964, -17.888, -17.775, -17.602, -17.315, -16.890, -16.315, -15.698,
    -18.033, -18.013, -17.969, -17.893, -17.779, -17.605, -17.318, -16.894,
    -16.319, -15.702, -18.039, -18.019, -17.974, -17.897, -17.783, -17.609,
    -17.322, -16.898, -16.323, -15.706, -18.045, -18.024, -17.979, -17.901,
    -17.787, -17.612, -17.326, -16.902, -16.327, -15.710, -18.051, -18.030,
    -17.984, -17.905, -17.791, -17.616, -17.329, -16.906, -16.331, -15.714,
    -18.057, -18.035, -17.988, -17.910, -17.794, -17.620, -17.333, -16.909,
    -16.335, -15.717, -18.063, -18.041, -17.993, -17.914, -17.798, -17.623,
    -17.336, -16.913, -16.339, -15.721, -18.070, -18.046, -17.998, -17.918,
    -17.802, -17.627, -17.340, -16.917, -16.343, -15.725, -18.076, -18.052,
    -18.003, -17.923, -17.806, -17.630, -17.344, -16.921, -16.346, -15.729,
    -18.082, -18.058, -18.008, -17.927, -17.810, -17.634, -17.347, -16.925,
    -16.350, -15.733, -18.088, -18.063, -18.013, -17.932, -17.814, -17.637,
    -17.351, -16.929, -16.354, -15.737, -18.095, -18.069, -18.018, -17.936,
    -17.818, -17.641, -17.355, -16.932, -16.358, -15.740, -18.101, -18.075,
    -18.023, -17.940, -17.822, -17.645, -17.358, -16.936, -16.362, -15.744,
    -18.107, -18.080, -18.028, -17.945, -17.826, -17.648, -17.362, -16.940,
    -16.366, -15.748, -18.114, -18.086, -18.034, -17.949, -17.830, -17.652,
    -17.366, -16.944, -16.370, -15.752, -18.120, -18.092, -18.039, -17.954,
    -17.834, -17.656, -17.369, -16.948, -16.374, -15.755, -18.126, -18.098,
    -18.044, -17.959, -17.838, -17.659, -17.373, -16.951, -16.378, -15.759,
    -18.133, -18.103, -18.049, -17.963, -17.842, -17.663, -17.377, -16.955,
    -16.381, -15.763, -18.139, -18.109, -18.054, -17.968, -17.846, -17.667,
    -17.380, -16.959, -16.385, -15.767, -18.146, -18.115, -18.060, -17.973,
    -17.850, -17.670, -17.384, -16.963, -16.389, -15.770, -18.152, -18.121,
    -18.065, -17.977, -17.854, -17.674, -17.388, -16.966, -16.393, -15.774,
    -18.159, -18.127, -18.070, -17.982, -17.858, -17.678, -17.391, -16.970,
    -16.397, -15.778, -18.166, -18.133, -18.076, -17.987, -17.863, -17.682,
    -17.395, -16.974, -16.401, -15.782, -18.172, -18.139, -18.081, -17.992,
    -17.867, -17.686, -17.399, -16.978, -16.405, -15.785, -18.179, -18.145,
    -18.087, -17.996, -17.871, -17.689, -17.403, -16.982, -16.408, -15.789,
    -18.186, -18.151, -18.092, -18.001, -17.875, -17.693, -17.406, -16.985,
    -16.412, -15.793, -18.192, -18.157, -18.097, -18.006, -17.880, -17.697,
    -17.410, -16.989, -16.416, -15.796, -18.199, -18.163, -18.103, -18.011,
    -17.884, -17.701, -17.414, -16.993, -16.420, -15.800, -18.206, -18.169,
    -18.108, -18.016, -17.888, -17.705, -17.418, -16.997, -16.424, -15.804,
    -18.212, -18.175, -18.114, -18.021, -17.893, -17.709, -17.421, -17.001,
    -16.428, -15.807, -18.219, -18.181, -18.119, -18.025, -17.897, -17.713,
    -17.425, -17.004, -16.431, -15.811, -18.225, -18.187, -18.125, -18.030,
    -17.901, -17.717, -17.429, -17.008, -16.435, -15.815, -18.232, -18.193,
    -18.130, -18.035, -17.906, -17.720, -17.433, -17.012, -16.439, -15.819,
    -18.239, -18.199, -18.136, -18.040, -17.910, -17.724, -17.437, -17.016,
    -16.443, -15.822, -18.245, -18.205, -18.141, -18.045, -17.914, -17.728,
    -17.440, -17.019, -16.446, -15.826, -18.252, -18.212, -18.147, -18.050,
    -17.919, -17.732, -17.444, -17.023, -16.450, -15.830, -18.259, -18.218,
    -18.152, -18.055, -17.923, -17.736, -17.448, -17.027, -16.454, -15.833,
    -18.266, -18.224, -18.158, -18.060, -17.927, -17.740, -17.452, -17.031,
    -16.457, -15.837, -18.272, -18.230, -18.164, -18.065, -17.932, -17.744,
    -17.455, -17.034, -16.461, -15.841, -18.279, -18.236, -18.169, -18.070,
    -17.936, -17.748, -17.459, -17.038, -16.465, -15.844, -18.286, -18.243,
    -18.175, -18.075, -17.941, -17.752, -17.463, -17.042, -16.468, -15.848,
    -18.293, -18.249, -18.180, -18.080, -17.945, -17.756, -17.467, -17.046,
    -16.472, -15.852, -18.299, -18.255, -18.186, -18.085, -17.950, -17.761,
    -17.471, -17.050, -16.476, -15.855, -18.306, -18.261, -18.192, -18.090,
    -17.954, -17.765, -17.474, -17.053, -16.479, -15.859, -18.313, -18.268,
    -18.198, -18.095, -17.959, -17.769, -17.478, -17.057, -16.483, -15.863,
    -18.320, -18.274, -18.203, -18.101, -17.964, -17.773, -17.482, -17.061,
    -16.487, -15.866, -18.327, -18.280, -18.209, -18.106, -17.968, -17.777,
    -17.486, -17.065, -16.490, -15.870, -18.333, -18.287, -18.215, -18.111,
    -17.973, -17.781, -17.490, -17.068, -16.494, -15.874, -18.340, -18.293,
    -18.221, -18.116, -17.977, -17.785, -17.494, -17.072, -16.498, -15.877,
    -18.347, -18.300, -18.226, -18.121, -17.982, -17.789, -17.497, -17.076,
    -16.501, -15.881, -18.354, -18.306, -18.232, -18.126, -17.987, -17.794,
    -17.501, -17.080, -16.505, -15.885, -18.361, -18.313, -18.238, -18.132,
    -17.991, -17.798, -17.505, -17.083, -16.508, -15.888, -18.368, -18.319,
    -18.244, -18.137, -17.996, -17.802, -17.509, -17.087, -16.512, -15.892,
    -18.375, -18.325, -18.250, -18.142, -18.001, -17.806, -17.513, -17.091,
    -16.516, -15.896, -18.382, -18.332, -18.256, -18.147, -18.005, -17.810,
    -17.517, -17.094, -16.519, -15.899, -18.388, -18.338, -18.261, -18.153,
    -18.010, -17.814, -17.521, -17.098, -16.523, -15.903, -18.395, -18.344,
    -18.267, -18.158, -18.015, -17.819, -17.524, -17.102, -16.527, -15.907,
    -18.402, -18.351, -18.273, -18.163, -18.019, -17.823, -17.528, -17.106,
    -16.530, -15.910, -18.409, -18.357, -18.279, -18.168, -18.024, -17.827,
    -17.532, -17.109, -16.534, -15.914, -18.416, -18.364, -18.285, -18.174,
    -18.029, -17.831, -17.536, -17.113, -16.537, -15.917, -18.423, -18.370,
    -18.290, -18.179, -18.033, -17.835, -17.540, -17.117, -16.541, -15.921,
    -18.430, -18.376, -18.296, -18.184, -18.038, -17.840, -17.544, -17.120,
    -16.545, -15.925, -18.436, -18.383, -18.302, -18.190, -18.043, -17.844,
    -17.548, -17.124, -16.548, -15.928, -18.443, -18.389, -18.308, -18.195,
    -18.047, -17.848, -17.551, -17.128, -16.552, -15.932, -18.450, -18.396,
    -18.314, -18.200, -18.052, -17.852, -17.555, -17.131, -16.555, -15.935,
    -18.457, -18.402, -18.320, -18.206, -18.057, -17.856, -17.559, -17.135,
    -16.559, -15.939, -18.464, -18.409, -18.326, -18.211, -18.062, -17.861,
    -17.563, -17.139, -16.563, -15.943, -18.471, -18.415, -18.331, -18.216,
    -18.067, -17.865, -17.567, -17.142, -16.566, -15.946, -18.478, -18.421,
    -18.337, -18.222, -18.071, -17.869, -17.571, -17.146, -16.570, -15.950,
    -18.485, -18.428, -18.343, -18.227, -18.076, -17.874, -17.575, -17.150,
    -16.573, -15.953, -18.492, -18.434, -18.349, -18.233, -18.081, -17.878,
    -17.579, -17.153, -16.577, -15.957, -18.498, -18.441, -18.355, -18.238,
    -18.086, -17.882, -17.583, -17.157, -16.581, -15.961, -18.505, -18.447,
    -18.361, -18.244, -18.091, -17.886, -17.586, -17.161, -16.584, -15.964,
    -18.512, -18.454, -18.367, -18.249, -18.095, -17.891, -17.590, -17.164,
    -16.588, -15.968, -18.519, -18.460, -18.373, -18.254, -18.100, -17.895,
    -17.594, -17.168, -16.591, -15.971, -18.526, -18.467, -18.379, -18.260,
    -18.105, -17.899, -17.598, -17.172, -16.595, -15.975, -18.533, -18.473,
    -18.385, -18.265, -18.110, -17.904, -17.602, -17.175, -16.598, -15.978,
    -18.540, -18.480, -18.391, -18.271, -18.115, -17.908, -17.606, -17.179,
    -16.602, -15.982, -18.547, -18.486, -18.397, -18.276, -18.120, -17.912,
    -17.610, -17.183, -16.606, -15.986, -18.554, -18.493, -18.403, -18.282,
    -18.125, -17.917, -17.614, -17.186, -16.609, -15.989, -18.561, -18.499,
    -18.409, -18.287, -18.130, -17.921, -17.618, -17.190, -16.613, -15.993,
    -18.568, -18.506, -18.415, -18.293, -18.135, -17.925, -17.622, -17.194,
    -16.616, -15.996, -18.575, -18.512, -18.421, -18.298, -18.140, -17.930,
    -17.626, -17.198, -16.620, -16.000, -18.582, -18.519, -18.427, -18.304,
    -18.144, -17.934, -17.630, -17.201, -16.624, -16.003, -18.589, -18.525,
    -18.433, -18.309, -18.149, -17.939, -17.634, -17.205, -16.627, -16.007,
    -18.595, -18.532, -18.439, -18.315, -18.154, -17.943, -17.638, -17.209,
    -16.631, -16.011, -18.602, -18.538, -18.445, -18.320, -18.159, -17.947,
    -17.642, -17.213, -16.635, -16.014, -18.609, -18.545, -18.451, -18.326,
    -18.164, -17.952, -17.646, -17.216, -16.638, -16.018, -18.616, -18.551,
    -18.457, -18.331, -18.169, -17.956, -17.650, -17.220, -16.642, -16.021,
    -18.623, -18.558, -18.463, -18.337, -18.174, -17.961, -17.654, -17.224,
    -16.646, -16.025, -18.630, -18.565, -18.469, -18.342, -18.179, -17.965,
    -17.658, -17.228, -16.649, -16.028, -18.637, -18.571, -18.476, -18.348,
    -18.184, -17.970, -17.662, -17.231, -16.653, -16.032, -18.644, -18.578,
    -18.482, -18.353, -18.189, -17.974, -17.666, -17.235, -16.656, -16.036,
    -18.651, -18.584, -18.488, -18.359, -18.194, -17.978, -17.670, -17.239,
    -16.660, -16.039, -18.658, -18.591, -18.494, -18.364, -18.199, -17.983,
    -17.674, -17.243, -16.664, -16.043, -18.665, -18.597, -18.500, -18.370,
    -18.204, -17.987, -17.678, -17.246, -16.667, -16.046, -18.672, -18.604,
    -18.506, -18.376, -18.209, -17.992, -17.682, -17.250, -16.671, -16.050,
    -18.679, -18.610, -18.512, -18.381, -18.215, -17.996, -17.686, -17.254,
    -16.675, -16.053, -18.686, -18.617, -18.518, -18.387, -18.220, -18.001,
    -17.690, -17.258, -16.678, -16.057, -18.693, -18.623, -18.524, -18.392,
    -18.225, -18.005, -17.694, -17.262, -16.682, -16.060, -18.700, -18.630,
    -18.531, -18.398, -18.230, -18.010, -17.698, -17.265, -16.686, -16.064,
    -18.707, -18.636, -18.537, -18.403, -18.235, -18.014, -17.702, -17.269,
    -16.689, -16.067, -18.714, -18.643, -18.543, -18.409, -18.240, -18.019,
    -17.706, -17.273, -16.693, -16.071, -18.721, -18.650, -18.549, -18.415,
    -18.245, -18.024, -17.710, -17.277, -16.697, -16.075, -18.728, -18.656,
    -18.555, -18.420, -18.250, -18.028, -17.714, -17.281, -16.700, -16.078,
    -18.735, -18.663, -18.561, -18.426, -18.255, -18.033, -17.718, -17.284,
    -16.704, -16.082, -18.742, -18.669, -18.568, -18.431, -18.260, -18.037,
    -17.722, -17.288, -16.708, -16.085, -18.749, -18.676, -18.574, -18.437,
    -18.265, -18.042, -17.726, -17.292, -16.711, -16.089, -18.756, -18.682,
    -18.580, -18.443, -18.270, -18.046, -17.730, -17.296, -16.715, -16.092,
    -18.763, -18.689, -18.586, -18.448, -18.276, -18.051, -17.735, -17.299,
    -16.719, -16.096, -18.770, -18.696, -18.592, -18.454, -18.281, -18.055,
    -17.739, -17.303, -16.722, -16.099, -18.777, -18.702, -18.598, -18.460,
    -18.286, -18.060, -17.743, -17.307, -16.726, -16.103, -18.784, -18.709,
    -18.604, -18.465, -18.291, -18.064, -17.747, -17.311, -16.730, -16.107,
    -18.791, -18.715, -18.611, -18.471, -18.296, -18.069, -17.751, -17.315,
    -16.733, -16.110, -18.798, -18.722, -18.617, -18.477, -18.301, -18.073,
    -17.755, -17.318, -16.737, -16.114, -18.804, -18.729, -18.623, -18.482,
    -18.306, -18.078, -17.759, -17.322, -16.741, -16.117, -18.811, -18.735,
    -18.629, -18.488, -18.311, -18.082, -17.763, -17.326, -16.744, -16.121,
    -18.818, -18.742, -18.635, -18.494, -18.316, -18.087, -17.767, -17.330,
    -16.748, -16.124, -18.825, -18.748, -18.641, -18.499, -18.322, -18.092,
    -17.772, -17.334, -16.752, -16.128, -18.832, -18.755, -18.648, -18.505,
    -18.327, -18.096, -17.776, -17.337, -16.755, -16.132, -18.839, -18.762,
    -18.654, -18.511, -18.332, -18.101, -17.780, -17.341, -16.759, -16.135,
    -18.846, -18.768, -18.660, -18.517, -18.337, -18.105, -17.784, -17.345,
    -16.763, -16.139, -18.853, -18.775, -18.666, -18.522, -18.342, -18.110,
    -17.788, -17.349, -16.767, -16.142, -18.860, -18.781, -18.672, -18.528,
    -18.347, -18.115, -17.792, -17.353, -16.770, -16.146, -18.867, -18.788,
    -18.678, -18.534, -18.352, -18.119, -17.796, -17.357, -16.774, -16.149,
    -18.874, -18.795, -18.685, -18.539, -18.358, -18.124, -17.801, -17.360,
    -16.778, -16.153, -18.881, -18.801, -18.691, -18.545, -18.363, -18.128,
    -17.805, -17.364, -16.781, -16.156, -18.888, -18.808, -18.697, -18.551,
    -18.368, -18.133, -17.809, -17.368, -16.785, -16.160, -18.895, -18.815,
    -18.703, -18.557, -18.373, -18.138, -17.813, -17.372, -16.789, -16.164,
    -18.902, -18.821, -18.709, -18.562, -18.378, -18.142, -17.817, -17.376,
    -16.792, -16.167, -18.909, -18.828, -18.716, -18.568, -18.384, -18.147,
    -17.821, -17.380, -16.796, -16.171, -18.916, -18.834, -18.722, -18.574,
    -18.389, -18.152, -17.826, -17.383, -16.800, -16.174, -18.923, -18.841,
    -18.728, -18.580, -18.394, -18.156, -17.830, -17.387, -16.804, -16.178,
    -18.929, -18.848, -18.734, -18.585, -18.399, -18.161, -17.834, -17.391,
    -16.807, -16.182, -18.936, -18.854, -18.740, -18.591, -18.404, -18.165,
    -17.838, -17.395, -16.811, -16.185, -18.943, -18.861, -18.746, -18.597,
    -18.410, -18.170, -17.842, -17.399, -16.815, -16.189, -18.950, -18.867,
    -18.753, -18.602, -18.415, -18.175, -17.846, -17.403, -16.819, -16.193,
    -18.957, -18.874, -18.759, -18.608, -18.420, -18.179, -17.850, -17.407,
    -16.822, -16.196, -18.964, -18.881, -18.765, -18.614, -18.425, -18.184,
    -17.855, -17.410, -16.826, -16.200, -18.971, -18.887, -18.771, -18.620,
    -18.431, -18.189, -17.859, -17.414, -16.830, -16.203, -18.978, -18.894,
    -18.777, -18.625, -18.436, -18.193, -17.863, -17.418, -16.834, -16.207,
    -18.985, -18.900, -18.784, -18.631, -18.441, -18.198, -17.867, -17.422,
    -16.837, -16.211, -18.991, -18.907, -18.790, -18.637, -18.446, -18.203,
    -17.871, -17.426, -16.841, -16.214, -18.998, -18.914, -18.796, -18.642,
    -18.451, -18.208, -17.875, -17.430, -16.845, -16.218, -19.005, -18.920,
    -18.802, -18.648, -18.457, -18.212, -17.880, -17.434, -16.849, -16.222,
    -19.012, -18.927, -18.808, -18.654, -18.462, -18.217, -17.884, -17.438,
    -16.852, -16.225, -19.019, -18.933, -18.815, -18.660, -18.467, -18.222,
    -17.888, -17.442, -16.856, -16.229, -19.026, -18.940, -18.821, -18.665,
    -18.473, -18.226, -17.892, -17.445, -16.860, -16.233, -19.033, -18.947,
    -18.827, -18.671, -18.478, -18.231, -17.896, -17.449, -16.864, -16.236,
    -19.039, -18.953, -18.833, -18.677, -18.483, -18.236, -17.900, -17.453,
    -16.868, -16.240, -19.046, -18.960, -18.840, -18.683, -18.488, -18.241,
    -17.905, -17.457, -16.871, -16.244, -19.053, -18.966, -18.846, -18.688,
    -18.494, -18.245, -17.909, -17.461, -16.875, -16.247, -19.060, -18.973,
    -18.852, -18.694, -18.499, -18.250, -17.913, -17.465, -16.879, -16.251,
    -19.067, -18.980, -18.858, -18.700, -18.504, -18.255, -17.917, -17.469,
    -16.883, -16.255, -19.074, -18.986, -18.864, -18.706, -18.509, -18.259,
    -17.921, -17.473, -16.887, -16.258, -19.081, -18.993, -18.871, -18.711,
    -18.515, -18.264, -17.925, -17.477, -16.890, -16.262, -19.087, -18.999,
    -18.877, -18.717, -18.520, -18.269, -17.930, -17.481, -16.894, -16.266,
    -19.094, -19.006, -18.883, -18.723, -18.525, -18.273, -17.934, -17.484,
    -16.898, -16.269, -19.101, -19.012, -18.889, -18.729, -18.530, -18.278,
    -17.938, -17.488, -16.902, -16.273, -19.108, -19.019, -18.896, -18.734,
    -18.536, -18.283, -17.942, -17.492, -16.906, -16.277, -19.115, -19.026,
    -18.902, -18.740, -18.541, -18.287, -17.946, -17.496, -16.910, -16.280,
    -19.122, -19.032, -18.908, -18.746, -18.546, -18.292, -17.950, -17.500,
    -16.913, -16.284, -19.129, -19.039, -18.914, -18.752, -18.551, -18.297,
    -17.955, -17.504, -16.917, -16.288, -19.135, -19.045, -18.921, -18.758,
    -18.557, -18.301, -17.959, -17.508, -16.921, -16.291, -19.142, -19.052,
    -18.927, -18.763, -18.562, -18.306, -17.963, -17.512, -16.925, -16.295,
    -19.149, -19.058, -18.933, -18.769, -18.567, -18.310, -17.967, -17.515,
    -16.929, -16.299, -19.156, -19.065, -18.939, -18.775, -18.572, -18.315,
    -17.971, -17.519, -16.933, -16.302, -19.163, -19.072, -18.945, -18.781,
    -18.578, -18.320, -17.975, -17.523, -16.936, -16.306, -19.170, -19.078,
    -18.952, -18.787, -18.583, -18.324, -17.980, -17.527, -16.940, -16.310,
    -19.176, -19.085, -18.958, -18.792, -18.588, -18.329, -17.984, -17.531,
    -16.944, -16.313, -19.183, -19.091, -18.964, -18.798, -18.593, -18.334,
    -17.988, -17.535, -16.948, -16.317, -19.190, -19.098, -18.970, -18.804,
    -18.599, -18.338, -17.992, -17.539, -16.952, -16.321, -19.197, -19.104,
    -18.977, -18.810, -18.604, -18.343, -17.996, -17.543, -16.956, -16.325,
    -19.204, -19.111, -18.983, -18.816, -18.609, -18.348, -18.000, -17.546,
    -16.960, -16.328, -19.211, -19.117, -18.989, -18.821, -18.614, -18.352,
    -18.005, -17.550, -16.963, -16.332, -19.217, -19.124, -18.995, -18.827,
    -18.620, -18.357, -18.009, -17.554, -16.967, -16.336, -19.224, -19.130,
    -19.002, -18.833, -18.625, -18.361, -18.013, -17.558, -16.971, -16.339,
    -19.231, -19.137, -19.008, -18.839, -18.630, -18.366, -18.017, -17.562,
    -16.975, -16.343, -19.238, -19.144, -19.014, -18.845, -18.635, -18.371,
    -18.021, -17.566, -16.979, -16.347, -19.245, -19.150, -19.020, -18.851,
    -18.640, -18.375, -18.025, -17.570, -16.983, -16.350, -19.251, -19.157,
    -19.027, -18.856, -18.646, -18.380, -18.029, -17.574, -16.986, -16.354,
    -19.258, -19.163, -19.033, -18.862, -18.651, -18.384, -18.033, -17.577,
    -16.990, -16.358, -19.265, -19.170, -19.039, -18.868, -18.656, -18.389,
    -18.037, -17.581, -16.994, -16.361, -19.272, -19.176, -19.045, -18.874,
    -18.661, -18.394, -18.042, -17.585, -16.998, -16.365, -19.279, -19.183,
    -19.052, -18.879, -18.667, -18.398, -18.046, -17.589, -17.002, -16.368,
    -19.286, -19.189, -19.058, -18.885, -18.672, -18.403, -18.050, -17.593,
    -17.005, -16.372, -19.292, -19.196, -19.064, -18.891, -18.677, -18.408,
    -18.054, -17.597, -17.009, -16.376, -19.299, -19.202, -19.070, -18.897,
    -18.682, -18.412, -18.058, -17.601, -17.013, -16.379, -19.306, -19.209,
    -19.076, -18.902, -18.688, -18.417, -18.062, -17.604, -17.017, -16.383,
    -19.313, -19.215, -19.083, -18.908, -18.693, -18.421, -18.066, -17.608,
    -17.020, -16.386, -19.320, -19.222, -19.089, -18.914, -18.698, -18.426,
    -18.070, -17.612, -17.024, -16.390, -19.326, -19.228, -19.095, -18.920,
    -18.703, -18.431, -18.074, -17.616, -17.028, -16.394, -19.333, -19.235,
    -19.101, -18.926, -18.709, -18.435, -18.078, -17.620, -17.032, -16.397,
    -19.340, -19.241, -19.107, -18.931, -18.714, -18.440, -18.082, -17.624,
    -17.035, -16.401, -19.347, -19.248, -19.114, -18.937, -18.719, -18.444,
    -18.086, -17.627, -17.039, -16.404, -19.354, -19.254, -19.120, -18.943,
    -18.724, -18.449, -18.090, -17.631, -17.043, -16.408, -19.360, -19.261,
    -19.126, -18.949, -18.730, -18.453, -18.094, -17.635, -17.047, -16.412,
    -19.367, -19.267, -19.132, -18.954, -18.735, -18.458, -18.098, -17.639,
    -17.050, -16.415, -19.374, -19.274, -19.138, -18.960, -18.740, -18.463,
    -18.102, -17.643, -17.054, -16.419, -19.381, -19.280, -19.144, -18.966,
    -18.745, -18.467, -18.106, -17.647, -17.058, -16.422, -19.387, -19.287,
    -19.151, -18.972, -18.751, -18.472, -18.110, -17.650, -17.062, -16.426,
    -19.394, -19.293, -19.157, -18.977, -18.756, -18.476, -18.114, -17.654,
    -17.065, -16.429, -19.401, -19.300, -19.163, -18.983, -18.761, -18.481,
    -18.118, -17.658, -17.069, -16.433, -19.408, -19.306, -19.169, -18.989,
    -18.766, -18.486, -18.122, -17.662, -17.073, -16.437, -19.414, -19.313,
    -19.175, -18.995, -18.771, -18.490, -18.126, -17.665, -17.076, -16.440,
    -19.421, -19.319, -19.181, -19.000, -18.777, -18.495, -18.130, -17.669,
    -17.080, -16.444, -19.428, -19.326, -19.188, -19.006, -18.782, -18.499,
    -18.134, -17.673, -17.084, -16.447, -19.435, -19.332, -19.194, -19.012,
    -18.787, -18.504, -18.138, -17.676, -17.087, -16.451, -19.441, -19.339,
    -19.200, -19.018, -18.792, -18.508, -18.142, -17.680, -17.091, -16.454,
    -19.448, -19.345, -19.206, -19.023, -18.797, -18.513, -18.146, -17.684,
    -17.094, -16.457, -19.455, -19.352, -19.212, -19.029, -18.802, -18.517,
    -18.150, -17.687, -17.098, -16.461, -19.461, -19.358, -19.218, -19.035,
    -18.808, -18.522, -18.154, -17.691, -17.102, -16.464, -19.468, -19.365,
    -19.224, -19.041, -18.813, -18.526, -18.158, -17.695, -17.105, -16.468,
    -19.475, -19.371, -19.231, -19.046, -18.818, -18.531, -18.161, -17.698,
    -17.109, -16.471, -19.481, -19.378, -19.237, -19.052, -18.823, -18.535,
    -18.165, -17.702, -17.112, -16.475, -19.488, -19.384, -19.243, -19.058,
    -18.828, -18.540, -18.169, -17.706, -17.116, -16.478, -19.494, -19.391,
    -19.249, -19.064, -18.833, -18.544, -18.173, -17.709, -17.119, -16.482,
    -19.501, -19.397, -19.255, -19.069, -18.839, -18.549, -18.177, -17.713,
    -17.123, -16.485, -19.508, -19.404, -19.261, -19.075, -18.844, -18.553,
    -18.181, -17.716, -17.127, -16.489, -19.514, -19.410, -19.267, -19.081,
    -18.849, -18.557, -18.185, -17.720, -17.130, -16.492, -19.521, -19.417,
    -19.273, -19.087, -18.854, -18.562, -18.189, -17.723, -17.134, -16.495,
    -19.528, -19.423, -19.279, -19.092, -18.859, -18.566, -18.193, -17.727,
    -17.137, -16.499, -19.534, -19.430, -19.286, -19.098, -18.864, -18.571,
    -18.197, -17.730, -17.141, -16.502, -19.541, -19.436, -19.292, -19.104,
    -18.869, -18.575, -18.201, -17.734, -17.144, -16.506, -19.547, -19.443,
    -19.298, -19.110, -18.875, -18.580, -18.204, -17.738, -17.148, -16.509,
    -19.554, -19.449, -19.304, -19.115, -18.880, -18.584, -18.208, -17.741,
    -17.151, -16.512, -19.560, -19.456, -19.310, -19.121, -18.885, -18.589,
    -18.212, -17.745, -17.155, -16.516, -19.567, -19.462, -19.316, -19.127,
    -18.890, -18.593, -18.216, -17.748, -17.158, -16.519, -19.574, -19.468,
    -19.322, -19.133, -18.895, -18.597, -18.220, -17.752, -17.161, -16.522,
    -19.580, -19.475, -19.328, -19.138, -18.900, -18.602, -18.224, -17.755,
    -17.165, -16.526, -19.587, -19.481, -19.334, -19.144, -18.905, -18.606,
    -18.227, -17.759, -17.168, -16.529, -19.593, -19.488, -19.340, -19.150,
    -18.910, -18.611, -18.231, -17.762, -17.172, -16.532, -19.600, -19.494,
    -19.347, -19.156, -18.915, -18.615, -18.235, -17.766, -17.175, -16.536,
    -19.607, -19.500, -19.353, -19.161, -18.921, -18.619, -18.239, -17.769,
    -17.179, -16.539, -19.613, -19.507, -19.359, -19.167, -18.926, -18.624,
    -18.243, -17.772, -17.182, -16.543, -19.620, -19.513, -19.365, -19.173,
    -18.931, -18.628, -18.246, -17.776, -17.185, -16.546, -19.627, -19.520,
    -19.371, -19.178, -18.936, -18.632, -18.250, -17.779, -17.189, -16.549,
    -19.633, -19.526, -19.377, -19.184, -18.941, -18.637, -18.254, -17.783,
    -17.192, -16.553, -19.640, -19.532, -19.383, -19.190, -18.946, -18.641,
    -18.258, -17.786, -17.196, -16.556, -19.646, -19.539, -19.389, -19.195,
    -18.951, -18.645, -18.261, -17.790, -17.199, -16.559, -19.653, -19.545,
    -19.396, -19.201, -18.956, -18.650, -18.265, -17.793, -17.202, -16.563,
    -19.660, -19.551, -19.402, -19.207, -18.961, -18.654, -18.269, -17.797,
    -17.206, -16.566, -19.666, -19.558, -19.408, -19.212, -18.966, -18.658,
    -18.272, -17.800, -17.209, -16.569, -19.673, -19.564, -19.414, -19.218,
    -18.971, -18.663, -18.276, -17.803, -17.212, -16.572, -19.679, -19.570,
    -19.420, -19.224, -18.976, -18.667, -18.280, -17.807, -17.216, -16.576,
    -19.686, -19.577, -19.426, -19.229, -18.981, -18.671, -18.283, -17.810,
    -17.219, -16.579, -19.692, -19.583, -19.432, -19.235, -18.986, -18.675,
    -18.287, -17.814, -17.222, -16.582, -19.699, -19.589, -19.438, -19.241,
    -18.991, -18.680, -18.291, -17.817, -17.226, -16.586, -19.706, -19.596,
    -19.445, -19.246, -18.996, -18.684, -18.294, -17.821, -17.229, -16.589,
    -19.712, -19.602, -19.451, -19.252, -19.001, -18.688, -18.298, -17.824,
    -17.232, -16.592, -19.719, -19.608, -19.457, -19.258, -19.006, -18.693,
    -18.302, -17.827, -17.235, -16.595, -19.725, -19.615, -19.463, -19.263,
    -19.011, -18.697, -18.305, -17.831, -17.239, -16.599, -19.732, -19.621,
    -19.469, -19.269, -19.016, -18.701, -18.309, -17.834, -17.242, -16.602,
    -19.739, -19.627, -19.475, -19.275, -19.021, -18.705, -18.313, -17.837,
    -17.245, -16.605, -19.745, -19.634, -19.481, -19.280, -19.026, -18.709,
    -18.316, -17.841, -17.249, -16.608, -19.752, -19.640, -19.487, -19.286,
    -19.031, -18.714, -18.320, -17.844, -17.252, -16.612, -19.758, -19.646,
    -19.493, -19.292, -19.036, -18.718, -18.323, -17.847, -17.255, -16.615,
    -19.765, -19.653, -19.499, -19.297, -19.041, -18.722, -18.327, -17.850,
    -17.258, -16.618, -19.771, -19.659, -19.505, -19.303, -19.046, -18.726,
    -18.331, -17.854, -17.262, -16.621, -19.778, -19.665, -19.511, -19.308,
    -19.051, -18.730, -18.334, -17.857, -17.265, -16.624, -19.784, -19.672,
    -19.518, -19.314, -19.056, -18.734, -18.338, -17.860, -17.268, -16.627,
    -19.791, -19.678, -19.524, -19.320, -19.061, -18.738, -18.341, -17.863,
    -17.271, -16.630, -19.797, -19.684, -19.530, -19.325, -19.066, -18.743,
    -18.345, -17.867, -17.274, -16.634, -19.804, -19.691, -19.536, -19.331,
    -19.070, -18.747, -18.348, -17.870, -17.278, -16.637, -19.810, -19.697,
    -19.542, -19.336, -19.075, -18.751, -18.352, -17.873, -17.281, -16.640,
    -19.817, -19.703, -19.548, -19.342, -19.080, -18.755, -18.355, -17.876,
    -17.284, -16.643, -19.823, -19.710, -19.554, -19.348, -19.085, -18.759,
    -18.359, -17.880, -17.287, -16.646, -19.830, -19.716, -19.560, -19.353,
    -19.090, -18.763, -18.362, -17.883, -17.290, -16.649, -19.836, -19.722,
    -19.566, -19.359, -19.095, -18.767, -18.366, -17.886, -17.294, -16.652,
    -19.842, -19.729, -19.572, -19.364, -19.100, -18.771, -18.369, -17.889,
    -17.297, -16.655, -19.849, -19.735, -19.578, -19.370, -19.105, -18.775,
    -18.373, -17.892, -17.300, -16.658, -19.855, -19.741, -19.584, -19.376,
    -19.110, -18.779, -18.376, -17.895, -17.303, -16.661, -19.862, -19.747,
    -19.590, -19.381, -19.115, -18.783, -18.380, -17.898, -17.306, -16.664,
    -19.868, -19.754, -19.596, -19.387, -19.120, -18.787, -18.383, -17.902,
    -17.309, -16.667, -19.875, -19.760, -19.602, -19.392, -19.124, -18.791,
    -18.387, -17.905, -17.313, -16.670, -19.881, -19.766, -19.608, -19.398,
    -19.129, -18.795, -18.390, -17.908, -17.316, -16.673, -19.888, -19.773,
    -19.614, -19.403, -19.134, -18.799, -18.394, -17.911, -17.319, -16.676,
    -19.894, -19.779, -19.620, -19.409, -19.139, -18.803, -18.397, -17.914,
    -17.322, -16.679, -19.900, -19.785, -19.626, -19.415, -19.144, -18.807,
    -18.400, -17.917, -17.325, -16.682, -19.907, -19.792, -19.632, -19.420,
    -19.149, -18.811, -18.404, -17.920, -17.328, -16.685, -19.913, -19.798,
    -19.638, -19.426, -19.153, -18.815, -18.407, -17.923, -17.331, -16.688,
    -19.920, -19.804, -19.644, -19.431, -19.158, -18.819, -18.411, -17.927,
    -17.334, -16.691, -19.926, -19.810, -19.650, -19.437, -19.163, -18.823,
    -18.414, -17.930, -17.337, -16.694, -19.933, -19.817, -19.656, -19.442,
    -19.168, -18.827, -18.418, -17.933, -17.340, -16.697, -19.939, -19.823,
    -19.662, -19.448, -19.173, -18.831, -18.421, -17.936, -17.344, -16.700,
    -19.945, -19.829, -19.668, -19.453, -19.177, -18.835, -18.424, -17.939,
    -17.347, -16.703, -19.952, -19.835, -19.674, -19.459, -19.182, -18.839,
    -18.428, -17.942, -17.350, -16.706, -19.958, -19.842, -19.680, -19.464,
    -19.187, -18.843, -18.431, -17.945, -17.353, -16.709, -19.965, -19.848,
    -19.685, -19.470, -19.192, -18.847, -18.434, -17.948, -17.356, -16.712,
    -19.971, -19.854, -19.691, -19.475, -19.196, -18.851, -18.438, -17.951,
    -17.359, -16.715, -19.977, -19.860, -19.697, -19.481, -19.201, -18.854,
    -18.441, -17.954, -17.362, -16.718, -19.984, -19.867, -19.703, -19.486,
    -19.206, -18.858, -18.445, -17.957, -17.365, -16.721, -19.990, -19.873,
    -19.709, -19.492, -19.210, -18.862, -18.448, -17.961, -17.368, -16.724,
    -19.997, -19.879, -19.715, -19.497, -19.215, -18.866, -18.451, -17.964,
    -17.371, -16.727, -20.003, -19.885, -19.721, -19.502, -19.220, -18.870,
    -18.455, -17.967, -17.374, -16.730, -20.009, -19.892, -19.727, -19.508,
    -19.224, -18.874, -18.458, -17.970, -17.377, -16.733, -20.016, -19.898,
    -19.733, -19.513, -19.229, -18.878, -18.461, -17.973, -17.379, -16.736,
    -20.022, -19.904, -19.739, -19.519, -19.234, -18.882, -18.465, -17.976,
    -17.382, -16.739, -20.029, -19.910, -19.745, -19.524, -19.238, -18.886,
    -18.468, -17.979, -17.385, -16.742, -20.035, -19.916, -19.750, -19.530,
    -19.243, -18.889, -18.471, -17.982, -17.388, -16.745, -20.041, -19.923,
    -19.756, -19.535, -19.248, -18.893, -18.474, -17.985, -17.391, -16.748,
    -20.048, -19.929, -19.762, -19.541, -19.252, -18.897, -18.478, -17.988,
    -17.394, -16.751, -20.054, -19.935, -19.768, -19.546, -19.257, -18.901,
    -18.481, -17.991, -17.397, -16.754, -20.060, -19.941, -19.774, -19.551,
    -19.262, -18.905, -18.484, -17.994, -17.400, -16.757, -20.067, -19.947,
    -19.780, -19.557, -19.266, -18.909, -18.488, -17.997, -17.403, -16.760,
    -20.073, -19.954, -19.786, -19.562, -19.271, -18.913, -18.491, -18.000,
    -17.406, -16.763, -20.079, -19.960, -19.792, -19.568, -19.276, -18.917,
    -18.494, -18.003, -17.409, -16.766, -20.086, -19.966, -19.798, -19.573,
    -19.281, -18.921, -18.498, -18.007, -17.412, -16.769, -20.092, -19.972,
    -19.804, -19.579, -19.285, -18.925, -18.501, -18.010, -17.415, -16.773,
    -20.098, -19.978, -19.809, -19.584, -19.290, -18.928, -18.505, -18.013,
    -17.419, -16.776, -20.105, -19.984, -19.815, -19.589, -19.295, -18.932,
    -18.508, -18.016, -17.422, -16.779, -20.111, -19.991, -19.821, -19.595,
    -19.299, -18.936, -18.511, -18.019, -17.425, -16.782, -20.117, -19.997,
    -19.827, -19.600, -19.304, -18.940, -18.515, -18.022, -17.428, -16.785,
    -20.124, -20.003, -19.833, -19.606, -19.309, -18.944, -18.518, -18.025,
    -17.431, -16.788, -20.130, -20.009, -19.839, -19.611, -19.314, -18.948,
    -18.521, -18.028, -17.434, -16.791, -20.136, -20.015, -19.845, -19.616,
    -19.318, -18.952, -18.525, -18.031, -17.437, -16.795, -20.143, -20.021,
    -19.851, -19.622, -19.323, -18.956, -18.528, -18.035, -17.440, -16.798,
    -20.149, -20.028, -19.857, -19.627, -19.328, -18.960, -18.531, -18.038,
    -17.443, -16.801, -20.155, -20.034, -19.863, -19.633, -19.333, -18.964,
    -18.535, -18.041, -17.447, -16.804, -20.162, -20.040, -19.869, -19.638,
    -19.337, -18.968, -18.538, -18.044, -17.450, -16.807, -20.168, -20.046,
    -19.875, -19.643, -19.342, -18.972, -18.541, -18.047, -17.453, -16.810,
    -20.174, -20.052, -19.880, -19.649, -19.347, -18.975, -18.545, -18.050,
    -17.456, -16.814, -20.181, -20.058, -19.886, -19.654, -19.351, -18.979,
    -18.548, -18.053, -17.459, -16.817, -20.187, -20.064, -19.892, -19.660,
    -19.356, -18.983, -18.552, -18.056, -17.462, -16.820, -20.193, -20.071,
    -19.898, -19.665, -19.361, -18.987, -18.555, -18.060, -17.465, -16.823,
    -20.199, -20.077, -19.904, -19.670, -19.366, -18.991, -18.558, -18.063,
    -17.469, -16.827, -20.206, -20.083, -19.910, -19.676, -19.370, -18.995,
    -18.562, -18.066, -17.472, -16.830, -20.212, -20.089, -19.916, -19.681,
    -19.375, -18.999, -18.565, -18.069, -17.475, -16.833, -20.218, -20.095,
    -19.922, -19.686, -19.380, -19.003, -18.568, -18.072, -17.478, -16.836,
    -20.225, -20.101, -19.928, -19.692, -19.384, -19.007, -18.572, -18.075,
    -17.481, -16.839, -20.231, -20.107, -19.934, -19.697, -19.389, -19.011,
    -18.575, -18.078, -17.484, -16.842, -20.237, -20.113, -19.939, -19.703,
    -19.394, -19.015, -18.579, -18.082, -17.488, -16.845, -20.243, -20.120,
    -19.945, -19.708, -19.399, -19.019, -18.582, -18.085, -17.491, -16.849,
    -20.249, -20.126, -19.951, -19.713, -19.403, -19.023, -18.585, -18.088,
    -17.494, -16.852, -20.256, -20.132, -19.957, -19.719, -19.408, -19.027,
    -18.589, -18.091, -17.497, -16.855, -20.262, -20.138, -19.963, -19.724,
    -19.413, -19.031, -18.592, -18.094, -17.500, -16.858, -20.268, -20.144,
    -19.969, -19.730, -19.417, -19.035, -18.596, -18.098, -17.503, -16.861,
    -20.274, -20.150, -19.975, -19.735, -19.422, -19.039, -18.599, -18.101,
    -17.506, -16.864, -20.281, -20.156, -19.980, -19.740, -19.427, -19.043,
    -18.603, -18.104, -17.510, -16.867, -20.287, -20.162, -19.986, -19.746,
    -19.432, -19.047, -18.606, -18.107, -17.513, -16.870, -20.293, -20.168,
    -19.992, -19.751, -19.436, -19.051, -18.609, -18.110, -17.516, -16.873,
    -20.299, -20.174, -19.998, -19.757, -19.441, -19.055, -18.613, -18.114,
    -17.519, -16.876, -20.305, -20.180, -20.004, -19.762, -19.446, -19.059,
    -18.616, -18.117, -17.522, -16.879, -20.312, -20.187, -20.010, -19.767,
    -19.451, -19.063, -18.620, -18.120, -17.525, -16.883, -20.318, -20.193,
    -20.015, -19.773, -19.455, -19.067, -18.623, -18.123, -17.529, -16.886,
    -20.324, -20.199, -20.021, -19.778, -19.460, -19.071, -18.627, -18.126,
    -17.532, -16.889, -20.330, -20.205, -20.027, -19.784, -19.465, -19.075,
    -18.630, -18.130, -17.535, -16.892, -20.336, -20.211, -20.033, -19.789,
    -19.469, -19.079, -18.634, -18.133, -17.538, -16.895, -20.342, -20.217,
    -20.039, -19.794, -19.474, -19.083, -18.637, -18.136, -17.541, -16.898,
    -20.349, -20.223, -20.045, -19.800, -19.479, -19.087, -18.641, -18.139,
    -17.544, -16.901, -20.355, -20.229, -20.050, -19.805, -19.484, -19.091,
    -18.644, -18.143, -17.548, -16.904, -20.361, -20.235, -20.056, -19.811,
    -19.488, -19.095, -18.648, -18.146, -17.551, -16.907, -20.367, -20.241,
    -20.062, -19.816, -19.493, -19.099, -18.651, -18.149, -17.554, -16.910,
    -20.373, -20.247, -20.068, -19.821, -19.498, -19.103, -18.654, -18.152,
    -17.557, -16.913, -20.379, -20.253, -20.074, -19.827, -19.502, -19.107,
    -18.658, -18.156, -17.560, -16.916, -20.385, -20.259, -20.079, -19.832,
    -19.507, -19.111, -18.661, -18.159, -17.564, -16.920, -20.392, -20.265,
    -20.085, -19.838, -19.512, -19.115, -18.665, -18.162, -17.567, -16.923,
    -20.398, -20.271, -20.091, -19.843, -19.517, -19.119, -18.668, -18.165,
    -17.570, -16.926, -20.404, -20.277, -20.097, -19.848, -19.521, -19.123,
    -18.672, -18.169, -17.573, -16.929, -20.410, -20.283, -20.103, -19.854,
    -19.526, -19.127, -18.675, -18.172, -17.577, -16.932, -20.416, -20.289,
    -20.108, -19.859, -19.531, -19.131, -18.679, -18.175, -17.580, -16.935,
    -20.422, -20.295, -20.114, -19.865, -19.536, -19.135, -18.682, -18.179,
    -17.583, -16.939, -20.428, -20.301, -20.120, -19.870, -19.540, -19.139,
    -18.686, -18.182, -17.586, -16.942, -20.435, -20.307, -20.126, -19.875,
    -19.545, -19.143, -18.689, -18.185, -17.590, -16.945, -20.441, -20.313,
    -20.131, -19.881, -19.550, -19.147, -18.693, -18.188, -17.593, -16.948,
    -20.447, -20.319, -20.137, -19.886, -19.555, -19.151, -18.697, -18.192,
    -17.596, -16.952, -20.453, -20.325, -20.143, -19.892, -19.560, -19.155,
    -18.700, -18.195, -17.600, -16.955, -20.459, -20.331, -20.149, -19.897,
    -19.564, -19.159, -18.704, -18.198, -17.603, -16.958, -20.465, -20.337,
    -20.154, -19.902, -19.569, -19.164, -18.707, -18.202, -17.606, -16.961,
    -20.471, -20.343, -20.160, -19.908, -19.574, -19.168, -18.711, -18.205,
    -17.609, -16.965, -20.477, -20.349, -20.166, -19.913, -19.579, -19.172,
    -18.714, -18.208, -17.613, -16.968, -20.483, -20.355, -20.172, -19.919,
    -19.583, -19.176, -18.718, -18.212, -17.616, -16.971, -20.490, -20.361,
    -20.177, -19.924, -19.588, -19.180, -18.721, -18.215, -17.619, -16.974,
    -20.496, -20.367, -20.183, -19.929, -19.593, -19.184, -18.725, -18.219,
    -17.623, -16.978, -20.502, -20.373, -20.189, -19.935, -19.598, -19.188,
    -18.728, -18.222, -17.626, -16.981, -20.508, -20.379, -20.195, -19.940,
    -19.602, -19.192, -18.732, -18.225, -17.629, -16.984, -20.514, -20.385,
    -20.200, -19.946, -19.607, -19.196, -18.735, -18.229, -17.633, -16.988,
    -20.520, -20.391, -20.206, -19.951, -19.612, -19.200, -18.739, -18.232,
    -17.636, -16.991, -20.526, -20.397, -20.212, -19.956, -19.617, -19.204,
    -18.743, -18.235, -17.639, -16.994, -20.532, -20.403, -20.217, -19.962,
    -19.622, -19.208, -18.746, -18.239, -17.643, -16.998, -20.538, -20.409,
    -20.223, -19.967, -19.626, -19.212, -18.750, -18.242, -17.646, -17.001,
    -20.544, -20.415, -20.229, -19.973, -19.631, -19.217, -18.753, -18.245,
    -17.649, -17.004, -20.550, -20.421, -20.235, -19.978, -19.636, -19.221,
    -18.757, -18.249, -17.653, -17.007, -20.556, -20.427, -20.240, -19.983,
    -19.641, -19.225, -18.761, -18.252, -17.656, -17.011, -20.562, -20.432,
    -20.246, -19.989, -19.646, -19.229, -18.764, -18.256, -17.659, -17.014,
    -20.568, -20.438, -20.252, -19.994, -19.651, -19.233, -18.768, -18.259,
    -17.663, -17.017, -20.574, -20.444, -20.258, -19.999, -19.655, -19.237,
    -18.772, -18.262, -17.666, -17.021, -20.580, -20.450, -20.263, -20.005,
    -19.660, -19.242, -18.775, -18.266, -17.669, -17.024, -20.586, -20.456,
    -20.269, -20.010, -19.665, -19.246, -18.779, -18.269, -17.673, -17.027,
    -20.592, -20.462, -20.275, -20.015, -19.670, -19.250, -18.783, -18.273,
    -17.676, -17.031, -20.598, -20.468, -20.280, -20.021, -19.675, -19.254,
    -18.786, -18.276, -17.679, -17.034, -20.604, -20.474, -20.286, -20.026,
    -19.680, -19.258, -18.790, -18.279, -17.683, -17.037, -20.610, -20.480,
    -20.292, -20.032, -19.684, -19.263, -18.794, -18.283, -17.686, -17.041,
    -20.616, -20.485, -20.298, -20.037, -19.689, -19.267, -18.797, -18.286,
    -17.690, -17.044, -20.622, -20.491, -20.303, -20.042, -19.694, -19.271,
    -18.801, -18.290, -17.693, -17.047, -20.628, -20.497, -20.309, -20.048,
    -19.699, -19.275, -18.805, -18.293, -17.696, -17.051, -20.634, -20.503,
    -20.315, -20.053, -19.704, -19.279, -18.809, -18.296, -17.700, -17.054,
    -20.640, -20.509, -20.320, -20.058, -19.709, -19.284, -18.812, -18.300,
    -17.703, -17.057, -20.646, -20.515, -20.326, -20.064, -19.714, -19.288,
    -18.816, -18.303, -17.706, -17.061, -20.652, -20.521, -20.332, -20.069,
    -19.718, -19.292, -18.820, -18.307, -17.710, -17.064, -20.658, -20.526,
    -20.338, -20.074, -19.723, -19.296, -18.823, -18.310, -17.713, -17.067,
    -20.664, -20.532, -20.343, -20.080, -19.728, -19.301, -18.827, -18.314,
    -17.717, -17.071, -20.670, -20.538, -20.349, -20.085, -19.733, -19.305,
    -18.831, -18.317, -17.720, -17.074, -20.676, -20.544, -20.355, -20.090,
    -19.738, -19.309, -18.835, -18.320, -17.723, -17.077, -20.682, -20.550,
    -20.360, -20.096, -19.743, -19.314, -18.838, -18.324, -17.727, -17.081,
    -20.688, -20.556, -20.366, -20.101, -19.748, -19.318, -18.842, -18.327,
    -17.730, -17.084, -20.694, -20.561, -20.372, -20.107, -19.753, -19.322,
    -18.846, -18.331, -17.734, -17.087, -20.700, -20.567, -20.377, -20.112,
    -19.757, -19.326, -18.850, -18.334, -17.737, -17.090, -20.706, -20.573,
    -20.383, -20.117, -19.762, -19.331, -18.854, -18.338, -17.740, -17.094,
    -20.712, -20.579, -20.389, -20.123, -19.767, -19.335, -18.857, -18.341,
    -17.744, -17.097, -20.718, -20.585, -20.394, -20.128, -19.772, -19.339,
    -18.861, -18.345, -17.747, -17.100, -20.724, -20.591, -20.400, -20.133,
    -19.777, -19.344, -18.865, -18.348, -17.750, -17.104, -20.729, -20.597,
    -20.406, -20.139, -19.782, -19.348, -18.869, -18.352, -17.754, -17.107,
    -20.735, -20.602, -20.411, -20.144, -19.787, -19.352, -18.872, -18.355,
    -17.757, -17.110, -20.741, -20.608, -20.417, -20.150, -19.792, -19.356,
    -18.876, -18.359, -17.761, -17.113, -20.747, -20.614, -20.423, -20.155,
    -19.797, -19.361, -18.880, -18.362, -17.764, -17.117, -20.753, -20.620,
    -20.428, -20.160, -19.802, -19.365, -18.884, -18.365, -17.767, -17.120,
    -20.759, -20.626, -20.434, -20.166, -19.807, -19.369, -18.888, -18.369,
    -17.771, -17.123, -20.765, -20.632, -20.439, -20.171, -19.811, -19.374,
    -18.891, -18.372, -17.774, -17.127, -20.771, -20.637, -20.445, -20.177,
    -19.816, -19.378, -18.895, -18.376, -17.778, -17.130, -20.777, -20.643,
    -20.451, -20.182, -19.821, -19.382, -18.899, -18.379, -17.781, -17.133,
    -20.783, -20.649, -20.456, -20.188, -19.826, -19.387, -18.903, -18.383,
    -17.785, -17.136, -20.789, -20.655, -20.462, -20.193, -19.831, -19.391,
    -18.907, -18.386, -17.788, -17.140, -20.794, -20.661, -20.468, -20.198,
    -19.836, -19.396, -18.911, -18.390, -17.791, -17.143, -20.800, -20.667,
    -20.473, -20.204, -19.841, -19.400, -18.914, -18.393, -17.795, -17.146,
    -20.806, -20.672, -20.479, -20.209, -19.846, -19.404, -18.918, -18.397,
    -17.798, -17.149, -20.812, -20.678, -20.484, -20.215, -19.851, -19.409,
    -18.922, -18.400, -17.802, -17.153, -20.818, -20.684, -20.490, -20.220,
    -19.856, -19.413, -18.926, -18.404, -17.805, -17.156, -20.824, -20.690,
    -20.496, -20.225, -19.861, -19.417, -18.930, -18.408, -17.808, -17.159,
    -20.830, -20.696, -20.501, -20.231, -19.866, -19.422, -18.934, -18.411,
    -17.812, -17.163, -20.836, -20.701, -20.507, -20.236, -19.871, -19.426,
    -18.938, -18.415, -17.815, -17.166, -20.842, -20.707, -20.512, -20.242,
    -19.876, -19.431, -18.941, -18.418, -17.819, -17.169, -20.847, -20.713,
    -20.518, -20.247, -19.881, -19.435, -18.945, -18.422, -17.822, -17.172,
    -20.853, -20.719, -20.524, -20.252, -19.886, -19.440, -18.949, -18.425,
    -17.825, -17.176, -20.859, -20.725, -20.529, -20.258, -19.891, -19.444,
    -18.953, -18.429, -17.829, -17.179, -20.865, -20.730, -20.535, -20.263,
    -19.896, -19.448, -18.957, -18.432, -17.832, -17.182, -20.871, -20.736,
    -20.541, -20.268, -19.901, -19.453, -18.961, -18.436, -17.835, -17.185,
    -20.877, -20.742, -20.546, -20.274, -19.905, -19.457, -18.965, -18.439,
    -17.839, -17.189, -20.883, -20.748, -20.552, -20.279, -19.910, -19.462,
    -18.969, -18.443, -17.842, -17.192, -20.889, -20.753, -20.557, -20.284,
    -19.915, -19.466, -18.973, -18.446, -17.845, -17.195, -20.895, -20.759,
    -20.563, -20.290, -19.920, -19.471, -18.977, -18.450, -17.849, -17.198,
    -20.900, -20.765, -20.569, -20.295, -19.925, -19.475, -18.981, -18.454,
    -17.852, -17.201, -20.906, -20.771, -20.574, -20.301, -19.930, -19.480,
    -18.984, -18.457, -17.856, -17.205, -20.912, -20.776, -20.580, -20.306,
    -19.935, -19.484, -18.988, -18.461, -17.859, -17.208, -20.918, -20.782,
    -20.585, -20.311, -19.940, -19.489, -18.992, -18.464, -17.862, -17.211,
    -20.924, -20.788, -20.591, -20.317, -19.945, -19.493, -18.996, -18.468,
    -17.866, -17.214, -20.930, -20.794, -20.596, -20.322, -19.950, -19.498,
    -19.000, -18.471, -17.869, -17.218, -20.936, -20.799, -20.602, -20.327,
    -19.955, -19.502, -19.004, -18.475, -17.872, -17.221, -20.942, -20.805,
    -20.608, -20.333, -19.960, -19.507, -19.008, -18.479, -17.876, -17.224,
    -20.947, -20.811, -20.613, -20.338, -19.965, -19.511, -19.012, -18.482,
    -17.879, -17.227, -20.953, -20.817, -20.619, -20.343, -19.970, -19.516,
    -19.016, -18.486, -17.882, -17.231, -20.959, -20.822, -20.624, -20.349,
    -19.975, -19.520, -19.020, -18.489, -17.886, -17.234, -20.965, -20.828,
    -20.630, -20.354, -19.980, -19.525, -19.024, -18.493, -17.889, -17.237,
    -20.971, -20.834, -20.636, -20.359, -19.985, -19.529, -19.028, -18.497,
    -17.892, -17.240, -20.977, -20.839, -20.641, -20.365, -19.990, -19.534,
    -19.032, -18.500, -17.896, -17.243, -20.982, -20.845, -20.647, -20.370,
    -19.995, -19.539, -19.036, -18.504, -17.899, -17.247, -20.988, -20.851,
    -20.652, -20.375, -20.000, -19.543, -19.040, -18.507, -17.902, -17.250,
    -20.994, -20.857, -20.658, -20.381, -20.005, -19.548, -19.044, -18.511,
    -17.906, -17.253, -21.000, -20.862, -20.663, -20.386, -20.010, -19.552,
    -19.048, -18.514, -17.909, -17.256, -21.006, -20.868, -20.669, -20.391,
    -20.015, -19.557, -19.052, -18.518, -17.912, -17.259, -21.011, -20.874,
    -20.675, -20.397, -20.020, -19.561, -19.056, -18.521, -17.915, -17.262,
    -21.017, -20.879, -20.680, -20.402, -20.025, -19.566, -19.060, -18.525,
    -17.919, -17.265, -21.023, -20.885, -20.686, -20.407, -20.030, -19.570,
    -19.064, -18.529, -17.922, -17.269, -21.029, -20.891, -20.691, -20.413,
    -20.035, -19.575, -19.068, -18.532, -17.925, -17.272, -21.034, -20.896,
    -20.697, -20.418, -20.040, -19.579, -19.072, -18.536, -17.929, -17.275,
    -21.040, -20.902, -20.702, -20.423, -20.045, -19.584, -19.076, -18.539,
    -17.932, -17.278, -21.046, -20.908, -20.708, -20.429, -20.050, -19.588,
    -19.080, -18.543, -17.935, -17.281, -21.052, -20.913, -20.714, -20.434,
    -20.055, -19.593, -19.084, -18.546, -17.938, -17.284, -21.058, -20.919,
    -20.719, -20.439, -20.060, -19.597, -19.088, -18.550, -17.942, -17.287,
    -21.063, -20.925, -20.725, -20.445, -20.065, -19.602, -19.092, -18.554,
    -17.945, -17.290, -21.069, -20.930, -20.730, -20.450, -20.071, -19.606,
    -19.096, -18.557, -17.948, -17.293, -21.075, -20.936, -20.736, -20.455,
    -20.076, -19.611, -19.100, -18.561, -17.952, -17.297, -21.080, -20.942,
    -20.741, -20.460, -20.081, -19.615, -19.104, -18.564, -17.955, -17.300,
    -21.086, -20.947, -20.747, -20.466, -20.086, -19.620, -19.108, -18.568,
    -17.958, -17.303, -21.092, -20.953, -20.752, -20.471, -20.091, -19.624,
    -19.112, -18.571, -17.961, -17.306, -21.098, -20.959, -20.758, -20.476,
    -20.096, -19.629, -19.116, -18.575, -17.965, -17.309, -21.103, -20.964,
    -20.763, -20.482, -20.101, -19.633, -19.120, -18.578, -17.968, -17.312,
    -21.109, -20.970, -20.769, -20.487, -20.106, -19.638, -19.124, -18.582,
    -17.971, -17.315, -21.115, -20.976, -20.775, -20.492, -20.111, -19.643,
    -19.128, -18.586, -17.974, -17.318, -21.120, -20.981, -20.780, -20.498,
    -20.116, -19.647, -19.132, -18.589, -17.977, -17.321, -21.126, -20.987,
    -20.786, -20.503, -20.121, -19.652, -19.136, -18.593, -17.981, -17.324,
    -21.132, -20.993, -20.791, -20.508, -20.126, -19.656, -19.140, -18.596,
    -17.984, -17.327, -21.138, -20.998, -20.797, -20.514, -20.131, -19.661,
    -19.144, -18.600, -17.987, -17.330, -21.143, -21.004, -20.802, -20.519,
    -20.136, -19.665, -19.148, -18.603, -17.990, -17.333, -21.149, -21.009,
    -20.808, -20.524, -20.141, -19.670, -19.152, -18.607, -17.994, -17.336,
    -21.155, -21.015, -20.813, -20.530, -20.146, -19.675, -19.156, -18.610,
    -17.997, -17.339, -21.160, -21.021, -20.819, -20.535, -20.151, -19.679,
    -19.160, -18.614, -18.000, -17.342, -21.166, -21.026, -20.824, -20.540,
    -20.156, -19.684, -19.165, -18.617, -18.003, -17.345, -21.172, -21.032,
    -20.830, -20.546, -20.161, -19.688, -19.169, -18.621, -18.006, -17.348,
    -21.177, -21.037, -20.835, -20.551, -20.166, -19.693, -19.173, -18.624,
    -18.010, -17.351, -21.183, -21.043, -20.840, -20.556, -20.171, -19.698,
    -19.177, -18.628, -18.013, -17.354, -21.189, -21.049, -20.846, -20.561,
    -20.176, -19.702, -19.181, -18.631, -18.016, -17.357, -21.194, -21.054,
    -20.851, -20.567, -20.181, -19.707, -19.185, -18.635, -18.019, -17.360,
    -21.200, -21.060, -20.857, -20.572, -20.186, -19.711, -19.189, -18.638,
    -18.022, -17.363, -21.206, -21.065, -20.862, -20.577, -20.191, -19.716,
    -19.193, -18.642, -18.026, -17.366, -21.211, -21.071, -20.868, -20.583,
    -20.196, -19.721, -19.197, -18.645, -18.029, -17.369, -21.217, -21.077,
    -20.873, -20.588, -20.201, -19.725, -19.201, -18.649, -18.032, -17.372,
    -21.223, -21.082, -20.879, -20.593, -20.206, -19.730, -19.205, -18.652,
    -18.035, -17.375, -21.228, -21.088, -20.884, -20.599, -20.211, -19.734,
    -19.210, -18.656, -18.038, -17.377, -21.234, -21.093, -20.890, -20.604,
    -20.216, -19.739, -19.214, -18.659, -18.042, -17.380, -21.240, -21.099,
    -20.895, -20.609, -20.221, -19.744, -19.218, -18.663, -18.045, -17.383,
    -21.245, -21.104, -20.901, -20.615, -20.226, -19.748, -19.222, -18.666,
    -18.048, -17.386, -21.251, -21.110, -20.906, -20.620, -20.231, -19.753,
    -19.226, -18.670, -18.051, -17.389, -21.257, -21.116, -20.911, -20.625,
    -20.236, -19.758, -19.230, -18.674, -18.054, -17.392, -21.262, -21.121,
    -20.917, -20.631, -20.241, -19.762, -19.234, -18.677, -18.057, -17.395,
    -21.268, -21.127, -20.922, -20.636, -20.246, -19.767, -19.238, -18.681,
    -18.060, -17.398, -21.273, -21.132, -20.928, -20.641, -20.251, -19.771,
    -19.242, -18.684, -18.064, -17.401, -21.279, -21.138, -20.933, -20.646,
    -20.256, -19.776, -19.246, -18.688, -18.067, -17.404, -21.285, -21.143,
    -20.939, -20.652, -20.261, -19.781, -19.250, -18.691, -18.070, -17.406,
    -21.290, -21.149, -20.944, -20.657, -20.266, -19.785, -19.254, -18.695,
    -18.073, -17.409, -21.296, -21.154, -20.949, -20.662, -20.271, -19.790,
    -19.258, -18.698, -18.076, -17.412, -21.301, -21.160, -20.955, -20.668,
    -20.276, -19.794, -19.262, -18.702, -18.079, -17.415, -21.307, -21.165,
    -20.960, -20.673, -20.281, -19.799, -19.267, -18.705, -18.082, -17.418,
    -21.313, -21.171, -20.966, -20.678, -20.286, -19.804, -19.271, -18.709,
    -18.085, -17.421, -21.318, -21.176, -20.971, -20.683, -20.291, -19.808,
    -19.275, -18.712, -18.088, -17.424, -21.324, -21.182, -20.976, -20.689,
    -20.296, -19.813, -19.279, -18.716, -18.091, -17.427, -21.329, -21.188,
    -20.982, -20.694, -20.301, -19.817, -19.283, -18.719, -18.094, -17.430,
    -21.335, -21.193, -20.987, -20.699, -20.306, -19.822, -19.287, -18.723,
    -18.098, -17.432, -21.340, -21.199, -20.993, -20.705, -20.312, -19.827,
    -19.291, -18.726, -18.101, -17.435, -21.346, -21.204, -20.998, -20.710,
    -20.317, -19.831, -19.295, -18.730, -18.104, -17.438, -21.351, -21.210,
    -21.003, -20.715, -20.322, -19.836, -19.299, -18.733, -18.107, -17.441,
    -21.357, -21.215, -21.009, -20.720, -20.327, -19.840, -19.303, -18.737,
    -18.110, -17.444, -21.362, -21.221, -21.014, -20.726, -20.332, -19.845,
    -19.307, -18.740, -18.113, -17.447, -21.368, -21.226, -21.020, -20.731,
    -20.337, -19.850, -19.311, -18.744, -18.116, -17.450, -21.373, -21.232,
    -21.025, -20.736, -20.342, -19.854, -19.315, -18.747, -18.119, -17.452,
    -21.379, -21.237, -21.030, -20.741, -20.347, -19.859, -19.319, -18.751,
    -18.122, -17.455, -21.384, -21.243, -21.036, -20.747, -20.352, -19.863,
    -19.323, -18.754, -18.125, -17.458, -21.390, -21.248, -21.041, -20.752,
    -20.357, -19.868, -19.327, -18.758, -18.128, -17.461, -21.396, -21.253,
    -21.046, -20.757, -20.362, -19.873, -19.331, -18.761, -18.131, -17.464,
    -21.401, -21.259, -21.052, -20.762, -20.367, -19.877, -19.335, -18.765,
    -18.134, -17.467, -21.407, -21.264, -21.057, -20.768, -20.372, -19.882,
    -19.339, -18.768, -18.137, -17.469, -21.412, -21.270, -21.063, -20.773,
    -20.377, -19.886, -19.343, -18.772, -18.140, -17.472, -21.418, -21.275,
    -21.068, -20.778, -20.382, -19.891, -19.347, -18.775, -18.143, -17.475,
    -21.423, -21.281, -21.073, -20.783, -20.387, -19.896, -19.351, -18.779,
    -18.146, -17.478, -21.429, -21.286, -21.079, -20.789, -20.392, -19.900,
    -19.355, -18.782, -18.149, -17.481, -21.434, -21.292, -21.084, -20.794,
    -20.397, -19.905, -19.359, -18.785, -18.152, -17.483, -21.440, -21.297,
    -21.089, -20.799, -20.402, -19.909, -19.363, -18.789, -18.155, -17.486,
    -21.445, -21.303, -21.095, -20.804, -20.407, -19.914, -19.367, -18.792,
    -18.158, -17.489, -21.451, -21.308, -21.100, -20.809, -20.412, -19.919,
    -19.371, -18.795, -18.161, -17.492, -21.456, -21.314, -21.105, -20.815,
    -20.417, -19.923, -19.375, -18.799, -18.164, -17.494, -21.462, -21.319,
    -21.111, -20.820, -20.422, -19.928, -19.380, -18.802, -18.166, -17.497,
    -21.467, -21.324, -21.116, -20.825, -20.427, -19.932, -19.384, -18.806,
    -18.169, -17.500, -21.473, -21.330, -21.122, -20.830, -20.432, -19.937,
    -19.388, -18.809, -18.172, -17.503, -21.478, -21.335, -21.127, -20.835,
    -20.437, -19.942, -19.392, -18.812, -18.175, -17.506, -21.484, -21.341,
    -21.132, -20.841, -20.442, -19.946, -19.396, -18.816, -18.178, -17.508,
    -21.489, -21.346, -21.138, -20.846, -20.447, -19.951, -19.400, -18.819,
    -18.181, -17.511, -21.495, -21.352, -21.143, -20.851, -20.452, -19.955,
    -19.404, -18.822, -18.184, -17.514, -21.500, -21.357, -21.148, -20.856,
    -20.457, -19.960, -19.408, -18.826, -18.187, -17.516, -21.506, -21.362,
    -21.154, -20.861, -20.462, -19.965, -19.412, -18.829, -18.190, -17.519,
    -21.511, -21.368, -21.159, -20.867, -20.467, -19.969, -19.416, -18.832,
    -18.192, -17.522, -21.517, -21.373, -21.164, -20.872, -20.472, -19.974,
    -19.420, -18.835, -18.195, -17.525, -21.522, -21.379, -21.170, -20.877,
    -20.477, -19.978, -19.424, -18.839, -18.198, -17.527, -21.528, -21.384,
    -21.175, -20.882, -20.482, -19.983, -19.428, -18.842, -18.201, -17.530,
    -21.534, -21.389, -21.180, -20.887, -20.487, -19.988, -19.432, -18.845,
    -18.204, -17.533, -21.539, -21.395, -21.186, -20.892, -20.492, -19.992,
    -19.436, -18.849, -18.207, -17.535, -21.544, -21.400, -21.191, -20.898,
    -20.497, -19.997, -19.440, -18.852, -18.209, -17.538, -21.550, -21.406,
    -21.196, -20.903, -20.502, -20.001, -19.444, -18.855, -18.212, -17.541,
    -21.555, -21.411, -21.202, -20.908, -20.507, -20.006, -19.448, -18.858,
    -18.215, -17.543, -21.561, -21.416, -21.207, -20.913, -20.512, -20.010,
    -19.452, -18.862, -18.218, -17.546, -21.566, -21.422, -21.212, -20.918,
    -20.517, -20.015, -19.456, -18.865, -18.221, -17.549, -21.572, -21.427,
    -21.217, -20.924, -20.522, -20.020, -19.460, -18.868, -18.224, -17.551,
    -21.577, -21.432, -21.223, -20.929, -20.527, -20.024, -19.464, -18.871,
    -18.226, -17.554, -21.583, -21.438, -21.228, -20.934, -20.532, -20.029,
    -19.468, -18.874, -18.229, -17.557, -21.588, -21.443, -21.233, -20.939,
    -20.537, -20.033, -19.471, -18.878, -18.232, -17.559, -21.594, -21.449,
    -21.239, -20.944, -20.542, -20.038, -19.475, -18.881, -18.235, -17.562,
    -21.599, -21.454, -21.244, -20.950, -20.547, -20.042, -19.479, -18.884,
    -18.237, -17.565, -21.605, -21.459, -21.249, -20.955, -20.552, -20.047,
    -19.483, -18.887, -18.240, -17.567, -21.610, -21.465, -21.254, -20.960,
    -20.557, -20.052, -19.487, -18.890, -18.243, -17.570, -21.616, -21.470,
    -21.260, -20.965, -20.561, -20.056, -19.491, -18.894, -18.246, -17.573,
    -21.621, -21.475, -21.265, -20.970, -20.566, -20.061, -19.495, -18.897,
    -18.248, -17.575, -21.627, -21.481, -21.270, -20.976, -20.571, -20.065,
    -19.499, -18.900, -18.251, -17.578, -21.632, -21.486, -21.276, -20.981,
    -20.576, -20.070, -19.503, -18.903, -18.254, -17.580, -21.638, -21.491,
    -21.281, -20.986, -20.581, -20.074, -19.507, -18.906, -18.257, -17.583,
    -21.643, -21.497, -21.286, -20.991, -20.586, -20.079, -19.511, -18.909,
    -18.259, -17.586, -21.649, -21.502, -21.291, -20.996, -20.591, -20.083,
    -19.514, -18.913, -18.262, -17.588, -21.654, -21.507, -21.297, -21.002,
    -20.596, -20.088, -19.518, -18.916, -18.265, -17.591, -21.660, -21.513,
    -21.302, -21.007, -20.601, -20.092, -19.522, -18.919, -18.267, -17.593,
    -21.665, -21.518, -21.307, -21.012, -20.606, -20.097, -19.526, -18.922,
    -18.270, -17.596, -21.670, -21.523, -21.312, -21.017, -20.611, -20.102,
    -19.530, -18.925, -18.273, -17.599, -21.676, -21.529, -21.318, -21.022,
    -20.616, -20.106, -19.534, -18.928, -18.275, -17.601, -21.681, -21.534,
    -21.323, -21.027, -20.621, -20.111, -19.537, -18.931, -18.278, -17.604,
    -21.687, -21.539, -21.328, -21.033, -20.626, -20.115, -19.541, -18.934,
    -18.281, -17.606, -21.692, -21.545, -21.333, -21.038, -20.631, -20.120,
    -19.545, -18.937, -18.283, -17.609, -21.697, -21.550, -21.339, -21.043,
    -20.636, -20.124, -19.549, -18.940, -18.286, -17.611, -21.703, -21.555,
    -21.344, -21.048, -20.641, -20.129, -19.553, -18.944, -18.289, -17.614,
    -21.708, -21.561, -21.349, -21.053, -20.645, -20.133, -19.557, -18.947,
    -18.291, -17.616, -21.714, -21.566, -21.354, -21.058, -20.650, -20.138,
    -19.560, -18.950, -18.294, -17.619, -21.719, -21.571, -21.360, -21.063,
    -20.655, -20.142, -19.564, -18.953, -18.296, -17.622, -21.724, -21.577,
    -21.365, -21.069, -20.660, -20.147, -19.568, -18.956, -18.299, -17.624,
    -21.730, -21.582, -21.370, -21.074, -20.665, -20.151, -19.572, -18.959,
    -18.302, -17.627, -21.735, -21.587, -21.375, -21.079, -20.670, -20.155,
    -19.575, -18.962, -18.304, -17.629, -21.740, -21.593, -21.380, -21.084,
    -20.675, -20.160, -19.579, -18.965, -18.307, -17.632, -21.746, -21.598,
    -21.386, -21.089, -20.680, -20.164, -19.583, -18.968, -18.309, -17.634,
    -21.751, -21.603, -21.391, -21.094, -20.685, -20.169, -19.587, -18.971,
    -18.312, -17.637, -21.756, -21.609, -21.396, -21.099, -20.690, -20.173,
    -19.590, -18.974, -18.315, -17.639, -21.762, -21.614, -21.401, -21.104,
    -20.695, -20.178, -19.594, -18.977, -18.317, -17.642, -21.767, -21.619,
    -21.407, -21.109, -20.700, -20.182, -19.598, -18.979, -18.320, -17.644,
    -21.772, -21.624, -21.412, -21.115, -20.704, -20.187, -19.602, -18.982,
    -18.322, -17.647, -21.778, -21.630, -21.417, -21.120, -20.709, -20.191,
    -19.605, -18.985, -18.325, -17.649, -21.783, -21.635, -21.422, -21.125,
    -20.714, -20.196, -19.609, -18.988, -18.327, -17.652, -21.788, -21.640,
    -21.428, -21.130, -20.719, -20.200, -19.613, -18.991, -18.330, -17.654,
    -21.794, -21.646, -21.433, -21.135, -20.724, -20.205, -19.616, -18.994,
    -18.332, -17.657, -21.799, -21.651, -21.438, -21.140, -20.729, -20.209,
    -19.620, -18.997, -18.335, -17.659, -21.804, -21.656, -21.443, -21.145,
    -20.734, -20.213, -19.624, -19.000, -18.337, -17.661, -21.810, -21.662,
    -21.448, -21.150, -20.739, -20.218, -19.627, -19.003, -18.340, -17.664,
    -21.815, -21.667, -21.454, -21.155, -20.744, -20.222, -19.631, -19.006,
    -18.342, -17.666, -21.820, -21.672, -21.459, -21.160, -20.749, -20.227,
    -19.635, -19.008, -18.345, -17.669, -21.826, -21.677, -21.464, -21.165,
    -20.753, -20.231, -19.638, -19.011, -18.347, -17.671, -21.831, -21.683,
    -21.469, -21.171, -20.758, -20.235, -19.642, -19.014, -18.350, -17.673,
    -21.836, -21.688, -21.474, -21.176, -20.763, -20.240, -19.645, -19.017,
    -18.352, -17.676, -21.841, -21.693, -21.479, -21.181, -20.768, -20.244,
    -19.649, -19.020, -18.355, -17.678, -21.847, -21.698, -21.485, -21.186,
    -20.773, -20.249, -19.653, -19.023, -18.357, -17.680, -21.852, -21.704,
    -21.490, -21.191, -20.778, -20.253, -19.656, -19.025, -18.359, -17.683,
    -21.857, -21.709, -21.495, -21.196, -20.783, -20.257, -19.660, -19.028,
    -18.362, -17.685, -21.863, -21.714, -21.500, -21.201, -20.788, -20.262,
    -19.663, -19.031, -18.364, -17.687, -21.868, -21.719, -21.505, -21.206,
    -20.793, -20.266, -19.667, -19.034, -18.366, -17.690, -21.873, -21.725,
    -21.510, -21.211, -20.797, -20.270, -19.670, -19.036, -18.369, -17.692,
    -21.878, -21.730, -21.516, -21.216, -20.802, -20.275, -19.674, -19.039,
    -18.371, -17.694, -21.884, -21.735, -21.521, -21.221, -20.807, -20.279,
    -19.678, -19.042, -18.373, -17.697, -21.889, -21.740, -21.526, -21.226,
    -20.812, -20.283, -19.681, -19.045, -18.376, -17.699, -21.894, -21.746,
    -21.531, -21.232, -20.817, -20.288, -19.685, -19.047, -18.378, -17.701,
    -21.900, -21.751, -21.536, -21.237, -20.822, -20.292, -19.688, -19.050,
    -18.380, -17.704, -21.905, -21.756, -21.541, -21.242, -20.827, -20.296,
    -19.692, -19.053, -18.383, -17.706, -21.910, -21.761, -21.546, -21.247,
    -20.832, -20.301, -19.695, -19.055, -18.385, -17.708, -21.915, -21.766,
    -21.552, -21.252, -20.836, -20.305, -19.699, -19.058, -18.387, -17.710,
    -21.921, -21.772, -21.557, -21.257, -20.841, -20.309, -19.702, -19.061,
    -18.389, -17.713, -21.926, -21.777, -21.562, -21.262, -20.846, -20.314,
    -19.706, -19.063, -18.392, -17.715, -21.931, -21.782, -21.567, -21.267,
    -20.851, -20.318, -19.709, -19.066, -18.394, -17.717, -21.936, -21.787,
    -21.572, -21.272, -20.856, -20.322, -19.712, -19.069, -18.396, -17.719,
    -21.942, -21.792, -21.577, -21.277, -20.861, -20.326, -19.716, -19.071,
    -18.398, -17.722, -21.947, -21.798, -21.582, -21.282, -20.865, -20.331,
    -19.719, -19.074, -18.401, -17.724, -21.952, -21.803, -21.588, -21.287,
    -20.870, -20.335, -19.723, -19.076, -18.403, -17.726, -21.957, -21.808,
    -21.593, -21.292, -20.875, -20.339, -19.726, -19.079, -18.405, -17.728,
    -21.963, -21.813, -21.598, -21.297, -20.880, -20.343, -19.729, -19.082,
    -18.407, -17.730, -21.968, -21.818, -21.603, -21.302, -20.885, -20.347,
    -19.733, -19.084, -18.409, -17.733, -21.973, -21.823, -21.608, -21.307,
    -20.889, -20.352, -19.736, -19.087, -18.411, -17.735, -21.978, -21.829,
    -21.613, -21.312, -20.894, -20.356, -19.739, -19.089, -18.414, -17.737,
    -21.984, -21.834, -21.618, -21.317, -20.899, -20.360, -19.743, -19.092,
    -18.416, -17.739, -21.989, -21.839, -21.623, -21.322, -20.904, -20.364,
    -19.746, -19.094, -18.418, -17.742, -21.994, -21.844, -21.629, -21.327,
    -20.908, -20.368, -19.749, -19.097, -18.420, -17.744, -21.999, -21.849,
    -21.634, -21.332, -20.913, -20.372, -19.753, -19.099, -18.422, -17.746,
    -22.004, -21.854, -21.639, -21.337, -20.918, -20.376, -19.756, -19.102,
    -18.424, -17.748, -22.010, -21.860, -21.644, -21.342, -20.923, -20.381,
    -19.759, -19.104, -18.426, -17.750, -22.015, -21.865, -21.649, -21.347,
    -20.928, -20.385, -19.762, -19.106, -18.428, -17.753, -22.020, -21.870,
    -21.654, -21.352, -20.932, -20.389, -19.766, -19.109, -18.430, -17.755,
    -22.025, -21.875, -21.659, -21.357, -20.937, -20.393, -19.769, -19.111,
    -18.432, -17.757, -22.031, -21.880, -21.664, -21.362, -20.942, -20.397,
    -19.772, -19.114, -18.434, -17.759, -22.036, -21.885, -21.669, -21.367,
    -20.946, -20.401, -19.775, -19.116, -18.436, -17.761, -22.041, -21.890,
    -21.675, -21.372, -20.951, -20.405, -19.778, -19.119, -18.438, -17.763,
    -22.046, -21.896, -21.680, -21.377, -20.956, -20.409, -19.782, -19.121,
    -18.440, -17.766, -22.052, -21.901, -21.685, -21.382, -20.961, -20.413,
    -19.785, -19.123, -18.442, -17.768, -22.057, -21.906, -21.690, -21.387,
    -20.965, -20.417, -19.788, -19.126, -18.444, -17.770, -22.062, -21.911,
    -21.695, -21.392, -20.970, -20.421, -19.791, -19.128, -18.446, -17.772};

const double CIA_Orton::H2He_e[][10] = {
    -30.657, -30.746, -30.824, -30.907, -30.995, -31.083, -31.169, -31.253,
    -31.334, -31.415, -28.825, -28.914, -28.992, -29.074, -29.162, -29.250,
    -29.337, -29.421, -29.502, -29.582, -27.438, -27.527, -27.605, -27.688,
    -27.776, -27.864, -27.951, -28.034, -28.116, -28.196, -26.052, -26.141,
    -26.219, -26.302, -26.390, -26.477, -26.564, -26.648, -26.729, -26.810,
    -25.241, -25.330, -25.408, -25.491, -25.579, -25.667, -25.753, -25.837,
    -25.918, -25.999, -24.666, -24.755, -24.833, -24.915, -25.003, -25.091,
    -25.178, -25.262, -25.343, -25.423, -24.220, -24.309, -24.386, -24.469,
    -24.557, -24.645, -24.732, -24.815, -24.897, -24.977, -23.855, -23.944,
    -24.022, -24.104, -24.193, -24.280, -24.367, -24.451, -24.532, -24.613,
    -23.547, -23.636, -23.714, -23.796, -23.884, -23.972, -24.059, -24.142,
    -24.224, -24.304, -23.280, -23.369, -23.447, -23.529, -23.617, -23.705,
    -23.792, -23.875, -23.957, -24.037, -23.044, -23.133, -23.211, -23.294,
    -23.382, -23.469, -23.556, -23.640, -23.721, -23.802, -22.834, -22.923,
    -23.000, -23.083, -23.171, -23.259, -23.345, -23.429, -23.510, -23.591,
    -22.643, -22.732, -22.810, -22.892, -22.980, -23.068, -23.155, -23.238,
    -23.320, -23.400, -22.469, -22.558, -22.636, -22.718, -22.806, -22.894,
    -22.981, -23.064, -23.146, -23.226, -22.309, -22.398, -22.476, -22.558,
    -22.646, -22.734, -22.821, -22.904, -22.986, -23.066, -22.161, -22.250,
    -22.328, -22.410, -22.498, -22.586, -22.672, -22.756, -22.837, -22.918,
    -22.023, -22.112, -22.190, -22.272, -22.360, -22.448, -22.535, -22.618,
    -22.699, -22.780, -21.894, -21.983, -22.061, -22.143, -22.231, -22.319,
    -22.405, -22.489, -22.570, -22.651, -21.773, -21.862, -21.939, -22.022,
    -22.110, -22.198, -22.284, -22.368, -22.449, -22.530, -21.659, -21.747,
    -21.825, -21.908, -21.996, -22.083, -22.170, -22.254, -22.335, -22.415,
    -21.551, -21.639, -21.717, -21.799, -21.887, -21.975, -22.062, -22.145,
    -22.227, -22.307, -21.448, -21.537, -21.614, -21.697, -21.785, -21.873,
    -21.959, -22.043, -22.124, -22.205, -21.002, -21.091, -21.169, -21.251,
    -21.339, -21.426, -21.513, -21.597, -21.678, -21.758, -20.638, -20.727,
    -20.804, -20.887, -20.974, -21.062, -21.149, -21.232, -21.313, -21.394,
    -20.331, -20.419, -20.497, -20.579, -20.666, -20.754, -20.840, -20.924,
    -21.005, -21.086, -20.065, -20.153, -20.230, -20.312, -20.400, -20.487,
    -20.574, -20.657, -20.738, -20.819, -19.830, -19.918, -19.995, -20.077,
    -20.165, -20.252, -20.338, -20.422, -20.503, -20.583, -19.621, -19.708,
    -19.785, -19.867, -19.954, -20.041, -20.128, -20.211, -20.292, -20.373,
    -19.431, -19.519, -19.595, -19.677, -19.764, -19.851, -19.937, -20.021,
    -20.102, -20.182, -19.259, -19.346, -19.422, -19.503, -19.591, -19.677,
    -19.764, -19.847, -19.928, -20.008, -19.100, -19.187, -19.263, -19.344,
    -19.431, -19.518, -19.604, -19.687, -19.768, -19.848, -18.954, -19.040,
    -19.116, -19.197, -19.283, -19.370, -19.456, -19.539, -19.620, -19.700,
    -18.817, -18.903, -18.979, -19.060, -19.146, -19.233, -19.318, -19.401,
    -19.482, -19.562, -18.690, -18.776, -18.851, -18.931, -19.018, -19.104,
    -19.190, -19.273, -19.353, -19.433, -18.571, -18.656, -18.731, -18.811,
    -18.897, -18.983, -19.069, -19.152, -19.232, -19.312, -18.459, -18.544,
    -18.618, -18.698, -18.784, -18.870, -18.955, -19.038, -19.118, -19.198,
    -18.353, -18.437, -18.511, -18.591, -18.676, -18.762, -18.847, -18.930,
    -19.010, -19.090, -18.253, -18.336, -18.410, -18.489, -18.575, -18.660,
    -18.745, -18.828, -18.908, -18.988, -18.067, -18.150, -18.223, -18.301,
    -18.386, -18.471, -18.556, -18.638, -18.718, -18.798, -17.898, -17.980,
    -18.052, -18.130, -18.214, -18.298, -18.383, -18.465, -18.545, -18.624,
    -17.744, -17.825, -17.896, -17.972, -18.056, -18.140, -18.224, -18.306,
    -18.385, -18.465, -17.602, -17.681, -17.751, -17.827, -17.910, -17.994,
    -18.077, -18.158, -18.238, -18.317, -17.471, -17.549, -17.618, -17.693,
    -17.775, -17.857, -17.941, -18.022, -18.101, -18.180, -17.349, -17.425,
    -17.493, -17.567, -17.648, -17.730, -17.813, -17.894, -17.972, -18.051,
    -17.235, -17.310, -17.377, -17.450, -17.530, -17.611, -17.693, -17.774,
    -17.852, -17.931, -17.129, -17.202, -17.267, -17.339, -17.418, -17.499,
    -17.581, -17.661, -17.739, -17.817, -17.029, -17.101, -17.165, -17.235,
    -17.313, -17.394, -17.475, -17.554, -17.632, -17.710, -16.935, -17.005,
    -17.067, -17.137, -17.214, -17.294, -17.374, -17.453, -17.530, -17.608,
    -16.846, -16.914, -16.976, -17.044, -17.120, -17.199, -17.278, -17.357,
    -17.434, -17.511, -16.763, -16.829, -16.889, -16.955, -17.031, -17.108,
    -17.187, -17.265, -17.342, -17.419, -16.683, -16.748, -16.806, -16.871,
    -16.946, -17.022, -17.101, -17.178, -17.254, -17.331, -16.608, -16.671,
    -16.727, -16.791, -16.864, -16.940, -17.018, -17.095, -17.171, -17.247,
    -16.537, -16.597, -16.652, -16.715, -16.787, -16.862, -16.939, -17.015,
    -17.090, -17.166, -16.469, -16.527, -16.581, -16.642, -16.713, -16.787,
    -16.863, -16.938, -17.013, -17.089, -16.404, -16.460, -16.512, -16.572,
    -16.642, -16.714, -16.790, -16.865, -16.939, -17.015, -16.342, -16.397,
    -16.447, -16.505, -16.573, -16.645, -16.720, -16.794, -16.868, -16.943,
    -16.283, -16.336, -16.384, -16.441, -16.508, -16.579, -16.652, -16.726,
    -16.800, -16.874, -16.227, -16.277, -16.324, -16.379, -16.445, -16.515,
    -16.587, -16.660, -16.733, -16.808, -16.174, -16.221, -16.266, -16.320,
    -16.384, -16.453, -16.525, -16.597, -16.670, -16.743, -16.122, -16.168,
    -16.211, -16.263, -16.326, -16.393, -16.464, -16.536, -16.608, -16.681,
    -16.073, -16.116, -16.157, -16.208, -16.269, -16.336, -16.406, -16.477,
    -16.548, -16.621, -16.026, -16.067, -16.106, -16.155, -16.215, -16.280,
    -16.349, -16.420, -16.490, -16.563, -15.980, -16.019, -16.056, -16.104,
    -16.162, -16.226, -16.295, -16.364, -16.434, -16.506, -15.937, -15.973,
    -16.009, -16.054, -16.112, -16.174, -16.242, -16.310, -16.380, -16.451,
    -15.895, -15.929, -15.963, -16.007, -16.062, -16.124, -16.190, -16.258,
    -16.327, -16.398, -15.855, -15.887, -15.918, -15.961, -16.015, -16.075,
    -16.141, -16.208, -16.276, -16.346, -15.817, -15.846, -15.876, -15.916,
    -15.969, -16.028, -16.092, -16.158, -16.226, -16.296, -15.780, -15.806,
    -15.834, -15.873, -15.924, -15.982, -16.045, -16.111, -16.178, -16.247,
    -15.744, -15.768, -15.794, -15.831, -15.881, -15.937, -16.000, -16.064,
    -16.130, -16.199, -15.710, -15.732, -15.756, -15.791, -15.839, -15.894,
    -15.955, -16.019, -16.084, -16.153, -15.677, -15.696, -15.718, -15.752,
    -15.798, -15.852, -15.912, -15.975, -16.040, -16.107, -15.645, -15.662,
    -15.682, -15.714, -15.759, -15.811, -15.870, -15.932, -15.996, -16.063,
    -15.614, -15.629, -15.647, -15.677, -15.720, -15.771, -15.829, -15.890,
    -15.954, -16.020, -15.585, -15.597, -15.613, -15.641, -15.683, -15.733,
    -15.789, -15.850, -15.912, -15.978, -15.557, -15.566, -15.580, -15.607,
    -15.647, -15.695, -15.751, -15.810, -15.872, -15.937, -15.529, -15.536,
    -15.548, -15.573, -15.611, -15.658, -15.713, -15.771, -15.832, -15.897,
    -15.503, -15.508, -15.517, -15.540, -15.577, -15.623, -15.676, -15.733,
    -15.794, -15.857, -15.477, -15.480, -15.487, -15.508, -15.544, -15.588,
    -15.640, -15.696, -15.756, -15.819, -15.453, -15.453, -15.458, -15.478,
    -15.511, -15.554, -15.605, -15.660, -15.719, -15.782, -15.429, -15.426,
    -15.430, -15.448, -15.480, -15.521, -15.571, -15.625, -15.683, -15.745,
    -15.406, -15.401, -15.403, -15.418, -15.449, -15.489, -15.537, -15.590,
    -15.648, -15.709, -15.384, -15.377, -15.376, -15.390, -15.419, -15.457,
    -15.505, -15.557, -15.613, -15.674, -15.362, -15.353, -15.351, -15.362,
    -15.390, -15.427, -15.473, -15.524, -15.579, -15.639, -15.342, -15.330,
    -15.326, -15.336, -15.361, -15.397, -15.441, -15.492, -15.546, -15.605,
    -15.322, -15.308, -15.301, -15.309, -15.333, -15.367, -15.411, -15.460,
    -15.514, -15.572, -15.302, -15.286, -15.278, -15.284, -15.306, -15.339,
    -15.381, -15.429, -15.482, -15.540, -15.284, -15.265, -15.255, -15.259,
    -15.280, -15.311, -15.352, -15.399, -15.451, -15.508, -15.266, -15.245,
    -15.232, -15.235, -15.254, -15.284, -15.323, -15.369, -15.420, -15.477,
    -15.248, -15.225, -15.211, -15.212, -15.229, -15.257, -15.295, -15.340,
    -15.390, -15.446, -15.232, -15.206, -15.190, -15.189, -15.204, -15.231,
    -15.268, -15.312, -15.361, -15.416, -15.215, -15.188, -15.169, -15.167,
    -15.180, -15.206, -15.241, -15.284, -15.332, -15.386, -15.200, -15.170,
    -15.149, -15.145, -15.157, -15.181, -15.215, -15.257, -15.304, -15.357,
    -15.185, -15.152, -15.130, -15.124, -15.134, -15.156, -15.190, -15.230,
    -15.276, -15.329, -15.170, -15.135, -15.111, -15.103, -15.112, -15.133,
    -15.165, -15.204, -15.249, -15.301, -15.156, -15.119, -15.093, -15.083,
    -15.090, -15.109, -15.140, -15.178, -15.223, -15.274, -15.142, -15.103,
    -15.075, -15.063, -15.069, -15.087, -15.116, -15.153, -15.196, -15.247,
    -15.129, -15.088, -15.058, -15.044, -15.048, -15.064, -15.092, -15.128,
    -15.171, -15.220, -15.116, -15.073, -15.041, -15.026, -15.028, -15.043,
    -15.069, -15.104, -15.146, -15.194, -15.103, -15.058, -15.025, -15.007,
    -15.008, -15.021, -15.046, -15.080, -15.121, -15.168, -15.091, -15.044,
    -15.009, -14.990, -14.989, -15.000, -15.024, -15.057, -15.096, -15.143,
    -15.080, -15.031, -14.993, -14.972, -14.970, -14.980, -15.002, -15.034,
    -15.072, -15.119, -15.069, -15.018, -14.978, -14.956, -14.951, -14.960,
    -14.981, -15.011, -15.049, -15.094, -15.058, -15.005, -14.963, -14.939,
    -14.933, -14.940, -14.960, -14.989, -15.026, -15.070, -15.047, -14.992,
    -14.949, -14.923, -14.915, -14.921, -14.940, -14.967, -15.003, -15.047,
    -15.037, -14.980, -14.935, -14.907, -14.898, -14.902, -14.919, -14.946,
    -14.981, -15.023, -15.027, -14.969, -14.922, -14.892, -14.881, -14.884,
    -14.900, -14.925, -14.959, -15.001, -15.018, -14.957, -14.909, -14.877,
    -14.865, -14.866, -14.880, -14.905, -14.937, -14.978, -15.009, -14.946,
    -14.896, -14.863, -14.848, -14.848, -14.861, -14.884, -14.916, -14.956,
    -15.000, -14.936, -14.883, -14.849, -14.833, -14.831, -14.843, -14.864,
    -14.895, -14.934, -14.991, -14.925, -14.871, -14.835, -14.817, -14.814,
    -14.824, -14.845, -14.875, -14.913, -14.983, -14.915, -14.859, -14.821,
    -14.802, -14.797, -14.806, -14.826, -14.854, -14.892, -14.975, -14.906,
    -14.848, -14.808, -14.787, -14.781, -14.789, -14.807, -14.834, -14.871,
    -14.967, -14.896, -14.837, -14.795, -14.773, -14.765, -14.771, -14.788,
    -14.815, -14.850, -14.960, -14.887, -14.826, -14.783, -14.759, -14.749,
    -14.754, -14.770, -14.796, -14.830, -14.953, -14.878, -14.815, -14.770,
    -14.745, -14.734, -14.738, -14.752, -14.777, -14.810, -14.946, -14.869,
    -14.805, -14.759, -14.731, -14.719, -14.721, -14.735, -14.758, -14.791,
    -14.939, -14.861, -14.795, -14.747, -14.718, -14.704, -14.705, -14.717,
    -14.740, -14.771, -14.933, -14.853, -14.785, -14.736, -14.705, -14.690,
    -14.689, -14.700, -14.721, -14.752, -14.927, -14.845, -14.776, -14.724,
    -14.692, -14.676, -14.674, -14.684, -14.704, -14.734, -14.921, -14.838,
    -14.767, -14.714, -14.680, -14.662, -14.659, -14.667, -14.686, -14.715,
    -14.915, -14.830, -14.758, -14.703, -14.668, -14.648, -14.644, -14.651,
    -14.669, -14.697, -14.909, -14.823, -14.749, -14.693, -14.656, -14.635,
    -14.629, -14.635, -14.652, -14.679, -14.904, -14.816, -14.741, -14.683,
    -14.644, -14.622, -14.614, -14.619, -14.635, -14.661, -14.899, -14.810,
    -14.732, -14.673, -14.633, -14.609, -14.600, -14.604, -14.619, -14.644,
    -14.894, -14.803, -14.724, -14.663, -14.622, -14.596, -14.586, -14.589,
    -14.602, -14.627, -14.889, -14.797, -14.717, -14.654, -14.611, -14.584,
    -14.573, -14.574, -14.586, -14.610, -14.884, -14.791, -14.709, -14.645,
    -14.600, -14.572, -14.559, -14.559, -14.570, -14.593, -14.880, -14.785,
    -14.702, -14.636, -14.590, -14.560, -14.546, -14.545, -14.555, -14.576,
    -14.876, -14.780, -14.695, -14.627, -14.580, -14.548, -14.533, -14.530,
    -14.540, -14.560, -14.871, -14.774, -14.688, -14.619, -14.570, -14.537,
    -14.520, -14.516, -14.525, -14.544, -14.868, -14.769, -14.681, -14.611,
    -14.560, -14.526, -14.508, -14.503, -14.510, -14.528, -14.864, -14.764,
    -14.675, -14.603, -14.550, -14.515, -14.495, -14.489, -14.495, -14.513,
    -14.860, -14.759, -14.668, -14.595, -14.541, -14.504, -14.483, -14.476,
    -14.481, -14.497, -14.857, -14.754, -14.662, -14.587, -14.532, -14.493,
    -14.471, -14.463, -14.466, -14.482, -14.853, -14.750, -14.656, -14.580,
    -14.523, -14.483, -14.460, -14.450, -14.452, -14.467, -14.850, -14.745,
    -14.650, -14.573, -14.514, -14.473, -14.448, -14.437, -14.438, -14.452,
    -14.847, -14.741, -14.645, -14.565, -14.506, -14.463, -14.437, -14.425,
    -14.425, -14.438, -14.844, -14.737, -14.639, -14.559, -14.497, -14.453,
    -14.426, -14.412, -14.411, -14.423, -14.842, -14.733, -14.634, -14.552,
    -14.489, -14.443, -14.415, -14.400, -14.398, -14.409, -14.839, -14.729,
    -14.629, -14.545, -14.481, -14.434, -14.404, -14.388, -14.385, -14.395,
    -14.836, -14.726, -14.624, -14.539, -14.473, -14.425, -14.393, -14.376,
    -14.372, -14.381, -14.834, -14.722, -14.619, -14.533, -14.466, -14.416,
    -14.383, -14.365, -14.359, -14.367, -14.832, -14.719, -14.615, -14.527,
    -14.458, -14.407, -14.373, -14.353, -14.347, -14.354, -14.830, -14.716,
    -14.610, -14.521, -14.451, -14.398, -14.363, -14.342, -14.335, -14.340,
    -14.828, -14.713, -14.606, -14.515, -14.444, -14.390, -14.353, -14.331,
    -14.322, -14.327, -14.826, -14.710, -14.602, -14.510, -14.437, -14.381,
    -14.343, -14.320, -14.310, -14.314, -14.824, -14.707, -14.598, -14.504,
    -14.430, -14.373, -14.334, -14.309, -14.299, -14.301, -14.822, -14.704,
    -14.594, -14.499, -14.424, -14.365, -14.325, -14.299, -14.287, -14.289,
    -14.821, -14.702, -14.590, -14.494, -14.417, -14.357, -14.315, -14.288,
    -14.275, -14.276, -14.819, -14.699, -14.586, -14.489, -14.411, -14.349,
    -14.306, -14.278, -14.264, -14.264, -14.818, -14.697, -14.583, -14.484,
    -14.404, -14.342, -14.297, -14.268, -14.253, -14.252, -14.817, -14.695,
    -14.579, -14.479, -14.398, -14.334, -14.289, -14.258, -14.242, -14.240,
    -14.816, -14.693, -14.576, -14.475, -14.392, -14.327, -14.280, -14.248,
    -14.231, -14.228, -14.815, -14.691, -14.573, -14.470, -14.387, -14.320,
    -14.272, -14.239, -14.220, -14.216, -14.814, -14.689, -14.570, -14.466,
    -14.381, -14.313, -14.263, -14.229, -14.210, -14.204, -14.813, -14.687,
    -14.567, -14.462, -14.375, -14.306, -14.255, -14.220, -14.199, -14.193,
    -14.812, -14.685, -14.564, -14.458, -14.370, -14.299, -14.247, -14.211,
    -14.189, -14.181, -14.811, -14.684, -14.562, -14.454, -14.365, -14.293,
    -14.239, -14.202, -14.179, -14.170, -14.811, -14.682, -14.559, -14.450,
    -14.360, -14.286, -14.232, -14.193, -14.169, -14.159, -14.810, -14.681,
    -14.557, -14.446, -14.355, -14.280, -14.224, -14.184, -14.159, -14.148,
    -14.810, -14.680, -14.554, -14.443, -14.350, -14.274, -14.217, -14.175,
    -14.149, -14.137, -14.809, -14.678, -14.552, -14.439, -14.345, -14.268,
    -14.209, -14.167, -14.139, -14.127, -14.809, -14.677, -14.550, -14.436,
    -14.340, -14.262, -14.202, -14.158, -14.130, -14.116, -14.809, -14.676,
    -14.548, -14.433, -14.336, -14.256, -14.195, -14.150, -14.120, -14.106,
    -14.808, -14.675, -14.546, -14.430, -14.331, -14.250, -14.188, -14.142,
    -14.111, -14.096, -14.808, -14.674, -14.544, -14.427, -14.327, -14.245,
    -14.181, -14.134, -14.102, -14.085, -14.808, -14.674, -14.542, -14.424,
    -14.323, -14.239, -14.174, -14.126, -14.093, -14.075, -14.808, -14.673,
    -14.541, -14.421, -14.319, -14.234, -14.168, -14.118, -14.084, -14.066,
    -14.809, -14.672, -14.539, -14.418, -14.315, -14.229, -14.161, -14.110,
    -14.075, -14.056, -14.809, -14.672, -14.538, -14.416, -14.311, -14.223,
    -14.155, -14.103, -14.067, -14.046, -14.809, -14.671, -14.536, -14.413,
    -14.307, -14.218, -14.148, -14.095, -14.058, -14.037, -14.809, -14.671,
    -14.535, -14.411, -14.304, -14.213, -14.142, -14.088, -14.050, -14.027,
    -14.810, -14.671, -14.534, -14.408, -14.300, -14.209, -14.136, -14.081,
    -14.041, -14.018, -14.810, -14.671, -14.533, -14.406, -14.297, -14.204,
    -14.130, -14.074, -14.033, -14.009, -14.811, -14.670, -14.532, -14.404,
    -14.293, -14.199, -14.124, -14.067, -14.025, -13.999, -14.811, -14.670,
    -14.531, -14.402, -14.290, -14.195, -14.119, -14.060, -14.017, -13.990,
    -14.812, -14.670, -14.530, -14.400, -14.287, -14.190, -14.113, -14.053,
    -14.009, -13.982, -14.813, -14.670, -14.529, -14.398, -14.284, -14.186,
    -14.107, -14.046, -14.001, -13.973, -14.813, -14.671, -14.528, -14.396,
    -14.281, -14.182, -14.102, -14.040, -13.994, -13.964, -14.814, -14.671,
    -14.528, -14.394, -14.278, -14.178, -14.097, -14.033, -13.986, -13.955,
    -14.815, -14.671, -14.527, -14.393, -14.275, -14.173, -14.091, -14.027,
    -13.978, -13.947, -14.816, -14.671, -14.526, -14.391, -14.272, -14.170,
    -14.086, -14.020, -13.971, -13.939, -14.817, -14.672, -14.526, -14.390,
    -14.269, -14.166, -14.081, -14.014, -13.964, -13.930, -14.818, -14.672,
    -14.526, -14.388, -14.267, -14.162, -14.076, -14.008, -13.957, -13.922,
    -14.819, -14.673, -14.525, -14.387, -14.264, -14.158, -14.071, -14.002,
    -13.949, -13.914, -14.820, -14.673, -14.525, -14.386, -14.262, -14.155,
    -14.067, -13.996, -13.942, -13.906, -14.821, -14.674, -14.525, -14.384,
    -14.260, -14.151, -14.062, -13.990, -13.936, -13.898, -14.823, -14.675,
    -14.525, -14.383, -14.257, -14.148, -14.057, -13.984, -13.929, -13.890,
    -14.824, -14.675, -14.525, -14.382, -14.255, -14.144, -14.053, -13.979,
    -13.922, -13.883, -14.825, -14.676, -14.525, -14.381, -14.253, -14.141,
    -14.048, -13.973, -13.915, -13.875, -14.827, -14.677, -14.525, -14.380,
    -14.251, -14.138, -14.044, -13.968, -13.909, -13.867, -14.828, -14.678,
    -14.525, -14.380, -14.249, -14.135, -14.039, -13.962, -13.902, -13.860,
    -14.829, -14.679, -14.525, -14.379, -14.247, -14.131, -14.035, -13.957,
    -13.896, -13.852, -14.831, -14.680, -14.525, -14.378, -14.245, -14.128,
    -14.031, -13.952, -13.890, -13.845, -14.833, -14.681, -14.526, -14.377,
    -14.243, -14.126, -14.027, -13.946, -13.883, -13.838, -14.834, -14.682,
    -14.526, -14.377, -14.242, -14.123, -14.023, -13.941, -13.877, -13.831,
    -14.836, -14.683, -14.526, -14.376, -14.240, -14.120, -14.019, -13.936,
    -13.871, -13.824, -14.837, -14.684, -14.527, -14.376, -14.239, -14.117,
    -14.015, -13.931, -13.865, -13.817, -14.839, -14.685, -14.527, -14.375,
    -14.237, -14.115, -14.012, -13.927, -13.859, -13.810, -14.841, -14.687,
    -14.528, -14.375, -14.236, -14.112, -14.008, -13.922, -13.854, -13.803,
    -14.843, -14.688, -14.529, -14.375, -14.234, -14.110, -14.004, -13.917,
    -13.848, -13.796, -14.844, -14.689, -14.529, -14.374, -14.233, -14.107,
    -14.001, -13.912, -13.842, -13.790, -14.846, -14.691, -14.530, -14.374,
    -14.232, -14.105, -13.997, -13.908, -13.837, -13.783, -14.848, -14.692,
    -14.531, -14.374, -14.231, -14.102, -13.994, -13.903, -13.831, -13.777,
    -14.850, -14.694, -14.531, -14.374, -14.229, -14.100, -13.990, -13.899,
    -13.826, -13.770, -14.852, -14.695, -14.532, -14.374, -14.228, -14.098,
    -13.987, -13.895, -13.820, -13.764, -14.854, -14.697, -14.533, -14.374,
    -14.227, -14.096, -13.984, -13.890, -13.815, -13.758, -14.856, -14.698,
    -14.534, -14.374, -14.226, -14.094, -13.981, -13.886, -13.810, -13.752,
    -14.858, -14.700, -14.535, -14.374, -14.226, -14.092, -13.978, -13.882,
    -13.804, -13.745, -14.860, -14.702, -14.536, -14.374, -14.225, -14.090,
    -13.975, -13.878, -13.799, -13.739, -14.862, -14.703, -14.537, -14.375,
    -14.224, -14.088, -13.972, -13.874, -13.794, -13.733, -14.865, -14.705,
    -14.538, -14.375, -14.223, -14.086, -13.969, -13.870, -13.789, -13.727,
    -14.867, -14.707, -14.540, -14.375, -14.222, -14.085, -13.966, -13.866,
    -13.784, -13.722, -14.869, -14.709, -14.541, -14.375, -14.222, -14.083,
    -13.963, -13.862, -13.780, -13.716, -14.871, -14.711, -14.542, -14.376,
    -14.221, -14.081, -13.960, -13.858, -13.775, -13.710, -14.874, -14.713,
    -14.543, -14.376, -14.221, -14.080, -13.958, -13.855, -13.770, -13.704,
    -14.876, -14.714, -14.545, -14.377, -14.220, -14.078, -13.955, -13.851,
    -13.766, -13.699, -14.878, -14.716, -14.546, -14.377, -14.220, -14.076,
    -13.953, -13.847, -13.761, -13.693, -14.881, -14.718, -14.547, -14.378,
    -14.219, -14.075, -13.950, -13.844, -13.756, -13.688, -14.883, -14.720,
    -14.549, -14.378, -14.219, -14.074, -13.948, -13.840, -13.752, -13.682,
    -14.885, -14.722, -14.550, -14.379, -14.219, -14.072, -13.945, -13.837,
    -13.748, -13.677, -14.888, -14.725, -14.552, -14.380, -14.218, -14.071,
    -13.943, -13.834, -13.743, -13.672, -14.890, -14.727, -14.553, -14.380,
    -14.218, -14.070, -13.941, -13.830, -13.739, -13.667, -14.893, -14.729,
    -14.555, -14.381, -14.218, -14.068, -13.938, -13.827, -13.735, -13.662,
    -14.895, -14.731, -14.556, -14.382, -14.218, -14.067, -13.936, -13.824,
    -13.731, -13.656, -14.898, -14.733, -14.558, -14.383, -14.218, -14.066,
    -13.934, -13.821, -13.726, -13.651, -14.901, -14.735, -14.559, -14.384,
    -14.217, -14.065, -13.932, -13.818, -13.722, -13.646, -14.903, -14.738,
    -14.561, -14.384, -14.217, -14.064, -13.930, -13.815, -13.718, -13.642,
    -14.906, -14.740, -14.563, -14.385, -14.217, -14.063, -13.928, -13.812,
    -13.714, -13.637, -14.909, -14.742, -14.564, -14.386, -14.217, -14.062,
    -13.926, -13.809, -13.711, -13.632, -14.911, -14.744, -14.566, -14.387,
    -14.218, -14.061, -13.924, -13.806, -13.707, -13.627, -14.914, -14.747,
    -14.568, -14.388, -14.218, -14.060, -13.922, -13.803, -13.703, -13.622,
    -14.917, -14.749, -14.570, -14.389, -14.218, -14.059, -13.920, -13.800,
    -13.699, -13.618, -14.919, -14.752, -14.571, -14.390, -14.218, -14.059,
    -13.918, -13.797, -13.696, -13.613, -14.922, -14.754, -14.573, -14.391,
    -14.218, -14.058, -13.917, -13.795, -13.692, -13.609, -14.925, -14.756,
    -14.575, -14.393, -14.218, -14.057, -13.915, -13.792, -13.688, -13.604,
    -14.928, -14.759, -14.577, -14.394, -14.219, -14.056, -13.913, -13.789,
    -13.685, -13.600, -14.931, -14.761, -14.579, -14.395, -14.219, -14.056,
    -13.912, -13.787, -13.681, -13.595, -14.933, -14.764, -14.581, -14.396,
    -14.219, -14.055, -13.910, -13.784, -13.678, -13.591, -14.936, -14.766,
    -14.583, -14.397, -14.219, -14.055, -13.909, -13.782, -13.674, -13.587,
    -14.939, -14.769, -14.585, -14.398, -14.220, -14.054, -13.907, -13.779,
    -13.671, -13.582, -14.942, -14.771, -14.587, -14.400, -14.220, -14.053,
    -13.906, -13.777, -13.668, -13.578, -14.945, -14.774, -14.589, -14.401,
    -14.221, -14.053, -13.904, -13.775, -13.665, -13.574, -14.948, -14.776,
    -14.591, -14.402, -14.221, -14.052, -13.903, -13.772, -13.661, -13.570,
    -14.951, -14.779, -14.593, -14.403, -14.221, -14.052, -13.901, -13.770,
    -13.658, -13.566, -14.954, -14.781, -14.595, -14.405, -14.222, -14.052,
    -13.900, -13.768, -13.655, -13.562, -14.957, -14.784, -14.597, -14.406,
    -14.222, -14.051, -13.899, -13.766, -13.652, -13.558, -14.960, -14.787,
    -14.599, -14.407, -14.223, -14.051, -13.897, -13.763, -13.649, -13.554,
    -14.963, -14.789, -14.601, -14.409, -14.223, -14.051, -13.896, -13.761,
    -13.646, -13.550, -14.966, -14.792, -14.603, -14.410, -14.224, -14.050,
    -13.895, -13.759, -13.643, -13.547, -14.969, -14.794, -14.605, -14.411,
    -14.225, -14.050, -13.894, -13.757, -13.640, -13.543, -14.972, -14.797,
    -14.607, -14.413, -14.225, -14.050, -13.893, -13.755, -13.637, -13.539,
    -14.975, -14.800, -14.609, -14.414, -14.226, -14.049, -13.892, -13.753,
    -13.634, -13.535, -14.978, -14.802, -14.611, -14.415, -14.226, -14.049,
    -13.891, -13.751, -13.632, -13.532, -14.981, -14.805, -14.613, -14.417,
    -14.227, -14.049, -13.889, -13.749, -13.629, -13.528, -14.984, -14.808,
    -14.615, -14.418, -14.228, -14.049, -13.888, -13.747, -13.626, -13.525,
    -14.987, -14.810, -14.617, -14.420, -14.228, -14.049, -13.887, -13.746,
    -13.623, -13.521, -14.990, -14.813, -14.619, -14.421, -14.229, -14.048,
    -13.887, -13.744, -13.621, -13.518, -14.993, -14.816, -14.621, -14.422,
    -14.230, -14.048, -13.886, -13.742, -13.618, -13.514, -14.996, -14.818,
    -14.623, -14.424, -14.230, -14.048, -13.885, -13.740, -13.616, -13.511,
    -14.999, -14.821, -14.625, -14.425, -14.231, -14.048, -13.884, -13.739,
    -13.613, -13.508, -15.003, -14.823, -14.627, -14.427, -14.232, -14.048,
    -13.883, -13.737, -13.611, -13.504, -15.006, -14.826, -14.629, -14.428,
    -14.232, -14.048, -13.882, -13.735, -13.608, -13.501, -15.009, -14.829,
    -14.631, -14.429, -14.233, -14.048, -13.881, -13.734, -13.606, -13.498,
    -15.012, -14.831, -14.633, -14.431, -14.234, -14.048, -13.880, -13.732,
    -13.603, -13.495, -15.015, -14.834, -14.635, -14.432, -14.234, -14.048,
    -13.880, -13.730, -13.601, -13.491, -15.018, -14.836, -14.637, -14.433,
    -14.235, -14.048, -13.879, -13.729, -13.599, -13.488, -15.021, -14.839,
    -14.639, -14.435, -14.236, -14.048, -13.878, -13.727, -13.596, -13.485,
    -15.024, -14.841, -14.641, -14.436, -14.236, -14.048, -13.877, -13.726,
    -13.594, -13.482, -15.027, -14.844, -14.643, -14.437, -14.237, -14.048,
    -13.877, -13.724, -13.592, -13.479, -15.030, -14.846, -14.645, -14.439,
    -14.238, -14.048, -13.876, -13.723, -13.590, -13.476, -15.033, -14.849,
    -14.647, -14.440, -14.239, -14.048, -13.875, -13.722, -13.588, -13.473,
    -15.036, -14.851, -14.649, -14.441, -14.239, -14.048, -13.875, -13.720,
    -13.585, -13.470, -15.039, -14.854, -14.650, -14.442, -14.240, -14.048,
    -13.874, -13.719, -13.583, -13.468, -15.042, -14.856, -14.652, -14.444,
    -14.240, -14.048, -13.873, -13.718, -13.581, -13.465, -15.045, -14.858,
    -14.654, -14.445, -14.241, -14.048, -13.873, -13.716, -13.579, -13.462,
    -15.048, -14.860, -14.655, -14.446, -14.242, -14.048, -13.872, -13.715,
    -13.577, -13.459, -15.051, -14.863, -14.657, -14.447, -14.242, -14.048,
    -13.872, -13.714, -13.575, -13.457, -15.053, -14.865, -14.659, -14.448,
    -14.243, -14.048, -13.871, -13.713, -13.573, -13.454, -15.056, -14.867,
    -14.660, -14.449, -14.243, -14.048, -13.871, -13.711, -13.571, -13.451,
    -15.059, -14.869, -14.662, -14.450, -14.244, -14.048, -13.870, -13.710,
    -13.570, -13.449, -15.062, -14.871, -14.663, -14.451, -14.245, -14.048,
    -13.870, -13.709, -13.568, -13.446, -15.064, -14.873, -14.664, -14.452,
    -14.245, -14.048, -13.869, -13.708, -13.566, -13.443, -15.067, -14.875,
    -14.666, -14.453, -14.246, -14.049, -13.869, -13.707, -13.564, -13.441,
    -15.070, -14.877, -14.667, -14.454, -14.246, -14.049, -13.868, -13.706,
    -13.562, -13.438, -15.072, -14.878, -14.668, -14.454, -14.246, -14.049,
    -13.868, -13.705, -13.561, -13.436, -15.074, -14.880, -14.669, -14.455,
    -14.247, -14.049, -13.867, -13.704, -13.559, -13.434, -15.077, -14.881,
    -14.670, -14.456, -14.247, -14.049, -13.867, -13.702, -13.557, -13.431,
    -15.079, -14.883, -14.671, -14.456, -14.248, -14.049, -13.866, -13.701,
    -13.555, -13.429, -15.081, -14.884, -14.672, -14.457, -14.248, -14.049,
    -13.866, -13.700, -13.554, -13.426, -15.083, -14.885, -14.672, -14.457,
    -14.248, -14.049, -13.865, -13.699, -13.552, -13.424, -15.085, -14.887,
    -14.673, -14.458, -14.248, -14.049, -13.865, -13.699, -13.551, -13.422,
    -15.087, -14.888, -14.674, -14.458, -14.249, -14.049, -13.864, -13.698,
    -13.549, -13.420, -15.089, -14.888, -14.674, -14.458, -14.249, -14.048,
    -13.864, -13.697, -13.548, -13.417, -15.091, -14.889, -14.674, -14.458,
    -14.249, -14.048, -13.864, -13.696, -13.546, -13.415, -15.092, -14.890,
    -14.674, -14.458, -14.249, -14.048, -13.863, -13.695, -13.544, -13.413,
    -15.094, -14.890, -14.674, -14.458, -14.249, -14.048, -13.863, -13.694,
    -13.543, -13.411, -15.095, -14.891, -14.674, -14.458, -14.249, -14.048,
    -13.862, -13.693, -13.542, -13.409, -15.096, -14.891, -14.674, -14.458,
    -14.249, -14.048, -13.862, -13.692, -13.540, -13.407, -15.097, -14.891,
    -14.674, -14.458, -14.249, -14.048, -13.861, -13.691, -13.539, -13.405,
    -15.098, -14.891, -14.673, -14.458, -14.248, -14.048, -13.861, -13.691,
    -13.537, -13.403, -15.099, -14.890, -14.673, -14.457, -14.248, -14.047,
    -13.861, -13.690, -13.536, -13.401, -15.099, -14.890, -14.672, -14.457,
    -14.248, -14.047, -13.860, -13.689, -13.535, -13.399, -15.099, -14.889,
    -14.671, -14.456, -14.247, -14.047, -13.860, -13.688, -13.533, -13.397,
    -15.099, -14.888, -14.670, -14.455, -14.247, -14.047, -13.859, -13.687,
    -13.532, -13.395, -15.099, -14.887, -14.668, -14.454, -14.247, -14.046,
    -13.859, -13.687, -13.531, -13.393, -15.098, -14.885, -14.667, -14.453,
    -14.246, -14.046, -13.859, -13.686, -13.529, -13.391, -15.098, -14.883,
    -14.665, -14.452, -14.245, -14.046, -13.858, -13.685, -13.528, -13.389,
    -15.097, -14.881, -14.663, -14.451, -14.245, -14.045, -13.858, -13.684,
    -13.527, -13.387, -15.095, -14.879, -14.661, -14.449, -14.244, -14.045,
    -13.857, -13.684, -13.526, -13.385, -15.094, -14.876, -14.659, -14.448,
    -14.243, -14.044, -13.857, -13.683, -13.525, -13.384, -15.092, -14.874,
    -14.656, -14.446, -14.242, -14.044, -13.857, -13.682, -13.524, -13.382,
    -15.089, -14.870, -14.653, -14.444, -14.241, -14.043, -13.856, -13.682,
    -13.522, -13.380, -15.087, -14.867, -14.650, -14.442, -14.240, -14.043,
    -13.856, -13.681, -13.521, -13.378, -15.084, -14.863, -14.647, -14.440,
    -14.239, -14.042, -13.855, -13.680, -13.520, -13.377, -15.080, -14.859,
    -14.643, -14.438, -14.238, -14.042, -13.855, -13.680, -13.519, -13.375,
    -15.076, -14.854, -14.640, -14.435, -14.237, -14.041, -13.854, -13.679,
    -13.518, -13.374, -15.072, -14.849, -14.636, -14.433, -14.235, -14.040,
    -13.854, -13.678, -13.517, -13.372, -15.067, -14.844, -14.631, -14.430,
    -14.234, -14.040, -13.854, -13.678, -13.516, -13.370, -15.062, -14.838,
    -14.627, -14.427, -14.232, -14.039, -13.853, -13.677, -13.515, -13.369,
    -15.056, -14.832, -14.622, -14.424, -14.230, -14.038, -13.853, -13.677,
    -13.514, -13.367, -15.049, -14.825, -14.616, -14.420, -14.229, -14.037,
    -13.852, -13.676, -13.513, -13.366, -15.042, -14.818, -14.611, -14.417,
    -14.227, -14.037, -13.852, -13.676, -13.512, -13.364, -15.035, -14.810,
    -14.605, -14.413, -14.225, -14.036, -13.851, -13.675, -13.511, -13.363,
    -15.027, -14.802, -14.599, -14.409, -14.223, -14.035, -13.851, -13.674,
    -13.510, -13.361, -15.018, -14.794, -14.592, -14.405, -14.221, -14.034,
    -13.850, -13.674, -13.509, -13.360, -15.008, -14.785, -14.585, -14.401,
    -14.219, -14.033, -13.850, -13.673, -13.508, -13.358, -14.998, -14.775,
    -14.578, -14.397, -14.216, -14.032, -13.849, -13.673, -13.508, -13.357,
    -14.987, -14.765, -14.571, -14.392, -14.214, -14.031, -13.849, -13.672,
    -13.507, -13.356, -14.975, -14.754, -14.563, -14.387, -14.212, -14.030,
    -13.848, -13.672, -13.506, -13.354, -14.962, -14.743, -14.555, -14.382,
    -14.209, -14.029, -13.848, -13.671, -13.505, -13.353, -14.949, -14.731,
    -14.546, -14.377, -14.206, -14.028, -13.847, -13.671, -13.504, -13.352,
    -14.935, -14.719, -14.537, -14.372, -14.204, -14.026, -13.847, -13.670,
    -13.503, -13.350, -14.920, -14.706, -14.528, -14.366, -14.201, -14.025,
    -13.846, -13.670, -13.503, -13.349, -14.904, -14.692, -14.518, -14.360,
    -14.198, -14.024, -13.846, -13.670, -13.502, -13.348, -14.887, -14.678,
    -14.508, -14.354, -14.195, -14.023, -13.845, -13.669, -13.501, -13.347,
    -14.869, -14.663, -14.498, -14.348, -14.192, -14.021, -13.845, -13.669,
    -13.501, -13.345, -14.851, -14.648, -14.487, -14.342, -14.188, -14.020,
    -13.844, -13.668, -13.500, -13.344, -14.831, -14.632, -14.476, -14.335,
    -14.185, -14.019, -13.843, -13.668, -13.499, -13.343, -14.811, -14.615,
    -14.465, -14.329, -14.182, -14.017, -13.843, -13.667, -13.498, -13.342,
    -14.789, -14.598, -14.453, -14.322, -14.178, -14.016, -13.842, -13.667,
    -13.498, -13.341, -14.767, -14.581, -14.441, -14.315, -14.175, -14.014,
    -13.842, -13.667, -13.497, -13.340, -14.744, -14.562, -14.429, -14.307,
    -14.171, -14.013, -13.841, -13.666, -13.497, -13.338, -14.720, -14.543,
    -14.416, -14.300, -14.168, -14.011, -13.841, -13.666, -13.496, -13.337,
    -14.695, -14.524, -14.403, -14.292, -14.164, -14.010, -13.840, -13.666,
    -13.495, -13.336, -14.669, -14.504, -14.390, -14.285, -14.160, -14.008,
    -13.840, -13.665, -13.495, -13.335, -14.642, -14.484, -14.376, -14.277,
    -14.156, -14.007, -13.839, -13.665, -13.494, -13.334, -14.614, -14.463,
    -14.362, -14.269, -14.152, -14.005, -13.839, -13.665, -13.494, -13.333,
    -14.586, -14.441, -14.348, -14.261, -14.148, -14.004, -13.838, -13.664,
    -13.493, -13.332, -14.556, -14.419, -14.334, -14.252, -14.144, -14.002,
    -13.838, -13.664, -13.493, -13.331, -14.526, -14.397, -14.319, -14.244,
    -14.140, -14.000, -13.837, -13.664, -13.492, -13.330, -14.495, -14.375,
    -14.304, -14.236, -14.136, -13.999, -13.837, -13.664, -13.492, -13.329,
    -14.464, -14.352, -14.289, -14.227, -14.132, -13.997, -13.836, -13.663,
    -13.491, -13.328, -14.432, -14.328, -14.274, -14.219, -14.128, -13.996,
    -13.836, -13.663, -13.491, -13.327, -14.399, -14.305, -14.259, -14.210,
    -14.124, -13.994, -13.835, -13.663, -13.490, -13.327, -14.366, -14.281,
    -14.243, -14.201, -14.120, -13.992, -13.835, -13.663, -13.490, -13.326,
    -14.333, -14.257, -14.228, -14.192, -14.115, -13.991, -13.834, -13.662,
    -13.489, -13.325, -14.299, -14.233, -14.212, -14.184, -14.111, -13.989,
    -13.834, -13.662, -13.489, -13.324, -14.265, -14.208, -14.197, -14.175,
    -14.107, -13.988, -13.833, -13.662, -13.489, -13.323, -14.230, -14.184,
    -14.181, -14.166, -14.103, -13.986, -13.833, -13.662, -13.488, -13.322,
    -14.196, -14.160, -14.166, -14.157, -14.099, -13.984, -13.832, -13.662,
    -13.488, -13.322, -14.161, -14.135, -14.150, -14.149, -14.095, -13.983,
    -13.832, -13.662, -13.488, -13.321, -14.127, -14.111, -14.135, -14.140,
    -14.090, -13.981, -13.832, -13.662, -13.487, -13.320, -14.092, -14.087,
    -14.119, -14.131, -14.086, -13.980, -13.831, -13.661, -13.487, -13.319,
    -14.058, -14.063, -14.104, -14.123, -14.082, -13.978, -13.831, -13.661,
    -13.487, -13.319, -14.024, -14.039, -14.089, -14.115, -14.078, -13.977,
    -13.831, -13.661, -13.486, -13.318, -13.990, -14.016, -14.074, -14.106,
    -14.075, -13.976, -13.830, -13.661, -13.486, -13.317, -13.957, -13.993,
    -14.060, -14.098, -14.071, -13.974, -13.830, -13.661, -13.486, -13.316,
    -13.925, -13.971, -14.045, -14.090, -14.067, -13.973, -13.830, -13.661,
    -13.486, -13.316, -13.893, -13.949, -14.031, -14.082, -14.063, -13.972,
    -13.830, -13.661, -13.485, -13.315, -13.862, -13.927, -14.017, -14.074,
    -14.060, -13.971, -13.830, -13.661, -13.485, -13.315, -13.832, -13.906,
    -14.004, -14.067, -14.056, -13.969, -13.829, -13.661, -13.485, -13.314,
    -13.802, -13.886, -13.991, -14.060, -14.053, -13.968, -13.829, -13.661,
    -13.485, -13.313, -13.774, -13.866, -13.978, -14.053, -14.050, -13.967,
    -13.829, -13.661, -13.485, -13.313, -13.747, -13.847, -13.966, -14.046,
    -14.046, -13.966, -13.829, -13.662, -13.485, -13.312, -13.721, -13.829,
    -13.955, -14.039, -14.043, -13.965, -13.829, -13.662, -13.484, -13.312,
    -13.696, -13.811, -13.943, -14.033, -14.041, -13.964, -13.829, -13.662,
    -13.484, -13.311, -13.672, -13.795, -13.932, -14.027, -14.038, -13.964,
    -13.829, -13.662, -13.484, -13.310, -13.650, -13.779, -13.922, -14.021,
    -14.035, -13.963, -13.829, -13.662, -13.484, -13.310, -13.629, -13.764,
    -13.912, -14.015, -14.033, -13.962, -13.829, -13.662, -13.484, -13.309,
    -13.609, -13.750, -13.903, -14.010, -14.030, -13.962, -13.829, -13.663,
    -13.484, -13.309, -13.591, -13.737, -13.895, -14.005, -14.028, -13.961,
    -13.830, -13.663, -13.484, -13.308, -13.574, -13.725, -13.887, -14.001,
    -14.026, -13.961, -13.830, -13.663, -13.484, -13.308, -13.559, -13.714,
    -13.879, -13.996, -14.024, -13.960, -13.830, -13.663, -13.484, -13.308,
    -13.545, -13.703, -13.872, -13.992, -14.023, -13.960, -13.830, -13.664,
    -13.484, -13.307, -13.533, -13.694, -13.866, -13.989, -14.021, -13.960,
    -13.831, -13.664, -13.484, -13.307, -13.522, -13.686, -13.860, -13.986,
    -14.020, -13.960, -13.831, -13.664, -13.484, -13.306, -13.513, -13.679,
    -13.855, -13.983, -14.019, -13.960, -13.831, -13.665, -13.484, -13.306,
    -13.505, -13.672, -13.851, -13.980, -14.018, -13.960, -13.832, -13.665,
    -13.484, -13.305, -13.498, -13.667, -13.847, -13.978, -14.017, -13.960,
    -13.832, -13.665, -13.484, -13.305, -13.493, -13.662, -13.843, -13.976,
    -14.016, -13.960, -13.833, -13.666, -13.484, -13.305, -13.489, -13.659,
    -13.841, -13.974, -14.016, -13.960, -13.833, -13.666, -13.484, -13.304,
    -13.487, -13.656, -13.838, -13.973, -14.016, -13.961, -13.834, -13.667,
    -13.484, -13.304, -13.486, -13.655, -13.837, -13.972, -14.016, -13.961,
    -13.835, -13.667, -13.484, -13.304, -13.486, -13.654, -13.836, -13.972,
    -14.016, -13.962, -13.835, -13.668, -13.485, -13.303, -13.487, -13.654,
    -13.836, -13.972, -14.016, -13.962, -13.836, -13.668, -13.485, -13.303,
    -13.490, -13.655, -13.836, -13.972, -14.017, -13.963, -13.837, -13.669,
    -13.485, -13.303, -13.493, -13.656, -13.836, -13.972, -14.017, -13.964,
    -13.838, -13.669, -13.485, -13.302, -13.498, -13.659, -13.838, -13.973,
    -14.018, -13.965, -13.839, -13.670, -13.485, -13.302, -13.504, -13.662,
    -13.839, -13.974, -14.019, -13.966, -13.840, -13.670, -13.485, -13.302,
    -13.510, -13.666, -13.842, -13.976, -14.020, -13.967, -13.840, -13.671,
    -13.486, -13.302, -13.517, -13.670, -13.844, -13.977, -14.022, -13.969,
    -13.841, -13.672, -13.486, -13.301, -13.525, -13.675, -13.848, -13.980,
    -14.023, -13.970, -13.843, -13.672, -13.486, -13.301, -13.534, -13.681,
    -13.851, -13.982, -14.025, -13.971, -13.844, -13.673, -13.486, -13.301,
    -13.544, -13.688, -13.855, -13.985, -14.027, -13.973, -13.845, -13.674,
    -13.487, -13.301, -13.554, -13.695, -13.860, -13.988, -14.029, -13.974,
    -13.846, -13.675, -13.487, -13.301, -13.565, -13.702, -13.865, -13.991,
    -14.031, -13.976, -13.847, -13.675, -13.487, -13.300, -13.577, -13.710,
    -13.870, -13.994, -14.034, -13.978, -13.848, -13.676, -13.487, -13.300,
    -13.589, -13.718, -13.876, -13.998, -14.036, -13.980, -13.850, -13.677,
    -13.488, -13.300, -13.601, -13.727, -13.882, -14.002, -14.039, -13.982,
    -13.851, -13.678, -13.488, -13.300, -13.614, -13.737, -13.888, -14.006,
    -14.042, -13.984, -13.852, -13.678, -13.488, -13.300, -13.627, -13.746,
    -13.895, -14.011, -14.045, -13.986, -13.854, -13.679, -13.489, -13.300,
    -13.641, -13.756, -13.902, -14.015, -14.048, -13.988, -13.855, -13.680,
    -13.489, -13.299, -13.655, -13.767, -13.909, -14.020, -14.052, -13.990,
    -13.857, -13.681, -13.489, -13.299, -13.669, -13.778, -13.917, -14.026,
    -14.055, -13.992, -13.858, -13.682, -13.490, -13.299, -13.684, -13.788,
    -13.924, -14.031, -14.059, -13.995, -13.860, -13.683, -13.490, -13.299,
    -13.699, -13.800, -13.933, -14.036, -14.062, -13.997, -13.861, -13.684,
    -13.491, -13.299, -13.714, -13.811, -13.941, -14.042, -14.066, -14.000,
    -13.863, -13.685, -13.491, -13.299, -13.729, -13.823, -13.949, -14.048,
    -14.070, -14.002, -13.865, -13.686, -13.491, -13.299, -13.744, -13.835,
    -13.958, -14.054, -14.074, -14.005, -13.866, -13.687, -13.492, -13.299,
    -13.760, -13.847, -13.967, -14.060, -14.078, -14.008, -13.868, -13.688,
    -13.492, -13.299, -13.775, -13.859, -13.976, -14.067, -14.083, -14.011,
    -13.870, -13.689, -13.493, -13.299, -13.791, -13.872, -13.985, -14.073,
    -14.087, -14.014, -13.872, -13.690, -13.493, -13.299, -13.807, -13.884,
    -13.995, -14.080, -14.092, -14.016, -13.873, -13.691, -13.493, -13.299,
    -13.823, -13.897, -14.004, -14.087, -14.096, -14.019, -13.875, -13.692,
    -13.494, -13.298, -13.839, -13.910, -14.014, -14.094, -14.101, -14.023,
    -13.877, -13.693, -13.494, -13.298, -13.855, -13.923, -14.024, -14.101,
    -14.106, -14.026, -13.879, -13.694, -13.495, -13.298, -13.871, -13.936,
    -14.034, -14.108, -14.110, -14.029, -13.881, -13.695, -13.495, -13.298,
    -13.887, -13.949, -14.044, -14.115, -14.115, -14.032, -13.883, -13.696,
    -13.496, -13.298, -13.903, -13.962, -14.054, -14.122, -14.120, -14.035,
    -13.885, -13.697, -13.496, -13.298, -13.919, -13.975, -14.064, -14.130,
    -14.125, -14.038, -13.887, -13.698, -13.497, -13.298, -13.935, -13.988,
    -14.074, -14.137, -14.131, -14.042, -13.889, -13.699, -13.497, -13.298,
    -13.951, -14.002, -14.085, -14.145, -14.136, -14.045, -13.891, -13.700,
    -13.498, -13.298, -13.967, -14.015, -14.095, -14.152, -14.141, -14.048,
    -13.893, -13.701, -13.498, -13.298, -13.983, -14.028, -14.106, -14.160,
    -14.146, -14.052, -13.895, -13.703, -13.499, -13.298, -13.999, -14.042,
    -14.116, -14.168, -14.152, -14.055, -13.897, -13.704, -13.499, -13.298,
    -14.014, -14.055, -14.127, -14.176, -14.157, -14.059, -13.899, -13.705,
    -13.500, -13.299, -14.030, -14.068, -14.137, -14.183, -14.162, -14.062,
    -13.901, -13.706, -13.500, -13.299, -14.046, -14.081, -14.148, -14.191,
    -14.168, -14.066, -13.903, -13.707, -13.501, -13.299, -14.061, -14.095,
    -14.159, -14.199, -14.173, -14.069, -13.906, -13.708, -13.501, -13.299,
    -14.077, -14.108, -14.169, -14.207, -14.179, -14.073, -13.908, -13.710,
    -13.502, -13.299, -14.092, -14.121, -14.180, -14.215, -14.185, -14.077,
    -13.910, -13.711, -13.502, -13.299, -14.108, -14.134, -14.191, -14.223,
    -14.190, -14.080, -13.912, -13.712, -13.503, -13.299, -14.123, -14.148,
    -14.201, -14.231, -14.196, -14.084, -13.914, -13.713, -13.503, -13.299,
    -14.138, -14.161, -14.212, -14.240, -14.201, -14.088, -13.916, -13.714,
    -13.504, -13.299, -14.153, -14.174, -14.223, -14.248, -14.207, -14.091,
    -13.919, -13.715, -13.504, -13.299, -14.168, -14.187, -14.233, -14.256,
    -14.213, -14.095, -13.921, -13.716, -13.505, -13.299, -14.183, -14.200,
    -14.244, -14.264, -14.219, -14.099, -13.923, -13.718, -13.505, -13.299,
    -14.198, -14.213, -14.255, -14.272, -14.224, -14.102, -13.925, -13.719,
    -13.506, -13.299, -14.213, -14.226, -14.265, -14.280, -14.230, -14.106,
    -13.927, -13.720, -13.506, -13.299, -14.228, -14.239, -14.276, -14.288,
    -14.236, -14.110, -13.929, -13.721, -13.507, -13.299, -14.242, -14.251,
    -14.287, -14.296, -14.241, -14.113, -13.932, -13.722, -13.507, -13.299,
    -14.257, -14.264, -14.297, -14.305, -14.247, -14.117, -13.934, -13.723,
    -13.508, -13.300, -14.271, -14.277, -14.308, -14.313, -14.253, -14.121,
    -13.936, -13.724, -13.508, -13.300, -14.285, -14.290, -14.318, -14.321,
    -14.259, -14.124, -13.938, -13.726, -13.509, -13.300, -14.300, -14.302,
    -14.329, -14.329, -14.264, -14.128, -13.940, -13.727, -13.509, -13.300,
    -14.314, -14.315, -14.339, -14.337, -14.270, -14.132, -13.942, -13.728,
    -13.510, -13.300, -14.328, -14.327, -14.349, -14.345, -14.276, -14.135,
    -13.944, -13.729, -13.510, -13.300, -14.342, -14.340, -14.360, -14.353,
    -14.281, -14.139, -13.946, -13.730, -13.511, -13.300, -14.356, -14.352,
    -14.370, -14.361, -14.287, -14.142, -13.948, -13.731, -13.511, -13.300,
    -14.369, -14.364, -14.380, -14.369, -14.293, -14.146, -13.950, -13.732,
    -13.511, -13.300, -14.383, -14.376, -14.391, -14.377, -14.298, -14.150,
    -13.953, -13.733, -13.512, -13.300, -14.397, -14.388, -14.401, -14.385,
    -14.304, -14.153, -13.955, -13.734, -13.512, -13.300, -14.410, -14.401,
    -14.411, -14.393, -14.309, -14.157, -13.957, -13.735, -13.513, -13.301,
    -14.423, -14.413, -14.421, -14.401, -14.315, -14.160, -13.959, -13.736,
    -13.513, -13.301, -14.437, -14.425, -14.431, -14.408, -14.320, -14.164,
    -13.960, -13.737, -13.514, -13.301, -14.450, -14.436, -14.441, -14.416,
    -14.326, -14.167, -13.962, -13.738, -13.514, -13.301, -14.463, -14.448,
    -14.451, -14.424, -14.331, -14.170, -13.964, -13.739, -13.515, -13.301,
    -14.476, -14.460, -14.461, -14.432, -14.337, -14.174, -13.966, -13.740,
    -13.515, -13.301, -14.489, -14.472, -14.471, -14.439, -14.342, -14.177,
    -13.968, -13.741, -13.515, -13.301, -14.502, -14.483, -14.481, -14.447,
    -14.348, -14.181, -13.970, -13.742, -13.516, -13.301, -14.514, -14.495,
    -14.491, -14.455, -14.353, -14.184, -13.972, -13.743, -13.516, -13.301,
    -14.527, -14.506, -14.501, -14.462, -14.358, -14.187, -13.973, -13.743,
    -13.516, -13.301, -14.540, -14.518, -14.510, -14.470, -14.363, -14.190,
    -13.975, -13.744, -13.517, -13.301, -14.552, -14.529, -14.520, -14.477,
    -14.368, -14.193, -13.977, -13.745, -13.517, -13.301, -14.565, -14.540,
    -14.529, -14.485, -14.374, -14.196, -13.978, -13.746, -13.517, -13.301,
    -14.577, -14.551, -14.539, -14.492, -14.379, -14.199, -13.980, -13.747,
    -13.518, -13.302, -14.589, -14.563, -14.548, -14.499, -14.384, -14.203,
    -13.982, -13.747, -13.518, -13.302, -14.601, -14.574, -14.558, -14.506,
    -14.388, -14.205, -13.983, -13.748, -13.518, -13.302, -14.614, -14.585,
    -14.567, -14.514, -14.393, -14.208, -13.985, -13.749, -13.519, -13.302,
    -14.626, -14.596, -14.576, -14.521, -14.398, -14.211, -13.986, -13.749,
    -13.519, -13.302, -14.638, -14.606, -14.586, -14.528, -14.403, -14.214,
    -13.988, -13.750, -13.519, -13.302, -14.649, -14.617, -14.595, -14.535,
    -14.408, -14.217, -13.989, -13.751, -13.519, -13.302, -14.661, -14.628,
    -14.604, -14.542, -14.412, -14.219, -13.990, -13.751, -13.519, -13.302,
    -14.673, -14.639, -14.613, -14.549, -14.417, -14.222, -13.992, -13.752,
    -13.520, -13.302, -14.685, -14.649, -14.622, -14.555, -14.421, -14.225,
    -13.993, -13.752, -13.520, -13.302, -14.696, -14.660, -14.631, -14.562,
    -14.426, -14.227, -13.994, -13.753, -13.520, -13.302, -14.708, -14.670,
    -14.640, -14.569, -14.430, -14.229, -13.995, -13.753, -13.520, -13.302,
    -14.719, -14.681, -14.648, -14.575, -14.434, -14.232, -13.996, -13.754,
    -13.520, -13.302, -14.730, -14.691, -14.657, -14.582, -14.439, -14.234,
    -13.997, -13.754, -13.520, -13.302, -14.742, -14.701, -14.666, -14.588,
    -14.443, -14.236, -13.998, -13.754, -13.520, -13.302, -14.753, -14.712,
    -14.674, -14.595, -14.447, -14.238, -13.999, -13.755, -13.521, -13.302,
    -14.764, -14.722, -14.683, -14.601, -14.451, -14.240, -14.000, -13.755,
    -13.521, -13.302, -14.775, -14.732, -14.691, -14.607, -14.454, -14.242,
    -14.001, -13.755, -13.521, -13.302, -14.786, -14.742, -14.700, -14.613,
    -14.458, -14.244, -14.002, -13.755, -13.521, -13.302, -14.797, -14.752,
    -14.708, -14.619, -14.462, -14.246, -14.002, -13.755, -13.521, -13.302,
    -14.808, -14.762, -14.716, -14.625, -14.465, -14.248, -14.003, -13.756,
    -13.521, -13.302, -14.819, -14.772, -14.724, -14.631, -14.469, -14.250,
    -14.003, -13.756, -13.521, -13.302, -14.830, -14.781, -14.732, -14.637,
    -14.472, -14.251, -14.004, -13.756, -13.521, -13.302, -14.840, -14.791,
    -14.740, -14.642, -14.475, -14.253, -14.004, -13.756, -13.520, -13.302,
    -14.851, -14.801, -14.748, -14.648, -14.479, -14.254, -14.005, -13.756,
    -13.520, -13.302, -14.861, -14.810, -14.756, -14.653, -14.482, -14.255,
    -14.005, -13.756, -13.520, -13.302, -14.872, -14.820, -14.764, -14.659,
    -14.485, -14.256, -14.005, -13.755, -13.520, -13.302, -14.882, -14.829,
    -14.771, -14.664, -14.487, -14.257, -14.005, -13.755, -13.520, -13.301,
    -14.893, -14.839, -14.779, -14.669, -14.490, -14.258, -14.005, -13.755,
    -13.520, -13.301, -14.903, -14.848, -14.786, -14.674, -14.493, -14.259,
    -14.005, -13.755, -13.520, -13.301, -14.913, -14.857, -14.794, -14.679,
    -14.495, -14.260, -14.005, -13.755, -13.519, -13.301, -14.924, -14.866,
    -14.801, -14.683, -14.497, -14.261, -14.005, -13.754, -13.519, -13.301,
    -14.934, -14.876, -14.808, -14.688, -14.500, -14.261, -14.005, -13.754,
    -13.519, -13.301, -14.944, -14.885, -14.815, -14.693, -14.502, -14.262,
    -14.005, -13.753, -13.519, -13.301, -14.954, -14.894, -14.822, -14.697,
    -14.504, -14.262, -14.004, -13.753, -13.518, -13.301, -14.964, -14.902,
    -14.829, -14.701, -14.505, -14.262, -14.004, -13.752, -13.518, -13.301,
    -14.974, -14.911, -14.836, -14.705, -14.507, -14.263, -14.003, -13.752,
    -13.518, -13.300, -14.984, -14.920, -14.843, -14.709, -14.509, -14.263,
    -14.003, -13.751, -13.517, -13.300, -14.993, -14.929, -14.849, -14.713,
    -14.510, -14.262, -14.002, -13.751, -13.517, -13.300, -15.003, -14.937,
    -14.855, -14.716, -14.511, -14.262, -14.001, -13.750, -13.516, -13.300,
    -15.013, -14.946, -14.862, -14.720, -14.512, -14.262, -14.001, -13.749,
    -13.516, -13.300, -15.022, -14.954, -14.868, -14.723, -14.513, -14.261,
    -14.000, -13.748, -13.516, -13.300, -15.032, -14.962, -14.874, -14.726,
    -14.514, -14.261, -13.999, -13.748, -13.515, -13.299, -15.041, -14.971,
    -14.880, -14.729, -14.514, -14.260, -13.998, -13.747, -13.515, -13.299,
    -15.051, -14.979, -14.886, -14.732, -14.515, -14.259, -13.996, -13.746,
    -13.514, -13.299, -15.060, -14.987, -14.891, -14.735, -14.515, -14.258,
    -13.995, -13.745, -13.513, -13.299, -15.070, -14.995, -14.897, -14.737,
    -14.515, -14.257, -13.994, -13.744, -13.513, -13.299, -15.079, -15.003,
    -14.902, -14.739, -14.515, -14.256, -13.992, -13.743, -13.512, -13.298,
    -15.088, -15.011, -14.907, -14.742, -14.515, -14.254, -13.991, -13.742,
    -13.512, -13.298, -15.097, -15.018, -14.912, -14.743, -14.514, -14.253,
    -13.989, -13.740, -13.511, -13.298, -15.106, -15.026, -14.917, -14.745,
    -14.514, -14.251, -13.988, -13.739, -13.510, -13.298, -15.115, -15.033,
    -14.922, -14.747, -14.513, -14.250, -13.986, -13.738, -13.510, -13.297,
    -15.124, -15.041, -14.926, -14.748, -14.512, -14.248, -13.984, -13.737,
    -13.509, -13.297, -15.133, -15.048, -14.930, -14.749, -14.511, -14.246,
    -13.982, -13.735, -13.508, -13.297, -15.142, -15.055, -14.934, -14.750,
    -14.509, -14.243, -13.980, -13.734, -13.508, -13.297, -15.151, -15.062,
    -14.938, -14.750, -14.508, -14.241, -13.978, -13.732, -13.507, -13.296,
    -15.160, -15.069, -14.942, -14.751, -14.506, -14.239, -13.976, -13.731,
    -13.506, -13.296, -15.168, -15.076, -14.945, -14.751, -14.504, -14.236,
    -13.974, -13.729, -13.505, -13.296, -15.177, -15.082, -14.949, -14.751,
    -14.502, -14.233, -13.971, -13.728, -13.504, -13.296, -15.185, -15.089,
    -14.952, -14.750, -14.499, -14.230, -13.969, -13.726, -13.503, -13.295,
    -15.194, -15.095, -14.955, -14.750, -14.497, -14.227, -13.966, -13.724,
    -13.503, -13.295, -15.202, -15.101, -14.957, -14.749, -14.494, -14.224,
    -13.964, -13.723, -13.502, -13.295, -15.210, -15.107, -14.959, -14.747,
    -14.491, -14.221, -13.961, -13.721, -13.501, -13.294, -15.218, -15.113,
    -14.961, -14.746, -14.488, -14.217, -13.958, -13.719, -13.500, -13.294,
    -15.226, -15.119, -14.963, -14.744, -14.484, -14.214, -13.955, -13.717,
    -13.499, -13.294, -15.234, -15.124, -14.965, -14.742, -14.481, -14.210,
    -13.952, -13.715, -13.498, -13.293, -15.242, -15.129, -14.966, -14.740,
    -14.477, -14.206, -13.949, -13.713, -13.497, -13.293, -15.250, -15.134,
    -14.967, -14.738, -14.473, -14.202, -13.946, -13.711, -13.496, -13.293,
    -15.258, -15.139, -14.968, -14.735, -14.468, -14.198, -13.943, -13.709,
    -13.495, -13.292, -15.266, -15.144, -14.968, -14.732, -14.464, -14.193,
    -13.940, -13.707, -13.494, -13.292, -15.273, -15.148, -14.968, -14.728,
    -14.459, -14.189, -13.936, -13.705, -13.493, -13.292, -15.280, -15.153,
    -14.968, -14.725, -14.454, -14.184, -13.933, -13.703, -13.492, -13.291,
    -15.288, -15.156, -14.967, -14.721, -14.449, -14.180, -13.929, -13.701,
    -13.491, -13.291, -15.295, -15.160, -14.966, -14.716, -14.443, -14.175,
    -13.926, -13.699, -13.490, -13.291, -15.302, -15.164, -14.965, -14.712,
    -14.438, -14.170, -13.922, -13.697, -13.489, -13.290, -15.309, -15.167,
    -14.963, -14.707, -14.432, -14.165, -13.918, -13.694, -13.488, -13.290,
    -15.315, -15.170, -14.961, -14.701, -14.426, -14.160, -13.915, -13.692,
    -13.486, -13.290, -15.322, -15.172, -14.959, -14.696, -14.420, -14.154,
    -13.911, -13.690, -13.485, -13.289, -15.328, -15.175, -14.956, -14.690,
    -14.413, -14.149, -13.907, -13.687, -13.484, -13.289, -15.335, -15.177,
    -14.953, -14.684, -14.406, -14.143, -13.903, -13.685, -13.483, -13.288,
    -15.341, -15.178, -14.950, -14.677, -14.400, -14.137, -13.899, -13.682,
    -13.482, -13.288, -15.347, -15.179, -14.946, -14.671, -14.392, -14.132,
    -13.895, -13.680, -13.481, -13.288, -15.352, -15.180, -14.941, -14.663,
    -14.385, -14.126, -13.891, -13.678, -13.479, -13.287, -15.358, -15.181,
    -14.937, -14.656, -14.378, -14.120, -13.887, -13.675, -13.478, -13.287,
    -15.363, -15.181, -14.932, -14.648, -14.370, -14.114, -13.882, -13.673,
    -13.477, -13.287, -15.368, -15.181, -14.926, -14.640, -14.362, -14.107,
    -13.878, -13.670, -13.476, -13.286, -15.373, -15.180, -14.920, -14.632,
    -14.354, -14.101, -13.874, -13.667, -13.475, -13.286, -15.378, -15.179,
    -14.914, -14.623, -14.346, -14.095, -13.869, -13.665, -13.473, -13.285,
    -15.382, -15.178, -14.907, -14.614, -14.338, -14.088, -13.865, -13.662,
    -13.472, -13.285, -15.386, -15.176, -14.900, -14.605, -14.329, -14.081,
    -13.860, -13.660, -13.471, -13.285, -15.390, -15.173, -14.892, -14.596,
    -14.320, -14.075, -13.856, -13.657, -13.470, -13.284, -15.393, -15.171,
    -14.884, -14.586, -14.312, -14.068, -13.851, -13.654, -13.468, -13.284,
    -15.397, -15.167, -14.875, -14.576, -14.303, -14.061, -13.847, -13.652,
    -13.467, -13.284, -15.399, -15.163, -14.866, -14.565, -14.294, -14.054,
    -13.842, -13.649, -13.466, -13.283, -15.402, -15.159, -14.857, -14.555,
    -14.284, -14.047, -13.838, -13.647, -13.465, -13.283, -15.404, -15.154,
    -14.847, -14.544, -14.275, -14.040, -13.833, -13.644, -13.464, -13.283,
    -15.406, -15.149, -14.837, -14.533, -14.266, -14.033, -13.828, -13.641,
    -13.462, -13.282, -15.407, -15.143, -14.826, -14.522, -14.256, -14.026,
    -13.824, -13.639, -13.461, -13.282, -15.408, -15.137, -14.815, -14.510,
    -14.246, -14.019, -13.819, -13.636, -13.460, -13.282, -15.409, -15.130,
    -14.804, -14.498, -14.237, -14.012, -13.815, -13.633, -13.459, -13.281,
    -15.409, -15.122, -14.792, -14.487, -14.227, -14.005, -13.810, -13.631,
    -13.457, -13.281, -15.408, -15.114, -14.780, -14.475, -14.217, -13.998,
    -13.805, -13.628, -13.456, -13.281, -15.407, -15.106, -14.768, -14.462,
    -14.207, -13.991, -13.801, -13.625, -13.455, -13.280, -15.406, -15.097,
    -14.755, -14.450, -14.197, -13.984, -13.796, -13.623, -13.454, -13.280,
    -15.404, -15.087, -14.742, -14.438, -14.187, -13.976, -13.791, -13.620,
    -13.453, -13.280, -15.402, -15.077, -14.729, -14.425, -14.177, -13.969,
    -13.787, -13.618, -13.452, -13.279, -15.399, -15.067, -14.715, -14.412,
    -14.167, -13.962, -13.782, -13.615, -13.450, -13.279, -15.395, -15.056,
    -14.701, -14.400, -14.157, -13.955, -13.778, -13.613, -13.449, -13.279,
    -15.391, -15.044, -14.687, -14.387, -14.147, -13.948, -13.773, -13.610,
    -13.448, -13.279, -15.387, -15.032, -14.673, -14.374, -14.137, -13.941,
    -13.769, -13.608, -13.447, -13.278, -15.382, -15.020, -14.658, -14.361,
    -14.127, -13.934, -13.764, -13.605, -13.446, -13.278, -15.376, -15.007,
    -14.644, -14.348, -14.117, -13.927, -13.760, -13.603, -13.445, -13.278,
    -15.370, -14.994, -14.629, -14.335, -14.107, -13.920, -13.756, -13.600,
    -13.444, -13.278, -15.363, -14.980, -14.614, -14.322, -14.097, -13.913,
    -13.751, -13.598, -13.443, -13.277, -15.356, -14.967, -14.599, -14.310,
    -14.088, -13.907, -13.747, -13.595, -13.442, -13.277, -15.348, -14.953,
    -14.584, -14.297, -14.078, -13.900, -13.743, -13.593, -13.441, -13.277,
    -15.340, -14.938, -14.569, -14.284, -14.068, -13.893, -13.739, -13.591,
    -13.440, -13.277, -15.331, -14.924, -14.554, -14.272, -14.059, -13.887,
    -13.735, -13.589, -13.439, -13.277, -15.322, -14.909, -14.540, -14.259,
    -14.050, -13.880, -13.731, -13.586, -13.438, -13.277, -15.312, -14.894,
    -14.525, -14.247, -14.041, -13.874, -13.727, -13.584, -13.437, -13.276,
    -15.302, -14.879, -14.510, -14.235, -14.032, -13.868, -13.723, -13.582,
    -13.436, -13.276, -15.292, -14.864, -14.496, -14.223, -14.023, -13.862,
    -13.719, -13.580, -13.435, -13.276, -15.282, -14.850, -14.481, -14.212,
    -14.014, -13.856, -13.715, -13.578, -13.435, -13.276, -15.271, -14.835,
    -14.467, -14.200, -14.006, -13.850, -13.712, -13.576, -13.434, -13.276,
    -15.260, -14.820, -14.453, -14.189, -13.997, -13.845, -13.708, -13.574,
    -13.433, -13.276, -15.249, -14.806, -14.440, -14.178, -13.989, -13.839,
    -13.705, -13.572, -13.432, -13.276, -15.238, -14.791, -14.427, -14.167,
    -13.982, -13.834, -13.701, -13.571, -13.432, -13.276, -15.227, -14.777,
    -14.414, -14.157, -13.974, -13.828, -13.698, -13.569, -13.431, -13.276,
    -15.216, -14.764, -14.401, -14.147, -13.967, -13.823, -13.695, -13.567,
    -13.430, -13.276, -15.206, -14.750, -14.389, -14.137, -13.959, -13.819,
    -13.692, -13.566, -13.430, -13.276, -15.195, -14.738, -14.377, -14.128,
    -13.952, -13.814, -13.689, -13.564, -13.429, -13.276, -15.185, -14.725,
    -14.366, -14.119, -13.946, -13.809, -13.686, -13.562, -13.428, -13.276,
    -15.175, -14.713, -14.355, -14.110, -13.939, -13.805, -13.683, -13.561,
    -13.428, -13.276, -15.166, -14.702, -14.345, -14.102, -13.933, -13.801,
    -13.681, -13.560, -13.427, -13.276, -15.157, -14.691, -14.335, -14.094,
    -13.927, -13.797, -13.678, -13.558, -13.427, -13.276, -15.148, -14.681,
    -14.326, -14.087, -13.922, -13.793, -13.676, -13.557, -13.427, -13.276,
    -15.141, -14.671, -14.317, -14.080, -13.917, -13.789, -13.673, -13.556,
    -13.426, -13.277, -15.133, -14.663, -14.309, -14.073, -13.912, -13.786,
    -13.671, -13.555, -13.426, -13.277, -15.127, -14.654, -14.301, -14.067,
    -13.907, -13.782, -13.669, -13.554, -13.426, -13.277, -15.121, -14.647,
    -14.294, -14.061, -13.903, -13.779, -13.667, -13.553, -13.425, -13.277,
    -15.116, -14.640, -14.288, -14.056, -13.899, -13.776, -13.666, -13.552,
    -13.425, -13.277, -15.112, -14.635, -14.282, -14.051, -13.895, -13.774,
    -13.664, -13.551, -13.425, -13.278, -15.109, -14.629, -14.277, -14.047,
    -13.891, -13.771, -13.662, -13.550, -13.425, -13.278, -15.106, -14.625,
    -14.273, -14.043, -13.888, -13.769, -13.661, -13.550, -13.425, -13.278,
    -15.104, -14.622, -14.269, -14.039, -13.885, -13.767, -13.660, -13.549,
    -13.425, -13.278, -15.103, -14.619, -14.266, -14.036, -13.883, -13.765,
    -13.659, -13.549, -13.425, -13.279, -15.103, -14.617, -14.263, -14.033,
    -13.881, -13.764, -13.658, -13.548, -13.425, -13.279, -15.104, -14.616,
    -14.261, -14.031, -13.879, -13.762, -13.657, -13.548, -13.425, -13.280,
    -15.106, -14.615, -14.260, -14.030, -13.877, -13.761, -13.656, -13.547,
    -13.425, -13.280, -15.108, -14.615, -14.259, -14.028, -13.876, -13.760,
    -13.655, -13.547, -13.425, -13.280, -15.111, -14.616, -14.259, -14.028,
    -13.875, -13.759, -13.655, -13.547, -13.425, -13.281, -15.115, -14.618,
    -14.259, -14.027, -13.875, -13.759, -13.655, -13.547, -13.426, -13.281,
    -15.119, -14.621, -14.260, -14.027, -13.874, -13.758, -13.654, -13.547,
    -13.426, -13.282, -15.125, -14.624, -14.261, -14.028, -13.874, -13.758,
    -13.654, -13.547, -13.426, -13.282, -15.131, -14.628, -14.263, -14.029,
    -13.875, -13.758, -13.654, -13.547, -13.427, -13.283, -15.137, -14.632,
    -14.266, -14.030, -13.875, -13.759, -13.655, -13.548, -13.427, -13.283,
    -15.144, -14.637, -14.269, -14.032, -13.876, -13.759, -13.655, -13.548,
    -13.427, -13.284, -15.152, -14.643, -14.273, -14.034, -13.877, -13.760,
    -13.655, -13.549, -13.428, -13.284, -15.160, -14.649, -14.277, -14.037,
    -13.879, -13.761, -13.656, -13.549, -13.429, -13.285, -15.169, -14.656,
    -14.282, -14.040, -13.881, -13.762, -13.657, -13.550, -13.429, -13.286,
    -15.178, -14.663, -14.287, -14.043, -13.883, -13.763, -13.658, -13.550,
    -13.430, -13.286, -15.188, -14.671, -14.292, -14.047, -13.885, -13.764,
    -13.659, -13.551, -13.430, -13.287, -15.198, -14.679, -14.298, -14.051,
    -13.888, -13.766, -13.660, -13.552, -13.431, -13.288, -15.208, -14.687,
    -14.305, -14.055, -13.891, -13.768, -13.661, -13.553, -13.432, -13.289,
    -15.219, -14.697, -14.311, -14.060, -13.894, -13.770, -13.662, -13.554,
    -13.433, -13.289, -15.230, -14.706, -14.318, -14.065, -13.897, -13.772,
    -13.664, -13.555, -13.434, -13.290, -15.242, -14.716, -14.326, -14.070,
    -13.901, -13.775, -13.665, -13.556, -13.435, -13.291, -15.253, -14.726,
    -14.334, -14.076, -13.905, -13.777, -13.667, -13.557, -13.436, -13.292,
    -15.265, -14.736, -14.342, -14.082, -13.909, -13.780, -13.669, -13.559,
    -13.437, -13.293, -15.277, -14.747, -14.350, -14.088, -13.913, -13.783,
    -13.671, -13.560, -13.438, -13.294, -15.289, -14.758, -14.359, -14.094,
    -13.918, -13.786, -13.673, -13.562, -13.439, -13.294, -15.301, -14.769,
    -14.368, -14.101, -13.922, -13.789, -13.675, -13.563, -13.440, -13.295,
    -15.314, -14.781, -14.378, -14.108, -13.927, -13.793, -13.678, -13.565,
    -13.441, -13.296, -15.326, -14.792, -14.387, -14.115, -13.933, -13.797,
    -13.680, -13.566, -13.442, -13.297, -15.339, -14.804, -14.397, -14.123,
    -13.938, -13.800, -13.683, -13.568, -13.444, -13.298, -15.352, -14.816,
    -14.407, -14.131, -13.944, -13.804, -13.685, -13.570, -13.445, -13.299,
    -15.364, -14.829, -14.417, -14.139, -13.949, -13.808, -13.688, -13.572,
    -13.446, -13.300, -15.377, -14.841, -14.428, -14.147, -13.955, -13.813,
    -13.691, -13.574, -13.448, -13.301, -15.390, -14.853, -14.438, -14.155,
    -13.962, -13.817, -13.694, -13.576, -13.449, -13.303, -15.403, -14.866,
    -14.449, -14.163, -13.968, -13.821, -13.697, -13.578, -13.451, -13.304,
    -15.416, -14.879, -14.460, -14.172, -13.974, -13.826, -13.701, -13.580,
    -13.452, -13.305, -15.429, -14.892, -14.471, -14.181, -13.981, -13.831,
    -13.704, -13.583, -13.454, -13.306, -15.442, -14.904, -14.482, -14.190,
    -13.988, -13.836, -13.707, -13.585, -13.455, -13.307, -15.455, -14.917,
    -14.493, -14.199, -13.995, -13.841, -13.711, -13.587, -13.457, -13.308,
    -15.468, -14.930, -14.505, -14.208, -14.002, -13.846, -13.714, -13.590,
    -13.459, -13.310, -15.481, -14.943, -14.516, -14.217, -14.009, -13.851,
    -13.718, -13.592, -13.461, -13.311, -15.493, -14.956, -14.528, -14.227,
    -14.016, -13.857, -13.722, -13.595, -13.462, -13.312, -15.506, -14.970,
    -14.540, -14.237, -14.024, -13.862, -13.726, -13.598, -13.464, -13.313,
    -15.519, -14.983, -14.551, -14.246, -14.031, -13.868, -13.730, -13.600,
    -13.466, -13.315, -15.532, -14.996, -14.563, -14.256, -14.039, -13.873,
    -13.734, -13.603, -13.468, -13.316, -15.544, -15.009, -14.575, -14.266,
    -14.047, -13.879, -13.738, -13.606, -13.470, -13.318, -15.557, -15.022,
    -14.587, -14.276, -14.055, -13.885, -13.742, -13.609, -13.472, -13.319,
    -15.570, -15.035, -14.599, -14.286, -14.063, -13.891, -13.747, -13.612,
    -13.474, -13.320, -15.582, -15.048, -14.611, -14.296, -14.071, -13.897,
    -13.751, -13.615, -13.476, -13.322, -15.595, -15.061, -14.623, -14.307,
    -14.079, -13.903, -13.755, -13.618, -13.478, -13.323, -15.607, -15.075,
    -14.635, -14.317, -14.087, -13.910, -13.760, -13.621, -13.480, -13.325,
    -15.619, -15.088, -14.647, -14.327, -14.095, -13.916, -13.764, -13.624,
    -13.482, -13.326, -15.631, -15.101, -14.660, -14.338, -14.104, -13.922,
    -13.769, -13.628, -13.485, -13.328, -15.644, -15.114, -14.672, -14.348,
    -14.112, -13.929, -13.774, -13.631, -13.487, -13.329, -15.656, -15.127,
    -14.684, -14.359, -14.121, -13.935, -13.779, -13.634, -13.489, -13.331,
    -15.668, -15.140, -14.696, -14.369, -14.129, -13.942, -13.783, -13.638,
    -13.491, -13.332, -15.680, -15.152, -14.708, -14.380, -14.138, -13.949,
    -13.788, -13.641, -13.494, -13.334, -15.692, -15.165, -14.721, -14.390,
    -14.147, -13.955, -13.793, -13.645, -13.496, -13.336, -15.703, -15.178,
    -14.733, -14.401, -14.156, -13.962, -13.798, -13.648, -13.498, -13.337,
    -15.715, -15.191, -14.745, -14.412, -14.164, -13.969, -13.803, -13.652,
    -13.501, -13.339, -15.727, -15.204, -14.757, -14.423, -14.173, -13.976,
    -13.809, -13.655, -13.503, -13.340, -15.738, -15.216, -14.769, -14.433,
    -14.182, -13.983, -13.814, -13.659, -13.506, -13.342, -15.750, -15.229,
    -14.781, -14.444, -14.191, -13.990, -13.819, -13.663, -13.508, -13.344,
    -15.761, -15.241, -14.793, -14.455, -14.200, -13.997, -13.824, -13.666,
    -13.511, -13.346, -15.773, -15.254, -14.806, -14.466, -14.209, -14.004,
    -13.830, -13.670, -13.513, -13.347, -15.784, -15.266, -14.818, -14.476,
    -14.218, -14.012, -13.835, -13.674, -13.516, -13.349, -15.795, -15.279,
    -14.830, -14.487, -14.227, -14.019, -13.840, -13.678, -13.519, -13.351,
    -15.806, -15.291, -14.842, -14.498, -14.236, -14.026, -13.846, -13.682,
    -13.521, -13.353, -15.817, -15.303, -14.854, -14.509, -14.245, -14.033,
    -13.851, -13.686, -13.524, -13.354, -15.828, -15.315, -14.866, -14.519,
    -14.255, -14.041, -13.857, -13.689, -13.527, -13.356, -15.839, -15.328,
    -14.877, -14.530, -14.264, -14.048, -13.863, -13.693, -13.529, -13.358,
    -15.850, -15.340, -14.889, -14.541, -14.273, -14.055, -13.868, -13.697,
    -13.532, -13.360, -15.861, -15.352, -14.901, -14.552, -14.282, -14.063,
    -13.874, -13.702, -13.535, -13.362, -15.872, -15.364, -14.913, -14.562,
    -14.291, -14.070, -13.879, -13.706, -13.538, -13.364, -15.882, -15.376,
    -14.925, -14.573, -14.300, -14.078, -13.885, -13.710, -13.540, -13.366,
    -15.893, -15.387, -14.936, -14.584, -14.310, -14.085, -13.891, -13.714,
    -13.543, -13.368, -15.903, -15.399, -14.948, -14.595, -14.319, -14.093,
    -13.897, -13.718, -13.546, -13.369, -15.914, -15.411, -14.960, -14.605,
    -14.328, -14.100, -13.902, -13.722, -13.549, -13.371, -15.924, -15.422,
    -14.971, -14.616, -14.337, -14.108, -13.908, -13.726, -13.552, -13.373,
    -15.935, -15.434, -14.983, -14.627, -14.347, -14.115, -13.914, -13.731,
    -13.555, -13.375, -15.945, -15.446, -14.994, -14.637, -14.356, -14.123,
    -13.920, -13.735, -13.558, -13.377, -15.955, -15.457, -15.006, -14.648,
    -14.365, -14.131, -13.926, -13.739, -13.561, -13.379, -15.965, -15.468,
    -15.017, -14.658, -14.374, -14.138, -13.932, -13.743, -13.564, -13.381,
    -15.975, -15.480, -15.029, -14.669, -14.384, -14.146, -13.938, -13.748,
    -13.566, -13.383, -15.986, -15.491, -15.040, -14.680, -14.393, -14.153,
    -13.944, -13.752, -13.569, -13.385, -15.996, -15.502, -15.051, -14.690,
    -14.402, -14.161, -13.950, -13.756, -13.572, -13.387, -16.005, -15.514,
    -15.063, -14.701, -14.411, -14.169, -13.955, -13.761, -13.576, -13.389,
    -16.015, -15.525, -15.074, -14.711, -14.421, -14.176, -13.961, -13.765,
    -13.579, -13.392, -16.025, -15.536, -15.085, -14.722, -14.430, -14.184,
    -13.967, -13.769, -13.582, -13.394, -16.035, -15.547, -15.096, -14.732,
    -14.439, -14.192, -13.973, -13.774, -13.585, -13.396, -16.045, -15.558,
    -15.107, -14.742, -14.448, -14.199, -13.979, -13.778, -13.588, -13.398,
    -16.054, -15.568, -15.118, -14.753, -14.457, -14.207, -13.985, -13.783,
    -13.591, -13.400, -16.064, -15.579, -15.129, -14.763, -14.466, -14.215,
    -13.991, -13.787, -13.594, -13.402, -16.074, -15.590, -15.140, -14.773,
    -14.476, -14.222, -13.997, -13.792, -13.597, -13.404, -16.083, -15.601,
    -15.151, -14.784, -14.485, -14.230, -14.003, -13.796, -13.600, -13.406,
    -16.093, -15.611, -15.162, -14.794, -14.494, -14.238, -14.010, -13.800,
    -13.603, -13.408, -16.102, -15.622, -15.173, -14.804, -14.503, -14.245,
    -14.016, -13.805, -13.606, -13.410, -16.111, -15.633, -15.183, -14.814,
    -14.512, -14.253, -14.022, -13.809, -13.610, -13.413, -16.121, -15.643,
    -15.194, -14.824, -14.521, -14.260, -14.028, -13.814, -13.613, -13.415,
    -16.130, -15.654, -15.205, -14.834, -14.530, -14.268, -14.034, -13.818,
    -13.616, -13.417, -16.139, -15.664, -15.215, -14.844, -14.539, -14.276,
    -14.040, -13.823, -13.619, -13.419, -16.148, -15.674, -15.226, -14.855,
    -14.548, -14.283, -14.046, -13.827, -13.622, -13.421, -16.158, -15.685,
    -15.237, -14.865, -14.557, -14.291, -14.052, -13.832, -13.625, -13.424,
    -16.167, -15.695, -15.247, -14.875, -14.566, -14.299, -14.058, -13.836,
    -13.629, -13.426, -16.176, -15.705, -15.258, -14.884, -14.575, -14.306,
    -14.064, -13.841, -13.632, -13.428, -16.185, -15.715, -15.268, -14.894,
    -14.584, -14.314, -14.070, -13.845, -13.635, -13.430, -16.194, -15.725,
    -15.278, -14.904, -14.593, -14.321, -14.076, -13.850, -13.638, -13.432,
    -16.203, -15.735, -15.289, -14.914, -14.602, -14.329, -14.082, -13.854,
    -13.641, -13.435, -16.212, -15.745, -15.299, -14.924, -14.610, -14.336,
    -14.088, -13.859, -13.644, -13.437, -16.220, -15.755, -15.309, -14.934,
    -14.619, -14.344, -14.094, -13.863, -13.648, -13.439, -16.229, -15.765,
    -15.319, -14.943, -14.628, -14.351, -14.100, -13.868, -13.651, -13.441,
    -16.238, -15.775, -15.329, -14.953, -14.637, -14.359, -14.106, -13.872,
    -13.654, -13.443, -16.247, -15.785, -15.340, -14.963, -14.646, -14.366,
    -14.112, -13.877, -13.657, -13.446, -16.256, -15.795, -15.350, -14.972,
    -14.654, -14.374, -14.118, -13.881, -13.661, -13.448, -16.264, -15.805,
    -15.360, -14.982, -14.663, -14.381, -14.124, -13.886, -13.664, -13.450,
    -16.273, -15.814, -15.370, -14.992, -14.672, -14.388, -14.130, -13.890,
    -13.667, -13.452, -16.282, -15.824, -15.379, -15.001, -14.680, -14.396,
    -14.136, -13.895, -13.670, -13.455, -16.290, -15.834, -15.389, -15.011,
    -14.689, -14.403, -14.141, -13.899, -13.673, -13.457, -16.299, -15.843,
    -15.399, -15.020, -14.697, -14.410, -14.147, -13.904, -13.677, -13.459,
    -16.307, -15.853, -15.409, -15.030, -14.706, -14.418, -14.153, -13.908,
    -13.680, -13.461, -16.316, -15.862, -15.419, -15.039, -14.714, -14.425,
    -14.159, -13.913, -13.683, -13.464, -16.324, -15.872, -15.429, -15.048,
    -14.723, -14.432, -14.165, -13.917, -13.686, -13.466, -16.333, -15.881,
    -15.438, -15.058, -14.731, -14.439, -14.171, -13.921, -13.689, -13.468,
    -16.341, -15.890, -15.448, -15.067, -14.740, -14.447, -14.176, -13.926,
    -13.693, -13.470, -16.349, -15.900, -15.458, -15.076, -14.748, -14.454,
    -14.182, -13.930, -13.696, -13.473, -16.358, -15.909, -15.467, -15.085,
    -14.757, -14.461, -14.188, -13.935, -13.699, -13.475, -16.366, -15.918,
    -15.477, -15.095, -14.765, -14.468, -14.194, -13.939, -13.702, -13.477,
    -16.374, -15.928, -15.486, -15.104, -14.773, -14.475, -14.199, -13.943,
    -13.705, -13.480, -16.383, -15.937, -15.496, -15.113, -14.782, -14.482,
    -14.205, -13.948, -13.708, -13.482, -16.391, -15.946, -15.505, -15.122,
    -14.790, -14.489, -14.211, -13.952, -13.712, -13.484, -16.399, -15.955,
    -15.515, -15.131, -14.798, -14.496, -14.216, -13.956, -13.715, -13.486,
    -16.407, -15.964, -15.524, -15.140, -14.806, -14.503, -14.222, -13.961,
    -13.718, -13.489, -16.415, -15.973, -15.533, -15.149, -14.814, -14.510,
    -14.228, -13.965, -13.721, -13.491, -16.423, -15.982, -15.543, -15.158,
    -14.823, -14.517, -14.233, -13.969, -13.724, -13.493, -16.431, -15.991,
    -15.552, -15.167, -14.831, -14.524, -14.239, -13.973, -13.727, -13.495,
    -16.440, -16.000, -15.561, -15.176, -14.839, -14.531, -14.244, -13.978,
    -13.731, -13.498, -16.448, -16.009, -15.570, -15.185, -14.847, -14.538,
    -14.250, -13.982, -13.734, -13.500, -16.456, -16.018, -15.579, -15.194,
    -14.855, -14.545, -14.255, -13.986, -13.737, -13.502, -16.464, -16.027,
    -15.589, -15.202, -14.863, -14.552, -14.261, -13.990, -13.740, -13.504,
    -16.472, -16.036, -15.598, -15.211, -14.871, -14.558, -14.266, -13.994,
    -13.743, -13.507, -16.480, -16.045, -15.607, -15.220, -14.878, -14.565,
    -14.272, -13.999, -13.746, -13.509, -16.487, -16.053, -15.616, -15.229,
    -14.886, -14.572, -14.277, -14.003, -13.749, -13.511, -16.495, -16.062,
    -15.625, -15.237, -14.894, -14.578, -14.282, -14.007, -13.752, -13.513,
    -16.503, -16.071, -15.634, -15.246, -14.902, -14.585, -14.288, -14.011,
    -13.755, -13.516, -16.511, -16.079, -15.643, -15.254, -14.910, -14.591,
    -14.293, -14.015, -13.758, -13.518, -16.519, -16.088, -15.651, -15.263,
    -14.917, -14.598, -14.298, -14.019, -13.761, -13.520, -16.527, -16.097,
    -15.660, -15.271, -14.925, -14.605, -14.303, -14.023, -13.764, -13.522,
    -16.535, -16.105, -15.669, -15.280, -14.933, -14.611, -14.309, -14.027,
    -13.767, -13.525, -16.542, -16.114, -15.678, -15.288, -14.940, -14.617,
    -14.314, -14.031, -13.771, -13.527, -16.550, -16.122, -15.687, -15.297,
    -14.948, -14.624, -14.319, -14.035, -13.774, -13.529, -16.558, -16.131,
    -15.695, -15.305, -14.955, -14.630, -14.324, -14.039, -13.777, -13.531,
    -16.566, -16.139, -15.704, -15.313, -14.963, -14.636, -14.329, -14.043,
    -13.780, -13.534, -16.573, -16.148, -15.713, -15.322, -14.970, -14.643,
    -14.334, -14.047, -13.783, -13.536, -16.581, -16.156, -15.721, -15.330,
    -14.978, -14.649, -14.339, -14.051, -13.786, -13.538, -16.589, -16.165,
    -15.730, -15.338, -14.985, -14.655, -14.344, -14.055, -13.788, -13.540,
    -16.596, -16.173, -15.739, -15.346, -14.992, -14.661, -14.349, -14.059,
    -13.791, -13.543, -16.604, -16.181, -15.747, -15.355, -15.000, -14.667,
    -14.354, -14.062, -13.794, -13.545, -16.612, -16.190, -15.756, -15.363,
    -15.007, -14.674, -14.359, -14.066, -13.797, -13.547, -16.619, -16.198,
    -15.764, -15.371, -15.014, -14.680, -14.364, -14.070, -13.800, -13.549,
    -16.627, -16.206, -15.773, -15.379, -15.021, -14.686, -14.369, -14.074,
    -13.803, -13.552, -16.634, -16.215, -15.781, -15.387, -15.028, -14.692,
    -14.373, -14.078, -13.806, -13.554, -16.642, -16.223, -15.790, -15.395,
    -15.035, -14.697, -14.378, -14.081, -13.809, -13.556, -16.649, -16.231,
    -15.798, -15.403, -15.043, -14.703, -14.383, -14.085, -13.812, -13.558,
    -16.657, -16.239, -15.806, -15.411, -15.050, -14.709, -14.388, -14.089,
    -13.815, -13.560, -16.664, -16.248, -15.815, -15.419, -15.056, -14.715,
    -14.392, -14.093, -13.818, -13.563, -16.672, -16.256, -15.823, -15.427,
    -15.063, -14.721, -14.397, -14.096, -13.821, -13.565, -16.679, -16.264,
    -15.831, -15.434, -15.070, -14.726, -14.401, -14.100, -13.823, -13.567,
    -16.687, -16.272, -15.839, -15.442, -15.077, -14.732, -14.406, -14.103,
    -13.826, -13.569, -16.694, -16.280, -15.848, -15.450, -15.084, -14.738,
    -14.411, -14.107, -13.829, -13.571, -16.702, -16.288, -15.856, -15.458,
    -15.091, -14.743, -14.415, -14.111, -13.832, -13.574, -16.709, -16.296,
    -15.864, -15.465, -15.097, -14.749, -14.420, -14.114, -13.835, -13.576,
    -16.716, -16.304, -15.872, -15.473, -15.104, -14.754, -14.424, -14.118,
    -13.838, -13.578, -16.724, -16.312, -15.880, -15.480, -15.111, -14.760,
    -14.428, -14.121, -13.840, -13.580, -16.731, -16.320, -15.888, -15.488,
    -15.117, -14.765, -14.433, -14.125, -13.843, -13.582, -16.738, -16.328,
    -15.896, -15.495, -15.124, -14.771, -14.437, -14.128, -13.846, -13.584,
    -16.746, -16.336, -15.904, -15.503, -15.130, -14.776, -14.441, -14.132,
    -13.849, -13.587, -16.753, -16.344, -15.912, -15.510, -15.136, -14.781,
    -14.446, -14.135, -13.852, -13.589, -16.760, -16.352, -15.920, -15.518,
    -15.143, -14.786, -14.450, -14.139, -13.854, -13.591, -16.768, -16.360,
    -15.928, -15.525, -15.149, -14.792, -14.454, -14.142, -13.857, -13.593,
    -16.775, -16.368, -15.936, -15.532, -15.156, -14.797, -14.458, -14.146,
    -13.860, -13.595, -16.782, -16.375, -15.944, -15.540, -15.162, -14.802,
    -14.463, -14.149, -13.863, -13.598, -16.789, -16.383, -15.951, -15.547,
    -15.168, -14.807, -14.467, -14.152, -13.865, -13.600, -16.797, -16.391,
    -15.959, -15.554, -15.174, -14.812, -14.471, -14.156, -13.868, -13.602,
    -16.804, -16.399, -15.967, -15.561, -15.180, -14.817, -14.475, -14.159,
    -13.871, -13.604, -16.811, -16.407, -15.975, -15.568, -15.186, -14.822,
    -14.479, -14.162, -13.874, -13.606, -16.818, -16.414, -15.982, -15.575,
    -15.192, -14.827, -14.483, -14.166, -13.876, -13.608, -16.825, -16.422,
    -15.990, -15.582, -15.198, -14.832, -14.487, -14.169, -13.879, -13.610,
    -16.833, -16.430, -15.998, -15.589, -15.204, -14.837, -14.491, -14.172,
    -13.882, -13.613, -16.840, -16.438, -16.005, -15.596, -15.210, -14.842,
    -14.495, -14.176, -13.884, -13.615, -16.847, -16.445, -16.013, -15.603,
    -15.216, -14.847, -14.499, -14.179, -13.887, -13.617, -16.854, -16.453,
    -16.020, -15.610, -15.222, -14.851, -14.503, -14.182, -13.890, -13.619,
    -16.861, -16.461, -16.028, -15.617, -15.228, -14.856, -14.507, -14.186,
    -13.892, -13.621, -16.868, -16.468, -16.035, -15.624, -15.234, -14.861,
    -14.511, -14.189, -13.895, -13.623, -16.875, -16.476, -16.043, -15.631,
    -15.239, -14.866, -14.515, -14.192, -13.898, -13.626, -16.883, -16.483,
    -16.050, -15.637, -15.245, -14.870, -14.519, -14.195, -13.900, -13.628,
    -16.890, -16.491, -16.058, -15.644, -15.251, -14.875, -14.522, -14.198,
    -13.903, -13.630, -16.897, -16.499, -16.065, -15.651, -15.256, -14.880,
    -14.526, -14.202, -13.906, -13.632, -16.904, -16.506, -16.072, -15.657,
    -15.262, -14.884, -14.530, -14.205, -13.909, -13.634, -16.911, -16.514,
    -16.080, -15.664, -15.267, -14.889, -14.534, -14.208, -13.911, -13.636,
    -16.918, -16.521, -16.087, -15.670, -15.273, -14.893, -14.538, -14.211,
    -13.914, -13.638, -16.925, -16.529, -16.094, -15.677, -15.279, -14.898,
    -14.542, -14.215, -13.917, -13.641, -16.932, -16.536, -16.102, -15.684,
    -15.284, -14.903, -14.545, -14.218, -13.919, -13.643, -16.939, -16.544,
    -16.109, -15.690, -15.290, -14.907, -14.549, -14.221, -13.922, -13.645,
    -16.946, -16.551, -16.116, -15.697, -15.295, -14.912, -14.553, -14.224,
    -13.925, -13.647, -16.953, -16.559, -16.123, -15.703, -15.301, -14.916,
    -14.557, -14.227, -13.927, -13.649, -16.960, -16.566, -16.130, -15.709,
    -15.306, -14.921, -14.561, -14.231, -13.930, -13.651, -16.967, -16.574,
    -16.138, -15.716, -15.311, -14.925, -14.564, -14.234, -13.933, -13.653,
    -16.974, -16.581, -16.145, -15.722, -15.317, -14.930, -14.568, -14.237,
    -13.935, -13.656, -16.981, -16.589, -16.152, -15.729, -15.322, -14.934,
    -14.572, -14.240, -13.938, -13.658, -16.988, -16.596, -16.159, -15.735,
    -15.328, -14.939, -14.576, -14.243, -13.941, -13.660, -16.995, -16.603,
    -16.166, -15.741, -15.333, -14.943, -14.579, -14.247, -13.943, -13.662,
    -17.002, -16.611, -16.173, -15.748, -15.338, -14.948, -14.583, -14.250,
    -13.946, -13.664, -17.009, -16.618, -16.180, -15.754, -15.344, -14.952,
    -14.587, -14.253, -13.949, -13.666, -17.016, -16.625, -16.187, -15.760,
    -15.349, -14.957, -14.591, -14.256, -13.951, -13.668, -17.022, -16.633,
    -16.194, -15.767, -15.355, -14.961, -14.595, -14.260, -13.954, -13.671,
    -17.029, -16.640, -16.201, -15.773, -15.360, -14.966, -14.599, -14.263,
    -13.957, -13.673, -17.036, -16.647, -16.208, -15.779, -15.365, -14.970,
    -14.602, -14.266, -13.959, -13.675, -17.043, -16.655, -16.215, -15.786,
    -15.371, -14.975, -14.606, -14.269, -13.962, -13.677, -17.050, -16.662,
    -16.222, -15.792, -15.376, -14.979, -14.610, -14.273, -13.965, -13.679,
    -17.057, -16.669, -16.229, -15.798, -15.382, -14.984, -14.614, -14.276,
    -13.968, -13.681, -17.064, -16.677, -16.236, -15.805, -15.387, -14.988,
    -14.618, -14.279, -13.970, -13.683, -17.071, -16.684, -16.243, -15.811,
    -15.392, -14.993, -14.622, -14.283, -13.973, -13.686, -17.077, -16.691,
    -16.250, -15.817, -15.398, -14.998, -14.626, -14.286, -13.976, -13.688,
    -17.084, -16.698, -16.257, -15.824, -15.403, -15.002, -14.630, -14.289,
    -13.978, -13.690, -17.091, -16.705, -16.264, -15.830, -15.409, -15.007,
    -14.634, -14.293, -13.981, -13.692, -17.098, -16.713, -16.271, -15.837,
    -15.415, -15.012, -14.638, -14.296, -13.984, -13.694, -17.105, -16.720,
    -16.278, -15.843, -15.420, -15.016, -14.642, -14.299, -13.987, -13.696,
    -17.111, -16.727, -16.285, -15.849, -15.426, -15.021, -14.646, -14.303,
    -13.990, -13.699, -17.118, -16.734, -16.292, -15.856, -15.431, -15.026,
    -14.650, -14.306, -13.992, -13.701, -17.125, -16.741, -16.299, -15.862,
    -15.437, -15.031, -14.654, -14.309, -13.995, -13.703, -17.132, -16.749,
    -16.307, -15.869, -15.443, -15.036, -14.658, -14.313, -13.998, -13.705,
    -17.139, -16.756, -16.314, -15.875, -15.448, -15.040, -14.662, -14.316,
    -14.001, -13.707, -17.145, -16.763, -16.321, -15.882, -15.454, -15.045,
    -14.666, -14.320, -14.003, -13.710, -17.152, -16.770, -16.328, -15.888,
    -15.460, -15.050, -14.670, -14.323, -14.006, -13.712, -17.159, -16.777,
    -16.335, -15.895, -15.466, -15.055, -14.674, -14.327, -14.009, -13.714,
    -17.166, -16.784, -16.342, -15.902, -15.471, -15.060, -14.679, -14.330,
    -14.012, -13.716, -17.172, -16.791, -16.349, -15.908, -15.477, -15.065,
    -14.683, -14.334, -14.015, -13.718, -17.179, -16.798, -16.356, -15.915,
    -15.483, -15.070, -14.687, -14.337, -14.018, -13.721, -17.186, -16.805,
    -16.363, -15.922, -15.489, -15.075, -14.691, -14.341, -14.021, -13.723,
    -17.193, -16.812, -16.370, -15.928, -15.495, -15.081, -14.696, -14.345,
    -14.023, -13.725, -17.199, -16.820, -16.377, -15.935, -15.501, -15.086,
    -14.700, -14.348, -14.026, -13.727, -17.206, -16.827, -16.384, -15.942,
    -15.507, -15.091, -14.705, -14.352, -14.029, -13.729, -17.213, -16.834,
    -16.391, -15.948, -15.513, -15.096, -14.709, -14.355, -14.032, -13.732,
    -17.219, -16.841, -16.398, -15.955, -15.519, -15.101, -14.713, -14.359,
    -14.035, -13.734, -17.226, -16.848, -16.405, -15.962, -15.525, -15.107,
    -14.718, -14.363, -14.038, -13.736, -17.233, -16.855, -16.413, -15.969,
    -15.532, -15.112, -14.722, -14.366, -14.041, -13.738, -17.240, -16.862,
    -16.420, -15.976, -15.538, -15.117, -14.727, -14.370, -14.044, -13.740,
    -17.246, -16.869, -16.427, -15.982, -15.544, -15.123, -14.731, -14.374,
    -14.047, -13.743, -17.253, -16.876, -16.434, -15.989, -15.550, -15.128,
    -14.736, -14.378, -14.050, -13.745, -17.260, -16.883, -16.441, -15.996,
    -15.556, -15.134, -14.741, -14.381, -14.053, -13.747, -17.266, -16.890,
    -16.448, -16.003, -15.563, -15.139, -14.745, -14.385, -14.056, -13.749,
    -17.273, -16.897, -16.455, -16.010, -15.569, -15.145, -14.750, -14.389,
    -14.059, -13.752, -17.280, -16.904, -16.462, -16.017, -15.575, -15.150,
    -14.755, -14.393, -14.062, -13.754, -17.286, -16.910, -16.469, -16.024,
    -15.582, -15.156, -14.759, -14.397, -14.065, -13.756, -17.293, -16.917,
    -16.477, -16.031, -15.588, -15.162, -14.764, -14.401, -14.068, -13.758,
    -17.299, -16.924, -16.484, -16.038, -15.595, -15.167, -14.769, -14.405,
    -14.071, -13.760, -17.306, -16.931, -16.491, -16.045, -15.601, -15.173,
    -14.774, -14.408, -14.074, -13.763, -17.313, -16.938, -16.498, -16.052,
    -15.608, -15.179, -14.779, -14.412, -14.077, -13.765, -17.319, -16.945,
    -16.505, -16.059, -15.614, -15.184, -14.784, -14.416, -14.080, -13.767,
    -17.326, -16.952, -16.512, -16.066, -15.621, -15.190, -14.788, -14.420,
    -14.083, -13.769, -17.333, -16.959, -16.519, -16.073, -15.627, -15.196,
    -14.793, -14.424, -14.086, -13.772, -17.339, -16.966, -16.526, -16.080,
    -15.634, -15.202, -14.798, -14.428, -14.089, -13.774, -17.346, -16.973,
    -16.533, -16.087, -15.640, -15.208, -14.803, -14.432, -14.092, -13.776,
    -17.352, -16.979, -16.540, -16.094, -15.647, -15.214, -14.808, -14.436,
    -14.095, -13.778, -17.359, -16.986, -16.548, -16.101, -15.654, -15.220,
    -14.813, -14.440, -14.098, -13.781, -17.365, -16.993, -16.555, -16.108,
    -15.660, -15.225, -14.818, -14.444, -14.101, -13.783, -17.372, -17.000,
    -16.562, -16.115, -15.667, -15.231, -14.823, -14.448, -14.104, -13.785,
    -17.379, -17.007, -16.569, -16.122, -15.674, -15.237, -14.828, -14.453,
    -14.107, -13.787, -17.385, -17.014, -16.576, -16.129, -15.680, -15.243,
    -14.834, -14.457, -14.110, -13.790, -17.392, -17.020, -16.583, -16.136,
    -15.687, -15.249, -14.839, -14.461, -14.114, -13.792, -17.398, -17.027,
    -16.590, -16.143, -15.694, -15.255, -14.844, -14.465, -14.117, -13.794,
    -17.405, -17.034, -16.597, -16.150, -15.700, -15.261, -14.849, -14.469,
    -14.120, -13.796, -17.411, -17.041, -16.604, -16.157, -15.707, -15.268,
    -14.854, -14.473, -14.123, -13.799, -17.418, -17.048, -16.611, -16.164,
    -15.714, -15.274, -14.859, -14.477, -14.126, -13.801, -17.424, -17.054,
    -16.618, -16.171, -15.721, -15.280, -14.865, -14.482, -14.129, -13.803,
    -17.431, -17.061, -16.625, -16.178, -15.727, -15.286, -14.870, -14.486,
    -14.132, -13.805, -17.438, -17.068, -16.632, -16.185, -15.734, -15.292,
    -14.875, -14.490, -14.136, -13.808, -17.444, -17.075, -16.639, -16.192,
    -15.741, -15.298, -14.880, -14.494, -14.139, -13.810, -17.451, -17.082,
    -16.646, -16.199, -15.748, -15.304, -14.886, -14.498, -14.142, -13.812,
    -17.457, -17.088, -16.653, -16.206, -15.754, -15.311, -14.891, -14.503,
    -14.145, -13.814, -17.464, -17.095, -16.660, -16.213, -15.761, -15.317,
    -14.896, -14.507, -14.148, -13.817, -17.470, -17.102, -16.667, -16.220,
    -15.768, -15.323, -14.902, -14.511, -14.151, -13.819, -17.477, -17.109,
    -16.674, -16.227, -15.775, -15.329, -14.907, -14.516, -14.155, -13.821,
    -17.483, -17.115, -16.681, -16.234, -15.782, -15.335, -14.912, -14.520,
    -14.158, -13.823, -17.490, -17.122, -16.688, -16.241, -15.788, -15.342,
    -14.918, -14.524, -14.161, -13.826, -17.496, -17.129, -16.695, -16.248,
    -15.795, -15.348, -14.923, -14.528, -14.164, -13.828, -17.503, -17.135,
    -16.702, -16.255, -15.802, -15.354, -14.928, -14.533, -14.167, -13.830,
    -17.509, -17.142, -16.708, -16.262, -15.809, -15.360, -14.934, -14.537,
    -14.171, -13.832, -17.516, -17.149, -16.715, -16.269, -15.815, -15.367,
    -14.939, -14.541, -14.174, -13.835, -17.522, -17.156, -16.722, -16.276,
    -15.822, -15.373, -14.945, -14.546, -14.177, -13.837, -17.529, -17.162,
    -16.729, -16.283, -15.829, -15.379, -14.950, -14.550, -14.180, -13.839,
    -17.535, -17.169, -16.736, -16.289, -15.836, -15.385, -14.956, -14.554,
    -14.183, -13.841, -17.541, -17.176, -16.743, -16.296, -15.843, -15.392,
    -14.961, -14.559, -14.186, -13.843, -17.548, -17.182, -16.750, -16.303,
    -15.849, -15.398, -14.966, -14.563, -14.190, -13.846, -17.554, -17.189,
    -16.757, -16.310, -15.856, -15.404, -14.972, -14.567, -14.193, -13.848,
    -17.561, -17.196, -16.763, -16.317, -15.863, -15.411, -14.977, -14.572,
    -14.196, -13.850, -17.567, -17.202, -16.770, -16.324, -15.870, -15.417,
    -14.983, -14.576, -14.199, -13.852, -17.574, -17.209, -16.777, -16.331,
    -15.876, -15.423, -14.988, -14.580, -14.202, -13.854, -17.580, -17.215,
    -16.784, -16.338, -15.883, -15.429, -14.994, -14.585, -14.206, -13.857,
    -17.587, -17.222, -16.791, -16.345, -15.890, -15.436, -14.999, -14.589,
    -14.209, -13.859, -17.593, -17.229, -16.798, -16.351, -15.897, -15.442,
    -15.005, -14.593, -14.212, -13.861, -17.599, -17.235, -16.804, -16.358,
    -15.903, -15.448, -15.010, -14.598, -14.215, -13.863, -17.606, -17.242,
    -16.811, -16.365, -15.910, -15.455, -15.016, -14.602, -14.218, -13.865,
    -17.612, -17.249, -16.818, -16.372, -15.917, -15.461, -15.021, -14.606,
    -14.221, -13.868, -17.619, -17.255, -16.825, -16.379, -15.924, -15.467,
    -15.026, -14.611, -14.224, -13.870, -17.625, -17.262, -16.831, -16.386,
    -15.930, -15.473, -15.032, -14.615, -14.228, -13.872, -17.632, -17.268,
    -16.838, -16.392, -15.937, -15.480, -15.037, -14.619, -14.231, -13.874,
    -17.638, -17.275, -16.845, -16.399, -15.944, -15.486, -15.043, -14.624,
    -14.234, -13.876, -17.644, -17.282, -16.852, -16.406, -15.950, -15.492,
    -15.048, -14.628, -14.237, -13.878, -17.651, -17.288, -16.858, -16.413,
    -15.957, -15.499, -15.054, -14.632, -14.240, -13.880, -17.657, -17.295,
    -16.865, -16.420, -15.964, -15.505, -15.059, -14.637, -14.243, -13.882,
    -17.664, -17.301, -16.872, -16.426, -15.970, -15.511, -15.065, -14.641,
    -14.246, -13.885, -17.670, -17.308, -16.879, -16.433, -15.977, -15.517,
    -15.070, -14.645, -14.249, -13.887, -17.676, -17.315, -16.885, -16.440,
    -15.984, -15.524, -15.076, -14.650, -14.252, -13.889, -17.683, -17.321,
    -16.892, -16.447, -15.990, -15.530, -15.081, -14.654, -14.255, -13.891,
    -17.689, -17.328, -16.899, -16.453, -15.997, -15.536, -15.086, -14.658,
    -14.259, -13.893, -17.696, -17.334, -16.905, -16.460, -16.004, -15.542,
    -15.092, -14.662, -14.262, -13.895, -17.702, -17.341, -16.912, -16.467,
    -16.010, -15.548, -15.097, -14.667, -14.265, -13.897, -17.708, -17.347,
    -16.919, -16.473, -16.017, -15.555, -15.103, -14.671, -14.268, -13.899,
    -17.715, -17.354, -16.925, -16.480, -16.024, -15.561, -15.108, -14.675,
    -14.271, -13.901, -17.721, -17.360, -16.932, -16.487, -16.030, -15.567,
    -15.113, -14.679, -14.274, -13.903, -17.727, -17.367, -16.939, -16.493,
    -16.037, -15.573, -15.119, -14.684, -14.277, -13.905, -17.734, -17.373,
    -16.945, -16.500, -16.043, -15.579, -15.124, -14.688, -14.280, -13.907,
    -17.740, -17.380, -16.952, -16.507, -16.050, -15.586, -15.129, -14.692,
    -14.283, -13.909, -17.747, -17.386, -16.959, -16.513, -16.056, -15.592,
    -15.135, -14.696, -14.286, -13.911, -17.753, -17.393, -16.965, -16.520,
    -16.063, -15.598, -15.140, -14.700, -14.288, -13.913, -17.759, -17.399,
    -16.972, -16.527, -16.070, -15.604, -15.145, -14.704, -14.291, -13.915,
    -17.766, -17.406, -16.978, -16.533, -16.076, -15.610, -15.151, -14.709,
    -14.294, -13.917, -17.772, -17.412, -16.985, -16.540, -16.083, -15.616,
    -15.156, -14.713, -14.297, -13.919, -17.778, -17.419, -16.992, -16.546,
    -16.089, -15.622, -15.161, -14.717, -14.300, -13.921, -17.785, -17.425,
    -16.998, -16.553, -16.096, -15.628, -15.167, -14.721, -14.303, -13.923,
    -17.791, -17.432, -17.005, -16.560, -16.102, -15.634, -15.172, -14.725,
    -14.306, -13.925, -17.797, -17.438, -17.011, -16.566, -16.109, -15.641,
    -15.177, -14.729, -14.309, -13.927, -17.804, -17.445, -17.018, -16.573,
    -16.115, -15.647, -15.182, -14.733, -14.312, -13.929, -17.810, -17.451,
    -17.024, -16.579, -16.122, -15.653, -15.188, -14.737, -14.314, -13.930,
    -17.816, -17.458, -17.031, -16.586, -16.128, -15.659, -15.193, -14.741,
    -14.317, -13.932, -17.823, -17.464, -17.037, -16.592, -16.135, -15.665,
    -15.198, -14.745, -14.320, -13.934, -17.829, -17.471, -17.044, -16.599,
    -16.141, -15.671, -15.203, -14.749, -14.323, -13.936, -17.835, -17.477,
    -17.051, -16.605, -16.147, -15.677, -15.208, -14.753, -14.325, -13.938,
    -17.842, -17.484, -17.057, -16.612, -16.154, -15.683, -15.213, -14.757,
    -14.328, -13.940, -17.848, -17.490, -17.064, -16.618, -16.160, -15.689,
    -15.219, -14.761, -14.331, -13.942, -17.854, -17.497, -17.070, -16.625,
    -16.167, -15.695, -15.224, -14.765, -14.334, -13.943, -17.861, -17.503,
    -17.077, -16.632, -16.173, -15.701, -15.229, -14.769, -14.336, -13.945,
    -17.867, -17.509, -17.083, -16.638, -16.180, -15.707, -15.234, -14.773,
    -14.339, -13.947, -17.873, -17.516, -17.090, -16.644, -16.186, -15.713,
    -15.239, -14.776, -14.342, -13.949, -17.880, -17.522, -17.096, -16.651,
    -16.192, -15.718, -15.244, -14.780, -14.344, -13.950, -17.886, -17.529,
    -17.103, -16.657, -16.199, -15.724, -15.249, -14.784, -14.347, -13.952,
    -17.892, -17.535, -17.109, -16.664, -16.205, -15.730, -15.254, -14.788,
    -14.349, -13.954, -17.898, -17.542, -17.116, -16.670, -16.211, -15.736,
    -15.259, -14.791, -14.352, -13.956, -17.905, -17.548, -17.122, -16.677,
    -16.218, -15.742, -15.264, -14.795, -14.354, -13.957, -17.911, -17.554,
    -17.129, -16.683, -16.224, -15.748, -15.269, -14.799, -14.357, -13.959,
    -17.917, -17.561, -17.135, -16.690, -16.231, -15.754, -15.274, -14.803,
    -14.359, -13.961, -17.924, -17.567, -17.141, -16.696, -16.237, -15.759,
    -15.279, -14.806, -14.362, -13.962, -17.930, -17.574, -17.148, -16.702,
    -16.243, -15.765, -15.283, -14.810, -14.364, -13.964, -17.936, -17.580,
    -17.154, -16.709, -16.250, -15.771, -15.288, -14.813, -14.367, -13.966,
    -17.943, -17.586, -17.161, -16.715, -16.256, -15.777, -15.293, -14.817,
    -14.369, -13.967, -17.949, -17.593, -17.167, -16.722, -16.262, -15.782,
    -15.298, -14.821, -14.372, -13.969, -17.955, -17.599, -17.174, -16.728,
    -16.268, -15.788, -15.303, -14.824, -14.374, -13.971, -17.961, -17.606,
    -17.180, -16.734, -16.275, -15.794, -15.307, -14.828, -14.377, -13.972,
    -17.968, -17.612, -17.187, -16.741, -16.281, -15.800, -15.312, -14.831,
    -14.379, -13.974, -17.974, -17.618, -17.193, -16.747, -16.287, -15.805,
    -15.317, -14.835, -14.381, -13.976, -17.980, -17.625, -17.199, -16.754,
    -16.294, -15.811, -15.322, -14.838, -14.384, -13.977, -17.986, -17.631,
    -17.206, -16.760, -16.300, -15.817, -15.326, -14.841, -14.386, -13.979,
    -17.993, -17.638, -17.212, -16.766, -16.306, -15.822, -15.331, -14.845,
    -14.388, -13.980, -17.999, -17.644, -17.219, -16.773, -16.312, -15.828,
    -15.336, -14.848, -14.390, -13.982, -18.005, -17.650, -17.225, -16.779,
    -16.318, -15.833, -15.340, -14.852, -14.393, -13.983, -18.012, -17.657,
    -17.231, -16.785, -16.325, -15.839, -15.345, -14.855, -14.395, -13.985,
    -18.018, -17.663, -17.238, -16.792, -16.331, -15.845, -15.349, -14.858,
    -14.397, -13.986, -18.024, -17.669, -17.244, -16.798, -16.337, -15.850,
    -15.354, -14.861, -14.399, -13.988, -18.030, -17.676, -17.250, -16.804,
    -16.343, -15.856, -15.358, -14.865, -14.402, -13.989, -18.037, -17.682,
    -17.257, -16.811, -16.350, -15.861, -15.363, -14.868, -14.404, -13.991,
    -18.043, -17.688, -17.263, -16.817, -16.356, -15.867, -15.367, -14.871,
    -14.406, -13.992, -18.049, -17.695, -17.270, -16.823, -16.362, -15.872,
    -15.372, -14.874, -14.408, -13.994, -18.055, -17.701, -17.276, -16.830,
    -16.368, -15.878, -15.376, -14.877, -14.410, -13.995, -18.062, -17.707,
    -17.282, -16.836, -16.374, -15.883, -15.380, -14.881, -14.412, -13.997,
    -18.068, -17.714, -17.289, -16.842, -16.380, -15.889, -15.385, -14.884,
    -14.414, -13.998, -18.074, -17.720, -17.295, -16.848, -16.387, -15.894,
    -15.389, -14.887, -14.416, -14.000, -18.080, -17.726, -17.301, -16.855,
    -16.393, -15.899, -15.393, -14.890, -14.418, -14.001, -18.087, -17.733,
    -17.308, -16.861, -16.399, -15.905, -15.398, -14.893, -14.420, -14.003,
    -18.093, -17.739, -17.314, -16.867, -16.405, -15.910, -15.402, -14.896,
    -14.422, -14.004, -18.099, -17.745, -17.320, -16.873, -16.411, -15.915,
    -15.406, -14.899, -14.424, -14.005, -18.105, -17.752, -17.327, -16.880,
    -16.417, -15.921, -15.410, -14.902, -14.426, -14.007, -18.111, -17.758,
    -17.333, -16.886, -16.423, -15.926, -15.414, -14.905, -14.428, -14.008,
    -18.118, -17.764, -17.339, -16.892, -16.429, -15.931, -15.419, -14.907,
    -14.430, -14.010, -18.124, -17.771, -17.346, -16.898, -16.436, -15.936,
    -15.423, -14.910, -14.432, -14.011, -18.130, -17.777, -17.352, -16.905,
    -16.442, -15.942, -15.427, -14.913, -14.434, -14.012, -18.136, -17.783,
    -17.358, -16.911, -16.448, -15.947, -15.431, -14.916, -14.436, -14.014,
    -18.143, -17.790, -17.365, -16.917, -16.454, -15.952, -15.435, -14.919,
    -14.438, -14.015, -18.149, -17.796, -17.371, -16.923, -16.460, -15.957,
    -15.439, -14.922, -14.440, -14.017, -18.155, -17.802, -17.377, -16.929,
    -16.466, -15.962, -15.443, -14.924, -14.442, -14.018, -18.161, -17.809,
    -17.383, -16.936, -16.472, -15.968, -15.447, -14.927, -14.444, -14.019,
    -18.167, -17.815, -17.390, -16.942, -16.478, -15.973, -15.451, -14.930,
    -14.446, -14.021, -18.174, -17.821, -17.396, -16.948, -16.484, -15.978,
    -15.455, -14.933, -14.447, -14.022, -18.180, -17.827, -17.402, -16.954,
    -16.490, -15.983, -15.459, -14.935, -14.449, -14.023, -18.186, -17.834,
    -17.409, -16.960, -16.496, -15.988, -15.462, -14.938, -14.451, -14.025,
    -18.192, -17.840, -17.415, -16.967, -16.502, -15.993, -15.466, -14.941,
    -14.453, -14.026, -18.199, -17.846, -17.421, -16.973, -16.508, -15.998,
    -15.470, -14.943, -14.455, -14.027, -18.205, -17.852, -17.427, -16.979,
    -16.514, -16.003, -15.474, -14.946, -14.456, -14.029, -18.211, -17.859,
    -17.434, -16.985, -16.520, -16.008, -15.478, -14.949, -14.458, -14.030,
    -18.217, -17.865, -17.440, -16.991, -16.526, -16.013, -15.482, -14.951,
    -14.460, -14.031, -18.223, -17.871, -17.446, -16.997, -16.532, -16.018,
    -15.485, -14.954, -14.462, -14.033, -18.230, -17.878, -17.452, -17.004,
    -16.538, -16.023, -15.489, -14.956, -14.464, -14.034, -18.236, -17.884,
    -17.459, -17.010, -16.544, -16.028, -15.493, -14.959, -14.465, -14.035,
    -18.242, -17.890, -17.465, -17.016, -16.550, -16.033, -15.496, -14.962,
    -14.467, -14.037, -18.248, -17.896, -17.471, -17.022, -16.556, -16.038,
    -15.500, -14.964, -14.469, -14.038, -18.254, -17.903, -17.477, -17.028,
    -16.562, -16.042, -15.504, -14.967, -14.471, -14.040, -18.261, -17.909,
    -17.484, -17.034, -16.568, -16.047, -15.507, -14.969, -14.473, -14.041,
    -18.267, -17.915, -17.490, -17.040, -16.574, -16.052, -15.511, -14.972,
    -14.474, -14.042, -18.273, -17.921, -17.496, -17.046, -16.580, -16.057,
    -15.515, -14.974, -14.476, -14.044, -18.279, -17.928, -17.502, -17.053,
    -16.586, -16.062, -15.518, -14.977, -14.478, -14.045, -18.285, -17.934,
    -17.508, -17.059, -16.592, -16.067, -15.522, -14.979, -14.480, -14.046,
    -18.291, -17.940, -17.515, -17.065, -16.598, -16.071, -15.526, -14.982,
    -14.481, -14.048, -18.298, -17.946, -17.521, -17.071, -16.604, -16.076,
    -15.529, -14.984, -14.483, -14.049, -18.304, -17.953, -17.527, -17.077,
    -16.610, -16.081, -15.533, -14.987, -14.485, -14.050, -18.310, -17.959,
    -17.533, -17.083, -16.616, -16.086, -15.536, -14.989, -14.487, -14.052,
    -18.316, -17.965, -17.539, -17.089, -16.622, -16.090, -15.540, -14.992,
    -14.488, -14.053, -18.322, -17.971, -17.546, -17.095, -16.628, -16.095,
    -15.543, -14.994, -14.490, -14.055, -18.329, -17.978, -17.552, -17.101,
    -16.634, -16.100, -15.547, -14.997, -14.492, -14.056, -18.335, -17.984,
    -17.558, -17.107, -16.640, -16.105, -15.551, -14.999, -14.494, -14.057,
    -18.341, -17.990, -17.564, -17.113, -16.646, -16.109, -15.554, -15.002,
    -14.496, -14.059, -18.347, -17.996, -17.570, -17.120, -16.651, -16.114,
    -15.558, -15.004, -14.497, -14.060, -18.353, -18.002, -17.577, -17.126,
    -16.657, -16.119, -15.561, -15.007, -14.499, -14.062, -18.359, -18.009,
    -17.583, -17.132, -16.663, -16.123, -15.565, -15.009, -14.501, -14.063,
    -18.366, -18.015, -17.589, -17.138, -16.669, -16.128, -15.568, -15.012,
    -14.503, -14.065, -18.372, -18.021, -17.595, -17.144, -16.675, -16.133,
    -15.572, -15.014, -14.505, -14.066, -18.378, -18.027, -17.601, -17.150,
    -16.681, -16.138, -15.575, -15.017, -14.507, -14.068, -18.384, -18.034,
    -17.608, -17.156, -16.687, -16.142, -15.579, -15.019, -14.509, -14.069,
    -18.390, -18.040, -17.614, -17.162, -16.693, -16.147, -15.582, -15.022,
    -14.510, -14.071, -18.396, -18.046, -17.620, -17.168, -16.699, -16.152,
    -15.586, -15.024, -14.512, -14.072, -18.403, -18.052, -17.626, -17.174,
    -16.704, -16.156, -15.590, -15.027, -14.514, -14.074, -18.409, -18.058,
    -17.632, -17.180, -16.710, -16.161, -15.593, -15.030, -14.516, -14.075,
    -18.415, -18.065, -17.638, -17.186, -16.716, -16.166, -15.597, -15.032,
    -14.518, -14.077, -18.421, -18.071, -17.644, -17.192, -16.722, -16.170,
    -15.600, -15.035, -14.520, -14.078, -18.427, -18.077, -17.651, -17.198,
    -16.728, -16.175, -15.604, -15.038, -14.522, -14.080, -18.433, -18.083,
    -17.657, -17.204, -16.734, -16.180, -15.608, -15.040, -14.524, -14.081,
    -18.439, -18.089, -17.663, -17.210, -16.740, -16.185, -15.611, -15.043,
    -14.526, -14.083, -18.446, -18.096, -17.669, -17.216, -16.745, -16.189,
    -15.615, -15.046, -14.528, -14.085, -18.452, -18.102, -17.675, -17.222,
    -16.751, -16.194, -15.619, -15.048, -14.530, -14.086, -18.458, -18.108,
    -17.681, -17.228, -16.757, -16.199, -15.622, -15.051, -14.532, -14.088,
    -18.464, -18.114, -17.687, -17.234, -16.763, -16.203, -15.626, -15.054,
    -14.534, -14.089, -18.470, -18.120, -17.694, -17.240, -16.769, -16.208,
    -15.630, -15.057, -14.536, -14.091, -18.476, -18.126, -17.700, -17.246,
    -16.774, -16.213, -15.634, -15.059, -14.539, -14.093, -18.483, -18.133,
    -17.706, -17.252, -16.780, -16.218, -15.637, -15.062, -14.541, -14.095,
    -18.489, -18.139, -17.712, -17.258, -16.786, -16.223, -15.641, -15.065,
    -14.543, -14.096, -18.495, -18.145, -17.718, -17.264, -16.792, -16.227,
    -15.645, -15.068, -14.545, -14.098, -18.501, -18.151, -17.724, -17.270,
    -16.798, -16.232, -15.649, -15.071, -14.547, -14.100, -18.507, -18.157,
    -17.730, -17.276, -16.803, -16.237, -15.653, -15.074, -14.549, -14.102,
    -18.513, -18.164, -17.736, -17.282, -16.809, -16.242, -15.656, -15.077,
    -14.552, -14.103, -18.519, -18.170, -17.743, -17.288, -16.815, -16.247,
    -15.660, -15.080, -14.554, -14.105, -18.526, -18.176, -17.749, -17.294,
    -16.821, -16.252, -15.664, -15.083, -14.556, -14.107, -18.532, -18.182,
    -17.755, -17.300, -16.827, -16.257, -15.668, -15.086, -14.559, -14.109,
    -18.538, -18.188, -17.761, -17.306, -16.832, -16.261, -15.672, -15.089,
    -14.561, -14.111, -18.544, -18.194, -17.767, -17.312, -16.838, -16.266,
    -15.676, -15.092, -14.563, -14.113, -18.550, -18.201, -17.773, -17.318,
    -16.844, -16.271, -15.680, -15.095, -14.566, -14.114, -18.556, -18.207,
    -17.779, -17.324, -16.850, -16.276, -15.684, -15.098, -14.568, -14.116,
    -18.562, -18.213, -17.785, -17.330, -16.855, -16.281, -15.688, -15.101,
    -14.571, -14.118, -18.568, -18.219, -17.791, -17.336, -16.861, -16.286,
    -15.693, -15.105, -14.573, -14.120, -18.575, -18.225, -17.797, -17.342,
    -16.867, -16.291, -15.697, -15.108, -14.576, -14.122, -18.581, -18.231,
    -17.804, -17.348, -16.873, -16.296, -15.701, -15.111, -14.578, -14.124,
    -18.587, -18.237, -17.810, -17.354, -16.878, -16.301, -15.705, -15.114,
    -14.581, -14.126, -18.593, -18.244, -17.816, -17.359, -16.884, -16.306,
    -15.709, -15.118, -14.583, -14.128, -18.599, -18.250, -17.822, -17.365,
    -16.890, -16.312, -15.714, -15.121, -14.586, -14.130, -18.605, -18.256,
    -17.828, -17.371, -16.896, -16.317, -15.718, -15.125, -14.589, -14.132,
    -18.611, -18.262, -17.834, -17.377, -16.901, -16.322, -15.722, -15.128,
    -14.591, -14.134, -18.617, -18.268, -17.840, -17.383, -16.907, -16.327,
    -15.727, -15.132, -14.594, -14.137, -18.624, -18.274, -17.846, -17.389,
    -16.913, -16.332, -15.731, -15.135, -14.597, -14.139, -18.630, -18.280,
    -17.852, -17.395, -16.919, -16.337, -15.735, -15.139, -14.600, -14.141,
    -18.636, -18.287, -17.858, -17.401, -16.924, -16.342, -15.740, -15.142,
    -14.602, -14.143, -18.642, -18.293, -17.864, -17.407, -16.930, -16.348,
    -15.744, -15.146, -14.605, -14.145, -18.648, -18.299, -17.870, -17.413,
    -16.936, -16.353, -15.749, -15.149, -14.608, -14.147, -18.654, -18.305,
    -17.876, -17.419, -16.941, -16.358, -15.753, -15.153, -14.611, -14.150,
    -18.660, -18.311, -17.883, -17.425, -16.947, -16.363, -15.758, -15.157,
    -14.614, -14.152, -18.666, -18.317, -17.889, -17.430, -16.953, -16.369,
    -15.762, -15.160, -14.617, -14.154, -18.672, -18.323, -17.895, -17.436,
    -16.958, -16.374, -15.767, -15.164, -14.620, -14.157, -18.679, -18.330,
    -17.901, -17.442, -16.964, -16.379, -15.772, -15.168, -14.623, -14.159,
    -18.685, -18.336, -17.907, -17.448, -16.970, -16.385, -15.776, -15.172,
    -14.626, -14.161, -18.691, -18.342, -17.913, -17.454, -16.976, -16.390,
    -15.781, -15.176, -14.629, -14.164, -18.697, -18.348, -17.919, -17.460,
    -16.981, -16.395, -15.786, -15.180, -14.632, -14.166, -18.703, -18.354,
    -17.925, -17.466, -16.987, -16.401, -15.791, -15.184, -14.635, -14.168,
    -18.709, -18.360, -17.931, -17.472, -16.993, -16.406, -15.795, -15.187,
    -14.638, -14.171, -18.715, -18.366, -17.937, -17.478, -16.998, -16.412,
    -15.800, -15.191, -14.641, -14.173, -18.721, -18.372, -17.943, -17.483,
    -17.004, -16.417, -15.805, -15.195, -14.644, -14.176, -18.727, -18.379,
    -17.949, -17.489, -17.010, -16.422, -15.810, -15.200, -14.648, -14.178,
    -18.734, -18.385, -17.955, -17.495, -17.015, -16.428, -15.815, -15.204,
    -14.651, -14.181, -18.740, -18.391, -17.961, -17.501, -17.021, -16.433,
    -15.820, -15.208, -14.654, -14.183, -18.746, -18.397, -17.967, -17.507,
    -17.027, -16.439, -15.825, -15.212, -14.657, -14.186, -18.752, -18.403,
    -17.973, -17.513, -17.032, -16.444, -15.830, -15.216, -14.661, -14.188,
    -18.758, -18.409, -17.979, -17.519, -17.038, -16.450, -15.835, -15.220,
    -14.664, -14.191, -18.764, -18.415, -17.985, -17.525, -17.044, -16.455,
    -15.840, -15.224, -14.667, -14.194, -18.770, -18.421, -17.991, -17.530,
    -17.049, -16.461, -15.845, -15.229, -14.671, -14.196, -18.776, -18.427,
    -17.997, -17.536, -17.055, -16.466, -15.850, -15.233, -14.674, -14.199,
    -18.782, -18.434, -18.003, -17.542, -17.061, -16.472, -15.855, -15.237,
    -14.678, -14.202, -18.788, -18.440, -18.009, -17.548, -17.066, -16.477,
    -15.860, -15.242, -14.681, -14.204, -18.795, -18.446, -18.015, -17.554,
    -17.072, -16.483, -15.865, -15.246, -14.685, -14.207, -18.801, -18.452,
    -18.021, -17.560, -17.077, -16.488, -15.870, -15.250, -14.688, -14.210,
    -18.807, -18.458, -18.027, -17.566, -17.083, -16.494, -15.875, -15.255,
    -14.692, -14.212, -18.813, -18.464, -18.033, -17.571, -17.089, -16.499,
    -15.880, -15.259, -14.695, -14.215, -18.819, -18.470, -18.039, -17.577,
    -17.094, -16.505, -15.885, -15.264, -14.699, -14.218, -18.825, -18.476,
    -18.045, -17.583, -17.100, -16.511, -15.891, -15.268, -14.703, -14.221,
    -18.831, -18.482, -18.051, -17.589, -17.106, -16.516, -15.896, -15.273,
    -14.706, -14.224, -18.837, -18.488, -18.057, -17.595, -17.111, -16.522,
    -15.901, -15.277, -14.710, -14.226, -18.843, -18.495, -18.063, -17.601,
    -17.117, -16.527, -15.906, -15.282, -14.714, -14.229, -18.849, -18.501,
    -18.069, -17.606, -17.122, -16.533, -15.912, -15.286, -14.717, -14.232,
    -18.855, -18.507, -18.075, -17.612, -17.128, -16.539, -15.917, -15.291,
    -14.721, -14.235, -18.862, -18.513, -18.081, -17.618, -17.134, -16.544,
    -15.922, -15.296, -14.725, -14.238, -18.868, -18.519, -18.087, -17.624,
    -17.139, -16.550, -15.927, -15.300, -14.729, -14.241, -18.874, -18.525,
    -18.093, -17.630, -17.145, -16.555, -15.933, -15.305, -14.732, -14.244,
    -18.880, -18.531, -18.099, -17.636, -17.150, -16.561, -15.938, -15.309,
    -14.736, -14.247, -18.886, -18.537, -18.105, -17.641, -17.156, -16.567,
    -15.943, -15.314, -14.740, -14.250, -18.892, -18.543, -18.111, -17.647,
    -17.162, -16.572, -15.949, -15.319, -14.744, -14.253, -18.898, -18.549,
    -18.117, -17.653, -17.167, -16.578, -15.954, -15.324, -14.748, -14.256,
    -18.904, -18.555, -18.123, -17.659, -17.173, -16.584, -15.959, -15.328,
    -14.752, -14.259, -18.910, -18.561, -18.129, -17.665, -17.178, -16.589,
    -15.965, -15.333, -14.756, -14.262, -18.916, -18.568, -18.135, -17.670,
    -17.184, -16.595, -15.970, -15.338, -14.760, -14.265, -18.922, -18.574,
    -18.141, -17.676, -17.190, -16.600, -15.976, -15.343, -14.764, -14.268,
    -18.928, -18.580, -18.147, -17.682, -17.195, -16.606, -15.981, -15.348,
    -14.768, -14.271, -18.935, -18.586, -18.153, -17.688, -17.201, -16.612,
    -15.986, -15.352, -14.772, -14.275, -18.941, -18.592, -18.159, -17.694,
    -17.206, -16.617, -15.992, -15.357, -14.776, -14.278, -18.947, -18.598,
    -18.165, -17.699, -17.212, -16.623, -15.997, -15.362, -14.780, -14.281,
    -18.953, -18.604, -18.171, -17.705, -17.218, -16.629, -16.003, -15.367,
    -14.784, -14.284, -18.959, -18.610, -18.177, -17.711, -17.223, -16.634,
    -16.008, -15.372, -14.788, -14.287, -18.965, -18.616, -18.183, -17.717,
    -17.229, -16.640, -16.014, -15.377, -14.792, -14.290, -18.971, -18.622,
    -18.189, -17.723, -17.234, -16.645, -16.019, -15.382, -14.796, -14.294,
    -18.977, -18.628, -18.195, -17.728, -17.240, -16.651, -16.025, -15.387,
    -14.800, -14.297, -18.983, -18.634, -18.201, -17.734, -17.245, -16.657,
    -16.030, -15.392, -14.804, -14.300, -18.989, -18.640, -18.207, -17.740,
    -17.251, -16.662, -16.036, -15.397, -14.808, -14.303, -18.995, -18.646,
    -18.213, -17.746, -17.256, -16.668, -16.041, -15.401, -14.813, -14.307,
    -19.001, -18.652, -18.219, -17.752, -17.262, -16.674, -16.046, -15.406,
    -14.817, -14.310, -19.007, -18.659, -18.225, -17.757, -17.268, -16.679,
    -16.052, -15.411, -14.821, -14.313, -19.013, -18.665, -18.231, -17.763,
    -17.273, -16.685, -16.057, -15.416, -14.825, -14.317, -19.020, -18.671,
    -18.237, -17.769, -17.279, -16.691, -16.063, -15.421, -14.829, -14.320,
    -19.026, -18.677, -18.243, -17.775, -17.284, -16.696, -16.068, -15.426,
    -14.834, -14.323, -19.032, -18.683, -18.249, -17.780, -17.290, -16.702,
    -16.074, -15.431, -14.838, -14.327, -19.038, -18.689, -18.255, -17.786,
    -17.295, -16.707, -16.079, -15.437, -14.842, -14.330, -19.044, -18.695,
    -18.261, -17.792, -17.301, -16.713, -16.085, -15.442, -14.846, -14.333,
    -19.050, -18.701, -18.266, -17.798, -17.306, -16.719, -16.090, -15.447,
    -14.851, -14.337, -19.056, -18.707, -18.272, -17.803, -17.312, -16.724,
    -16.096, -15.452, -14.855, -14.340, -19.062, -18.713, -18.278, -17.809,
    -17.317, -16.730, -16.102, -15.457, -14.859, -14.344, -19.068, -18.719,
    -18.284, -17.815, -17.323, -16.735, -16.107, -15.462, -14.864, -14.347,
    -19.074, -18.725, -18.290, -17.821, -17.329, -16.741, -16.113, -15.467,
    -14.868, -14.350, -19.080, -18.731, -18.296, -17.827, -17.334, -16.747,
    -16.118, -15.472, -14.872, -14.354, -19.086, -18.737, -18.302, -17.832,
    -17.340, -16.752, -16.124, -15.477, -14.877, -14.357, -19.092, -18.743,
    -18.308, -17.838, -17.345, -16.758, -16.129, -15.482, -14.881, -14.361,
    -19.098, -18.749, -18.314, -17.844, -17.351, -16.763, -16.135, -15.487,
    -14.885, -14.364, -19.104, -18.755, -18.320, -17.850, -17.356, -16.769,
    -16.140, -15.492, -14.890, -14.368, -19.110, -18.761, -18.326, -17.855,
    -17.362, -16.775, -16.146, -15.497, -14.894, -14.371, -19.117, -18.767,
    -18.332, -17.861, -17.367, -16.780, -16.151, -15.503, -14.898, -14.375,
    -19.123, -18.773, -18.338, -17.867, -17.373, -16.786, -16.157, -15.508,
    -14.903, -14.378, -19.129, -18.780, -18.344, -17.873, -17.378, -16.791,
    -16.162, -15.513, -14.907, -14.382, -19.135, -18.786, -18.350, -17.878,
    -17.384, -16.797, -16.168, -15.518, -14.912, -14.385, -19.141, -18.792,
    -18.356, -17.884, -17.389, -16.802, -16.173, -15.523, -14.916, -14.389,
    -19.147, -18.798, -18.361, -17.890, -17.395, -16.808, -16.179, -15.528,
    -14.920, -14.392, -19.153, -18.804, -18.367, -17.895, -17.400, -16.814,
    -16.184, -15.533, -14.925, -14.396, -19.159, -18.810, -18.373, -17.901,
    -17.406, -16.819, -16.190, -15.538, -14.929, -14.399, -19.165, -18.816,
    -18.379, -17.907, -17.411, -16.825, -16.195, -15.543, -14.934, -14.403,
    -19.171, -18.822, -18.385, -17.913, -17.417, -16.830, -16.201, -15.549,
    -14.938, -14.407, -19.177, -18.828, -18.391, -17.918, -17.422, -16.836,
    -16.206, -15.554, -14.943, -14.410, -19.183, -18.834, -18.397, -17.924,
    -17.428, -16.841, -16.212, -15.559, -14.947, -14.414, -19.189, -18.840,
    -18.403, -17.930, -17.433, -16.847, -16.217, -15.564, -14.951, -14.417,
    -19.195, -18.846, -18.409, -17.936, -17.439, -16.852, -16.223, -15.569,
    -14.956, -14.421, -19.201, -18.852, -18.415, -17.941, -17.444, -16.858,
    -16.228, -15.574, -14.960, -14.425, -19.207, -18.858, -18.421, -17.947,
    -17.450, -16.863, -16.234, -15.579, -14.965, -14.428, -19.213, -18.864,
    -18.427, -17.953, -17.455, -16.869, -16.239, -15.585, -14.969, -14.432,
    -19.219, -18.870, -18.432, -17.959, -17.461, -16.875, -16.245, -15.590,
    -14.974, -14.435, -19.225, -18.876, -18.438, -17.964, -17.466, -16.880,
    -16.250, -15.595, -14.978, -14.439, -19.231, -18.882, -18.444, -17.970,
    -17.472, -16.886, -16.256, -15.600, -14.983, -14.443, -19.238, -18.888,
    -18.450, -17.976, -17.477, -16.891, -16.261, -15.605, -14.987, -14.446,
    -19.244, -18.894, -18.456, -17.981, -17.483, -16.897, -16.267, -15.610,
    -14.992, -14.450, -19.250, -18.900, -18.462, -17.987, -17.488, -16.902,
    -16.272, -15.615, -14.996, -14.454, -19.256, -18.906, -18.468, -17.993,
    -17.494, -16.908, -16.278, -15.620, -15.001, -14.457, -19.262, -18.912,
    -18.474, -17.999, -17.499, -16.913, -16.283, -15.626, -15.005, -14.461,
    -19.268, -18.918, -18.480, -18.004, -17.505, -16.919, -16.289, -15.631,
    -15.010, -14.465, -19.274, -18.924, -18.486, -18.010, -17.510, -16.924,
    -16.294, -15.636, -15.014, -14.468, -19.280, -18.930, -18.492, -18.016,
    -17.515, -16.930, -16.299, -15.641, -15.019, -14.472, -19.286, -18.936,
    -18.497, -18.021, -17.521, -16.935, -16.305, -15.646, -15.023, -14.476,
    -19.292, -18.942, -18.503, -18.027, -17.526, -16.941, -16.310, -15.651,
    -15.028, -14.479, -19.298, -18.948, -18.509, -18.033, -17.532, -16.946,
    -16.316, -15.656, -15.032, -14.483, -19.304, -18.954, -18.515, -18.038,
    -17.537, -16.951, -16.321, -15.661, -15.037, -14.487, -19.310, -18.960,
    -18.521, -18.044, -17.543, -16.957, -16.327, -15.667, -15.041, -14.490,
    -19.316, -18.966, -18.527, -18.050, -17.548, -16.962, -16.332, -15.672,
    -15.046, -14.494, -19.322, -18.972, -18.533, -18.056, -17.554, -16.968,
    -16.338, -15.677, -15.050, -14.498, -19.328, -18.978, -18.539, -18.061,
    -17.559, -16.973, -16.343, -15.682, -15.055, -14.501, -19.334, -18.984,
    -18.545, -18.067, -17.565, -16.979, -16.348, -15.687, -15.059, -14.505,
    -19.340, -18.990, -18.550, -18.073, -17.570, -16.984, -16.354, -15.692,
    -15.064, -14.509, -19.346, -18.996, -18.556, -18.078, -17.576, -16.990,
    -16.359, -15.697, -15.068, -14.513, -19.352, -19.002, -18.562, -18.084,
    -17.581, -16.995, -16.365, -15.702, -15.073, -14.516, -19.358, -19.008,
    -18.568, -18.090, -17.586, -17.001, -16.370, -15.707, -15.077, -14.520,
    -19.364, -19.014, -18.574, -18.095, -17.592, -17.006, -16.375, -15.712,
    -15.082, -14.524, -19.370, -19.020, -18.580, -18.101, -17.597, -17.011,
    -16.381, -15.717, -15.086, -14.527, -19.376, -19.026, -18.586, -18.107,
    -17.603, -17.017, -16.386, -15.723, -15.091, -14.531, -19.382, -19.032,
    -18.592, -18.112, -17.608, -17.022, -16.392, -15.728, -15.095, -14.535,
    -19.388, -19.038, -18.598, -18.118, -17.614, -17.028, -16.397, -15.733,
    -15.100, -14.539, -19.395, -19.044, -18.603, -18.124, -17.619, -17.033,
    -16.402, -15.738, -15.104, -14.542, -19.401, -19.050, -18.609, -18.129,
    -17.625, -17.038, -16.408, -15.743, -15.108, -14.546, -19.407, -19.056,
    -18.615, -18.135, -17.630, -17.044, -16.413, -15.748, -15.113, -14.550,
    -19.413, -19.062, -18.621, -18.141, -17.635, -17.049, -16.419, -15.753,
    -15.117, -14.553, -19.419, -19.068, -18.627, -18.147, -17.641, -17.055,
    -16.424, -15.758, -15.122, -14.557, -19.425, -19.074, -18.633, -18.152,
    -17.646, -17.060, -16.429, -15.763, -15.126, -14.561, -19.431, -19.080,
    -18.639, -18.158, -17.652, -17.065, -16.435, -15.768, -15.131, -14.565,
    -19.437, -19.086, -18.645, -18.164, -17.657, -17.071, -16.440, -15.773,
    -15.135, -14.568, -19.443, -19.092, -18.650, -18.169, -17.663, -17.076,
    -16.445, -15.778, -15.140, -14.572, -19.449, -19.098, -18.656, -18.175,
    -17.668, -17.082, -16.451, -15.783, -15.144, -14.576, -19.455, -19.104,
    -18.662, -18.181, -17.673, -17.087, -16.456, -15.788, -15.149, -14.580,
    -19.461, -19.110, -18.668, -18.186, -17.679, -17.092, -16.461, -15.793,
    -15.153, -14.583, -19.467, -19.116, -18.674, -18.192, -17.684, -17.098,
    -16.467, -15.798, -15.158, -14.587, -19.473, -19.122, -18.680, -18.198,
    -17.690, -17.103, -16.472, -15.803, -15.162, -14.591, -19.479, -19.128,
    -18.686, -18.203, -17.695, -17.109, -16.477, -15.808, -15.167, -14.594,
    -19.485, -19.134, -18.691, -18.209, -17.701, -17.114, -16.483, -15.813,
    -15.171, -14.598, -19.491, -19.140, -18.697, -18.215, -17.706, -17.119,
    -16.488, -15.818, -15.176, -14.602, -19.497, -19.146, -18.703, -18.220,
    -17.711, -17.125, -16.493, -15.823, -15.180, -14.606, -19.503, -19.152,
    -18.709, -18.226, -17.717, -17.130, -16.499, -15.828, -15.185, -14.609,
    -19.509, -19.158, -18.715, -18.232, -17.722, -17.135, -16.504, -15.833,
    -15.189, -14.613, -19.515, -19.164, -18.721, -18.237, -17.728, -17.141,
    -16.509, -15.838, -15.193, -14.617, -19.521, -19.170, -18.727, -18.243,
    -17.733, -17.146, -16.514, -15.843, -15.198, -14.621, -19.527, -19.176,
    -18.732, -18.249, -17.738, -17.151, -16.520, -15.848, -15.202, -14.624,
    -19.533, -19.182, -18.738, -18.254, -17.744, -17.157, -16.525, -15.853,
    -15.207, -14.628, -19.539, -19.188, -18.744, -18.260, -17.749, -17.162,
    -16.530, -15.858, -15.211, -14.632, -19.545, -19.194, -18.750, -18.265,
    -17.755, -17.167, -16.536, -15.863, -15.216, -14.636, -19.551, -19.200,
    -18.756, -18.271, -17.760, -17.173, -16.541, -15.868, -15.220, -14.639,
    -19.557, -19.206, -18.762, -18.277, -17.765, -17.178, -16.546, -15.873,
    -15.225, -14.643, -19.563, -19.212, -18.768, -18.282, -17.771, -17.183,
    -16.551, -15.878, -15.229, -14.647, -19.569, -19.218, -18.773, -18.288,
    -17.776, -17.189, -16.557, -15.883, -15.233, -14.651, -19.575, -19.224,
    -18.779, -18.294, -17.782, -17.194, -16.562, -15.888, -15.238, -14.654,
    -19.581, -19.230, -18.785, -18.299, -17.787, -17.199, -16.567, -15.893,
    -15.242, -14.658, -19.587, -19.236, -18.791, -18.305, -17.792, -17.205,
    -16.572, -15.897, -15.247, -14.662, -19.593, -19.242, -18.797, -18.311,
    -17.798, -17.210, -16.578, -15.902, -15.251, -14.666, -19.599, -19.248,
    -18.803, -18.316, -17.803, -17.215, -16.583, -15.907, -15.256, -14.669,
    -19.605, -19.254, -18.809, -18.322, -17.809, -17.221, -16.588, -15.912,
    -15.260, -14.673, -19.611, -19.260, -18.814, -18.328, -17.814, -17.226,
    -16.593, -15.917, -15.264, -14.677, -19.617, -19.266, -18.820, -18.333,
    -17.819, -17.231, -16.599, -15.922, -15.269, -14.681, -19.623, -19.272,
    -18.826, -18.339, -17.825, -17.236, -16.604, -15.927, -15.273, -14.684,
    -19.629, -19.278, -18.832, -18.344, -17.830, -17.242, -16.609, -15.932,
    -15.278, -14.688, -19.635, -19.284, -18.838, -18.350, -17.836, -17.247,
    -16.614, -15.937, -15.282, -14.692, -19.641, -19.290, -18.844, -18.356,
    -17.841, -17.252, -16.619, -15.942, -15.287, -14.696, -19.647, -19.296,
    -18.849, -18.361, -17.846, -17.258, -16.625, -15.947, -15.291, -14.699,
    -19.653, -19.302, -18.855, -18.367, -17.852, -17.263, -16.630, -15.951,
    -15.295, -14.703, -19.660, -19.308, -18.861, -18.373, -17.857, -17.268,
    -16.635, -15.956, -15.300, -14.707, -19.666, -19.314, -18.867, -18.378,
    -17.862, -17.273, -16.640, -15.961, -15.304, -14.711, -19.672, -19.320,
    -18.873, -18.384, -17.868, -17.279, -16.645, -15.966, -15.309, -14.714,
    -19.678, -19.326, -18.879, -18.390, -17.873, -17.284, -16.651, -15.971,
    -15.313, -14.718, -19.684, -19.332, -18.884, -18.395, -17.879, -17.289,
    -16.656, -15.976, -15.317, -14.722, -19.690, -19.338, -18.890, -18.401,
    -17.884, -17.294, -16.661, -15.981, -15.322, -14.726, -19.696, -19.343,
    -18.896, -18.406, -17.889, -17.300, -16.666, -15.986, -15.326, -14.730,
    -19.702, -19.349, -18.902, -18.412, -17.895, -17.305, -16.671, -15.990,
    -15.331, -14.733, -19.708, -19.355, -18.908, -18.418, -17.900, -17.310,
    -16.677, -15.995, -15.335, -14.737, -19.714, -19.361, -18.914, -18.423,
    -17.905, -17.316, -16.682, -16.000, -15.339, -14.741, -19.720, -19.367,
    -18.919, -18.429, -17.911, -17.321, -16.687, -16.005, -15.344, -14.745,
    -19.726, -19.373, -18.925, -18.435, -17.916, -17.326, -16.692, -16.010,
    -15.348, -14.748, -19.732, -19.379, -18.931, -18.440, -17.922, -17.331,
    -16.697, -16.015, -15.353, -14.752, -19.738, -19.385, -18.937, -18.446,
    -17.927, -17.336, -16.702, -16.020, -15.357, -14.756, -19.744, -19.391,
    -18.943, -18.451, -17.932, -17.342, -16.708, -16.024, -15.361, -14.760,
    -19.750, -19.397, -18.949, -18.457, -17.938, -17.347, -16.713, -16.029,
    -15.366, -14.763, -19.756, -19.403, -18.954, -18.463, -17.943, -17.352,
    -16.718, -16.034, -15.370, -14.767, -19.762, -19.409, -18.960, -18.468,
    -17.948, -17.357, -16.723, -16.039, -15.375, -14.771, -19.768, -19.415,
    -18.966, -18.474, -17.954, -17.363, -16.728, -16.044, -15.379, -14.775,
    -19.774, -19.421, -18.972, -18.479, -17.959, -17.368, -16.733, -16.049,
    -15.383, -14.778, -19.780, -19.427, -18.978, -18.485, -17.964, -17.373,
    -16.738, -16.054, -15.388, -14.782, -19.786, -19.433, -18.983, -18.491,
    -17.970, -17.378, -16.743, -16.058, -15.392, -14.786, -19.792, -19.439,
    -18.989, -18.496, -17.975, -17.384, -16.749, -16.063, -15.397, -14.790,
    -19.798, -19.445, -18.995, -18.502, -17.981, -17.389, -16.754, -16.068,
    -15.401, -14.794, -19.804, -19.451, -19.001, -18.507, -17.986, -17.394,
    -16.759, -16.073, -15.405, -14.797, -19.810, -19.457, -19.007, -18.513,
    -17.991, -17.399, -16.764, -16.078, -15.410, -14.801, -19.816, -19.463,
    -19.013, -18.519, -17.997, -17.404, -16.769, -16.083, -15.414, -14.805,
    -19.822, -19.469, -19.018, -18.524, -18.002, -17.410, -16.774, -16.087,
    -15.419, -14.809, -19.828, -19.475, -19.024, -18.530, -18.007, -17.415,
    -16.779, -16.092, -15.423, -14.812, -19.834, -19.481, -19.030, -18.536,
    -18.013, -17.420, -16.784, -16.097, -15.427, -14.816, -19.840, -19.487,
    -19.036, -18.541, -18.018, -17.425, -16.789, -16.102, -15.432, -14.820,
    -19.846, -19.492, -19.042, -18.547, -18.023, -17.431, -16.795, -16.107,
    -15.436, -14.824, -19.852, -19.498, -19.047, -18.552, -18.029, -17.436,
    -16.800, -16.112, -15.441, -14.828, -19.858, -19.504, -19.053, -18.558,
    -18.034, -17.441, -16.805, -16.116, -15.445, -14.831, -19.864, -19.510,
    -19.059, -18.564, -18.039, -17.446, -16.810, -16.121, -15.449, -14.835,
    -19.870, -19.516, -19.065, -18.569, -18.045, -17.451, -16.815, -16.126,
    -15.454, -14.839, -19.876, -19.522, -19.071, -18.575, -18.050, -17.457,
    -16.820, -16.131, -15.458, -14.843, -19.882, -19.528, -19.076, -18.580,
    -18.055, -17.462, -16.825, -16.136, -15.463, -14.847, -19.888, -19.534,
    -19.082, -18.586, -18.061, -17.467, -16.830, -16.140, -15.467, -14.850,
    -19.894, -19.540, -19.088, -18.592, -18.066, -17.472, -16.835, -16.145,
    -15.471, -14.854, -19.900, -19.546, -19.094, -18.597, -18.071, -17.477,
    -16.840, -16.150, -15.476, -14.858, -19.906, -19.552, -19.100, -18.603,
    -18.077, -17.482, -16.845, -16.155, -15.480, -14.862, -19.912, -19.558,
    -19.106, -18.608, -18.082, -17.488, -16.850, -16.160, -15.485, -14.865,
    -19.918, -19.564, -19.111, -18.614, -18.087, -17.493, -16.855, -16.165,
    -15.489, -14.869, -19.924, -19.570, -19.117, -18.619, -18.093, -17.498,
    -16.860, -16.169, -15.493, -14.873, -19.930, -19.576, -19.123, -18.625,
    -18.098, -17.503, -16.865, -16.174, -15.498, -14.877, -19.936, -19.582,
    -19.129, -18.631, -18.103, -17.508, -16.871, -16.179, -15.502, -14.881,
    -19.942, -19.588, -19.135, -18.636, -18.109, -17.514, -16.876, -16.184,
    -15.507, -14.884, -19.948, -19.594, -19.140, -18.642, -18.114, -17.519,
    -16.881, -16.189, -15.511, -14.888, -19.954, -19.600, -19.146, -18.647,
    -18.119, -17.524, -16.886, -16.194, -15.516, -14.892, -19.960, -19.605,
    -19.152, -18.653, -18.125, -17.529, -16.891, -16.198, -15.520, -14.896,
    -19.966, -19.611, -19.158, -18.659, -18.130, -17.534, -16.896, -16.203,
    -15.524, -14.900, -19.972, -19.617, -19.164, -18.664, -18.135, -17.539,
    -16.901, -16.208, -15.529, -14.903, -19.978, -19.623, -19.169, -18.670,
    -18.141, -17.544, -16.906, -16.213, -15.533, -14.907, -19.984, -19.629,
    -19.175, -18.675, -18.146, -17.550, -16.911, -16.218, -15.538, -14.911,
    -19.990, -19.635, -19.181, -18.681, -18.151, -17.555, -16.916, -16.222,
    -15.542, -14.915, -19.996, -19.641, -19.187, -18.687, -18.157, -17.560,
    -16.921, -16.227, -15.546, -14.919, -20.002, -19.647, -19.192, -18.692,
    -18.162, -17.565, -16.926, -16.232, -15.551, -14.923, -20.008, -19.653,
    -19.198, -18.698, -18.167, -17.570, -16.931, -16.237, -15.555, -14.926,
    -20.014, -19.659, -19.204, -18.703, -18.173, -17.575, -16.936, -16.242,
    -15.560, -14.930, -20.020, -19.665, -19.210, -18.709, -18.178, -17.581,
    -16.941, -16.247, -15.564, -14.934, -20.026, -19.671, -19.216, -18.714,
    -18.183, -17.586, -16.946, -16.251, -15.569, -14.938, -20.032, -19.677,
    -19.221, -18.720, -18.189, -17.591, -16.951, -16.256, -15.573, -14.942,
    -20.038, -19.683, -19.227, -18.726, -18.194, -17.596, -16.956, -16.261,
    -15.577, -14.945, -20.044, -19.689, -19.233, -18.731, -18.199, -17.601,
    -16.961, -16.266, -15.582, -14.949, -20.050, -19.695, -19.239, -18.737,
    -18.205, -17.606, -16.966, -16.271, -15.586, -14.953, -20.056, -19.700,
    -19.245, -18.742, -18.210, -17.611, -16.971, -16.275, -15.591, -14.957,
    -20.062, -19.706, -19.250, -18.748, -18.215, -17.617, -16.976, -16.280,
    -15.595, -14.961, -20.068, -19.712, -19.256, -18.753, -18.221, -17.622,
    -16.981, -16.285, -15.599, -14.964, -20.074, -19.718, -19.262, -18.759,
    -18.226, -17.627, -16.986, -16.290, -15.604, -14.968, -20.080, -19.724,
    -19.268, -18.765, -18.231, -17.632, -16.991, -16.295, -15.608, -14.972,
    -20.086, -19.730, -19.274, -18.770, -18.236, -17.637, -16.996, -16.299,
    -15.613, -14.976, -20.092, -19.736, -19.279, -18.776, -18.242, -17.642,
    -17.001, -16.304, -15.617, -14.980, -20.098, -19.742, -19.285, -18.781,
    -18.247, -17.647, -17.006, -16.309, -15.622, -14.984, -20.104, -19.748,
    -19.291, -18.787, -18.252, -17.652, -17.011, -16.314, -15.626, -14.987,
    -20.110, -19.754, -19.297, -18.792, -18.258, -17.658, -17.016, -16.319,
    -15.630, -14.991, -20.116, -19.760, -19.302, -18.798, -18.263, -17.663,
    -17.021, -16.323, -15.635, -14.995, -20.122, -19.766, -19.308, -18.804,
    -18.268, -17.668, -17.026, -16.328, -15.639, -14.999, -20.128, -19.772,
    -19.314, -18.809, -18.274, -17.673, -17.031, -16.333, -15.644, -15.003,
    -20.134, -19.778, -19.320, -18.815, -18.279, -17.678, -17.036, -16.338,
    -15.648, -15.006, -20.140, -19.784, -19.326, -18.820, -18.284, -17.683,
    -17.040, -16.343, -15.652, -15.010, -20.146, -19.789, -19.331, -18.826,
    -18.290, -17.688, -17.045, -16.347, -15.657, -15.014, -20.152, -19.795,
    -19.337, -18.831, -18.295, -17.693, -17.050, -16.352, -15.661, -15.018,
    -20.157, -19.801, -19.343, -18.837, -18.300, -17.699, -17.055, -16.357,
    -15.666, -15.022, -20.163, -19.807, -19.349, -18.842, -18.305, -17.704,
    -17.060, -16.362, -15.670, -15.026, -20.169, -19.813, -19.354, -18.848,
    -18.311, -17.709, -17.065, -16.367, -15.675, -15.029, -20.175, -19.819,
    -19.360, -18.854, -18.316, -17.714, -17.070, -16.371, -15.679, -15.033,
    -20.181, -19.825, -19.366, -18.859, -18.321, -17.719, -17.075, -16.376,
    -15.683, -15.037, -20.187, -19.831, -19.372, -18.865, -18.327, -17.724,
    -17.080, -16.381, -15.688, -15.041, -20.193, -19.837, -19.378, -18.870,
    -18.332, -17.729, -17.085, -16.386, -15.692, -15.045, -20.199, -19.843,
    -19.383, -18.876, -18.337, -17.734, -17.090, -16.391, -15.697, -15.048,
    -20.205, -19.849, -19.389, -18.881, -18.343, -17.739, -17.095, -16.395,
    -15.701, -15.052, -20.211, -19.855, -19.395, -18.887, -18.348, -17.744,
    -17.100, -16.400, -15.705, -15.056, -20.217, -19.861, -19.401, -18.892,
    -18.353, -17.750, -17.105, -16.405, -15.710, -15.060, -20.223, -19.866,
    -19.406, -18.898, -18.358, -17.755, -17.110, -16.410, -15.714, -15.064,
    -20.229, -19.872, -19.412, -18.904, -18.364, -17.760, -17.115, -16.415,
    -15.719, -15.067, -20.235, -19.878, -19.418, -18.909, -18.369, -17.765,
    -17.119, -16.419, -15.723, -15.071, -20.241, -19.884, -19.424, -18.915,
    -18.374, -17.770, -17.124, -16.424, -15.728, -15.075, -20.247, -19.890,
    -19.430, -18.920, -18.380, -17.775, -17.129, -16.429, -15.732, -15.079,
    -20.253, -19.896, -19.435, -18.926, -18.385, -17.780, -17.134, -16.434,
    -15.736, -15.083, -20.259, -19.902, -19.441, -18.931, -18.390, -17.785,
    -17.139, -16.438, -15.741, -15.087, -20.265, -19.908, -19.447, -18.937,
    -18.395, -17.790, -17.144, -16.443, -15.745, -15.090, -20.271, -19.914,
    -19.453, -18.942, -18.401, -17.795, -17.149, -16.448, -15.750, -15.094,
    -20.277, -19.920, -19.458, -18.948, -18.406, -17.800, -17.154, -16.453,
    -15.754, -15.098, -20.283, -19.926, -19.464, -18.954, -18.411, -17.806,
    -17.159, -16.458, -15.758, -15.102, -20.289, -19.932, -19.470, -18.959,
    -18.417, -17.811, -17.164, -16.462, -15.763, -15.106, -20.295, -19.937,
    -19.476, -18.965, -18.422, -17.816, -17.169, -16.467, -15.767, -15.109,
    -20.301, -19.943, -19.481, -18.970, -18.427, -17.821, -17.173, -16.472,
    -15.771, -15.113, -20.307, -19.949, -19.487, -18.976, -18.432, -17.826,
    -17.178, -16.477, -15.776, -15.117, -20.313, -19.955, -19.493, -18.981,
    -18.438, -17.831, -17.183, -16.481, -15.780, -15.121, -20.319, -19.961,
    -19.499, -18.987, -18.443, -17.836, -17.188, -16.486, -15.785, -15.124,
    -20.325, -19.967, -19.504, -18.992, -18.448, -17.841, -17.193, -16.491,
    -15.789, -15.128, -20.331, -19.973, -19.510, -18.998, -18.454, -17.846,
    -17.198, -16.496, -15.793, -15.132, -20.337, -19.979, -19.516, -19.003,
    -18.459, -17.851, -17.203, -16.500, -15.798, -15.136, -20.343, -19.985,
    -19.522, -19.009, -18.464, -17.856, -17.208, -16.505, -15.802, -15.140,
    -20.349, -19.991, -19.527, -19.014, -18.469, -17.861, -17.213, -16.510,
    -15.807, -15.143, -20.355, -19.997, -19.533, -19.020, -18.475, -17.866,
    -17.217, -16.515, -15.811, -15.147, -20.361, -20.003, -19.539, -19.026,
    -18.480, -17.871, -17.222, -16.519, -15.815, -15.151, -20.367, -20.008,
    -19.545, -19.031, -18.485, -17.877, -17.227, -16.524, -15.820, -15.155,
    -20.373, -20.014, -19.551, -19.037, -18.491, -17.882, -17.232, -16.529,
    -15.824, -15.158, -20.379, -20.020, -19.556, -19.042, -18.496, -17.887,
    -17.237, -16.534, -15.828, -15.162, -20.385, -20.026, -19.562, -19.048,
    -18.501, -17.892, -17.242, -16.538, -15.833, -15.166, -20.391, -20.032,
    -19.568, -19.053, -18.506, -17.897, -17.247, -16.543, -15.837, -15.170,
    -20.397, -20.038, -19.574, -19.059, -18.512, -17.902, -17.252, -16.548,
    -15.841, -15.174, -20.403, -20.044, -19.579, -19.064, -18.517, -17.907,
    -17.256, -16.553, -15.846, -15.177, -20.409, -20.050, -19.585, -19.070,
    -18.522, -17.912, -17.261, -16.557, -15.850, -15.181, -20.415, -20.056,
    -19.591, -19.075, -18.527, -17.917, -17.266, -16.562, -15.854, -15.185,
    -20.421, -20.062, -19.597, -19.081, -18.533, -17.922, -17.271, -16.567,
    -15.859, -15.189, -20.427, -20.068, -19.602, -19.086, -18.538, -17.927,
    -17.276, -16.572, -15.863, -15.192, -20.433, -20.073, -19.608, -19.092,
    -18.543, -17.932, -17.281, -16.576, -15.868, -15.196, -20.439, -20.079,
    -19.614, -19.098, -18.549, -17.937, -17.286, -16.581, -15.872, -15.200,
    -20.445, -20.085, -19.620, -19.103, -18.554, -17.942, -17.290, -16.586,
    -15.876, -15.204, -20.451, -20.091, -19.625, -19.109, -18.559, -17.947,
    -17.295, -16.590, -15.881, -15.207, -20.457, -20.097, -19.631, -19.114,
    -18.564, -17.952, -17.300, -16.595, -15.885, -15.211, -20.463, -20.103,
    -19.637, -19.120, -18.570, -17.957, -17.305, -16.600, -15.889, -15.215,
    -20.469, -20.109, -19.643, -19.125, -18.575, -17.962, -17.310, -16.605,
    -15.894, -15.219, -20.475, -20.115, -19.648, -19.131, -18.580, -17.967,
    -17.315, -16.609, -15.898, -15.222, -20.481, -20.121, -19.654, -19.136,
    -18.585, -17.973, -17.320, -16.614, -15.902, -15.226, -20.486, -20.127,
    -19.660, -19.142, -18.591, -17.978, -17.324, -16.619, -15.906, -15.230,
    -20.492, -20.133, -19.666, -19.147, -18.596, -17.983, -17.329, -16.623,
    -15.911, -15.233, -20.498, -20.138, -19.671, -19.153, -18.601, -17.988,
    -17.334, -16.628, -15.915, -15.237, -20.504, -20.144, -19.677, -19.158,
    -18.606, -17.993, -17.339, -16.633, -15.919, -15.241, -20.510, -20.150,
    -19.683, -19.164, -18.612, -17.998, -17.344, -16.638, -15.924, -15.245,
    -20.516, -20.156, -19.689, -19.169, -18.617, -18.003, -17.349, -16.642,
    -15.928, -15.248, -20.522, -20.162, -19.694, -19.175, -18.622, -18.008,
    -17.353, -16.647, -15.932, -15.252, -20.528, -20.168, -19.700, -19.180,
    -18.627, -18.013, -17.358, -16.652, -15.937, -15.256, -20.534, -20.174,
    -19.706, -19.186, -18.633, -18.018, -17.363, -16.656, -15.941, -15.259,
    -20.540, -20.180, -19.712, -19.191, -18.638, -18.023, -17.368, -16.661,
    -15.945, -15.263, -20.546, -20.186, -19.717, -19.197, -18.643, -18.028,
    -17.373, -16.666, -15.949, -15.267, -20.552, -20.192, -19.723, -19.203,
    -18.648, -18.033, -17.378, -16.670, -15.954, -15.271, -20.558, -20.197,
    -19.729, -19.208, -18.654, -18.038, -17.382, -16.675, -15.958, -15.274,
    -20.564, -20.203, -19.734, -19.214, -18.659, -18.043, -17.387, -16.680,
    -15.962, -15.278, -20.570, -20.209, -19.740, -19.219, -18.664, -18.048,
    -17.392, -16.684, -15.967, -15.282, -20.576, -20.215, -19.746, -19.225,
    -18.670, -18.053, -17.397, -16.689, -15.971, -15.285, -20.582, -20.221,
    -19.752, -19.230, -18.675, -18.058, -17.402, -16.694, -15.975, -15.289,
    -20.588, -20.227, -19.757, -19.236, -18.680, -18.063, -17.406, -16.699,
    -15.979, -15.293, -20.594, -20.233, -19.763, -19.241, -18.685, -18.068,
    -17.411, -16.703, -15.984, -15.296, -20.600, -20.239, -19.769, -19.247,
    -18.691, -18.073, -17.416, -16.708, -15.988, -15.300, -20.606, -20.245,
    -19.775, -19.252, -18.696, -18.078, -17.421, -16.713, -15.992, -15.304,
    -20.612, -20.251, -19.780, -19.258, -18.701, -18.083, -17.426, -16.717,
    -15.996, -15.307, -20.618, -20.256, -19.786, -19.263, -18.706, -18.088,
    -17.431, -16.722, -16.001, -15.311, -20.624, -20.262, -19.792, -19.269,
    -18.712, -18.093, -17.435, -16.727, -16.005, -15.315, -20.630, -20.268,
    -19.798, -19.274, -18.717, -18.098, -17.440, -16.731, -16.009, -15.318,
    -20.636, -20.274, -19.803, -19.280, -18.722, -18.103, -17.445, -16.736,
    -16.013, -15.322, -20.642, -20.280, -19.809, -19.285, -18.727, -18.108,
    -17.450, -16.741, -16.018, -15.326, -20.648, -20.286, -19.815, -19.291,
    -18.733, -18.113, -17.455, -16.745, -16.022, -15.329, -20.654, -20.292,
    -19.821, -19.296, -18.738, -18.118, -17.459, -16.750, -16.026, -15.333,
    -20.660, -20.298, -19.826, -19.302, -18.743, -18.123, -17.464, -16.754,
    -16.030, -15.337, -20.666, -20.304, -19.832, -19.307, -18.748, -18.128,
    -17.469, -16.759, -16.035, -15.340, -20.672, -20.310, -19.838, -19.313,
    -18.753, -18.133, -17.474, -16.764, -16.039, -15.344, -20.678, -20.315,
    -19.843, -19.318, -18.759, -18.138, -17.479, -16.768, -16.043, -15.348,
    -20.684, -20.321, -19.849, -19.324, -18.764, -18.143, -17.483, -16.773,
    -16.047, -15.351, -20.690, -20.327, -19.855, -19.329, -18.769, -18.148,
    -17.488, -16.778, -16.052, -15.355, -20.696, -20.333, -19.861, -19.335,
    -18.774, -18.153, -17.493, -16.782, -16.056, -15.359, -20.701, -20.339,
    -19.866, -19.340, -18.780, -18.158, -17.498, -16.787, -16.060, -15.362,
    -20.707, -20.345, -19.872, -19.346, -18.785, -18.163, -17.502, -16.792,
    -16.064, -15.366, -20.713, -20.351, -19.878, -19.351, -18.790, -18.168,
    -17.507, -16.796, -16.068, -15.370, -20.719, -20.357, -19.884, -19.357,
    -18.795, -18.173, -17.512, -16.801, -16.073, -15.373, -20.725, -20.363,
    -19.889, -19.362, -18.801, -18.178, -17.517, -16.806, -16.077, -15.377,
    -20.731, -20.368, -19.895, -19.368, -18.806, -18.183, -17.522, -16.810,
    -16.081, -15.381, -20.737, -20.374, -19.901, -19.373, -18.811, -18.188,
    -17.526, -16.815, -16.085, -15.384, -20.743, -20.380, -19.907, -19.379,
    -18.816, -18.193, -17.531, -16.819, -16.089, -15.388, -20.749, -20.386,
    -19.912, -19.384, -18.822, -18.198, -17.536, -16.824, -16.094, -15.392,
    -20.755, -20.392, -19.918, -19.390, -18.827, -18.203, -17.541, -16.829,
    -16.098, -15.395, -20.761, -20.398, -19.924, -19.395, -18.832, -18.208,
    -17.545, -16.833, -16.102, -15.399, -20.767, -20.404, -19.929, -19.401,
    -18.837, -18.213, -17.550, -16.838, -16.106, -15.403, -20.773, -20.410,
    -19.935, -19.406, -18.843, -18.218, -17.555, -16.843, -16.110, -15.406,
    -20.779, -20.416, -19.941, -19.412, -18.848, -18.223, -17.560, -16.847,
    -16.115, -15.410, -20.785, -20.421, -19.947, -19.417, -18.853, -18.228,
    -17.565, -16.852, -16.119, -15.414, -20.791, -20.427, -19.952, -19.423,
    -18.858, -18.233, -17.569, -16.856, -16.123, -15.417, -20.797, -20.433,
    -19.958, -19.428, -18.864, -18.238, -17.574, -16.861, -16.127, -15.421,
    -20.803, -20.439, -19.964, -19.434, -18.869, -18.243, -17.579, -16.866,
    -16.131, -15.424, -20.809, -20.445, -19.970, -19.439, -18.874, -18.248,
    -17.584, -16.870, -16.136, -15.428, -20.815, -20.451, -19.975, -19.445,
    -18.879, -18.253, -17.588, -16.875, -16.140, -15.432, -20.821, -20.457,
    -19.981, -19.450, -18.884, -18.258, -17.593, -16.879, -16.144, -15.435,
    -20.827, -20.463, -19.987, -19.456, -18.890, -18.263, -17.598, -16.884,
    -16.148, -15.439, -20.833, -20.469, -19.992, -19.461, -18.895, -18.268,
    -17.603, -16.889, -16.152, -15.443, -20.839, -20.474, -19.998, -19.467,
    -18.900, -18.273, -17.607, -16.893, -16.156, -15.446, -20.845, -20.480,
    -20.004, -19.472, -18.905, -18.278, -17.612, -16.898, -16.161, -15.450,
    -20.851, -20.486, -20.010, -19.478, -18.911, -18.283, -17.617, -16.902,
    -16.165, -15.454, -20.857, -20.492, -20.015, -19.483, -18.916, -18.288,
    -17.622, -16.907, -16.169, -15.457, -20.863, -20.498, -20.021, -19.489,
    -18.921, -18.293, -17.626, -16.912, -16.173, -15.461, -20.869, -20.504,
    -20.027, -19.494, -18.926, -18.298, -17.631, -16.916, -16.177, -15.465,
    -20.874, -20.510, -20.032, -19.500, -18.932, -18.303, -17.636, -16.921,
    -16.182, -15.468, -20.880, -20.516, -20.038, -19.505, -18.937, -18.308,
    -17.641, -16.925, -16.186, -15.472, -20.886, -20.522, -20.044, -19.511,
    -18.942, -18.313, -17.646, -16.930, -16.190, -15.476, -20.892, -20.527,
    -20.050, -19.516, -18.947, -18.318, -17.650, -16.935, -16.194, -15.479,
    -20.898, -20.533, -20.055, -19.522, -18.952, -18.323, -17.655, -16.939,
    -16.198, -15.483, -20.904, -20.539, -20.061, -19.527, -18.958, -18.328,
    -17.660, -16.944, -16.202, -15.487, -20.910, -20.545, -20.067, -19.533,
    -18.963, -18.333, -17.665, -16.948, -16.207, -15.490, -20.916, -20.551,
    -20.072, -19.538, -18.968, -18.338, -17.669, -16.953, -16.211, -15.494,
    -20.922, -20.557, -20.078, -19.544, -18.973, -18.343, -17.674, -16.957,
    -16.215, -15.497, -20.928, -20.563, -20.084, -19.549, -18.979, -18.348,
    -17.679, -16.962, -16.219, -15.501, -20.934, -20.569, -20.090, -19.555,
    -18.984, -18.353, -17.683, -16.967, -16.223, -15.505, -20.940, -20.574,
    -20.095, -19.560, -18.989, -18.358, -17.688, -16.971, -16.228, -15.508,
    -20.946, -20.580, -20.101, -19.566, -18.994, -18.363, -17.693, -16.976,
    -16.232, -15.512, -20.952, -20.586, -20.107, -19.571, -18.999, -18.368,
    -17.698, -16.980, -16.236, -15.516, -20.958, -20.592, -20.112, -19.577,
    -19.005, -18.373, -17.702, -16.985, -16.240, -15.520, -20.964, -20.598,
    -20.118, -19.582, -19.010, -18.378, -17.707, -16.989, -16.244, -15.523,
    -20.970, -20.604, -20.124, -19.588, -19.015, -18.383, -17.712, -16.994,
    -16.248, -15.527, -20.976, -20.610, -20.130, -19.593, -19.020, -18.388,
    -17.717, -16.999, -16.253, -15.531, -20.982, -20.616, -20.135, -19.599,
    -19.026, -18.393, -17.721, -17.003, -16.257, -15.534, -20.988, -20.622,
    -20.141, -19.604, -19.031, -18.398, -17.726, -17.008, -16.261, -15.538,
    -20.994, -20.627, -20.147, -19.610, -19.036, -18.403, -17.731, -17.012,
    -16.265, -15.542, -21.000, -20.633, -20.152, -19.615, -19.041, -18.408,
    -17.736, -17.017, -16.269, -15.545, -21.006, -20.639, -20.158, -19.621,
    -19.046, -18.413, -17.740, -17.021, -16.274, -15.549, -21.012, -20.645,
    -20.164, -19.626, -19.052, -18.418, -17.745, -17.026, -16.278, -15.553,
    -21.018, -20.651, -20.170, -19.632, -19.057, -18.423, -17.750, -17.030,
    -16.282, -15.556, -21.024, -20.657, -20.175, -19.637, -19.062, -18.428,
    -17.754, -17.035, -16.286, -15.560, -21.030, -20.663, -20.181, -19.642,
    -19.067, -18.432, -17.759, -17.040, -16.290, -15.564, -21.035, -20.669,
    -20.187, -19.648, -19.072, -18.437, -17.764, -17.044, -16.294, -15.568,
    -21.041, -20.674, -20.192, -19.653, -19.078, -18.442, -17.769, -17.049,
    -16.299, -15.571, -21.047, -20.680, -20.198, -19.659, -19.083, -18.447,
    -17.773, -17.053, -16.303, -15.575, -21.053, -20.686, -20.204, -19.664,
    -19.088, -18.452, -17.778, -17.058, -16.307, -15.579, -21.059, -20.692,
    -20.210, -19.670, -19.093, -18.457, -17.783, -17.062, -16.311, -15.582,
    -21.065, -20.698, -20.215, -19.675, -19.099, -18.462, -17.788, -17.067,
    -16.315, -15.586, -21.071, -20.704, -20.221, -19.681, -19.104, -18.467,
    -17.792, -17.071, -16.320, -15.590, -21.077, -20.710, -20.227, -19.686,
    -19.109, -18.472, -17.797, -17.076, -16.324, -15.594, -21.083, -20.716,
    -20.232, -19.692, -19.114, -18.477, -17.802, -17.080, -16.328, -15.597,
    -21.089, -20.721, -20.238, -19.697, -19.119, -18.482, -17.806, -17.085,
    -16.332, -15.601, -21.095, -20.727, -20.244, -19.703, -19.125, -18.487,
    -17.811, -17.089, -16.336, -15.605, -21.101, -20.733, -20.250, -19.708,
    -19.130, -18.492, -17.816, -17.094, -16.341, -15.609, -21.107, -20.739,
    -20.255, -19.714, -19.135, -18.497, -17.821, -17.099, -16.345, -15.612,
    -21.113, -20.745, -20.261, -19.719, -19.140, -18.502, -17.825, -17.103,
    -16.349, -15.616, -21.119, -20.751, -20.267, -19.725, -19.145, -18.507,
    -17.830, -17.108, -16.353, -15.620, -21.125, -20.757, -20.272, -19.730,
    -19.151, -18.512, -17.835, -17.112, -16.357, -15.624, -21.131, -20.763,
    -20.278, -19.736, -19.156, -18.517, -17.839, -17.117, -16.362, -15.627,
    -21.137, -20.768, -20.284, -19.741, -19.161, -18.522, -17.844, -17.121,
    -16.366, -15.631, -21.143, -20.774, -20.289, -19.747, -19.166, -18.527,
    -17.849, -17.126, -16.370, -15.635, -21.149, -20.780, -20.295, -19.752,
    -19.171, -18.532, -17.854, -17.130, -16.374, -15.639, -21.155, -20.786,
    -20.301, -19.758, -19.177, -18.537, -17.858, -17.135, -16.378, -15.642,
    -21.161, -20.792, -20.307, -19.763, -19.182, -18.542, -17.863, -17.139,
    -16.383, -15.646, -21.167, -20.798, -20.312, -19.768, -19.187, -18.546,
    -17.868, -17.144, -16.387, -15.650, -21.173, -20.804, -20.318, -19.774,
    -19.192, -18.551, -17.872, -17.148, -16.391, -15.654, -21.178, -20.810,
    -20.324, -19.779, -19.197, -18.556, -17.877, -17.153, -16.395, -15.657,
    -21.184, -20.815, -20.329, -19.785, -19.203, -18.561, -17.882, -17.157,
    -16.399, -15.661, -21.190, -20.821, -20.335, -19.790, -19.208, -18.566,
    -17.886, -17.162, -16.404, -15.665, -21.196, -20.827, -20.341, -19.796,
    -19.213, -18.571, -17.891, -17.166, -16.408, -15.669, -21.202, -20.833,
    -20.346, -19.801, -19.218, -18.576, -17.896, -17.171, -16.412, -15.673,
    -21.208, -20.839, -20.352, -19.807, -19.224, -18.581, -17.901, -17.175,
    -16.416, -15.676, -21.214, -20.845, -20.358, -19.812, -19.229, -18.586,
    -17.905, -17.180, -16.421, -15.680, -21.220, -20.851, -20.364, -19.818,
    -19.234, -18.591, -17.910, -17.184, -16.425, -15.684, -21.226, -20.856,
    -20.369, -19.823, -19.239, -18.596, -17.915, -17.189, -16.429, -15.688,
    -21.232, -20.862, -20.375, -19.829, -19.244, -18.601, -17.919, -17.193,
    -16.433, -15.692, -21.238, -20.868, -20.381, -19.834, -19.250, -18.606,
    -17.924, -17.198, -16.437, -15.696, -21.244, -20.874, -20.386, -19.840,
    -19.255, -18.611, -17.929, -17.202, -16.442, -15.699, -21.250, -20.880,
    -20.392, -19.845, -19.260, -18.616, -17.933, -17.207, -16.446, -15.703,
    -21.256, -20.886, -20.398, -19.851, -19.265, -18.621, -17.938, -17.211,
    -16.450, -15.707, -21.262, -20.892, -20.403, -19.856, -19.270, -18.626,
    -17.943, -17.216, -16.454, -15.711, -21.268, -20.898, -20.409, -19.861,
    -19.276, -18.631, -17.948, -17.220, -16.459, -15.715, -21.274, -20.903,
    -20.415, -19.867, -19.281, -18.635, -17.952, -17.225, -16.463, -15.719,
    -21.280, -20.909, -20.421, -19.872, -19.286, -18.640, -17.957, -17.229,
    -16.467, -15.722, -21.286, -20.915, -20.426, -19.878, -19.291, -18.645,
    -17.962, -17.234, -16.471, -15.726, -21.292, -20.921, -20.432, -19.883,
    -19.296, -18.650, -17.966, -17.238, -16.476, -15.730, -21.298, -20.927,
    -20.438, -19.889, -19.301, -18.655, -17.971, -17.243, -16.480, -15.734,
    -21.304, -20.933, -20.443, -19.894, -19.307, -18.660, -17.976, -17.247,
    -16.484, -15.738, -21.309, -20.939, -20.449, -19.900, -19.312, -18.665,
    -17.980, -17.252, -16.488, -15.742, -21.315, -20.944, -20.455, -19.905,
    -19.317, -18.670, -17.985, -17.256, -16.493, -15.746, -21.321, -20.950,
    -20.460, -19.911, -19.322, -18.675, -17.990, -17.261, -16.497, -15.749,
    -21.327, -20.956, -20.466, -19.916, -19.327, -18.680, -17.994, -17.265,
    -16.501, -15.753, -21.333, -20.962, -20.472, -19.922, -19.333, -18.685,
    -17.999, -17.270, -16.505, -15.757, -21.339, -20.968, -20.477, -19.927,
    -19.338, -18.690, -18.004, -17.274, -16.510, -15.761, -21.345, -20.974,
    -20.483, -19.932, -19.343, -18.695, -18.008, -17.278, -16.514, -15.765,
    -21.351, -20.980, -20.489, -19.938, -19.348, -18.700, -18.013, -17.283,
    -16.518, -15.769, -21.357, -20.986, -20.495, -19.943, -19.353, -18.705,
    -18.018, -17.287, -16.522, -15.773, -21.363, -20.991, -20.500, -19.949,
    -19.359, -18.710, -18.022, -17.292, -16.526, -15.777, -21.369, -20.997,
    -20.506, -19.954, -19.364, -18.714, -18.027, -17.296, -16.531, -15.781,
    -21.375, -21.003, -20.512, -19.960, -19.369, -18.719, -18.032, -17.301,
    -16.535, -15.784, -21.381, -21.009, -20.517, -19.965, -19.374, -18.724,
    -18.036, -17.305, -16.539, -15.788, -21.387, -21.015, -20.523, -19.971,
    -19.379, -18.729, -18.041, -17.310, -16.543, -15.792, -21.393, -21.021,
    -20.529, -19.976, -19.385, -18.734, -18.046, -17.314, -16.548, -15.796,
    -21.399, -21.027, -20.534, -19.982, -19.390, -18.739, -18.051, -17.319,
    -16.552, -15.800, -21.405, -21.032, -20.540, -19.987, -19.395, -18.744,
    -18.055, -17.323, -16.556, -15.804, -21.411, -21.038, -20.546, -19.993,
    -19.400, -18.749, -18.060, -17.328, -16.560, -15.808, -21.417, -21.044,
    -20.551, -19.998, -19.405, -18.754, -18.065, -17.332, -16.565, -15.812,
    -21.423, -21.050, -20.557, -20.003, -19.411, -18.759, -18.069, -17.337,
    -16.569, -15.816, -21.429, -21.056, -20.563, -20.009, -19.416, -18.764,
    -18.074, -17.341, -16.573, -15.820, -21.434, -21.062, -20.569, -20.014,
    -19.421, -18.769, -18.079, -17.345, -16.578, -15.824, -21.440, -21.068,
    -20.574, -20.020, -19.426, -18.774, -18.083, -17.350, -16.582, -15.828,
    -21.446, -21.073, -20.580, -20.025, -19.431, -18.778, -18.088, -17.354,
    -16.586, -15.831, -21.452, -21.079, -20.586, -20.031, -19.436, -18.783,
    -18.093, -17.359, -16.590, -15.835, -21.458, -21.085, -20.591, -20.036,
    -19.442, -18.788, -18.097, -17.363, -16.595, -15.839, -21.464, -21.091,
    -20.597, -20.042, -19.447, -18.793, -18.102, -17.368, -16.599, -15.843,
    -21.470, -21.097, -20.603, -20.047, -19.452, -18.798, -18.107, -17.372,
    -16.603, -15.847, -21.476, -21.103, -20.608, -20.053, -19.457, -18.803,
    -18.111, -17.377, -16.607, -15.851, -21.482, -21.109, -20.614, -20.058,
    -19.462, -18.808, -18.116, -17.381, -16.612, -15.855, -21.488, -21.114,
    -20.620, -20.063, -19.468, -18.813, -18.121, -17.386, -16.616, -15.859,
    -21.494, -21.120, -20.625, -20.069, -19.473, -18.818, -18.125, -17.390,
    -16.620, -15.863, -21.500, -21.126, -20.631, -20.074, -19.478, -18.823,
    -18.130, -17.394, -16.624, -15.867, -21.506, -21.132, -20.637, -20.080,
    -19.483, -18.828, -18.135, -17.399, -16.629, -15.871, -21.512, -21.138,
    -20.642, -20.085, -19.488, -18.833, -18.139, -17.403, -16.633, -15.875,
    -21.518, -21.144, -20.648, -20.091, -19.494, -18.838, -18.144, -17.408,
    -16.637, -15.879, -21.524, -21.150, -20.654, -20.096, -19.499, -18.842,
    -18.148, -17.412, -16.641, -15.883, -21.530, -21.155, -20.659, -20.102,
    -19.504, -18.847, -18.153, -17.417, -16.646, -15.887, -21.536, -21.161,
    -20.665, -20.107, -19.509, -18.852, -18.158, -17.421, -16.650, -15.891,
    -21.542, -21.167, -20.671, -20.113, -19.514, -18.857, -18.162, -17.426,
    -16.654, -15.895, -21.548, -21.173, -20.677, -20.118, -19.519, -18.862,
    -18.167, -17.430, -16.658, -15.899, -21.554, -21.179, -20.682, -20.123,
    -19.525, -18.867, -18.172, -17.434, -16.663, -15.903, -21.559, -21.185,
    -20.688, -20.129, -19.530, -18.872, -18.176, -17.439, -16.667, -15.907,
    -21.565, -21.191, -20.694, -20.134, -19.535, -18.877, -18.181, -17.443,
    -16.671, -15.911, -21.571, -21.196, -20.699, -20.140, -19.540, -18.882,
    -18.186, -17.448, -16.675, -15.914, -21.577, -21.202, -20.705, -20.145,
    -19.545, -18.887, -18.190, -17.452, -16.680, -15.918, -21.583, -21.208,
    -20.711, -20.151, -19.550, -18.892, -18.195, -17.457, -16.684, -15.922,
    -21.589, -21.214, -20.716, -20.156, -19.556, -18.897, -18.200, -17.461,
    -16.688, -15.926, -21.595, -21.220, -20.722, -20.162, -19.561, -18.901,
    -18.204, -17.465, -16.692, -15.930, -21.601, -21.226, -20.728, -20.167,
    -19.566, -18.906, -18.209, -17.470, -16.697, -15.934, -21.607, -21.232,
    -20.733, -20.172, -19.571, -18.911, -18.214, -17.474, -16.701, -15.938,
    -21.613, -21.237, -20.739, -20.178, -19.576, -18.916, -18.218, -17.479,
    -16.705, -15.942, -21.619, -21.243, -20.745, -20.183, -19.582, -18.921,
    -18.223, -17.483, -16.709, -15.946, -21.625, -21.249, -20.750, -20.189,
    -19.587, -18.926, -18.228, -17.488, -16.714, -15.950, -21.631, -21.255,
    -20.756, -20.194, -19.592, -18.931, -18.232, -17.492, -16.718, -15.954,
    -21.637, -21.261, -20.762, -20.200, -19.597, -18.936, -18.237, -17.496,
    -16.722, -15.958, -21.643, -21.267, -20.767, -20.205, -19.602, -18.941,
    -18.242, -17.501, -16.726, -15.962, -21.649, -21.273, -20.773, -20.211,
    -19.607, -18.946, -18.246, -17.505, -16.731, -15.966, -21.655, -21.278,
    -20.779, -20.216, -19.613, -18.951, -18.251, -17.510, -16.735, -15.970,
    -21.661, -21.284, -20.784, -20.221, -19.618, -18.955, -18.255, -17.514,
    -16.739, -15.974, -21.667, -21.290, -20.790, -20.227, -19.623, -18.960,
    -18.260, -17.519, -16.743, -15.978, -21.672, -21.296, -20.796, -20.232,
    -19.628, -18.965, -18.265, -17.523, -16.748, -15.982, -21.678, -21.302,
    -20.801, -20.238, -19.633, -18.970, -18.269, -17.527, -16.752, -15.986,
    -21.684, -21.308, -20.807, -20.243, -19.638, -18.975, -18.274, -17.532,
    -16.756, -15.990, -21.690, -21.314, -20.813, -20.249, -19.644, -18.980,
    -18.279, -17.536, -16.760, -15.994, -21.696, -21.319, -20.818, -20.254,
    -19.649, -18.985, -18.283, -17.541, -16.765, -15.998, -21.702, -21.325,
    -20.824, -20.260, -19.654, -18.990, -18.288, -17.545, -16.769, -16.002,
    -21.708, -21.331, -20.830, -20.265, -19.659, -18.995, -18.293, -17.549,
    -16.773, -16.006, -21.714, -21.337, -20.836, -20.270, -19.664, -19.000,
    -18.297, -17.554, -16.777, -16.010, -21.720, -21.343, -20.841, -20.276,
    -19.670, -19.004, -18.302, -17.558, -16.782, -16.014, -21.726, -21.349,
    -20.847, -20.281, -19.675, -19.009, -18.307, -17.563, -16.786, -16.018,
    -21.732, -21.355, -20.853, -20.287, -19.680, -19.014, -18.311, -17.567,
    -16.790, -16.022, -21.738, -21.360, -20.858, -20.292, -19.685, -19.019,
    -18.316, -17.571, -16.794, -16.026, -21.744, -21.366, -20.864, -20.298,
    -19.690, -19.024, -18.320, -17.576, -16.799, -16.030, -21.750, -21.372,
    -20.870, -20.303, -19.695, -19.029, -18.325, -17.580, -16.803, -16.034,
    -21.756, -21.378, -20.875, -20.309, -19.701, -19.034, -18.330, -17.585,
    -16.807, -16.038, -21.762, -21.384, -20.881, -20.314, -19.706, -19.039,
    -18.334, -17.589, -16.811, -16.042, -21.768, -21.390, -20.887, -20.319,
    -19.711, -19.044, -18.339, -17.594, -16.815, -16.046, -21.774, -21.395,
    -20.892, -20.325, -19.716, -19.049, -18.344, -17.598, -16.820, -16.050,
    -21.780, -21.401, -20.898, -20.330, -19.721, -19.053, -18.348, -17.602,
    -16.824, -16.054, -21.785, -21.407, -20.904, -20.336, -19.726, -19.058,
    -18.353, -17.607, -16.828, -16.058, -21.791, -21.413, -20.909, -20.341,
    -19.732, -19.063, -18.358, -17.611, -16.832, -16.062, -21.797, -21.419,
    -20.915, -20.347, -19.737, -19.068, -18.362, -17.616, -16.837, -16.066,
    -21.803, -21.425, -20.921, -20.352, -19.742, -19.073, -18.367, -17.620,
    -16.841, -16.070, -21.809, -21.431, -20.926, -20.357, -19.747, -19.078,
    -18.371, -17.624, -16.845, -16.074, -21.815, -21.436, -20.932, -20.363,
    -19.752, -19.083, -18.376, -17.629, -16.849, -16.078, -21.821, -21.442,
    -20.938, -20.368, -19.757, -19.088, -18.381, -17.633, -16.854, -16.082,
    -21.827, -21.448, -20.943, -20.374, -19.763, -19.093, -18.385, -17.637,
    -16.858, -16.086, -21.833, -21.454, -20.949, -20.379, -19.768, -19.098,
    -18.390, -17.642, -16.862, -16.090, -21.839, -21.460, -20.955, -20.385,
    -19.773, -19.102, -18.395, -17.646, -16.866, -16.094, -21.845, -21.466,
    -20.960, -20.390, -19.778, -19.107, -18.399, -17.651, -16.870, -16.098,
    -21.851, -21.471, -20.966, -20.395, -19.783, -19.112, -18.404, -17.655,
    -16.875, -16.102, -21.857, -21.477, -20.972, -20.401, -19.788, -19.117,
    -18.408, -17.659, -16.879, -16.106, -21.863, -21.483, -20.977, -20.406,
    -19.794, -19.122, -18.413, -17.664, -16.883, -16.110, -21.869, -21.489,
    -20.983, -20.412, -19.799, -19.127, -18.418, -17.668, -16.887, -16.114,
    -21.875, -21.495, -20.989, -20.417, -19.804, -19.132, -18.422, -17.673,
    -16.892, -16.118, -21.881, -21.501, -20.994, -20.423, -19.809, -19.137,
    -18.427, -17.677, -16.896, -16.122, -21.886, -21.507, -21.000, -20.428,
    -19.814, -19.142, -18.432, -17.681, -16.900, -16.126, -21.892, -21.512,
    -21.006, -20.434, -19.819, -19.146, -18.436, -17.686, -16.904, -16.130,
    -21.898, -21.518, -21.011, -20.439, -19.825, -19.151, -18.441, -17.690,
    -16.908, -16.134, -21.904, -21.524, -21.017, -20.444, -19.830, -19.156,
    -18.445, -17.695, -16.913, -16.138, -21.910, -21.530, -21.023, -20.450,
    -19.835, -19.161, -18.450, -17.699, -16.917, -16.142, -21.916, -21.536,
    -21.028, -20.455, -19.840, -19.166, -18.455, -17.703, -16.921, -16.146,
    -21.922, -21.542, -21.034, -20.461, -19.845, -19.171, -18.459, -17.708,
    -16.925, -16.150, -21.928, -21.547, -21.040, -20.466, -19.850, -19.176,
    -18.464, -17.712, -16.930, -16.154, -21.934, -21.553, -21.045, -20.472,
    -19.855, -19.181, -18.469, -17.716, -16.934, -16.158, -21.940, -21.559,
    -21.051, -20.477, -19.861, -19.186, -18.473, -17.721, -16.938, -16.162,
    -21.946, -21.565, -21.057, -20.482, -19.866, -19.190, -18.478, -17.725,
    -16.942, -16.166, -21.952, -21.571, -21.062, -20.488, -19.871, -19.195,
    -18.482, -17.730, -16.946, -16.170, -21.958, -21.577, -21.068, -20.493,
    -19.876, -19.200, -18.487, -17.734, -16.951, -16.174, -21.964, -21.583,
    -21.074, -20.499, -19.881, -19.205, -18.492, -17.738, -16.955, -16.178,
    -21.970, -21.588, -21.079, -20.504, -19.886, -19.210, -18.496, -17.743,
    -16.959, -16.182, -21.976, -21.594, -21.085, -20.510, -19.892, -19.215,
    -18.501, -17.747, -16.963, -16.186, -21.982, -21.600, -21.091, -20.515,
    -19.897, -19.220, -18.505, -17.752, -16.967, -16.190, -21.988, -21.606,
    -21.096, -20.520, -19.902, -19.225, -18.510, -17.756, -16.972, -16.194,
    -21.993, -21.612, -21.102, -20.526, -19.907, -19.230, -18.515, -17.760,
    -16.976, -16.198, -21.999, -21.618, -21.108, -20.531, -19.912, -19.234,
    -18.519, -17.765, -16.980, -16.202, -22.005, -21.623, -21.113, -20.537,
    -19.917, -19.239, -18.524, -17.769, -16.984, -16.206, -22.011, -21.629,
    -21.119, -20.542, -19.923, -19.244, -18.529, -17.773, -16.988, -16.210,
    -22.017, -21.635, -21.125, -20.548, -19.928, -19.249, -18.533, -17.778,
    -16.993, -16.214, -22.023, -21.641, -21.130, -20.553, -19.933, -19.254,
    -18.538, -17.782, -16.997, -16.218, -22.029, -21.647, -21.136, -20.558,
    -19.938, -19.259, -18.542, -17.786, -17.001, -16.222, -22.035, -21.653,
    -21.142, -20.564, -19.943, -19.264, -18.547, -17.791, -17.005, -16.226,
    -22.041, -21.658, -21.147, -20.569, -19.948, -19.269, -18.552, -17.795,
    -17.009, -16.230, -22.047, -21.664, -21.153, -20.575, -19.953, -19.274,
    -18.556, -17.800, -17.014, -16.234, -22.053, -21.670, -21.159, -20.580,
    -19.959, -19.278, -18.561, -17.804, -17.018, -16.238, -22.059, -21.676,
    -21.164, -20.586, -19.964, -19.283, -18.565, -17.808, -17.022, -16.242,
    -22.065, -21.682, -21.170, -20.591, -19.969, -19.288, -18.570, -17.813,
    -17.026, -16.246, -22.071, -21.688, -21.175, -20.596, -19.974, -19.293,
    -18.575, -17.817, -17.030, -16.250, -22.077, -21.694, -21.181, -20.602,
    -19.979, -19.298, -18.579, -17.821, -17.035, -16.254, -22.083, -21.699,
    -21.187, -20.607, -19.984, -19.303, -18.584, -17.826, -17.039, -16.258,
    -22.089, -21.705, -21.192, -20.613, -19.990, -19.308, -18.588, -17.830,
    -17.043, -16.262, -22.094, -21.711, -21.198, -20.618, -19.995, -19.313,
    -18.593, -17.835, -17.047, -16.266, -22.100, -21.717, -21.204, -20.623,
    -20.000, -19.317, -18.598, -17.839, -17.051, -16.270, -22.106, -21.723,
    -21.209, -20.629, -20.005, -19.322, -18.602, -17.843, -17.055, -16.274,
    -22.112, -21.729, -21.215, -20.634, -20.010, -19.327, -18.607, -17.848,
    -17.060, -16.278, -22.118, -21.734, -21.221, -20.640, -20.015, -19.332,
    -18.612, -17.852, -17.064, -16.282, -22.124, -21.740, -21.226, -20.645,
    -20.020, -19.337, -18.616, -17.856, -17.068, -16.286, -22.130, -21.746,
    -21.232, -20.651, -20.026, -19.342, -18.621, -17.861, -17.072, -16.290,
    -22.136, -21.752, -21.238, -20.656, -20.031, -19.347, -18.625, -17.865,
    -17.076, -16.294, -22.142, -21.758, -21.243, -20.661, -20.036, -19.352,
    -18.630, -17.869, -17.081, -16.298, -22.148, -21.764, -21.249, -20.667,
    -20.041, -19.356, -18.635, -17.874, -17.085, -16.302, -22.154, -21.769,
    -21.255, -20.672, -20.046, -19.361, -18.639, -17.878, -17.089, -16.306,
    -22.160, -21.775, -21.260, -20.678, -20.051, -19.366, -18.644, -17.882,
    -17.093, -16.310, -22.166, -21.781, -21.266, -20.683, -20.057, -19.371,
    -18.648, -17.887, -17.097, -16.314, -22.172, -21.787, -21.272, -20.689,
    -20.062, -19.376, -18.653, -17.891, -17.101, -16.318, -22.178, -21.793,
    -21.277, -20.694, -20.067, -19.381, -18.658, -17.896, -17.106, -16.322,
    -22.184, -21.799, -21.283, -20.699, -20.072, -19.386, -18.662, -17.900,
    -17.110, -16.326, -22.189, -21.804, -21.289, -20.705, -20.077, -19.391,
    -18.667, -17.904, -17.114, -16.330, -22.195, -21.810, -21.294, -20.710,
    -20.082, -19.395, -18.671, -17.909, -17.118, -16.334, -22.201, -21.816,
    -21.300, -20.716, -20.087, -19.400, -18.676, -17.913, -17.122, -16.338,
    -22.207, -21.822, -21.306, -20.721, -20.093, -19.405, -18.681, -17.917,
    -17.126, -16.342, -22.213, -21.828, -21.311, -20.726, -20.098, -19.410,
    -18.685, -17.922, -17.131, -16.346, -22.219, -21.834, -21.317, -20.732,
    -20.103, -19.415, -18.690, -17.926, -17.135, -16.350, -22.225, -21.839,
    -21.323, -20.737, -20.108, -19.420, -18.694, -17.930, -17.139, -16.354,
    -22.231, -21.845, -21.328, -20.743, -20.113, -19.425, -18.699, -17.935,
    -17.143, -16.358, -22.237, -21.851, -21.334, -20.748, -20.118, -19.430,
    -18.704, -17.939, -17.147, -16.362, -22.243, -21.857, -21.340, -20.754,
    -20.123, -19.434, -18.708, -17.943, -17.151, -16.366, -22.249, -21.863,
    -21.345, -20.759, -20.129, -19.439, -18.713, -17.948, -17.156, -16.370,
    -22.255, -21.869, -21.351, -20.764, -20.134, -19.444, -18.717, -17.952,
    -17.160, -16.374, -22.261, -21.874, -21.356, -20.770, -20.139, -19.449,
    -18.722, -17.956, -17.164, -16.378, -22.267, -21.880, -21.362, -20.775,
    -20.144, -19.454, -18.726, -17.961, -17.168, -16.382, -22.273, -21.886,
    -21.368, -20.781, -20.149, -19.459, -18.731, -17.965, -17.172, -16.386,
    -22.279, -21.892, -21.373, -20.786, -20.154, -19.464, -18.736, -17.969,
    -17.176, -16.389, -22.285, -21.898, -21.379, -20.791, -20.159, -19.468,
    -18.740, -17.974, -17.180, -16.393, -22.290, -21.904, -21.385, -20.797,
    -20.165, -19.473, -18.745, -17.978, -17.185, -16.397, -22.296, -21.909,
    -21.390, -20.802, -20.170, -19.478, -18.749, -17.982, -17.189, -16.401,
    -22.302, -21.915, -21.396, -20.808, -20.175, -19.483, -18.754, -17.987,
    -17.193, -16.405, -22.308, -21.921, -21.402, -20.813, -20.180, -19.488,
    -18.759, -17.991, -17.197, -16.409, -22.314, -21.927, -21.407, -20.819,
    -20.185, -19.493, -18.763, -17.995, -17.201, -16.413, -22.320, -21.933,
    -21.413, -20.824, -20.190, -19.498, -18.768, -18.000, -17.205, -16.417,
    -22.326, -21.939, -21.419, -20.829, -20.195, -19.503, -18.772, -18.004,
    -17.210, -16.421, -22.332, -21.944, -21.424, -20.835, -20.201, -19.507,
    -18.777, -18.008, -17.214, -16.425, -22.338, -21.950, -21.430, -20.840,
    -20.206, -19.512, -18.782, -18.013, -17.218, -16.429, -22.344, -21.956,
    -21.436, -20.846, -20.211, -19.517, -18.786, -18.017, -17.222, -16.433,
    -22.350, -21.962, -21.441, -20.851, -20.216, -19.522, -18.791, -18.022,
    -17.226, -16.437, -22.356, -21.968, -21.447, -20.856, -20.221, -19.527,
    -18.795, -18.026, -17.230, -16.441, -22.362, -21.974, -21.453, -20.862,
    -20.226, -19.532, -18.800, -18.030, -17.234, -16.445, -22.368, -21.979,
    -21.458, -20.867, -20.231, -19.537, -18.805, -18.035, -17.239, -16.449,
    -22.374, -21.985, -21.464, -20.873, -20.237, -19.541, -18.809, -18.039,
    -17.243, -16.453, -22.380, -21.991, -21.469, -20.878, -20.242, -19.546,
    -18.814, -18.043, -17.247, -16.457, -22.385, -21.997, -21.475, -20.883,
    -20.247, -19.551, -18.818, -18.048, -17.251, -16.461, -22.391, -22.003,
    -21.481, -20.889, -20.252, -19.556, -18.823, -18.052, -17.255, -16.465,
    -22.397, -22.009, -21.486, -20.894, -20.257, -19.561, -18.827, -18.056,
    -17.259, -16.469, -22.403, -22.014, -21.492, -20.900, -20.262, -19.566,
    -18.832, -18.061, -17.263, -16.472, -22.409, -22.020, -21.498, -20.905,
    -20.267, -19.571, -18.837, -18.065, -17.267, -16.476, -22.415, -22.026,
    -21.503, -20.911, -20.273, -19.576, -18.841, -18.069, -17.272, -16.480,
    -22.421, -22.032, -21.509, -20.916, -20.278, -19.580, -18.846, -18.074,
    -17.276, -16.484, -22.427, -22.038, -21.515, -20.921, -20.283, -19.585,
    -18.850, -18.078, -17.280, -16.488, -22.433, -22.044, -21.520, -20.927,
    -20.288, -19.590, -18.855, -18.082, -17.284, -16.492, -22.439, -22.049,
    -21.526, -20.932, -20.293, -19.595, -18.860, -18.086, -17.288, -16.496,
    -22.445, -22.055, -21.532, -20.938, -20.298, -19.600, -18.864, -18.091,
    -17.292, -16.500, -22.451, -22.061, -21.537, -20.943, -20.303, -19.605,
    -18.869, -18.095, -17.296, -16.504, -22.457, -22.067, -21.543, -20.948,
    -20.309, -19.610, -18.873, -18.099, -17.301, -16.508, -22.463, -22.073,
    -21.549, -20.954, -20.314, -19.614, -18.878, -18.104, -17.305, -16.512,
    -22.469, -22.079, -21.554, -20.959, -20.319, -19.619, -18.882, -18.108,
    -17.309, -16.516, -22.474, -22.084, -21.560, -20.965, -20.324, -19.624,
    -18.887, -18.112, -17.313, -16.520, -22.480, -22.090, -21.565, -20.970,
    -20.329, -19.629, -18.892, -18.117, -17.317, -16.524, -22.486, -22.096,
    -21.571, -20.975, -20.334, -19.634, -18.896, -18.121, -17.321, -16.528,
    -22.492, -22.102, -21.577, -20.981, -20.339, -19.639, -18.901, -18.125,
    -17.325, -16.532, -22.498, -22.108, -21.582, -20.986, -20.345, -19.644,
    -18.905, -18.130, -17.329, -16.535, -22.504, -22.114, -21.588, -20.992,
    -20.350, -19.648, -18.910, -18.134, -17.334, -16.539, -22.510, -22.119,
    -21.594, -20.997, -20.355, -19.653, -18.915, -18.138, -17.338, -16.543,
    -22.516, -22.125, -21.599, -21.002, -20.360, -19.658, -18.919, -18.143,
    -17.342, -16.547, -22.522, -22.131, -21.605, -21.008, -20.365, -19.663,
    -18.924, -18.147, -17.346, -16.551, -22.528, -22.137, -21.611, -21.013,
    -20.370, -19.668, -18.928, -18.151, -17.350, -16.555, -22.534, -22.143,
    -21.616, -21.019, -20.375, -19.673, -18.933, -18.156, -17.354, -16.559,
    -22.540, -22.149, -21.622, -21.024, -20.380, -19.678, -18.937, -18.160,
    -17.358, -16.563, -22.546, -22.154, -21.628, -21.029, -20.386, -19.682,
    -18.942, -18.164, -17.362, -16.567, -22.552, -22.160, -21.633, -21.035,
    -20.391, -19.687, -18.947, -18.169, -17.366, -16.571, -22.558, -22.166,
    -21.639, -21.040, -20.396, -19.692, -18.951, -18.173, -17.371, -16.575,
    -22.563, -22.172, -21.644, -21.046, -20.401, -19.697, -18.956, -18.177,
    -17.375, -16.579, -22.569, -22.178, -21.650, -21.051, -20.406, -19.702,
    -18.960, -18.182, -17.379, -16.583, -22.575, -22.183, -21.656, -21.057,
    -20.411, -19.707, -18.965, -18.186, -17.383, -16.586, -22.581, -22.189,
    -21.661, -21.062, -20.416, -19.712, -18.969, -18.190, -17.387, -16.590,
    -22.587, -22.195, -21.667, -21.067, -20.422, -19.716, -18.974, -18.195,
    -17.391, -16.594, -22.593, -22.201, -21.673, -21.073, -20.427, -19.721,
    -18.979, -18.199, -17.395, -16.598, -22.599, -22.207, -21.678, -21.078,
    -20.432, -19.726, -18.983, -18.203, -17.399, -16.602, -22.605, -22.213,
    -21.684, -21.084, -20.437, -19.731, -18.988, -18.207, -17.403, -16.606,
    -22.611, -22.218, -21.690, -21.089, -20.442, -19.736, -18.992, -18.212,
    -17.408, -16.610, -22.617, -22.224, -21.695, -21.094, -20.447, -19.741,
    -18.997, -18.216, -17.412, -16.614, -22.623, -22.230, -21.701, -21.100,
    -20.452, -19.745, -19.001, -18.220, -17.416, -16.618, -22.629, -22.236,
    -21.706, -21.105, -20.457, -19.750, -19.006, -18.225, -17.420, -16.622,
    -22.635, -22.242, -21.712, -21.111, -20.463, -19.755, -19.011, -18.229,
    -17.424, -16.626, -22.641, -22.248, -21.718, -21.116, -20.468, -19.760,
    -19.015, -18.233, -17.428, -16.629, -22.647, -22.253, -21.723, -21.121,
    -20.473, -19.765, -19.020, -18.238, -17.432, -16.633, -22.652, -22.259,
    -21.729, -21.127, -20.478, -19.770, -19.024, -18.242, -17.436, -16.637,
    -22.658, -22.265, -21.735, -21.132, -20.483, -19.775, -19.029, -18.246,
    -17.440, -16.641, -22.664, -22.271, -21.740, -21.138, -20.488, -19.779,
    -19.033, -18.251, -17.445, -16.645, -22.670, -22.277, -21.746, -21.143,
    -20.493, -19.784, -19.038, -18.255, -17.449, -16.649, -22.676, -22.283,
    -21.752, -21.148, -20.498, -19.789, -19.043, -18.259, -17.453, -16.653,
    -22.682, -22.288, -21.757, -21.154, -20.504, -19.794, -19.047, -18.264,
    -17.457, -16.657, -22.688, -22.294, -21.763, -21.159, -20.509, -19.799,
    -19.052, -18.268, -17.461, -16.661, -22.694, -22.300, -21.768, -21.165,
    -20.514, -19.804, -19.056, -18.272, -17.465, -16.665, -22.700, -22.306,
    -21.774, -21.170, -20.519, -19.809, -19.061, -18.276, -17.469, -16.668,
    -22.706, -22.312, -21.780, -21.175, -20.524, -19.813, -19.065, -18.281,
    -17.473, -16.672, -22.712, -22.317, -21.785, -21.181, -20.529, -19.818,
    -19.070, -18.285, -17.477, -16.676, -22.718, -22.323, -21.791, -21.186,
    -20.534, -19.823, -19.075, -18.289, -17.481, -16.680, -22.724, -22.329,
    -21.797, -21.192, -20.539, -19.828, -19.079, -18.294, -17.485, -16.684,
    -22.730, -22.335, -21.802, -21.197, -20.545, -19.833, -19.084, -18.298,
    -17.490, -16.688, -22.736, -22.341, -21.808, -21.202, -20.550, -19.838,
    -19.088, -18.302, -17.494, -16.692, -22.741, -22.347, -21.814, -21.208,
    -20.555, -19.842, -19.093, -18.307, -17.498, -16.696, -22.747, -22.352,
    -21.819, -21.213, -20.560, -19.847, -19.097, -18.311, -17.502, -16.700,
    -22.753, -22.358, -21.825, -21.218, -20.565, -19.852, -19.102, -18.315,
    -17.506, -16.704, -22.759, -22.364, -21.830, -21.224, -20.570, -19.857,
    -19.107, -18.319, -17.510, -16.707, -22.765, -22.370, -21.836, -21.229,
    -20.575, -19.862, -19.111, -18.324, -17.514, -16.711, -22.771, -22.376,
    -21.842, -21.235, -20.580, -19.867, -19.116, -18.328, -17.518, -16.715,
    -22.777, -22.381, -21.847, -21.240, -20.586, -19.872, -19.120, -18.332,
    -17.522, -16.719, -22.783, -22.387, -21.853, -21.245, -20.591, -19.876,
    -19.125, -18.337, -17.526, -16.723, -22.789, -22.393, -21.859, -21.251,
    -20.596, -19.881, -19.129, -18.341, -17.530, -16.727, -22.795, -22.399,
    -21.864, -21.256, -20.601, -19.886, -19.134, -18.345, -17.535, -16.731,
    -22.801, -22.405, -21.870, -21.262, -20.606, -19.891, -19.138, -18.350,
    -17.539, -16.735, -22.807, -22.411, -21.876, -21.267, -20.611, -19.896,
    -19.143, -18.354, -17.543, -16.738, -22.813, -22.416, -21.881, -21.272,
    -20.616, -19.901, -19.148, -18.358, -17.547, -16.742, -22.819, -22.422,
    -21.887, -21.278, -20.621, -19.905, -19.152, -18.362, -17.551, -16.746,
    -22.824, -22.428, -21.892, -21.283, -20.627, -19.910, -19.157, -18.367,
    -17.555, -16.750, -22.830, -22.434, -21.898, -21.289, -20.632, -19.915,
    -19.161, -18.371, -17.559, -16.754, -22.836, -22.440, -21.904, -21.294,
    -20.637, -19.920, -19.166, -18.375, -17.563, -16.758, -22.842, -22.446,
    -21.909, -21.299, -20.642, -19.925, -19.170, -18.380, -17.567, -16.762,
    -22.848, -22.451, -21.915, -21.305, -20.647, -19.930, -19.175, -18.384,
    -17.571, -16.766, -22.854, -22.457, -21.921, -21.310, -20.652, -19.935,
    -19.180, -18.388, -17.575, -16.770, -22.860, -22.463, -21.926, -21.316,
    -20.657, -19.939, -19.184, -18.393, -17.579, -16.773, -22.866, -22.469,
    -21.932, -21.321, -20.662, -19.944, -19.189, -18.397, -17.583, -16.777,
    -22.872, -22.475, -21.937, -21.326, -20.668, -19.949, -19.193, -18.401,
    -17.588, -16.781, -22.878, -22.480, -21.943, -21.332, -20.673, -19.954,
    -19.198, -18.405, -17.592, -16.785, -22.884, -22.486, -21.949, -21.337,
    -20.678, -19.959, -19.202, -18.410, -17.596, -16.789, -22.890, -22.492,
    -21.954, -21.343, -20.683, -19.964, -19.207, -18.414, -17.600, -16.793,
    -22.896, -22.498, -21.960, -21.348, -20.688, -19.968, -19.211, -18.418,
    -17.604, -16.797, -22.902, -22.504, -21.966, -21.353, -20.693, -19.973,
    -19.216, -18.423, -17.608, -16.800, -22.908, -22.510, -21.971, -21.359,
    -20.698, -19.978, -19.221, -18.427, -17.612, -16.804, -22.913, -22.515,
    -21.977, -21.364, -20.703, -19.983, -19.225, -18.431, -17.616, -16.808,
    -22.919, -22.521, -21.983, -21.370, -20.709, -19.988, -19.230, -18.435,
    -17.620, -16.812, -22.925, -22.527, -21.988, -21.375, -20.714, -19.993,
    -19.234, -18.440, -17.624, -16.816, -22.931, -22.533, -21.994, -21.380,
    -20.719, -19.997, -19.239, -18.444, -17.628, -16.820, -22.937, -22.539,
    -21.999, -21.386, -20.724, -20.002, -19.243, -18.448, -17.632, -16.824,
    -22.943, -22.544, -22.005, -21.391, -20.729, -20.007, -19.248, -18.453,
    -17.636, -16.828, -22.949, -22.550, -22.011, -21.396, -20.734, -20.012,
    -19.252, -18.457, -17.640, -16.831, -22.955, -22.556, -22.016, -21.402,
    -20.739, -20.017, -19.257, -18.461, -17.645, -16.835, -22.961, -22.562,
    -22.022, -21.407, -20.744, -20.022, -19.262, -18.466, -17.649, -16.839,
    -22.967, -22.568, -22.028, -21.413, -20.749, -20.026, -19.266, -18.470,
    -17.653, -16.843, -22.973, -22.574, -22.033, -21.418, -20.755, -20.031,
    -19.271, -18.474, -17.657, -16.847, -22.979, -22.579, -22.039, -21.423,
    -20.760, -20.036, -19.275, -18.478, -17.661, -16.851, -22.985, -22.585,
    -22.044, -21.429, -20.765, -20.041, -19.280, -18.483, -17.665, -16.855,
    -22.991, -22.591, -22.050, -21.434, -20.770, -20.046, -19.284, -18.487,
    -17.669, -16.858, -22.996, -22.597, -22.056, -21.440, -20.775, -20.051,
    -19.289, -18.491, -17.673, -16.862, -23.002, -22.603, -22.061, -21.445,
    -20.780, -20.055, -19.293, -18.496, -17.677, -16.866, -23.008, -22.608,
    -22.067, -21.450, -20.785, -20.060, -19.298, -18.500, -17.681, -16.870,
    -23.014, -22.614, -22.073, -21.456, -20.790, -20.065, -19.303, -18.504,
    -17.685, -16.874, -23.020, -22.620, -22.078, -21.461, -20.796, -20.070,
    -19.307, -18.508, -17.689, -16.878, -23.026, -22.626, -22.084, -21.467,
    -20.801, -20.075, -19.312, -18.513, -17.693, -16.882, -23.032, -22.632,
    -22.089, -21.472, -20.806, -20.080, -19.316, -18.517, -17.697, -16.885,
    -23.038, -22.637, -22.095, -21.477, -20.811, -20.084, -19.321, -18.521,
    -17.701, -16.889, -23.044, -22.643, -22.101, -21.483, -20.816, -20.089,
    -19.325, -18.526, -17.705, -16.893, -23.050, -22.649, -22.106, -21.488,
    -20.821, -20.094, -19.330, -18.530, -17.709, -16.897, -23.056, -22.655,
    -22.112, -21.493, -20.826, -20.099, -19.334, -18.534, -17.714, -16.901,
    -23.062, -22.661, -22.118, -21.499, -20.831, -20.104, -19.339, -18.538,
    -17.718, -16.905, -23.068, -22.667, -22.123, -21.504, -20.836, -20.109,
    -19.344, -18.543, -17.722, -16.908, -23.074, -22.672, -22.129, -21.510,
    -20.842, -20.113, -19.348, -18.547, -17.726, -16.912, -23.079, -22.678,
    -22.134, -21.515, -20.847, -20.118, -19.353, -18.551, -17.730, -16.916,
    -23.085, -22.684, -22.140, -21.520, -20.852, -20.123, -19.357, -18.555,
    -17.734, -16.920, -23.091, -22.690, -22.146, -21.526, -20.857, -20.128,
    -19.362, -18.560, -17.738, -16.924, -23.097, -22.696, -22.151, -21.531,
    -20.862, -20.133, -19.366, -18.564, -17.742, -16.928, -23.103, -22.701,
    -22.157, -21.537, -20.867, -20.138, -19.371, -18.568, -17.746, -16.932,
    -23.109, -22.707, -22.163, -21.542, -20.872, -20.142, -19.375, -18.573,
    -17.750, -16.935, -23.115, -22.713, -22.168, -21.547, -20.877, -20.147,
    -19.380, -18.577, -17.754, -16.939, -23.121, -22.719, -22.174, -21.553,
    -20.882, -20.152, -19.384, -18.581, -17.758, -16.943, -23.127, -22.725,
    -22.179, -21.558, -20.888, -20.157, -19.389, -18.585, -17.762, -16.947,
    -23.133, -22.731, -22.185, -21.563, -20.893, -20.162, -19.394, -18.590,
    -17.766, -16.951, -23.139, -22.736, -22.191, -21.569, -20.898, -20.167,
    -19.398, -18.594, -17.770, -16.955, -23.145, -22.742, -22.196, -21.574,
    -20.903, -20.171, -19.403, -18.598, -17.774, -16.958, -23.151, -22.748,
    -22.202, -21.580, -20.908, -20.176, -19.407, -18.603, -17.778, -16.962,
    -23.156, -22.754, -22.208, -21.585, -20.913, -20.181, -19.412, -18.607,
    -17.782, -16.966, -23.162, -22.760, -22.213, -21.590, -20.918, -20.186,
    -19.416, -18.611, -17.786, -16.970, -23.168, -22.765, -22.219, -21.596,
    -20.923, -20.191, -19.421, -18.615, -17.790, -16.974, -23.174, -22.771,
    -22.224, -21.601, -20.929, -20.196, -19.425, -18.620, -17.795, -16.978,
    -23.180, -22.777, -22.230, -21.607, -20.934, -20.200, -19.430, -18.624,
    -17.799, -16.981, -23.186, -22.783, -22.236, -21.612, -20.939, -20.205,
    -19.434, -18.628, -17.803, -16.985, -23.192, -22.789, -22.241, -21.617,
    -20.944, -20.210, -19.439, -18.632, -17.807, -16.989, -23.198, -22.794,
    -22.247, -21.623, -20.949, -20.215, -19.444, -18.637, -17.811, -16.993,
    -23.204, -22.800, -22.253, -21.628, -20.954, -20.220, -19.448, -18.641,
    -17.815, -16.997, -23.210, -22.806, -22.258, -21.633, -20.959, -20.225,
    -19.453, -18.645, -17.819, -17.001, -23.216, -22.812, -22.264, -21.639,
    -20.964, -20.229, -19.457, -18.650, -17.823, -17.004, -23.222, -22.818,
    -22.269, -21.644, -20.969, -20.234, -19.462, -18.654, -17.827, -17.008,
    -23.228, -22.823, -22.275, -21.650, -20.974, -20.239, -19.466, -18.658,
    -17.831, -17.012, -23.234, -22.829, -22.281, -21.655, -20.980, -20.244,
    -19.471, -18.662, -17.835, -17.016, -23.239, -22.835, -22.286, -21.660,
    -20.985, -20.249, -19.475, -18.667, -17.839, -17.020, -23.245, -22.841,
    -22.292, -21.666, -20.990, -20.254, -19.480, -18.671, -17.843, -17.023,
    -23.251, -22.847, -22.297, -21.671, -20.995, -20.258, -19.484, -18.675,
    -17.847, -17.027, -23.257, -22.853, -22.303, -21.676, -21.000, -20.263,
    -19.489, -18.679, -17.851, -17.031, -23.263, -22.858, -22.309, -21.682,
    -21.005, -20.268, -19.493, -18.684, -17.855, -17.035, -23.269, -22.864,
    -22.314, -21.687, -21.010, -20.273, -19.498, -18.688, -17.859, -17.039,
    -23.275, -22.870, -22.320, -21.693, -21.015, -20.278, -19.503, -18.692,
    -17.863, -17.043, -23.281, -22.876, -22.326, -21.698, -21.020, -20.282,
    -19.507, -18.697, -17.867, -17.046, -23.287, -22.882, -22.331, -21.703,
    -21.026, -20.287, -19.512, -18.701, -17.871, -17.050, -23.293, -22.887,
    -22.337, -21.709, -21.031, -20.292, -19.516, -18.705, -17.875, -17.054,
    -23.299, -22.893, -22.342, -21.714, -21.036, -20.297, -19.521, -18.709,
    -17.879, -17.058, -23.305, -22.899, -22.348, -21.719, -21.041, -20.302,
    -19.525, -18.714, -17.883, -17.062, -23.311, -22.905, -22.354, -21.725,
    -21.046, -20.307, -19.530, -18.718, -17.887, -17.066, -23.316, -22.911,
    -22.359, -21.730, -21.051, -20.311, -19.534, -18.722, -17.891, -17.069,
    -23.322, -22.916, -22.365, -21.736, -21.056, -20.316, -19.539, -18.726,
    -17.895, -17.073, -23.328, -22.922, -22.371, -21.741, -21.061, -20.321,
    -19.543, -18.731, -17.899, -17.077, -23.334, -22.928, -22.376, -21.746,
    -21.066, -20.326, -19.548, -18.735, -17.903, -17.081, -23.340, -22.934,
    -22.382, -21.752, -21.072, -20.331, -19.553, -18.739, -17.908, -17.085,
    -23.346, -22.940, -22.387, -21.757, -21.077, -20.336, -19.557, -18.743,
    -17.912, -17.088, -23.352, -22.945, -22.393, -21.762, -21.082, -20.340,
    -19.562, -18.748, -17.916, -17.092, -23.358, -22.951, -22.399, -21.768,
    -21.087, -20.345, -19.566, -18.752, -17.920, -17.096, -23.364, -22.957,
    -22.404, -21.773, -21.092, -20.350, -19.571, -18.756, -17.924, -17.100,
    -23.370, -22.963, -22.410, -21.779, -21.097, -20.355, -19.575, -18.761,
    -17.928, -17.104, -23.376, -22.969, -22.415, -21.784, -21.102, -20.360,
    -19.580, -18.765, -17.932, -17.107, -23.382, -22.975, -22.421, -21.789,
    -21.107, -20.364, -19.584, -18.769, -17.936, -17.111, -23.388, -22.980,
    -22.427, -21.795, -21.112, -20.369, -19.589, -18.773, -17.940, -17.115,
    -23.393, -22.986, -22.432, -21.800, -21.117, -20.374, -19.593, -18.778,
    -17.944, -17.119, -23.399, -22.992, -22.438, -21.805, -21.123, -20.379,
    -19.598, -18.782, -17.948, -17.123, -23.405, -22.998, -22.444, -21.811,
    -21.128, -20.384, -19.602, -18.786, -17.952, -17.126, -23.411, -23.004,
    -22.449, -21.816, -21.133, -20.389, -19.607, -18.790, -17.956, -17.130,
    -23.417, -23.009, -22.455, -21.822, -21.138, -20.393, -19.611, -18.795,
    -17.960, -17.134, -23.423, -23.015, -22.460, -21.827, -21.143, -20.398,
    -19.616, -18.799, -17.964, -17.138, -23.429, -23.021, -22.466, -21.832,
    -21.148, -20.403, -19.621, -18.803, -17.968, -17.142, -23.435, -23.027,
    -22.472, -21.838, -21.153, -20.408, -19.625, -18.807, -17.972, -17.146,
    -23.441, -23.033, -22.477, -21.843, -21.158, -20.413, -19.630, -18.812,
    -17.976, -17.149, -23.447, -23.038, -22.483, -21.848, -21.163, -20.418,
    -19.634, -18.816, -17.980, -17.153, -23.453, -23.044, -22.488, -21.854,
    -21.169, -20.422, -19.639, -18.820, -17.984, -17.157, -23.459, -23.050,
    -22.494, -21.859, -21.174, -20.427, -19.643, -18.824, -17.988, -17.161,
    -23.465, -23.056, -22.500, -21.865, -21.179, -20.432, -19.648, -18.829,
    -17.992, -17.165, -23.470, -23.062, -22.505, -21.870, -21.184, -20.437,
    -19.652, -18.833, -17.996, -17.168, -23.476, -23.067, -22.511, -21.875,
    -21.189, -20.442, -19.657, -18.837, -18.000, -17.172, -23.482, -23.073,
    -22.516, -21.881, -21.194, -20.446, -19.661, -18.842, -18.004, -17.176,
    -23.488, -23.079, -22.522, -21.886, -21.199, -20.451, -19.666, -18.846,
    -18.008, -17.180, -23.494, -23.085, -22.528, -21.891, -21.204, -20.456,
    -19.670, -18.850, -18.012, -17.184, -23.500, -23.091, -22.533, -21.897,
    -21.209, -20.461, -19.675, -18.854, -18.016, -17.187, -23.506, -23.096,
    -22.539, -21.902, -21.214, -20.466, -19.679, -18.859, -18.020, -17.191,
    -23.512, -23.102, -22.545, -21.908, -21.220, -20.471, -19.684, -18.863,
    -18.024, -17.195, -23.518, -23.108, -22.550, -21.913, -21.225, -20.475,
    -19.689, -18.867, -18.028, -17.199, -23.524, -23.114, -22.556, -21.918,
    -21.230, -20.480, -19.693, -18.871, -18.032, -17.202, -23.530, -23.120,
    -22.561, -21.924, -21.235, -20.485, -19.698, -18.876, -18.036, -17.206,
    -23.536, -23.125, -22.567, -21.929, -21.240, -20.490, -19.702, -18.880,
    -18.040, -17.210, -23.542, -23.131, -22.573, -21.934, -21.245, -20.495,
    -19.707, -18.884, -18.044, -17.214, -23.547, -23.137, -22.578, -21.940,
    -21.250, -20.499, -19.711, -18.888, -18.048, -17.218, -23.553, -23.143,
    -22.584, -21.945, -21.255, -20.504, -19.716, -18.893, -18.052, -17.221,
    -23.559, -23.149, -22.589, -21.951, -21.260, -20.509, -19.720, -18.897,
    -18.056, -17.225, -23.565, -23.154, -22.595, -21.956, -21.265, -20.514,
    -19.725, -18.901, -18.060, -17.229, -23.571, -23.160, -22.601, -21.961,
    -21.271, -20.519, -19.729, -18.905, -18.064, -17.233, -23.577, -23.166,
    -22.606, -21.967, -21.276, -20.523, -19.734, -18.910, -18.068, -17.237,
    -23.583, -23.172, -22.612, -21.972, -21.281, -20.528, -19.738, -18.914,
    -18.072, -17.240, -23.589, -23.178, -22.617, -21.977, -21.286, -20.533,
    -19.743, -18.918, -18.076, -17.244, -23.595, -23.183, -22.623, -21.983,
    -21.291, -20.538, -19.747, -18.922, -18.080, -17.248, -23.601, -23.189,
    -22.629, -21.988, -21.296, -20.543, -19.752, -18.927, -18.084, -17.252,
    -23.607, -23.195, -22.634, -21.993, -21.301, -20.548, -19.756, -18.931,
    -18.088, -17.256, -23.613, -23.201, -22.640, -21.999, -21.306, -20.552,
    -19.761, -18.935, -18.092, -17.259, -23.619, -23.207, -22.646, -22.004,
    -21.311, -20.557, -19.766, -18.939, -18.096, -17.263, -23.624, -23.212,
    -22.651, -22.010, -21.316, -20.562, -19.770, -18.944, -18.100, -17.267,
    -23.630, -23.218, -22.657, -22.015, -21.321, -20.567, -19.775, -18.948,
    -18.104, -17.271, -23.636, -23.224, -22.662, -22.020, -21.327, -20.572,
    -19.779, -18.952, -18.108, -17.274, -23.642, -23.230, -22.668, -22.026,
    -21.332, -20.576, -19.784, -18.956, -18.112, -17.278, -23.648, -23.236,
    -22.674, -22.031, -21.337, -20.581, -19.788, -18.961, -18.116, -17.282,
    -23.654, -23.241, -22.679, -22.036, -21.342, -20.586, -19.793, -18.965,
    -18.120, -17.286, -23.660, -23.247, -22.685, -22.042, -21.347, -20.591,
    -19.797, -18.969, -18.124, -17.290, -23.666, -23.253, -22.690, -22.047,
    -21.352, -20.596, -19.802, -18.973, -18.128, -17.293, -23.672, -23.259,
    -22.696, -22.053, -21.357, -20.601, -19.806, -18.978, -18.132, -17.297,
    -23.678, -23.265, -22.702, -22.058, -21.362, -20.605, -19.811, -18.982,
    -18.136, -17.301, -23.684, -23.270, -22.707, -22.063, -21.367, -20.610,
    -19.815, -18.986, -18.140, -17.305, -23.690, -23.276, -22.713, -22.069,
    -21.372, -20.615, -19.820, -18.990, -18.144, -17.309, -23.695, -23.282,
    -22.718, -22.074, -21.378, -20.620, -19.824, -18.995, -18.148, -17.312,
    -23.701, -23.288, -22.724, -22.079, -21.383, -20.625, -19.829, -18.999,
    -18.152, -17.316, -23.707, -23.294, -22.730, -22.085, -21.388, -20.629,
    -19.833, -19.003, -18.156, -17.320, -23.713, -23.299, -22.735, -22.090,
    -21.393, -20.634, -19.838, -19.007, -18.160, -17.324, -23.719, -23.305,
    -22.741, -22.095, -21.398, -20.639, -19.842, -19.012, -18.164, -17.327,
    -23.725, -23.311, -22.746, -22.101, -21.403, -20.644, -19.847, -19.016,
    -18.168, -17.331, -23.731, -23.317, -22.752, -22.106, -21.408, -20.649,
    -19.851, -19.020, -18.172, -17.335, -23.737, -23.323, -22.758, -22.112,
    -21.413, -20.653, -19.856, -19.024, -18.176, -17.339, -23.743, -23.328,
    -22.763, -22.117, -21.418, -20.658, -19.861, -19.029, -18.180, -17.342,
    -23.749, -23.334, -22.769, -22.122, -21.423, -20.663, -19.865, -19.033,
    -18.184, -17.346, -23.755, -23.340, -22.774, -22.128, -21.428, -20.668,
    -19.870, -19.037, -18.188, -17.350, -23.761, -23.346, -22.780, -22.133,
    -21.434, -20.673, -19.874, -19.041, -18.192, -17.354, -23.767, -23.352,
    -22.786, -22.138, -21.439, -20.677, -19.879, -19.046, -18.196, -17.358,
    -23.772, -23.357, -22.791, -22.144, -21.444, -20.682, -19.883, -19.050,
    -18.200, -17.361, -23.778, -23.363, -22.797, -22.149, -21.449, -20.687,
    -19.888, -19.054, -18.204, -17.365, -23.784, -23.369, -22.803, -22.154,
    -21.454, -20.692, -19.892, -19.058, -18.208, -17.369, -23.790, -23.375,
    -22.808, -22.160, -21.459, -20.697, -19.897, -19.063, -18.212, -17.373,
    -23.796, -23.381, -22.814, -22.165, -21.464, -20.702, -19.901, -19.067,
    -18.216, -17.376, -23.802, -23.386, -22.819, -22.171, -21.469, -20.706,
    -19.906, -19.071, -18.220, -17.380, -23.808, -23.392, -22.825, -22.176,
    -21.474, -20.711, -19.910, -19.075, -18.224, -17.384, -23.814, -23.398,
    -22.831, -22.181, -21.479, -20.716, -19.915, -19.080, -18.228, -17.388,
    -23.820, -23.404, -22.836, -22.187, -21.484, -20.721, -19.919, -19.084,
    -18.232, -17.391, -23.826, -23.410, -22.842, -22.192, -21.490, -20.726,
    -19.924, -19.088, -18.236, -17.395, -23.832, -23.415, -22.847, -22.197,
    -21.495, -20.730, -19.928, -19.092, -18.240, -17.399, -23.838, -23.421,
    -22.853, -22.203, -21.500, -20.735, -19.933, -19.097, -18.244, -17.403,
    -23.843, -23.427, -22.859, -22.208, -21.505, -20.740, -19.937, -19.101,
    -18.248, -17.407, -23.849, -23.433, -22.864, -22.213, -21.510, -20.745,
    -19.942, -19.105, -18.252, -17.410, -23.855, -23.439, -22.870, -22.219,
    -21.515, -20.750, -19.946, -19.109, -18.256, -17.414, -23.861, -23.444,
    -22.875, -22.224, -21.520, -20.754, -19.951, -19.113, -18.260, -17.418,
    -23.867, -23.450, -22.881, -22.230, -21.525, -20.759, -19.955, -19.118,
    -18.264, -17.422, -23.873, -23.456, -22.887, -22.235, -21.530, -20.764,
    -19.960, -19.122, -18.268, -17.425, -23.879, -23.462, -22.892, -22.240,
    -21.535, -20.769, -19.964, -19.126, -18.272, -17.429, -23.885, -23.468,
    -22.898, -22.246, -21.540, -20.774, -19.969, -19.130, -18.276, -17.433,
    -23.891, -23.473, -22.903, -22.251, -21.546, -20.778, -19.974, -19.135,
    -18.280, -17.437, -23.897, -23.479, -22.909, -22.256, -21.551, -20.783,
    -19.978, -19.139, -18.284, -17.440, -23.903, -23.485, -22.915, -22.262,
    -21.556, -20.788, -19.983, -19.143, -18.288, -17.444, -23.909, -23.491,
    -22.920, -22.267, -21.561, -20.793, -19.987, -19.147, -18.292, -17.448,
    -23.914, -23.497, -22.926, -22.272, -21.566, -20.798, -19.992, -19.152,
    -18.296, -17.452, -23.920, -23.502, -22.931, -22.278, -21.571, -20.802,
    -19.996, -19.156, -18.300, -17.455, -23.926, -23.508, -22.937, -22.283,
    -21.576, -20.807, -20.001, -19.160, -18.304, -17.459, -23.932, -23.514,
    -22.943, -22.288, -21.581, -20.812, -20.005, -19.164, -18.308, -17.463,
    -23.938, -23.520, -22.948, -22.294, -21.586, -20.817, -20.010, -19.169,
    -18.312, -17.467, -23.944, -23.526, -22.954, -22.299, -21.591, -20.822,
    -20.014, -19.173, -18.316, -17.470, -23.950, -23.531, -22.959, -22.305,
    -21.596, -20.826, -20.019, -19.177, -18.320, -17.474, -23.956, -23.537,
    -22.965, -22.310, -21.602, -20.831, -20.023, -19.181, -18.324, -17.478,
    -23.962, -23.543, -22.971, -22.315, -21.607, -20.836, -20.028, -19.186,
    -18.328, -17.482, -23.968, -23.549, -22.976, -22.321, -21.612, -20.841,
    -20.032, -19.190, -18.332, -17.486, -23.974, -23.555, -22.982, -22.326,
    -21.617, -20.846, -20.037, -19.194, -18.336, -17.489, -23.980, -23.560,
    -22.987, -22.331, -21.622, -20.850, -20.041, -19.198, -18.340, -17.493,
    -23.985, -23.566, -22.993, -22.337, -21.627, -20.855, -20.046, -19.202,
    -18.344, -17.497, -23.991, -23.572, -22.999, -22.342, -21.632, -20.860,
    -20.050, -19.207, -18.348, -17.501, -23.997, -23.578, -23.004, -22.347,
    -21.637, -20.865, -20.055, -19.211, -18.352, -17.504, -24.003, -23.584,
    -23.010, -22.353, -21.642, -20.870, -20.059, -19.215, -18.356, -17.508,
    -24.009, -23.589, -23.015, -22.358, -21.647, -20.874, -20.064, -19.219,
    -18.360, -17.512, -24.015, -23.595, -23.021, -22.363, -21.652, -20.879,
    -20.068, -19.224, -18.364, -17.516, -24.021, -23.601, -23.027, -22.369,
    -21.657, -20.884, -20.073, -19.228, -18.368, -17.519, -24.027, -23.607,
    -23.032, -22.374, -21.663, -20.889, -20.077, -19.232, -18.372, -17.523,
    -24.033, -23.613, -23.038, -22.380, -21.668, -20.894, -20.082, -19.236,
    -18.376, -17.527, -24.039, -23.618, -23.043, -22.385, -21.673, -20.898,
    -20.086, -19.241, -18.380, -17.531, -24.045, -23.624, -23.049, -22.390,
    -21.678, -20.903, -20.091, -19.245, -18.384, -17.534, -24.051, -23.630,
    -23.055, -22.396, -21.683, -20.908, -20.095, -19.249, -18.388, -17.538,
    -24.056, -23.636, -23.060, -22.401, -21.688, -20.913, -20.100, -19.253,
    -18.392, -17.542, -24.062, -23.641, -23.066, -22.406, -21.693, -20.918,
    -20.104, -19.258, -18.396, -17.546, -24.068, -23.647, -23.071, -22.412,
    -21.698, -20.922, -20.109, -19.262, -18.400, -17.549, -24.074, -23.653,
    -23.077, -22.417, -21.703, -20.927, -20.113, -19.266, -18.404, -17.553,
    -24.080, -23.659, -23.083, -22.422, -21.708, -20.932, -20.118, -19.270,
    -18.408, -17.557, -24.086, -23.665, -23.088, -22.428, -21.713, -20.937,
    -20.122, -19.274, -18.412, -17.561, -24.092, -23.670, -23.094, -22.433,
    -21.718, -20.942, -20.127, -19.279, -18.416, -17.564, -24.098, -23.676,
    -23.099, -22.438, -21.724, -20.946, -20.132, -19.283, -18.420, -17.568,
    -24.104, -23.682, -23.105, -22.444, -21.729, -20.951, -20.136, -19.287,
    -18.424, -17.572, -24.110, -23.688, -23.110, -22.449, -21.734, -20.956,
    -20.141, -19.291, -18.428, -17.575, -24.116, -23.694, -23.116, -22.455,
    -21.739, -20.961, -20.145, -19.296, -18.432, -17.579, -24.122, -23.699,
    -23.122, -22.460, -21.744, -20.966, -20.150, -19.300, -18.436, -17.583,
    -24.127, -23.705, -23.127, -22.465, -21.749, -20.970, -20.154, -19.304,
    -18.440, -17.587, -24.133, -23.711, -23.133, -22.471, -21.754, -20.975,
    -20.159, -19.308, -18.444, -17.590, -24.139, -23.717, -23.138, -22.476,
    -21.759, -20.980, -20.163, -19.313, -18.448, -17.594, -24.145, -23.723,
    -23.144, -22.481, -21.764, -20.985, -20.168, -19.317, -18.451, -17.598,
    -24.151, -23.728, -23.150, -22.487, -21.769, -20.990, -20.172, -19.321,
    -18.455, -17.602, -24.157, -23.734, -23.155, -22.492, -21.774, -20.994,
    -20.177, -19.325, -18.459, -17.605, -24.163, -23.740, -23.161, -22.497,
    -21.779, -20.999, -20.181, -19.329, -18.463, -17.609, -24.169, -23.746,
    -23.166, -22.503, -21.785, -21.004, -20.186, -19.334, -18.467, -17.613,
    -24.175, -23.752, -23.172, -22.508, -21.790, -21.009, -20.190, -19.338,
    -18.471, -17.617, -24.181, -23.757, -23.178, -22.513, -21.795, -21.014,
    -20.195, -19.342, -18.475, -17.620, -24.187, -23.763, -23.183, -22.519,
    -21.800, -21.018, -20.199, -19.346, -18.479, -17.624, -24.193, -23.769,
    -23.189, -22.524, -21.805, -21.023, -20.204, -19.351, -18.483, -17.628,
    -24.198, -23.775, -23.194, -22.529, -21.810, -21.028, -20.208, -19.355,
    -18.487, -17.632, -24.204, -23.780, -23.200, -22.535, -21.815, -21.033,
    -20.213, -19.359, -18.491, -17.635, -24.210, -23.786, -23.206, -22.540,
    -21.820, -21.038, -20.217, -19.363, -18.495, -17.639, -24.216, -23.792,
    -23.211, -22.545, -21.825, -21.042, -20.222, -19.368, -18.499, -17.643,
    -24.222, -23.798, -23.217, -22.551, -21.830, -21.047, -20.226, -19.372,
    -18.503, -17.647, -24.228, -23.804, -23.222, -22.556, -21.835, -21.052,
    -20.231, -19.376, -18.507, -17.650, -24.234, -23.809, -23.228, -22.562,
    -21.840, -21.057, -20.235, -19.380, -18.511, -17.654, -24.240, -23.815,
    -23.234, -22.567, -21.845, -21.062, -20.240, -19.384, -18.515, -17.658,
    -24.246, -23.821, -23.239, -22.572, -21.851, -21.066, -20.244, -19.389,
    -18.519, -17.661, -24.252, -23.827, -23.245, -22.578, -21.856, -21.071,
    -20.249, -19.393, -18.523, -17.665, -24.258, -23.833, -23.250, -22.583,
    -21.861, -21.076, -20.253, -19.397, -18.527, -17.669, -24.263, -23.838,
    -23.256, -22.588, -21.866, -21.081, -20.258, -19.401, -18.531, -17.673,
    -24.269, -23.844, -23.262, -22.594, -21.871, -21.086, -20.262, -19.406,
    -18.535, -17.676, -24.275, -23.850, -23.267, -22.599, -21.876, -21.090,
    -20.267, -19.410, -18.539, -17.680, -24.281, -23.856, -23.273, -22.604,
    -21.881, -21.095, -20.271, -19.414, -18.543, -17.684, -24.287, -23.861,
    -23.278, -22.610, -21.886, -21.100, -20.276, -19.418, -18.547, -17.688,
    -24.293, -23.867, -23.284, -22.615, -21.891, -21.105, -20.280, -19.422,
    -18.551, -17.691, -24.299, -23.873, -23.289, -22.620, -21.896, -21.110,
    -20.285, -19.427, -18.555, -17.695, -24.305, -23.879, -23.295, -22.626,
    -21.901, -21.114, -20.289, -19.431, -18.559, -17.699, -24.311, -23.885,
    -23.301, -22.631, -21.906, -21.119, -20.294, -19.435, -18.563, -17.703,
    -24.317, -23.890, -23.306, -22.636, -21.912, -21.124, -20.298, -19.439,
    -18.567, -17.706, -24.323, -23.896, -23.312, -22.642, -21.917, -21.129,
    -20.303, -19.444, -18.571, -17.710, -24.329, -23.902, -23.317, -22.647,
    -21.922, -21.133, -20.307, -19.448, -18.575, -17.714, -24.334, -23.908,
    -23.323, -22.652, -21.927, -21.138, -20.312, -19.452, -18.579, -17.717,
    -24.340, -23.914, -23.329, -22.658, -21.932, -21.143, -20.316, -19.456,
    -18.582, -17.721, -24.346, -23.919, -23.334, -22.663, -21.937, -21.148,
    -20.321, -19.460, -18.586, -17.725, -24.352, -23.925, -23.340, -22.669,
    -21.942, -21.153, -20.325, -19.465, -18.590, -17.729, -24.358, -23.931,
    -23.345, -22.674, -21.947, -21.157, -20.330, -19.469, -18.594, -17.732,
    -24.364, -23.937, -23.351, -22.679, -21.952, -21.162, -20.334, -19.473,
    -18.598, -17.736, -24.370, -23.943, -23.357, -22.685, -21.957, -21.167,
    -20.339, -19.477, -18.602, -17.740, -24.376, -23.948, -23.362, -22.690,
    -21.962, -21.172, -20.343, -19.482, -18.606, -17.744, -24.382, -23.954,
    -23.368, -22.695, -21.967, -21.177, -20.348, -19.486, -18.610, -17.747,
    -24.388, -23.960, -23.373, -22.701, -21.972, -21.181, -20.352, -19.490,
    -18.614, -17.751, -24.394, -23.966, -23.379, -22.706, -21.977, -21.186,
    -20.357, -19.494, -18.618, -17.755, -24.399, -23.971, -23.384, -22.711,
    -21.983, -21.191, -20.361, -19.498, -18.622, -17.758, -24.405, -23.977,
    -23.390, -22.717, -21.988, -21.196, -20.366, -19.503, -18.626, -17.762,
    -24.411, -23.983, -23.396, -22.722, -21.993, -21.201, -20.370, -19.507,
    -18.630, -17.766, -24.417, -23.989, -23.401, -22.727, -21.998, -21.205,
    -20.375, -19.511, -18.634, -17.770, -24.423, -23.995, -23.407, -22.733,
    -22.003, -21.210, -20.379, -19.515, -18.638, -17.773, -24.429, -24.000,
    -23.412, -22.738, -22.008, -21.215, -20.384, -19.520, -18.642, -17.777,
    -24.435, -24.006, -23.418, -22.743, -22.013, -21.220, -20.388, -19.524,
    -18.646, -17.781, -24.441, -24.012, -23.424, -22.749, -22.018, -21.225,
    -20.393, -19.528, -18.650, -17.785, -24.447, -24.018, -23.429, -22.754,
    -22.023, -21.229, -20.397, -19.532, -18.654, -17.788, -24.453, -24.024,
    -23.435, -22.759, -22.028, -21.234, -20.402, -19.536, -18.658, -17.792,
    -24.459, -24.029, -23.440, -22.765, -22.033, -21.239, -20.406, -19.541,
    -18.662, -17.796, -24.464, -24.035, -23.446, -22.770, -22.038, -21.244,
    -20.411, -19.545, -18.666, -17.799, -24.470, -24.041, -23.452, -22.775,
    -22.043, -21.248, -20.415, -19.549, -18.670, -17.803, -24.476, -24.047,
    -23.457, -22.781, -22.049, -21.253, -20.420, -19.553, -18.674, -17.807,
    -24.482, -24.052, -23.463, -22.786, -22.054, -21.258, -20.424, -19.557,
    -18.678, -17.811, -24.488, -24.058, -23.468, -22.791, -22.059, -21.263,
    -20.429, -19.562, -18.681, -17.814, -24.494, -24.064, -23.474, -22.797,
    -22.064, -21.268, -20.433, -19.566, -18.685, -17.818, -24.500, -24.070,
    -23.479, -22.802, -22.069, -21.272, -20.438, -19.570, -18.689, -17.822,
    -24.506, -24.076, -23.485, -22.807, -22.074, -21.277, -20.442, -19.574,
    -18.693, -17.825, -24.512, -24.081, -23.491, -22.813, -22.079, -21.282,
    -20.447, -19.579, -18.697, -17.829, -24.518, -24.087, -23.496, -22.818,
    -22.084, -21.287, -20.451, -19.583, -18.701, -17.833, -24.524, -24.093,
    -23.502, -22.823, -22.089, -21.292, -20.456, -19.587, -18.705, -17.837,
    -24.530, -24.099, -23.507, -22.829, -22.094, -21.296, -20.460, -19.591,
    -18.709, -17.840, -24.535, -24.104, -23.513, -22.834, -22.099, -21.301,
    -20.465, -19.595, -18.713, -17.844, -24.541, -24.110, -23.519, -22.840,
    -22.104, -21.306, -20.469, -19.600, -18.717, -17.848, -24.547, -24.116,
    -23.524, -22.845, -22.109, -21.311, -20.474, -19.604, -18.721, -17.851,
    -24.553, -24.122, -23.530, -22.850, -22.114, -21.315, -20.478, -19.608,
    -18.725, -17.855, -24.559, -24.128, -23.535, -22.856, -22.120, -21.320,
    -20.483, -19.612, -18.729, -17.859, -24.565, -24.133, -23.541, -22.861,
    -22.125, -21.325, -20.487, -19.617, -18.733, -17.863, -24.571, -24.139,
    -23.546, -22.866, -22.130, -21.330, -20.492, -19.621, -18.737, -17.866,
    -24.577, -24.145, -23.552, -22.872, -22.135, -21.335, -20.496, -19.625,
    -18.741, -17.870, -24.583, -24.151, -23.558, -22.877, -22.140, -21.339,
    -20.501, -19.629, -18.745, -17.874, -24.589, -24.157, -23.563, -22.882,
    -22.145, -21.344, -20.505, -19.633, -18.749, -17.877, -24.595, -24.162,
    -23.569, -22.888, -22.150, -21.349, -20.510, -19.638, -18.753, -17.881,
    -24.600, -24.168, -23.574, -22.893, -22.155, -21.354, -20.514, -19.642,
    -18.757, -17.885, -24.606, -24.174, -23.580, -22.898, -22.160, -21.359,
    -20.519, -19.646, -18.761, -17.889, -24.612, -24.180, -23.586, -22.904,
    -22.165, -21.363, -20.523, -19.650, -18.765, -17.892, -24.618, -24.185,
    -23.591, -22.909, -22.170, -21.368, -20.528, -19.654, -18.768, -17.896,
    -24.624, -24.191, -23.597, -22.914, -22.175, -21.373, -20.532, -19.659,
    -18.772, -17.900, -24.630, -24.197, -23.602, -22.920, -22.180, -21.378,
    -20.537, -19.663, -18.776, -17.903, -24.636, -24.203, -23.608, -22.925,
    -22.185, -21.382, -20.541, -19.667, -18.780, -17.907, -24.642, -24.209,
    -23.613, -22.930, -22.190, -21.387, -20.546, -19.671, -18.784, -17.911,
    -24.648, -24.214, -23.619, -22.936, -22.196, -21.392, -20.550, -19.675,
    -18.788, -17.915, -24.654, -24.220, -23.625, -22.941, -22.201, -21.397,
    -20.555, -19.680, -18.792, -17.918, -24.660, -24.226, -23.630, -22.946,
    -22.206, -21.402, -20.559, -19.684, -18.796, -17.922, -24.665, -24.232,
    -23.636, -22.952, -22.211, -21.406, -20.564, -19.688, -18.800, -17.926,
    -24.671, -24.237, -23.641, -22.957, -22.216, -21.411, -20.568, -19.692,
    -18.804, -17.929, -24.677, -24.243, -23.647, -22.962, -22.221, -21.416,
    -20.573, -19.697, -18.808, -17.933, -24.683, -24.249, -23.653, -22.968,
    -22.226, -21.421, -20.577, -19.701, -18.812, -17.937, -24.689, -24.255,
    -23.658, -22.973, -22.231, -21.426, -20.582, -19.705, -18.816, -17.941,
    -24.695, -24.261, -23.664, -22.978, -22.236, -21.430, -20.586, -19.709,
    -18.820, -17.944, -24.701, -24.266, -23.669, -22.984, -22.241, -21.435,
    -20.591, -19.713, -18.824, -17.948, -24.707, -24.272, -23.675, -22.989,
    -22.246, -21.440, -20.595, -19.718, -18.828, -17.952, -24.713, -24.278,
    -23.680, -22.994, -22.251, -21.445, -20.600, -19.722, -18.832, -17.955,
    -24.719, -24.284, -23.686, -23.000, -22.256, -21.449, -20.604, -19.726,
    -18.836, -17.959, -24.725, -24.289, -23.692, -23.005, -22.261, -21.454,
    -20.609, -19.730, -18.840, -17.963, -24.730, -24.295, -23.697, -23.010,
    -22.266, -21.459, -20.613, -19.734, -18.844, -17.966, -24.736, -24.301,
    -23.703, -23.016, -22.272, -21.464, -20.618, -19.739, -18.847, -17.970,
    -24.742, -24.307, -23.708, -23.021, -22.277, -21.469, -20.622, -19.743,
    -18.851, -17.974, -24.748, -24.313, -23.714, -23.026, -22.282, -21.473,
    -20.627, -19.747, -18.855, -17.978, -24.754, -24.318, -23.719, -23.032,
    -22.287, -21.478, -20.631, -19.751, -18.859, -17.981, -24.760, -24.324,
    -23.725, -23.037, -22.292, -21.483, -20.636, -19.755, -18.863, -17.985,
    -24.766, -24.330, -23.731, -23.042, -22.297, -21.488, -20.640, -19.760,
    -18.867, -17.989, -24.772, -24.336, -23.736, -23.048, -22.302, -21.492,
    -20.645, -19.764, -18.871, -17.992, -24.778, -24.341, -23.742, -23.053,
    -22.307, -21.497, -20.649, -19.768, -18.875, -17.996, -24.784, -24.347,
    -23.747, -23.058, -22.312, -21.502, -20.654, -19.772, -18.879, -18.000,
    -24.790, -24.353, -23.753, -23.064, -22.317, -21.507, -20.658, -19.777,
    -18.883, -18.003, -24.795, -24.359, -23.759, -23.069, -22.322, -21.512,
    -20.662, -19.781, -18.887, -18.007, -24.801, -24.365, -23.764, -23.074,
    -22.327, -21.516, -20.667, -19.785, -18.891, -18.011, -24.807, -24.370,
    -23.770, -23.080, -22.332, -21.521, -20.671, -19.789, -18.895, -18.015,
    -24.813, -24.376, -23.775, -23.085, -22.337, -21.526, -20.676, -19.793,
    -18.899, -18.018, -24.819, -24.382, -23.781, -23.090, -22.342, -21.531,
    -20.680, -19.798, -18.903, -18.022, -24.825, -24.388, -23.786, -23.096,
    -22.348, -21.535, -20.685, -19.802, -18.907, -18.026, -24.831, -24.393,
    -23.792, -23.101, -22.353, -21.540, -20.689, -19.806, -18.911, -18.029,
    -24.837, -24.399, -23.798, -23.106, -22.358, -21.545, -20.694, -19.810,
    -18.915, -18.033, -24.843, -24.405, -23.803, -23.112, -22.363, -21.550,
    -20.698, -19.814, -18.918, -18.037, -24.849, -24.411, -23.809, -23.117,
    -22.368, -21.555, -20.703, -19.819, -18.922, -18.040, -24.854, -24.417,
    -23.814, -23.122, -22.373, -21.559, -20.707, -19.823, -18.926, -18.044,
    -24.860, -24.422, -23.820, -23.128, -22.378, -21.564, -20.712, -19.827,
    -18.930, -18.048, -24.866, -24.428, -23.825, -23.133, -22.383, -21.569,
    -20.716, -19.831, -18.934, -18.052, -24.872, -24.434, -23.831, -23.138,
    -22.388, -21.574, -20.721, -19.835, -18.938, -18.055, -24.878, -24.440,
    -23.837, -23.144, -22.393, -21.578, -20.725, -19.840, -18.942, -18.059,
    -24.884, -24.445, -23.842, -23.149, -22.398, -21.583, -20.730, -19.844,
    -18.946, -18.063, -24.890, -24.451, -23.848, -23.154, -22.403, -21.588,
    -20.734, -19.848, -18.950, -18.066, -24.896, -24.457, -23.853, -23.160,
    -22.408, -21.593, -20.739, -19.852, -18.954, -18.070, -24.902, -24.463,
    -23.859, -23.165, -22.413, -21.598, -20.743, -19.856, -18.958, -18.074,
    -24.908, -24.469, -23.864, -23.170, -22.418, -21.602, -20.748, -19.861,
    -18.962, -18.077, -24.914, -24.474, -23.870, -23.176, -22.423, -21.607,
    -20.752, -19.865, -18.966, -18.081, -24.919, -24.480, -23.876, -23.181,
    -22.429, -21.612, -20.757, -19.869, -18.970, -18.085, -24.925, -24.486,
    -23.881, -23.186, -22.434, -21.617, -20.761, -19.873, -18.974, -18.089,
    -24.931, -24.492, -23.887, -23.192, -22.439, -21.621, -20.766, -19.877,
    -18.978, -18.092, -24.937, -24.497, -23.892, -23.197, -22.444, -21.626,
    -20.770, -19.882, -18.981, -18.096, -24.943, -24.503, -23.898, -23.202,
    -22.449, -21.631, -20.775, -19.886, -18.985, -18.100, -24.949, -24.509,
    -23.904, -23.208, -22.454, -21.636, -20.779, -19.890, -18.989, -18.103,
    -24.955, -24.515, -23.909, -23.213, -22.459, -21.641, -20.784, -19.894,
    -18.993, -18.107, -24.961, -24.521, -23.915, -23.218, -22.464, -21.645,
    -20.788, -19.898, -18.997, -18.111, -24.967, -24.526, -23.920, -23.224,
    -22.469, -21.650, -20.793, -19.903, -19.001, -18.114, -24.973, -24.532,
    -23.926, -23.229, -22.474, -21.655, -20.797, -19.907, -19.005, -18.118,
    -24.979, -24.538, -23.931, -23.234, -22.479, -21.660, -20.802, -19.911,
    -19.009, -18.122, -24.984, -24.544, -23.937, -23.240, -22.484, -21.664,
    -20.806, -19.915, -19.013, -18.125, -24.990, -24.549, -23.943, -23.245,
    -22.489, -21.669, -20.811, -19.919, -19.017, -18.129, -24.996, -24.555,
    -23.948, -23.250, -22.494, -21.674, -20.815, -19.924, -19.021, -18.133,
    -25.002, -24.561, -23.954, -23.256, -22.499, -21.679, -20.820, -19.928,
    -19.025, -18.137, -25.008, -24.567, -23.959, -23.261, -22.504, -21.684,
    -20.824, -19.932, -19.029, -18.140, -25.014, -24.572, -23.965, -23.266,
    -22.509, -21.688, -20.829, -19.936, -19.033, -18.144, -25.020, -24.578,
    -23.970, -23.272, -22.515, -21.693, -20.833, -19.940, -19.037, -18.148,
    -25.026, -24.584, -23.976, -23.277, -22.520, -21.698, -20.837, -19.945,
    -19.041, -18.151, -25.032, -24.590, -23.982, -23.282, -22.525, -21.703,
    -20.842, -19.949, -19.044, -18.155, -25.038, -24.596, -23.987, -23.288,
    -22.530, -21.707, -20.846, -19.953, -19.048, -18.159, -25.043, -24.601,
    -23.993, -23.293, -22.535, -21.712, -20.851, -19.957, -19.052, -18.162,
    -25.049, -24.607, -23.998, -23.298, -22.540, -21.717, -20.855, -19.961,
    -19.056, -18.166, -25.055, -24.613, -24.004, -23.304, -22.545, -21.722,
    -20.860, -19.966, -19.060, -18.170, -25.061, -24.619, -24.009, -23.309,
    -22.550, -21.726, -20.864, -19.970, -19.064, -18.173, -25.067, -24.624,
    -24.015, -23.314, -22.555, -21.731, -20.869, -19.974, -19.068, -18.177,
    -25.073, -24.630, -24.021, -23.320, -22.560, -21.736, -20.873, -19.978,
    -19.072, -18.181, -25.079, -24.636, -24.026, -23.325, -22.565, -21.741,
    -20.878, -19.982, -19.076, -18.185, -25.085, -24.642, -24.032, -23.330,
    -22.570, -21.746, -20.882, -19.987, -19.080, -18.188, -25.091, -24.648,
    -24.037, -23.336, -22.575, -21.750, -20.887, -19.991, -19.084, -18.192,
    -25.097, -24.653, -24.043, -23.341, -22.580, -21.755, -20.891, -19.995,
    -19.088, -18.196, -25.102, -24.659, -24.048, -23.346, -22.585, -21.760,
    -20.896, -19.999, -19.092, -18.199, -25.108, -24.665, -24.054, -23.352,
    -22.590, -21.765, -20.900, -20.003, -19.096, -18.203, -25.114, -24.671,
    -24.060, -23.357, -22.595, -21.769, -20.905, -20.008, -19.100, -18.207,
    -25.120, -24.676, -24.065, -23.362, -22.601, -21.774, -20.909, -20.012,
    -19.103, -18.210, -25.126, -24.682, -24.071, -23.368, -22.606, -21.779,
    -20.914, -20.016, -19.107, -18.214, -25.132, -24.688, -24.076, -23.373,
    -22.611, -21.784, -20.918, -20.020, -19.111, -18.218, -25.138, -24.694,
    -24.082, -23.378, -22.616, -21.789, -20.923, -20.024, -19.115, -18.221,
    -25.144, -24.699, -24.087, -23.384, -22.621, -21.793, -20.927, -20.029,
    -19.119, -18.225, -25.150, -24.705, -24.093, -23.389, -22.626, -21.798,
    -20.932, -20.033, -19.123, -18.229, -25.156, -24.711, -24.099, -23.394,
    -22.631, -21.803, -20.936, -20.037, -19.127, -18.232, -25.162, -24.717,
    -24.104, -23.400, -22.636, -21.808, -20.941, -20.041, -19.131, -18.236,
    -25.167, -24.723, -24.110, -23.405, -22.641, -21.812, -20.945, -20.045,
    -19.135, -18.240, -25.173, -24.728, -24.115, -23.410, -22.646, -21.817,
    -20.950, -20.050, -19.139, -18.243, -25.179, -24.734, -24.121, -23.415,
    -22.651, -21.822, -20.954, -20.054, -19.143, -18.247, -25.185, -24.740,
    -24.126, -23.421, -22.656, -21.827, -20.958, -20.058, -19.147, -18.251,
    -25.191, -24.746, -24.132, -23.426, -22.661, -21.831, -20.963, -20.062,
    -19.151, -18.255, -25.197, -24.751, -24.137, -23.431, -22.666, -21.836,
    -20.967, -20.066, -19.155, -18.258, -25.203, -24.757, -24.143, -23.437,
    -22.671, -21.841, -20.972, -20.071, -19.158, -18.262, -25.209, -24.763,
    -24.149, -23.442, -22.676, -21.846, -20.976, -20.075, -19.162, -18.266,
    -25.215, -24.769, -24.154, -23.447, -22.681, -21.851, -20.981, -20.079,
    -19.166, -18.269, -25.221, -24.774, -24.160, -23.453, -22.686, -21.855,
    -20.985, -20.083, -19.170, -18.273, -25.226, -24.780, -24.165, -23.458,
    -22.691, -21.860, -20.990, -20.087, -19.174, -18.277, -25.232, -24.786,
    -24.171, -23.463, -22.697, -21.865, -20.994, -20.092, -19.178, -18.280,
    -25.238, -24.792, -24.176, -23.469, -22.702, -21.870, -20.999, -20.096,
    -19.182, -18.284, -25.244, -24.798, -24.182, -23.474, -22.707, -21.874,
    -21.003, -20.100, -19.186, -18.288, -25.250, -24.803, -24.188, -23.479,
    -22.712, -21.879, -21.008, -20.104, -19.190, -18.291, -25.256, -24.809,
    -24.193, -23.485, -22.717, -21.884, -21.012, -20.108, -19.194, -18.295,
    -25.262, -24.815, -24.199, -23.490, -22.722, -21.889, -21.017, -20.112,
    -19.198, -18.299, -25.268, -24.821, -24.204, -23.495, -22.727, -21.893,
    -21.021, -20.117, -19.202, -18.302, -25.274, -24.826, -24.210, -23.501,
    -22.732, -21.898, -21.026, -20.121, -19.206, -18.306, -25.280, -24.832,
    -24.215, -23.506, -22.737, -21.903, -21.030, -20.125, -19.210, -18.310,
    -25.285, -24.838, -24.221, -23.511, -22.742, -21.908, -21.035, -20.129,
    -19.213, -18.313, -25.291, -24.844, -24.227, -23.517, -22.747, -21.913,
    -21.039, -20.133, -19.217, -18.317, -25.297, -24.849, -24.232, -23.522,
    -22.752, -21.917, -21.044, -20.138, -19.221, -18.321, -25.303, -24.855,
    -24.238, -23.527, -22.757, -21.922, -21.048, -20.142, -19.225, -18.324,
    -25.309, -24.861, -24.243, -23.533, -22.762, -21.927, -21.053, -20.146,
    -19.229, -18.328, -25.315, -24.867, -24.249, -23.538, -22.767, -21.932,
    -21.057, -20.150, -19.233, -18.332, -25.321, -24.873, -24.254, -23.543,
    -22.772, -21.936, -21.061, -20.154, -19.237, -18.335, -25.327, -24.878,
    -24.260, -23.549, -22.777, -21.941, -21.066, -20.159, -19.241, -18.339,
    -25.333, -24.884, -24.266, -23.554, -22.782, -21.946, -21.070, -20.163,
    -19.245, -18.343, -25.339, -24.890, -24.271, -23.559, -22.787, -21.951,
    -21.075, -20.167, -19.249, -18.346, -25.344, -24.896, -24.277, -23.565,
    -22.793, -21.955, -21.079, -20.171, -19.253, -18.350, -25.350, -24.901,
    -24.282, -23.570, -22.798, -21.960, -21.084, -20.175, -19.257, -18.354,
    -25.356, -24.907, -24.288, -23.575, -22.803, -21.965, -21.088, -20.180,
    -19.261, -18.358, -25.362, -24.913, -24.293, -23.580, -22.808, -21.970,
    -21.093, -20.184, -19.264, -18.361, -25.368, -24.919, -24.299, -23.586,
    -22.813, -21.974, -21.097, -20.188, -19.268, -18.365, -25.374, -24.924,
    -24.304, -23.591, -22.818, -21.979, -21.102, -20.192, -19.272, -18.369,
    -25.380, -24.930, -24.310, -23.596, -22.823, -21.984, -21.106, -20.196,
    -19.276, -18.372, -25.386, -24.936, -24.316, -23.602, -22.828, -21.989,
    -21.111, -20.201, -19.280, -18.376, -25.392, -24.942, -24.321, -23.607,
    -22.833, -21.994, -21.115, -20.205, -19.284, -18.380, -25.398, -24.947,
    -24.327, -23.612, -22.838, -21.998, -21.120, -20.209, -19.288, -18.383,
    -25.403, -24.953, -24.332, -23.618, -22.843, -22.003, -21.124, -20.213,
    -19.292, -18.387, -25.409, -24.959, -24.338, -23.623, -22.848, -22.008,
    -21.129, -20.217, -19.296, -18.391, -25.415, -24.965, -24.343, -23.628,
    -22.853, -22.013, -21.133, -20.221, -19.300, -18.394, -25.421, -24.971,
    -24.349, -23.634, -22.858, -22.017, -21.138, -20.226, -19.304, -18.398,
    -25.427, -24.976, -24.355, -23.639, -22.863, -22.022, -21.142, -20.230,
    -19.308, -18.402, -25.433, -24.982, -24.360, -23.644, -22.868, -22.027,
    -21.146, -20.234, -19.312, -18.405, -25.439, -24.988, -24.366, -23.650,
    -22.873, -22.032, -21.151, -20.238, -19.315, -18.409, -25.445, -24.994,
    -24.371, -23.655, -22.878, -22.036, -21.155, -20.242, -19.319, -18.413,
    -25.451, -24.999, -24.377, -23.660, -22.883, -22.041, -21.160, -20.247,
    -19.323, -18.416, -25.457, -25.005, -24.382, -23.666, -22.888, -22.046,
    -21.164, -20.251, -19.327, -18.420, -25.462, -25.011, -24.388, -23.671,
    -22.893, -22.051, -21.169, -20.255, -19.331, -18.424, -25.468, -25.017,
    -24.393, -23.676, -22.899, -22.055, -21.173, -20.259, -19.335, -18.427,
    -25.474, -25.022, -24.399, -23.682, -22.904, -22.060, -21.178, -20.263,
    -19.339, -18.431, -25.480, -25.028, -24.405, -23.687, -22.909, -22.065,
    -21.182, -20.268, -19.343, -18.435, -25.486, -25.034, -24.410, -23.692,
    -22.914, -22.070, -21.187, -20.272, -19.347, -18.438, -25.492, -25.040,
    -24.416, -23.698, -22.919, -22.075, -21.191, -20.276, -19.351, -18.442,
    -25.498, -25.045, -24.421, -23.703, -22.924, -22.079, -21.196, -20.280,
    -19.355, -18.446, -25.504, -25.051, -24.427, -23.708, -22.929, -22.084,
    -21.200, -20.284, -19.359, -18.449, -25.510, -25.057, -24.432, -23.713,
    -22.934, -22.089, -21.205, -20.288, -19.362, -18.453, -25.516, -25.063,
    -24.438, -23.719, -22.939, -22.094, -21.209, -20.293, -19.366, -18.457,
    -25.521, -25.069, -24.444, -23.724, -22.944, -22.098, -21.214, -20.297,
    -19.370, -18.460, -25.527, -25.074, -24.449, -23.729, -22.949, -22.103,
    -21.218, -20.301, -19.374, -18.464, -25.533, -25.080, -24.455, -23.735,
    -22.954, -22.108, -21.223, -20.305, -19.378, -18.468, -25.539, -25.086,
    -24.460, -23.740, -22.959, -22.113, -21.227, -20.309, -19.382, -18.471,
    -25.545, -25.092, -24.466, -23.745, -22.964, -22.117, -21.231, -20.314,
    -19.386, -18.475, -25.551, -25.097, -24.471, -23.751, -22.969, -22.122,
    -21.236, -20.318, -19.390, -18.479, -25.557, -25.103, -24.477, -23.756,
    -22.974, -22.127, -21.240, -20.322, -19.394, -18.482, -25.563, -25.109,
    -24.482, -23.761, -22.979, -22.132, -21.245, -20.326, -19.398, -18.486,
    -25.569, -25.115, -24.488, -23.767, -22.984, -22.136, -21.249, -20.330,
    -19.402, -18.490, -25.575, -25.120, -24.494, -23.772, -22.989, -22.141,
    -21.254, -20.334, -19.406, -18.493, -25.580, -25.126, -24.499, -23.777,
    -22.994, -22.146, -21.258, -20.339, -19.409, -18.497, -25.586, -25.132,
    -24.505, -23.783, -22.999, -22.151, -21.263, -20.343, -19.413, -18.501,
    -25.592, -25.138, -24.510, -23.788, -23.005, -22.155, -21.267, -20.347,
    -19.417, -18.504, -25.598, -25.143, -24.516, -23.793, -23.010, -22.160,
    -21.272, -20.351, -19.421, -18.508, -25.604, -25.149, -24.521, -23.799,
    -23.015, -22.165, -21.276, -20.355, -19.425, -18.512, -25.610, -25.155,
    -24.527, -23.804, -23.020, -22.170, -21.281, -20.360, -19.429, -18.515,
    -25.616, -25.161, -24.532, -23.809, -23.025, -22.174, -21.285, -20.364,
    -19.433, -18.519, -25.622, -25.167, -24.538, -23.814, -23.030, -22.179,
    -21.290, -20.368, -19.437, -18.523, -25.628, -25.172, -24.544, -23.820,
    -23.035, -22.184, -21.294, -20.372, -19.441, -18.526, -25.634, -25.178,
    -24.549, -23.825, -23.040, -22.189, -21.299, -20.376, -19.445, -18.530,
    -25.639, -25.184, -24.555, -23.830, -23.045, -22.194, -21.303, -20.381,
    -19.449, -18.534, -25.645, -25.190, -24.560, -23.836, -23.050, -22.198,
    -21.307, -20.385, -19.452, -18.537, -25.651, -25.195, -24.566, -23.841,
    -23.055, -22.203, -21.312, -20.389, -19.456, -18.541, -25.657, -25.201,
    -24.571, -23.846, -23.060, -22.208, -21.316, -20.393, -19.460, -18.545,
    -25.663, -25.207, -24.577, -23.852, -23.065, -22.213, -21.321, -20.397,
    -19.464, -18.548, -25.669, -25.213, -24.582, -23.857, -23.070, -22.217,
    -21.325, -20.401, -19.468, -18.552, -25.675, -25.218, -24.588, -23.862,
    -23.075, -22.222, -21.330, -20.406, -19.472, -18.556, -25.681, -25.224,
    -24.594, -23.868, -23.080, -22.227, -21.334, -20.410, -19.476, -18.559,
    -25.687, -25.230, -24.599, -23.873, -23.085, -22.232, -21.339, -20.414,
    -19.480, -18.563, -25.692, -25.236, -24.605, -23.878, -23.090, -22.236,
    -21.343, -20.418, -19.484, -18.567, -25.698, -25.241, -24.610, -23.884,
    -23.095, -22.241, -21.348, -20.422, -19.488, -18.570, -25.704, -25.247,
    -24.616, -23.889, -23.100, -22.246, -21.352, -20.427, -19.492, -18.574,
    -25.710, -25.253, -24.621, -23.894, -23.105, -22.251, -21.357, -20.431,
    -19.496, -18.578, -25.716, -25.259, -24.627, -23.899, -23.110, -22.255,
    -21.361, -20.435, -19.499, -18.581, -25.722, -25.264, -24.633, -23.905,
    -23.115, -22.260, -21.366, -20.439, -19.503, -18.585, -25.728, -25.270,
    -24.638, -23.910, -23.120, -22.265, -21.370, -20.443, -19.507, -18.589,
    -25.734, -25.276, -24.644, -23.915, -23.126, -22.270, -21.374, -20.447,
    -19.511, -18.592, -25.740, -25.282, -24.649, -23.921, -23.131, -22.274,
    -21.379, -20.452, -19.515, -18.596, -25.746, -25.287, -24.655, -23.926,
    -23.136, -22.279, -21.383, -20.456, -19.519, -18.600, -25.751, -25.293,
    -24.660, -23.931, -23.141, -22.284, -21.388, -20.460, -19.523, -18.603,
    -25.757, -25.299, -24.666, -23.937, -23.146, -22.289, -21.392, -20.464,
    -19.527, -18.607, -25.763, -25.305, -24.671, -23.942, -23.151, -22.293,
    -21.397, -20.468, -19.531, -18.611, -25.769, -25.311, -24.677, -23.947,
    -23.156, -22.298, -21.401, -20.472, -19.535, -18.614, -25.775, -25.316,
    -24.683, -23.953, -23.161, -22.303, -21.406, -20.477, -19.539, -18.618,
    -25.781, -25.322, -24.688, -23.958, -23.166, -22.308, -21.410, -20.481,
    -19.542, -18.621, -25.787, -25.328, -24.694, -23.963, -23.171, -22.312,
    -21.415, -20.485, -19.546, -18.625, -25.793, -25.334, -24.699, -23.968,
    -23.176, -22.317, -21.419, -20.489, -19.550, -18.629, -25.799, -25.339,
    -24.705, -23.974, -23.181, -22.322, -21.424, -20.493, -19.554, -18.632,
    -25.805, -25.345, -24.710, -23.979, -23.186, -22.327, -21.428, -20.498,
    -19.558, -18.636, -25.810, -25.351, -24.716, -23.984, -23.191, -22.331,
    -21.432, -20.502, -19.562, -18.640, -25.816, -25.357, -24.721, -23.990,
    -23.196, -22.336, -21.437, -20.506, -19.566, -18.643, -25.822, -25.362,
    -24.727, -23.995, -23.201, -22.341, -21.441, -20.510, -19.570, -18.647,
    -25.828, -25.368, -24.732, -24.000, -23.206, -22.346, -21.446, -20.514,
    -19.574, -18.651, -25.834, -25.374, -24.738, -24.006, -23.211, -22.350,
    -21.450, -20.518, -19.578, -18.654, -25.840, -25.380, -24.744, -24.011,
    -23.216, -22.355, -21.455, -20.523, -19.582, -18.658, -25.846, -25.385,
    -24.749, -24.016, -23.221, -22.360, -21.459, -20.527, -19.585, -18.662,
    -25.852, -25.391, -24.755, -24.022, -23.226, -22.365, -21.464, -20.531,
    -19.589, -18.665, -25.858, -25.397, -24.760, -24.027, -23.231, -22.370,
    -21.468, -20.535, -19.593, -18.669, -25.863, -25.403, -24.766, -24.032,
    -23.236, -22.374, -21.473, -20.539, -19.597, -18.673, -25.869, -25.408,
    -24.771, -24.038, -23.241, -22.379, -21.477, -20.544, -19.601, -18.676,
    -25.875, -25.414, -24.777, -24.043, -23.246, -22.384, -21.482, -20.548,
    -19.605, -18.680, -25.881, -25.420, -24.782, -24.048, -23.252, -22.389,
    -21.486, -20.552, -19.609, -18.684, -25.887, -25.426, -24.788, -24.053,
    -23.257, -22.393, -21.491, -20.556, -19.613, -18.687, -25.893, -25.431,
    -24.794, -24.059, -23.262, -22.398, -21.495, -20.560, -19.617, -18.691,
    -25.899, -25.437, -24.799, -24.064, -23.267, -22.403, -21.499, -20.564,
    -19.621, -18.695, -25.905, -25.443, -24.805, -24.069, -23.272, -22.408,
    -21.504, -20.569, -19.624, -18.698, -25.911, -25.449, -24.810, -24.075,
    -23.277, -22.412, -21.508, -20.573, -19.628, -18.702, -25.917, -25.454,
    -24.816, -24.080, -23.282, -22.417, -21.513, -20.577, -19.632, -18.706,
    -25.922, -25.460, -24.821, -24.085, -23.287, -22.422, -21.517, -20.581,
    -19.636, -18.709, -25.928, -25.466, -24.827, -24.091, -23.292, -22.427,
    -21.522, -20.585, -19.640, -18.713, -25.934, -25.472, -24.832, -24.096,
    -23.297, -22.431, -21.526, -20.589, -19.644, -18.717, -25.940, -25.477,
    -24.838, -24.101, -23.302, -22.436, -21.531, -20.594, -19.648, -18.720,
    -25.946, -25.483, -24.844, -24.106, -23.307, -22.441, -21.535, -20.598,
    -19.652, -18.724, -25.952, -25.489, -24.849, -24.112, -23.312, -22.446,
    -21.540, -20.602, -19.656, -18.727, -25.958, -25.495, -24.855, -24.117,
    -23.317, -22.450, -21.544, -20.606, -19.660, -18.731, -25.964, -25.500,
    -24.860, -24.122, -23.322, -22.455, -21.549, -20.610, -19.664, -18.735,
    -25.970, -25.506, -24.866, -24.128, -23.327, -22.460, -21.553, -20.615,
    -19.667, -18.738, -25.975, -25.512, -24.871, -24.133, -23.332, -22.465,
    -21.557, -20.619, -19.671, -18.742, -25.981, -25.518, -24.877, -24.138,
    -23.337, -22.469, -21.562, -20.623, -19.675, -18.746, -25.987, -25.523,
    -24.882, -24.144, -23.342, -22.474, -21.566, -20.627, -19.679, -18.749,
    -25.993, -25.529, -24.888, -24.149, -23.347, -22.479, -21.571, -20.631,
    -19.683, -18.753, -25.999, -25.535, -24.893, -24.154, -23.352, -22.484,
    -21.575, -20.635, -19.687, -18.757, -26.005, -25.541, -24.899, -24.160,
    -23.357, -22.488, -21.580, -20.640, -19.691, -18.760, -26.011, -25.547,
    -24.905, -24.165, -23.362, -22.493, -21.584, -20.644, -19.695, -18.764,
    -26.017, -25.552, -24.910, -24.170, -23.367, -22.498, -21.589, -20.648,
    -19.699, -18.768, -26.023, -25.558, -24.916, -24.175, -23.372, -22.503,
    -21.593, -20.652, -19.703, -18.771, -26.028, -25.564, -24.921, -24.181,
    -23.377, -22.507, -21.598, -20.656, -19.706, -18.775, -26.034, -25.570,
    -24.927, -24.186, -23.382, -22.512, -21.602, -20.660, -19.710, -18.779,
    -26.040, -25.575, -24.932, -24.191, -23.387, -22.517, -21.607, -20.665,
    -19.714, -18.782, -26.046, -25.581, -24.938, -24.197, -23.393, -22.522,
    -21.611, -20.669, -19.718, -18.786, -26.052, -25.587, -24.943, -24.202,
    -23.398, -22.526, -21.615, -20.673, -19.722, -18.790, -26.058, -25.593,
    -24.949, -24.207, -23.403, -22.531, -21.620, -20.677, -19.726, -18.793,
    -26.064, -25.598, -24.955, -24.213, -23.408, -22.536, -21.624, -20.681,
    -19.730, -18.797, -26.070, -25.604, -24.960, -24.218, -23.413, -22.541,
    -21.629, -20.685, -19.734, -18.800, -26.076, -25.610, -24.966, -24.223,
    -23.418, -22.545, -21.633, -20.690, -19.738, -18.804, -26.081, -25.616,
    -24.971, -24.228, -23.423, -22.550, -21.638, -20.694, -19.742, -18.808,
    -26.087, -25.621, -24.977, -24.234, -23.428, -22.555, -21.642, -20.698,
    -19.745, -18.811, -26.093, -25.627, -24.982, -24.239, -23.433, -22.560,
    -21.647, -20.702, -19.749, -18.815, -26.099, -25.633, -24.988, -24.244,
    -23.438, -22.564, -21.651, -20.706, -19.753, -18.819, -26.105, -25.639,
    -24.993, -24.250, -23.443, -22.569, -21.656, -20.710, -19.757, -18.822,
    -26.111, -25.644, -24.999, -24.255, -23.448, -22.574, -21.660, -20.715,
    -19.761, -18.826, -26.117, -25.650, -25.004, -24.260, -23.453, -22.579,
    -21.664, -20.719, -19.765, -18.830, -26.123, -25.656, -25.010, -24.266,
    -23.458, -22.583, -21.669, -20.723, -19.769, -18.833, -26.129, -25.662,
    -25.016, -24.271, -23.463, -22.588, -21.673, -20.727, -19.773, -18.837,
    -26.135, -25.667, -25.021, -24.276, -23.468, -22.593, -21.678, -20.731,
    -19.777, -18.841, -26.140, -25.673, -25.027, -24.282, -23.473, -22.598,
    -21.682, -20.736, -19.781, -18.844, -26.146, -25.679, -25.032, -24.287,
    -23.478, -22.602, -21.687, -20.740, -19.784, -18.848, -26.152, -25.685,
    -25.038, -24.292, -23.483, -22.607, -21.691, -20.744, -19.788, -18.851,
    -26.158, -25.690, -25.043, -24.297, -23.488, -22.612, -21.696, -20.748,
    -19.792, -18.855, -26.164, -25.696, -25.049, -24.303, -23.493, -22.616,
    -21.700, -20.752, -19.796, -18.859, -26.170, -25.702, -25.054, -24.308,
    -23.498, -22.621, -21.705, -20.756, -19.800, -18.862, -26.176, -25.708,
    -25.060, -24.313, -23.503, -22.626, -21.709, -20.761, -19.804, -18.866,
    -26.182, -25.713, -25.065, -24.319, -23.508, -22.631, -21.713, -20.765,
    -19.808, -18.870, -26.188, -25.719, -25.071, -24.324, -23.513, -22.635,
    -21.718, -20.769, -19.812, -18.873, -26.193, -25.725, -25.077, -24.329,
    -23.518, -22.640, -21.722, -20.773, -19.816, -18.877, -26.199, -25.731,
    -25.082, -24.335, -23.523, -22.645, -21.727, -20.777, -19.820, -18.881,
    -26.205, -25.736, -25.088, -24.340, -23.528, -22.650, -21.731, -20.781,
    -19.823, -18.884, -26.211, -25.742, -25.093, -24.345, -23.533, -22.654,
    -21.736, -20.786, -19.827, -18.888, -26.217, -25.748, -25.099, -24.350,
    -23.538, -22.659, -21.740, -20.790, -19.831, -18.892, -26.223, -25.754,
    -25.104, -24.356, -23.543, -22.664, -21.745, -20.794, -19.835, -18.895,
    -26.229, -25.759, -25.110, -24.361, -23.548, -22.669, -21.749, -20.798,
    -19.839, -18.899, -26.235, -25.765, -25.115, -24.366, -23.554, -22.673,
    -21.754, -20.802, -19.843, -18.903, -26.241, -25.771, -25.121, -24.372,
    -23.559, -22.678, -21.758, -20.806, -19.847, -18.906, -26.246, -25.777,
    -25.126, -24.377, -23.564, -22.683, -21.762, -20.811, -19.851, -18.910,
    -26.252, -25.782, -25.132, -24.382, -23.569, -22.688, -21.767, -20.815,
    -19.855, -18.913, -26.258, -25.788, -25.138, -24.388, -23.574, -22.692,
    -21.771, -20.819, -19.858, -18.917, -26.264, -25.794, -25.143, -24.393,
    -23.579, -22.697, -21.776, -20.823, -19.862, -18.921, -26.270, -25.800,
    -25.149, -24.398, -23.584, -22.702, -21.780, -20.827, -19.866, -18.924,
    -26.276, -25.805, -25.154, -24.403, -23.589, -22.707, -21.785, -20.831,
    -19.870, -18.928, -26.282, -25.811, -25.160, -24.409, -23.594, -22.711,
    -21.789, -20.836, -19.874, -18.932, -26.288, -25.817, -25.165, -24.414,
    -23.599, -22.716, -21.794, -20.840, -19.878, -18.935, -26.294, -25.823,
    -25.171, -24.419, -23.604, -22.721, -21.798, -20.844, -19.882, -18.939,
    -26.299, -25.828, -25.176, -24.425, -23.609, -22.726, -21.803, -20.848,
    -19.886, -18.943, -26.305, -25.834, -25.182, -24.430, -23.614, -22.730,
    -21.807, -20.852, -19.890, -18.946, -26.311, -25.840, -25.187, -24.435,
    -23.619, -22.735, -21.811, -20.856, -19.894, -18.950, -26.317, -25.846,
    -25.193, -24.440, -23.624, -22.740, -21.816, -20.861, -19.897, -18.953,
    -26.323, -25.851, -25.198, -24.446, -23.629, -22.745, -21.820, -20.865,
    -19.901, -18.957, -26.329, -25.857, -25.204, -24.451, -23.634, -22.749,
    -21.825, -20.869, -19.905, -18.961, -26.335, -25.863, -25.210, -24.456,
    -23.639, -22.754, -21.829, -20.873, -19.909, -18.964, -26.341, -25.869,
    -25.215, -24.462, -23.644, -22.759, -21.834, -20.877, -19.913, -18.968,
    -26.347, -25.874, -25.221, -24.467, -23.649, -22.764, -21.838, -20.881,
    -19.917, -18.972, -26.352, -25.880, -25.226, -24.472, -23.654, -22.768,
    -21.843, -20.886, -19.921, -18.975, -26.358, -25.886, -25.232, -24.478,
    -23.659, -22.773, -21.847, -20.890, -19.925, -18.979, -26.364, -25.892,
    -25.237, -24.483, -23.664, -22.778, -21.852, -20.894, -19.929, -18.983,
    -26.370, -25.897, -25.243, -24.488, -23.669, -22.783, -21.856, -20.898,
    -19.932, -18.986, -26.376, -25.903, -25.248, -24.493, -23.674, -22.787,
    -21.860, -20.902, -19.936, -18.990, -26.382, -25.909, -25.254, -24.499,
    -23.679, -22.792, -21.865, -20.906, -19.940, -18.993, -26.388, -25.915,
    -25.259, -24.504, -23.684, -22.797, -21.869, -20.911, -19.944, -18.997,
    -26.394, -25.920, -25.265, -24.509, -23.689, -22.802, -21.874, -20.915,
    -19.948, -19.001, -26.400, -25.926, -25.271, -24.515, -23.694, -22.806,
    -21.878, -20.919, -19.952, -19.004, -26.405, -25.932, -25.276, -24.520,
    -23.699, -22.811, -21.883, -20.923, -19.956, -19.008, -26.411, -25.937,
    -25.282, -24.525, -23.704, -22.816, -21.887, -20.927, -19.960, -19.012,
    -26.417, -25.943, -25.287, -24.531, -23.709, -22.820, -21.892, -20.931,
    -19.964, -19.015, -26.423, -25.949, -25.293, -24.536, -23.714, -22.825,
    -21.896, -20.936, -19.967, -19.019, -26.429, -25.955, -25.298, -24.541,
    -23.719, -22.830, -21.901, -20.940, -19.971, -19.023, -26.435, -25.960,
    -25.304, -24.546, -23.724, -22.835, -21.905, -20.944, -19.975, -19.026,
    -26.441, -25.966, -25.309, -24.552, -23.729, -22.839, -21.909, -20.948,
    -19.979, -19.030, -26.447, -25.972, -25.315, -24.557, -23.734, -22.844,
    -21.914, -20.952, -19.983, -19.033, -26.453, -25.978, -25.320, -24.562,
    -23.739, -22.849, -21.918, -20.956, -19.987, -19.037, -26.458, -25.983,
    -25.326, -24.568, -23.745, -22.854, -21.923, -20.960, -19.991, -19.041,
    -26.464, -25.989, -25.331, -24.573, -23.750, -22.858, -21.927, -20.965,
    -19.995, -19.044, -26.470, -25.995, -25.337, -24.578, -23.755, -22.863,
    -21.932, -20.969, -19.999, -19.048, -26.476, -26.001, -25.343, -24.583,
    -23.760, -22.868, -21.936, -20.973, -20.002, -19.052, -26.482, -26.006,
    -25.348, -24.589, -23.765, -22.873, -21.941, -20.977, -20.006, -19.055,
    -26.488, -26.012, -25.354, -24.594, -23.770, -22.877, -21.945, -20.981,
    -20.010, -19.059, -26.494, -26.018, -25.359, -24.599, -23.775, -22.882,
    -21.949, -20.985, -20.014, -19.063, -26.500, -26.024, -25.365, -24.605,
    -23.780, -22.887, -21.954, -20.990, -20.018, -19.066, -26.505, -26.029,
    -25.370, -24.610, -23.785, -22.892, -21.958, -20.994, -20.022, -19.070,
    -26.511, -26.035, -25.376, -24.615, -23.790, -22.896, -21.963, -20.998,
    -20.026, -19.073, -26.517, -26.041, -25.381, -24.621, -23.795, -22.901,
    -21.967, -21.002, -20.030, -19.077, -26.523, -26.047, -25.387, -24.626,
    -23.800, -22.906, -21.972, -21.006, -20.034, -19.081, -26.529, -26.052,
    -25.392, -24.631, -23.805, -22.911, -21.976, -21.010, -20.037, -19.084,
    -26.535, -26.058, -25.398, -24.636, -23.810, -22.915, -21.981, -21.015,
    -20.041, -19.088, -26.541, -26.064, -25.403, -24.642, -23.815, -22.920,
    -21.985, -21.019, -20.045, -19.092, -26.547, -26.070, -25.409, -24.647,
    -23.820, -22.925, -21.989, -21.023, -20.049, -19.095, -26.553, -26.075,
    -25.415, -24.652, -23.825, -22.929, -21.994, -21.027, -20.053, -19.099,
    -26.558, -26.081, -25.420, -24.658, -23.830, -22.934, -21.998, -21.031,
    -20.057, -19.103, -26.564, -26.087, -25.426, -24.663, -23.835, -22.939,
    -22.003, -21.035, -20.061, -19.106, -26.570, -26.093, -25.431, -24.668,
    -23.840, -22.944, -22.007, -21.040, -20.065, -19.110, -26.576, -26.098,
    -25.437, -24.673, -23.845, -22.948, -22.012, -21.044, -20.069, -19.113,
    -26.582, -26.104, -25.442, -24.679, -23.850, -22.953, -22.016, -21.048,
    -20.072, -19.117, -26.588, -26.110, -25.448, -24.684, -23.855, -22.958,
    -22.021, -21.052, -20.076, -19.121, -26.594, -26.116, -25.453, -24.689,
    -23.860, -22.963, -22.025, -21.056, -20.080, -19.124, -26.600, -26.121,
    -25.459, -24.695, -23.865, -22.967, -22.030, -21.060, -20.084, -19.128,
};

const double CIA_Orton::H2He_n[][10] = {
    -30.428, -30.595, -30.740, -30.866, -30.978, -31.077, -31.168, -31.253,
    -31.334, -31.415, -28.595, -28.763, -28.908, -29.033, -29.146, -29.244,
    -29.335, -29.420, -29.502, -29.582, -27.209, -27.376, -27.521, -27.647,
    -27.759, -27.858, -27.949, -28.034, -28.115, -28.196, -25.823, -25.990,
    -26.135, -26.261, -26.373, -26.472, -26.563, -26.648, -26.729, -26.810,
    -25.012, -25.179, -25.324, -25.450, -25.562, -25.661, -25.752, -25.837,
    -25.918, -25.999, -24.436, -24.604, -24.749, -24.874, -24.987, -25.086,
    -25.177, -25.261, -25.343, -25.424, -23.990, -24.158, -24.302, -24.428,
    -24.540, -24.639, -24.730, -24.815, -24.897, -24.977, -23.625, -23.793,
    -23.938, -24.063, -24.176, -24.275, -24.366, -24.450, -24.532, -24.613,
    -23.317, -23.485, -23.629, -23.755, -23.868, -23.966, -24.057, -24.142,
    -24.224, -24.304, -23.050, -23.218, -23.362, -23.488, -23.600, -23.699,
    -23.790, -23.875, -23.957, -24.037, -22.815, -22.982, -23.127, -23.253,
    -23.365, -23.464, -23.555, -23.640, -23.721, -23.802, -22.604, -22.771,
    -22.916, -23.042, -23.154, -23.253, -23.344, -23.429, -23.510, -23.591,
    -22.413, -22.581, -22.726, -22.851, -22.964, -23.062, -23.153, -23.238,
    -23.320, -23.400, -22.240, -22.407, -22.552, -22.677, -22.790, -22.888,
    -22.979, -23.064, -23.146, -23.226, -22.080, -22.247, -22.392, -22.517,
    -22.630, -22.728, -22.819, -22.904, -22.986, -23.066, -21.931, -22.099,
    -22.243, -22.369, -22.481, -22.580, -22.671, -22.756, -22.837, -22.918,
    -21.794, -21.961, -22.106, -22.231, -22.343, -22.442, -22.533, -22.618,
    -22.699, -22.780, -21.665, -21.832, -21.976, -22.102, -22.214, -22.313,
    -22.404, -22.489, -22.570, -22.651, -21.543, -21.711, -21.855, -21.981,
    -22.093, -22.192, -22.283, -22.368, -22.449, -22.530, -21.429, -21.596,
    -21.741, -21.867, -21.979, -22.078, -22.169, -22.253, -22.335, -22.415,
    -21.321, -21.488, -21.633, -21.759, -21.871, -21.970, -22.060, -22.145,
    -22.227, -22.307, -21.219, -21.386, -21.530, -21.656, -21.768, -21.867,
    -21.958, -22.043, -22.124, -22.205, -20.773, -20.940, -21.085, -21.210,
    -21.322, -21.421, -21.512, -21.596, -21.678, -21.759, -20.410, -20.576,
    -20.720, -20.846, -20.958, -21.056, -21.147, -21.232, -21.313, -21.394,
    -20.103, -20.269, -20.413, -20.538, -20.650, -20.748, -20.839, -20.924,
    -21.005, -21.086, -19.837, -20.003, -20.146, -20.271, -20.383, -20.481,
    -20.572, -20.657, -20.738, -20.819, -19.603, -19.768, -19.911, -20.036,
    -20.148, -20.246, -20.337, -20.421, -20.503, -20.583, -19.394, -19.559,
    -19.702, -19.826, -19.938, -20.036, -20.126, -20.211, -20.292, -20.373,
    -19.205, -19.369, -19.512, -19.636, -19.747, -19.846, -19.936, -20.020,
    -20.102, -20.182, -19.033, -19.197, -19.339, -19.463, -19.574, -19.672,
    -19.762, -19.847, -19.928, -20.008, -18.875, -19.038, -19.180, -19.303,
    -19.414, -19.512, -19.602, -19.687, -19.768, -19.848, -18.730, -18.892,
    -19.033, -19.156, -19.267, -19.364, -19.455, -19.539, -19.620, -19.700,
    -18.594, -18.755, -18.896, -19.019, -19.129, -19.227, -19.317, -19.401,
    -19.482, -19.562, -18.468, -18.628, -18.768, -18.891, -19.001, -19.098,
    -19.188, -19.272, -19.353, -19.433, -18.350, -18.509, -18.649, -18.771,
    -18.881, -18.978, -19.067, -19.151, -19.232, -19.312, -18.238, -18.397,
    -18.536, -18.657, -18.767, -18.864, -18.954, -19.037, -19.118, -19.198,
    -18.134, -18.291, -18.429, -18.550, -18.660, -18.756, -18.846, -18.930,
    -19.010, -19.090, -18.034, -18.191, -18.328, -18.449, -18.558, -18.655,
    -18.744, -18.827, -18.908, -18.988, -17.851, -18.005, -18.141, -18.261,
    -18.369, -18.465, -18.554, -18.638, -18.718, -18.798, -17.685, -17.837,
    -17.971, -18.090, -18.197, -18.293, -18.381, -18.464, -18.545, -18.624,
    -17.533, -17.683, -17.815, -17.933, -18.040, -18.134, -18.223, -18.305,
    -18.385, -18.465, -17.394, -17.541, -17.672, -17.788, -17.894, -17.988,
    -18.076, -18.158, -18.238, -18.317, -17.265, -17.409, -17.538, -17.653,
    -17.758, -17.852, -17.939, -18.021, -18.101, -18.180, -17.146, -17.287,
    -17.414, -17.528, -17.632, -17.725, -17.812, -17.893, -17.972, -18.051,
    -17.036, -17.174, -17.299, -17.411, -17.514, -17.606, -17.692, -17.773,
    -17.852, -17.931, -16.932, -17.067, -17.190, -17.300, -17.402, -17.494,
    -17.579, -17.660, -17.739, -17.817, -16.835, -16.967, -17.088, -17.197,
    -17.297, -17.388, -17.473, -17.554, -17.632, -17.710, -16.745, -16.873,
    -16.991, -17.099, -17.198, -17.288, -17.373, -17.453, -17.530, -17.608,
    -16.659, -16.784, -16.900, -17.006, -17.104, -17.193, -17.277, -17.356,
    -17.434, -17.511, -16.578, -16.701, -16.814, -16.918, -17.015, -17.103,
    -17.186, -17.265, -17.342, -17.419, -16.502, -16.621, -16.732, -16.834,
    -16.930, -17.017, -17.099, -17.178, -17.254, -17.331, -16.430, -16.546,
    -16.654, -16.754, -16.849, -16.935, -17.017, -17.094, -17.171, -17.247,
    -16.362, -16.474, -16.580, -16.678, -16.771, -16.856, -16.937, -17.015,
    -17.090, -17.166, -16.297, -16.406, -16.509, -16.606, -16.697, -16.781,
    -16.861, -16.938, -17.013, -17.089, -16.236, -16.341, -16.441, -16.536,
    -16.626, -16.709, -16.789, -16.865, -16.939, -17.015, -16.177, -16.279,
    -16.377, -16.469, -16.558, -16.640, -16.718, -16.794, -16.868, -16.943,
    -16.122, -16.220, -16.315, -16.405, -16.493, -16.573, -16.651, -16.726,
    -16.800, -16.874, -16.068, -16.163, -16.256, -16.344, -16.430, -16.509,
    -16.586, -16.660, -16.733, -16.808, -16.018, -16.109, -16.199, -16.285,
    -16.369, -16.448, -16.523, -16.597, -16.670, -16.743, -15.969, -16.057,
    -16.144, -16.228, -16.311, -16.388, -16.463, -16.536, -16.608, -16.681,
    -15.923, -16.007, -16.091, -16.174, -16.255, -16.331, -16.405, -16.477,
    -16.548, -16.621, -15.879, -15.959, -16.041, -16.121, -16.200, -16.275,
    -16.348, -16.419, -16.490, -16.563, -15.836, -15.913, -15.992, -16.070,
    -16.148, -16.221, -16.293, -16.364, -16.434, -16.506, -15.795, -15.869,
    -15.945, -16.021, -16.097, -16.169, -16.240, -16.310, -16.380, -16.451,
    -15.756, -15.827, -15.900, -15.974, -16.048, -16.119, -16.189, -16.258,
    -16.327, -16.398, -15.719, -15.786, -15.857, -15.928, -16.001, -16.070,
    -16.139, -16.207, -16.276, -16.346, -15.683, -15.747, -15.815, -15.884,
    -15.955, -16.023, -16.091, -16.158, -16.226, -16.296, -15.649, -15.709,
    -15.774, -15.841, -15.910, -15.977, -16.044, -16.110, -16.178, -16.247,
    -15.615, -15.672, -15.735, -15.800, -15.867, -15.933, -15.998, -16.064,
    -16.130, -16.199, -15.584, -15.637, -15.697, -15.760, -15.825, -15.889,
    -15.954, -16.019, -16.084, -16.153, -15.553, -15.603, -15.660, -15.721,
    -15.785, -15.847, -15.911, -15.975, -16.040, -16.107, -15.523, -15.571,
    -15.625, -15.683, -15.745, -15.806, -15.869, -15.932, -15.996, -16.063,
    -15.495, -15.539, -15.591, -15.647, -15.707, -15.767, -15.828, -15.890,
    -15.954, -16.020, -15.468, -15.508, -15.557, -15.611, -15.670, -15.728,
    -15.788, -15.849, -15.912, -15.978, -15.441, -15.479, -15.525, -15.577,
    -15.634, -15.690, -15.749, -15.810, -15.872, -15.937, -15.416, -15.451,
    -15.494, -15.544, -15.598, -15.654, -15.712, -15.771, -15.832, -15.897,
    -15.392, -15.423, -15.464, -15.511, -15.564, -15.618, -15.675, -15.733,
    -15.793, -15.858, -15.368, -15.396, -15.435, -15.480, -15.531, -15.583,
    -15.639, -15.696, -15.756, -15.819, -15.345, -15.371, -15.406, -15.449,
    -15.499, -15.550, -15.604, -15.660, -15.719, -15.782, -15.323, -15.346,
    -15.379, -15.419, -15.467, -15.516, -15.569, -15.625, -15.683, -15.745,
    -15.302, -15.322, -15.352, -15.391, -15.436, -15.484, -15.536, -15.590,
    -15.647, -15.709, -15.282, -15.298, -15.326, -15.363, -15.406, -15.453,
    -15.503, -15.557, -15.613, -15.674, -15.262, -15.276, -15.301, -15.335,
    -15.377, -15.422, -15.471, -15.524, -15.579, -15.639, -15.243, -15.254,
    -15.277, -15.309, -15.349, -15.392, -15.440, -15.491, -15.546, -15.605,
    -15.225, -15.233, -15.253, -15.283, -15.321, -15.363, -15.410, -15.460,
    -15.514, -15.572, -15.207, -15.212, -15.230, -15.258, -15.294, -15.335,
    -15.380, -15.429, -15.482, -15.540, -15.190, -15.192, -15.208, -15.233,
    -15.268, -15.307, -15.351, -15.399, -15.451, -15.508, -15.173, -15.173,
    -15.186, -15.209, -15.242, -15.279, -15.322, -15.369, -15.420, -15.477,
    -15.158, -15.154, -15.165, -15.186, -15.217, -15.253, -15.294, -15.340,
    -15.390, -15.446, -15.142, -15.136, -15.145, -15.164, -15.193, -15.227,
    -15.267, -15.312, -15.361, -15.416, -15.127, -15.119, -15.125, -15.142,
    -15.169, -15.201, -15.240, -15.284, -15.332, -15.386, -15.113, -15.102,
    -15.105, -15.120, -15.146, -15.177, -15.214, -15.257, -15.304, -15.358,
    -15.099, -15.086, -15.087, -15.099, -15.123, -15.152, -15.189, -15.230,
    -15.276, -15.329, -15.085, -15.070, -15.068, -15.079, -15.101, -15.129,
    -15.163, -15.204, -15.249, -15.301, -15.073, -15.054, -15.051, -15.059,
    -15.079, -15.105, -15.139, -15.178, -15.223, -15.274, -15.060, -15.039,
    -15.033, -15.040, -15.058, -15.083, -15.115, -15.153, -15.196, -15.247,
    -15.048, -15.025, -15.017, -15.021, -15.037, -15.060, -15.091, -15.128,
    -15.171, -15.220, -15.036, -15.011, -15.000, -15.002, -15.017, -15.039,
    -15.068, -15.104, -15.146, -15.194, -15.025, -14.997, -14.984, -14.985,
    -14.997, -15.017, -15.045, -15.080, -15.121, -15.169, -15.014, -14.984,
    -14.969, -14.967, -14.978, -14.996, -15.023, -15.057, -15.096, -15.143,
    -15.003, -14.971, -14.954, -14.950, -14.959, -14.976, -15.001, -15.034,
    -15.072, -15.119, -14.993, -14.958, -14.939, -14.933, -14.941, -14.956,
    -14.980, -15.011, -15.049, -15.094, -14.983, -14.946, -14.925, -14.917,
    -14.923, -14.936, -14.959, -14.989, -15.026, -15.070, -14.974, -14.935,
    -14.911, -14.901, -14.905, -14.917, -14.939, -14.967, -15.003, -15.047,
    -14.965, -14.923, -14.898, -14.886, -14.888, -14.898, -14.918, -14.946,
    -14.981, -15.023, -14.956, -14.912, -14.885, -14.871, -14.871, -14.880,
    -14.899, -14.925, -14.959, -15.001, -14.947, -14.902, -14.872, -14.856,
    -14.855, -14.862, -14.879, -14.904, -14.937, -14.978, -14.939, -14.891,
    -14.860, -14.842, -14.839, -14.844, -14.860, -14.884, -14.916, -14.956,
    -14.931, -14.881, -14.847, -14.828, -14.823, -14.827, -14.842, -14.864,
    -14.895, -14.934, -14.923, -14.871, -14.836, -14.814, -14.808, -14.810,
    -14.823, -14.845, -14.874, -14.913, -14.916, -14.862, -14.824, -14.801,
    -14.792, -14.794, -14.805, -14.826, -14.854, -14.892, -14.908, -14.853,
    -14.813, -14.788, -14.778, -14.777, -14.788, -14.807, -14.834, -14.871,
    -14.902, -14.844, -14.802, -14.775, -14.763, -14.761, -14.770, -14.788,
    -14.815, -14.850, -14.895, -14.835, -14.792, -14.763, -14.749, -14.746,
    -14.753, -14.770, -14.796, -14.830, -14.888, -14.827, -14.782, -14.751,
    -14.736, -14.730, -14.737, -14.752, -14.777, -14.810, -14.882, -14.819,
    -14.772, -14.739, -14.722, -14.715, -14.720, -14.735, -14.758, -14.791,
    -14.876, -14.811, -14.762, -14.728, -14.709, -14.701, -14.704, -14.717,
    -14.740, -14.771, -14.870, -14.804, -14.753, -14.717, -14.696, -14.686,
    -14.688, -14.700, -14.721, -14.752, -14.865, -14.797, -14.744, -14.706,
    -14.683, -14.672, -14.673, -14.683, -14.704, -14.734, -14.859, -14.789,
    -14.735, -14.695, -14.671, -14.658, -14.658, -14.667, -14.686, -14.715,
    -14.854, -14.783, -14.726, -14.685, -14.659, -14.645, -14.643, -14.651,
    -14.669, -14.697, -14.849, -14.776, -14.718, -14.674, -14.647, -14.631,
    -14.628, -14.635, -14.652, -14.679, -14.845, -14.770, -14.710, -14.665,
    -14.636, -14.618, -14.614, -14.619, -14.635, -14.661, -14.840, -14.763,
    -14.702, -14.655, -14.624, -14.606, -14.599, -14.604, -14.619, -14.644,
    -14.836, -14.757, -14.694, -14.646, -14.613, -14.593, -14.585, -14.589,
    -14.602, -14.627, -14.831, -14.752, -14.687, -14.636, -14.603, -14.581,
    -14.572, -14.574, -14.586, -14.610, -14.827, -14.746, -14.679, -14.627,
    -14.592, -14.569, -14.558, -14.559, -14.570, -14.593, -14.823, -14.741,
    -14.672, -14.619, -14.582, -14.557, -14.545, -14.545, -14.555, -14.576,
    -14.820, -14.735, -14.665, -14.610, -14.572, -14.545, -14.532, -14.530,
    -14.540, -14.560, -14.816, -14.730, -14.659, -14.602, -14.562, -14.534,
    -14.519, -14.516, -14.525, -14.544, -14.813, -14.726, -14.652, -14.594,
    -14.552, -14.523, -14.507, -14.503, -14.510, -14.528, -14.809, -14.721,
    -14.646, -14.586, -14.542, -14.512, -14.494, -14.489, -14.495, -14.513,
    -14.806, -14.716, -14.640, -14.578, -14.533, -14.501, -14.482, -14.476,
    -14.481, -14.497, -14.803, -14.712, -14.634, -14.571, -14.524, -14.490,
    -14.470, -14.463, -14.466, -14.482, -14.801, -14.708, -14.628, -14.563,
    -14.515, -14.480, -14.459, -14.450, -14.452, -14.467, -14.798, -14.704,
    -14.623, -14.556, -14.507, -14.470, -14.447, -14.437, -14.438, -14.452,
    -14.795, -14.700, -14.618, -14.549, -14.498, -14.460, -14.436, -14.424,
    -14.425, -14.438, -14.793, -14.696, -14.612, -14.543, -14.490, -14.450,
    -14.425, -14.412, -14.411, -14.423, -14.791, -14.693, -14.607, -14.536,
    -14.482, -14.441, -14.414, -14.400, -14.398, -14.409, -14.788, -14.689,
    -14.602, -14.530, -14.474, -14.431, -14.403, -14.388, -14.385, -14.395,
    -14.786, -14.686, -14.598, -14.523, -14.466, -14.422, -14.393, -14.376,
    -14.372, -14.381, -14.784, -14.683, -14.593, -14.517, -14.459, -14.413,
    -14.382, -14.365, -14.359, -14.367, -14.782, -14.680, -14.589, -14.511,
    -14.451, -14.404, -14.372, -14.353, -14.347, -14.354, -14.781, -14.677,
    -14.585, -14.506, -14.444, -14.395, -14.362, -14.342, -14.335, -14.340,
    -14.779, -14.674, -14.580, -14.500, -14.437, -14.387, -14.352, -14.331,
    -14.322, -14.327, -14.778, -14.671, -14.576, -14.495, -14.430, -14.379,
    -14.343, -14.320, -14.310, -14.314, -14.776, -14.669, -14.573, -14.489,
    -14.423, -14.370, -14.333, -14.309, -14.299, -14.301, -14.775, -14.667,
    -14.569, -14.484, -14.417, -14.362, -14.324, -14.299, -14.287, -14.289,
    -14.774, -14.664, -14.565, -14.479, -14.410, -14.355, -14.315, -14.288,
    -14.275, -14.276, -14.773, -14.662, -14.562, -14.474, -14.404, -14.347,
    -14.306, -14.278, -14.264, -14.264, -14.772, -14.660, -14.559, -14.470,
    -14.398, -14.339, -14.297, -14.268, -14.253, -14.252, -14.771, -14.658,
    -14.555, -14.465, -14.392, -14.332, -14.288, -14.258, -14.242, -14.240,
    -14.770, -14.656, -14.552, -14.461, -14.386, -14.325, -14.279, -14.248,
    -14.231, -14.228, -14.769, -14.655, -14.549, -14.456, -14.380, -14.317,
    -14.271, -14.239, -14.220, -14.216, -14.768, -14.653, -14.546, -14.452,
    -14.374, -14.311, -14.263, -14.229, -14.210, -14.204, -14.768, -14.651,
    -14.544, -14.448, -14.369, -14.304, -14.255, -14.220, -14.199, -14.193,
    -14.767, -14.650, -14.541, -14.444, -14.364, -14.297, -14.247, -14.211,
    -14.189, -14.181, -14.767, -14.649, -14.539, -14.440, -14.358, -14.290,
    -14.239, -14.201, -14.179, -14.170, -14.767, -14.647, -14.536, -14.437,
    -14.353, -14.284, -14.231, -14.193, -14.169, -14.159, -14.766, -14.646,
    -14.534, -14.433, -14.348, -14.278, -14.223, -14.184, -14.159, -14.148,
    -14.766, -14.645, -14.532, -14.429, -14.343, -14.271, -14.216, -14.175,
    -14.149, -14.138, -14.766, -14.644, -14.530, -14.426, -14.339, -14.265,
    -14.209, -14.167, -14.139, -14.127, -14.766, -14.643, -14.528, -14.423,
    -14.334, -14.259, -14.201, -14.158, -14.130, -14.116, -14.766, -14.643,
    -14.526, -14.420, -14.330, -14.254, -14.194, -14.150, -14.120, -14.106,
    -14.766, -14.642, -14.524, -14.417, -14.325, -14.248, -14.187, -14.142,
    -14.111, -14.096, -14.766, -14.641, -14.522, -14.414, -14.321, -14.242,
    -14.180, -14.134, -14.102, -14.086, -14.767, -14.641, -14.521, -14.411,
    -14.317, -14.237, -14.174, -14.126, -14.093, -14.076, -14.767, -14.640,
    -14.519, -14.408, -14.313, -14.232, -14.167, -14.118, -14.084, -14.066,
    -14.767, -14.640, -14.518, -14.406, -14.309, -14.226, -14.161, -14.110,
    -14.075, -14.056, -14.768, -14.639, -14.516, -14.403, -14.305, -14.221,
    -14.154, -14.103, -14.067, -14.046, -14.768, -14.639, -14.515, -14.401,
    -14.301, -14.216, -14.148, -14.095, -14.058, -14.037, -14.769, -14.639,
    -14.514, -14.398, -14.298, -14.211, -14.142, -14.088, -14.050, -14.027,
    -14.770, -14.639, -14.513, -14.396, -14.294, -14.206, -14.136, -14.081,
    -14.041, -14.018, -14.770, -14.639, -14.512, -14.394, -14.291, -14.202,
    -14.130, -14.074, -14.033, -14.009, -14.771, -14.639, -14.511, -14.392,
    -14.288, -14.197, -14.124, -14.066, -14.025, -13.999, -14.772, -14.639,
    -14.510, -14.390, -14.284, -14.193, -14.118, -14.060, -14.017, -13.990,
    -14.773, -14.639, -14.509, -14.388, -14.281, -14.188, -14.113, -14.053,
    -14.009, -13.982, -14.774, -14.640, -14.509, -14.386, -14.278, -14.184,
    -14.107, -14.046, -14.001, -13.973, -14.775, -14.640, -14.508, -14.384,
    -14.275, -14.180, -14.102, -14.039, -13.994, -13.964, -14.776, -14.640,
    -14.508, -14.383, -14.272, -14.176, -14.096, -14.033, -13.986, -13.955,
    -14.777, -14.641, -14.507, -14.381, -14.269, -14.172, -14.091, -14.027,
    -13.978, -13.947, -14.778, -14.641, -14.507, -14.380, -14.267, -14.168,
    -14.086, -14.020, -13.971, -13.939, -14.779, -14.642, -14.506, -14.378,
    -14.264, -14.164, -14.081, -14.014, -13.964, -13.930, -14.780, -14.642,
    -14.506, -14.377, -14.262, -14.160, -14.076, -14.008, -13.957, -13.922,
    -14.781, -14.643, -14.506, -14.376, -14.259, -14.156, -14.071, -14.002,
    -13.949, -13.914, -14.783, -14.644, -14.506, -14.374, -14.257, -14.153,
    -14.066, -13.996, -13.942, -13.906, -14.784, -14.644, -14.506, -14.373,
    -14.255, -14.149, -14.061, -13.990, -13.936, -13.898, -14.786, -14.645,
    -14.506, -14.372, -14.252, -14.146, -14.057, -13.984, -13.929, -13.890,
    -14.787, -14.646, -14.506, -14.371, -14.250, -14.142, -14.052, -13.979,
    -13.922, -13.883, -14.788, -14.647, -14.506, -14.370, -14.248, -14.139,
    -14.048, -13.973, -13.915, -13.875, -14.790, -14.648, -14.506, -14.370,
    -14.246, -14.136, -14.043, -13.968, -13.909, -13.867, -14.792, -14.649,
    -14.506, -14.369, -14.244, -14.133, -14.039, -13.962, -13.902, -13.860,
    -14.793, -14.650, -14.507, -14.368, -14.242, -14.130, -14.035, -13.957,
    -13.896, -13.853, -14.795, -14.651, -14.507, -14.368, -14.241, -14.127,
    -14.031, -13.952, -13.890, -13.845, -14.797, -14.653, -14.508, -14.367,
    -14.239, -14.124, -14.027, -13.946, -13.883, -13.838, -14.798, -14.654,
    -14.508, -14.366, -14.237, -14.121, -14.023, -13.941, -13.877, -13.831,
    -14.800, -14.655, -14.509, -14.366, -14.236, -14.118, -14.019, -13.936,
    -13.871, -13.824, -14.802, -14.656, -14.509, -14.366, -14.234, -14.116,
    -14.015, -13.931, -13.865, -13.817, -14.804, -14.658, -14.510, -14.365,
    -14.233, -14.113, -14.011, -13.927, -13.859, -13.810, -14.806, -14.659,
    -14.510, -14.365, -14.231, -14.111, -14.008, -13.922, -13.854, -13.803,
    -14.808, -14.661, -14.511, -14.365, -14.230, -14.108, -14.004, -13.917,
    -13.848, -13.797, -14.810, -14.662, -14.512, -14.365, -14.229, -14.106,
    -14.000, -13.912, -13.842, -13.790, -14.812, -14.664, -14.513, -14.365,
    -14.228, -14.103, -13.997, -13.908, -13.837, -13.783, -14.814, -14.665,
    -14.514, -14.365, -14.227, -14.101, -13.994, -13.903, -13.831, -13.777,
    -14.816, -14.667, -14.515, -14.365, -14.226, -14.099, -13.990, -13.899,
    -13.826, -13.770, -14.818, -14.669, -14.516, -14.365, -14.225, -14.097,
    -13.987, -13.895, -13.820, -13.764, -14.820, -14.670, -14.517, -14.365,
    -14.224, -14.095, -13.984, -13.890, -13.815, -13.758, -14.822, -14.672,
    -14.518, -14.365, -14.223, -14.093, -13.981, -13.886, -13.810, -13.752,
    -14.825, -14.674, -14.519, -14.365, -14.222, -14.091, -13.977, -13.882,
    -13.804, -13.745, -14.827, -14.676, -14.520, -14.365, -14.221, -14.089,
    -13.974, -13.878, -13.799, -13.739, -14.829, -14.677, -14.521, -14.366,
    -14.220, -14.087, -13.972, -13.874, -13.794, -13.733, -14.831, -14.679,
    -14.522, -14.366, -14.220, -14.085, -13.969, -13.870, -13.789, -13.728,
    -14.834, -14.681, -14.524, -14.366, -14.219, -14.083, -13.966, -13.866,
    -13.784, -13.722, -14.836, -14.683, -14.525, -14.367, -14.218, -14.082,
    -13.963, -13.862, -13.780, -13.716, -14.839, -14.685, -14.526, -14.367,
    -14.218, -14.080, -13.960, -13.858, -13.775, -13.710, -14.841, -14.687,
    -14.528, -14.368, -14.217, -14.079, -13.958, -13.855, -13.770, -13.705,
    -14.843, -14.689, -14.529, -14.369, -14.217, -14.077, -13.955, -13.851,
    -13.766, -13.699, -14.846, -14.691, -14.531, -14.369, -14.217, -14.076,
    -13.952, -13.847, -13.761, -13.693, -14.848, -14.694, -14.532, -14.370,
    -14.216, -14.074, -13.950, -13.844, -13.756, -13.688, -14.851, -14.696,
    -14.534, -14.371, -14.216, -14.073, -13.948, -13.840, -13.752, -13.683,
    -14.854, -14.698, -14.535, -14.371, -14.216, -14.072, -13.945, -13.837,
    -13.748, -13.677, -14.856, -14.700, -14.537, -14.372, -14.216, -14.070,
    -13.943, -13.834, -13.743, -13.672, -14.859, -14.702, -14.539, -14.373,
    -14.215, -14.069, -13.941, -13.830, -13.739, -13.667, -14.861, -14.705,
    -14.540, -14.374, -14.215, -14.068, -13.938, -13.827, -13.735, -13.662,
    -14.864, -14.707, -14.542, -14.375, -14.215, -14.067, -13.936, -13.824,
    -13.731, -13.656, -14.867, -14.709, -14.544, -14.376, -14.215, -14.066,
    -13.934, -13.821, -13.726, -13.651, -14.870, -14.712, -14.546, -14.377,
    -14.215, -14.065, -13.932, -13.818, -13.722, -13.646, -14.872, -14.714,
    -14.547, -14.378, -14.215, -14.064, -13.930, -13.815, -13.718, -13.642,
    -14.875, -14.717, -14.549, -14.379, -14.215, -14.063, -13.928, -13.812,
    -13.714, -13.637, -14.878, -14.719, -14.551, -14.380, -14.215, -14.062,
    -13.926, -13.809, -13.711, -13.632, -14.881, -14.721, -14.553, -14.381,
    -14.216, -14.061, -13.924, -13.806, -13.707, -13.627, -14.884, -14.724,
    -14.555, -14.382, -14.216, -14.060, -13.922, -13.803, -13.703, -13.622,
    -14.886, -14.727, -14.557, -14.383, -14.216, -14.059, -13.920, -13.800,
    -13.699, -13.618, -14.889, -14.729, -14.559, -14.384, -14.216, -14.059,
    -13.919, -13.797, -13.696, -13.613, -14.892, -14.732, -14.561, -14.386,
    -14.217, -14.058, -13.917, -13.795, -13.692, -13.609, -14.895, -14.734,
    -14.563, -14.387, -14.217, -14.057, -13.915, -13.792, -13.688, -13.604,
    -14.898, -14.737, -14.565, -14.388, -14.217, -14.056, -13.913, -13.789,
    -13.685, -13.600, -14.901, -14.740, -14.567, -14.390, -14.218, -14.056,
    -13.912, -13.787, -13.681, -13.595, -14.904, -14.742, -14.569, -14.391,
    -14.218, -14.055, -13.910, -13.784, -13.678, -13.591, -14.907, -14.745,
    -14.571, -14.392, -14.219, -14.055, -13.909, -13.782, -13.674, -13.587,
    -14.910, -14.748, -14.574, -14.394, -14.219, -14.054, -13.907, -13.779,
    -13.671, -13.583, -14.913, -14.750, -14.576, -14.395, -14.220, -14.054,
    -13.906, -13.777, -13.668, -13.578, -14.916, -14.753, -14.578, -14.397,
    -14.220, -14.053, -13.904, -13.775, -13.665, -13.574, -14.919, -14.756,
    -14.580, -14.398, -14.221, -14.053, -13.903, -13.772, -13.661, -13.570,
    -14.922, -14.759, -14.582, -14.400, -14.221, -14.053, -13.902, -13.770,
    -13.658, -13.566, -14.926, -14.761, -14.585, -14.401, -14.222, -14.052,
    -13.900, -13.768, -13.655, -13.562, -14.929, -14.764, -14.587, -14.403,
    -14.223, -14.052, -13.899, -13.766, -13.652, -13.558, -14.932, -14.767,
    -14.589, -14.404, -14.223, -14.052, -13.898, -13.763, -13.649, -13.554,
    -14.935, -14.770, -14.592, -14.406, -14.224, -14.051, -13.897, -13.761,
    -13.646, -13.550, -14.938, -14.773, -14.594, -14.408, -14.225, -14.051,
    -13.896, -13.759, -13.643, -13.547, -14.942, -14.776, -14.596, -14.409,
    -14.226, -14.051, -13.894, -13.757, -13.640, -13.543, -14.945, -14.779,
    -14.599, -14.411, -14.226, -14.051, -13.893, -13.755, -13.637, -13.539,
    -14.948, -14.782, -14.601, -14.413, -14.227, -14.051, -13.892, -13.753,
    -13.634, -13.535, -14.951, -14.785, -14.604, -14.414, -14.228, -14.050,
    -13.891, -13.751, -13.632, -13.532, -14.955, -14.788, -14.606, -14.416,
    -14.229, -14.050, -13.890, -13.749, -13.629, -13.528, -14.958, -14.791,
    -14.608, -14.418, -14.230, -14.050, -13.889, -13.748, -13.626, -13.525,
    -14.961, -14.794, -14.611, -14.419, -14.230, -14.050, -13.888, -13.746,
    -13.623, -13.521, -14.965, -14.797, -14.613, -14.421, -14.231, -14.050,
    -13.887, -13.744, -13.621, -13.518, -14.968, -14.800, -14.616, -14.423,
    -14.232, -14.050, -13.886, -13.742, -13.618, -13.514, -14.971, -14.803,
    -14.618, -14.425, -14.233, -14.050, -13.885, -13.740, -13.616, -13.511,
    -14.975, -14.806, -14.621, -14.426, -14.234, -14.050, -13.885, -13.739,
    -13.613, -13.508, -14.978, -14.809, -14.623, -14.428, -14.235, -14.050,
    -13.884, -13.737, -13.611, -13.504, -14.981, -14.812, -14.626, -14.430,
    -14.236, -14.050, -13.883, -13.735, -13.608, -13.501, -14.985, -14.815,
    -14.628, -14.432, -14.237, -14.050, -13.882, -13.734, -13.606, -13.498,
    -14.988, -14.818, -14.631, -14.434, -14.238, -14.050, -13.881, -13.732,
    -13.603, -13.495, -14.992, -14.821, -14.634, -14.435, -14.239, -14.050,
    -13.881, -13.731, -13.601, -13.492, -14.995, -14.824, -14.636, -14.437,
    -14.240, -14.051, -13.880, -13.729, -13.599, -13.488, -14.998, -14.827,
    -14.639, -14.439, -14.241, -14.051, -13.879, -13.728, -13.596, -13.485,
    -15.002, -14.830, -14.641, -14.441, -14.242, -14.051, -13.878, -13.726,
    -13.594, -13.482, -15.005, -14.833, -14.644, -14.443, -14.243, -14.051,
    -13.878, -13.725, -13.592, -13.479, -15.009, -14.836, -14.646, -14.444,
    -14.244, -14.051, -13.877, -13.723, -13.590, -13.476, -15.012, -14.840,
    -14.649, -14.446, -14.245, -14.051, -13.876, -13.722, -13.588, -13.473,
    -15.016, -14.843, -14.651, -14.448, -14.246, -14.051, -13.876, -13.721,
    -13.585, -13.471, -15.019, -14.846, -14.654, -14.450, -14.247, -14.052,
    -13.875, -13.719, -13.583, -13.468, -15.023, -14.849, -14.656, -14.452,
    -14.248, -14.052, -13.875, -13.718, -13.581, -13.465, -15.026, -14.852,
    -14.659, -14.453, -14.248, -14.052, -13.874, -13.717, -13.579, -13.462,
    -15.030, -14.855, -14.661, -14.455, -14.249, -14.052, -13.874, -13.715,
    -13.577, -13.459, -15.033, -14.858, -14.664, -14.457, -14.250, -14.052,
    -13.873, -13.714, -13.575, -13.457, -15.037, -14.861, -14.666, -14.459,
    -14.251, -14.053, -13.873, -13.713, -13.573, -13.454, -15.040, -14.864,
    -14.669, -14.461, -14.252, -14.053, -13.872, -13.712, -13.571, -13.451,
    -15.044, -14.867, -14.671, -14.462, -14.253, -14.053, -13.872, -13.710,
    -13.570, -13.449, -15.047, -14.870, -14.674, -14.464, -14.254, -14.053,
    -13.871, -13.709, -13.568, -13.446, -15.051, -14.873, -14.676, -14.466,
    -14.255, -14.053, -13.871, -13.708, -13.566, -13.443, -15.054, -14.877,
    -14.679, -14.467, -14.256, -14.054, -13.870, -13.707, -13.564, -13.441,
    -15.058, -14.880, -14.681, -14.469, -14.257, -14.054, -13.870, -13.706,
    -13.562, -13.438, -15.061, -14.883, -14.683, -14.471, -14.258, -14.054,
    -13.869, -13.705, -13.561, -13.436, -15.064, -14.886, -14.686, -14.472,
    -14.259, -14.054, -13.869, -13.704, -13.559, -13.434, -15.068, -14.888,
    -14.688, -14.474, -14.260, -14.054, -13.868, -13.703, -13.557, -13.431,
    -15.071, -14.891, -14.690, -14.475, -14.261, -14.055, -13.868, -13.702,
    -13.556, -13.429, -15.075, -14.894, -14.693, -14.477, -14.261, -14.055,
    -13.868, -13.701, -13.554, -13.426, -15.078, -14.897, -14.695, -14.479,
    -14.262, -14.055, -13.867, -13.700, -13.552, -13.424, -15.081, -14.900,
    -14.697, -14.480, -14.263, -14.055, -13.867, -13.699, -13.551, -13.422,
    -15.085, -14.903, -14.699, -14.481, -14.264, -14.055, -13.867, -13.698,
    -13.549, -13.420, -15.088, -14.906, -14.701, -14.483, -14.265, -14.056,
    -13.866, -13.697, -13.548, -13.417, -15.091, -14.908, -14.703, -14.484,
    -14.265, -14.056, -13.866, -13.696, -13.546, -13.415, -15.095, -14.911,
    -14.705, -14.486, -14.266, -14.056, -13.865, -13.695, -13.545, -13.413,
    -15.098, -14.914, -14.707, -14.487, -14.267, -14.056, -13.865, -13.694,
    -13.543, -13.411, -15.101, -14.916, -14.709, -14.488, -14.267, -14.056,
    -13.865, -13.693, -13.542, -13.409, -15.104, -14.919, -14.711, -14.489,
    -14.268, -14.056, -13.864, -13.693, -13.540, -13.407, -15.108, -14.922,
    -14.713, -14.490, -14.268, -14.056, -13.864, -13.692, -13.539, -13.405,
    -15.111, -14.924, -14.715, -14.492, -14.269, -14.056, -13.864, -13.691,
    -13.537, -13.403, -15.114, -14.926, -14.716, -14.493, -14.269, -14.056,
    -13.863, -13.690, -13.536, -13.401, -15.117, -14.929, -14.718, -14.494,
    -14.270, -14.056, -13.863, -13.689, -13.535, -13.399, -15.120, -14.931,
    -14.720, -14.494, -14.270, -14.057, -13.863, -13.689, -13.533, -13.397,
    -15.123, -14.933, -14.721, -14.495, -14.271, -14.057, -13.862, -13.688,
    -13.532, -13.395, -15.125, -14.935, -14.723, -14.496, -14.271, -14.056,
    -13.862, -13.687, -13.531, -13.393, -15.128, -14.937, -14.724, -14.497,
    -14.271, -14.056, -13.862, -13.686, -13.530, -13.391, -15.131, -14.939,
    -14.725, -14.498, -14.272, -14.056, -13.861, -13.686, -13.528, -13.389,
    -15.133, -14.941, -14.726, -14.498, -14.272, -14.056, -13.861, -13.685,
    -13.527, -13.387, -15.136, -14.943, -14.727, -14.499, -14.272, -14.056,
    -13.861, -13.684, -13.526, -13.386, -15.138, -14.944, -14.728, -14.499,
    -14.272, -14.056, -13.860, -13.684, -13.525, -13.384, -15.141, -14.946,
    -14.729, -14.499, -14.272, -14.056, -13.860, -13.683, -13.524, -13.382,
    -15.143, -14.947, -14.730, -14.500, -14.272, -14.056, -13.860, -13.682,
    -13.522, -13.380, -15.145, -14.949, -14.730, -14.500, -14.272, -14.056,
    -13.859, -13.682, -13.521, -13.379, -15.147, -14.950, -14.731, -14.500,
    -14.272, -14.055, -13.859, -13.681, -13.520, -13.377, -15.149, -14.951,
    -14.731, -14.500, -14.272, -14.055, -13.858, -13.680, -13.519, -13.375,
    -15.151, -14.952, -14.731, -14.500, -14.271, -14.055, -13.858, -13.680,
    -13.518, -13.374, -15.152, -14.952, -14.732, -14.499, -14.271, -14.055,
    -13.858, -13.679, -13.517, -13.372, -15.154, -14.953, -14.732, -14.499,
    -14.271, -14.054, -13.857, -13.679, -13.516, -13.370, -15.155, -14.953,
    -14.731, -14.499, -14.270, -14.054, -13.857, -13.678, -13.515, -13.369,
    -15.156, -14.954, -14.731, -14.498, -14.270, -14.053, -13.857, -13.677,
    -13.514, -13.367, -15.157, -14.954, -14.731, -14.498, -14.269, -14.053,
    -13.856, -13.677, -13.513, -13.366, -15.158, -14.954, -14.730, -14.497,
    -14.269, -14.053, -13.856, -13.676, -13.512, -13.364, -15.158, -14.953,
    -14.729, -14.496, -14.268, -14.052, -13.856, -13.676, -13.511, -13.363,
    -15.158, -14.953, -14.728, -14.495, -14.267, -14.052, -13.855, -13.675,
    -13.510, -13.361, -15.158, -14.952, -14.727, -14.494, -14.266, -14.051,
    -13.855, -13.675, -13.509, -13.360, -15.158, -14.951, -14.726, -14.493,
    -14.265, -14.050, -13.854, -13.674, -13.509, -13.358, -15.158, -14.949,
    -14.724, -14.491, -14.264, -14.050, -13.854, -13.674, -13.508, -13.357,
    -15.157, -14.948, -14.722, -14.490, -14.263, -14.049, -13.854, -13.673,
    -13.507, -13.356, -15.156, -14.946, -14.720, -14.488, -14.262, -14.049,
    -13.853, -13.673, -13.506, -13.354, -15.154, -14.944, -14.718, -14.486,
    -14.261, -14.048, -13.853, -13.672, -13.505, -13.353, -15.152, -14.942,
    -14.716, -14.484, -14.260, -14.047, -13.852, -13.672, -13.504, -13.352,
    -15.150, -14.939, -14.713, -14.482, -14.258, -14.046, -13.852, -13.671,
    -13.504, -13.350, -15.148, -14.936, -14.711, -14.480, -14.257, -14.046,
    -13.851, -13.671, -13.503, -13.349, -15.145, -14.933, -14.708, -14.478,
    -14.255, -14.045, -13.851, -13.670, -13.502, -13.348, -15.142, -14.929,
    -14.704, -14.476, -14.254, -14.044, -13.851, -13.670, -13.501, -13.347,
    -15.138, -14.925, -14.701, -14.473, -14.252, -14.043, -13.850, -13.670,
    -13.501, -13.345, -15.134, -14.921, -14.697, -14.470, -14.251, -14.042,
    -13.850, -13.669, -13.500, -13.344, -15.129, -14.916, -14.693, -14.467,
    -14.249, -14.041, -13.849, -13.669, -13.499, -13.343, -15.124, -14.911,
    -14.689, -14.464, -14.247, -14.040, -13.849, -13.668, -13.499, -13.342,
    -15.119, -14.906, -14.685, -14.461, -14.245, -14.040, -13.848, -13.668,
    -13.498, -13.341, -15.112, -14.900, -14.680, -14.458, -14.243, -14.039,
    -13.848, -13.668, -13.497, -13.340, -15.106, -14.894, -14.675, -14.455,
    -14.241, -14.038, -13.847, -13.667, -13.497, -13.338, -15.098, -14.887,
    -14.670, -14.451, -14.239, -14.037, -13.847, -13.667, -13.496, -13.337,
    -15.091, -14.881, -14.665, -14.448, -14.237, -14.035, -13.847, -13.667,
    -13.495, -13.336, -15.082, -14.873, -14.659, -14.444, -14.235, -14.034,
    -13.846, -13.666, -13.495, -13.335, -15.073, -14.866, -14.653, -14.440,
    -14.232, -14.033, -13.846, -13.666, -13.494, -13.334, -15.064, -14.857,
    -14.647, -14.436, -14.230, -14.032, -13.845, -13.666, -13.494, -13.333,
    -15.053, -14.849, -14.641, -14.432, -14.228, -14.031, -13.845, -13.665,
    -13.493, -13.332, -15.042, -14.840, -14.635, -14.428, -14.225, -14.030,
    -13.844, -13.665, -13.493, -13.331, -15.031, -14.831, -14.628, -14.423,
    -14.223, -14.029, -13.844, -13.665, -13.492, -13.330, -15.019, -14.821,
    -14.621, -14.419, -14.220, -14.028, -13.844, -13.665, -13.492, -13.329,
    -15.006, -14.811, -14.614, -14.414, -14.218, -14.027, -13.843, -13.664,
    -13.491, -13.328, -14.993, -14.801, -14.607, -14.410, -14.215, -14.026,
    -13.843, -13.664, -13.491, -13.328, -14.979, -14.790, -14.599, -14.405,
    -14.213, -14.024, -13.843, -13.664, -13.490, -13.327, -14.964, -14.779,
    -14.592, -14.401, -14.210, -14.023, -13.842, -13.664, -13.490, -13.326,
    -14.949, -14.768, -14.584, -14.396, -14.208, -14.022, -13.842, -13.664,
    -13.490, -13.325, -14.933, -14.756, -14.576, -14.391, -14.205, -14.021,
    -13.841, -13.663, -13.489, -13.324, -14.917, -14.745, -14.568, -14.386,
    -14.203, -14.020, -13.841, -13.663, -13.489, -13.323, -14.900, -14.733,
    -14.560, -14.381, -14.200, -14.019, -13.841, -13.663, -13.488, -13.322,
    -14.883, -14.720, -14.552, -14.376, -14.197, -14.018, -13.840, -13.663,
    -13.488, -13.322, -14.866, -14.708, -14.544, -14.371, -14.195, -14.017,
    -13.840, -13.663, -13.488, -13.321, -14.848, -14.695, -14.536, -14.367,
    -14.192, -14.016, -13.840, -13.663, -13.487, -13.320, -14.829, -14.683,
    -14.528, -14.362, -14.190, -14.015, -13.840, -13.663, -13.487, -13.319,
    -14.811, -14.670, -14.519, -14.357, -14.187, -14.014, -13.839, -13.663,
    -13.487, -13.319, -14.792, -14.657, -14.511, -14.352, -14.185, -14.013,
    -13.839, -13.663, -13.487, -13.318, -14.774, -14.644, -14.503, -14.347,
    -14.182, -14.012, -13.839, -13.663, -13.486, -13.317, -14.755, -14.632,
    -14.495, -14.343, -14.180, -14.011, -13.839, -13.663, -13.486, -13.317,
    -14.736, -14.619, -14.487, -14.338, -14.178, -14.010, -13.839, -13.663,
    -13.486, -13.316, -14.717, -14.606, -14.479, -14.333, -14.175, -14.009,
    -13.839, -13.663, -13.486, -13.315, -14.699, -14.594, -14.471, -14.329,
    -14.173, -14.008, -13.839, -13.663, -13.485, -13.315, -14.681, -14.582,
    -14.464, -14.325, -14.171, -14.008, -13.838, -13.663, -13.485, -13.314,
    -14.663, -14.570, -14.456, -14.320, -14.169, -14.007, -13.838, -13.663,
    -13.485, -13.313, -14.645, -14.559, -14.449, -14.316, -14.167, -14.006,
    -13.838, -13.663, -13.485, -13.313, -14.628, -14.547, -14.442, -14.312,
    -14.165, -14.006, -13.838, -13.663, -13.485, -13.312, -14.612, -14.537,
    -14.436, -14.309, -14.164, -14.005, -13.838, -13.663, -13.485, -13.312,
    -14.596, -14.526, -14.429, -14.305, -14.162, -14.005, -13.839, -13.663,
    -13.484, -13.311, -14.581, -14.516, -14.423, -14.302, -14.160, -14.004,
    -13.839, -13.663, -13.484, -13.311, -14.567, -14.507, -14.417, -14.298,
    -14.159, -14.004, -13.839, -13.664, -13.484, -13.310, -14.553, -14.498,
    -14.412, -14.295, -14.157, -14.004, -13.839, -13.664, -13.484, -13.309,
    -14.541, -14.490, -14.406, -14.293, -14.156, -14.003, -13.839, -13.664,
    -13.484, -13.309, -14.529, -14.482, -14.402, -14.290, -14.155, -14.003,
    -13.839, -13.664, -13.484, -13.308, -14.518, -14.475, -14.397, -14.288,
    -14.154, -14.003, -13.840, -13.665, -13.484, -13.308, -14.508, -14.468,
    -14.393, -14.285, -14.153, -14.003, -13.840, -13.665, -13.484, -13.308,
    -14.500, -14.462, -14.390, -14.283, -14.153, -14.003, -13.840, -13.665,
    -13.484, -13.307, -14.492, -14.457, -14.386, -14.282, -14.152, -14.003,
    -13.841, -13.665, -13.484, -13.307, -14.485, -14.452, -14.383, -14.280,
    -14.151, -14.003, -13.841, -13.666, -13.484, -13.306, -14.479, -14.448,
    -14.381, -14.279, -14.151, -14.004, -13.842, -13.666, -13.484, -13.306,
    -14.475, -14.445, -14.379, -14.278, -14.151, -14.004, -13.842, -13.667,
    -13.484, -13.305, -14.471, -14.442, -14.377, -14.277, -14.151, -14.004,
    -13.843, -13.667, -13.484, -13.305, -14.469, -14.440, -14.376, -14.277,
    -14.151, -14.005, -13.843, -13.667, -13.484, -13.305, -14.467, -14.439,
    -14.375, -14.277, -14.151, -14.005, -13.844, -13.668, -13.484, -13.304,
    -14.466, -14.439, -14.375, -14.277, -14.152, -14.006, -13.845, -13.668,
    -13.484, -13.304, -14.467, -14.439, -14.375, -14.277, -14.152, -14.007,
    -13.845, -13.669, -13.485, -13.304, -14.468, -14.439, -14.376, -14.278,
    -14.153, -14.008, -13.846, -13.669, -13.485, -13.303, -14.470, -14.440,
    -14.377, -14.279, -14.154, -14.009, -13.847, -13.670, -13.485, -13.303,
    -14.473, -14.442, -14.378, -14.280, -14.155, -14.010, -13.848, -13.670,
    -13.485, -13.303, -14.477, -14.445, -14.379, -14.281, -14.156, -14.011,
    -13.849, -13.671, -13.485, -13.302, -14.482, -14.448, -14.382, -14.282,
    -14.157, -14.012, -13.849, -13.671, -13.485, -13.302, -14.487, -14.451,
    -14.384, -14.284, -14.159, -14.013, -13.850, -13.672, -13.486, -13.302,
    -14.493, -14.455, -14.387, -14.286, -14.161, -14.014, -13.851, -13.673,
    -13.486, -13.302, -14.500, -14.460, -14.390, -14.289, -14.162, -14.016,
    -13.852, -13.673, -13.486, -13.301, -14.507, -14.465, -14.394, -14.291,
    -14.164, -14.017, -13.854, -13.674, -13.486, -13.301, -14.515, -14.471,
    -14.397, -14.294, -14.166, -14.019, -13.855, -13.675, -13.487, -13.301,
    -14.523, -14.476, -14.402, -14.297, -14.168, -14.020, -13.856, -13.675,
    -13.487, -13.301, -14.532, -14.483, -14.406, -14.300, -14.171, -14.022,
    -13.857, -13.676, -13.487, -13.301, -14.542, -14.490, -14.411, -14.304,
    -14.173, -14.024, -13.858, -13.677, -13.487, -13.300, -14.551, -14.497,
    -14.416, -14.307, -14.176, -14.026, -13.860, -13.678, -13.488, -13.300,
    -14.562, -14.504, -14.421, -14.311, -14.178, -14.027, -13.861, -13.679,
    -13.488, -13.300, -14.572, -14.512, -14.427, -14.315, -14.181, -14.029,
    -13.862, -13.679, -13.488, -13.300, -14.583, -14.520, -14.433, -14.319,
    -14.184, -14.032, -13.864, -13.680, -13.489, -13.300, -14.594, -14.528,
    -14.439, -14.323, -14.187, -14.034, -13.865, -13.681, -13.489, -13.300,
    -14.605, -14.537, -14.445, -14.328, -14.191, -14.036, -13.866, -13.682,
    -13.489, -13.299, -14.617, -14.546, -14.452, -14.333, -14.194, -14.038,
    -13.868, -13.683, -13.490, -13.299, -14.629, -14.555, -14.458, -14.338,
    -14.197, -14.041, -13.869, -13.684, -13.490, -13.299, -14.641, -14.564,
    -14.465, -14.343, -14.201, -14.043, -13.871, -13.685, -13.490, -13.299,
    -14.653, -14.573, -14.472, -14.348, -14.204, -14.045, -13.873, -13.685,
    -13.491, -13.299, -14.665, -14.583, -14.480, -14.353, -14.208, -14.048,
    -13.874, -13.686, -13.491, -13.299, -14.678, -14.593, -14.487, -14.358,
    -14.212, -14.051, -13.876, -13.687, -13.491, -13.299, -14.690, -14.603,
    -14.494, -14.364, -14.216, -14.053, -13.878, -13.688, -13.492, -13.299,
    -14.703, -14.613, -14.502, -14.370, -14.220, -14.056, -13.879, -13.689,
    -13.492, -13.299, -14.715, -14.623, -14.510, -14.375, -14.224, -14.059,
    -13.881, -13.690, -13.493, -13.299, -14.728, -14.633, -14.518, -14.381,
    -14.228, -14.061, -13.883, -13.691, -13.493, -13.299, -14.741, -14.643,
    -14.526, -14.387, -14.232, -14.064, -13.885, -13.692, -13.494, -13.299,
    -14.754, -14.654, -14.534, -14.393, -14.237, -14.067, -13.887, -13.693,
    -13.494, -13.299, -14.766, -14.664, -14.542, -14.399, -14.241, -14.070,
    -13.888, -13.694, -13.494, -13.299, -14.779, -14.675, -14.551, -14.406,
    -14.246, -14.073, -13.890, -13.695, -13.495, -13.298, -14.792, -14.685,
    -14.559, -14.412, -14.250, -14.076, -13.892, -13.697, -13.495, -13.298,
    -14.805, -14.696, -14.567, -14.418, -14.255, -14.079, -13.894, -13.698,
    -13.496, -13.298, -14.818, -14.707, -14.576, -14.425, -14.259, -14.083,
    -13.896, -13.699, -13.496, -13.298, -14.830, -14.717, -14.584, -14.431,
    -14.264, -14.086, -13.898, -13.700, -13.497, -13.298, -14.843, -14.728,
    -14.593, -14.438, -14.269, -14.089, -13.900, -13.701, -13.497, -13.298,
    -14.855, -14.739, -14.602, -14.445, -14.274, -14.092, -13.902, -13.702,
    -13.498, -13.298, -14.868, -14.749, -14.610, -14.451, -14.278, -14.095,
    -13.904, -13.703, -13.498, -13.298, -14.881, -14.760, -14.619, -14.458,
    -14.283, -14.099, -13.906, -13.704, -13.499, -13.299, -14.893, -14.771,
    -14.628, -14.465, -14.288, -14.102, -13.908, -13.705, -13.499, -13.299,
    -14.905, -14.781, -14.636, -14.471, -14.293, -14.105, -13.910, -13.707,
    -13.500, -13.299, -14.918, -14.792, -14.645, -14.478, -14.298, -14.109,
    -13.912, -13.708, -13.500, -13.299, -14.930, -14.803, -14.654, -14.485,
    -14.303, -14.112, -13.914, -13.709, -13.501, -13.299, -14.942, -14.813,
    -14.663, -14.492, -14.308, -14.115, -13.916, -13.710, -13.501, -13.299,
    -14.954, -14.824, -14.671, -14.499, -14.313, -14.119, -13.919, -13.711,
    -13.502, -13.299, -14.966, -14.834, -14.680, -14.506, -14.318, -14.122,
    -13.921, -13.712, -13.502, -13.299, -14.978, -14.845, -14.689, -14.513,
    -14.323, -14.126, -13.923, -13.713, -13.503, -13.299, -14.990, -14.855,
    -14.698, -14.520, -14.329, -14.129, -13.925, -13.715, -13.503, -13.299,
    -15.002, -14.865, -14.706, -14.526, -14.334, -14.133, -13.927, -13.716,
    -13.504, -13.299, -15.014, -14.876, -14.715, -14.533, -14.339, -14.136,
    -13.929, -13.717, -13.504, -13.299, -15.025, -14.886, -14.724, -14.540,
    -14.344, -14.140, -13.931, -13.718, -13.505, -13.299, -15.037, -14.896,
    -14.732, -14.547, -14.349, -14.143, -13.933, -13.719, -13.505, -13.299,
    -15.049, -14.907, -14.741, -14.554, -14.354, -14.147, -13.935, -13.720,
    -13.506, -13.299, -15.060, -14.917, -14.750, -14.561, -14.359, -14.150,
    -13.938, -13.721, -13.506, -13.299, -15.071, -14.927, -14.758, -14.568,
    -14.364, -14.153, -13.940, -13.723, -13.507, -13.299, -15.082, -14.937,
    -14.767, -14.575, -14.369, -14.157, -13.942, -13.724, -13.507, -13.300,
    -15.094, -14.947, -14.775, -14.582, -14.374, -14.160, -13.944, -13.725,
    -13.508, -13.300, -15.105, -14.957, -14.784, -14.588, -14.380, -14.164,
    -13.946, -13.726, -13.508, -13.300, -15.116, -14.967, -14.792, -14.595,
    -14.385, -14.167, -13.948, -13.727, -13.509, -13.300, -15.127, -14.977,
    -14.801, -14.602, -14.390, -14.171, -13.950, -13.728, -13.509, -13.300,
    -15.137, -14.986, -14.809, -14.609, -14.395, -14.174, -13.952, -13.729,
    -13.510, -13.300, -15.148, -14.996, -14.817, -14.615, -14.400, -14.177,
    -13.954, -13.730, -13.510, -13.300, -15.159, -15.006, -14.826, -14.622,
    -14.405, -14.181, -13.956, -13.731, -13.511, -13.300, -15.169, -15.015,
    -14.834, -14.629, -14.410, -14.184, -13.958, -13.732, -13.511, -13.300,
    -15.180, -15.025, -14.842, -14.635, -14.415, -14.187, -13.960, -13.734,
    -13.512, -13.300, -15.190, -15.034, -14.850, -14.642, -14.419, -14.191,
    -13.962, -13.735, -13.512, -13.300, -15.201, -15.044, -14.858, -14.648,
    -14.424, -14.194, -13.964, -13.736, -13.513, -13.300, -15.211, -15.053,
    -14.866, -14.655, -14.429, -14.197, -13.966, -13.737, -13.513, -13.301,
    -15.221, -15.062, -14.874, -14.661, -14.434, -14.200, -13.968, -13.738,
    -13.513, -13.301, -15.231, -15.071, -14.882, -14.668, -14.439, -14.204,
    -13.970, -13.739, -13.514, -13.301, -15.241, -15.081, -14.890, -14.674,
    -14.443, -14.207, -13.972, -13.739, -13.514, -13.301, -15.251, -15.090,
    -14.898, -14.680, -14.448, -14.210, -13.974, -13.740, -13.515, -13.301,
    -15.261, -15.099, -14.906, -14.687, -14.453, -14.213, -13.975, -13.741,
    -13.515, -13.301, -15.271, -15.108, -14.913, -14.693, -14.457, -14.216,
    -13.977, -13.742, -13.515, -13.301, -15.281, -15.117, -14.921, -14.699,
    -14.462, -14.219, -13.979, -13.743, -13.516, -13.301, -15.291, -15.125,
    -14.929, -14.705, -14.466, -14.222, -13.981, -13.744, -13.516, -13.301,
    -15.300, -15.134, -14.936, -14.711, -14.471, -14.225, -13.982, -13.745,
    -13.516, -13.301, -15.310, -15.143, -14.944, -14.717, -14.475, -14.228,
    -13.984, -13.746, -13.517, -13.301, -15.319, -15.151, -14.951, -14.723,
    -14.479, -14.231, -13.985, -13.746, -13.517, -13.301, -15.329, -15.160,
    -14.958, -14.729, -14.484, -14.233, -13.987, -13.747, -13.517, -13.302,
    -15.338, -15.168, -14.966, -14.734, -14.488, -14.236, -13.989, -13.748,
    -13.518, -13.302, -15.347, -15.177, -14.973, -14.740, -14.492, -14.239,
    -13.990, -13.749, -13.518, -13.302, -15.356, -15.185, -14.980, -14.746,
    -14.496, -14.241, -13.992, -13.749, -13.518, -13.302, -15.365, -15.193,
    -14.987, -14.751, -14.500, -14.244, -13.993, -13.750, -13.519, -13.302,
    -15.374, -15.202, -14.994, -14.757, -14.504, -14.246, -13.994, -13.751,
    -13.519, -13.302, -15.383, -15.210, -15.001, -14.762, -14.508, -14.249,
    -13.996, -13.751, -13.519, -13.302, -15.392, -15.218, -15.007, -14.767,
    -14.511, -14.251, -13.997, -13.752, -13.519, -13.302, -15.401, -15.226,
    -15.014, -14.773, -14.515, -14.253, -13.998, -13.752, -13.520, -13.302,
    -15.410, -15.234, -15.021, -14.778, -14.519, -14.256, -13.999, -13.753,
    -13.520, -13.302, -15.419, -15.241, -15.027, -14.783, -14.522, -14.258,
    -14.000, -13.753, -13.520, -13.302, -15.427, -15.249, -15.034, -14.788,
    -14.526, -14.260, -14.001, -13.754, -13.520, -13.302, -15.436, -15.257,
    -15.040, -14.793, -14.529, -14.262, -14.003, -13.754, -13.520, -13.302,
    -15.444, -15.264, -15.046, -14.797, -14.532, -14.264, -14.003, -13.755,
    -13.520, -13.302, -15.453, -15.272, -15.053, -14.802, -14.535, -14.266,
    -14.004, -13.755, -13.520, -13.302, -15.461, -15.279, -15.059, -14.807,
    -14.539, -14.268, -14.005, -13.755, -13.521, -13.302, -15.469, -15.287,
    -15.065, -14.811, -14.542, -14.269, -14.006, -13.756, -13.521, -13.302,
    -15.477, -15.294, -15.070, -14.815, -14.544, -14.271, -14.007, -13.756,
    -13.521, -13.302, -15.486, -15.301, -15.076, -14.820, -14.547, -14.273,
    -14.008, -13.756, -13.521, -13.302, -15.494, -15.308, -15.082, -14.824,
    -14.550, -14.274, -14.008, -13.756, -13.521, -13.302, -15.502, -15.315,
    -15.087, -14.828, -14.552, -14.275, -14.009, -13.756, -13.521, -13.302,
    -15.509, -15.322, -15.093, -14.831, -14.555, -14.277, -14.009, -13.756,
    -13.521, -13.302, -15.517, -15.328, -15.098, -14.835, -14.557, -14.278,
    -14.010, -13.756, -13.521, -13.302, -15.525, -15.335, -15.103, -14.839,
    -14.559, -14.279, -14.010, -13.757, -13.521, -13.302, -15.533, -15.342,
    -15.108, -14.842, -14.561, -14.280, -14.010, -13.757, -13.521, -13.302,
    -15.540, -15.348, -15.113, -14.846, -14.563, -14.281, -14.011, -13.756,
    -13.520, -13.302, -15.548, -15.354, -15.118, -14.849, -14.565, -14.282,
    -14.011, -13.756, -13.520, -13.302, -15.555, -15.360, -15.122, -14.852,
    -14.567, -14.282, -14.011, -13.756, -13.520, -13.302, -15.562, -15.366,
    -15.127, -14.855, -14.568, -14.283, -14.011, -13.756, -13.520, -13.301,
    -15.569, -15.372, -15.131, -14.857, -14.570, -14.284, -14.011, -13.756,
    -13.520, -13.301, -15.576, -15.378, -15.135, -14.860, -14.571, -14.284,
    -14.011, -13.756, -13.520, -13.301, -15.583, -15.384, -15.139, -14.862,
    -14.572, -14.284, -14.011, -13.755, -13.519, -13.301, -15.590, -15.389,
    -15.143, -14.865, -14.573, -14.284, -14.010, -13.755, -13.519, -13.301,
    -15.597, -15.395, -15.147, -14.867, -14.574, -14.284, -14.010, -13.754,
    -13.519, -13.301, -15.604, -15.400, -15.150, -14.869, -14.575, -14.284,
    -14.010, -13.754, -13.519, -13.301, -15.610, -15.405, -15.154, -14.870,
    -14.576, -14.284, -14.009, -13.754, -13.518, -13.301, -15.617, -15.410,
    -15.157, -14.872, -14.576, -14.284, -14.008, -13.753, -13.518, -13.301,
    -15.623, -15.414, -15.160, -14.873, -14.576, -14.284, -14.008, -13.752,
    -13.518, -13.300, -15.629, -15.419, -15.162, -14.874, -14.576, -14.283,
    -14.007, -13.752, -13.517, -13.300, -15.635, -15.423, -15.165, -14.875,
    -14.576, -14.282, -14.006, -13.751, -13.517, -13.300, -15.641, -15.427,
    -15.167, -14.876, -14.576, -14.282, -14.005, -13.750, -13.516, -13.300,
    -15.647, -15.431, -15.169, -14.877, -14.576, -14.281, -14.004, -13.750,
    -13.516, -13.300, -15.652, -15.435, -15.171, -14.877, -14.575, -14.280,
    -14.003, -13.749, -13.516, -13.300, -15.658, -15.438, -15.172, -14.877,
    -14.574, -14.279, -14.002, -13.748, -13.515, -13.299, -15.663, -15.442,
    -15.174, -14.877, -14.573, -14.277, -14.001, -13.747, -13.515, -13.299,
    -15.668, -15.445, -15.175, -14.877, -14.572, -14.276, -14.000, -13.746,
    -13.514, -13.299, -15.673, -15.447, -15.176, -14.876, -14.571, -14.274,
    -13.998, -13.745, -13.513, -13.299, -15.677, -15.450, -15.176, -14.875,
    -14.569, -14.273, -13.997, -13.744, -13.513, -13.299, -15.682, -15.452,
    -15.176, -14.874, -14.567, -14.271, -13.995, -13.743, -13.512, -13.298,
    -15.686, -15.454, -15.176, -14.873, -14.566, -14.269, -13.994, -13.742,
    -13.512, -13.298, -15.690, -15.456, -15.176, -14.871, -14.563, -14.267,
    -13.992, -13.741, -13.511, -13.298, -15.693, -15.457, -15.175, -14.869,
    -14.561, -14.265, -13.990, -13.739, -13.510, -13.298, -15.697, -15.458,
    -15.175, -14.867, -14.559, -14.262, -13.988, -13.738, -13.510, -13.298,
    -15.700, -15.459, -15.173, -14.865, -14.556, -14.260, -13.986, -13.737,
    -13.509, -13.297, -15.703, -15.459, -15.172, -14.862, -14.553, -14.257,
    -13.984, -13.735, -13.508, -13.297, -15.706, -15.459, -15.170, -14.859,
    -14.550, -14.254, -13.982, -13.734, -13.508, -13.297, -15.708, -15.459,
    -15.167, -14.856, -14.546, -14.251, -13.980, -13.732, -13.507, -13.296,
    -15.710, -15.458, -15.165, -14.852, -14.543, -14.248, -13.977, -13.731,
    -13.506, -13.296, -15.711, -15.457, -15.162, -14.849, -14.539, -14.245,
    -13.975, -13.729, -13.505, -13.296, -15.712, -15.456, -15.158, -14.844,
    -14.535, -14.242, -13.972, -13.728, -13.504, -13.296, -15.713, -15.454,
    -15.155, -14.840, -14.531, -14.238, -13.970, -13.726, -13.503, -13.295,
    -15.714, -15.451, -15.150, -14.835, -14.526, -14.234, -13.967, -13.724,
    -13.503, -13.295, -15.714, -15.448, -15.146, -14.830, -14.522, -14.230,
    -13.964, -13.723, -13.502, -13.295, -15.713, -15.445, -15.141, -14.825,
    -14.517, -14.227, -13.961, -13.721, -13.501, -13.294, -15.712, -15.441,
    -15.135, -14.819, -14.511, -14.222, -13.958, -13.719, -13.500, -13.294,
    -15.711, -15.437, -15.130, -14.813, -14.506, -14.218, -13.955, -13.717,
    -13.499, -13.294, -15.709, -15.432, -15.123, -14.807, -14.501, -14.214,
    -13.952, -13.715, -13.498, -13.293, -15.706, -15.427, -15.117, -14.800,
    -14.495, -14.209, -13.949, -13.713, -13.497, -13.293, -15.703, -15.421,
    -15.109, -14.793, -14.489, -14.205, -13.946, -13.711, -13.496, -13.293,
    -15.700, -15.414, -15.102, -14.785, -14.482, -14.200, -13.942, -13.709,
    -13.495, -13.292, -15.695, -15.407, -15.094, -14.778, -14.476, -14.195,
    -13.939, -13.707, -13.494, -13.292, -15.690, -15.400, -15.085, -14.770,
    -14.469, -14.190, -13.936, -13.705, -13.493, -13.292, -15.685, -15.391,
    -15.076, -14.761, -14.462, -14.185, -13.932, -13.703, -13.492, -13.291,
    -15.679, -15.382, -15.066, -14.752, -14.455, -14.179, -13.928, -13.701,
    -13.491, -13.291, -15.672, -15.373, -15.056, -14.743, -14.448, -14.174,
    -13.925, -13.698, -13.490, -13.291, -15.664, -15.363, -15.046, -14.734,
    -14.440, -14.168, -13.921, -13.696, -13.489, -13.290, -15.655, -15.352,
    -15.035, -14.724, -14.432, -14.162, -13.917, -13.694, -13.487, -13.290,
    -15.646, -15.340, -15.023, -14.714, -14.424, -14.156, -13.913, -13.691,
    -13.486, -13.290, -15.636, -15.328, -15.011, -14.704, -14.416, -14.150,
    -13.909, -13.689, -13.485, -13.289, -15.624, -15.315, -14.998, -14.693,
    -14.407, -14.144, -13.905, -13.687, -13.484, -13.289, -15.612, -15.301,
    -14.985, -14.682, -14.399, -14.138, -13.901, -13.684, -13.483, -13.288,
    -15.600, -15.287, -14.972, -14.670, -14.390, -14.132, -13.897, -13.682,
    -13.482, -13.288, -15.586, -15.272, -14.958, -14.658, -14.381, -14.125,
    -13.892, -13.679, -13.481, -13.288, -15.571, -15.256, -14.943, -14.646,
    -14.372, -14.119, -13.888, -13.677, -13.479, -13.287, -15.555, -15.240,
    -14.928, -14.634, -14.362, -14.112, -13.884, -13.674, -13.478, -13.287,
    -15.538, -15.222, -14.913, -14.621, -14.353, -14.105, -13.879, -13.672,
    -13.477, -13.287, -15.521, -15.204, -14.897, -14.608, -14.343, -14.098,
    -13.875, -13.669, -13.476, -13.286, -15.502, -15.186, -14.880, -14.595,
    -14.333, -14.091, -13.870, -13.667, -13.475, -13.286, -15.482, -15.166,
    -14.863, -14.581, -14.323, -14.084, -13.866, -13.664, -13.473, -13.286,
    -15.461, -15.146, -14.846, -14.568, -14.312, -14.077, -13.861, -13.662,
    -13.472, -13.285, -15.439, -15.125, -14.828, -14.553, -14.302, -14.070,
    -13.857, -13.659, -13.471, -13.285, -15.416, -15.103, -14.809, -14.539,
    -14.292, -14.063, -13.852, -13.656, -13.470, -13.284, -15.392, -15.081,
    -14.791, -14.525, -14.281, -14.056, -13.847, -13.654, -13.468, -13.284,
    -15.367, -15.058, -14.772, -14.510, -14.270, -14.048, -13.843, -13.651,
    -13.467, -13.284, -15.341, -15.035, -14.752, -14.495, -14.259, -14.041,
    -13.838, -13.648, -13.466, -13.283, -15.314, -15.010, -14.732, -14.479,
    -14.248, -14.033, -13.833, -13.646, -13.465, -13.283, -15.285, -14.985,
    -14.712, -14.464, -14.237, -14.026, -13.829, -13.643, -13.463, -13.283,
    -15.256, -14.960, -14.691, -14.448, -14.226, -14.018, -13.824, -13.640,
    -13.462, -13.282, -15.226, -14.934, -14.670, -14.432, -14.214, -14.010,
    -13.819, -13.638, -13.461, -13.282, -15.195, -14.907, -14.649, -14.416,
    -14.203, -14.003, -13.814, -13.635, -13.460, -13.282, -15.163, -14.880,
    -14.628, -14.400, -14.192, -13.995, -13.810, -13.632, -13.459, -13.281,
    -15.130, -14.852, -14.606, -14.384, -14.180, -13.987, -13.805, -13.630,
    -13.457, -13.281, -15.096, -14.824, -14.584, -14.368, -14.168, -13.980,
    -13.800, -13.627, -13.456, -13.281, -15.061, -14.796, -14.562, -14.351,
    -14.157, -13.972, -13.795, -13.624, -13.455, -13.280, -15.026, -14.767,
    -14.539, -14.335, -14.145, -13.964, -13.790, -13.622, -13.454, -13.280,
    -14.990, -14.737, -14.517, -14.318, -14.134, -13.957, -13.786, -13.619,
    -13.453, -13.280, -14.953, -14.708, -14.494, -14.302, -14.122, -13.949,
    -13.781, -13.616, -13.451, -13.279, -14.915, -14.678, -14.471, -14.285,
    -14.111, -13.941, -13.776, -13.614, -13.450, -13.279, -14.877, -14.648,
    -14.448, -14.268, -14.099, -13.934, -13.772, -13.611, -13.449, -13.279,
    -14.839, -14.617, -14.425, -14.252, -14.088, -13.926, -13.767, -13.609,
    -13.448, -13.279, -14.800, -14.587, -14.402, -14.235, -14.076, -13.919,
    -13.762, -13.606, -13.447, -13.278, -14.760, -14.556, -14.379, -14.219,
    -14.065, -13.911, -13.758, -13.604, -13.446, -13.278, -14.721, -14.525,
    -14.357, -14.202, -14.053, -13.904, -13.753, -13.601, -13.445, -13.278,
    -14.681, -14.495, -14.334, -14.186, -14.042, -13.897, -13.749, -13.599,
    -13.444, -13.278, -14.641, -14.464, -14.311, -14.170, -14.031, -13.889,
    -13.744, -13.597, -13.443, -13.278, -14.601, -14.434, -14.289, -14.154,
    -14.020, -13.882, -13.740, -13.594, -13.442, -13.277, -14.561, -14.403,
    -14.266, -14.138, -14.009, -13.875, -13.736, -13.592, -13.441, -13.277,
    -14.521, -14.373, -14.244, -14.122, -13.999, -13.868, -13.732, -13.590,
    -13.440, -13.277, -14.481, -14.343, -14.222, -14.107, -13.988, -13.861,
    -13.727, -13.587, -13.439, -13.277, -14.442, -14.314, -14.201, -14.091,
    -13.977, -13.854, -13.723, -13.585, -13.438, -13.277, -14.402, -14.284,
    -14.179, -14.076, -13.967, -13.848, -13.719, -13.583, -13.437, -13.277,
    -14.364, -14.255, -14.158, -14.061, -13.957, -13.841, -13.715, -13.581,
    -13.436, -13.276, -14.325, -14.227, -14.138, -14.047, -13.947, -13.835,
    -13.711, -13.579, -13.435, -13.276, -14.288, -14.199, -14.118, -14.033,
    -13.938, -13.829, -13.708, -13.577, -13.434, -13.276, -14.251, -14.172,
    -14.098, -14.019, -13.928, -13.822, -13.704, -13.575, -13.434, -13.276,
    -14.215, -14.145, -14.079, -14.005, -13.919, -13.816, -13.700, -13.573,
    -13.433, -13.276, -14.179, -14.119, -14.060, -13.992, -13.910, -13.811,
    -13.697, -13.571, -13.432, -13.276, -14.145, -14.094, -14.042, -13.979,
    -13.901, -13.805, -13.693, -13.569, -13.431, -13.276, -14.112, -14.070,
    -14.024, -13.967, -13.892, -13.799, -13.690, -13.567, -13.431, -13.276,
    -14.080, -14.046, -14.007, -13.954, -13.884, -13.794, -13.687, -13.566,
    -13.430, -13.276, -14.048, -14.023, -13.990, -13.943, -13.876, -13.789,
    -13.683, -13.564, -13.429, -13.276, -14.019, -14.001, -13.974, -13.931,
    -13.868, -13.784, -13.680, -13.562, -13.429, -13.276, -13.990, -13.980,
    -13.959, -13.921, -13.861, -13.779, -13.678, -13.561, -13.428, -13.276,
    -13.963, -13.960, -13.944, -13.910, -13.854, -13.774, -13.675, -13.559,
    -13.428, -13.276, -13.937, -13.941, -13.931, -13.900, -13.847, -13.770,
    -13.672, -13.558, -13.427, -13.276, -13.913, -13.923, -13.917, -13.891,
    -13.841, -13.765, -13.669, -13.557, -13.427, -13.276, -13.890, -13.906,
    -13.905, -13.882, -13.834, -13.761, -13.667, -13.555, -13.426, -13.276,
    -13.869, -13.890, -13.893, -13.873, -13.828, -13.757, -13.665, -13.554,
    -13.426, -13.277, -13.849, -13.875, -13.882, -13.865, -13.823, -13.754,
    -13.662, -13.553, -13.426, -13.277, -13.831, -13.862, -13.872, -13.858,
    -13.817, -13.750, -13.660, -13.552, -13.425, -13.277, -13.815, -13.849,
    -13.862, -13.851, -13.813, -13.747, -13.658, -13.551, -13.425, -13.277,
    -13.800, -13.838, -13.853, -13.844, -13.808, -13.744, -13.656, -13.550,
    -13.425, -13.277, -13.787, -13.827, -13.845, -13.838, -13.804, -13.741,
    -13.655, -13.549, -13.425, -13.278, -13.775, -13.818, -13.838, -13.833,
    -13.800, -13.738, -13.653, -13.549, -13.425, -13.278, -13.765, -13.810,
    -13.832, -13.828, -13.796, -13.736, -13.652, -13.548, -13.425, -13.278,
    -13.757, -13.803, -13.826, -13.823, -13.793, -13.734, -13.650, -13.547,
    -13.424, -13.278, -13.750, -13.797, -13.821, -13.819, -13.790, -13.732,
    -13.649, -13.547, -13.424, -13.279, -13.745, -13.792, -13.817, -13.816,
    -13.787, -13.730, -13.648, -13.546, -13.425, -13.279, -13.741, -13.788,
    -13.813, -13.813, -13.785, -13.728, -13.647, -13.546, -13.425, -13.280,
    -13.738, -13.785, -13.810, -13.811, -13.783, -13.727, -13.646, -13.546,
    -13.425, -13.280, -13.737, -13.783, -13.808, -13.809, -13.781, -13.726,
    -13.646, -13.545, -13.425, -13.280, -13.738, -13.782, -13.807, -13.807,
    -13.780, -13.725, -13.645, -13.545, -13.425, -13.281, -13.739, -13.782,
    -13.806, -13.806, -13.779, -13.724, -13.645, -13.545, -13.425, -13.281,
    -13.742, -13.783, -13.806, -13.806, -13.779, -13.724, -13.645, -13.545,
    -13.426, -13.282, -13.746, -13.785, -13.807, -13.806, -13.778, -13.723,
    -13.644, -13.545, -13.426, -13.282, -13.751, -13.788, -13.808, -13.806,
    -13.778, -13.723, -13.644, -13.546, -13.426, -13.283, -13.757, -13.792,
    -13.810, -13.807, -13.779, -13.723, -13.645, -13.546, -13.427, -13.283,
    -13.764, -13.796, -13.813, -13.808, -13.779, -13.724, -13.645, -13.546,
    -13.427, -13.284, -13.772, -13.802, -13.816, -13.810, -13.780, -13.724,
    -13.645, -13.547, -13.428, -13.285, -13.781, -13.807, -13.820, -13.812,
    -13.782, -13.725, -13.646, -13.547, -13.428, -13.285, -13.791, -13.814,
    -13.824, -13.815, -13.783, -13.726, -13.647, -13.548, -13.429, -13.286,
    -13.802, -13.821, -13.829, -13.818, -13.785, -13.727, -13.647, -13.548,
    -13.430, -13.286, -13.813, -13.829, -13.834, -13.822, -13.787, -13.729,
    -13.648, -13.549, -13.430, -13.287, -13.825, -13.838, -13.840, -13.825,
    -13.790, -13.730, -13.650, -13.550, -13.431, -13.288, -13.838, -13.847,
    -13.846, -13.830, -13.792, -13.732, -13.651, -13.551, -13.432, -13.289,
    -13.851, -13.857, -13.853, -13.834, -13.795, -13.734, -13.652, -13.552,
    -13.433, -13.289, -13.865, -13.867, -13.860, -13.839, -13.799, -13.736,
    -13.654, -13.553, -13.433, -13.290, -13.879, -13.878, -13.868, -13.844,
    -13.802, -13.739, -13.655, -13.554, -13.434, -13.291, -13.894, -13.889,
    -13.876, -13.850, -13.806, -13.741, -13.657, -13.555, -13.435, -13.292,
    -13.909, -13.900, -13.884, -13.856, -13.810, -13.744, -13.659, -13.557,
    -13.436, -13.293, -13.924, -13.912, -13.893, -13.862, -13.814, -13.747,
    -13.661, -13.558, -13.437, -13.294, -13.940, -13.924, -13.902, -13.869,
    -13.819, -13.750, -13.663, -13.560, -13.439, -13.294, -13.956, -13.937,
    -13.911, -13.875, -13.824, -13.753, -13.665, -13.561, -13.440, -13.295,
    -13.973, -13.950, -13.921, -13.882, -13.829, -13.757, -13.667, -13.563,
    -13.441, -13.296, -13.989, -13.963, -13.931, -13.890, -13.834, -13.760,
    -13.670, -13.564, -13.442, -13.297, -14.006, -13.976, -13.941, -13.897,
    -13.839, -13.764, -13.672, -13.566, -13.443, -13.298, -14.023, -13.990,
    -13.952, -13.905, -13.845, -13.768, -13.675, -13.568, -13.445, -13.299,
    -14.041, -14.004, -13.963, -13.913, -13.851, -13.772, -13.678, -13.570,
    -13.446, -13.300, -14.058, -14.018, -13.974, -13.921, -13.857, -13.776,
    -13.681, -13.572, -13.448, -13.301, -14.075, -14.032, -13.985, -13.930,
    -13.863, -13.780, -13.684, -13.574, -13.449, -13.303, -14.093, -14.047,
    -13.996, -13.938, -13.869, -13.785, -13.687, -13.576, -13.450, -13.304,
    -14.111, -14.061, -14.008, -13.947, -13.876, -13.790, -13.690, -13.578,
    -13.452, -13.305, -14.129, -14.076, -14.020, -13.956, -13.882, -13.794,
    -13.693, -13.581, -13.454, -13.306, -14.146, -14.091, -14.032, -13.966,
    -13.889, -13.799, -13.697, -13.583, -13.455, -13.307, -14.164, -14.106,
    -14.044, -13.975, -13.896, -13.804, -13.700, -13.585, -13.457, -13.308,
    -14.182, -14.121, -14.056, -13.984, -13.903, -13.809, -13.704, -13.588,
    -13.459, -13.310, -14.200, -14.136, -14.068, -13.994, -13.910, -13.815,
    -13.708, -13.590, -13.460, -13.311, -14.218, -14.151, -14.081, -14.004,
    -13.918, -13.820, -13.711, -13.593, -13.462, -13.312, -14.236, -14.167,
    -14.093, -14.014, -13.925, -13.826, -13.715, -13.596, -13.464, -13.313,
    -14.254, -14.182, -14.106, -14.024, -13.933, -13.831, -13.719, -13.598,
    -13.466, -13.315, -14.272, -14.197, -14.119, -14.034, -13.941, -13.837,
    -13.723, -13.601, -13.468, -13.316, -14.290, -14.213, -14.132, -14.044,
    -13.949, -13.843, -13.728, -13.604, -13.470, -13.318, -14.307, -14.228,
    -14.144, -14.055, -13.957, -13.849, -13.732, -13.607, -13.472, -13.319,
    -14.325, -14.244, -14.157, -14.065, -13.965, -13.855, -13.736, -13.610,
    -13.474, -13.320, -14.343, -14.259, -14.170, -14.076, -13.973, -13.861,
    -13.740, -13.613, -13.476, -13.322, -14.361, -14.275, -14.184, -14.086,
    -13.981, -13.867, -13.745, -13.616, -13.478, -13.323, -14.378, -14.290,
    -14.197, -14.097, -13.990, -13.873, -13.749, -13.619, -13.480, -13.325,
    -14.396, -14.306, -14.210, -14.108, -13.998, -13.880, -13.754, -13.622,
    -13.482, -13.326, -14.413, -14.321, -14.223, -14.118, -14.007, -13.886,
    -13.759, -13.626, -13.484, -13.328, -14.430, -14.336, -14.236, -14.129,
    -14.015, -13.893, -13.763, -13.629, -13.487, -13.329, -14.448, -14.352,
    -14.249, -14.140, -14.024, -13.899, -13.768, -13.632, -13.489, -13.331,
    -14.465, -14.367, -14.263, -14.151, -14.033, -13.906, -13.773, -13.636,
    -13.491, -13.332, -14.482, -14.383, -14.276, -14.162, -14.041, -13.913,
    -13.778, -13.639, -13.493, -13.334, -14.499, -14.398, -14.289, -14.173,
    -14.050, -13.919, -13.783, -13.642, -13.496, -13.336, -14.516, -14.413,
    -14.302, -14.184, -14.059, -13.926, -13.788, -13.646, -13.498, -13.337,
    -14.533, -14.428, -14.316, -14.195, -14.068, -13.933, -13.793, -13.650,
    -13.501, -13.339, -14.550, -14.443, -14.329, -14.207, -14.077, -13.940,
    -13.798, -13.653, -13.503, -13.341, -14.566, -14.458, -14.342, -14.218,
    -14.086, -13.947, -13.804, -13.657, -13.505, -13.342, -14.583, -14.474,
    -14.355, -14.229, -14.095, -13.954, -13.809, -13.661, -13.508, -13.344,
    -14.600, -14.489, -14.368, -14.240, -14.105, -13.962, -13.814, -13.664,
    -13.511, -13.346, -14.616, -14.503, -14.382, -14.251, -14.114, -13.969,
    -13.819, -13.668, -13.513, -13.347, -14.632, -14.518, -14.395, -14.263,
    -14.123, -13.976, -13.825, -13.672, -13.516, -13.349, -14.648, -14.533,
    -14.408, -14.274, -14.132, -13.983, -13.830, -13.676, -13.518, -13.351,
    -14.665, -14.548, -14.421, -14.285, -14.141, -13.991, -13.836, -13.680,
    -13.521, -13.353, -14.681, -14.563, -14.434, -14.296, -14.151, -13.998,
    -13.841, -13.684, -13.524, -13.355, -14.696, -14.577, -14.447, -14.308,
    -14.160, -14.005, -13.847, -13.688, -13.526, -13.356, -14.712, -14.592,
    -14.460, -14.319, -14.169, -14.013, -13.852, -13.691, -13.529, -13.358,
    -14.728, -14.606, -14.473, -14.330, -14.179, -14.020, -13.858, -13.696,
    -13.532, -13.360, -14.744, -14.621, -14.486, -14.341, -14.188, -14.028,
    -13.864, -13.700, -13.535, -13.362, -14.759, -14.635, -14.499, -14.353,
    -14.198, -14.035, -13.869, -13.704, -13.537, -13.364, -14.775, -14.649,
    -14.512, -14.364, -14.207, -14.043, -13.875, -13.708, -13.540, -13.366,
    -14.790, -14.664, -14.525, -14.375, -14.216, -14.050, -13.881, -13.712,
    -13.543, -13.368, -14.805, -14.678, -14.538, -14.386, -14.226, -14.058,
    -13.887, -13.716, -13.546, -13.370, -14.820, -14.692, -14.550, -14.398,
    -14.235, -14.066, -13.893, -13.720, -13.549, -13.371, -14.835, -14.706,
    -14.563, -14.409, -14.245, -14.073, -13.898, -13.724, -13.552, -13.373,
    -14.850, -14.720, -14.576, -14.420, -14.254, -14.081, -13.904, -13.729,
    -13.554, -13.375, -14.865, -14.734, -14.588, -14.431, -14.264, -14.089,
    -13.910, -13.733, -13.557, -13.377, -14.880, -14.748, -14.601, -14.442,
    -14.273, -14.096, -13.916, -13.737, -13.560, -13.379, -14.895, -14.761,
    -14.614, -14.453, -14.283, -14.104, -13.922, -13.741, -13.563, -13.381,
    -14.909, -14.775, -14.626, -14.464, -14.292, -14.112, -13.928, -13.746,
    -13.566, -13.383, -14.924, -14.789, -14.639, -14.475, -14.301, -14.119,
    -13.934, -13.750, -13.569, -13.385, -14.938, -14.802, -14.651, -14.486,
    -14.311, -14.127, -13.940, -13.754, -13.572, -13.387, -14.952, -14.816,
    -14.663, -14.497, -14.320, -14.135, -13.946, -13.759, -13.575, -13.389,
    -14.967, -14.829, -14.676, -14.508, -14.330, -14.142, -13.952, -13.763,
    -13.578, -13.392, -14.981, -14.843, -14.688, -14.519, -14.339, -14.150,
    -13.958, -13.768, -13.581, -13.394, -14.995, -14.856, -14.700, -14.530,
    -14.349, -14.158, -13.964, -13.772, -13.584, -13.396, -15.009, -14.869,
    -14.712, -14.541, -14.358, -14.166, -13.970, -13.776, -13.588, -13.398,
    -15.023, -14.882, -14.724, -14.552, -14.367, -14.173, -13.976, -13.781,
    -13.591, -13.400, -15.036, -14.895, -14.737, -14.563, -14.377, -14.181,
    -13.982, -13.785, -13.594, -13.402, -15.050, -14.908, -14.749, -14.574,
    -14.386, -14.189, -13.988, -13.790, -13.597, -13.404, -15.064, -14.921,
    -14.761, -14.584, -14.395, -14.197, -13.994, -13.794, -13.600, -13.406,
    -15.077, -14.934, -14.772, -14.595, -14.405, -14.204, -14.000, -13.799,
    -13.603, -13.408, -15.091, -14.947, -14.784, -14.606, -14.414, -14.212,
    -14.006, -13.803, -13.606, -13.411, -15.104, -14.960, -14.796, -14.616,
    -14.423, -14.220, -14.012, -13.808, -13.609, -13.413, -15.118, -14.972,
    -14.808, -14.627, -14.433, -14.228, -14.018, -13.812, -13.612, -13.415,
    -15.131, -14.985, -14.820, -14.638, -14.442, -14.235, -14.024, -13.817,
    -13.616, -13.417, -15.144, -14.997, -14.831, -14.648, -14.451, -14.243,
    -14.031, -13.821, -13.619, -13.419, -15.157, -15.010, -14.843, -14.659,
    -14.460, -14.251, -14.037, -13.826, -13.622, -13.421, -15.170, -15.022,
    -14.855, -14.669, -14.470, -14.259, -14.043, -13.830, -13.625, -13.424,
    -15.183, -15.035, -14.866, -14.680, -14.479, -14.266, -14.049, -13.835,
    -13.628, -13.426, -15.196, -15.047, -14.878, -14.690, -14.488, -14.274,
    -14.055, -13.839, -13.632, -13.428, -15.209, -15.059, -14.889, -14.701,
    -14.497, -14.282, -14.061, -13.844, -13.635, -13.430, -15.221, -15.071,
    -14.901, -14.711, -14.506, -14.289, -14.067, -13.848, -13.638, -13.432,
    -15.234, -15.084, -14.912, -14.721, -14.515, -14.297, -14.073, -13.853,
    -13.641, -13.435, -15.247, -15.096, -14.923, -14.732, -14.525, -14.305,
    -14.079, -13.857, -13.644, -13.437, -15.259, -15.108, -14.934, -14.742,
    -14.534, -14.312, -14.085, -13.862, -13.648, -13.439, -15.272, -15.120,
    -14.946, -14.752, -14.543, -14.320, -14.091, -13.866, -13.651, -13.441,
    -15.284, -15.132, -14.957, -14.762, -14.552, -14.327, -14.097, -13.871,
    -13.654, -13.443, -15.296, -15.143, -14.968, -14.773, -14.561, -14.335,
    -14.103, -13.875, -13.657, -13.446, -15.309, -15.155, -14.979, -14.783,
    -14.570, -14.343, -14.109, -13.880, -13.660, -13.448, -15.321, -15.167,
    -14.990, -14.793, -14.579, -14.350, -14.115, -13.884, -13.664, -13.450,
    -15.333, -15.179, -15.001, -14.803, -14.587, -14.358, -14.121, -13.889,
    -13.667, -13.452, -15.345, -15.190, -15.012, -14.813, -14.596, -14.365,
    -14.127, -13.893, -13.670, -13.455, -15.357, -15.202, -15.023, -14.823,
    -14.605, -14.373, -14.133, -13.898, -13.673, -13.457, -15.369, -15.213,
    -15.034, -14.833, -14.614, -14.380, -14.139, -13.902, -13.676, -13.459,
    -15.381, -15.225, -15.045, -14.843, -14.623, -14.387, -14.145, -13.907,
    -13.680, -13.461, -15.393, -15.236, -15.055, -14.853, -14.632, -14.395,
    -14.151, -13.911, -13.683, -13.464, -15.404, -15.247, -15.066, -14.863,
    -14.640, -14.402, -14.157, -13.915, -13.686, -13.466, -15.416, -15.259,
    -15.077, -14.872, -14.649, -14.410, -14.163, -13.920, -13.689, -13.468,
    -15.428, -15.270, -15.087, -14.882, -14.658, -14.417, -14.169, -13.924,
    -13.692, -13.470, -15.439, -15.281, -15.098, -14.892, -14.667, -14.424,
    -14.174, -13.929, -13.696, -13.473, -15.451, -15.292, -15.109, -14.902,
    -14.675, -14.432, -14.180, -13.933, -13.699, -13.475, -15.462, -15.303,
    -15.119, -14.911, -14.684, -14.439, -14.186, -13.938, -13.702, -13.477,
    -15.474, -15.314, -15.130, -14.921, -14.692, -14.446, -14.192, -13.942,
    -13.705, -13.480, -15.485, -15.325, -15.140, -14.931, -14.701, -14.453,
    -14.198, -13.946, -13.708, -13.482, -15.497, -15.336, -15.150, -14.940,
    -14.710, -14.461, -14.203, -13.951, -13.711, -13.484, -15.508, -15.347,
    -15.161, -14.950, -14.718, -14.468, -14.209, -13.955, -13.715, -13.486,
    -15.519, -15.358, -15.171, -14.959, -14.727, -14.475, -14.215, -13.959,
    -13.718, -13.489, -15.530, -15.369, -15.181, -14.969, -14.735, -14.482,
    -14.220, -13.964, -13.721, -13.491, -15.541, -15.380, -15.192, -14.978,
    -14.743, -14.489, -14.226, -13.968, -13.724, -13.493, -15.552, -15.391,
    -15.202, -14.988, -14.752, -14.496, -14.232, -13.972, -13.727, -13.495,
    -15.563, -15.401, -15.212, -14.997, -14.760, -14.503, -14.237, -13.976,
    -13.730, -13.498, -15.574, -15.412, -15.222, -15.007, -14.769, -14.510,
    -14.243, -13.981, -13.734, -13.500, -15.585, -15.422, -15.232, -15.016,
    -14.777, -14.517, -14.248, -13.985, -13.737, -13.502, -15.596, -15.433,
    -15.242, -15.025, -14.785, -14.524, -14.254, -13.989, -13.740, -13.504,
    -15.607, -15.444, -15.252, -15.034, -14.793, -14.531, -14.259, -13.993,
    -13.743, -13.507, -15.618, -15.454, -15.262, -15.044, -14.802, -14.538,
    -14.265, -13.997, -13.746, -13.509, -15.629, -15.464, -15.272, -15.053,
    -14.810, -14.545, -14.270, -14.002, -13.749, -13.511, -15.639, -15.475,
    -15.282, -15.062, -14.818, -14.552, -14.276, -14.006, -13.752, -13.514,
    -15.650, -15.485, -15.292, -15.071, -14.826, -14.559, -14.281, -14.010,
    -13.755, -13.516, -15.661, -15.495, -15.302, -15.080, -14.834, -14.566,
    -14.287, -14.014, -13.758, -13.518, -15.671, -15.506, -15.311, -15.089,
    -14.842, -14.572, -14.292, -14.018, -13.761, -13.520, -15.682, -15.516,
    -15.321, -15.098, -14.850, -14.579, -14.297, -14.022, -13.764, -13.523,
    -15.692, -15.526, -15.331, -15.107, -14.858, -14.586, -14.303, -14.026,
    -13.767, -13.525, -15.703, -15.536, -15.341, -15.116, -14.866, -14.592,
    -14.308, -14.030, -13.770, -13.527, -15.713, -15.546, -15.350, -15.125,
    -14.874, -14.599, -14.313, -14.034, -13.773, -13.529, -15.723, -15.557,
    -15.360, -15.134, -14.882, -14.605, -14.318, -14.038, -13.776, -13.532,
    -15.734, -15.567, -15.369, -15.143, -14.890, -14.612, -14.323, -14.042,
    -13.779, -13.534, -15.744, -15.577, -15.379, -15.152, -14.898, -14.619,
    -14.329, -14.046, -13.782, -13.536, -15.754, -15.587, -15.389, -15.161,
    -14.905, -14.625, -14.334, -14.050, -13.785, -13.538, -15.764, -15.596,
    -15.398, -15.169, -14.913, -14.631, -14.339, -14.054, -13.788, -13.540,
    -15.774, -15.606, -15.408, -15.178, -14.921, -14.638, -14.344, -14.058,
    -13.791, -13.543, -15.785, -15.616, -15.417, -15.187, -14.929, -14.644,
    -14.349, -14.062, -13.794, -13.545, -15.795, -15.626, -15.426, -15.196,
    -14.936, -14.651, -14.354, -14.066, -13.797, -13.547, -15.805, -15.636,
    -15.436, -15.204, -14.944, -14.657, -14.359, -14.069, -13.800, -13.549,
    -15.815, -15.646, -15.445, -15.213, -14.951, -14.663, -14.364, -14.073,
    -13.803, -13.552, -15.825, -15.655, -15.454, -15.221, -14.959, -14.669,
    -14.369, -14.077, -13.806, -13.554, -15.835, -15.665, -15.464, -15.230,
    -14.966, -14.675, -14.373, -14.081, -13.809, -13.556, -15.845, -15.675,
    -15.473, -15.238, -14.974, -14.682, -14.378, -14.085, -13.812, -13.558,
    -15.854, -15.684, -15.482, -15.247, -14.981, -14.688, -14.383, -14.088,
    -13.815, -13.560, -15.864, -15.694, -15.491, -15.255, -14.989, -14.694,
    -14.388, -14.092, -13.818, -13.563, -15.874, -15.703, -15.500, -15.264,
    -14.996, -14.700, -14.393, -14.096, -13.821, -13.565, -15.884, -15.713,
    -15.510, -15.272, -15.003, -14.706, -14.397, -14.099, -13.823, -13.567,
    -15.894, -15.722, -15.519, -15.281, -15.011, -14.712, -14.402, -14.103,
    -13.826, -13.569, -15.903, -15.732, -15.528, -15.289, -15.018, -14.718,
    -14.407, -14.107, -13.829, -13.571, -15.913, -15.741, -15.537, -15.297,
    -15.025, -14.724, -14.411, -14.110, -13.832, -13.574, -15.923, -15.751,
    -15.546, -15.306, -15.032, -14.729, -14.416, -14.114, -13.835, -13.576,
    -15.932, -15.760, -15.555, -15.314, -15.040, -14.735, -14.420, -14.117,
    -13.838, -13.578, -15.942, -15.770, -15.564, -15.322, -15.047, -14.741,
    -14.425, -14.121, -13.840, -13.580, -15.951, -15.779, -15.573, -15.330,
    -15.054, -14.747, -14.429, -14.125, -13.843, -13.582, -15.961, -15.788,
    -15.582, -15.338, -15.061, -14.752, -14.434, -14.128, -13.846, -13.585,
    -15.970, -15.797, -15.591, -15.346, -15.068, -14.758, -14.438, -14.132,
    -13.849, -13.587, -15.980, -15.807, -15.599, -15.354, -15.075, -14.764,
    -14.443, -14.135, -13.852, -13.589, -15.989, -15.816, -15.608, -15.363,
    -15.082, -14.769, -14.447, -14.139, -13.854, -13.591, -15.998, -15.825,
    -15.617, -15.371, -15.089, -14.775, -14.451, -14.142, -13.857, -13.593,
    -16.008, -15.834, -15.626, -15.379, -15.095, -14.780, -14.456, -14.145,
    -13.860, -13.595, -16.017, -15.843, -15.635, -15.386, -15.102, -14.786,
    -14.460, -14.149, -13.863, -13.598, -16.026, -15.852, -15.643, -15.394,
    -15.109, -14.791, -14.464, -14.152, -13.865, -13.600, -16.036, -15.861,
    -15.652, -15.402, -15.116, -14.797, -14.469, -14.156, -13.868, -13.602,
    -16.045, -15.870, -15.661, -15.410, -15.122, -14.802, -14.473, -14.159,
    -13.871, -13.604, -16.054, -15.880, -15.670, -15.418, -15.129, -14.807,
    -14.477, -14.162, -13.874, -13.606, -16.063, -15.889, -15.678, -15.426,
    -15.136, -14.813, -14.481, -14.166, -13.876, -13.608, -16.073, -15.897,
    -15.687, -15.434, -15.142, -14.818, -14.485, -14.169, -13.879, -13.611,
    -16.082, -15.906, -15.695, -15.441, -15.149, -14.823, -14.489, -14.172,
    -13.882, -13.613, -16.091, -15.915, -15.704, -15.449, -15.155, -14.828,
    -14.494, -14.176, -13.884, -13.615, -16.100, -15.924, -15.713, -15.457,
    -15.162, -14.834, -14.498, -14.179, -13.887, -13.617, -16.109, -15.933,
    -15.721, -15.464, -15.168, -14.839, -14.502, -14.182, -13.890, -13.619,
    -16.118, -15.942, -15.730, -15.472, -15.175, -14.844, -14.506, -14.186,
    -13.892, -13.621, -16.127, -15.951, -15.738, -15.480, -15.181, -14.849,
    -14.510, -14.189, -13.895, -13.623, -16.136, -15.960, -15.747, -15.487,
    -15.188, -14.854, -14.514, -14.192, -13.898, -13.626, -16.145, -15.968,
    -15.755, -15.495, -15.194, -14.859, -14.518, -14.196, -13.901, -13.628,
    -16.154, -15.977, -15.763, -15.502, -15.200, -14.864, -14.522, -14.199,
    -13.903, -13.630, -16.163, -15.986, -15.772, -15.510, -15.207, -14.869,
    -14.526, -14.202, -13.906, -13.632, -16.172, -15.995, -15.780, -15.517,
    -15.213, -14.874, -14.530, -14.205, -13.909, -13.634, -16.181, -16.003,
    -15.789, -15.525, -15.219, -14.879, -14.534, -14.209, -13.911, -13.636,
    -16.189, -16.012, -15.797, -15.532, -15.225, -14.884, -14.538, -14.212,
    -13.914, -13.638, -16.198, -16.021, -15.805, -15.539, -15.232, -14.889,
    -14.542, -14.215, -13.917, -13.641, -16.207, -16.029, -15.814, -15.547,
    -15.238, -14.894, -14.546, -14.218, -13.919, -13.643, -16.216, -16.038,
    -15.822, -15.554, -15.244, -14.899, -14.549, -14.222, -13.922, -13.645,
    -16.225, -16.046, -15.830, -15.561, -15.250, -14.904, -14.553, -14.225,
    -13.925, -13.647, -16.233, -16.055, -15.838, -15.569, -15.256, -14.909,
    -14.557, -14.228, -13.927, -13.649, -16.242, -16.063, -15.847, -15.576,
    -15.262, -14.913, -14.561, -14.231, -13.930, -13.651, -16.251, -16.072,
    -15.855, -15.583, -15.268, -14.918, -14.565, -14.234, -13.933, -13.653,
    -16.259, -16.080, -15.863, -15.590, -15.274, -14.923, -14.569, -14.238,
    -13.935, -13.656, -16.268, -16.089, -15.871, -15.598, -15.281, -14.928,
    -14.573, -14.241, -13.938, -13.658, -16.277, -16.097, -15.879, -15.605,
    -15.287, -14.933, -14.577, -14.244, -13.941, -13.660, -16.285, -16.106,
    -15.887, -15.612, -15.293, -14.938, -14.581, -14.247, -13.943, -13.662,
    -16.294, -16.114, -15.896, -15.619, -15.299, -14.943, -14.585, -14.251,
    -13.946, -13.664, -16.302, -16.123, -15.904, -15.626, -15.305, -14.947,
    -14.589, -14.254, -13.949, -13.666, -16.311, -16.131, -15.912, -15.634,
    -15.311, -14.952, -14.593, -14.257, -13.951, -13.668, -16.320, -16.139,
    -15.920, -15.641, -15.317, -14.957, -14.596, -14.260, -13.954, -13.671,
    -16.328, -16.148, -15.928, -15.648, -15.323, -14.962, -14.600, -14.264,
    -13.957, -13.673, -16.336, -16.156, -15.936, -15.655, -15.329, -14.967,
    -14.604, -14.267, -13.960, -13.675, -16.345, -16.164, -15.944, -15.662,
    -15.335, -14.972, -14.608, -14.270, -13.962, -13.677, -16.353, -16.173,
    -15.952, -15.669, -15.341, -14.977, -14.612, -14.274, -13.965, -13.679,
    -16.362, -16.181, -15.960, -15.676, -15.347, -14.981, -14.616, -14.277,
    -13.968, -13.681, -16.370, -16.189, -15.968, -15.683, -15.353, -14.986,
    -14.620, -14.280, -13.970, -13.684, -16.379, -16.197, -15.976, -15.690,
    -15.359, -14.991, -14.624, -14.284, -13.973, -13.686, -16.387, -16.206,
    -15.984, -15.697, -15.365, -14.996, -14.628, -14.287, -13.976, -13.688,
    -16.395, -16.214, -15.992, -15.705, -15.371, -15.001, -14.632, -14.290,
    -13.979, -13.690, -16.404, -16.222, -15.999, -15.712, -15.377, -15.006,
    -14.636, -14.294, -13.981, -13.692, -16.412, -16.230, -16.007, -15.719,
    -15.383, -15.011, -14.641, -14.297, -13.984, -13.694, -16.420, -16.238,
    -16.015, -15.726, -15.389, -15.016, -14.645, -14.300, -13.987, -13.697,
    -16.429, -16.246, -16.023, -15.733, -15.395, -15.021, -14.649, -14.304,
    -13.990, -13.699, -16.437, -16.254, -16.031, -15.740, -15.401, -15.026,
    -14.653, -14.307, -13.992, -13.701, -16.445, -16.263, -16.039, -15.747,
    -15.407, -15.031, -14.657, -14.311, -13.995, -13.703, -16.453, -16.271,
    -16.046, -15.754, -15.414, -15.036, -14.661, -14.314, -13.998, -13.705,
    -16.462, -16.279, -16.054, -15.761, -15.420, -15.041, -14.665, -14.318,
    -14.001, -13.707, -16.470, -16.287, -16.062, -15.768, -15.426, -15.047,
    -14.670, -14.321, -14.004, -13.710, -16.478, -16.295, -16.070, -15.775,
    -15.432, -15.052, -14.674, -14.325, -14.006, -13.712, -16.486, -16.303,
    -16.077, -15.782, -15.438, -15.057, -14.678, -14.328, -14.009, -13.714,
    -16.494, -16.311, -16.085, -15.789, -15.445, -15.062, -14.683, -14.332,
    -14.012, -13.716, -16.502, -16.319, -16.093, -15.797, -15.451, -15.067,
    -14.687, -14.335, -14.015, -13.718, -16.511, -16.327, -16.101, -15.804,
    -15.457, -15.073, -14.691, -14.339, -14.018, -13.721, -16.519, -16.335,
    -16.108, -15.811, -15.463, -15.078, -14.696, -14.342, -14.021, -13.723,
    -16.527, -16.343, -16.116, -15.818, -15.470, -15.083, -14.700, -14.346,
    -14.024, -13.725, -16.535, -16.351, -16.124, -15.825, -15.476, -15.089,
    -14.704, -14.350, -14.026, -13.727, -16.543, -16.358, -16.131, -15.832,
    -15.482, -15.094, -14.709, -14.353, -14.029, -13.729, -16.551, -16.366,
    -16.139, -15.839, -15.489, -15.099, -14.713, -14.357, -14.032, -13.732,
    -16.559, -16.374, -16.147, -15.846, -15.495, -15.105, -14.718, -14.360,
    -14.035, -13.734, -16.567, -16.382, -16.154, -15.853, -15.501, -15.110,
    -14.722, -14.364, -14.038, -13.736, -16.575, -16.390, -16.162, -15.861,
    -15.508, -15.116, -14.727, -14.368, -14.041, -13.738, -16.583, -16.398,
    -16.169, -15.868, -15.514, -15.121, -14.732, -14.372, -14.044, -13.740,
    -16.591, -16.406, -16.177, -15.875, -15.521, -15.127, -14.736, -14.375,
    -14.047, -13.743, -16.599, -16.413, -16.184, -15.882, -15.527, -15.132,
    -14.741, -14.379, -14.050, -13.745, -16.607, -16.421, -16.192, -15.889,
    -15.534, -15.138, -14.745, -14.383, -14.053, -13.747, -16.615, -16.429,
    -16.199, -15.896, -15.540, -15.144, -14.750, -14.387, -14.056, -13.749,
    -16.623, -16.437, -16.207, -15.903, -15.547, -15.149, -14.755, -14.391,
    -14.059, -13.752, -16.630, -16.445, -16.214, -15.910, -15.553, -15.155,
    -14.760, -14.394, -14.062, -13.754, -16.638, -16.452, -16.222, -15.918,
    -15.560, -15.161, -14.764, -14.398, -14.065, -13.756, -16.646, -16.460,
    -16.229, -15.925, -15.566, -15.166, -14.769, -14.402, -14.068, -13.758,
    -16.654, -16.468, -16.237, -15.932, -15.573, -15.172, -14.774, -14.406,
    -14.071, -13.761, -16.662, -16.475, -16.244, -15.939, -15.579, -15.178,
    -14.779, -14.410, -14.074, -13.763, -16.670, -16.483, -16.252, -15.946,
    -15.586, -15.184, -14.784, -14.414, -14.077, -13.765, -16.678, -16.491,
    -16.259, -15.953, -15.593, -15.190, -14.789, -14.418, -14.080, -13.767,
    -16.685, -16.498, -16.267, -15.960, -15.599, -15.196, -14.794, -14.422,
    -14.083, -13.770, -16.693, -16.506, -16.274, -15.967, -15.606, -15.201,
    -14.799, -14.426, -14.086, -13.772, -16.701, -16.514, -16.281, -15.975,
    -15.612, -15.207, -14.804, -14.430, -14.089, -13.774, -16.709, -16.521,
    -16.289, -15.982, -15.619, -15.213, -14.809, -14.434, -14.092, -13.776,
    -16.717, -16.529, -16.296, -15.989, -15.626, -15.219, -14.814, -14.438,
    -14.095, -13.779, -16.724, -16.537, -16.303, -15.996, -15.632, -15.225,
    -14.819, -14.442, -14.098, -13.781, -16.732, -16.544, -16.311, -16.003,
    -15.639, -15.231, -14.824, -14.446, -14.101, -13.783, -16.740, -16.552,
    -16.318, -16.010, -15.646, -15.237, -14.829, -14.450, -14.104, -13.785,
    -16.747, -16.559, -16.325, -16.017, -15.653, -15.243, -14.834, -14.454,
    -14.108, -13.788, -16.755, -16.567, -16.333, -16.024, -15.659, -15.249,
    -14.839, -14.458, -14.111, -13.790, -16.763, -16.574, -16.340, -16.031,
    -15.666, -15.255, -14.844, -14.462, -14.114, -13.792, -16.770, -16.582,
    -16.347, -16.039, -15.673, -15.261, -14.849, -14.467, -14.117, -13.794,
    -16.778, -16.590, -16.355, -16.046, -15.679, -15.267, -14.855, -14.471,
    -14.120, -13.797, -16.786, -16.597, -16.362, -16.053, -15.686, -15.273,
    -14.860, -14.475, -14.123, -13.799, -16.793, -16.605, -16.369, -16.060,
    -15.693, -15.279, -14.865, -14.479, -14.126, -13.801, -16.801, -16.612,
    -16.376, -16.067, -15.700, -15.286, -14.870, -14.483, -14.130, -13.803,
    -16.809, -16.620, -16.384, -16.074, -15.706, -15.292, -14.875, -14.488,
    -14.133, -13.806, -16.816, -16.627, -16.391, -16.081, -15.713, -15.298,
    -14.881, -14.492, -14.136, -13.808, -16.824, -16.634, -16.398, -16.088,
    -15.720, -15.304, -14.886, -14.496, -14.139, -13.810, -16.831, -16.642,
    -16.405, -16.095, -15.726, -15.310, -14.891, -14.500, -14.142, -13.812,
    -16.839, -16.649, -16.412, -16.102, -15.733, -15.316, -14.897, -14.504,
    -14.145, -13.815, -16.847, -16.657, -16.420, -16.109, -15.740, -15.323,
    -14.902, -14.509, -14.149, -13.817, -16.854, -16.664, -16.427, -16.116,
    -15.747, -15.329, -14.907, -14.513, -14.152, -13.819, -16.862, -16.672,
    -16.434, -16.123, -15.753, -15.335, -14.913, -14.517, -14.155, -13.821,
    -16.869, -16.679, -16.441, -16.130, -15.760, -15.341, -14.918, -14.522,
    -14.158, -13.824, -16.877, -16.686, -16.448, -16.137, -15.767, -15.347,
    -14.923, -14.526, -14.161, -13.826, -16.884, -16.694, -16.455, -16.144,
    -15.774, -15.354, -14.929, -14.530, -14.164, -13.828, -16.892, -16.701,
    -16.463, -16.151, -15.780, -15.360, -14.934, -14.534, -14.168, -13.830,
    -16.899, -16.708, -16.470, -16.158, -15.787, -15.366, -14.939, -14.539,
    -14.171, -13.832, -16.907, -16.716, -16.477, -16.165, -15.794, -15.372,
    -14.945, -14.543, -14.174, -13.835, -16.914, -16.723, -16.484, -16.172,
    -15.800, -15.378, -14.950, -14.547, -14.177, -13.837, -16.922, -16.731,
    -16.491, -16.179, -15.807, -15.385, -14.956, -14.552, -14.180, -13.839,
    -16.929, -16.738, -16.498, -16.186, -15.814, -15.391, -14.961, -14.556,
    -14.184, -13.841, -16.937, -16.745, -16.505, -16.193, -15.821, -15.397,
    -14.967, -14.560, -14.187, -13.844, -16.944, -16.752, -16.512, -16.200,
    -15.827, -15.403, -14.972, -14.565, -14.190, -13.846, -16.951, -16.760,
    -16.519, -16.206, -15.834, -15.410, -14.977, -14.569, -14.193, -13.848,
    -16.959, -16.767, -16.526, -16.213, -15.841, -15.416, -14.983, -14.573,
    -14.196, -13.850, -16.966, -16.774, -16.533, -16.220, -15.847, -15.422,
    -14.988, -14.578, -14.199, -13.852, -16.974, -16.782, -16.540, -16.227,
    -15.854, -15.428, -14.994, -14.582, -14.203, -13.855, -16.981, -16.789,
    -16.547, -16.234, -15.861, -15.435, -14.999, -14.586, -14.206, -13.857,
    -16.988, -16.796, -16.554, -16.241, -15.867, -15.441, -15.005, -14.591,
    -14.209, -13.859, -16.996, -16.803, -16.561, -16.248, -15.874, -15.447,
    -15.010, -14.595, -14.212, -13.861, -17.003, -16.811, -16.568, -16.255,
    -15.881, -15.453, -15.015, -14.599, -14.215, -13.863, -17.010, -16.818,
    -16.575, -16.261, -15.887, -15.459, -15.021, -14.604, -14.218, -13.865,
    -17.018, -16.825, -16.582, -16.268, -15.894, -15.466, -15.026, -14.608,
    -14.222, -13.868, -17.025, -16.832, -16.589, -16.275, -15.900, -15.472,
    -15.032, -14.612, -14.225, -13.870, -17.032, -16.839, -16.596, -16.282,
    -15.907, -15.478, -15.037, -14.617, -14.228, -13.872, -17.040, -16.847,
    -16.603, -16.289, -15.914, -15.484, -15.043, -14.621, -14.231, -13.874,
    -17.047, -16.854, -16.610, -16.295, -15.920, -15.491, -15.048, -14.625,
    -14.234, -13.876, -17.054, -16.861, -16.617, -16.302, -15.927, -15.497,
    -15.054, -14.630, -14.237, -13.878, -17.062, -16.868, -16.624, -16.309,
    -15.934, -15.503, -15.059, -14.634, -14.240, -13.880, -17.069, -16.875,
    -16.631, -16.316, -15.940, -15.509, -15.064, -14.638, -14.243, -13.883,
    -17.076, -16.882, -16.638, -16.323, -15.947, -15.515, -15.070, -14.643,
    -14.247, -13.885, -17.083, -16.890, -16.645, -16.329, -15.953, -15.522,
    -15.075, -14.647, -14.250, -13.887, -17.091, -16.897, -16.652, -16.336,
    -15.960, -15.528, -15.081, -14.651, -14.253, -13.889, -17.098, -16.904,
    -16.659, -16.343, -15.966, -15.534, -15.086, -14.656, -14.256, -13.891,
    -17.105, -16.911, -16.665, -16.349, -15.973, -15.540, -15.091, -14.660,
    -14.259, -13.893, -17.112, -16.918, -16.672, -16.356, -15.979, -15.546,
    -15.097, -14.664, -14.262, -13.895, -17.120, -16.925, -16.679, -16.363,
    -15.986, -15.552, -15.102, -14.668, -14.265, -13.897, -17.127, -16.932,
    -16.686, -16.370, -15.993, -15.558, -15.108, -14.673, -14.268, -13.899,
    -17.134, -16.939, -16.693, -16.376, -15.999, -15.565, -15.113, -14.677,
    -14.271, -13.901, -17.141, -16.946, -16.700, -16.383, -16.006, -15.571,
    -15.118, -14.681, -14.274, -13.903, -17.148, -16.953, -16.707, -16.390,
    -16.012, -15.577, -15.124, -14.685, -14.277, -13.905, -17.156, -16.960,
    -16.713, -16.396, -16.019, -15.583, -15.129, -14.689, -14.280, -13.907,
    -17.163, -16.968, -16.720, -16.403, -16.025, -15.589, -15.134, -14.694,
    -14.283, -13.909, -17.170, -16.975, -16.727, -16.410, -16.032, -15.595,
    -15.140, -14.698, -14.286, -13.911, -17.177, -16.982, -16.734, -16.416,
    -16.038, -15.601, -15.145, -14.702, -14.289, -13.913, -17.184, -16.989,
    -16.741, -16.423, -16.045, -15.607, -15.150, -14.706, -14.292, -13.915,
    -17.191, -16.996, -16.748, -16.430, -16.051, -15.613, -15.156, -14.710,
    -14.295, -13.917, -17.199, -17.003, -16.754, -16.436, -16.057, -15.619,
    -15.161, -14.714, -14.297, -13.919, -17.206, -17.010, -16.761, -16.443,
    -16.064, -15.625, -15.166, -14.718, -14.300, -13.921, -17.213, -17.017,
    -16.768, -16.449, -16.070, -15.631, -15.171, -14.722, -14.303, -13.923,
    -17.220, -17.024, -16.775, -16.456, -16.077, -15.637, -15.177, -14.727,
    -14.306, -13.925, -17.227, -17.031, -16.781, -16.463, -16.083, -15.643,
    -15.182, -14.731, -14.309, -13.927, -17.234, -17.038, -16.788, -16.469,
    -16.090, -15.649, -15.187, -14.735, -14.312, -13.929, -17.241, -17.045,
    -16.795, -16.476, -16.096, -15.655, -15.192, -14.739, -14.315, -13.931,
    -17.248, -17.052, -16.802, -16.482, -16.102, -15.661, -15.197, -14.743,
    -14.317, -13.932, -17.255, -17.059, -16.808, -16.489, -16.109, -15.667,
    -15.202, -14.747, -14.320, -13.934, -17.263, -17.066, -16.815, -16.495,
    -16.115, -15.673, -15.208, -14.751, -14.323, -13.936, -17.270, -17.073,
    -16.822, -16.502, -16.121, -15.679, -15.213, -14.754, -14.326, -13.938,
    -17.277, -17.079, -16.829, -16.509, -16.128, -15.685, -15.218, -14.758,
    -14.328, -13.940, -17.284, -17.086, -16.835, -16.515, -16.134, -15.691,
    -15.223, -14.762, -14.331, -13.942, -17.291, -17.093, -16.842, -16.522,
    -16.141, -15.697, -15.228, -14.766, -14.334, -13.943, -17.298, -17.100,
    -16.849, -16.528, -16.147, -15.703, -15.233, -14.770, -14.336, -13.945,
    -17.305, -17.107, -16.855, -16.535, -16.153, -15.709, -15.238, -14.774,
    -14.339, -13.947, -17.312, -17.114, -16.862, -16.541, -16.160, -15.714,
    -15.243, -14.778, -14.342, -13.949, -17.319, -17.121, -16.869, -16.548,
    -16.166, -15.720, -15.248, -14.782, -14.344, -13.951, -17.326, -17.128,
    -16.876, -16.554, -16.172, -15.726, -15.253, -14.785, -14.347, -13.952,
    -17.333, -17.135, -16.882, -16.561, -16.178, -15.732, -15.258, -14.789,
    -14.350, -13.954, -17.340, -17.142, -16.889, -16.567, -16.185, -15.738,
    -15.263, -14.793, -14.352, -13.956, -17.347, -17.149, -16.896, -16.574,
    -16.191, -15.744, -15.268, -14.796, -14.355, -13.957, -17.354, -17.155,
    -16.902, -16.580, -16.197, -15.749, -15.273, -14.800, -14.357, -13.959,
    -17.361, -17.162, -16.909, -16.586, -16.204, -15.755, -15.278, -14.804,
    -14.360, -13.961, -17.368, -17.169, -16.915, -16.593, -16.210, -15.761,
    -15.282, -14.807, -14.362, -13.963, -17.375, -17.176, -16.922, -16.599,
    -16.216, -15.766, -15.287, -14.811, -14.365, -13.964, -17.382, -17.183,
    -16.929, -16.606, -16.222, -15.772, -15.292, -14.815, -14.367, -13.966,
    -17.389, -17.190, -16.935, -16.612, -16.229, -15.778, -15.297, -14.818,
    -14.370, -13.968, -17.396, -17.197, -16.942, -16.619, -16.235, -15.784,
    -15.302, -14.822, -14.372, -13.969, -17.403, -17.203, -16.949, -16.625,
    -16.241, -15.789, -15.306, -14.825, -14.374, -13.971, -17.410, -17.210,
    -16.955, -16.632, -16.247, -15.795, -15.311, -14.829, -14.377, -13.972,
    -17.417, -17.217, -16.962, -16.638, -16.254, -15.800, -15.316, -14.832,
    -14.379, -13.974, -17.424, -17.224, -16.968, -16.644, -16.260, -15.806,
    -15.320, -14.836, -14.381, -13.976, -17.431, -17.231, -16.975, -16.651,
    -16.266, -15.812, -15.325, -14.839, -14.384, -13.977, -17.437, -17.238,
    -16.982, -16.657, -16.272, -15.817, -15.330, -14.843, -14.386, -13.979,
    -17.444, -17.244, -16.988, -16.663, -16.278, -15.823, -15.334, -14.846,
    -14.388, -13.980, -17.451, -17.251, -16.995, -16.670, -16.285, -15.828,
    -15.339, -14.849, -14.391, -13.982, -17.458, -17.258, -17.001, -16.676,
    -16.291, -15.834, -15.343, -14.853, -14.393, -13.983, -17.465, -17.265,
    -17.008, -16.683, -16.297, -15.839, -15.348, -14.856, -14.395, -13.985,
    -17.472, -17.271, -17.015, -16.689, -16.303, -15.845, -15.352, -14.859,
    -14.397, -13.986, -17.479, -17.278, -17.021, -16.695, -16.309, -15.850,
    -15.357, -14.862, -14.400, -13.988, -17.486, -17.285, -17.028, -16.702,
    -16.315, -15.856, -15.361, -14.866, -14.402, -13.990, -17.493, -17.292,
    -17.034, -16.708, -16.322, -15.861, -15.366, -14.869, -14.404, -13.991,
    -17.500, -17.299, -17.041, -16.714, -16.328, -15.867, -15.370, -14.872,
    -14.406, -13.992, -17.506, -17.305, -17.047, -16.721, -16.334, -15.872,
    -15.375, -14.875, -14.408, -13.994, -17.513, -17.312, -17.054, -16.727,
    -16.340, -15.877, -15.379, -14.878, -14.410, -13.995, -17.520, -17.319,
    -17.060, -16.733, -16.346, -15.883, -15.383, -14.881, -14.412, -13.997,
    -17.527, -17.326, -17.067, -16.740, -16.352, -15.888, -15.387, -14.885,
    -14.414, -13.998, -17.534, -17.332, -17.073, -16.746, -16.358, -15.893,
    -15.392, -14.888, -14.416, -14.000, -17.541, -17.339, -17.080, -16.752,
    -16.364, -15.899, -15.396, -14.891, -14.418, -14.001, -17.548, -17.346,
    -17.086, -16.759, -16.370, -15.904, -15.400, -14.894, -14.421, -14.003,
    -17.554, -17.352, -17.093, -16.765, -16.377, -15.909, -15.404, -14.897,
    -14.423, -14.004, -17.561, -17.359, -17.099, -16.771, -16.383, -15.914,
    -15.409, -14.900, -14.425, -14.006, -17.568, -17.366, -17.106, -16.777,
    -16.389, -15.920, -15.413, -14.902, -14.427, -14.007, -17.575, -17.373,
    -17.112, -16.784, -16.395, -15.925, -15.417, -14.905, -14.428, -14.008,
    -17.582, -17.379, -17.119, -16.790, -16.401, -15.930, -15.421, -14.908,
    -14.430, -14.010, -17.588, -17.386, -17.125, -16.796, -16.407, -15.935,
    -15.425, -14.911, -14.432, -14.011, -17.595, -17.393, -17.132, -16.803,
    -16.413, -15.940, -15.429, -14.914, -14.434, -14.012, -17.602, -17.399,
    -17.138, -16.809, -16.419, -15.946, -15.433, -14.917, -14.436, -14.014,
    -17.609, -17.406, -17.145, -16.815, -16.425, -15.951, -15.437, -14.920,
    -14.438, -14.015, -17.616, -17.413, -17.151, -16.821, -16.431, -15.956,
    -15.441, -14.922, -14.440, -14.017, -17.622, -17.419, -17.158, -16.828,
    -16.437, -15.961, -15.445, -14.925, -14.442, -14.018, -17.629, -17.426,
    -17.164, -16.834, -16.443, -15.966, -15.449, -14.928, -14.444, -14.019,
    -17.636, -17.433, -17.171, -16.840, -16.449, -15.971, -15.453, -14.931,
    -14.446, -14.021, -17.643, -17.439, -17.177, -16.846, -16.455, -15.976,
    -15.457, -14.933, -14.447, -14.022, -17.650, -17.446, -17.184, -16.852,
    -16.461, -15.981, -15.460, -14.936, -14.449, -14.023, -17.656, -17.453,
    -17.190, -16.859, -16.467, -15.986, -15.464, -14.939, -14.451, -14.025,
    -17.663, -17.459, -17.196, -16.865, -16.473, -15.991, -15.468, -14.941,
    -14.453, -14.026, -17.670, -17.466, -17.203, -16.871, -16.479, -15.996,
    -15.472, -14.944, -14.455, -14.027, -17.677, -17.473, -17.209, -16.877,
    -16.485, -16.001, -15.476, -14.947, -14.457, -14.029, -17.683, -17.479,
    -17.216, -16.884, -16.491, -16.006, -15.479, -14.949, -14.458, -14.030,
    -17.690, -17.486, -17.222, -16.890, -16.497, -16.010, -15.483, -14.952,
    -14.460, -14.032, -17.697, -17.493, -17.229, -16.896, -16.503, -16.015,
    -15.487, -14.954, -14.462, -14.033, -17.704, -17.499, -17.235, -16.902,
    -16.509, -16.020, -15.491, -14.957, -14.464, -14.034, -17.710, -17.506,
    -17.241, -16.908, -16.515, -16.025, -15.494, -14.959, -14.465, -14.036,
    -17.717, -17.512, -17.248, -16.915, -16.521, -16.030, -15.498, -14.962,
    -14.467, -14.037, -17.724, -17.519, -17.254, -16.921, -16.527, -16.035,
    -15.501, -14.965, -14.469, -14.038, -17.730, -17.526, -17.261, -16.927,
    -16.533, -16.039, -15.505, -14.967, -14.471, -14.040, -17.737, -17.532,
    -17.267, -16.933, -16.539, -16.044, -15.509, -14.970, -14.473, -14.041,
    -17.744, -17.539, -17.273, -16.939, -16.545, -16.049, -15.512, -14.972,
    -14.474, -14.042, -17.751, -17.545, -17.280, -16.945, -16.551, -16.054,
    -15.516, -14.975, -14.476, -14.044, -17.757, -17.552, -17.286, -16.952,
    -16.557, -16.058, -15.519, -14.977, -14.478, -14.045, -17.764, -17.559,
    -17.293, -16.958, -16.563, -16.063, -15.523, -14.980, -14.480, -14.046,
    -17.771, -17.565, -17.299, -16.964, -16.568, -16.068, -15.527, -14.982,
    -14.481, -14.048, -17.777, -17.572, -17.305, -16.970, -16.574, -16.072,
    -15.530, -14.985, -14.483, -14.049, -17.784, -17.578, -17.312, -16.976,
    -16.580, -16.077, -15.534, -14.987, -14.485, -14.051, -17.791, -17.585,
    -17.318, -16.982, -16.586, -16.082, -15.537, -14.990, -14.487, -14.052,
    -17.797, -17.591, -17.324, -16.988, -16.592, -16.086, -15.541, -14.992,
    -14.488, -14.053, -17.804, -17.598, -17.331, -16.995, -16.598, -16.091,
    -15.544, -14.994, -14.490, -14.055, -17.811, -17.605, -17.337, -17.001,
    -16.604, -16.096, -15.548, -14.997, -14.492, -14.056, -17.817, -17.611,
    -17.343, -17.007, -16.610, -16.100, -15.551, -14.999, -14.494, -14.058,
    -17.824, -17.618, -17.350, -17.013, -16.616, -16.105, -15.555, -15.002,
    -14.496, -14.059, -17.831, -17.624, -17.356, -17.019, -16.622, -16.110,
    -15.558, -15.004, -14.497, -14.060, -17.837, -17.631, -17.362, -17.025,
    -16.627, -16.114, -15.562, -15.007, -14.499, -14.062, -17.844, -17.637,
    -17.369, -17.031, -16.633, -16.119, -15.565, -15.009, -14.501, -14.063,
    -17.851, -17.644, -17.375, -17.037, -16.639, -16.124, -15.569, -15.012,
    -14.503, -14.065, -17.857, -17.650, -17.381, -17.043, -16.645, -16.128,
    -15.572, -15.014, -14.505, -14.066, -17.864, -17.657, -17.388, -17.049,
    -16.651, -16.133, -15.576, -15.017, -14.507, -14.068, -17.871, -17.663,
    -17.394, -17.056, -16.657, -16.137, -15.579, -15.019, -14.509, -14.069,
    -17.877, -17.670, -17.400, -17.062, -16.663, -16.142, -15.583, -15.022,
    -14.510, -14.071, -17.884, -17.676, -17.407, -17.068, -16.668, -16.147,
    -15.586, -15.025, -14.512, -14.072, -17.891, -17.683, -17.413, -17.074,
    -16.674, -16.151, -15.590, -15.027, -14.514, -14.074, -17.897, -17.690,
    -17.419, -17.080, -16.680, -16.156, -15.593, -15.030, -14.516, -14.075,
    -17.904, -17.696, -17.426, -17.086, -16.686, -16.160, -15.597, -15.032,
    -14.518, -14.077, -17.910, -17.703, -17.432, -17.092, -16.692, -16.165,
    -15.600, -15.035, -14.520, -14.078, -17.917, -17.709, -17.438, -17.098,
    -16.698, -16.170, -15.604, -15.038, -14.522, -14.080, -17.924, -17.716,
    -17.445, -17.104, -16.703, -16.174, -15.607, -15.040, -14.524, -14.081,
    -17.930, -17.722, -17.451, -17.110, -16.709, -16.179, -15.611, -15.043,
    -14.526, -14.083, -17.937, -17.728, -17.457, -17.116, -16.715, -16.184,
    -15.615, -15.046, -14.528, -14.085, -17.943, -17.735, -17.463, -17.122,
    -16.721, -16.188, -15.618, -15.048, -14.530, -14.086, -17.950, -17.741,
    -17.470, -17.128, -16.727, -16.193, -15.622, -15.051, -14.532, -14.088,
    -17.957, -17.748, -17.476, -17.134, -16.732, -16.198, -15.626, -15.054,
    -14.534, -14.090, -17.963, -17.754, -17.482, -17.140, -16.738, -16.202,
    -15.629, -15.057, -14.536, -14.091, -17.970, -17.761, -17.488, -17.146,
    -16.744, -16.207, -15.633, -15.059, -14.538, -14.093, -17.976, -17.767,
    -17.495, -17.153, -16.750, -16.212, -15.637, -15.062, -14.541, -14.095,
    -17.983, -17.774, -17.501, -17.159, -16.756, -16.216, -15.641, -15.065,
    -14.543, -14.096, -17.990, -17.780, -17.507, -17.165, -16.761, -16.221,
    -15.644, -15.068, -14.545, -14.098, -17.996, -17.787, -17.514, -17.171,
    -16.767, -16.226, -15.648, -15.071, -14.547, -14.100, -18.003, -17.793,
    -17.520, -17.177, -16.773, -16.231, -15.652, -15.074, -14.549, -14.102,
    -18.009, -17.800, -17.526, -17.183, -16.779, -16.236, -15.656, -15.077,
    -14.552, -14.103, -18.016, -17.806, -17.532, -17.189, -16.785, -16.240,
    -15.660, -15.080, -14.554, -14.105, -18.022, -17.813, -17.539, -17.195,
    -16.790, -16.245, -15.664, -15.083, -14.556, -14.107, -18.029, -17.819,
    -17.545, -17.201, -16.796, -16.250, -15.667, -15.086, -14.559, -14.109,
    -18.035, -17.825, -17.551, -17.207, -16.802, -16.255, -15.671, -15.089,
    -14.561, -14.111, -18.042, -17.832, -17.557, -17.213, -16.808, -16.260,
    -15.675, -15.092, -14.563, -14.113, -18.049, -17.838, -17.563, -17.219,
    -16.813, -16.265, -15.679, -15.095, -14.566, -14.114, -18.055, -17.845,
    -17.570, -17.225, -16.819, -16.270, -15.683, -15.098, -14.568, -14.116,
    -18.062, -17.851, -17.576, -17.231, -16.825, -16.274, -15.687, -15.101,
    -14.571, -14.118, -18.068, -17.858, -17.582, -17.237, -16.831, -16.279,
    -15.692, -15.104, -14.573, -14.120, -18.075, -17.864, -17.588, -17.243,
    -16.836, -16.284, -15.696, -15.108, -14.576, -14.122, -18.081, -17.870,
    -17.595, -17.249, -16.842, -16.289, -15.700, -15.111, -14.578, -14.124,
    -18.088, -17.877, -17.601, -17.255, -16.848, -16.294, -15.704, -15.114,
    -14.581, -14.126, -18.094, -17.883, -17.607, -17.261, -16.853, -16.299,
    -15.708, -15.118, -14.583, -14.128, -18.101, -17.890, -17.613, -17.267,
    -16.859, -16.304, -15.712, -15.121, -14.586, -14.130, -18.107, -17.896,
    -17.619, -17.273, -16.865, -16.309, -15.717, -15.124, -14.589, -14.132,
    -18.114, -17.903, -17.626, -17.279, -16.871, -16.315, -15.721, -15.128,
    -14.591, -14.134, -18.120, -17.909, -17.632, -17.284, -16.876, -16.320,
    -15.725, -15.131, -14.594, -14.137, -18.127, -17.915, -17.638, -17.290,
    -16.882, -16.325, -15.730, -15.135, -14.597, -14.139, -18.133, -17.922,
    -17.644, -17.296, -16.888, -16.330, -15.734, -15.138, -14.599, -14.141,
    -18.140, -17.928, -17.650, -17.302, -16.893, -16.335, -15.739, -15.142,
    -14.602, -14.143, -18.146, -17.935, -17.657, -17.308, -16.899, -16.340,
    -15.743, -15.145, -14.605, -14.145, -18.153, -17.941, -17.663, -17.314,
    -16.905, -16.345, -15.747, -15.149, -14.608, -14.147, -18.159, -17.947,
    -17.669, -17.320, -16.911, -16.351, -15.752, -15.153, -14.611, -14.150,
    -18.166, -17.954, -17.675, -17.326, -16.916, -16.356, -15.756, -15.156,
    -14.614, -14.152, -18.172, -17.960, -17.681, -17.332, -16.922, -16.361,
    -15.761, -15.160, -14.617, -14.154, -18.179, -17.966, -17.688, -17.338,
    -16.928, -16.366, -15.766, -15.164, -14.620, -14.157, -18.185, -17.973,
    -17.694, -17.344, -16.933, -16.372, -15.770, -15.168, -14.623, -14.159,
    -18.192, -17.979, -17.700, -17.350, -16.939, -16.377, -15.775, -15.171,
    -14.626, -14.161, -18.198, -17.986, -17.706, -17.356, -16.945, -16.382,
    -15.780, -15.175, -14.629, -14.164, -18.205, -17.992, -17.712, -17.362,
    -16.950, -16.388, -15.784, -15.179, -14.632, -14.166, -18.211, -17.998,
    -17.718, -17.368, -16.956, -16.393, -15.789, -15.183, -14.635, -14.168,
    -18.218, -18.005, -17.725, -17.374, -16.962, -16.398, -15.794, -15.187,
    -14.638, -14.171, -18.224, -18.011, -17.731, -17.380, -16.967, -16.404,
    -15.799, -15.191, -14.641, -14.173, -18.231, -18.017, -17.737, -17.386,
    -16.973, -16.409, -15.803, -15.195, -14.644, -14.176, -18.237, -18.024,
    -17.743, -17.391, -16.979, -16.415, -15.808, -15.199, -14.648, -14.178,
    -18.244, -18.030, -17.749, -17.397, -16.984, -16.420, -15.813, -15.203,
    -14.651, -14.181, -18.250, -18.036, -17.755, -17.403, -16.990, -16.425,
    -15.818, -15.207, -14.654, -14.183, -18.257, -18.043, -17.762, -17.409,
    -16.996, -16.431, -15.823, -15.211, -14.657, -14.186, -18.263, -18.049,
    -17.768, -17.415, -17.001, -16.436, -15.828, -15.216, -14.661, -14.188,
    -18.269, -18.056, -17.774, -17.421, -17.007, -16.442, -15.833, -15.220,
    -14.664, -14.191, -18.276, -18.062, -17.780, -17.427, -17.013, -16.447,
    -15.838, -15.224, -14.667, -14.194, -18.282, -18.068, -17.786, -17.433,
    -17.018, -16.453, -15.843, -15.228, -14.671, -14.196, -18.289, -18.075,
    -17.792, -17.439, -17.024, -16.458, -15.848, -15.233, -14.674, -14.199,
    -18.295, -18.081, -17.798, -17.445, -17.030, -16.464, -15.853, -15.237,
    -14.678, -14.202, -18.302, -18.087, -17.805, -17.450, -17.035, -16.469,
    -15.858, -15.241, -14.681, -14.204, -18.308, -18.094, -17.811, -17.456,
    -17.041, -16.475, -15.863, -15.245, -14.685, -14.207, -18.315, -18.100,
    -17.817, -17.462, -17.046, -16.480, -15.868, -15.250, -14.688, -14.210,
    -18.321, -18.106, -17.823, -17.468, -17.052, -16.486, -15.873, -15.254,
    -14.692, -14.212, -18.327, -18.112, -17.829, -17.474, -17.058, -16.492,
    -15.878, -15.259, -14.695, -14.215, -18.334, -18.119, -17.835, -17.480,
    -17.063, -16.497, -15.884, -15.263, -14.699, -14.218, -18.340, -18.125,
    -17.841, -17.486, -17.069, -16.503, -15.889, -15.268, -14.703, -14.221,
    -18.347, -18.131, -17.847, -17.492, -17.075, -16.508, -15.894, -15.272,
    -14.706, -14.224, -18.353, -18.138, -17.854, -17.498, -17.080, -16.514,
    -15.899, -15.277, -14.710, -14.227, -18.360, -18.144, -17.860, -17.503,
    -17.086, -16.519, -15.905, -15.281, -14.714, -14.229, -18.366, -18.150,
    -17.866, -17.509, -17.091, -16.525, -15.910, -15.286, -14.717, -14.232,
    -18.372, -18.157, -17.872, -17.515, -17.097, -16.531, -15.915, -15.290,
    -14.721, -14.235, -18.379, -18.163, -17.878, -17.521, -17.103, -16.536,
    -15.920, -15.295, -14.725, -14.238, -18.385, -18.169, -17.884, -17.527,
    -17.108, -16.542, -15.926, -15.300, -14.729, -14.241, -18.392, -18.176,
    -17.890, -17.533, -17.114, -16.547, -15.931, -15.304, -14.732, -14.244,
    -18.398, -18.182, -17.896, -17.539, -17.119, -16.553, -15.936, -15.309,
    -14.736, -14.247, -18.404, -18.188, -17.902, -17.544, -17.125, -16.559,
    -15.942, -15.314, -14.740, -14.250, -18.411, -18.194, -17.908, -17.550,
    -17.131, -16.564, -15.947, -15.318, -14.744, -14.253, -18.417, -18.201,
    -17.915, -17.556, -17.136, -16.570, -15.952, -15.323, -14.748, -14.256,
    -18.424, -18.207, -17.921, -17.562, -17.142, -16.576, -15.958, -15.328,
    -14.752, -14.259, -18.430, -18.213, -17.927, -17.568, -17.147, -16.581,
    -15.963, -15.333, -14.756, -14.262, -18.436, -18.220, -17.933, -17.574,
    -17.153, -16.587, -15.968, -15.337, -14.760, -14.265, -18.443, -18.226,
    -17.939, -17.580, -17.159, -16.592, -15.974, -15.342, -14.764, -14.268,
    -18.449, -18.232, -17.945, -17.585, -17.164, -16.598, -15.979, -15.347,
    -14.768, -14.271, -18.456, -18.238, -17.951, -17.591, -17.170, -16.604,
    -15.985, -15.352, -14.772, -14.275, -18.462, -18.245, -17.957, -17.597,
    -17.175, -16.609, -15.990, -15.357, -14.776, -14.278, -18.468, -18.251,
    -17.963, -17.603, -17.181, -16.615, -15.995, -15.362, -14.780, -14.281,
    -18.475, -18.257, -17.969, -17.609, -17.186, -16.621, -16.001, -15.366,
    -14.784, -14.284, -18.481, -18.264, -17.975, -17.615, -17.192, -16.626,
    -16.006, -15.371, -14.788, -14.287, -18.487, -18.270, -17.981, -17.620,
    -17.198, -16.632, -16.012, -15.376, -14.792, -14.290, -18.494, -18.276,
    -17.987, -17.626, -17.203, -16.637, -16.017, -15.381, -14.796, -14.294,
    -18.500, -18.282, -17.994, -17.632, -17.209, -16.643, -16.023, -15.386,
    -14.800, -14.297, -18.507, -18.289, -18.000, -17.638, -17.214, -16.649,
    -16.028, -15.391, -14.804, -14.300, -18.513, -18.295, -18.006, -17.644,
    -17.220, -16.654, -16.034, -15.396, -14.808, -14.303, -18.519, -18.301,
    -18.012, -17.650, -17.225, -16.660, -16.039, -15.401, -14.813, -14.307,
    -18.526, -18.307, -18.018, -17.655, -17.231, -16.666, -16.045, -15.406,
    -14.817, -14.310, -18.532, -18.314, -18.024, -17.661, -17.236, -16.671,
    -16.050, -15.411, -14.821, -14.313, -18.538, -18.320, -18.030, -17.667,
    -17.242, -16.677, -16.056, -15.416, -14.825, -14.317, -18.545, -18.326,
    -18.036, -17.673, -17.247, -16.683, -16.061, -15.421, -14.829, -14.320,
    -18.551, -18.332, -18.042, -17.679, -17.253, -16.688, -16.067, -15.426,
    -14.834, -14.323, -18.557, -18.339, -18.048, -17.684, -17.259, -16.694,
    -16.072, -15.431, -14.838, -14.327, -18.564, -18.345, -18.054, -17.690,
    -17.264, -16.699, -16.078, -15.436, -14.842, -14.330, -18.570, -18.351,
    -18.060, -17.696, -17.270, -16.705, -16.083, -15.441, -14.846, -14.333,
    -18.577, -18.357, -18.066, -17.702, -17.275, -16.711, -16.089, -15.446,
    -14.851, -14.337, -18.583, -18.364, -18.072, -17.708, -17.281, -16.716,
    -16.094, -15.451, -14.855, -14.340, -18.589, -18.370, -18.078, -17.713,
    -17.286, -16.722, -16.100, -15.456, -14.859, -14.344, -18.596, -18.376,
    -18.084, -17.719, -17.292, -16.728, -16.105, -15.461, -14.864, -14.347,
    -18.602, -18.382, -18.090, -17.725, -17.297, -16.733, -16.111, -15.466,
    -14.868, -14.350, -18.608, -18.388, -18.096, -17.731, -17.303, -16.739,
    -16.116, -15.471, -14.872, -14.354, -18.615, -18.395, -18.102, -17.737,
    -17.308, -16.744, -16.122, -15.477, -14.876, -14.357, -18.621, -18.401,
    -18.108, -17.742, -17.314, -16.750, -16.127, -15.482, -14.881, -14.361,
    -18.627, -18.407, -18.114, -17.748, -17.319, -16.756, -16.133, -15.487,
    -14.885, -14.364, -18.634, -18.413, -18.121, -17.754, -17.325, -16.761,
    -16.138, -15.492, -14.890, -14.368, -18.640, -18.420, -18.127, -17.760,
    -17.330, -16.767, -16.144, -15.497, -14.894, -14.371, -18.646, -18.426,
    -18.133, -17.766, -17.336, -16.772, -16.149, -15.502, -14.898, -14.375,
    -18.653, -18.432, -18.139, -17.771, -17.341, -16.778, -16.155, -15.507,
    -14.903, -14.378, -18.659, -18.438, -18.145, -17.777, -17.347, -16.783,
    -16.160, -15.512, -14.907, -14.382, -18.665, -18.444, -18.151, -17.783,
    -17.352, -16.789, -16.166, -15.517, -14.911, -14.385, -18.672, -18.451,
    -18.157, -17.789, -17.358, -16.795, -16.171, -15.523, -14.916, -14.389,
    -18.678, -18.457, -18.163, -17.794, -17.364, -16.800, -16.177, -15.528,
    -14.920, -14.392, -18.684, -18.463, -18.169, -17.800, -17.369, -16.806,
    -16.183, -15.533, -14.925, -14.396, -18.690, -18.469, -18.175, -17.806,
    -17.375, -16.811, -16.188, -15.538, -14.929, -14.400, -18.697, -18.476,
    -18.181, -17.812, -17.380, -16.817, -16.194, -15.543, -14.934, -14.403,
    -18.703, -18.482, -18.187, -17.817, -17.386, -16.822, -16.199, -15.548,
    -14.938, -14.407, -18.709, -18.488, -18.193, -17.823, -17.391, -16.828,
    -16.205, -15.553, -14.942, -14.410, -18.716, -18.494, -18.199, -17.829,
    -17.396, -16.834, -16.210, -15.558, -14.947, -14.414, -18.722, -18.500,
    -18.205, -17.835, -17.402, -16.839, -16.216, -15.564, -14.951, -14.417,
    -18.728, -18.507, -18.211, -17.841, -17.407, -16.845, -16.221, -15.569,
    -14.956, -14.421, -18.735, -18.513, -18.217, -17.846, -17.413, -16.850,
    -16.227, -15.574, -14.960, -14.425, -18.741, -18.519, -18.223, -17.852,
    -17.418, -16.856, -16.232, -15.579, -14.965, -14.428, -18.747, -18.525,
    -18.229, -17.858, -17.424, -16.861, -16.238, -15.584, -14.969, -14.432,
    -18.754, -18.531, -18.235, -17.864, -17.429, -16.867, -16.243, -15.589,
    -14.974, -14.435, -18.760, -18.537, -18.241, -17.869, -17.435, -16.872,
    -16.249, -15.594, -14.978, -14.439, -18.766, -18.544, -18.247, -17.875,
    -17.440, -16.878, -16.254, -15.600, -14.983, -14.443, -18.772, -18.550,
    -18.253, -17.881, -17.446, -16.883, -16.260, -15.605, -14.987, -14.446,
    -18.779, -18.556, -18.259, -17.887, -17.451, -16.889, -16.265, -15.610,
    -14.992, -14.450, -18.785, -18.562, -18.265, -17.892, -17.457, -16.894,
    -16.270, -15.615, -14.996, -14.454, -18.791, -18.568, -18.271, -17.898,
    -17.462, -16.900, -16.276, -15.620, -15.001, -14.457, -18.798, -18.575,
    -18.277, -17.904, -17.468, -16.905, -16.281, -15.625, -15.005, -14.461,
    -18.804, -18.581, -18.283, -17.910, -17.473, -16.911, -16.287, -15.630,
    -15.010, -14.465, -18.810, -18.587, -18.289, -17.915, -17.479, -16.916,
    -16.292, -15.635, -15.014, -14.468, -18.816, -18.593, -18.295, -17.921,
    -17.484, -16.922, -16.298, -15.641, -15.019, -14.472, -18.823, -18.599,
    -18.301, -17.927, -17.490, -16.927, -16.303, -15.646, -15.023, -14.476,
    -18.829, -18.605, -18.307, -17.932, -17.495, -16.933, -16.309, -15.651,
    -15.028, -14.479, -18.835, -18.612, -18.313, -17.938, -17.501, -16.938,
    -16.314, -15.656, -15.032, -14.483, -18.842, -18.618, -18.318, -17.944,
    -17.506, -16.944, -16.320, -15.661, -15.037, -14.487, -18.848, -18.624,
    -18.324, -17.950, -17.511, -16.949, -16.325, -15.666, -15.041, -14.490,
    -18.854, -18.630, -18.330, -17.955, -17.517, -16.955, -16.330, -15.671,
    -15.046, -14.494, -18.860, -18.636, -18.336, -17.961, -17.522, -16.960,
    -16.336, -15.676, -15.050, -14.498, -18.867, -18.642, -18.342, -17.967,
    -17.528, -16.966, -16.341, -15.681, -15.054, -14.501, -18.873, -18.649,
    -18.348, -17.973, -17.533, -16.971, -16.347, -15.687, -15.059, -14.505,
    -18.879, -18.655, -18.354, -17.978, -17.539, -16.977, -16.352, -15.692,
    -15.063, -14.509, -18.885, -18.661, -18.360, -17.984, -17.544, -16.982,
    -16.358, -15.697, -15.068, -14.513, -18.892, -18.667, -18.366, -17.990,
    -17.550, -16.987, -16.363, -15.702, -15.072, -14.516, -18.898, -18.673,
    -18.372, -17.995, -17.555, -16.993, -16.368, -15.707, -15.077, -14.520,
    -18.904, -18.679, -18.378, -18.001, -17.561, -16.998, -16.374, -15.712,
    -15.081, -14.524, -18.910, -18.685, -18.384, -18.007, -17.566, -17.004,
    -16.379, -15.717, -15.086, -14.527, -18.917, -18.692, -18.390, -18.013,
    -17.571, -17.009, -16.385, -15.722, -15.090, -14.531, -18.923, -18.698,
    -18.396, -18.018, -17.577, -17.015, -16.390, -15.727, -15.095, -14.535,
    -18.929, -18.704, -18.402, -18.024, -17.582, -17.020, -16.395, -15.732,
    -15.099, -14.539, -18.935, -18.710, -18.408, -18.030, -17.588, -17.025,
    -16.401, -15.737, -15.104, -14.542, -18.942, -18.716, -18.414, -18.035,
    -17.593, -17.031, -16.406, -15.742, -15.108, -14.546, -18.948, -18.722,
    -18.420, -18.041, -17.599, -17.036, -16.412, -15.747, -15.113, -14.550,
    -18.954, -18.729, -18.426, -18.047, -17.604, -17.042, -16.417, -15.753,
    -15.117, -14.553, -18.960, -18.735, -18.432, -18.053, -17.610, -17.047,
    -16.422, -15.758, -15.122, -14.557, -18.967, -18.741, -18.438, -18.058,
    -17.615, -17.052, -16.428, -15.763, -15.126, -14.561, -18.973, -18.747,
    -18.444, -18.064, -17.620, -17.058, -16.433, -15.768, -15.131, -14.565,
    -18.979, -18.753, -18.450, -18.070, -17.626, -17.063, -16.438, -15.773,
    -15.135, -14.568, -18.985, -18.759, -18.455, -18.075, -17.631, -17.069,
    -16.444, -15.778, -15.140, -14.572, -18.992, -18.765, -18.461, -18.081,
    -17.637, -17.074, -16.449, -15.783, -15.144, -14.576, -18.998, -18.771,
    -18.467, -18.087, -17.642, -17.079, -16.454, -15.788, -15.149, -14.580,
    -19.004, -18.778, -18.473, -18.092, -17.647, -17.085, -16.460, -15.793,
    -15.153, -14.583, -19.010, -18.784, -18.479, -18.098, -17.653, -17.090,
    -16.465, -15.798, -15.158, -14.587, -19.017, -18.790, -18.485, -18.104,
    -17.658, -17.096, -16.470, -15.803, -15.162, -14.591, -19.023, -18.796,
    -18.491, -18.110, -17.664, -17.101, -16.476, -15.808, -15.167, -14.595,
    -19.029, -18.802, -18.497, -18.115, -17.669, -17.106, -16.481, -15.813,
    -15.171, -14.598, -19.035, -18.808, -18.503, -18.121, -17.675, -17.112,
    -16.486, -15.818, -15.176, -14.602, -19.042, -18.814, -18.509, -18.127,
    -17.680, -17.117, -16.492, -15.823, -15.180, -14.606, -19.048, -18.820,
    -18.515, -18.132, -17.685, -17.122, -16.497, -15.828, -15.184, -14.610,
    -19.054, -18.827, -18.521, -18.138, -17.691, -17.128, -16.502, -15.833,
    -15.189, -14.613, -19.060, -18.833, -18.527, -18.144, -17.696, -17.133,
    -16.508, -15.838, -15.193, -14.617, -19.066, -18.839, -18.533, -18.149,
    -17.702, -17.138, -16.513, -15.843, -15.198, -14.621, -19.073, -18.845,
    -18.539, -18.155, -17.707, -17.144, -16.518, -15.848, -15.202, -14.624,
    -19.079, -18.851, -18.544, -18.161, -17.712, -17.149, -16.524, -15.853,
    -15.207, -14.628, -19.085, -18.857, -18.550, -18.166, -17.718, -17.154,
    -16.529, -15.858, -15.211, -14.632, -19.091, -18.863, -18.556, -18.172,
    -17.723, -17.160, -16.534, -15.863, -15.216, -14.636, -19.098, -18.869,
    -18.562, -18.178, -17.729, -17.165, -16.539, -15.868, -15.220, -14.639,
    -19.104, -18.875, -18.568, -18.183, -17.734, -17.170, -16.545, -15.873,
    -15.225, -14.643, -19.110, -18.882, -18.574, -18.189, -17.739, -17.176,
    -16.550, -15.877, -15.229, -14.647, -19.116, -18.888, -18.580, -18.195,
    -17.745, -17.181, -16.555, -15.882, -15.233, -14.651, -19.122, -18.894,
    -18.586, -18.200, -17.750, -17.186, -16.560, -15.887, -15.238, -14.654,
    -19.129, -18.900, -18.592, -18.206, -17.756, -17.192, -16.566, -15.892,
    -15.242, -14.658, -19.135, -18.906, -18.598, -18.212, -17.761, -17.197,
    -16.571, -15.897, -15.247, -14.662, -19.141, -18.912, -18.604, -18.217,
    -17.766, -17.202, -16.576, -15.902, -15.251, -14.666, -19.147, -18.918,
    -18.610, -18.223, -17.772, -17.208, -16.581, -15.907, -15.256, -14.669,
    -19.153, -18.924, -18.615, -18.229, -17.777, -17.213, -16.587, -15.912,
    -15.260, -14.673, -19.160, -18.930, -18.621, -18.234, -17.783, -17.218,
    -16.592, -15.917, -15.264, -14.677, -19.166, -18.936, -18.627, -18.240,
    -17.788, -17.224, -16.597, -15.922, -15.269, -14.681, -19.172, -18.943,
    -18.633, -18.246, -17.793, -17.229, -16.602, -15.927, -15.273, -14.684,
    -19.178, -18.949, -18.639, -18.251, -17.799, -17.234, -16.608, -15.932,
    -15.278, -14.688, -19.184, -18.955, -18.645, -18.257, -17.804, -17.239,
    -16.613, -15.937, -15.282, -14.692, -19.191, -18.961, -18.651, -18.263,
    -17.810, -17.245, -16.618, -15.941, -15.287, -14.696, -19.197, -18.967,
    -18.657, -18.268, -17.815, -17.250, -16.623, -15.946, -15.291, -14.699,
    -19.203, -18.973, -18.663, -18.274, -17.820, -17.255, -16.628, -15.951,
    -15.295, -14.703, -19.209, -18.979, -18.669, -18.280, -17.826, -17.261,
    -16.634, -15.956, -15.300, -14.707, -19.215, -18.985, -18.674, -18.285,
    -17.831, -17.266, -16.639, -15.961, -15.304, -14.711, -19.222, -18.991,
    -18.680, -18.291, -17.836, -17.271, -16.644, -15.966, -15.309, -14.715,
    -19.228, -18.997, -18.686, -18.297, -17.842, -17.276, -16.649, -15.971,
    -15.313, -14.718, -19.234, -19.003, -18.692, -18.302, -17.847, -17.282,
    -16.654, -15.976, -15.317, -14.722, -19.240, -19.010, -18.698, -18.308,
    -17.853, -17.287, -16.660, -15.981, -15.322, -14.726, -19.246, -19.016,
    -18.704, -18.314, -17.858, -17.292, -16.665, -15.985, -15.326, -14.730,
    -19.253, -19.022, -18.710, -18.319, -17.863, -17.298, -16.670, -15.990,
    -15.331, -14.733, -19.259, -19.028, -18.716, -18.325, -17.869, -17.303,
    -16.675, -15.995, -15.335, -14.737, -19.265, -19.034, -18.722, -18.331,
    -17.874, -17.308, -16.680, -16.000, -15.339, -14.741, -19.271, -19.040,
    -18.727, -18.336, -17.879, -17.313, -16.686, -16.005, -15.344, -14.745,
    -19.277, -19.046, -18.733, -18.342, -17.885, -17.319, -16.691, -16.010,
    -15.348, -14.748, -19.284, -19.052, -18.739, -18.347, -17.890, -17.324,
    -16.696, -16.015, -15.353, -14.752, -19.290, -19.058, -18.745, -18.353,
    -17.896, -17.329, -16.701, -16.019, -15.357, -14.756, -19.296, -19.064,
    -18.751, -18.359, -17.901, -17.334, -16.706, -16.024, -15.361, -14.760,
    -19.302, -19.070, -18.757, -18.364, -17.906, -17.340, -16.711, -16.029,
    -15.366, -14.763, -19.308, -19.076, -18.763, -18.370, -17.912, -17.345,
    -16.716, -16.034, -15.370, -14.767, -19.314, -19.082, -18.769, -18.376,
    -17.917, -17.350, -16.722, -16.039, -15.375, -14.771, -19.321, -19.089,
    -18.774, -18.381, -17.922, -17.355, -16.727, -16.044, -15.379, -14.775,
    -19.327, -19.095, -18.780, -18.387, -17.928, -17.360, -16.732, -16.049,
    -15.383, -14.779, -19.333, -19.101, -18.786, -18.393, -17.933, -17.366,
    -16.737, -16.053, -15.388, -14.782, -19.339, -19.107, -18.792, -18.398,
    -17.938, -17.371, -16.742, -16.058, -15.392, -14.786, -19.345, -19.113,
    -18.798, -18.404, -17.944, -17.376, -16.747, -16.063, -15.397, -14.790,
    -19.351, -19.119, -18.804, -18.409, -17.949, -17.381, -16.752, -16.068,
    -15.401, -14.794, -19.358, -19.125, -18.810, -18.415, -17.955, -17.387,
    -16.757, -16.073, -15.405, -14.797, -19.364, -19.131, -18.816, -18.421,
    -17.960, -17.392, -16.763, -16.078, -15.410, -14.801, -19.370, -19.137,
    -18.821, -18.426, -17.965, -17.397, -16.768, -16.082, -15.414, -14.805,
    -19.376, -19.143, -18.827, -18.432, -17.971, -17.402, -16.773, -16.087,
    -15.419, -14.809, -19.382, -19.149, -18.833, -18.438, -17.976, -17.407,
    -16.778, -16.092, -15.423, -14.812, -19.388, -19.155, -18.839, -18.443,
    -17.981, -17.413, -16.783, -16.097, -15.427, -14.816, -19.395, -19.161,
    -18.845, -18.449, -17.987, -17.418, -16.788, -16.102, -15.432, -14.820,
    -19.401, -19.167, -18.851, -18.455, -17.992, -17.423, -16.793, -16.107,
    -15.436, -14.824, -19.407, -19.173, -18.857, -18.460, -17.997, -17.428,
    -16.798, -16.111, -15.441, -14.828, -19.413, -19.179, -18.863, -18.466,
    -18.003, -17.433, -16.803, -16.116, -15.445, -14.831, -19.419, -19.185,
    -18.868, -18.471, -18.008, -17.439, -16.808, -16.121, -15.449, -14.835,
    -19.425, -19.191, -18.874, -18.477, -18.013, -17.444, -16.814, -16.126,
    -15.454, -14.839, -19.432, -19.198, -18.880, -18.483, -18.019, -17.449,
    -16.819, -16.131, -15.458, -14.843, -19.438, -19.204, -18.886, -18.488,
    -18.024, -17.454, -16.824, -16.136, -15.463, -14.847, -19.444, -19.210,
    -18.892, -18.494, -18.029, -17.459, -16.829, -16.140, -15.467, -14.850,
    -19.450, -19.216, -18.898, -18.499, -18.035, -17.465, -16.834, -16.145,
    -15.471, -14.854, -19.456, -19.222, -18.904, -18.505, -18.040, -17.470,
    -16.839, -16.150, -15.476, -14.858, -19.462, -19.228, -18.909, -18.511,
    -18.045, -17.475, -16.844, -16.155, -15.480, -14.862, -19.469, -19.234,
    -18.915, -18.516, -18.051, -17.480, -16.849, -16.160, -15.485, -14.866,
    -19.475, -19.240, -18.921, -18.522, -18.056, -17.485, -16.854, -16.165,
    -15.489, -14.869, -19.481, -19.246, -18.927, -18.528, -18.061, -17.491,
    -16.859, -16.169, -15.494, -14.873, -19.487, -19.252, -18.933, -18.533,
    -18.067, -17.496, -16.864, -16.174, -15.498, -14.877, -19.493, -19.258,
    -18.939, -18.539, -18.072, -17.501, -16.869, -16.179, -15.502, -14.881,
    -19.499, -19.264, -18.944, -18.544, -18.077, -17.506, -16.874, -16.184,
    -15.507, -14.885, -19.505, -19.270, -18.950, -18.550, -18.083, -17.511,
    -16.879, -16.189, -15.511, -14.888, -19.512, -19.276, -18.956, -18.556,
    -18.088, -17.516, -16.884, -16.193, -15.516, -14.892, -19.518, -19.282,
    -18.962, -18.561, -18.093, -17.522, -16.889, -16.198, -15.520, -14.896,
    -19.524, -19.288, -18.968, -18.567, -18.099, -17.527, -16.894, -16.203,
    -15.524, -14.900, -19.530, -19.294, -18.974, -18.572, -18.104, -17.532,
    -16.899, -16.208, -15.529, -14.904, -19.536, -19.300, -18.980, -18.578,
    -18.109, -17.537, -16.905, -16.213, -15.533, -14.907, -19.542, -19.306,
    -18.985, -18.584, -18.115, -17.542, -16.910, -16.218, -15.538, -14.911,
    -19.548, -19.312, -18.991, -18.589, -18.120, -17.547, -16.915, -16.222,
    -15.542, -14.915, -19.555, -19.318, -18.997, -18.595, -18.125, -17.553,
    -16.920, -16.227, -15.546, -14.919, -19.561, -19.324, -19.003, -18.600,
    -18.131, -17.558, -16.925, -16.232, -15.551, -14.923, -19.567, -19.330,
    -19.009, -18.606, -18.136, -17.563, -16.930, -16.237, -15.555, -14.926,
    -19.573, -19.336, -19.015, -18.612, -18.141, -17.568, -16.935, -16.242,
    -15.560, -14.930, -19.579, -19.342, -19.020, -18.617, -18.147, -17.573,
    -16.940, -16.246, -15.564, -14.934, -19.585, -19.348, -19.026, -18.623,
    -18.152, -17.578, -16.945, -16.251, -15.569, -14.938, -19.591, -19.354,
    -19.032, -18.628, -18.157, -17.583, -16.950, -16.256, -15.573, -14.942,
    -19.597, -19.360, -19.038, -18.634, -18.163, -17.589, -16.955, -16.261,
    -15.577, -14.945, -19.604, -19.367, -19.044, -18.640, -18.168, -17.594,
    -16.960, -16.266, -15.582, -14.949, -19.610, -19.373, -19.050, -18.645,
    -18.173, -17.599, -16.965, -16.271, -15.586, -14.953, -19.616, -19.379,
    -19.055, -18.651, -18.179, -17.604, -16.970, -16.275, -15.591, -14.957,
    -19.622, -19.385, -19.061, -18.656, -18.184, -17.609, -16.975, -16.280,
    -15.595, -14.961, -19.628, -19.391, -19.067, -18.662, -18.189, -17.614,
    -16.980, -16.285, -15.599, -14.965, -19.634, -19.397, -19.073, -18.667,
    -18.195, -17.619, -16.985, -16.290, -15.604, -14.968, -19.640, -19.403,
    -19.079, -18.673, -18.200, -17.625, -16.990, -16.295, -15.608, -14.972,
    -19.647, -19.409, -19.085, -18.679, -18.205, -17.630, -16.995, -16.299,
    -15.613, -14.976, -19.653, -19.415, -19.090, -18.684, -18.211, -17.635,
    -17.000, -16.304, -15.617, -14.980, -19.659, -19.421, -19.096, -18.690,
    -18.216, -17.640, -17.005, -16.309, -15.622, -14.984, -19.665, -19.427,
    -19.102, -18.695, -18.221, -17.645, -17.009, -16.314, -15.626, -14.987,
    -19.671, -19.433, -19.108, -18.701, -18.226, -17.650, -17.014, -16.319,
    -15.630, -14.991, -19.677, -19.439, -19.114, -18.707, -18.232, -17.655,
    -17.019, -16.323, -15.635, -14.995, -19.683, -19.445, -19.120, -18.712,
    -18.237, -17.660, -17.024, -16.328, -15.639, -14.999, -19.689, -19.451,
    -19.125, -18.718, -18.242, -17.666, -17.029, -16.333, -15.644, -15.003,
    -19.695, -19.457, -19.131, -18.723, -18.248, -17.671, -17.034, -16.338,
    -15.648, -15.007, -19.702, -19.463, -19.137, -18.729, -18.253, -17.676,
    -17.039, -16.343, -15.653, -15.010, -19.708, -19.469, -19.143, -18.734,
    -18.258, -17.681, -17.044, -16.347, -15.657, -15.014, -19.714, -19.475,
    -19.149, -18.740, -18.264, -17.686, -17.049, -16.352, -15.661, -15.018,
    -19.720, -19.481, -19.154, -18.746, -18.269, -17.691, -17.054, -16.357,
    -15.666, -15.022, -19.726, -19.487, -19.160, -18.751, -18.274, -17.696,
    -17.059, -16.362, -15.670, -15.026, -19.732, -19.493, -19.166, -18.757,
    -18.279, -17.701, -17.064, -16.367, -15.675, -15.029, -19.738, -19.499,
    -19.172, -18.762, -18.285, -17.706, -17.069, -16.371, -15.679, -15.033,
    -19.744, -19.505, -19.178, -18.768, -18.290, -17.712, -17.074, -16.376,
    -15.683, -15.037, -19.750, -19.511, -19.184, -18.774, -18.295, -17.717,
    -17.079, -16.381, -15.688, -15.041, -19.757, -19.517, -19.189, -18.779,
    -18.301, -17.722, -17.084, -16.386, -15.692, -15.045, -19.763, -19.523,
    -19.195, -18.785, -18.306, -17.727, -17.089, -16.391, -15.697, -15.048,
    -19.769, -19.529, -19.201, -18.790, -18.311, -17.732, -17.094, -16.395,
    -15.701, -15.052, -19.775, -19.535, -19.207, -18.796, -18.317, -17.737,
    -17.099, -16.400, -15.706, -15.056, -19.781, -19.541, -19.213, -18.801,
    -18.322, -17.742, -17.104, -16.405, -15.710, -15.060, -19.787, -19.547,
    -19.218, -18.807, -18.327, -17.747, -17.108, -16.410, -15.714, -15.064,
    -19.793, -19.553, -19.224, -18.812, -18.332, -17.752, -17.113, -16.415,
    -15.719, -15.068, -19.799, -19.559, -19.230, -18.818, -18.338, -17.757,
    -17.118, -16.419, -15.723, -15.071, -19.805, -19.565, -19.236, -18.824,
    -18.343, -17.763, -17.123, -16.424, -15.728, -15.075, -19.812, -19.571,
    -19.242, -18.829, -18.348, -17.768, -17.128, -16.429, -15.732, -15.079,
    -19.818, -19.577, -19.247, -18.835, -18.354, -17.773, -17.133, -16.434,
    -15.736, -15.083, -19.824, -19.583, -19.253, -18.840, -18.359, -17.778,
    -17.138, -16.438, -15.741, -15.087, -19.830, -19.589, -19.259, -18.846,
    -18.364, -17.783, -17.143, -16.443, -15.745, -15.090, -19.836, -19.595,
    -19.265, -18.851, -18.369, -17.788, -17.148, -16.448, -15.750, -15.094,
    -19.842, -19.601, -19.271, -18.857, -18.375, -17.793, -17.153, -16.453,
    -15.754, -15.098, -19.848, -19.607, -19.276, -18.863, -18.380, -17.798,
    -17.158, -16.458, -15.758, -15.102, -19.854, -19.613, -19.282, -18.868,
    -18.385, -17.803, -17.162, -16.462, -15.763, -15.106, -19.860, -19.619,
    -19.288, -18.874, -18.391, -17.808, -17.167, -16.467, -15.767, -15.109,
    -19.866, -19.625, -19.294, -18.879, -18.396, -17.813, -17.172, -16.472,
    -15.772, -15.113, -19.873, -19.631, -19.300, -18.885, -18.401, -17.818,
    -17.177, -16.477, -15.776, -15.117, -19.879, -19.637, -19.305, -18.890,
    -18.406, -17.824, -17.182, -16.481, -15.780, -15.121, -19.885, -19.643,
    -19.311, -18.896, -18.412, -17.829, -17.187, -16.486, -15.785, -15.125,
    -19.891, -19.649, -19.317, -18.901, -18.417, -17.834, -17.192, -16.491,
    -15.789, -15.128, -19.897, -19.655, -19.323, -18.907, -18.422, -17.839,
    -17.197, -16.496, -15.793, -15.132, -19.903, -19.661, -19.329, -18.913,
    -18.428, -17.844, -17.202, -16.500, -15.798, -15.136, -19.909, -19.667,
    -19.334, -18.918, -18.433, -17.849, -17.207, -16.505, -15.802, -15.140,
    -19.915, -19.673, -19.340, -18.924, -18.438, -17.854, -17.211, -16.510,
    -15.807, -15.143, -19.921, -19.679, -19.346, -18.929, -18.443, -17.859,
    -17.216, -16.515, -15.811, -15.147, -19.927, -19.685, -19.352, -18.935,
    -18.449, -17.864, -17.221, -16.519, -15.815, -15.151, -19.933, -19.690,
    -19.358, -18.940, -18.454, -17.869, -17.226, -16.524, -15.820, -15.155,
    -19.940, -19.696, -19.363, -18.946, -18.459, -17.874, -17.231, -16.529,
    -15.824, -15.159, -19.946, -19.702, -19.369, -18.951, -18.465, -17.879,
    -17.236, -16.534, -15.828, -15.162, -19.952, -19.708, -19.375, -18.957,
    -18.470, -17.884, -17.241, -16.538, -15.833, -15.166, -19.958, -19.714,
    -19.381, -18.963, -18.475, -17.889, -17.246, -16.543, -15.837, -15.170,
    -19.964, -19.720, -19.387, -18.968, -18.480, -17.894, -17.250, -16.548,
    -15.841, -15.174, -19.970, -19.726, -19.392, -18.974, -18.486, -17.900,
    -17.255, -16.553, -15.846, -15.177, -19.976, -19.732, -19.398, -18.979,
    -18.491, -17.905, -17.260, -16.557, -15.850, -15.181, -19.982, -19.738,
    -19.404, -18.985, -18.496, -17.910, -17.265, -16.562, -15.855, -15.185,
    -19.988, -19.744, -19.410, -18.990, -18.501, -17.915, -17.270, -16.567,
    -15.859, -15.189, -19.994, -19.750, -19.415, -18.996, -18.507, -17.920,
    -17.275, -16.572, -15.863, -15.192, -20.000, -19.756, -19.421, -19.001,
    -18.512, -17.925, -17.280, -16.576, -15.868, -15.196, -20.006, -19.762,
    -19.427, -19.007, -18.517, -17.930, -17.284, -16.581, -15.872, -15.200,
    -20.013, -19.768, -19.433, -19.012, -18.523, -17.935, -17.289, -16.586,
    -15.876, -15.204, -20.019, -19.774, -19.439, -19.018, -18.528, -17.940,
    -17.294, -16.590, -15.881, -15.207, -20.025, -19.780, -19.444, -19.024,
    -18.533, -17.945, -17.299, -16.595, -15.885, -15.211, -20.031, -19.786,
    -19.450, -19.029, -18.538, -17.950, -17.304, -16.600, -15.889, -15.215,
    -20.037, -19.792, -19.456, -19.035, -18.544, -17.955, -17.309, -16.605,
    -15.894, -15.219, -20.043, -19.798, -19.462, -19.040, -18.549, -17.960,
    -17.314, -16.609, -15.898, -15.222, -20.049, -19.804, -19.468, -19.046,
    -18.554, -17.965, -17.318, -16.614, -15.902, -15.226, -20.055, -19.810,
    -19.473, -19.051, -18.559, -17.970, -17.323, -16.619, -15.906, -15.230,
    -20.061, -19.816, -19.479, -19.057, -18.565, -17.975, -17.328, -16.623,
    -15.911, -15.233, -20.067, -19.822, -19.485, -19.062, -18.570, -17.980,
    -17.333, -16.628, -15.915, -15.237, -20.073, -19.828, -19.491, -19.068,
    -18.575, -17.985, -17.338, -16.633, -15.919, -15.241, -20.079, -19.834,
    -19.496, -19.073, -18.580, -17.990, -17.343, -16.638, -15.924, -15.245,
    -20.085, -19.840, -19.502, -19.079, -18.586, -17.995, -17.347, -16.642,
    -15.928, -15.248, -20.092, -19.846, -19.508, -19.084, -18.591, -18.000,
    -17.352, -16.647, -15.932, -15.252, -20.098, -19.852, -19.514, -19.090,
    -18.596, -18.005, -17.357, -16.652, -15.937, -15.256, -20.104, -19.858,
    -19.519, -19.096, -18.601, -18.011, -17.362, -16.656, -15.941, -15.260,
    -20.110, -19.864, -19.525, -19.101, -18.607, -18.016, -17.367, -16.661,
    -15.945, -15.263, -20.116, -19.870, -19.531, -19.107, -18.612, -18.021,
    -17.372, -16.666, -15.949, -15.267, -20.122, -19.875, -19.537, -19.112,
    -18.617, -18.026, -17.376, -16.670, -15.954, -15.271, -20.128, -19.881,
    -19.543, -19.118, -18.622, -18.031, -17.381, -16.675, -15.958, -15.274,
    -20.134, -19.887, -19.548, -19.123, -18.628, -18.036, -17.386, -16.680,
    -15.962, -15.278, -20.140, -19.893, -19.554, -19.129, -18.633, -18.041,
    -17.391, -16.685, -15.967, -15.282, -20.146, -19.899, -19.560, -19.134,
    -18.638, -18.046, -17.396, -16.689, -15.971, -15.285, -20.152, -19.905,
    -19.566, -19.140, -18.644, -18.051, -17.401, -16.694, -15.975, -15.289,
    -20.158, -19.911, -19.571, -19.145, -18.649, -18.056, -17.405, -16.699,
    -15.979, -15.293, -20.164, -19.917, -19.577, -19.151, -18.654, -18.061,
    -17.410, -16.703, -15.984, -15.296, -20.170, -19.923, -19.583, -19.156,
    -18.659, -18.066, -17.415, -16.708, -15.988, -15.300, -20.176, -19.929,
    -19.589, -19.162, -18.665, -18.071, -17.420, -16.713, -15.992, -15.304,
    -20.183, -19.935, -19.594, -19.167, -18.670, -18.076, -17.425, -16.717,
    -15.996, -15.308, -20.189, -19.941, -19.600, -19.173, -18.675, -18.081,
    -17.429, -16.722, -16.001, -15.311, -20.195, -19.947, -19.606, -19.178,
    -18.680, -18.086, -17.434, -16.727, -16.005, -15.315, -20.201, -19.953,
    -19.612, -19.184, -18.686, -18.091, -17.439, -16.731, -16.009, -15.319,
    -20.207, -19.959, -19.618, -19.189, -18.691, -18.096, -17.444, -16.736,
    -16.013, -15.322, -20.213, -19.965, -19.623, -19.195, -18.696, -18.101,
    -17.449, -16.741, -16.018, -15.326, -20.219, -19.971, -19.629, -19.201,
    -18.701, -18.106, -17.453, -16.745, -16.022, -15.330, -20.225, -19.977,
    -19.635, -19.206, -18.707, -18.111, -17.458, -16.750, -16.026, -15.333,
    -20.231, -19.983, -19.641, -19.212, -18.712, -18.116, -17.463, -16.755,
    -16.030, -15.337, -20.237, -19.989, -19.646, -19.217, -18.717, -18.121,
    -17.468, -16.759, -16.035, -15.341, -20.243, -19.995, -19.652, -19.223,
    -18.722, -18.126, -17.473, -16.764, -16.039, -15.344, -20.249, -20.000,
    -19.658, -19.228, -18.727, -18.131, -17.477, -16.768, -16.043, -15.348,
    -20.255, -20.006, -19.664, -19.234, -18.733, -18.136, -17.482, -16.773,
    -16.047, -15.352, -20.261, -20.012, -19.669, -19.239, -18.738, -18.141,
    -17.487, -16.778, -16.052, -15.355, -20.267, -20.018, -19.675, -19.245,
    -18.743, -18.146, -17.492, -16.782, -16.056, -15.359, -20.273, -20.024,
    -19.681, -19.250, -18.748, -18.151, -17.497, -16.787, -16.060, -15.362,
    -20.279, -20.030, -19.687, -19.256, -18.754, -18.156, -17.501, -16.792,
    -16.064, -15.366, -20.285, -20.036, -19.692, -19.261, -18.759, -18.161,
    -17.506, -16.796, -16.068, -15.370, -20.292, -20.042, -19.698, -19.267,
    -18.764, -18.166, -17.511, -16.801, -16.073, -15.373, -20.298, -20.048,
    -19.704, -19.272, -18.769, -18.171, -17.516, -16.806, -16.077, -15.377,
    -20.304, -20.054, -19.710, -19.278, -18.775, -18.176, -17.520, -16.810,
    -16.081, -15.381, -20.310, -20.060, -19.715, -19.283, -18.780, -18.181,
    -17.525, -16.815, -16.085, -15.384, -20.316, -20.066, -19.721, -19.289,
    -18.785, -18.186, -17.530, -16.819, -16.089, -15.388, -20.322, -20.072,
    -19.727, -19.294, -18.790, -18.191, -17.535, -16.824, -16.094, -15.392,
    -20.328, -20.078, -19.733, -19.300, -18.796, -18.196, -17.540, -16.829,
    -16.098, -15.395, -20.334, -20.084, -19.738, -19.305, -18.801, -18.201,
    -17.544, -16.833, -16.102, -15.399, -20.340, -20.090, -19.744, -19.311,
    -18.806, -18.206, -17.549, -16.838, -16.106, -15.403, -20.346, -20.095,
    -19.750, -19.316, -18.811, -18.211, -17.554, -16.843, -16.110, -15.406,
    -20.352, -20.101, -19.756, -19.322, -18.817, -18.216, -17.559, -16.847,
    -16.115, -15.410, -20.358, -20.107, -19.761, -19.327, -18.822, -18.221,
    -17.563, -16.852, -16.119, -15.414, -20.364, -20.113, -19.767, -19.333,
    -18.827, -18.226, -17.568, -16.856, -16.123, -15.417, -20.370, -20.119,
    -19.773, -19.338, -18.832, -18.231, -17.573, -16.861, -16.127, -15.421,
    -20.376, -20.125, -19.779, -19.344, -18.838, -18.236, -17.578, -16.866,
    -16.131, -15.425, -20.382, -20.131, -19.784, -19.349, -18.843, -18.241,
    -17.583, -16.870, -16.136, -15.428, -20.388, -20.137, -19.790, -19.355,
    -18.848, -18.246, -17.587, -16.875, -16.140, -15.432, -20.394, -20.143,
    -19.796, -19.360, -18.853, -18.251, -17.592, -16.879, -16.144, -15.435,
    -20.400, -20.149, -19.802, -19.366, -18.858, -18.256, -17.597, -16.884,
    -16.148, -15.439, -20.406, -20.155, -19.807, -19.371, -18.864, -18.261,
    -17.602, -16.889, -16.152, -15.443, -20.412, -20.161, -19.813, -19.377,
    -18.869, -18.266, -17.606, -16.893, -16.157, -15.446, -20.418, -20.167,
    -19.819, -19.382, -18.874, -18.271, -17.611, -16.898, -16.161, -15.450,
    -20.425, -20.173, -19.825, -19.388, -18.879, -18.276, -17.616, -16.902,
    -16.165, -15.454, -20.431, -20.179, -19.830, -19.393, -18.885, -18.281,
    -17.621, -16.907, -16.169, -15.457, -20.437, -20.184, -19.836, -19.399,
    -18.890, -18.286, -17.625, -16.912, -16.173, -15.461, -20.443, -20.190,
    -19.842, -19.404, -18.895, -18.291, -17.630, -16.916, -16.177, -15.465,
    -20.449, -20.196, -19.848, -19.410, -18.900, -18.296, -17.635, -16.921,
    -16.182, -15.468, -20.455, -20.202, -19.853, -19.415, -18.905, -18.301,
    -17.640, -16.925, -16.186, -15.472, -20.461, -20.208, -19.859, -19.421,
    -18.911, -18.306, -17.644, -16.930, -16.190, -15.476, -20.467, -20.214,
    -19.865, -19.426, -18.916, -18.311, -17.649, -16.935, -16.194, -15.479,
    -20.473, -20.220, -19.870, -19.432, -18.921, -18.316, -17.654, -16.939,
    -16.198, -15.483, -20.479, -20.226, -19.876, -19.437, -18.926, -18.321,
    -17.659, -16.944, -16.202, -15.487, -20.485, -20.232, -19.882, -19.443,
    -18.932, -18.326, -17.663, -16.948, -16.207, -15.490, -20.491, -20.238,
    -19.888, -19.448, -18.937, -18.331, -17.668, -16.953, -16.211, -15.494,
    -20.497, -20.244, -19.893, -19.454, -18.942, -18.336, -17.673, -16.957,
    -16.215, -15.498, -20.503, -20.250, -19.899, -19.459, -18.947, -18.341,
    -17.678, -16.962, -16.219, -15.501, -20.509, -20.256, -19.905, -19.465,
    -18.953, -18.346, -17.682, -16.967, -16.223, -15.505, -20.515, -20.261,
    -19.911, -19.470, -18.958, -18.351, -17.687, -16.971, -16.228, -15.509,
    -20.521, -20.267, -19.916, -19.476, -18.963, -18.356, -17.692, -16.976,
    -16.232, -15.512, -20.527, -20.273, -19.922, -19.481, -18.968, -18.361,
    -17.697, -16.980, -16.236, -15.516, -20.533, -20.279, -19.928, -19.487,
    -18.973, -18.366, -17.701, -16.985, -16.240, -15.520, -20.539, -20.285,
    -19.934, -19.492, -18.979, -18.371, -17.706, -16.989, -16.244, -15.523,
    -20.545, -20.291, -19.939, -19.498, -18.984, -18.375, -17.711, -16.994,
    -16.248, -15.527, -20.551, -20.297, -19.945, -19.503, -18.989, -18.380,
    -17.716, -16.999, -16.253, -15.531, -20.557, -20.303, -19.951, -19.509,
    -18.994, -18.385, -17.720, -17.003, -16.257, -15.534, -20.563, -20.309,
    -19.956, -19.514, -18.999, -18.390, -17.725, -17.008, -16.261, -15.538,
    -20.569, -20.315, -19.962, -19.520, -19.005, -18.395, -17.730, -17.012,
    -16.265, -15.542, -20.575, -20.321, -19.968, -19.525, -19.010, -18.400,
    -17.734, -17.017, -16.269, -15.545, -20.581, -20.327, -19.974, -19.531,
    -19.015, -18.405, -17.739, -17.021, -16.274, -15.549, -20.587, -20.332,
    -19.979, -19.536, -19.020, -18.410, -17.744, -17.026, -16.278, -15.553,
    -20.593, -20.338, -19.985, -19.542, -19.026, -18.415, -17.749, -17.030,
    -16.282, -15.556, -20.599, -20.344, -19.991, -19.547, -19.031, -18.420,
    -17.753, -17.035, -16.286, -15.560, -20.605, -20.350, -19.997, -19.553,
    -19.036, -18.425, -17.758, -17.040, -16.290, -15.564, -20.612, -20.356,
    -20.002, -19.558, -19.041, -18.430, -17.763, -17.044, -16.294, -15.568,
    -20.618, -20.362, -20.008, -19.564, -19.046, -18.435, -17.768, -17.049,
    -16.299, -15.571, -20.624, -20.368, -20.014, -19.569, -19.052, -18.440,
    -17.772, -17.053, -16.303, -15.575, -20.630, -20.374, -20.019, -19.575,
    -19.057, -18.445, -17.777, -17.058, -16.307, -15.579, -20.636, -20.380,
    -20.025, -19.580, -19.062, -18.450, -17.782, -17.062, -16.311, -15.582,
    -20.642, -20.386, -20.031, -19.586, -19.067, -18.455, -17.786, -17.067,
    -16.315, -15.586, -20.648, -20.392, -20.037, -19.591, -19.072, -18.460,
    -17.791, -17.071, -16.320, -15.590, -20.654, -20.398, -20.042, -19.597,
    -19.078, -18.465, -17.796, -17.076, -16.324, -15.594, -20.660, -20.403,
    -20.048, -19.602, -19.083, -18.470, -17.801, -17.080, -16.328, -15.597,
    -20.666, -20.409, -20.054, -19.608, -19.088, -18.475, -17.805, -17.085,
    -16.332, -15.601, -20.672, -20.415, -20.060, -19.613, -19.093, -18.480,
    -17.810, -17.089, -16.336, -15.605, -20.678, -20.421, -20.065, -19.619,
    -19.099, -18.485, -17.815, -17.094, -16.341, -15.609, -20.684, -20.427,
    -20.071, -19.624, -19.104, -18.490, -17.819, -17.099, -16.345, -15.612,
    -20.690, -20.433, -20.077, -19.630, -19.109, -18.495, -17.824, -17.103,
    -16.349, -15.616, -20.696, -20.439, -20.082, -19.635, -19.114, -18.500,
    -17.829, -17.108, -16.353, -15.620, -20.702, -20.445, -20.088, -19.641,
    -19.119, -18.504, -17.834, -17.112, -16.357, -15.624, -20.708, -20.451,
    -20.094, -19.646, -19.125, -18.509, -17.838, -17.117, -16.362, -15.627,
    -20.714, -20.457, -20.100, -19.651, -19.130, -18.514, -17.843, -17.121,
    -16.366, -15.631, -20.720, -20.462, -20.105, -19.657, -19.135, -18.519,
    -17.848, -17.126, -16.370, -15.635, -20.726, -20.468, -20.111, -19.662,
    -19.140, -18.524, -17.852, -17.130, -16.374, -15.639, -20.732, -20.474,
    -20.117, -19.668, -19.145, -18.529, -17.857, -17.135, -16.378, -15.642,
    -20.738, -20.480, -20.122, -19.673, -19.151, -18.534, -17.862, -17.139,
    -16.383, -15.646, -20.744, -20.486, -20.128, -19.679, -19.156, -18.539,
    -17.867, -17.144, -16.387, -15.650, -20.750, -20.492, -20.134, -19.684,
    -19.161, -18.544, -17.871, -17.148, -16.391, -15.654, -20.756, -20.498,
    -20.140, -19.690, -19.166, -18.549, -17.876, -17.153, -16.395, -15.658,
    -20.762, -20.504, -20.145, -19.695, -19.171, -18.554, -17.881, -17.157,
    -16.399, -15.661, -20.768, -20.510, -20.151, -19.701, -19.177, -18.559,
    -17.885, -17.162, -16.404, -15.665, -20.774, -20.516, -20.157, -19.706,
    -19.182, -18.564, -17.890, -17.166, -16.408, -15.669, -20.780, -20.521,
    -20.162, -19.712, -19.187, -18.569, -17.895, -17.171, -16.412, -15.673,
    -20.786, -20.527, -20.168, -19.717, -19.192, -18.574, -17.900, -17.175,
    -16.416, -15.677, -20.792, -20.533, -20.174, -19.723, -19.197, -18.579,
    -17.904, -17.180, -16.421, -15.680, -20.798, -20.539, -20.180, -19.728,
    -19.203, -18.584, -17.909, -17.184, -16.425, -15.684, -20.804, -20.545,
    -20.185, -19.734, -19.208, -18.589, -17.914, -17.189, -16.429, -15.688,
    -20.810, -20.551, -20.191, -19.739, -19.213, -18.594, -17.918, -17.193,
    -16.433, -15.692, -20.816, -20.557, -20.197, -19.745, -19.218, -18.598,
    -17.923, -17.198, -16.437, -15.696, -20.822, -20.563, -20.202, -19.750,
    -19.223, -18.603, -17.928, -17.202, -16.442, -15.699, -20.828, -20.569,
    -20.208, -19.756, -19.229, -18.608, -17.932, -17.207, -16.446, -15.703,
    -20.834, -20.575, -20.214, -19.761, -19.234, -18.613, -17.937, -17.211,
    -16.450, -15.707, -20.840, -20.580, -20.220, -19.766, -19.239, -18.618,
    -17.942, -17.216, -16.454, -15.711, -20.846, -20.586, -20.225, -19.772,
    -19.244, -18.623, -17.946, -17.220, -16.459, -15.715, -20.852, -20.592,
    -20.231, -19.777, -19.249, -18.628, -17.951, -17.225, -16.463, -15.719,
    -20.858, -20.598, -20.237, -19.783, -19.255, -18.633, -17.956, -17.229,
    -16.467, -15.722, -20.864, -20.604, -20.242, -19.788, -19.260, -18.638,
    -17.961, -17.234, -16.471, -15.726, -20.870, -20.610, -20.248, -19.794,
    -19.265, -18.643, -17.965, -17.238, -16.476, -15.730, -20.876, -20.616,
    -20.254, -19.799, -19.270, -18.648, -17.970, -17.243, -16.480, -15.734,
    -20.882, -20.622, -20.259, -19.805, -19.275, -18.653, -17.975, -17.247,
    -16.484, -15.738, -20.888, -20.628, -20.265, -19.810, -19.281, -18.658,
    -17.979, -17.252, -16.488, -15.742, -20.894, -20.633, -20.271, -19.816,
    -19.286, -18.663, -17.984, -17.256, -16.493, -15.746, -20.900, -20.639,
    -20.277, -19.821, -19.291, -18.668, -17.989, -17.261, -16.497, -15.750,
    -20.906, -20.645, -20.282, -19.827, -19.296, -18.673, -17.993, -17.265,
    -16.501, -15.753, -20.912, -20.651, -20.288, -19.832, -19.301, -18.677,
    -17.998, -17.270, -16.505, -15.757, -20.918, -20.657, -20.294, -19.838,
    -19.307, -18.682, -18.003, -17.274, -16.510, -15.761, -20.924, -20.663,
    -20.299, -19.843, -19.312, -18.687, -18.007, -17.278, -16.514, -15.765,
    -20.930, -20.669, -20.305, -19.848, -19.317, -18.692, -18.012, -17.283,
    -16.518, -15.769, -20.936, -20.675, -20.311, -19.854, -19.322, -18.697,
    -18.017, -17.287, -16.522, -15.773, -20.942, -20.681, -20.317, -19.859,
    -19.327, -18.702, -18.021, -17.292, -16.526, -15.777, -20.948, -20.686,
    -20.322, -19.865, -19.333, -18.707, -18.026, -17.296, -16.531, -15.781,
    -20.954, -20.692, -20.328, -19.870, -19.338, -18.712, -18.031, -17.301,
    -16.535, -15.784, -20.960, -20.698, -20.334, -19.876, -19.343, -18.717,
    -18.035, -17.305, -16.539, -15.788, -20.966, -20.704, -20.339, -19.881,
    -19.348, -18.722, -18.040, -17.310, -16.543, -15.792, -20.972, -20.710,
    -20.345, -19.887, -19.353, -18.727, -18.045, -17.314, -16.548, -15.796,
    -20.978, -20.716, -20.351, -19.892, -19.358, -18.732, -18.049, -17.319,
    -16.552, -15.800, -20.984, -20.722, -20.356, -19.898, -19.364, -18.737,
    -18.054, -17.323, -16.556, -15.804, -20.990, -20.728, -20.362, -19.903,
    -19.369, -18.742, -18.059, -17.328, -16.560, -15.808, -20.996, -20.734,
    -20.368, -19.909, -19.374, -18.746, -18.063, -17.332, -16.565, -15.812,
    -21.002, -20.739, -20.374, -19.914, -19.379, -18.751, -18.068, -17.337,
    -16.569, -15.816, -21.008, -20.745, -20.379, -19.919, -19.384, -18.756,
    -18.073, -17.341, -16.573, -15.820, -21.014, -20.751, -20.385, -19.925,
    -19.390, -18.761, -18.077, -17.345, -16.577, -15.824, -21.020, -20.757,
    -20.391, -19.930, -19.395, -18.766, -18.082, -17.350, -16.582, -15.828,
    -21.026, -20.763, -20.396, -19.936, -19.400, -18.771, -18.087, -17.354,
    -16.586, -15.831, -21.032, -20.769, -20.402, -19.941, -19.405, -18.776,
    -18.091, -17.359, -16.590, -15.835, -21.038, -20.775, -20.408, -19.947,
    -19.410, -18.781, -18.096, -17.363, -16.595, -15.839, -21.044, -20.781,
    -20.413, -19.952, -19.416, -18.786, -18.101, -17.368, -16.599, -15.843,
    -21.050, -20.787, -20.419, -19.958, -19.421, -18.791, -18.105, -17.372,
    -16.603, -15.847, -21.056, -20.792, -20.425, -19.963, -19.426, -18.796,
    -18.110, -17.377, -16.607, -15.851, -21.062, -20.798, -20.430, -19.969,
    -19.431, -18.801, -18.115, -17.381, -16.612, -15.855, -21.068, -20.804,
    -20.436, -19.974, -19.436, -18.806, -18.119, -17.386, -16.616, -15.859,
    -21.074, -20.810, -20.442, -19.979, -19.441, -18.810, -18.124, -17.390,
    -16.620, -15.863, -21.080, -20.816, -20.448, -19.985, -19.447, -18.815,
    -18.129, -17.394, -16.624, -15.867, -21.086, -20.822, -20.453, -19.990,
    -19.452, -18.820, -18.133, -17.399, -16.629, -15.871, -21.092, -20.828,
    -20.459, -19.996, -19.457, -18.825, -18.138, -17.403, -16.633, -15.875,
    -21.098, -20.834, -20.465, -20.001, -19.462, -18.830, -18.143, -17.408,
    -16.637, -15.879, -21.104, -20.839, -20.470, -20.007, -19.467, -18.835,
    -18.147, -17.412, -16.641, -15.883, -21.110, -20.845, -20.476, -20.012,
    -19.473, -18.840, -18.152, -17.417, -16.646, -15.887, -21.116, -20.851,
    -20.482, -20.018, -19.478, -18.845, -18.157, -17.421, -16.650, -15.891,
    -21.122, -20.857, -20.487, -20.023, -19.483, -18.850, -18.161, -17.426,
    -16.654, -15.895, -21.128, -20.863, -20.493, -20.029, -19.488, -18.855,
    -18.166, -17.430, -16.658, -15.899, -21.134, -20.869, -20.499, -20.034,
    -19.493, -18.860, -18.171, -17.434, -16.663, -15.903, -21.140, -20.875,
    -20.504, -20.039, -19.498, -18.865, -18.175, -17.439, -16.667, -15.907,
    -21.146, -20.881, -20.510, -20.045, -19.504, -18.869, -18.180, -17.443,
    -16.671, -15.911, -21.152, -20.886, -20.516, -20.050, -19.509, -18.874,
    -18.185, -17.448, -16.675, -15.915, -21.158, -20.892, -20.522, -20.056,
    -19.514, -18.879, -18.189, -17.452, -16.680, -15.919, -21.164, -20.898,
    -20.527, -20.061, -19.519, -18.884, -18.194, -17.457, -16.684, -15.922,
    -21.170, -20.904, -20.533, -20.067, -19.524, -18.889, -18.199, -17.461,
    -16.688, -15.926, -21.176, -20.910, -20.539, -20.072, -19.529, -18.894,
    -18.203, -17.465, -16.692, -15.930, -21.182, -20.916, -20.544, -20.078,
    -19.535, -18.899, -18.208, -17.470, -16.697, -15.934, -21.188, -20.922,
    -20.550, -20.083, -19.540, -18.904, -18.213, -17.474, -16.701, -15.938,
    -21.194, -20.928, -20.556, -20.089, -19.545, -18.909, -18.217, -17.479,
    -16.705, -15.942, -21.200, -20.933, -20.561, -20.094, -19.550, -18.914,
    -18.222, -17.483, -16.709, -15.946, -21.206, -20.939, -20.567, -20.099,
    -19.555, -18.919, -18.227, -17.488, -16.714, -15.950, -21.212, -20.945,
    -20.573, -20.105, -19.561, -18.923, -18.231, -17.492, -16.718, -15.954,
    -21.218, -20.951, -20.578, -20.110, -19.566, -18.928, -18.236, -17.496,
    -16.722, -15.958, -21.224, -20.957, -20.584, -20.116, -19.571, -18.933,
    -18.240, -17.501, -16.726, -15.962, -21.230, -20.963, -20.590, -20.121,
    -19.576, -18.938, -18.245, -17.505, -16.731, -15.966, -21.236, -20.969,
    -20.595, -20.127, -19.581, -18.943, -18.250, -17.510, -16.735, -15.970,
    -21.242, -20.974, -20.601, -20.132, -19.586, -18.948, -18.254, -17.514,
    -16.739, -15.974, -21.248, -20.980, -20.607, -20.138, -19.592, -18.953,
    -18.259, -17.519, -16.743, -15.978, -21.254, -20.986, -20.612, -20.143,
    -19.597, -18.958, -18.264, -17.523, -16.748, -15.982, -21.260, -20.992,
    -20.618, -20.148, -19.602, -18.963, -18.268, -17.527, -16.752, -15.986,
    -21.266, -20.998, -20.624, -20.154, -19.607, -18.968, -18.273, -17.532,
    -16.756, -15.990, -21.272, -21.004, -20.629, -20.159, -19.612, -18.973,
    -18.278, -17.536, -16.760, -15.994, -21.278, -21.010, -20.635, -20.165,
    -19.617, -18.977, -18.282, -17.541, -16.765, -15.998, -21.284, -21.016,
    -20.641, -20.170, -19.623, -18.982, -18.287, -17.545, -16.769, -16.002,
    -21.290, -21.021, -20.647, -20.176, -19.628, -18.987, -18.292, -17.549,
    -16.773, -16.006, -21.296, -21.027, -20.652, -20.181, -19.633, -18.992,
    -18.296, -17.554, -16.777, -16.010, -21.302, -21.033, -20.658, -20.187,
    -19.638, -18.997, -18.301, -17.558, -16.782, -16.014, -21.307, -21.039,
    -20.664, -20.192, -19.643, -19.002, -18.305, -17.563, -16.786, -16.018,
    -21.313, -21.045, -20.669, -20.197, -19.648, -19.007, -18.310, -17.567,
    -16.790, -16.022, -21.319, -21.051, -20.675, -20.203, -19.654, -19.012,
    -18.315, -17.571, -16.794, -16.026, -21.325, -21.057, -20.681, -20.208,
    -19.659, -19.017, -18.319, -17.576, -16.799, -16.030, -21.331, -21.062,
    -20.686, -20.214, -19.664, -19.022, -18.324, -17.580, -16.803, -16.034,
    -21.337, -21.068, -20.692, -20.219, -19.669, -19.026, -18.329, -17.585,
    -16.807, -16.038, -21.343, -21.074, -20.698, -20.225, -19.674, -19.031,
    -18.333, -17.589, -16.811, -16.042, -21.349, -21.080, -20.703, -20.230,
    -19.679, -19.036, -18.338, -17.593, -16.815, -16.046, -21.355, -21.086,
    -20.709, -20.235, -19.685, -19.041, -18.343, -17.598, -16.820, -16.050,
    -21.361, -21.092, -20.715, -20.241, -19.690, -19.046, -18.347, -17.602,
    -16.824, -16.054, -21.367, -21.098, -20.720, -20.246, -19.695, -19.051,
    -18.352, -17.607, -16.828, -16.058, -21.373, -21.104, -20.726, -20.252,
    -19.700, -19.056, -18.356, -17.611, -16.832, -16.062, -21.379, -21.109,
    -20.732, -20.257, -19.705, -19.061, -18.361, -17.615, -16.837, -16.066,
    -21.385, -21.115, -20.737, -20.263, -19.710, -19.066, -18.366, -17.620,
    -16.841, -16.070, -21.391, -21.121, -20.743, -20.268, -19.716, -19.071,
    -18.370, -17.624, -16.845, -16.074, -21.397, -21.127, -20.749, -20.274,
    -19.721, -19.075, -18.375, -17.629, -16.849, -16.078, -21.403, -21.133,
    -20.754, -20.279, -19.726, -19.080, -18.380, -17.633, -16.854, -16.082,
    -21.409, -21.139, -20.760, -20.284, -19.731, -19.085, -18.384, -17.637,
    -16.858, -16.086, -21.415, -21.145, -20.766, -20.290, -19.736, -19.090,
    -18.389, -17.642, -16.862, -16.090, -21.421, -21.150, -20.771, -20.295,
    -19.741, -19.095, -18.394, -17.646, -16.866, -16.094, -21.427, -21.156,
    -20.777, -20.301, -19.747, -19.100, -18.398, -17.651, -16.870, -16.098,
    -21.433, -21.162, -20.783, -20.306, -19.752, -19.105, -18.403, -17.655,
    -16.875, -16.102, -21.439, -21.168, -20.788, -20.312, -19.757, -19.110,
    -18.407, -17.659, -16.879, -16.106, -21.445, -21.174, -20.794, -20.317,
    -19.762, -19.115, -18.412, -17.664, -16.883, -16.110, -21.451, -21.180,
    -20.800, -20.322, -19.767, -19.119, -18.417, -17.668, -16.887, -16.114,
    -21.457, -21.186, -20.805, -20.328, -19.772, -19.124, -18.421, -17.673,
    -16.892, -16.118, -21.463, -21.191, -20.811, -20.333, -19.778, -19.129,
    -18.426, -17.677, -16.896, -16.122, -21.469, -21.197, -20.817, -20.339,
    -19.783, -19.134, -18.431, -17.681, -16.900, -16.126, -21.475, -21.203,
    -20.822, -20.344, -19.788, -19.139, -18.435, -17.686, -16.904, -16.130,
    -21.481, -21.209, -20.828, -20.350, -19.793, -19.144, -18.440, -17.690,
    -16.908, -16.134, -21.487, -21.215, -20.834, -20.355, -19.798, -19.149,
    -18.444, -17.695, -16.913, -16.138, -21.493, -21.221, -20.839, -20.360,
    -19.803, -19.154, -18.449, -17.699, -16.917, -16.142, -21.499, -21.227,
    -20.845, -20.366, -19.809, -19.159, -18.454, -17.703, -16.921, -16.146,
    -21.505, -21.232, -20.851, -20.371, -19.814, -19.163, -18.458, -17.708,
    -16.925, -16.150, -21.511, -21.238, -20.856, -20.377, -19.819, -19.168,
    -18.463, -17.712, -16.930, -16.154, -21.517, -21.244, -20.862, -20.382,
    -19.824, -19.173, -18.467, -17.716, -16.934, -16.158, -21.523, -21.250,
    -20.868, -20.388, -19.829, -19.178, -18.472, -17.721, -16.938, -16.162,
    -21.528, -21.256, -20.873, -20.393, -19.834, -19.183, -18.477, -17.725,
    -16.942, -16.166, -21.534, -21.262, -20.879, -20.398, -19.840, -19.188,
    -18.481, -17.730, -16.946, -16.170, -21.540, -21.267, -20.885, -20.404,
    -19.845, -19.193, -18.486, -17.734, -16.951, -16.174, -21.546, -21.273,
    -20.890, -20.409, -19.850, -19.198, -18.491, -17.738, -16.955, -16.178,
    -21.552, -21.279, -20.896, -20.415, -19.855, -19.203, -18.495, -17.743,
    -16.959, -16.182, -21.558, -21.285, -20.902, -20.420, -19.860, -19.207,
    -18.500, -17.747, -16.963, -16.186, -21.564, -21.291, -20.907, -20.426,
    -19.865, -19.212, -18.504, -17.751, -16.967, -16.190, -21.570, -21.297,
    -20.913, -20.431, -19.870, -19.217, -18.509, -17.756, -16.972, -16.194,
    -21.576, -21.303, -20.919, -20.436, -19.876, -19.222, -18.514, -17.760,
    -16.976, -16.198, -21.582, -21.308, -20.924, -20.442, -19.881, -19.227,
    -18.518, -17.765, -16.980, -16.202, -21.588, -21.314, -20.930, -20.447,
    -19.886, -19.232, -18.523, -17.769, -16.984, -16.206, -21.594, -21.320,
    -20.936, -20.453, -19.891, -19.237, -18.527, -17.773, -16.988, -16.210,
    -21.600, -21.326, -20.941, -20.458, -19.896, -19.242, -18.532, -17.778,
    -16.993, -16.214, -21.606, -21.332, -20.947, -20.464, -19.901, -19.247,
    -18.537, -17.782, -16.997, -16.218, -21.612, -21.338, -20.953, -20.469,
    -19.907, -19.251, -18.541, -17.786, -17.001, -16.222, -21.618, -21.344,
    -20.958, -20.474, -19.912, -19.256, -18.546, -17.791, -17.005, -16.226,
    -21.624, -21.349, -20.964, -20.480, -19.917, -19.261, -18.551, -17.795,
    -17.009, -16.230, -21.630, -21.355, -20.970, -20.485, -19.922, -19.266,
    -18.555, -17.800, -17.014, -16.234, -21.636, -21.361, -20.975, -20.491,
    -19.927, -19.271, -18.560, -17.804, -17.018, -16.238, -21.642, -21.367,
    -20.981, -20.496, -19.932, -19.276, -18.564, -17.808, -17.022, -16.242,
    -21.648, -21.373, -20.987, -20.502, -19.937, -19.281, -18.569, -17.813,
    -17.026, -16.246, -21.654, -21.379, -20.992, -20.507, -19.943, -19.286,
    -18.574, -17.817, -17.030, -16.250, -21.660, -21.384, -20.998, -20.512,
    -19.948, -19.290, -18.578, -17.821, -17.035, -16.254, -21.666, -21.390,
    -21.004, -20.518, -19.953, -19.295, -18.583, -17.826, -17.039, -16.258,
    -21.672, -21.396, -21.009, -20.523, -19.958, -19.300, -18.587, -17.830,
    -17.043, -16.262, -21.678, -21.402, -21.015, -20.529, -19.963, -19.305,
    -18.592, -17.834, -17.047, -16.266, -21.684, -21.408, -21.021, -20.534,
    -19.968, -19.310, -18.597, -17.839, -17.051, -16.270, -21.689, -21.414,
    -21.026, -20.540, -19.974, -19.315, -18.601, -17.843, -17.055, -16.274,
    -21.695, -21.420, -21.032, -20.545, -19.979, -19.320, -18.606, -17.848,
    -17.060, -16.278, -21.701, -21.425, -21.038, -20.550, -19.984, -19.325,
    -18.610, -17.852, -17.064, -16.282, -21.707, -21.431, -21.043, -20.556,
    -19.989, -19.329, -18.615, -17.856, -17.068, -16.286, -21.713, -21.437,
    -21.049, -20.561, -19.994, -19.334, -18.620, -17.861, -17.072, -16.290,
    -21.719, -21.443, -21.055, -20.567, -19.999, -19.339, -18.624, -17.865,
    -17.076, -16.294, -21.725, -21.449, -21.060, -20.572, -20.004, -19.344,
    -18.629, -17.869, -17.081, -16.298, -21.731, -21.455, -21.066, -20.577,
    -20.010, -19.349, -18.633, -17.874, -17.085, -16.302, -21.737, -21.460,
    -21.072, -20.583, -20.015, -19.354, -18.638, -17.878, -17.089, -16.306,
    -21.743, -21.466, -21.077, -20.588, -20.020, -19.359, -18.643, -17.882,
    -17.093, -16.310, -21.749, -21.472, -21.083, -20.594, -20.025, -19.364,
    -18.647, -17.887, -17.097, -16.314, -21.755, -21.478, -21.088, -20.599,
    -20.030, -19.368, -18.652, -17.891, -17.101, -16.318, -21.761, -21.484,
    -21.094, -20.605, -20.035, -19.373, -18.656, -17.895, -17.106, -16.322,
    -21.767, -21.490, -21.100, -20.610, -20.040, -19.378, -18.661, -17.900,
    -17.110, -16.326, -21.773, -21.495, -21.105, -20.615, -20.046, -19.383,
    -18.666, -17.904, -17.114, -16.330, -21.779, -21.501, -21.111, -20.621,
    -20.051, -19.388, -18.670, -17.909, -17.118, -16.334, -21.785, -21.507,
    -21.117, -20.626, -20.056, -19.393, -18.675, -17.913, -17.122, -16.338,
    -21.791, -21.513, -21.122, -20.632, -20.061, -19.398, -18.679, -17.917,
    -17.126, -16.342, -21.797, -21.519, -21.128, -20.637, -20.066, -19.403,
    -18.684, -17.922, -17.131, -16.346, -21.803, -21.525, -21.134, -20.642,
    -20.071, -19.407, -18.689, -17.926, -17.135, -16.350, -21.809, -21.530,
    -21.139, -20.648, -20.076, -19.412, -18.693, -17.930, -17.139, -16.354,
    -21.815, -21.536, -21.145, -20.653, -20.082, -19.417, -18.698, -17.935,
    -17.143, -16.358, -21.821, -21.542, -21.151, -20.659, -20.087, -19.422,
    -18.702, -17.939, -17.147, -16.362, -21.826, -21.548, -21.156, -20.664,
    -20.092, -19.427, -18.707, -17.943, -17.151, -16.366, -21.832, -21.554,
    -21.162, -20.670, -20.097, -19.432, -18.712, -17.948, -17.156, -16.370,
    -21.838, -21.560, -21.168, -20.675, -20.102, -19.437, -18.716, -17.952,
    -17.160, -16.374, -21.844, -21.566, -21.173, -20.680, -20.107, -19.442,
    -18.721, -17.956, -17.164, -16.378, -21.850, -21.571, -21.179, -20.686,
    -20.112, -19.446, -18.725, -17.961, -17.168, -16.382, -21.856, -21.577,
    -21.185, -20.691, -20.118, -19.451, -18.730, -17.965, -17.172, -16.386,
    -21.862, -21.583, -21.190, -20.697, -20.123, -19.456, -18.735, -17.969,
    -17.176, -16.390, -21.868, -21.589, -21.196, -20.702, -20.128, -19.461,
    -18.739, -17.974, -17.180, -16.393, -21.874, -21.595, -21.202, -20.707,
    -20.133, -19.466, -18.744, -17.978, -17.185, -16.397, -21.880, -21.601,
    -21.207, -20.713, -20.138, -19.471, -18.748, -17.982, -17.189, -16.401,
    -21.886, -21.606, -21.213, -20.718, -20.143, -19.476, -18.753, -17.987,
    -17.193, -16.405, -21.892, -21.612, -21.218, -20.724, -20.148, -19.480,
    -18.758, -17.991, -17.197, -16.409, -21.898, -21.618, -21.224, -20.729,
    -20.154, -19.485, -18.762, -17.995, -17.201, -16.413, -21.904, -21.624,
    -21.230, -20.734, -20.159, -19.490, -18.767, -18.000, -17.205, -16.417,
    -21.910, -21.630, -21.235, -20.740, -20.164, -19.495, -18.771, -18.004,
    -17.210, -16.421, -21.916, -21.636, -21.241, -20.745, -20.169, -19.500,
    -18.776, -18.008, -17.214, -16.425, -21.922, -21.641, -21.247, -20.751,
    -20.174, -19.505, -18.781, -18.013, -17.218, -16.429, -21.928, -21.647,
    -21.252, -20.756, -20.179, -19.510, -18.785, -18.017, -17.222, -16.433,
    -21.934, -21.653, -21.258, -20.762, -20.184, -19.515, -18.790, -18.021,
    -17.226, -16.437, -21.940, -21.659, -21.264, -20.767, -20.190, -19.519,
    -18.794, -18.026, -17.230, -16.441, -21.945, -21.665, -21.269, -20.772,
    -20.195, -19.524, -18.799, -18.030, -17.234, -16.445, -21.951, -21.671,
    -21.275, -20.778, -20.200, -19.529, -18.803, -18.034, -17.239, -16.449,
    -21.957, -21.676, -21.281, -20.783, -20.205, -19.534, -18.808, -18.039,
    -17.243, -16.453, -21.963, -21.682, -21.286, -20.789, -20.210, -19.539,
    -18.813, -18.043, -17.247, -16.457, -21.969, -21.688, -21.292, -20.794,
    -20.215, -19.544, -18.817, -18.047, -17.251, -16.461, -21.975, -21.694,
    -21.297, -20.799, -20.220, -19.549, -18.822, -18.052, -17.255, -16.465,
    -21.981, -21.700, -21.303, -20.805, -20.226, -19.553, -18.826, -18.056,
    -17.259, -16.469, -21.987, -21.706, -21.309, -20.810, -20.231, -19.558,
    -18.831, -18.060, -17.263, -16.473, -21.993, -21.711, -21.314, -20.816,
    -20.236, -19.563, -18.836, -18.065, -17.267, -16.476, -21.999, -21.717,
    -21.320, -20.821, -20.241, -19.568, -18.840, -18.069, -17.272, -16.480,
    -22.005, -21.723, -21.326, -20.826, -20.246, -19.573, -18.845, -18.073,
    -17.276, -16.484, -22.011, -21.729, -21.331, -20.832, -20.251, -19.578,
    -18.849, -18.078, -17.280, -16.488, -22.017, -21.735, -21.337, -20.837,
    -20.256, -19.583, -18.854, -18.082, -17.284, -16.492, -22.023, -21.740,
    -21.343, -20.843, -20.261, -19.587, -18.859, -18.086, -17.288, -16.496,
    -22.029, -21.746, -21.348, -20.848, -20.267, -19.592, -18.863, -18.091,
    -17.292, -16.500, -22.035, -21.752, -21.354, -20.853, -20.272, -19.597,
    -18.868, -18.095, -17.296, -16.504, -22.041, -21.758, -21.360, -20.859,
    -20.277, -19.602, -18.872, -18.099, -17.301, -16.508, -22.047, -21.764,
    -21.365, -20.864, -20.282, -19.607, -18.877, -18.104, -17.305, -16.512,
    -22.053, -21.770, -21.371, -20.870, -20.287, -19.612, -18.881, -18.108,
    -17.309, -16.516, -22.058, -21.775, -21.376, -20.875, -20.292, -19.617,
    -18.886, -18.112, -17.313, -16.520, -22.064, -21.781, -21.382, -20.880,
    -20.297, -19.621, -18.891, -18.117, -17.317, -16.524, -22.070, -21.787,
    -21.388, -20.886, -20.303, -19.626, -18.895, -18.121, -17.321, -16.528,
    -22.076, -21.793, -21.393, -20.891, -20.308, -19.631, -18.900, -18.125,
    -17.325, -16.532, -22.082, -21.799, -21.399, -20.897, -20.313, -19.636,
    -18.904, -18.130, -17.329, -16.536, -22.088, -21.805, -21.405, -20.902,
    -20.318, -19.641, -18.909, -18.134, -17.334, -16.539, -22.094, -21.810,
    -21.410, -20.907, -20.323, -19.646, -18.913, -18.138, -17.338, -16.543,
    -22.100, -21.816, -21.416, -20.913, -20.328, -19.651, -18.918, -18.143,
    -17.342, -16.547, -22.106, -21.822, -21.422, -20.918, -20.333, -19.655,
    -18.923, -18.147, -17.346, -16.551, -22.112, -21.828, -21.427, -20.924,
    -20.338, -19.660, -18.927, -18.151, -17.350, -16.555, -22.118, -21.834,
    -21.433, -20.929, -20.344, -19.665, -18.932, -18.156, -17.354, -16.559,
    -22.124, -21.840, -21.438, -20.934, -20.349, -19.670, -18.936, -18.160,
    -17.358, -16.563, -22.130, -21.845, -21.444, -20.940, -20.354, -19.675,
    -18.941, -18.164, -17.362, -16.567, -22.136, -21.851, -21.450, -20.945,
    -20.359, -19.680, -18.946, -18.169, -17.366, -16.571, -22.142, -21.857,
    -21.455, -20.951, -20.364, -19.685, -18.950, -18.173, -17.371, -16.575,
    -22.148, -21.863, -21.461, -20.956, -20.369, -19.689, -18.955, -18.177,
    -17.375, -16.579, -22.154, -21.869, -21.467, -20.961, -20.374, -19.694,
    -18.959, -18.182, -17.379, -16.583, -22.159, -21.874, -21.472, -20.967,
    -20.380, -19.699, -18.964, -18.186, -17.383, -16.587, -22.165, -21.880,
    -21.478, -20.972, -20.385, -19.704, -18.968, -18.190, -17.387, -16.590,
    -22.171, -21.886, -21.484, -20.978, -20.390, -19.709, -18.973, -18.195,
    -17.391, -16.594, -22.177, -21.892, -21.489, -20.983, -20.395, -19.714,
    -18.978, -18.199, -17.395, -16.598, -22.183, -21.898, -21.495, -20.988,
    -20.400, -19.719, -18.982, -18.203, -17.399, -16.602, -22.189, -21.904,
    -21.500, -20.994, -20.405, -19.723, -18.987, -18.207, -17.403, -16.606,
    -22.195, -21.909, -21.506, -20.999, -20.410, -19.728, -18.991, -18.212,
    -17.408, -16.610, -22.201, -21.915, -21.512, -21.005, -20.415, -19.733,
    -18.996, -18.216, -17.412, -16.614, -22.207, -21.921, -21.517, -21.010,
    -20.421, -19.738, -19.000, -18.220, -17.416, -16.618, -22.213, -21.927,
    -21.523, -21.015, -20.426, -19.743, -19.005, -18.225, -17.420, -16.622,
    -22.219, -21.933, -21.529, -21.021, -20.431, -19.748, -19.010, -18.229,
    -17.424, -16.626, -22.225, -21.938, -21.534, -21.026, -20.436, -19.752,
    -19.014, -18.233, -17.428, -16.630, -22.231, -21.944, -21.540, -21.032,
    -20.441, -19.757, -19.019, -18.238, -17.432, -16.633, -22.237, -21.950,
    -21.546, -21.037, -20.446, -19.762, -19.023, -18.242, -17.436, -16.637,
    -22.243, -21.956, -21.551, -21.042, -20.451, -19.767, -19.028, -18.246,
    -17.440, -16.641, -22.249, -21.962, -21.557, -21.048, -20.456, -19.772,
    -19.032, -18.251, -17.445, -16.645, -22.254, -21.968, -21.562, -21.053,
    -20.462, -19.777, -19.037, -18.255, -17.449, -16.649, -22.260, -21.973,
    -21.568, -21.059, -20.467, -19.782, -19.042, -18.259, -17.453, -16.653,
    -22.266, -21.979, -21.574, -21.064, -20.472, -19.786, -19.046, -18.263,
    -17.457, -16.657, -22.272, -21.985, -21.579, -21.069, -20.477, -19.791,
    -19.051, -18.268, -17.461, -16.661, -22.278, -21.991, -21.585, -21.075,
    -20.482, -19.796, -19.055, -18.272, -17.465, -16.665, -22.284, -21.997,
    -21.591, -21.080, -20.487, -19.801, -19.060, -18.276, -17.469, -16.669,
    -22.290, -22.002, -21.596, -21.086, -20.492, -19.806, -19.064, -18.281,
    -17.473, -16.672, -22.296, -22.008, -21.602, -21.091, -20.497, -19.811,
    -19.069, -18.285, -17.477, -16.676, -22.302, -22.014, -21.607, -21.096,
    -20.503, -19.816, -19.074, -18.289, -17.481, -16.680, -22.308, -22.020,
    -21.613, -21.102, -20.508, -19.820, -19.078, -18.294, -17.485, -16.684,
    -22.314, -22.026, -21.619, -21.107, -20.513, -19.825, -19.083, -18.298,
    -17.490, -16.688, -22.320, -22.032, -21.624, -21.113, -20.518, -19.830,
    -19.087, -18.302, -17.494, -16.692, -22.326, -22.037, -21.630, -21.118,
    -20.523, -19.835, -19.092, -18.307, -17.498, -16.696, -22.332, -22.043,
    -21.636, -21.123, -20.528, -19.840, -19.096, -18.311, -17.502, -16.700,
    -22.338, -22.049, -21.641, -21.129, -20.533, -19.845, -19.101, -18.315,
    -17.506, -16.704, -22.343, -22.055, -21.647, -21.134, -20.538, -19.849,
    -19.105, -18.319, -17.510, -16.707, -22.349, -22.061, -21.652, -21.140,
    -20.544, -19.854, -19.110, -18.324, -17.514, -16.711, -22.355, -22.066,
    -21.658, -21.145, -20.549, -19.859, -19.115, -18.328, -17.518, -16.715,
    -22.361, -22.072, -21.664, -21.150, -20.554, -19.864, -19.119, -18.332,
    -17.522, -16.719, -22.367, -22.078, -21.669, -21.156, -20.559, -19.869,
    -19.124, -18.337, -17.526, -16.723, -22.373, -22.084, -21.675, -21.161,
    -20.564, -19.874, -19.128, -18.341, -17.530, -16.727, -22.379, -22.090,
    -21.681, -21.166, -20.569, -19.879, -19.133, -18.345, -17.535, -16.731,
    -22.385, -22.096, -21.686, -21.172, -20.574, -19.883, -19.137, -18.350,
    -17.539, -16.735, -22.391, -22.101, -21.692, -21.177, -20.579, -19.888,
    -19.142, -18.354, -17.543, -16.739, -22.397, -22.107, -21.697, -21.183,
    -20.585, -19.893, -19.147, -18.358, -17.547, -16.742, -22.403, -22.113,
    -21.703, -21.188, -20.590, -19.898, -19.151, -18.362, -17.551, -16.746,
    -22.409, -22.119, -21.709, -21.193, -20.595, -19.903, -19.156, -18.367,
    -17.555, -16.750, -22.415, -22.125, -21.714, -21.199, -20.600, -19.908,
    -19.160, -18.371, -17.559, -16.754, -22.421, -22.130, -21.720, -21.204,
    -20.605, -19.912, -19.165, -18.375, -17.563, -16.758, -22.427, -22.136,
    -21.726, -21.210, -20.610, -19.917, -19.169, -18.380, -17.567, -16.762,
    -22.432, -22.142, -21.731, -21.215, -20.615, -19.922, -19.174, -18.384,
    -17.571, -16.766, -22.438, -22.148, -21.737, -21.220, -20.620, -19.927,
    -19.178, -18.388, -17.575, -16.770, -22.444, -22.154, -21.742, -21.226,
    -20.625, -19.932, -19.183, -18.393, -17.579, -16.773, -22.450, -22.159,
    -21.748, -21.231, -20.631, -19.937, -19.188, -18.397, -17.583, -16.777,
    -22.456, -22.165, -21.754, -21.237, -20.636, -19.941, -19.192, -18.401,
    -17.588, -16.781, -22.462, -22.171, -21.759, -21.242, -20.641, -19.946,
    -19.197, -18.405, -17.592, -16.785, -22.468, -22.177, -21.765, -21.247,
    -20.646, -19.951, -19.201, -18.410, -17.596, -16.789, -22.474, -22.183,
    -21.771, -21.253, -20.651, -19.956, -19.206, -18.414, -17.600, -16.793,
    -22.480, -22.188, -21.776, -21.258, -20.656, -19.961, -19.210, -18.418,
    -17.604, -16.797, -22.486, -22.194, -21.782, -21.263, -20.661, -19.966,
    -19.215, -18.423, -17.608, -16.801, -22.492, -22.200, -21.787, -21.269,
    -20.666, -19.970, -19.220, -18.427, -17.612, -16.804, -22.498, -22.206,
    -21.793, -21.274, -20.672, -19.975, -19.224, -18.431, -17.616, -16.808,
    -22.504, -22.212, -21.799, -21.280, -20.677, -19.980, -19.229, -18.435,
    -17.620, -16.812, -22.510, -22.218, -21.804, -21.285, -20.682, -19.985,
    -19.233, -18.440, -17.624, -16.816, -22.515, -22.223, -21.810, -21.290,
    -20.687, -19.990, -19.238, -18.444, -17.628, -16.820, -22.521, -22.229,
    -21.816, -21.296, -20.692, -19.995, -19.242, -18.448, -17.632, -16.824,
    -22.527, -22.235, -21.821, -21.301, -20.697, -20.000, -19.247, -18.453,
    -17.636, -16.828, -22.533, -22.241, -21.827, -21.307, -20.702, -20.004,
    -19.251, -18.457, -17.640, -16.831, -22.539, -22.247, -21.832, -21.312,
    -20.707, -20.009, -19.256, -18.461, -17.645, -16.835, -22.545, -22.252,
    -21.838, -21.317, -20.712, -20.014, -19.261, -18.465, -17.649, -16.839,
    -22.551, -22.258, -21.844, -21.323, -20.718, -20.019, -19.265, -18.470,
    -17.653, -16.843, -22.557, -22.264, -21.849, -21.328, -20.723, -20.024,
    -19.270, -18.474, -17.657, -16.847, -22.563, -22.270, -21.855, -21.333,
    -20.728, -20.029, -19.274, -18.478, -17.661, -16.851, -22.569, -22.276,
    -21.861, -21.339, -20.733, -20.033, -19.279, -18.483, -17.665, -16.855,
    -22.575, -22.281, -21.866, -21.344, -20.738, -20.038, -19.283, -18.487,
    -17.669, -16.858, -22.581, -22.287, -21.872, -21.350, -20.743, -20.043,
    -19.288, -18.491, -17.673, -16.862, -22.587, -22.293, -21.877, -21.355,
    -20.748, -20.048, -19.292, -18.495, -17.677, -16.866, -22.592, -22.299,
    -21.883, -21.360, -20.753, -20.053, -19.297, -18.500, -17.681, -16.870,
    -22.598, -22.305, -21.889, -21.366, -20.759, -20.058, -19.301, -18.504,
    -17.685, -16.874, -22.604, -22.310, -21.894, -21.371, -20.764, -20.062,
    -19.306, -18.508, -17.689, -16.878, -22.610, -22.316, -21.900, -21.377,
    -20.769, -20.067, -19.311, -18.513, -17.693, -16.882, -22.616, -22.322,
    -21.905, -21.382, -20.774, -20.072, -19.315, -18.517, -17.697, -16.885,
    -22.622, -22.328, -21.911, -21.387, -20.779, -20.077, -19.320, -18.521,
    -17.701, -16.889, -22.628, -22.334, -21.917, -21.393, -20.784, -20.082,
    -19.324, -18.525, -17.705, -16.893, -22.634, -22.339, -21.922, -21.398,
    -20.789, -20.087, -19.329, -18.530, -17.709, -16.897, -22.640, -22.345,
    -21.928, -21.403, -20.794, -20.091, -19.333, -18.534, -17.714, -16.901,
    -22.646, -22.351, -21.934, -21.409, -20.799, -20.096, -19.338, -18.538,
    -17.718, -16.905, -22.652, -22.357, -21.939, -21.414, -20.805, -20.101,
    -19.342, -18.543, -17.722, -16.909, -22.658, -22.363, -21.945, -21.420,
    -20.810, -20.106, -19.347, -18.547, -17.726, -16.912, -22.664, -22.368,
    -21.950, -21.425, -20.815, -20.111, -19.352, -18.551, -17.730, -16.916,
    -22.670, -22.374, -21.956, -21.430, -20.820, -20.116, -19.356, -18.555,
    -17.734, -16.920, -22.675, -22.380, -21.962, -21.436, -20.825, -20.120,
    -19.361, -18.560, -17.738, -16.924, -22.681, -22.386, -21.967, -21.441,
    -20.830, -20.125, -19.365, -18.564, -17.742, -16.928, -22.687, -22.392,
    -21.973, -21.446, -20.835, -20.130, -19.370, -18.568, -17.746, -16.932,
    -22.693, -22.397, -21.978, -21.452, -20.840, -20.135, -19.374, -18.573,
    -17.750, -16.935, -22.699, -22.403, -21.984, -21.457, -20.845, -20.140,
    -19.379, -18.577, -17.754, -16.939, -22.705, -22.409, -21.990, -21.463,
    -20.851, -20.145, -19.383, -18.581, -17.758, -16.943, -22.711, -22.415,
    -21.995, -21.468, -20.856, -20.149, -19.388, -18.585, -17.762, -16.947,
    -22.717, -22.421, -22.001, -21.473, -20.861, -20.154, -19.392, -18.590,
    -17.766, -16.951, -22.723, -22.426, -22.007, -21.479, -20.866, -20.159,
    -19.397, -18.594, -17.770, -16.955, -22.729, -22.432, -22.012, -21.484,
    -20.871, -20.164, -19.402, -18.598, -17.774, -16.958, -22.735, -22.438,
    -22.018, -21.489, -20.876, -20.169, -19.406, -18.603, -17.778, -16.962,
    -22.741, -22.444, -22.023, -21.495, -20.881, -20.173, -19.411, -18.607,
    -17.782, -16.966, -22.746, -22.450, -22.029, -21.500, -20.886, -20.178,
    -19.415, -18.611, -17.786, -16.970, -22.752, -22.455, -22.035, -21.506,
    -20.891, -20.183, -19.420, -18.615, -17.790, -16.974, -22.758, -22.461,
    -22.040, -21.511, -20.896, -20.188, -19.424, -18.620, -17.795, -16.978,
    -22.764, -22.467, -22.046, -21.516, -20.902, -20.193, -19.429, -18.624,
    -17.799, -16.981, -22.770, -22.473, -22.051, -21.522, -20.907, -20.198,
    -19.433, -18.628, -17.803, -16.985, -22.776, -22.479, -22.057, -21.527,
    -20.912, -20.202, -19.438, -18.632, -17.807, -16.989, -22.782, -22.484,
    -22.063, -21.532, -20.917, -20.207, -19.442, -18.637, -17.811, -16.993,
    -22.788, -22.490, -22.068, -21.538, -20.922, -20.212, -19.447, -18.641,
    -17.815, -16.997, -22.794, -22.496, -22.074, -21.543, -20.927, -20.217,
    -19.452, -18.645, -17.819, -17.001, -22.800, -22.502, -22.079, -21.549,
    -20.932, -20.222, -19.456, -18.650, -17.823, -17.004, -22.806, -22.508,
    -22.085, -21.554, -20.937, -20.227, -19.461, -18.654, -17.827, -17.008,
    -22.812, -22.513, -22.091, -21.559, -20.942, -20.231, -19.465, -18.658,
    -17.831, -17.012, -22.817, -22.519, -22.096, -21.565, -20.948, -20.236,
    -19.470, -18.662, -17.835, -17.016, -22.823, -22.525, -22.102, -21.570,
    -20.953, -20.241, -19.474, -18.667, -17.839, -17.020, -22.829, -22.531,
    -22.108, -21.575, -20.958, -20.246, -19.479, -18.671, -17.843, -17.024,
    -22.835, -22.537, -22.113, -21.581, -20.963, -20.251, -19.483, -18.675,
    -17.847, -17.027, -22.841, -22.542, -22.119, -21.586, -20.968, -20.256,
    -19.488, -18.679, -17.851, -17.031, -22.847, -22.548, -22.124, -21.592,
    -20.973, -20.260, -19.492, -18.684, -17.855, -17.035, -22.853, -22.554,
    -22.130, -21.597, -20.978, -20.265, -19.497, -18.688, -17.859, -17.039,
    -22.859, -22.560, -22.136, -21.602, -20.983, -20.270, -19.502, -18.692,
    -17.863, -17.043, -22.865, -22.566, -22.141, -21.608, -20.988, -20.275,
    -19.506, -18.697, -17.867, -17.046, -22.871, -22.571, -22.147, -21.613,
    -20.993, -20.280, -19.511, -18.701, -17.871, -17.050, -22.877, -22.577,
    -22.152, -21.618, -20.999, -20.284, -19.515, -18.705, -17.875, -17.054,
    -22.883, -22.583, -22.158, -21.624, -21.004, -20.289, -19.520, -18.709,
    -17.879, -17.058, -22.888, -22.589, -22.164, -21.629, -21.009, -20.294,
    -19.524, -18.714, -17.883, -17.062, -22.894, -22.595, -22.169, -21.635,
    -21.014, -20.299, -19.529, -18.718, -17.887, -17.066, -22.900, -22.600,
    -22.175, -21.640, -21.019, -20.304, -19.533, -18.722, -17.891, -17.069,
    -22.906, -22.606, -22.180, -21.645, -21.024, -20.309, -19.538, -18.726,
    -17.895, -17.073, -22.912, -22.612, -22.186, -21.651, -21.029, -20.313,
    -19.542, -18.731, -17.899, -17.077, -22.918, -22.618, -22.192, -21.656,
    -21.034, -20.318, -19.547, -18.735, -17.903, -17.081, -22.924, -22.624,
    -22.197, -21.661, -21.039, -20.323, -19.551, -18.739, -17.908, -17.085,
    -22.930, -22.629, -22.203, -21.667, -21.045, -20.328, -19.556, -18.743,
    -17.912, -17.088, -22.936, -22.635, -22.208, -21.672, -21.050, -20.333,
    -19.561, -18.748, -17.916, -17.092, -22.942, -22.641, -22.214, -21.678,
    -21.055, -20.338, -19.565, -18.752, -17.920, -17.096, -22.948, -22.647,
    -22.220, -21.683, -21.060, -20.342, -19.570, -18.756, -17.924, -17.100,
    -22.954, -22.653, -22.225, -21.688, -21.065, -20.347, -19.574, -18.761,
    -17.928, -17.104, -22.959, -22.658, -22.231, -21.694, -21.070, -20.352,
    -19.579, -18.765, -17.932, -17.108, -22.965, -22.664, -22.236, -21.699,
    -21.075, -20.357, -19.583, -18.769, -17.936, -17.111, -22.971, -22.670,
    -22.242, -21.704, -21.080, -20.362, -19.588, -18.773, -17.940, -17.115,
    -22.977, -22.676, -22.248, -21.710, -21.085, -20.366, -19.592, -18.778,
    -17.944, -17.119, -22.983, -22.681, -22.253, -21.715, -21.090, -20.371,
    -19.597, -18.782, -17.948, -17.123, -22.989, -22.687, -22.259, -21.720,
    -21.096, -20.376, -19.601, -18.786, -17.952, -17.127, -22.995, -22.693,
    -22.264, -21.726, -21.101, -20.381, -19.606, -18.790, -17.956, -17.130,
    -23.001, -22.699, -22.270, -21.731, -21.106, -20.386, -19.610, -18.795,
    -17.960, -17.134, -23.007, -22.705, -22.276, -21.737, -21.111, -20.391,
    -19.615, -18.799, -17.964, -17.138, -23.013, -22.710, -22.281, -21.742,
    -21.116, -20.395, -19.619, -18.803, -17.968, -17.142, -23.019, -22.716,
    -22.287, -21.747, -21.121, -20.400, -19.624, -18.807, -17.972, -17.146,
    -23.024, -22.722, -22.292, -21.753, -21.126, -20.405, -19.629, -18.812,
    -17.976, -17.149, -23.030, -22.728, -22.298, -21.758, -21.131, -20.410,
    -19.633, -18.816, -17.980, -17.153, -23.036, -22.734, -22.304, -21.763,
    -21.136, -20.415, -19.638, -18.820, -17.984, -17.157, -23.042, -22.739,
    -22.309, -21.769, -21.141, -20.419, -19.642, -18.824, -17.988, -17.161,
    -23.048, -22.745, -22.315, -21.774, -21.147, -20.424, -19.647, -18.829,
    -17.992, -17.165, -23.054, -22.751, -22.320, -21.779, -21.152, -20.429,
    -19.651, -18.833, -17.996, -17.168, -23.060, -22.757, -22.326, -21.785,
    -21.157, -20.434, -19.656, -18.837, -18.000, -17.172, -23.066, -22.763,
    -22.332, -21.790, -21.162, -20.439, -19.660, -18.841, -18.004, -17.176,
    -23.072, -22.768, -22.337, -21.796, -21.167, -20.444, -19.665, -18.846,
    -18.008, -17.180, -23.078, -22.774, -22.343, -21.801, -21.172, -20.448,
    -19.669, -18.850, -18.012, -17.184, -23.084, -22.780, -22.348, -21.806,
    -21.177, -20.453, -19.674, -18.854, -18.016, -17.187, -23.090, -22.786,
    -22.354, -21.812, -21.182, -20.458, -19.678, -18.859, -18.020, -17.191,
    -23.095, -22.791, -22.360, -21.817, -21.187, -20.463, -19.683, -18.863,
    -18.024, -17.195, -23.101, -22.797, -22.365, -21.822, -21.192, -20.468,
    -19.687, -18.867, -18.028, -17.199, -23.107, -22.803, -22.371, -21.828,
    -21.197, -20.472, -19.692, -18.871, -18.032, -17.203, -23.113, -22.809,
    -22.376, -21.833, -21.203, -20.477, -19.697, -18.876, -18.036, -17.206,
    -23.119, -22.815, -22.382, -21.838, -21.208, -20.482, -19.701, -18.880,
    -18.040, -17.210, -23.125, -22.820, -22.388, -21.844, -21.213, -20.487,
    -19.706, -18.884, -18.044, -17.214, -23.131, -22.826, -22.393, -21.849,
    -21.218, -20.492, -19.710, -18.888, -18.048, -17.218, -23.137, -22.832,
    -22.399, -21.855, -21.223, -20.497, -19.715, -18.893, -18.052, -17.222,
    -23.143, -22.838, -22.404, -21.860, -21.228, -20.501, -19.719, -18.897,
    -18.056, -17.225, -23.149, -22.844, -22.410, -21.865, -21.233, -20.506,
    -19.724, -18.901, -18.060, -17.229, -23.155, -22.849, -22.416, -21.871,
    -21.238, -20.511, -19.728, -18.905, -18.064, -17.233, -23.160, -22.855,
    -22.421, -21.876, -21.243, -20.516, -19.733, -18.910, -18.068, -17.237,
    -23.166, -22.861, -22.427, -21.881, -21.248, -20.521, -19.737, -18.914,
    -18.072, -17.240, -23.172, -22.867, -22.432, -21.887, -21.254, -20.525,
    -19.742, -18.918, -18.076, -17.244, -23.178, -22.872, -22.438, -21.892,
    -21.259, -20.530, -19.746, -18.922, -18.080, -17.248, -23.184, -22.878,
    -22.444, -21.897, -21.264, -20.535, -19.751, -18.927, -18.084, -17.252,
    -23.190, -22.884, -22.449, -21.903, -21.269, -20.540, -19.755, -18.931,
    -18.088, -17.256, -23.196, -22.890, -22.455, -21.908, -21.274, -20.545,
    -19.760, -18.935, -18.092, -17.259, -23.202, -22.896, -22.460, -21.914,
    -21.279, -20.549, -19.764, -18.939, -18.096, -17.263, -23.208, -22.901,
    -22.466, -21.919, -21.284, -20.554, -19.769, -18.944, -18.100, -17.267,
    -23.214, -22.907, -22.472, -21.924, -21.289, -20.559, -19.774, -18.948,
    -18.104, -17.271, -23.219, -22.913, -22.477, -21.930, -21.294, -20.564,
    -19.778, -18.952, -18.108, -17.275, -23.225, -22.919, -22.483, -21.935,
    -21.299, -20.569, -19.783, -18.956, -18.112, -17.278, -23.231, -22.925,
    -22.488, -21.940, -21.304, -20.574, -19.787, -18.961, -18.116, -17.282,
    -23.237, -22.930, -22.494, -21.946, -21.310, -20.578, -19.792, -18.965,
    -18.120, -17.286, -23.243, -22.936, -22.500, -21.951, -21.315, -20.583,
    -19.796, -18.969, -18.124, -17.290, -23.249, -22.942, -22.505, -21.956,
    -21.320, -20.588, -19.801, -18.973, -18.128, -17.293, -23.255, -22.948,
    -22.511, -21.962, -21.325, -20.593, -19.805, -18.978, -18.132, -17.297,
    -23.261, -22.953, -22.516, -21.967, -21.330, -20.598, -19.810, -18.982,
    -18.136, -17.301, -23.267, -22.959, -22.522, -21.972, -21.335, -20.602,
    -19.814, -18.986, -18.140, -17.305, -23.273, -22.965, -22.528, -21.978,
    -21.340, -20.607, -19.819, -18.990, -18.144, -17.309, -23.279, -22.971,
    -22.533, -21.983, -21.345, -20.612, -19.823, -18.995, -18.148, -17.312,
    -23.284, -22.977, -22.539, -21.989, -21.350, -20.617, -19.828, -18.999,
    -18.152, -17.316, -23.290, -22.982, -22.544, -21.994, -21.355, -20.622,
    -19.832, -19.003, -18.156, -17.320, -23.296, -22.988, -22.550, -21.999,
    -21.360, -20.626, -19.837, -19.007, -18.160, -17.324, -23.302, -22.994,
    -22.556, -22.005, -21.366, -20.631, -19.841, -19.012, -18.164, -17.327,
    -23.308, -23.000, -22.561, -22.010, -21.371, -20.636, -19.846, -19.016,
    -18.168, -17.331, -23.314, -23.005, -22.567, -22.015, -21.376, -20.641,
    -19.850, -19.020, -18.172, -17.335, -23.320, -23.011, -22.572, -22.021,
    -21.381, -20.646, -19.855, -19.024, -18.176, -17.339, -23.326, -23.017,
    -22.578, -22.026, -21.386, -20.650, -19.859, -19.029, -18.180, -17.343,
    -23.332, -23.023, -22.583, -22.031, -21.391, -20.655, -19.864, -19.033,
    -18.184, -17.346, -23.338, -23.029, -22.589, -22.037, -21.396, -20.660,
    -19.869, -19.037, -18.188, -17.350, -23.343, -23.034, -22.595, -22.042,
    -21.401, -20.665, -19.873, -19.041, -18.192, -17.354, -23.349, -23.040,
    -22.600, -22.047, -21.406, -20.670, -19.878, -19.046, -18.196, -17.358,
    -23.355, -23.046, -22.606, -22.053, -21.411, -20.675, -19.882, -19.050,
    -18.200, -17.361, -23.361, -23.052, -22.611, -22.058, -21.416, -20.679,
    -19.887, -19.054, -18.204, -17.365, -23.367, -23.058, -22.617, -22.064,
    -21.422, -20.684, -19.891, -19.058, -18.208, -17.369, -23.373, -23.063,
    -22.623, -22.069, -21.427, -20.689, -19.896, -19.063, -18.212, -17.373,
    -23.379, -23.069, -22.628, -22.074, -21.432, -20.694, -19.900, -19.067,
    -18.216, -17.376, -23.385, -23.075, -22.634, -22.080, -21.437, -20.699,
    -19.905, -19.071, -18.220, -17.380, -23.391, -23.081, -22.639, -22.085,
    -21.442, -20.703, -19.909, -19.075, -18.224, -17.384, -23.397, -23.086,
    -22.645, -22.090, -21.447, -20.708, -19.914, -19.080, -18.228, -17.388,
    -23.402, -23.092, -22.651, -22.096, -21.452, -20.713, -19.918, -19.084,
    -18.232, -17.392, -23.408, -23.098, -22.656, -22.101, -21.457, -20.718,
    -19.923, -19.088, -18.236, -17.395, -23.414, -23.104, -22.662, -22.106,
    -21.462, -20.723, -19.927, -19.092, -18.240, -17.399, -23.420, -23.110,
    -22.667, -22.112, -21.467, -20.727, -19.932, -19.096, -18.244, -17.403,
    -23.426, -23.115, -22.673, -22.117, -21.472, -20.732, -19.936, -19.101,
    -18.248, -17.407, -23.432, -23.121, -22.678, -22.122, -21.477, -20.737,
    -19.941, -19.105, -18.252, -17.410, -23.438, -23.127, -22.684, -22.128,
    -21.483, -20.742, -19.945, -19.109, -18.256, -17.414, -23.444, -23.133,
    -22.690, -22.133, -21.488, -20.747, -19.950, -19.113, -18.260, -17.418,
    -23.450, -23.138, -22.695, -22.139, -21.493, -20.751, -19.954, -19.118,
    -18.264, -17.422, -23.456, -23.144, -22.701, -22.144, -21.498, -20.756,
    -19.959, -19.122, -18.268, -17.425, -23.461, -23.150, -22.706, -22.149,
    -21.503, -20.761, -19.963, -19.126, -18.272, -17.429, -23.467, -23.156,
    -22.712, -22.155, -21.508, -20.766, -19.968, -19.130, -18.276, -17.433,
    -23.473, -23.162, -22.718, -22.160, -21.513, -20.771, -19.972, -19.135,
    -18.280, -17.437, -23.479, -23.167, -22.723, -22.165, -21.518, -20.775,
    -19.977, -19.139, -18.284, -17.440, -23.485, -23.173, -22.729, -22.171,
    -21.523, -20.780, -19.981, -19.143, -18.288, -17.444, -23.491, -23.179,
    -22.734, -22.176, -21.528, -20.785, -19.986, -19.147, -18.292, -17.448,
    -23.497, -23.185, -22.740, -22.181, -21.533, -20.790, -19.991, -19.152,
    -18.296, -17.452, -23.503, -23.190, -22.746, -22.187, -21.539, -20.795,
    -19.995, -19.156, -18.300, -17.456, -23.509, -23.196, -22.751, -22.192,
    -21.544, -20.799, -20.000, -19.160, -18.304, -17.459, -23.515, -23.202,
    -22.757, -22.197, -21.549, -20.804, -20.004, -19.164, -18.308, -17.463,
    -23.520, -23.208, -22.762, -22.203, -21.554, -20.809, -20.009, -19.169,
    -18.312, -17.467, -23.526, -23.213, -22.768, -22.208, -21.559, -20.814,
    -20.013, -19.173, -18.316, -17.471, -23.532, -23.219, -22.773, -22.213,
    -21.564, -20.819, -20.018, -19.177, -18.320, -17.474, -23.538, -23.225,
    -22.779, -22.219, -21.569, -20.823, -20.022, -19.181, -18.324, -17.478,
    -23.544, -23.231, -22.785, -22.224, -21.574, -20.828, -20.027, -19.186,
    -18.328, -17.482, -23.550, -23.237, -22.790, -22.229, -21.579, -20.833,
    -20.031, -19.190, -18.332, -17.486, -23.556, -23.242, -22.796, -22.235,
    -21.584, -20.838, -20.036, -19.194, -18.336, -17.489, -23.562, -23.248,
    -22.801, -22.240, -21.589, -20.843, -20.040, -19.198, -18.340, -17.493,
    -23.568, -23.254, -22.807, -22.246, -21.594, -20.847, -20.045, -19.202,
    -18.344, -17.497, -23.574, -23.260, -22.813, -22.251, -21.600, -20.852,
    -20.049, -19.207, -18.348, -17.501, -23.579, -23.265, -22.818, -22.256,
    -21.605, -20.857, -20.054, -19.211, -18.352, -17.504, -23.585, -23.271,
    -22.824, -22.262, -21.610, -20.862, -20.058, -19.215, -18.356, -17.508,
    -23.591, -23.277, -22.829, -22.267, -21.615, -20.867, -20.063, -19.219,
    -18.360, -17.512, -23.597, -23.283, -22.835, -22.272, -21.620, -20.871,
    -20.067, -19.224, -18.364, -17.516, -23.603, -23.289, -22.840, -22.278,
    -21.625, -20.876, -20.072, -19.228, -18.368, -17.519, -23.609, -23.294,
    -22.846, -22.283, -21.630, -20.881, -20.076, -19.232, -18.372, -17.523,
    -23.615, -23.300, -22.852, -22.288, -21.635, -20.886, -20.081, -19.236,
    -18.376, -17.527, -23.621, -23.306, -22.857, -22.294, -21.640, -20.891,
    -20.085, -19.241, -18.380, -17.531, -23.627, -23.312, -22.863, -22.299,
    -21.645, -20.895, -20.090, -19.245, -18.384, -17.534, -23.633, -23.317,
    -22.868, -22.304, -21.650, -20.900, -20.094, -19.249, -18.388, -17.538,
    -23.638, -23.323, -22.874, -22.310, -21.655, -20.905, -20.099, -19.253,
    -18.392, -17.542, -23.644, -23.329, -22.880, -22.315, -21.660, -20.910,
    -20.103, -19.258, -18.396, -17.546, -23.650, -23.335, -22.885, -22.320,
    -21.666, -20.915, -20.108, -19.262, -18.400, -17.549, -23.656, -23.340,
    -22.891, -22.326, -21.671, -20.919, -20.112, -19.266, -18.404, -17.553,
    -23.662, -23.346, -22.896, -22.331, -21.676, -20.924, -20.117, -19.270,
    -18.408, -17.557, -23.668, -23.352, -22.902, -22.336, -21.681, -20.929,
    -20.121, -19.274, -18.412, -17.561, -23.674, -23.358, -22.907, -22.342,
    -21.686, -20.934, -20.126, -19.279, -18.416, -17.564, -23.680, -23.364,
    -22.913, -22.347, -21.691, -20.939, -20.130, -19.283, -18.420, -17.568,
    -23.686, -23.369, -22.919, -22.352, -21.696, -20.943, -20.135, -19.287,
    -18.424, -17.572, -23.691, -23.375, -22.924, -22.358, -21.701, -20.948,
    -20.139, -19.291, -18.428, -17.576, -23.697, -23.381, -22.930, -22.363,
    -21.706, -20.953, -20.144, -19.296, -18.432, -17.579, -23.703, -23.387,
    -22.935, -22.368, -21.711, -20.958, -20.148, -19.300, -18.436, -17.583,
    -23.709, -23.392, -22.941, -22.374, -21.716, -20.963, -20.153, -19.304,
    -18.440, -17.587, -23.715, -23.398, -22.947, -22.379, -21.721, -20.967,
    -20.158, -19.308, -18.444, -17.591, -23.721, -23.404, -22.952, -22.385,
    -21.727, -20.972, -20.162, -19.313, -18.448, -17.594, -23.727, -23.410,
    -22.958, -22.390, -21.732, -20.977, -20.167, -19.317, -18.451, -17.598,
    -23.733, -23.415, -22.963, -22.395, -21.737, -20.982, -20.171, -19.321,
    -18.455, -17.602, -23.739, -23.421, -22.969, -22.401, -21.742, -20.987,
    -20.176, -19.325, -18.459, -17.605, -23.744, -23.427, -22.974, -22.406,
    -21.747, -20.991, -20.180, -19.329, -18.463, -17.609, -23.750, -23.433,
    -22.980, -22.411, -21.752, -20.996, -20.185, -19.334, -18.467, -17.613,
    -23.756, -23.439, -22.986, -22.417, -21.757, -21.001, -20.189, -19.338,
    -18.471, -17.617, -23.762, -23.444, -22.991, -22.422, -21.762, -21.006,
    -20.194, -19.342, -18.475, -17.620, -23.768, -23.450, -22.997, -22.427,
    -21.767, -21.011, -20.198, -19.346, -18.479, -17.624, -23.774, -23.456,
    -23.002, -22.433, -21.772, -21.015, -20.203, -19.351, -18.483, -17.628,
    -23.780, -23.462, -23.008, -22.438, -21.777, -21.020, -20.207, -19.355,
    -18.487, -17.632, -23.786, -23.467, -23.013, -22.443, -21.782, -21.025,
    -20.212, -19.359, -18.491, -17.635, -23.792, -23.473, -23.019, -22.449,
    -21.787, -21.030, -20.216, -19.363, -18.495, -17.639, -23.798, -23.479,
    -23.025, -22.454, -21.792, -21.035, -20.221, -19.367, -18.499, -17.643,
    -23.803, -23.485, -23.030, -22.459, -21.798, -21.039, -20.225, -19.372,
    -18.503, -17.647, -23.809, -23.490, -23.036, -22.465, -21.803, -21.044,
    -20.230, -19.376, -18.507, -17.650, -23.815, -23.496, -23.041, -22.470,
    -21.808, -21.049, -20.234, -19.380, -18.511, -17.654, -23.821, -23.502,
    -23.047, -22.475, -21.813, -21.054, -20.239, -19.384, -18.515, -17.658,
    -23.827, -23.508, -23.052, -22.481, -21.818, -21.059, -20.243, -19.389,
    -18.519, -17.662, -23.833, -23.514, -23.058, -22.486, -21.823, -21.063,
    -20.248, -19.393, -18.523, -17.665, -23.839, -23.519, -23.064, -22.491,
    -21.828, -21.068, -20.252, -19.397, -18.527, -17.669, -23.845, -23.525,
    -23.069, -22.497, -21.833, -21.073, -20.257, -19.401, -18.531, -17.673,
    -23.851, -23.531, -23.075, -22.502, -21.838, -21.078, -20.261, -19.406,
    -18.535, -17.676, -23.856, -23.537, -23.080, -22.507, -21.843, -21.083,
    -20.266, -19.410, -18.539, -17.680, -23.862, -23.542, -23.086, -22.513,
    -21.848, -21.087, -20.270, -19.414, -18.543, -17.684, -23.868, -23.548,
    -23.091, -22.518, -21.853, -21.092, -20.275, -19.418, -18.547, -17.688,
    -23.874, -23.554, -23.097, -22.523, -21.858, -21.097, -20.279, -19.422,
    -18.551, -17.691, -23.880, -23.560, -23.103, -22.529, -21.864, -21.102,
    -20.284, -19.427, -18.555, -17.695, -23.886, -23.565, -23.108, -22.534,
    -21.869, -21.106, -20.288, -19.431, -18.559, -17.699, -23.892, -23.571,
    -23.114, -22.539, -21.874, -21.111, -20.293, -19.435, -18.563, -17.703,
    -23.898, -23.577, -23.119, -22.545, -21.879, -21.116, -20.297, -19.439,
    -18.567, -17.706, -23.904, -23.583, -23.125, -22.550, -21.884, -21.121,
    -20.302, -19.444, -18.571, -17.710, -23.909, -23.588, -23.131, -22.555,
    -21.889, -21.126, -20.306, -19.448, -18.575, -17.714, -23.915, -23.594,
    -23.136, -22.561, -21.894, -21.130, -20.311, -19.452, -18.579, -17.718,
    -23.921, -23.600, -23.142, -22.566, -21.899, -21.135, -20.315, -19.456,
    -18.582, -17.721, -23.927, -23.606, -23.147, -22.571, -21.904, -21.140,
    -20.320, -19.460, -18.586, -17.725, -23.933, -23.611, -23.153, -22.577,
    -21.909, -21.145, -20.324, -19.465, -18.590, -17.729, -23.939, -23.617,
    -23.158, -22.582, -21.914, -21.150, -20.329, -19.469, -18.594, -17.732,
    -23.945, -23.623, -23.164, -22.587, -21.919, -21.154, -20.333, -19.473,
    -18.598, -17.736, -23.951, -23.629, -23.170, -22.593, -21.924, -21.159,
    -20.338, -19.477, -18.602, -17.740, -23.957, -23.635, -23.175, -22.598,
    -21.929, -21.164, -20.342, -19.482, -18.606, -17.744, -23.962, -23.640,
    -23.181, -22.603, -21.935, -21.169, -20.347, -19.486, -18.610, -17.747,
    -23.968, -23.646, -23.186, -22.609, -21.940, -21.174, -20.351, -19.490,
    -18.614, -17.751, -23.974, -23.652, -23.192, -22.614, -21.945, -21.178,
    -20.356, -19.494, -18.618, -17.755, -23.980, -23.658, -23.197, -22.619,
    -21.950, -21.183, -20.360, -19.498, -18.622, -17.758, -23.986, -23.663,
    -23.203, -22.625, -21.955, -21.188, -20.365, -19.503, -18.626, -17.762,
    -23.992, -23.669, -23.209, -22.630, -21.960, -21.193, -20.369, -19.507,
    -18.630, -17.766, -23.998, -23.675, -23.214, -22.635, -21.965, -21.197,
    -20.374, -19.511, -18.634, -17.770, -24.004, -23.681, -23.220, -22.641,
    -21.970, -21.202, -20.378, -19.515, -18.638, -17.773, -24.010, -23.686,
    -23.225, -22.646, -21.975, -21.207, -20.383, -19.519, -18.642, -17.777,
    -24.015, -23.692, -23.231, -22.651, -21.980, -21.212, -20.387, -19.524,
    -18.646, -17.781, -24.021, -23.698, -23.236, -22.657, -21.985, -21.217,
    -20.392, -19.528, -18.650, -17.785, -24.027, -23.704, -23.242, -22.662,
    -21.990, -21.221, -20.396, -19.532, -18.654, -17.788, -24.033, -23.709,
    -23.248, -22.667, -21.995, -21.226, -20.401, -19.536, -18.658, -17.792,
    -24.039, -23.715, -23.253, -22.673, -22.000, -21.231, -20.405, -19.541,
    -18.662, -17.796, -24.045, -23.721, -23.259, -22.678, -22.006, -21.236,
    -20.410, -19.545, -18.666, -17.799, -24.051, -23.727, -23.264, -22.683,
    -22.011, -21.241, -20.414, -19.549, -18.670, -17.803, -24.057, -23.732,
    -23.270, -22.689, -22.016, -21.245, -20.419, -19.553, -18.674, -17.807,
    -24.062, -23.738, -23.275, -22.694, -22.021, -21.250, -20.423, -19.557,
    -18.678, -17.811, -24.068, -23.744, -23.281, -22.699, -22.026, -21.255,
    -20.428, -19.562, -18.681, -17.814, -24.074, -23.750, -23.286, -22.705,
    -22.031, -21.260, -20.432, -19.566, -18.685, -17.818, -24.080, -23.756,
    -23.292, -22.710, -22.036, -21.265, -20.437, -19.570, -18.689, -17.822,
    -24.086, -23.761, -23.298, -22.715, -22.041, -21.269, -20.441, -19.574,
    -18.693, -17.825, -24.092, -23.767, -23.303, -22.721, -22.046, -21.274,
    -20.446, -19.579, -18.697, -17.829, -24.098, -23.773, -23.309, -22.726,
    -22.051, -21.279, -20.450, -19.583, -18.701, -17.833, -24.104, -23.779,
    -23.314, -22.731, -22.056, -21.284, -20.455, -19.587, -18.705, -17.837,
    -24.110, -23.784, -23.320, -22.737, -22.061, -21.288, -20.459, -19.591,
    -18.709, -17.840, -24.115, -23.790, -23.325, -22.742, -22.066, -21.293,
    -20.464, -19.595, -18.713, -17.844, -24.121, -23.796, -23.331, -22.747,
    -22.071, -21.298, -20.468, -19.600, -18.717, -17.848, -24.127, -23.802,
    -23.337, -22.753, -22.076, -21.303, -20.473, -19.604, -18.721, -17.851,
    -24.133, -23.807, -23.342, -22.758, -22.082, -21.308, -20.477, -19.608,
    -18.725, -17.855, -24.139, -23.813, -23.348, -22.763, -22.087, -21.312,
    -20.482, -19.612, -18.729, -17.859, -24.145, -23.819, -23.353, -22.769,
    -22.092, -21.317, -20.486, -19.616, -18.733, -17.863, -24.151, -23.825,
    -23.359, -22.774, -22.097, -21.322, -20.491, -19.621, -18.737, -17.866,
    -24.157, -23.830, -23.364, -22.779, -22.102, -21.327, -20.495, -19.625,
    -18.741, -17.870, -24.163, -23.836, -23.370, -22.785, -22.107, -21.331,
    -20.500, -19.629, -18.745, -17.874, -24.168, -23.842, -23.376, -22.790,
    -22.112, -21.336, -20.504, -19.633, -18.749, -17.877, -24.174, -23.848,
    -23.381, -22.795, -22.117, -21.341, -20.509, -19.638, -18.753, -17.881,
    -24.180, -23.853, -23.387, -22.801, -22.122, -21.346, -20.513, -19.642,
    -18.757, -17.885, -24.186, -23.859, -23.392, -22.806, -22.127, -21.351,
    -20.518, -19.646, -18.761, -17.889, -24.192, -23.865, -23.398, -22.811,
    -22.132, -21.355, -20.522, -19.650, -18.765, -17.892, -24.198, -23.871,
    -23.403, -22.817, -22.137, -21.360, -20.527, -19.654, -18.768, -17.896,
    -24.204, -23.876, -23.409, -22.822, -22.142, -21.365, -20.531, -19.659,
    -18.772, -17.900, -24.210, -23.882, -23.415, -22.827, -22.147, -21.370,
    -20.536, -19.663, -18.776, -17.903, -24.215, -23.888, -23.420, -22.833,
    -22.152, -21.375, -20.540, -19.667, -18.780, -17.907, -24.221, -23.894,
    -23.426, -22.838, -22.158, -21.379, -20.545, -19.671, -18.784, -17.911,
    -24.227, -23.899, -23.431, -22.843, -22.163, -21.384, -20.549, -19.675,
    -18.788, -17.915, -24.233, -23.905, -23.437, -22.849, -22.168, -21.389,
    -20.554, -19.680, -18.792, -17.918, -24.239, -23.911, -23.442, -22.854,
    -22.173, -21.394, -20.558, -19.684, -18.796, -17.922, -24.245, -23.917,
    -23.448, -22.859, -22.178, -21.398, -20.563, -19.688, -18.800, -17.926,
    -24.251, -23.922, -23.453, -22.865, -22.183, -21.403, -20.567, -19.692,
    -18.804, -17.929, -24.257, -23.928, -23.459, -22.870, -22.188, -21.408,
    -20.572, -19.697, -18.808, -17.933, -24.263, -23.934, -23.465, -22.875,
    -22.193, -21.413, -20.576, -19.701, -18.812, -17.937, -24.268, -23.940,
    -23.470, -22.881, -22.198, -21.418, -20.581, -19.705, -18.816, -17.941,
    -24.274, -23.945, -23.476, -22.886, -22.203, -21.422, -20.585, -19.709,
    -18.820, -17.944, -24.280, -23.951, -23.481, -22.891, -22.208, -21.427,
    -20.590, -19.713, -18.824, -17.948, -24.286, -23.957, -23.487, -22.897,
    -22.213, -21.432, -20.594, -19.718, -18.828, -17.952, -24.292, -23.963,
    -23.492, -22.902, -22.218, -21.437, -20.599, -19.722, -18.832, -17.955,
    -24.298, -23.968, -23.498, -22.907, -22.223, -21.441, -20.603, -19.726,
    -18.836, -17.959, -24.304, -23.974, -23.504, -22.913, -22.228, -21.446,
    -20.608, -19.730, -18.840, -17.963, -24.310, -23.980, -23.509, -22.918,
    -22.233, -21.451, -20.612, -19.734, -18.844, -17.967, -24.315, -23.986,
    -23.515, -22.923, -22.239, -21.456, -20.616, -19.739, -18.847, -17.970,
    -24.321, -23.991, -23.520, -22.929, -22.244, -21.461, -20.621, -19.743,
    -18.851, -17.974, -24.327, -23.997, -23.526, -22.934, -22.249, -21.465,
    -20.625, -19.747, -18.855, -17.978, -24.333, -24.003, -23.531, -22.939,
    -22.254, -21.470, -20.630, -19.751, -18.859, -17.981, -24.339, -24.009,
    -23.537, -22.945, -22.259, -21.475, -20.634, -19.755, -18.863, -17.985,
    -24.345, -24.014, -23.542, -22.950, -22.264, -21.480, -20.639, -19.760,
    -18.867, -17.989, -24.351, -24.020, -23.548, -22.955, -22.269, -21.485,
    -20.643, -19.764, -18.871, -17.992, -24.357, -24.026, -23.554, -22.961,
    -22.274, -21.489, -20.648, -19.768, -18.875, -17.996, -24.362, -24.032,
    -23.559, -22.966, -22.279, -21.494, -20.652, -19.772, -18.879, -18.000,
    -24.368, -24.037, -23.565, -22.971, -22.284, -21.499, -20.657, -19.776,
    -18.883, -18.004, -24.374, -24.043, -23.570, -22.977, -22.289, -21.504,
    -20.661, -19.781, -18.887, -18.007, -24.380, -24.049, -23.576, -22.982,
    -22.294, -21.508, -20.666, -19.785, -18.891, -18.011, -24.386, -24.055,
    -23.581, -22.987, -22.299, -21.513, -20.670, -19.789, -18.895, -18.015,
    -24.392, -24.060, -23.587, -22.993, -22.304, -21.518, -20.675, -19.793,
    -18.899, -18.018, -24.398, -24.066, -23.593, -22.998, -22.309, -21.523,
    -20.679, -19.798, -18.903, -18.022, -24.404, -24.072, -23.598, -23.003,
    -22.314, -21.528, -20.684, -19.802, -18.907, -18.026, -24.409, -24.078,
    -23.604, -23.009, -22.319, -21.532, -20.688, -19.806, -18.911, -18.029,
    -24.415, -24.084, -23.609, -23.014, -22.325, -21.537, -20.693, -19.810,
    -18.915, -18.033, -24.421, -24.089, -23.615, -23.019, -22.330, -21.542,
    -20.697, -19.814, -18.918, -18.037, -24.427, -24.095, -23.620, -23.025,
    -22.335, -21.547, -20.702, -19.819, -18.922, -18.041, -24.433, -24.101,
    -23.626, -23.030, -22.340, -21.551, -20.706, -19.823, -18.926, -18.044,
    -24.439, -24.106, -23.631, -23.035, -22.345, -21.556, -20.711, -19.827,
    -18.930, -18.048, -24.445, -24.112, -23.637, -23.041, -22.350, -21.561,
    -20.715, -19.831, -18.934, -18.052, -24.451, -24.118, -23.643, -23.046,
    -22.355, -21.566, -20.720, -19.835, -18.938, -18.055, -24.457, -24.124,
    -23.648, -23.051, -22.360, -21.570, -20.724, -19.840, -18.942, -18.059,
    -24.462, -24.129, -23.654, -23.056, -22.365, -21.575, -20.729, -19.844,
    -18.946, -18.063, -24.468, -24.135, -23.659, -23.062, -22.370, -21.580,
    -20.733, -19.848, -18.950, -18.066, -24.474, -24.141, -23.665, -23.067,
    -22.375, -21.585, -20.738, -19.852, -18.954, -18.070, -24.480, -24.147,
    -23.670, -23.072, -22.380, -21.590, -20.742, -19.856, -18.958, -18.074,
    -24.486, -24.152, -23.676, -23.078, -22.385, -21.594, -20.747, -19.861,
    -18.962, -18.078, -24.492, -24.158, -23.681, -23.083, -22.390, -21.599,
    -20.751, -19.865, -18.966, -18.081, -24.498, -24.164, -23.687, -23.088,
    -22.395, -21.604, -20.756, -19.869, -18.970, -18.085, -24.504, -24.170,
    -23.693, -23.094, -22.400, -21.609, -20.760, -19.873, -18.974, -18.089,
    -24.509, -24.175, -23.698, -23.099, -22.405, -21.613, -20.765, -19.877,
    -18.978, -18.092, -24.515, -24.181, -23.704, -23.104, -22.411, -21.618,
    -20.769, -19.882, -18.981, -18.096, -24.521, -24.187, -23.709, -23.110,
    -22.416, -21.623, -20.774, -19.886, -18.985, -18.100, -24.527, -24.193,
    -23.715, -23.115, -22.421, -21.628, -20.778, -19.890, -18.989, -18.103,
    -24.533, -24.198, -23.720, -23.120, -22.426, -21.633, -20.783, -19.894,
    -18.993, -18.107, -24.539, -24.204, -23.726, -23.126, -22.431, -21.637,
    -20.787, -19.898, -18.997, -18.111, -24.545, -24.210, -23.731, -23.131,
    -22.436, -21.642, -20.792, -19.903, -19.001, -18.114, -24.551, -24.216,
    -23.737, -23.136, -22.441, -21.647, -20.796, -19.907, -19.005, -18.118,
    -24.556, -24.221, -23.743, -23.142, -22.446, -21.652, -20.801, -19.911,
    -19.009, -18.122, -24.562, -24.227, -23.748, -23.147, -22.451, -21.656,
    -20.805, -19.915, -19.013, -18.126, -24.568, -24.233, -23.754, -23.152,
    -22.456, -21.661, -20.809, -19.919, -19.017, -18.129, -24.574, -24.239,
    -23.759, -23.158, -22.461, -21.666, -20.814, -19.924, -19.021, -18.133,
    -24.580, -24.244, -23.765, -23.163, -22.466, -21.671, -20.818, -19.928,
    -19.025, -18.137, -24.586, -24.250, -23.770, -23.168, -22.471, -21.676,
    -20.823, -19.932, -19.029, -18.140, -24.592, -24.256, -23.776, -23.174,
    -22.476, -21.680, -20.827, -19.936, -19.033, -18.144, -24.598, -24.262,
    -23.781, -23.179, -22.481, -21.685, -20.832, -19.940, -19.037, -18.148,
    -24.603, -24.267, -23.787, -23.184, -22.486, -21.690, -20.836, -19.945,
    -19.041, -18.151, -24.609, -24.273, -23.793, -23.189, -22.491, -21.695,
    -20.841, -19.949, -19.044, -18.155, -24.615, -24.279, -23.798, -23.195,
    -22.496, -21.699, -20.845, -19.953, -19.048, -18.159, -24.621, -24.285,
    -23.804, -23.200, -22.501, -21.704, -20.850, -19.957, -19.052, -18.162,
    -24.627, -24.290, -23.809, -23.205, -22.507, -21.709, -20.854, -19.961,
    -19.056, -18.166, -24.633, -24.296, -23.815, -23.211, -22.512, -21.714,
    -20.859, -19.966, -19.060, -18.170, -24.639, -24.302, -23.820, -23.216,
    -22.517, -21.718, -20.863, -19.970, -19.064, -18.173, -24.644, -24.308,
    -23.826, -23.221, -22.522, -21.723, -20.868, -19.974, -19.068, -18.177,
    -24.650, -24.313, -23.831, -23.227, -22.527, -21.728, -20.872, -19.978,
    -19.072, -18.181, -24.656, -24.319, -23.837, -23.232, -22.532, -21.733,
    -20.877, -19.982, -19.076, -18.185, -24.662, -24.325, -23.843, -23.237,
    -22.537, -21.738, -20.881, -19.987, -19.080, -18.188, -24.668, -24.331,
    -23.848, -23.243, -22.542, -21.742, -20.886, -19.991, -19.084, -18.192,
    -24.674, -24.336, -23.854, -23.248, -22.547, -21.747, -20.890, -19.995,
    -19.088, -18.196, -24.680, -24.342, -23.859, -23.253, -22.552, -21.752,
    -20.895, -19.999, -19.092, -18.199, -24.686, -24.348, -23.865, -23.259,
    -22.557, -21.757, -20.899, -20.003, -19.096, -18.203, -24.691, -24.354,
    -23.870, -23.264, -22.562, -21.761, -20.904, -20.008, -19.100, -18.207,
    -24.697, -24.359, -23.876, -23.269, -22.567, -21.766, -20.908, -20.012,
    -19.103, -18.210, -24.703, -24.365, -23.881, -23.275, -22.572, -21.771,
    -20.913, -20.016, -19.107, -18.214, -24.709, -24.371, -23.887, -23.280,
    -22.577, -21.776, -20.917, -20.020, -19.111, -18.218, -24.715, -24.377,
    -23.892, -23.285, -22.582, -21.781, -20.922, -20.024, -19.115, -18.221,
    -24.721, -24.382, -23.898, -23.291, -22.587, -21.785, -20.926, -20.029,
    -19.119, -18.225, -24.727, -24.388, -23.904, -23.296, -22.592, -21.790,
    -20.930, -20.033, -19.123, -18.229, -24.733, -24.394, -23.909, -23.301,
    -22.597, -21.795, -20.935, -20.037, -19.127, -18.232, -24.738, -24.400,
    -23.915, -23.306, -22.603, -21.800, -20.939, -20.041, -19.131, -18.236,
    -24.744, -24.405, -23.920, -23.312, -22.608, -21.804, -20.944, -20.045,
    -19.135, -18.240, -24.750, -24.411, -23.926, -23.317, -22.613, -21.809,
    -20.948, -20.050, -19.139, -18.244, -24.756, -24.417, -23.931, -23.322,
    -22.618, -21.814, -20.953, -20.054, -19.143, -18.247, -24.762, -24.422,
    -23.937, -23.328, -22.623, -21.819, -20.957, -20.058, -19.147, -18.251,
    -24.768, -24.428, -23.942, -23.333, -22.628, -21.823, -20.962, -20.062,
    -19.151, -18.255, -24.774, -24.434, -23.948, -23.338, -22.633, -21.828,
    -20.966, -20.066, -19.155, -18.258, -24.780, -24.440, -23.954, -23.344,
    -22.638, -21.833, -20.971, -20.071, -19.158, -18.262, -24.785, -24.445,
    -23.959, -23.349, -22.643, -21.838, -20.975, -20.075, -19.162, -18.266,
    -24.791, -24.451, -23.965, -23.354, -22.648, -21.842, -20.980, -20.079,
    -19.166, -18.269, -24.797, -24.457, -23.970, -23.360, -22.653, -21.847,
    -20.984, -20.083, -19.170, -18.273, -24.803, -24.463, -23.976, -23.365,
    -22.658, -21.852, -20.989, -20.087, -19.174, -18.277, -24.809, -24.468,
    -23.981, -23.370, -22.663, -21.857, -20.993, -20.091, -19.178, -18.280,
    -24.815, -24.474, -23.987, -23.376, -22.668, -21.862, -20.998, -20.096,
    -19.182, -18.284, -24.821, -24.480, -23.992, -23.381, -22.673, -21.866,
    -21.002, -20.100, -19.186, -18.288, -24.826, -24.486, -23.998, -23.386,
    -22.678, -21.871, -21.007, -20.104, -19.190, -18.291, -24.832, -24.491,
    -24.003, -23.391, -22.683, -21.876, -21.011, -20.108, -19.194, -18.295,
    -24.838, -24.497, -24.009, -23.397, -22.688, -21.881, -21.016, -20.112,
    -19.198, -18.299, -24.844, -24.503, -24.015, -23.402, -22.693, -21.885,
    -21.020, -20.117, -19.202, -18.302, -24.850, -24.509, -24.020, -23.407,
    -22.698, -21.890, -21.025, -20.121, -19.206, -18.306, -24.856, -24.514,
    -24.026, -23.413, -22.704, -21.895, -21.029, -20.125, -19.210, -18.310,
    -24.862, -24.520, -24.031, -23.418, -22.709, -21.900, -21.034, -20.129,
    -19.213, -18.313, -24.868, -24.526, -24.037, -23.423, -22.714, -21.904,
    -21.038, -20.133, -19.217, -18.317, -24.873, -24.532, -24.042, -23.429,
    -22.719, -21.909, -21.042, -20.138, -19.221, -18.321, -24.879, -24.537,
    -24.048, -23.434, -22.724, -21.914, -21.047, -20.142, -19.225, -18.324,
    -24.885, -24.543, -24.053, -23.439, -22.729, -21.919, -21.051, -20.146,
    -19.229, -18.328, -24.891, -24.549, -24.059, -23.445, -22.734, -21.924,
    -21.056, -20.150, -19.233, -18.332, -24.897, -24.555, -24.064, -23.450,
    -22.739, -21.928, -21.060, -20.154, -19.237, -18.336, -24.903, -24.560,
    -24.070, -23.455, -22.744, -21.933, -21.065, -20.159, -19.241, -18.339,
    -24.909, -24.566, -24.076, -23.461, -22.749, -21.938, -21.069, -20.163,
    -19.245, -18.343, -24.915, -24.572, -24.081, -23.466, -22.754, -21.943,
    -21.074, -20.167, -19.249, -18.347, -24.920, -24.577, -24.087, -23.471,
    -22.759, -21.947, -21.078, -20.171, -19.253, -18.350, -24.926, -24.583,
    -24.092, -23.476, -22.764, -21.952, -21.083, -20.175, -19.257, -18.354,
    -24.932, -24.589, -24.098, -23.482, -22.769, -21.957, -21.087, -20.180,
    -19.261, -18.358, -24.938, -24.595, -24.103, -23.487, -22.774, -21.962,
    -21.092, -20.184, -19.264, -18.361, -24.944, -24.600, -24.109, -23.492,
    -22.779, -21.966, -21.096, -20.188, -19.268, -18.365, -24.950, -24.606,
    -24.114, -23.498, -22.784, -21.971, -21.101, -20.192, -19.272, -18.369,
    -24.956, -24.612, -24.120, -23.503, -22.789, -21.976, -21.105, -20.196,
    -19.276, -18.372, -24.961, -24.618, -24.125, -23.508, -22.794, -21.981,
    -21.110, -20.200, -19.280, -18.376, -24.967, -24.623, -24.131, -23.514,
    -22.799, -21.985, -21.114, -20.205, -19.284, -18.380, -24.973, -24.629,
    -24.137, -23.519, -22.804, -21.990, -21.119, -20.209, -19.288, -18.383,
    -24.979, -24.635, -24.142, -23.524, -22.809, -21.995, -21.123, -20.213,
    -19.292, -18.387, -24.985, -24.641, -24.148, -23.530, -22.815, -22.000,
    -21.127, -20.217, -19.296, -18.391, -24.991, -24.646, -24.153, -23.535,
    -22.820, -22.005, -21.132, -20.221, -19.300, -18.394, -24.997, -24.652,
    -24.159, -23.540, -22.825, -22.009, -21.136, -20.226, -19.304, -18.398,
    -25.003, -24.658, -24.164, -23.545, -22.830, -22.014, -21.141, -20.230,
    -19.308, -18.402, -25.008, -24.663, -24.170, -23.551, -22.835, -22.019,
    -21.145, -20.234, -19.312, -18.405, -25.014, -24.669, -24.175, -23.556,
    -22.840, -22.024, -21.150, -20.238, -19.315, -18.409, -25.020, -24.675,
    -24.181, -23.561, -22.845, -22.028, -21.154, -20.242, -19.319, -18.413,
    -25.026, -24.681, -24.186, -23.567, -22.850, -22.033, -21.159, -20.247,
    -19.323, -18.416, -25.032, -24.686, -24.192, -23.572, -22.855, -22.038,
    -21.163, -20.251, -19.327, -18.420, -25.038, -24.692, -24.198, -23.577,
    -22.860, -22.043, -21.168, -20.255, -19.331, -18.424, -25.044, -24.698,
    -24.203, -23.583, -22.865, -22.047, -21.172, -20.259, -19.335, -18.427,
    -25.049, -24.704, -24.209, -23.588, -22.870, -22.052, -21.177, -20.263,
    -19.339, -18.431, -25.055, -24.709, -24.214, -23.593, -22.875, -22.057,
    -21.181, -20.267, -19.343, -18.435, -25.061, -24.715, -24.220, -23.599,
    -22.880, -22.062, -21.186, -20.272, -19.347, -18.438, -25.067, -24.721,
    -24.225, -23.604, -22.885, -22.066, -21.190, -20.276, -19.351, -18.442,
    -25.073, -24.727, -24.231, -23.609, -22.890, -22.071, -21.195, -20.280,
    -19.355, -18.446, -25.079, -24.732, -24.236, -23.614, -22.895, -22.076,
    -21.199, -20.284, -19.359, -18.449, -25.085, -24.738, -24.242, -23.620,
    -22.900, -22.081, -21.204, -20.288, -19.362, -18.453, -25.090, -24.744,
    -24.247, -23.625, -22.905, -22.085, -21.208, -20.293, -19.366, -18.457,
    -25.096, -24.750, -24.253, -23.630, -22.910, -22.090, -21.212, -20.297,
    -19.370, -18.460, -25.102, -24.755, -24.259, -23.636, -22.915, -22.095,
    -21.217, -20.301, -19.374, -18.464, -25.108, -24.761, -24.264, -23.641,
    -22.920, -22.100, -21.221, -20.305, -19.378, -18.468, -25.114, -24.767,
    -24.270, -23.646, -22.925, -22.104, -21.226, -20.309, -19.382, -18.471,
    -25.120, -24.772, -24.275, -23.652, -22.930, -22.109, -21.230, -20.314,
    -19.386, -18.475, -25.126, -24.778, -24.281, -23.657, -22.936, -22.114,
    -21.235, -20.318, -19.390, -18.479, -25.132, -24.784, -24.286, -23.662,
    -22.941, -22.119, -21.239, -20.322, -19.394, -18.482, -25.137, -24.790,
    -24.292, -23.668, -22.946, -22.124, -21.244, -20.326, -19.398, -18.486,
    -25.143, -24.795, -24.297, -23.673, -22.951, -22.128, -21.248, -20.330,
    -19.402, -18.490, -25.149, -24.801, -24.303, -23.678, -22.956, -22.133,
    -21.253, -20.334, -19.406, -18.493, -25.155, -24.807, -24.308, -23.683,
    -22.961, -22.138, -21.257, -20.339, -19.409, -18.497, -25.161, -24.813,
    -24.314, -23.689, -22.966, -22.143, -21.262, -20.343, -19.413, -18.501,
    -25.167, -24.818, -24.319, -23.694, -22.971, -22.147, -21.266, -20.347,
    -19.417, -18.504, -25.173, -24.824, -24.325, -23.699, -22.976, -22.152,
    -21.271, -20.351, -19.421, -18.508, -25.178, -24.830, -24.331, -23.705,
    -22.981, -22.157, -21.275, -20.355, -19.425, -18.512, -25.184, -24.835,
    -24.336, -23.710, -22.986, -22.162, -21.280, -20.360, -19.429, -18.515,
    -25.190, -24.841, -24.342, -23.715, -22.991, -22.166, -21.284, -20.364,
    -19.433, -18.519, -25.196, -24.847, -24.347, -23.721, -22.996, -22.171,
    -21.288, -20.368, -19.437, -18.523, -25.202, -24.853, -24.353, -23.726,
    -23.001, -22.176, -21.293, -20.372, -19.441, -18.526, -25.208, -24.858,
    -24.358, -23.731, -23.006, -22.181, -21.297, -20.376, -19.445, -18.530,
    -25.214, -24.864, -24.364, -23.736, -23.011, -22.185, -21.302, -20.380,
    -19.449, -18.534, -25.219, -24.870, -24.369, -23.742, -23.016, -22.190,
    -21.306, -20.385, -19.452, -18.537, -25.225, -24.876, -24.375, -23.747,
    -23.021, -22.195, -21.311, -20.389, -19.456, -18.541, -25.231, -24.881,
    -24.380, -23.752, -23.026, -22.200, -21.315, -20.393, -19.460, -18.545,
    -25.237, -24.887, -24.386, -23.758, -23.031, -22.204, -21.320, -20.397,
    -19.464, -18.548, -25.243, -24.893, -24.391, -23.763, -23.036, -22.209,
    -21.324, -20.401, -19.468, -18.552, -25.249, -24.899, -24.397, -23.768,
    -23.041, -22.214, -21.329, -20.406, -19.472, -18.556, -25.255, -24.904,
    -24.403, -23.774, -23.046, -22.219, -21.333, -20.410, -19.476, -18.559,
    -25.261, -24.910, -24.408, -23.779, -23.051, -22.223, -21.338, -20.414,
    -19.480, -18.563, -25.266, -24.916, -24.414, -23.784, -23.056, -22.228,
    -21.342, -20.418, -19.484, -18.567, -25.272, -24.921, -24.419, -23.790,
    -23.062, -22.233, -21.347, -20.422, -19.488, -18.570, -25.278, -24.927,
    -24.425, -23.795, -23.067, -22.238, -21.351, -20.426, -19.492, -18.574,
    -25.284, -24.933, -24.430, -23.800, -23.072, -22.242, -21.355, -20.431,
    -19.496, -18.578, -25.290, -24.939, -24.436, -23.805, -23.077, -22.247,
    -21.360, -20.435, -19.499, -18.581, -25.296, -24.944, -24.441, -23.811,
    -23.082, -22.252, -21.364, -20.439, -19.503, -18.585, -25.302, -24.950,
    -24.447, -23.816, -23.087, -22.257, -21.369, -20.443, -19.507, -18.589,
    -25.307, -24.956, -24.452, -23.821, -23.092, -22.261, -21.373, -20.447,
    -19.511, -18.592, -25.313, -24.962, -24.458, -23.827, -23.097, -22.266,
    -21.378, -20.452, -19.515, -18.596, -25.319, -24.967, -24.463, -23.832,
    -23.102, -22.271, -21.382, -20.456, -19.519, -18.600, -25.325, -24.973,
    -24.469, -23.837, -23.107, -22.276, -21.387, -20.460, -19.523, -18.603,
    -25.331, -24.979, -24.474, -23.843, -23.112, -22.281, -21.391, -20.464,
    -19.527, -18.607, -25.337, -24.984, -24.480, -23.848, -23.117, -22.285,
    -21.396, -20.468, -19.531, -18.611, -25.343, -24.990, -24.486, -23.853,
    -23.122, -22.290, -21.400, -20.472, -19.535, -18.614, -25.348, -24.996,
    -24.491, -23.858, -23.127, -22.295, -21.405, -20.477, -19.539, -18.618,
    -25.354, -25.002, -24.497, -23.864, -23.132, -22.300, -21.409, -20.481,
    -19.542, -18.622, -25.360, -25.007, -24.502, -23.869, -23.137, -22.304,
    -21.414, -20.485, -19.546, -18.625, -25.366, -25.013, -24.508, -23.874,
    -23.142, -22.309, -21.418, -20.489, -19.550, -18.629, -25.372, -25.019,
    -24.513, -23.880, -23.147, -22.314, -21.422, -20.493, -19.554, -18.632,
    -25.378, -25.025, -24.519, -23.885, -23.152, -22.319, -21.427, -20.498,
    -19.558, -18.636, -25.384, -25.030, -24.524, -23.890, -23.157, -22.323,
    -21.431, -20.502, -19.562, -18.640, -25.389, -25.036, -24.530, -23.896,
    -23.162, -22.328, -21.436, -20.506, -19.566, -18.643, -25.395, -25.042,
    -24.535, -23.901, -23.167, -22.333, -21.440, -20.510, -19.570, -18.647,
    -25.401, -25.047, -24.541, -23.906, -23.172, -22.338, -21.445, -20.514,
    -19.574, -18.651, -25.407, -25.053, -24.546, -23.911, -23.177, -22.342,
    -21.449, -20.518, -19.578, -18.654, -25.413, -25.059, -24.552, -23.917,
    -23.182, -22.347, -21.454, -20.523, -19.582, -18.658, -25.419, -25.065,
    -24.558, -23.922, -23.187, -22.352, -21.458, -20.527, -19.585, -18.662,
    -25.425, -25.070, -24.563, -23.927, -23.192, -22.357, -21.463, -20.531,
    -19.589, -18.665, -25.430, -25.076, -24.569, -23.933, -23.197, -22.361,
    -21.467, -20.535, -19.593, -18.669, -25.436, -25.082, -24.574, -23.938,
    -23.202, -22.366, -21.472, -20.539, -19.597, -18.673, -25.442, -25.088,
    -24.580, -23.943, -23.208, -22.371, -21.476, -20.543, -19.601, -18.676,
    -25.448, -25.093, -24.585, -23.948, -23.213, -22.376, -21.480, -20.548,
    -19.605, -18.680, -25.454, -25.099, -24.591, -23.954, -23.218, -22.380,
    -21.485, -20.552, -19.609, -18.684, -25.460, -25.105, -24.596, -23.959,
    -23.223, -22.385, -21.489, -20.556, -19.613, -18.687, -25.466, -25.110,
    -24.602, -23.964, -23.228, -22.390, -21.494, -20.560, -19.617, -18.691,
    -25.471, -25.116, -24.607, -23.970, -23.233, -22.395, -21.498, -20.564,
    -19.621, -18.695, -25.477, -25.122, -24.613, -23.975, -23.238, -22.399,
    -21.503, -20.569, -19.624, -18.698, -25.483, -25.128, -24.618, -23.980,
    -23.243, -22.404, -21.507, -20.573, -19.628, -18.702, -25.489, -25.133,
    -24.624, -23.986, -23.248, -22.409, -21.512, -20.577, -19.632, -18.706,
    -25.495, -25.139, -24.629, -23.991, -23.253, -22.414, -21.516, -20.581,
    -19.636, -18.709, -25.501, -25.145, -24.635, -23.996, -23.258, -22.418,
    -21.521, -20.585, -19.640, -18.713, -25.507, -25.151, -24.640, -24.001,
    -23.263, -22.423, -21.525, -20.589, -19.644, -18.717, -25.512, -25.156,
    -24.646, -24.007, -23.268, -22.428, -21.530, -20.594, -19.648, -18.720,
    -25.518, -25.162, -24.652, -24.012, -23.273, -22.433, -21.534, -20.598,
    -19.652, -18.724, -25.524, -25.168, -24.657, -24.017, -23.278, -22.437,
    -21.538, -20.602, -19.656, -18.728, -25.530, -25.173, -24.663, -24.023,
    -23.283, -22.442, -21.543, -20.606, -19.660, -18.731, -25.536, -25.179,
    -24.668, -24.028, -23.288, -22.447, -21.547, -20.610, -19.664, -18.735,
    -25.542, -25.185, -24.674, -24.033, -23.293, -22.452, -21.552, -20.614,
    -19.667, -18.738, -25.548, -25.191, -24.679, -24.039, -23.298, -22.456,
    -21.556, -20.619, -19.671, -18.742, -25.553, -25.196, -24.685, -24.044,
    -23.303, -22.461, -21.561, -20.623, -19.675, -18.746, -25.559, -25.202,
    -24.690, -24.049, -23.308, -22.466, -21.565, -20.627, -19.679, -18.749,
    -25.565, -25.208, -24.696, -24.054, -23.313, -22.471, -21.570, -20.631,
    -19.683, -18.753, -25.571, -25.213, -24.701, -24.060, -23.318, -22.475,
    -21.574, -20.635, -19.687, -18.757, -25.577, -25.219, -24.707, -24.065,
    -23.323, -22.480, -21.579, -20.640, -19.691, -18.760, -25.583, -25.225,
    -24.712, -24.070, -23.328, -22.485, -21.583, -20.644, -19.695, -18.764,
    -25.589, -25.231, -24.718, -24.076, -23.333, -22.490, -21.588, -20.648,
    -19.699, -18.768, -25.594, -25.236, -24.723, -24.081, -23.338, -22.494,
    -21.592, -20.652, -19.703, -18.771, -25.600, -25.242, -24.729, -24.086,
    -23.343, -22.499, -21.596, -20.656, -19.706, -18.775, -25.606, -25.248,
    -24.734, -24.091, -23.348, -22.504, -21.601, -20.660, -19.710, -18.779,
    -25.612, -25.254, -24.740, -24.097, -23.353, -22.509, -21.605, -20.665,
    -19.714, -18.782, -25.618, -25.259, -24.746, -24.102, -23.358, -22.513,
    -21.610, -20.669, -19.718, -18.786, -25.624, -25.265, -24.751, -24.107,
    -23.363, -22.518, -21.614, -20.673, -19.722, -18.790, -25.630, -25.271,
    -24.757, -24.113, -23.369, -22.523, -21.619, -20.677, -19.726, -18.793,
    -25.635, -25.276, -24.762, -24.118, -23.374, -22.528, -21.623, -20.681,
    -19.730, -18.797, -25.641, -25.282, -24.768, -24.123, -23.379, -22.532,
    -21.628, -20.685, -19.734, -18.801, -25.647, -25.288, -24.773, -24.128,
    -23.384, -22.537, -21.632, -20.690, -19.738, -18.804, -25.653, -25.294,
    -24.779, -24.134, -23.389, -22.542, -21.637, -20.694, -19.742, -18.808,
    -25.659, -25.299, -24.784, -24.139, -23.394, -22.547, -21.641, -20.698,
    -19.745, -18.811, -25.665, -25.305, -24.790, -24.144, -23.399, -22.551,
    -21.645, -20.702, -19.749, -18.815, -25.670, -25.311, -24.795, -24.150,
    -23.404, -22.556, -21.650, -20.706, -19.753, -18.819, -25.676, -25.316,
    -24.801, -24.155, -23.409, -22.561, -21.654, -20.710, -19.757, -18.822,
    -25.682, -25.322, -24.806, -24.160, -23.414, -22.566, -21.659, -20.715,
    -19.761, -18.826, -25.688, -25.328, -24.812, -24.166, -23.419, -22.570,
    -21.663, -20.719, -19.765, -18.830, -25.694, -25.334, -24.817, -24.171,
    -23.424, -22.575, -21.668, -20.723, -19.769, -18.833, -25.700, -25.339,
    -24.823, -24.176, -23.429, -22.580, -21.672, -20.727, -19.773, -18.837,
    -25.706, -25.345, -24.828, -24.181, -23.434, -22.585, -21.677, -20.731,
    -19.777, -18.841, -25.711, -25.351, -24.834, -24.187, -23.439, -22.589,
    -21.681, -20.735, -19.781, -18.844, -25.717, -25.357, -24.840, -24.192,
    -23.444, -22.594, -21.686, -20.740, -19.784, -18.848, -25.723, -25.362,
    -24.845, -24.197, -23.449, -22.599, -21.690, -20.744, -19.788, -18.852,
    -25.729, -25.368, -24.851, -24.203, -23.454, -22.603, -21.695, -20.748,
    -19.792, -18.855, -25.735, -25.374, -24.856, -24.208, -23.459, -22.608,
    -21.699, -20.752, -19.796, -18.859, -25.741, -25.379, -24.862, -24.213,
    -23.464, -22.613, -21.703, -20.756, -19.800, -18.862, -25.747, -25.385,
    -24.867, -24.218, -23.469, -22.618, -21.708, -20.760, -19.804, -18.866,
    -25.752, -25.391, -24.873, -24.224, -23.474, -22.622, -21.712, -20.765,
    -19.808, -18.870, -25.758, -25.397, -24.878, -24.229, -23.479, -22.627,
    -21.717, -20.769, -19.812, -18.873, -25.764, -25.402, -24.884, -24.234,
    -23.484, -22.632, -21.721, -20.773, -19.816, -18.877, -25.770, -25.408,
    -24.889, -24.240, -23.489, -22.637, -21.726, -20.777, -19.820, -18.881,
    -25.776, -25.414, -24.895, -24.245, -23.494, -22.641, -21.730, -20.781,
    -19.823, -18.884, -25.782, -25.419, -24.900, -24.250, -23.499, -22.646,
    -21.735, -20.785, -19.827, -18.888, -25.788, -25.425, -24.906, -24.255,
    -23.504, -22.651, -21.739, -20.790, -19.831, -18.892, -25.793, -25.431,
    -24.911, -24.261, -23.509, -22.656, -21.744, -20.794, -19.835, -18.895,
    -25.799, -25.437, -24.917, -24.266, -23.514, -22.660, -21.748, -20.798,
    -19.839, -18.899, -25.805, -25.442, -24.922, -24.271, -23.519, -22.665,
    -21.752, -20.802, -19.843, -18.903, -25.811, -25.448, -24.928, -24.277,
    -23.524, -22.670, -21.757, -20.806, -19.847, -18.906, -25.817, -25.454,
    -24.933, -24.282, -23.529, -22.675, -21.761, -20.811, -19.851, -18.910,
    -25.823, -25.459, -24.939, -24.287, -23.534, -22.679, -21.766, -20.815,
    -19.855, -18.913, -25.828, -25.465, -24.944, -24.292, -23.539, -22.684,
    -21.770, -20.819, -19.858, -18.917, -25.834, -25.471, -24.950, -24.298,
    -23.544, -22.689, -21.775, -20.823, -19.862, -18.921, -25.840, -25.477,
    -24.956, -24.303, -23.549, -22.694, -21.779, -20.827, -19.866, -18.924,
    -25.846, -25.482, -24.961, -24.308, -23.554, -22.698, -21.784, -20.831,
    -19.870, -18.928, -25.852, -25.488, -24.967, -24.314, -23.559, -22.703,
    -21.788, -20.836, -19.874, -18.932, -25.858, -25.494, -24.972, -24.319,
    -23.565, -22.708, -21.793, -20.840, -19.878, -18.935, -25.864, -25.499,
    -24.978, -24.324, -23.570, -22.713, -21.797, -20.844, -19.882, -18.939,
    -25.869, -25.505, -24.983, -24.329, -23.575, -22.717, -21.801, -20.848,
    -19.886, -18.943, -25.875, -25.511, -24.989, -24.335, -23.580, -22.722,
    -21.806, -20.852, -19.890, -18.946, -25.881, -25.517, -24.994, -24.340,
    -23.585, -22.727, -21.810, -20.856, -19.894, -18.950, -25.887, -25.522,
    -25.000, -24.345, -23.590, -22.732, -21.815, -20.861, -19.897, -18.954,
    -25.893, -25.528, -25.005, -24.351, -23.595, -22.736, -21.819, -20.865,
    -19.901, -18.957, -25.899, -25.534, -25.011, -24.356, -23.600, -22.741,
    -21.824, -20.869, -19.905, -18.961, -25.905, -25.539, -25.016, -24.361,
    -23.605, -22.746, -21.828, -20.873, -19.909, -18.964, -25.910, -25.545,
    -25.022, -24.366, -23.610, -22.751, -21.833, -20.877, -19.913, -18.968,
    -25.916, -25.551, -25.027, -24.372, -23.615, -22.755, -21.837, -20.881,
    -19.917, -18.972, -25.922, -25.557, -25.033, -24.377, -23.620, -22.760,
    -21.842, -20.885, -19.921, -18.975, -25.928, -25.562, -25.038, -24.382,
    -23.625, -22.765, -21.846, -20.890, -19.925, -18.979, -25.934, -25.568,
    -25.044, -24.388, -23.630, -22.770, -21.850, -20.894, -19.929, -18.983,
    -25.940, -25.574, -25.049, -24.393, -23.635, -22.774, -21.855, -20.898,
    -19.932, -18.986, -25.945, -25.579, -25.055, -24.398, -23.640, -22.779,
    -21.859, -20.902, -19.936, -18.990, -25.951, -25.585, -25.060, -24.403,
    -23.645, -22.784, -21.864, -20.906, -19.940, -18.994, -25.957, -25.591,
    -25.066, -24.409, -23.650, -22.788, -21.868, -20.910, -19.944, -18.997,
    -25.963, -25.597, -25.071, -24.414, -23.655, -22.793, -21.873, -20.915,
    -19.948, -19.001, -25.969, -25.602, -25.077, -24.419, -23.660, -22.798,
    -21.877, -20.919, -19.952, -19.004, -25.975, -25.608, -25.083, -24.425,
    -23.665, -22.803, -21.882, -20.923, -19.956, -19.008, -25.981, -25.614,
    -25.088, -24.430, -23.670, -22.807, -21.886, -20.927, -19.960, -19.012,
    -25.986, -25.619, -25.094, -24.435, -23.675, -22.812, -21.890, -20.931,
    -19.964, -19.015, -25.992, -25.625, -25.099, -24.440, -23.680, -22.817,
    -21.895, -20.935, -19.967, -19.019, -25.998, -25.631, -25.105, -24.446,
    -23.685, -22.822, -21.899, -20.940, -19.971, -19.023, -26.004, -25.637,
    -25.110, -24.451, -23.690, -22.826, -21.904, -20.944, -19.975, -19.026,
    -26.010, -25.642, -25.116, -24.456, -23.695, -22.831, -21.908, -20.948,
    -19.979, -19.030, -26.016, -25.648, -25.121, -24.462, -23.700, -22.836,
    -21.913, -20.952, -19.983, -19.034, -26.022, -25.654, -25.127, -24.467,
    -23.705, -22.841, -21.917, -20.956, -19.987, -19.037, -26.027, -25.659,
    -25.132, -24.472, -23.710, -22.845, -21.922, -20.960, -19.991, -19.041,
    -26.033, -25.665, -25.138, -24.477, -23.715, -22.850, -21.926, -20.965,
    -19.995, -19.044, -26.039, -25.671, -25.143, -24.483, -23.720, -22.855,
    -21.931, -20.969, -19.999, -19.048, -26.045, -25.677, -25.149, -24.488,
    -23.725, -22.860, -21.935, -20.973, -20.002, -19.052, -26.051, -25.682,
    -25.154, -24.493, -23.730, -22.864, -21.939, -20.977, -20.006, -19.055,
    -26.057, -25.688, -25.160, -24.499, -23.735, -22.869, -21.944, -20.981,
    -20.010, -19.059, -26.062, -25.694, -25.165, -24.504, -23.740, -22.874,
    -21.948, -20.985, -20.014, -19.063, -26.068, -25.699, -25.171, -24.509,
    -23.745, -22.878, -21.953, -20.990, -20.018, -19.066, -26.074, -25.705,
    -25.176, -24.514, -23.750, -22.883, -21.957, -20.994, -20.022, -19.070,
    -26.080, -25.711, -25.182, -24.520, -23.755, -22.888, -21.962, -20.998,
    -20.026, -19.074, -26.086, -25.717, -25.187, -24.525, -23.760, -22.893,
    -21.966, -21.002, -20.030, -19.077, -26.092, -25.722, -25.193, -24.530,
    -23.765, -22.897, -21.971, -21.006, -20.034, -19.081, -26.098, -25.728,
    -25.198, -24.536, -23.770, -22.902, -21.975, -21.010, -20.037, -19.084,
    -26.103, -25.734, -25.204, -24.541, -23.775, -22.907, -21.979, -21.015,
    -20.041, -19.088, -26.109, -25.739, -25.209, -24.546, -23.780, -22.912,
    -21.984, -21.019, -20.045, -19.092, -26.115, -25.745, -25.215, -24.551,
    -23.785, -22.916, -21.988, -21.023, -20.049, -19.095, -26.121, -25.751,
    -25.220, -24.557, -23.790, -22.921, -21.993, -21.027, -20.053, -19.099,
    -26.127, -25.757, -25.226, -24.562, -23.795, -22.926, -21.997, -21.031,
    -20.057, -19.103, -26.133, -25.762, -25.231, -24.567, -23.800, -22.931,
    -22.002, -21.035, -20.061, -19.106, -26.138, -25.768, -25.237, -24.573,
    -23.805, -22.935, -22.006, -21.040, -20.065, -19.110, -26.144, -25.774,
    -25.243, -24.578, -23.810, -22.940, -22.011, -21.044, -20.069, -19.113,
    -26.150, -25.779, -25.248, -24.583, -23.816, -22.945, -22.015, -21.048,
    -20.072, -19.117, -26.156, -25.785, -25.254, -24.588, -23.821, -22.950,
    -22.019, -21.052, -20.076, -19.121, -26.162, -25.791, -25.259, -24.594,
    -23.826, -22.954, -22.024, -21.056, -20.080, -19.124, -26.168, -25.797,
    -25.265, -24.599, -23.831, -22.959, -22.028, -21.060, -20.084, -19.128,
};

const double CIA_Orton::H2CH4_e[][10] = {
    -27.768, -27.867, -28.002, -28.211, -28.479, -28.782, -29.103, -29.436,
    -29.775, -30.115, -25.935, -26.034, -26.169, -26.379, -26.646, -26.949,
    -27.271, -27.604, -27.943, -28.282, -24.549, -24.648, -24.782, -24.992,
    -25.260, -25.563, -25.884, -26.217, -26.556, -26.896, -23.162, -23.261,
    -23.395, -23.605, -23.873, -24.176, -24.498, -24.831, -25.170, -25.510,
    -22.347, -22.446, -22.582, -22.792, -23.061, -23.364, -23.686, -24.019,
    -24.358, -24.698, -21.766, -21.865, -22.001, -22.212, -22.482, -22.786,
    -23.109, -23.443, -23.782, -24.122, -21.314, -21.412, -21.549, -21.762,
    -22.033, -22.337, -22.661, -22.995, -23.335, -23.675, -20.943, -21.041,
    -21.179, -21.393, -21.665, -21.970, -22.294, -22.629, -22.969, -23.310,
    -20.631, -20.729, -20.867, -21.082, -21.354, -21.660, -21.985, -22.320,
    -22.660, -23.001, -20.359, -20.457, -20.596, -20.812, -21.084, -21.391,
    -21.716, -22.052, -22.392, -22.734, -20.114, -20.211, -20.352, -20.569,
    -20.843, -21.152, -21.478, -21.814, -22.155, -22.497, -19.894, -19.990,
    -20.132, -20.351, -20.627, -20.936, -21.264, -21.601, -21.943, -22.285,
    -19.694, -19.790, -19.933, -20.153, -20.431, -20.742, -21.070, -21.408,
    -21.750, -22.093, -19.511, -19.606, -19.750, -19.972, -20.251, -20.564,
    -20.893, -21.232, -21.575, -21.918, -19.340, -19.435, -19.580, -19.804,
    -20.085, -20.399, -20.729, -21.069, -21.413, -21.756, -19.178, -19.273,
    -19.419, -19.645, -19.929, -20.244, -20.577, -20.917, -21.262, -21.606,
    -19.018, -19.115, -19.265, -19.495, -19.782, -20.100, -20.433, -20.776,
    -21.121, -21.466, -18.874, -18.972, -19.123, -19.356, -19.644, -19.964,
    -20.300, -20.643, -20.990, -21.335, -18.738, -18.835, -18.989, -19.224,
    -19.515, -19.836, -20.174, -20.518, -20.866, -21.212, -18.608, -18.706,
    -18.861, -19.099, -19.392, -19.716, -20.054, -20.400, -20.749, -21.096,
    -18.485, -18.584, -18.741, -18.980, -19.276, -19.601, -19.942, -20.289,
    -20.638, -20.986, -18.367, -18.466, -18.625, -18.867, -19.164, -19.492,
    -19.834, -20.182, -20.533, -20.881, -17.835, -17.935, -18.103, -18.357,
    -18.668, -19.006, -19.357, -19.713, -20.069, -20.422, -17.398, -17.496,
    -17.671, -17.935, -18.255, -18.604, -18.963, -19.326, -19.688, -20.045,
    -17.010, -17.105, -17.285, -17.558, -17.890, -18.249, -18.618, -18.990,
    -19.358, -19.721, -16.684, -16.777, -16.958, -17.236, -17.575, -17.942,
    -18.319, -18.697, -19.072, -19.439, -16.393, -16.483, -16.665, -16.948,
    -17.294, -17.668, -18.052, -18.436, -18.816, -19.189, -16.155, -16.243,
    -16.423, -16.707, -17.055, -17.433, -17.821, -18.209, -18.593, -18.968,
    -15.940, -16.025, -16.204, -16.488, -16.838, -17.219, -17.610, -18.002,
    -18.389, -18.768, -15.767, -15.849, -16.025, -16.306, -16.655, -17.036,
    -17.428, -17.821, -18.209, -18.589, -15.602, -15.679, -15.851, -16.131,
    -16.480, -16.862, -17.256, -17.651, -18.041, -18.422, -15.462, -15.536,
    -15.705, -15.982, -16.330, -16.711, -17.105, -17.500, -17.890, -18.272,
    -15.350, -15.426, -15.592, -15.865, -16.208, -16.585, -16.976, -17.369,
    -17.757, -18.138, -15.257, -15.340, -15.504, -15.772, -16.110, -16.482,
    -16.867, -17.256, -17.641, -18.018, -15.182, -15.272, -15.434, -15.697,
    -16.027, -16.393, -16.772, -17.155, -17.536, -17.909, -15.110, -15.209,
    -15.371, -15.628, -15.952, -16.311, -16.684, -17.061, -17.437, -17.807,
    -15.030, -15.136, -15.299, -15.554, -15.874, -16.227, -16.596, -16.969,
    -17.341, -17.708, -14.934, -15.050, -15.217, -15.473, -15.791, -16.143,
    -16.508, -16.879, -17.248, -17.613, -14.793, -14.929, -15.101, -15.353,
    -15.662, -16.004, -16.360, -16.721, -17.083, -17.442, -14.649, -14.799,
    -14.976, -15.226, -15.530, -15.866, -16.215, -16.571, -16.927, -17.281,
    -14.464, -14.611, -14.792, -15.047, -15.357, -15.696, -16.048, -16.406,
    -16.764, -17.119, -14.312, -14.457, -14.639, -14.896, -15.207, -15.547,
    -15.900, -16.258, -16.616, -16.971, -14.205, -14.353, -14.532, -14.785,
    -15.092, -15.428, -15.777, -16.132, -16.487, -16.840, -14.058, -14.193,
    -14.370, -14.626, -14.937, -15.279, -15.633, -15.991, -16.349, -16.704,
    -13.953, -14.074, -14.245, -14.499, -14.811, -15.153, -15.508, -15.867,
    -16.226, -16.581, -13.882, -13.990, -14.152, -14.399, -14.707, -15.046,
    -15.398, -15.756, -16.114, -16.468, -13.764, -13.856, -14.012, -14.260,
    -14.571, -14.914, -15.272, -15.633, -15.995, -16.352, -13.686, -13.777,
    -13.929, -14.173, -14.481, -14.822, -15.177, -15.537, -15.896, -16.252,
    -13.646, -13.743, -13.891, -14.126, -14.424, -14.757, -15.104, -15.458,
    -15.813, -16.165, -13.598, -13.703, -13.849, -14.078, -14.369, -14.694,
    -15.035, -15.384, -15.734, -16.083, -13.591, -13.724, -13.871, -14.088,
    -14.363, -14.672, -14.999, -15.336, -15.676, -16.017, -13.561, -13.729,
    -13.883, -14.092, -14.353, -14.647, -14.961, -15.287, -15.618, -15.951,
    -13.520, -13.709, -13.868, -14.070, -14.322, -14.607, -14.912, -15.230,
    -15.555, -15.884, -13.477, -13.688, -13.852, -14.049, -14.292, -14.568,
    -14.864, -15.175, -15.494, -15.818, -13.410, -13.620, -13.786, -13.984,
    -14.226, -14.501, -14.796, -15.106, -15.424, -15.746, -13.390, -13.583,
    -13.739, -13.932, -14.170, -14.442, -14.735, -15.042, -15.358, -15.679,
    -13.366, -13.547, -13.695, -13.883, -14.118, -14.386, -14.676, -14.981,
    -15.294, -15.613, -13.345, -13.513, -13.654, -13.836, -14.067, -14.333,
    -14.620, -14.922, -15.234, -15.551, -13.324, -13.484, -13.618, -13.795,
    -14.022, -14.283, -14.567, -14.867, -15.176, -15.491, -13.273, -13.437,
    -13.572, -13.749, -13.974, -14.233, -14.515, -14.812, -15.119, -15.433,
    -13.211, -13.386, -13.526, -13.703, -13.927, -14.185, -14.465, -14.759,
    -15.065, -15.377, -13.142, -13.331, -13.479, -13.659, -13.882, -14.138,
    -14.416, -14.709, -15.013, -15.322, -13.079, -13.282, -13.436, -13.617,
    -13.840, -14.094, -14.370, -14.661, -14.962, -15.270, -12.999, -13.222,
    -13.387, -13.573, -13.797, -14.051, -14.325, -14.614, -14.913, -15.220,
    -12.946, -13.180, -13.349, -13.536, -13.759, -14.010, -14.282, -14.569,
    -14.867, -15.171, -12.914, -13.150, -13.319, -13.504, -13.724, -13.973,
    -14.242, -14.526, -14.822, -15.124, -12.879, -13.119, -13.288, -13.472,
    -13.690, -13.937, -14.203, -14.485, -14.778, -15.079, -12.856, -13.096,
    -13.263, -13.444, -13.659, -13.903, -14.167, -14.446, -14.737, -15.035,
    -12.819, -13.064, -13.234, -13.414, -13.628, -13.869, -14.131, -14.408,
    -14.696, -14.992, -12.774, -13.028, -13.203, -13.384, -13.597, -13.837,
    -14.096, -14.371, -14.657, -14.951, -12.739, -12.997, -13.174, -13.356,
    -13.568, -13.806, -14.063, -14.335, -14.619, -14.911, -12.690, -12.954,
    -13.138, -13.324, -13.536, -13.774, -14.030, -14.300, -14.582, -14.872,
    -12.631, -12.907, -13.100, -13.291, -13.505, -13.743, -13.998, -14.266,
    -14.546, -14.835, -12.578, -12.861, -13.062, -13.258, -13.475, -13.712,
    -13.966, -14.233, -14.511, -14.798, -12.523, -12.816, -13.026, -13.227,
    -13.446, -13.684, -13.936, -14.202, -14.478, -14.762, -12.484, -12.782,
    -12.997, -13.201, -13.420, -13.657, -13.908, -14.172, -14.446, -14.728,
    -12.470, -12.764, -12.978, -13.180, -13.398, -13.633, -13.882, -14.143,
    -14.415, -14.695, -12.453, -12.744, -12.958, -13.160, -13.376, -13.610,
    -13.856, -14.115, -14.385, -14.663, -12.447, -12.733, -12.943, -13.143,
    -13.357, -13.588, -13.832, -14.089, -14.356, -14.631, -12.415, -12.705,
    -12.919, -13.121, -13.335, -13.565, -13.808, -14.062, -14.327, -14.601,
    -12.388, -12.679, -12.896, -13.099, -13.314, -13.544, -13.785, -14.037,
    -14.300, -14.571, -12.362, -12.654, -12.873, -13.078, -13.293, -13.522,
    -13.762, -14.012, -14.273, -14.542, -12.328, -12.626, -12.849, -13.056,
    -13.272, -13.501, -13.739, -13.988, -14.247, -14.514, -12.304, -12.602,
    -12.827, -13.036, -13.252, -13.480, -13.717, -13.964, -14.221, -14.486,
    -12.271, -12.576, -12.806, -13.016, -13.233, -13.460, -13.696, -13.941,
    -14.196, -14.459, -12.262, -12.565, -12.793, -13.002, -13.217, -13.442,
    -13.676, -13.920, -14.172, -14.433, -12.279, -12.568, -12.788, -12.993,
    -13.205, -13.427, -13.659, -13.899, -14.150, -14.408, -12.301, -12.576,
    -12.787, -12.986, -13.194, -13.413, -13.642, -13.880, -14.127, -14.384,
    -12.306, -12.571, -12.778, -12.975, -13.181, -13.398, -13.625, -13.861,
    -14.106, -14.360, -12.306, -12.563, -12.767, -12.962, -13.167, -13.383,
    -13.608, -13.842, -14.085, -14.337, -12.278, -12.540, -12.749, -12.946,
    -13.152, -13.367, -13.591, -13.823, -14.064, -14.314, -12.257, -12.522,
    -12.733, -12.932, -13.137, -13.352, -13.575, -13.805, -14.044, -14.292,
    -12.205, -12.483, -12.703, -12.908, -13.118, -13.334, -13.557, -13.786,
    -14.024, -14.270, -12.152, -12.444, -12.674, -12.886, -13.099, -13.316,
    -13.539, -13.768, -14.004, -14.248, -12.103, -12.411, -12.650, -12.867,
    -13.082, -13.300, -13.523, -13.750, -13.985, -14.227, -12.055, -12.381,
    -12.630, -12.851, -13.068, -13.286, -13.507, -13.734, -13.966, -14.207,
    -12.025, -12.363, -12.617, -12.840, -13.056, -13.274, -13.494, -13.718,
    -13.949, -14.187, -12.036, -12.374, -12.624, -12.841, -13.053, -13.266,
    -13.483, -13.704, -13.932, -14.168, -12.016, -12.365, -12.618, -12.835,
    -13.045, -13.257, -13.471, -13.690, -13.916, -14.150, -12.024, -12.376,
    -12.626, -12.837, -13.043, -13.250, -13.461, -13.677, -13.901, -14.132,
    -11.982, -12.345, -12.604, -12.821, -13.029, -13.237, -13.448, -13.663,
    -13.884, -14.114, -11.947, -12.319, -12.585, -12.806, -13.016, -13.225,
    -13.435, -13.649, -13.869, -14.096, -11.917, -12.295, -12.567, -12.792,
    -13.003, -13.212, -13.422, -13.634, -13.853, -14.079, -11.885, -12.269,
    -12.547, -12.775, -12.989, -13.199, -13.408, -13.620, -13.837, -14.062,
    -11.855, -12.242, -12.524, -12.757, -12.974, -13.185, -13.394, -13.606,
    -13.822, -14.045, -11.830, -12.217, -12.504, -12.741, -12.959, -13.171,
    -13.381, -13.592, -13.807, -14.028, -11.814, -12.200, -12.488, -12.726,
    -12.946, -13.159, -13.369, -13.579, -13.793, -14.012, -11.813, -12.194,
    -12.480, -12.718, -12.938, -13.150, -13.358, -13.567, -13.779, -13.997,
    -11.820, -12.194, -12.477, -12.713, -12.931, -13.142, -13.349, -13.556,
    -13.766, -13.982, -11.809, -12.185, -12.469, -12.705, -12.923, -13.133,
    -13.339, -13.545, -13.753, -13.967, -11.793, -12.174, -12.461, -12.698,
    -12.915, -13.125, -13.329, -13.534, -13.741, -13.953, -11.775, -12.159,
    -12.448, -12.687, -12.905, -13.114, -13.319, -13.522, -13.728, -13.938,
    -11.765, -12.149, -12.439, -12.678, -12.896, -13.105, -13.309, -13.511,
    -13.715, -13.924, -11.747, -12.132, -12.423, -12.663, -12.882, -13.092,
    -13.297, -13.499, -13.702, -13.910, -11.732, -12.118, -12.411, -12.652,
    -12.871, -13.081, -13.285, -13.487, -13.690, -13.896, -11.733, -12.115,
    -12.403, -12.642, -12.861, -13.071, -13.275, -13.476, -13.677, -13.882,
    -11.744, -12.120, -12.405, -12.640, -12.856, -13.064, -13.266, -13.466,
    -13.666, -13.869, -11.768, -12.136, -12.412, -12.642, -12.854, -13.059,
    -13.259, -13.457, -13.655, -13.857, -11.845, -12.183, -12.437, -12.653,
    -12.857, -13.057, -13.254, -13.449, -13.645, -13.845, -11.903, -12.216,
    -12.454, -12.660, -12.857, -13.053, -13.247, -13.441, -13.635, -13.833,
    -11.990, -12.263, -12.477, -12.669, -12.859, -13.051, -13.242, -13.433,
    -13.625, -13.821, -11.976, -12.251, -12.466, -12.660, -12.850, -13.042,
    -13.233, -13.423, -13.614, -13.809, -11.960, -12.237, -12.455, -12.650,
    -12.841, -13.033, -13.224, -13.414, -13.604, -13.797, -11.926, -12.211,
    -12.436, -12.636, -12.830, -13.024, -13.215, -13.404, -13.594, -13.786,
    -11.864, -12.166, -12.406, -12.616, -12.816, -13.012, -13.204, -13.394,
    -13.583, -13.774, -11.800, -12.119, -12.373, -12.593, -12.799, -12.999,
    -13.193, -13.383, -13.572, -13.762, -11.738, -12.073, -12.340, -12.569,
    -12.782, -12.986, -13.182, -13.373, -13.561, -13.751, -11.692, -12.038,
    -12.314, -12.551, -12.768, -12.974, -13.172, -13.363, -13.551, -13.740,
    -11.683, -12.029, -12.305, -12.542, -12.760, -12.967, -13.165, -13.355,
    -13.542, -13.729, -11.693, -12.032, -12.303, -12.539, -12.756, -12.962,
    -13.159, -13.348, -13.534, -13.719, -11.708, -12.039, -12.306, -12.538,
    -12.754, -12.958, -13.153, -13.341, -13.525, -13.709, -11.725, -12.047,
    -12.309, -12.539, -12.752, -12.954, -13.148, -13.335, -13.517, -13.700,
    -11.705, -12.031, -12.297, -12.529, -12.744, -12.947, -13.140, -13.326,
    -13.508, -13.690, -11.682, -12.015, -12.286, -12.522, -12.737, -12.940,
    -13.133, -13.319, -13.500, -13.680, -11.652, -11.994, -12.270, -12.509,
    -12.727, -12.931, -13.124, -13.310, -13.490, -13.670, -11.617, -11.970,
    -12.254, -12.497, -12.717, -12.923, -13.116, -13.301, -13.481, -13.660,
    -11.600, -11.957, -12.243, -12.488, -12.709, -12.915, -13.108, -13.293,
    -13.472, -13.650, -11.594, -11.952, -12.240, -12.485, -12.706, -12.910,
    -13.102, -13.286, -13.464, -13.641, -11.602, -11.958, -12.244, -12.487,
    -12.705, -12.907, -13.098, -13.280, -13.457, -13.632, -11.618, -11.970,
    -12.253, -12.493, -12.708, -12.907, -13.095, -13.275, -13.450, -13.624,
    -11.636, -11.985, -12.264, -12.500, -12.710, -12.906, -13.092, -13.270,
    -13.444, -13.615, -11.640, -11.989, -12.268, -12.502, -12.711, -12.905,
    -13.089, -13.265, -13.437, -13.607, -11.626, -11.981, -12.263, -12.498,
    -12.706, -12.899, -13.082, -13.258, -13.429, -13.598, -11.629, -11.982,
    -12.261, -12.494, -12.701, -12.894, -13.076, -13.251, -13.422, -13.590,
    -11.630, -11.979, -12.256, -12.488, -12.695, -12.888, -13.070, -13.244,
    -13.414, -13.581, -11.600, -11.954, -12.235, -12.472, -12.683, -12.877,
    -13.061, -13.236, -13.405, -13.572, -11.597, -11.946, -12.225, -12.461,
    -12.672, -12.868, -13.052, -13.228, -13.397, -13.563, -11.587, -11.934,
    -12.211, -12.447, -12.660, -12.858, -13.044, -13.220, -13.389, -13.555,
    -11.579, -11.925, -12.201, -12.437, -12.650, -12.849, -13.035, -13.212,
    -13.381, -13.546, -11.609, -11.944, -12.210, -12.439, -12.649, -12.845,
    -13.030, -13.206, -13.374, -13.538, -11.665, -11.979, -12.228, -12.447,
    -12.650, -12.844, -13.027, -13.201, -13.368, -13.531, -11.687, -11.991,
    -12.233, -12.446, -12.647, -12.839, -13.021, -13.195, -13.361, -13.523,
    -11.735, -12.019, -12.247, -12.452, -12.648, -12.837, -13.018, -13.190,
    -13.355, -13.516, -11.732, -12.013, -12.239, -12.444, -12.641, -12.830,
    -13.011, -13.183, -13.348, -13.508, -11.721, -12.004, -12.231, -12.437,
    -12.634, -12.824, -13.005, -13.177, -13.341, -13.500, -11.708, -11.984,
    -12.210, -12.418, -12.618, -12.811, -12.994, -13.168, -13.332, -13.492,
    -11.692, -11.967, -12.195, -12.405, -12.607, -12.802, -12.986, -13.160,
    -13.325, -13.484, -11.650, -11.933, -12.169, -12.386, -12.593, -12.791,
    -12.977, -13.152, -13.317, -13.476, -11.625, -11.916, -12.159, -12.380,
    -12.589, -12.787, -12.973, -13.147, -13.311, -13.468, -11.601, -11.902,
    -12.151, -12.376, -12.586, -12.784, -12.969, -13.142, -13.305, -13.462,
    -11.588, -11.894, -12.147, -12.374, -12.585, -12.782, -12.966, -13.138,
    -13.300, -13.455, -11.576, -11.887, -12.144, -12.372, -12.584, -12.781,
    -12.963, -13.134, -13.294, -13.449, -11.566, -11.878, -12.137, -12.367,
    -12.579, -12.776, -12.959, -13.129, -13.289, -13.442, -11.543, -11.859,
    -12.123, -12.356, -12.571, -12.770, -12.953, -13.123, -13.282, -13.435,
    -11.534, -11.849, -12.113, -12.348, -12.564, -12.764, -12.947, -13.117,
    -13.276, -13.428, -11.537, -11.847, -12.109, -12.344, -12.560, -12.759,
    -12.942, -13.111, -13.270, -13.421, -11.551, -11.847, -12.103, -12.335,
    -12.551, -12.751, -12.935, -13.104, -13.263, -13.414, -11.539, -11.834,
    -12.090, -12.323, -12.540, -12.742, -12.926, -13.097, -13.255, -13.406,
    -11.531, -11.824, -12.079, -12.312, -12.530, -12.732, -12.918, -13.089,
    -13.248, -13.399, -11.529, -11.820, -12.072, -12.303, -12.521, -12.724,
    -12.910, -13.082, -13.241, -13.392, -11.540, -11.827, -12.076, -12.304,
    -12.519, -12.720, -12.906, -13.077, -13.235, -13.385, -11.565, -11.845,
    -12.088, -12.311, -12.522, -12.720, -12.904, -13.073, -13.230, -13.379,
    -11.578, -11.856, -12.096, -12.316, -12.524, -12.720, -12.901, -13.069,
    -13.225, -13.373, -11.586, -11.865, -12.105, -12.323, -12.528, -12.721,
    -12.900, -13.066, -13.221, -13.367, -11.586, -11.865, -12.104, -12.321,
    -12.525, -12.717, -12.895, -13.061, -13.215, -13.361, -11.584, -11.866,
    -12.106, -12.322, -12.525, -12.715, -12.892, -13.056, -13.209, -13.355,
    -11.569, -11.856, -12.099, -12.317, -12.519, -12.710, -12.886, -13.050,
    -13.203, -13.348, -11.556, -11.851, -12.098, -12.316, -12.518, -12.707,
    -12.882, -13.045, -13.197, -13.341, -11.553, -11.850, -12.098, -12.316,
    -12.517, -12.704, -12.878, -13.040, -13.192, -13.335, -11.562, -11.859,
    -12.106, -12.322, -12.520, -12.705, -12.877, -13.037, -13.187, -13.329,
    -11.587, -11.881, -12.123, -12.334, -12.528, -12.709, -12.877, -13.035,
    -13.183, -13.324, -11.658, -11.935, -12.162, -12.360, -12.543, -12.716,
    -12.880, -13.034, -13.180, -13.319, -11.712, -11.977, -12.190, -12.377,
    -12.552, -12.720, -12.880, -13.031, -13.176, -13.314, -11.791, -12.033,
    -12.227, -12.398, -12.563, -12.724, -12.880, -13.029, -13.171, -13.308,
    -11.795, -12.036, -12.228, -12.398, -12.561, -12.721, -12.876, -13.024,
    -13.166, -13.302, -11.803, -12.040, -12.229, -12.396, -12.558, -12.717,
    -12.871, -13.019, -13.160, -13.296, -11.791, -12.024, -12.213, -12.382,
    -12.547, -12.708, -12.864, -13.012, -13.154, -13.290, -11.764, -11.998,
    -12.189, -12.363, -12.531, -12.696, -12.854, -13.005, -13.147, -13.283,
    -11.723, -11.961, -12.159, -12.339, -12.513, -12.682, -12.844, -12.996,
    -13.139, -13.276, -11.681, -11.923, -12.128, -12.315, -12.495, -12.668,
    -12.833, -12.987, -13.132, -13.269, -11.643, -11.892, -12.104, -12.297,
    -12.481, -12.657, -12.824, -12.980, -13.125, -13.262, -11.633, -11.882,
    -12.095, -12.289, -12.474, -12.652, -12.819, -12.974, -13.119, -13.256,
    -11.625, -11.873, -12.087, -12.283, -12.469, -12.647, -12.814, -12.969,
    -13.114, -13.250, -11.610, -11.863, -12.081, -12.279, -12.466, -12.644,
    -12.810, -12.965, -13.109, -13.244, -11.602, -11.854, -12.074, -12.275,
    -12.463, -12.641, -12.807, -12.961, -13.104, -13.239, -11.592, -11.844,
    -12.065, -12.267, -12.456, -12.635, -12.801, -12.955, -13.098, -13.233,
    -11.581, -11.834, -12.056, -12.259, -12.450, -12.629, -12.796, -12.950,
    -13.093, -13.227, -11.591, -11.836, -12.055, -12.256, -12.445, -12.624,
    -12.790, -12.944, -13.087, -13.220, -11.588, -11.831, -12.048, -12.249,
    -12.439, -12.618, -12.784, -12.938, -13.081, -13.214, -11.599, -11.838,
    -12.052, -12.250, -12.437, -12.615, -12.780, -12.933, -13.075, -13.208,
    -11.613, -11.848, -12.059, -12.254, -12.439, -12.615, -12.778, -12.930,
    -13.071, -13.203, -11.620, -11.856, -12.064, -12.258, -12.441, -12.615,
    -12.777, -12.927, -13.067, -13.198, -11.636, -11.869, -12.074, -12.264,
    -12.445, -12.617, -12.776, -12.925, -13.063, -13.193, -11.647, -11.877,
    -12.081, -12.270, -12.448, -12.618, -12.776, -12.923, -13.059, -13.188,
    -11.643, -11.874, -12.078, -12.268, -12.447, -12.616, -12.773, -12.919,
    -13.055, -13.183, -11.638, -11.869, -12.075, -12.265, -12.444, -12.613,
    -12.769, -12.914, -13.050, -13.178, -11.642, -11.871, -12.076, -12.265,
    -12.443, -12.610, -12.766, -12.910, -13.045, -13.172, -11.654, -11.879,
    -12.081, -12.267, -12.443, -12.608, -12.762, -12.906, -13.040, -13.166,
    -11.660, -11.877, -12.076, -12.262, -12.437, -12.602, -12.756, -12.899,
    -13.034, -13.160, -11.667, -11.884, -12.082, -12.265, -12.437, -12.600,
    -12.752, -12.894, -13.028, -13.154, -11.670, -11.888, -12.086, -12.267,
    -12.436, -12.597, -12.747, -12.889, -13.022, -13.148, -11.678, -11.897,
    -12.094, -12.273, -12.439, -12.596, -12.744, -12.884, -13.017, -13.142,
    -11.721, -11.937, -12.126, -12.296, -12.453, -12.603, -12.746, -12.883,
    -13.013, -13.137, -11.785, -11.990, -12.167, -12.324, -12.470, -12.612,
    -12.749, -12.882, -13.010, -13.133, -11.842, -12.036, -12.200, -12.346,
    -12.483, -12.618, -12.750, -12.880, -13.006, -13.128, -11.909, -12.089,
    -12.239, -12.372, -12.499, -12.626, -12.753, -12.879, -13.003, -13.124,
    -11.917, -12.095, -12.242, -12.373, -12.498, -12.623, -12.749, -12.875,
    -12.998, -13.118, -11.917, -12.096, -12.243, -12.373, -12.496, -12.620,
    -12.746, -12.871, -12.993, -13.112, -11.898, -12.080, -12.230, -12.361,
    -12.486, -12.611, -12.737, -12.863, -12.986, -13.106, -11.884, -12.069,
    -12.221, -12.353, -12.479, -12.605, -12.732, -12.857, -12.980, -13.100,
    -11.860, -12.046, -12.200, -12.336, -12.466, -12.595, -12.723, -12.850,
    -12.974, -13.094, -11.850, -12.033, -12.188, -12.326, -12.458, -12.589,
    -12.718, -12.846, -12.969, -13.088, -11.852, -12.030, -12.183, -12.321,
    -12.454, -12.585, -12.715, -12.842, -12.964, -13.083, -11.863, -12.034,
    -12.183, -12.319, -12.451, -12.582, -12.712, -12.838, -12.960, -13.078,
    -11.872, -12.035, -12.181, -12.316, -12.448, -12.579, -12.708, -12.834,
    -12.955, -13.073, -11.881, -12.040, -12.182, -12.315, -12.446, -12.576,
    -12.704, -12.829, -12.951, -13.068, -11.875, -12.033, -12.175, -12.309,
    -12.440, -12.571, -12.699, -12.824, -12.945, -13.062, -11.885, -12.038,
    -12.176, -12.307, -12.437, -12.567, -12.694, -12.819, -12.939, -13.056,
    -11.892, -12.041, -12.177, -12.306, -12.435, -12.564, -12.690, -12.814,
    -12.934, -13.050, -11.884, -12.035, -12.171, -12.301, -12.429, -12.558,
    -12.684, -12.808, -12.928, -13.044, -11.891, -12.040, -12.174, -12.302,
    -12.428, -12.554, -12.679, -12.802, -12.922, -13.038, -11.895, -12.046,
    -12.178, -12.304, -12.427, -12.551, -12.674, -12.796, -12.916, -13.032,
    -11.892, -12.047, -12.182, -12.306, -12.427, -12.548, -12.670, -12.791,
    -12.910, -13.025, -11.902, -12.057, -12.189, -12.311, -12.429, -12.548,
    -12.668, -12.787, -12.905, -13.020, -11.915, -12.067, -12.197, -12.316,
    -12.432, -12.548, -12.666, -12.784, -12.900, -13.015, -11.906, -12.060,
    -12.193, -12.314, -12.430, -12.546, -12.662, -12.779, -12.895, -13.009,
    -11.910, -12.061, -12.192, -12.313, -12.428, -12.544, -12.659, -12.775,
    -12.891, -13.004, -11.917, -12.066, -12.195, -12.314, -12.428, -12.542,
    -12.656, -12.771, -12.885, -12.998, -11.918, -12.067, -12.197, -12.315,
    -12.428, -12.540, -12.652, -12.766, -12.880, -12.993, -11.943, -12.083,
    -12.206, -12.319, -12.429, -12.538, -12.648, -12.761, -12.874, -12.986,
    -11.955, -12.092, -12.213, -12.323, -12.430, -12.537, -12.646, -12.757,
    -12.869, -12.981, -11.963, -12.098, -12.217, -12.326, -12.431, -12.536,
    -12.643, -12.753, -12.864, -12.975, -11.979, -12.112, -12.228, -12.334,
    -12.435, -12.537, -12.642, -12.750, -12.860, -12.970, -11.984, -12.121,
    -12.238, -12.342, -12.441, -12.540, -12.642, -12.748, -12.856, -12.965,
    -12.009, -12.148, -12.263, -12.362, -12.454, -12.547, -12.644, -12.747,
    -12.853, -12.961, -12.034, -12.172, -12.283, -12.377, -12.463, -12.551,
    -12.644, -12.744, -12.849, -12.956, -12.062, -12.202, -12.310, -12.396,
    -12.475, -12.556, -12.645, -12.742, -12.845, -12.951, -12.062, -12.201,
    -12.309, -12.394, -12.472, -12.552, -12.640, -12.737, -12.839, -12.945,
    -12.078, -12.211, -12.313, -12.396, -12.471, -12.549, -12.636, -12.731,
    -12.833, -12.939, -12.094, -12.217, -12.313, -12.391, -12.464, -12.541,
    -12.628, -12.724, -12.826, -12.932, -12.130, -12.235, -12.319, -12.390,
    -12.459, -12.534, -12.620, -12.716, -12.819, -12.925, -12.121, -12.224,
    -12.308, -12.378, -12.448, -12.524, -12.610, -12.707, -12.811, -12.918,
    -12.116, -12.219, -12.301, -12.371, -12.440, -12.516, -12.603, -12.699,
    -12.803, -12.911, -12.119, -12.220, -12.301, -12.369, -12.436, -12.511,
    -12.596, -12.693, -12.796, -12.904, -12.136, -12.234, -12.309, -12.373,
    -12.436, -12.509, -12.593, -12.688, -12.791, -12.898, -12.169, -12.258,
    -12.324, -12.380, -12.439, -12.508, -12.590, -12.683, -12.785, -12.892,
    -12.203, -12.283, -12.340, -12.389, -12.442, -12.508, -12.587, -12.679,
    -12.780, -12.887, -12.229, -12.300, -12.349, -12.392, -12.442, -12.506,
    -12.583, -12.674, -12.775, -12.881, -12.237, -12.304, -12.350, -12.391,
    -12.439, -12.501, -12.578, -12.668, -12.769, -12.874, -12.226, -12.295,
    -12.342, -12.384, -12.433, -12.496, -12.572, -12.662, -12.762, -12.868,
    -12.203, -12.277, -12.329, -12.374, -12.425, -12.488, -12.565, -12.655,
    -12.755, -12.861, -12.173, -12.253, -12.311, -12.361, -12.415, -12.480,
    -12.557, -12.648, -12.748, -12.855, -12.163, -12.244, -12.303, -12.354,
    -12.408, -12.473, -12.550, -12.641, -12.741, -12.848, -12.169, -12.246,
    -12.303, -12.351, -12.404, -12.468, -12.545, -12.635, -12.735, -12.842,
    -12.184, -12.255, -12.308, -12.353, -12.403, -12.465, -12.540, -12.629,
    -12.729, -12.836, -12.207, -12.272, -12.318, -12.358, -12.404, -12.463,
    -12.537, -12.625, -12.724, -12.830, -12.231, -12.288, -12.327, -12.362,
    -12.404, -12.461, -12.533, -12.620, -12.718, -12.824, -12.247, -12.299,
    -12.333, -12.363, -12.403, -12.457, -12.528, -12.614, -12.712, -12.818,
    -12.251, -12.301, -12.333, -12.362, -12.401, -12.454, -12.523, -12.608,
    -12.706, -12.811, -12.271, -12.315, -12.341, -12.366, -12.401, -12.452,
    -12.519, -12.602, -12.699, -12.805, -12.281, -12.321, -12.344, -12.366,
    -12.399, -12.447, -12.513, -12.596, -12.693, -12.798, -12.270, -12.310,
    -12.335, -12.358, -12.391, -12.440, -12.505, -12.588, -12.685, -12.791,
    -12.278, -12.315, -12.337, -12.358, -12.389, -12.436, -12.500, -12.582,
    -12.678, -12.784, -12.274, -12.312, -12.334, -12.355, -12.386, -12.432,
    -12.494, -12.575, -12.671, -12.777, -12.261, -12.302, -12.327, -12.351,
    -12.382, -12.427, -12.489, -12.569, -12.664, -12.770, -12.274, -12.311,
    -12.335, -12.357, -12.386, -12.428, -12.487, -12.565, -12.659, -12.764,
    -12.285, -12.319, -12.341, -12.361, -12.388, -12.427, -12.484, -12.560,
    -12.653, -12.758, -12.278, -12.312, -12.336, -12.357, -12.385, -12.424,
    -12.479, -12.554, -12.647, -12.752, -12.292, -12.322, -12.342, -12.361,
    -12.385, -12.421, -12.475, -12.549, -12.640, -12.745, -12.297, -12.322,
    -12.339, -12.356, -12.379, -12.415, -12.467, -12.541, -12.633, -12.738,
    -12.299, -12.321, -12.335, -12.349, -12.371, -12.405, -12.458, -12.532,
    -12.624, -12.730, -12.329, -12.339, -12.342, -12.348, -12.365, -12.397,
    -12.449, -12.523, -12.616, -12.722, -12.340, -12.342, -12.339, -12.341,
    -12.355, -12.387, -12.439, -12.513, -12.607, -12.714, -12.346, -12.340,
    -12.331, -12.329, -12.342, -12.374, -12.426, -12.502, -12.597, -12.706,
    -12.352, -12.340, -12.326, -12.321, -12.332, -12.363, -12.417, -12.493,
    -12.589, -12.699, -12.343, -12.329, -12.315, -12.310, -12.321, -12.353,
    -12.407, -12.484, -12.581, -12.691, -12.345, -12.325, -12.307, -12.300,
    -12.311, -12.343, -12.397, -12.475, -12.573, -12.684, -12.350, -12.324,
    -12.302, -12.293, -12.303, -12.335, -12.389, -12.467, -12.565, -12.677,
    -12.350, -12.318, -12.292, -12.282, -12.292, -12.324, -12.379, -12.458,
    -12.557, -12.669, -12.346, -12.307, -12.278, -12.267, -12.278, -12.312,
    -12.368, -12.448, -12.548, -12.661, -12.358, -12.307, -12.271, -12.257,
    -12.267, -12.300, -12.357, -12.438, -12.539, -12.653, -12.370, -12.306,
    -12.263, -12.246, -12.255, -12.289, -12.347, -12.428, -12.530, -12.645,
    -12.407, -12.319, -12.262, -12.238, -12.244, -12.277, -12.335, -12.418,
    -12.520, -12.637, -12.398, -12.305, -12.247, -12.224, -12.232, -12.266,
    -12.324, -12.407, -12.511, -12.629, -12.390, -12.293, -12.234, -12.212,
    -12.220, -12.255, -12.314, -12.397, -12.502, -12.620, -12.386, -12.284,
    -12.224, -12.202, -12.211, -12.245, -12.304, -12.388, -12.493, -12.612,
    -12.387, -12.280, -12.218, -12.196, -12.203, -12.237, -12.296, -12.380,
    -12.485, -12.605, -12.395, -12.281, -12.216, -12.191, -12.197, -12.230,
    -12.288, -12.372, -12.477, -12.597, -12.401, -12.279, -12.209, -12.182,
    -12.188, -12.220, -12.279, -12.363, -12.469, -12.590, -12.400, -12.270,
    -12.197, -12.169, -12.175, -12.209, -12.268, -12.353, -12.460, -12.582,
    -12.390, -12.254, -12.180, -12.153, -12.161, -12.196, -12.256, -12.343,
    -12.451, -12.574, -12.369, -12.230, -12.157, -12.133, -12.144, -12.181,
    -12.244, -12.332, -12.441, -12.566, -12.342, -12.200, -12.129, -12.109,
    -12.124, -12.165, -12.230, -12.320, -12.431, -12.557, -12.312, -12.167,
    -12.100, -12.085, -12.104, -12.148, -12.216, -12.308, -12.421, -12.549,
    -12.289, -12.140, -12.074, -12.062, -12.085, -12.133, -12.203, -12.297,
    -12.412, -12.540, -12.272, -12.118, -12.052, -12.043, -12.068, -12.118,
    -12.191, -12.286, -12.402, -12.532, -12.257, -12.098, -12.032, -12.025,
    -12.053, -12.105, -12.179, -12.276, -12.394, -12.525, -12.245, -12.083,
    -12.017, -12.013, -12.042, -12.096, -12.170, -12.268, -12.386, -12.518,
    -12.231, -12.065, -12.000, -11.997, -12.029, -12.084, -12.159, -12.258,
    -12.377, -12.510, -12.212, -12.044, -11.982, -11.982, -12.017, -12.073,
    -12.149, -12.248, -12.368, -12.502, -12.186, -12.017, -11.957, -11.960,
    -11.998, -12.057, -12.135, -12.236, -12.358, -12.493, -12.163, -11.992,
    -11.934, -11.940, -11.980, -12.041, -12.122, -12.224, -12.348, -12.485,
    -12.134, -11.963, -11.906, -11.916, -11.959, -12.023, -12.106, -12.211,
    -12.337, -12.475, -12.097, -11.927, -11.875, -11.889, -11.937, -12.004,
    -12.090, -12.197, -12.325, -12.466, -12.065, -11.894, -11.845, -11.863,
    -11.914, -11.984, -12.073, -12.183, -12.313, -12.456, -12.026, -11.857,
    -11.812, -11.835, -11.890, -11.964, -12.056, -12.169, -12.302, -12.447,
    -11.983, -11.817, -11.777, -11.805, -11.865, -11.943, -12.038, -12.155,
    -12.290, -12.437, -11.947, -11.783, -11.746, -11.778, -11.842, -11.924,
    -12.022, -12.141, -12.279, -12.428, -11.907, -11.745, -11.712, -11.749,
    -11.818, -11.903, -12.006, -12.128, -12.268, -12.419, -11.862, -11.704,
    -11.676, -11.719, -11.792, -11.883, -11.989, -12.114, -12.257, -12.410,
    -11.820, -11.666, -11.643, -11.690, -11.768, -11.863, -11.972, -12.100,
    -12.246, -12.401, -11.775, -11.626, -11.608, -11.661, -11.744, -11.842,
    -11.955, -12.086, -12.234, -12.392, -11.729, -11.585, -11.573, -11.631,
    -11.719, -11.822, -11.938, -12.073, -12.223, -12.383, -11.686, -11.547,
    -11.541, -11.604, -11.696, -11.802, -11.922, -12.059, -12.212, -12.374,
    -11.638, -11.505, -11.505, -11.574, -11.671, -11.782, -11.905, -12.045,
    -12.201, -12.365, -11.588, -11.462, -11.468, -11.543, -11.646, -11.761,
    -11.888, -12.031, -12.190, -12.356, -11.536, -11.418, -11.431, -11.513,
    -11.621, -11.741, -11.872, -12.018, -12.179, -12.347, -11.480, -11.371,
    -11.392, -11.480, -11.595, -11.720, -11.855, -12.004, -12.168, -12.338,
    -11.426, -11.325, -11.353, -11.448, -11.569, -11.699, -11.838, -11.991,
    -12.157, -12.330, -11.371, -11.279, -11.314, -11.416, -11.542, -11.677,
    -11.821, -11.977, -12.146, -12.321, -11.314, -11.232, -11.274, -11.383,
    -11.515, -11.655, -11.803, -11.963, -12.135, -12.312, -11.257, -11.184,
    -11.233, -11.349, -11.487, -11.633, -11.785, -11.949, -12.124, -12.303,
    -11.201, -11.137, -11.194, -11.316, -11.461, -11.611, -11.768, -11.935,
    -12.113, -12.294, -11.145, -11.090, -11.155, -11.283, -11.434, -11.589,
    -11.750, -11.921, -12.101, -12.285, -11.091, -11.045, -11.117, -11.252,
    -11.407, -11.567, -11.732, -11.906, -12.090, -12.276, -11.031, -10.996,
    -11.076, -11.217, -11.379, -11.545, -11.714, -11.892, -12.079, -12.267,
    -10.971, -10.946, -11.034, -11.183, -11.352, -11.523, -11.697, -11.879,
    -12.068, -12.259, -10.911, -10.897, -10.993, -11.150, -11.324, -11.501,
    -11.680, -11.865, -12.057, -12.250, -10.851, -10.848, -10.953, -11.117,
    -11.298, -11.481, -11.664, -11.853, -12.048, -12.242, -10.792, -10.800,
    -10.913, -11.084, -11.272, -11.460, -11.648, -11.840, -12.038, -12.234,
    -10.733, -10.751, -10.873, -11.052, -11.247, -11.440, -11.632, -11.828,
    -12.028, -12.227, -10.673, -10.703, -10.832, -11.018, -11.219, -11.418,
    -11.615, -11.814, -12.018, -12.218, -10.614, -10.654, -10.792, -10.984,
    -11.192, -11.396, -11.597, -11.800, -12.007, -12.210, -10.555, -10.605,
    -10.751, -10.950, -11.164, -11.373, -11.579, -11.786, -11.996, -12.201,
    -10.497, -10.557, -10.710, -10.916, -11.136, -11.351, -11.561, -11.772,
    -11.985, -12.192, -10.439, -10.509, -10.669, -10.882, -11.108, -11.328,
    -11.543, -11.758, -11.973, -12.183, -10.382, -10.462, -10.630, -10.849,
    -11.080, -11.306, -11.525, -11.743, -11.962, -12.175, -10.326, -10.415,
    -10.591, -10.817, -11.054, -11.284, -11.508, -11.730, -11.952, -12.167,
    -10.271, -10.370, -10.553, -10.785, -11.028, -11.263, -11.491, -11.717,
    -11.942, -12.159, -10.217, -10.326, -10.516, -10.754, -11.003, -11.243,
    -11.475, -11.705, -11.932, -12.151, -10.164, -10.282, -10.479, -10.724,
    -10.978, -11.223, -11.459, -11.692, -11.922, -12.143, -10.112, -10.239,
    -10.444, -10.694, -10.954, -11.204, -11.444, -11.680, -11.913, -12.136,
    -10.062, -10.198, -10.409, -10.665, -10.930, -11.184, -11.429, -11.668,
    -11.904, -12.128, -10.013, -10.157, -10.375, -10.637, -10.907, -11.166,
    -11.414, -11.657, -11.895, -12.121, -9.966,  -10.118, -10.342, -10.610,
    -10.884, -11.148, -11.400, -11.646, -11.886, -12.114, -9.921,  -10.080,
    -10.310, -10.583, -10.862, -11.130, -11.386, -11.634, -11.877, -12.106,
    -9.877,  -10.044, -10.279, -10.557, -10.841, -11.113, -11.372, -11.624,
    -11.868, -12.099, -9.583,  -9.808,  -10.096, -10.418, -10.738, -11.038,
    -11.318, -11.586, -11.842, -12.082, -9.360,  -9.625,  -9.950,  -10.305,
    -10.651, -10.973, -11.272, -11.553, -11.819, -12.066, -9.087,  -9.397,
    -9.765,  -10.158, -10.538, -10.889, -11.210, -11.510, -11.790, -12.046,
    -8.895,  -9.233,  -9.630,  -10.049, -10.452, -10.823, -11.161, -11.474,
    -11.765, -12.029, -8.787,  -9.147,  -9.562,  -9.995,  -10.410, -10.791,
    -11.137, -11.457, -11.752, -12.019, -8.651,  -9.037,  -9.473,  -9.925,
    -10.356, -10.750, -11.107, -11.435, -11.736, -12.008, -8.679,  -9.071,
    -9.507,  -9.955,  -10.381, -10.769, -11.120, -11.443, -11.741, -12.009,
    -8.669,  -9.074,  -9.516,  -9.966,  -10.391, -10.777, -11.126, -11.447,
    -11.742, -12.009, -8.596,  -9.019,  -9.475,  -9.936,  -10.369, -10.760,
    -11.114, -11.437, -11.735, -12.002, -8.660,  -9.082,  -9.531,  -9.982,
    -10.405, -10.787, -11.133, -11.450, -11.742, -12.005, -8.644,  -9.075,
    -9.528,  -9.981,  -10.405, -10.787, -11.132, -11.448, -11.740, -12.002,
    -8.574,  -9.023,  -9.490,  -9.953,  -10.384, -10.772, -11.121, -11.440,
    -11.733, -11.997, -8.676,  -9.114,  -9.565,  -10.013, -10.429, -10.805,
    -11.144, -11.455, -11.742, -12.001, -8.637,  -9.079,  -9.535,  -9.987,
    -10.408, -10.788, -11.131, -11.445, -11.734, -11.994, -8.478,  -8.937,
    -9.414,  -9.887,  -10.328, -10.727, -11.085, -11.412, -11.711, -11.979,
    -8.409,  -8.864,  -9.344,  -9.824,  -10.275, -10.683, -11.052, -11.387,
    -11.693, -11.966, -8.222,  -8.693,  -9.195,  -9.698,  -10.172, -10.603,
    -10.991, -11.343, -11.662, -11.944, -8.023,  -8.517,  -9.044,  -9.572,
    -10.070, -10.522, -10.929, -11.297, -11.630, -11.922, -7.935,  -8.449,
    -8.991,  -9.532,  -10.040, -10.500, -10.913, -11.285, -11.620, -11.914,
    -7.896,  -8.428,  -8.984,  -9.534,  -10.046, -10.507, -10.919, -11.289,
    -11.622, -11.914, -7.914,  -8.463,  -9.026,  -9.577,  -10.086, -10.541,
    -10.946, -11.308, -11.634, -11.920, -8.024,  -8.579,  -9.138,  -9.678,
    -10.171, -10.609, -10.997, -11.345, -11.657, -11.933, -8.204,  -8.747,
    -9.286,  -9.801,  -10.269, -10.684, -11.051, -11.381, -11.680, -11.945,
    -8.422,  -8.937,  -9.440,  -9.921,  -10.360, -10.749, -11.097, -11.411,
    -11.698, -11.955, -8.549,  -9.040,  -9.519,  -9.979,  -10.402, -10.779,
    -11.116, -11.423, -11.704, -11.957, -8.498,  -8.999,  -9.487,  -9.955,
    -10.383, -10.764, -11.105, -11.415, -11.698, -11.951, -8.293,  -8.828,
    -9.352,  -9.851,  -10.305, -10.708, -11.065, -11.386, -11.678, -11.937,
    -8.044,  -8.610,  -9.170,  -9.705,  -10.192, -10.623, -11.003, -11.343,
    -11.648, -11.917, -7.817,  -8.399,  -8.984,  -9.548,  -10.064, -10.522,
    -10.927, -11.288, -11.610, -11.891, -7.670,  -8.251,  -8.844,  -9.422,
    -9.956,  -10.433, -10.857, -11.236, -11.573, -11.866, -7.613,  -8.183,
    -8.773,  -9.353,  -9.893,  -10.379, -10.813, -11.201, -11.547, -11.847,
    -7.601,  -8.161,  -8.744,  -9.321,  -9.862,  -10.350, -10.788, -11.181,
    -11.531, -11.835, -7.659,  -8.206,  -8.777,  -9.345,  -9.878,  -10.361,
    -10.795, -11.184, -11.532, -11.834, -7.797,  -8.332,  -8.887,  -9.439,
    -9.956,  -10.423, -10.841, -11.217, -11.553, -11.846, -7.898,  -8.431,
    -8.979,  -9.520,  -10.023, -10.477, -10.883, -11.247, -11.573, -11.857,
    -7.958,  -8.498,  -9.045,  -9.580,  -10.076, -10.520, -10.916, -11.270,
    -11.587, -11.865, -8.066,  -8.602,  -9.139,  -9.661,  -10.142, -10.573,
    -10.955, -11.298, -11.604, -11.874, -8.048,  -8.592,  -9.136,  -9.662,
    -10.145, -10.576, -10.957, -11.299, -11.604, -11.872, -7.953,  -8.513,
    -9.073,  -9.613,  -10.109, -10.549, -10.938, -11.285, -11.594, -11.863,
    -8.036,  -8.581,  -9.128,  -9.657,  -10.142, -10.574, -10.955, -11.296,
    -11.599, -11.865, -7.984,  -8.535,  -9.089,  -9.625,  -10.117, -10.555,
    -10.941, -11.285, -11.591, -11.858, -7.891,  -8.452,  -9.018,  -9.567,
    -10.071, -10.519, -10.914, -11.265, -11.576, -11.847, -7.953,  -8.498,
    -9.052,  -9.592,  -10.089, -10.531, -10.922, -11.269, -11.577, -11.846,
    -7.901,  -8.446,  -9.003,  -9.549,  -10.053, -10.503, -10.900, -11.252,
    -11.565, -11.836, -7.868,  -8.412,  -8.971,  -9.521,  -10.029, -10.483,
    -10.884, -11.241, -11.556, -11.829, -7.972,  -8.497,  -9.039,  -9.575,
    -10.071, -10.515, -10.907, -11.255, -11.564, -11.832, -8.020,  -8.535,
    -9.069,  -9.598,  -10.089, -10.528, -10.916, -11.261, -11.566, -11.832,
    -8.218,  -8.706,  -9.213,  -9.715,  -10.182, -10.599, -10.968, -11.296,
    -11.588, -11.844, -8.471,  -8.918,  -9.385,  -9.851,  -10.287, -10.677,
    -11.024, -11.334, -11.611, -11.856, -8.624,  -9.045,  -9.487,  -9.931,
    -10.348, -10.722, -11.056, -11.355, -11.623, -11.861, -8.848,  -9.225,
    -9.627,  -10.038, -10.428, -10.781, -11.096, -11.381, -11.639, -11.868,
    -8.852,  -9.231,  -9.633,  -10.045, -10.434, -10.785, -11.099, -11.382,
    -11.638, -11.866, -8.859,  -9.238,  -9.641,  -10.052, -10.441, -10.791,
    -11.103, -11.384, -11.637, -11.863, -8.867,  -9.246,  -9.649,  -10.061,
    -10.448, -10.797, -11.107, -11.386, -11.637, -11.861, -8.877,  -9.256,
    -9.659,  -10.070, -10.457, -10.804, -11.112, -11.388, -11.637, -11.859,
    -8.889,  -9.267,  -9.670,  -10.080, -10.466, -10.811, -11.117, -11.391,
    -11.638, -11.857, -8.902,  -9.280,  -9.682,  -10.092, -10.476, -10.819,
    -11.123, -11.394, -11.638, -11.856, -8.917,  -9.293,  -9.695,  -10.104,
    -10.486, -10.828, -11.129, -11.398, -11.639, -11.854, -8.933,  -9.308,
    -9.709,  -10.117, -10.498, -10.837, -11.136, -11.402, -11.640, -11.853,
    -8.951,  -9.324,  -9.724,  -10.130, -10.510, -10.847, -11.143, -11.406,
    -11.641, -11.851, -8.970,  -9.341,  -9.740,  -10.145, -10.522, -10.857,
    -11.150, -11.410, -11.643, -11.850, -8.990,  -9.358,  -9.757,  -10.160,
    -10.536, -10.868, -11.158, -11.415, -11.644, -11.849, -9.011,  -9.377,
    -9.774,  -10.176, -10.549, -10.879, -11.167, -11.420, -11.646, -11.848,
    -9.033,  -9.396,  -9.792,  -10.192, -10.564, -10.891, -11.175, -11.425,
    -11.648, -11.847, -9.056,  -9.417,  -9.810,  -10.209, -10.578, -10.903,
    -11.184, -11.431, -11.651, -11.847, -9.080,  -9.437,  -9.829,  -10.226,
    -10.594, -10.916, -11.194, -11.437, -11.653, -11.847, -9.105,  -9.459,
    -9.849,  -10.244, -10.609, -10.929, -11.204, -11.443, -11.656, -11.846,
    -9.130,  -9.481,  -9.869,  -10.262, -10.625, -10.942, -11.214, -11.450,
    -11.659, -11.846, -9.157,  -9.504,  -9.890,  -10.281, -10.642, -10.956,
    -11.224, -11.456, -11.662, -11.846, -9.184,  -9.527,  -9.911,  -10.300,
    -10.658, -10.969, -11.234, -11.463, -11.664, -11.845, -9.212,  -9.551,
    -9.932,  -10.319, -10.675, -10.983, -11.244, -11.469, -11.667, -11.845,
    -9.240,  -9.575,  -9.954,  -10.339, -10.692, -10.997, -11.254, -11.475,
    -11.670, -11.845, -9.269,  -9.599,  -9.976,  -10.359, -10.709, -11.011,
    -11.265, -11.482, -11.673, -11.844, -9.298,  -9.624,  -9.999,  -10.379,
    -10.727, -11.025, -11.276, -11.489, -11.676, -11.844, -9.328,  -9.649,
    -10.022, -10.399, -10.744, -11.040, -11.287, -11.496, -11.679, -11.844,
    -9.359,  -9.675,  -10.045, -10.420, -10.763, -11.055, -11.298, -11.503,
    -11.682, -11.844, -9.390,  -9.701,  -10.068, -10.441, -10.781, -11.070,
    -11.310, -11.511, -11.686, -11.844, -9.421,  -9.727,  -10.092, -10.463,
    -10.800, -11.086, -11.321, -11.519, -11.690, -11.845, -9.452,  -9.754,
    -10.115, -10.484, -10.818, -11.101, -11.333, -11.526, -11.694, -11.845,
    -9.484,  -9.780,  -10.139, -10.506, -10.837, -11.117, -11.345, -11.534,
    -11.697, -11.845, -9.516,  -9.807,  -10.163, -10.527, -10.856, -11.132,
    -11.357, -11.541, -11.700, -11.845, -9.549,  -9.834,  -10.188, -10.549,
    -10.875, -11.148, -11.368, -11.549, -11.703, -11.844, -9.582,  -9.861,
    -10.212, -10.571, -10.894, -11.164, -11.380, -11.556, -11.706, -11.844,
    -9.615,  -9.888,  -10.237, -10.593, -10.914, -11.179, -11.392, -11.563,
    -11.709, -11.843, -9.648,  -9.916,  -10.261, -10.615, -10.933, -11.195,
    -11.403, -11.570, -11.712, -11.843, -9.681,  -9.943,  -10.286, -10.637,
    -10.952, -11.211, -11.415, -11.577, -11.715, -11.842, -9.715,  -9.971,
    -10.311, -10.659, -10.971, -11.227, -11.426, -11.585, -11.718, -11.842,
    -9.748,  -9.999,  -10.335, -10.682, -10.991, -11.242, -11.438, -11.592,
    -11.721, -11.842, -9.782,  -10.026, -10.360, -10.704, -11.010, -11.258,
    -11.449, -11.599, -11.724, -11.841, -9.816,  -10.054, -10.385, -10.726,
    -11.029, -11.274, -11.461, -11.606, -11.727, -11.840, -9.850,  -10.082,
    -10.410, -10.749, -11.049, -11.290, -11.472, -11.613, -11.730, -11.840,
    -9.884,  -10.110, -10.435, -10.771, -11.068, -11.306, -11.484, -11.620,
    -11.733, -11.839, -9.918,  -10.138, -10.460, -10.794, -11.088, -11.322,
    -11.496, -11.627, -11.736, -11.839, -9.952,  -10.166, -10.485, -10.816,
    -11.108, -11.338, -11.507, -11.634, -11.738, -11.838, -9.987,  -10.194,
    -10.510, -10.839, -11.127, -11.353, -11.519, -11.641, -11.741, -11.837,
    -10.021, -10.221, -10.535, -10.861, -11.147, -11.369, -11.530, -11.647,
    -11.743, -11.836, -10.055, -10.249, -10.560, -10.883, -11.166, -11.385,
    -11.541, -11.654, -11.746, -11.835, -10.089, -10.277, -10.585, -10.906,
    -11.185, -11.400, -11.552, -11.660, -11.748, -11.834, -10.123, -10.305,
    -10.609, -10.928, -11.204, -11.415, -11.563, -11.667, -11.750, -11.833,
    -10.158, -10.332, -10.634, -10.950, -11.223, -11.431, -11.574, -11.673,
    -11.752, -11.832, -10.192, -10.360, -10.659, -10.972, -11.242, -11.446,
    -11.584, -11.678, -11.754, -11.831, -10.226, -10.387, -10.684, -10.994,
    -11.261, -11.461, -11.595, -11.684, -11.755, -11.829, -10.259, -10.415,
    -10.708, -11.016, -11.280, -11.476, -11.605, -11.690, -11.757, -11.828,
    -10.293, -10.442, -10.733, -11.039, -11.299, -11.491, -11.615, -11.695,
    -11.758, -11.826, -10.327, -10.469, -10.757, -11.060, -11.318, -11.505,
    -11.625, -11.700, -11.759, -11.824, -10.361, -10.497, -10.782, -11.082,
    -11.337, -11.520, -11.635, -11.706, -11.761, -11.822, -10.394, -10.524,
    -10.806, -11.104, -11.355, -11.534, -11.645, -11.711, -11.762, -11.820,
    -10.427, -10.551, -10.830, -11.126, -11.374, -11.548, -11.654, -11.716,
    -11.763, -11.819, -10.460, -10.577, -10.854, -11.147, -11.392, -11.563,
    -11.664, -11.721, -11.764, -11.817, -10.494, -10.604, -10.878, -11.169,
    -11.410, -11.577, -11.673, -11.726, -11.765, -11.815, -10.526, -10.631,
    -10.902, -11.190, -11.428, -11.590, -11.682, -11.730, -11.765, -11.812,
    -10.559, -10.657, -10.926, -11.211, -11.446, -11.604, -11.690, -11.733,
    -11.765, -11.810, -10.591, -10.684, -10.950, -11.232, -11.463, -11.616,
    -11.698, -11.736, -11.764, -11.806, -10.624, -10.710, -10.973, -11.253,
    -11.480, -11.629, -11.705, -11.739, -11.764, -11.803, -10.656, -10.736,
    -10.997, -11.274, -11.497, -11.641, -11.712, -11.741, -11.762, -11.799,
    -10.688, -10.762, -11.020, -11.294, -11.514, -11.653, -11.719, -11.744,
    -11.761, -11.796, -10.719, -10.787, -11.043, -11.315, -11.531, -11.665,
    -11.726, -11.746, -11.760, -11.792, -10.751, -10.813, -11.066, -11.335,
    -11.547, -11.676, -11.733, -11.748, -11.758, -11.788, -10.782, -10.838,
    -11.089, -11.355, -11.564, -11.688, -11.739, -11.750, -11.757, -11.785,
    -10.813, -10.864, -11.112, -11.375, -11.580, -11.699, -11.745, -11.751,
    -11.755, -11.781, -10.843, -10.889, -11.135, -11.395, -11.596, -11.710,
    -11.751, -11.753, -11.753, -11.777, -10.873, -10.914, -11.157, -11.415,
    -11.611, -11.721, -11.756, -11.754, -11.751, -11.772, -10.903, -10.938,
    -11.180, -11.434, -11.627, -11.731, -11.762, -11.754, -11.748, -11.768,
    -10.933, -10.963, -11.202, -11.453, -11.642, -11.741, -11.766, -11.755,
    -11.745, -11.763, -10.963, -10.987, -11.224, -11.472, -11.656, -11.750,
    -11.770, -11.754, -11.742, -11.758, -10.992, -11.012, -11.246, -11.491,
    -11.671, -11.759, -11.774, -11.754, -11.739, -11.753, -11.021, -11.036,
    -11.267, -11.510, -11.685, -11.768, -11.777, -11.753, -11.735, -11.748,
    -11.049, -11.060, -11.289, -11.528, -11.698, -11.776, -11.780, -11.752,
    -11.732, -11.743, -11.078, -11.083, -11.310, -11.546, -11.712, -11.784,
    -11.783, -11.751, -11.728, -11.738, -11.105, -11.107, -11.331, -11.564,
    -11.725, -11.791, -11.785, -11.749, -11.724, -11.732, -11.133, -11.130,
    -11.352, -11.582, -11.738, -11.798, -11.787, -11.747, -11.719, -11.726,
    -11.160, -11.153, -11.373, -11.599, -11.750, -11.805, -11.788, -11.745,
    -11.714, -11.720, -11.187, -11.176, -11.393, -11.616, -11.762, -11.811,
    -11.789, -11.742, -11.709, -11.714, -11.214, -11.199, -11.414, -11.633,
    -11.774, -11.817, -11.790, -11.739, -11.704, -11.708, -11.240, -11.221,
    -11.434, -11.650, -11.785, -11.823, -11.791, -11.736, -11.699, -11.701,
    -11.266, -11.244, -11.454, -11.666, -11.796, -11.828, -11.791, -11.732,
    -11.693, -11.695, -11.291, -11.266, -11.474, -11.682, -11.806, -11.832,
    -11.790, -11.728, -11.687, -11.688, -11.316, -11.288, -11.493, -11.697,
    -11.816, -11.836, -11.789, -11.724, -11.681, -11.681, -11.341, -11.309,
    -11.512, -11.712, -11.826, -11.839, -11.787, -11.719, -11.675, -11.674,
    -11.366, -11.331, -11.531, -11.727, -11.834, -11.842, -11.785, -11.714,
    -11.668, -11.667, -11.390, -11.352, -11.550, -11.742, -11.843, -11.844,
    -11.783, -11.709, -11.661, -11.660, -11.413, -11.373, -11.569, -11.756,
    -11.851, -11.846, -11.780, -11.703, -11.654, -11.653, -11.437, -11.394,
    -11.587, -11.769, -11.858, -11.847, -11.776, -11.697, -11.646, -11.645,
    -11.459, -11.414, -11.605, -11.783, -11.865, -11.848, -11.772, -11.690,
    -11.639, -11.637, -11.482, -11.435, -11.623, -11.795, -11.871, -11.848,
    -11.768, -11.683, -11.630, -11.628, -11.504, -11.455, -11.640, -11.808,
    -11.877, -11.847, -11.763, -11.676, -11.622, -11.620, -11.526, -11.475,
    -11.657, -11.820, -11.882, -11.846, -11.758, -11.668, -11.613, -11.611,
    -11.547, -11.494, -11.674, -11.831, -11.887, -11.845, -11.752, -11.660,
    -11.604, -11.602, -11.569, -11.514, -11.690, -11.842, -11.891, -11.842,
    -11.745, -11.651, -11.595, -11.594, -11.589, -11.533, -11.706, -11.852,
    -11.894, -11.840, -11.739, -11.643, -11.586, -11.585, -11.610, -11.552,
    -11.722, -11.862, -11.897, -11.836, -11.732, -11.634, -11.577, -11.576,
    -11.630, -11.570, -11.737, -11.872, -11.899, -11.832, -11.724, -11.625,
    -11.567, -11.566, -11.649, -11.589, -11.752, -11.880, -11.900, -11.828,
    -11.716, -11.615, -11.557, -11.557, -11.668, -11.607, -11.767, -11.888,
    -11.901, -11.822, -11.707, -11.604, -11.547, -11.547, -11.687, -11.624,
    -11.781, -11.896, -11.900, -11.816, -11.697, -11.594, -11.536, -11.537,
    -11.706, -11.642, -11.795, -11.902, -11.899, -11.809, -11.687, -11.582,
    -11.525, -11.527, -11.724, -11.659, -11.808, -11.909, -11.898, -11.801,
    -11.676, -11.571, -11.514, -11.517, -11.742, -11.676, -11.821, -11.914,
    -11.895, -11.793, -11.666, -11.559, -11.502, -11.507, -11.759, -11.693,
    -11.833, -11.919, -11.892, -11.785, -11.654, -11.547, -11.491, -11.496,
    -11.776, -11.709, -11.845, -11.923, -11.888, -11.775, -11.643, -11.535,
    -11.479, -11.486, -11.793, -11.725, -11.856, -11.926, -11.884, -11.766,
    -11.631, -11.523, -11.468, -11.476, -11.809, -11.741, -11.867, -11.929,
    -11.878, -11.755, -11.618, -11.510, -11.456, -11.465, -11.825, -11.756,
    -11.877, -11.930, -11.872, -11.744, -11.605, -11.496, -11.443, -11.454,
    -11.841, -11.771, -11.886, -11.931, -11.865, -11.733, -11.591, -11.483,
    -11.431, -11.443, -11.856, -11.786, -11.895, -11.931, -11.857, -11.720,
    -11.578, -11.469, -11.418, -11.432, -11.871, -11.800, -11.903, -11.931,
    -11.849, -11.707, -11.563, -11.455, -11.405, -11.421, -11.885, -11.814,
    -11.911, -11.929, -11.840, -11.694, -11.548, -11.441, -11.392, -11.409,
    -11.899, -11.827, -11.918, -11.927, -11.829, -11.680, -11.533, -11.426,
    -11.379, -11.398, -11.913, -11.840, -11.924, -11.923, -11.818, -11.665,
    -11.517, -11.411, -11.365, -11.386, -11.927, -11.852, -11.929, -11.919,
    -11.807, -11.649, -11.501, -11.396, -11.352, -11.374, -11.940, -11.864,
    -11.934, -11.914, -11.794, -11.634, -11.485, -11.380, -11.338, -11.363,
    -11.953, -11.876, -11.938, -11.907, -11.781, -11.617, -11.468, -11.364,
    -11.324, -11.351, -11.965, -11.887, -11.941, -11.900, -11.767, -11.600,
    -11.450, -11.348, -11.310, -11.339, -11.977, -11.897, -11.943, -11.892,
    -11.752, -11.582, -11.433, -11.332, -11.295, -11.327, -11.989, -11.907,
    -11.944, -11.883, -11.736, -11.564, -11.415, -11.315, -11.281, -11.314,
    -12.000, -11.917, -11.944, -11.873, -11.720, -11.546, -11.397, -11.299,
    -11.266, -11.302, -12.011, -11.925, -11.944, -11.863, -11.703, -11.527,
    -11.378, -11.282, -11.252, -11.290, -12.022, -11.934, -11.942, -11.851,
    -11.685, -11.507, -11.359, -11.265, -11.237, -11.277, -12.032, -11.941,
    -11.939, -11.838, -11.666, -11.487, -11.340, -11.247, -11.222, -11.265,
    -12.041, -11.948, -11.936, -11.824, -11.647, -11.466, -11.320, -11.230,
    -11.207, -11.252, -12.051, -11.954, -11.931, -11.809, -11.627, -11.445,
    -11.300, -11.212, -11.191, -11.240, -12.060, -11.959, -11.925, -11.793,
    -11.606, -11.424, -11.280, -11.194, -11.176, -11.227, -12.068, -11.964,
    -11.919, -11.777, -11.585, -11.402, -11.259, -11.176, -11.161, -11.214,
    -12.076, -11.967, -11.911, -11.759, -11.563, -11.380, -11.239, -11.157,
    -11.145, -11.201, -12.084, -11.970, -11.902, -11.740, -11.541, -11.357,
    -11.218, -11.139, -11.129, -11.188, -12.091, -11.972, -11.892, -11.721,
    -11.517, -11.334, -11.196, -11.120, -11.113, -11.175, -12.097, -11.973,
    -11.880, -11.701, -11.494, -11.310, -11.175, -11.102, -11.098, -11.162,
    -12.103, -11.973, -11.868, -11.679, -11.469, -11.287, -11.153, -11.083,
    -11.082, -11.149, -12.108, -11.972, -11.855, -11.657, -11.444, -11.262,
    -11.131, -11.064, -11.066, -11.136, -12.113, -11.970, -11.840, -11.634,
    -11.419, -11.238, -11.109, -11.045, -11.050, -11.123, -12.117, -11.967,
    -11.824, -11.610, -11.393, -11.213, -11.087, -11.025, -11.034, -11.111,
    -12.121, -11.963, -11.807, -11.586, -11.367, -11.188, -11.065, -11.006,
    -11.018, -11.098, -12.124, -11.958, -11.789, -11.560, -11.340, -11.163,
    -11.042, -10.987, -11.002, -11.085, -12.126, -11.952, -11.770, -11.534,
    -11.313, -11.137, -11.020, -10.968, -10.986, -11.072, -12.128, -11.944,
    -11.750, -11.508, -11.285, -11.111, -10.997, -10.948, -10.970, -11.059,
    -12.129, -11.936, -11.728, -11.480, -11.257, -11.085, -10.974, -10.929,
    -10.953, -11.046, -12.129, -11.926, -11.706, -11.452, -11.228, -11.059,
    -10.951, -10.909, -10.937, -11.033, -12.128, -11.915, -11.682, -11.423,
    -11.199, -11.033, -10.927, -10.889, -10.921, -11.020, -12.126, -11.902,
    -11.658, -11.394, -11.170, -11.006, -10.904, -10.869, -10.905, -11.007,
    -12.124, -11.889, -11.632, -11.364, -11.141, -10.979, -10.881, -10.850,
    -10.888, -10.994, -12.120, -11.874, -11.606, -11.333, -11.111, -10.952,
    -10.858, -10.830, -10.872, -10.981, -12.116, -11.858, -11.579, -11.302,
    -11.081, -10.925, -10.834, -10.810, -10.856, -10.968, -12.110, -11.841,
    -11.550, -11.271, -11.051, -10.898, -10.811, -10.791, -10.840, -10.956,
    -12.104, -11.822, -11.521, -11.239, -11.021, -10.871, -10.788, -10.771,
    -10.825, -10.943, -12.096, -11.802, -11.491, -11.207, -10.990, -10.844,
    -10.764, -10.752, -10.809, -10.931, -12.088, -11.781, -11.461, -11.174,
    -10.960, -10.817, -10.741, -10.733, -10.793, -10.918, -12.078, -11.759,
    -11.429, -11.141, -10.929, -10.790, -10.718, -10.713, -10.777, -10.906,
    -12.067, -11.735, -11.397, -11.108, -10.898, -10.763, -10.695, -10.694,
    -10.762, -10.894, -12.056, -11.711, -11.365, -11.075, -10.868, -10.736,
    -10.672, -10.675, -10.746, -10.881, -12.042, -11.685, -11.332, -11.041,
    -10.837, -10.709, -10.649, -10.656, -10.731, -10.869, -12.028, -11.659,
    -11.298, -11.007, -10.806, -10.682, -10.626, -10.637, -10.716, -10.857,
    -12.013, -11.631, -11.264, -10.973, -10.775, -10.655, -10.603, -10.618,
    -10.700, -10.845, -11.996, -11.603, -11.230, -10.939, -10.744, -10.628,
    -10.581, -10.599, -10.685, -10.834, -11.979, -11.573, -11.195, -10.906,
    -10.714, -10.602, -10.558, -10.581, -10.671, -10.822, -11.960, -11.543,
    -11.160, -10.872, -10.683, -10.576, -10.536, -10.563, -10.656, -10.811,
    -11.940, -11.512, -11.125, -10.838, -10.653, -10.549, -10.514, -10.545,
    -10.642, -10.799, -11.920, -11.481, -11.089, -10.804, -10.623, -10.524,
    -10.492, -10.527, -10.627, -10.788, -11.898, -11.449, -11.054, -10.771,
    -10.593, -10.498, -10.471, -10.510, -10.613, -10.777, -11.875, -11.416,
    -11.019, -10.737, -10.564, -10.473, -10.450, -10.492, -10.600, -10.767,
    -11.851, -11.384, -10.983, -10.704, -10.535, -10.448, -10.429, -10.475,
    -10.586, -10.756, -11.827, -11.350, -10.948, -10.672, -10.506, -10.423,
    -10.408, -10.459, -10.573, -10.746, -11.802, -11.317, -10.913, -10.639,
    -10.477, -10.399, -10.388, -10.442, -10.560, -10.735, -11.776, -11.284,
    -10.879, -10.607, -10.449, -10.375, -10.368, -10.426, -10.547, -10.726,
    -11.750, -11.250, -10.844, -10.576, -10.422, -10.352, -10.349, -10.410,
    -10.534, -10.716, -11.723, -11.217, -10.810, -10.545, -10.395, -10.329,
    -10.330, -10.395, -10.522, -10.706, -11.696, -11.184, -10.777, -10.514,
    -10.368, -10.306, -10.312, -10.380, -10.510, -10.697, -11.669, -11.151,
    -10.744, -10.485, -10.343, -10.285, -10.294, -10.365, -10.499, -10.688,
    -11.642, -11.119, -10.712, -10.456, -10.317, -10.263, -10.276, -10.351,
    -10.487, -10.679, -11.615, -11.087, -10.681, -10.427, -10.293, -10.243,
    -10.259, -10.337, -10.477, -10.671, -11.588, -11.056, -10.651, -10.400,
    -10.269, -10.223, -10.242, -10.324, -10.466, -10.663, -11.562, -11.025,
    -10.621, -10.374, -10.246, -10.203, -10.227, -10.311, -10.456, -10.655,
    -11.536, -10.996, -10.592, -10.348, -10.224, -10.185, -10.211, -10.299,
    -10.446, -10.647, -11.396, -10.854, -10.470, -10.249, -10.147, -10.125,
    -10.165, -10.264, -10.420, -10.628, -11.261, -10.718, -10.353, -10.153,
    -10.070, -10.065, -10.119, -10.229, -10.394, -10.608, -11.056, -10.520,
    -10.182, -10.014, -9.959,  -9.979,  -10.053, -10.179, -10.356, -10.581,
    -10.853, -10.325, -10.013, -9.873,  -9.846,  -9.889,  -9.983,  -10.125,
    -10.316, -10.551, -10.760, -10.239, -9.939,  -9.812,  -9.796,  -9.849,
    -9.951,  -10.101, -10.297, -10.537, -10.635, -10.128, -9.846,  -9.737,
    -9.736,  -9.801,  -9.914,  -10.072, -10.275, -10.520, -10.580, -10.085,
    -9.813,  -9.711,  -9.716,  -9.786,  -9.902,  -10.063, -10.268, -10.515,
    -10.612, -10.123, -9.851,  -9.746,  -9.746,  -9.810,  -9.921,  -10.077,
    -10.278, -10.522, -10.549, -10.075, -9.816,  -9.721,  -9.727,  -9.796,
    -9.911,  -10.069, -10.272, -10.517, -10.552, -10.085, -9.829,  -9.733,
    -9.738,  -9.805,  -9.917,  -10.074, -10.275, -10.519, -10.603, -10.138,
    -9.877,  -9.775,  -9.772,  -9.831,  -9.937,  -10.088, -10.286, -10.526,
    -10.535, -10.083, -9.834,  -9.741,  -9.746,  -9.812,  -9.922,  -10.077,
    -10.277, -10.519, -10.552, -10.104, -9.855,  -9.759,  -9.761,  -9.823,
    -9.930,  -10.082, -10.280, -10.521, -10.605, -10.156, -9.899,  -9.795,
    -9.788,  -9.843,  -9.946,  -10.093, -10.288, -10.526, -10.528, -10.080,
    -9.831,  -9.736,  -9.740,  -9.804,  -9.914,  -10.069, -10.269, -10.511,
    -10.414, -9.967,  -9.728,  -9.648,  -9.666,  -9.744,  -9.866,  -10.031,
    -10.240, -10.489, -10.315, -9.865,  -9.632,  -9.562,  -9.593,  -9.684,
    -9.818,  -9.993,  -10.210, -10.467, -10.123, -9.683,  -9.470,  -9.423,
    -9.477,  -9.589,  -9.742,  -9.934,  -10.165, -10.432, -9.968,  -9.546,
    -9.354,  -9.328,  -9.399,  -9.527,  -9.693,  -9.895,  -10.135, -10.410,
    -9.930,  -9.521,  -9.340,  -9.320,  -9.396,  -9.527,  -9.694,  -9.897,
    -10.136, -10.411, -9.906,  -9.516,  -9.347,  -9.335,  -9.414,  -9.544,
    -9.710,  -9.910,  -10.147, -10.418, -9.964,  -9.587,  -9.422,  -9.406,
    -9.478,  -9.599,  -9.755,  -9.946,  -10.174, -10.438, -10.108, -9.734,
    -9.560,  -9.528,  -9.581,  -9.683,  -9.822,  -9.997,  -10.212, -10.466,
    -10.293, -9.907,  -9.710,  -9.652,  -9.679,  -9.759,  -9.879,  -10.040,
    -10.244, -10.488, -10.466, -10.061, -9.836,  -9.750,  -9.754,  -9.815,
    -9.920,  -10.069, -10.264, -10.502, -10.524, -10.110, -9.873,  -9.777,
    -9.773,  -9.828,  -9.929,  -10.075, -10.268, -10.504, -10.411, -10.014,
    -9.796,  -9.718,  -9.728,  -9.794,  -9.904,  -10.056, -10.253, -10.493,
    -10.188, -9.816,  -9.633,  -9.588,  -9.628,  -9.718,  -9.846,  -10.012,
    -10.221, -10.469, -9.963,  -9.606,  -9.450,  -9.436,  -9.505,  -9.621,
    -9.770,  -9.955,  -10.178, -10.437, -9.779,  -9.424,  -9.281,  -9.287,
    -9.378,  -9.516,  -9.686,  -9.889,  -10.127, -10.398, -9.680,  -9.317,
    -9.175,  -9.188,  -9.289,  -9.440,  -9.623,  -9.837,  -10.086, -10.367,
    -9.654,  -9.279,  -9.130,  -9.141,  -9.244,  -9.399,  -9.587,  -9.807,
    -10.061, -10.348, -9.645,  -9.261,  -9.108,  -9.117,  -9.220,  -9.376,
    -9.566,  -9.789,  -10.047, -10.336, -9.750,  -9.354,  -9.186,  -9.181,
    -9.272,  -9.417,  -9.599,  -9.814,  -10.065, -10.349, -9.888,  -9.485,
    -9.304,  -9.283,  -9.358,  -9.488,  -9.656,  -9.859,  -10.099, -10.374,
    -9.941,  -9.547,  -9.366,  -9.342,  -9.410,  -9.533,  -9.692,  -9.888,
    -10.121, -10.390, -10.044, -9.650,  -9.462,  -9.426,  -9.482,  -9.592,
    -9.739,  -9.924,  -10.149, -10.410, -10.103, -9.712,  -9.522,  -9.480,
    -9.529,  -9.630,  -9.771,  -9.948,  -10.167, -10.423, -10.033, -9.654,
    -9.477,  -9.446,  -9.503,  -9.611,  -9.756,  -9.938,  -10.158, -10.416,
    -10.022, -9.644,  -9.469,  -9.441,  -9.500,  -9.609,  -9.755,  -9.936,
    -10.157, -10.414, -10.066, -9.682,  -9.500,  -9.465,  -9.519,  -9.624,
    -9.766,  -9.945,  -10.162, -10.418, -9.981,  -9.603,  -9.432,  -9.409,
    -9.474,  -9.589,  -9.739,  -9.923,  -10.146, -10.405, -9.964,  -9.584,
    -9.414,  -9.393,  -9.461,  -9.578,  -9.729,  -9.916,  -10.140, -10.400,
    -9.984,  -9.593,  -9.416,  -9.392,  -9.458,  -9.574,  -9.726,  -9.912,
    -10.136, -10.396, -9.916,  -9.527,  -9.355,  -9.339,  -9.414,  -9.538,
    -9.697,  -9.890,  -10.119, -10.383, -9.956,  -9.557,  -9.378,  -9.357,
    -9.427,  -9.548,  -9.704,  -9.894,  -10.121, -10.384, -10.051, -9.637,
    -9.443,  -9.408,  -9.467,  -9.580,  -9.728,  -9.912,  -10.134, -10.393,
    -10.138, -9.714,  -9.508,  -9.463,  -9.511,  -9.615,  -9.755,  -9.933,
    -10.149, -10.403, -10.343, -9.896,  -9.662,  -9.588,  -9.612,  -9.694,
    -9.816,  -9.978,  -10.183, -10.427, -10.551, -10.076, -9.810,  -9.707,
    -9.705,  -9.766,  -9.871,  -10.019, -10.212, -10.447, -10.831, -10.317,
    -10.003, -9.857,  -9.821,  -9.853,  -9.936,  -10.067, -10.247, -10.471,
    -10.836, -10.323, -10.010, -9.864,  -9.827,  -9.859,  -9.941,  -10.070,
    -10.249, -10.472, -10.842, -10.330, -10.017, -9.872,  -9.835,  -9.865,
    -9.946,  -10.074, -10.251, -10.473, -10.851, -10.339, -10.026, -9.881,
    -9.843,  -9.872,  -9.952,  -10.079, -10.254, -10.474, -10.861, -10.349,
    -10.037, -9.891,  -9.852,  -9.880,  -9.958,  -10.083, -10.257, -10.476,
    -10.872, -10.360, -10.048, -9.902,  -9.862,  -9.889,  -9.965,  -10.089,
    -10.261, -10.478, -10.885, -10.373, -10.060, -9.914,  -9.873,  -9.898,
    -9.973,  -10.095, -10.265, -10.481, -10.900, -10.386, -10.074, -9.926,
    -9.884,  -9.908,  -9.981,  -10.101, -10.269, -10.483, -10.915, -10.401,
    -10.088, -9.940,  -9.897,  -9.919,  -9.990,  -10.108, -10.274, -10.486,
    -10.932, -10.417, -10.104, -9.954,  -9.910,  -9.930,  -9.999,  -10.115,
    -10.279, -10.489, -10.950, -10.434, -10.120, -9.969,  -9.923,  -9.942,
    -10.009, -10.122, -10.284, -10.493, -10.969, -10.452, -10.137, -9.985,
    -9.937,  -9.954,  -10.019, -10.130, -10.290, -10.496, -10.990, -10.471,
    -10.154, -10.002, -9.952,  -9.967,  -10.030, -10.139, -10.296, -10.500,
    -11.011, -10.490, -10.173, -10.019, -9.968,  -9.980,  -10.041, -10.147,
    -10.303, -10.504, -11.032, -10.511, -10.192, -10.036, -9.983,  -9.994,
    -10.052, -10.157, -10.309, -10.509, -11.055, -10.531, -10.211, -10.054,
    -10.000, -10.008, -10.064, -10.166, -10.316, -10.513, -11.079, -10.553,
    -10.232, -10.073, -10.016, -10.023, -10.076, -10.176, -10.324, -10.518,
    -11.103, -10.575, -10.252, -10.092, -10.034, -10.038, -10.089, -10.186,
    -10.331, -10.523, -11.128, -10.598, -10.273, -10.111, -10.051, -10.053,
    -10.102, -10.196, -10.339, -10.528, -11.153, -10.621, -10.295, -10.131,
    -10.069, -10.069, -10.115, -10.207, -10.347, -10.534, -11.179, -10.645,
    -10.317, -10.152, -10.087, -10.085, -10.129, -10.218, -10.355, -10.539,
    -11.206, -10.669, -10.339, -10.172, -10.106, -10.101, -10.143, -10.229,
    -10.363, -10.545, -11.233, -10.693, -10.362, -10.193, -10.125, -10.118,
    -10.157, -10.240, -10.372, -10.551, -11.260, -10.718, -10.385, -10.214,
    -10.144, -10.135, -10.171, -10.252, -10.381, -10.557, -11.288, -10.744,
    -10.409, -10.236, -10.163, -10.152, -10.185, -10.264, -10.390, -10.564,
    -11.316, -10.769, -10.432, -10.258, -10.183, -10.169, -10.200, -10.276,
    -10.399, -10.570, -11.345, -10.795, -10.456, -10.280, -10.203, -10.187,
    -10.215, -10.288, -10.409, -10.577, -11.373, -10.821, -10.480, -10.302,
    -10.223, -10.205, -10.230, -10.300, -10.418, -10.584, -11.403, -10.847,
    -10.505, -10.324, -10.243, -10.223, -10.246, -10.313, -10.428, -10.591,
    -11.432, -10.874, -10.529, -10.347, -10.264, -10.241, -10.261, -10.326,
    -10.438, -10.598, -11.461, -10.901, -10.554, -10.370, -10.285, -10.259,
    -10.277, -10.339, -10.448, -10.605, -11.491, -10.928, -10.579, -10.393,
    -10.305, -10.277, -10.293, -10.352, -10.458, -10.612, -11.521, -10.955,
    -10.604, -10.416, -10.326, -10.296, -10.309, -10.365, -10.469, -10.620,
    -11.551, -10.982, -10.629, -10.439, -10.347, -10.315, -10.325, -10.379,
    -10.479, -10.627, -11.581, -11.009, -10.654, -10.462, -10.369, -10.334,
    -10.342, -10.392, -10.490, -10.635, -11.611, -11.036, -10.679, -10.485,
    -10.390, -10.353, -10.358, -10.406, -10.501, -10.643, -11.642, -11.064,
    -10.705, -10.509, -10.411, -10.372, -10.374, -10.419, -10.511, -10.651,
    -11.672, -11.091, -10.730, -10.532, -10.433, -10.391, -10.391, -10.433,
    -10.522, -10.659, -11.702, -11.119, -10.756, -10.556, -10.454, -10.410,
    -10.408, -10.447, -10.533, -10.667, -11.733, -11.146, -10.781, -10.580,
    -10.476, -10.429, -10.425, -10.461, -10.545, -10.675, -11.763, -11.174,
    -10.807, -10.603, -10.497, -10.449, -10.441, -10.475, -10.556, -10.683,
    -11.793, -11.201, -10.832, -10.627, -10.519, -10.468, -10.458, -10.490,
    -10.567, -10.692, -11.824, -11.229, -10.858, -10.651, -10.541, -10.488,
    -10.475, -10.504, -10.578, -10.700, -11.854, -11.256, -10.884, -10.674,
    -10.563, -10.507, -10.493, -10.518, -10.590, -10.708, -11.884, -11.284,
    -10.909, -10.698, -10.584, -10.527, -10.510, -10.533, -10.601, -10.717,
    -11.914, -11.311, -10.935, -10.722, -10.606, -10.546, -10.527, -10.547,
    -10.613, -10.726, -11.944, -11.339, -10.960, -10.746, -10.628, -10.566,
    -10.544, -10.562, -10.624, -10.734, -11.974, -11.366, -10.986, -10.769,
    -10.650, -10.586, -10.561, -10.576, -10.636, -10.743, -12.004, -11.393,
    -11.011, -10.793, -10.672, -10.605, -10.579, -10.591, -10.648, -10.752,
    -12.034, -11.421, -11.037, -10.817, -10.693, -10.625, -10.596, -10.606,
    -10.660, -10.760, -12.063, -11.448, -11.062, -10.841, -10.715, -10.645,
    -10.613, -10.620, -10.671, -10.769, -12.093, -11.475, -11.087, -10.864,
    -10.737, -10.665, -10.631, -10.635, -10.683, -10.778, -12.122, -11.502,
    -11.113, -10.888, -10.759, -10.684, -10.648, -10.650, -10.695, -10.787,
    -12.151, -11.529, -11.138, -10.911, -10.781, -10.704, -10.665, -10.664,
    -10.707, -10.796, -12.180, -11.555, -11.163, -10.935, -10.802, -10.724,
    -10.683, -10.679, -10.719, -10.805, -12.209, -11.582, -11.188, -10.958,
    -10.824, -10.743, -10.700, -10.694, -10.731, -10.814, -12.237, -11.609,
    -11.213, -10.982, -10.846, -10.763, -10.717, -10.709, -10.743, -10.823,
    -12.265, -11.635, -11.238, -11.005, -10.867, -10.783, -10.735, -10.723,
    -10.755, -10.831, -12.294, -11.661, -11.263, -11.029, -10.889, -10.802,
    -10.752, -10.738, -10.766, -10.840, -12.321, -11.688, -11.288, -11.052,
    -10.911, -10.822, -10.769, -10.753, -10.778, -10.849, -12.349, -11.714,
    -11.312, -11.075, -10.932, -10.842, -10.787, -10.768, -10.790, -10.858,
    -12.377, -11.740, -11.337, -11.098, -10.954, -10.861, -10.804, -10.782,
    -10.802, -10.867, -12.404, -11.765, -11.361, -11.121, -10.975, -10.881,
    -10.821, -10.797, -10.814, -10.876, -12.431, -11.791, -11.386, -11.144,
    -10.996, -10.900, -10.839, -10.812, -10.826, -10.885, -12.458, -11.816,
    -11.410, -11.167, -11.018, -10.920, -10.856, -10.826, -10.838, -10.894,
    -12.484, -11.842, -11.434, -11.190, -11.039, -10.939, -10.873, -10.841,
    -10.850, -10.903, -12.510, -11.867, -11.458, -11.213, -11.060, -10.958,
    -10.890, -10.856, -10.861, -10.912, -12.536, -11.892, -11.482, -11.235,
    -11.081, -10.978, -10.907, -10.870, -10.873, -10.921, -12.562, -11.917,
    -11.506, -11.258, -11.102, -10.997, -10.924, -10.885, -10.885, -10.930,
    -12.587, -11.942, -11.530, -11.280, -11.123, -11.016, -10.941, -10.899,
    -10.897, -10.939, -12.613, -11.966, -11.553, -11.303, -11.144, -11.035,
    -10.958, -10.914, -10.908, -10.947, -12.638, -11.991, -11.577, -11.325,
    -11.165, -11.054, -10.975, -10.928, -10.920, -10.956, -12.662, -12.015,
    -11.600, -11.347, -11.186, -11.073, -10.992, -10.942, -10.932, -10.965,
    -12.687, -12.039, -11.623, -11.370, -11.207, -11.092, -11.009, -10.957,
    -10.943, -10.974, -12.711, -12.063, -11.646, -11.392, -11.227, -11.111,
    -11.025, -10.971, -10.955, -10.982, -12.735, -12.086, -11.669, -11.414,
    -11.248, -11.130, -11.042, -10.985, -10.966, -10.991, -12.758, -12.110,
    -11.692, -11.435, -11.269, -11.149, -11.059, -10.999, -10.977, -10.999,
    -12.782, -12.133, -11.715, -11.457, -11.289, -11.167, -11.075, -11.013,
    -10.989, -11.008, -12.805, -12.157, -11.738, -11.479, -11.309, -11.186,
    -11.092, -11.027, -11.000, -11.016, -12.827, -12.180, -11.760, -11.501,
    -11.330, -11.205, -11.108, -11.041, -11.011, -11.025, -12.850, -12.202,
    -11.783, -11.522, -11.350, -11.223, -11.125, -11.055, -11.022, -11.033,
    -12.872, -12.225, -11.805, -11.543, -11.370, -11.242, -11.141, -11.069,
    -11.033, -11.041, -12.894, -12.248, -11.827, -11.565, -11.390, -11.260,
    -11.157, -11.082, -11.044, -11.049, -12.915, -12.270, -11.849, -11.586,
    -11.410, -11.278, -11.173, -11.096, -11.055, -11.057, -12.937, -12.292,
    -11.871, -11.607, -11.430, -11.296, -11.189, -11.109, -11.066, -11.065,
    -12.958, -12.314, -11.893, -11.628, -11.450, -11.314, -11.205, -11.123,
    -11.076, -11.073, -12.979, -12.336, -11.914, -11.649, -11.469, -11.332,
    -11.221, -11.136, -11.087, -11.081, -12.999, -12.357, -11.936, -11.670,
    -11.489, -11.350, -11.237, -11.149, -11.097, -11.089, -13.020, -12.379,
    -11.957, -11.690, -11.508, -11.368, -11.252, -11.162, -11.108, -11.097,
    -13.040, -12.400, -11.978, -11.711, -11.528, -11.386, -11.268, -11.175,
    -11.118, -11.104, -13.059, -12.421, -11.999, -11.731, -11.547, -11.404,
    -11.283, -11.188, -11.128, -11.112, -13.079, -12.442, -12.020, -11.752,
    -11.566, -11.421, -11.299, -11.201, -11.138, -11.119, -13.098, -12.463,
    -12.041, -11.772, -11.586, -11.439, -11.314, -11.214, -11.148, -11.127,
    -13.117, -12.484, -12.062, -11.792, -11.605, -11.456, -11.329, -11.226,
    -11.158, -11.134, -13.136, -12.504, -12.082, -11.812, -11.624, -11.473,
    -11.344, -11.239, -11.168, -11.141, -13.154, -12.524, -12.103, -11.832,
    -11.642, -11.490, -11.359, -11.251, -11.178, -11.148, -13.173, -12.544,
    -12.123, -11.852, -11.661, -11.507, -11.374, -11.263, -11.187, -11.155,
    -13.191, -12.564, -12.143, -11.872, -11.680, -11.524, -11.388, -11.275,
    -11.196, -11.162, -13.209, -12.584, -12.163, -11.891, -11.698, -11.541,
    -11.403, -11.287, -11.206, -11.169, -13.226, -12.604, -12.183, -11.911,
    -11.717, -11.558, -11.417, -11.299, -11.215, -11.175, -13.243, -12.623,
    -12.203, -11.930, -11.735, -11.574, -11.432, -11.311, -11.224, -11.182,
    -13.261, -12.642, -12.223, -11.950, -11.753, -11.591, -11.446, -11.322,
    -11.232, -11.188, -13.277, -12.662, -12.242, -11.969, -11.772, -11.607,
    -11.460, -11.333, -11.241, -11.195, -13.294, -12.681, -12.262, -11.988,
    -11.790, -11.624, -11.474, -11.345, -11.250, -11.201, -13.311, -12.699,
    -12.281, -12.007, -11.808, -11.640, -11.487, -11.356, -11.258, -11.207,
    -13.327, -12.718, -12.300, -12.026, -11.825, -11.656, -11.501, -11.366,
    -11.266, -11.213, -13.343, -12.736, -12.319, -12.044, -11.843, -11.671,
    -11.514, -11.377, -11.274, -11.218, -13.359, -12.755, -12.338, -12.063,
    -11.861, -11.687, -11.528, -11.388, -11.282, -11.224, -13.374, -12.773,
    -12.357, -12.082, -11.878, -11.703, -11.541, -11.398, -11.290, -11.230,
    -13.390, -12.791, -12.376, -12.100, -11.895, -11.718, -11.554, -11.408,
    -11.298, -11.235, -13.405, -12.809, -12.394, -12.118, -11.913, -11.733,
    -11.566, -11.418, -11.305, -11.240, -13.420, -12.827, -12.413, -12.137,
    -11.930, -11.749, -11.579, -11.428, -11.312, -11.245, -13.435, -12.844,
    -12.431, -12.155, -11.947, -11.764, -11.591, -11.438, -11.319, -11.250,
    -13.450, -12.862, -12.449, -12.173, -11.963, -11.778, -11.603, -11.447,
    -11.326, -11.255, -13.465, -12.879, -12.467, -12.190, -11.980, -11.793,
    -11.615, -11.456, -11.333, -11.260, -13.479, -12.896, -12.485, -12.208,
    -11.997, -11.807, -11.627, -11.465, -11.340, -11.265, -13.493, -12.913,
    -12.503, -12.226, -12.013, -11.822, -11.639, -11.474, -11.346, -11.269,
    -13.507, -12.930, -12.521, -12.243, -12.029, -11.836, -11.650, -11.482,
    -11.352, -11.273, -13.521, -12.947, -12.539, -12.261, -12.045, -11.850,
    -11.661, -11.491, -11.358, -11.277, -13.535, -12.964, -12.556, -12.278,
    -12.061, -11.863, -11.672, -11.499, -11.364, -11.281, -13.549, -12.980,
    -12.573, -12.295, -12.077, -11.877, -11.683, -11.507, -11.370, -11.285,
    -13.562, -12.997, -12.591, -12.312, -12.093, -11.890, -11.693, -11.515,
    -11.375, -11.289, -13.576, -13.013, -12.608, -12.329, -12.108, -11.903,
    -11.703, -11.522, -11.380, -11.292, -13.589, -13.029, -12.625, -12.346,
    -12.124, -11.916, -11.713, -11.529, -11.385, -11.296, -13.602, -13.045,
    -12.642, -12.362, -12.139, -11.929, -11.723, -11.536, -11.390, -11.299,
    -13.615, -13.061, -12.659, -12.379, -12.154, -11.942, -11.733, -11.543,
    -11.394, -11.302, -13.628, -13.077, -12.675, -12.395, -12.169, -11.954,
    -11.742, -11.549, -11.399, -11.305, -13.641, -13.092, -12.692, -12.411,
    -12.184, -11.966, -11.751, -11.555, -11.403, -11.308, -13.654, -13.108,
    -12.708, -12.428, -12.198, -11.978, -11.759, -11.561, -11.407, -11.310,
    -13.666, -13.123, -12.725, -12.443, -12.212, -11.989, -11.768, -11.567,
    -11.411, -11.313, -13.679, -13.139, -12.741, -12.459, -12.226, -12.000,
    -11.776, -11.573, -11.414, -11.315, -13.691, -13.154, -12.757, -12.475,
    -12.240, -12.011, -11.784, -11.578, -11.417, -11.317, -13.703, -13.169,
    -12.773, -12.491, -12.254, -12.022, -11.791, -11.583, -11.421, -11.319,
    -13.715, -13.184, -12.789, -12.506, -12.267, -12.032, -11.799, -11.587,
    -11.423, -11.321, -13.727, -13.199, -12.805, -12.521, -12.281, -12.043,
    -11.805, -11.592, -11.426, -11.323, -13.739, -13.214, -12.820, -12.536,
    -12.294, -12.053, -11.812, -11.596, -11.429, -11.325, -13.751, -13.228,
    -12.836, -12.551, -12.306, -12.062, -11.818, -11.600, -11.431, -11.326,
    -13.763, -13.243, -12.851, -12.566, -12.319, -12.071, -11.824, -11.603,
    -11.433, -11.327, -13.775, -13.257, -12.867, -12.580, -12.331, -12.080,
    -11.830, -11.606, -11.435, -11.329, -13.787, -13.272, -12.882, -12.595,
    -12.343, -12.089, -11.835, -11.609, -11.436, -11.330, -13.798, -13.286,
    -12.897, -12.609, -12.355, -12.097, -11.841, -11.612, -11.438, -11.330,
    -13.810, -13.300, -12.912, -12.623, -12.367, -12.105, -11.845, -11.614,
    -11.439, -11.331, -13.821, -13.314, -12.927, -12.637, -12.378, -12.113,
    -11.850, -11.617, -11.440, -11.332, -13.833, -13.328, -12.941, -12.651,
    -12.389, -12.120, -11.854, -11.619, -11.441, -11.332, -13.844, -13.342,
    -12.956, -12.664, -12.400, -12.127, -11.857, -11.620, -11.442, -11.333,
    -13.855, -13.356, -12.970, -12.678, -12.410, -12.134, -11.861, -11.622,
    -11.442, -11.333, -13.867, -13.370, -12.985, -12.691, -12.420, -12.140,
    -11.864, -11.623, -11.443, -11.333, -13.878, -13.384, -12.999, -12.704,
    -12.430, -12.146, -11.867, -11.624, -11.443, -11.333, -13.889, -13.397,
    -13.013, -12.717, -12.440, -12.152, -11.869, -11.624, -11.443, -11.333,
    -13.900, -13.411, -13.027, -12.729, -12.449, -12.157, -11.871, -11.625,
    -11.442, -11.333, -13.911, -13.424, -13.041, -12.741, -12.458, -12.162,
    -11.873, -11.625, -11.442, -11.333, -13.922, -13.438, -13.055, -12.754,
    -12.466, -12.167, -11.874, -11.625, -11.441, -11.332, -13.933, -13.451,
    -13.068, -12.765, -12.475, -12.171, -11.876, -11.624, -11.441, -11.332,
    -13.944, -13.464, -13.082, -12.777, -12.482, -12.175, -11.876, -11.624,
    -11.440, -11.331, -13.955, -13.477, -13.095, -12.788, -12.490, -12.178,
    -11.877, -11.623, -11.439, -11.331, -13.966, -13.490, -13.108, -12.800,
    -12.497, -12.181, -11.877, -11.622, -11.438, -11.330, -13.976, -13.503,
    -13.121, -12.811, -12.504, -12.184, -11.877, -11.621, -11.437, -11.329,
    -13.987, -13.516, -13.134, -12.821, -12.511, -12.186, -11.877, -11.619,
    -11.435, -11.329, -13.998, -13.529, -13.147, -12.832, -12.517, -12.188,
    -11.876, -11.618, -11.434, -11.328, -14.009, -13.542, -13.160, -12.842,
    -12.523, -12.190, -11.875, -11.616, -11.432, -11.327, -14.019, -13.554,
    -13.172, -12.852, -12.528, -12.191, -11.874, -11.614, -11.431, -11.326,
    -14.030, -13.567, -13.185, -12.862, -12.534, -12.193, -11.873, -11.612,
    -11.429, -11.325, -14.041, -13.579, -13.197, -12.871, -12.539, -12.193,
    -11.872, -11.610, -11.428, -11.324, -14.051, -13.592, -13.209, -12.880,
    -12.543, -12.194, -11.870, -11.608, -11.426, -11.323, -14.062, -13.604,
    -13.221, -12.889, -12.548, -12.194, -11.868, -11.606, -11.424, -11.322,
    -14.072, -13.616, -13.233, -12.898, -12.552, -12.194, -11.866, -11.603,
    -11.422, -11.321, -14.083, -13.629, -13.244, -12.906, -12.555, -12.194,
    -11.864, -11.601, -11.420, -11.320, -14.093, -13.641, -13.256, -12.915,
    -12.559, -12.194, -11.862, -11.598, -11.418, -11.319, -14.104, -13.653,
    -13.267, -12.923, -12.562, -12.193, -11.859, -11.596, -11.417, -11.318,
    -14.114, -13.665, -13.279, -12.930, -12.565, -12.192, -11.857, -11.593,
    -11.415, -11.317, -14.125, -13.677, -13.290, -12.938, -12.568, -12.192,
    -11.855, -11.591, -11.413, -11.316, -14.135, -13.689, -13.301, -12.945,
    -12.570, -12.191, -11.852, -11.588, -11.411, -11.315, -14.146, -13.701,
    -13.312, -12.953, -12.573, -12.189, -11.850, -11.586, -11.410, -11.314,
    -14.156, -13.712, -13.323, -12.960, -12.575, -12.188, -11.847, -11.584,
    -11.408, -11.314, -14.167, -13.724, -13.333, -12.967, -12.577, -12.187,
    -11.845, -11.581, -11.407, -11.313, -14.177, -13.736, -13.344, -12.974,
    -12.579, -12.186, -11.843, -11.579, -11.405, -11.312, -14.188, -13.748,
    -13.354, -12.980, -12.582, -12.185, -11.840, -11.577, -11.404, -11.312,
    -14.198, -13.759, -13.365, -12.987, -12.584, -12.184, -11.838, -11.575,
    -11.403, -11.311, -14.208, -13.767, -13.361, -12.964, -12.546, -12.143,
    -11.801, -11.546, -11.381, -11.297, -14.218, -13.775, -13.358, -12.944,
    -12.512, -12.106, -11.768, -11.520, -11.362, -11.284, -14.227, -13.781,
    -13.351, -12.918, -12.473, -12.063, -11.731, -11.490, -11.340, -11.269,
    -14.237, -13.790, -13.353, -12.909, -12.456, -12.045, -11.714, -11.477,
    -11.331, -11.262, -14.247, -13.798, -13.353, -12.897, -12.437, -12.025,
    -11.697, -11.464, -11.321, -11.256, -14.257, -13.808, -13.358, -12.896,
    -12.432, -12.019, -11.692, -11.460, -11.319, -11.255, -14.268, -13.821,
    -13.375, -12.916, -12.454, -12.039, -11.709, -11.474, -11.330, -11.263,
    -14.278, -13.830, -13.382, -12.919, -12.454, -12.039, -11.709, -11.474,
    -11.331, -11.264, -14.289, -13.842, -13.393, -12.930, -12.464, -12.048,
    -11.716, -11.480, -11.336, -11.268, -14.299, -13.856, -13.412, -12.953,
    -12.489, -12.070, -11.735, -11.495, -11.347, -11.277, -14.309, -13.865,
    -13.417, -12.953, -12.484, -12.065, -11.730, -11.492, -11.346, -11.276,
    -14.320, -13.877, -13.429, -12.965, -12.496, -12.075, -11.739, -11.499,
    -11.351, -11.280, -14.331, -13.891, -13.447, -12.987, -12.518, -12.095,
    -11.755, -11.512, -11.361, -11.287, -14.341, -13.899, -13.449, -12.978,
    -12.503, -12.078, -11.740, -11.500, -11.352, -11.282, -14.350, -13.905,
    -13.442, -12.956, -12.471, -12.045, -11.711, -11.477, -11.335, -11.270,
    -14.360, -13.912, -13.437, -12.935, -12.440, -12.013, -11.684, -11.455,
    -11.319, -11.259, -14.369, -13.912, -13.415, -12.889, -12.383, -11.957,
    -11.635, -11.417, -11.291, -11.239, -14.378, -13.914, -13.398, -12.854,
    -12.340, -11.917, -11.602, -11.391, -11.271, -11.225, -14.388, -13.922,
    -13.404, -12.857, -12.343, -11.920, -11.605, -11.395, -11.275, -11.228,
    -14.398, -13.933, -13.414, -12.869, -12.358, -11.935, -11.620, -11.407,
    -11.285, -11.236, -14.409, -13.948, -13.440, -12.906, -12.399, -11.976,
    -11.655, -11.436, -11.308, -11.252, -14.421, -13.969, -13.480, -12.964,
    -12.464, -12.036, -11.706, -11.476, -11.338, -11.274, -14.433, -13.990,
    -13.519, -13.020, -12.524, -12.091, -11.751, -11.511, -11.364, -11.293,
    -14.445, -14.010, -13.552, -13.065, -12.571, -12.132, -11.785, -11.537,
    -11.383, -11.306, -14.456, -14.023, -13.569, -13.084, -12.590, -12.147,
    -11.796, -11.546, -11.389, -11.310, -14.466, -14.030, -13.567, -13.073,
    -12.574, -12.132, -11.784, -11.537, -11.383, -11.306, -14.474, -14.030,
    -13.547, -13.032, -12.526, -12.088, -11.748, -11.510, -11.364, -11.293,
    -14.482, -14.027, -13.517, -12.976, -12.460, -12.027, -11.698, -11.471,
    -11.336, -11.274, -14.490, -14.022, -13.483, -12.914, -12.387, -11.957,
    -11.639, -11.424, -11.301, -11.249, -14.499, -14.021, -13.460, -12.869,
    -12.333, -11.905, -11.593, -11.387, -11.273, -11.229, -14.509, -14.028,
    -13.454, -12.850, -12.307, -11.877, -11.567, -11.366, -11.257, -11.217,
    -14.519, -14.036, -13.455, -12.842, -12.293, -11.862, -11.553, -11.354,
    -11.247, -11.210, -14.532, -14.055, -13.486, -12.880, -12.330, -11.894,
    -11.580, -11.375, -11.263, -11.222, -14.544, -14.078, -13.529, -12.938,
    -12.391, -11.950, -11.627, -11.412, -11.291, -11.242, -14.556, -14.095,
    -13.556, -12.975, -12.431, -11.987, -11.659, -11.438, -11.311, -11.257,
    -14.567, -14.114, -13.589, -13.020, -12.480, -12.032, -11.697, -11.469,
    -11.334, -11.273, -14.579, -14.130, -13.613, -13.053, -12.515, -12.065,
    -11.725, -11.491, -11.351, -11.285, -14.589, -14.136, -13.612, -13.046,
    -12.507, -12.058, -11.719, -11.487, -11.349, -11.283, -14.599, -14.145,
    -13.618, -13.048, -12.507, -12.057, -11.719, -11.488, -11.349, -11.284,
    -14.610, -14.159, -13.635, -13.067, -12.525, -12.073, -11.732, -11.498,
    -11.357, -11.289, -14.620, -14.163, -13.627, -13.048, -12.503, -12.053,
    -11.715, -11.485, -11.348, -11.283, -14.630, -14.171, -13.629, -13.044,
    -12.495, -12.045, -11.709, -11.480, -11.344, -11.280, -14.641, -14.183,
    -13.640, -13.052, -12.501, -12.049, -11.712, -11.482, -11.345, -11.281,
    -14.651, -14.187, -13.631, -13.032, -12.476, -12.026, -11.692, -11.467,
    -11.334, -11.273, -14.662, -14.200, -13.644, -13.043, -12.486, -12.033,
    -11.698, -11.472, -11.338, -11.275, -14.674, -14.218, -13.672, -13.077,
    -12.519, -12.062, -11.721, -11.489, -11.351, -11.284, -14.686, -14.234,
    -13.697, -13.107, -12.549, -12.088, -11.743, -11.506, -11.363, -11.292,
    -14.699, -14.259, -13.744, -13.173, -12.617, -12.149, -11.792, -11.544,
    -11.391, -11.312, -14.712, -14.282, -13.788, -13.234, -12.681, -12.205,
    -11.838, -11.579, -11.416, -11.329, -14.724, -14.300, -13.818, -13.275,
    -12.724, -12.243, -11.869, -11.603, -11.433, -11.340, -14.736, -14.318,
    -13.848, -13.315, -12.765, -12.280, -11.898, -11.625, -11.450, -11.351,
    -14.747, -14.329, -13.859, -13.325, -12.775, -12.288, -11.905, -11.631,
    -11.454, -11.354, -14.758, -14.340, -13.870, -13.336, -12.785, -12.297,
    -11.913, -11.637, -11.458, -11.357, -14.769, -14.351, -13.881, -13.347,
    -12.795, -12.306, -11.921, -11.643, -11.463, -11.360, -14.780, -14.363,
    -13.892, -13.358, -12.806, -12.316, -11.929, -11.650, -11.468, -11.364,
    -14.791, -14.374, -13.904, -13.370, -12.817, -12.326, -11.938, -11.657,
    -11.474, -11.367, -14.802, -14.386, -13.916, -13.382, -12.829, -12.337,
    -11.947, -11.665, -11.479, -11.371, -14.813, -14.397, -13.928, -13.395,
    -12.841, -12.348, -11.957, -11.673, -11.485, -11.374, -14.824, -14.409,
    -13.940, -13.407, -12.854, -12.360, -11.967, -11.681, -11.491, -11.378,
    -14.835, -14.420, -13.953, -13.420, -12.866, -12.372, -11.977, -11.689,
    -11.497, -11.383, -14.846, -14.432, -13.965, -13.433, -12.880, -12.384,
    -11.988, -11.698, -11.504, -11.387, -14.857, -14.444, -13.978, -13.447,
    -12.893, -12.397, -11.999, -11.707, -11.511, -11.391, -14.869, -14.456,
    -13.990, -13.461, -12.907, -12.409, -12.010, -11.716, -11.518, -11.396,
    -14.880, -14.467, -14.003, -13.475, -12.921, -12.423, -12.022, -11.726,
    -11.525, -11.401, -14.891, -14.479, -14.016, -13.489, -12.936, -12.436,
    -12.034, -11.736, -11.532, -11.405, -14.902, -14.491, -14.029, -13.503,
    -12.950, -12.450, -12.046, -11.746, -11.540, -11.410, -14.913, -14.503,
    -14.042, -13.517, -12.965, -12.464, -12.058, -11.756, -11.547, -11.415,
    -14.925, -14.515, -14.055, -13.532, -12.980, -12.478, -12.071, -11.766,
    -11.555, -11.420, -14.936, -14.527, -14.068, -13.547, -12.995, -12.493,
    -12.084, -11.777, -11.563, -11.426, -14.947, -14.539, -14.081, -13.561,
    -13.011, -12.508, -12.097, -11.788, -11.571, -11.431, -14.959, -14.550,
    -14.095, -13.576, -13.026, -12.523, -12.110, -11.799, -11.580, -11.437,
    -14.970, -14.562, -14.108, -13.591, -13.042, -12.538, -12.124, -11.810,
    -11.588, -11.442, -14.981, -14.574, -14.121, -13.606, -13.058, -12.553,
    -12.137, -11.821, -11.597, -11.448, -14.993, -14.586, -14.135, -13.621,
    -13.074, -12.568, -12.151, -11.833, -11.605, -11.453, -15.004, -14.598,
    -14.148, -13.637, -13.090, -12.584, -12.165, -11.844, -11.614, -11.459,
    -15.015, -14.610, -14.161, -13.652, -13.106, -12.600, -12.179, -11.856,
    -11.623, -11.465, -15.027, -14.622, -14.175, -13.667, -13.122, -12.615,
    -12.193, -11.868, -11.632, -11.471, -15.038, -14.634, -14.188, -13.682,
    -13.139, -12.631, -12.208, -11.880, -11.641, -11.477, -15.050, -14.646,
    -14.201, -13.698, -13.155, -12.647, -12.222, -11.892, -11.650, -11.483,
    -15.061, -14.658, -14.215, -13.713, -13.172, -12.663, -12.237, -11.904,
    -11.659, -11.489, -15.073, -14.670, -14.228, -13.728, -13.188, -12.680,
    -12.251, -11.916, -11.669, -11.495, -15.084, -14.682, -14.241, -13.744,
    -13.205, -12.696, -12.266, -11.929, -11.678, -11.501, -15.096, -14.694,
    -14.254, -13.759, -13.221, -12.712, -12.281, -11.941, -11.687, -11.507,
    -15.107, -14.706, -14.268, -13.774, -13.238, -12.728, -12.296, -11.953,
    -11.697, -11.513, -15.119, -14.718, -14.281, -13.789, -13.254, -12.745,
    -12.311, -11.966, -11.706, -11.519, -15.130, -14.730, -14.294, -13.805,
    -13.271, -12.761, -12.326, -11.979, -11.716, -11.525, -15.142, -14.742,
    -14.307, -13.820, -13.287, -12.778, -12.341, -11.991, -11.725, -11.531,
    -15.153, -14.754, -14.321, -13.835, -13.304, -12.794, -12.356, -12.004,
    -11.735, -11.537, -15.165, -14.766, -14.334, -13.850, -13.321, -12.811,
    -12.371, -12.016, -11.744, -11.544, -15.176, -14.777, -14.347, -13.865,
    -13.337, -12.827, -12.386, -12.029, -11.754, -11.550, -15.188, -14.789,
    -14.360, -13.881, -13.354, -12.844, -12.401, -12.042, -11.763, -11.556,
    -15.200, -14.801, -14.373, -13.896, -13.370, -12.860, -12.416, -12.055,
    -11.773, -11.562, -15.211, -14.813, -14.386, -13.911, -13.387, -12.877,
    -12.431, -12.067, -11.783, -11.568, -15.223, -14.825, -14.399, -13.926,
    -13.403, -12.893, -12.447, -12.080, -11.792, -11.574, -15.235, -14.837,
    -14.412, -13.941, -13.420, -12.910, -12.462, -12.093, -11.802, -11.580,
    -15.246, -14.849, -14.425, -13.955, -13.436, -12.926, -12.477, -12.106,
    -11.811, -11.586, -15.258, -14.861, -14.438, -13.970, -13.452, -12.943,
    -12.492, -12.118, -11.820, -11.591, -15.270, -14.873, -14.451, -13.985,
    -13.469, -12.959, -12.507, -12.131, -11.830, -11.597, -15.281, -14.884,
    -14.464, -14.000, -13.485, -12.976, -12.522, -12.144, -11.839, -11.603,
    -15.293, -14.896, -14.477, -14.015, -13.501, -12.992, -12.538, -12.156,
    -11.848, -11.609, -15.305, -14.908, -14.489, -14.029, -13.517, -13.008,
    -12.553, -12.169, -11.858, -11.614, -15.316, -14.920, -14.502, -14.044,
    -13.533, -13.025, -12.568, -12.181, -11.867, -11.620, -15.328, -14.932,
    -14.515, -14.058, -13.549, -13.041, -12.583, -12.194, -11.876, -11.625,
    -15.340, -14.944, -14.527, -14.073, -13.565, -13.057, -12.598, -12.206,
    -11.885, -11.631, -15.352, -14.956, -14.540, -14.087, -13.581, -13.074,
    -12.613, -12.219, -11.894, -11.636, -15.364, -14.967, -14.553, -14.101,
    -13.597, -13.090, -12.628, -12.231, -11.903, -11.641, -15.375, -14.979,
    -14.565, -14.116, -13.613, -13.106, -12.642, -12.243, -11.912, -11.646,
    -15.387, -14.991, -14.578, -14.130, -13.629, -13.122, -12.657, -12.256,
    -11.920, -11.651, -15.399, -15.003, -14.590, -14.144, -13.645, -13.138,
    -12.672, -12.268, -11.929, -11.656, -15.411, -15.015, -14.603, -14.158,
    -13.660, -13.154, -12.687, -12.280, -11.937, -11.661, -15.423, -15.026,
    -14.615, -14.172, -13.676, -13.169, -12.701, -12.292, -11.946, -11.666,
    -15.434, -15.038, -14.628, -14.186, -13.691, -13.185, -12.716, -12.303,
    -11.954, -11.670, -15.446, -15.050, -14.640, -14.200, -13.707, -13.201,
    -12.730, -12.315, -11.962, -11.675, -15.458, -15.062, -14.652, -14.214,
    -13.722, -13.217, -12.744, -12.327, -11.970, -11.679, -15.470, -15.074,
    -14.665, -14.228, -13.737, -13.232, -12.759, -12.338, -11.978, -11.683,
    -15.482, -15.085, -14.677, -14.241, -13.752, -13.248, -12.773, -12.350,
    -11.986, -11.687, -15.494, -15.097, -14.689, -14.255, -13.768, -13.263,
    -12.787, -12.361, -11.993, -11.691, -15.506, -15.109, -14.702, -14.269,
    -13.783, -13.279, -12.801, -12.372, -12.001, -11.695, -15.518, -15.121,
    -14.714, -14.282, -13.798, -13.294, -12.815, -12.383, -12.008, -11.699,
    -15.530, -15.133, -14.726, -14.296, -13.812, -13.309, -12.829, -12.394,
    -12.015, -11.702, -15.542, -15.144, -14.738, -14.309, -13.827, -13.324,
    -12.843, -12.405, -12.022, -11.706, -15.553, -15.156, -14.750, -14.323,
    -13.842, -13.339, -12.856, -12.416, -12.029, -11.709, -15.565, -15.168,
    -14.762, -14.336, -13.857, -13.354, -12.870, -12.426, -12.036, -11.712,
    -15.577, -15.180, -14.774, -14.349, -13.871, -13.369, -12.883, -12.436,
    -12.042, -11.715, -15.589, -15.191, -14.786, -14.362, -13.886, -13.384,
    -12.897, -12.447, -12.049, -11.717, -15.601, -15.203, -14.798, -14.376,
    -13.901, -13.399, -12.910, -12.457, -12.055, -11.720, -15.613, -15.215,
    -14.810, -14.389, -13.915, -13.413, -12.923, -12.467, -12.061, -11.722,
    -15.625, -15.227, -14.822, -14.402, -13.929, -13.428, -12.936, -12.476,
    -12.066, -11.725, -15.637, -15.238, -14.834, -14.415, -13.944, -13.442,
    -12.948, -12.486, -12.072, -11.727, -15.650, -15.250, -14.846, -14.428,
    -13.958, -13.457, -12.961, -12.495, -12.077, -11.729, -15.662, -15.262,
    -14.858, -14.441, -13.972, -13.471, -12.974, -12.504, -12.083, -11.730,
    -15.674, -15.274, -14.870, -14.453, -13.986, -13.485, -12.986, -12.513,
    -12.088, -11.732, -15.686, -15.285, -14.882, -14.466, -14.000, -13.499,
    -12.998, -12.522, -12.092, -11.733, -15.698, -15.297, -14.894, -14.479,
    -14.014, -13.513, -13.010, -12.531, -12.097, -11.734, -15.710, -15.309,
    -14.906, -14.492, -14.028, -13.527, -13.022, -12.539, -12.101, -11.735,
    -15.722, -15.321, -14.917, -14.504, -14.041, -13.541, -13.034, -12.547,
    -12.105, -11.736, -15.734, -15.332, -14.929, -14.517, -14.055, -13.554,
    -13.046, -12.555, -12.109, -11.737, -15.746, -15.344, -14.941, -14.530,
    -14.069, -13.568, -13.057, -12.563, -12.113, -11.737, -15.758, -15.356,
    -14.953, -14.542, -14.082, -13.581, -13.069, -12.570, -12.116, -11.738,
    -15.770, -15.368, -14.964, -14.555, -14.096, -13.595, -13.080, -12.578,
    -12.120, -11.738, -15.783, -15.379, -14.976, -14.567, -14.109, -13.608,
    -13.091, -12.585, -12.123, -11.737, -15.795, -15.391, -14.988, -14.579,
    -14.122, -13.621, -13.101, -12.592, -12.125, -11.737, -15.807, -15.403,
    -14.999, -14.592, -14.136, -13.634, -13.112, -12.598, -12.128, -11.737,
    -15.819, -15.414, -15.011, -14.604, -14.149, -13.647, -13.122, -12.604,
    -12.130, -11.736, -15.831, -15.426, -15.023, -14.616, -14.162, -13.659,
    -13.133, -12.610, -12.132, -11.735, -15.844, -15.438, -15.034, -14.628,
    -14.175, -13.672, -13.143, -12.616, -12.134, -11.734, -15.856, -15.450,
    -15.046, -14.641, -14.188, -13.685, -13.152, -12.622, -12.135, -11.733,
    -15.868, -15.461, -15.058, -14.653, -14.201, -13.697, -13.162, -12.627,
    -12.136, -11.731, -15.880, -15.473, -15.069, -14.665, -14.214, -13.709,
    -13.171, -12.632, -12.137, -11.729, -15.892, -15.485, -15.081, -14.677,
    -14.226, -13.721, -13.180, -12.637, -12.138, -11.727, -15.905, -15.497,
    -15.092, -14.689, -14.239, -13.733, -13.189, -12.641, -12.139, -11.725,
    -15.917, -15.508, -15.104, -14.701, -14.251, -13.745, -13.198, -12.645,
    -12.139, -11.723, -15.929, -15.520, -15.115, -14.713, -14.264, -13.757,
    -13.206, -12.649, -12.139, -11.721, -15.941, -15.532, -15.127, -14.725,
    -14.276, -13.768, -13.214, -12.653, -12.138, -11.718, -15.954, -15.544,
    -15.138, -14.737, -14.289, -13.779, -13.222, -12.656, -12.138, -11.715,
    -15.966, -15.555, -15.150, -14.748, -14.301, -13.791, -13.230, -12.659,
    -12.137, -11.712, -15.978, -15.567, -15.161, -14.760, -14.313, -13.802,
    -13.237, -12.662, -12.136, -11.709, -15.991, -15.579, -15.173, -14.772,
    -14.325, -13.813, -13.245, -12.664, -12.134, -11.705, -16.003, -15.591,
    -15.184, -14.784, -14.337, -13.823, -13.251, -12.666, -12.133, -11.702,
    -16.015, -15.603, -15.196, -14.795, -14.349, -13.834, -13.258, -12.668,
    -12.131, -11.698, -16.028, -15.614, -15.207, -14.807, -14.361, -13.844,
    -13.264, -12.669, -12.129, -11.694, -16.040, -15.626, -15.218, -14.818,
    -14.373, -13.854, -13.270, -12.670, -12.126, -11.690, -16.052, -15.638,
    -15.230, -14.830, -14.385, -13.864, -13.276, -12.671, -12.124, -11.686,
    -16.065, -15.650, -15.241, -14.842, -14.396, -13.874, -13.282, -12.672,
    -12.121, -11.681, -16.077, -15.661, -15.253, -14.853, -14.408, -13.884,
    -13.287, -12.672, -12.117, -11.676, -16.089, -15.673, -15.264, -14.864,
    -14.419, -13.893, -13.291, -12.672, -12.114, -11.672, -16.102, -15.685,
    -15.275, -14.876, -14.430, -13.902, -13.296, -12.671, -12.110, -11.667,
    -16.114, -15.697, -15.287, -14.887, -14.442, -13.911, -13.300, -12.670,
    -12.106, -11.662, -16.127, -15.708, -15.298, -14.899, -14.453, -13.920,
    -13.304, -12.669, -12.102, -11.656, -16.139, -15.720, -15.309, -14.910,
    -14.464, -13.929, -13.308, -12.668, -12.098, -11.651, -16.151, -15.732,
    -15.321, -14.921, -14.475, -13.937, -13.311, -12.666, -12.093, -11.645,
    -16.164, -15.744, -15.332, -14.932, -14.486, -13.945, -13.314, -12.664,
    -12.088, -11.640, -16.176, -15.756, -15.343, -14.944, -14.496, -13.953,
    -13.316, -12.661, -12.083, -11.634, -16.189, -15.767, -15.354, -14.955,
    -14.507, -13.961, -13.318, -12.659, -12.078, -11.628, -16.201, -15.779,
    -15.366, -14.966, -14.518, -13.968, -13.320, -12.656, -12.072, -11.622,
    -16.214, -15.791, -15.377, -14.977, -14.528, -13.976, -13.322, -12.652,
    -12.066, -11.616, -16.226, -15.803, -15.388, -14.988, -14.538, -13.983,
    -13.323, -12.649, -12.060, -11.609, -16.239, -15.815, -15.400, -14.999,
    -14.549, -13.989, -13.324, -12.645, -12.054, -11.603, -16.251, -15.826,
    -15.411, -15.010, -14.559, -13.996, -13.324, -12.640, -12.048, -11.596,
    -16.264, -15.838, -15.422, -15.021, -14.569, -14.002, -13.324, -12.636,
    -12.041, -11.590, -16.276, -15.850, -15.433, -15.032, -14.579, -14.008,
    -13.324, -12.631, -12.034, -11.583, -16.289, -15.862, -15.444, -15.043,
    -14.589, -14.013, -13.323, -12.626, -12.028, -11.576, -16.301, -15.874,
    -15.456, -15.054, -14.598, -14.019, -13.323, -12.621, -12.020, -11.569,
    -16.314, -15.885, -15.467, -15.065, -14.608, -14.024, -13.321, -12.615,
    -12.013, -11.562, -16.326, -15.897, -15.478, -15.076, -14.617, -14.028,
    -13.320, -12.609, -12.006, -11.555, -16.339, -15.909, -15.489, -15.086,
    -14.626, -14.033, -13.318, -12.603, -11.998, -11.548, -16.351, -15.921,
    -15.500, -15.097, -14.635, -14.037, -13.315, -12.596, -11.991, -11.541,
    -16.364, -15.933, -15.512, -15.108, -14.644, -14.041, -13.313, -12.590,
    -11.983, -11.534, -16.376, -15.944, -15.523, -15.118, -14.653, -14.044,
    -13.310, -12.583, -11.975, -11.526, -16.389, -15.956, -15.534, -15.129,
    -14.662, -14.048, -13.307, -12.576, -11.967, -11.519, -16.401, -15.968,
    -15.545, -15.140, -14.671, -14.051, -13.303, -12.568, -11.959, -11.512,
    -16.414, -15.980, -15.556, -15.150, -14.679, -14.053, -13.299, -12.561,
    -11.951, -11.505, -16.427, -15.992, -15.567, -15.161, -14.687, -14.056,
    -13.295, -12.553, -11.943, -11.497, -16.439, -16.004, -15.579, -15.171,
    -14.695, -14.058, -13.290, -12.546, -11.934, -11.490, -16.452, -16.015,
    -15.590, -15.181, -14.703, -14.060, -13.286, -12.538, -11.926, -11.483,
    -16.464, -16.027, -15.601, -15.192, -14.711, -14.061, -13.281, -12.529,
    -11.918, -11.475, -16.477, -16.039, -15.612, -15.202, -14.719, -14.063,
    -13.275, -12.521, -11.909, -11.468, -16.490, -16.051, -15.623, -15.212,
    -14.726, -14.063, -13.270, -12.513, -11.901, -11.461, -16.502, -16.063,
    -15.634, -15.223, -14.734, -14.064, -13.264, -12.505, -11.892, -11.454,
    -16.515, -16.075, -15.645, -15.233, -14.741, -14.065, -13.258, -12.496,
    -11.884, -11.447, -16.527, -16.086, -15.656, -15.243, -14.748, -14.065,
    -13.252, -12.487, -11.876, -11.440, -16.540, -16.098, -15.667, -15.253,
    -14.755, -14.065, -13.246, -12.479, -11.867, -11.433, -16.553, -16.110,
    -15.678, -15.263, -14.761, -14.064, -13.240, -12.470, -11.859, -11.426,
    -16.565, -16.122, -15.689, -15.273, -14.768, -14.064, -13.233, -12.462,
    -11.851, -11.419, -16.578, -16.134, -15.700, -15.283, -14.774, -14.063,
    -13.226, -12.453, -11.843, -11.412, -16.591, -16.146, -15.712, -15.293,
    -14.781, -14.062, -13.220, -12.444, -11.835, -11.406, -16.603, -16.158,
    -15.723, -15.303, -14.787, -14.061, -13.213, -12.436, -11.827, -11.399,
    -16.616, -16.169, -15.734, -15.313, -14.793, -14.059, -13.206, -12.427,
    -11.819, -11.393, -16.629, -16.181, -15.745, -15.323, -14.799, -14.058,
    -13.199, -12.419, -11.811, -11.386, -16.641, -16.193, -15.756, -15.332,
    -14.804, -14.056, -13.192, -12.411, -11.803, -11.380, -16.654, -16.205,
    -15.767, -15.342, -14.810, -14.054, -13.185, -12.402, -11.796, -11.374,
    -16.667, -16.217, -15.778, -15.352, -14.816, -14.052, -13.178, -12.394,
    -11.789, -11.368, -16.679, -16.229, -15.789, -15.361, -14.821, -14.051,
    -13.172, -12.387, -11.782, -11.363, -16.692, -16.240, -15.800, -15.371,
    -14.826, -14.049, -13.165, -12.379, -11.775, -11.357, -16.705, -16.252,
    -15.811, -15.381, -14.831, -14.047, -13.159, -12.372, -11.768, -11.352,
    -16.717, -16.264, -15.821, -15.390, -14.837, -14.045, -13.152, -12.364,
    -11.762, -11.347, -16.730, -16.276, -15.832, -15.400, -14.842, -14.043,
    -13.146, -12.357, -11.756, -11.342, -16.743, -16.288, -15.843, -15.409,
    -14.847, -14.041, -13.140, -12.351, -11.750, -11.338, -16.755, -16.300,
    -15.854, -15.419, -14.852, -14.039, -13.134, -12.344, -11.744, -11.333,
    -16.768, -16.311, -15.865, -15.424, -14.840, -14.006, -13.094, -12.308,
    -11.716, -11.311, -16.781, -16.323, -15.875, -15.430, -14.830, -13.977,
    -13.060, -12.277, -11.691, -11.293, -16.794, -16.335, -15.886, -15.434,
    -14.812, -13.936, -13.013, -12.236, -11.657, -11.267, -16.806, -16.347,
    -15.896, -15.439, -14.800, -13.905, -12.978, -12.204, -11.632, -11.248,
    -16.819, -16.359, -15.907, -15.446, -14.796, -13.892, -12.961, -12.190,
    -11.620, -11.239, -16.832, -16.371, -15.918, -15.452, -14.790, -13.874,
    -12.940, -12.171, -11.605, -11.228, -16.844, -16.382, -15.929, -15.464,
    -14.806, -13.892, -12.957, -12.185, -11.617, -11.236, -16.857, -16.394,
    -15.940, -15.474, -14.816, -13.902, -12.966, -12.192, -11.622, -11.241,
    -16.870, -16.406, -15.950, -15.483, -14.818, -13.898, -12.960, -12.187,
    -11.618, -11.237, -16.882, -16.418, -15.961, -15.495, -14.836, -13.920,
    -12.981, -12.204, -11.631, -11.247, -16.895, -16.430, -15.972, -15.505,
    -14.843, -13.924, -12.983, -12.205, -11.632, -11.248, -16.908, -16.441,
    -15.983, -15.513, -14.846, -13.921, -12.979, -12.201, -11.629, -11.245,
    -16.920, -16.453, -15.994, -15.526, -14.867, -13.947, -13.002, -12.220,
    -11.643, -11.256, -16.933, -16.465, -16.005, -15.534, -14.868, -13.940,
    -12.993, -12.212, -11.636, -11.251, -16.946, -16.477, -16.015, -15.538,
    -14.851, -13.906, -12.957, -12.180, -11.610, -11.231, -16.959, -16.488,
    -16.026, -15.543, -14.843, -13.884, -12.931, -12.157, -11.592, -11.216,
    -16.971, -16.500, -16.035, -15.544, -14.816, -13.835, -12.880, -12.112,
    -11.556, -11.189, -16.984, -16.512, -16.045, -15.543, -14.787, -13.787,
    -12.830, -12.069, -11.521, -11.162, -16.997, -16.523, -16.055, -15.549,
    -14.785, -13.778, -12.821, -12.062, -11.516, -11.158, -17.009, -16.535,
    -16.066, -15.558, -14.794, -13.788, -12.831, -12.071, -11.524, -11.165,
    -17.022, -16.547, -16.077, -15.572, -14.817, -13.818, -12.861, -12.097,
    -11.545, -11.181, -17.034, -16.559, -16.089, -15.590, -14.855, -13.870,
    -12.912, -12.140, -11.579, -11.207, -17.047, -16.570, -16.101, -15.610,
    -14.895, -13.924, -12.963, -12.182, -11.612, -11.232, -17.060, -16.582,
    -16.112, -15.628, -14.931, -13.972, -13.007, -12.218, -11.639, -11.252,
    -17.072, -16.593, -16.123, -15.641, -14.951, -13.994, -13.027, -12.233,
    -11.650, -11.260, -17.085, -16.605, -16.134, -15.648, -14.952, -13.989,
    -13.020, -12.227, -11.645, -11.256, -17.097, -16.617, -16.144, -15.650,
    -14.933, -13.955, -12.986, -12.199, -11.623, -11.239, -17.110, -16.628,
    -16.153, -15.648, -14.901, -13.903, -12.934, -12.155, -11.589, -11.214,
    -17.122, -16.640, -16.162, -15.643, -14.860, -13.838, -12.869, -12.100,
    -11.546, -11.181, -17.135, -16.651, -16.171, -15.639, -14.824, -13.780,
    -12.810, -12.049, -11.505, -11.149, -17.147, -16.663, -16.181, -15.640,
    -14.805, -13.745, -12.774, -12.017, -11.478, -11.128, -17.160, -16.674,
    -16.191, -15.646, -14.798, -13.729, -12.755, -11.999, -11.463, -11.116,
    -17.172, -16.685, -16.202, -15.658, -14.813, -13.743, -12.766, -12.007,
    -11.469, -11.120, -17.184, -16.697, -16.213, -15.678, -14.853, -13.794,
    -12.813, -12.047, -11.501, -11.144, -17.197, -16.708, -16.225, -15.696,
    -14.887, -13.836, -12.854, -12.081, -11.528, -11.165, -17.209, -16.719,
    -16.236, -15.711, -14.915, -13.873, -12.888, -12.110, -11.551, -11.183,
    -17.221, -16.731, -16.247, -15.728, -14.947, -13.913, -12.926, -12.142,
    -11.576, -11.202, -17.233, -16.742, -16.257, -15.737, -14.953, -13.918,
    -12.931, -12.146, -11.579, -11.204, -17.246, -16.753, -16.267, -15.741,
    -14.947, -13.905, -12.918, -12.135, -11.570, -11.197, -17.258, -16.764,
    -16.278, -15.754, -14.966, -13.926, -12.936, -12.150, -11.582, -11.205,
    -17.270, -16.775, -16.288, -15.760, -14.962, -13.916, -12.925, -12.141,
    -11.574, -11.200, -17.282, -16.786, -16.297, -15.763, -14.950, -13.895,
    -12.905, -12.123, -11.560, -11.189, -17.294, -16.797, -16.308, -15.775,
    -14.964, -13.908, -12.916, -12.131, -11.566, -11.193, -17.305, -16.808,
    -16.317, -15.779, -14.955, -13.891, -12.898, -12.116, -11.554, -11.183,
    -17.317, -16.819, -16.327, -15.785, -14.952, -13.882, -12.888, -12.107,
    -11.547, -11.178, -17.329, -16.829, -16.338, -15.799, -14.976, -13.910,
    -12.912, -12.127, -11.562, -11.189, -17.340, -16.840, -16.348, -15.811,
    -14.991, -13.924, -12.925, -12.137, -11.569, -11.194, -17.352, -16.851,
    -16.359, -15.832, -15.035, -13.980, -12.977, -12.179, -11.601, -11.218,
    -17.363, -16.861, -16.370, -15.853, -15.082, -14.041, -13.032, -12.224,
    -11.636, -11.243, -17.374, -16.871, -16.381, -15.868, -15.110, -14.076,
    -13.064, -12.249, -11.655, -11.257, -17.386, -16.882, -16.391, -15.885,
    -15.144, -14.119, -13.104, -12.281, -11.679, -11.275, -17.397, -16.892,
    -16.401, -15.894, -15.152, -14.127, -13.110, -12.286, -11.683, -11.277,
    -17.407, -16.902, -16.411, -15.904, -15.161, -14.135, -13.117, -12.292,
    -11.687, -11.280, -17.418, -16.912, -16.420, -15.913, -15.171, -14.144,
    -13.125, -12.298, -11.692, -11.284, -17.429, -16.922, -16.430, -15.922,
    -15.180, -14.153, -13.132, -12.304, -11.696, -11.287, -17.439, -16.931,
    -16.439, -15.932, -15.190, -14.162, -13.141, -12.311, -11.702, -11.291,
    -17.449, -16.941, -16.449, -15.941, -15.200, -14.172, -13.150, -12.318,
    -11.707, -11.295, -17.459, -16.950, -16.458, -15.951, -15.210, -14.183,
    -13.159, -12.326, -11.713, -11.299, -17.469, -16.960, -16.467, -15.960,
    -15.221, -14.194, -13.169, -12.334, -11.720, -11.304, -17.479, -16.969,
    -16.476, -15.970, -15.231, -14.205, -13.179, -12.342, -11.726, -11.308,
    -17.488, -16.978, -16.485, -15.979, -15.242, -14.216, -13.189, -12.351,
    -11.733, -11.314, -17.497, -16.986, -16.494, -15.989, -15.253, -14.228,
    -13.200, -12.360, -11.740, -11.319, -17.506, -16.995, -16.503, -15.998,
    -15.264, -14.240, -13.211, -12.370, -11.748, -11.324, -17.514, -17.003,
    -16.512, -16.007, -15.275, -14.252, -13.223, -12.379, -11.755, -11.330,
    -17.523, -17.012, -16.520, -16.017, -15.287, -14.265, -13.235, -12.389,
    -11.763, -11.336, -17.531, -17.020, -16.529, -16.026, -15.298, -14.278,
    -13.247, -12.400, -11.772, -11.342, -17.538, -17.027, -16.537, -16.035,
    -15.310, -14.291, -13.259, -12.410, -11.780, -11.348, -17.545, -17.035,
    -16.546, -16.045, -15.321, -14.304, -13.272, -12.421, -11.789, -11.355,
    -17.552, -17.042, -16.554, -16.054, -15.333, -14.317, -13.285, -12.432,
    -11.798, -11.362, -17.558, -17.049, -16.562, -16.063, -15.345, -14.331,
    -13.298, -12.444, -11.807, -11.369, -17.564, -17.056, -16.570, -16.072,
    -15.356, -14.345, -13.311, -12.455, -11.816, -11.376, -17.570, -17.062,
    -16.577, -16.081, -15.368, -14.359, -13.325, -12.467, -11.826, -11.383,
    -17.575, -17.068, -16.585, -16.090, -15.380, -14.373, -13.339, -12.479,
    -11.836, -11.391, -17.579, -17.074, -16.592, -16.099, -15.392, -14.387,
    -13.353, -12.491, -11.846, -11.398, -17.583, -17.079, -16.599, -16.108,
    -15.404, -14.402, -13.367, -12.504, -11.856, -11.406, -17.586, -17.084,
    -16.606, -16.117, -15.416, -14.416, -13.381, -12.516, -11.866, -11.414,
    -17.589, -17.089, -16.613, -16.126, -15.428, -14.431, -13.396, -12.529,
    -11.877, -11.422, -17.591, -17.094, -16.620, -16.134, -15.440, -14.445,
    -13.410, -12.542, -11.887, -11.430, -17.592, -17.098, -16.626, -16.143,
    -15.452, -14.460, -13.425, -12.555, -11.898, -11.439, -17.593, -17.101,
    -16.633, -16.151, -15.464, -14.475, -13.440, -12.568, -11.909, -11.447,
    -17.593, -17.104, -16.639, -16.160, -15.475, -14.490, -13.455, -12.582,
    -11.920, -11.456, -17.592, -17.107, -16.644, -16.168, -15.487, -14.504,
    -13.470, -12.595, -11.931, -11.465, -17.591, -17.109, -16.650, -16.176,
    -15.499, -14.519, -13.485, -12.609, -11.943, -11.474, -17.589, -17.111,
    -16.655, -16.184, -15.511, -14.534, -13.500, -12.623, -11.954, -11.483,
    -17.586, -17.113, -16.661, -16.192, -15.523, -14.549, -13.515, -12.636,
    -11.966, -11.492, -17.582, -17.114, -16.665, -16.200, -15.534, -14.564,
    -13.531, -12.650, -11.978, -11.501, -17.577, -17.114, -16.670, -16.208,
    -15.546, -14.579, -13.546, -12.664, -11.989, -11.510, -17.571, -17.114,
    -16.675, -16.215, -15.557, -14.594, -13.562, -12.678, -12.001, -11.520,
    -17.564, -17.113, -16.679, -16.223, -15.569, -14.609, -13.577, -12.693,
    -12.013, -11.529, -17.557, -17.112, -16.683, -16.230, -15.580, -14.624,
    -13.593, -12.707, -12.025, -11.539, -17.548, -17.111, -16.686, -16.237,
    -15.592, -14.639, -13.609, -12.721, -12.037, -11.548, -17.539, -17.108,
    -16.690, -16.244, -15.603, -14.654, -13.624, -12.736, -12.050, -11.558,
    -17.528, -17.106, -16.693, -16.251, -15.614, -14.669, -13.640, -12.750,
    -12.062, -11.568, -17.517, -17.102, -16.696, -16.258, -15.625, -14.684,
    -13.656, -12.765, -12.074, -11.578, -17.504, -17.099, -16.698, -16.265,
    -15.636, -14.699, -13.672, -12.779, -12.087, -11.588, -17.491, -17.094,
    -16.701, -16.272, -15.647, -14.714, -13.687, -12.794, -12.099, -11.598,
    -17.477, -17.089, -16.703, -16.278, -15.658, -14.729, -13.703, -12.808,
    -12.112, -11.608, -17.461, -17.084, -16.704, -16.284, -15.669, -14.744,
    -13.719, -12.823, -12.124, -11.618, -17.445, -17.078, -16.706, -16.291,
    -15.680, -14.759, -13.735, -12.838, -12.137, -11.628, -17.428, -17.071,
    -16.707, -16.297, -15.691, -14.773, -13.751, -12.852, -12.150, -11.638,
    -17.410, -17.064, -16.708, -16.303, -15.701, -14.788, -13.766, -12.867,
    -12.162, -11.648, -17.391, -17.056, -16.708, -16.308, -15.712, -14.803,
    -13.782, -12.882, -12.175, -11.659, -17.371, -17.048, -16.709, -16.314,
    -15.722, -14.817, -13.798, -12.897, -12.188, -11.669, -17.351, -17.039,
    -16.709, -16.320, -15.732, -14.832, -13.814, -12.911, -12.201, -11.679,
    -17.329, -17.030, -16.709, -16.325, -15.743, -14.846, -13.830, -12.926,
    -12.214, -11.690, -17.307, -17.021, -16.708, -16.330, -15.753, -14.861,
    -13.845, -12.941, -12.226, -11.700, -17.284, -17.011, -16.708, -16.335,
    -15.763, -14.875, -13.861, -12.956, -12.239, -11.711, -17.261, -17.000,
    -16.707, -16.340, -15.773, -14.889, -13.877, -12.971, -12.252, -11.721,
    -17.237, -16.989, -16.705, -16.345, -15.783, -14.903, -13.892, -12.986,
    -12.265, -11.732, -17.212, -16.978, -16.704, -16.350, -15.793, -14.918,
    -13.908, -13.000, -12.278, -11.742, -17.187, -16.966, -16.703, -16.355,
    -15.802, -14.932, -13.924, -13.015, -12.291, -11.753, -17.162, -16.955,
    -16.701, -16.360, -15.812, -14.946, -13.939, -13.030, -12.304, -11.763,
    -17.136, -16.942, -16.699, -16.364, -15.822, -14.960, -13.955, -13.045,
    -12.317, -11.774, -17.110, -16.930, -16.697, -16.369, -15.831, -14.974,
    -13.970, -13.060, -12.330, -11.785, -17.084, -16.918, -16.695, -16.373,
    -15.841, -14.987, -13.986, -13.074, -12.343, -11.795, -17.058, -16.905,
    -16.692, -16.377, -15.850, -15.001, -14.001, -13.089, -12.356, -11.806,
    -17.031, -16.892, -16.690, -16.382, -15.859, -15.015, -14.016, -13.104,
    -12.368, -11.816, -17.005, -16.879, -16.688, -16.386, -15.869, -15.029,
    -14.032, -13.118, -12.381, -11.827, -16.979, -16.866, -16.685, -16.390,
    -15.878, -15.042, -14.047, -13.133, -12.394, -11.837, -16.953, -16.854,
    -16.682, -16.394, -15.887, -15.056, -14.062, -13.148, -12.407, -11.848,
    -16.927, -16.841, -16.680, -16.398, -15.896, -15.069, -14.077, -13.162,
    -12.420, -11.858, -16.902, -16.828, -16.677, -16.402, -15.905, -15.082,
    -14.093, -13.177, -12.433, -11.869, -16.877, -16.816, -16.675, -16.406,
    -15.914, -15.096, -14.108, -13.191, -12.446, -11.880, -16.852, -16.804,
    -16.673, -16.411, -15.923, -15.109, -14.123, -13.206, -12.458, -11.890,
    -16.829, -16.792, -16.670, -16.415, -15.932, -15.122, -14.138, -13.220,
    -12.471, -11.901, -16.805, -16.781, -16.668, -16.419, -15.941, -15.135,
    -14.153, -13.235, -12.484, -11.911, -16.506, -16.584, -16.565, -16.378,
    -15.934, -15.145, -14.167, -13.249, -12.496, -11.921, -16.391, -16.507,
    -16.524, -16.364, -15.937, -15.156, -14.181, -13.263, -12.509, -11.932,
    -16.309, -16.452, -16.495, -16.356, -15.941, -15.168, -14.196, -13.278,
    -12.522, -11.942, -16.278, -16.435, -16.490, -16.359, -15.950, -15.181,
    -14.211, -13.292, -12.534, -11.953, -16.248, -16.420, -16.486, -16.362,
    -15.958, -15.193, -14.225, -13.306, -12.547, -11.963, -16.218, -16.404,
    -16.481, -16.365, -15.966, -15.206, -14.240, -13.320, -12.559, -11.973,
    -16.188, -16.388, -16.476, -16.368, -15.975, -15.219, -14.254, -13.335,
    -12.572, -11.983, -16.230, -16.423, -16.501, -16.384, -15.988, -15.232,
    -14.269, -13.349, -12.584, -11.994, -16.197, -16.402, -16.492, -16.385,
    -15.995, -15.245, -14.283, -13.363, -12.597, -12.004, -16.142, -16.364,
    -16.472, -16.380, -16.000, -15.256, -14.298, -13.377, -12.609, -12.014,
    -16.072, -16.313, -16.443, -16.370, -16.003, -15.267, -14.312, -13.390,
    -12.621, -12.024, -15.915, -16.195, -16.371, -16.339, -15.998, -15.276,
    -14.325, -13.404, -12.633, -12.034, -15.898, -16.181, -16.363, -16.338,
    -16.005, -15.288, -14.339, -13.418, -12.646, -12.044, -15.863, -16.159,
    -16.353, -16.338, -16.011, -15.300, -14.353, -13.432, -12.658, -12.054,
    -15.805, -16.119, -16.331, -16.331, -16.016, -15.311, -14.367, -13.445,
    -12.670, -12.064, -15.835, -16.150, -16.357, -16.350, -16.030, -15.324,
    -14.381, -13.459, -12.682, -12.074, -15.783, -16.120, -16.345, -16.350,
    -16.037, -15.336, -14.395, -13.473, -12.694, -12.084, -15.743, -16.095,
    -16.334, -16.349, -16.043, -15.347, -14.409, -13.486, -12.706, -12.093,
    -15.825, -16.165, -16.384, -16.379, -16.061, -15.362, -14.423, -13.500,
    -12.718, -12.103, -15.861, -16.199, -16.409, -16.397, -16.074, -15.375,
    -14.437, -13.513, -12.729, -12.113, -15.842, -16.190, -16.408, -16.401,
    -16.082, -15.387, -14.451, -13.527, -12.741, -12.122, -15.927, -16.260,
    -16.456, -16.429, -16.098, -15.400, -14.465, -13.540, -12.753, -12.132,
    -15.869, -16.217, -16.431, -16.420, -16.102, -15.411, -14.478, -13.553,
    -12.764, -12.141, -15.747, -16.120, -16.368, -16.391, -16.097, -15.419,
    -14.491, -13.566, -12.776, -12.151, -15.743, -16.103, -16.352, -16.384,
    -16.099, -15.429, -14.504, -13.579, -12.787, -12.160, -15.628, -16.003,
    -16.281, -16.348, -16.091, -15.436, -14.516, -13.592, -12.798, -12.169,
    -15.466, -15.870, -16.189, -16.300, -16.078, -15.441, -14.528, -13.604,
    -12.810, -12.178, -15.355, -15.787, -16.137, -16.277, -16.075, -15.450,
    -14.541, -13.617, -12.821, -12.187, -15.287, -15.746, -16.117, -16.273,
    -16.080, -15.460, -14.554, -13.630, -12.832, -12.196, -15.277, -15.754,
    -16.134, -16.290, -16.094, -15.474, -14.568, -13.643, -12.843, -12.205,
    -15.368, -15.847, -16.211, -16.340, -16.121, -15.490, -14.582, -13.656,
    -12.854, -12.214, -15.629, -16.064, -16.362, -16.425, -16.160, -15.509,
    -14.596, -13.668, -12.865, -12.223, -15.924, -16.288, -16.505, -16.498,
    -16.192, -15.527, -14.610, -13.681, -12.876, -12.232, -16.145, -16.440,
    -16.593, -16.542, -16.214, -15.541, -14.624, -13.694, -12.887, -12.241,
    -16.092, -16.406, -16.577, -16.539, -16.219, -15.551, -14.636, -13.706,
    -12.897, -12.249, -15.795, -16.198, -16.457, -16.487, -16.206, -15.557,
    -14.648, -13.718, -12.908, -12.258, -15.474, -15.950, -16.299, -16.411,
    -16.184, -15.560, -14.659, -13.730, -12.918, -12.266, -15.211, -15.731,
    -16.144, -16.328, -16.157, -15.561, -14.669, -13.741, -12.928, -12.274,
    -15.102, -15.626, -16.061, -16.280, -16.141, -15.565, -14.680, -13.753,
    -12.939, -12.283, -15.094, -15.607, -16.041, -16.267, -16.140, -15.573,
    -14.692, -13.765, -12.949, -12.291, -15.133, -15.632, -16.054, -16.275,
    -16.147, -15.583, -14.704, -13.776, -12.959, -12.299, -15.237, -15.714,
    -16.114, -16.312, -16.168, -15.597, -14.717, -13.788, -12.969, -12.307,
    -15.358, -15.824, -16.199, -16.366, -16.197, -15.614, -14.730, -13.800,
    -12.979, -12.315, -15.408, -15.880, -16.249, -16.401, -16.218, -15.628,
    -14.743, -13.812, -12.989, -12.323, -15.408, -15.896, -16.270, -16.420,
    -16.232, -15.641, -14.756, -13.823, -12.998, -12.330, -15.485, -15.966,
    -16.324, -16.455, -16.252, -15.655, -14.768, -13.835, -13.008, -12.338,
    -15.478, -15.960, -16.322, -16.457, -16.258, -15.665, -14.780, -13.846,
    -13.017, -12.346, -15.348, -15.855, -16.251, -16.422, -16.250, -15.670,
    -14.791, -13.857, -13.027, -12.353, -15.356, -15.859, -16.254, -16.426,
    -16.257, -15.680, -14.803, -13.868, -13.036, -12.361, -15.255, -15.776,
    -16.196, -16.396, -16.250, -15.686, -14.813, -13.879, -13.045, -12.368,
    -15.135, -15.676, -16.124, -16.357, -16.238, -15.691, -14.824, -13.889,
    -13.054, -12.375, -15.185, -15.713, -16.149, -16.374, -16.250, -15.702,
    -14.835, -13.900, -13.063, -12.382, -15.170, -15.695, -16.134, -16.366,
    -16.252, -15.710, -14.846, -13.911, -13.072, -12.390, -15.122, -15.651,
    -16.101, -16.348, -16.249, -15.717, -14.857, -13.921, -13.081, -12.397,
    -15.151, -15.674, -16.118, -16.361, -16.259, -15.728, -14.868, -13.931,
    -13.090, -12.403, -15.130, -15.657, -16.107, -16.357, -16.262, -15.736,
    -14.879, -13.942, -13.098, -12.410, -15.159, -15.686, -16.133, -16.377,
    -16.276, -15.748, -14.890, -13.952, -13.107, -12.417, -15.276, -15.786,
    -16.209, -16.425, -16.302, -15.763, -14.902, -13.962, -13.115, -12.424,
    -15.304, -15.811, -16.229, -16.440, -16.314, -15.774, -14.913, -13.972,
    -13.123, -12.430, -15.328, -15.833, -16.247, -16.454, -16.325, -15.785,
    -14.924, -13.982, -13.132, -12.437, -15.321, -15.826, -16.243, -16.455,
    -16.330, -15.794, -14.935, -13.992, -13.140, -12.443, -15.244, -15.764,
    -16.200, -16.433, -16.325, -15.800, -14.945, -14.002, -13.148, -12.450,
    -15.240, -15.758, -16.195, -16.432, -16.330, -15.808, -14.955, -14.011,
    -13.156, -12.456, -15.237, -15.753, -16.192, -16.432, -16.334, -15.817,
    -14.965, -14.021, -13.163, -12.462, -15.236, -15.749, -16.189, -16.432,
    -16.339, -15.825, -14.975, -14.030, -13.171, -12.469, -15.234, -15.745,
    -16.186, -16.433, -16.344, -15.834, -14.986, -14.040, -13.179, -12.475,
    -15.314, -15.810, -16.234, -16.464, -16.362, -15.846, -14.996, -14.049,
    -13.186, -12.481, -15.412, -15.889, -16.291, -16.500, -16.382, -15.859,
    -15.007, -14.058, -13.194, -12.487, -15.709, -16.117, -16.448, -16.590,
    -16.425, -15.879, -15.020, -14.068, -13.202, -12.493, -15.705, -16.113,
    -16.447, -16.593, -16.431, -15.887, -15.030, -14.077, -13.209, -12.499,
    -15.702, -16.110, -16.447, -16.596, -16.437, -15.896, -15.039, -14.086,
    -13.216, -12.505, -15.700, -16.109, -16.447, -16.599, -16.443, -15.905,
    -15.049, -14.094, -13.223, -12.511, -15.699, -16.108, -16.448, -16.602,
    -16.449, -15.914, -15.059, -14.103, -13.231, -12.516, -15.700, -16.108,
    -16.449, -16.606, -16.456, -15.922, -15.068, -14.112, -13.238, -12.522,
    -15.702, -16.109, -16.452, -16.610, -16.462, -15.931, -15.078, -14.120,
    -13.245, -12.528, -15.704, -16.111, -16.454, -16.615, -16.469, -15.940,
    -15.088, -14.129, -13.252, -12.534, -15.708, -16.114, -16.457, -16.620,
    -16.476, -15.949, -15.097, -14.137, -13.259, -12.539, -15.712, -16.117,
    -16.461, -16.625, -16.483, -15.957, -15.107, -14.146, -13.266, -12.545,
    -15.718, -16.121, -16.465, -16.630, -16.490, -15.966, -15.116, -14.154,
    -13.272, -12.551, -15.724, -16.126, -16.470, -16.636, -16.497, -15.975,
    -15.125, -14.163, -13.279, -12.556, -15.732, -16.131, -16.475, -16.642,
    -16.504, -15.984, -15.135, -14.171, -13.286, -12.562, -15.739, -16.137,
    -16.481, -16.648, -16.511, -15.993, -15.144, -14.179, -13.293, -12.568,
    -15.748, -16.144, -16.487, -16.654, -16.519, -16.001, -15.153, -14.187,
    -13.300, -12.573, -15.758, -16.151, -16.494, -16.661, -16.526, -16.009,
    -15.160, -14.191, -13.302, -12.574, -15.768, -16.159, -16.501, -16.668,
    -16.534, -16.017, -15.167, -14.196, -13.304, -12.576, -15.778, -16.167,
    -16.508, -16.675, -16.542, -16.025, -15.171, -14.196, -13.303, -12.575,
    -15.790, -16.176, -16.516, -16.683, -16.550, -16.032, -15.175, -14.196,
    -13.301, -12.573, -15.802, -16.185, -16.524, -16.690, -16.558, -16.040,
    -15.183, -14.202, -13.305, -12.576, -15.814, -16.195, -16.532, -16.698,
    -16.566, -16.048, -15.189, -14.206, -13.308, -12.579, -15.827, -16.205,
    -16.541, -16.706, -16.574, -16.057, -15.199, -14.216, -13.316, -12.586,
    -15.841, -16.216, -16.550, -16.715, -16.582, -16.067, -15.210, -14.227,
    -13.327, -12.595, -15.855, -16.227, -16.560, -16.723, -16.591, -16.076,
    -15.218, -14.234, -13.333, -12.600, -15.869, -16.238, -16.569, -16.732,
    -16.599, -16.085, -15.230, -14.246, -13.344, -12.609, -15.884, -16.250,
    -16.579, -16.741, -16.607, -16.095, -15.241, -14.257, -13.354, -12.618,
    -15.899, -16.262, -16.590, -16.750, -16.616, -16.103, -15.249, -14.264,
    -13.359, -12.622, -15.915, -16.275, -16.600, -16.759, -16.625, -16.113,
    -15.260, -14.276, -13.370, -12.632, -15.931, -16.287, -16.611, -16.769,
    -16.634, -16.122, -15.271, -14.286, -13.379, -12.639, -15.947, -16.300,
    -16.622, -16.778, -16.642, -16.130, -15.276, -14.289, -13.380, -12.640,
    -15.964, -16.314, -16.633, -16.788, -16.651, -16.138, -15.281, -14.290,
    -13.380, -12.639, -15.981, -16.327, -16.645, -16.798, -16.660, -16.145,
    -15.284, -14.289, -13.378, -12.637, -15.998, -16.341, -16.657, -16.808,
    -16.669, -16.151, -15.284, -14.284, -13.371, -12.631, -16.016, -16.355,
    -16.668, -16.818, -16.679, -16.159, -15.289, -14.286, -13.371, -12.631,
    -16.034, -16.369, -16.681, -16.829, -16.688, -16.168, -15.299, -14.295,
    -13.380, -12.639, -16.052, -16.384, -16.693, -16.839, -16.697, -16.178,
    -15.311, -14.308, -13.392, -12.649, -16.070, -16.398, -16.705, -16.849,
    -16.706, -16.189, -15.327, -14.327, -13.410, -12.665, -16.088, -16.413,
    -16.718, -16.860, -16.716, -16.201, -15.344, -14.348, -13.430, -12.682,
    -16.107, -16.428, -16.731, -16.871, -16.725, -16.212, -15.359, -14.366,
    -13.447, -12.696, -16.126, -16.444, -16.743, -16.882, -16.735, -16.223,
    -15.373, -14.380, -13.461, -12.708, -16.145, -16.459, -16.756, -16.893,
    -16.744, -16.232, -15.382, -14.389, -13.468, -12.714, -16.164, -16.475,
    -16.770, -16.904, -16.754, -16.240, -15.388, -14.392, -13.470, -12.715,
    -16.183, -16.490, -16.783, -16.915, -16.763, -16.247, -15.389, -14.389,
    -13.465, -12.711, -16.203, -16.506, -16.796, -16.926, -16.773, -16.252,
    -15.388, -14.382, -13.457, -12.704, -16.223, -16.522, -16.810, -16.938,
    -16.783, -16.258, -15.386, -14.373, -13.447, -12.695, -16.242, -16.538,
    -16.823, -16.949, -16.792, -16.264, -15.386, -14.368, -13.440, -12.690,
    -16.262, -16.554, -16.837, -16.960, -16.802, -16.272, -15.390, -14.368,
    -13.439, -12.688, -16.282, -16.571, -16.851, -16.972, -16.812, -16.280,
    -15.397, -14.373, -13.442, -12.691, -16.302, -16.587, -16.865, -16.984,
    -16.822, -16.292, -15.412, -14.391, -13.459, -12.705, -16.322, -16.603,
    -16.879, -16.995, -16.832, -16.304, -15.429, -14.411, -13.478, -12.721,
    -16.343, -16.620, -16.893, -17.007, -16.842, -16.314, -15.443, -14.426,
    -13.492, -12.733, -16.363, -16.637, -16.907, -17.019, -16.852, -16.326,
    -15.459, -14.445, -13.510, -12.749, -16.383, -16.653, -16.921, -17.031,
    -16.862, -16.336, -15.471, -14.458, -13.522, -12.758, -16.404, -16.670,
    -16.936, -17.043, -16.872, -16.344, -15.477, -14.462, -13.525, -12.760,
    -16.424, -16.687, -16.950, -17.055, -16.882, -16.353, -15.486, -14.471,
    -13.533, -12.767, -16.445, -16.704, -16.964, -17.067, -16.892, -16.363,
    -15.495, -14.479, -13.540, -12.773, -16.465, -16.721, -16.979, -17.079,
    -16.902, -16.370, -15.500, -14.481, -13.540, -12.773, -16.486, -16.738,
    -16.993, -17.091, -16.912, -16.380, -15.509, -14.489, -13.547, -12.779,
    -16.506, -16.755, -17.008, -17.103, -16.922, -16.388, -15.515, -14.493,
    -13.550, -12.781, -16.527, -16.772, -17.022, -17.115, -16.932, -16.396,
    -15.519, -14.494, -13.550, -12.781, -16.547, -16.789, -17.037, -17.127,
    -16.943, -16.406, -15.530, -14.505, -13.560, -12.789, -16.568, -16.806,
    -17.052, -17.140, -16.953, -16.416, -15.542, -14.517, -13.571, -12.798,
    -16.589, -16.823, -17.066, -17.152, -16.963, -16.427, -15.556, -14.534,
    -13.586, -12.810, -16.609, -16.840, -17.081, -17.164, -16.973, -16.439,
    -15.574, -14.555, -13.606, -12.827, -16.630, -16.858, -17.096, -17.177,
    -16.983, -16.450, -15.589, -14.572, -13.622, -12.840, -16.650, -16.875,
    -17.111, -17.189, -16.994, -16.462, -15.606, -14.592, -13.641, -12.855,
    -16.671, -16.892, -17.126, -17.201, -17.004, -16.471, -15.615, -14.600,
    -13.648, -12.861, -16.691, -16.909, -17.140, -17.214, -17.014, -16.481,
    -15.624, -14.609, -13.656, -12.867, -16.712, -16.927, -17.155, -17.226,
    -17.025, -16.490, -15.634, -14.618, -13.663, -12.874, -16.732, -16.944,
    -17.170, -17.238, -17.035, -16.499, -15.643, -14.627, -13.671, -12.880,
    -16.753, -16.961, -17.185, -17.251, -17.045, -16.509, -15.653, -14.636,
    -13.679, -12.887, -16.773, -16.978, -17.200, -17.263, -17.056, -16.518,
    -15.662, -14.645, -13.687, -12.893, -16.794, -16.995, -17.215, -17.276,
    -17.066, -16.528, -15.672, -14.655, -13.695, -12.900, -16.814, -17.013,
    -17.230, -17.288, -17.076, -16.537, -15.681, -14.664, -13.704, -12.906,
    -16.834, -17.030, -17.245, -17.301, -17.087, -16.546, -15.691, -14.674,
    -13.712, -12.913, -16.854, -17.047, -17.259, -17.313, -17.097, -16.556,
    -15.701, -14.683, -13.721, -12.920, -16.875, -17.064, -17.274, -17.326,
    -17.107, -16.565, -15.710, -14.693, -13.729, -12.927, -16.895, -17.081,
    -17.289, -17.338, -17.118, -16.575, -15.720, -14.703, -13.738, -12.934,
    -16.915, -17.098, -17.304, -17.351, -17.128, -16.584, -15.730, -14.712,
    -13.747, -12.942, -16.935, -17.116, -17.319, -17.363, -17.138, -16.594,
    -15.740, -14.722, -13.756, -12.949, -16.955, -17.133, -17.334, -17.376,
    -17.149, -16.603, -15.750, -14.732, -13.765, -12.956, -16.974, -17.150,
    -17.349, -17.388, -17.159, -16.613, -15.760, -14.742, -13.774, -12.963,
    -16.994, -17.167, -17.364, -17.401, -17.170, -16.622, -15.770, -14.753,
    -13.783, -12.971, -17.014, -17.184, -17.379, -17.413, -17.180, -16.632,
    -15.780, -14.763, -13.793, -12.979, -17.034, -17.201, -17.393, -17.426,
    -17.190, -16.641, -15.790, -14.773, -13.802, -12.986, -17.053, -17.218,
    -17.408, -17.438, -17.201, -16.651, -15.800, -14.783, -13.811, -12.994,
    -17.073, -17.234, -17.423, -17.451, -17.211, -16.660, -15.810, -14.794,
    -13.821, -13.001, -17.092, -17.251, -17.438, -17.464, -17.221, -16.670,
    -15.820, -14.804, -13.830, -13.009, -17.111, -17.268, -17.453, -17.476,
    -17.232, -16.679, -15.830, -14.815, -13.840, -13.017, -17.131, -17.285,
    -17.468, -17.489, -17.242, -16.689, -15.840, -14.825, -13.850, -13.025,
    -17.150, -17.302, -17.482, -17.501, -17.253, -16.698, -15.850, -14.836,
    -13.859, -13.033, -17.169, -17.318, -17.497, -17.514, -17.263, -16.708,
    -15.860, -14.846, -13.869, -13.041, -17.188, -17.335, -17.512, -17.526,
    -17.273, -16.717, -15.871, -14.857, -13.879, -13.049, -17.207, -17.352,
    -17.526, -17.538, -17.284, -16.727, -15.881, -14.868, -13.889, -13.057,
    -17.226, -17.368, -17.541, -17.551, -17.294, -16.736, -15.891, -14.878,
    -13.899, -13.065, -17.244, -17.385, -17.556, -17.563, -17.304, -16.746,
    -15.901, -14.889, -13.909, -13.073, -17.263, -17.401, -17.571, -17.576,
    -17.315, -16.755, -15.911, -14.900, -13.919, -13.081, -17.281, -17.418,
    -17.585, -17.588, -17.325, -16.765, -15.921, -14.910, -13.929, -13.089,
    -17.300, -17.434, -17.600, -17.601, -17.335, -16.774, -15.931, -14.921,
    -13.939, -13.098, -17.318, -17.451, -17.614, -17.613, -17.346, -16.784,
    -15.941, -14.932, -13.949, -13.106, -17.336, -17.467, -17.629, -17.626,
    -17.356, -16.793, -15.952, -14.943, -13.959, -13.114, -17.355, -17.483,
    -17.643, -17.638, -17.366, -16.802, -15.962, -14.953, -13.969, -13.122,
    -17.373, -17.500, -17.658, -17.650, -17.377, -16.812, -15.972, -14.964,
    -13.979, -13.130, -17.391, -17.516, -17.672, -17.663, -17.387, -16.821,
    -15.982, -14.975, -13.989, -13.139, -17.408, -17.532, -17.687, -17.675,
    -17.397, -16.831, -15.992, -14.986, -13.999, -13.147, -17.426, -17.548,
    -17.701, -17.687, -17.408, -16.840, -16.002, -14.996, -14.009, -13.155,
    -17.444, -17.564, -17.716, -17.700, -17.418, -16.849, -16.012, -15.007,
    -14.019, -13.163, -17.461, -17.580, -17.730, -17.712, -17.428, -16.859,
    -16.022, -15.018, -14.030, -13.172, -17.479, -17.596, -17.744, -17.724,
    -17.438, -16.868, -16.032, -15.029, -14.040, -13.180, -17.496, -17.612,
    -17.759, -17.737, -17.449, -16.877, -16.042, -15.039, -14.050, -13.188,
    -17.514, -17.628, -17.773, -17.749, -17.459, -16.887, -16.052, -15.050,
    -14.060, -13.196, -17.531, -17.644, -17.787, -17.761, -17.469, -16.896,
    -16.062, -15.061, -14.070, -13.205, -17.548, -17.659, -17.802, -17.773,
    -17.479, -16.905, -16.072, -15.071, -14.080, -13.213, -17.565, -17.675,
    -17.816, -17.786, -17.490, -16.915, -16.082, -15.082, -14.090, -13.221,
    -17.582, -17.691, -17.830, -17.798, -17.500, -16.924, -16.092, -15.093,
    -14.100, -13.229, -17.598, -17.706, -17.844, -17.810, -17.510, -16.933,
    -16.101, -15.103, -14.110, -13.237, -17.615, -17.722, -17.858, -17.822,
    -17.520, -16.942, -16.111, -15.114, -14.120, -13.245, -17.632, -17.737,
    -17.872, -17.834, -17.530, -16.952, -16.121, -15.125, -14.131, -13.253,
    -17.648, -17.753, -17.886, -17.846, -17.541, -16.961, -16.131, -15.135,
    -14.141, -13.261, -17.664, -17.768, -17.900, -17.859, -17.551, -16.970,
    -16.141, -15.146, -14.150, -13.269, -17.681, -17.783, -17.914, -17.871,
    -17.561, -16.979, -16.150, -15.156, -14.160, -13.277, -17.697, -17.799,
    -17.928, -17.883, -17.571, -16.989, -16.160, -15.167, -14.170, -13.285,
    -17.713, -17.814, -17.942, -17.895, -17.581, -16.998, -16.170, -15.177,
    -14.180, -13.293, -17.729, -17.829, -17.956, -17.907, -17.591, -17.007,
    -16.180, -15.188, -14.190, -13.301, -17.745, -17.844, -17.970, -17.919,
    -17.602, -17.016, -16.189, -15.198, -14.200, -13.309, -17.760, -17.859,
    -17.984, -17.931, -17.612, -17.025, -16.199, -15.208, -14.210, -13.317,
    -17.776, -17.874, -17.998, -17.943, -17.622, -17.034, -16.209, -15.219,
    -14.219, -13.324, -17.792, -17.889, -18.011, -17.955, -17.632, -17.043,
    -16.218, -15.229, -14.229, -13.332, -17.807, -17.904, -18.025, -17.967,
    -17.642, -17.053, -16.228, -15.239, -14.239, -13.339, -17.822, -17.919,
    -18.039, -17.979, -17.652, -17.062, -16.237, -15.250, -14.248, -13.347,
    -17.838, -17.934, -18.052, -17.991, -17.662, -17.071, -16.247, -15.260,
    -14.258, -13.354, -17.853, -17.948, -18.066, -18.003, -17.672, -17.080,
    -16.256, -15.270, -14.267, -13.362, -17.868, -17.963, -18.080, -18.015,
    -17.682, -17.089, -16.266, -15.280, -14.277, -13.369, -17.883, -17.978,
    -18.093, -18.026, -17.692, -17.098, -16.275, -15.290, -14.286, -13.376,
    -17.898, -17.992, -18.107, -18.038, -17.702, -17.107, -16.285, -15.300,
    -14.296, -13.384, -17.913, -18.007, -18.120, -18.050, -17.712, -17.116,
    -16.294, -15.310, -14.305, -13.391, -17.927, -18.021, -18.134, -18.062,
    -17.722, -17.125, -16.304, -15.320, -14.314, -13.398, -17.942, -18.036,
    -18.147, -18.074, -17.732, -17.134, -16.313, -15.330, -14.323, -13.405,
    -17.957, -18.050, -18.161, -18.085, -17.742, -17.143, -16.322, -15.340,
    -14.333, -13.412, -17.971, -18.064, -18.174, -18.097, -17.752, -17.152,
    -16.332, -15.350, -14.342, -13.418, -17.985, -18.079, -18.187, -18.109,
    -17.762, -17.161, -16.341, -15.360, -14.351, -13.425, -18.000, -18.093,
    -18.201, -18.121, -17.772, -17.170, -16.350, -15.370, -14.360, -13.432,
    -18.014, -18.107, -18.214, -18.132, -17.782, -17.179, -16.359, -15.379,
    -14.368, -13.438, -18.028, -18.121, -18.227, -18.144, -17.792, -17.188,
    -16.369, -15.389, -14.377, -13.445, -18.042, -18.135, -18.240, -18.156,
    -17.802, -17.196, -16.378, -15.399, -14.386, -13.451, -18.056, -18.149,
    -18.254, -18.167, -17.812, -17.205, -16.387, -15.408, -14.395, -13.458,
    -18.070, -18.163, -18.267, -18.179, -17.822, -17.214, -16.396, -15.418,
    -14.403, -13.464, -18.083, -18.177, -18.280, -18.191, -17.831, -17.223,
    -16.405, -15.427, -14.412, -13.470, -18.097, -18.191, -18.293, -18.202,
    -17.841, -17.232, -16.414, -15.437, -14.420, -13.476, -18.111, -18.205,
    -18.306, -18.214, -17.851, -17.241, -16.423, -15.446, -14.429, -13.482,
    -18.124, -18.219, -18.319, -18.225, -17.861, -17.249, -16.432, -15.455,
    -14.437, -13.488, -18.137, -18.232, -18.332, -18.237, -17.871, -17.258,
    -16.441, -15.465, -14.445, -13.494, -18.151, -18.246, -18.345, -18.248,
    -17.881, -17.267, -16.450, -15.474, -14.453, -13.499, -18.164, -18.260,
    -18.358, -18.260, -17.890, -17.276, -16.459, -15.483, -14.461, -13.505,
    -18.177, -18.273, -18.371, -18.271, -17.900, -17.284, -16.468, -15.492,
    -14.469, -13.510, -18.190, -18.287, -18.384, -18.283, -17.910, -17.293,
    -16.477, -15.501, -14.477, -13.516, -18.203, -18.300, -18.397, -18.294,
    -17.920, -17.302, -16.486, -15.510, -14.485, -13.521, -18.216, -18.314,
    -18.409, -18.306, -17.929, -17.311, -16.495, -15.519, -14.492, -13.526,
    -18.229, -18.327, -18.422, -18.317, -17.939, -17.319, -16.504, -15.528,
    -14.500, -13.531, -18.242, -18.340, -18.435, -18.328, -17.949, -17.328,
    -16.513, -15.537, -14.508, -13.536, -18.255, -18.354, -18.448, -18.340,
    -17.959, -17.337, -16.522, -15.546, -14.515, -13.541, -18.268, -18.367,
    -18.460, -18.351, -17.968, -17.345, -16.530, -15.555, -14.522, -13.546,
    -18.280, -18.380, -18.473, -18.362, -17.978, -17.354, -16.539, -15.564,
    -14.530, -13.550, -18.293, -18.393, -18.486, -18.374, -17.988, -17.363,
    -16.548, -15.572, -14.537, -13.555, -18.305, -18.406, -18.498, -18.385,
    -17.997, -17.371, -16.557, -15.581, -14.544, -13.559, -18.318, -18.419,
    -18.511, -18.396, -18.007, -17.380, -16.565, -15.589, -14.551, -13.563,
    -18.330, -18.432, -18.523, -18.407, -18.016, -17.388, -16.574, -15.598,
    -14.557, -13.568, -18.342, -18.445, -18.536, -18.418, -18.026, -17.397,
    -16.583, -15.606, -14.564, -13.572, -18.354, -18.458, -18.548, -18.430,
    -18.036, -17.405, -16.591, -15.615, -14.571, -13.576, -18.367, -18.471,
    -18.561, -18.441, -18.045, -17.414, -16.600, -15.623, -14.577, -13.579,
    -18.379, -18.484, -18.573, -18.452, -18.055, -17.422, -16.609, -15.631,
    -14.584, -13.583, -18.391, -18.497, -18.585, -18.463, -18.064, -17.431,
    -16.617, -15.639, -14.590, -13.587, -18.403, -18.510, -18.598, -18.474,
    -18.074, -17.439, -16.626, -15.647, -14.596, -13.590, -18.415, -18.522,
    -18.610, -18.485, -18.083, -17.448, -16.634, -15.655, -14.603, -13.594,
    -18.427, -18.535, -18.622, -18.496, -18.093, -17.456, -16.643, -15.663,
    -14.609, -13.597, -18.438, -18.548, -18.635, -18.507, -18.102, -17.465,
    -16.651, -15.671, -14.614, -13.600, -18.450, -18.560, -18.647, -18.518,
    -18.112, -17.473, -16.660, -15.679, -14.620, -13.604, -18.462, -18.573,
    -18.659, -18.529, -18.121, -17.481, -16.668, -15.687, -14.626, -13.607,
    -18.473, -18.585, -18.671, -18.540, -18.130, -17.490, -16.676, -15.695,
    -14.632, -13.610, -18.485, -18.598, -18.683, -18.551, -18.140, -17.498,
    -16.685, -15.702, -14.637, -13.612, -18.497, -18.610, -18.695, -18.562,
    -18.149, -17.506, -16.693, -15.710, -14.643, -13.615, -18.508, -18.623,
    -18.707, -18.573, -18.158, -17.515, -16.701, -15.717, -14.648, -13.618,
    -18.520, -18.635, -18.720, -18.584, -18.168, -17.523, -16.710, -15.725,
    -14.653, -13.620, -18.531, -18.648, -18.732, -18.595, -18.177, -17.531,
    -16.718, -15.732, -14.658, -13.623, -18.542, -18.660, -18.744, -18.605,
    -18.186, -17.539, -16.726, -15.739, -14.663, -13.625, -18.554, -18.672,
    -18.755, -18.616, -18.196, -17.548, -16.735, -15.747, -14.668, -13.628,
    -18.565, -18.684, -18.767, -18.627, -18.205, -17.556, -16.743, -15.754,
    -14.673, -13.630, -18.576, -18.696, -18.779, -18.638, -18.214, -17.564,
    -16.751, -15.761, -14.678, -13.632, -18.587, -18.709, -18.791, -18.648,
    -18.223, -17.572, -16.759, -15.768, -14.682, -13.634, -18.598, -18.721,
    -18.803, -18.659, -18.232, -17.580, -16.767, -15.775, -14.687, -13.636,
    -18.609, -18.733, -18.815, -18.670, -18.242, -17.589, -16.775, -15.782,
    -14.691, -13.638, -18.621, -18.745, -18.827, -18.680, -18.251, -17.597,
    -16.783, -15.789, -14.696, -13.640, -18.632, -18.757, -18.838, -18.691,
    -18.260, -17.605, -16.791, -15.796, -14.700, -13.642, -18.642, -18.769,
    -18.850, -18.702, -18.269, -17.613, -16.800, -15.802, -14.704, -13.644,
    -18.653, -18.781, -18.862, -18.712, -18.278, -17.621, -16.808, -15.809,
    -14.709, -13.646, -18.664, -18.793, -18.873, -18.723, -18.287, -17.629,
    -16.815, -15.816, -14.713, -13.647, -18.675, -18.804, -18.885, -18.733,
    -18.296, -17.637, -16.823, -15.822, -14.717, -13.649, -18.686, -18.816,
    -18.897, -18.744, -18.305, -17.645, -16.831, -15.829, -14.721, -13.651,
    -18.697, -18.828, -18.908, -18.754, -18.314, -17.653, -16.839, -15.835,
    -14.725, -13.652, -18.707, -18.840, -18.920, -18.765, -18.323, -17.661,
    -16.847, -15.841, -14.729, -13.654, -18.718, -18.852, -18.931, -18.775,
    -18.332, -17.668, -16.855, -15.848, -14.733, -13.656, -18.729, -18.863,
    -18.943, -18.785, -18.340, -17.676, -16.863, -15.854, -14.736, -13.657,
    -18.739, -18.875, -18.954, -18.796, -18.349, -17.684, -16.871, -15.860,
    -14.740, -13.659, -18.750, -18.887, -18.966, -18.806, -18.358, -17.692,
    -16.878, -15.867, -14.744, -13.661, -18.761, -18.898, -18.977, -18.816,
    -18.367, -17.700, -16.886, -15.873, -14.748, -13.662, -18.771, -18.910,
    -18.988, -18.826, -18.375, -17.707, -16.894, -15.879, -14.751, -13.664,
    -18.782, -18.921, -19.000, -18.836, -18.384, -17.715, -16.902, -15.885,
    -14.755, -13.666, -18.792, -18.933, -19.011, -18.847, -18.393, -17.723,
    -16.909, -15.891, -14.759, -13.668, -18.803, -18.944, -19.022, -18.857,
    -18.401, -17.730, -16.917, -15.897, -14.763, -13.669, -18.813, -18.956,
    -19.033, -18.867, -18.410, -17.738, -16.925, -15.903, -14.766, -13.671,
    -18.823, -18.967, -19.045, -18.877, -18.418, -17.746, -16.932, -15.909,
    -14.770, -13.673, -18.834, -18.979, -19.056, -18.887, -18.427, -17.753,
    -16.940, -15.916, -14.774, -13.675, -18.844, -18.990, -19.067, -18.897,
    -18.435, -17.761, -16.947, -15.919, -14.773, -13.672, -18.854, -19.001,
    -19.078, -18.907, -18.444, -17.768, -16.955, -15.922, -14.771, -13.668,
    -18.865, -19.013, -19.089, -18.916, -18.452, -17.776, -16.962, -15.923,
    -14.766, -13.660, -18.875, -19.024, -19.100, -18.926, -18.460, -17.783,
    -16.970, -15.923, -14.759, -13.650, -18.885, -19.035, -19.111, -18.936,
    -18.469, -17.790, -16.977, -15.927, -14.759, -13.648, -18.895, -19.046,
    -19.122, -18.946, -18.477, -17.798, -16.984, -15.930, -14.758, -13.645,
    -18.906, -19.057, -19.133, -18.955, -18.485, -17.805, -16.992, -15.935,
    -14.761, -13.647, -18.916, -19.069, -19.144, -18.965, -18.493, -17.812,
    -16.999, -15.944, -14.771, -13.655, -18.926, -19.080, -19.155, -18.974,
    -18.501, -17.819, -17.006, -15.950, -14.775, -13.658, -18.936, -19.091,
    -19.165, -18.984, -18.509, -17.826, -17.013, -15.958, -14.782, -13.664,
    -18.946, -19.102, -19.176, -18.993, -18.517, -17.834, -17.021, -15.967,
    -14.792, -13.673, -18.956, -19.113, -19.187, -19.003, -18.525, -17.841,
    -17.028, -15.972, -14.795, -13.675, -18.966, -19.124, -19.198, -19.012,
    -18.533, -17.848, -17.035, -15.980, -14.802, -13.680, -18.976, -19.135,
    -19.208, -19.021, -18.540, -17.855, -17.042, -15.988, -14.811, -13.688,
    -18.986, -19.146, -19.219, -19.031, -18.548, -17.862, -17.049, -15.992,
    -14.811, -13.686, -18.996, -19.157, -19.229, -19.040, -18.556, -17.868,
    -17.056, -15.992, -14.806, -13.679, -19.006, -19.168, -19.240, -19.049,
    -18.563, -17.875, -17.063, -15.993, -14.801, -13.671, -19.016, -19.178,
    -19.250, -19.058, -18.571, -17.882, -17.070, -15.990, -14.789, -13.657,
    -19.026, -19.189, -19.261, -19.067, -18.578, -17.889, -17.077, -15.989,
    -14.783, -13.649, -19.036, -19.200, -19.271, -19.076, -18.586, -17.895,
    -17.084, -15.996, -14.789, -13.654, -19.046, -19.211, -19.281, -19.084,
    -18.593, -17.902, -17.091, -16.004, -14.798, -13.663, -19.056, -19.222,
    -19.292, -19.093, -18.600, -17.909, -17.097, -16.017, -14.814, -13.678,
    -19.066, -19.232, -19.302, -19.102, -18.607, -17.915, -17.104, -16.033,
    -14.835, -13.699, -19.076, -19.243, -19.312, -19.110, -18.614, -17.921,
    -17.111, -16.046, -14.853, -13.716, -19.086, -19.254, -19.322, -19.119,
    -18.621, -17.928, -17.118, -16.058, -14.867, -13.729, -19.096, -19.264,
    -19.332, -19.127, -18.628, -17.934, -17.124, -16.065, -14.874, -13.734,
    -19.106, -19.275, -19.342, -19.136, -18.635, -17.940, -17.131, -16.069,
    -14.874, -13.733, -19.115, -19.286, -19.352, -19.144, -18.641, -17.946,
    -17.137, -16.068, -14.867, -13.725, -19.125, -19.296, -19.362, -19.152,
    -18.648, -17.953, -17.144, -16.064, -14.855, -13.711, -19.135, -19.307,
    -19.372, -19.160, -18.655, -17.959, -17.150, -16.058, -14.840, -13.693,
    -19.145, -19.317, -19.381, -19.168, -18.661, -17.965, -17.157, -16.054,
    -14.828, -13.678, -19.155, -19.327, -19.391, -19.175, -18.667, -17.970,
    -17.163, -16.053, -14.822, -13.671, -19.164, -19.338, -19.401, -19.183,
    -18.673, -17.976, -17.169, -16.055, -14.820, -13.667, -19.174, -19.348,
    -19.410, -19.191, -18.679, -17.982, -17.175, -16.066, -14.833, -13.679,
    -19.184, -19.359, -19.419, -19.198, -18.685, -17.988, -17.182, -16.081,
    -14.853, -13.699, -19.194, -19.369, -19.429, -19.205, -18.691, -17.993,
    -17.188, -16.093, -14.868, -13.713, -19.203, -19.379, -19.438, -19.213,
    -18.697, -17.999, -17.194, -16.106, -14.885, -13.729, -19.213, -19.389,
    -19.447, -19.220, -18.703, -18.004, -17.200, -16.117, -14.898, -13.741,
    -19.223, -19.400, -19.456, -19.227, -18.708, -18.009, -17.206, -16.121,
    -14.900, -13.742, -19.232, -19.410, -19.465, -19.233, -18.713, -18.015,
    -17.212, -16.127, -14.904, -13.745, -19.242, -19.420, -19.474, -19.240,
    -18.718, -18.020, -17.218, -16.134, -14.911, -13.751, -19.252, -19.430,
    -19.483, -19.246, -18.724, -18.025, -17.223, -16.136, -14.910, -13.748,
    -19.261, -19.440, -19.492, -19.253, -18.728, -18.030, -17.229, -16.140,
    -14.911, -13.748, -19.271, -19.450, -19.500, -19.259, -18.733, -18.034,
    -17.235, -16.145, -14.915, -13.751, -19.281, -19.460, -19.509, -19.265,
    -18.738, -18.039, -17.241, -16.146, -14.912, -13.746, -19.290, -19.470,
    -19.517, -19.271, -18.742, -18.044, -17.246, -16.152, -14.918, -13.751,
    -19.300, -19.480, -19.525, -19.276, -18.747, -18.048, -17.252, -16.162,
    -14.929, -13.761, -19.309, -19.490, -19.533, -19.282, -18.751, -18.053,
    -17.257, -16.171, -14.940, -13.770, -19.319, -19.499, -19.541, -19.287,
    -18.755, -18.057, -17.262, -16.186, -14.959, -13.789, -19.329, -19.509,
    -19.549, -19.292, -18.758, -18.061, -17.268, -16.199, -14.977, -13.806,
    -19.338, -19.519, -19.557, -19.297, -18.762, -18.065, -17.273, -16.214,
    -14.997, -13.825, -19.348, -19.528, -19.565, -19.302, -18.766, -18.069,
    -17.278, -16.220, -15.003, -13.830, -19.357, -19.538, -19.572, -19.306,
    -18.769, -18.073, -17.283, -16.227, -15.009, -13.834, -19.367, -19.547,
    -19.579, -19.311, -18.772, -18.077, -17.288, -16.233, -15.015, -13.839,
    -19.376, -19.557, -19.586, -19.315, -18.775, -18.080, -17.293, -16.239,
    -15.021, -13.844, -19.386, -19.566, -19.593, -19.318, -18.777, -18.084,
    -17.298, -16.246, -15.028, -13.850, -19.395, -19.575, -19.600, -19.322,
    -18.780, -18.087, -17.303, -16.252, -15.034, -13.855, -19.405, -19.585,
    -19.606, -19.325, -18.782, -18.090, -17.308, -16.259, -15.041, -13.861,
    -19.414, -19.594, -19.613, -19.328, -18.784, -18.093, -17.312, -16.265,
    -15.048, -13.867, -19.424, -19.603, -19.619, -19.331, -18.786, -18.096,
    -17.317, -16.272, -15.055, -13.872, -19.433, -19.612, -19.625, -19.334,
    -18.788, -18.099, -17.321, -16.278, -15.062, -13.879, -19.443, -19.621,
    -19.631, -19.336, -18.789, -18.102, -17.326, -16.285, -15.069, -13.885,
    -19.452, -19.629, -19.636, -19.338, -18.790, -18.104, -17.330, -16.292,
    -15.076, -13.891, -19.462, -19.638, -19.642, -19.340, -18.791, -18.106,
    -17.334, -16.298, -15.084, -13.898, -19.471, -19.647, -19.647, -19.341,
    -18.792, -18.109, -17.339, -16.305, -15.091, -13.904, -19.480, -19.655,
    -19.651, -19.342, -18.792, -18.111, -17.343, -16.312, -15.099, -13.911,
    -19.490, -19.664, -19.656, -19.343, -18.793, -18.112, -17.347, -16.319,
    -15.107, -13.918, -19.499, -19.672, -19.660, -19.343, -18.793, -18.114,
    -17.351, -16.325, -15.115, -13.925, -19.508, -19.680, -19.664, -19.343,
    -18.792, -18.116, -17.354, -16.332, -15.122, -13.932, -19.518, -19.688,
    -19.668, -19.343, -18.792, -18.117, -17.358, -16.339, -15.130, -13.940,
    -19.527, -19.696, -19.672, -19.343, -18.791, -18.118, -17.362, -16.346,
    -15.139, -13.947, -19.536, -19.704, -19.675, -19.342, -18.790, -18.119,
    -17.365, -16.352, -15.147, -13.954, -19.545, -19.712, -19.678, -19.340,
    -18.789, -18.120, -17.369, -16.359, -15.155, -13.962, -19.555, -19.719,
    -19.680, -19.339, -18.787, -18.121, -17.372, -16.366, -15.163, -13.970,
    -19.564, -19.726, -19.682, -19.337, -18.785, -18.121, -17.375, -16.373,
    -15.172, -13.978, -19.573, -19.734, -19.684, -19.335, -18.783, -18.122,
    -17.378, -16.379, -15.180, -13.985, -19.582, -19.741, -19.686, -19.332,
    -18.780, -18.122, -17.381, -16.386, -15.189, -13.993, -19.591, -19.747,
    -19.687, -19.329, -18.778, -18.122, -17.384, -16.393, -15.197, -14.002,
    -19.600, -19.754, -19.687, -19.325, -18.774, -18.122, -17.387, -16.400,
    -15.206, -14.010, -19.609, -19.761, -19.688, -19.321, -18.771, -18.121,
    -17.390, -16.406, -15.215, -14.018, -19.618, -19.767, -19.688, -19.317,
    -18.767, -18.120, -17.392, -16.413, -15.223, -14.026, -19.627, -19.773,
    -19.687, -19.312, -18.763, -18.120, -17.395, -16.419, -15.232, -14.035,
    -19.636, -19.779, -19.686, -19.307, -18.759, -18.119, -17.397, -16.426,
    -15.241, -14.043, -19.645, -19.784, -19.685, -19.302, -18.755, -18.118,
    -17.400, -16.433, -15.250, -14.052, -19.653, -19.790, -19.683, -19.296,
    -18.750, -18.116, -17.402, -16.439, -15.259, -14.060, -19.662, -19.795,
    -19.681, -19.289, -18.744, -18.115, -17.404, -16.445, -15.267, -14.069,
    -19.671, -19.799, -19.678, -19.283, -18.739, -18.113, -17.406, -16.452,
    -15.276, -14.078, -19.679, -19.804, -19.675, -19.275, -18.733, -18.111,
    -17.408, -16.458, -15.285, -14.087, -19.688, -19.808, -19.671, -19.268,
    -18.727, -18.109, -17.409, -16.465, -15.294, -14.095, -19.696, -19.812,
    -19.667, -19.259, -18.721, -18.106, -17.411, -16.471, -15.303, -14.104,
    -19.705, -19.816, -19.662, -19.251, -18.714, -18.104, -17.412, -16.477,
    -15.312, -14.113, -19.713, -19.819, -19.657, -19.242, -18.707, -18.101,
    -17.414, -16.483, -15.321, -14.122, -19.721, -19.822, -19.652, -19.233,
    -18.699, -18.098, -17.415, -16.489, -15.330, -14.131, -19.729, -19.824,
    -19.645, -19.223, -18.692, -18.094, -17.416, -16.495, -15.339, -14.140,
    -19.737, -19.826, -19.638, -19.212, -18.684, -18.091, -17.417, -16.501,
    -15.348, -14.150, -19.745, -19.828, -19.631, -19.202, -18.675, -18.087,
    -17.418, -16.507, -15.357, -14.159, -19.753, -19.829, -19.623, -19.190,
    -18.667, -18.083, -17.419, -16.513, -15.366, -14.168, -19.760, -19.830,
    -19.615, -19.179, -18.658, -18.079, -17.420, -16.519, -15.375, -14.177,
    -19.768, -19.831, -19.606, -19.167, -18.648, -18.075, -17.420, -16.525,
    -15.384, -14.186, -19.775, -19.830, -19.596, -19.154, -18.639, -18.071,
    -17.420, -16.530, -15.393, -14.196, -19.782, -19.830, -19.586, -19.141,
    -18.629, -18.066, -17.421, -16.536, -15.402, -14.205, -19.789, -19.829,
    -19.575, -19.128, -18.619, -18.061, -17.421, -16.541, -15.411, -14.214,
    -19.796, -19.827, -19.564, -19.114, -18.608, -18.056, -17.421, -16.547,
    -15.420, -14.223, -19.803, -19.825, -19.552, -19.100, -18.598, -18.051,
    -17.421, -16.552, -15.428, -14.233, -19.809, -19.823, -19.540, -19.085,
    -18.586, -18.045, -17.421, -16.558, -15.437, -14.242, -19.815, -19.820,
    -19.527, -19.070, -18.575, -18.040, -17.420, -16.563, -15.446, -14.252,
    -19.821, -19.816, -19.513, -19.055, -18.563, -18.034, -17.420, -16.568,
    -15.455, -14.261, -19.827, -19.811, -19.499, -19.039, -18.552, -18.028,
    -17.419, -16.573, -15.464, -14.270, -19.833, -19.806, -19.484, -19.023,
    -18.539, -18.021, -17.419, -16.578, -15.473, -14.280, -19.838, -19.801,
    -19.469, -19.007, -18.527, -18.015, -17.418, -16.583, -15.481, -14.289,
    -19.843, -19.795, -19.453, -18.990, -18.514, -18.008, -17.417, -16.588,
    -15.490, -14.298, -19.848, -19.788, -19.436, -18.972, -18.501, -18.001,
    -17.416, -16.593, -15.499, -14.308, -19.852, -19.780, -19.419, -18.955,
    -18.488, -17.994, -17.415, -16.598, -15.507, -14.317, -19.856, -19.772,
    -19.402, -18.937, -18.475, -17.987, -17.414, -16.602, -15.516, -14.327,
    -19.860, -19.763, -19.383, -18.919, -18.461, -17.980, -17.412, -16.607,
    -15.525, -14.336, -19.863, -19.753, -19.365, -18.900, -18.447, -17.972,
    -17.411, -16.612, -15.533, -14.345, -19.866, -19.743, -19.346, -18.881,
    -18.433, -17.964, -17.409, -16.616, -15.542, -14.355, -19.868, -19.732,
    -19.326, -18.862, -18.419, -17.956, -17.407, -16.620, -15.550, -14.364,
    -19.870, -19.720, -19.306, -18.842, -18.404, -17.948, -17.406, -16.625,
    -15.559, -14.374, -19.872, -19.708, -19.285, -18.822, -18.389, -17.940,
    -17.404, -16.629, -15.567, -14.383, -19.873, -19.694, -19.264, -18.802,
    -18.374, -17.932, -17.402, -16.633, -15.575, -14.392, -19.873, -19.680,
    -19.242, -18.782, -18.359, -17.923, -17.399, -16.637, -15.584, -14.402,
    -19.873, -19.666, -19.220, -18.762, -18.344, -17.915, -17.397, -16.641,
    -15.592, -14.411, -19.873, -19.650, -19.198, -18.741, -18.329, -17.906,
    -17.395, -16.645, -15.600, -14.420, -19.872, -19.634, -19.175, -18.720,
    -18.313, -17.897, -17.392, -16.649, -15.608, -14.429, -19.870, -19.617,
    -19.152, -18.698, -18.297, -17.888, -17.390, -16.653, -15.617, -14.439,
    -19.868, -19.599, -19.128, -18.677, -18.281, -17.879, -17.387, -16.657,
    -15.625, -14.448, -19.865, -19.581, -19.104, -18.656, -18.265, -17.869,
    -17.385, -16.661, -15.633, -14.457, -19.861, -19.562, -19.080, -18.634,
    -18.249, -17.860, -17.382, -16.664, -15.641, -14.466, -19.857, -19.542,
    -19.055, -18.612, -18.233, -17.851, -17.379, -16.668, -15.649, -14.476,
    -19.851, -19.522, -19.030, -18.590, -18.217, -17.841, -17.376, -16.671,
    -15.657, -14.485, -19.846, -19.501, -19.005, -18.568, -18.200, -17.831,
    -17.373, -16.675, -15.665, -14.494, -19.839, -19.479, -18.979, -18.545,
    -18.184, -17.822, -17.370, -16.678, -15.673, -14.503, -19.832, -19.457,
    -18.954, -18.523, -18.167, -17.812, -17.367, -16.681, -15.680, -14.512,
    -19.823, -19.434, -18.928, -18.501, -18.151, -17.802, -17.364, -16.685,
    -15.688, -14.521, -19.814, -19.410, -18.902, -18.478, -18.134, -17.792,
    -17.360, -16.688, -15.696, -14.530, -19.805, -19.386, -18.875, -18.456,
    -18.117, -17.782, -17.357, -16.691, -15.704, -14.539, -19.794, -19.362,
    -18.849, -18.433, -18.100, -17.772, -17.354, -16.694, -15.711, -14.548,
    -19.783, -19.337, -18.822, -18.410, -18.084, -17.762, -17.350, -16.697,
    -15.719, -14.557, -19.770, -19.311, -18.795, -18.388, -18.067, -17.752,
    -17.347, -16.700, -15.727, -14.566, -19.757, -19.285, -18.768, -18.365,
    -18.050, -17.742, -17.343, -16.703, -15.734, -14.575, -19.743, -19.259,
    -18.741, -18.343, -18.034, -17.732, -17.340, -16.706, -15.742, -14.584,
    -19.728, -19.232, -18.714, -18.320, -18.017, -17.722, -17.336, -16.709,
    -15.749, -14.593, -19.712, -19.205, -18.687, -18.298, -18.001, -17.712,
    -17.333, -16.712, -15.756, -14.602, -19.696, -19.177, -18.660, -18.275,
    -17.984, -17.702, -17.329, -16.714, -15.764, -14.610, -19.679, -19.150,
    -18.633, -18.253, -17.968, -17.692, -17.325, -16.717, -15.771, -14.619,
    -19.660, -19.122, -18.606, -18.231, -17.951, -17.682, -17.322, -16.720,
    -15.778, -14.628, -19.642, -19.093, -18.580, -18.209, -17.935, -17.672,
    -17.318, -16.723, -15.786, -14.636, -19.622, -19.065, -18.553, -18.187,
    -17.919, -17.662, -17.315, -16.725, -15.793, -14.645, -19.602, -19.036,
    -18.526, -18.165, -17.903, -17.652, -17.311, -16.728, -15.800, -14.654,
    -19.581, -19.008, -18.500, -18.144, -17.887, -17.642, -17.308, -16.730,
    -15.807, -14.662, -19.559, -18.979, -18.473, -18.122, -17.871, -17.633,
    -17.304, -16.733, -15.814, -14.671, -19.537, -18.950, -18.447, -18.101,
    -17.856, -17.623, -17.300, -16.736, -15.821, -14.679, -19.514, -18.922,
    -18.421, -18.080, -17.841, -17.614, -17.297, -16.738, -15.828, -14.688,
    -19.491, -18.893, -18.395, -18.060, -17.826, -17.604, -17.294, -16.741,
    -15.835, -14.696, -19.467, -18.865, -18.370, -18.040, -17.811, -17.595,
    -17.290, -16.743, -15.842, -14.705, -19.443, -18.836, -18.345, -18.019,
    -17.796, -17.586, -17.287, -16.746, -15.849, -14.713, -19.418, -18.808,
    -18.320, -18.000, -17.782, -17.577, -17.284, -16.748, -15.856, -14.722,
    -19.394, -18.780, -18.296, -17.980, -17.767, -17.568, -17.280, -16.751,
    -15.863, -14.730, -19.369, -18.752, -18.272, -17.961, -17.754, -17.560,
    -17.277, -16.753, -15.870, -14.738, -19.343, -18.725, -18.248, -17.943,
    -17.740, -17.551, -17.274, -16.756, -15.877, -14.747, -19.318, -18.698,
    -18.225, -17.925, -17.727, -17.543, -17.271, -16.758, -15.883, -14.755,
    -19.293, -18.672, -18.203, -17.907, -17.714, -17.535, -17.268, -16.761,
    -15.890, -14.763, -19.268, -18.646, -18.181, -17.889, -17.701, -17.527,
    -17.266, -16.763, -15.897, -14.771, -19.243, -18.620, -18.159, -17.872,
    -17.689, -17.520, -17.263, -16.766, -15.904, -14.779, -19.218, -18.595,
    -18.138, -17.856, -17.677, -17.513, -17.260, -16.768, -15.910, -14.788,
    -19.193, -18.571, -18.118, -17.840, -17.665, -17.505, -17.258, -16.771,
    -15.917, -14.796, -19.169, -18.547, -18.098, -17.825, -17.654, -17.499,
    -17.256, -16.773, -15.923, -14.804, -19.145, -18.524, -18.079, -17.810,
    -17.643, -17.492, -17.253, -16.776, -15.930, -14.812, -19.122, -18.502,
    -18.060, -17.796, -17.633, -17.486, -17.251, -16.779, -15.937, -14.820,
    -19.099, -18.480, -18.043, -17.782, -17.623, -17.480, -17.249, -16.781,
    -15.943, -14.828, -19.077, -18.460, -18.026, -17.769, -17.613, -17.474,
    -17.248, -16.784, -15.950, -14.836, -18.780, -18.185, -17.801, -17.594,
    -17.484, -17.386, -17.198, -16.766, -15.950, -14.843, -18.669, -18.084,
    -17.717, -17.527, -17.434, -17.351, -17.179, -16.761, -15.954, -14.850,
    -18.596, -18.019, -17.664, -17.486, -17.402, -17.330, -17.167, -16.759,
    -15.960, -14.858, -18.567, -17.996, -17.648, -17.474, -17.395, -17.326,
    -17.166, -16.762, -15.966, -14.866, -18.538, -17.974, -17.632, -17.463,
    -17.387, -17.321, -17.165, -16.765, -15.972, -14.873, -18.509, -17.953,
    -17.617, -17.452, -17.380, -17.318, -17.165, -16.768, -15.979, -14.881,
    -18.486, -17.937, -17.606, -17.445, -17.376, -17.315, -17.165, -16.771,
    -15.985, -14.889, -18.523, -17.972, -17.637, -17.471, -17.395, -17.330,
    -17.175, -16.779, -15.993, -14.897, -18.486, -17.940, -17.611, -17.450,
    -17.380, -17.319, -17.170, -16.780, -15.999, -14.905, -18.433, -17.892,
    -17.570, -17.417, -17.355, -17.302, -17.160, -16.779, -16.004, -14.912,
    -18.359, -17.826, -17.514, -17.371, -17.319, -17.276, -17.146, -16.775,
    -16.008, -14.920, -18.211, -17.690, -17.398, -17.277, -17.245, -17.223,
    -17.114, -16.763, -16.010, -14.926, -18.202, -17.680, -17.388, -17.267,
    -17.237, -17.217, -17.111, -16.765, -16.015, -14.932, -18.160, -17.646,
    -17.361, -17.247, -17.222, -17.207, -17.106, -16.765, -16.020, -14.939,
    -18.109, -17.604, -17.328, -17.222, -17.203, -17.195, -17.100, -16.765,
    -16.025, -14.944, -18.141, -17.641, -17.365, -17.254, -17.231, -17.216,
    -17.115, -16.776, -16.033, -14.952, -18.077, -17.593, -17.331, -17.231,
    -17.215, -17.207, -17.111, -16.777, -16.038, -14.959, -18.055, -17.579,
    -17.323, -17.226, -17.212, -17.206, -17.112, -16.780, -16.044, -14.966,
    -18.138, -17.659, -17.392, -17.284, -17.258, -17.239, -17.134, -16.794,
    -16.054, -14.975, -18.161, -17.683, -17.414, -17.303, -17.273, -17.251,
    -17.142, -16.800, -16.061, -14.983, -18.156, -17.684, -17.418, -17.306,
    -17.277, -17.254, -17.145, -16.805, -16.068, -14.991, -18.231, -17.754,
    -17.477, -17.354, -17.313, -17.280, -17.162, -16.816, -16.077, -15.000,
    -18.163, -17.694, -17.426, -17.313, -17.281, -17.257, -17.149, -16.812,
    -16.081, -15.007, -18.047, -17.584, -17.331, -17.233, -17.217, -17.210,
    -17.120, -16.801, -16.083, -15.014, -18.047, -17.568, -17.306, -17.205,
    -17.191, -17.189, -17.107, -16.797, -16.086, -15.022, -17.922, -17.446,
    -17.195, -17.111, -17.114, -17.131, -17.070, -16.781, -16.086, -15.027,
    -17.762, -17.301, -17.070, -17.007, -17.031, -17.069, -17.030, -16.764,
    -16.085, -15.032, -17.664, -17.222, -17.010, -16.960, -16.997, -17.045,
    -17.016, -16.760, -16.088, -15.037, -17.599, -17.178, -16.983, -16.945,
    -16.989, -17.042, -17.017, -16.763, -16.093, -15.042, -17.599, -17.196,
    -17.009, -16.975, -17.018, -17.067, -17.036, -16.776, -16.101, -15.049,
    -17.705, -17.307, -17.115, -17.069, -17.097, -17.129, -17.077, -16.800,
    -16.114, -15.058, -17.968, -17.546, -17.317, -17.232, -17.222, -17.220,
    -17.136, -16.831, -16.129, -15.068, -18.262, -17.795, -17.513, -17.380,
    -17.331, -17.294, -17.183, -16.855, -16.142, -15.079, -18.458, -17.949,
    -17.624, -17.458, -17.385, -17.331, -17.205, -16.868, -16.152, -15.089,
    -18.378, -17.887, -17.578, -17.426, -17.362, -17.316, -17.197, -16.867,
    -16.158, -15.099, -18.071, -17.637, -17.390, -17.287, -17.263, -17.249,
    -17.158, -16.852, -16.159, -15.106, -17.759, -17.367, -17.172, -17.118,
    -17.136, -17.160, -17.104, -16.828, -16.156, -15.112, -17.517, -17.144,
    -16.979, -16.959, -17.010, -17.067, -17.044, -16.801, -16.152, -15.116,
    -17.418, -17.042, -16.882, -16.872, -16.937, -17.009, -17.006, -16.783,
    -16.150, -15.120, -17.420, -17.030, -16.862, -16.848, -16.913, -16.990,
    -16.992, -16.778, -16.152, -15.124, -17.459, -17.056, -16.877, -16.856,
    -16.917, -16.991, -16.993, -16.780, -16.155, -15.128, -17.568, -17.149,
    -16.953, -16.917, -16.965, -17.026, -17.017, -16.795, -16.164, -15.133,
    -17.687, -17.265, -17.058, -17.008, -17.039, -17.084, -17.057, -16.819,
    -16.175, -15.140, -17.723, -17.313, -17.110, -17.056, -17.082, -17.118,
    -17.082, -16.834, -16.185, -15.147, -17.735, -17.339, -17.141, -17.088,
    -17.110, -17.142, -17.100, -16.846, -16.194, -15.156, -17.804, -17.405,
    -17.200, -17.138, -17.151, -17.173, -17.121, -16.859, -16.204, -15.167,
    -17.786, -17.388, -17.185, -17.126, -17.142, -17.167, -17.118, -16.860,
    -16.210, -15.175, -17.666, -17.283, -17.099, -17.058, -17.090, -17.130,
    -17.095, -16.851, -16.213, -15.184, -17.669, -17.282, -17.096, -17.055,
    -17.087, -17.128, -17.095, -16.853, -16.219, -15.192, -17.562, -17.187,
    -17.017, -16.990, -17.037, -17.091, -17.072, -16.844, -16.220, -15.198,
    -17.453, -17.088, -16.932, -16.921, -16.982, -17.050, -17.046, -16.833,
    -16.221, -15.204, -17.513, -17.136, -16.968, -16.948, -17.002, -17.065,
    -17.056, -16.840, -16.228, -15.211, -17.485, -17.105, -16.938, -16.920,
    -16.978, -17.046, -17.044, -16.836, -16.230, -15.217, -17.447, -17.067,
    -16.904, -16.890, -16.954, -17.028, -17.032, -16.832, -16.233, -15.223,
    -17.475, -17.091, -16.922, -16.905, -16.965, -17.036, -17.039, -16.838,
    -16.239, -15.230, -17.451, -17.069, -16.904, -16.890, -16.953, -17.028,
    -17.034, -16.838, -16.243, -15.235, -17.496, -17.114, -16.946, -16.927,
    -16.984, -17.052, -17.051, -16.849, -16.251, -15.243, -17.603, -17.210,
    -17.027, -16.994, -17.037, -17.092, -17.079, -16.866, -16.262, -15.252,
    -17.628, -17.232, -17.047, -17.011, -17.051, -17.103, -17.087, -16.872,
    -16.269, -15.260, -17.652, -17.255, -17.067, -17.028, -17.065, -17.114,
    -17.095, -16.879, -16.277, -15.270, -17.637, -17.241, -17.055, -17.017,
    -17.057, -17.108, -17.092, -16.879, -16.282, -15.279, -17.566, -17.178,
    -17.003, -16.976, -17.024, -17.084, -17.077, -16.874, -16.285, -15.287,
    -17.563, -17.172, -16.996, -16.969, -17.018, -17.079, -17.075, -16.875,
    -16.290, -15.294, -17.561, -17.167, -16.990, -16.963, -17.013, -17.075,
    -17.073, -16.875, -16.295, -15.301, -17.559, -17.163, -16.985, -16.957,
    -17.008, -17.071, -17.071, -16.876, -16.299, -15.308, -17.560, -17.160,
    -16.981, -16.953, -17.004, -17.068, -17.069, -16.878, -16.304, -15.315,
    -17.646, -17.234, -17.041, -17.002, -17.042, -17.096, -17.089, -16.890,
    -16.313, -15.323, -17.745, -17.317, -17.108, -17.055, -17.083, -17.127,
    -17.109, -16.903, -16.321, -15.330, -18.028, -17.548, -17.290, -17.195,
    -17.189, -17.203, -17.159, -16.931, -16.335, -15.339, -18.024, -17.543,
    -17.287, -17.193, -17.186, -17.201, -17.159, -16.933, -16.340, -15.346,
    -18.021, -17.540, -17.284, -17.190, -17.185, -17.200, -17.159, -16.935,
    -16.345, -15.353, -18.020, -17.538, -17.282, -17.189, -17.183, -17.200,
    -17.159, -16.937, -16.350, -15.360, -18.020, -17.537, -17.281, -17.188,
    -17.183, -17.199, -17.160, -16.939, -16.355, -15.368, -18.021, -17.537,
    -17.280, -17.187, -17.182, -17.199, -17.161, -16.942, -16.360, -15.375,
    -18.023, -17.537, -17.281, -17.188, -17.182, -17.199, -17.161, -16.944,
    -16.366, -15.382, -18.026, -17.539, -17.282, -17.188, -17.183, -17.200,
    -17.163, -16.947, -16.371, -15.389, -18.030, -17.541, -17.284, -17.190,
    -17.184, -17.201, -17.164, -16.949, -16.376, -15.396, -18.035, -17.544,
    -17.286, -17.192, -17.186, -17.202, -17.166, -16.952, -16.381, -15.403,
    -18.040, -17.548, -17.289, -17.194, -17.187, -17.204, -17.167, -16.955,
    -16.386, -15.410, -18.047, -17.553, -17.293, -17.197, -17.190, -17.206,
    -17.169, -16.958, -16.392, -15.417, -18.055, -17.558, -17.297, -17.200,
    -17.192, -17.208, -17.172, -16.961, -16.397, -15.425, -18.063, -17.564,
    -17.302, -17.204, -17.196, -17.210, -17.174, -16.965, -16.402, -15.432,
    -18.072, -17.570, -17.307, -17.209, -17.199, -17.213, -17.177, -16.968,
    -16.408, -15.439, -18.081, -17.578, -17.313, -17.213, -17.203, -17.216,
    -17.179, -16.971, -16.413, -15.446, -18.092, -17.585, -17.319, -17.218,
    -17.207, -17.220, -17.182, -16.975, -16.419, -15.453, -18.103, -17.594,
    -17.326, -17.224, -17.211, -17.223, -17.185, -16.979, -16.424, -15.461,
    -18.114, -17.603, -17.333, -17.230, -17.216, -17.227, -17.189, -16.982,
    -16.429, -15.468, -18.126, -17.612, -17.341, -17.236, -17.221, -17.231,
    -17.192, -16.986, -16.435, -15.475, -18.139, -17.622, -17.349, -17.243,
    -17.227, -17.236, -17.196, -16.990, -16.441, -15.482, -18.152, -17.632,
    -17.358, -17.250, -17.233, -17.240, -17.200, -16.994, -16.446, -15.490,
    -18.166, -17.643, -17.367, -17.258, -17.239, -17.245, -17.204, -16.998,
    -16.452, -15.497, -18.180, -17.654, -17.376, -17.266, -17.245, -17.250,
    -17.208, -17.002, -16.457, -15.504, -18.194, -17.665, -17.386, -17.274,
    -17.252, -17.256, -17.213, -17.007, -16.463, -15.512, -18.209, -17.677,
    -17.396, -17.282, -17.259, -17.261, -17.217, -17.011, -16.469, -15.519,
    -18.225, -17.690, -17.406, -17.291, -17.266, -17.267, -17.222, -17.016,
    -16.474, -15.526, -18.240, -17.702, -17.417, -17.300, -17.273, -17.273,
    -17.227, -17.020, -16.480, -15.533, -18.257, -17.715, -17.427, -17.309,
    -17.281, -17.279, -17.232, -17.025, -16.486, -15.541, -18.273, -17.728,
    -17.439, -17.318, -17.289, -17.285, -17.237, -17.030, -16.492, -15.548,
    -18.290, -17.742, -17.450, -17.328, -17.297, -17.292, -17.242, -17.034,
    -16.498, -15.555, -18.307, -17.756, -17.462, -17.338, -17.305, -17.299,
    -17.248, -17.039, -16.503, -15.563, -18.324, -17.770, -17.474, -17.348,
    -17.313, -17.305, -17.253, -17.044, -16.509, -15.570, -18.342, -17.784,
    -17.486, -17.359, -17.322, -17.312, -17.259, -17.049, -16.515, -15.578,
    -18.360, -17.799, -17.499, -17.369, -17.331, -17.320, -17.265, -17.055,
    -16.521, -15.585, -18.378, -17.813, -17.511, -17.380, -17.340, -17.327,
    -17.271, -17.060, -16.527, -15.592, -18.396, -17.828, -17.524, -17.391,
    -17.349, -17.335, -17.277, -17.065, -16.533, -15.600, -18.414, -17.844,
    -17.537, -17.402, -17.359, -17.342, -17.283, -17.071, -16.539, -15.607,
    -18.433, -17.859, -17.551, -17.414, -17.368, -17.350, -17.289, -17.076,
    -16.545, -15.614, -18.452, -17.875, -17.564, -17.425, -17.378, -17.358,
    -17.296, -17.082, -16.551, -15.622, -18.471, -17.890, -17.578, -17.437,
    -17.388, -17.366, -17.302, -17.087, -16.557, -15.629, -18.490, -17.906,
    -17.591, -17.449, -17.398, -17.375, -17.309, -17.093, -16.563, -15.636,
    -18.510, -17.922, -17.605, -17.461, -17.408, -17.383, -17.316, -17.098,
    -16.569, -15.644, -18.529, -17.938, -17.619, -17.473, -17.419, -17.392,
    -17.323, -17.104, -16.575, -15.651, -18.549, -17.955, -17.633, -17.485,
    -17.429, -17.400, -17.330, -17.110, -16.581, -15.659, -18.568, -17.971,
    -17.648, -17.498, -17.440, -17.409, -17.337, -17.116, -16.588, -15.666,
    -18.588, -17.988, -17.662, -17.510, -17.450, -17.418, -17.344, -17.122,
    -16.594, -15.673, -18.608, -18.004, -17.677, -17.523, -17.461, -17.427,
    -17.351, -17.128, -16.600, -15.681, -18.628, -18.021, -17.691, -17.536,
    -17.472, -17.436, -17.359, -17.134, -16.606, -15.688, -18.648, -18.038,
    -17.706, -17.549, -17.483, -17.445, -17.366, -17.140, -16.612, -15.695,
    -18.668, -18.055, -17.721, -17.562, -17.494, -17.455, -17.374, -17.147,
    -16.619, -15.703, -18.688, -18.072, -17.736, -17.575, -17.506, -17.464,
    -17.381, -17.153, -16.625, -15.710, -18.709, -18.089, -17.751, -17.588,
    -17.517, -17.473, -17.389, -17.159, -16.631, -15.718, -18.729, -18.106,
    -17.766, -17.601, -17.528, -17.483, -17.397, -17.166, -16.638, -15.725,
    -18.749, -18.124, -17.781, -17.615, -17.540, -17.493, -17.405, -17.172,
    -16.644, -15.732, -18.770, -18.141, -17.797, -17.628, -17.552, -17.502,
    -17.412, -17.178, -16.650, -15.740, -18.790, -18.158, -17.812, -17.642,
    -17.563, -17.512, -17.420, -17.185, -16.657, -15.747, -18.811, -18.176,
    -17.827, -17.655, -17.575, -17.522, -17.429, -17.192, -16.663, -15.754,
    -18.831, -18.193, -17.843, -17.669, -17.587, -17.532, -17.437, -17.198,
    -16.669, -15.762, -18.852, -18.210, -17.858, -17.683, -17.599, -17.542,
    -17.445, -17.205, -16.676, -15.769, -18.872, -18.228, -17.874, -17.697,
    -17.611, -17.552, -17.453, -17.212, -16.682, -15.776, -18.892, -18.246,
    -17.890, -17.710, -17.623, -17.563, -17.461, -17.218, -16.689, -15.784,
    -18.913, -18.263, -17.905, -17.724, -17.635, -17.573, -17.470, -17.225,
    -16.695, -15.791, -18.933, -18.281, -17.921, -17.738, -17.647, -17.583,
    -17.478, -17.232, -16.702, -15.798, -18.954, -18.298, -17.937, -17.752,
    -17.660, -17.594, -17.487, -17.239, -16.708, -15.806, -18.974, -18.316,
    -17.952, -17.766, -17.672, -17.604, -17.495, -17.246, -16.715, -15.813,
    -18.995, -18.334, -17.968, -17.781, -17.684, -17.615, -17.504, -17.253,
    -16.721, -15.820, -19.015, -18.351, -17.984, -17.795, -17.697, -17.625,
    -17.513, -17.260, -16.728, -15.828, -19.035, -18.369, -18.000, -17.809,
    -17.709, -17.636, -17.521, -17.267, -16.734, -15.835, -19.056, -18.386,
    -18.016, -17.823, -17.722, -17.646, -17.530, -17.274, -16.741, -15.842,
    -19.076, -18.404, -18.032, -17.837, -17.734, -17.657, -17.539, -17.281,
    -16.747, -15.850, -19.096, -18.422, -18.048, -17.852, -17.747, -17.668,
    -17.548, -17.288, -16.754, -15.857, -19.116, -18.439, -18.064, -17.866,
    -17.759, -17.679, -17.556, -17.295, -16.760, -15.864, -19.137, -18.457,
    -18.079, -17.880, -17.772, -17.690, -17.565, -17.302, -16.767, -15.872,
    -19.157, -18.475, -18.095, -17.895, -17.785, -17.700, -17.574, -17.310,
    -16.774, -15.879, -19.177, -18.492, -18.111, -17.909, -17.797, -17.711,
    -17.583, -17.317, -16.780, -15.886, -19.197, -18.510, -18.127, -17.923,
    -17.810, -17.722, -17.592, -17.324, -16.787, -15.893, -19.217, -18.527,
    -18.143, -17.938, -17.823, -17.733, -17.601, -17.331, -16.794, -15.901,
    -19.236, -18.545, -18.159, -17.952, -17.836, -17.744, -17.611, -17.339,
    -16.800, -15.908, -19.256, -18.562, -18.175, -17.967, -17.849, -17.755,
    -17.620, -17.346, -16.807, -15.915, -19.276, -18.580, -18.191, -17.981,
    -17.861, -17.766, -17.629, -17.354, -16.813, -15.922, -19.296, -18.597,
    -18.207, -17.995, -17.874, -17.778, -17.638, -17.361, -16.820, -15.930,
    -19.315, -18.615, -18.223, -18.010, -17.887, -17.789, -17.647, -17.368,
    -16.827, -15.937, -19.335, -18.632, -18.239, -18.024, -17.900, -17.800,
    -17.657, -17.376, -16.834, -15.944, -19.354, -18.650, -18.255, -18.039,
    -17.913, -17.811, -17.666, -17.383, -16.840, -15.951, -19.374, -18.667,
    -18.271, -18.053, -17.926, -17.822, -17.675, -17.391, -16.847, -15.959,
    -19.393, -18.684, -18.286, -18.068, -17.939, -17.834, -17.685, -17.398,
    -16.854, -15.966, -19.412, -18.701, -18.302, -18.082, -17.952, -17.845,
    -17.694, -17.406, -16.860, -15.973, -19.431, -18.719, -18.318, -18.097,
    -17.965, -17.856, -17.703, -17.413, -16.867, -15.980, -19.450, -18.736,
    -18.334, -18.111, -17.978, -17.867, -17.713, -17.421, -16.874, -15.987,
    -19.469, -18.753, -18.350, -18.126, -17.991, -17.879, -17.722, -17.429,
    -16.880, -15.994, -19.488, -18.770, -18.366, -18.140, -18.004, -17.890,
    -17.732, -17.436, -16.887, -16.002, -19.507, -18.787, -18.381, -18.155,
    -18.017, -17.901, -17.741, -17.444, -16.894, -16.009, -19.525, -18.804,
    -18.397, -18.169, -18.030, -17.913, -17.751, -17.452, -16.901, -16.016,
    -19.544, -18.821, -18.413, -18.184, -18.043, -17.924, -17.760, -17.459,
    -16.907, -16.023, -19.562, -18.838, -18.429, -18.198, -18.056, -17.936,
    -17.770, -17.467, -16.914, -16.030, -19.581, -18.855, -18.444, -18.213,
    -18.069, -17.947, -17.779, -17.475, -16.921, -16.037, -19.599, -18.872,
    -18.460, -18.227, -18.082, -17.958, -17.789, -17.482, -16.928, -16.045,
    -19.617, -18.889, -18.476, -18.241, -18.095, -17.970, -17.798, -17.490,
    -16.934, -16.052, -19.636, -18.905, -18.491, -18.256, -18.108, -17.981,
    -17.808, -17.498, -16.941, -16.059, -19.654, -18.922, -18.507, -18.270,
    -18.121, -17.993, -17.817, -17.505, -16.948, -16.066, -19.672, -18.939,
    -18.522, -18.285, -18.134, -18.004, -17.827, -17.513, -16.955, -16.073,
    -19.689, -18.955, -18.538, -18.299, -18.147, -18.016, -17.837, -17.521,
    -16.961, -16.080, -19.707, -18.972, -18.553, -18.313, -18.160, -18.027,
    -17.846, -17.529, -16.968, -16.087, -19.725, -18.988, -18.569, -18.328,
    -18.173, -18.039, -17.856, -17.537, -16.975, -16.094, -19.742, -19.005,
    -18.584, -18.342, -18.186, -18.050, -17.866, -17.544, -16.982, -16.101,
    -19.760, -19.021, -18.600, -18.356, -18.199, -18.061, -17.875, -17.552,
    -16.989, -16.108, -19.777, -19.038, -18.615, -18.371, -18.212, -18.073,
    -17.885, -17.560, -16.995, -16.115, -19.794, -19.054, -18.630, -18.385,
    -18.225, -18.084, -17.895, -17.568, -17.002, -16.122, -19.811, -19.070,
    -18.646, -18.399, -18.238, -18.096, -17.904, -17.576, -17.009, -16.129,
    -19.829, -19.086, -18.661, -18.413, -18.251, -18.107, -17.914, -17.583,
    -17.016, -16.137, -19.845, -19.102, -18.676, -18.427, -18.264, -18.119,
    -17.924, -17.591, -17.023, -16.144, -19.862, -19.118, -18.692, -18.442,
    -18.277, -18.130, -17.933, -17.599, -17.029, -16.151, -19.879, -19.134,
    -18.707, -18.456, -18.290, -18.142, -17.943, -17.607, -17.036, -16.158,
    -19.896, -19.150, -18.722, -18.470, -18.303, -18.153, -17.953, -17.615,
    -17.043, -16.165, -19.912, -19.166, -18.737, -18.484, -18.316, -18.165,
    -17.962, -17.623, -17.050, -16.172, -19.929, -19.182, -18.752, -18.498,
    -18.329, -18.176, -17.972, -17.631, -17.056, -16.179, -19.945, -19.198,
    -18.767, -18.512, -18.342, -18.188, -17.982, -17.638, -17.063, -16.185,
    -19.961, -19.214, -18.782, -18.526, -18.355, -18.199, -17.992, -17.646,
    -17.070, -16.192, -19.977, -19.229, -18.797, -18.540, -18.368, -18.211,
    -18.001, -17.654, -17.077, -16.199, -19.994, -19.245, -18.812, -18.554,
    -18.380, -18.222, -18.011, -17.662, -17.084, -16.206, -20.010, -19.260,
    -18.827, -18.568, -18.393, -18.234, -18.021, -17.670, -17.090, -16.213,
    -20.025, -19.276, -18.842, -18.582, -18.406, -18.245, -18.030, -17.678,
    -17.097, -16.220, -20.041, -19.291, -18.856, -18.596, -18.419, -18.257,
    -18.040, -17.686, -17.104, -16.227, -20.057, -19.307, -18.871, -18.610,
    -18.432, -18.268, -18.050, -17.694, -17.111, -16.234, -20.072, -19.322,
    -18.886, -18.624, -18.445, -18.279, -18.060, -17.701, -17.117, -16.241,
    -20.088, -19.337, -18.901, -18.638, -18.457, -18.291, -18.069, -17.709,
    -17.124, -16.248, -20.103, -19.353, -18.915, -18.652, -18.470, -18.302,
    -18.079, -17.717, -17.131, -16.255, -20.119, -19.368, -18.930, -18.666,
    -18.483, -18.314, -18.089, -17.725, -17.138, -16.262, -20.134, -19.383,
    -18.944, -18.679, -18.496, -18.325, -18.098, -17.733, -17.145, -16.269,
    -20.149, -19.398, -18.959, -18.693, -18.508, -18.336, -18.108, -17.741,
    -17.151, -16.275, -20.164, -19.413, -18.973, -18.707, -18.521, -18.348,
    -18.118, -17.749, -17.158, -16.282, -20.179, -19.428, -18.988, -18.721,
    -18.534, -18.359, -18.128, -17.757, -17.165, -16.289, -20.194, -19.443,
    -19.002, -18.734, -18.547, -18.371, -18.137, -17.764, -17.172, -16.296,
    -20.208, -19.458, -19.017, -18.748, -18.559, -18.382, -18.147, -17.772,
    -17.178, -16.303, -20.223, -19.472, -19.031, -18.762, -18.572, -18.393,
    -18.157, -17.780, -17.185, -16.310, -20.238, -19.487, -19.045, -18.775,
    -18.585, -18.405, -18.166, -17.788, -17.192, -16.316, -20.252, -19.502,
    -19.060, -18.789, -18.597, -18.416, -18.176, -17.796, -17.198, -16.323,
    -20.266, -19.516, -19.074, -18.802, -18.610, -18.427, -18.186, -17.804,
    -17.205, -16.330, -20.281, -19.531, -19.088, -18.816, -18.622, -18.439,
    -18.195, -17.812, -17.212, -16.337, -20.295, -19.545, -19.102, -18.829,
    -18.635, -18.450, -18.205, -17.820, -17.219, -16.344, -20.309, -19.560,
    -19.116, -18.843, -18.647, -18.461, -18.215, -17.827, -17.225, -16.350,
    -20.323, -19.574, -19.130, -18.856, -18.660, -18.472, -18.224, -17.835,
    -17.232, -16.357, -20.337, -19.588, -19.144, -18.870, -18.672, -18.484,
    -18.234, -17.843, -17.239, -16.364, -20.351, -19.603, -19.158, -18.883,
    -18.685, -18.495, -18.244, -17.851, -17.245, -16.371, -20.365, -19.617,
    -19.172, -18.896, -18.697, -18.506, -18.253, -17.859, -17.252, -16.377,
    -20.378, -19.631, -19.186, -18.910, -18.710, -18.517, -18.263, -17.867,
    -17.259, -16.384, -20.392, -19.645, -19.200, -18.923, -18.722, -18.529,
    -18.272, -17.874, -17.266, -16.391, -20.406, -19.659, -19.214, -18.936,
    -18.735, -18.540, -18.282, -17.882, -17.272, -16.397, -20.419, -19.673,
    -19.228, -18.949, -18.747, -18.551, -18.292, -17.890, -17.279, -16.404,
    -20.433, -19.687, -19.241, -18.963, -18.759, -18.562, -18.301, -17.898,
    -17.286, -16.411, -20.446, -19.701, -19.255, -18.976, -18.772, -18.573,
    -18.311, -17.906, -17.292, -16.417, -20.459, -19.715, -19.269, -18.989,
    -18.784, -18.584, -18.320, -17.914, -17.299, -16.424, -20.472, -19.729,
    -19.282, -19.002, -18.796, -18.596, -18.330, -17.921, -17.306, -16.431,
    -20.485, -19.742, -19.296, -19.015, -18.809, -18.607, -18.339, -17.929,
    -17.312, -16.437, -20.498, -19.756, -19.310, -19.028, -18.821, -18.618,
    -18.349, -17.937, -17.319, -16.444, -20.511, -19.770, -19.323, -19.041,
    -18.833, -18.629, -18.358, -17.945, -17.325, -16.451, -20.524, -19.783,
    -19.337, -19.054, -18.845, -18.640, -18.368, -17.952, -17.332, -16.457,
    -20.537, -19.797, -19.350, -19.067, -18.857, -18.651, -18.377, -17.960,
    -17.339, -16.464, -20.550, -19.810, -19.363, -19.080, -18.870, -18.662,
    -18.387, -17.968, -17.345, -16.471, -20.563, -19.824, -19.377, -19.093,
    -18.882, -18.673, -18.396, -17.976, -17.352, -16.477, -20.575, -19.837,
    -19.390, -19.106, -18.894, -18.684, -18.406, -17.983, -17.358, -16.484,
    -20.588, -19.851, -19.403, -19.119, -18.906, -18.695, -18.415, -17.991,
    -17.365, -16.490, -20.600, -19.864, -19.417, -19.131, -18.918, -18.706,
    -18.425, -17.999, -17.372, -16.497, -20.613, -19.877, -19.430, -19.144,
    -18.930, -18.717, -18.434, -18.006, -17.378, -16.503, -20.625, -19.890,
    -19.443, -19.157, -18.942, -18.728, -18.443, -18.014, -17.385, -16.510,
    -20.637, -19.903, -19.456, -19.170, -18.954, -18.739, -18.453, -18.022,
    -17.391, -16.517, -20.650, -19.917, -19.469, -19.182, -18.966, -18.750,
    -18.462, -18.029, -17.398, -16.523, -20.662, -19.930, -19.482, -19.195,
    -18.978, -18.761, -18.472, -18.037, -17.404, -16.530, -20.674, -19.943,
    -19.495, -19.208, -18.990, -18.771, -18.481, -18.045, -17.411, -16.536,
    -20.686, -19.956, -19.508, -19.220, -19.002, -18.782, -18.490, -18.052,
    -17.417, -16.543, -20.698, -19.969, -19.521, -19.233, -19.014, -18.793,
    -18.500, -18.060, -17.424, -16.549, -20.710, -19.981, -19.534, -19.246,
    -19.026, -18.804, -18.509, -18.068, -17.430, -16.556, -20.722, -19.994,
    -19.547, -19.258, -19.038, -18.815, -18.518, -18.075, -17.437, -16.562,
    -20.734, -20.007, -19.560, -19.271, -19.049, -18.825, -18.528, -18.083,
    -17.443, -16.569, -20.746, -20.020, -19.572, -19.283, -19.061, -18.836,
    -18.537, -18.090, -17.450, -16.575, -20.757, -20.032, -19.585, -19.295,
    -19.073, -18.847, -18.546, -18.098, -17.456, -16.581, -20.769, -20.045,
    -19.598, -19.308, -19.085, -18.858, -18.555, -18.106, -17.463, -16.588,
    -20.781, -20.058, -19.611, -19.320, -19.096, -18.868, -18.564, -18.113,
    -17.469, -16.594, -20.792, -20.070, -19.623, -19.333, -19.108, -18.879,
    -18.574, -18.121, -17.476, -16.601, -20.804, -20.083, -19.636, -19.345,
    -19.120, -18.890, -18.583, -18.128, -17.482, -16.607, -20.815, -20.095,
    -19.648, -19.357, -19.132, -18.900, -18.592, -18.136, -17.488, -16.613,
    -20.827, -20.108, -19.661, -19.369, -19.143, -18.911, -18.601, -18.143,
    -17.495, -16.620, -20.838, -20.120, -19.674, -19.382, -19.155, -18.922,
    -18.610, -18.151, -17.501, -16.626, -20.850, -20.133, -19.686, -19.394,
    -19.166, -18.932, -18.619, -18.158, -17.508, -16.633, -20.861, -20.145,
    -19.698, -19.406, -19.178, -18.943, -18.629, -18.165, -17.514, -16.639,
    -20.872, -20.157, -19.711, -19.418, -19.190, -18.953, -18.638, -18.173,
    -17.520, -16.645, -20.883, -20.169, -19.723, -19.430, -19.201, -18.964,
    -18.647, -18.180, -17.527, -16.652, -20.895, -20.182, -19.736, -19.442,
    -19.213, -18.974, -18.656, -18.188, -17.533, -16.658, -20.906, -20.194,
    -19.748, -19.454, -19.224, -18.985, -18.665, -18.195, -17.539, -16.664,
    -20.917, -20.206, -19.760, -19.466, -19.236, -18.995, -18.674, -18.202,
    -17.546, -16.670, -20.928, -20.218, -19.772, -19.478, -19.247, -19.006,
    -18.683, -18.210, -17.552, -16.677, -20.939, -20.230, -19.785, -19.490,
    -19.258, -19.016, -18.692, -18.217, -17.558, -16.683, -20.950, -20.242,
    -19.797, -19.502, -19.270, -19.027, -18.701, -18.224, -17.564, -16.689,
    -20.961, -20.254, -19.809, -19.514, -19.281, -19.037, -18.710, -18.232,
    -17.571, -16.696, -20.972, -20.266, -19.821, -19.526, -19.292, -19.047,
    -18.718, -18.239, -17.577, -16.702, -20.983, -20.278, -19.833, -19.538,
    -19.304, -19.058, -18.727, -18.246, -17.583, -16.708, -20.994, -20.290,
    -19.845, -19.550, -19.315, -19.068, -18.736, -18.254, -17.589, -16.714,
    -21.005, -20.302, -19.857, -19.562, -19.326, -19.078, -18.745, -18.261,
    -17.596, -16.720, -21.015, -20.314, -19.869, -19.573, -19.338, -19.089,
    -18.754, -18.268, -17.602, -16.727, -21.026, -20.325, -19.881, -19.585,
    -19.349, -19.099, -18.763, -18.275, -17.608, -16.733, -21.037, -20.337,
    -19.893, -19.597, -19.360, -19.109, -18.771, -18.282, -17.614, -16.739,
    -21.048, -20.349, -19.905, -19.609, -19.371, -19.119, -18.780, -18.289,
    -17.620, -16.745, -21.058, -20.360, -19.917, -19.620, -19.382, -19.129,
    -18.789, -18.297, -17.626, -16.751, -21.069, -20.372, -19.929, -19.632,
    -19.393, -19.140, -18.798, -18.304, -17.633, -16.757, -21.080, -20.384,
    -19.940, -19.644, -19.405, -19.150, -18.806, -18.311, -17.639, -16.764,
    -21.090, -20.395, -19.952, -19.655, -19.416, -19.160, -18.815, -18.318,
    -17.645, -16.770, -21.101, -20.407, -19.964, -19.667, -19.427, -19.170,
    -18.824, -18.325, -17.651, -16.776, -21.111, -20.418, -19.976, -19.678,
    -19.438, -19.180, -18.832, -18.332, -17.657, -16.782, -21.122, -20.430,
    -19.987, -19.690, -19.449, -19.190, -18.841, -18.339, -17.663, -16.788,
    -21.132, -20.441, -19.999, -19.701, -19.460, -19.200, -18.849, -18.346,
    -17.669, -16.794, -21.143, -20.453, -20.010, -19.713, -19.471, -19.210,
    -18.858, -18.353, -17.675, -16.800, -21.153, -20.464, -20.022, -19.724,
    -19.482, -19.220, -18.866, -18.360, -17.681, -16.806, -21.164, -20.476,
    -20.034, -19.736, -19.492, -19.230, -18.875, -18.367, -17.687, -16.812,
    -21.174, -20.487, -20.045, -19.747, -19.503, -19.240, -18.883, -18.374,
    -17.693, -16.818, -21.184, -20.498, -20.057, -19.758, -19.514, -19.250,
    -18.892, -18.380, -17.699, -16.824, -21.195, -20.510, -20.068, -19.770,
    -19.525, -19.259, -18.900, -18.387, -17.705, -16.830, -21.205, -20.521,
    -20.080, -19.781, -19.536, -19.269, -18.908, -18.394, -17.711, -16.836,
    -21.215, -20.532, -20.091, -19.792, -19.547, -19.279, -18.917, -18.401,
    -17.717, -16.842, -21.226, -20.543, -20.102, -19.803, -19.557, -19.289,
    -18.925, -18.408, -17.722, -16.848, -21.236, -20.554, -20.114, -19.815,
    -19.568, -19.298, -18.933, -18.414, -17.728, -16.854, -21.246, -20.566,
    -20.125, -19.826, -19.579, -19.308, -18.941, -18.421, -17.734, -16.860,
    -21.256, -20.577, -20.136, -19.837, -19.589, -19.318, -18.950, -18.428,
    -17.740, -16.866, -21.267, -20.588, -20.148, -19.848, -19.600, -19.327,
    -18.958, -18.434, -17.746, -16.872, -21.277, -20.599, -20.159, -19.859,
    -19.611, -19.337, -18.966, -18.441, -17.752, -16.878, -21.287, -20.610,
    -20.170, -19.870, -19.621, -19.347, -18.974, -18.448, -17.757, -16.884,
    -21.297, -20.621, -20.181, -19.881, -19.632, -19.356, -18.982, -18.454,
    -17.763, -16.890, -21.307, -20.632, -20.192, -19.892, -19.642, -19.366,
    -18.990, -18.461, -17.769, -16.895, -21.317, -20.643, -20.204, -19.903,
    -19.653, -19.375, -18.998, -18.467, -17.775, -16.901, -21.328, -20.654,
    -20.215, -19.914, -19.663, -19.385, -19.006, -18.474, -17.780, -16.907,
    -21.338, -20.665, -20.226, -19.925, -19.674, -19.394, -19.014, -18.480,
    -17.786, -16.913, -21.348, -20.676, -20.237, -19.936, -19.684, -19.403,
    -19.022, -18.487, -17.792, -16.919, -21.358, -20.687, -20.248, -19.947,
    -19.694, -19.413, -19.030, -18.493, -17.797, -16.925, -21.368, -20.698,
    -20.259, -19.958, -19.705, -19.422, -19.037, -18.499, -17.803, -16.930,
    -21.378, -20.708, -20.270, -19.969, -19.715, -19.431, -19.045, -18.506,
    -17.808, -16.936, -21.388, -20.719, -20.281, -19.980, -19.725, -19.440,
    -19.053, -18.512, -17.814, -16.942, -21.398, -20.730, -20.292, -19.990,
    -19.736, -19.450, -19.060, -18.518, -17.819, -16.948, -21.408, -20.741,
    -20.303, -20.001, -19.746, -19.459, -19.068, -18.525, -17.825, -16.953,
    -21.418, -20.752, -20.314, -20.012, -19.756, -19.468, -19.076, -18.531,
    -17.830, -16.959, -21.428, -20.762, -20.325, -20.023, -19.766, -19.477,
    -19.083, -18.537, -17.836, -16.965, -21.438, -20.773, -20.336, -20.033,
    -19.776, -19.486, -19.091, -18.543, -17.841, -16.970, -21.448, -20.784,
    -20.346, -20.044, -19.786, -19.495, -19.098, -18.549, -17.847, -16.976,
    -21.458, -20.794, -20.357, -20.055, -19.796, -19.504, -19.106, -18.555,
    -17.852, -16.982, -21.468, -20.805, -20.368, -20.065, -19.806, -19.513,
    -19.113, -18.561, -17.858, -16.987, -21.478, -20.816, -20.379, -20.076,
    -19.816, -19.522, -19.120, -18.567, -17.863, -16.993, -21.488, -20.826,
    -20.389, -20.086, -19.826, -19.530, -19.127, -18.573, -17.868, -16.999,
    -21.498, -20.837, -20.400, -20.097, -19.836, -19.539, -19.135, -18.579,
    -17.874, -17.004, -21.507, -20.848, -20.411, -20.107, -19.846, -19.548,
    -19.142, -18.585, -17.879, -17.010, -21.517, -20.858, -20.422, -20.118,
    -19.856, -19.556, -19.149, -18.590, -17.884, -17.015, -21.527, -20.869,
    -20.432, -20.128, -19.866, -19.565, -19.156, -18.596, -17.889, -17.021,
    -21.537, -20.879, -20.443, -20.139, -19.876, -19.574, -19.163, -18.602,
    -17.895, -17.026, -21.547, -20.890, -20.453, -20.149, -19.885, -19.582,
    -19.170, -18.608, -17.900, -17.032, -21.557, -20.900, -20.464, -20.160,
    -19.895, -19.591, -19.177, -18.613, -17.905, -17.038, -21.567, -20.911,
    -20.475, -20.170, -19.905, -19.599, -19.184, -18.619, -17.910, -17.043,
    -21.577, -20.921, -20.485, -20.180, -19.914, -19.607, -19.190, -18.624,
    -17.915, -17.048, -21.586, -20.932, -20.496, -20.190, -19.924, -19.616,
    -19.197, -18.630, -17.920, -17.054, -21.596, -20.942, -20.506, -20.201,
    -19.934, -19.624, -19.204, -18.635, -17.925, -17.059, -21.606, -20.952,
    -20.517, -20.211, -19.943, -19.632, -19.210, -18.641, -17.930, -17.065,
    -21.616, -20.963, -20.527, -20.221, -19.952, -19.640, -19.217, -18.646,
    -17.935, -17.070, -21.626, -20.973, -20.537, -20.231, -19.962, -19.648,
    -19.223, -18.651, -17.940, -17.076, -21.636, -20.984, -20.548, -20.241,
    -19.971, -19.656, -19.230, -18.657, -17.945, -17.081, -21.645, -20.994,
    -20.558, -20.252, -19.981, -19.664, -19.236, -18.662, -17.950, -17.086,
    -21.655, -21.004, -20.569, -20.262, -19.990, -19.672, -19.242, -18.667,
    -17.955, -17.092, -21.665, -21.015, -20.579, -20.272, -19.999, -19.680,
    -19.248, -18.672, -17.960, -17.097, -21.675, -21.025, -20.589, -20.282,
    -20.008, -19.687, -19.255, -18.677, -17.965, -17.102, -21.685, -21.035,
    -20.600, -20.292, -20.017, -19.695, -19.261, -18.682, -17.970, -17.108,
    -21.694, -21.046, -20.610, -20.302, -20.026, -19.703, -19.267, -18.687,
    -17.974, -17.113, -21.704, -21.056, -20.620, -20.312, -20.035, -19.710,
    -19.273, -18.692, -17.979, -17.118, -21.714, -21.066, -20.630, -20.321,
    -20.044, -19.718, -19.278, -18.697, -17.984, -17.124, -21.724, -21.076,
    -20.641, -20.331, -20.053, -19.725, -19.284, -18.702, -17.989, -17.129,
    -21.734, -21.086, -20.651, -20.341, -20.062, -19.732, -19.290, -18.707,
    -17.993, -17.134, -21.743, -21.097, -20.661, -20.351, -20.071, -19.739,
    -19.295, -18.711, -17.998, -17.139, -21.753, -21.107, -20.671, -20.361,
    -20.080, -19.747, -19.301, -18.716, -18.002, -17.144, -21.763, -21.117,
    -20.681, -20.370, -20.088, -19.754, -19.306, -18.721, -18.007, -17.150,
    -21.773, -21.127, -20.691, -20.380, -20.097, -19.761, -19.312, -18.725,
    -18.012, -17.155, -21.782, -21.137, -20.701, -20.390, -20.106, -19.767,
    -19.317, -18.730, -18.016, -17.160, -21.792, -21.148, -20.712, -20.399,
    -20.114, -19.774, -19.322, -18.734, -18.020, -17.165, -21.802, -21.158,
    -20.722, -20.409, -20.123, -19.781, -19.327, -18.738, -18.025, -17.170,
    -21.812, -21.168, -20.732, -20.418, -20.131, -19.788, -19.332, -18.743,
    -18.029, -17.175, -21.822, -21.178, -20.742, -20.428, -20.139, -19.794,
    -19.337, -18.747, -18.034, -17.180, -21.831, -21.188, -20.752, -20.437,
    -20.147, -19.800, -19.342, -18.751, -18.038, -17.185, -21.841, -21.198,
    -20.762, -20.447, -20.156, -19.807, -19.347, -18.755, -18.042, -17.190,
    -21.851, -21.208, -20.772, -20.456, -20.164, -19.813, -19.352, -18.759,
    -18.047, -17.195, -21.861, -21.218, -20.781, -20.465, -20.172, -19.819,
    -19.356, -18.763, -18.051, -17.200, -21.870, -21.228, -20.791, -20.475,
    -20.180, -19.825, -19.361, -18.767, -18.055, -17.205, -21.880, -21.238,
    -20.801, -20.484, -20.187, -19.831, -19.365, -18.771, -18.059, -17.210,
    -21.890, -21.248, -20.811, -20.493, -20.195, -19.837, -19.369, -18.775,
    -18.063, -17.215, -21.900, -21.258, -20.821, -20.502, -20.203, -19.843,
    -19.374, -18.778, -18.067, -17.220, -21.909, -21.268, -20.831, -20.511,
    -20.211, -19.848, -19.378, -18.782, -18.071, -17.225, -21.919, -21.278,
    -20.841, -20.521, -20.218, -19.854, -19.382, -18.786, -18.076, -17.230,
    -21.929, -21.288, -20.850, -20.530, -20.226, -19.859, -19.386, -18.789,
    -18.080, -17.235, -21.939, -21.298, -20.860, -20.539, -20.233, -19.865,
    -19.389, -18.793, -18.084, -17.240, -21.949, -21.308, -20.870, -20.547,
    -20.240, -19.870, -19.393, -18.796, -18.087, -17.245, -21.958, -21.318,
    -20.879, -20.556, -20.247, -19.875, -19.397, -18.800, -18.091, -17.250,
    -21.968, -21.328, -20.889, -20.565, -20.254, -19.880, -19.400, -18.803,
    -18.095, -17.254, -21.978, -21.338, -20.899, -20.574, -20.261, -19.884,
    -19.404, -18.806, -18.099, -17.259, -21.988, -21.348, -20.908, -20.582,
    -20.268, -19.889, -19.407, -18.809, -18.103, -17.264, -21.997, -21.358,
    -20.918, -20.591, -20.275, -19.894, -19.410, -18.812, -18.107, -17.269,
    -22.007, -21.368, -20.928, -20.600, -20.282, -19.898, -19.413, -18.815,
    -18.110, -17.274, -22.017, -21.378, -20.937, -20.608, -20.288, -19.902,
    -19.416, -18.818, -18.114, -17.278, -22.027, -21.388, -20.947, -20.617,
    -20.295, -19.907, -19.419, -18.821, -18.118, -17.283, -22.036, -21.398,
    -20.956, -20.625, -20.301, -19.911, -19.422, -18.824, -18.121, -17.288,
    -22.046, -21.407, -20.966, -20.633, -20.307, -19.915, -19.424, -18.827,
    -18.125, -17.292, -22.056, -21.417, -20.975, -20.642, -20.313, -19.918,
    -19.427, -18.829, -18.128, -17.297, -22.066, -21.427, -20.984, -20.650,
    -20.319, -19.922, -19.429, -18.832, -18.132, -17.302, -22.076, -21.437,
    -20.994, -20.658, -20.325, -19.926, -19.432, -18.835, -18.136, -17.306,
    -22.085, -21.447, -21.003, -20.666, -20.331, -19.929, -19.434, -18.837,
    -18.139, -17.311, -22.095, -21.457, -21.012, -20.674, -20.337, -19.932,
    -19.436, -18.840, -18.142, -17.316, -22.105, -21.466, -21.022, -20.682,
    -20.342, -19.935, -19.438, -18.842, -18.146, -17.320, -22.115, -21.476,
    -21.031, -20.690, -20.347, -19.938, -19.440, -18.844, -18.149, -17.325,
    -22.124, -21.486, -21.040, -20.697, -20.353, -19.941, -19.442, -18.846,
    -18.153, -17.329, -22.134, -21.496, -21.049, -20.705, -20.358, -19.944,
    -19.443, -18.849, -18.156, -17.334, -22.144, -21.505, -21.058, -20.712,
    -20.363, -19.946, -19.445, -18.851, -18.159, -17.338, -22.154, -21.515,
    -21.067, -20.720, -20.368, -19.949, -19.446, -18.853, -18.162, -17.343,
    -22.164, -21.525, -21.076, -20.727, -20.372, -19.951, -19.448, -18.855,
    -18.166, -17.347, -22.173, -21.535, -21.085, -20.735, -20.377, -19.953,
    -19.449, -18.857, -18.169, -17.352, -22.183, -21.544, -21.094, -20.742,
    -20.381, -19.955, -19.450, -18.859, -18.172, -17.356, -22.193, -21.554,
    -21.103, -20.749, -20.386, -19.957, -19.451, -18.860, -18.175, -17.361,
    -22.203, -21.564, -21.112, -20.756, -20.390, -19.959, -19.452, -18.862,
    -18.178, -17.365, -22.213, -21.573, -21.121, -20.763, -20.394, -19.960,
    -19.453, -18.864, -18.181, -17.370, -22.222, -21.583, -21.130, -20.770,
    -20.397, -19.962, -19.454, -18.865, -18.184, -17.374, -22.232, -21.593,
    -21.139, -20.776, -20.401, -19.963, -19.455, -18.867, -18.187, -17.379,
    -22.242, -21.602, -21.147, -20.783, -20.405, -19.964, -19.455, -18.869,
    -18.190, -17.383, -22.252, -21.612, -21.156, -20.789, -20.408, -19.965,
    -19.456, -18.870, -18.193, -17.387, -22.262, -21.622, -21.165, -20.796,
    -20.411, -19.966, -19.456, -18.871, -18.196, -17.392, -22.272, -21.631,
    -21.173, -20.802, -20.414, -19.967, -19.456, -18.873, -18.199, -17.396,
    -22.281, -21.641, -21.182, -20.808, -20.417, -19.967, -19.456, -18.874,
    -18.202, -17.400, -22.291, -21.650, -21.190, -20.814, -20.420, -19.968,
    -19.456, -18.875, -18.205, -17.405, -22.301, -21.660, -21.199, -20.820,
    -20.423, -19.968, -19.457, -18.877, -18.208, -17.409, -22.311, -21.669,
    -21.207, -20.826, -20.425, -19.968, -19.456, -18.878, -18.211, -17.413,
    -22.321, -21.679, -21.215, -20.831, -20.428, -19.968, -19.456, -18.879,
    -18.213, -17.418, -22.330, -21.688, -21.223, -20.837, -20.430, -19.968,
    -19.456, -18.880, -18.216, -17.422, -22.340, -21.698, -21.232, -20.842,
    -20.432, -19.968, -19.456, -18.881, -18.219, -17.426, -22.350, -21.707,
    -21.240, -20.848, -20.434, -19.968, -19.455, -18.882, -18.222, -17.430,
    -22.360, -21.717, -21.248, -20.853, -20.435, -19.967, -19.455, -18.883,
    -18.224, -17.435, -22.370, -21.726, -21.256, -20.858, -20.437, -19.967,
    -19.455, -18.884, -18.227, -17.439, -22.380, -21.736, -21.264, -20.863,
    -20.438, -19.966, -19.454, -18.885, -18.230, -17.443, -22.389, -21.745,
    -21.271, -20.868, -20.440, -19.965, -19.453, -18.886, -18.233, -17.447,
    -22.399, -21.754, -21.279, -20.872, -20.441, -19.965, -19.453, -18.887,
    -18.235, -17.452, -22.409, -21.764, -21.287, -20.877, -20.442, -19.964,
    -19.452, -18.888, -18.238, -17.456, -22.419, -21.773, -21.295, -20.881,
    -20.443, -19.963, -19.451, -18.889, -18.241, -17.460, -22.429, -21.782,
    -21.302, -20.886, -20.444, -19.962, -19.451, -18.890, -18.243, -17.464,
    -22.439, -21.791, -21.310, -20.890, -20.444, -19.961, -19.450, -18.890,
    -18.246, -17.468, -22.449, -21.801, -21.317, -20.894, -20.445, -19.959,
    -19.449, -18.891, -18.248, -17.473, -22.458, -21.810, -21.324, -20.898,
    -20.445, -19.958, -19.448, -18.892, -18.251, -17.477, -22.468, -21.819,
    -21.332, -20.902, -20.446, -19.957, -19.447, -18.893, -18.254, -17.481,
    -22.478, -21.828, -21.339, -20.906, -20.446, -19.955, -19.446, -18.894,
    -18.256, -17.485, -22.488, -21.838, -21.346, -20.909, -20.446, -19.954,
    -19.445, -18.894, -18.259, -17.489, -22.498, -21.847, -21.353, -20.913,
    -20.446, -19.953, -19.444, -18.895, -18.262, -17.493, -22.508, -21.856,
    -21.360, -20.916, -20.446, -19.951, -19.444, -18.896, -18.264, -17.497,
    -22.517, -21.865, -21.367, -20.919, -20.446, -19.950, -19.443, -18.897,
    -18.267, -17.502, -22.527, -21.874, -21.374, -20.923, -20.446, -19.948,
    -19.442, -18.898, -18.269, -17.506, -22.537, -21.883, -21.381, -20.926,
    -20.446, -19.946, -19.441, -18.898, -18.272, -17.510, -22.547, -21.892,
    -21.388, -20.929, -20.445, -19.945, -19.440, -18.899, -18.275, -17.514,
    -22.557, -21.901, -21.394, -20.932, -20.445, -19.943, -19.439, -18.900,
    -18.277, -17.518, -22.567, -21.910, -21.401, -20.935, -20.445, -19.942,
    -19.438, -18.901, -18.280, -17.522, -22.577, -21.919, -21.408, -20.938,
    -20.444, -19.941, -19.438, -18.902, -18.283, -17.526, -22.586, -21.928,
    -21.414, -20.940, -20.444, -19.939, -19.437, -18.903, -18.285, -17.531,
    -22.596, -21.937, -21.421, -20.943, -20.444, -19.938, -19.436, -18.904,
    -18.288, -17.535, -22.606, -21.946, -21.427, -20.946, -20.443, -19.936,
    -19.436, -18.905, -18.291, -17.539, -22.616, -21.954, -21.433, -20.949,
    -20.443, -19.935, -19.435, -18.906, -18.293, -17.543, -22.626, -21.963,
    -21.440, -20.951, -20.443, -19.934, -19.435, -18.907, -18.296, -17.547,
    -22.636, -21.972, -21.446, -20.954, -20.443, -19.933, -19.434, -18.908,
    -18.299, -17.551, -22.645, -21.981, -21.452, -20.957, -20.443, -19.932,
    -19.434, -18.909, -18.302, -17.555, -22.655, -21.990, -21.459, -20.959,
    -20.443, -19.931, -19.434, -18.910, -18.304, -17.559, -22.665, -21.999,
    -21.465, -20.962, -20.443, -19.930, -19.433, -18.912, -18.307, -17.564,
    -22.675, -22.007, -21.471, -20.965, -20.443, -19.929, -19.433, -18.913,
    -18.310, -17.568, -22.685, -22.016, -21.478, -20.968, -20.443, -19.929,
    -19.433, -18.914, -18.313, -17.572, -22.694, -22.018, -21.462, -20.930,
    -20.395, -19.884, -19.400, -18.895, -18.306, -17.572, -22.703, -22.023,
    -21.458, -20.915, -20.374, -19.864, -19.385, -18.887, -18.304, -17.575,
    -22.712, -22.028, -21.454, -20.900, -20.353, -19.844, -19.371, -18.880,
    -18.302, -17.577, -22.722, -22.037, -21.458, -20.901, -20.352, -19.843,
    -19.371, -18.881, -18.305, -17.581, -22.731, -22.045, -21.463, -20.903,
    -20.352, -19.843, -19.372, -18.883, -18.308, -17.585, -22.741, -22.053,
    -21.469, -20.905, -20.352, -19.843, -19.372, -18.885, -18.311, -17.590,
    -22.751, -22.061, -21.474, -20.907, -20.353, -19.843, -19.373, -18.887,
    -18.314, -17.594, -22.761, -22.072, -21.486, -20.921, -20.366, -19.855,
    -19.383, -18.894, -18.320, -17.599, -22.771, -22.080, -21.492, -20.925,
    -20.369, -19.857, -19.384, -18.896, -18.323, -17.603, -22.780, -22.087,
    -21.493, -20.918, -20.359, -19.847, -19.378, -18.893, -18.324, -17.606,
    -22.790, -22.093, -21.493, -20.912, -20.349, -19.838, -19.371, -18.890,
    -18.324, -17.610, -22.798, -22.096, -21.481, -20.886, -20.317, -19.809,
    -19.349, -18.878, -18.320, -17.611, -22.807, -22.100, -21.475, -20.868,
    -20.295, -19.788, -19.334, -18.870, -18.318, -17.613, -22.817, -22.108,
    -21.481, -20.873, -20.298, -19.791, -19.336, -18.872, -18.321, -17.617,
    -22.826, -22.114, -21.480, -20.864, -20.288, -19.781, -19.330, -18.870,
    -18.322, -17.620, -22.836, -22.123, -21.489, -20.874, -20.297, -19.790,
    -19.337, -18.875, -18.327, -17.625, -22.846, -22.132, -21.499, -20.884,
    -20.307, -19.799, -19.345, -18.882, -18.333, -17.630, -22.855, -22.136,
    -21.492, -20.870, -20.291, -19.786, -19.336, -18.878, -18.332, -17.633,
    -22.866, -22.150, -21.513, -20.896, -20.318, -19.809, -19.354, -18.890,
    -18.341, -17.639, -22.877, -22.163, -21.533, -20.922, -20.344, -19.831,
    -19.371, -18.902, -18.349, -17.645, -22.886, -22.171, -21.536, -20.922,
    -20.342, -19.830, -19.370, -18.903, -18.351, -17.649, -22.897, -22.185,
    -21.557, -20.947, -20.368, -19.852, -19.387, -18.914, -18.359, -17.655,
    -22.907, -22.193, -21.563, -20.951, -20.370, -19.853, -19.388, -18.916,
    -18.362, -17.659, -22.916, -22.195, -21.552, -20.928, -20.343, -19.828,
    -19.370, -18.906, -18.359, -17.661, -22.925, -22.200, -21.546, -20.911,
    -20.321, -19.807, -19.354, -18.896, -18.355, -17.662, -22.934, -22.202,
    -21.535, -20.887, -20.291, -19.779, -19.332, -18.883, -18.351, -17.663,
    -22.941, -22.198, -21.508, -20.840, -20.238, -19.732, -19.296, -18.861,
    -18.341, -17.662, -22.948, -22.193, -21.485, -20.804, -20.201, -19.700,
    -19.273, -18.848, -18.336, -17.663, -22.956, -22.195, -21.479, -20.795,
    -20.192, -19.694, -19.270, -18.847, -18.338, -17.667, -22.965, -22.201,
    -21.484, -20.801, -20.201, -19.703, -19.279, -18.855, -18.344, -17.672,
    -22.976, -22.218, -21.513, -20.841, -20.243, -19.741, -19.309, -18.876,
    -18.357, -17.680, -22.991, -22.247, -21.568, -20.914, -20.318, -19.806,
    -19.358, -18.908, -18.375, -17.690, -23.005, -22.278, -21.625, -20.989,
    -20.393, -19.868, -19.404, -18.938, -18.392, -17.699, -23.019, -22.303,
    -21.670, -21.048, -20.449, -19.915, -19.437, -18.959, -18.404, -17.707,
    -23.030, -22.315, -21.683, -21.061, -20.460, -19.923, -19.443, -18.964,
    -18.409, -17.712, -23.037, -22.312, -21.662, -21.027, -20.425, -19.894,
    -19.423, -18.953, -18.405, -17.713, -23.041, -22.296, -21.614, -20.956,
    -20.352, -19.834, -19.381, -18.928, -18.395, -17.712, -23.044, -22.275,
    -21.554, -20.869, -20.264, -19.759, -19.326, -18.895, -18.380, -17.709,
    -23.050, -22.264, -21.515, -20.811, -20.202, -19.705, -19.285, -18.870,
    -18.369, -17.707, -23.058, -22.265, -21.503, -20.787, -20.174, -19.679,
    -19.265, -18.857, -18.364, -17.708, -23.069, -22.275, -21.510, -20.789,
    -20.173, -19.675, -19.262, -18.855, -18.365, -17.711, -23.081, -22.290,
    -21.529, -20.808, -20.189, -19.688, -19.271, -18.862, -18.370, -17.716,
    -23.094, -22.315, -21.571, -20.861, -20.241, -19.733, -19.305, -18.886,
    -18.384, -17.724, -23.107, -22.336, -21.607, -20.908, -20.289, -19.775,
    -19.338, -18.908, -18.398, -17.733, -23.116, -22.343, -21.614, -20.918,
    -20.301, -19.787, -19.349, -18.917, -18.405, -17.738, -23.128, -22.361,
    -21.644, -20.955, -20.338, -19.819, -19.374, -18.934, -18.416, -17.745,
    -23.139, -22.374, -21.660, -20.973, -20.355, -19.834, -19.385, -18.943,
    -18.422, -17.750, -23.146, -22.371, -21.641, -20.944, -20.325, -19.809,
    -19.368, -18.933, -18.419, -17.752, -23.155, -22.375, -21.638, -20.935,
    -20.316, -19.801, -19.362, -18.930, -18.419, -17.755, -23.163, -22.377,
    -21.630, -20.921, -20.300, -19.787, -19.353, -18.926, -18.419, -17.758,
    -23.169, -22.369, -21.603, -20.882, -20.261, -19.754, -19.329, -18.912,
    -18.413, -17.758, -23.178, -22.374, -21.602, -20.876, -20.254, -19.748,
    -19.324, -18.910, -18.414, -17.761, -23.190, -22.386, -21.613, -20.885,
    -20.260, -19.752, -19.327, -18.913, -18.417, -17.765, -23.197, -22.384,
    -21.596, -20.859, -20.233, -19.729, -19.310, -18.902, -18.414, -17.766,
    -23.207, -22.394, -21.605, -20.867, -20.238, -19.733, -19.313, -18.905,
    -18.417, -17.770, -23.218, -22.403, -21.613, -20.873, -20.243, -19.737,
    -19.316, -18.908, -18.421, -17.774, -23.226, -22.408, -21.614, -20.872,
    -20.242, -19.736, -19.317, -18.910, -18.423, -17.778, -23.241, -22.434,
    -21.656, -20.921, -20.289, -19.775, -19.347, -18.930, -18.436, -17.786,
    -23.252, -22.449, -21.675, -20.942, -20.309, -19.792, -19.360, -18.940,
    -18.442, -17.791, -23.263, -22.460, -21.688, -20.955, -20.321, -19.802,
    -19.368, -18.946, -18.447, -17.796, -23.273, -22.469, -21.696, -20.962,
    -20.326, -19.807, -19.372, -18.949, -18.451, -17.800, -23.281, -22.468,
    -21.682, -20.942, -20.306, -19.790, -19.359, -18.943, -18.449, -17.802,
    -23.290, -22.474, -21.683, -20.938, -20.301, -19.785, -19.356, -18.942,
    -18.451, -17.805, -23.300, -22.481, -21.686, -20.938, -20.299, -19.783,
    -19.355, -18.942, -18.452, -17.809, -23.310, -22.488, -21.689, -20.938,
    -20.298, -19.782, -19.354, -18.942, -18.454, -17.812, -23.319, -22.495,
    -21.692, -20.938, -20.297, -19.780, -19.353, -18.942, -18.456, -17.815,
    -23.331, -22.510, -21.710, -20.956, -20.313, -19.793, -19.363, -18.950,
    -18.461, -17.820, -23.344, -22.530, -21.740, -20.990, -20.344, -19.819,
    -19.382, -18.963, -18.470, -17.826, -23.364, -22.576, -21.822, -21.089,
    -20.436, -19.893, -19.437, -18.998, -18.490, -17.837, -23.374, -22.584,
    -21.826, -21.091, -20.436, -19.894, -19.437, -18.999, -18.492, -17.840,
    -23.384, -22.592, -21.831, -21.093, -20.438, -19.894, -19.438, -19.001,
    -18.495, -17.844, -23.394, -22.600, -21.837, -21.096, -20.439, -19.895,
    -19.439, -19.003, -18.497, -17.847, -23.404, -22.608, -21.842, -21.099,
    -20.441, -19.896, -19.440, -19.004, -18.500, -17.851, -23.414, -22.616,
    -21.848, -21.102, -20.443, -19.898, -19.442, -19.006, -18.503, -17.855,
    -23.424, -22.625, -21.854, -21.106, -20.445, -19.899, -19.443, -19.008,
    -18.505, -17.858, -23.434, -22.633, -21.860, -21.110, -20.448, -19.901,
    -19.445, -19.010, -18.508, -17.862, -23.444, -22.642, -21.866, -21.114,
    -20.451, -19.904, -19.447, -19.012, -18.511, -17.866, -23.454, -22.651,
    -21.873, -21.119, -20.454, -19.906, -19.449, -19.015, -18.514, -17.869,
    -23.464, -22.659, -21.879, -21.124, -20.458, -19.909, -19.452, -19.017,
    -18.517, -17.873, -23.474, -22.668, -21.886, -21.129, -20.461, -19.912,
    -19.454, -19.020, -18.520, -17.877, -23.485, -22.677, -21.894, -21.135,
    -20.466, -19.915, -19.457, -19.022, -18.523, -17.881, -23.495, -22.686,
    -21.901, -21.140, -20.470, -19.919, -19.460, -19.025, -18.526, -17.884,
    -23.505, -22.696, -21.909, -21.146, -20.475, -19.922, -19.463, -19.028,
    -18.529, -17.888, -23.516, -22.705, -21.916, -21.153, -20.480, -19.926,
    -19.466, -19.031, -18.532, -17.892, -23.526, -22.714, -21.924, -21.159,
    -20.485, -19.930, -19.469, -19.034, -18.535, -17.896, -23.536, -22.724,
    -21.932, -21.166, -20.490, -19.935, -19.473, -19.037, -18.539, -17.900,
    -23.547, -22.733, -21.941, -21.173, -20.496, -19.939, -19.477, -19.041,
    -18.542, -17.904, -23.557, -22.743, -21.949, -21.180, -20.502, -19.944,
    -19.481, -19.044, -18.546, -17.907, -23.567, -22.752, -21.958, -21.187,
    -20.508, -19.949, -19.485, -19.048, -18.549, -17.911, -23.578, -22.762,
    -21.967, -21.195, -20.514, -19.954, -19.489, -19.051, -18.553, -17.915,
    -23.588, -22.772, -21.975, -21.203, -20.521, -19.959, -19.493, -19.055,
    -18.556, -17.919, -23.599, -22.781, -21.984, -21.211, -20.527, -19.965,
    -19.498, -19.059, -18.560, -17.923, -23.609, -22.791, -21.994, -21.219,
    -20.534, -19.971, -19.502, -19.063, -18.564, -17.927, -23.620, -22.801,
    -22.003, -21.227, -20.542, -19.977, -19.507, -19.067, -18.567, -17.931,
    -23.630, -22.811, -22.012, -21.236, -20.549, -19.983, -19.512, -19.071,
    -18.571, -17.935, -23.641, -22.821, -22.022, -21.244, -20.556, -19.989,
    -19.517, -19.075, -18.575, -17.939, -23.651, -22.831, -22.031, -21.253,
    -20.564, -19.995, -19.523, -19.080, -18.579, -17.943, -23.662, -22.841,
    -22.041, -21.262, -20.572, -20.002, -19.528, -19.084, -18.583, -17.947,
    -23.672, -22.851, -22.051, -21.271, -20.580, -20.009, -19.533, -19.089,
    -18.587, -17.951, -23.683, -22.862, -22.061, -21.280, -20.588, -20.016,
    -19.539, -19.093, -18.591, -17.956, -23.693, -22.872, -22.071, -21.290,
    -20.596, -20.023, -19.545, -19.098, -18.595, -17.960, -23.704, -22.882,
    -22.081, -21.299, -20.605, -20.030, -19.551, -19.103, -18.600, -17.964,
    -23.714, -22.892, -22.091, -21.309, -20.613, -20.037, -19.557, -19.108,
    -18.604, -17.968, -23.725, -22.902, -22.101, -21.319, -20.622, -20.044,
    -19.563, -19.113, -18.608, -17.972, -23.736, -22.913, -22.111, -21.328,
    -20.631, -20.052, -19.569, -19.118, -18.612, -17.976, -23.746, -22.923,
    -22.122, -21.338, -20.640, -20.060, -19.575, -19.123, -18.617, -17.981,
    -23.757, -22.933, -22.132, -21.348, -20.649, -20.068, -19.582, -19.128,
    -18.621, -17.985, -23.767, -22.944, -22.142, -21.359, -20.658, -20.075,
    -19.588, -19.134, -18.626, -17.989, -23.778, -22.954, -22.153, -21.369,
    -20.668, -20.084, -19.595, -19.139, -18.630, -17.993, -23.789, -22.965,
    -22.164, -21.379, -20.677, -20.092, -19.602, -19.145, -18.635, -17.998,
    -23.799, -22.975, -22.174, -21.390, -20.687, -20.100, -19.609, -19.150,
    -18.639, -18.002, -23.810, -22.986, -22.185, -21.400, -20.696, -20.108,
    -19.616, -19.156, -18.644, -18.006, -23.821, -22.996, -22.195, -21.411,
    -20.706, -20.117, -19.623, -19.162, -18.649, -18.011, -23.831, -23.006,
    -22.206, -21.421, -20.716, -20.125, -19.630, -19.167, -18.654, -18.015,
    -23.842, -23.017, -22.217, -21.432, -20.726, -20.134, -19.637, -19.173,
    -18.658, -18.019, -23.852, -23.027, -22.228, -21.443, -20.736, -20.143,
    -19.644, -19.179, -18.663, -18.024, -23.863, -23.038, -22.239, -21.454,
    -20.746, -20.152, -19.652, -19.185, -18.668, -18.028, -23.874, -23.048,
    -22.249, -21.464, -20.756, -20.161, -19.659, -19.191, -18.673, -18.033,
    -23.884, -23.059, -22.260, -21.475, -20.766, -20.170, -19.667, -19.197,
    -18.678, -18.037, -23.895, -23.069, -22.271, -21.486, -20.777, -20.179,
    -19.674, -19.203, -18.683, -18.041, -23.906, -23.080, -22.282, -21.497,
    -20.787, -20.188, -19.682, -19.209, -18.688, -18.046, -23.916, -23.090,
    -22.293, -21.508, -20.798, -20.197, -19.690, -19.216, -18.693, -18.050,
    -23.927, -23.101, -22.304, -21.520, -20.808, -20.206, -19.698, -19.222,
    -18.698, -18.055, -23.938, -23.111, -22.315, -21.531, -20.819, -20.216,
    -19.705, -19.228, -18.703, -18.059, -23.948, -23.122, -22.326, -21.542,
    -20.829, -20.225, -19.713, -19.235, -18.708, -18.064, -23.959, -23.132,
    -22.337, -21.553, -20.840, -20.235, -19.721, -19.241, -18.713, -18.068,
    -23.970, -23.143, -22.348, -21.564, -20.851, -20.244, -19.730, -19.248,
    -18.718, -18.073, -23.980, -23.154, -22.359, -21.576, -20.861, -20.254,
    -19.738, -19.254, -18.724, -18.077, -23.991, -23.164, -22.370, -21.587,
    -20.872, -20.263, -19.746, -19.261, -18.729, -18.082, -24.002, -23.175,
    -22.381, -21.598, -20.883, -20.273, -19.754, -19.268, -18.734, -18.086,
    -24.012, -23.185, -22.392, -21.610, -20.894, -20.283, -19.762, -19.274,
    -18.739, -18.091, -24.023, -23.196, -22.403, -21.621, -20.905, -20.293,
    -19.771, -19.281, -18.745, -18.096, -24.034, -23.206, -22.414, -21.632,
    -20.916, -20.303, -19.779, -19.288, -18.750, -18.100, -24.044, -23.217,
    -22.425, -21.644, -20.927, -20.312, -19.788, -19.295, -18.756, -18.105,
    -24.055, -23.227, -22.436, -21.655, -20.938, -20.322, -19.796, -19.301,
    -18.761, -18.109, -24.066, -23.238, -22.447, -21.666, -20.949, -20.332,
    -19.805, -19.308, -18.766, -18.114, -24.076, -23.248, -22.458, -21.678,
    -20.960, -20.342, -19.813, -19.315, -18.772, -18.119, -24.087, -23.259,
    -22.469, -21.689, -20.971, -20.352, -19.822, -19.322, -18.777, -18.123,
    -24.098, -23.269, -22.480, -21.701, -20.982, -20.363, -19.830, -19.329,
    -18.783, -18.128, -24.108, -23.279, -22.491, -21.712, -20.993, -20.373,
    -19.839, -19.336, -18.788, -18.132, -24.119, -23.290, -22.502, -21.724,
    -21.004, -20.383, -19.848, -19.343, -18.794, -18.137, -24.130, -23.300,
    -22.513, -21.735, -21.015, -20.393, -19.857, -19.351, -18.799, -18.142,
    -24.140, -23.311, -22.524, -21.746, -21.027, -20.403, -19.865, -19.358,
    -18.805, -18.146, -24.151, -23.321, -22.535, -21.758, -21.038, -20.414,
    -19.874, -19.365, -18.811, -18.151, -24.162, -23.332, -22.545, -21.769,
    -21.049, -20.424, -19.883, -19.372, -18.816, -18.156, -24.172, -23.342,
    -22.556, -21.781, -21.060, -20.434, -19.892, -19.379, -18.822, -18.160,
    -24.183, -23.353, -22.567, -21.792, -21.071, -20.444, -19.901, -19.387,
    -18.828, -18.165, -24.194, -23.363, -22.578, -21.804, -21.083, -20.455,
    -19.910, -19.394, -18.833, -18.170, -24.204, -23.373, -22.589, -21.815,
    -21.094, -20.465, -19.919, -19.401, -18.839, -18.175, -24.215, -23.384,
    -22.600, -21.826, -21.105, -20.476, -19.928, -19.408, -18.845, -18.179,
    -24.226, -23.394, -22.611, -21.838, -21.117, -20.486, -19.937, -19.416,
    -18.850, -18.184, -24.237, -23.405, -22.622, -21.849, -21.128, -20.496,
    -19.946, -19.423, -18.856, -18.189, -24.247, -23.415, -22.632, -21.861,
    -21.139, -20.507, -19.955, -19.431, -18.862, -18.193, -24.258, -23.425,
    -22.643, -21.872, -21.150, -20.517, -19.964, -19.438, -18.868, -18.198,
    -24.269, -23.436, -22.654, -21.883, -21.162, -20.528, -19.973, -19.445,
    -18.873, -18.203, -24.279, -23.446, -22.665, -21.895, -21.173, -20.538,
    -19.983, -19.453, -18.879, -18.207, -24.290, -23.456, -22.676, -21.906,
    -21.184, -20.549, -19.992, -19.460, -18.885, -18.212, -24.301, -23.467,
    -22.686, -21.918, -21.196, -20.559, -20.001, -19.468, -18.891, -18.217,
    -24.311, -23.477, -22.697, -21.929, -21.207, -20.570, -20.010, -19.475,
    -18.896, -18.222, -24.322, -23.487, -22.708, -21.940, -21.218, -20.580,
    -20.019, -19.483, -18.902, -18.226, -24.333, -23.498, -22.719, -21.951,
    -21.229, -20.591, -20.028, -19.490, -18.908, -18.231, -24.343, -23.508,
    -22.729, -21.963, -21.241, -20.601, -20.038, -19.498, -18.914, -18.236,
    -24.354, -23.518, -22.740, -21.974, -21.252, -20.612, -20.047, -19.505,
    -18.920, -18.240, -24.365, -23.529, -22.751, -21.985, -21.263, -20.623,
    -20.056, -19.513, -18.925, -18.245, -24.375, -23.539, -22.761, -21.997,
    -21.274, -20.633, -20.065, -19.521, -18.931, -18.250, -24.386, -23.549,
    -22.772, -22.008, -21.286, -20.644, -20.075, -19.528, -18.937, -18.255,
    -24.397, -23.559, -22.783, -22.019, -21.297, -20.654, -20.084, -19.536,
    -18.943, -18.259, -24.407, -23.570, -22.793, -22.030, -21.308, -20.665,
    -20.093, -19.543, -18.949, -18.264, -24.418, -23.580, -22.804, -22.041,
    -21.319, -20.675, -20.103, -19.551, -18.955, -18.269, -24.429, -23.590,
    -22.814, -22.052, -21.331, -20.686, -20.112, -19.559, -18.961, -18.273,
    -24.439, -23.600, -22.825, -22.064, -21.342, -20.696, -20.121, -19.566,
    -18.966, -18.278, -24.450, -23.611, -22.835, -22.075, -21.353, -20.707,
    -20.131, -19.574, -18.972, -18.283, -24.461, -23.621, -22.846, -22.086,
    -21.364, -20.718, -20.140, -19.581, -18.978, -18.288, -24.471, -23.631,
    -22.856, -22.097, -21.375, -20.728, -20.149, -19.589, -18.984, -18.292,
    -24.482, -23.641, -22.867, -22.108, -21.387, -20.739, -20.159, -19.597,
    -18.990, -18.297, -24.493, -23.651, -22.877, -22.119, -21.398, -20.749,
    -20.168, -19.604, -18.996, -18.302, -24.503, -23.662, -22.888, -22.130,
    -21.409, -20.760, -20.177, -19.612, -19.002, -18.306, -24.514, -23.672,
    -22.898, -22.141, -21.420, -20.770, -20.187, -19.620, -19.007, -18.311,
    -24.525, -23.682, -22.909, -22.152, -21.431, -20.781, -20.196, -19.627,
    -19.013, -18.316, -24.535, -23.692, -22.919, -22.163, -21.442, -20.791,
    -20.205, -19.635, -19.019, -18.320, -24.546, -23.702, -22.929, -22.174,
    -21.453, -20.802, -20.215, -19.642, -19.025, -18.325, -24.557, -23.712,
    -22.940, -22.185, -21.464, -20.813, -20.224, -19.650, -19.031, -18.330,
    -24.567, -23.723, -22.950, -22.196, -21.475, -20.823, -20.233, -19.658,
    -19.037, -18.334, -24.578, -23.733, -22.960, -22.207, -21.486, -20.834,
    -20.243, -19.665, -19.042, -18.339, -24.589, -23.743, -22.971, -22.217,
    -21.497, -20.844, -20.252, -19.673, -19.048, -18.343, -24.600, -23.753,
    -22.981, -22.228, -21.508, -20.855, -20.261, -19.681, -19.054, -18.348,
    -24.610, -23.763, -22.991, -22.239, -21.519, -20.865, -20.271, -19.688,
    -19.060, -18.353, -24.621, -23.773, -23.001, -22.250, -21.530, -20.875,
    -20.280, -19.696, -19.066, -18.357, -24.632, -23.783, -23.012, -22.261,
    -21.541, -20.886, -20.289, -19.703, -19.071, -18.362, -24.642, -23.794,
    -23.022, -22.271, -21.552, -20.896, -20.299, -19.711, -19.077, -18.366,
    -24.653, -23.804, -23.032, -22.282, -21.563, -20.907, -20.308, -19.719,
    -19.083, -18.371, -24.664, -23.814, -23.042, -22.293, -21.574, -20.917,
    -20.317, -19.726, -19.089, -18.375, -24.674, -23.824, -23.052, -22.303,
    -21.585, -20.928, -20.326, -19.734, -19.094, -18.380, -24.685, -23.834,
    -23.063, -22.314, -21.596, -20.938, -20.336, -19.741, -19.100, -18.385,
    -24.696, -23.844, -23.073, -22.325, -21.607, -20.949, -20.345, -19.749,
    -19.106, -18.389, -24.706, -23.854, -23.083, -22.335, -21.618, -20.959,
    -20.354, -19.756, -19.112, -18.394, -24.717, -23.864, -23.093, -22.346,
    -21.628, -20.969, -20.363, -19.764, -19.117, -18.398, -24.728, -23.874,
    -23.103, -22.357, -21.639, -20.980, -20.373, -19.771, -19.123, -18.402,
    -24.738, -23.884, -23.113, -22.367, -21.650, -20.990, -20.382, -19.779,
    -19.129, -18.407, -24.749, -23.894, -23.123, -22.378, -21.661, -21.000,
    -20.391, -19.786, -19.134, -18.411, -24.760, -23.904, -23.133, -22.388,
    -21.671, -21.011, -20.400, -19.794, -19.140, -18.416, -24.770, -23.914,
    -23.143, -22.399, -21.682, -21.021, -20.410, -19.801, -19.145, -18.420,
    -24.781, -23.924, -23.153, -22.409, -21.693, -21.031, -20.419, -19.809,
    -19.151, -18.425, -24.792, -23.934, -23.163, -22.420, -21.704, -21.042,
    -20.428, -19.816, -19.157, -18.429, -24.803, -23.944, -23.173, -22.430,
    -21.714, -21.052, -20.437, -19.824, -19.162, -18.433, -24.813, -23.954,
    -23.183, -22.440, -21.725, -21.062, -20.446, -19.831, -19.168, -18.438,
    -24.824, -23.964, -23.193, -22.451, -21.735, -21.072, -20.455, -19.838,
    -19.173, -18.442, -24.835, -23.974, -23.203, -22.461, -21.746, -21.083,
    -20.464, -19.846, -19.179, -18.446, -24.845, -23.984, -23.213, -22.472,
    -21.757, -21.093, -20.474, -19.853, -19.184, -18.451, -24.856, -23.994,
    -23.223, -22.482, -21.767, -21.103, -20.483, -19.860, -19.190, -18.455,
    -24.867, -24.004, -23.233, -22.492, -21.778, -21.113, -20.492, -19.868,
    -19.195, -18.459, -24.877, -24.014, -23.243, -22.502, -21.788, -21.123,
    -20.501, -19.875, -19.201, -18.464, -24.888, -24.024, -23.253, -22.513,
    -21.799, -21.133, -20.510, -19.882, -19.206, -18.468, -24.899, -24.034,
    -23.263, -22.523, -21.809, -21.144, -20.519, -19.890, -19.211, -18.472,
    -24.910, -24.044, -23.273, -22.533, -21.820, -21.154, -20.528, -19.897,
    -19.217, -18.476, -24.920, -24.054, -23.282, -22.543, -21.830, -21.164,
    -20.537, -19.904, -19.222, -18.480, -24.931, -24.064, -23.292, -22.554,
    -21.841, -21.174, -20.546, -19.911, -19.227, -18.485, -24.942, -24.074,
    -23.302, -22.564, -21.851, -21.184, -20.555, -19.918, -19.233, -18.489,
    -24.952, -24.084, -23.312, -22.574, -21.861, -21.194, -20.564, -19.925,
    -19.238, -18.493, -24.963, -24.094, -23.322, -22.584, -21.872, -21.204,
    -20.572, -19.932, -19.243, -18.497, -24.974, -24.104, -23.331, -22.594,
    -21.882, -21.214, -20.581, -19.940, -19.248, -18.501, -24.984, -24.114,
    -23.341, -22.604, -21.892, -21.224, -20.590, -19.947, -19.253, -18.505,
    -24.995, -24.124, -23.351, -22.614, -21.903, -21.234, -20.599, -19.954,
    -19.259, -18.509, -25.006, -24.134, -23.361, -22.624, -21.913, -21.244,
    -20.608, -19.961, -19.264, -18.513, -25.017, -24.144, -23.370, -22.634,
    -21.923, -21.254, -20.617, -19.968, -19.269, -18.517, -25.027, -24.154,
    -23.380, -22.644, -21.933, -21.264, -20.625, -19.974, -19.274, -18.521,
    -25.038, -24.164, -23.390, -22.654, -21.944, -21.273, -20.634, -19.981,
    -19.279, -18.525, -25.049, -24.174, -23.399, -22.664, -21.954, -21.283,
    -20.643, -19.988, -19.284, -18.529, -25.059, -24.184, -23.409, -22.674,
    -21.964, -21.293, -20.651, -19.995, -19.289, -18.533, -25.070, -24.194,
    -23.419, -22.684, -21.974, -21.303, -20.660, -20.002, -19.294, -18.536,
    -25.081, -24.204, -23.428, -22.694, -21.984, -21.313, -20.669, -20.009,
    -19.298, -18.540, -25.092, -24.213, -23.438, -22.704, -21.994, -21.322,
    -20.677, -20.015, -19.303, -18.544, -25.102, -24.223, -23.448, -22.714,
    -22.004, -21.332, -20.686, -20.022, -19.308, -18.548, -25.113, -24.233,
    -23.457, -22.724, -22.014, -21.342, -20.694, -20.029, -19.313, -18.552,
    -25.124, -24.243, -23.467, -22.734, -22.024, -21.352, -20.703, -20.035,
    -19.318, -18.555, -25.134, -24.253, -23.477, -22.743, -22.035, -21.361,
    -20.711, -20.042, -19.322, -18.559, -25.145, -24.263, -23.486, -22.753,
    -22.045, -21.371, -20.720, -20.048, -19.327, -18.563, -25.156, -24.273,
    -23.496, -22.763, -22.055, -21.381, -20.728, -20.055, -19.332, -18.566,
    -25.167, -24.283, -23.505, -22.773, -22.064, -21.390, -20.737, -20.061,
    -19.336, -18.570, -25.177, -24.293, -23.515, -22.782, -22.074, -21.400,
    -20.745, -20.068, -19.341, -18.574, -25.188, -24.303, -23.524, -22.792,
    -22.084, -21.409, -20.753, -20.074, -19.345, -18.577, -25.199, -24.312,
    -23.534, -22.802, -22.094, -21.419, -20.762, -20.080, -19.350, -18.581,
    -25.209, -24.322, -23.543, -22.812, -22.104, -21.428, -20.770, -20.087,
    -19.354, -18.584, -25.220, -24.332, -23.553, -22.821, -22.114, -21.438,
    -20.778, -20.093, -19.359, -18.588, -25.231, -24.342, -23.563, -22.831,
    -22.124, -21.447, -20.786, -20.099, -19.363, -18.591, -25.242, -24.352,
    -23.572, -22.841, -22.134, -21.457, -20.795, -20.105, -19.367, -18.594,
    -25.252, -24.362, -23.582, -22.850, -22.143, -21.466, -20.803, -20.111,
    -19.372, -18.598, -25.263, -24.372, -23.591, -22.860, -22.153, -21.476,
    -20.811, -20.117, -19.376, -18.601, -25.274, -24.382, -23.600, -22.869,
    -22.163, -21.485, -20.819, -20.123, -19.380, -18.604, -25.285, -24.391,
    -23.610, -22.879, -22.173, -21.494, -20.827, -20.129, -19.384, -18.608,
    -25.295, -24.401, -23.619, -22.888, -22.182, -21.504, -20.835, -20.135,
    -19.388, -18.611, -25.306, -24.411, -23.629, -22.898, -22.192, -21.513,
    -20.843, -20.141, -19.392, -18.614, -25.317, -24.421, -23.638, -22.908,
    -22.202, -21.522, -20.851, -20.147, -19.396, -18.617, -25.327, -24.431,
    -23.648, -22.917, -22.211, -21.531, -20.858, -20.153, -19.400, -18.620,
    -25.338, -24.441, -23.657, -22.927, -22.221, -21.541, -20.866, -20.158,
    -19.404, -18.623, -25.349, -24.451, -23.667, -22.936, -22.231, -21.550,
    -20.874, -20.164, -19.408, -18.626, -25.360, -24.460, -23.676, -22.946,
    -22.240, -21.559, -20.882, -20.170, -19.411, -18.629, -25.370, -24.470,
    -23.685, -22.955, -22.250, -21.568, -20.889, -20.175, -19.415, -18.632,
    -25.381, -24.480, -23.695, -22.964, -22.259, -21.577, -20.897, -20.181,
    -19.419, -18.635, -25.392, -24.490, -23.704, -22.974, -22.269, -21.586,
    -20.905, -20.186, -19.422, -18.638, -25.403, -24.500, -23.713, -22.983,
    -22.278, -21.595, -20.912, -20.191, -19.426, -18.641, -25.413, -24.510,
    -23.723, -22.993, -22.288, -21.604, -20.920, -20.197, -19.430, -18.644,
    -25.424, -24.519, -23.732, -23.002, -22.297, -21.613, -20.927, -20.202,
    -19.433, -18.647, -25.435, -24.529, -23.742, -23.011, -22.307, -21.622,
    -20.934, -20.207, -19.436, -18.650, -25.446, -24.539, -23.751, -23.021,
    -22.316, -21.631, -20.942, -20.212, -19.440, -18.652, -25.456, -24.549,
    -23.760, -23.030, -22.326, -21.640, -20.949, -20.217, -19.443, -18.655,
    -25.467, -24.559, -23.770, -23.039, -22.335, -21.649, -20.956, -20.222,
    -19.446, -18.658, -25.478, -24.569, -23.779, -23.049, -22.344, -21.657,
    -20.963, -20.227, -19.449, -18.660, -25.489, -24.578, -23.788, -23.058,
    -22.354, -21.666, -20.970, -20.232, -19.452, -18.663, -25.499, -24.588,
    -23.797, -23.067, -22.363, -21.675, -20.977, -20.237, -19.456, -18.665,
    -25.510, -24.598, -23.807, -23.076, -22.372, -21.684, -20.984, -20.241,
    -19.459, -18.668, -25.521, -24.608, -23.816, -23.086, -22.381, -21.692,
    -20.991, -20.246, -19.461, -18.670, -25.532, -24.618, -23.825, -23.095,
    -22.391, -21.701, -20.998, -20.250, -19.464, -18.673, -25.542, -24.628,
    -23.835, -23.104, -22.400, -21.709, -21.005, -20.255, -19.467, -18.675,
    -25.553, -24.637, -23.844, -23.113, -22.409, -21.718, -21.012, -20.259,
    -19.470, -18.677, -25.564, -24.647, -23.853, -23.122, -22.418, -21.726,
    -21.018, -20.263, -19.472, -18.679, -25.575, -24.657, -23.862, -23.132,
    -22.427, -21.735, -21.025, -20.268, -19.475, -18.682, -25.585, -24.667,
    -23.872, -23.141, -22.437, -21.743, -21.031, -20.272, -19.478, -18.684,
    -25.596, -24.677, -23.881, -23.150, -22.446, -21.752, -21.038, -20.276,
    -19.480, -18.686, -25.607, -24.686, -23.890, -23.159, -22.455, -21.760,
    -21.044, -20.280, -19.483, -18.688, -25.618, -24.696, -23.899, -23.168,
    -22.464, -21.768, -21.051, -20.284, -19.485, -18.690, -25.628, -24.706,
    -23.909, -23.177, -22.473, -21.776, -21.057, -20.288, -19.487, -18.692,
    -25.639, -24.716, -23.918, -23.186, -22.482, -21.785, -21.063, -20.291,
    -19.489, -18.694, -25.650, -24.726, -23.927, -23.195, -22.491, -21.793,
    -21.069, -20.295, -19.491, -18.696, -25.661, -24.735, -23.936, -23.204,
    -22.500, -21.801, -21.075, -20.298, -19.494, -18.698, -25.671, -24.745,
    -23.945, -23.214, -22.509, -21.809, -21.081, -20.302, -19.496, -18.700,
    -25.682, -24.755, -23.955, -23.223, -22.518, -21.817, -21.087, -20.305,
    -19.497, -18.702, -25.693, -24.765, -23.964, -23.232, -22.527, -21.825,
    -21.093, -20.309, -19.499, -18.703, -25.704, -24.775, -23.973, -23.241,
    -22.535, -21.833, -21.098, -20.312, -19.501, -18.705, -25.715, -24.784,
    -23.982, -23.250, -22.544, -21.841, -21.104, -20.315, -19.503, -18.707,
    -25.725, -24.794, -23.991, -23.259, -22.553, -21.848, -21.109, -20.318,
    -19.504, -18.708, -25.736, -24.804, -24.000, -23.268, -22.562, -21.856,
    -21.115, -20.321, -19.506, -18.710, -25.747, -24.814, -24.010, -23.277,
    -22.571, -21.864, -21.120, -20.324, -19.508, -18.711, -25.758, -24.824,
    -24.019, -23.285, -22.579, -21.871, -21.125, -20.326, -19.509, -18.713,
    -25.768, -24.833, -24.028, -23.294, -22.588, -21.879, -21.131, -20.329,
    -19.510, -18.714, -25.779, -24.843, -24.037, -23.303, -22.597, -21.887,
    -21.136, -20.332, -19.512, -18.715, -25.790, -24.853, -24.046, -23.312,
    -22.605, -21.894, -21.141, -20.334, -19.513, -18.717, -25.801, -24.863,
    -24.055, -23.321, -22.614, -21.901, -21.145, -20.336, -19.514, -18.718,
    -25.811, -24.873, -24.065, -23.330, -22.623, -21.909, -21.150, -20.339,
    -19.515, -18.719, -25.822, -24.882, -24.074, -23.339, -22.631, -21.916,
    -21.155, -20.341, -19.516, -18.720, -25.833, -24.892, -24.083, -23.348,
    -22.640, -21.923, -21.159, -20.343, -19.517, -18.722, -25.844, -24.902,
    -24.092, -23.357, -22.648, -21.930, -21.164, -20.345, -19.518, -18.723,
    -25.855, -24.912, -24.101, -23.365, -22.657, -21.937, -21.168, -20.346,
    -19.518, -18.724, -25.865, -24.922, -24.110, -23.374, -22.665, -21.945,
    -21.173, -20.348, -19.519, -18.725, -25.876, -24.931, -24.119, -23.383,
    -22.674, -21.951, -21.177, -20.350, -19.520, -18.726, -25.887, -24.941,
    -24.128, -23.392, -22.682, -21.958, -21.181, -20.351, -19.520, -18.727,
    -25.898, -24.951, -24.137, -23.401, -22.690, -21.965, -21.185, -20.353,
    -19.521, -18.727, -25.909, -24.961, -24.146, -23.409, -22.699, -21.972,
    -21.189, -20.354, -19.521, -18.728, -25.919, -24.970, -24.156, -23.418,
    -22.707, -21.979, -21.192, -20.355, -19.521, -18.729, -25.930, -24.980,
    -24.165, -23.427, -22.715, -21.985, -21.196, -20.356, -19.521, -18.730,
    -25.941, -24.990, -24.174, -23.436, -22.723, -21.992, -21.199, -20.357,
    -19.522, -18.730, -25.952, -25.000, -24.183, -23.444, -22.732, -21.998,
    -21.203, -20.358, -19.522, -18.731, -25.962, -25.010, -24.192, -23.453,
    -22.740, -22.004, -21.206, -20.359, -19.522, -18.732, -25.973, -25.019,
    -24.201, -23.462, -22.748, -22.011, -21.209, -20.359, -19.521, -18.732,
    -25.984, -25.029, -24.210, -23.470, -22.756, -22.017, -21.212, -20.360,
    -19.521, -18.733, -25.995, -25.039, -24.219, -23.479, -22.764, -22.023,
    -21.215, -20.360, -19.521, -18.733, -26.006, -25.049, -24.228, -23.488,
    -22.772, -22.029, -21.218, -20.360, -19.521, -18.733, -26.016, -25.058,
    -24.237, -23.496, -22.780, -22.035, -21.220, -20.361, -19.520, -18.734,
    -26.027, -25.068, -24.246, -23.505, -22.788, -22.041, -21.223, -20.361,
    -19.520, -18.734, -26.038, -25.078, -24.255, -23.514, -22.796, -22.046,
    -21.225, -20.361, -19.519, -18.734, -26.049, -25.088, -24.264, -23.522,
    -22.804, -22.052, -21.227, -20.360, -19.519, -18.734, -26.060, -25.098,
    -24.273, -23.531, -22.812, -22.057, -21.229, -20.360, -19.518, -18.735,
    -26.070, -25.107, -24.282, -23.539, -22.819, -22.063, -21.231, -20.360,
    -19.517, -18.735, -26.081, -25.117, -24.291, -23.548, -22.827, -22.068,
    -21.233, -20.359, -19.516, -18.735, -26.092, -25.127, -24.300, -23.556,
    -22.835, -22.073, -21.235, -20.359, -19.515, -18.735, -26.103, -25.137,
    -24.309, -23.565, -22.842, -22.078, -21.236, -20.358, -19.514, -18.735,
    -26.114, -25.146, -24.318, -23.574, -22.850, -22.083, -21.238, -20.357,
    -19.513, -18.735, -26.124, -25.156, -24.327, -23.582, -22.858, -22.088,
    -21.239, -20.356, -19.512, -18.735, -26.135, -25.166, -24.336, -23.591,
    -22.865, -22.093, -21.240, -20.355, -19.511, -18.735, -26.146, -25.176,
    -24.345, -23.599, -22.873, -22.098, -21.241, -20.354, -19.510, -18.734,
    -26.157, -25.185, -24.354, -23.607, -22.880, -22.102, -21.242, -20.352,
    -19.508, -18.734, -26.168, -25.195, -24.363, -23.616, -22.887, -22.107,
    -21.243, -20.351, -19.507, -18.734, -26.178, -25.205, -24.372, -23.624,
    -22.895, -22.111, -21.243, -20.350, -19.505, -18.734, -26.189, -25.215,
    -24.381, -23.633, -22.902, -22.115, -21.244, -20.348, -19.504, -18.733,
    -26.200, -25.225, -24.390, -23.641, -22.909, -22.119, -21.244, -20.346,
    -19.502, -18.733, -26.211, -25.234, -24.399, -23.650, -22.916, -22.123,
    -21.244, -20.344, -19.500, -18.732, -26.222, -25.244, -24.408, -23.658,
    -22.923, -22.127, -21.244, -20.342, -19.499, -18.732, -26.232, -25.254,
    -24.417, -23.666, -22.930, -22.131, -21.244, -20.340, -19.497, -18.731,
    -26.243, -25.264, -24.426, -23.675, -22.937, -22.135, -21.244, -20.338,
    -19.495, -18.731, -26.254, -25.273, -24.435, -23.683, -22.944, -22.138,
    -21.243, -20.336, -19.493, -18.730, -26.265, -25.283, -24.444, -23.691,
    -22.951, -22.141, -21.243, -20.333, -19.491, -18.730, -26.276, -25.293,
    -24.453, -23.699, -22.958, -22.145, -21.242, -20.331, -19.489, -18.729,
    -26.286, -25.303, -24.462, -23.708, -22.964, -22.148, -21.241, -20.328,
    -19.487, -18.728, -26.297, -25.312, -24.471, -23.716, -22.971, -22.150,
    -21.240, -20.326, -19.484, -18.728, -26.308, -25.322, -24.480, -23.724,
    -22.977, -22.153, -21.239, -20.323, -19.482, -18.727, -26.319, -25.332,
    -24.489, -23.732, -22.984, -22.156, -21.237, -20.320, -19.480, -18.726,
    -26.330, -25.342, -24.497, -23.741, -22.990, -22.158, -21.236, -20.317,
    -19.477, -18.725, -26.340, -25.351, -24.506, -23.749, -22.997, -22.161,
    -21.234, -20.314, -19.475, -18.725, -26.351, -25.361, -24.515, -23.757,
    -23.003, -22.163, -21.233, -20.311, -19.472, -18.724, -26.362, -25.371,
    -24.524, -23.765, -23.009, -22.165, -21.231, -20.307, -19.470, -18.723,
    -26.373, -25.381, -24.533, -23.773, -23.015, -22.167, -21.229, -20.304,
    -19.467, -18.722, -26.384, -25.391, -24.542, -23.781, -23.021, -22.169,
    -21.226, -20.301, -19.465, -18.721, -26.395, -25.400, -24.551, -23.789,
    -23.027, -22.170, -21.224, -20.297, -19.462, -18.720, -26.405, -25.410,
    -24.560, -23.797, -23.033, -22.172, -21.222, -20.293, -19.459, -18.719,
    -26.416, -25.420, -24.569, -23.805, -23.039, -22.173, -21.219, -20.290,
    -19.457, -18.718, -26.427, -25.430, -24.578, -23.813, -23.044, -22.174,
    -21.216, -20.286, -19.454, -18.717, -26.438, -25.439, -24.587, -23.821,
    -23.050, -22.175, -21.214, -20.282, -19.451, -18.716, -26.449, -25.449,
    -24.595, -23.829, -23.055, -22.176, -21.211, -20.278, -19.448, -18.715,
    -26.459, -25.459, -24.604, -23.837, -23.061, -22.177, -21.207, -20.274,
    -19.445, -18.714, -26.470, -25.469, -24.613, -23.845, -23.066, -22.177,
    -21.204, -20.270, -19.442, -18.713, -26.481, -25.478, -24.622, -23.853,
    -23.071, -22.178, -21.201, -20.266, -19.439, -18.712, -26.492, -25.488,
    -24.631, -23.861, -23.076, -22.178, -21.197, -20.262, -19.436, -18.710,
    -26.503, -25.498, -24.640, -23.868, -23.082, -22.178, -21.194, -20.258,
    -19.433, -18.709, -26.514, -25.508, -24.649, -23.876, -23.086, -22.178,
    -21.190, -20.253, -19.430, -18.708, -26.524, -25.517, -24.657, -23.884,
    -23.091, -22.178, -21.186, -20.249, -19.427, -18.707, -26.535, -25.527,
    -24.666, -23.892, -23.096, -22.177, -21.182, -20.244, -19.424, -18.706,
    -26.546, -25.537, -24.675, -23.899, -23.101, -22.177, -21.178, -20.240,
    -19.421, -18.704, -26.557, -25.547, -24.684, -23.907, -23.105, -22.176,
    -21.174, -20.235, -19.417, -18.703, -26.568, -25.556, -24.693, -23.915,
    -23.110, -22.175, -21.170, -20.231, -19.414, -18.702, -26.579, -25.566,
    -24.702, -23.922, -23.114, -22.174, -21.166, -20.226, -19.411, -18.701,
    -26.589, -25.576, -24.710, -23.930, -23.118, -22.173, -21.161, -20.222,
    -19.408, -18.700, -26.600, -25.586, -24.719, -23.937, -23.122, -22.172,
    -21.157, -20.217, -19.405, -18.698, -26.611, -25.595, -24.728, -23.945,
    -23.126, -22.171, -21.152, -20.212, -19.401, -18.697, -26.622, -25.605,
    -24.737, -23.952, -23.130, -22.169, -21.148, -20.207, -19.398, -18.696,
    -26.633, -25.615, -24.746, -23.960, -23.134, -22.167, -21.143, -20.203,
    -19.395, -18.695, -26.644, -25.625, -24.754, -23.967, -23.137, -22.166,
    -21.138, -20.198, -19.392, -18.693, -26.654, -25.634, -24.763, -23.974,
    -23.141, -22.164, -21.133, -20.193, -19.388, -18.692, -26.665, -25.644,
    -24.772, -23.982, -23.144, -22.162, -21.128, -20.188, -19.385, -18.691,
};

const double CIA_Orton::H2CH4_n[][10] = {
    -26.500, -27.061, -27.542, -27.974, -28.371, -28.741, -29.092, -29.434,
    -29.775, -30.115, -24.667, -25.228, -25.709, -26.141, -26.538, -26.908,
    -27.259, -27.601, -27.942, -28.282, -23.280, -23.842, -24.323, -24.754,
    -25.152, -25.522, -25.873, -26.215, -26.556, -26.896, -21.893, -22.454,
    -22.936, -23.368, -23.765, -24.135, -24.486, -24.829, -25.170, -25.510,
    -21.076, -21.639, -22.122, -22.554, -22.952, -23.323, -23.674, -24.017,
    -24.358, -24.698, -20.490, -21.055, -21.540, -21.974, -22.373, -22.745,
    -23.097, -23.440, -23.782, -24.122, -20.034, -20.602, -21.088, -21.524,
    -21.924, -22.296, -22.649, -22.993, -23.335, -23.675, -19.659, -20.229,
    -20.717, -21.154, -21.556, -21.929, -22.283, -22.627, -22.969, -23.310,
    -19.344, -19.915, -20.404, -20.843, -21.245, -21.619, -21.973, -22.317,
    -22.660, -23.001, -19.070, -19.642, -20.133, -20.573, -20.976, -21.350,
    -21.705, -22.049, -22.392, -22.734, -18.818, -19.393, -19.887, -20.329,
    -20.734, -21.110, -21.466, -21.812, -22.155, -22.497, -18.590, -19.169,
    -19.666, -20.111, -20.518, -20.895, -21.252, -21.599, -21.942, -22.285,
    -18.385, -18.966, -19.466, -19.913, -20.322, -20.701, -21.058, -21.406,
    -21.750, -22.093, -18.195, -18.779, -19.282, -19.731, -20.142, -20.522,
    -20.881, -21.229, -21.574, -21.918, -18.017, -18.605, -19.111, -19.563,
    -19.976, -20.358, -20.717, -21.067, -21.412, -21.756, -17.848, -18.439,
    -18.949, -19.404, -19.819, -20.203, -20.565, -20.915, -21.262, -21.606,
    -17.686, -18.281, -18.794, -19.253, -19.672, -20.058, -20.422, -20.773,
    -21.121, -21.466, -17.535, -18.134, -18.651, -19.113, -19.534, -19.923,
    -20.288, -20.641, -20.989, -21.335, -17.393, -17.995, -18.515, -18.981,
    -19.404, -19.795, -20.162, -20.516, -20.865, -21.212, -17.257, -17.862,
    -18.387, -18.855, -19.281, -19.674, -20.042, -20.398, -20.748, -21.096,
    -17.130, -17.738, -18.265, -18.736, -19.165, -19.560, -19.930, -20.286,
    -20.638, -20.986, -17.006, -17.617, -18.148, -18.622, -19.054, -19.450,
    -19.822, -20.180, -20.532, -20.881, -16.446, -17.072, -17.619, -18.110,
    -18.556, -18.964, -19.345, -19.711, -20.069, -20.422, -15.987, -16.622,
    -17.181, -17.685, -18.143, -18.562, -18.951, -19.324, -19.688, -20.045,
    -15.577, -16.218, -16.790, -17.306, -17.776, -18.207, -18.606, -18.987,
    -19.358, -19.721, -15.242, -15.883, -16.459, -16.982, -17.461, -17.900,
    -18.307, -18.695, -19.071, -19.439, -14.944, -15.584, -16.163, -16.692,
    -17.178, -17.625, -18.040, -18.434, -18.816, -19.189, -14.708, -15.343,
    -15.920, -16.451, -16.939, -17.390, -17.808, -18.206, -18.592, -18.968,
    -14.495, -15.125, -15.700, -16.231, -16.722, -17.176, -17.598, -18.000,
    -18.389, -18.768, -14.330, -14.951, -15.521, -16.049, -16.539, -16.992,
    -17.415, -17.818, -18.209, -18.589, -14.164, -14.780, -15.347, -15.874,
    -16.364, -16.819, -17.244, -17.648, -18.041, -18.422, -14.026, -14.637,
    -15.200, -15.725, -16.213, -16.668, -17.092, -17.497, -17.890, -18.272,
    -13.934, -14.535, -15.090, -15.608, -16.092, -16.542, -16.964, -17.366,
    -17.757, -18.138, -13.869, -14.462, -15.008, -15.518, -15.994, -16.438,
    -16.855, -17.253, -17.640, -18.018, -13.820, -14.406, -14.944, -15.445,
    -15.913, -16.350, -16.760, -17.153, -17.535, -17.909, -13.778, -14.356,
    -14.886, -15.379, -15.839, -16.268, -16.672, -17.059, -17.437, -17.807,
    -13.715, -14.293, -14.818, -15.306, -15.761, -16.185, -16.584, -16.967,
    -17.341, -17.708, -13.634, -14.215, -14.741, -15.227, -15.680, -16.101,
    -16.496, -16.876, -17.248, -17.613, -13.545, -14.120, -14.636, -15.111,
    -15.552, -15.962, -16.348, -16.719, -17.083, -17.442, -13.440, -14.012,
    -14.522, -14.989, -15.422, -15.825, -16.203, -16.568, -16.927, -17.281,
    -13.237, -13.817, -14.335, -14.809, -15.248, -15.655, -16.036, -16.403,
    -16.764, -17.119, -13.086, -13.665, -14.183, -14.659, -15.099, -15.506,
    -15.889, -16.256, -16.616, -16.971, -13.005, -13.573, -14.082, -14.550,
    -14.984, -15.387, -15.765, -16.129, -16.487, -16.840, -12.822, -13.395,
    -13.912, -14.388, -14.829, -15.238, -15.621, -15.989, -16.349, -16.704,
    -12.699, -13.268, -13.783, -14.260, -14.702, -15.112, -15.496, -15.865,
    -16.226, -16.581, -12.631, -13.185, -13.690, -14.160, -14.597, -15.004,
    -15.386, -15.754, -16.114, -16.468, -12.481, -13.034, -13.543, -14.017,
    -14.460, -14.873, -15.260, -15.631, -15.994, -16.352, -12.421, -12.963,
    -13.464, -13.932, -14.371, -14.780, -15.165, -15.534, -15.896, -16.252,
    -12.428, -12.950, -13.433, -13.888, -14.316, -14.716, -15.093, -15.456,
    -15.813, -16.165, -12.418, -12.927, -13.398, -13.843, -14.261, -14.653,
    -15.024, -15.381, -15.734, -16.083, -12.513, -12.995, -13.440, -13.860,
    -14.258, -14.632, -14.988, -15.333, -15.676, -16.017, -12.605, -13.059,
    -13.476, -13.873, -14.250, -14.608, -14.950, -15.284, -15.618, -15.951,
    -12.642, -13.077, -13.476, -13.857, -14.221, -14.568, -14.901, -15.228,
    -15.555, -15.884, -12.683, -13.097, -13.478, -13.843, -14.194, -14.530,
    -14.853, -15.173, -15.494, -15.818, -12.624, -13.035, -13.415, -13.778,
    -14.128, -14.463, -14.785, -15.103, -15.423, -15.746, -12.589, -12.992,
    -13.366, -13.726, -14.072, -14.404, -14.724, -15.039, -15.357, -15.679,
    -12.554, -12.952, -13.321, -13.677, -14.020, -14.348, -14.665, -14.978,
    -15.294, -15.614, -12.522, -12.914, -13.278, -13.630, -13.969, -14.295,
    -14.609, -14.920, -15.233, -15.551, -12.493, -12.880, -13.240, -13.588,
    -13.924, -14.245, -14.556, -14.864, -15.176, -15.491, -12.452, -12.838,
    -13.197, -13.543, -13.876, -14.195, -14.504, -14.810, -15.119, -15.433,
    -12.409, -12.795, -13.154, -13.499, -13.830, -14.147, -14.453, -14.757,
    -15.065, -15.377, -12.363, -12.751, -13.111, -13.455, -13.785, -14.101,
    -14.405, -14.707, -15.012, -15.323, -12.322, -12.711, -13.072, -13.415,
    -13.744, -14.057, -14.359, -14.659, -14.962, -15.270, -12.273, -12.666,
    -13.029, -13.374, -13.702, -14.013, -14.314, -14.612, -14.913, -15.220,
    -12.238, -12.633, -12.995, -13.338, -13.664, -13.973, -14.271, -14.567,
    -14.866, -15.171, -12.215, -12.607, -12.967, -13.307, -13.629, -13.936,
    -14.231, -14.524, -14.821, -15.124, -12.191, -12.581, -12.939, -13.276,
    -13.596, -13.900, -14.193, -14.483, -14.778, -15.079, -12.174, -12.561,
    -12.915, -13.249, -13.565, -13.866, -14.156, -14.444, -14.736, -15.035,
    -12.150, -12.536, -12.889, -13.221, -13.535, -13.833, -14.120, -14.405,
    -14.696, -14.992, -12.124, -12.510, -12.862, -13.192, -13.505, -13.801,
    -14.086, -14.369, -14.657, -14.951, -12.103, -12.487, -12.838, -13.166,
    -13.476, -13.770, -14.053, -14.333, -14.619, -14.911, -12.076, -12.458,
    -12.808, -13.136, -13.446, -13.738, -14.019, -14.298, -14.582, -14.872,
    -12.043, -12.425, -12.777, -13.106, -13.416, -13.707, -13.987, -14.264,
    -14.546, -14.835, -12.013, -12.394, -12.747, -13.076, -13.386, -13.677,
    -13.956, -14.231, -14.511, -14.798, -11.982, -12.364, -12.718, -13.049,
    -13.358, -13.649, -13.926, -14.200, -14.478, -14.762, -11.960, -12.341,
    -12.694, -13.025, -13.333, -13.623, -13.898, -14.169, -14.445, -14.728,
    -11.953, -12.328, -12.678, -13.006, -13.312, -13.599, -13.872, -14.141,
    -14.414, -14.695, -11.945, -12.315, -12.662, -12.987, -13.291, -13.575,
    -13.846, -14.113, -14.384, -14.663, -11.944, -12.307, -12.650, -12.971,
    -13.272, -13.554, -13.822, -14.086, -14.355, -14.631, -11.928, -12.289,
    -12.630, -12.951, -13.251, -13.532, -13.798, -14.060, -14.327, -14.601,
    -11.914, -12.273, -12.613, -12.932, -13.231, -13.510, -13.775, -14.035,
    -14.299, -14.571, -11.901, -12.257, -12.595, -12.913, -13.211, -13.489,
    -13.752, -14.010, -14.273, -14.542, -11.884, -12.239, -12.576, -12.894,
    -13.191, -13.468, -13.729, -13.986, -14.246, -14.514, -11.872, -12.224,
    -12.559, -12.876, -13.172, -13.447, -13.707, -13.962, -14.221, -14.486,
    -11.855, -12.208, -12.543, -12.859, -13.153, -13.428, -13.686, -13.939,
    -14.196, -14.459, -11.853, -12.201, -12.533, -12.846, -13.139, -13.410,
    -13.667, -13.918, -14.172, -14.433, -11.869, -12.206, -12.530, -12.838,
    -13.127, -13.395, -13.649, -13.897, -14.149, -14.408, -11.888, -12.213,
    -12.530, -12.832, -13.116, -13.381, -13.632, -13.878, -14.127, -14.384,
    -11.897, -12.212, -12.524, -12.823, -13.104, -13.367, -13.615, -13.859,
    -14.106, -14.360, -11.902, -12.208, -12.516, -12.812, -13.091, -13.352,
    -13.598, -13.840, -14.085, -14.337, -11.888, -12.195, -12.502, -12.798,
    -13.076, -13.336, -13.581, -13.821, -14.064, -14.314, -11.879, -12.185,
    -12.491, -12.785, -13.063, -13.322, -13.565, -13.803, -14.044, -14.292,
    -11.849, -12.159, -12.469, -12.765, -13.045, -13.304, -13.547, -13.784,
    -14.023, -14.270, -11.816, -12.133, -12.447, -12.746, -13.027, -13.287,
    -13.530, -13.766, -14.004, -14.248, -11.786, -12.112, -12.429, -12.730,
    -13.011, -13.271, -13.513, -13.748, -13.984, -14.227, -11.755, -12.092,
    -12.414, -12.717, -12.998, -13.257, -13.498, -13.732, -13.966, -14.207,
    -11.738, -12.081, -12.405, -12.707, -12.987, -13.245, -13.485, -13.716,
    -13.949, -14.187, -11.751, -12.093, -12.412, -12.709, -12.984, -13.237,
    -13.474, -13.702, -13.932, -14.168, -11.741, -12.088, -12.409, -12.704,
    -12.977, -13.228, -13.462, -13.688, -13.916, -14.150, -11.752, -12.099,
    -12.416, -12.706, -12.974, -13.222, -13.452, -13.675, -13.900, -14.132,
    -11.724, -12.079, -12.400, -12.693, -12.962, -13.209, -13.439, -13.661,
    -13.884, -14.114, -11.702, -12.062, -12.386, -12.680, -12.950, -13.197,
    -13.426, -13.647, -13.868, -14.096, -11.682, -12.046, -12.373, -12.668,
    -12.938, -13.185, -13.413, -13.633, -13.853, -14.079, -11.661, -12.029,
    -12.357, -12.654, -12.925, -13.172, -13.400, -13.618, -13.837, -14.062,
    -11.641, -12.010, -12.341, -12.639, -12.911, -13.158, -13.386, -13.604,
    -13.822, -14.045, -11.625, -11.993, -12.325, -12.625, -12.897, -13.145,
    -13.373, -13.590, -13.807, -14.028, -11.616, -11.982, -12.313, -12.613,
    -12.885, -13.133, -13.361, -13.577, -13.792, -14.012, -11.619, -11.980,
    -12.308, -12.606, -12.877, -13.124, -13.350, -13.565, -13.779, -13.997,
    -11.628, -11.983, -12.308, -12.603, -12.872, -13.117, -13.341, -13.554,
    -13.766, -13.982, -11.623, -11.978, -12.302, -12.596, -12.864, -13.108,
    -13.331, -13.543, -13.753, -13.967, -11.613, -11.972, -12.297, -12.590,
    -12.857, -13.100, -13.322, -13.532, -13.740, -13.953, -11.602, -11.962,
    -12.288, -12.581, -12.848, -13.090, -13.311, -13.521, -13.728, -13.938,
    -11.597, -11.957, -12.282, -12.574, -12.840, -13.081, -13.301, -13.509,
    -13.715, -13.924, -11.585, -11.945, -12.269, -12.562, -12.827, -13.068,
    -13.289, -13.497, -13.702, -13.910, -11.575, -11.936, -12.260, -12.552,
    -12.816, -13.057, -13.278, -13.485, -13.689, -13.896, -11.580, -11.935,
    -12.256, -12.544, -12.807, -13.048, -13.267, -13.474, -13.677, -13.882,
    -11.592, -11.943, -12.258, -12.543, -12.803, -13.041, -13.259, -13.464,
    -13.665, -13.869, -11.615, -11.958, -12.267, -12.545, -12.801, -13.036,
    -13.251, -13.455, -13.655, -13.857, -11.685, -12.000, -12.290, -12.557,
    -12.805, -13.034, -13.246, -13.447, -13.645, -13.845, -11.737, -12.030,
    -12.306, -12.563, -12.805, -13.031, -13.240, -13.439, -13.635, -13.833,
    -11.814, -12.071, -12.327, -12.573, -12.807, -13.028, -13.235, -13.431,
    -13.625, -13.821, -11.806, -12.064, -12.320, -12.565, -12.799, -13.020,
    -13.226, -13.422, -13.614, -13.809, -11.796, -12.055, -12.312, -12.557,
    -12.791, -13.011, -13.217, -13.412, -13.604, -13.798, -11.770, -12.035,
    -12.297, -12.545, -12.781, -13.002, -13.208, -13.403, -13.593, -13.786,
    -11.719, -12.000, -12.272, -12.527, -12.767, -12.991, -13.197, -13.392,
    -13.583, -13.774, -11.667, -11.962, -12.244, -12.507, -12.752, -12.978,
    -13.186, -13.382, -13.572, -13.762, -11.615, -11.925, -12.217, -12.487,
    -12.736, -12.965, -13.175, -13.371, -13.561, -13.751, -11.576, -11.897,
    -12.196, -12.470, -12.723, -12.954, -13.165, -13.362, -13.551, -13.740,
    -11.571, -11.891, -12.189, -12.463, -12.716, -12.947, -13.158, -13.354,
    -13.542, -13.729, -11.581, -11.895, -12.189, -12.461, -12.712, -12.943,
    -13.152, -13.347, -13.534, -13.719, -11.597, -11.903, -12.192, -12.461,
    -12.710, -12.939, -13.147, -13.340, -13.525, -13.710, -11.614, -11.913,
    -12.197, -12.462, -12.709, -12.935, -13.142, -13.333, -13.517, -13.700,
    -11.599, -11.901, -12.187, -12.454, -12.701, -12.928, -13.134, -13.325,
    -13.508, -13.690, -11.580, -11.888, -12.179, -12.448, -12.695, -12.922,
    -13.127, -13.317, -13.500, -13.680, -11.555, -11.871, -12.166, -12.437,
    -12.686, -12.912, -13.118, -13.308, -13.490, -13.670, -11.524, -11.852,
    -12.152, -12.427, -12.677, -12.904, -13.110, -13.300, -13.481, -13.660,
    -11.510, -11.841, -12.144, -12.419, -12.670, -12.897, -13.102, -13.292,
    -13.472, -13.650, -11.507, -11.839, -12.142, -12.417, -12.666, -12.892,
    -13.097, -13.285, -13.464, -13.641, -11.515, -11.846, -12.147, -12.420,
    -12.666, -12.890, -13.092, -13.279, -13.457, -13.632, -11.531, -11.858,
    -12.157, -12.426, -12.669, -12.890, -13.089, -13.274, -13.450, -13.624,
    -11.549, -11.873, -12.167, -12.433, -12.672, -12.889, -13.086, -13.269,
    -13.443, -13.615, -11.554, -11.879, -12.172, -12.436, -12.673, -12.888,
    -13.083, -13.264, -13.437, -13.607, -11.543, -11.873, -12.169, -12.432,
    -12.669, -12.883, -13.077, -13.257, -13.429, -13.598, -11.547, -11.875,
    -12.168, -12.429, -12.665, -12.877, -13.071, -13.250, -13.421, -13.590,
    -11.549, -11.874, -12.165, -12.425, -12.659, -12.871, -13.064, -13.243,
    -13.414, -13.581, -11.524, -11.853, -12.147, -12.410, -12.647, -12.861,
    -13.056, -13.235, -13.405, -13.572, -11.522, -11.847, -12.138, -12.400,
    -12.637, -12.852, -13.047, -13.227, -13.397, -13.563, -11.514, -11.837,
    -12.127, -12.388, -12.626, -12.843, -13.038, -13.218, -13.389, -13.555,
    -11.508, -11.830, -12.118, -12.379, -12.617, -12.834, -13.030, -13.211,
    -13.381, -13.546, -11.537, -11.849, -12.128, -12.382, -12.616, -12.830,
    -13.025, -13.205, -13.374, -13.538, -11.589, -11.881, -12.146, -12.390,
    -12.618, -12.829, -13.022, -13.200, -13.368, -13.531, -11.611, -11.894,
    -12.151, -12.390, -12.615, -12.824, -13.016, -13.194, -13.361, -13.523,
    -11.656, -11.921, -12.165, -12.396, -12.616, -12.822, -13.013, -13.189,
    -13.355, -13.516, -11.655, -11.916, -12.159, -12.389, -12.609, -12.816,
    -13.006, -13.182, -13.348, -13.508, -11.646, -11.909, -12.152, -12.383,
    -12.603, -12.810, -13.000, -13.176, -13.341, -13.500, -11.635, -11.892,
    -12.134, -12.366, -12.588, -12.798, -12.990, -13.166, -13.332, -13.492,
    -11.621, -11.878, -12.120, -12.354, -12.578, -12.789, -12.982, -13.159,
    -13.325, -13.484, -11.583, -11.848, -12.098, -12.336, -12.565, -12.778,
    -12.973, -13.151, -13.317, -13.476, -11.559, -11.834, -12.089, -12.331,
    -12.561, -12.774, -12.969, -13.146, -13.311, -13.468, -11.538, -11.821,
    -12.083, -12.328, -12.558, -12.772, -12.965, -13.141, -13.305, -13.462,
    -11.526, -11.815, -12.079, -12.326, -12.557, -12.770, -12.962, -13.137,
    -13.300, -13.455, -11.516, -11.809, -12.077, -12.325, -12.556, -12.768,
    -12.959, -13.133, -13.294, -13.449, -11.508, -11.802, -12.071, -12.321,
    -12.552, -12.764, -12.955, -13.128, -13.289, -13.442, -11.486, -11.786,
    -12.059, -12.311, -12.545, -12.758, -12.949, -13.122, -13.282, -13.435,
    -11.478, -11.777, -12.051, -12.304, -12.538, -12.752, -12.943, -13.116,
    -13.276, -13.428, -11.482, -11.776, -12.048, -12.301, -12.534, -12.747,
    -12.938, -13.110, -13.270, -13.421, -11.496, -11.777, -12.043, -12.293,
    -12.526, -12.740, -12.931, -13.103, -13.263, -13.414, -11.484, -11.765,
    -12.031, -12.281, -12.516, -12.730, -12.923, -13.096, -13.255, -13.406,
    -11.478, -11.757, -12.021, -12.271, -12.506, -12.721, -12.914, -13.088,
    -13.248, -13.399, -11.477, -11.754, -12.015, -12.263, -12.497, -12.713,
    -12.907, -13.081, -13.241, -13.392, -11.488, -11.762, -12.020, -12.264,
    -12.496, -12.710, -12.902, -13.076, -13.235, -13.385, -11.511, -11.779,
    -12.032, -12.272, -12.499, -12.710, -12.900, -13.072, -13.230, -13.379,
    -11.525, -11.790, -12.040, -12.277, -12.501, -12.709, -12.897, -13.068,
    -13.225, -13.373, -11.532, -11.799, -12.049, -12.285, -12.506, -12.711,
    -12.896, -13.065, -13.220, -13.367, -11.533, -11.800, -12.049, -12.283,
    -12.503, -12.707, -12.892, -13.060, -13.215, -13.361, -11.532, -11.802,
    -12.051, -12.285, -12.503, -12.705, -12.889, -13.055, -13.209, -13.355,
    -11.518, -11.793, -12.046, -12.280, -12.498, -12.700, -12.883, -13.049,
    -13.203, -13.348, -11.506, -11.789, -12.045, -12.279, -12.497, -12.697,
    -12.879, -13.044, -13.197, -13.341, -11.503, -11.789, -12.046, -12.280,
    -12.496, -12.695, -12.875, -13.040, -13.192, -13.335, -11.512, -11.798,
    -12.054, -12.286, -12.499, -12.696, -12.874, -13.036, -13.187, -13.329,
    -11.536, -11.819, -12.071, -12.299, -12.507, -12.700, -12.874, -13.034,
    -13.183, -13.324, -11.604, -11.871, -12.109, -12.324, -12.522, -12.707,
    -12.877, -13.033, -13.179, -13.319, -11.656, -11.910, -12.136, -12.341,
    -12.532, -12.711, -12.877, -13.031, -13.175, -13.314, -11.731, -11.964,
    -12.171, -12.362, -12.543, -12.716, -12.877, -13.028, -13.171, -13.308,
    -11.735, -11.968, -12.173, -12.362, -12.541, -12.713, -12.873, -13.024,
    -13.166, -13.302, -11.743, -11.972, -12.174, -12.361, -12.539, -12.709,
    -12.869, -13.019, -13.160, -13.296, -11.732, -11.958, -12.160, -12.348,
    -12.528, -12.700, -12.861, -13.012, -13.154, -13.290, -11.707, -11.934,
    -12.138, -12.329, -12.513, -12.688, -12.852, -13.004, -13.147, -13.283,
    -11.669, -11.900, -12.110, -12.306, -12.495, -12.674, -12.841, -12.995,
    -13.139, -13.276, -11.630, -11.865, -12.081, -12.284, -12.478, -12.661,
    -12.831, -12.987, -13.132, -13.269, -11.594, -11.836, -12.058, -12.266,
    -12.464, -12.650, -12.822, -12.979, -13.125, -13.262, -11.585, -11.827,
    -12.051, -12.260, -12.458, -12.645, -12.817, -12.974, -13.119, -13.256,
    -11.577, -11.819, -12.043, -12.254, -12.453, -12.640, -12.812, -12.969,
    -13.114, -13.250, -11.563, -11.810, -12.038, -12.251, -12.451, -12.637,
    -12.808, -12.965, -13.109, -13.244, -11.555, -11.802, -12.032, -12.247,
    -12.447, -12.634, -12.805, -12.960, -13.104, -13.239, -11.546, -11.793,
    -12.024, -12.239, -12.441, -12.628, -12.799, -12.955, -13.098, -13.233,
    -11.536, -11.784, -12.016, -12.233, -12.435, -12.623, -12.794, -12.950,
    -13.093, -13.227, -11.546, -11.787, -12.015, -12.230, -12.431, -12.618,
    -12.788, -12.943, -13.087, -13.220, -11.543, -11.782, -12.009, -12.224,
    -12.425, -12.612, -12.783, -12.938, -13.081, -13.214, -11.554, -11.789,
    -12.013, -12.225, -12.424, -12.609, -12.778, -12.933, -13.075, -13.208,
    -11.568, -11.800, -12.020, -12.229, -12.426, -12.609, -12.777, -12.930,
    -13.071, -13.203, -11.575, -11.807, -12.026, -12.233, -12.428, -12.610,
    -12.776, -12.927, -13.067, -13.198, -11.591, -11.820, -12.036, -12.240,
    -12.432, -12.611, -12.775, -12.925, -13.063, -13.193, -11.601, -11.829,
    -12.043, -12.245, -12.436, -12.613, -12.775, -12.922, -13.059, -13.188,
    -11.597, -11.826, -12.041, -12.244, -12.434, -12.611, -12.772, -12.919,
    -13.055, -13.183, -11.593, -11.822, -12.039, -12.242, -12.432, -12.608,
    -12.768, -12.914, -13.050, -13.178, -11.597, -11.825, -12.040, -12.242,
    -12.431, -12.606, -12.765, -12.910, -13.045, -13.172, -11.609, -11.833,
    -12.045, -12.245, -12.432, -12.604, -12.761, -12.906, -13.040, -13.167,
    -11.615, -11.831, -12.041, -12.240, -12.426, -12.598, -12.755, -12.899,
    -13.034, -13.160, -11.622, -11.838, -12.047, -12.243, -12.426, -12.596,
    -12.751, -12.894, -13.028, -13.154, -11.625, -11.842, -12.051, -12.246,
    -12.426, -12.593, -12.747, -12.889, -13.022, -13.148, -11.632, -11.852,
    -12.060, -12.252, -12.429, -12.592, -12.744, -12.884, -13.017, -13.142,
    -11.674, -11.890, -12.092, -12.275, -12.444, -12.600, -12.746, -12.883,
    -13.013, -13.137, -11.735, -11.942, -12.131, -12.303, -12.461, -12.609,
    -12.749, -12.882, -13.010, -13.133, -11.789, -11.985, -12.164, -12.325,
    -12.474, -12.615, -12.750, -12.880, -13.006, -13.128, -11.854, -12.037,
    -12.202, -12.351, -12.490, -12.623, -12.753, -12.880, -13.003, -13.124,
    -11.862, -12.043, -12.206, -12.352, -12.489, -12.621, -12.749, -12.875,
    -12.998, -13.118, -11.862, -12.044, -12.207, -12.353, -12.488, -12.618,
    -12.746, -12.871, -12.993, -13.112, -11.844, -12.030, -12.195, -12.342,
    -12.478, -12.609, -12.738, -12.863, -12.986, -13.106, -11.831, -12.020,
    -12.187, -12.334, -12.472, -12.603, -12.732, -12.858, -12.981, -13.100,
    -11.808, -11.999, -12.167, -12.319, -12.459, -12.594, -12.724, -12.851,
    -12.974, -13.094, -11.799, -11.986, -12.156, -12.309, -12.452, -12.588,
    -12.719, -12.846, -12.969, -13.088, -11.802, -11.984, -12.152, -12.305,
    -12.448, -12.584, -12.715, -12.842, -12.964, -13.083, -11.812, -11.988,
    -12.152, -12.303, -12.446, -12.582, -12.712, -12.838, -12.960, -13.078,
    -11.821, -11.991, -12.151, -12.301, -12.443, -12.579, -12.709, -12.834,
    -12.956, -13.073, -11.830, -11.996, -12.153, -12.301, -12.442, -12.576,
    -12.705, -12.830, -12.951, -13.068, -11.824, -11.990, -12.147, -12.295,
    -12.436, -12.571, -12.700, -12.825, -12.945, -13.062, -11.834, -11.994,
    -12.148, -12.294, -12.434, -12.567, -12.695, -12.819, -12.940, -13.056,
    -11.842, -11.998, -12.149, -12.294, -12.432, -12.564, -12.691, -12.815,
    -12.934, -13.050, -11.834, -11.993, -12.145, -12.289, -12.427, -12.559,
    -12.685, -12.808, -12.928, -13.044, -11.841, -11.999, -12.148, -12.290,
    -12.425, -12.555, -12.680, -12.803, -12.922, -13.038, -11.845, -12.004,
    -12.153, -12.293, -12.425, -12.552, -12.676, -12.797, -12.916, -13.032,
    -11.842, -12.006, -12.157, -12.296, -12.426, -12.550, -12.671, -12.792,
    -12.910, -13.026, -11.853, -12.016, -12.165, -12.301, -12.429, -12.550,
    -12.669, -12.788, -12.905, -13.020, -11.865, -12.026, -12.173, -12.308,
    -12.432, -12.551, -12.668, -12.784, -12.900, -13.015, -11.857, -12.020,
    -12.170, -12.305, -12.430, -12.549, -12.664, -12.780, -12.895, -13.009,
    -11.861, -12.022, -12.170, -12.305, -12.430, -12.547, -12.661, -12.776,
    -12.891, -13.004, -11.868, -12.027, -12.174, -12.307, -12.430, -12.545,
    -12.658, -12.771, -12.885, -12.998, -11.869, -12.029, -12.176, -12.309,
    -12.430, -12.544, -12.655, -12.767, -12.880, -12.993, -11.894, -12.045,
    -12.186, -12.314, -12.432, -12.542, -12.651, -12.762, -12.874, -12.986,
    -11.906, -12.054, -12.193, -12.319, -12.434, -12.542, -12.648, -12.758,
    -12.869, -12.981, -11.914, -12.061, -12.199, -12.323, -12.435, -12.541,
    -12.646, -12.754, -12.864, -12.975, -11.929, -12.075, -12.210, -12.331,
    -12.441, -12.543, -12.645, -12.751, -12.860, -12.970, -11.934, -12.085,
    -12.221, -12.341, -12.447, -12.546, -12.645, -12.749, -12.856, -12.965,
    -11.959, -12.111, -12.246, -12.361, -12.461, -12.554, -12.648, -12.748,
    -12.853, -12.961, -11.982, -12.136, -12.268, -12.377, -12.471, -12.558,
    -12.648, -12.745, -12.849, -12.956, -12.009, -12.165, -12.295, -12.398,
    -12.484, -12.564, -12.649, -12.743, -12.845, -12.951, -12.011, -12.165,
    -12.295, -12.397, -12.482, -12.560, -12.644, -12.738, -12.839, -12.945,
    -12.026, -12.176, -12.301, -12.400, -12.481, -12.558, -12.640, -12.732,
    -12.833, -12.939, -12.041, -12.182, -12.301, -12.396, -12.475, -12.550,
    -12.632, -12.725, -12.826, -12.932, -12.077, -12.201, -12.309, -12.397,
    -12.471, -12.544, -12.625, -12.717, -12.819, -12.925, -12.068, -12.192,
    -12.299, -12.386, -12.461, -12.534, -12.615, -12.708, -12.811, -12.918,
    -12.065, -12.188, -12.294, -12.380, -12.454, -12.526, -12.607, -12.701,
    -12.803, -12.911, -12.068, -12.190, -12.295, -12.379, -12.451, -12.522,
    -12.601, -12.694, -12.797, -12.904, -12.085, -12.205, -12.305, -12.385,
    -12.452, -12.520, -12.598, -12.689, -12.791, -12.898, -12.117, -12.229,
    -12.321, -12.394, -12.456, -12.520, -12.595, -12.685, -12.786, -12.892,
    -12.150, -12.255, -12.339, -12.404, -12.461, -12.521, -12.593, -12.681,
    -12.780, -12.887, -12.175, -12.273, -12.350, -12.409, -12.462, -12.519,
    -12.589, -12.676, -12.775, -12.881, -12.183, -12.278, -12.352, -12.409,
    -12.460, -12.515, -12.584, -12.670, -12.769, -12.875, -12.174, -12.271,
    -12.347, -12.405, -12.455, -12.510, -12.578, -12.664, -12.762, -12.868,
    -12.153, -12.255, -12.335, -12.396, -12.449, -12.504, -12.572, -12.657,
    -12.756, -12.861, -12.125, -12.234, -12.320, -12.385, -12.439, -12.495,
    -12.564, -12.650, -12.748, -12.855, -12.117, -12.226, -12.314, -12.379,
    -12.434, -12.489, -12.557, -12.643, -12.741, -12.848, -12.123, -12.230,
    -12.316, -12.379, -12.431, -12.485, -12.552, -12.637, -12.735, -12.842,
    -12.138, -12.241, -12.323, -12.383, -12.432, -12.483, -12.548, -12.631,
    -12.730, -12.836, -12.162, -12.260, -12.336, -12.391, -12.434, -12.482,
    -12.544, -12.627, -12.724, -12.830, -12.186, -12.278, -12.348, -12.397,
    -12.436, -12.480, -12.541, -12.622, -12.718, -12.824, -12.202, -12.291,
    -12.356, -12.401, -12.436, -12.478, -12.536, -12.616, -12.712, -12.818,
    -12.207, -12.296, -12.360, -12.402, -12.436, -12.475, -12.532, -12.610,
    -12.706, -12.811, -12.227, -12.312, -12.371, -12.409, -12.438, -12.474,
    -12.528, -12.605, -12.700, -12.805, -12.239, -12.321, -12.378, -12.412,
    -12.438, -12.470, -12.523, -12.598, -12.693, -12.798, -12.230, -12.313,
    -12.371, -12.406, -12.432, -12.464, -12.515, -12.591, -12.685, -12.791,
    -12.239, -12.320, -12.377, -12.409, -12.432, -12.461, -12.510, -12.584,
    -12.678, -12.784, -12.237, -12.321, -12.378, -12.410, -12.431, -12.458,
    -12.505, -12.578, -12.671, -12.777, -12.227, -12.314, -12.375, -12.409,
    -12.429, -12.454, -12.500, -12.572, -12.665, -12.770, -12.241, -12.327,
    -12.387, -12.419, -12.435, -12.456, -12.498, -12.568, -12.659, -12.764,
    -12.253, -12.339, -12.397, -12.427, -12.440, -12.457, -12.495, -12.563,
    -12.653, -12.758, -12.249, -12.336, -12.397, -12.427, -12.439, -12.454,
    -12.491, -12.557, -12.647, -12.752, -12.266, -12.350, -12.409, -12.435,
    -12.442, -12.453, -12.487, -12.552, -12.641, -12.745, -12.273, -12.355,
    -12.411, -12.434, -12.439, -12.447, -12.480, -12.544, -12.633, -12.738,
    -12.278, -12.359, -12.412, -12.432, -12.433, -12.439, -12.471, -12.535,
    -12.625, -12.730, -12.311, -12.383, -12.425, -12.435, -12.429, -12.432,
    -12.462, -12.526, -12.616, -12.722, -12.325, -12.392, -12.428, -12.432,
    -12.422, -12.422, -12.452, -12.516, -12.607, -12.714, -12.334, -12.395,
    -12.426, -12.425, -12.411, -12.410, -12.440, -12.505, -12.598, -12.706,
    -12.345, -12.402, -12.427, -12.421, -12.404, -12.401, -12.431, -12.496,
    -12.590, -12.699, -12.339, -12.398, -12.422, -12.414, -12.396, -12.392,
    -12.421, -12.488, -12.582, -12.691, -12.346, -12.400, -12.421, -12.409,
    -12.388, -12.383, -12.412, -12.479, -12.573, -12.684, -12.356, -12.407,
    -12.423, -12.407, -12.383, -12.376, -12.404, -12.471, -12.566, -12.677,
    -12.361, -12.408, -12.421, -12.402, -12.375, -12.367, -12.395, -12.462,
    -12.557, -12.669, -12.363, -12.405, -12.413, -12.392, -12.364, -12.355,
    -12.384, -12.452, -12.548, -12.661, -12.381, -12.414, -12.415, -12.387,
    -12.356, -12.346, -12.374, -12.442, -12.540, -12.653, -12.401, -12.423,
    -12.415, -12.382, -12.347, -12.335, -12.363, -12.432, -12.531, -12.645,
    -12.447, -12.449, -12.424, -12.380, -12.339, -12.325, -12.353, -12.422,
    -12.521, -12.637, -12.445, -12.444, -12.418, -12.372, -12.330, -12.315,
    -12.342, -12.412, -12.512, -12.629, -12.447, -12.442, -12.414, -12.366,
    -12.321, -12.305, -12.332, -12.402, -12.503, -12.620, -12.451, -12.445,
    -12.414, -12.363, -12.316, -12.297, -12.323, -12.393, -12.494, -12.612,
    -12.463, -12.454, -12.419, -12.364, -12.312, -12.291, -12.315, -12.384,
    -12.486, -12.605, -12.484, -12.470, -12.429, -12.367, -12.310, -12.285,
    -12.308, -12.376, -12.478, -12.598, -12.502, -12.484, -12.436, -12.367,
    -12.305, -12.277, -12.299, -12.367, -12.470, -12.590, -12.516, -12.491,
    -12.436, -12.362, -12.296, -12.267, -12.288, -12.358, -12.461, -12.582,
    -12.519, -12.489, -12.430, -12.353, -12.285, -12.256, -12.277, -12.347,
    -12.452, -12.574, -12.512, -12.477, -12.417, -12.339, -12.272, -12.243,
    -12.265, -12.337, -12.442, -12.566, -12.498, -12.459, -12.398, -12.321,
    -12.255, -12.227, -12.252, -12.325, -12.432, -12.557, -12.480, -12.438,
    -12.377, -12.302, -12.238, -12.212, -12.238, -12.313, -12.422, -12.549,
    -12.472, -12.425, -12.361, -12.286, -12.223, -12.198, -12.226, -12.302,
    -12.412, -12.540, -12.472, -12.418, -12.351, -12.274, -12.210, -12.185,
    -12.214, -12.291, -12.403, -12.533, -12.477, -12.417, -12.345, -12.265,
    -12.199, -12.174, -12.203, -12.282, -12.394, -12.525, -12.486, -12.421,
    -12.345, -12.261, -12.194, -12.166, -12.194, -12.273, -12.387, -12.518,
    -12.495, -12.424, -12.343, -12.256, -12.185, -12.156, -12.184, -12.263,
    -12.378, -12.510, -12.498, -12.425, -12.341, -12.251, -12.178, -12.147,
    -12.174, -12.254, -12.369, -12.502, -12.495, -12.416, -12.329, -12.237,
    -12.163, -12.133, -12.161, -12.242, -12.359, -12.493, -12.497, -12.413,
    -12.321, -12.225, -12.149, -12.119, -12.148, -12.230, -12.349, -12.485,
    -12.492, -12.403, -12.307, -12.209, -12.132, -12.102, -12.132, -12.217,
    -12.337, -12.476, -12.478, -12.384, -12.288, -12.190, -12.114, -12.085,
    -12.117, -12.204, -12.326, -12.466, -12.471, -12.371, -12.271, -12.171,
    -12.094, -12.066, -12.100, -12.189, -12.314, -12.456, -12.456, -12.352,
    -12.249, -12.149, -12.074, -12.047, -12.084, -12.175, -12.302, -12.447,
    -12.436, -12.328, -12.225, -12.126, -12.052, -12.028, -12.067, -12.161,
    -12.291, -12.437, -12.427, -12.314, -12.208, -12.107, -12.033, -12.010,
    -12.051, -12.148, -12.280, -12.428, -12.415, -12.296, -12.187, -12.085,
    -12.012, -11.991, -12.035, -12.134, -12.269, -12.419, -12.395, -12.273,
    -12.162, -12.062, -11.990, -11.972, -12.018, -12.121, -12.258, -12.410,
    -12.382, -12.255, -12.142, -12.041, -11.970, -11.954, -12.002, -12.107,
    -12.247, -12.401, -12.364, -12.234, -12.119, -12.018, -11.949, -11.934,
    -11.986, -12.093, -12.235, -12.392, -12.346, -12.213, -12.097, -11.996,
    -11.928, -11.915, -11.969, -12.080, -12.224, -12.383, -12.335, -12.198,
    -12.080, -11.977, -11.909, -11.898, -11.954, -12.066, -12.213, -12.374,
    -12.315, -12.176, -12.056, -11.954, -11.888, -11.879, -11.937, -12.052,
    -12.202, -12.365, -12.293, -12.152, -12.032, -11.931, -11.866, -11.859,
    -11.921, -12.039, -12.191, -12.356, -12.268, -12.126, -12.007, -11.908,
    -11.845, -11.841, -11.904, -12.025, -12.180, -12.347, -12.236, -12.094,
    -11.978, -11.882, -11.822, -11.821, -11.888, -12.012, -12.169, -12.338,
    -12.207, -12.065, -11.950, -11.856, -11.799, -11.801, -11.872, -11.998,
    -12.158, -12.330, -12.177, -12.036, -11.922, -11.831, -11.776, -11.781,
    -11.855, -11.985, -12.147, -12.321, -12.145, -12.005, -11.892, -11.803,
    -11.752, -11.760, -11.837, -11.971, -12.136, -12.312, -12.111, -11.972,
    -11.861, -11.775, -11.727, -11.739, -11.820, -11.956, -12.125, -12.303,
    -12.080, -11.942, -11.833, -11.749, -11.704, -11.719, -11.803, -11.943,
    -12.114, -12.294, -12.048, -11.911, -11.804, -11.722, -11.680, -11.698,
    -11.786, -11.929, -12.103, -12.285, -12.021, -11.885, -11.778, -11.697,
    -11.657, -11.678, -11.768, -11.914, -12.091, -12.276, -11.981, -11.849,
    -11.745, -11.668, -11.632, -11.657, -11.751, -11.901, -12.080, -12.267,
    -11.939, -11.812, -11.712, -11.639, -11.607, -11.636, -11.734, -11.887,
    -12.069, -12.259, -11.897, -11.775, -11.679, -11.610, -11.582, -11.615,
    -11.717, -11.873, -12.059, -12.250, -11.855, -11.738, -11.647, -11.583,
    -11.559, -11.596, -11.701, -11.861, -12.049, -12.242, -11.812, -11.702,
    -11.616, -11.556, -11.536, -11.577, -11.686, -11.848, -12.039, -12.234,
    -11.769, -11.664, -11.584, -11.529, -11.513, -11.558, -11.671, -11.836,
    -12.029, -12.227, -11.724, -11.626, -11.550, -11.500, -11.489, -11.537,
    -11.653, -11.823, -12.019, -12.218, -11.678, -11.587, -11.516, -11.470,
    -11.463, -11.516, -11.636, -11.809, -12.008, -12.210, -11.631, -11.546,
    -11.481, -11.439, -11.437, -11.494, -11.618, -11.795, -11.997, -12.201,
    -11.584, -11.505, -11.445, -11.409, -11.411, -11.473, -11.601, -11.781,
    -11.986, -12.192, -11.536, -11.464, -11.409, -11.378, -11.385, -11.450,
    -11.583, -11.767, -11.975, -12.184, -11.489, -11.424, -11.374, -11.347,
    -11.359, -11.429, -11.565, -11.752, -11.964, -12.175, -11.443, -11.385,
    -11.340, -11.318, -11.334, -11.408, -11.548, -11.739, -11.953, -12.167,
    -11.397, -11.346, -11.307, -11.290, -11.310, -11.388, -11.532, -11.726,
    -11.943, -12.159, -11.352, -11.309, -11.275, -11.262, -11.287, -11.369,
    -11.516, -11.714, -11.933, -12.151, -11.307, -11.271, -11.243, -11.235,
    -11.263, -11.349, -11.501, -11.701, -11.924, -12.143, -11.263, -11.234,
    -11.211, -11.208, -11.241, -11.331, -11.485, -11.689, -11.914, -12.136,
    -11.219, -11.198, -11.180, -11.181, -11.218, -11.312, -11.470, -11.678,
    -11.905, -12.128, -11.177, -11.162, -11.150, -11.156, -11.197, -11.294,
    -11.456, -11.666, -11.896, -12.121, -11.135, -11.128, -11.120, -11.131,
    -11.176, -11.277, -11.442, -11.655, -11.887, -12.114, -11.095, -11.093,
    -11.091, -11.105, -11.155, -11.260, -11.428, -11.644, -11.878, -12.107,
    -11.056, -11.061, -11.063, -11.082, -11.135, -11.243, -11.415, -11.633,
    -11.870, -12.100, -10.784, -10.842, -10.892, -10.951, -11.036, -11.170,
    -11.362, -11.596, -11.844, -12.082, -10.575, -10.671, -10.755, -10.843,
    -10.953, -11.107, -11.316, -11.563, -11.821, -12.066, -10.314, -10.454,
    -10.579, -10.703, -10.844, -11.024, -11.255, -11.520, -11.791, -12.046,
    -10.129, -10.297, -10.449, -10.598, -10.761, -10.960, -11.206, -11.485,
    -11.767, -12.029, -10.025, -10.214, -10.384, -10.546, -10.720, -10.929,
    -11.183, -11.467, -11.754, -12.020, -9.893,  -10.107, -10.299, -10.479,
    -10.668, -10.888, -11.152, -11.445, -11.738, -12.008, -9.920,  -10.141,
    -10.331, -10.508, -10.692, -10.907, -11.166, -11.454, -11.742, -12.010,
    -9.912,  -10.144, -10.341, -10.519, -10.702, -10.916, -11.172, -11.457,
    -11.744, -12.009, -9.840,  -10.091, -10.302, -10.490, -10.680, -10.899,
    -11.160, -11.448, -11.736, -12.003, -9.903,  -10.152, -10.356, -10.535,
    -10.716, -10.926, -11.178, -11.460, -11.743, -12.005, -9.887,  -10.146,
    -10.354, -10.535, -10.716, -10.926, -11.178, -11.459, -11.741, -12.003,
    -9.820,  -10.096, -10.317, -10.508, -10.696, -10.911, -11.167, -11.450,
    -11.735, -11.997, -9.919,  -10.184, -10.390, -10.566, -10.740, -10.943,
    -11.189, -11.465, -11.743, -12.001, -9.882,  -10.151, -10.361, -10.541,
    -10.720, -10.927, -11.176, -11.455, -11.735, -11.995, -9.726,  -10.013,
    -10.244, -10.445, -10.643, -10.867, -11.132, -11.423, -11.713, -11.979,
    -9.659,  -9.942,  -10.176, -10.383, -10.590, -10.824, -11.098, -11.398,
    -11.695, -11.966, -9.475,  -9.775,  -10.030, -10.261, -10.490, -10.745,
    -11.038, -11.353, -11.663, -11.944, -9.278,  -9.602,  -9.883,  -10.138,
    -10.390, -10.666, -10.977, -11.308, -11.631, -11.922, -9.192,  -9.534,
    -9.831,  -10.099, -10.361, -10.644, -10.960, -11.296, -11.622, -11.914,
    -9.153,  -9.515,  -9.824,  -10.100, -10.366, -10.651, -10.966, -11.300,
    -11.624, -11.914, -9.171,  -9.549,  -9.866,  -10.143, -10.406, -10.684,
    -10.993, -11.319, -11.636, -11.920, -9.280,  -9.663,  -9.976,  -10.242,
    -10.490, -10.752, -11.044, -11.355, -11.659, -11.933, -9.458,  -9.829,
    -10.121, -10.362, -10.585, -10.825, -11.097, -11.392, -11.682, -11.946,
    -9.673,  -10.015, -10.272, -10.480, -10.674, -10.890, -11.143, -11.422,
    -11.700, -11.955, -9.798,  -10.116, -10.349, -10.537, -10.716, -10.919,
    -11.162, -11.434, -11.706, -11.957, -9.748,  -10.076, -10.318, -10.513,
    -10.698, -10.905, -11.152, -11.425, -11.699, -11.951, -9.547,  -9.910,
    -10.187, -10.413, -10.622, -10.849, -11.111, -11.397, -11.679, -11.937,
    -9.301,  -9.695,  -10.009, -10.270, -10.511, -10.766, -11.050, -11.353,
    -11.649, -11.917, -9.076,  -9.487,  -9.826,  -10.116, -10.386, -10.667,
    -10.975, -11.299, -11.611, -11.891, -8.930,  -9.341,  -9.688,  -9.992,
    -10.279, -10.579, -10.906, -11.247, -11.574, -11.866, -8.873,  -9.274,
    -9.618,  -9.924,  -10.217, -10.525, -10.861, -11.212, -11.548, -11.848,
    -8.862,  -9.252,  -9.589,  -9.893,  -10.186, -10.497, -10.837, -11.192,
    -11.533, -11.835, -8.919,  -9.296,  -9.622,  -9.916,  -10.203, -10.507,
    -10.843, -11.195, -11.533, -11.834, -9.056,  -9.421,  -9.731,  -10.009,
    -10.279, -10.568, -10.890, -11.228, -11.555, -11.846, -9.157,  -9.519,
    -9.822,  -10.088, -10.346, -10.622, -10.931, -11.258, -11.574, -11.857,
    -9.217,  -9.586,  -9.887,  -10.148, -10.397, -10.664, -10.963, -11.281,
    -11.589, -11.865, -9.324,  -9.688,  -9.980,  -10.228, -10.463, -10.717,
    -11.003, -11.308, -11.606, -11.874, -9.306,  -9.679,  -9.977,  -10.229,
    -10.466, -10.720, -11.005, -11.309, -11.606, -11.872, -9.211,  -9.601,
    -9.915,  -10.182, -10.430, -10.693, -10.986, -11.295, -11.595, -11.864,
    -9.293,  -9.668,  -9.969,  -10.224, -10.463, -10.718, -11.003, -11.306,
    -11.601, -11.865, -9.242,  -9.623,  -9.931,  -10.193, -10.439, -10.699,
    -10.989, -11.295, -11.592, -11.858, -9.151,  -9.541,  -9.861,  -10.136,
    -10.393, -10.664, -10.962, -11.275, -11.578, -11.847, -9.212,  -9.587,
    -9.895,  -10.161, -10.411, -10.676, -10.970, -11.280, -11.579, -11.846,
    -9.161,  -9.535,  -9.847,  -10.119, -10.376, -10.648, -10.948, -11.263,
    -11.566, -11.836, -9.128,  -9.501,  -9.815,  -10.091, -10.352, -10.629,
    -10.933, -11.251, -11.557, -11.829, -9.231,  -9.585,  -9.883,  -10.144,
    -10.394, -10.660, -10.955, -11.266, -11.566, -11.832, -9.279,  -9.623,
    -9.912,  -10.167, -10.411, -10.673, -10.964, -11.272, -11.568, -11.832,
    -9.475,  -9.792,  -10.054, -10.282, -10.503, -10.743, -11.015, -11.307,
    -11.590, -11.844, -9.726,  -10.002, -10.223, -10.416, -10.606, -10.820,
    -11.071, -11.344, -11.613, -11.856, -9.876,  -10.127, -10.323, -10.495,
    -10.666, -10.865, -11.103, -11.365, -11.625, -11.861, -10.097, -10.304,
    -10.461, -10.600, -10.745, -10.922, -11.143, -11.392, -11.640, -11.869,
    -10.101, -10.309, -10.467, -10.606, -10.751, -10.927, -11.146, -11.393,
    -11.639, -11.866, -10.108, -10.316, -10.474, -10.613, -10.758, -10.933,
    -11.150, -11.394, -11.639, -11.863, -10.116, -10.325, -10.483, -10.622,
    -10.765, -10.939, -11.154, -11.396, -11.639, -11.861, -10.126, -10.335,
    -10.493, -10.631, -10.773, -10.945, -11.159, -11.399, -11.639, -11.859,
    -10.138, -10.346, -10.504, -10.642, -10.783, -10.953, -11.164, -11.401,
    -11.639, -11.857, -10.151, -10.358, -10.516, -10.653, -10.792, -10.961,
    -11.170, -11.405, -11.640, -11.856, -10.166, -10.372, -10.529, -10.665,
    -10.803, -10.969, -11.176, -11.408, -11.641, -11.854, -10.182, -10.386,
    -10.543, -10.678, -10.814, -10.978, -11.182, -11.412, -11.641, -11.853,
    -10.200, -10.402, -10.558, -10.691, -10.826, -10.988, -11.189, -11.416,
    -11.643, -11.852, -10.219, -10.419, -10.573, -10.706, -10.838, -10.998,
    -11.197, -11.420, -11.644, -11.850, -10.238, -10.437, -10.590, -10.721,
    -10.852, -11.009, -11.205, -11.425, -11.646, -11.849, -10.259, -10.455,
    -10.607, -10.736, -10.865, -11.020, -11.213, -11.430, -11.647, -11.848,
    -10.281, -10.474, -10.625, -10.752, -10.879, -11.031, -11.221, -11.435,
    -11.650, -11.848, -10.304, -10.494, -10.643, -10.769, -10.894, -11.043,
    -11.230, -11.441, -11.652, -11.847, -10.328, -10.515, -10.662, -10.786,
    -10.909, -11.056, -11.240, -11.447, -11.655, -11.847, -10.352, -10.536,
    -10.681, -10.804, -10.924, -11.069, -11.249, -11.453, -11.657, -11.846,
    -10.378, -10.558, -10.701, -10.822, -10.940, -11.082, -11.259, -11.460,
    -11.660, -11.846, -10.404, -10.580, -10.722, -10.840, -10.956, -11.095,
    -11.269, -11.466, -11.663, -11.846, -10.431, -10.603, -10.742, -10.859,
    -10.972, -11.109, -11.279, -11.473, -11.666, -11.846, -10.458, -10.627,
    -10.763, -10.878, -10.989, -11.122, -11.289, -11.479, -11.669, -11.845,
    -10.486, -10.650, -10.785, -10.897, -11.005, -11.136, -11.299, -11.485,
    -11.671, -11.845, -10.514, -10.675, -10.807, -10.916, -11.022, -11.149,
    -11.310, -11.492, -11.674, -11.844, -10.543, -10.699, -10.829, -10.936,
    -11.039, -11.163, -11.320, -11.498, -11.677, -11.844, -10.573, -10.724,
    -10.851, -10.956, -11.057, -11.178, -11.331, -11.505, -11.680, -11.844,
    -10.603, -10.749, -10.874, -10.976, -11.074, -11.192, -11.342, -11.513,
    -11.684, -11.844, -10.633, -10.775, -10.897, -10.997, -11.092, -11.207,
    -11.354, -11.521, -11.688, -11.845, -10.664, -10.801, -10.920, -11.018,
    -11.111, -11.223, -11.365, -11.528, -11.691, -11.845, -10.695, -10.827,
    -10.943, -11.039, -11.129, -11.238, -11.377, -11.536, -11.695, -11.845,
    -10.726, -10.853, -10.967, -11.060, -11.147, -11.253, -11.389, -11.543,
    -11.698, -11.845, -10.758, -10.879, -10.990, -11.081, -11.166, -11.269,
    -11.400, -11.551, -11.702, -11.845, -10.790, -10.906, -11.014, -11.102,
    -11.184, -11.284, -11.412, -11.558, -11.705, -11.844, -10.822, -10.932,
    -11.038, -11.123, -11.203, -11.299, -11.423, -11.565, -11.708, -11.844,
    -10.855, -10.959, -11.062, -11.145, -11.221, -11.314, -11.434, -11.572,
    -11.711, -11.843, -10.887, -10.986, -11.086, -11.166, -11.240, -11.329,
    -11.446, -11.579, -11.714, -11.843, -10.920, -11.013, -11.110, -11.188,
    -11.258, -11.345, -11.457, -11.586, -11.716, -11.842, -10.953, -11.040,
    -11.134, -11.209, -11.277, -11.360, -11.469, -11.593, -11.720, -11.842,
    -10.986, -11.067, -11.158, -11.231, -11.296, -11.375, -11.480, -11.600,
    -11.722, -11.842, -11.019, -11.094, -11.182, -11.252, -11.315, -11.390,
    -11.491, -11.607, -11.725, -11.841, -11.052, -11.121, -11.207, -11.274,
    -11.333, -11.406, -11.503, -11.614, -11.728, -11.841, -11.085, -11.149,
    -11.231, -11.296, -11.352, -11.421, -11.514, -11.621, -11.731, -11.840,
    -11.118, -11.176, -11.255, -11.317, -11.371, -11.437, -11.525, -11.628,
    -11.734, -11.839, -11.152, -11.203, -11.279, -11.339, -11.390, -11.452,
    -11.537, -11.636, -11.737, -11.839, -11.185, -11.230, -11.304, -11.361,
    -11.409, -11.467, -11.548, -11.642, -11.739, -11.838, -11.218, -11.257,
    -11.328, -11.382, -11.427, -11.483, -11.559, -11.649, -11.742, -11.837,
    -11.252, -11.285, -11.352, -11.404, -11.446, -11.498, -11.570, -11.656,
    -11.744, -11.836, -11.285, -11.312, -11.376, -11.425, -11.464, -11.513,
    -11.581, -11.662, -11.747, -11.835, -11.318, -11.339, -11.400, -11.446,
    -11.483, -11.527, -11.592, -11.668, -11.749, -11.835, -11.352, -11.366,
    -11.424, -11.468, -11.501, -11.542, -11.602, -11.675, -11.751, -11.833,
    -11.385, -11.392, -11.448, -11.489, -11.519, -11.557, -11.613, -11.681,
    -11.753, -11.832, -11.418, -11.419, -11.471, -11.510, -11.537, -11.571,
    -11.623, -11.686, -11.755, -11.831, -11.451, -11.446, -11.495, -11.531,
    -11.555, -11.585, -11.633, -11.692, -11.756, -11.829, -11.484, -11.472,
    -11.519, -11.552, -11.573, -11.599, -11.643, -11.697, -11.758, -11.828,
    -11.516, -11.499, -11.543, -11.573, -11.591, -11.614, -11.653, -11.703,
    -11.759, -11.826, -11.549, -11.525, -11.566, -11.594, -11.608, -11.627,
    -11.662, -11.708, -11.760, -11.824, -11.582, -11.551, -11.589, -11.614,
    -11.625, -11.641, -11.671, -11.713, -11.761, -11.822, -11.614, -11.578,
    -11.612, -11.634, -11.643, -11.654, -11.681, -11.718, -11.763, -11.821,
    -11.646, -11.604, -11.635, -11.655, -11.660, -11.668, -11.690, -11.723,
    -11.764, -11.819, -11.678, -11.629, -11.658, -11.675, -11.677, -11.681,
    -11.699, -11.728, -11.765, -11.817, -11.710, -11.655, -11.681, -11.695,
    -11.694, -11.695, -11.708, -11.733, -11.766, -11.815, -11.741, -11.680,
    -11.703, -11.715, -11.710, -11.707, -11.717, -11.736, -11.766, -11.813,
    -11.773, -11.706, -11.726, -11.734, -11.726, -11.719, -11.724, -11.740,
    -11.766, -11.810, -11.804, -11.731, -11.748, -11.753, -11.742, -11.731,
    -11.732, -11.743, -11.765, -11.807, -11.835, -11.756, -11.770, -11.772,
    -11.757, -11.742, -11.739, -11.746, -11.764, -11.803, -11.865, -11.781,
    -11.791, -11.791, -11.772, -11.753, -11.745, -11.748, -11.763, -11.800,
    -11.896, -11.805, -11.813, -11.809, -11.787, -11.764, -11.751, -11.750,
    -11.762, -11.796, -11.926, -11.829, -11.834, -11.827, -11.802, -11.775,
    -11.758, -11.752, -11.760, -11.792, -11.956, -11.854, -11.855, -11.845,
    -11.816, -11.785, -11.764, -11.754, -11.759, -11.788, -11.986, -11.878,
    -11.876, -11.863, -11.831, -11.796, -11.770, -11.756, -11.757, -11.785,
    -12.015, -11.901, -11.897, -11.881, -11.845, -11.806, -11.775, -11.757,
    -11.756, -11.781, -12.044, -11.925, -11.918, -11.898, -11.859, -11.815,
    -11.781, -11.758, -11.753, -11.777, -12.073, -11.948, -11.938, -11.915,
    -11.872, -11.824, -11.786, -11.759, -11.751, -11.772, -12.101, -11.971,
    -11.958, -11.932, -11.885, -11.833, -11.790, -11.760, -11.749, -11.768,
    -12.129, -11.994, -11.978, -11.948, -11.898, -11.842, -11.794, -11.760,
    -11.746, -11.763, -12.157, -12.016, -11.997, -11.964, -11.910, -11.849,
    -11.798, -11.759, -11.743, -11.758, -12.184, -12.039, -12.016, -11.980,
    -11.921, -11.857, -11.801, -11.759, -11.740, -11.753, -12.211, -12.060,
    -12.035, -11.995, -11.933, -11.864, -11.804, -11.758, -11.736, -11.748,
    -12.238, -12.082, -12.053, -12.010, -11.943, -11.870, -11.806, -11.757,
    -11.732, -11.743, -12.264, -12.103, -12.071, -12.025, -11.954, -11.877,
    -11.808, -11.756, -11.728, -11.738, -12.290, -12.125, -12.089, -12.039,
    -11.964, -11.883, -11.810, -11.754, -11.724, -11.732, -12.316, -12.145,
    -12.106, -12.052, -11.973, -11.888, -11.811, -11.751, -11.719, -11.726,
    -12.341, -12.166, -12.123, -12.066, -11.983, -11.893, -11.812, -11.749,
    -11.715, -11.720, -12.365, -12.186, -12.140, -12.078, -11.991, -11.897,
    -11.812, -11.746, -11.710, -11.714, -12.390, -12.206, -12.156, -12.091,
    -11.999, -11.901, -11.812, -11.743, -11.705, -11.708, -12.414, -12.225,
    -12.172, -12.103, -12.007, -11.905, -11.812, -11.740, -11.699, -11.702,
    -12.437, -12.244, -12.187, -12.114, -12.014, -11.908, -11.811, -11.736,
    -11.693, -11.695, -12.460, -12.262, -12.202, -12.125, -12.021, -11.910,
    -11.810, -11.732, -11.688, -11.688, -12.482, -12.281, -12.217, -12.135,
    -12.027, -11.912, -11.808, -11.727, -11.681, -11.681, -12.504, -12.298,
    -12.231, -12.145, -12.032, -11.913, -11.806, -11.722, -11.675, -11.674,
    -12.526, -12.316, -12.244, -12.154, -12.037, -11.914, -11.803, -11.717,
    -11.668, -11.667, -12.547, -12.332, -12.257, -12.162, -12.041, -11.914,
    -11.800, -11.712, -11.662, -11.660, -12.567, -12.349, -12.269, -12.170,
    -12.045, -11.914, -11.797, -11.706, -11.654, -11.653, -12.587, -12.365,
    -12.281, -12.177, -12.047, -11.913, -11.792, -11.699, -11.647, -11.645,
    -12.607, -12.380, -12.292, -12.184, -12.050, -11.911, -11.788, -11.693,
    -11.639, -11.637, -12.626, -12.395, -12.302, -12.189, -12.051, -11.909,
    -11.782, -11.685, -11.631, -11.628, -12.644, -12.409, -12.312, -12.195,
    -12.052, -11.907, -11.777, -11.678, -11.622, -11.620, -12.662, -12.423,
    -12.321, -12.199, -12.053, -11.903, -11.771, -11.670, -11.614, -11.611,
    -12.679, -12.436, -12.329, -12.203, -12.052, -11.899, -11.764, -11.662,
    -11.605, -11.602, -12.695, -12.448, -12.337, -12.205, -12.051, -11.895,
    -11.757, -11.653, -11.596, -11.594, -12.711, -12.460, -12.344, -12.207,
    -12.049, -11.890, -11.750, -11.644, -11.586, -11.585, -12.726, -12.471,
    -12.350, -12.209, -12.046, -11.884, -11.742, -11.635, -11.577, -11.576,
    -12.741, -12.481, -12.355, -12.209, -12.043, -11.878, -11.734, -11.626,
    -11.567, -11.567, -12.755, -12.491, -12.359, -12.209, -12.039, -11.871,
    -11.725, -11.616, -11.557, -11.557, -12.767, -12.499, -12.363, -12.207,
    -12.033, -11.863, -11.715, -11.605, -11.547, -11.547, -12.780, -12.507,
    -12.365, -12.205, -12.028, -11.854, -11.705, -11.594, -11.536, -11.537,
    -12.791, -12.514, -12.367, -12.202, -12.021, -11.845, -11.694, -11.583,
    -11.525, -11.527, -12.801, -12.520, -12.367, -12.198, -12.013, -11.835,
    -11.683, -11.571, -11.514, -11.517, -12.811, -12.525, -12.367, -12.193,
    -12.005, -11.825, -11.671, -11.560, -11.502, -11.507, -12.820, -12.529,
    -12.365, -12.187, -11.996, -11.814, -11.659, -11.547, -11.491, -11.496,
    -12.827, -12.533, -12.363, -12.180, -11.986, -11.802, -11.647, -11.535,
    -11.479, -11.486, -12.834, -12.535, -12.359, -12.172, -11.976, -11.790,
    -11.634, -11.523, -11.468, -11.476, -12.840, -12.536, -12.355, -12.163,
    -11.964, -11.778, -11.621, -11.510, -11.455, -11.465, -12.844, -12.535,
    -12.349, -12.153, -11.952, -11.764, -11.608, -11.496, -11.443, -11.454,
    -12.848, -12.534, -12.342, -12.143, -11.939, -11.751, -11.593, -11.483,
    -11.431, -11.443, -12.850, -12.532, -12.334, -12.131, -11.926, -11.736,
    -11.579, -11.469, -11.418, -11.432, -12.851, -12.528, -12.325, -12.118,
    -11.911, -11.721, -11.564, -11.455, -11.405, -11.421, -12.850, -12.523,
    -12.314, -12.104, -11.896, -11.705, -11.548, -11.440, -11.392, -11.409,
    -12.848, -12.516, -12.303, -12.090, -11.880, -11.689, -11.533, -11.425,
    -11.379, -11.398, -12.845, -12.508, -12.290, -12.074, -11.863, -11.672,
    -11.516, -11.410, -11.365, -11.386, -12.841, -12.499, -12.276, -12.057,
    -11.845, -11.655, -11.500, -11.395, -11.351, -11.374, -12.834, -12.489,
    -12.261, -12.040, -11.827, -11.637, -11.483, -11.379, -11.338, -11.363,
    -12.827, -12.477, -12.244, -12.021, -11.808, -11.618, -11.465, -11.363,
    -11.324, -11.351, -12.817, -12.463, -12.227, -12.002, -11.788, -11.599,
    -11.447, -11.347, -11.310, -11.339, -12.806, -12.448, -12.208, -11.981,
    -11.768, -11.580, -11.429, -11.331, -11.295, -11.327, -12.793, -12.432,
    -12.188, -11.960, -11.747, -11.560, -11.411, -11.314, -11.281, -11.315,
    -12.779, -12.414, -12.167, -11.938, -11.725, -11.539, -11.392, -11.297,
    -11.266, -11.302, -12.762, -12.395, -12.144, -11.915, -11.703, -11.518,
    -11.373, -11.280, -11.251, -11.290, -12.744, -12.374, -12.121, -11.891,
    -11.680, -11.497, -11.354, -11.263, -11.237, -11.277, -12.724, -12.351,
    -12.096, -11.866, -11.656, -11.475, -11.334, -11.246, -11.221, -11.265,
    -12.702, -12.327, -12.070, -11.840, -11.632, -11.453, -11.314, -11.228,
    -11.206, -11.252, -12.677, -12.302, -12.043, -11.814, -11.607, -11.430,
    -11.293, -11.210, -11.191, -11.240, -12.651, -12.275, -12.015, -11.786,
    -11.582, -11.407, -11.273, -11.192, -11.176, -11.227, -12.623, -12.246,
    -11.986, -11.758, -11.556, -11.384, -11.252, -11.174, -11.160, -11.214,
    -12.594, -12.217, -11.956, -11.730, -11.529, -11.360, -11.231, -11.155,
    -11.145, -11.201, -12.562, -12.185, -11.925, -11.700, -11.503, -11.336,
    -11.209, -11.137, -11.129, -11.188, -12.528, -12.153, -11.893, -11.670,
    -11.475, -11.311, -11.188, -11.118, -11.113, -11.175, -12.492, -12.119,
    -11.860, -11.639, -11.447, -11.286, -11.166, -11.099, -11.097, -11.162,
    -12.455, -12.084, -11.826, -11.608, -11.419, -11.261, -11.144, -11.080,
    -11.081, -11.149, -12.415, -12.047, -11.791, -11.576, -11.390, -11.236,
    -11.122, -11.061, -11.065, -11.136, -12.374, -12.009, -11.756, -11.543,
    -11.361, -11.210, -11.099, -11.042, -11.049, -11.123, -12.331, -11.970,
    -11.719, -11.510, -11.331, -11.184, -11.077, -11.023, -11.033, -11.111,
    -12.287, -11.930, -11.682, -11.477, -11.302, -11.158, -11.054, -11.004,
    -11.017, -11.098, -12.241, -11.889, -11.644, -11.443, -11.271, -11.131,
    -11.031, -10.984, -11.001, -11.085, -12.193, -11.847, -11.606, -11.408,
    -11.241, -11.105, -11.008, -10.965, -10.985, -11.072, -12.144, -11.804,
    -11.567, -11.373, -11.210, -11.078, -10.985, -10.945, -10.969, -11.059,
    -12.093, -11.760, -11.527, -11.338, -11.179, -11.051, -10.962, -10.926,
    -10.953, -11.046, -12.041, -11.715, -11.487, -11.302, -11.147, -11.024,
    -10.939, -10.906, -10.937, -11.033, -11.988, -11.669, -11.446, -11.266,
    -11.116, -10.996, -10.915, -10.886, -10.920, -11.020, -11.934, -11.623,
    -11.405, -11.229, -11.084, -10.969, -10.892, -10.866, -10.904, -11.007,
    -11.878, -11.576, -11.363, -11.192, -11.052, -10.941, -10.868, -10.847,
    -10.888, -10.994, -11.822, -11.528, -11.321, -11.155, -11.020, -10.913,
    -10.845, -10.827, -10.872, -10.981, -11.765, -11.480, -11.278, -11.118,
    -10.988, -10.886, -10.821, -10.807, -10.856, -10.968, -11.707, -11.431,
    -11.236, -11.081, -10.955, -10.858, -10.797, -10.788, -10.840, -10.956,
    -11.648, -11.381, -11.193, -11.043, -10.923, -10.830, -10.774, -10.768,
    -10.824, -10.943, -11.588, -11.332, -11.149, -11.006, -10.890, -10.802,
    -10.750, -10.749, -10.808, -10.931, -11.528, -11.282, -11.106, -10.968,
    -10.858, -10.775, -10.727, -10.729, -10.793, -10.918, -11.468, -11.231,
    -11.062, -10.930, -10.825, -10.747, -10.704, -10.710, -10.777, -10.906,
    -11.407, -11.181, -11.018, -10.892, -10.793, -10.719, -10.680, -10.691,
    -10.761, -10.894, -11.346, -11.130, -10.974, -10.854, -10.760, -10.691,
    -10.657, -10.671, -10.746, -10.881, -11.285, -11.079, -10.931, -10.816,
    -10.728, -10.664, -10.634, -10.652, -10.730, -10.869, -11.223, -11.028,
    -10.887, -10.779, -10.695, -10.636, -10.611, -10.633, -10.715, -10.857,
    -11.162, -10.977, -10.843, -10.741, -10.663, -10.609, -10.588, -10.614,
    -10.700, -10.845, -11.100, -10.926, -10.799, -10.703, -10.631, -10.582,
    -10.565, -10.596, -10.685, -10.834, -11.039, -10.875, -10.756, -10.666,
    -10.599, -10.555, -10.543, -10.577, -10.670, -10.822, -10.978, -10.825,
    -10.712, -10.629, -10.567, -10.528, -10.521, -10.559, -10.655, -10.811,
    -10.917, -10.774, -10.669, -10.592, -10.536, -10.502, -10.499, -10.541,
    -10.641, -10.799, -10.856, -10.724, -10.627, -10.555, -10.505, -10.475,
    -10.477, -10.523, -10.627, -10.788, -10.796, -10.674, -10.584, -10.519,
    -10.474, -10.449, -10.455, -10.506, -10.613, -10.777, -10.737, -10.625,
    -10.542, -10.483, -10.443, -10.424, -10.434, -10.488, -10.599, -10.766,
    -10.678, -10.576, -10.501, -10.447, -10.413, -10.398, -10.413, -10.471,
    -10.585, -10.756, -10.620, -10.528, -10.459, -10.412, -10.383, -10.373,
    -10.392, -10.455, -10.572, -10.746, -10.562, -10.481, -10.419, -10.378,
    -10.354, -10.349, -10.372, -10.438, -10.559, -10.735, -10.506, -10.434,
    -10.379, -10.344, -10.325, -10.325, -10.352, -10.422, -10.546, -10.725,
    -10.450, -10.388, -10.340, -10.310, -10.297, -10.301, -10.333, -10.406,
    -10.534, -10.716, -10.396, -10.342, -10.301, -10.277, -10.269, -10.278,
    -10.314, -10.391, -10.522, -10.706, -10.342, -10.298, -10.263, -10.245,
    -10.242, -10.255, -10.295, -10.376, -10.510, -10.697, -10.290, -10.255,
    -10.227, -10.214, -10.215, -10.233, -10.277, -10.361, -10.498, -10.688,
    -10.240, -10.213, -10.191, -10.183, -10.190, -10.212, -10.259, -10.347,
    -10.487, -10.679, -10.190, -10.172, -10.156, -10.154, -10.164, -10.191,
    -10.242, -10.333, -10.476, -10.671, -10.143, -10.132, -10.122, -10.125,
    -10.140, -10.171, -10.226, -10.320, -10.465, -10.663, -10.097, -10.094,
    -10.089, -10.097, -10.117, -10.151, -10.210, -10.307, -10.455, -10.655,
    -10.053, -10.057, -10.058, -10.070, -10.094, -10.132, -10.194, -10.295,
    -10.445, -10.647, -9.845,  -9.890,  -9.927,  -9.968,  -10.015, -10.072,
    -10.148, -10.260, -10.419, -10.628, -9.656,  -9.735,  -9.802,  -9.869,
    -9.938,  -10.012, -10.102, -10.225, -10.393, -10.608, -9.388,  -9.512,
    -9.622,  -9.726,  -9.826,  -9.925,  -10.036, -10.175, -10.356, -10.581,
    -9.137,  -9.300,  -9.446,  -9.583,  -9.711,  -9.835,  -9.966,  -10.121,
    -10.315, -10.551, -9.026,  -9.206,  -9.370,  -9.521,  -9.661,  -9.794,
    -9.934,  -10.096, -10.296, -10.537, -8.878,  -9.087,  -9.274,  -9.444,
    -9.600,  -9.747,  -9.897,  -10.068, -10.275, -10.520, -8.815,  -9.040,
    -9.239,  -9.418,  -9.580,  -9.731,  -9.885,  -10.058, -10.267, -10.515,
    -8.850,  -9.080,  -9.278,  -9.453,  -9.609,  -9.755,  -9.903,  -10.073,
    -10.278, -10.522, -8.777,  -9.028,  -9.242,  -9.427,  -9.591,  -9.742,
    -9.893,  -10.065, -10.272, -10.517, -8.779,  -9.038,  -9.254,  -9.439,
    -9.602,  -9.750,  -9.900,  -10.069, -10.275, -10.519, -8.835,  -9.094,
    -9.304,  -9.481,  -9.635,  -9.777,  -9.920,  -10.084, -10.285, -10.526,
    -8.758,  -9.035,  -9.259,  -9.448,  -9.610,  -9.757,  -9.905,  -10.073,
    -10.276, -10.519, -8.775,  -9.056,  -9.280,  -9.465,  -9.624,  -9.768,
    -9.913,  -10.078, -10.280, -10.521, -8.835,  -9.111,  -9.325,  -9.501,
    -9.652,  -9.789,  -9.928,  -10.089, -10.287, -10.526, -8.747,  -9.030,
    -9.255,  -9.442,  -9.603,  -9.749,  -9.897,  -10.064, -10.268, -10.511,
    -8.618,  -8.911,  -9.149,  -9.352,  -9.528,  -9.689,  -9.849,  -10.027,
    -10.239, -10.489, -8.509,  -8.804,  -9.051,  -9.265,  -9.455,  -9.628,
    -9.800,  -9.989,  -10.210, -10.467, -8.299,  -8.614,  -8.886,  -9.125,
    -9.338,  -9.534,  -9.724,  -9.929,  -10.164, -10.432, -8.133,  -8.473,
    -8.768,  -9.028,  -9.260,  -9.471,  -9.675,  -9.891,  -10.134, -10.410,
    -8.092,  -8.446,  -8.753,  -9.021,  -9.257,  -9.471,  -9.676,  -9.892,
    -10.136, -10.410, -8.067,  -8.441,  -8.760,  -9.035,  -9.275,  -9.488,
    -9.692,  -9.906,  -10.146, -10.418, -8.128,  -8.514,  -8.835,  -9.107,
    -9.339,  -9.544,  -9.737,  -9.942,  -10.174, -10.438, -8.281,  -8.666,
    -8.976,  -9.230,  -9.442,  -9.627,  -9.804,  -9.993,  -10.212, -10.466,
    -8.479,  -8.845,  -9.129,  -9.355,  -9.541,  -9.704,  -9.862,  -10.036,
    -10.243, -10.488, -8.667,  -9.005,  -9.257,  -9.454,  -9.616,  -9.760,
    -9.902,  -10.065, -10.264, -10.502, -8.731,  -9.056,  -9.295,  -9.481,
    -9.636,  -9.773,  -9.912,  -10.071, -10.267, -10.504, -8.606,  -8.954,
    -9.216,  -9.421,  -9.590,  -9.739,  -9.886,  -10.052, -10.253, -10.493,
    -8.363,  -8.748,  -9.049,  -9.290,  -9.489,  -9.662,  -9.828,  -10.008,
    -10.220, -10.469, -8.123,  -8.532,  -8.863,  -9.137,  -9.366,  -9.565,
    -9.752,  -9.950,  -10.177, -10.437, -7.929,  -8.345,  -8.692,  -8.986,
    -9.238,  -9.460,  -9.668,  -9.884,  -10.126, -10.398, -7.824,  -8.235,
    -8.584,  -8.886,  -9.149,  -9.383,  -9.604,  -9.833,  -10.085, -10.367,
    -7.797,  -8.196,  -8.539,  -8.839,  -9.104,  -9.342,  -9.569,  -9.803,
    -10.061, -10.348, -7.787,  -8.178,  -8.516,  -8.815,  -9.079,  -9.319,
    -9.548,  -9.785,  -10.046, -10.336, -7.897,  -8.272,  -8.595,  -8.879,
    -9.132,  -9.361,  -9.581,  -9.810,  -10.064, -10.349, -8.041,  -8.406,
    -8.714,  -8.982,  -9.218,  -9.432,  -9.638,  -9.855,  -10.099, -10.374,
    -8.096,  -8.469,  -8.777,  -9.041,  -9.270,  -9.476,  -9.674,  -9.883,
    -10.120, -10.390, -8.204,  -8.574,  -8.874,  -9.126,  -9.342,  -9.535,
    -9.721,  -9.920,  -10.148, -10.410, -8.266,  -8.638,  -8.935,  -9.180,
    -9.389,  -9.574,  -9.752,  -9.944,  -10.166, -10.423, -8.192,  -8.578,
    -8.888,  -9.146,  -9.363,  -9.555,  -9.738,  -9.933,  -10.158, -10.416,
    -8.179,  -8.567,  -8.881,  -9.140,  -9.360,  -9.553,  -9.736,  -9.932,
    -10.156, -10.414, -8.226,  -8.606,  -8.912,  -9.165,  -9.380,  -9.568,
    -9.748,  -9.940,  -10.162, -10.418, -8.135,  -8.525,  -8.843,  -9.108,
    -9.334,  -9.533,  -9.720,  -9.919,  -10.145, -10.405, -8.117,  -8.505,
    -8.824,  -9.092,  -9.320,  -9.521,  -9.711,  -9.911,  -10.139, -10.400,
    -8.137,  -8.514,  -8.827,  -9.091,  -9.318,  -9.518,  -9.708,  -9.908,
    -10.136, -10.396, -8.066,  -8.446,  -8.765,  -9.038,  -9.273,  -9.482,
    -9.679,  -9.885,  -10.118, -10.383, -8.107,  -8.477,  -8.788,  -9.055,
    -9.286,  -9.491,  -9.686,  -9.890,  -10.121, -10.384, -8.207,  -8.558,
    -8.853,  -9.107,  -9.327,  -9.523,  -9.710,  -9.908,  -10.134, -10.393,
    -8.298,  -8.637,  -8.919,  -9.162,  -9.371,  -9.558,  -9.737,  -9.928,
    -10.148, -10.403, -8.516,  -8.824,  -9.075,  -9.288,  -9.473,  -9.637,
    -9.798,  -9.974,  -10.182, -10.426, -8.739,  -9.010,  -9.225,  -9.408,
    -9.566,  -9.710,  -9.853,  -10.015, -10.211, -10.447, -9.047,  -9.260,
    -9.422,  -9.559,  -9.682,  -9.797,  -9.918,  -10.063, -10.246, -10.471,
    -9.051,  -9.265,  -9.428,  -9.566,  -9.688,  -9.803,  -9.923,  -10.066,
    -10.248, -10.472, -9.058,  -9.272,  -9.436,  -9.574,  -9.696,  -9.809,
    -9.928,  -10.070, -10.250, -10.473, -9.066,  -9.281,  -9.445,  -9.583,
    -9.704,  -9.816,  -9.934,  -10.074, -10.253, -10.474, -9.076,  -9.291,
    -9.455,  -9.593,  -9.713,  -9.824,  -9.940,  -10.079, -10.256, -10.476,
    -9.088,  -9.302,  -9.466,  -9.604,  -9.723,  -9.833,  -9.947,  -10.084,
    -10.260, -10.478, -9.101,  -9.315,  -9.479,  -9.616,  -9.734,  -9.842,
    -9.955,  -10.090, -10.264, -10.481, -9.116,  -9.329,  -9.492,  -9.628,
    -9.745,  -9.852,  -9.963,  -10.097, -10.269, -10.483, -9.133,  -9.344,
    -9.507,  -9.642,  -9.758,  -9.863,  -9.972,  -10.103, -10.273, -10.486,
    -9.150,  -9.360,  -9.522,  -9.656,  -9.771,  -9.874,  -9.981,  -10.111,
    -10.278, -10.489, -9.170,  -9.377,  -9.538,  -9.671,  -9.784,  -9.886,
    -9.991,  -10.118, -10.284, -10.493, -9.190,  -9.395,  -9.555,  -9.687,
    -9.799,  -9.898,  -10.001, -10.126, -10.290, -10.496, -9.211,  -9.414,
    -9.573,  -9.704,  -9.813,  -9.911,  -10.012, -10.134, -10.296, -10.500,
    -9.234,  -9.434,  -9.592,  -9.721,  -9.829,  -9.924,  -10.023, -10.143,
    -10.302, -10.504, -9.257,  -9.455,  -9.611,  -9.738,  -9.845,  -9.938,
    -10.034, -10.152, -10.309, -10.509, -9.281,  -9.476,  -9.631,  -9.757,
    -9.861,  -9.952,  -10.046, -10.162, -10.316, -10.513, -9.307,  -9.499,
    -9.651,  -9.775,  -9.878,  -9.967,  -10.058, -10.171, -10.323, -10.518,
    -9.333,  -9.521,  -9.672,  -9.794,  -9.895,  -9.982,  -10.071, -10.181,
    -10.330, -10.523, -9.360,  -9.545,  -9.693,  -9.814,  -9.912,  -9.997,
    -10.084, -10.192, -10.338, -10.528, -9.387,  -9.568,  -9.715,  -9.834,
    -9.930,  -10.013, -10.097, -10.202, -10.346, -10.534, -9.416,  -9.593,
    -9.738,  -9.854,  -9.949,  -10.029, -10.111, -10.213, -10.354, -10.539,
    -9.445,  -9.618,  -9.760,  -9.875,  -9.967,  -10.045, -10.124, -10.224,
    -10.363, -10.545, -9.474,  -9.643,  -9.783,  -9.896,  -9.986,  -10.062,
    -10.139, -10.236, -10.371, -10.551, -9.504,  -9.669,  -9.807,  -9.918,
    -10.006, -10.079, -10.153, -10.247, -10.380, -10.557, -9.535,  -9.695,
    -9.830,  -9.939,  -10.025, -10.096, -10.167, -10.259, -10.389, -10.564,
    -9.566,  -9.721,  -9.854,  -9.961,  -10.045, -10.113, -10.182, -10.271,
    -10.399, -10.570, -9.597,  -9.748,  -9.879,  -9.983,  -10.065, -10.131,
    -10.197, -10.284, -10.408, -10.577, -9.629,  -9.775,  -9.903,  -10.006,
    -10.085, -10.149, -10.212, -10.296, -10.418, -10.584, -9.661,  -9.802,
    -9.928,  -10.028, -10.105, -10.167, -10.228, -10.309, -10.427, -10.591,
    -9.694,  -9.829,  -9.953,  -10.051, -10.126, -10.185, -10.243, -10.322,
    -10.437, -10.598, -9.727,  -9.857,  -9.978,  -10.074, -10.147, -10.203,
    -10.259, -10.335, -10.447, -10.605, -9.760,  -9.885,  -10.003, -10.097,
    -10.167, -10.222, -10.275, -10.348, -10.458, -10.612, -9.793,  -9.913,
    -10.028, -10.120, -10.188, -10.240, -10.291, -10.361, -10.468, -10.620,
    -9.827,  -9.941,  -10.054, -10.144, -10.210, -10.259, -10.307, -10.374,
    -10.479, -10.627, -9.861,  -9.969,  -10.079, -10.167, -10.231, -10.278,
    -10.324, -10.388, -10.489, -10.635, -9.895,  -9.998,  -10.105, -10.190,
    -10.252, -10.297, -10.340, -10.401, -10.500, -10.643, -9.929,  -10.026,
    -10.131, -10.214, -10.274, -10.316, -10.357, -10.415, -10.511, -10.651,
    -9.964,  -10.055, -10.157, -10.238, -10.295, -10.335, -10.373, -10.429,
    -10.522, -10.659, -9.998,  -10.083, -10.183, -10.262, -10.317, -10.355,
    -10.390, -10.443, -10.533, -10.667, -10.033, -10.112, -10.209, -10.285,
    -10.338, -10.374, -10.407, -10.457, -10.544, -10.675, -10.068, -10.141,
    -10.235, -10.309, -10.360, -10.393, -10.424, -10.471, -10.555, -10.683,
    -10.103, -10.170, -10.261, -10.333, -10.382, -10.413, -10.441, -10.486,
    -10.566, -10.692, -10.138, -10.198, -10.287, -10.357, -10.404, -10.432,
    -10.458, -10.500, -10.578, -10.700, -10.172, -10.227, -10.313, -10.381,
    -10.426, -10.452, -10.475, -10.514, -10.589, -10.708, -10.207, -10.256,
    -10.339, -10.405, -10.447, -10.472, -10.492, -10.529, -10.601, -10.717,
    -10.243, -10.285, -10.365, -10.429, -10.469, -10.491, -10.509, -10.543,
    -10.612, -10.726, -10.278, -10.313, -10.391, -10.453, -10.491, -10.511,
    -10.526, -10.558, -10.624, -10.734, -10.313, -10.342, -10.417, -10.477,
    -10.513, -10.531, -10.544, -10.572, -10.636, -10.743, -10.348, -10.371,
    -10.443, -10.501, -10.535, -10.551, -10.561, -10.587, -10.647, -10.752,
    -10.382, -10.399, -10.469, -10.525, -10.557, -10.570, -10.578, -10.601,
    -10.659, -10.760, -10.417, -10.428, -10.495, -10.549, -10.579, -10.590,
    -10.596, -10.616, -10.671, -10.769, -10.452, -10.456, -10.521, -10.572,
    -10.601, -10.610, -10.613, -10.631, -10.683, -10.778, -10.487, -10.485,
    -10.546, -10.596, -10.623, -10.630, -10.630, -10.646, -10.694, -10.787,
    -10.522, -10.513, -10.572, -10.620, -10.645, -10.649, -10.648, -10.660,
    -10.706, -10.796, -10.556, -10.541, -10.598, -10.644, -10.666, -10.669,
    -10.665, -10.675, -10.718, -10.805, -10.591, -10.569, -10.623, -10.668,
    -10.688, -10.689, -10.683, -10.690, -10.730, -10.814, -10.625, -10.597,
    -10.649, -10.691, -10.710, -10.709, -10.700, -10.705, -10.742, -10.823,
    -10.659, -10.625, -10.674, -10.715, -10.732, -10.728, -10.717, -10.719,
    -10.754, -10.831, -10.693, -10.653, -10.700, -10.738, -10.754, -10.748,
    -10.735, -10.734, -10.766, -10.840, -10.727, -10.681, -10.725, -10.762,
    -10.775, -10.768, -10.752, -10.749, -10.778, -10.849, -10.761, -10.708,
    -10.750, -10.785, -10.797, -10.787, -10.769, -10.764, -10.790, -10.858,
    -10.794, -10.736, -10.775, -10.809, -10.818, -10.807, -10.787, -10.778,
    -10.802, -10.867, -10.828, -10.763, -10.801, -10.832, -10.840, -10.826,
    -10.804, -10.793, -10.814, -10.876, -10.861, -10.790, -10.825, -10.855,
    -10.861, -10.846, -10.821, -10.808, -10.825, -10.885, -10.894, -10.818,
    -10.850, -10.878, -10.883, -10.866, -10.839, -10.822, -10.837, -10.894,
    -10.926, -10.844, -10.875, -10.902, -10.904, -10.885, -10.856, -10.837,
    -10.849, -10.903, -10.959, -10.871, -10.900, -10.925, -10.926, -10.904,
    -10.873, -10.852, -10.861, -10.912, -10.991, -10.898, -10.924, -10.947,
    -10.947, -10.924, -10.890, -10.866, -10.873, -10.921, -11.023, -10.924,
    -10.949, -10.970, -10.968, -10.943, -10.907, -10.881, -10.884, -10.930,
    -11.055, -10.951, -10.973, -10.993, -10.989, -10.962, -10.924, -10.895,
    -10.896, -10.939, -11.087, -10.977, -10.997, -11.016, -11.010, -10.982,
    -10.941, -10.910, -10.908, -10.947, -11.118, -11.003, -11.021, -11.038,
    -11.031, -11.001, -10.958, -10.924, -10.919, -10.956, -11.149, -11.029,
    -11.045, -11.061, -11.052, -11.020, -10.975, -10.939, -10.931, -10.965,
    -11.179, -11.054, -11.069, -11.083, -11.073, -11.039, -10.992, -10.953,
    -10.943, -10.974, -11.210, -11.080, -11.093, -11.106, -11.094, -11.058,
    -11.009, -10.967, -10.954, -10.982, -11.240, -11.105, -11.116, -11.128,
    -11.115, -11.077, -11.026, -10.981, -10.965, -10.991, -11.270, -11.130,
    -11.140, -11.150, -11.136, -11.096, -11.042, -10.995, -10.977, -10.999,
    -11.299, -11.155, -11.163, -11.172, -11.156, -11.115, -11.059, -11.010,
    -10.988, -11.008, -11.328, -11.180, -11.187, -11.194, -11.177, -11.133,
    -11.075, -11.024, -10.999, -11.016, -11.357, -11.205, -11.210, -11.216,
    -11.197, -11.152, -11.092, -11.038, -11.011, -11.025, -11.386, -11.229,
    -11.233, -11.238, -11.217, -11.171, -11.108, -11.051, -11.022, -11.033,
    -11.414, -11.254, -11.256, -11.259, -11.238, -11.189, -11.125, -11.065,
    -11.033, -11.041, -11.442, -11.278, -11.278, -11.281, -11.258, -11.208,
    -11.141, -11.079, -11.044, -11.049, -11.469, -11.302, -11.301, -11.303,
    -11.278, -11.226, -11.157, -11.092, -11.054, -11.057, -11.497, -11.325,
    -11.324, -11.324, -11.298, -11.244, -11.173, -11.106, -11.065, -11.065,
    -11.524, -11.349, -11.346, -11.345, -11.318, -11.263, -11.189, -11.119,
    -11.076, -11.073, -11.550, -11.372, -11.368, -11.366, -11.338, -11.281,
    -11.205, -11.133, -11.086, -11.081, -11.576, -11.396, -11.390, -11.388,
    -11.358, -11.299, -11.221, -11.146, -11.097, -11.089, -11.602, -11.419,
    -11.412, -11.409, -11.378, -11.317, -11.237, -11.159, -11.107, -11.097,
    -11.627, -11.441, -11.434, -11.429, -11.397, -11.335, -11.252, -11.172,
    -11.118, -11.104, -11.652, -11.464, -11.456, -11.450, -11.417, -11.353,
    -11.268, -11.185, -11.128, -11.112, -11.677, -11.487, -11.478, -11.471,
    -11.436, -11.370, -11.283, -11.198, -11.138, -11.119, -11.702, -11.509,
    -11.499, -11.492, -11.456, -11.388, -11.299, -11.211, -11.148, -11.127,
    -11.726, -11.531, -11.520, -11.512, -11.475, -11.406, -11.314, -11.223,
    -11.158, -11.134, -11.749, -11.553, -11.542, -11.532, -11.494, -11.423,
    -11.329, -11.236, -11.168, -11.141, -11.772, -11.575, -11.563, -11.553,
    -11.513, -11.440, -11.344, -11.248, -11.177, -11.148, -11.795, -11.596,
    -11.584, -11.573, -11.532, -11.458, -11.359, -11.260, -11.187, -11.155,
    -11.818, -11.618, -11.605, -11.593, -11.551, -11.475, -11.374, -11.272,
    -11.196, -11.162, -11.840, -11.639, -11.625, -11.613, -11.570, -11.492,
    -11.389, -11.284, -11.205, -11.169, -11.862, -11.660, -11.646, -11.633,
    -11.589, -11.509, -11.403, -11.296, -11.214, -11.175, -11.883, -11.681,
    -11.666, -11.653, -11.608, -11.526, -11.418, -11.308, -11.223, -11.182,
    -11.905, -11.702, -11.687, -11.672, -11.626, -11.543, -11.432, -11.319,
    -11.232, -11.188, -11.925, -11.722, -11.707, -11.692, -11.645, -11.559,
    -11.446, -11.331, -11.241, -11.195, -11.946, -11.743, -11.727, -11.711,
    -11.663, -11.576, -11.460, -11.342, -11.249, -11.201, -11.966, -11.763,
    -11.747, -11.731, -11.681, -11.592, -11.474, -11.353, -11.258, -11.207,
    -11.986, -11.783, -11.767, -11.750, -11.699, -11.608, -11.488, -11.364,
    -11.266, -11.213, -12.005, -11.803, -11.786, -11.769, -11.717, -11.625,
    -11.501, -11.375, -11.274, -11.219, -12.024, -11.822, -11.806, -11.788,
    -11.735, -11.641, -11.515, -11.385, -11.282, -11.224, -12.043, -11.842,
    -11.825, -11.807, -11.753, -11.657, -11.528, -11.396, -11.290, -11.230,
    -12.061, -11.861, -11.845, -11.826, -11.771, -11.672, -11.541, -11.406,
    -11.298, -11.235, -12.079, -11.880, -11.864, -11.845, -11.789, -11.688,
    -11.554, -11.416, -11.305, -11.240, -12.097, -11.899, -11.883, -11.864,
    -11.806, -11.704, -11.567, -11.426, -11.312, -11.245, -12.115, -11.918,
    -11.902, -11.882, -11.824, -11.719, -11.580, -11.436, -11.319, -11.250,
    -12.132, -11.937, -11.921, -11.901, -11.841, -11.734, -11.592, -11.445,
    -11.326, -11.255, -12.149, -11.955, -11.940, -11.919, -11.858, -11.749,
    -11.604, -11.454, -11.333, -11.260, -12.165, -11.974, -11.958, -11.937,
    -11.875, -11.764, -11.616, -11.463, -11.339, -11.265, -12.182, -11.992,
    -11.977, -11.955, -11.892, -11.779, -11.628, -11.472, -11.346, -11.269,
    -12.198, -12.010, -11.995, -11.973, -11.909, -11.794, -11.640, -11.481,
    -11.352, -11.273, -12.213, -12.028, -12.013, -11.991, -11.926, -11.808,
    -11.651, -11.489, -11.358, -11.277, -12.229, -12.045, -12.031, -12.009,
    -11.943, -11.822, -11.662, -11.498, -11.364, -11.281, -12.244, -12.063,
    -12.050, -12.027, -11.959, -11.837, -11.673, -11.506, -11.369, -11.285,
    -12.259, -12.081, -12.067, -12.044, -11.975, -11.850, -11.684, -11.513,
    -11.375, -11.289, -12.274, -12.098, -12.085, -12.062, -11.992, -11.864,
    -11.695, -11.521, -11.380, -11.292, -12.288, -12.115, -12.103, -12.079,
    -12.008, -11.878, -11.705, -11.528, -11.385, -11.296, -12.303, -12.132,
    -12.120, -12.097, -12.024, -11.891, -11.715, -11.535, -11.390, -11.299,
    -12.317, -12.149, -12.138, -12.114, -12.040, -11.904, -11.725, -11.542,
    -11.394, -11.302, -12.330, -12.166, -12.155, -12.131, -12.055, -11.917,
    -11.735, -11.549, -11.399, -11.305, -12.344, -12.182, -12.173, -12.148,
    -12.071, -11.930, -11.744, -11.555, -11.403, -11.308, -12.357, -12.199,
    -12.190, -12.165, -12.086, -11.943, -11.753, -11.561, -11.407, -11.310,
    -12.371, -12.215, -12.207, -12.182, -12.102, -11.955, -11.762, -11.567,
    -11.411, -11.313, -12.383, -12.231, -12.224, -12.198, -12.117, -11.967,
    -11.770, -11.572, -11.414, -11.315, -12.396, -12.247, -12.240, -12.215,
    -12.132, -11.979, -11.779, -11.578, -11.418, -11.317, -12.409, -12.263,
    -12.257, -12.231, -12.146, -11.991, -11.786, -11.583, -11.421, -11.319,
    -12.421, -12.279, -12.274, -12.247, -12.161, -12.002, -11.794, -11.587,
    -11.424, -11.321, -12.434, -12.295, -12.290, -12.264, -12.176, -12.013,
    -11.802, -11.592, -11.426, -11.323, -12.446, -12.311, -12.307, -12.280,
    -12.190, -12.024, -11.809, -11.596, -11.429, -11.325, -12.458, -12.326,
    -12.323, -12.296, -12.204, -12.035, -11.816, -11.600, -11.431, -11.326,
    -12.469, -12.342, -12.339, -12.311, -12.218, -12.045, -11.822, -11.604,
    -11.433, -11.327, -12.481, -12.357, -12.355, -12.327, -12.232, -12.055,
    -11.828, -11.607, -11.435, -11.329, -12.492, -12.372, -12.371, -12.343,
    -12.245, -12.065, -11.834, -11.610, -11.437, -11.330, -12.504, -12.387,
    -12.387, -12.358, -12.259, -12.075, -11.840, -11.613, -11.438, -11.331,
    -12.515, -12.402, -12.403, -12.374, -12.272, -12.084, -11.845, -11.616,
    -11.439, -11.331, -12.526, -12.417, -12.419, -12.389, -12.285, -12.093,
    -11.850, -11.618, -11.441, -11.332, -12.537, -12.432, -12.434, -12.404,
    -12.297, -12.102, -11.855, -11.620, -11.441, -11.332, -12.548, -12.446,
    -12.450, -12.419, -12.310, -12.110, -11.859, -11.622, -11.442, -11.333,
    -12.558, -12.461, -12.465, -12.434, -12.322, -12.118, -11.863, -11.623,
    -11.443, -11.333, -12.569, -12.475, -12.480, -12.449, -12.334, -12.126,
    -11.866, -11.625, -11.443, -11.333, -12.580, -12.490, -12.495, -12.463,
    -12.346, -12.133, -11.870, -11.626, -11.443, -11.333, -12.590, -12.504,
    -12.511, -12.478, -12.358, -12.140, -11.873, -11.626, -11.443, -11.333,
    -12.600, -12.518, -12.526, -12.492, -12.369, -12.147, -11.876, -11.627,
    -11.443, -11.333, -12.611, -12.532, -12.541, -12.506, -12.381, -12.154,
    -11.878, -11.627, -11.443, -11.333, -12.621, -12.546, -12.555, -12.520,
    -12.392, -12.160, -11.880, -11.627, -11.442, -11.333, -12.631, -12.560,
    -12.570, -12.534, -12.402, -12.166, -11.882, -11.627, -11.441, -11.332,
    -12.641, -12.574, -12.585, -12.548, -12.413, -12.171, -11.883, -11.627,
    -11.441, -11.332, -12.651, -12.588, -12.599, -12.561, -12.423, -12.176,
    -11.884, -11.626, -11.440, -11.331, -12.660, -12.601, -12.614, -12.575,
    -12.433, -12.181, -11.885, -11.625, -11.439, -11.330, -12.670, -12.615,
    -12.628, -12.588, -12.442, -12.185, -11.886, -11.624, -11.437, -11.330,
    -12.680, -12.628, -12.642, -12.601, -12.452, -12.190, -11.886, -11.623,
    -11.436, -11.329, -12.690, -12.642, -12.657, -12.614, -12.461, -12.193,
    -11.886, -11.622, -11.435, -11.328, -12.699, -12.655, -12.671, -12.627,
    -12.470, -12.197, -11.886, -11.620, -11.433, -11.327, -12.709, -12.668,
    -12.685, -12.640, -12.478, -12.200, -11.886, -11.618, -11.432, -11.326,
    -12.718, -12.682, -12.699, -12.652, -12.487, -12.203, -11.885, -11.616,
    -11.430, -11.325, -12.728, -12.695, -12.712, -12.665, -12.495, -12.206,
    -11.884, -11.615, -11.428, -11.324, -12.737, -12.708, -12.726, -12.677,
    -12.503, -12.208, -11.883, -11.612, -11.427, -11.323, -12.747, -12.721,
    -12.740, -12.689, -12.510, -12.211, -11.882, -11.610, -11.425, -11.322,
    -12.756, -12.734, -12.753, -12.701, -12.518, -12.212, -11.881, -11.608,
    -11.423, -11.321, -12.765, -12.747, -12.767, -12.713, -12.525, -12.214,
    -11.879, -11.606, -11.421, -11.320, -12.775, -12.759, -12.780, -12.724,
    -12.532, -12.216, -11.878, -11.603, -11.419, -11.319, -12.784, -12.772,
    -12.793, -12.736, -12.538, -12.217, -11.876, -11.601, -11.417, -11.318,
    -12.793, -12.785, -12.807, -12.747, -12.545, -12.218, -11.874, -11.599,
    -11.416, -11.317, -12.802, -12.797, -12.820, -12.758, -12.551, -12.219,
    -11.872, -11.596, -11.414, -11.316, -12.811, -12.810, -12.833, -12.769,
    -12.557, -12.220, -11.870, -11.594, -11.412, -11.315, -12.821, -12.823,
    -12.846, -12.780, -12.563, -12.221, -11.868, -11.592, -11.411, -11.314,
    -12.830, -12.835, -12.859, -12.791, -12.569, -12.221, -11.867, -11.589,
    -11.409, -11.314, -12.839, -12.848, -12.872, -12.802, -12.574, -12.222,
    -11.865, -11.587, -11.407, -11.313, -12.848, -12.860, -12.884, -12.812,
    -12.580, -12.222, -11.863, -11.585, -11.406, -11.312, -12.857, -12.872,
    -12.897, -12.823, -12.585, -12.223, -11.861, -11.583, -11.405, -11.312,
    -12.866, -12.885, -12.910, -12.833, -12.591, -12.224, -11.860, -11.582,
    -11.404, -11.312, -12.875, -12.896, -12.918, -12.830, -12.568, -12.188,
    -11.824, -11.552, -11.382, -11.297, -12.884, -12.908, -12.927, -12.827,
    -12.546, -12.157, -11.793, -11.527, -11.363, -11.284, -12.893, -12.919,
    -12.935, -12.820, -12.520, -12.120, -11.757, -11.497, -11.341, -11.269,
    -12.902, -12.931, -12.945, -12.822, -12.511, -12.104, -11.741, -11.484,
    -11.332, -11.263, -12.911, -12.943, -12.954, -12.823, -12.500, -12.087,
    -11.724, -11.471, -11.322, -11.256, -12.920, -12.954, -12.965, -12.829,
    -12.500, -12.084, -11.720, -11.468, -11.320, -11.255, -12.930, -12.967,
    -12.979, -12.846, -12.520, -12.103, -11.737, -11.481, -11.331, -11.263,
    -12.939, -12.978, -12.991, -12.854, -12.523, -12.104, -11.737, -11.482,
    -11.332, -11.265, -12.948, -12.990, -13.003, -12.866, -12.534, -12.113,
    -11.745, -11.488, -11.337, -11.269, -12.957, -13.003, -13.018, -12.885,
    -12.556, -12.134, -11.763, -11.503, -11.349, -11.277, -12.966, -13.014,
    -13.028, -12.891, -12.556, -12.131, -11.759, -11.500, -11.347, -11.276,
    -12.975, -13.026, -13.041, -12.904, -12.568, -12.141, -11.768, -11.507,
    -11.352, -11.280, -12.984, -13.038, -13.056, -12.921, -12.588, -12.160,
    -11.784, -11.519, -11.362, -11.287, -12.993, -13.050, -13.065, -12.923,
    -12.579, -12.146, -11.769, -11.508, -11.353, -11.282, -13.002, -13.061,
    -13.072, -12.917, -12.557, -12.117, -11.742, -11.485, -11.337, -11.270,
    -13.011, -13.072, -13.079, -12.912, -12.537, -12.089, -11.715, -11.463,
    -11.320, -11.259, -13.020, -13.082, -13.082, -12.892, -12.494, -12.038,
    -11.668, -11.425, -11.292, -11.239, -13.029, -13.092, -13.085, -12.878,
    -12.463, -12.002, -11.636, -11.399, -11.273, -11.225, -13.039, -13.104,
    -13.096, -12.885, -12.467, -12.006, -11.640, -11.403, -11.276, -11.228,
    -13.048, -13.115, -13.108, -12.898, -12.481, -12.021, -11.654, -11.416,
    -11.286, -11.236, -13.057, -13.127, -13.124, -12.925, -12.518, -12.059,
    -11.689, -11.444, -11.309, -11.253, -13.066, -13.140, -13.145, -12.963,
    -12.572, -12.116, -11.739, -11.485, -11.339, -11.275, -13.076, -13.153,
    -13.165, -12.999, -12.622, -12.167, -11.783, -11.520, -11.365, -11.293,
    -13.085, -13.166, -13.184, -13.029, -12.662, -12.207, -11.816, -11.545,
    -11.384, -11.306, -13.094, -13.178, -13.198, -13.046, -12.679, -12.222,
    -11.828, -11.554, -11.391, -11.311, -13.103, -13.189, -13.206, -13.045,
    -12.669, -12.209, -11.816, -11.545, -11.384, -11.306, -13.113, -13.199,
    -13.209, -13.029, -12.634, -12.169, -11.781, -11.518, -11.365, -11.293,
    -13.122, -13.208, -13.208, -13.003, -12.584, -12.113, -11.732, -11.480,
    -11.337, -11.274, -13.131, -13.217, -13.205, -12.972, -12.527, -12.049,
    -11.675, -11.433, -11.302, -11.249, -13.140, -13.227, -13.206, -12.950,
    -12.485, -12.000, -11.630, -11.397, -11.274, -11.229, -13.149, -13.238,
    -13.212, -12.944, -12.465, -11.975, -11.605, -11.376, -11.258, -11.218,
    -13.159, -13.249, -13.221, -12.943, -12.455, -11.961, -11.591, -11.364,
    -11.249, -11.211, -13.168, -13.261, -13.239, -12.971, -12.488, -11.992,
    -11.617, -11.384, -11.264, -11.222, -13.178, -13.274, -13.261, -13.011,
    -12.541, -12.046, -11.664, -11.422, -11.293, -11.242, -13.188, -13.287,
    -13.278, -13.038, -12.577, -12.082, -11.696, -11.448, -11.313, -11.257,
    -13.197, -13.299, -13.297, -13.070, -12.619, -12.124, -11.733, -11.478,
    -11.336, -11.273, -13.207, -13.311, -13.313, -13.094, -12.650, -12.156,
    -11.761, -11.500, -11.353, -11.285, -13.216, -13.322, -13.321, -13.095,
    -12.645, -12.150, -11.755, -11.497, -11.350, -11.284, -13.226, -13.333,
    -13.330, -13.101, -12.647, -12.150, -11.756, -11.497, -11.351, -11.284,
    -13.235, -13.345, -13.344, -13.117, -12.664, -12.166, -11.769, -11.507,
    -11.358, -11.289, -13.245, -13.355, -13.350, -13.111, -12.648, -12.147,
    -11.752, -11.494, -11.349, -11.283, -13.254, -13.366, -13.358, -13.113,
    -12.644, -12.141, -11.747, -11.490, -11.346, -11.280, -13.264, -13.377,
    -13.370, -13.123, -12.651, -12.145, -11.749, -11.492, -11.347, -11.281,
    -13.274, -13.387, -13.375, -13.116, -12.633, -12.124, -11.730, -11.477,
    -11.336, -11.273, -13.283, -13.399, -13.387, -13.128, -12.643, -12.132,
    -11.736, -11.481, -11.339, -11.275, -13.293, -13.411, -13.404, -13.153,
    -12.672, -12.160, -11.759, -11.499, -11.352, -11.284, -13.303, -13.423,
    -13.420, -13.175, -12.699, -12.185, -11.780, -11.515, -11.364, -11.292,
    -13.313, -13.437, -13.443, -13.219, -12.757, -12.243, -11.829, -11.554,
    -11.392, -11.312, -13.323, -13.450, -13.465, -13.259, -12.812, -12.297,
    -11.875, -11.589, -11.418, -11.329, -13.333, -13.462, -13.482, -13.288,
    -12.849, -12.333, -11.905, -11.612, -11.435, -11.341, -13.343, -13.474,
    -13.499, -13.315, -12.884, -12.368, -11.934, -11.634, -11.451, -11.352,
    -13.352, -13.485, -13.510, -13.326, -12.894, -12.376, -11.941, -11.640,
    -11.455, -11.354, -13.362, -13.497, -13.522, -13.337, -12.905, -12.385,
    -11.949, -11.646, -11.460, -11.357, -13.372, -13.508, -13.533, -13.349,
    -12.915, -12.395, -11.957, -11.653, -11.465, -11.360, -13.382, -13.519,
    -13.545, -13.360, -12.926, -12.405, -11.965, -11.660, -11.470, -11.364,
    -13.392, -13.530, -13.557, -13.372, -12.938, -12.415, -11.974, -11.667,
    -11.475, -11.367, -13.402, -13.542, -13.568, -13.384, -12.949, -12.426,
    -11.983, -11.674, -11.481, -11.371, -13.412, -13.553, -13.580, -13.396,
    -12.961, -12.437, -11.993, -11.682, -11.487, -11.375, -13.422, -13.564,
    -13.592, -13.408, -12.974, -12.448, -12.003, -11.690, -11.493, -11.379,
    -13.432, -13.576, -13.603, -13.420, -12.986, -12.460, -12.013, -11.699,
    -11.499, -11.383, -13.442, -13.587, -13.615, -13.433, -12.999, -12.472,
    -12.024, -11.707, -11.505, -11.387, -13.452, -13.598, -13.627, -13.445,
    -13.012, -12.485, -12.035, -11.716, -11.512, -11.391, -13.463, -13.609,
    -13.639, -13.458, -13.026, -12.498, -12.046, -11.726, -11.519, -11.396,
    -13.473, -13.621, -13.651, -13.471, -13.039, -12.511, -12.058, -11.735,
    -11.526, -11.401, -13.483, -13.632, -13.662, -13.484, -13.053, -12.524,
    -12.070, -11.745, -11.534, -11.406, -13.493, -13.643, -13.674, -13.497,
    -13.067, -12.538, -12.082, -11.755, -11.541, -11.410, -13.503, -13.655,
    -13.686, -13.510, -13.081, -12.552, -12.094, -11.765, -11.549, -11.416,
    -13.514, -13.666, -13.698, -13.523, -13.095, -12.566, -12.107, -11.775,
    -11.557, -11.421, -13.524, -13.677, -13.710, -13.536, -13.110, -12.580,
    -12.120, -11.786, -11.565, -11.426, -13.534, -13.689, -13.722, -13.549,
    -13.124, -12.595, -12.133, -11.797, -11.573, -11.431, -13.544, -13.700,
    -13.734, -13.563, -13.139, -12.609, -12.146, -11.808, -11.581, -11.437,
    -13.555, -13.711, -13.746, -13.576, -13.154, -12.624, -12.159, -11.819,
    -11.590, -11.442, -13.565, -13.723, -13.757, -13.589, -13.169, -12.639,
    -12.173, -11.830, -11.598, -11.448, -13.576, -13.734, -13.769, -13.602,
    -13.184, -12.654, -12.187, -11.842, -11.607, -11.454, -13.586, -13.745,
    -13.781, -13.616, -13.199, -12.669, -12.201, -11.853, -11.616, -11.459,
    -13.596, -13.757, -13.793, -13.629, -13.214, -12.685, -12.215, -11.865,
    -11.624, -11.465, -13.607, -13.768, -13.805, -13.643, -13.229, -12.700,
    -12.229, -11.877, -11.633, -11.471, -13.617, -13.779, -13.817, -13.656,
    -13.244, -12.716, -12.243, -11.889, -11.642, -11.477, -13.628, -13.791,
    -13.829, -13.669, -13.260, -12.731, -12.257, -11.901, -11.652, -11.483,
    -13.638, -13.802, -13.841, -13.683, -13.275, -12.747, -12.272, -11.913,
    -11.661, -11.489, -13.649, -13.813, -13.852, -13.696, -13.290, -12.763,
    -12.286, -11.925, -11.670, -11.495, -13.660, -13.825, -13.864, -13.709,
    -13.306, -12.779, -12.301, -11.938, -11.679, -11.501, -13.670, -13.836,
    -13.876, -13.723, -13.321, -12.795, -12.316, -11.950, -11.689, -11.507,
    -13.681, -13.847, -13.888, -13.736, -13.336, -12.811, -12.330, -11.962,
    -11.698, -11.513, -13.691, -13.859, -13.900, -13.749, -13.352, -12.826,
    -12.345, -11.975, -11.708, -11.519, -13.702, -13.870, -13.911, -13.763,
    -13.367, -12.842, -12.360, -11.987, -11.717, -11.525, -13.713, -13.881,
    -13.923, -13.776, -13.382, -12.859, -12.375, -12.000, -11.727, -11.531,
    -13.724, -13.893, -13.935, -13.789, -13.398, -12.875, -12.390, -12.013,
    -11.736, -11.538, -13.734, -13.904, -13.947, -13.802, -13.413, -12.891,
    -12.405, -12.025, -11.746, -11.544, -13.745, -13.915, -13.958, -13.815,
    -13.428, -12.907, -12.420, -12.038, -11.755, -11.550, -13.756, -13.927,
    -13.970, -13.829, -13.444, -12.923, -12.435, -12.051, -11.765, -11.556,
    -13.767, -13.938, -13.982, -13.842, -13.459, -12.939, -12.450, -12.063,
    -11.774, -11.562, -13.777, -13.949, -13.993, -13.855, -13.474, -12.955,
    -12.465, -12.076, -11.784, -11.568, -13.788, -13.961, -14.005, -13.868,
    -13.489, -12.971, -12.480, -12.089, -11.793, -11.574, -13.799, -13.972,
    -14.017, -13.881, -13.504, -12.987, -12.495, -12.102, -11.803, -11.580,
    -13.810, -13.983, -14.028, -13.894, -13.519, -13.003, -12.510, -12.114,
    -11.812, -11.586, -13.821, -13.995, -14.040, -13.907, -13.534, -13.019,
    -12.525, -12.127, -11.822, -11.592, -13.832, -14.006, -14.051, -13.919,
    -13.549, -13.035, -12.540, -12.140, -11.831, -11.597, -13.843, -14.017,
    -14.063, -13.932, -13.564, -13.051, -12.555, -12.152, -11.840, -11.603,
    -13.854, -14.029, -14.075, -13.945, -13.579, -13.067, -12.570, -12.165,
    -11.850, -11.609, -13.865, -14.040, -14.086, -13.958, -13.594, -13.082,
    -12.585, -12.177, -11.859, -11.614, -13.876, -14.052, -14.098, -13.971,
    -13.609, -13.098, -12.600, -12.190, -11.868, -11.620, -13.887, -14.063,
    -14.109, -13.983, -13.624, -13.114, -12.615, -12.202, -11.877, -11.625,
    -13.898, -14.074, -14.121, -13.996, -13.638, -13.130, -12.630, -12.215,
    -11.886, -11.631, -13.909, -14.086, -14.132, -14.008, -13.653, -13.145,
    -12.645, -12.227, -11.895, -11.636, -13.920, -14.097, -14.144, -14.021,
    -13.668, -13.161, -12.659, -12.239, -11.904, -11.641, -13.931, -14.109,
    -14.155, -14.034, -13.682, -13.177, -12.674, -12.252, -11.913, -11.646,
    -13.942, -14.120, -14.167, -14.046, -13.697, -13.192, -12.689, -12.264,
    -11.921, -11.651, -13.954, -14.131, -14.178, -14.058, -13.711, -13.208,
    -12.703, -12.276, -11.930, -11.656, -13.965, -14.143, -14.189, -14.071,
    -13.725, -13.223, -12.718, -12.288, -11.938, -11.661, -13.976, -14.154,
    -14.201, -14.083, -13.740, -13.238, -12.732, -12.300, -11.947, -11.666,
    -13.987, -14.166, -14.212, -14.096, -13.754, -13.253, -12.746, -12.311,
    -11.955, -11.670, -13.998, -14.177, -14.224, -14.108, -13.768, -13.269,
    -12.761, -12.323, -11.963, -11.675, -14.010, -14.188, -14.235, -14.120,
    -13.782, -13.284, -12.775, -12.335, -11.971, -11.679, -14.021, -14.200,
    -14.246, -14.132, -13.796, -13.299, -12.789, -12.346, -11.979, -11.683,
    -14.032, -14.211, -14.258, -14.144, -13.810, -13.314, -12.803, -12.357,
    -11.987, -11.687, -14.043, -14.223, -14.269, -14.157, -13.824, -13.329,
    -12.817, -12.369, -11.994, -11.691, -14.055, -14.234, -14.281, -14.169,
    -13.838, -13.343, -12.831, -12.380, -12.002, -11.695, -14.066, -14.246,
    -14.292, -14.181, -13.852, -13.358, -12.844, -12.391, -12.009, -11.699,
    -14.077, -14.257, -14.303, -14.193, -13.865, -13.373, -12.858, -12.402,
    -12.016, -11.702, -14.089, -14.268, -14.315, -14.205, -13.879, -13.387,
    -12.872, -12.412, -12.023, -11.706, -14.100, -14.280, -14.326, -14.217,
    -13.892, -13.402, -12.885, -12.423, -12.030, -11.709, -14.112, -14.291,
    -14.337, -14.229, -13.906, -13.416, -12.898, -12.433, -12.037, -11.712,
    -14.123, -14.303, -14.348, -14.240, -13.919, -13.431, -12.912, -12.444,
    -12.043, -11.715, -14.135, -14.314, -14.360, -14.252, -13.933, -13.445,
    -12.925, -12.454, -12.050, -11.718, -14.146, -14.326, -14.371, -14.264,
    -13.946, -13.459, -12.938, -12.464, -12.056, -11.720, -14.158, -14.337,
    -14.382, -14.276, -13.959, -13.473, -12.950, -12.474, -12.062, -11.723,
    -14.169, -14.349, -14.393, -14.288, -13.973, -13.487, -12.963, -12.483,
    -12.067, -11.725, -14.181, -14.360, -14.405, -14.299, -13.986, -13.501,
    -12.976, -12.493, -12.073, -11.727, -14.192, -14.372, -14.416, -14.311,
    -13.999, -13.515, -12.988, -12.502, -12.078, -11.729, -14.204, -14.383,
    -14.427, -14.323, -14.012, -13.528, -13.000, -12.511, -12.084, -11.730,
    -14.215, -14.395, -14.438, -14.334, -14.025, -13.542, -13.013, -12.520,
    -12.089, -11.732, -14.227, -14.406, -14.450, -14.346, -14.038, -13.556,
    -13.025, -12.529, -12.093, -11.733, -14.238, -14.418, -14.461, -14.357,
    -14.050, -13.569, -13.036, -12.537, -12.098, -11.735, -14.250, -14.429,
    -14.472, -14.369, -14.063, -13.582, -13.048, -12.545, -12.102, -11.736,
    -14.262, -14.441, -14.483, -14.380, -14.076, -13.595, -13.060, -12.553,
    -12.106, -11.736, -14.273, -14.452, -14.494, -14.392, -14.088, -13.608,
    -13.071, -12.561, -12.110, -11.737, -14.285, -14.464, -14.506, -14.403,
    -14.101, -13.621, -13.082, -12.569, -12.114, -11.737, -14.297, -14.475,
    -14.517, -14.415, -14.114, -13.634, -13.093, -12.576, -12.117, -11.738,
    -14.308, -14.487, -14.528, -14.426, -14.126, -13.647, -13.104, -12.584,
    -12.120, -11.738, -14.320, -14.498, -14.539, -14.438, -14.138, -13.660,
    -13.115, -12.591, -12.123, -11.738, -14.332, -14.510, -14.550, -14.449,
    -14.151, -13.672, -13.125, -12.597, -12.126, -11.737, -14.344, -14.521,
    -14.561, -14.460, -14.163, -13.684, -13.136, -12.604, -12.129, -11.737,
    -14.355, -14.533, -14.572, -14.471, -14.175, -13.697, -13.146, -12.610,
    -12.131, -11.736, -14.367, -14.545, -14.584, -14.483, -14.187, -13.709,
    -13.156, -12.616, -12.133, -11.735, -14.379, -14.556, -14.595, -14.494,
    -14.199, -13.721, -13.165, -12.622, -12.134, -11.734, -14.391, -14.568,
    -14.606, -14.505, -14.211, -13.733, -13.175, -12.627, -12.136, -11.733,
    -14.403, -14.579, -14.617, -14.516, -14.223, -13.744, -13.184, -12.632,
    -12.137, -11.731, -14.414, -14.591, -14.628, -14.527, -14.235, -13.756,
    -13.193, -12.637, -12.138, -11.729, -14.426, -14.602, -14.639, -14.539,
    -14.247, -13.767, -13.202, -12.642, -12.139, -11.728, -14.438, -14.614,
    -14.650, -14.550, -14.258, -13.779, -13.210, -12.646, -12.139, -11.725,
    -14.450, -14.626, -14.661, -14.561, -14.270, -13.790, -13.219, -12.650,
    -12.139, -11.723, -14.462, -14.637, -14.673, -14.572, -14.281, -13.801,
    -13.227, -12.654, -12.139, -11.721, -14.474, -14.649, -14.684, -14.583,
    -14.293, -13.812, -13.235, -12.657, -12.139, -11.718, -14.486, -14.660,
    -14.695, -14.594, -14.304, -13.823, -13.242, -12.661, -12.138, -11.715,
    -14.498, -14.672, -14.706, -14.605, -14.316, -13.833, -13.250, -12.663,
    -12.137, -11.712, -14.510, -14.684, -14.717, -14.616, -14.327, -13.843,
    -13.257, -12.666, -12.136, -11.709, -14.522, -14.695, -14.728, -14.627,
    -14.338, -13.854, -13.263, -12.668, -12.135, -11.705, -14.534, -14.707,
    -14.739, -14.638, -14.349, -13.864, -13.270, -12.670, -12.133, -11.702,
    -14.546, -14.719, -14.750, -14.648, -14.360, -13.874, -13.276, -12.672,
    -12.131, -11.698, -14.558, -14.730, -14.761, -14.659, -14.371, -13.883,
    -13.282, -12.673, -12.129, -11.694, -14.570, -14.742, -14.772, -14.670,
    -14.382, -13.893, -13.288, -12.674, -12.127, -11.690, -14.582, -14.754,
    -14.783, -14.681, -14.393, -13.902, -13.293, -12.675, -12.124, -11.686,
    -14.594, -14.765, -14.795, -14.692, -14.404, -13.911, -13.298, -12.675,
    -12.121, -11.681, -14.606, -14.777, -14.806, -14.703, -14.414, -13.920,
    -13.303, -12.675, -12.118, -11.677, -14.618, -14.789, -14.817, -14.713,
    -14.425, -13.929, -13.307, -12.675, -12.114, -11.672, -14.630, -14.800,
    -14.828, -14.724, -14.435, -13.937, -13.311, -12.674, -12.111, -11.667,
    -14.642, -14.812, -14.839, -14.735, -14.446, -13.946, -13.315, -12.673,
    -12.107, -11.662, -14.654, -14.824, -14.850, -14.745, -14.456, -13.954,
    -13.319, -12.672, -12.102, -11.656, -14.666, -14.835, -14.861, -14.756,
    -14.466, -13.961, -13.322, -12.671, -12.098, -11.651, -14.678, -14.847,
    -14.872, -14.767, -14.476, -13.969, -13.325, -12.669, -12.093, -11.645,
    -14.691, -14.859, -14.883, -14.777, -14.486, -13.976, -13.327, -12.666,
    -12.088, -11.640, -14.703, -14.870, -14.894, -14.788, -14.496, -13.983,
    -13.329, -12.664, -12.083, -11.634, -14.715, -14.882, -14.905, -14.798,
    -14.506, -13.990, -13.331, -12.661, -12.078, -11.628, -14.727, -14.894,
    -14.916, -14.809, -14.516, -13.997, -13.332, -12.658, -12.072, -11.622,
    -14.739, -14.906, -14.927, -14.819, -14.525, -14.003, -13.333, -12.654,
    -12.066, -11.616, -14.752, -14.917, -14.938, -14.830, -14.535, -14.009,
    -13.334, -12.651, -12.060, -11.609, -14.764, -14.929, -14.949, -14.840,
    -14.544, -14.015, -13.334, -12.646, -12.054, -11.603, -14.776, -14.941,
    -14.960, -14.851, -14.553, -14.021, -13.334, -12.642, -12.048, -11.596,
    -14.788, -14.952, -14.971, -14.861, -14.563, -14.026, -13.334, -12.637,
    -12.041, -11.590, -14.800, -14.964, -14.982, -14.872, -14.572, -14.031,
    -13.333, -12.632, -12.035, -11.583, -14.813, -14.976, -14.993, -14.882,
    -14.580, -14.036, -13.332, -12.627, -12.028, -11.576, -14.825, -14.988,
    -15.005, -14.892, -14.589, -14.040, -13.331, -12.622, -12.021, -11.569,
    -14.837, -14.999, -15.016, -14.903, -14.598, -14.045, -13.329, -12.616,
    -12.013, -11.562, -14.850, -15.011, -15.027, -14.913, -14.606, -14.049,
    -13.327, -12.610, -12.006, -11.555, -14.862, -15.023, -15.038, -14.923,
    -14.615, -14.052, -13.325, -12.604, -11.998, -11.548, -14.874, -15.035,
    -15.049, -14.933, -14.623, -14.055, -13.322, -12.597, -11.991, -11.541,
    -14.887, -15.047, -15.060, -14.943, -14.631, -14.058, -13.319, -12.590,
    -11.983, -11.534, -14.899, -15.058, -15.071, -14.954, -14.639, -14.061,
    -13.316, -12.583, -11.975, -11.527, -14.911, -15.070, -15.082, -14.964,
    -14.647, -14.063, -13.312, -12.576, -11.967, -11.519, -14.924, -15.082,
    -15.093, -14.974, -14.655, -14.065, -13.308, -12.569, -11.959, -11.512,
    -14.936, -15.094, -15.104, -14.984, -14.662, -14.067, -13.304, -12.561,
    -11.951, -11.505, -14.948, -15.106, -15.115, -14.994, -14.670, -14.069,
    -13.299, -12.553, -11.943, -11.497, -14.961, -15.117, -15.126, -15.004,
    -14.677, -14.070, -13.294, -12.546, -11.934, -11.490, -14.973, -15.129,
    -15.137, -15.014, -14.684, -14.071, -13.289, -12.537, -11.926, -11.483,
    -14.986, -15.141, -15.148, -15.024, -14.691, -14.071, -13.283, -12.529,
    -11.918, -11.476, -14.998, -15.153, -15.159, -15.034, -14.698, -14.071,
    -13.278, -12.521, -11.909, -11.468, -15.011, -15.165, -15.170, -15.043,
    -14.704, -14.071, -13.272, -12.513, -11.901, -11.461, -15.023, -15.176,
    -15.181, -15.053, -14.711, -14.071, -13.266, -12.504, -11.892, -11.454,
    -15.035, -15.188, -15.192, -15.063, -14.717, -14.071, -13.259, -12.495,
    -11.884, -11.447, -15.048, -15.200, -15.203, -15.073, -14.723, -14.070,
    -13.253, -12.487, -11.875, -11.440, -15.060, -15.212, -15.214, -15.082,
    -14.729, -14.069, -13.246, -12.478, -11.867, -11.433, -15.073, -15.224,
    -15.225, -15.092, -14.735, -14.067, -13.239, -12.469, -11.859, -11.426,
    -15.085, -15.236, -15.236, -15.102, -14.740, -14.066, -13.232, -12.461,
    -11.851, -11.419, -15.098, -15.248, -15.247, -15.111, -14.746, -14.064,
    -13.225, -12.452, -11.842, -11.412, -15.110, -15.259, -15.258, -15.121,
    -14.751, -14.062, -13.218, -12.443, -11.834, -11.406, -15.123, -15.271,
    -15.269, -15.130, -14.756, -14.060, -13.211, -12.435, -11.826, -11.399,
    -15.136, -15.283, -15.280, -15.140, -14.761, -14.058, -13.204, -12.426,
    -11.818, -11.393, -15.148, -15.295, -15.291, -15.149, -14.766, -14.055,
    -13.197, -12.418, -11.811, -11.386, -15.161, -15.307, -15.302, -15.158,
    -14.771, -14.053, -13.189, -12.409, -11.803, -11.380, -15.173, -15.319,
    -15.313, -15.168, -14.776, -14.050, -13.182, -12.401, -11.796, -11.374,
    -15.186, -15.331, -15.324, -15.177, -14.780, -14.047, -13.175, -12.393,
    -11.788, -11.368, -15.198, -15.343, -15.335, -15.186, -14.785, -14.045,
    -13.168, -12.385, -11.781, -11.363, -15.211, -15.354, -15.346, -15.195,
    -14.789, -14.042, -13.161, -12.377, -11.775, -11.357, -15.224, -15.366,
    -15.357, -15.205, -14.793, -14.039, -13.154, -12.370, -11.768, -11.352,
    -15.236, -15.378, -15.368, -15.214, -14.798, -14.036, -13.148, -12.363,
    -11.762, -11.347, -15.249, -15.390, -15.379, -15.223, -14.802, -14.034,
    -13.141, -12.356, -11.755, -11.342, -15.261, -15.402, -15.390, -15.232,
    -14.806, -14.031, -13.135, -12.349, -11.750, -11.338, -15.274, -15.414,
    -15.401, -15.241, -14.810, -14.029, -13.129, -12.342, -11.744, -11.333,
    -15.287, -15.426, -15.413, -15.246, -14.796, -13.993, -13.088, -12.306,
    -11.715, -11.312, -15.299, -15.438, -15.424, -15.251, -14.783, -13.962,
    -13.053, -12.275, -11.690, -11.293, -15.312, -15.450, -15.435, -15.254,
    -14.762, -13.919, -13.005, -12.233, -11.657, -11.267, -15.325, -15.462,
    -15.446, -15.258, -14.747, -13.886, -12.969, -12.202, -11.632, -11.248,
    -15.337, -15.474, -15.457, -15.264, -14.743, -13.872, -12.953, -12.187,
    -11.620, -11.239, -15.350, -15.486, -15.468, -15.270, -14.735, -13.852,
    -12.932, -12.169, -11.605, -11.228, -15.363, -15.497, -15.479, -15.281,
    -14.750, -13.870, -12.949, -12.183, -11.616, -11.236, -15.375, -15.509,
    -15.490, -15.292, -14.761, -13.880, -12.957, -12.190, -11.622, -11.241,
    -15.388, -15.521, -15.501, -15.299, -14.762, -13.876, -12.951, -12.184,
    -11.617, -11.237, -15.401, -15.533, -15.512, -15.312, -14.780, -13.898,
    -12.972, -12.202, -11.631, -11.247, -15.414, -15.545, -15.523, -15.322,
    -14.787, -13.902, -12.974, -12.203, -11.632, -11.248, -15.426, -15.557,
    -15.534, -15.330, -14.789, -13.899, -12.970, -12.199, -11.628, -11.245,
    -15.439, -15.569, -15.545, -15.343, -14.810, -13.925, -12.993, -12.218,
    -11.643, -11.256, -15.452, -15.581, -15.556, -15.351, -14.810, -13.918,
    -12.984, -12.209, -11.636, -11.251, -15.465, -15.593, -15.567, -15.353,
    -14.791, -13.882, -12.947, -12.177, -11.610, -11.231, -15.477, -15.605,
    -15.578, -15.358, -14.781, -13.859, -12.921, -12.154, -11.591, -11.216,
    -15.490, -15.617, -15.589, -15.357, -14.750, -13.808, -12.869, -12.109,
    -11.556, -11.189, -15.503, -15.629, -15.600, -15.354, -14.719, -13.757,
    -12.819, -12.066, -11.521, -11.162, -15.516, -15.641, -15.611, -15.360,
    -14.715, -13.748, -12.810, -12.059, -11.516, -11.158, -15.528, -15.653,
    -15.622, -15.370, -14.724, -13.758, -12.820, -12.068, -11.523, -11.165,
    -15.541, -15.665, -15.633, -15.384, -14.748, -13.789, -12.849, -12.094,
    -11.544, -11.181, -15.554, -15.677, -15.644, -15.403, -14.788, -13.842,
    -12.901, -12.137, -11.578, -11.207, -15.567, -15.689, -15.655, -15.423,
    -14.831, -13.897, -12.953, -12.180, -11.611, -11.232, -15.580, -15.701,
    -15.666, -15.442, -14.869, -13.946, -12.997, -12.215, -11.639, -11.252,
    -15.592, -15.713, -15.677, -15.456, -14.889, -13.969, -13.017, -12.230,
    -11.650, -11.260, -15.605, -15.725, -15.687, -15.463, -14.889, -13.963,
    -13.009, -12.224, -11.645, -11.256, -15.618, -15.737, -15.698, -15.464,
    -14.868, -13.928, -12.975, -12.196, -11.623, -11.239, -15.631, -15.749,
    -15.709, -15.460, -14.833, -13.873, -12.922, -12.152, -11.589, -11.214,
    -15.644, -15.761, -15.720, -15.453, -14.788, -13.806, -12.857, -12.097,
    -11.545, -11.181, -15.657, -15.773, -15.731, -15.448, -14.748, -13.746,
    -12.797, -12.046, -11.504, -11.149, -15.669, -15.785, -15.742, -15.448,
    -14.727, -13.711, -12.761, -12.013, -11.477, -11.128, -15.682, -15.797,
    -15.753, -15.452, -14.718, -13.693, -12.742, -11.996, -11.462, -11.116,
    -15.695, -15.809, -15.764, -15.465, -14.733, -13.708, -12.753, -12.004,
    -11.468, -11.120, -15.708, -15.821, -15.775, -15.487, -14.776, -13.759,
    -12.801, -12.044, -11.500, -11.144, -15.721, -15.832, -15.786, -15.505,
    -14.811, -13.803, -12.841, -12.078, -11.527, -11.165, -15.734, -15.844,
    -15.797, -15.522, -14.841, -13.839, -12.876, -12.107, -11.550, -11.183,
    -15.747, -15.856, -15.808, -15.539, -14.874, -13.881, -12.914, -12.139,
    -11.575, -11.202, -15.760, -15.868, -15.819, -15.548, -14.880, -13.885,
    -12.918, -12.143, -11.578, -11.204, -15.772, -15.880, -15.830, -15.552,
    -14.873, -13.872, -12.905, -12.132, -11.570, -11.197, -15.785, -15.892,
    -15.840, -15.566, -14.892, -13.893, -12.924, -12.147, -11.581, -11.205,
    -15.798, -15.904, -15.851, -15.571, -14.887, -13.882, -12.913, -12.137,
    -11.574, -11.200, -15.811, -15.916, -15.862, -15.573, -14.874, -13.860,
    -12.892, -12.120, -11.560, -11.189, -15.824, -15.928, -15.873, -15.585,
    -14.888, -13.874, -12.903, -12.128, -11.566, -11.193, -15.837, -15.940,
    -15.884, -15.589, -14.878, -13.856, -12.885, -12.112, -11.553, -11.183,
    -15.850, -15.952, -15.895, -15.595, -14.874, -13.847, -12.875, -12.104,
    -11.546, -11.178, -15.863, -15.964, -15.906, -15.610, -14.899, -13.875,
    -12.899, -12.123, -11.561, -11.189, -15.875, -15.976, -15.916, -15.622,
    -14.914, -13.890, -12.912, -12.133, -11.569, -11.194, -15.888, -15.988,
    -15.927, -15.645, -14.960, -13.946, -12.964, -12.175, -11.601, -11.218,
    -15.901, -16.000, -15.938, -15.668, -15.010, -14.008, -13.020, -12.220,
    -11.635, -11.243, -15.914, -16.011, -15.949, -15.684, -15.039, -14.044,
    -13.052, -12.246, -11.655, -11.257, -15.927, -16.023, -15.959, -15.702,
    -15.075, -14.088, -13.092, -12.278, -11.679, -11.275, -15.940, -16.035,
    -15.970, -15.712, -15.084, -14.096, -13.098, -12.283, -11.682, -11.277,
    -15.953, -16.047, -15.981, -15.721, -15.093, -14.104, -13.105, -12.288,
    -11.686, -11.280, -15.966, -16.059, -15.992, -15.731, -15.102, -14.113,
    -13.112, -12.294, -11.691, -11.284, -15.979, -16.071, -16.002, -15.741,
    -15.112, -14.122, -13.120, -12.301, -11.696, -11.287, -15.991, -16.082,
    -16.013, -15.751, -15.121, -14.131, -13.129, -12.308, -11.701, -11.291,
    -16.004, -16.094, -16.023, -15.761, -15.131, -14.141, -13.138, -12.315,
    -11.707, -11.295, -16.017, -16.106, -16.034, -15.771, -15.142, -14.152,
    -13.147, -12.323, -11.713, -11.299, -16.030, -16.118, -16.045, -15.781,
    -15.152, -14.162, -13.157, -12.331, -11.719, -11.304, -16.043, -16.129,
    -16.055, -15.791, -15.163, -14.173, -13.167, -12.339, -11.726, -11.308,
    -16.056, -16.141, -16.066, -15.801, -15.174, -14.185, -13.177, -12.348,
    -11.732, -11.314, -16.068, -16.153, -16.076, -15.811, -15.185, -14.197,
    -13.188, -12.357, -11.740, -11.319, -16.081, -16.164, -16.087, -15.821,
    -15.197, -14.209, -13.199, -12.366, -11.747, -11.324, -16.094, -16.176,
    -16.097, -15.831, -15.208, -14.221, -13.211, -12.376, -11.755, -11.330,
    -16.107, -16.188, -16.108, -15.841, -15.220, -14.234, -13.223, -12.386,
    -11.763, -11.336, -16.119, -16.199, -16.118, -15.852, -15.231, -14.247,
    -13.235, -12.397, -11.771, -11.342, -16.132, -16.211, -16.128, -15.862,
    -15.243, -14.260, -13.247, -12.407, -11.780, -11.348, -16.145, -16.222,
    -16.139, -15.872, -15.255, -14.273, -13.260, -12.418, -11.788, -11.355,
    -16.157, -16.234, -16.149, -15.882, -15.267, -14.287, -13.273, -12.429,
    -11.797, -11.362, -16.170, -16.245, -16.159, -15.892, -15.279, -14.301,
    -13.286, -12.441, -11.806, -11.369, -16.183, -16.256, -16.169, -15.902,
    -15.292, -14.314, -13.299, -12.452, -11.816, -11.376, -16.195, -16.268,
    -16.179, -15.912, -15.304, -14.329, -13.313, -12.464, -11.825, -11.383,
    -16.208, -16.279, -16.189, -15.923, -15.316, -14.343, -13.327, -12.476,
    -11.835, -11.391, -16.220, -16.290, -16.200, -15.933, -15.328, -14.357,
    -13.341, -12.488, -11.845, -11.398, -16.233, -16.301, -16.209, -15.943,
    -15.341, -14.372, -13.355, -12.501, -11.855, -11.406, -16.245, -16.313,
    -16.219, -15.953, -15.353, -14.386, -13.369, -12.513, -11.866, -11.414,
    -16.257, -16.324, -16.229, -15.963, -15.366, -14.401, -13.384, -12.526,
    -11.876, -11.422, -16.270, -16.335, -16.239, -15.972, -15.378, -14.416,
    -13.398, -12.539, -11.887, -11.430, -16.282, -16.345, -16.249, -15.982,
    -15.390, -14.431, -13.413, -12.552, -11.898, -11.439, -16.294, -16.356,
    -16.259, -15.992, -15.403, -14.446, -13.428, -12.565, -11.909, -11.447,
    -16.306, -16.367, -16.268, -16.002, -15.415, -14.461, -13.443, -12.579,
    -11.920, -11.456, -16.318, -16.378, -16.278, -16.012, -15.428, -14.476,
    -13.458, -12.592, -11.931, -11.465, -16.330, -16.388, -16.287, -16.021,
    -15.440, -14.491, -13.473, -12.606, -11.942, -11.474, -16.342, -16.399,
    -16.297, -16.031, -15.452, -14.506, -13.488, -12.619, -11.954, -11.483,
    -16.354, -16.409, -16.306, -16.040, -15.465, -14.521, -13.504, -12.633,
    -11.965, -11.492, -16.366, -16.420, -16.315, -16.050, -15.477, -14.536,
    -13.519, -12.647, -11.977, -11.501, -16.378, -16.430, -16.325, -16.059,
    -15.489, -14.552, -13.535, -12.661, -11.989, -11.510, -16.389, -16.440,
    -16.334, -16.069, -15.501, -14.567, -13.550, -12.675, -12.001, -11.520,
    -16.401, -16.450, -16.343, -16.078, -15.513, -14.582, -13.566, -12.689,
    -12.013, -11.529, -16.412, -16.460, -16.352, -16.087, -15.525, -14.597,
    -13.582, -12.704, -12.025, -11.539, -16.423, -16.470, -16.361, -16.097,
    -15.538, -14.613, -13.597, -12.718, -12.037, -11.548, -16.434, -16.480,
    -16.369, -16.106, -15.549, -14.628, -13.613, -12.732, -12.049, -11.558,
    -16.445, -16.489, -16.378, -16.115, -15.561, -14.643, -13.629, -12.747,
    -12.061, -11.568, -16.456, -16.499, -16.386, -16.124, -15.573, -14.658,
    -13.645, -12.761, -12.074, -11.578, -16.467, -16.508, -16.395, -16.133,
    -15.585, -14.673, -13.660, -12.776, -12.086, -11.588, -16.477, -16.517,
    -16.403, -16.141, -15.597, -14.688, -13.676, -12.791, -12.099, -11.598,
    -16.488, -16.526, -16.412, -16.150, -15.608, -14.703, -13.692, -12.805,
    -12.111, -11.608, -16.498, -16.535, -16.420, -16.159, -15.620, -14.719,
    -13.708, -12.820, -12.124, -11.618, -16.508, -16.544, -16.428, -16.167,
    -15.632, -14.734, -13.724, -12.835, -12.136, -11.628, -16.518, -16.552,
    -16.436, -16.176, -15.643, -14.749, -13.740, -12.849, -12.149, -11.638,
    -16.528, -16.561, -16.444, -16.184, -15.654, -14.763, -13.756, -12.864,
    -12.162, -11.648, -16.537, -16.569, -16.451, -16.193, -15.666, -14.778,
    -13.771, -12.879, -12.175, -11.659, -16.546, -16.577, -16.459, -16.201,
    -15.677, -14.793, -13.787, -12.894, -12.187, -11.669, -16.555, -16.585,
    -16.466, -16.209, -15.688, -14.808, -13.803, -12.908, -12.200, -11.679,
    -16.564, -16.593, -16.474, -16.217, -15.699, -14.823, -13.819, -12.923,
    -12.213, -11.690, -16.573, -16.600, -16.481, -16.225, -15.710, -14.837,
    -13.835, -12.938, -12.226, -11.700, -16.581, -16.607, -16.488, -16.233,
    -15.721, -14.852, -13.850, -12.953, -12.239, -11.711, -16.589, -16.615,
    -16.495, -16.241, -15.732, -14.867, -13.866, -12.968, -12.252, -11.721,
    -16.597, -16.622, -16.502, -16.249, -15.743, -14.881, -13.882, -12.983,
    -12.265, -11.732, -16.604, -16.628, -16.509, -16.257, -15.753, -14.895,
    -13.898, -12.997, -12.277, -11.742, -16.612, -16.635, -16.516, -16.264,
    -15.764, -14.910, -13.913, -13.012, -12.290, -11.753, -16.619, -16.641,
    -16.522, -16.272, -15.775, -14.924, -13.929, -13.027, -12.303, -11.763,
    -16.625, -16.648, -16.529, -16.280, -15.785, -14.938, -13.944, -13.042,
    -12.316, -11.774, -16.632, -16.654, -16.535, -16.287, -15.796, -14.953,
    -13.960, -13.057, -12.329, -11.785, -16.638, -16.660, -16.541, -16.295,
    -15.806, -14.967, -13.976, -13.071, -12.342, -11.795, -16.644, -16.666,
    -16.548, -16.302, -15.816, -14.981, -13.991, -13.086, -12.355, -11.806,
    -16.650, -16.671, -16.554, -16.309, -15.827, -14.995, -14.006, -13.101,
    -12.368, -11.816, -16.656, -16.677, -16.560, -16.317, -15.837, -15.009,
    -14.022, -13.116, -12.381, -11.827, -16.661, -16.682, -16.566, -16.324,
    -15.847, -15.023, -14.037, -13.130, -12.394, -11.837, -16.666, -16.688,
    -16.572, -16.331, -15.857, -15.036, -14.053, -13.145, -12.407, -11.848,
    -16.671, -16.693, -16.578, -16.338, -15.867, -15.050, -14.068, -13.160,
    -12.419, -11.859, -16.676, -16.698, -16.584, -16.346, -15.877, -15.064,
    -14.083, -13.174, -12.432, -11.869, -16.680, -16.703, -16.590, -16.353,
    -15.887, -15.077, -14.098, -13.189, -12.445, -11.880, -16.685, -16.708,
    -16.596, -16.360, -15.897, -15.091, -14.113, -13.203, -12.458, -11.890,
    -16.689, -16.713, -16.602, -16.367, -15.907, -15.104, -14.128, -13.218,
    -12.471, -11.901, -16.693, -16.718, -16.607, -16.374, -15.917, -15.118,
    -14.143, -13.232, -12.483, -11.911, -16.618, -16.660, -16.571, -16.357,
    -15.916, -15.128, -14.158, -13.246, -12.496, -11.921, -16.590, -16.639,
    -16.560, -16.354, -15.921, -15.140, -14.172, -13.261, -12.509, -11.932,
    -16.569, -16.625, -16.553, -16.355, -15.928, -15.152, -14.187, -13.275,
    -12.521, -11.942, -16.566, -16.626, -16.557, -16.361, -15.937, -15.165,
    -14.202, -13.289, -12.534, -11.953, -16.563, -16.627, -16.561, -16.367,
    -15.946, -15.178, -14.216, -13.304, -12.546, -11.963, -16.559, -16.628,
    -16.565, -16.374, -15.956, -15.191, -14.231, -13.318, -12.559, -11.973,
    -16.555, -16.628, -16.569, -16.380, -15.965, -15.204, -14.246, -13.332,
    -12.571, -11.983, -16.580, -16.650, -16.586, -16.394, -15.977, -15.218,
    -14.260, -13.346, -12.584, -11.994, -16.574, -16.648, -16.588, -16.399,
    -15.986, -15.230, -14.275, -13.360, -12.596, -12.004, -16.559, -16.639,
    -16.585, -16.400, -15.993, -15.242, -14.289, -13.374, -12.609, -12.014,
    -16.536, -16.623, -16.577, -16.399, -15.999, -15.254, -14.303, -13.388,
    -12.621, -12.024, -16.470, -16.574, -16.547, -16.385, -15.998, -15.264,
    -14.317, -13.402, -12.633, -12.034, -16.468, -16.574, -16.548, -16.389,
    -16.006, -15.276, -14.331, -13.415, -12.645, -12.044, -16.457, -16.569,
    -16.548, -16.394, -16.014, -15.288, -14.345, -13.429, -12.657, -12.054,
    -16.433, -16.554, -16.542, -16.394, -16.021, -15.299, -14.359, -13.443,
    -12.669, -12.064, -16.455, -16.577, -16.561, -16.410, -16.034, -15.313,
    -14.373, -13.457, -12.681, -12.074, -16.434, -16.567, -16.560, -16.414,
    -16.042, -15.325, -14.387, -13.470, -12.693, -12.084, -16.418, -16.560,
    -16.560, -16.417, -16.050, -15.337, -14.401, -13.484, -12.705, -12.093,
    -16.469, -16.602, -16.591, -16.440, -16.066, -15.351, -14.416, -13.497,
    -12.717, -12.103, -16.495, -16.626, -16.610, -16.454, -16.078, -15.364,
    -14.430, -13.511, -12.729, -12.113, -16.490, -16.627, -16.615, -16.461,
    -16.087, -15.376, -14.443, -13.524, -12.741, -12.122, -16.542, -16.669,
    -16.645, -16.481, -16.102, -15.390, -14.458, -13.538, -12.752, -12.132,
    -16.517, -16.653, -16.637, -16.481, -16.107, -15.401, -14.471, -13.551,
    -12.764, -12.141, -16.456, -16.607, -16.609, -16.468, -16.107, -15.410,
    -14.484, -13.564, -12.775, -12.151, -16.459, -16.604, -16.605, -16.467,
    -16.112, -15.420, -14.497, -13.577, -12.787, -12.160, -16.396, -16.553,
    -16.571, -16.449, -16.109, -15.428, -14.509, -13.590, -12.798, -12.169,
    -16.300, -16.479, -16.522, -16.424, -16.102, -15.435, -14.522, -13.602,
    -12.809, -12.178, -16.230, -16.432, -16.496, -16.413, -16.103, -15.444,
    -14.534, -13.615, -12.821, -12.187, -16.188, -16.410, -16.488, -16.414,
    -16.110, -15.455, -14.547, -13.628, -12.832, -12.196, -16.185, -16.420,
    -16.503, -16.428, -16.123, -15.468, -14.561, -13.641, -12.843, -12.205,
    -16.253, -16.483, -16.553, -16.464, -16.145, -15.484, -14.575, -13.653,
    -12.854, -12.214, -16.428, -16.619, -16.646, -16.520, -16.176, -15.503,
    -14.590, -13.666, -12.865, -12.223, -16.608, -16.749, -16.729, -16.569,
    -16.203, -15.519, -14.604, -13.679, -12.876, -12.232, -16.732, -16.833,
    -16.781, -16.599, -16.221, -15.533, -14.617, -13.692, -12.886, -12.241,
    -16.711, -16.822, -16.778, -16.602, -16.228, -15.544, -14.630, -13.704,
    -12.897, -12.249, -16.550, -16.717, -16.719, -16.575, -16.222, -15.551,
    -14.642, -13.716, -12.907, -12.258, -16.350, -16.576, -16.633, -16.531,
    -16.209, -15.555, -14.653, -13.728, -12.918, -12.266, -16.169, -16.438,
    -16.543, -16.482, -16.191, -15.558, -14.664, -13.739, -12.928, -12.274,
    -16.091, -16.370, -16.494, -16.454, -16.183, -15.563, -14.675, -13.751,
    -12.938, -12.283, -16.088, -16.360, -16.484, -16.449, -16.184, -15.572,
    -14.686, -13.763, -12.948, -12.291, -16.121, -16.381, -16.497, -16.458,
    -16.193, -15.582, -14.699, -13.775, -12.958, -12.299, -16.201, -16.442,
    -16.539, -16.486, -16.211, -15.596, -14.712, -13.786, -12.969, -12.307,
    -16.294, -16.520, -16.598, -16.525, -16.235, -15.612, -14.725, -13.798,
    -12.978, -12.315, -16.333, -16.562, -16.634, -16.552, -16.253, -15.626,
    -14.738, -13.810, -12.988, -12.323, -16.337, -16.577, -16.651, -16.568,
    -16.266, -15.639, -14.751, -13.822, -12.998, -12.331, -16.395, -16.626,
    -16.689, -16.593, -16.284, -15.652, -14.763, -13.833, -13.008, -12.338,
    -16.394, -16.627, -16.692, -16.599, -16.291, -15.663, -14.775, -13.844,
    -13.017, -12.346, -16.305, -16.562, -16.652, -16.579, -16.287, -15.669,
    -14.786, -13.855, -13.027, -12.353, -16.314, -16.569, -16.658, -16.586,
    -16.295, -15.680, -14.798, -13.866, -13.036, -12.361, -16.242, -16.515,
    -16.624, -16.570, -16.293, -15.687, -14.809, -13.877, -13.045, -12.368,
    -16.154, -16.448, -16.580, -16.547, -16.287, -15.692, -14.819, -13.888,
    -13.054, -12.375, -16.195, -16.478, -16.601, -16.562, -16.298, -15.703,
    -14.831, -13.899, -13.063, -12.382, -16.187, -16.468, -16.595, -16.560,
    -16.302, -15.712, -14.842, -13.909, -13.072, -12.390, -16.152, -16.440,
    -16.576, -16.551, -16.302, -15.720, -14.853, -13.920, -13.081, -12.397,
    -16.177, -16.460, -16.591, -16.563, -16.313, -15.730, -14.864, -13.930,
    -13.089, -12.404, -16.164, -16.451, -16.587, -16.563, -16.317, -15.739,
    -14.875, -13.940, -13.098, -12.410, -16.189, -16.476, -16.609, -16.580,
    -16.331, -15.751, -14.887, -13.951, -13.107, -12.417, -16.284, -16.552,
    -16.663, -16.616, -16.353, -15.766, -14.899, -13.961, -13.115, -12.424,
    -16.308, -16.573, -16.681, -16.630, -16.364, -15.777, -14.910, -13.971,
    -13.123, -12.430, -16.330, -16.592, -16.696, -16.643, -16.375, -15.788,
    -14.921, -13.981, -13.132, -12.437, -16.328, -16.590, -16.698, -16.646,
    -16.381, -15.797, -14.932, -13.991, -13.140, -12.443, -16.271, -16.549,
    -16.672, -16.634, -16.380, -15.804, -14.942, -14.001, -13.148, -12.450,
    -16.271, -16.548, -16.672, -16.637, -16.386, -15.813, -14.952, -14.010,
    -13.156, -12.456, -16.271, -16.548, -16.674, -16.641, -16.392, -15.822,
    -14.963, -14.020, -13.163, -12.462, -16.273, -16.548, -16.675, -16.644,
    -16.398, -15.831, -14.973, -14.029, -13.171, -12.469, -16.274, -16.548,
    -16.676, -16.647, -16.404, -15.839, -14.983, -14.039, -13.179, -12.475,
    -16.340, -16.600, -16.713, -16.672, -16.420, -15.852, -14.994, -14.048,
    -13.186, -12.481, -16.420, -16.660, -16.756, -16.700, -16.438, -15.865,
    -15.005, -14.058, -13.194, -12.487, -16.644, -16.823, -16.863, -16.764,
    -16.472, -15.883, -15.017, -14.067, -13.202, -12.493, -16.645, -16.824,
    -16.866, -16.769, -16.479, -15.892, -15.027, -14.076, -13.209, -12.499,
    -16.647, -16.827, -16.870, -16.774, -16.486, -15.901, -15.037, -14.085,
    -13.216, -12.505, -16.649, -16.830, -16.874, -16.780, -16.493, -15.910,
    -15.047, -14.094, -13.223, -12.511, -16.652, -16.833, -16.879, -16.786,
    -16.501, -15.919, -15.057, -14.103, -13.231, -12.516, -16.656, -16.837,
    -16.883, -16.792, -16.508, -15.928, -15.067, -14.112, -13.238, -12.522,
    -16.661, -16.841, -16.889, -16.798, -16.515, -15.937, -15.077, -14.120,
    -13.245, -12.528, -16.667, -16.847, -16.894, -16.804, -16.523, -15.946,
    -15.086, -14.129, -13.252, -12.534, -16.673, -16.852, -16.900, -16.811,
    -16.530, -15.955, -15.096, -14.137, -13.259, -12.539, -16.680, -16.858,
    -16.907, -16.817, -16.538, -15.964, -15.106, -14.146, -13.266, -12.545,
    -16.687, -16.865, -16.913, -16.824, -16.545, -15.973, -15.115, -14.154,
    -13.273, -12.551, -16.695, -16.872, -16.920, -16.831, -16.553, -15.983,
    -15.125, -14.163, -13.279, -12.556, -16.704, -16.879, -16.927, -16.839,
    -16.561, -15.992, -15.134, -14.171, -13.286, -12.562, -16.714, -16.887,
    -16.935, -16.846, -16.569, -16.001, -15.144, -14.179, -13.293, -12.568,
    -16.724, -16.896, -16.943, -16.854, -16.577, -16.010, -15.153, -14.188,
    -13.300, -12.573, -16.734, -16.904, -16.951, -16.862, -16.585, -16.018,
    -15.160, -14.192, -13.302, -12.575, -16.745, -16.913, -16.959, -16.870,
    -16.593, -16.026, -15.167, -14.196, -13.305, -12.576, -16.756, -16.923,
    -16.968, -16.878, -16.602, -16.034, -15.172, -14.197, -13.303, -12.575,
    -16.768, -16.933, -16.977, -16.886, -16.610, -16.042, -15.176, -14.197,
    -13.301, -12.573, -16.781, -16.943, -16.986, -16.895, -16.618, -16.051,
    -15.184, -14.203, -13.306, -12.577, -16.793, -16.954, -16.995, -16.903,
    -16.627, -16.059, -15.191, -14.207, -13.308, -12.579, -16.806, -16.964,
    -17.005, -16.912, -16.635, -16.068, -15.201, -14.217, -13.317, -12.586,
    -16.820, -16.975, -17.015, -16.921, -16.644, -16.078, -15.212, -14.228,
    -13.328, -12.595, -16.834, -16.987, -17.025, -16.930, -16.653, -16.087,
    -15.220, -14.235, -13.333, -12.600, -16.848, -16.999, -17.035, -16.939,
    -16.661, -16.096, -15.232, -14.247, -13.344, -12.609, -16.863, -17.010,
    -17.045, -16.949, -16.670, -16.106, -15.243, -14.259, -13.354, -12.618,
    -16.877, -17.023, -17.056, -16.958, -16.679, -16.115, -15.251, -14.265,
    -13.360, -12.622, -16.893, -17.035, -17.067, -16.968, -16.688, -16.125,
    -15.263, -14.277, -13.371, -12.632, -16.908, -17.048, -17.078, -16.977,
    -16.697, -16.134, -15.273, -14.287, -13.379, -12.639, -16.924, -17.061,
    -17.089, -16.987, -16.706, -16.142, -15.279, -14.290, -13.381, -12.640,
    -16.939, -17.074, -17.100, -16.997, -16.715, -16.150, -15.284, -14.291,
    -13.380, -12.640, -16.956, -17.087, -17.111, -17.007, -16.724, -16.158,
    -15.288, -14.291, -13.378, -12.637, -16.972, -17.100, -17.123, -17.017,
    -16.733, -16.165, -15.288, -14.286, -13.371, -12.631, -16.988, -17.114,
    -17.135, -17.027, -16.743, -16.173, -15.293, -14.288, -13.372, -12.631,
    -17.005, -17.128, -17.146, -17.037, -16.752, -16.182, -15.303, -14.297,
    -13.380, -12.639, -17.022, -17.141, -17.158, -17.048, -16.761, -16.192,
    -15.315, -14.310, -13.393, -12.650, -17.039, -17.155, -17.170, -17.058,
    -16.771, -16.203, -15.331, -14.329, -13.411, -12.665, -17.056, -17.170,
    -17.183, -17.069, -16.780, -16.214, -15.348, -14.349, -13.431, -12.682,
    -17.074, -17.184, -17.195, -17.079, -16.790, -16.226, -15.363, -14.367,
    -13.448, -12.696, -17.091, -17.198, -17.207, -17.090, -16.799, -16.236,
    -15.376, -14.382, -13.461, -12.708, -17.109, -17.213, -17.220, -17.101,
    -16.809, -16.245, -15.386, -14.391, -13.469, -12.714, -17.126, -17.227,
    -17.232, -17.111, -16.818, -16.254, -15.392, -14.394, -13.470, -12.715,
    -17.144, -17.242, -17.245, -17.122, -16.828, -16.261, -15.393, -14.391,
    -13.466, -12.711, -17.162, -17.257, -17.258, -17.133, -16.837, -16.267,
    -15.393, -14.384, -13.458, -12.704, -17.180, -17.272, -17.270, -17.144,
    -16.847, -16.273, -15.391, -14.376, -13.447, -12.695, -17.198, -17.287,
    -17.283, -17.155, -16.857, -16.280, -15.392, -14.371, -13.441, -12.690,
    -17.216, -17.302, -17.296, -17.166, -16.866, -16.288, -15.396, -14.371,
    -13.439, -12.688, -17.235, -17.317, -17.309, -17.177, -16.876, -16.297,
    -15.403, -14.376, -13.443, -12.691, -17.253, -17.332, -17.322, -17.189,
    -16.886, -16.308, -15.419, -14.394, -13.459, -12.705, -17.271, -17.348,
    -17.335, -17.200, -16.896, -16.319, -15.435, -14.414, -13.479, -12.722,
    -17.290, -17.363, -17.348, -17.211, -16.905, -16.330, -15.449, -14.429,
    -13.493, -12.733, -17.308, -17.378, -17.362, -17.222, -16.915, -16.341,
    -15.465, -14.448, -13.511, -12.749, -17.327, -17.394, -17.375, -17.234,
    -16.925, -16.351, -15.476, -14.460, -13.523, -12.758, -17.345, -17.409,
    -17.388, -17.245, -16.935, -16.359, -15.482, -14.464, -13.525, -12.761,
    -17.364, -17.425, -17.401, -17.256, -16.945, -16.369, -15.492, -14.473,
    -13.534, -12.767, -17.382, -17.440, -17.415, -17.268, -16.955, -16.378,
    -15.501, -14.482, -13.541, -12.773, -17.401, -17.456, -17.428, -17.279,
    -16.965, -16.386, -15.506, -14.483, -13.541, -12.773, -17.419, -17.471,
    -17.442, -17.291, -16.975, -16.395, -15.515, -14.491, -13.548, -12.779,
    -17.438, -17.487, -17.455, -17.302, -16.985, -16.404, -15.521, -14.496,
    -13.551, -12.781, -17.456, -17.503, -17.468, -17.314, -16.995, -16.412,
    -15.526, -14.497, -13.550, -12.781, -17.475, -17.518, -17.482, -17.325,
    -17.005, -16.422, -15.537, -14.508, -13.561, -12.789, -17.494, -17.534,
    -17.495, -17.337, -17.015, -16.432, -15.548, -14.520, -13.571, -12.798,
    -17.512, -17.550, -17.509, -17.348, -17.025, -16.443, -15.563, -14.536,
    -13.586, -12.810, -17.531, -17.565, -17.523, -17.360, -17.035, -16.455,
    -15.580, -14.557, -13.606, -12.827, -17.549, -17.581, -17.536, -17.372,
    -17.045, -16.466, -15.595, -14.575, -13.623, -12.840, -17.568, -17.597,
    -17.550, -17.383, -17.055, -16.477, -15.612, -14.594, -13.641, -12.855,
    -17.586, -17.612, -17.563, -17.395, -17.065, -16.486, -15.621, -14.603,
    -13.649, -12.861, -17.605, -17.628, -17.577, -17.407, -17.075, -16.496,
    -15.630, -14.612, -13.656, -12.868, -17.623, -17.644, -17.591, -17.418,
    -17.085, -16.505, -15.639, -14.621, -13.664, -12.874, -17.641, -17.660,
    -17.604, -17.430, -17.095, -16.514, -15.649, -14.630, -13.672, -12.880,
    -17.660, -17.675, -17.618, -17.442, -17.105, -16.524, -15.658, -14.639,
    -13.680, -12.887, -17.678, -17.691, -17.631, -17.453, -17.115, -16.533,
    -15.668, -14.648, -13.688, -12.893, -17.696, -17.707, -17.645, -17.465,
    -17.125, -16.543, -15.677, -14.657, -13.696, -12.900, -17.715, -17.722,
    -17.659, -17.477, -17.135, -16.552, -15.687, -14.667, -13.704, -12.907,
    -17.733, -17.738, -17.672, -17.488, -17.145, -16.561, -15.697, -14.676,
    -13.713, -12.913, -17.751, -17.754, -17.686, -17.500, -17.156, -16.571,
    -15.707, -14.686, -13.721, -12.920, -17.769, -17.769, -17.700, -17.512,
    -17.166, -16.580, -15.716, -14.695, -13.730, -12.927, -17.787, -17.785,
    -17.713, -17.524, -17.176, -16.590, -15.726, -14.705, -13.739, -12.934,
    -17.805, -17.800, -17.727, -17.535, -17.186, -16.599, -15.736, -14.715,
    -13.748, -12.942, -17.823, -17.816, -17.740, -17.547, -17.196, -16.609,
    -15.746, -14.725, -13.757, -12.949, -17.841, -17.832, -17.754, -17.559,
    -17.206, -16.618, -15.756, -14.735, -13.766, -12.956, -17.859, -17.847,
    -17.768, -17.571, -17.216, -16.628, -15.766, -14.745, -13.775, -12.964,
    -17.877, -17.863, -17.781, -17.582, -17.226, -16.637, -15.776, -14.755,
    -13.784, -12.971, -17.894, -17.878, -17.795, -17.594, -17.236, -16.646,
    -15.786, -14.765, -13.793, -12.979, -17.912, -17.893, -17.808, -17.606,
    -17.246, -16.656, -15.796, -14.776, -13.802, -12.986, -17.930, -17.909,
    -17.822, -17.617, -17.256, -16.665, -15.806, -14.786, -13.812, -12.994,
    -17.947, -17.924, -17.835, -17.629, -17.267, -16.675, -15.816, -14.796,
    -13.821, -13.002, -17.965, -17.940, -17.849, -17.641, -17.277, -16.684,
    -15.826, -14.807, -13.831, -13.009, -17.982, -17.955, -17.863, -17.653,
    -17.287, -16.694, -15.836, -14.817, -13.841, -13.017, -18.000, -17.970,
    -17.876, -17.664, -17.297, -16.703, -15.846, -14.828, -13.850, -13.025,
    -18.017, -17.986, -17.890, -17.676, -17.307, -16.713, -15.856, -14.838,
    -13.860, -13.033, -18.034, -18.001, -17.903, -17.688, -17.317, -16.722,
    -15.866, -14.849, -13.870, -13.041, -18.051, -18.016, -17.917, -17.699,
    -17.327, -16.731, -15.876, -14.859, -13.879, -13.049, -18.069, -18.031,
    -17.930, -17.711, -17.337, -16.741, -15.886, -14.870, -13.889, -13.057,
    -18.086, -18.046, -17.943, -17.723, -17.347, -16.750, -15.897, -14.881,
    -13.899, -13.065, -18.103, -18.062, -17.957, -17.734, -17.357, -16.760,
    -15.907, -14.891, -13.909, -13.073, -18.120, -18.077, -17.970, -17.746,
    -17.367, -16.769, -15.917, -14.902, -13.919, -13.081, -18.136, -18.092,
    -17.984, -17.758, -17.377, -16.779, -15.927, -14.913, -13.929, -13.090,
    -18.153, -18.107, -17.997, -17.769, -17.387, -16.788, -15.937, -14.924,
    -13.939, -13.098, -18.170, -18.122, -18.010, -17.781, -17.397, -16.797,
    -15.947, -14.934, -13.949, -13.106, -18.187, -18.137, -18.024, -17.792,
    -17.408, -16.807, -15.957, -14.945, -13.959, -13.114, -18.203, -18.152,
    -18.037, -17.804, -17.418, -16.816, -15.967, -14.956, -13.969, -13.122,
    -18.220, -18.167, -18.050, -17.816, -17.428, -16.825, -15.977, -14.967,
    -13.979, -13.131, -18.236, -18.182, -18.064, -17.827, -17.438, -16.835,
    -15.987, -14.977, -13.990, -13.139, -18.252, -18.196, -18.077, -17.839,
    -17.448, -16.844, -15.997, -14.988, -14.000, -13.147, -18.269, -18.211,
    -18.090, -17.850, -17.458, -16.853, -16.007, -14.999, -14.010, -13.155,
    -18.285, -18.226, -18.103, -17.862, -17.468, -16.863, -16.017, -15.009,
    -14.020, -13.164, -18.301, -18.241, -18.116, -17.874, -17.478, -16.872,
    -16.027, -15.020, -14.030, -13.172, -18.317, -18.255, -18.130, -17.885,
    -17.488, -16.881, -16.037, -15.031, -14.040, -13.180, -18.333, -18.270,
    -18.143, -17.897, -17.498, -16.890, -16.047, -15.042, -14.050, -13.188,
    -18.349, -18.285, -18.156, -17.908, -17.508, -16.900, -16.057, -15.052,
    -14.061, -13.197, -18.365, -18.299, -18.169, -17.920, -17.517, -16.909,
    -16.067, -15.063, -14.071, -13.205, -18.381, -18.314, -18.182, -17.931,
    -17.527, -16.918, -16.077, -15.074, -14.081, -13.213, -18.397, -18.328,
    -18.195, -17.943, -17.537, -16.927, -16.087, -15.084, -14.091, -13.221,
    -18.413, -18.343, -18.208, -17.954, -17.547, -16.937, -16.096, -15.095,
    -14.101, -13.229, -18.428, -18.357, -18.221, -17.965, -17.557, -16.946,
    -16.106, -15.106, -14.111, -13.237, -18.444, -18.372, -18.234, -17.977,
    -17.567, -16.955, -16.116, -15.116, -14.121, -13.245, -18.459, -18.386,
    -18.247, -17.988, -17.577, -16.964, -16.126, -15.127, -14.131, -13.253,
    -18.475, -18.400, -18.260, -18.000, -17.587, -16.973, -16.136, -15.137,
    -14.141, -13.262, -18.490, -18.415, -18.273, -18.011, -17.597, -16.982,
    -16.145, -15.148, -14.151, -13.270, -18.505, -18.429, -18.286, -18.023,
    -17.607, -16.992, -16.155, -15.158, -14.161, -13.277, -18.521, -18.443,
    -18.299, -18.034, -17.617, -17.001, -16.165, -15.169, -14.171, -13.285,
    -18.536, -18.457, -18.312, -18.045, -17.627, -17.010, -16.175, -15.179,
    -14.181, -13.293, -18.551, -18.471, -18.325, -18.057, -17.636, -17.019,
    -16.184, -15.190, -14.191, -13.301, -18.566, -18.486, -18.338, -18.068,
    -17.646, -17.028, -16.194, -15.200, -14.200, -13.309, -18.581, -18.500,
    -18.351, -18.079, -17.656, -17.037, -16.203, -15.211, -14.210, -13.317,
    -18.596, -18.514, -18.363, -18.091, -17.666, -17.046, -16.213, -15.221,
    -14.220, -13.324, -18.610, -18.528, -18.376, -18.102, -17.676, -17.055,
    -16.223, -15.231, -14.230, -13.332, -18.625, -18.542, -18.389, -18.113,
    -17.686, -17.064, -16.232, -15.241, -14.239, -13.340, -18.640, -18.556,
    -18.402, -18.124, -17.695, -17.073, -16.242, -15.252, -14.249, -13.347,
    -18.655, -18.570, -18.414, -18.136, -17.705, -17.082, -16.251, -15.262,
    -14.258, -13.355, -18.669, -18.583, -18.427, -18.147, -17.715, -17.091,
    -16.261, -15.272, -14.268, -13.362, -18.684, -18.597, -18.440, -18.158,
    -17.725, -17.100, -16.270, -15.282, -14.277, -13.369, -18.698, -18.611,
    -18.452, -18.169, -17.735, -17.109, -16.279, -15.292, -14.287, -13.376,
    -18.713, -18.625, -18.465, -18.181, -17.744, -17.118, -16.289, -15.302,
    -14.296, -13.384, -18.727, -18.639, -18.478, -18.192, -17.754, -17.127,
    -16.298, -15.312, -14.305, -13.391, -18.741, -18.652, -18.490, -18.203,
    -17.764, -17.136, -16.308, -15.322, -14.315, -13.398, -18.756, -18.666,
    -18.503, -18.214, -17.773, -17.145, -16.317, -15.332, -14.324, -13.405,
    -18.770, -18.680, -18.515, -18.225, -17.783, -17.154, -16.326, -15.342,
    -14.333, -13.412, -18.784, -18.693, -18.528, -18.236, -17.793, -17.163,
    -16.336, -15.352, -14.342, -13.419, -18.798, -18.707, -18.540, -18.247,
    -17.803, -17.172, -16.345, -15.362, -14.351, -13.425, -18.812, -18.720,
    -18.553, -18.259, -17.812, -17.180, -16.354, -15.371, -14.360, -13.432,
    -18.826, -18.734, -18.565, -18.270, -17.822, -17.189, -16.363, -15.381,
    -14.369, -13.439, -18.840, -18.747, -18.578, -18.281, -17.832, -17.198,
    -16.372, -15.391, -14.378, -13.445, -18.854, -18.761, -18.590, -18.292,
    -17.841, -17.207, -16.382, -15.400, -14.386, -13.451, -18.867, -18.774,
    -18.602, -18.303, -17.851, -17.216, -16.391, -15.410, -14.395, -13.458,
    -18.881, -18.787, -18.615, -18.314, -17.861, -17.224, -16.400, -15.419,
    -14.404, -13.464, -18.895, -18.801, -18.627, -18.325, -17.870, -17.233,
    -16.409, -15.429, -14.412, -13.470, -18.908, -18.814, -18.640, -18.336,
    -17.880, -17.242, -16.418, -15.438, -14.421, -13.476, -18.922, -18.827,
    -18.652, -18.347, -17.889, -17.251, -16.427, -15.448, -14.429, -13.482,
    -18.935, -18.841, -18.664, -18.358, -17.899, -17.259, -16.436, -15.457,
    -14.437, -13.488, -18.949, -18.854, -18.676, -18.369, -17.909, -17.268,
    -16.445, -15.466, -14.445, -13.494, -18.962, -18.867, -18.689, -18.380,
    -17.918, -17.277, -16.454, -15.476, -14.454, -13.499, -18.976, -18.880,
    -18.701, -18.391, -17.928, -17.286, -16.463, -15.485, -14.462, -13.505,
    -18.989, -18.893, -18.713, -18.401, -17.937, -17.294, -16.472, -15.494,
    -14.470, -13.510, -19.002, -18.906, -18.725, -18.412, -17.947, -17.303,
    -16.481, -15.503, -14.477, -13.516, -19.016, -18.919, -18.737, -18.423,
    -17.956, -17.312, -16.490, -15.512, -14.485, -13.521, -19.029, -18.932,
    -18.750, -18.434, -17.966, -17.320, -16.499, -15.521, -14.493, -13.526,
    -19.042, -18.945, -18.762, -18.445, -17.975, -17.329, -16.507, -15.530,
    -14.500, -13.531, -19.055, -18.958, -18.774, -18.456, -17.985, -17.337,
    -16.516, -15.539, -14.508, -13.536, -19.068, -18.971, -18.786, -18.467,
    -17.994, -17.346, -16.525, -15.548, -14.515, -13.541, -19.081, -18.984,
    -18.798, -18.477, -18.004, -17.355, -16.534, -15.556, -14.523, -13.546,
    -19.094, -18.997, -18.810, -18.488, -18.013, -17.363, -16.542, -15.565,
    -14.530, -13.550, -19.107, -19.010, -18.822, -18.499, -18.022, -17.372,
    -16.551, -15.574, -14.537, -13.555, -19.120, -19.023, -18.834, -18.510,
    -18.032, -17.380, -16.560, -15.582, -14.544, -13.559, -19.133, -19.036,
    -18.846, -18.520, -18.041, -17.389, -16.569, -15.591, -14.551, -13.563,
    -19.145, -19.048, -18.858, -18.531, -18.050, -17.397, -16.577, -15.599,
    -14.558, -13.568, -19.158, -19.061, -18.870, -18.542, -18.060, -17.406,
    -16.586, -15.608, -14.565, -13.572, -19.171, -19.074, -18.882, -18.552,
    -18.069, -17.414, -16.594, -15.616, -14.571, -13.576, -19.183, -19.087,
    -18.893, -18.563, -18.079, -17.423, -16.603, -15.624, -14.578, -13.580,
    -19.196, -19.099, -18.905, -18.574, -18.088, -17.431, -16.612, -15.632,
    -14.584, -13.583, -19.209, -19.112, -18.917, -18.584, -18.097, -17.439,
    -16.620, -15.640, -14.590, -13.587, -19.221, -19.124, -18.929, -18.595,
    -18.106, -17.448, -16.629, -15.649, -14.597, -13.591, -19.234, -19.137,
    -18.941, -18.605, -18.116, -17.456, -16.637, -15.657, -14.603, -13.594,
    -19.246, -19.150, -18.953, -18.616, -18.125, -17.465, -16.646, -15.664,
    -14.609, -13.597, -19.259, -19.162, -18.964, -18.627, -18.134, -17.473,
    -16.654, -15.672, -14.615, -13.601, -19.271, -19.175, -18.976, -18.637,
    -18.143, -17.481, -16.662, -15.680, -14.621, -13.604, -19.283, -19.187,
    -18.988, -18.648, -18.152, -17.490, -16.671, -15.688, -14.626, -13.607,
    -19.296, -19.199, -18.999, -18.658, -18.162, -17.498, -16.679, -15.696,
    -14.632, -13.610, -19.308, -19.212, -19.011, -18.668, -18.171, -17.506,
    -16.688, -15.703, -14.637, -13.612, -19.320, -19.224, -19.023, -18.679,
    -18.180, -17.514, -16.696, -15.711, -14.643, -13.615, -19.332, -19.237,
    -19.034, -18.689, -18.189, -17.523, -16.704, -15.718, -14.648, -13.618,
    -19.345, -19.249, -19.046, -18.700, -18.198, -17.531, -16.712, -15.726,
    -14.653, -13.620, -19.357, -19.261, -19.057, -18.710, -18.207, -17.539,
    -16.721, -15.733, -14.658, -13.623, -19.369, -19.274, -19.069, -18.720,
    -18.216, -17.547, -16.729, -15.740, -14.663, -13.625, -19.381, -19.286,
    -19.080, -18.731, -18.225, -17.555, -16.737, -15.748, -14.668, -13.628,
    -19.393, -19.298, -19.092, -18.741, -18.234, -17.563, -16.745, -15.755,
    -14.673, -13.630, -19.405, -19.310, -19.103, -18.751, -18.243, -17.572,
    -16.753, -15.762, -14.678, -13.632, -19.417, -19.322, -19.115, -18.761,
    -18.252, -17.580, -16.762, -15.769, -14.683, -13.634, -19.429, -19.335,
    -19.126, -18.772, -18.261, -17.588, -16.770, -15.776, -14.687, -13.636,
    -19.441, -19.347, -19.138, -18.782, -18.270, -17.596, -16.778, -15.783,
    -14.692, -13.638, -19.453, -19.359, -19.149, -18.792, -18.279, -17.604,
    -16.786, -15.790, -14.696, -13.640, -19.465, -19.371, -19.160, -18.802,
    -18.288, -17.612, -16.794, -15.796, -14.700, -13.642, -19.477, -19.383,
    -19.172, -18.812, -18.296, -17.620, -16.802, -15.803, -14.705, -13.644,
    -19.488, -19.395, -19.183, -18.822, -18.305, -17.628, -16.810, -15.810,
    -14.709, -13.646, -19.500, -19.407, -19.194, -18.832, -18.314, -17.636,
    -16.818, -15.816, -14.713, -13.647, -19.512, -19.419, -19.205, -18.842,
    -18.323, -17.644, -16.826, -15.823, -14.717, -13.649, -19.524, -19.431,
    -19.217, -18.852, -18.331, -17.651, -16.834, -15.829, -14.721, -13.651,
    -19.535, -19.443, -19.228, -18.862, -18.340, -17.659, -16.842, -15.836,
    -14.725, -13.653, -19.547, -19.455, -19.239, -18.872, -18.349, -17.667,
    -16.849, -15.842, -14.729, -13.654, -19.559, -19.466, -19.250, -18.882,
    -18.357, -17.675, -16.857, -15.848, -14.733, -13.656, -19.570, -19.478,
    -19.261, -18.892, -18.366, -17.683, -16.865, -15.855, -14.737, -13.657,
    -19.582, -19.490, -19.272, -18.902, -18.374, -17.690, -16.873, -15.861,
    -14.740, -13.659, -19.593, -19.502, -19.283, -18.911, -18.383, -17.698,
    -16.881, -15.867, -14.744, -13.661, -19.605, -19.514, -19.294, -18.921,
    -18.391, -17.706, -16.888, -15.873, -14.748, -13.662, -19.617, -19.525,
    -19.305, -18.931, -18.400, -17.713, -16.896, -15.880, -14.752, -13.664,
    -19.628, -19.537, -19.316, -18.941, -18.408, -17.721, -16.904, -15.886,
    -14.755, -13.666, -19.640, -19.549, -19.327, -18.950, -18.417, -17.729,
    -16.911, -15.892, -14.759, -13.668, -19.651, -19.560, -19.338, -18.960,
    -18.425, -17.736, -16.919, -15.898, -14.763, -13.669, -19.662, -19.572,
    -19.348, -18.969, -18.433, -17.744, -16.926, -15.904, -14.766, -13.671,
    -19.674, -19.583, -19.359, -18.979, -18.441, -17.751, -16.934, -15.910,
    -14.770, -13.673, -19.685, -19.595, -19.370, -18.988, -18.450, -17.759,
    -16.942, -15.916, -14.774, -13.675, -19.697, -19.606, -19.380, -18.997,
    -18.458, -17.766, -16.949, -15.919, -14.773, -13.672, -19.708, -19.618,
    -19.391, -19.007, -18.466, -17.773, -16.956, -15.922, -14.771, -13.668,
    -19.719, -19.629, -19.402, -19.016, -18.474, -17.781, -16.964, -15.923,
    -14.766, -13.660, -19.730, -19.641, -19.412, -19.025, -18.482, -17.788,
    -16.971, -15.923, -14.759, -13.650, -19.742, -19.652, -19.423, -19.034,
    -18.490, -17.795, -16.979, -15.927, -14.759, -13.648, -19.753, -19.663,
    -19.433, -19.043, -18.498, -17.802, -16.986, -15.930, -14.758, -13.645,
    -19.764, -19.675, -19.443, -19.052, -18.505, -17.810, -16.993, -15.935,
    -14.761, -13.647, -19.775, -19.686, -19.454, -19.061, -18.513, -17.817,
    -17.001, -15.944, -14.771, -13.655, -19.787, -19.697, -19.464, -19.070,
    -18.521, -17.824, -17.008, -15.950, -14.775, -13.658, -19.798, -19.708,
    -19.474, -19.079, -18.528, -17.831, -17.015, -15.958, -14.782, -13.664,
    -19.809, -19.720, -19.484, -19.088, -18.536, -17.838, -17.022, -15.967,
    -14.792, -13.673, -19.820, -19.731, -19.494, -19.096, -18.544, -17.845,
    -17.029, -15.972, -14.795, -13.675, -19.831, -19.742, -19.504, -19.105,
    -18.551, -17.852, -17.036, -15.980, -14.802, -13.680, -19.842, -19.753,
    -19.514, -19.114, -18.558, -17.859, -17.043, -15.989, -14.811, -13.688,
    -19.853, -19.764, -19.524, -19.122, -18.566, -17.865, -17.050, -15.992,
    -14.811, -13.686, -19.864, -19.775, -19.534, -19.130, -18.573, -17.872,
    -17.057, -15.992, -14.806, -13.679, -19.875, -19.785, -19.544, -19.139,
    -18.580, -17.879, -17.064, -15.993, -14.801, -13.672, -19.886, -19.796,
    -19.553, -19.147, -18.587, -17.885, -17.071, -15.990, -14.789, -13.657,
    -19.897, -19.807, -19.563, -19.155, -18.594, -17.892, -17.078, -15.989,
    -14.783, -13.649, -19.908, -19.818, -19.572, -19.163, -18.601, -17.898,
    -17.085, -15.996, -14.789, -13.654, -19.919, -19.828, -19.582, -19.171,
    -18.608, -17.905, -17.092, -16.004, -14.798, -13.663, -19.930, -19.839,
    -19.591, -19.179, -18.615, -17.911, -17.099, -16.017, -14.814, -13.678,
    -19.940, -19.850, -19.600, -19.186, -18.621, -17.918, -17.105, -16.033,
    -14.835, -13.699, -19.951, -19.860, -19.609, -19.194, -18.628, -17.924,
    -17.112, -16.046, -14.853, -13.716, -19.962, -19.870, -19.618, -19.201,
    -18.634, -17.930, -17.119, -16.058, -14.867, -13.729, -19.973, -19.881,
    -19.627, -19.209, -18.641, -17.936, -17.125, -16.065, -14.874, -13.735,
    -19.983, -19.891, -19.636, -19.216, -18.647, -17.942, -17.132, -16.069,
    -14.874, -13.733, -19.994, -19.901, -19.645, -19.223, -18.653, -17.948,
    -17.138, -16.068, -14.867, -13.725, -20.005, -19.911, -19.653, -19.230,
    -18.659, -17.954, -17.145, -16.064, -14.855, -13.711, -20.015, -19.921,
    -19.662, -19.237, -18.665, -17.960, -17.151, -16.057, -14.840, -13.693,
    -20.026, -19.931, -19.670, -19.243, -18.671, -17.966, -17.157, -16.053,
    -14.827, -13.678, -20.036, -19.941, -19.678, -19.250, -18.677, -17.972,
    -17.164, -16.053, -14.822, -13.671, -20.047, -19.951, -19.686, -19.256,
    -18.682, -17.977, -17.170, -16.055, -14.820, -13.667, -20.057, -19.961,
    -19.694, -19.263, -18.688, -17.983, -17.176, -16.066, -14.833, -13.679,
    -20.068, -19.970, -19.702, -19.269, -18.693, -17.988, -17.182, -16.081,
    -14.853, -13.699, -20.078, -19.980, -19.709, -19.275, -18.698, -17.994,
    -17.188, -16.093, -14.868, -13.713, -20.089, -19.989, -19.716, -19.280,
    -18.703, -17.999, -17.194, -16.106, -14.885, -13.729, -20.099, -19.998,
    -19.724, -19.286, -18.708, -18.004, -17.200, -16.117, -14.898, -13.741,
    -20.109, -20.007, -19.731, -19.291, -18.713, -18.009, -17.206, -16.121,
    -14.899, -13.742, -20.119, -20.016, -19.738, -19.296, -18.718, -18.014,
    -17.212, -16.126, -14.904, -13.745, -20.129, -20.025, -19.744, -19.301,
    -18.722, -18.019, -17.218, -16.134, -14.911, -13.751, -20.139, -20.034,
    -19.751, -19.306, -18.726, -18.024, -17.224, -16.135, -14.909, -13.748,
    -20.149, -20.042, -19.757, -19.311, -18.731, -18.029, -17.230, -16.139,
    -14.911, -13.748, -20.159, -20.051, -19.763, -19.315, -18.735, -18.033,
    -17.235, -16.145, -14.915, -13.751, -20.169, -20.059, -19.769, -19.319,
    -18.738, -18.038, -17.241, -16.145, -14.912, -13.746, -20.179, -20.067,
    -19.774, -19.323, -18.742, -18.042, -17.246, -16.152, -14.918, -13.751,
    -20.189, -20.075, -19.780, -19.327, -18.746, -18.046, -17.252, -16.161,
    -14.929, -13.761, -20.198, -20.083, -19.785, -19.330, -18.749, -18.051,
    -17.257, -16.171, -14.939, -13.770, -20.208, -20.090, -19.789, -19.334,
    -18.752, -18.055, -17.262, -16.185, -14.959, -13.789, -20.217, -20.097,
    -19.794, -19.336, -18.755, -18.059, -17.268, -16.199, -14.977, -13.806,
    -20.226, -20.104, -19.798, -19.339, -18.758, -18.062, -17.273, -16.214,
    -14.997, -13.825, -20.236, -20.111, -19.802, -19.342, -18.760, -18.066,
    -17.278, -16.220, -15.003, -13.830, -20.245, -20.118, -19.806, -19.344,
    -18.763, -18.070, -17.283, -16.226, -15.009, -13.834, -20.254, -20.124,
    -19.809, -19.346, -18.765, -18.073, -17.288, -16.232, -15.015, -13.839,
    -20.263, -20.130, -19.812, -19.347, -18.767, -18.076, -17.293, -16.239,
    -15.021, -13.844, -20.271, -20.136, -19.815, -19.348, -18.768, -18.079,
    -17.298, -16.245, -15.028, -13.850, -20.280, -20.141, -19.817, -19.349,
    -18.770, -18.083, -17.302, -16.252, -15.034, -13.855, -20.288, -20.147,
    -19.819, -19.350, -18.771, -18.085, -17.307, -16.258, -15.041, -13.861,
    -20.297, -20.152, -19.821, -19.350, -18.772, -18.088, -17.312, -16.265,
    -15.048, -13.867, -20.305, -20.156, -19.822, -19.350, -18.773, -18.091,
    -17.316, -16.271, -15.055, -13.873, -20.313, -20.160, -19.823, -19.350,
    -18.773, -18.093, -17.321, -16.278, -15.062, -13.879, -20.321, -20.164,
    -19.823, -19.349, -18.774, -18.096, -17.325, -16.284, -15.069, -13.885,
    -20.328, -20.168, -19.823, -19.348, -18.774, -18.098, -17.329, -16.291,
    -15.076, -13.891, -20.335, -20.171, -19.823, -19.347, -18.774, -18.100,
    -17.333, -16.298, -15.084, -13.898, -20.343, -20.173, -19.822, -19.345,
    -18.773, -18.102, -17.338, -16.304, -15.091, -13.904, -20.349, -20.176,
    -19.821, -19.343, -18.772, -18.103, -17.342, -16.311, -15.099, -13.911,
    -20.356, -20.178, -19.819, -19.340, -18.771, -18.105, -17.345, -16.318,
    -15.106, -13.918, -20.362, -20.179, -19.817, -19.338, -18.770, -18.106,
    -17.349, -16.325, -15.114, -13.925, -20.368, -20.180, -19.814, -19.334,
    -18.769, -18.107, -17.353, -16.331, -15.122, -13.932, -20.374, -20.180,
    -19.811, -19.331, -18.767, -18.108, -17.357, -16.338, -15.130, -13.940,
    -20.379, -20.180, -19.807, -19.327, -18.765, -18.109, -17.360, -16.345,
    -15.138, -13.947, -20.385, -20.179, -19.803, -19.322, -18.762, -18.110,
    -17.364, -16.352, -15.147, -13.954, -20.389, -20.178, -19.798, -19.317,
    -18.760, -18.110, -17.367, -16.359, -15.155, -13.962, -20.394, -20.177,
    -19.792, -19.312, -18.757, -18.111, -17.370, -16.365, -15.163, -13.970,
    -20.397, -20.174, -19.787, -19.306, -18.753, -18.111, -17.373, -16.372,
    -15.172, -13.978, -20.401, -20.171, -19.780, -19.300, -18.750, -18.111,
    -17.376, -16.379, -15.180, -13.986, -20.404, -20.168, -19.773, -19.293,
    -18.746, -18.110, -17.379, -16.386, -15.189, -13.994, -20.407, -20.163,
    -19.766, -19.286, -18.742, -18.110, -17.382, -16.392, -15.197, -14.002,
    -20.409, -20.159, -19.758, -19.279, -18.737, -18.109, -17.385, -16.399,
    -15.206, -14.010, -20.410, -20.153, -19.749, -19.271, -18.733, -18.108,
    -17.388, -16.406, -15.214, -14.018, -20.411, -20.147, -19.740, -19.263,
    -18.728, -18.107, -17.390, -16.412, -15.223, -14.026, -20.411, -20.140,
    -19.730, -19.254, -18.722, -18.106, -17.393, -16.419, -15.232, -14.035,
    -20.411, -20.132, -19.719, -19.244, -18.716, -18.105, -17.395, -16.425,
    -15.241, -14.043, -20.410, -20.124, -19.708, -19.235, -18.710, -18.103,
    -17.397, -16.432, -15.250, -14.052, -20.409, -20.114, -19.696, -19.225,
    -18.704, -18.101, -17.399, -16.438, -15.258, -14.060, -20.406, -20.104,
    -19.684, -19.214, -18.697, -18.099, -17.401, -16.445, -15.267, -14.069,
    -20.403, -20.094, -19.671, -19.203, -18.690, -18.097, -17.403, -16.451,
    -15.276, -14.078, -20.399, -20.082, -19.657, -19.191, -18.683, -18.095,
    -17.405, -16.458, -15.285, -14.087, -20.395, -20.070, -19.643, -19.179,
    -18.675, -18.092, -17.406, -16.464, -15.294, -14.096, -20.389, -20.057,
    -19.628, -19.167, -18.668, -18.089, -17.408, -16.470, -15.303, -14.104,
    -20.383, -20.043, -19.613, -19.154, -18.659, -18.086, -17.409, -16.476,
    -15.312, -14.113, -20.376, -20.028, -19.597, -19.141, -18.651, -18.083,
    -17.410, -16.482, -15.321, -14.122, -20.367, -20.012, -19.580, -19.127,
    -18.642, -18.079, -17.412, -16.488, -15.330, -14.131, -20.358, -19.996,
    -19.563, -19.113, -18.633, -18.075, -17.413, -16.495, -15.339, -14.140,
    -20.348, -19.978, -19.545, -19.098, -18.623, -18.072, -17.414, -16.500,
    -15.348, -14.150, -20.337, -19.960, -19.527, -19.083, -18.613, -18.067,
    -17.414, -16.506, -15.357, -14.159, -20.325, -19.941, -19.508, -19.068,
    -18.603, -18.063, -17.415, -16.512, -15.366, -14.168, -20.312, -19.921,
    -19.488, -19.052, -18.593, -18.059, -17.416, -16.518, -15.375, -14.177,
    -20.297, -19.901, -19.468, -19.036, -18.582, -18.054, -17.416, -16.524,
    -15.384, -14.186, -20.282, -19.879, -19.447, -19.019, -18.571, -18.049,
    -17.416, -16.529, -15.393, -14.196, -20.265, -19.857, -19.426, -19.002,
    -18.559, -18.044, -17.416, -16.535, -15.402, -14.205, -20.247, -19.834,
    -19.404, -18.985, -18.548, -18.038, -17.417, -16.541, -15.410, -14.214,
    -20.228, -19.810, -19.382, -18.967, -18.536, -18.033, -17.416, -16.546,
    -15.419, -14.223, -20.208, -19.785, -19.359, -18.948, -18.524, -18.027,
    -17.416, -16.551, -15.428, -14.233, -20.187, -19.759, -19.335, -18.930,
    -18.511, -18.021, -17.416, -16.557, -15.437, -14.242, -20.164, -19.733,
    -19.311, -18.911, -18.498, -18.015, -17.416, -16.562, -15.446, -14.252,
    -20.141, -19.706, -19.287, -18.892, -18.485, -18.008, -17.415, -16.567,
    -15.455, -14.261, -20.116, -19.678, -19.262, -18.872, -18.472, -18.002,
    -17.414, -16.572, -15.464, -14.270, -20.090, -19.650, -19.237, -18.852,
    -18.458, -17.995, -17.413, -16.577, -15.472, -14.280, -20.062, -19.621,
    -19.211, -18.832, -18.445, -17.988, -17.413, -16.582, -15.481, -14.289,
    -20.034, -19.591, -19.185, -18.811, -18.431, -17.981, -17.412, -16.587,
    -15.490, -14.298, -20.004, -19.560, -19.158, -18.790, -18.416, -17.974,
    -17.410, -16.592, -15.498, -14.308, -19.974, -19.529, -19.131, -18.769,
    -18.402, -17.966, -17.409, -16.597, -15.507, -14.317, -19.942, -19.497,
    -19.104, -18.748, -18.387, -17.958, -17.408, -16.601, -15.516, -14.327,
    -19.909, -19.465, -19.076, -18.726, -18.372, -17.951, -17.406, -16.606,
    -15.524, -14.336, -19.875, -19.432, -19.048, -18.704, -18.357, -17.943,
    -17.405, -16.611, -15.533, -14.345, -19.840, -19.398, -19.020, -18.682,
    -18.342, -17.934, -17.403, -16.615, -15.541, -14.355, -19.804, -19.364,
    -18.991, -18.659, -18.326, -17.926, -17.401, -16.619, -15.550, -14.364,
    -19.767, -19.330, -18.962, -18.637, -18.310, -17.918, -17.399, -16.624,
    -15.558, -14.374, -19.729, -19.295, -18.932, -18.614, -18.294, -17.909,
    -17.397, -16.628, -15.567, -14.383, -19.690, -19.259, -18.903, -18.591,
    -18.278, -17.900, -17.395, -16.632, -15.575, -14.392, -19.650, -19.223,
    -18.873, -18.567, -18.262, -17.891, -17.393, -16.636, -15.583, -14.402,
    -19.609, -19.187, -18.843, -18.544, -18.245, -17.882, -17.390, -16.640,
    -15.592, -14.411, -19.568, -19.151, -18.813, -18.520, -18.229, -17.873,
    -17.388, -16.644, -15.600, -14.420, -19.526, -19.114, -18.782, -18.497,
    -18.212, -17.863, -17.385, -16.648, -15.608, -14.429, -19.483, -19.076,
    -18.751, -18.473, -18.195, -17.854, -17.383, -16.652, -15.616, -14.439,
    -19.439, -19.039, -18.721, -18.449, -18.178, -17.844, -17.380, -16.656,
    -15.625, -14.448, -19.395, -19.001, -18.690, -18.425, -18.161, -17.835,
    -17.377, -16.659, -15.633, -14.457, -19.350, -18.962, -18.658, -18.401,
    -18.144, -17.825, -17.374, -16.663, -15.641, -14.466, -19.305, -18.924,
    -18.627, -18.376, -18.127, -17.815, -17.371, -16.667, -15.649, -14.476,
    -19.259, -18.885, -18.596, -18.352, -18.110, -17.805, -17.368, -16.670,
    -15.657, -14.485, -19.212, -18.847, -18.564, -18.328, -18.092, -17.795,
    -17.365, -16.673, -15.665, -14.494, -19.165, -18.808, -18.533, -18.303,
    -18.075, -17.785, -17.362, -16.677, -15.672, -14.503, -19.118, -18.769,
    -18.501, -18.279, -18.057, -17.775, -17.359, -16.680, -15.680, -14.512,
    -19.071, -18.730, -18.470, -18.254, -18.040, -17.764, -17.355, -16.683,
    -15.688, -14.521, -19.023, -18.690, -18.438, -18.230, -18.022, -17.754,
    -17.352, -16.687, -15.696, -14.530, -18.975, -18.651, -18.407, -18.205,
    -18.005, -17.744, -17.348, -16.690, -15.704, -14.539, -18.926, -18.612,
    -18.375, -18.181, -17.987, -17.733, -17.345, -16.693, -15.711, -14.548,
    -18.878, -18.572, -18.344, -18.156, -17.970, -17.723, -17.341, -16.696,
    -15.719, -14.557, -18.829, -18.533, -18.312, -18.132, -17.952, -17.713,
    -17.338, -16.699, -15.726, -14.566, -18.780, -18.494, -18.281, -18.108,
    -17.935, -17.702, -17.334, -16.702, -15.734, -14.575, -18.732, -18.455,
    -18.250, -18.084, -17.918, -17.692, -17.331, -16.705, -15.741, -14.584,
    -18.683, -18.415, -18.219, -18.060, -17.900, -17.681, -17.327, -16.708,
    -15.749, -14.593, -18.634, -18.376, -18.188, -18.036, -17.883, -17.671,
    -17.323, -16.710, -15.756, -14.602, -18.585, -18.338, -18.157, -18.012,
    -17.866, -17.661, -17.320, -16.713, -15.764, -14.610, -18.536, -18.299,
    -18.126, -17.988, -17.849, -17.650, -17.316, -16.716, -15.771, -14.619,
    -18.488, -18.260, -18.096, -17.965, -17.832, -17.640, -17.312, -16.719,
    -15.778, -14.628, -18.439, -18.222, -18.065, -17.941, -17.815, -17.630,
    -17.308, -16.721, -15.786, -14.637, -18.391, -18.184, -18.035, -17.918,
    -17.798, -17.620, -17.305, -16.724, -15.793, -14.645, -18.343, -18.146,
    -18.006, -17.895, -17.782, -17.609, -17.301, -16.726, -15.800, -14.654,
    -18.295, -18.109, -17.976, -17.873, -17.765, -17.599, -17.297, -16.729,
    -15.807, -14.662, -18.248, -18.072, -17.947, -17.850, -17.749, -17.590,
    -17.294, -16.732, -15.814, -14.671, -18.201, -18.035, -17.918, -17.828,
    -17.733, -17.580, -17.290, -16.734, -15.821, -14.680, -18.154, -17.999,
    -17.890, -17.806, -17.717, -17.570, -17.287, -16.737, -15.828, -14.688,
    -18.108, -17.963, -17.861, -17.785, -17.702, -17.560, -17.283, -16.739,
    -15.835, -14.696, -18.063, -17.927, -17.834, -17.763, -17.686, -17.551,
    -17.280, -16.742, -15.842, -14.705, -18.017, -17.892, -17.806, -17.742,
    -17.671, -17.542, -17.276, -16.744, -15.849, -14.713, -17.973, -17.858,
    -17.779, -17.722, -17.656, -17.533, -17.273, -16.747, -15.856, -14.722,
    -17.929, -17.824, -17.753, -17.702, -17.642, -17.524, -17.270, -16.749,
    -15.863, -14.730, -17.886, -17.791, -17.727, -17.682, -17.628, -17.515,
    -17.266, -16.752, -15.870, -14.738, -17.843, -17.758, -17.702, -17.662,
    -17.614, -17.506, -17.263, -16.754, -15.877, -14.747, -17.801, -17.726,
    -17.677, -17.643, -17.600, -17.498, -17.260, -16.757, -15.883, -14.755,
    -17.760, -17.694, -17.652, -17.625, -17.586, -17.490, -17.257, -16.759,
    -15.890, -14.763, -17.720, -17.664, -17.629, -17.607, -17.573, -17.482,
    -17.254, -16.761, -15.897, -14.771, -17.680, -17.634, -17.606, -17.589,
    -17.561, -17.474, -17.252, -16.764, -15.903, -14.779, -17.642, -17.605,
    -17.583, -17.572, -17.549, -17.466, -17.249, -16.767, -15.910, -14.788,
    -17.605, -17.576, -17.561, -17.556, -17.537, -17.459, -17.246, -16.769,
    -15.917, -14.796, -17.568, -17.549, -17.540, -17.540, -17.525, -17.452,
    -17.244, -16.772, -15.923, -14.804, -17.533, -17.522, -17.520, -17.524,
    -17.514, -17.445, -17.242, -16.774, -15.930, -14.812, -17.499, -17.497,
    -17.500, -17.509, -17.503, -17.439, -17.240, -16.777, -15.936, -14.820,
    -17.466, -17.472, -17.482, -17.495, -17.493, -17.432, -17.238, -16.780,
    -15.943, -14.828, -17.435, -17.449, -17.464, -17.482, -17.484, -17.426,
    -17.236, -16.782, -15.950, -14.836, -17.064, -17.151, -17.231, -17.303,
    -17.352, -17.338, -17.185, -16.764, -15.950, -14.843, -16.931, -17.042,
    -17.144, -17.236, -17.301, -17.303, -17.166, -16.759, -15.954, -14.850,
    -16.845, -16.973, -17.089, -17.193, -17.270, -17.281, -17.155, -16.757,
    -15.959, -14.858, -16.809, -16.949, -17.072, -17.181, -17.262, -17.277,
    -17.154, -16.760, -15.966, -14.866, -16.774, -16.925, -17.055, -17.170,
    -17.254, -17.272, -17.153, -16.763, -15.972, -14.873, -16.740, -16.902,
    -17.040, -17.159, -17.247, -17.268, -17.152, -16.766, -15.979, -14.881,
    -16.713, -16.884, -17.028, -17.151, -17.242, -17.266, -17.152, -16.769,
    -15.985, -14.889, -16.753, -16.921, -17.060, -17.177, -17.262, -17.281,
    -17.162, -16.777, -15.993, -14.897, -16.710, -16.887, -17.033, -17.156,
    -17.246, -17.270, -17.157, -16.778, -15.999, -14.905, -16.649, -16.836,
    -16.991, -17.123, -17.221, -17.252, -17.147, -16.777, -16.004, -14.912,
    -16.566, -16.766, -16.933, -17.076, -17.184, -17.226, -17.133, -16.773,
    -16.008, -14.920, -16.402, -16.625, -16.815, -16.980, -17.110, -17.173,
    -17.101, -16.761, -16.010, -14.926, -16.391, -16.614, -16.804, -16.970,
    -17.101, -17.167, -17.098, -16.763, -16.015, -14.932, -16.344, -16.578,
    -16.777, -16.950, -17.086, -17.156, -17.093, -16.763, -16.020, -14.939,
    -16.289, -16.535, -16.744, -16.924, -17.067, -17.144, -17.086, -16.763,
    -16.025, -14.944, -16.323, -16.573, -16.780, -16.957, -17.095, -17.165,
    -17.101, -16.774, -16.033, -14.952, -16.252, -16.523, -16.746, -16.933,
    -17.079, -17.156, -17.097, -16.775, -16.038, -14.959, -16.229, -16.508,
    -16.737, -16.928, -17.077, -17.155, -17.098, -16.778, -16.044, -14.966,
    -16.317, -16.590, -16.808, -16.987, -17.122, -17.188, -17.120, -16.792,
    -16.054, -14.975, -16.341, -16.615, -16.830, -17.006, -17.137, -17.200,
    -17.129, -16.798, -16.061, -14.983, -16.335, -16.615, -16.833, -17.009,
    -17.141, -17.203, -17.132, -16.803, -16.068, -14.991, -16.415, -16.687,
    -16.893, -17.057, -17.177, -17.229, -17.149, -16.813, -16.077, -15.000,
    -16.341, -16.625, -16.842, -17.015, -17.145, -17.206, -17.136, -16.810,
    -16.080, -15.007, -16.216, -16.512, -16.745, -16.935, -17.081, -17.159,
    -17.107, -16.799, -16.082, -15.014, -16.215, -16.495, -16.719, -16.907,
    -17.055, -17.138, -17.093, -16.794, -16.086, -15.022, -16.082, -16.369,
    -16.607, -16.811, -16.977, -17.079, -17.056, -16.779, -16.086, -15.027,
    -15.913, -16.220, -16.480, -16.706, -16.893, -17.016, -17.016, -16.761,
    -16.085, -15.032, -15.810, -16.139, -16.419, -16.660, -16.859, -16.993,
    -17.002, -16.757, -16.088, -15.037, -15.742, -16.095, -16.391, -16.644,
    -16.851, -16.990, -17.002, -16.761, -16.093, -15.042, -15.741, -16.112,
    -16.418, -16.674, -16.880, -17.015, -17.021, -16.773, -16.101, -15.049,
    -15.851, -16.226, -16.525, -16.769, -16.959, -17.076, -17.063, -16.797,
    -16.114, -15.058, -16.126, -16.470, -16.729, -16.933, -17.086, -17.168,
    -17.122, -16.828, -16.129, -15.068, -16.440, -16.727, -16.928, -17.082,
    -17.195, -17.243, -17.169, -16.853, -16.142, -15.079, -16.651, -16.885,
    -17.041, -17.162, -17.250, -17.280, -17.191, -16.866, -16.152, -15.089,
    -16.564, -16.821, -16.995, -17.129, -17.226, -17.265, -17.184, -16.865,
    -16.158, -15.099, -16.233, -16.562, -16.803, -16.989, -17.126, -17.197,
    -17.144, -16.849, -16.158, -15.106, -15.904, -16.286, -16.582, -16.817,
    -16.998, -17.107, -17.089, -16.826, -16.156, -15.112, -15.653, -16.059,
    -16.387, -16.657, -16.872, -17.014, -17.030, -16.798, -16.152, -15.116,
    -15.551, -15.955, -16.289, -16.570, -16.798, -16.956, -16.991, -16.781,
    -16.150, -15.120, -15.552, -15.942, -16.268, -16.546, -16.774, -16.936,
    -16.977, -16.775, -16.151, -15.124, -15.593, -15.968, -16.284, -16.554,
    -16.778, -16.937, -16.978, -16.778, -16.155, -15.128, -15.705, -16.063,
    -16.361, -16.616, -16.826, -16.973, -17.002, -16.793, -16.164, -15.133,
    -15.828, -16.181, -16.467, -16.706, -16.901, -17.031, -17.043, -16.816,
    -16.175, -15.140, -15.865, -16.230, -16.518, -16.755, -16.943, -17.065,
    -17.067, -16.832, -16.185, -15.147, -15.877, -16.255, -16.550, -16.787,
    -16.972, -17.089, -17.085, -16.843, -16.194, -15.156, -15.948, -16.323,
    -16.610, -16.838, -17.013, -17.121, -17.107, -16.857, -16.204, -15.167,
    -15.929, -16.305, -16.595, -16.826, -17.004, -17.115, -17.104, -16.858,
    -16.210, -15.175, -15.804, -16.198, -16.507, -16.756, -16.952, -17.077,
    -17.081, -16.849, -16.213, -15.184, -15.807, -16.197, -16.505, -16.754,
    -16.949, -17.075, -17.080, -16.851, -16.218, -15.192, -15.696, -16.100,
    -16.424, -16.688, -16.898, -17.038, -17.057, -16.842, -16.220, -15.198,
    -15.583, -16.000, -16.339, -16.618, -16.843, -16.997, -17.031, -16.831,
    -16.220, -15.204, -15.645, -16.048, -16.375, -16.646, -16.863, -17.011,
    -17.041, -16.838, -16.227, -15.212, -15.616, -16.016, -16.344, -16.618,
    -16.839, -16.993, -17.029, -16.834, -16.230, -15.217, -15.576, -15.978,
    -16.310, -16.588, -16.814, -16.974, -17.017, -16.830, -16.233, -15.223,
    -15.605, -16.002, -16.329, -16.603, -16.826, -16.983, -17.024, -16.835,
    -16.239, -15.230, -15.580, -15.980, -16.310, -16.588, -16.814, -16.974,
    -17.019, -16.835, -16.243, -15.235, -15.626, -16.026, -16.352, -16.624,
    -16.844, -16.998, -17.036, -16.846, -16.251, -15.243, -15.736, -16.122,
    -16.434, -16.692, -16.898, -17.039, -17.064, -16.863, -16.262, -15.252,
    -15.761, -16.145, -16.454, -16.709, -16.912, -17.049, -17.072, -16.870,
    -16.269, -15.260, -15.786, -16.168, -16.474, -16.726, -16.926, -17.060,
    -17.080, -16.876, -16.276, -15.270, -15.770, -16.153, -16.461, -16.715,
    -16.917, -17.054, -17.077, -16.877, -16.282, -15.279, -15.696, -16.090,
    -16.409, -16.673, -16.885, -17.030, -17.062, -16.871, -16.285, -15.287,
    -15.693, -16.084, -16.402, -16.666, -16.879, -17.026, -17.060, -16.872,
    -16.290, -15.294, -15.690, -16.078, -16.396, -16.660, -16.873, -17.021,
    -17.057, -16.873, -16.294, -15.301, -15.688, -16.074, -16.390, -16.655,
    -16.868, -17.018, -17.055, -16.874, -16.299, -15.308, -15.689, -16.071,
    -16.387, -16.651, -16.865, -17.014, -17.054, -16.875, -16.304, -15.315,
    -15.777, -16.146, -16.447, -16.699, -16.902, -17.043, -17.073, -16.887,
    -16.312, -15.323, -15.879, -16.229, -16.515, -16.753, -16.944, -17.073,
    -17.094, -16.900, -16.321, -15.330, -16.171, -16.464, -16.699, -16.894,
    -17.050, -17.149, -17.144, -16.928, -16.335, -15.339, -16.167, -16.459,
    -16.695, -16.891, -17.048, -17.148, -17.144, -16.930, -16.340, -15.346,
    -16.164, -16.456, -16.692, -16.889, -17.046, -17.147, -17.144, -16.932,
    -16.345, -15.353, -16.162, -16.454, -16.690, -16.887, -17.044, -17.146,
    -17.144, -16.934, -16.350, -15.360, -16.162, -16.452, -16.689, -16.886,
    -17.044, -17.146, -17.145, -16.937, -16.355, -15.368, -16.162, -16.452,
    -16.688, -16.886, -17.043, -17.145, -17.145, -16.939, -16.360, -15.375,
    -16.164, -16.452, -16.688, -16.886, -17.043, -17.146, -17.146, -16.942,
    -16.365, -15.382, -16.166, -16.454, -16.689, -16.887, -17.044, -17.146,
    -17.147, -16.944, -16.371, -15.389, -16.170, -16.456, -16.691, -16.888,
    -17.045, -17.147, -17.149, -16.947, -16.376, -15.396, -16.175, -16.459,
    -16.693, -16.890, -17.046, -17.148, -17.150, -16.950, -16.381, -15.403,
    -16.181, -16.463, -16.696, -16.892, -17.048, -17.150, -17.152, -16.953,
    -16.386, -15.410, -16.187, -16.467, -16.700, -16.895, -17.051, -17.152,
    -17.154, -16.956, -16.392, -15.417, -16.194, -16.472, -16.704, -16.898,
    -17.053, -17.154, -17.156, -16.959, -16.397, -15.425, -16.203, -16.478,
    -16.709, -16.902, -17.056, -17.157, -17.159, -16.962, -16.402, -15.432,
    -16.211, -16.485, -16.714, -16.906, -17.060, -17.159, -17.161, -16.965,
    -16.408, -15.439, -16.221, -16.492, -16.720, -16.911, -17.064, -17.162,
    -17.164, -16.969, -16.413, -15.446, -16.231, -16.500, -16.726, -16.916,
    -17.068, -17.166, -17.167, -16.972, -16.418, -15.453, -16.242, -16.508,
    -16.733, -16.922, -17.072, -17.169, -17.170, -16.976, -16.424, -15.461,
    -16.254, -16.517, -16.740, -16.928, -17.077, -17.173, -17.173, -16.980,
    -16.429, -15.468, -16.266, -16.526, -16.748, -16.934, -17.082, -17.177,
    -17.177, -16.983, -16.435, -15.475, -16.279, -16.536, -16.756, -16.941,
    -17.088, -17.182, -17.181, -16.987, -16.440, -15.482, -16.292, -16.546,
    -16.765, -16.948, -17.093, -17.186, -17.184, -16.991, -16.446, -15.490,
    -16.306, -16.557, -16.774, -16.956, -17.099, -17.191, -17.189, -16.996,
    -16.452, -15.497, -16.320, -16.568, -16.783, -16.963, -17.106, -17.196,
    -17.193, -17.000, -16.457, -15.504, -16.335, -16.580, -16.793, -16.971,
    -17.112, -17.201, -17.197, -17.004, -16.463, -15.512, -16.350, -16.592,
    -16.803, -16.980, -17.119, -17.207, -17.202, -17.008, -16.469, -15.519,
    -16.366, -16.604, -16.813, -16.988, -17.126, -17.213, -17.206, -17.013,
    -16.474, -15.526, -16.382, -16.617, -16.824, -16.997, -17.134, -17.219,
    -17.211, -17.018, -16.480, -15.534, -16.398, -16.630, -16.835, -17.007,
    -17.141, -17.225, -17.216, -17.022, -16.486, -15.541, -16.415, -16.643,
    -16.846, -17.016, -17.149, -17.231, -17.221, -17.027, -16.491, -15.548,
    -16.432, -16.657, -16.858, -17.026, -17.157, -17.238, -17.227, -17.032,
    -16.497, -15.556, -16.449, -16.671, -16.869, -17.036, -17.165, -17.244,
    -17.232, -17.037, -16.503, -15.563, -16.466, -16.685, -16.881, -17.046,
    -17.174, -17.251, -17.238, -17.042, -16.509, -15.570, -16.484, -16.699,
    -16.894, -17.056, -17.183, -17.258, -17.243, -17.047, -16.515, -15.578,
    -16.502, -16.714, -16.906, -17.067, -17.192, -17.266, -17.249, -17.052,
    -16.521, -15.585, -16.521, -16.729, -16.919, -17.078, -17.201, -17.273,
    -17.255, -17.057, -16.527, -15.592, -16.539, -16.744, -16.932, -17.089,
    -17.210, -17.281, -17.261, -17.062, -16.533, -15.600, -16.558, -16.759,
    -16.945, -17.100, -17.219, -17.288, -17.268, -17.068, -16.539, -15.607,
    -16.577, -16.775, -16.958, -17.112, -17.229, -17.296, -17.274, -17.073,
    -16.545, -15.614, -16.597, -16.790, -16.972, -17.123, -17.239, -17.304,
    -17.280, -17.079, -16.551, -15.622, -16.616, -16.806, -16.985, -17.135,
    -17.249, -17.312, -17.287, -17.084, -16.557, -15.629, -16.636, -16.822,
    -16.999, -17.147, -17.259, -17.321, -17.294, -17.090, -16.563, -15.637,
    -16.655, -16.838, -17.013, -17.159, -17.269, -17.329, -17.300, -17.096,
    -16.569, -15.644, -16.675, -16.855, -17.027, -17.171, -17.279, -17.337,
    -17.307, -17.102, -16.575, -15.651, -16.695, -16.871, -17.041, -17.183,
    -17.290, -17.346, -17.314, -17.107, -16.581, -15.659, -16.715, -16.888,
    -17.056, -17.196, -17.300, -17.355, -17.321, -17.113, -16.587, -15.666,
    -16.735, -16.904, -17.070, -17.209, -17.311, -17.364, -17.328, -17.119,
    -16.594, -15.673, -16.756, -16.921, -17.085, -17.221, -17.322, -17.373,
    -17.336, -17.125, -16.600, -15.681, -16.776, -16.938, -17.100, -17.234,
    -17.333, -17.382, -17.343, -17.131, -16.606, -15.688, -16.797, -16.955,
    -17.114, -17.247, -17.344, -17.391, -17.351, -17.138, -16.612, -15.696,
    -16.817, -16.972, -17.129, -17.260, -17.355, -17.400, -17.358, -17.144,
    -16.619, -15.703, -16.838, -16.989, -17.144, -17.273, -17.366, -17.410,
    -17.366, -17.150, -16.625, -15.710, -16.859, -17.006, -17.159, -17.286,
    -17.378, -17.419, -17.373, -17.156, -16.631, -15.718, -16.880, -17.024,
    -17.175, -17.300, -17.389, -17.429, -17.381, -17.163, -16.637, -15.725,
    -16.900, -17.041, -17.190, -17.313, -17.401, -17.439, -17.389, -17.169,
    -16.644, -15.732, -16.921, -17.059, -17.205, -17.327, -17.412, -17.448,
    -17.397, -17.176, -16.650, -15.740, -16.942, -17.076, -17.221, -17.340,
    -17.424, -17.458, -17.405, -17.182, -16.656, -15.747, -16.963, -17.094,
    -17.236, -17.354, -17.436, -17.468, -17.413, -17.189, -16.663, -15.754,
    -16.984, -17.111, -17.252, -17.368, -17.448, -17.478, -17.421, -17.195,
    -16.669, -15.762, -17.005, -17.129, -17.267, -17.381, -17.460, -17.488,
    -17.429, -17.202, -16.676, -15.769, -17.026, -17.147, -17.283, -17.395,
    -17.472, -17.498, -17.438, -17.209, -16.682, -15.777, -17.047, -17.164,
    -17.299, -17.409, -17.484, -17.509, -17.446, -17.216, -16.689, -15.784,
    -17.068, -17.182, -17.314, -17.423, -17.496, -17.519, -17.454, -17.222,
    -16.695, -15.791, -17.089, -17.200, -17.330, -17.437, -17.508, -17.529,
    -17.463, -17.229, -16.701, -15.799, -17.110, -17.217, -17.346, -17.451,
    -17.521, -17.540, -17.471, -17.236, -16.708, -15.806, -17.131, -17.235,
    -17.362, -17.465, -17.533, -17.550, -17.480, -17.243, -16.714, -15.813,
    -17.152, -17.253, -17.378, -17.479, -17.545, -17.561, -17.488, -17.250,
    -16.721, -15.820, -17.173, -17.271, -17.394, -17.494, -17.558, -17.571,
    -17.497, -17.257, -16.727, -15.828, -17.194, -17.289, -17.409, -17.508,
    -17.570, -17.582, -17.506, -17.264, -16.734, -15.835, -17.215, -17.307,
    -17.425, -17.522, -17.583, -17.593, -17.515, -17.271, -16.741, -15.842,
    -17.236, -17.324, -17.441, -17.536, -17.595, -17.603, -17.523, -17.278,
    -16.747, -15.850, -17.256, -17.342, -17.457, -17.551, -17.608, -17.614,
    -17.532, -17.285, -16.754, -15.857, -17.277, -17.360, -17.473, -17.565,
    -17.621, -17.625, -17.541, -17.292, -16.760, -15.864, -17.298, -17.378,
    -17.489, -17.579, -17.633, -17.636, -17.550, -17.300, -16.767, -15.872,
    -17.319, -17.396, -17.505, -17.594, -17.646, -17.647, -17.559, -17.307,
    -16.773, -15.879, -17.339, -17.413, -17.521, -17.608, -17.659, -17.658,
    -17.568, -17.314, -16.780, -15.886, -17.360, -17.431, -17.537, -17.623,
    -17.671, -17.669, -17.577, -17.321, -16.787, -15.893, -17.380, -17.449,
    -17.553, -17.637, -17.684, -17.680, -17.586, -17.329, -16.793, -15.901,
    -17.401, -17.467, -17.570, -17.651, -17.697, -17.691, -17.595, -17.336,
    -16.800, -15.908, -17.421, -17.484, -17.586, -17.666, -17.710, -17.702,
    -17.604, -17.343, -16.807, -15.915, -17.441, -17.502, -17.602, -17.680,
    -17.723, -17.713, -17.614, -17.351, -16.813, -15.922, -17.462, -17.520,
    -17.618, -17.695, -17.736, -17.724, -17.623, -17.358, -16.820, -15.930,
    -17.482, -17.537, -17.634, -17.709, -17.749, -17.735, -17.632, -17.366,
    -16.827, -15.937, -17.502, -17.555, -17.650, -17.724, -17.762, -17.746,
    -17.641, -17.373, -16.833, -15.944, -17.522, -17.572, -17.666, -17.738,
    -17.774, -17.757, -17.651, -17.381, -16.840, -15.951, -17.542, -17.590,
    -17.682, -17.753, -17.787, -17.769, -17.660, -17.388, -16.847, -15.959,
    -17.562, -17.607, -17.698, -17.768, -17.800, -17.780, -17.669, -17.396,
    -16.853, -15.966, -17.582, -17.625, -17.714, -17.782, -17.813, -17.791,
    -17.679, -17.403, -16.860, -15.973, -17.601, -17.642, -17.729, -17.797,
    -17.826, -17.803, -17.688, -17.411, -16.867, -15.980, -17.621, -17.660,
    -17.745, -17.811, -17.839, -17.814, -17.697, -17.418, -16.874, -15.987,
    -17.641, -17.677, -17.761, -17.826, -17.852, -17.825, -17.707, -17.426,
    -16.880, -15.995, -17.660, -17.694, -17.777, -17.840, -17.865, -17.837,
    -17.716, -17.434, -16.887, -16.002, -17.679, -17.712, -17.793, -17.855,
    -17.879, -17.848, -17.726, -17.441, -16.894, -16.009, -17.699, -17.729,
    -17.809, -17.869, -17.892, -17.859, -17.735, -17.449, -16.900, -16.016,
    -17.718, -17.746, -17.825, -17.884, -17.905, -17.871, -17.745, -17.456,
    -16.907, -16.023, -17.737, -17.763, -17.841, -17.898, -17.918, -17.882,
    -17.754, -17.464, -16.914, -16.030, -17.756, -17.780, -17.856, -17.913,
    -17.931, -17.894, -17.764, -17.472, -16.921, -16.037, -17.775, -17.797,
    -17.872, -17.927, -17.944, -17.905, -17.773, -17.480, -16.927, -16.045,
    -17.794, -17.814, -17.888, -17.942, -17.957, -17.916, -17.783, -17.487,
    -16.934, -16.052, -17.812, -17.831, -17.904, -17.956, -17.970, -17.928,
    -17.793, -17.495, -16.941, -16.059, -17.831, -17.848, -17.919, -17.971,
    -17.983, -17.939, -17.802, -17.503, -16.948, -16.066, -17.849, -17.865,
    -17.935, -17.985, -17.996, -17.951, -17.812, -17.511, -16.955, -16.073,
    -17.868, -17.882, -17.951, -17.999, -18.009, -17.962, -17.822, -17.518,
    -16.961, -16.080, -17.886, -17.899, -17.966, -18.014, -18.022, -17.974,
    -17.831, -17.526, -16.968, -16.087, -17.904, -17.915, -17.982, -18.028,
    -18.035, -17.985, -17.841, -17.534, -16.975, -16.094, -17.922, -17.932,
    -17.997, -18.043, -18.048, -17.997, -17.850, -17.542, -16.982, -16.101,
    -17.940, -17.949, -18.013, -18.057, -18.061, -18.008, -17.860, -17.549,
    -16.988, -16.108, -17.958, -17.965, -18.028, -18.071, -18.074, -18.020,
    -17.870, -17.557, -16.995, -16.115, -17.976, -17.982, -18.044, -18.086,
    -18.087, -18.031, -17.880, -17.565, -17.002, -16.122, -17.994, -17.998,
    -18.059, -18.100, -18.100, -18.043, -17.889, -17.573, -17.009, -16.130,
    -18.011, -18.014, -18.075, -18.114, -18.113, -18.054, -17.899, -17.581,
    -17.016, -16.137, -18.029, -18.031, -18.090, -18.128, -18.126, -18.066,
    -17.909, -17.589, -17.022, -16.144, -18.046, -18.047, -18.105, -18.143,
    -18.139, -18.077, -17.918, -17.596, -17.029, -16.151, -18.064, -18.063,
    -18.120, -18.157, -18.152, -18.089, -17.928, -17.604, -17.036, -16.158,
    -18.081, -18.079, -18.136, -18.171, -18.165, -18.100, -17.938, -17.612,
    -17.043, -16.165, -18.098, -18.095, -18.151, -18.185, -18.178, -18.112,
    -17.947, -17.620, -17.049, -16.172, -18.115, -18.112, -18.166, -18.199,
    -18.191, -18.123, -17.957, -17.628, -17.056, -16.179, -18.132, -18.127,
    -18.181, -18.214, -18.204, -18.135, -17.967, -17.636, -17.063, -16.186,
    -18.148, -18.143, -18.196, -18.228, -18.217, -18.146, -17.977, -17.644,
    -17.070, -16.193, -18.165, -18.159, -18.211, -18.242, -18.230, -18.158,
    -17.986, -17.652, -17.077, -16.199, -18.182, -18.175, -18.226, -18.256,
    -18.243, -18.169, -17.996, -17.659, -17.083, -16.206, -18.198, -18.191,
    -18.241, -18.270, -18.256, -18.181, -18.006, -17.667, -17.090, -16.213,
    -18.214, -18.207, -18.256, -18.284, -18.269, -18.192, -18.016, -17.675,
    -17.097, -16.220, -18.231, -18.222, -18.271, -18.298, -18.282, -18.204,
    -18.025, -17.683, -17.104, -16.227, -18.247, -18.238, -18.286, -18.312,
    -18.295, -18.215, -18.035, -17.691, -17.110, -16.234, -18.263, -18.253,
    -18.301, -18.326, -18.308, -18.227, -18.045, -17.699, -17.117, -16.241,
    -18.279, -18.269, -18.316, -18.340, -18.320, -18.238, -18.055, -17.707,
    -17.124, -16.248, -18.295, -18.284, -18.330, -18.354, -18.333, -18.250,
    -18.064, -17.715, -17.131, -16.255, -18.310, -18.300, -18.345, -18.368,
    -18.346, -18.261, -18.074, -17.722, -17.138, -16.262, -18.326, -18.315,
    -18.360, -18.381, -18.359, -18.272, -18.084, -17.730, -17.144, -16.269,
    -18.341, -18.330, -18.374, -18.395, -18.372, -18.284, -18.093, -17.738,
    -17.151, -16.275, -18.357, -18.345, -18.389, -18.409, -18.384, -18.295,
    -18.103, -17.746, -17.158, -16.282, -18.372, -18.360, -18.404, -18.423,
    -18.397, -18.307, -18.113, -17.754, -17.165, -16.289, -18.387, -18.375,
    -18.418, -18.437, -18.410, -18.318, -18.123, -17.762, -17.171, -16.296,
    -18.403, -18.390, -18.433, -18.450, -18.422, -18.329, -18.132, -17.770,
    -17.178, -16.303, -18.418, -18.405, -18.447, -18.464, -18.435, -18.341,
    -18.142, -17.778, -17.185, -16.310, -18.433, -18.420, -18.461, -18.478,
    -18.448, -18.352, -18.152, -17.786, -17.192, -16.316, -18.447, -18.435,
    -18.476, -18.491, -18.460, -18.364, -18.161, -17.793, -17.198, -16.323,
    -18.462, -18.450, -18.490, -18.505, -18.473, -18.375, -18.171, -17.801,
    -17.205, -16.330, -18.477, -18.465, -18.504, -18.518, -18.486, -18.386,
    -18.181, -17.809, -17.212, -16.337, -18.491, -18.479, -18.519, -18.532,
    -18.498, -18.398, -18.190, -17.817, -17.218, -16.344, -18.506, -18.494,
    -18.533, -18.545, -18.511, -18.409, -18.200, -17.825, -17.225, -16.350,
    -18.520, -18.508, -18.547, -18.559, -18.523, -18.420, -18.210, -17.833,
    -17.232, -16.357, -18.535, -18.523, -18.561, -18.572, -18.536, -18.431,
    -18.219, -17.841, -17.239, -16.364, -18.549, -18.537, -18.575, -18.586,
    -18.548, -18.443, -18.229, -17.848, -17.245, -16.371, -18.563, -18.552,
    -18.589, -18.599, -18.561, -18.454, -18.239, -17.856, -17.252, -16.377,
    -18.577, -18.566, -18.603, -18.613, -18.573, -18.465, -18.248, -17.864,
    -17.259, -16.384, -18.591, -18.580, -18.617, -18.626, -18.586, -18.476,
    -18.258, -17.872, -17.265, -16.391, -18.605, -18.595, -18.631, -18.639,
    -18.598, -18.488, -18.268, -17.880, -17.272, -16.397, -18.618, -18.609,
    -18.645, -18.653, -18.611, -18.499, -18.277, -17.888, -17.279, -16.404,
    -18.632, -18.623, -18.659, -18.666, -18.623, -18.510, -18.287, -17.895,
    -17.285, -16.411, -18.646, -18.637, -18.673, -18.679, -18.635, -18.521,
    -18.296, -17.903, -17.292, -16.418, -18.659, -18.651, -18.686, -18.692,
    -18.648, -18.532, -18.306, -17.911, -17.299, -16.424, -18.673, -18.665,
    -18.700, -18.705, -18.660, -18.544, -18.315, -17.919, -17.305, -16.431,
    -18.686, -18.679, -18.714, -18.719, -18.672, -18.555, -18.325, -17.927,
    -17.312, -16.438, -18.699, -18.693, -18.727, -18.732, -18.685, -18.566,
    -18.335, -17.934, -17.319, -16.444, -18.713, -18.706, -18.741, -18.745,
    -18.697, -18.577, -18.344, -17.942, -17.325, -16.451, -18.726, -18.720,
    -18.755, -18.758, -18.709, -18.588, -18.354, -17.950, -17.332, -16.457,
    -18.739, -18.734, -18.768, -18.771, -18.722, -18.599, -18.363, -17.958,
    -17.338, -16.464, -18.752, -18.748, -18.782, -18.784, -18.734, -18.610,
    -18.373, -17.965, -17.345, -16.471, -18.765, -18.761, -18.795, -18.797,
    -18.746, -18.621, -18.382, -17.973, -17.352, -16.477, -18.778, -18.775,
    -18.809, -18.810, -18.758, -18.632, -18.392, -17.981, -17.358, -16.484,
    -18.790, -18.788, -18.822, -18.823, -18.770, -18.643, -18.401, -17.989,
    -17.365, -16.490, -18.803, -18.802, -18.835, -18.836, -18.782, -18.654,
    -18.411, -17.996, -17.371, -16.497, -18.816, -18.815, -18.849, -18.848,
    -18.794, -18.665, -18.420, -18.004, -17.378, -16.504, -18.828, -18.828,
    -18.862, -18.861, -18.807, -18.676, -18.429, -18.012, -17.385, -16.510,
    -18.841, -18.842, -18.875, -18.874, -18.819, -18.687, -18.439, -18.019,
    -17.391, -16.517, -18.853, -18.855, -18.888, -18.887, -18.831, -18.698,
    -18.448, -18.027, -17.398, -16.523, -18.865, -18.868, -18.901, -18.899,
    -18.843, -18.709, -18.458, -18.035, -17.404, -16.530, -18.878, -18.881,
    -18.914, -18.912, -18.855, -18.720, -18.467, -18.042, -17.411, -16.536,
    -18.890, -18.894, -18.927, -18.925, -18.867, -18.731, -18.476, -18.050,
    -17.417, -16.543, -18.902, -18.907, -18.941, -18.937, -18.879, -18.742,
    -18.486, -18.058, -17.424, -16.549, -18.914, -18.920, -18.954, -18.950,
    -18.890, -18.753, -18.495, -18.065, -17.430, -16.556, -18.926, -18.933,
    -18.966, -18.963, -18.902, -18.764, -18.504, -18.073, -17.437, -16.562,
    -18.938, -18.946, -18.979, -18.975, -18.914, -18.774, -18.514, -18.081,
    -17.443, -16.569, -18.950, -18.959, -18.992, -18.988, -18.926, -18.785,
    -18.523, -18.088, -17.450, -16.575, -18.962, -18.972, -19.005, -19.000,
    -18.938, -18.796, -18.532, -18.096, -17.456, -16.581, -18.974, -18.985,
    -19.018, -19.013, -18.950, -18.807, -18.542, -18.103, -17.463, -16.588,
    -18.986, -18.998, -19.031, -19.025, -18.962, -18.817, -18.551, -18.111,
    -17.469, -16.594, -18.997, -19.010, -19.043, -19.038, -18.973, -18.828,
    -18.560, -18.118, -17.475, -16.601, -19.009, -19.023, -19.056, -19.050,
    -18.985, -18.839, -18.569, -18.126, -17.482, -16.607, -19.021, -19.036,
    -19.069, -19.062, -18.997, -18.850, -18.578, -18.133, -17.488, -16.613,
    -19.032, -19.048, -19.081, -19.075, -19.008, -18.860, -18.588, -18.141,
    -17.495, -16.620, -19.044, -19.061, -19.094, -19.087, -19.020, -18.871,
    -18.597, -18.148, -17.501, -16.626, -19.055, -19.073, -19.107, -19.099,
    -19.032, -18.882, -18.606, -18.156, -17.507, -16.633, -19.067, -19.086,
    -19.119, -19.111, -19.043, -18.892, -18.615, -18.163, -17.514, -16.639,
    -19.078, -19.098, -19.132, -19.124, -19.055, -18.903, -18.624, -18.171,
    -17.520, -16.645, -19.089, -19.110, -19.144, -19.136, -19.067, -18.913,
    -18.633, -18.178, -17.526, -16.652, -19.100, -19.123, -19.156, -19.148,
    -19.078, -18.924, -18.642, -18.186, -17.533, -16.658, -19.112, -19.135,
    -19.169, -19.160, -19.090, -18.935, -18.651, -18.193, -17.539, -16.664,
    -19.123, -19.147, -19.181, -19.172, -19.101, -18.945, -18.660, -18.200,
    -17.545, -16.671, -19.134, -19.159, -19.194, -19.184, -19.113, -18.956,
    -18.669, -18.208, -17.552, -16.677, -19.145, -19.172, -19.206, -19.196,
    -19.124, -18.966, -18.678, -18.215, -17.558, -16.683, -19.156, -19.184,
    -19.218, -19.208, -19.136, -18.976, -18.687, -18.222, -17.564, -16.689,
    -19.167, -19.196, -19.230, -19.220, -19.147, -18.987, -18.696, -18.230,
    -17.571, -16.696, -19.178, -19.208, -19.243, -19.232, -19.158, -18.997,
    -18.705, -18.237, -17.577, -16.702, -19.189, -19.220, -19.255, -19.244,
    -19.170, -19.008, -18.714, -18.244, -17.583, -16.708, -19.200, -19.232,
    -19.267, -19.256, -19.181, -19.018, -18.723, -18.251, -17.589, -16.714,
    -19.211, -19.244, -19.279, -19.268, -19.193, -19.028, -18.732, -18.259,
    -17.595, -16.721, -19.222, -19.256, -19.291, -19.280, -19.204, -19.039,
    -18.741, -18.266, -17.602, -16.727, -19.232, -19.268, -19.303, -19.292,
    -19.215, -19.049, -18.750, -18.273, -17.608, -16.733, -19.243, -19.279,
    -19.315, -19.303, -19.226, -19.059, -18.759, -18.280, -17.614, -16.739,
    -19.254, -19.291, -19.327, -19.315, -19.238, -19.070, -18.767, -18.287,
    -17.620, -16.745, -19.265, -19.303, -19.339, -19.327, -19.249, -19.080,
    -18.776, -18.295, -17.626, -16.751, -19.275, -19.315, -19.351, -19.339,
    -19.260, -19.090, -18.785, -18.302, -17.632, -16.758, -19.286, -19.327,
    -19.363, -19.350, -19.271, -19.100, -18.794, -18.309, -17.638, -16.764,
    -19.296, -19.338, -19.374, -19.362, -19.282, -19.111, -18.802, -18.316,
    -17.645, -16.770, -19.307, -19.350, -19.386, -19.374, -19.293, -19.121,
    -18.811, -18.323, -17.651, -16.776, -19.318, -19.361, -19.398, -19.385,
    -19.305, -19.131, -18.820, -18.330, -17.657, -16.782, -19.328, -19.373,
    -19.410, -19.397, -19.316, -19.141, -18.828, -18.337, -17.663, -16.788,
    -19.338, -19.385, -19.422, -19.408, -19.327, -19.151, -18.837, -18.344,
    -17.669, -16.794, -19.349, -19.396, -19.433, -19.420, -19.338, -19.161,
    -18.845, -18.351, -17.675, -16.800, -19.359, -19.408, -19.445, -19.431,
    -19.349, -19.171, -18.854, -18.358, -17.681, -16.806, -19.370, -19.419,
    -19.457, -19.443, -19.360, -19.181, -18.862, -18.365, -17.687, -16.812,
    -19.380, -19.430, -19.468, -19.454, -19.371, -19.191, -18.871, -18.372,
    -17.693, -16.818, -19.390, -19.442, -19.480, -19.466, -19.382, -19.201,
    -18.879, -18.379, -17.699, -16.824, -19.401, -19.453, -19.491, -19.477,
    -19.393, -19.211, -18.888, -18.386, -17.705, -16.830, -19.411, -19.465,
    -19.503, -19.488, -19.403, -19.221, -18.896, -18.392, -17.711, -16.836,
    -19.421, -19.476, -19.514, -19.500, -19.414, -19.231, -18.905, -18.399,
    -17.716, -16.842, -19.432, -19.487, -19.526, -19.511, -19.425, -19.241,
    -18.913, -18.406, -17.722, -16.848, -19.442, -19.498, -19.537, -19.522,
    -19.436, -19.251, -18.921, -18.413, -17.728, -16.854, -19.452, -19.510,
    -19.549, -19.534, -19.447, -19.260, -18.930, -18.419, -17.734, -16.860,
    -19.462, -19.521, -19.560, -19.545, -19.457, -19.270, -18.938, -18.426,
    -17.740, -16.866, -19.472, -19.532, -19.571, -19.556, -19.468, -19.280,
    -18.946, -18.433, -17.746, -16.872, -19.482, -19.543, -19.583, -19.567,
    -19.479, -19.290, -18.954, -18.439, -17.751, -16.878, -19.493, -19.554,
    -19.594, -19.578, -19.490, -19.299, -18.962, -18.446, -17.757, -16.884,
    -19.503, -19.565, -19.605, -19.590, -19.500, -19.309, -18.971, -18.453,
    -17.763, -16.890, -19.513, -19.576, -19.616, -19.601, -19.511, -19.319,
    -18.979, -18.459, -17.769, -16.895, -19.523, -19.588, -19.628, -19.612,
    -19.522, -19.328, -18.987, -18.466, -17.774, -16.901, -19.533, -19.599,
    -19.639, -19.623, -19.532, -19.338, -18.995, -18.472, -17.780, -16.907,
    -19.543, -19.610, -19.650, -19.634, -19.543, -19.347, -19.003, -18.479,
    -17.786, -16.913, -19.553, -19.621, -19.661, -19.645, -19.553, -19.357,
    -19.011, -18.485, -17.791, -16.919, -19.563, -19.631, -19.672, -19.656,
    -19.564, -19.366, -19.019, -18.492, -17.797, -16.925, -19.573, -19.642,
    -19.683, -19.667, -19.574, -19.376, -19.026, -18.498, -17.803, -16.930,
    -19.583, -19.653, -19.694, -19.678, -19.585, -19.385, -19.034, -18.504,
    -17.808, -16.936, -19.593, -19.664, -19.706, -19.689, -19.595, -19.394,
    -19.042, -18.511, -17.814, -16.942, -19.603, -19.675, -19.717, -19.700,
    -19.606, -19.404, -19.050, -18.517, -17.819, -16.948, -19.613, -19.686,
    -19.728, -19.710, -19.616, -19.413, -19.058, -18.523, -17.825, -16.953,
    -19.623, -19.697, -19.739, -19.721, -19.626, -19.422, -19.065, -18.529,
    -17.830, -16.959, -19.633, -19.708, -19.749, -19.732, -19.637, -19.432,
    -19.073, -18.535, -17.836, -16.965, -19.642, -19.718, -19.760, -19.743,
    -19.647, -19.441, -19.080, -18.542, -17.841, -16.970, -19.652, -19.729,
    -19.771, -19.754, -19.657, -19.450, -19.088, -18.548, -17.847, -16.976,
    -19.662, -19.740, -19.782, -19.764, -19.667, -19.459, -19.095, -18.554,
    -17.852, -16.982, -19.672, -19.750, -19.793, -19.775, -19.678, -19.468,
    -19.103, -18.560, -17.858, -16.987, -19.682, -19.761, -19.804, -19.786,
    -19.688, -19.477, -19.110, -18.566, -17.863, -16.993, -19.692, -19.772,
    -19.815, -19.796, -19.698, -19.486, -19.118, -18.572, -17.868, -16.999,
    -19.702, -19.783, -19.825, -19.807, -19.708, -19.495, -19.125, -18.578,
    -17.874, -17.004, -19.711, -19.793, -19.836, -19.818, -19.718, -19.504,
    -19.132, -18.584, -17.879, -17.010, -19.721, -19.804, -19.847, -19.828,
    -19.728, -19.513, -19.139, -18.589, -17.884, -17.015, -19.731, -19.814,
    -19.858, -19.839, -19.738, -19.522, -19.147, -18.595, -17.889, -17.021,
    -19.741, -19.825, -19.868, -19.850, -19.748, -19.531, -19.154, -18.601,
    -17.895, -17.027, -19.751, -19.835, -19.879, -19.860, -19.758, -19.539,
    -19.161, -18.607, -17.900, -17.032, -19.760, -19.846, -19.890, -19.871,
    -19.768, -19.548, -19.168, -18.612, -17.905, -17.038, -19.770, -19.857,
    -19.900, -19.881, -19.778, -19.557, -19.175, -18.618, -17.910, -17.043,
    -19.780, -19.867, -19.911, -19.891, -19.788, -19.565, -19.182, -18.623,
    -17.915, -17.049, -19.790, -19.878, -19.922, -19.902, -19.798, -19.574,
    -19.189, -18.629, -17.920, -17.054, -19.799, -19.888, -19.932, -19.912,
    -19.808, -19.582, -19.195, -18.634, -17.925, -17.059, -19.809, -19.898,
    -19.943, -19.923, -19.817, -19.591, -19.202, -18.640, -17.930, -17.065,
    -19.819, -19.909, -19.953, -19.933, -19.827, -19.599, -19.209, -18.645,
    -17.935, -17.070, -19.829, -19.919, -19.964, -19.943, -19.837, -19.607,
    -19.215, -18.651, -17.940, -17.076, -19.838, -19.930, -19.974, -19.954,
    -19.847, -19.616, -19.222, -18.656, -17.945, -17.081, -19.848, -19.940,
    -19.985, -19.964, -19.856, -19.624, -19.228, -18.661, -17.950, -17.086,
    -19.858, -19.951, -19.995, -19.974, -19.866, -19.632, -19.235, -18.666,
    -17.955, -17.092, -19.868, -19.961, -20.006, -19.985, -19.875, -19.640,
    -19.241, -18.672, -17.960, -17.097, -19.877, -19.971, -20.016, -19.995,
    -19.885, -19.648, -19.247, -18.677, -17.965, -17.102, -19.887, -19.982,
    -20.026, -20.005, -19.894, -19.656, -19.254, -18.682, -17.970, -17.108,
    -19.897, -19.992, -20.037, -20.015, -19.904, -19.664, -19.260, -18.687,
    -17.974, -17.113, -19.906, -20.002, -20.047, -20.025, -19.913, -19.672,
    -19.266, -18.692, -17.979, -17.118, -19.916, -20.012, -20.058, -20.035,
    -19.923, -19.680, -19.272, -18.697, -17.984, -17.124, -19.926, -20.023,
    -20.068, -20.046, -19.932, -19.688, -19.278, -18.701, -17.989, -17.129,
    -19.935, -20.033, -20.078, -20.056, -19.941, -19.695, -19.284, -18.706,
    -17.993, -17.134, -19.945, -20.043, -20.088, -20.066, -19.951, -19.703,
    -19.289, -18.711, -17.998, -17.139, -19.955, -20.054, -20.099, -20.076,
    -19.960, -19.711, -19.295, -18.716, -18.002, -17.144, -19.965, -20.064,
    -20.109, -20.086, -19.969, -19.718, -19.301, -18.720, -18.007, -17.150,
    -19.974, -20.074, -20.119, -20.096, -19.978, -19.726, -19.306, -18.725,
    -18.012, -17.155, -19.984, -20.084, -20.129, -20.106, -19.987, -19.733,
    -19.312, -18.729, -18.016, -17.160, -19.994, -20.094, -20.140, -20.116,
    -19.996, -19.740, -19.317, -18.734, -18.021, -17.165, -20.003, -20.105,
    -20.150, -20.125, -20.005, -19.747, -19.322, -18.738, -18.025, -17.170,
    -20.013, -20.115, -20.160, -20.135, -20.014, -19.754, -19.328, -18.743,
    -18.029, -17.175, -20.023, -20.125, -20.170, -20.145, -20.023, -19.761,
    -19.333, -18.747, -18.034, -17.180, -20.032, -20.135, -20.180, -20.155,
    -20.032, -19.768, -19.338, -18.751, -18.038, -17.185, -20.042, -20.145,
    -20.190, -20.165, -20.040, -19.775, -19.343, -18.755, -18.042, -17.190,
    -20.052, -20.155, -20.200, -20.174, -20.049, -19.782, -19.348, -18.759,
    -18.047, -17.195, -20.061, -20.165, -20.210, -20.184, -20.058, -19.789,
    -19.353, -18.763, -18.051, -17.201, -20.071, -20.175, -20.221, -20.194,
    -20.066, -19.795, -19.357, -18.767, -18.055, -17.205, -20.081, -20.185,
    -20.231, -20.204, -20.075, -19.802, -19.362, -18.771, -18.059, -17.210,
    -20.090, -20.196, -20.241, -20.213, -20.084, -19.808, -19.367, -18.775,
    -18.063, -17.215, -20.100, -20.206, -20.251, -20.223, -20.092, -19.815,
    -19.371, -18.779, -18.068, -17.220, -20.110, -20.216, -20.261, -20.232,
    -20.100, -19.821, -19.375, -18.783, -18.072, -17.225, -20.119, -20.226,
    -20.271, -20.242, -20.109, -19.827, -19.380, -18.786, -18.076, -17.230,
    -20.129, -20.236, -20.281, -20.252, -20.117, -19.833, -19.384, -18.790,
    -18.080, -17.235, -20.139, -20.246, -20.290, -20.261, -20.125, -19.839,
    -19.388, -18.793, -18.084, -17.240, -20.148, -20.256, -20.300, -20.270,
    -20.133, -19.845, -19.392, -18.797, -18.088, -17.245, -20.158, -20.266,
    -20.310, -20.280, -20.141, -19.851, -19.396, -18.800, -18.091, -17.250,
    -20.168, -20.276, -20.320, -20.289, -20.149, -19.856, -19.399, -18.804,
    -18.095, -17.254, -20.177, -20.286, -20.330, -20.299, -20.157, -19.862,
    -19.403, -18.807, -18.099, -17.259, -20.187, -20.296, -20.340, -20.308,
    -20.165, -19.867, -19.407, -18.810, -18.103, -17.264, -20.197, -20.306,
    -20.350, -20.317, -20.173, -19.873, -19.410, -18.813, -18.107, -17.269,
    -20.206, -20.316, -20.360, -20.327, -20.181, -19.878, -19.413, -18.816,
    -18.110, -17.274, -20.216, -20.326, -20.369, -20.336, -20.188, -19.883,
    -19.417, -18.819, -18.114, -17.278, -20.226, -20.336, -20.379, -20.345,
    -20.196, -19.888, -19.420, -18.822, -18.118, -17.283, -20.235, -20.345,
    -20.389, -20.354, -20.203, -19.893, -19.423, -18.825, -18.122, -17.288,
    -20.245, -20.355, -20.399, -20.363, -20.211, -19.898, -19.426, -18.828,
    -18.125, -17.292, -20.255, -20.365, -20.408, -20.372, -20.218, -19.902,
    -19.429, -18.831, -18.129, -17.297, -20.264, -20.375, -20.418, -20.381,
    -20.225, -19.907, -19.432, -18.833, -18.132, -17.302, -20.274, -20.385,
    -20.428, -20.390, -20.232, -19.911, -19.434, -18.836, -18.136, -17.306,
    -20.284, -20.395, -20.437, -20.399, -20.239, -19.915, -19.437, -18.838,
    -18.139, -17.311, -20.293, -20.405, -20.447, -20.408, -20.246, -19.920,
    -19.439, -18.841, -18.143, -17.316, -20.303, -20.415, -20.457, -20.417,
    -20.253, -19.924, -19.441, -18.843, -18.146, -17.320, -20.313, -20.425,
    -20.466, -20.426, -20.260, -19.927, -19.444, -18.846, -18.149, -17.325,
    -20.323, -20.434, -20.476, -20.435, -20.267, -19.931, -19.446, -18.848,
    -18.153, -17.329, -20.332, -20.444, -20.486, -20.443, -20.273, -19.935,
    -19.448, -18.850, -18.156, -17.334, -20.342, -20.454, -20.495, -20.452,
    -20.280, -19.938, -19.450, -18.852, -18.159, -17.339, -20.352, -20.464,
    -20.505, -20.461, -20.286, -19.942, -19.452, -18.855, -18.163, -17.343,
    -20.361, -20.474, -20.514, -20.469, -20.292, -19.945, -19.453, -18.857,
    -18.166, -17.348, -20.371, -20.484, -20.524, -20.478, -20.299, -19.948,
    -19.455, -18.859, -18.169, -17.352, -20.381, -20.493, -20.533, -20.487,
    -20.305, -19.951, -19.456, -18.861, -18.172, -17.356, -20.390, -20.503,
    -20.543, -20.495, -20.311, -19.954, -19.458, -18.862, -18.175, -17.361,
    -20.400, -20.513, -20.552, -20.503, -20.316, -19.957, -19.459, -18.864,
    -18.178, -17.365, -20.410, -20.523, -20.562, -20.512, -20.322, -19.960,
    -19.460, -18.866, -18.181, -17.370, -20.420, -20.533, -20.571, -20.520,
    -20.328, -19.962, -19.461, -18.868, -18.185, -17.374, -20.429, -20.542,
    -20.581, -20.528, -20.333, -19.964, -19.462, -18.869, -18.188, -17.379,
    -20.439, -20.552, -20.590, -20.536, -20.339, -19.967, -19.463, -18.871,
    -18.191, -17.383, -20.449, -20.562, -20.599, -20.545, -20.344, -19.969,
    -19.464, -18.872, -18.194, -17.387, -20.459, -20.572, -20.609, -20.553,
    -20.349, -19.971, -19.465, -18.874, -18.196, -17.392, -20.468, -20.582,
    -20.618, -20.561, -20.355, -19.973, -19.465, -18.875, -18.199, -17.396,
    -20.478, -20.591, -20.627, -20.569, -20.360, -19.974, -19.466, -18.877,
    -18.202, -17.400, -20.488, -20.601, -20.637, -20.577, -20.364, -19.976,
    -19.466, -18.878, -18.205, -17.405, -20.498, -20.611, -20.646, -20.584,
    -20.369, -19.978, -19.467, -18.879, -18.208, -17.409, -20.507, -20.620,
    -20.655, -20.592, -20.374, -19.979, -19.467, -18.881, -18.211, -17.413,
    -20.517, -20.630, -20.664, -20.600, -20.378, -19.980, -19.467, -18.882,
    -18.214, -17.418, -20.527, -20.640, -20.674, -20.608, -20.383, -19.981,
    -19.468, -18.883, -18.216, -17.422, -20.537, -20.650, -20.683, -20.615,
    -20.387, -19.982, -19.468, -18.884, -18.219, -17.426, -20.546, -20.659,
    -20.692, -20.623, -20.391, -19.983, -19.468, -18.885, -18.222, -17.430,
    -20.556, -20.669, -20.701, -20.630, -20.395, -19.984, -19.468, -18.886,
    -18.225, -17.435, -20.566, -20.679, -20.710, -20.638, -20.399, -19.985,
    -19.467, -18.887, -18.227, -17.439, -20.576, -20.688, -20.719, -20.645,
    -20.403, -19.985, -19.467, -18.888, -18.230, -17.443, -20.586, -20.698,
    -20.728, -20.652, -20.407, -19.986, -19.467, -18.889, -18.233, -17.447,
    -20.595, -20.708, -20.737, -20.659, -20.410, -19.986, -19.467, -18.890,
    -18.236, -17.452, -20.605, -20.717, -20.746, -20.666, -20.414, -19.987,
    -19.466, -18.891, -18.238, -17.456, -20.615, -20.727, -20.755, -20.673,
    -20.417, -19.987, -19.466, -18.892, -18.241, -17.460, -20.625, -20.737,
    -20.764, -20.680, -20.420, -19.987, -19.466, -18.893, -18.244, -17.464,
    -20.635, -20.746, -20.773, -20.687, -20.424, -19.987, -19.465, -18.894,
    -18.246, -17.468, -20.644, -20.756, -20.782, -20.694, -20.427, -19.987,
    -19.465, -18.895, -18.249, -17.473, -20.654, -20.766, -20.791, -20.701,
    -20.430, -19.987, -19.464, -18.896, -18.252, -17.477, -20.664, -20.775,
    -20.800, -20.708, -20.433, -19.987, -19.464, -18.897, -18.254, -17.481,
    -20.674, -20.785, -20.809, -20.714, -20.435, -19.987, -19.463, -18.897,
    -18.257, -17.485, -20.684, -20.795, -20.818, -20.721, -20.438, -19.986,
    -19.462, -18.898, -18.259, -17.489, -20.694, -20.804, -20.827, -20.728,
    -20.441, -19.986, -19.462, -18.899, -18.262, -17.493, -20.703, -20.814,
    -20.835, -20.734, -20.443, -19.986, -19.461, -18.900, -18.265, -17.498,
    -20.713, -20.823, -20.844, -20.740, -20.446, -19.985, -19.461, -18.901,
    -18.267, -17.502, -20.723, -20.833, -20.853, -20.747, -20.448, -19.985,
    -19.460, -18.902, -18.270, -17.506, -20.733, -20.843, -20.862, -20.753,
    -20.451, -19.985, -19.460, -18.903, -18.273, -17.510, -20.743, -20.852,
    -20.870, -20.759, -20.453, -19.984, -19.459, -18.903, -18.275, -17.514,
    -20.753, -20.862, -20.879, -20.765, -20.455, -19.984, -19.458, -18.904,
    -18.278, -17.518, -20.762, -20.871, -20.888, -20.772, -20.457, -19.984,
    -19.458, -18.905, -18.281, -17.522, -20.772, -20.881, -20.896, -20.778,
    -20.460, -19.983, -19.458, -18.906, -18.283, -17.526, -20.782, -20.891,
    -20.905, -20.784, -20.462, -19.983, -19.457, -18.907, -18.286, -17.531,
    -20.792, -20.900, -20.913, -20.790, -20.464, -19.982, -19.457, -18.908,
    -18.289, -17.535, -20.802, -20.910, -20.922, -20.796, -20.466, -19.982,
    -19.456, -18.909, -18.291, -17.539, -20.812, -20.919, -20.930, -20.802,
    -20.468, -19.982, -19.456, -18.910, -18.294, -17.543, -20.822, -20.929,
    -20.939, -20.808, -20.470, -19.982, -19.456, -18.912, -18.297, -17.547,
    -20.832, -20.938, -20.948, -20.814, -20.473, -19.982, -19.456, -18.913,
    -18.299, -17.551, -20.842, -20.948, -20.956, -20.820, -20.475, -19.982,
    -19.456, -18.914, -18.302, -17.555, -20.851, -20.957, -20.965, -20.826,
    -20.477, -19.982, -19.456, -18.915, -18.305, -17.560, -20.861, -20.967,
    -20.973, -20.831, -20.479, -19.982, -19.456, -18.917, -18.308, -17.564,
    -20.871, -20.977, -20.981, -20.837, -20.482, -19.982, -19.456, -18.918,
    -18.311, -17.568, -20.881, -20.986, -20.990, -20.843, -20.484, -19.982,
    -19.456, -18.919, -18.314, -17.572, -20.891, -20.995, -20.993, -20.829,
    -20.450, -19.942, -19.424, -18.900, -18.306, -17.572, -20.901, -21.004,
    -20.998, -20.826, -20.436, -19.925, -19.410, -18.893, -18.305, -17.575,
    -20.911, -21.013, -21.004, -20.822, -20.423, -19.908, -19.397, -18.885,
    -18.303, -17.577, -20.921, -21.022, -21.012, -20.827, -20.424, -19.908,
    -19.397, -18.887, -18.306, -17.581, -20.931, -21.032, -21.020, -20.832,
    -20.426, -19.908, -19.398, -18.889, -18.309, -17.585, -20.941, -21.041,
    -21.028, -20.838, -20.428, -19.909, -19.398, -18.891, -18.312, -17.590,
    -20.951, -21.051, -21.036, -20.843, -20.431, -19.910, -19.399, -18.892,
    -18.315, -17.594, -20.960, -21.060, -21.046, -20.854, -20.443, -19.922,
    -19.409, -18.900, -18.321, -17.599, -20.970, -21.070, -21.055, -20.861,
    -20.447, -19.924, -19.411, -18.902, -18.324, -17.603, -20.980, -21.079,
    -21.061, -20.862, -20.441, -19.916, -19.405, -18.899, -18.325, -17.606,
    -20.990, -21.088, -21.068, -20.862, -20.435, -19.908, -19.398, -18.896,
    -18.325, -17.610, -21.000, -21.097, -21.072, -20.853, -20.412, -19.882,
    -19.377, -18.884, -18.321, -17.611, -21.010, -21.106, -21.076, -20.847,
    -20.397, -19.864, -19.363, -18.876, -18.319, -17.613, -21.020, -21.116,
    -21.085, -20.854, -20.401, -19.867, -19.365, -18.878, -18.322, -17.617,
    -21.030, -21.125, -21.091, -20.853, -20.395, -19.858, -19.359, -18.876,
    -18.323, -17.621, -21.040, -21.134, -21.100, -20.862, -20.404, -19.867,
    -19.366, -18.882, -18.328, -17.625, -21.050, -21.144, -21.109, -20.872,
    -20.414, -19.877, -19.374, -18.888, -18.333, -17.630, -21.060, -21.153,
    -21.114, -20.868, -20.403, -19.864, -19.365, -18.884, -18.333, -17.633,
    -21.070, -21.163, -21.126, -20.886, -20.426, -19.887, -19.383, -18.897,
    -18.342, -17.639, -21.080, -21.173, -21.139, -20.904, -20.448, -19.908,
    -19.400, -18.908, -18.349, -17.645, -21.090, -21.182, -21.146, -20.908,
    -20.449, -19.907, -19.400, -18.909, -18.352, -17.649, -21.100, -21.192,
    -21.158, -20.926, -20.471, -19.928, -19.416, -18.921, -18.360, -17.655,
    -21.110, -21.202, -21.167, -20.932, -20.474, -19.930, -19.418, -18.923,
    -18.363, -17.659, -21.120, -21.211, -21.170, -20.924, -20.454, -19.908,
    -19.400, -18.912, -18.359, -17.661, -21.130, -21.219, -21.175, -20.918,
    -20.438, -19.889, -19.384, -18.903, -18.356, -17.662, -21.140, -21.228,
    -21.178, -20.908, -20.416, -19.863, -19.363, -18.890, -18.351, -17.663,
    -21.150, -21.236, -21.176, -20.885, -20.375, -19.820, -19.329, -18.868,
    -18.342, -17.662, -21.160, -21.244, -21.175, -20.867, -20.345, -19.790,
    -19.306, -18.855, -18.337, -17.663, -21.170, -21.252, -21.180, -20.865,
    -20.340, -19.785, -19.303, -18.854, -18.339, -17.667, -21.180, -21.262,
    -21.188, -20.872, -20.348, -19.795, -19.312, -18.862, -18.345, -17.672,
    -21.190, -21.272, -21.203, -20.899, -20.384, -19.831, -19.342, -18.883,
    -18.358, -17.680, -21.200, -21.284, -21.226, -20.946, -20.447, -19.891,
    -19.390, -18.915, -18.376, -17.690, -21.211, -21.296, -21.249, -20.992,
    -20.509, -19.950, -19.434, -18.944, -18.393, -17.700, -21.221, -21.308,
    -21.269, -21.028, -20.556, -19.993, -19.467, -18.965, -18.405, -17.707,
    -21.231, -21.318, -21.279, -21.039, -20.566, -20.002, -19.473, -18.970,
    -18.410, -17.712, -21.241, -21.326, -21.279, -21.024, -20.540, -19.975,
    -19.454, -18.959, -18.406, -17.713, -21.251, -21.332, -21.272, -20.988,
    -20.483, -19.920, -19.413, -18.935, -18.396, -17.712, -21.260, -21.338,
    -21.259, -20.941, -20.412, -19.851, -19.360, -18.902, -18.381, -17.709,
    -21.270, -21.345, -21.253, -20.909, -20.362, -19.801, -19.320, -18.877,
    -18.370, -17.708, -21.280, -21.353, -21.255, -20.897, -20.340, -19.776,
    -19.300, -18.864, -18.365, -17.708, -21.290, -21.363, -21.263, -20.902,
    -20.340, -19.774, -19.297, -18.863, -18.366, -17.711, -21.301, -21.373,
    -21.276, -20.917, -20.355, -19.786, -19.306, -18.870, -18.371, -17.716,
    -21.311, -21.385, -21.295, -20.953, -20.400, -19.829, -19.340, -18.893,
    -18.385, -17.724, -21.321, -21.396, -21.313, -20.985, -20.441, -19.869,
    -19.372, -18.916, -18.399, -17.733, -21.331, -21.405, -21.322, -20.994,
    -20.453, -19.881, -19.383, -18.924, -18.406, -17.738, -21.341, -21.416,
    -21.337, -21.020, -20.485, -19.911, -19.408, -18.942, -18.416, -17.745,
    -21.352, -21.426, -21.348, -21.034, -20.500, -19.926, -19.419, -18.950,
    -18.423, -17.751, -21.362, -21.434, -21.348, -21.020, -20.477, -19.903,
    -19.402, -18.940, -18.420, -17.752, -21.372, -21.443, -21.354, -21.018,
    -20.470, -19.895, -19.396, -18.938, -18.420, -17.755, -21.382, -21.451,
    -21.357, -21.013, -20.459, -19.883, -19.388, -18.933, -18.420, -17.758,
    -21.392, -21.458, -21.354, -20.992, -20.427, -19.853, -19.364, -18.919,
    -18.414, -17.758, -21.402, -21.467, -21.360, -20.992, -20.423, -19.847,
    -19.360, -18.918, -18.415, -17.761, -21.412, -21.477, -21.370, -21.001,
    -20.429, -19.852, -19.363, -18.920, -18.418, -17.765, -21.422, -21.485,
    -21.370, -20.987, -20.407, -19.830, -19.346, -18.910, -18.415, -17.767,
    -21.432, -21.495, -21.379, -20.995, -20.413, -19.834, -19.350, -18.913,
    -18.418, -17.771, -21.442, -21.505, -21.388, -21.002, -20.419, -19.838,
    -19.353, -18.916, -18.421, -17.775, -21.452, -21.513, -21.394, -21.005,
    -20.419, -19.838, -19.353, -18.918, -18.424, -17.778, -21.463, -21.525,
    -21.414, -21.039, -20.460, -19.876, -19.383, -18.938, -18.436, -17.786,
    -21.473, -21.535, -21.426, -21.055, -20.478, -19.892, -19.396, -18.948,
    -18.443, -17.791, -21.483, -21.545, -21.436, -21.067, -20.490, -19.902,
    -19.404, -18.954, -18.448, -17.796, -21.493, -21.555, -21.445, -21.074,
    -20.495, -19.907, -19.408, -18.957, -18.452, -17.800, -21.503, -21.563,
    -21.447, -21.064, -20.479, -19.891, -19.396, -18.950, -18.450, -17.802,
    -21.514, -21.572, -21.453, -21.065, -20.476, -19.887, -19.393, -18.950,
    -18.451, -17.805, -21.524, -21.581, -21.460, -21.068, -20.476, -19.886,
    -19.392, -18.950, -18.453, -17.809, -21.534, -21.591, -21.467, -21.071,
    -20.476, -19.885, -19.391, -18.950, -18.455, -17.812, -21.544, -21.600,
    -21.474, -21.075, -20.476, -19.884, -19.390, -18.950, -18.457, -17.815,
    -21.554, -21.610, -21.486, -21.089, -20.491, -19.896, -19.400, -18.958,
    -18.462, -17.820, -21.565, -21.621, -21.502, -21.114, -20.519, -19.921,
    -19.419, -18.971, -18.471, -17.826, -21.576, -21.635, -21.534, -21.178,
    -20.598, -19.992, -19.473, -19.006, -18.491, -17.837, -21.586, -21.645,
    -21.542, -21.183, -20.600, -19.993, -19.473, -19.007, -18.493, -17.840,
    -21.596, -21.654, -21.550, -21.187, -20.602, -19.994, -19.474, -19.009,
    -18.496, -17.844, -21.606, -21.664, -21.557, -21.192, -20.604, -19.995,
    -19.475, -19.010, -18.498, -17.848, -21.616, -21.673, -21.565, -21.197,
    -20.607, -19.997, -19.477, -19.012, -18.501, -17.851, -21.627, -21.682,
    -21.573, -21.203, -20.610, -19.998, -19.478, -19.014, -18.504, -17.855,
    -21.637, -21.692, -21.581, -21.208, -20.613, -20.000, -19.480, -19.016,
    -18.506, -17.858, -21.647, -21.701, -21.589, -21.214, -20.617, -20.003,
    -19.482, -19.018, -18.509, -17.862, -21.657, -21.711, -21.597, -21.220,
    -20.620, -20.005, -19.484, -19.020, -18.512, -17.866, -21.668, -21.720,
    -21.605, -21.226, -20.624, -20.008, -19.486, -19.023, -18.515, -17.870,
    -21.678, -21.730, -21.614, -21.232, -20.629, -20.011, -19.488, -19.025,
    -18.518, -17.873, -21.688, -21.739, -21.622, -21.238, -20.633, -20.014,
    -19.491, -19.028, -18.521, -17.877, -21.698, -21.749, -21.630, -21.245,
    -20.638, -20.018, -19.494, -19.030, -18.524, -17.881, -21.709, -21.758,
    -21.639, -21.252, -20.643, -20.021, -19.497, -19.033, -18.527, -17.885,
    -21.719, -21.768, -21.647, -21.259, -20.648, -20.025, -19.500, -19.036,
    -18.530, -17.888, -21.729, -21.777, -21.656, -21.266, -20.653, -20.029,
    -19.503, -19.039, -18.533, -17.892, -21.739, -21.787, -21.664, -21.273,
    -20.659, -20.034, -19.507, -19.042, -18.536, -17.896, -21.750, -21.796,
    -21.673, -21.280, -20.665, -20.038, -19.510, -19.045, -18.540, -17.900,
    -21.760, -21.806, -21.681, -21.288, -20.671, -20.043, -19.514, -19.049,
    -18.543, -17.904, -21.770, -21.816, -21.690, -21.295, -20.677, -20.048,
    -19.518, -19.052, -18.547, -17.908, -21.781, -21.825, -21.699, -21.303,
    -20.683, -20.053, -19.522, -19.056, -18.550, -17.911, -21.791, -21.835,
    -21.708, -21.311, -20.690, -20.058, -19.527, -19.059, -18.554, -17.915,
    -21.801, -21.844, -21.717, -21.319, -20.697, -20.064, -19.531, -19.063,
    -18.557, -17.919, -21.812, -21.854, -21.726, -21.327, -20.704, -20.069,
    -19.536, -19.067, -18.561, -17.923, -21.822, -21.864, -21.735, -21.335,
    -20.711, -20.075, -19.540, -19.071, -18.565, -17.927, -21.832, -21.873,
    -21.744, -21.343, -20.718, -20.081, -19.545, -19.075, -18.568, -17.931,
    -21.843, -21.883, -21.753, -21.352, -20.725, -20.087, -19.550, -19.079,
    -18.572, -17.935, -21.853, -21.893, -21.762, -21.360, -20.733, -20.094,
    -19.555, -19.084, -18.576, -17.939, -21.863, -21.902, -21.771, -21.369,
    -20.741, -20.100, -19.561, -19.088, -18.580, -17.943, -21.874, -21.912,
    -21.780, -21.378, -20.749, -20.107, -19.566, -19.093, -18.584, -17.948,
    -21.884, -21.921, -21.789, -21.386, -20.757, -20.113, -19.572, -19.097,
    -18.588, -17.952, -21.894, -21.931, -21.798, -21.395, -20.765, -20.120,
    -19.577, -19.102, -18.592, -17.956, -21.905, -21.941, -21.808, -21.404,
    -20.773, -20.127, -19.583, -19.107, -18.596, -17.960, -21.915, -21.951,
    -21.817, -21.413, -20.781, -20.135, -19.589, -19.111, -18.601, -17.964,
    -21.925, -21.960, -21.826, -21.422, -20.790, -20.142, -19.595, -19.116,
    -18.605, -17.968, -21.936, -21.970, -21.835, -21.432, -20.798, -20.149,
    -19.601, -19.121, -18.609, -17.972, -21.946, -21.980, -21.845, -21.441,
    -20.807, -20.157, -19.607, -19.126, -18.613, -17.977, -21.956, -21.989,
    -21.854, -21.450, -20.816, -20.165, -19.614, -19.132, -18.618, -17.981,
    -21.967, -21.999, -21.863, -21.460, -20.825, -20.172, -19.620, -19.137,
    -18.622, -17.985, -21.977, -22.009, -21.873, -21.469, -20.834, -20.180,
    -19.627, -19.142, -18.627, -17.989, -21.988, -22.018, -21.882, -21.479,
    -20.843, -20.188, -19.633, -19.148, -18.631, -17.994, -21.998, -22.028,
    -21.892, -21.488, -20.852, -20.196, -19.640, -19.153, -18.636, -17.998,
    -22.008, -22.038, -21.901, -21.498, -20.861, -20.205, -19.647, -19.159,
    -18.640, -18.002, -22.019, -22.047, -21.910, -21.507, -20.871, -20.213,
    -19.654, -19.164, -18.645, -18.006, -22.029, -22.057, -21.920, -21.517,
    -20.880, -20.221, -19.661, -19.170, -18.650, -18.011, -22.040, -22.067,
    -21.929, -21.527, -20.890, -20.230, -19.668, -19.176, -18.655, -18.015,
    -22.050, -22.077, -21.939, -21.536, -20.899, -20.239, -19.675, -19.182,
    -18.659, -18.019, -22.060, -22.086, -21.948, -21.546, -20.909, -20.247,
    -19.683, -19.187, -18.664, -18.024, -22.071, -22.096, -21.958, -21.556,
    -20.919, -20.256, -19.690, -19.193, -18.669, -18.028, -22.081, -22.106,
    -21.967, -21.566, -20.929, -20.265, -19.697, -19.199, -18.674, -18.033,
    -22.092, -22.116, -21.977, -21.576, -20.938, -20.274, -19.705, -19.205,
    -18.679, -18.037, -22.102, -22.125, -21.986, -21.586, -20.948, -20.283,
    -19.713, -19.212, -18.684, -18.041, -22.112, -22.135, -21.996, -21.596,
    -20.958, -20.292, -19.720, -19.218, -18.689, -18.046, -22.123, -22.145,
    -22.005, -21.606, -20.968, -20.301, -19.728, -19.224, -18.694, -18.050,
    -22.133, -22.154, -22.015, -21.616, -20.978, -20.310, -19.736, -19.230,
    -18.699, -18.055, -22.144, -22.164, -22.024, -21.626, -20.989, -20.319,
    -19.744, -19.237, -18.704, -18.059, -22.154, -22.174, -22.034, -21.636,
    -20.999, -20.329, -19.752, -19.243, -18.709, -18.064, -22.165, -22.184,
    -22.043, -21.646, -21.009, -20.338, -19.760, -19.250, -18.714, -18.068,
    -22.175, -22.193, -22.053, -21.656, -21.019, -20.348, -19.768, -19.256,
    -18.719, -18.073, -22.186, -22.203, -22.062, -21.666, -21.029, -20.357,
    -19.776, -19.263, -18.725, -18.077, -22.196, -22.213, -22.072, -21.676,
    -21.040, -20.367, -19.784, -19.269, -18.730, -18.082, -22.206, -22.222,
    -22.081, -21.686, -21.050, -20.376, -19.792, -19.276, -18.735, -18.087,
    -22.217, -22.232, -22.091, -21.696, -21.060, -20.386, -19.800, -19.283,
    -18.740, -18.091, -22.227, -22.242, -22.100, -21.706, -21.071, -20.396,
    -19.809, -19.289, -18.746, -18.096, -22.238, -22.252, -22.110, -21.716,
    -21.081, -20.405, -19.817, -19.296, -18.751, -18.100, -22.248, -22.261,
    -22.119, -21.726, -21.092, -20.415, -19.826, -19.303, -18.757, -18.105,
    -22.259, -22.271, -22.129, -21.736, -21.102, -20.425, -19.834, -19.310,
    -18.762, -18.109, -22.269, -22.281, -22.138, -21.746, -21.113, -20.435,
    -19.842, -19.317, -18.767, -18.114, -22.280, -22.291, -22.148, -21.757,
    -21.123, -20.445, -19.851, -19.324, -18.773, -18.119, -22.290, -22.300,
    -22.157, -21.767, -21.134, -20.455, -19.860, -19.331, -18.778, -18.123,
    -22.301, -22.310, -22.167, -21.777, -21.144, -20.464, -19.868, -19.338,
    -18.784, -18.128, -22.311, -22.320, -22.176, -21.787, -21.155, -20.474,
    -19.877, -19.345, -18.789, -18.133, -22.322, -22.330, -22.186, -21.797,
    -21.166, -20.484, -19.886, -19.352, -18.795, -18.137, -22.332, -22.339,
    -22.195, -21.807, -21.176, -20.495, -19.894, -19.359, -18.800, -18.142,
    -22.343, -22.349, -22.205, -21.817, -21.187, -20.505, -19.903, -19.366,
    -18.806, -18.147, -22.353, -22.359, -22.214, -21.827, -21.197, -20.515,
    -19.912, -19.373, -18.812, -18.151, -22.364, -22.368, -22.223, -21.838,
    -21.208, -20.525, -19.921, -19.380, -18.817, -18.156, -22.374, -22.378,
    -22.233, -21.848, -21.219, -20.535, -19.930, -19.388, -18.823, -18.161,
    -22.385, -22.388, -22.242, -21.858, -21.229, -20.545, -19.939, -19.395,
    -18.829, -18.165, -22.395, -22.398, -22.252, -21.868, -21.240, -20.555,
    -19.947, -19.402, -18.834, -18.170, -22.406, -22.407, -22.261, -21.878,
    -21.251, -20.565, -19.956, -19.409, -18.840, -18.175, -22.416, -22.417,
    -22.271, -21.888, -21.261, -20.576, -19.965, -19.417, -18.846, -18.179,
    -22.427, -22.427, -22.280, -21.898, -21.272, -20.586, -19.974, -19.424,
    -18.851, -18.184, -22.437, -22.436, -22.290, -21.908, -21.283, -20.596,
    -19.983, -19.431, -18.857, -18.189, -22.448, -22.446, -22.299, -21.918,
    -21.293, -20.606, -19.992, -19.439, -18.863, -18.193, -22.458, -22.456,
    -22.308, -21.928, -21.304, -20.617, -20.001, -19.446, -18.868, -18.198,
    -22.469, -22.466, -22.318, -21.938, -21.315, -20.627, -20.011, -19.454,
    -18.874, -18.203, -22.479, -22.475, -22.327, -21.948, -21.325, -20.637,
    -20.020, -19.461, -18.880, -18.208, -22.490, -22.485, -22.337, -21.958,
    -21.336, -20.648, -20.029, -19.469, -18.886, -18.212, -22.500, -22.495,
    -22.346, -21.968, -21.347, -20.658, -20.038, -19.476, -18.892, -18.217,
    -22.511, -22.505, -22.355, -21.978, -21.357, -20.668, -20.047, -19.484,
    -18.897, -18.222, -22.521, -22.514, -22.365, -21.988, -21.368, -20.679,
    -20.056, -19.491, -18.903, -18.226, -22.532, -22.524, -22.374, -21.998,
    -21.379, -20.689, -20.065, -19.499, -18.909, -18.231, -22.542, -22.534,
    -22.383, -22.008, -21.389, -20.699, -20.075, -19.506, -18.915, -18.236,
    -22.553, -22.543, -22.393, -22.018, -21.400, -20.710, -20.084, -19.514,
    -18.921, -18.241, -22.563, -22.553, -22.402, -22.028, -21.410, -20.720,
    -20.093, -19.521, -18.926, -18.245, -22.574, -22.563, -22.411, -22.038,
    -21.421, -20.730, -20.102, -19.529, -18.932, -18.250, -22.585, -22.573,
    -22.421, -22.048, -21.432, -20.741, -20.111, -19.536, -18.938, -18.255,
    -22.595, -22.582, -22.430, -22.058, -21.442, -20.751, -20.121, -19.544,
    -18.944, -18.259, -22.606, -22.592, -22.439, -22.067, -21.453, -20.761,
    -20.130, -19.551, -18.950, -18.264, -22.616, -22.602, -22.449, -22.077,
    -21.463, -20.772, -20.139, -19.559, -18.956, -18.269, -22.627, -22.611,
    -22.458, -22.087, -21.474, -20.782, -20.148, -19.567, -18.962, -18.274,
    -22.637, -22.621, -22.467, -22.097, -21.484, -20.793, -20.158, -19.574,
    -18.967, -18.278, -22.648, -22.631, -22.477, -22.107, -21.495, -20.803,
    -20.167, -19.582, -18.973, -18.283, -22.658, -22.641, -22.486, -22.117,
    -21.506, -20.813, -20.176, -19.589, -18.979, -18.288, -22.669, -22.650,
    -22.495, -22.126, -21.516, -20.824, -20.185, -19.597, -18.985, -18.292,
    -22.680, -22.660, -22.505, -22.136, -21.527, -20.834, -20.195, -19.605,
    -18.991, -18.297, -22.690, -22.670, -22.514, -22.146, -21.537, -20.844,
    -20.204, -19.612, -18.997, -18.302, -22.701, -22.679, -22.523, -22.156,
    -21.548, -20.855, -20.213, -19.620, -19.002, -18.306, -22.711, -22.689,
    -22.532, -22.165, -21.558, -20.865, -20.223, -19.628, -19.008, -18.311,
    -22.722, -22.699, -22.542, -22.175, -21.568, -20.875, -20.232, -19.635,
    -19.014, -18.316, -22.732, -22.709, -22.551, -22.185, -21.579, -20.886,
    -20.241, -19.643, -19.020, -18.320, -22.743, -22.718, -22.560, -22.194,
    -21.589, -20.896, -20.250, -19.650, -19.026, -18.325, -22.754, -22.728,
    -22.569, -22.204, -21.600, -20.906, -20.260, -19.658, -19.032, -18.330,
    -22.764, -22.738, -22.579, -22.214, -21.610, -20.917, -20.269, -19.666,
    -19.038, -18.334, -22.775, -22.747, -22.588, -22.223, -21.621, -20.927,
    -20.278, -19.673, -19.043, -18.339, -22.785, -22.757, -22.597, -22.233,
    -21.631, -20.937, -20.288, -19.681, -19.049, -18.343, -22.796, -22.767,
    -22.606, -22.243, -21.641, -20.947, -20.297, -19.688, -19.055, -18.348,
    -22.807, -22.777, -22.615, -22.252, -21.652, -20.958, -20.306, -19.696,
    -19.061, -18.353, -22.817, -22.786, -22.625, -22.262, -21.662, -20.968,
    -20.315, -19.704, -19.067, -18.357, -22.828, -22.796, -22.634, -22.271,
    -21.672, -20.978, -20.325, -19.711, -19.072, -18.362, -22.838, -22.806,
    -22.643, -22.281, -21.682, -20.988, -20.334, -19.719, -19.078, -18.366,
    -22.849, -22.815, -22.652, -22.291, -21.693, -20.999, -20.343, -19.726,
    -19.084, -18.371, -22.860, -22.825, -22.661, -22.300, -21.703, -21.009,
    -20.352, -19.734, -19.090, -18.376, -22.870, -22.835, -22.671, -22.310,
    -21.713, -21.019, -20.361, -19.741, -19.095, -18.380, -22.881, -22.845,
    -22.680, -22.319, -21.723, -21.029, -20.371, -19.749, -19.101, -18.385,
    -22.891, -22.854, -22.689, -22.329, -21.734, -21.039, -20.380, -19.757,
    -19.107, -18.389, -22.902, -22.864, -22.698, -22.338, -21.744, -21.050,
    -20.389, -19.764, -19.112, -18.394, -22.913, -22.874, -22.707, -22.347,
    -21.754, -21.060, -20.398, -19.772, -19.118, -18.398, -22.923, -22.883,
    -22.716, -22.357, -21.764, -21.070, -20.407, -19.779, -19.124, -18.403,
    -22.934, -22.893, -22.725, -22.366, -21.774, -21.080, -20.417, -19.787,
    -19.129, -18.407, -22.945, -22.903, -22.735, -22.376, -21.784, -21.090,
    -20.426, -19.794, -19.135, -18.411, -22.955, -22.913, -22.744, -22.385,
    -21.794, -21.100, -20.435, -19.802, -19.141, -18.416, -22.966, -22.922,
    -22.753, -22.394, -21.804, -21.110, -20.444, -19.809, -19.146, -18.420,
    -22.977, -22.932, -22.762, -22.404, -21.814, -21.120, -20.453, -19.816,
    -19.152, -18.425, -22.987, -22.942, -22.771, -22.413, -21.825, -21.130,
    -20.462, -19.824, -19.157, -18.429, -22.998, -22.951, -22.780, -22.422,
    -21.835, -21.140, -20.471, -19.831, -19.163, -18.433, -23.008, -22.961,
    -22.789, -22.432, -21.845, -21.150, -20.480, -19.839, -19.169, -18.438,
    -23.019, -22.971, -22.798, -22.441, -21.855, -21.160, -20.489, -19.846,
    -19.174, -18.442, -23.030, -22.981, -22.808, -22.450, -21.864, -21.170,
    -20.498, -19.853, -19.180, -18.447, -23.040, -22.990, -22.817, -22.460,
    -21.874, -21.180, -20.507, -19.861, -19.185, -18.451, -23.051, -23.000,
    -22.826, -22.469, -21.884, -21.190, -20.516, -19.868, -19.191, -18.455,
    -23.062, -23.010, -22.835, -22.478, -21.894, -21.200, -20.525, -19.875,
    -19.196, -18.459, -23.072, -23.019, -22.844, -22.487, -21.904, -21.210,
    -20.534, -19.882, -19.201, -18.464, -23.083, -23.029, -22.853, -22.497,
    -21.914, -21.220, -20.543, -19.890, -19.207, -18.468, -23.094, -23.039,
    -22.862, -22.506, -21.924, -21.230, -20.552, -19.897, -19.212, -18.472,
    -23.104, -23.049, -22.871, -22.515, -21.934, -21.240, -20.561, -19.904,
    -19.218, -18.476, -23.115, -23.058, -22.880, -22.524, -21.943, -21.250,
    -20.570, -19.911, -19.223, -18.480, -23.126, -23.068, -22.889, -22.533,
    -21.953, -21.260, -20.579, -19.918, -19.228, -18.485, -23.136, -23.078,
    -22.898, -22.543, -21.963, -21.269, -20.588, -19.926, -19.233, -18.489,
    -23.147, -23.087, -22.907, -22.552, -21.973, -21.279, -20.597, -19.933,
    -19.239, -18.493, -23.158, -23.097, -22.916, -22.561, -21.983, -21.289,
    -20.606, -19.940, -19.244, -18.497, -23.168, -23.107, -22.925, -22.570,
    -21.992, -21.299, -20.614, -19.947, -19.249, -18.501, -23.179, -23.117,
    -22.934, -22.579, -22.002, -21.308, -20.623, -19.954, -19.254, -18.505,
    -23.190, -23.126, -22.943, -22.588, -22.012, -21.318, -20.632, -19.961,
    -19.259, -18.509, -23.200, -23.136, -22.952, -22.597, -22.021, -21.328,
    -20.641, -19.968, -19.264, -18.513, -23.211, -23.146, -22.961, -22.606,
    -22.031, -21.337, -20.649, -19.975, -19.270, -18.517, -23.222, -23.155,
    -22.971, -22.615, -22.041, -21.347, -20.658, -19.982, -19.275, -18.521,
    -23.232, -23.165, -22.980, -22.624, -22.050, -21.357, -20.667, -19.988,
    -19.280, -18.525, -23.243, -23.175, -22.989, -22.634, -22.060, -21.366,
    -20.675, -19.995, -19.285, -18.529, -23.254, -23.185, -22.998, -22.643,
    -22.069, -21.376, -20.684, -20.002, -19.289, -18.533, -23.264, -23.194,
    -23.007, -22.652, -22.079, -21.386, -20.692, -20.009, -19.294, -18.537,
    -23.275, -23.204, -23.016, -22.661, -22.088, -21.395, -20.701, -20.016,
    -19.299, -18.540, -23.286, -23.214, -23.025, -22.670, -22.098, -21.405,
    -20.709, -20.022, -19.304, -18.544, -23.296, -23.223, -23.034, -22.679,
    -22.107, -21.414, -20.718, -20.029, -19.309, -18.548, -23.307, -23.233,
    -23.043, -22.687, -22.117, -21.424, -20.726, -20.036, -19.314, -18.552,
    -23.318, -23.243, -23.052, -22.696, -22.126, -21.433, -20.735, -20.042,
    -19.318, -18.555, -23.328, -23.253, -23.061, -22.705, -22.136, -21.442,
    -20.743, -20.049, -19.323, -18.559, -23.339, -23.262, -23.070, -22.714,
    -22.145, -21.452, -20.752, -20.055, -19.328, -18.563, -23.350, -23.272,
    -23.079, -22.723, -22.155, -21.461, -20.760, -20.062, -19.332, -18.566,
    -23.361, -23.282, -23.087, -22.732, -22.164, -21.471, -20.768, -20.068,
    -19.337, -18.570, -23.371, -23.291, -23.096, -22.741, -22.173, -21.480,
    -20.776, -20.074, -19.342, -18.574, -23.382, -23.301, -23.105, -22.750,
    -22.183, -21.489, -20.785, -20.081, -19.346, -18.577, -23.393, -23.311,
    -23.114, -22.759, -22.192, -21.499, -20.793, -20.087, -19.351, -18.581,
    -23.403, -23.321, -23.123, -22.768, -22.201, -21.508, -20.801, -20.093,
    -19.355, -18.584, -23.414, -23.330, -23.132, -22.777, -22.211, -21.517,
    -20.809, -20.100, -19.359, -18.588, -23.425, -23.340, -23.141, -22.785,
    -22.220, -21.526, -20.817, -20.106, -19.364, -18.591, -23.436, -23.350,
    -23.150, -22.794, -22.229, -21.535, -20.825, -20.112, -19.368, -18.594,
    -23.446, -23.359, -23.159, -22.803, -22.238, -21.545, -20.833, -20.118,
    -19.372, -18.598, -23.457, -23.369, -23.168, -22.812, -22.247, -21.554,
    -20.841, -20.124, -19.376, -18.601, -23.468, -23.379, -23.177, -22.821,
    -22.257, -21.563, -20.849, -20.130, -19.381, -18.604, -23.478, -23.389,
    -23.186, -22.829, -22.266, -21.572, -20.857, -20.136, -19.385, -18.608,
    -23.489, -23.398, -23.195, -22.838, -22.275, -21.581, -20.865, -20.142,
    -19.389, -18.611, -23.500, -23.408, -23.204, -22.847, -22.284, -21.590,
    -20.873, -20.148, -19.393, -18.614, -23.511, -23.418, -23.213, -22.856,
    -22.293, -21.599, -20.881, -20.153, -19.397, -18.617, -23.521, -23.427,
    -23.222, -22.864, -22.302, -21.608, -20.888, -20.159, -19.401, -18.620,
    -23.532, -23.437, -23.231, -22.873, -22.311, -21.617, -20.896, -20.165,
    -19.405, -18.623, -23.543, -23.447, -23.240, -22.882, -22.320, -21.626,
    -20.904, -20.170, -19.408, -18.627, -23.553, -23.457, -23.249, -22.891,
    -22.329, -21.635, -20.911, -20.176, -19.412, -18.630, -23.564, -23.466,
    -23.257, -22.899, -22.338, -21.643, -20.919, -20.181, -19.416, -18.633,
    -23.575, -23.476, -23.266, -22.908, -22.347, -21.652, -20.926, -20.187,
    -19.420, -18.635, -23.586, -23.486, -23.275, -22.917, -22.356, -21.661,
    -20.934, -20.192, -19.423, -18.638, -23.596, -23.496, -23.284, -22.925,
    -22.365, -21.670, -20.941, -20.197, -19.427, -18.641, -23.607, -23.505,
    -23.293, -22.934, -22.374, -21.679, -20.948, -20.203, -19.430, -18.644,
    -23.618, -23.515, -23.302, -22.943, -22.383, -21.687, -20.956, -20.208,
    -19.434, -18.647, -23.629, -23.525, -23.311, -22.951, -22.392, -21.696,
    -20.963, -20.213, -19.437, -18.650, -23.639, -23.534, -23.320, -22.960,
    -22.401, -21.704, -20.970, -20.218, -19.440, -18.652, -23.650, -23.544,
    -23.329, -22.969, -22.410, -21.713, -20.977, -20.223, -19.444, -18.655,
    -23.661, -23.554, -23.338, -22.977, -22.419, -21.722, -20.984, -20.228,
    -19.447, -18.658, -23.672, -23.564, -23.347, -22.986, -22.428, -21.730,
    -20.991, -20.233, -19.450, -18.660, -23.682, -23.573, -23.355, -22.995,
    -22.436, -21.739, -20.998, -20.237, -19.453, -18.663, -23.693, -23.583,
    -23.364, -23.003, -22.445, -21.747, -21.005, -20.242, -19.456, -18.665,
    -23.704, -23.593, -23.373, -23.012, -22.454, -21.755, -21.012, -20.247,
    -19.459, -18.668, -23.715, -23.602, -23.382, -23.020, -22.463, -21.764,
    -21.019, -20.251, -19.462, -18.670, -23.725, -23.612, -23.391, -23.029,
    -22.471, -21.772, -21.025, -20.256, -19.465, -18.673, -23.736, -23.622,
    -23.400, -23.038, -22.480, -21.780, -21.032, -20.260, -19.468, -18.675,
    -23.747, -23.632, -23.409, -23.046, -22.489, -21.788, -21.039, -20.265,
    -19.470, -18.677, -23.758, -23.641, -23.418, -23.055, -22.497, -21.797,
    -21.045, -20.269, -19.473, -18.680, -23.768, -23.651, -23.427, -23.063,
    -22.506, -21.805, -21.052, -20.273, -19.476, -18.682, -23.779, -23.661,
    -23.435, -23.072, -22.515, -21.813, -21.058, -20.277, -19.478, -18.684,
    -23.790, -23.671, -23.444, -23.080, -22.523, -21.821, -21.064, -20.281,
    -19.481, -18.686, -23.801, -23.680, -23.453, -23.089, -22.532, -21.829,
    -21.071, -20.285, -19.483, -18.688, -23.811, -23.690, -23.462, -23.097,
    -22.540, -21.837, -21.077, -20.289, -19.485, -18.690, -23.822, -23.700,
    -23.471, -23.106, -22.549, -21.845, -21.083, -20.293, -19.488, -18.692,
    -23.833, -23.709, -23.480, -23.114, -22.557, -21.853, -21.089, -20.296,
    -19.490, -18.694, -23.844, -23.719, -23.489, -23.123, -22.566, -21.861,
    -21.095, -20.300, -19.492, -18.696, -23.854, -23.729, -23.498, -23.131,
    -22.574, -21.868, -21.101, -20.303, -19.494, -18.698, -23.865, -23.739,
    -23.506, -23.140, -22.583, -21.876, -21.106, -20.307, -19.496, -18.700,
    -23.876, -23.748, -23.515, -23.148, -22.591, -21.884, -21.112, -20.310,
    -19.498, -18.702, -23.887, -23.758, -23.524, -23.157, -22.599, -21.891,
    -21.118, -20.313, -19.500, -18.703, -23.898, -23.768, -23.533, -23.165,
    -22.608, -21.899, -21.123, -20.317, -19.502, -18.705, -23.908, -23.778,
    -23.542, -23.173, -22.616, -21.906, -21.128, -20.320, -19.503, -18.707,
    -23.919, -23.787, -23.551, -23.182, -22.624, -21.914, -21.134, -20.323,
    -19.505, -18.708, -23.930, -23.797, -23.560, -23.190, -22.633, -21.921,
    -21.139, -20.325, -19.506, -18.710, -23.941, -23.807, -23.568, -23.199,
    -22.641, -21.929, -21.144, -20.328, -19.508, -18.711, -23.951, -23.816,
    -23.577, -23.207, -22.649, -21.936, -21.149, -20.331, -19.509, -18.713,
    -23.962, -23.826, -23.586, -23.215, -22.657, -21.943, -21.154, -20.333,
    -19.511, -18.714, -23.973, -23.836, -23.595, -23.224, -22.666, -21.950,
    -21.159, -20.336, -19.512, -18.716, -23.984, -23.846, -23.604, -23.232,
    -22.674, -21.957, -21.164, -20.338, -19.513, -18.717, -23.995, -23.855,
    -23.613, -23.241, -22.682, -21.964, -21.169, -20.341, -19.514, -18.718,
    -24.005, -23.865, -23.621, -23.249, -22.690, -21.971, -21.173, -20.343,
    -19.515, -18.719, -24.016, -23.875, -23.630, -23.257, -22.698, -21.978,
    -21.178, -20.345, -19.516, -18.721, -24.027, -23.885, -23.639, -23.266,
    -22.706, -21.985, -21.182, -20.347, -19.517, -18.722, -24.038, -23.894,
    -23.648, -23.274, -22.714, -21.992, -21.186, -20.349, -19.518, -18.723,
    -24.048, -23.904, -23.657, -23.282, -22.722, -21.998, -21.190, -20.350,
    -19.519, -18.724, -24.059, -23.914, -23.666, -23.290, -22.730, -22.005,
    -21.194, -20.352, -19.519, -18.725, -24.070, -23.924, -23.674, -23.299,
    -22.738, -22.011, -21.198, -20.354, -19.520, -18.726, -24.081, -23.933,
    -23.683, -23.307, -22.746, -22.018, -21.202, -20.355, -19.520, -18.727,
    -24.092, -23.943, -23.692, -23.315, -22.753, -22.024, -21.206, -20.356,
    -19.521, -18.727, -24.102, -23.953, -23.701, -23.324, -22.761, -22.031,
    -21.209, -20.358, -19.521, -18.728, -24.113, -23.962, -23.710, -23.332,
    -22.769, -22.037, -21.213, -20.359, -19.522, -18.729, -24.124, -23.972,
    -23.719, -23.340, -22.777, -22.043, -21.216, -20.360, -19.522, -18.730,
    -24.135, -23.982, -23.727, -23.348, -22.785, -22.049, -21.219, -20.361,
    -19.522, -18.730, -24.146, -23.992, -23.736, -23.357, -22.792, -22.055,
    -21.223, -20.361, -19.522, -18.731, -24.156, -24.001, -23.745, -23.365,
    -22.800, -22.061, -21.226, -20.362, -19.522, -18.732, -24.167, -24.011,
    -23.754, -23.373, -22.807, -22.067, -21.228, -20.363, -19.522, -18.732,
    -24.178, -24.021, -23.763, -23.381, -22.815, -22.072, -21.231, -20.363,
    -19.522, -18.733, -24.189, -24.031, -23.772, -23.389, -22.822, -22.078,
    -21.234, -20.363, -19.521, -18.733, -24.200, -24.040, -23.780, -23.398,
    -22.830, -22.083, -21.236, -20.364, -19.521, -18.733, -24.210, -24.050,
    -23.789, -23.406, -22.837, -22.089, -21.239, -20.364, -19.521, -18.734,
    -24.221, -24.060, -23.798, -23.414, -22.845, -22.094, -21.241, -20.364,
    -19.520, -18.734, -24.232, -24.070, -23.807, -23.422, -22.852, -22.099,
    -21.243, -20.364, -19.520, -18.734, -24.243, -24.079, -23.816, -23.430,
    -22.859, -22.104, -21.245, -20.363, -19.519, -18.735, -24.254, -24.089,
    -23.824, -23.438, -22.867, -22.109, -21.247, -20.363, -19.518, -18.735,
    -24.264, -24.099, -23.833, -23.447, -22.874, -22.114, -21.248, -20.363,
    -19.517, -18.735, -24.275, -24.109, -23.842, -23.455, -22.881, -22.119,
    -21.250, -20.362, -19.516, -18.735, -24.286, -24.118, -23.851, -23.463,
    -22.888, -22.124, -21.251, -20.361, -19.516, -18.735, -24.297, -24.128,
    -23.860, -23.471, -22.895, -22.128, -21.253, -20.360, -19.515, -18.735,
    -24.308, -24.138, -23.869, -23.479, -22.902, -22.133, -21.254, -20.360,
    -19.513, -18.735, -24.319, -24.147, -23.877, -23.487, -22.909, -22.137,
    -21.255, -20.358, -19.512, -18.735, -24.329, -24.157, -23.886, -23.495,
    -22.916, -22.141, -21.256, -20.357, -19.511, -18.735, -24.340, -24.167,
    -23.895, -23.503, -22.923, -22.145, -21.256, -20.356, -19.510, -18.734,
    -24.351, -24.177, -23.904, -23.511, -22.930, -22.149, -21.257, -20.355,
    -19.508, -18.734, -24.362, -24.186, -23.913, -23.519, -22.936, -22.153,
    -21.257, -20.353, -19.507, -18.734, -24.373, -24.196, -23.921, -23.527,
    -22.943, -22.157, -21.257, -20.352, -19.505, -18.734, -24.383, -24.206,
    -23.930, -23.535, -22.950, -22.161, -21.258, -20.350, -19.504, -18.733,
    -24.394, -24.216, -23.939, -23.543, -22.956, -22.164, -21.258, -20.348,
    -19.502, -18.733, -24.405, -24.225, -23.948, -23.551, -22.963, -22.167,
    -21.257, -20.346, -19.500, -18.732, -24.416, -24.235, -23.957, -23.559,
    -22.969, -22.171, -21.257, -20.344, -19.499, -18.732, -24.427, -24.245,
    -23.965, -23.567, -22.976, -22.174, -21.257, -20.342, -19.497, -18.732,
    -24.438, -24.255, -23.974, -23.575, -22.982, -22.177, -21.256, -20.340,
    -19.495, -18.731, -24.448, -24.264, -23.983, -23.583, -22.988, -22.179,
    -21.255, -20.337, -19.493, -18.730, -24.459, -24.274, -23.992, -23.591,
    -22.994, -22.182, -21.254, -20.335, -19.491, -18.730, -24.470, -24.284,
    -24.001, -23.599, -23.000, -22.185, -21.253, -20.332, -19.489, -18.729,
    -24.481, -24.294, -24.009, -23.606, -23.006, -22.187, -21.252, -20.330,
    -19.487, -18.729, -24.492, -24.303, -24.018, -23.614, -23.012, -22.189,
    -21.251, -20.327, -19.484, -18.728, -24.503, -24.313, -24.027, -23.622,
    -23.018, -22.191, -21.249, -20.324, -19.482, -18.727, -24.513, -24.323,
    -24.036, -23.630, -23.024, -22.193, -21.248, -20.321, -19.480, -18.726,
    -24.524, -24.333, -24.044, -23.638, -23.030, -22.195, -21.246, -20.318,
    -19.477, -18.726, -24.535, -24.342, -24.053, -23.645, -23.035, -22.197,
    -21.244, -20.315, -19.475, -18.725, -24.546, -24.352, -24.062, -23.653,
    -23.041, -22.198, -21.242, -20.312, -19.472, -18.724, -24.557, -24.362,
    -24.071, -23.661, -23.046, -22.199, -21.240, -20.308, -19.470, -18.723,
    -24.568, -24.372, -24.080, -23.668, -23.052, -22.200, -21.237, -20.305,
    -19.467, -18.722, -24.578, -24.381, -24.088, -23.676, -23.057, -22.202,
    -21.235, -20.301, -19.465, -18.721, -24.589, -24.391, -24.097, -23.684,
    -23.062, -22.202, -21.232, -20.298, -19.462, -18.720, -24.600, -24.401,
    -24.106, -23.691, -23.067, -22.203, -21.229, -20.294, -19.459, -18.719,
    -24.611, -24.411, -24.115, -23.699, -23.072, -22.204, -21.227, -20.290,
    -19.457, -18.718, -24.622, -24.420, -24.124, -23.707, -23.077, -22.204,
    -21.223, -20.287, -19.454, -18.717, -24.633, -24.430, -24.132, -23.714,
    -23.082, -22.204, -21.220, -20.283, -19.451, -18.716, -24.643, -24.440,
    -24.141, -23.722, -23.087, -22.204, -21.217, -20.279, -19.448, -18.715,
    -24.654, -24.450, -24.150, -23.729, -23.091, -22.204, -21.214, -20.275,
    -19.445, -18.714, -24.665, -24.459, -24.159, -23.737, -23.096, -22.204,
    -21.210, -20.270, -19.442, -18.713, -24.676, -24.469, -24.167, -23.744,
    -23.100, -22.204, -21.206, -20.266, -19.439, -18.712, -24.687, -24.479,
    -24.176, -23.751, -23.104, -22.203, -21.203, -20.262, -19.436, -18.710,
    -24.698, -24.489, -24.185, -23.759, -23.108, -22.202, -21.199, -20.258,
    -19.433, -18.709, -24.708, -24.498, -24.194, -23.766, -23.113, -22.201,
    -21.195, -20.253, -19.430, -18.708, -24.719, -24.508, -24.203, -23.774,
    -23.116, -22.200, -21.191, -20.249, -19.427, -18.707, -24.730, -24.518,
    -24.211, -23.781, -23.120, -22.199, -21.186, -20.244, -19.424, -18.706,
    -24.741, -24.528, -24.220, -23.788, -23.124, -22.198, -21.182, -20.240,
    -19.420, -18.705, -24.752, -24.537, -24.229, -23.795, -23.128, -22.196,
    -21.178, -20.235, -19.417, -18.703, -24.763, -24.547, -24.238, -23.802,
    -23.131, -22.195, -21.173, -20.231, -19.414, -18.702, -24.774, -24.557,
    -24.246, -23.810, -23.134, -22.193, -21.169, -20.226, -19.411, -18.701,
    -24.784, -24.567, -24.255, -23.817, -23.138, -22.191, -21.164, -20.221,
    -19.408, -18.700, -24.795, -24.576, -24.264, -23.824, -23.141, -22.189,
    -21.159, -20.217, -19.404, -18.698, -24.806, -24.586, -24.273, -23.831,
    -23.144, -22.187, -21.154, -20.212, -19.401, -18.697, -24.817, -24.596,
    -24.282, -23.838, -23.147, -22.184, -21.149, -20.207, -19.398, -18.696,
    -24.828, -24.606, -24.290, -23.845, -23.149, -22.182, -21.144, -20.202,
    -19.395, -18.695, -24.839, -24.615, -24.299, -23.852, -23.152, -22.179,
    -21.139, -20.197, -19.391, -18.693, -24.849, -24.625, -24.308, -23.859,
    -23.155, -22.177, -21.134, -20.192, -19.388, -18.692, -24.860, -24.635,
    -24.317, -23.865, -23.157, -22.174, -21.129, -20.188, -19.385, -18.691,
};

/*
const double CIA_Orton::N2N2[][10] = {
 -24.146, -24.892, -25.543, -26.124, -26.652, -27.139, -27.593, -28.021,
-28.427, -28.815, -22.313, -23.059, -23.711, -24.292, -24.819, -25.306, -25.760,
-26.188, -26.594, -26.982, -20.926, -21.673, -22.324, -22.905, -23.433, -23.919,
-24.374, -24.802, -25.208, -25.596, -19.539, -20.286, -20.937, -21.518, -22.046,
-22.533, -22.987, -23.415, -23.821, -24.209, -18.727, -19.474, -20.125, -20.706,
-21.234, -21.721, -22.176, -22.604, -23.010, -23.398, -18.150, -18.897, -19.548,
-20.130, -20.657, -21.145, -21.599, -22.028, -22.434, -22.822, -17.702, -18.448,
-19.100, -19.682, -20.210, -20.697, -21.152, -21.581, -21.987, -22.375, -17.335,
-18.082, -18.734, -19.316, -19.844, -20.331, -20.787, -21.215, -21.622, -22.010,
 -17.025, -17.772, -18.424, -19.006, -19.534, -20.022, -20.477, -20.906,
-21.313, -21.701, -16.757, -17.503, -18.155, -18.737, -19.266, -19.754, -20.209,
-20.638, -21.045, -21.434, -16.519, -17.266, -17.918, -18.500, -19.029, -19.517,
-19.973, -20.402, -20.809, -21.198, -16.307, -17.054, -17.706, -18.288, -18.817,
-19.305, -19.761, -20.190, -20.598, -20.987, -16.115, -16.862, -17.514, -18.096,
-18.625, -19.113, -19.570, -19.999, -20.406, -20.796, -15.940, -16.687, -17.339,
-17.921, -18.450, -18.938, -19.395, -19.824, -20.232, -20.621, -15.780, -16.526,
-17.178, -17.760, -18.289, -18.777, -19.234, -19.664, -20.071, -20.461, -15.632,
-16.377, -17.029, -17.611, -18.140, -18.628, -19.085, -19.515, -19.922, -20.312,
 -15.493, -16.239, -16.890, -17.472, -18.001, -18.489, -18.946, -19.376,
-19.784, -20.173, -15.364, -16.109, -16.760, -17.342, -17.871, -18.359, -18.816,
-19.246, -19.654, -20.044, -15.243, -15.988, -16.638, -17.220, -17.749, -18.237,
-18.694, -19.124, -19.532, -19.922, -15.130, -15.874, -16.524, -17.105, -17.634,
-18.123, -18.579, -19.010, -19.418, -19.808, -15.023, -15.766, -16.416, -16.997,
-17.526, -18.014, -18.471, -18.901, -19.309, -19.699, -14.923, -15.665, -16.314,
-16.895, -17.423, -17.911, -18.368, -18.798, -19.206, -19.596, -14.489, -15.227,
-15.873, -16.452, -16.978, -17.466, -17.922, -18.352, -18.760, -19.150, -14.144,
-14.878, -15.520, -16.095, -16.620, -17.106, -17.560, -17.989, -18.397, -18.786,
 -13.860, -14.591, -15.230, -15.802, -16.324, -16.807, -17.260, -17.687,
-18.093, -18.481, -13.622, -14.350, -14.985, -15.554, -16.073, -16.553, -17.003,
-17.428, -17.832, -18.219, -13.424, -14.148, -14.780, -15.345, -15.860, -16.337,
-16.783, -17.205, -17.607, -17.991, -13.257, -13.978, -14.606, -15.167, -15.678,
-16.150, -16.593, -17.011, -17.409, -17.791, -13.110, -13.828, -14.453, -15.010,
-15.516, -15.984, -16.423, -16.837, -17.232, -17.611, -12.977, -13.693, -14.315,
-14.869, -15.371, -15.835, -16.270, -16.680, -17.072, -17.447, -12.850, -13.564,
-14.185, -14.737, -15.237, -15.698, -16.129, -16.536, -16.924, -17.297, -12.729,
-13.443, -14.064, -14.615, -15.113, -15.571, -15.999, -16.403, -16.789, -17.159,
 -12.613, -13.328, -13.949, -14.499, -14.996, -15.452, -15.878, -16.279,
-16.662, -17.030, -12.506, -13.220, -13.841, -14.390, -14.886, -15.340, -15.764,
-16.163, -16.544, -16.910, -12.408, -13.122, -13.741, -14.290, -14.784, -15.237,
-15.658, -16.055, -16.433, -16.797, -12.318, -13.031, -13.649, -14.197, -14.689,
-15.139, -15.558, -15.953, -16.328, -16.690, -12.237, -12.949, -13.566, -14.111,
-14.600, -15.048, -15.465, -15.856, -16.230, -16.590, -12.164, -12.876, -13.491,
-14.033, -14.519, -14.964, -15.378, -15.767, -16.138, -16.496, -12.040, -12.752,
-13.362, -13.897, -14.376, -14.814, -15.221, -15.605, -15.971, -16.324, -11.930,
-12.643, -13.249, -13.779, -14.252, -14.684, -15.085, -15.463, -15.823, -16.171,
 -11.805, -12.526, -13.136, -13.666, -14.138, -14.567, -14.964, -15.336,
-15.691, -16.034, -11.677, -12.408, -13.027, -13.563, -14.037, -14.465, -14.859,
-15.227, -15.576, -15.913, -11.545, -12.283, -12.911, -13.456, -13.936, -14.367,
-14.760, -15.125, -15.470, -15.802, -11.419, -12.162, -12.799, -13.353, -13.839,
-14.273, -14.666, -15.029, -15.370, -15.698, -11.306, -12.052, -12.696, -13.256,
-13.747, -14.183, -14.577, -14.938, -15.276, -15.600, -11.199, -11.948, -12.595,
-13.159, -13.654, -14.093, -14.488, -14.848, -15.184, -15.506, -11.099, -11.849,
-12.498, -13.066, -13.564, -14.005, -14.401, -14.762, -15.097, -15.416, -10.998,
-11.747, -12.398, -12.969, -13.471, -13.917, -14.316, -14.678, -15.013, -15.331,
 -10.883, -11.631, -12.286, -12.862, -13.371, -13.823, -14.229, -14.594,
-14.932, -15.250, -10.767, -11.511, -12.166, -12.748, -13.266, -13.727, -14.140,
-14.512, -14.852, -15.171, -10.649, -11.389, -12.046, -12.634, -13.160, -13.631,
-14.053, -14.430, -14.775, -15.096, -10.542, -11.280, -11.938, -12.531, -13.064,
-13.542, -13.971, -14.354, -14.702, -15.024, -10.458, -11.196, -11.855, -12.450,
-12.985, -13.466, -13.898, -14.284, -14.634, -14.957, -10.391, -11.131, -11.792,
-12.387, -12.921, -13.402, -13.834, -14.221, -14.571, -14.894, -10.340, -11.081,
-11.740, -12.331, -12.863, -13.342, -13.773, -14.160, -14.511, -14.834, -10.296,
-11.032, -11.685, -12.273, -12.801, -13.279, -13.711, -14.099, -14.451, -14.775,
 -10.231, -10.959, -11.608, -12.194, -12.724, -13.205, -13.641, -14.034,
-14.389, -14.715, -10.174, -10.888, -11.528, -12.110, -12.641, -13.126, -13.568,
-13.966, -14.327, -14.656, -10.135, -10.830, -11.458, -12.033, -12.563, -13.050,
-13.497, -13.901, -14.266, -14.599, -10.108, -10.787, -11.403, -11.971, -12.497,
-12.985, -13.433, -13.840, -14.208, -14.544, -10.104, -10.767, -11.370, -11.928,
-12.448, -12.932, -13.379, -13.786, -14.156, -14.493, -10.108, -10.757, -11.348,
-11.898, -12.410, -12.888, -13.332, -13.737, -14.106, -14.444, -10.106, -10.742,
-11.324, -11.865, -12.370, -12.844, -13.284, -13.688, -14.058, -14.396, -10.093,
-10.716, -11.288, -11.822, -12.322, -12.792, -13.232, -13.637, -14.007, -14.347,
 -10.066, -10.674, -11.236, -11.764, -12.261, -12.732, -13.173, -13.580,
-13.954, -14.297, -10.040, -10.627, -11.177, -11.699, -12.194, -12.665, -13.110,
-13.522, -13.900, -14.247, -10.019, -10.586, -11.123, -11.637, -12.130, -12.602,
-13.049, -13.464, -13.847, -14.197, -10.009, -10.559, -11.084, -11.591, -12.080,
-12.549, -12.996, -13.413, -13.797, -14.150, -10.014, -10.551, -11.066, -11.564,
-12.046, -12.511, -12.955, -13.370, -13.754, -14.107, -10.026, -10.557, -11.063,
-11.553, -12.027, -12.484, -12.922, -13.333, -13.715, -14.067, -10.039, -10.563,
-11.062, -11.543, -12.008, -12.458, -12.889, -13.296, -13.676, -14.028, -10.040,
-10.556, -11.047, -11.521, -11.978, -12.422, -12.849, -13.254, -13.633, -13.986,
 -10.019, -10.526, -11.009, -11.477, -11.930, -12.372, -12.799, -13.205,
-13.586, -13.942, -10.008, -10.498, -10.969, -11.428, -11.877, -12.317, -12.745,
-13.153, -13.538, -13.896, -10.018, -10.487, -10.941, -11.389, -11.831, -12.268,
-12.695, -13.105, -13.491, -13.853, -10.046, -10.495, -10.933, -11.368, -11.801,
-12.232, -12.655, -13.063, -13.450, -13.812, -10.098, -10.527, -10.947, -11.367,
-11.788, -12.210, -12.627, -13.030, -13.414, -13.776, -10.154, -10.566, -10.969,
-11.375, -11.784, -12.195, -12.604, -13.001, -13.382, -13.742, -10.199, -10.595,
-10.985, -11.379, -11.776, -12.178, -12.579, -12.971, -13.349, -13.707, -10.229,
-10.609, -10.987, -11.369, -11.758, -12.152, -12.549, -12.937, -13.313, -13.671,
 -10.251, -10.611, -10.973, -11.345, -11.726, -12.116, -12.510, -12.898,
-13.274, -13.633, -10.277, -10.612, -10.955, -11.315, -11.689, -12.075, -12.467,
-12.855, -13.232, -13.593, -10.314, -10.621, -10.945, -11.291, -11.656, -12.038,
-12.428, -12.815, -13.193, -13.555, -10.369, -10.648, -10.950, -11.281, -11.637,
-12.012, -12.397, -12.782, -13.158, -13.520, -10.433, -10.689, -10.973, -11.290,
-11.634, -11.999, -12.377, -12.756, -13.129, -13.489, -10.498, -10.737, -11.007,
-11.311, -11.642, -11.997, -12.366, -12.737, -13.104, -13.461, -10.552, -10.780,
-11.038, -11.329, -11.649, -11.993, -12.353, -12.717, -13.079, -13.433, -10.583,
-10.802, -11.051, -11.333, -11.643, -11.978, -12.331, -12.691, -13.050, -13.402,
 -10.597, -10.805, -11.043, -11.316, -11.620, -11.950, -12.300, -12.657,
-13.016, -13.368, -10.616, -10.809, -11.033, -11.295, -11.591, -11.917, -12.264,
-12.621, -12.979, -13.333, -10.658, -10.831, -11.037, -11.286, -11.572, -11.891,
-12.234, -12.588, -12.946, -13.299, -10.719, -10.872, -11.060, -11.293, -11.567,
-11.877, -12.213, -12.563, -12.917, -13.270, -10.797, -10.933, -11.103, -11.319,
-11.579, -11.876, -12.203, -12.546, -12.895, -13.244, -10.876, -10.998, -11.152,
-11.352, -11.597, -11.882, -12.198, -12.532, -12.876, -13.221, -10.936, -11.049,
-11.191, -11.378, -11.610, -11.884, -12.191, -12.518, -12.856, -13.197, -10.982,
-11.085, -11.215, -11.390, -11.612, -11.877, -12.176, -12.498, -12.832, -13.171,
 -11.026, -11.113, -11.228, -11.390, -11.602, -11.859, -12.154, -12.472,
-12.804, -13.142, -11.076, -11.144, -11.239, -11.387, -11.587, -11.838, -12.128,
-12.443, -12.774, -13.112, -11.144, -11.187, -11.260, -11.390, -11.579, -11.821,
-12.106, -12.418, -12.747, -13.083, -11.234, -11.250, -11.298, -11.409, -11.583,
-11.815, -12.093, -12.399, -12.724, -13.059, -11.332, -11.323, -11.350, -11.443,
-11.602, -11.822, -12.090, -12.389, -12.708, -13.038, -11.424, -11.399, -11.409,
-11.486, -11.631, -11.838, -12.094, -12.383, -12.695, -13.020, -11.498, -11.464,
-11.461, -11.524, -11.656, -11.850, -12.096, -12.376, -12.682, -13.001, -11.545,
-11.506, -11.494, -11.546, -11.667, -11.851, -12.089, -12.363, -12.663, -12.980,
 -11.583, -11.534, -11.510, -11.552, -11.663, -11.840, -12.072, -12.342,
-12.640, -12.955, -11.623, -11.562, -11.524, -11.553, -11.655, -11.825, -12.051,
-12.318, -12.614, -12.928, -11.678, -11.606, -11.552, -11.566, -11.656, -11.817,
-12.037, -12.299, -12.592, -12.904, -11.748, -11.665, -11.596, -11.595, -11.671,
-11.820, -12.032, -12.287, -12.575, -12.884, -11.825, -11.738, -11.656, -11.639,
-11.699, -11.836, -12.036, -12.283, -12.564, -12.868, -11.898, -11.810, -11.718,
-11.687, -11.733, -11.856, -12.045, -12.282, -12.555, -12.854, -11.954, -11.867,
-11.769, -11.727, -11.760, -11.871, -12.050, -12.278, -12.545, -12.838, -12.001,
-11.913, -11.807, -11.754, -11.777, -11.878, -12.048, -12.269, -12.531, -12.821,
 -12.052, -11.957, -11.839, -11.773, -11.784, -11.876, -12.038, -12.255,
-12.513, -12.800, -12.109, -12.004, -11.872, -11.790, -11.788, -11.871, -12.027,
-12.239, -12.493, -12.779, -12.181, -12.066, -11.915, -11.815, -11.799, -11.870,
-12.019, -12.225, -12.476, -12.759, -12.267, -12.142, -11.973, -11.854, -11.822,
-11.881, -12.020, -12.219, -12.465, -12.743, -12.358, -12.225, -12.041, -11.906,
-11.858, -11.903, -12.030, -12.220, -12.458, -12.731, -12.433, -12.303, -12.111,
-11.962, -11.900, -11.931, -12.046, -12.226, -12.455, -12.722, -12.496, -12.369,
-12.172, -12.012, -11.937, -11.956, -12.059, -12.229, -12.451, -12.711, -12.539,
-12.416, -12.215, -12.047, -11.961, -11.969, -12.063, -12.225, -12.441, -12.697,
 -12.581, -12.457, -12.248, -12.069, -11.973, -11.972, -12.058, -12.215,
-12.427, -12.680, -12.624, -12.497, -12.279, -12.088, -11.981, -11.971, -12.051,
-12.202, -12.411, -12.661, -12.670, -12.548, -12.321, -12.118, -11.998, -11.978,
-12.049, -12.195, -12.398, -12.646, -12.720, -12.605, -12.375, -12.161, -12.028,
-11.995, -12.057, -12.194, -12.391, -12.634, -12.768, -12.667, -12.436, -12.214,
-12.069, -12.023, -12.072, -12.200, -12.389, -12.625, -12.805, -12.722, -12.496,
-12.269, -12.112, -12.053, -12.091, -12.208, -12.389, -12.619, -12.832, -12.765,
-12.545, -12.314, -12.148, -12.078, -12.105, -12.213, -12.386, -12.611, -12.860,
-12.803, -12.586, -12.351, -12.175, -12.095, -12.113, -12.213, -12.380, -12.600,
 -12.892, -12.844, -12.625, -12.382, -12.196, -12.106, -12.115, -12.209,
-12.371, -12.587, -12.928, -12.890, -12.667, -12.413, -12.215, -12.114, -12.115,
-12.202, -12.360, -12.572, -12.970, -12.947, -12.719, -12.453, -12.241, -12.128,
-12.119, -12.199, -12.352, -12.560, -13.014, -13.008, -12.778, -12.503, -12.277,
-12.151, -12.132, -12.204, -12.349, -12.552, -13.064, -13.070, -12.842, -12.561,
-12.323, -12.184, -12.153, -12.214, -12.351, -12.548, -13.105, -13.127, -12.903,
-12.621, -12.373, -12.221, -12.178, -12.228, -12.356, -12.545, -13.148, -13.179,
-12.959, -12.674, -12.418, -12.254, -12.199, -12.239, -12.358, -12.541, -13.188,
-13.226, -13.005, -12.715, -12.451, -12.277, -12.212, -12.243, -12.356, -12.533,
 -13.234, -13.275, -13.047, -12.750, -12.475, -12.291, -12.217, -12.242,
-12.349, -12.523, -13.277, -13.322, -13.089, -12.782, -12.497, -12.303, -12.221,
-12.239, -12.342, -12.511, -13.313, -13.366, -13.135, -12.822, -12.527, -12.322,
-12.231, -12.241, -12.338, -12.503, -13.349, -13.408, -13.185, -12.870, -12.566,
-12.350, -12.248, -12.250, -12.339, -12.498, -13.385, -13.449, -13.238, -12.923,
-12.613, -12.386, -12.273, -12.264, -12.344, -12.497, -13.419, -13.486, -13.287,
-12.975, -12.661, -12.423, -12.299, -12.279, -12.351, -12.496, -13.454, -13.520,
-13.330, -13.020, -12.702, -12.455, -12.320, -12.292, -12.355, -12.494, -13.494,
-13.557, -13.372, -13.060, -12.736, -12.481, -12.336, -12.299, -12.356, -12.489,
 -13.537, -13.597, -13.417, -13.099, -12.767, -12.501, -12.347, -12.303,
-12.353, -12.482, -13.582, -13.640, -13.464, -13.139, -12.797, -12.520, -12.357,
-12.305, -12.350, -12.475, -13.627, -13.685, -13.517, -13.186, -12.835, -12.545,
-12.372, -12.311, -12.350, -12.469, -13.668, -13.725, -13.568, -13.236, -12.879,
-12.578, -12.394, -12.324, -12.354, -12.468, -13.709, -13.763, -13.619, -13.290,
-12.930, -12.619, -12.423, -12.342, -12.364, -12.470, -13.744, -13.796, -13.664,
-13.341, -12.981, -12.661, -12.454, -12.363, -12.375, -12.473, -13.781, -13.830,
-13.706, -13.388, -13.028, -12.699, -12.482, -12.380, -12.383, -12.474, -13.819,
-13.866, -13.747, -13.429, -13.066, -12.728, -12.501, -12.391, -12.387, -12.472,
 -13.861, -13.906, -13.789, -13.468, -13.100, -12.751, -12.515, -12.397,
-12.387, -12.467, -13.904, -13.949, -13.833, -13.507, -13.135, -12.774, -12.528,
-12.403, -12.386, -12.462, -13.937, -13.985, -13.874, -13.550, -13.174, -12.802,
-12.547, -12.413, -12.390, -12.460, -13.967, -14.017, -13.913, -13.595, -13.220,
-12.838, -12.573, -12.429, -12.398, -12.461, -13.991, -14.043, -13.947, -13.640,
-13.270, -12.878, -12.604, -12.450, -12.410, -12.465, -14.007, -14.061, -13.974,
-13.681, -13.318, -12.919, -12.634, -12.471, -12.422, -12.470, -14.024, -14.078,
-13.997, -13.717, -13.361, -12.954, -12.661, -12.489, -12.431, -12.473, -14.046,
-14.100, -14.025, -13.753, -13.401, -12.985, -12.683, -12.503, -12.438, -12.473,
 -14.069, -14.125, -14.056, -13.791, -13.441, -13.012, -12.702, -12.513,
-12.441, -12.471, -14.096, -14.154, -14.091, -13.832, -13.483, -13.041, -12.720,
-12.523, -12.445, -12.470, -14.123, -14.186, -14.128, -13.877, -13.531, -13.074,
-12.743, -12.538, -12.452, -12.470, -14.154, -14.211, -14.157, -13.919, -13.580,
-13.113, -12.772, -12.557, -12.463, -12.474, -14.190, -14.239, -14.184, -13.959,
-13.631, -13.156, -12.806, -12.581, -12.478, -12.482, -14.228, -14.269, -14.212,
-13.998, -13.680, -13.200, -12.841, -12.607, -12.494, -12.489, -14.270, -14.305,
-14.244, -14.037, -13.727, -13.241, -12.873, -12.629, -12.507, -12.495, -14.316,
-14.348, -14.282, -14.077, -13.771, -13.276, -12.898, -12.645, -12.516, -12.498,
 -14.362, -14.393, -14.324, -14.118, -13.815, -13.308, -12.919, -12.658,
-12.522, -12.498, -14.406, -14.436, -14.364, -14.159, -13.857, -13.340, -12.940,
-12.671, -12.528, -12.499, -14.440, -14.471, -14.399, -14.197, -13.899, -13.377,
-12.965, -12.688, -12.538, -12.502, -14.473, -14.502, -14.430, -14.231, -13.940,
-13.418, -12.996, -12.711, -12.551, -12.509, -14.509, -14.533, -14.458, -14.264,
-13.979, -13.461, -13.031, -12.736, -12.568, -12.518, -14.543, -14.561, -14.484,
-14.294, -14.015, -13.504, -13.064, -12.762, -12.585, -12.527, -14.585, -14.595,
-14.514, -14.324, -14.051, -13.544, -13.094, -12.784, -12.599, -12.534, -14.632,
-14.633, -14.547, -14.358, -14.087, -13.581, -13.121, -12.802, -12.610, -12.539,
 -14.682, -14.676, -14.586, -14.395, -14.125, -13.619, -13.145, -12.819,
-12.619, -12.542, -14.732, -14.720, -14.627, -14.433, -14.163, -13.658, -13.170,
-12.836, -12.629, -12.545, -14.780, -14.764, -14.667, -14.471, -14.202, -13.701,
-13.199, -12.856, -12.641, -12.551, -14.826, -14.799, -14.697, -14.502, -14.236,
-13.745, -13.232, -12.881, -12.658, -12.560, -14.870, -14.833, -14.723, -14.529,
-14.268, -13.791, -13.269, -12.909, -12.678, -12.572, -14.912, -14.865, -14.748,
-14.554, -14.297, -13.834, -13.305, -12.938, -12.698, -12.584, -14.956, -14.901,
-14.777, -14.582, -14.327, -13.875, -13.337, -12.963, -12.715, -12.594, -15.003,
-14.942, -14.812, -14.614, -14.360, -13.915, -13.366, -12.984, -12.728, -12.600,
 -15.050, -14.986, -14.851, -14.649, -14.394, -13.954, -13.392, -13.002,
-12.739, -12.605, -15.098, -15.034, -14.894, -14.688, -14.430, -13.993, -13.419,
-13.022, -12.751, -12.611, -15.138, -15.073, -14.931, -14.722, -14.464, -14.033,
-13.449, -13.044, -12.766, -12.619, -15.173, -15.106, -14.962, -14.752, -14.495,
-14.071, -13.483, -13.071, -12.785, -12.630, -15.204, -15.134, -14.987, -14.777,
-14.522, -14.107, -13.518, -13.100, -12.806, -12.643, -15.227, -15.151, -15.003,
-14.796, -14.546, -14.141, -13.551, -13.128, -12.826, -12.656, -15.256, -15.173,
-15.021, -14.816, -14.570, -14.174, -13.582, -13.153, -12.843, -12.666, -15.287,
-15.198, -15.045, -14.841, -14.596, -14.207, -13.611, -13.175, -12.859, -12.675,
 -15.320, -15.226, -15.074, -14.871, -14.627, -14.241, -13.639, -13.196,
-12.872, -12.682, -15.356, -15.259, -15.107, -14.904, -14.658, -14.276, -13.669,
-13.218, -12.887, -12.691, -15.388, -15.292, -15.140, -14.935, -14.689, -14.311,
-13.701, -13.243, -12.905, -12.701, -15.430, -15.321, -15.163, -14.957, -14.713,
-14.342, -13.735, -13.272, -12.925, -12.714, -15.476, -15.356, -15.187, -14.978,
-14.735, -14.371, -13.771, -13.303, -12.949, -12.730, -15.524, -15.394, -15.214,
-15.000, -14.758, -14.400, -13.806, -13.333, -12.971, -12.745, -15.573, -15.437,
-15.247, -15.028, -14.784, -14.430, -13.839, -13.360, -12.992, -12.758, -15.624,
-15.485, -15.289, -15.064, -14.815, -14.462, -13.870, -13.385, -13.009, -12.768,
 -15.670, -15.531, -15.332, -15.101, -14.849, -14.495, -13.901, -13.407,
-13.024, -12.777, -15.713, -15.575, -15.373, -15.137, -14.881, -14.527, -13.932,
-13.431, -13.041, -12.787, -15.750, -15.612, -15.407, -15.167, -14.909, -14.557,
-13.964, -13.458, -13.060, -12.799, -15.786, -15.646, -15.437, -15.193, -14.933,
-14.585, -13.998, -13.487, -13.083, -12.815, -15.825, -15.680, -15.465, -15.217,
-14.956, -14.611, -14.033, -13.518, -13.107, -12.831, -15.864, -15.713, -15.491,
-15.240, -14.977, -14.636, -14.066, -13.547, -13.129, -12.846, -15.910, -15.753,
-15.523, -15.266, -15.001, -14.662, -14.098, -13.574, -13.150, -12.860, -15.961,
-15.797, -15.561, -15.297, -15.029, -14.691, -14.130, -13.599, -13.169, -12.872,
 -16.012, -15.844, -15.603, -15.334, -15.060, -14.721, -14.162, -13.624,
-13.186, -12.883, -16.065, -15.893, -15.647, -15.371, -15.092, -14.751, -14.196,
-13.651, -13.206, -12.895, -16.113, -15.939, -15.688, -15.406, -15.122, -14.780,
-14.230, -13.679, -13.227, -12.910, -16.161, -15.981, -15.721, -15.433, -15.145,
-14.804, -14.264, -13.710, -13.251, -12.927, -16.208, -16.020, -15.749, -15.455,
-15.165, -14.826, -14.298, -13.742, -13.277, -12.945, -16.256, -16.061, -15.779,
-15.477, -15.185, -14.848, -14.330, -13.773, -13.302, -12.962, -16.303, -16.103,
-15.812, -15.504, -15.208, -14.872, -14.362, -13.802, -13.324, -12.978, -16.351,
-16.150, -15.852, -15.537, -15.236, -14.899, -14.394, -13.829, -13.345, -12.991,
 -16.398, -16.199, -15.897, -15.575, -15.268, -14.928, -14.426, -13.856,
-13.364, -13.004, -16.446, -16.250, -15.945, -15.615, -15.302, -14.958, -14.458,
-13.885, -13.385, -13.018, -16.488, -16.293, -15.984, -15.650, -15.331, -14.985,
-14.490, -13.915, -13.408, -13.034, -16.528, -16.332, -16.019, -15.679, -15.357,
-15.010, -14.520, -13.947, -13.434, -13.052, -16.564, -16.365, -16.046, -15.702,
-15.377, -15.031, -14.549, -13.978, -13.459, -13.070, -16.597, -16.392, -16.066,
-15.719, -15.393, -15.050, -14.576, -14.009, -13.484, -13.088, -16.637, -16.424,
-16.091, -15.740, -15.412, -15.070, -14.603, -14.038, -13.507, -13.104, -16.675,
-16.458, -16.120, -15.765, -15.436, -15.093, -14.631, -14.067, -13.529, -13.119,
 -16.714, -16.493, -16.153, -15.796, -15.463, -15.119, -14.660, -14.096,
-13.551, -13.133, -16.756, -16.532, -16.190, -15.829, -15.493, -15.146, -14.689,
-14.127, -13.574, -13.149, -16.795, -16.569, -16.223, -15.859, -15.519, -15.170,
-14.717, -14.159, -13.599, -13.167, -16.841, -16.608, -16.252, -15.881, -15.538,
-15.190, -14.742, -14.190, -13.626, -13.187, -16.890, -16.651, -16.283, -15.904,
-15.556, -15.208, -14.766, -14.221, -13.654, -13.207, -16.939, -16.697, -16.318,
-15.929, -15.577, -15.228, -14.791, -14.252, -13.681, -13.226, -16.988, -16.745,
-16.357, -15.960, -15.603, -15.251, -14.816, -14.282, -13.706, -13.243, -17.035,
-16.794, -16.403, -15.999, -15.635, -15.279, -14.844, -14.313, -13.730, -13.260,
 -17.078, -16.841, -16.448, -16.038, -15.667, -15.307, -14.872, -14.343,
-13.755, -13.275, -17.120, -16.886, -16.491, -16.075, -15.698, -15.334, -14.899,
-14.372, -13.780, -13.293, -17.159, -16.925, -16.527, -16.106, -15.724, -15.357,
-14.924, -14.401, -13.807, -13.311, -17.199, -16.964, -16.561, -16.134, -15.747,
-15.378, -14.947, -14.430, -13.835, -13.332, -17.239, -17.003, -16.593, -16.159,
-15.767, -15.397, -14.969, -14.457, -13.864, -13.352, -17.281, -17.041, -16.625,
-16.184, -15.788, -15.415, -14.991, -14.484, -13.891, -13.372, -17.328, -17.087,
-16.663, -16.214, -15.812, -15.437, -15.014, -14.512, -13.918, -13.390, -17.376,
-17.135, -16.705, -16.249, -15.840, -15.462, -15.038, -14.539, -13.944, -13.407,
 -17.423, -17.184, -16.751, -16.289, -15.873, -15.489, -15.065, -14.567,
-13.971, -13.424, -17.471, -17.234, -16.798, -16.329, -15.906, -15.516, -15.090,
-14.595, -13.999, -13.443, -17.516, -17.282, -16.840, -16.364, -15.934, -15.540,
-15.114, -14.622, -14.028, -13.463, -17.562, -17.329, -16.878, -16.394, -15.957,
-15.560, -15.135, -14.647, -14.057, -13.485, -17.607, -17.374, -16.912, -16.419,
-15.976, -15.577, -15.154, -14.671, -14.086, -13.507, -17.651, -17.421, -16.949,
-16.446, -15.997, -15.596, -15.174, -14.695, -14.114, -13.527, -17.695, -17.469,
-16.989, -16.478, -16.022, -15.617, -15.196, -14.720, -14.142, -13.547, -17.738,
-17.519, -17.034, -16.515, -16.053, -15.642, -15.220, -14.745, -14.169, -13.565,
 -17.781, -17.572, -17.083, -16.558, -16.087, -15.671, -15.245, -14.771,
-14.197, -13.584, -17.824, -17.623, -17.131, -16.601, -16.122, -15.699, -15.271,
-14.797, -14.224, -13.604, -17.865, -17.671, -17.175, -16.638, -16.153, -15.724,
-15.294, -14.821, -14.252, -13.625, -17.904, -17.716, -17.213, -16.670, -16.178,
-15.746, -15.315, -14.845, -14.279, -13.647, -17.943, -17.759, -17.246, -16.696,
-16.199, -15.764, -15.333, -14.867, -14.305, -13.669, -17.981, -17.800, -17.274,
-16.718, -16.216, -15.779, -15.351, -14.888, -14.330, -13.690, -18.021, -17.845,
-17.308, -16.745, -16.238, -15.798, -15.370, -14.910, -14.356, -13.710, -18.061,
-17.889, -17.343, -16.775, -16.263, -15.821, -15.391, -14.933, -14.382, -13.729,
 -18.100, -17.933, -17.381, -16.809, -16.293, -15.846, -15.415, -14.956,
-14.408, -13.749, -18.140, -17.979, -17.422, -16.846, -16.323, -15.872, -15.438,
-14.980, -14.434, -13.771, -18.179, -18.022, -17.458, -16.876, -16.348, -15.893,
-15.458, -15.002, -14.459, -13.793, -18.221, -18.072, -17.495, -16.904, -16.369,
-15.911, -15.476, -15.022, -14.483, -13.817, -18.263, -18.122, -17.533, -16.932,
-16.390, -15.928, -15.493, -15.041, -14.507, -13.840, -18.305, -18.172, -17.575,
-16.964, -16.414, -15.948, -15.512, -15.062, -14.530, -13.862, -18.346, -18.221,
-17.620, -17.001, -16.444, -15.972, -15.533, -15.083, -14.554, -13.884, -18.385,
-18.267, -17.666, -17.042, -16.478, -16.000, -15.557, -15.106, -14.579, -13.905,
 -18.423, -18.311, -17.713, -17.083, -16.513, -16.029, -15.582, -15.129,
-14.603, -13.927, -18.461, -18.354, -17.756, -17.121, -16.544, -16.055, -15.604,
-15.151, -14.627, -13.949, -18.498, -18.396, -17.796, -17.155, -16.572, -16.077,
-15.625, -15.172, -14.650, -13.973, -18.535, -18.438, -17.834, -17.186, -16.596,
-16.097, -15.642, -15.190, -14.672, -13.997, -18.573, -18.479, -17.872, -17.216,
-16.619, -16.115, -15.660, -15.209, -14.694, -14.020, -18.611, -18.520, -17.910,
-17.246, -16.642, -16.134, -15.677, -15.227, -14.715, -14.043, -18.650, -18.564,
-17.954, -17.282, -16.671, -16.157, -15.697, -15.247, -14.737, -14.066, -18.690,
-18.607, -17.999, -17.321, -16.703, -16.183, -15.720, -15.268, -14.760, -14.089,
 -18.729, -18.650, -18.047, -17.363, -16.738, -16.211, -15.744, -15.290,
-14.782, -14.112, -18.768, -18.693, -18.096, -17.406, -16.773, -16.239, -15.767,
-15.312, -14.804, -14.136, -18.805, -18.734, -18.139, -17.442, -16.802, -16.263,
-15.787, -15.331, -14.826, -14.160, -18.843, -18.777, -18.183, -17.476, -16.827,
-16.283, -15.804, -15.349, -14.846, -14.184, -18.880, -18.821, -18.225, -17.507,
-16.850, -16.300, -15.820, -15.366, -14.865, -14.208, -18.917, -18.864, -18.269,
-17.540, -16.876, -16.320, -15.838, -15.383, -14.885, -14.231, -18.954, -18.906,
-18.315, -17.577, -16.905, -16.343, -15.858, -15.402, -14.906, -14.254, -18.992,
-18.947, -18.363, -17.619, -16.939, -16.371, -15.881, -15.423, -14.927, -14.277,
 -19.029, -18.987, -18.416, -17.665, -16.978, -16.402, -15.906, -15.445,
-14.948, -14.301, -19.066, -19.026, -18.466, -17.709, -17.015, -16.431, -15.929,
-15.465, -14.969, -14.324, -19.102, -19.066, -18.513, -17.749, -17.048, -16.457,
-15.951, -15.485, -14.989, -14.347, -19.139, -19.105, -18.558, -17.784, -17.076,
-16.479, -15.969, -15.502, -15.008, -14.370, -19.176, -19.145, -18.600, -17.814,
-17.100, -16.497, -15.985, -15.519, -15.026, -14.392, -19.213, -19.185, -18.642,
-17.843, -17.122, -16.514, -16.001, -15.535, -15.044, -14.414, -19.249, -19.226,
-18.685, -17.877, -17.149, -16.536, -16.020, -15.552, -15.063, -14.436, -19.286,
-19.267, -18.728, -17.912, -17.180, -16.561, -16.041, -15.571, -15.082, -14.458,
 -19.322, -19.307, -18.772, -17.950, -17.213, -16.588, -16.063, -15.592,
-15.102, -14.480, -19.358, -19.347, -18.815, -17.988, -17.246, -16.614, -16.086,
-15.611, -15.121, -14.502, -19.395, -19.387, -18.859, -18.021, -17.273, -16.635,
-16.104, -15.628, -15.139, -14.524, -19.431, -19.426, -18.907, -18.056, -17.299,
-16.655, -16.120, -15.644, -15.156, -14.545, -19.467, -19.465, -18.955, -18.090,
-17.325, -16.675, -16.136, -15.660, -15.173, -14.565, -19.503, -19.504, -19.002,
-18.128, -17.356, -16.698, -16.155, -15.677, -15.191, -14.586, -19.539, -19.543,
-19.049, -18.170, -17.391, -16.725, -16.178, -15.696, -15.210, -14.607, -19.575,
-19.582, -19.092, -18.212, -17.428, -16.755, -16.202, -15.717, -15.229, -14.628,
 -19.611, -19.621, -19.137, -18.255, -17.467, -16.785, -16.226, -15.737,
-15.248, -14.649, -19.646, -19.659, -19.181, -18.295, -17.502, -16.812, -16.248,
-15.756, -15.266, -14.669, -19.682, -19.698, -19.224, -18.332, -17.534, -16.836,
-16.267, -15.773, -15.283, -14.689, -19.718, -19.736, -19.266, -18.369, -17.565,
-16.857, -16.285, -15.789, -15.300, -14.708, -19.753, -19.775, -19.307, -18.404,
-17.594, -16.878, -16.302, -15.805, -15.316, -14.728, -19.788, -19.813, -19.348,
-18.441, -17.625, -16.901, -16.320, -15.821, -15.332, -14.747, -19.824, -19.851,
-19.391, -18.481, -17.659, -16.926, -16.341, -15.839, -15.350, -14.766, -19.859,
-19.889, -19.433, -18.524, -17.698, -16.956, -16.364, -15.859, -15.368, -14.786,
 -19.894, -19.927, -19.475, -18.568, -17.738, -16.986, -16.389, -15.879,
-15.386, -14.806, -19.929, -19.965, -19.517, -18.612, -17.778, -17.016, -16.412,
-15.899, -15.404, -14.825, -19.965, -20.002, -19.558, -18.653, -17.813, -17.040,
-16.431, -15.916, -15.421, -14.843, -20.000, -20.040, -19.601, -18.692, -17.847,
-17.062, -16.449, -15.931, -15.436, -14.861, -20.034, -20.077, -19.643, -18.730,
-17.880, -17.083, -16.465, -15.946, -15.451, -14.879, -20.069, -20.115, -19.686,
-18.771, -17.915, -17.107, -16.484, -15.962, -15.467, -14.897, -20.104, -20.152,
-19.727, -18.814, -17.954, -17.135, -16.506, -15.980, -15.484, -14.915, -20.139,
-20.189, -19.767, -18.860, -17.997, -17.166, -16.530, -16.000, -15.502, -14.934,
 -20.174, -20.226, -19.807, -18.909, -18.044, -17.199, -16.556, -16.021,
-15.520, -14.952, -20.208, -20.263, -19.847, -18.955, -18.087, -17.230, -16.580,
-16.041, -15.538, -14.970, -20.243, -20.300, -19.886, -19.000, -18.129, -17.258,
-16.602, -16.059, -15.554, -14.988, -20.277, -20.337, -19.926, -19.042, -18.168,
-17.283, -16.621, -16.075, -15.569, -15.005, -20.312, -20.374, -19.965, -19.081,
-18.204, -17.304, -16.637, -16.089, -15.584, -15.021, -20.346, -20.410, -20.005,
-19.122, -18.241, -17.327, -16.655, -16.105, -15.599, -15.038, -20.380, -20.447,
-20.045, -19.162, -18.279, -17.353, -16.675, -16.122, -15.615, -15.055, -20.414,
-20.483, -20.085, -19.204, -18.318, -17.382, -16.698, -16.141, -15.632, -15.072,
 -20.449, -20.520, -20.125, -19.245, -18.357, -17.412, -16.722, -16.161,
-15.649, -15.090, -20.483, -20.556, -20.165, -19.286, -18.396, -17.440, -16.744,
-16.179, -15.666, -15.107, -20.517, -20.592, -20.203, -19.329, -18.436, -17.465,
-16.763, -16.195, -15.680, -15.123, -20.551, -20.628, -20.242, -19.373, -18.478,
-17.489, -16.780, -16.210, -15.695, -15.138, -20.585, -20.664, -20.281, -19.417,
-18.519, -17.514, -16.799, -16.225, -15.709, -15.154, -20.619, -20.700, -20.320,
-19.460, -18.560, -17.542, -16.820, -16.243, -15.725, -15.170, -20.652, -20.736,
-20.358, -19.504, -18.601, -17.574, -16.845, -16.263, -15.742, -15.187, -20.686,
-20.771, -20.396, -19.546, -18.642, -17.607, -16.870, -16.283, -15.760, -15.204,
 -20.720, -20.807, -20.434, -19.589, -18.685, -17.641, -16.895, -16.304,
-15.777, -15.221, -20.753, -20.843, -20.473, -19.631, -18.726, -17.672, -16.918,
-16.322, -15.793, -15.237, -20.787, -20.878, -20.511, -19.672, -18.767, -17.701,
-16.938, -16.338, -15.808, -15.252, -20.821, -20.913, -20.548, -19.713, -18.807,
-17.728, -16.957, -16.354, -15.822, -15.267, -20.854, -20.949, -20.586, -19.752,
-18.845, -17.755, -16.976, -16.369, -15.836, -15.282, -20.888, -20.984, -20.624,
-19.793, -18.884, -17.784, -16.996, -16.386, -15.851, -15.297, -20.921, -21.019,
-20.661, -19.833, -18.923, -17.817, -17.020, -16.405, -15.867, -15.313, -20.954,
-21.054, -20.698, -19.873, -18.962, -17.851, -17.045, -16.426, -15.885, -15.329,
 -20.988, -21.089, -20.736, -19.913, -19.001, -17.887, -17.071, -16.446,
-15.902, -15.345, -21.021, -21.124, -20.773, -19.953, -19.039, -17.921, -17.095,
-16.465, -15.918, -15.361, -21.054, -21.159, -20.810, -19.993, -19.078, -17.953,
-17.115, -16.482, -15.932, -15.375, -21.087, -21.193, -20.847, -20.034, -19.118,
-17.982, -17.134, -16.497, -15.946, -15.389, -21.120, -21.228, -20.884, -20.075,
-19.159, -18.012, -17.153, -16.512, -15.960, -15.404, -21.153, -21.263, -20.920,
-20.115, -19.199, -18.045, -17.175, -16.530, -15.975, -15.419, -21.186, -21.297,
-20.957, -20.154, -19.238, -18.081, -17.200, -16.549, -15.991, -15.434, -21.219,
-21.332, -20.993, -20.193, -19.276, -18.119, -17.226, -16.570, -16.009, -15.450,
 -21.252, -21.366, -21.029, -20.232, -19.314, -18.160, -17.254, -16.592,
-16.026, -15.466, -21.285, -21.400, -21.066, -20.270, -19.351, -18.198, -17.279,
-16.612, -16.043, -15.481, -21.318, -21.435, -21.102, -20.308, -19.389, -18.234,
-17.302, -16.630, -16.058, -15.495, -21.351, -21.469, -21.138, -20.346, -19.427,
-18.269, -17.322, -16.646, -16.072, -15.509, -21.383, -21.503, -21.174, -20.384,
-19.465, -18.301, -17.341, -16.661, -16.085, -15.523, -21.416, -21.537, -21.209,
-20.423, -19.503, -18.336, -17.362, -16.678, -16.100, -15.537, -21.449, -21.571,
-21.245, -20.461, -19.541, -18.370, -17.385, -16.697, -16.116, -15.551, -21.481,
-21.605, -21.281, -20.500, -19.580, -18.406, -17.410, -16.717, -16.132, -15.567,
 -21.514, -21.639, -21.316, -20.538, -19.617, -18.441, -17.435, -16.737,
-16.149, -15.582, -21.546, -21.672, -21.352, -20.576, -19.655, -18.477, -17.458,
-16.755, -16.164, -15.596, -21.579, -21.706, -21.387, -20.613, -19.692, -18.513,
-17.479, -16.771, -16.178, -15.610, -21.611, -21.740, -21.422, -20.650, -19.730,
-18.550, -17.499, -16.787, -16.192, -15.623, -21.644, -21.773, -21.457, -20.688,
-19.767, -18.586, -17.521, -16.804, -16.206, -15.637, -21.676, -21.807, -21.492,
-20.725, -19.804, -18.623, -17.545, -16.824, -16.222, -15.651, -21.708, -21.840,
-21.527, -20.762, -19.841, -18.660, -17.572, -16.845, -16.239, -15.666, -21.741,
-21.874, -21.562, -20.799, -19.877, -18.698, -17.600, -16.866, -16.257, -15.681,
 -21.773, -21.907, -21.596, -20.835, -19.914, -18.737, -17.626, -16.887,
-16.273, -15.696, -21.805, -21.940, -21.631, -20.872, -19.950, -18.775, -17.650,
-16.906, -16.289, -15.710, -21.837, -21.974, -21.666, -20.909, -19.987, -18.812,
-17.672, -16.923, -16.303, -15.723, -21.870, -22.007, -21.700, -20.945, -20.023,
-18.848, -17.693, -16.940, -16.317, -15.736, -21.902, -22.040, -21.734, -20.981,
-20.059, -18.884, -17.715, -16.957, -16.331, -15.750, -21.934, -22.073, -21.769,
-21.017, -20.095, -18.920, -17.740, -16.976, -16.347, -15.764, -21.966, -22.106,
-21.803, -21.053, -20.131, -18.956, -17.766, -16.996, -16.363, -15.778, -21.998,
-22.139, -21.837, -21.089, -20.167, -18.992, -17.793, -17.018, -16.381, -15.793,
 -22.030, -22.172, -21.871, -21.125, -20.203, -19.028, -17.821, -17.039,
-16.398, -15.808, -22.062, -22.205, -21.905, -21.161, -20.238, -19.064, -17.845,
-17.059, -16.413, -15.822, -22.094, -22.238, -21.939, -21.196, -20.274, -19.101,
-17.868, -17.076, -16.428, -15.835, -22.126, -22.271, -21.973, -21.231, -20.309,
-19.138, -17.890, -17.093, -16.441, -15.847, -22.158, -22.303, -22.006, -21.267,
-20.344, -19.175, -17.913, -17.110, -16.456, -15.861, -22.189, -22.336, -22.040,
-21.302, -20.379, -19.212, -17.938, -17.130, -16.472, -15.875, -22.221, -22.369,
-22.073, -21.337, -20.414, -19.248, -17.966, -17.151, -16.489, -15.889, -22.253,
-22.401, -22.107, -21.372, -20.449, -19.284, -17.995, -17.174, -16.507, -15.904,
 -22.285, -22.434, -22.140, -21.407, -20.484, -19.320, -18.024, -17.197,
-16.524, -15.919, -22.316, -22.467, -22.174, -21.441, -20.518, -19.356, -18.051,
-17.217, -16.541, -15.933, -22.348, -22.499, -22.207, -21.476, -20.553, -19.391,
-18.076, -17.236, -16.556, -15.946, -22.380, -22.531, -22.240, -21.511, -20.587,
-19.427, -18.099, -17.253, -16.570, -15.959, -22.411, -22.564, -22.273, -21.545,
-20.621, -19.463, -18.122, -17.271, -16.584, -15.971, -22.443, -22.596, -22.306,
-21.579, -20.655, -19.499, -18.147, -17.290, -16.599, -15.985, -22.475, -22.628,
-22.339, -21.614, -20.689, -19.536, -18.173, -17.311, -16.616, -15.999, -22.506,
-22.661, -22.372, -21.648, -20.723, -19.572, -18.201, -17.333, -16.633, -16.014,
 -22.538, -22.693, -22.405, -21.682, -20.757, -19.607, -18.229, -17.354,
-16.650, -16.028, -22.569, -22.725, -22.438, -21.716, -20.791, -19.643, -18.253,
-17.373, -16.665, -16.041, -22.601, -22.757, -22.471, -21.749, -20.824, -19.679,
-18.277, -17.391, -16.679, -16.054, -22.632, -22.789, -22.504, -21.783, -20.858,
-19.714, -18.302, -17.409, -16.694, -16.066, -22.663, -22.821, -22.536, -21.817,
-20.891, -19.749, -18.328, -17.428, -16.709, -16.080, -22.695, -22.854, -22.569,
-21.850, -20.924, -19.785, -18.357, -17.450, -16.727, -16.094, -22.726, -22.885,
-22.601, -21.884, -20.957, -19.820, -18.386, -17.473, -16.745, -16.109, -22.757,
-22.917, -22.634, -21.917, -20.990, -19.855, -18.417, -17.496, -16.762, -16.124,
 -22.789, -22.949, -22.666, -21.950, -21.023, -19.890, -18.446, -17.518,
-16.779, -16.138, -22.820, -22.981, -22.699, -21.984, -21.056, -19.925, -18.473,
-17.538, -16.795, -16.151, -22.851, -23.013, -22.731, -22.017, -21.089, -19.960,
-18.500, -17.557, -16.810, -16.163, -22.883, -23.045, -22.763, -22.050, -21.122,
-19.995, -18.526, -17.576, -16.825, -16.176, -22.914, -23.077, -22.796, -22.083,
-21.154, -20.030, -18.553, -17.596, -16.840, -16.189, -22.945, -23.108, -22.828,
-22.116, -21.187, -20.064, -18.582, -17.617, -16.857, -16.203, -22.976, -23.140,
-22.860, -22.148, -21.219, -20.099, -18.612, -17.640, -16.875, -16.218, -23.007,
-23.172, -22.892, -22.181, -21.252, -20.133, -18.644, -17.664, -16.893, -16.233,
 -23.038, -23.204, -22.924, -22.214, -21.284, -20.168, -18.674, -17.686,
-16.910, -16.247, -23.069, -23.235, -22.956, -22.246, -21.316, -20.202, -18.703,
-17.707, -16.926, -16.260, -23.101, -23.267, -22.988, -22.279, -21.348, -20.236,
-18.731, -17.726, -16.941, -16.273, -23.132, -23.298, -23.020, -22.311, -21.380,
-20.270, -18.759, -17.746, -16.956, -16.285, -23.163, -23.330, -23.052, -22.344,
-21.412, -20.304, -18.788, -17.766, -16.972, -16.298, -23.194, -23.361, -23.084,
-22.376, -21.444, -20.338, -18.820, -17.789, -16.990, -16.313, -23.225, -23.393,
-23.115, -22.408, -21.476, -20.372, -18.852, -17.813, -17.008, -16.327, -23.256,
-23.424, -23.147, -22.440, -21.507, -20.405, -18.887, -17.838, -17.027, -16.343,
 -23.287, -23.456, -23.179, -22.473, -21.539, -20.439, -18.920, -17.862,
-17.045, -16.357, -23.318, -23.487, -23.210, -22.505, -21.571, -20.472, -18.952,
-17.884, -17.062, -16.371, -23.348, -23.518, -23.242, -22.537, -21.602, -20.506,
-18.983, -17.905, -17.078, -16.384, -23.379, -23.550, -23.274, -22.569, -21.634,
-20.539, -19.013, -17.925, -17.093, -16.396, -23.410, -23.581, -23.305, -22.600,
-21.665, -20.572, -19.043, -17.946, -17.109, -16.409, -23.441, -23.612, -23.337,
-22.632, -21.696, -20.606, -19.074, -17.969, -17.126, -16.423, -23.472, -23.643,
-23.368, -22.664, -21.728, -20.639, -19.105, -17.993, -17.144, -16.438, -23.503,
-23.675, -23.400, -22.696, -21.759, -20.672, -19.137, -18.017, -17.163, -16.452,
 -23.534, -23.706, -23.431, -22.727, -21.790, -20.705, -19.168, -18.040,
-17.180, -16.466, -23.564, -23.737, -23.462, -22.759, -21.821, -20.737, -19.200,
-18.061, -17.196, -16.479, -23.595, -23.768, -23.494, -22.791, -21.852, -20.770,
-19.232, -18.082, -17.212, -16.492, -23.626, -23.799, -23.525, -22.822, -21.883,
-20.803, -19.264, -18.104, -17.228, -16.505, -23.657, -23.830, -23.556, -22.853,
-21.914, -20.835, -19.296, -18.128, -17.246, -16.519, -23.687, -23.861, -23.587,
-22.885, -21.945, -20.868, -19.329, -18.153, -17.265, -16.534, -23.718, -23.892,
-23.619, -22.916, -21.976, -20.900, -19.361, -18.179, -17.284, -16.549, -23.749,
-23.923, -23.650, -22.948, -22.006, -20.933, -19.394, -18.205, -17.303, -16.564,
 -23.779, -23.954, -23.681, -22.979, -22.037, -20.965, -19.427, -18.230,
-17.321, -16.577, -23.810, -23.985, -23.712, -23.010, -22.068, -20.997, -19.460,
-18.253, -17.337, -16.591, -23.840, -24.016, -23.743, -23.041, -22.098, -21.030,
-19.492, -18.277, -17.354, -16.604, -23.871, -24.047, -23.774, -23.072, -22.129,
-21.062, -19.524, -18.300, -17.371, -16.617, -23.902, -24.078, -23.805, -23.103,
-22.159, -21.094, -19.556, -18.324, -17.389, -16.631, -23.932, -24.109, -23.836,
-23.135, -22.190, -21.126, -19.588, -18.350, -17.408, -16.645, -23.963, -24.140,
-23.867, -23.166, -22.220, -21.158, -19.619, -18.377, -17.427, -16.661, -23.993,
-24.171, -23.898, -23.196, -22.251, -21.190, -19.651, -18.404, -17.447, -16.676,
 -24.024, -24.202, -23.929, -23.227, -22.281, -21.222, -19.683, -18.430,
-17.465, -16.690, -24.054, -24.232, -23.960, -23.258, -22.311, -21.254, -19.715,
-18.455, -17.483, -16.703, -24.085, -24.263, -23.990, -23.289, -22.342, -21.285,
-19.748, -18.480, -17.499, -16.716, -24.115, -24.294, -24.021, -23.320, -22.372,
-21.317, -19.780, -18.505, -17.517, -16.729, -24.146, -24.325, -24.052, -23.351,
-22.402, -21.349, -19.813, -18.531, -17.535, -16.743, -24.176, -24.355, -24.083,
-23.382, -22.432, -21.380, -19.845, -18.559, -17.555, -16.758, -24.207, -24.386,
-24.113, -23.412, -22.462, -21.412, -19.877, -18.588, -17.576, -16.774, -24.237,
-24.417, -24.144, -23.443, -22.493, -21.443, -19.909, -18.617, -17.597, -16.789,
 -24.267, -24.447, -24.175, -23.474, -22.523, -21.475, -19.941, -18.645,
-17.616, -16.804, -24.298, -24.478, -24.205, -23.504, -22.553, -21.506, -19.972,
-18.673, -17.635, -16.818, -24.328, -24.508, -24.236, -23.535, -22.583, -21.538,
-20.004, -18.700, -17.653, -16.831, -24.359, -24.539, -24.267, -23.565, -22.613,
-21.569, -20.036, -18.726, -17.670, -16.844, -24.389, -24.570, -24.297, -23.596,
-22.643, -21.601, -20.068, -18.753, -17.690, -16.858, -24.419, -24.600, -24.328,
-23.626, -22.673, -21.632, -20.100, -18.781, -17.710, -16.873, -24.450, -24.631,
-24.358, -23.657, -22.702, -21.663, -20.132, -18.808, -17.731, -16.889, -24.480,
-24.661, -24.389, -23.687, -22.732, -21.694, -20.164, -18.836, -17.751, -16.904,
 -24.510, -24.692, -24.419, -23.718, -22.762, -21.726, -20.196, -18.864,
-17.770, -16.917, -24.540, -24.722, -24.450, -23.748, -22.792, -21.757, -20.227,
-18.892, -17.788, -16.931, -24.571, -24.753, -24.480, -23.779, -22.822, -21.788,
-20.259, -18.920, -17.807, -16.944, -24.601, -24.783, -24.511, -23.809, -22.851,
-21.819, -20.290, -18.948, -17.827, -16.959, -24.631, -24.814, -24.541, -23.839,
-22.881, -21.850, -20.322, -18.976, -17.848, -16.974, -24.661, -24.844, -24.572,
-23.870, -22.911, -21.881, -20.353, -19.005, -17.870, -16.990, -24.692, -24.874,
-24.602, -23.900, -22.941, -21.913, -20.385, -19.034, -17.892, -17.005, -24.722,
-24.905, -24.632, -23.930, -22.970, -21.944, -20.416, -19.063, -17.914, -17.020,
 -24.752, -24.935, -24.663, -23.960, -23.000, -21.975, -20.447, -19.092,
-17.935, -17.035, -24.782, -24.965, -24.693, -23.990, -23.029, -22.006, -20.479,
-19.121, -17.955, -17.048, -24.812, -24.996, -24.723, -24.021, -23.059, -22.037,
-20.510, -19.149, -17.975, -17.062, -24.842, -25.026, -24.754, -24.051, -23.089,
-22.068, -20.541, -19.178, -17.996, -17.077, -24.873, -25.056, -24.784, -24.081,
-23.118, -22.099, -20.572, -19.206, -18.018, -17.092, -24.903, -25.087, -24.814,
-24.111, -23.148, -22.130, -20.603, -19.234, -18.040, -17.108, -24.933, -25.117,
-24.844, -24.141, -23.177, -22.160, -20.634, -19.263, -18.064, -17.124, -24.963,
-25.147, -24.875, -24.171, -23.207, -22.191, -20.665, -19.291, -18.087, -17.139,
 -24.993, -25.177, -24.905, -24.201, -23.236, -22.222, -20.695, -19.320,
-18.109, -17.153, -25.023, -25.208, -24.935, -24.231, -23.266, -22.253, -20.726,
-19.349, -18.130, -17.167, -25.053, -25.238, -24.965, -24.261, -23.295, -22.284,
-20.757, -19.378, -18.151, -17.182, -25.083, -25.268, -24.995, -24.291, -23.324,
-22.315, -20.787, -19.407, -18.173, -17.196, -25.113, -25.298, -25.025, -24.321,
-23.354, -22.346, -20.818, -19.436, -18.197, -17.212, -25.143, -25.328, -25.056,
-24.351, -23.383, -22.377, -20.848, -19.465, -18.221, -17.228, -25.173, -25.358,
-25.086, -24.381, -23.413, -22.407, -20.879, -19.493, -18.246, -17.245, -25.203,
-25.389, -25.116, -24.411, -23.442, -22.438, -20.909, -19.522, -18.270, -17.261,
 -25.233, -25.419, -25.146, -24.441, -23.471, -22.469, -20.939, -19.551,
-18.294, -17.276, -25.263, -25.449, -25.176, -24.471, -23.501, -22.500, -20.969,
-19.579, -18.318, -17.290, -25.293, -25.479, -25.206, -24.501, -23.530, -22.531,
-20.999, -19.608, -18.341, -17.305, -25.323, -25.509, -25.236, -24.531, -23.559,
-22.561, -21.029, -19.637, -18.364, -17.320, -25.353, -25.539, -25.266, -24.560,
-23.588, -22.592, -21.059, -19.666, -18.388, -17.335, -25.383, -25.569, -25.296,
-24.590, -23.618, -22.623, -21.089, -19.695, -18.412, -17.351, -25.413, -25.599,
-25.326, -24.620, -23.647, -22.654, -21.119, -19.724, -18.436, -17.368, -25.443,
-25.629, -25.356, -24.650, -23.676, -22.685, -21.149, -19.752, -18.460, -17.383,
 -25.473, -25.659, -25.386, -24.680, -23.705, -22.715, -21.178, -19.781,
-18.484, -17.398, -25.503, -25.689, -25.416, -24.709, -23.735, -22.746, -21.208,
-19.810, -18.508, -17.413, -25.532, -25.719, -25.446, -24.739, -23.764, -22.777,
-21.238, -19.839, -18.532, -17.428, -25.562, -25.749, -25.476, -24.769, -23.793,
-22.807, -21.267, -19.867, -18.557, -17.444, -25.592, -25.779, -25.506, -24.798,
-23.822, -22.838, -21.296, -19.896, -18.581, -17.461, -25.622, -25.809, -25.535,
-24.828, -23.851, -22.869, -21.326, -19.925, -18.606, -17.478, -25.652, -25.839,
-25.565, -24.858, -23.880, -22.900, -21.355, -19.953, -18.632, -17.494, -25.682,
-25.869, -25.595, -24.888, -23.910, -22.930, -21.384, -19.982, -18.657, -17.510,
 -25.712, -25.899, -25.625, -24.917, -23.939, -22.961, -21.414, -20.010,
-18.682, -17.525, -25.741, -25.929, -25.655, -24.947, -23.968, -22.992, -21.443,
-20.039, -18.707, -17.541, -25.771, -25.959, -25.685, -24.976, -23.997, -23.023,
-21.472, -20.067, -18.731, -17.556, -25.801, -25.989, -25.715, -25.006, -24.026,
-23.053, -21.501, -20.096, -18.756, -17.572, -25.831, -26.018, -25.744, -25.036,
-24.055, -23.084, -21.530, -20.124, -18.781, -17.589, -25.861, -26.048, -25.774,
-25.065, -24.084, -23.115, -21.559, -20.152, -18.806, -17.607, -25.890, -26.078,
-25.804, -25.095, -24.113, -23.145, -21.588, -20.181, -18.831, -17.624, -25.920,
-26.108, -25.834, -25.125, -24.142, -23.176, -21.617, -20.209, -18.856, -17.640,
 -25.950, -26.138, -25.863, -25.154, -24.171, -23.207, -21.645, -20.237,
-18.881, -17.656, -25.980, -26.168, -25.893, -25.184, -24.200, -23.237, -21.674,
-20.265, -18.907, -17.672, -26.009, -26.197, -25.923, -25.213, -24.229, -23.268,
-21.703, -20.293, -18.932, -17.688, -26.039, -26.227, -25.953, -25.243, -24.258,
-23.299, -21.731, -20.321, -18.957, -17.705, -26.069, -26.257, -25.982, -25.272,
-24.287, -23.330, -21.760, -20.349, -18.983, -17.722, -26.098, -26.287, -26.012,
-25.302, -24.316, -23.360, -21.789, -20.377, -19.008, -17.740, -26.128, -26.317,
-26.042, -25.331, -24.345, -23.391, -21.817, -20.405, -19.033, -17.758, -26.158,
-26.346, -26.072, -25.361, -24.374, -23.422, -21.846, -20.433, -19.058, -17.775,
 -26.187, -26.376, -26.101, -25.390, -24.403, -23.452, -21.874, -20.461,
-19.084, -17.792, -26.217, -26.406, -26.131, -25.420, -24.432, -23.483, -21.902,
-20.489, -19.109, -17.808, -26.247, -26.436, -26.161, -25.449, -24.461, -23.514,
-21.931, -20.516, -19.135, -17.825, -26.276, -26.465, -26.190, -25.479, -24.490,
-23.544, -21.959, -20.544, -19.160, -17.842, -26.306, -26.495, -26.220, -25.508,
-24.519, -23.575, -21.988, -20.571, -19.186, -17.860, -26.336, -26.525, -26.250,
-25.537, -24.548, -23.606, -22.016, -20.599, -19.211, -17.879, -26.365, -26.554,
-26.279, -25.567, -24.577, -23.637, -22.044, -20.626, -19.237, -17.896, -26.395,
-26.584, -26.309, -25.596, -24.606, -23.667, -22.072, -20.654, -19.262, -17.913,
 -26.424, -26.614, -26.338, -25.626, -24.635, -23.698, -22.100, -20.681,
-19.288, -17.930, -26.454, -26.643, -26.368, -25.655, -24.664, -23.729, -22.129,
-20.708, -19.313, -17.947, -26.484, -26.673, -26.398, -25.685, -24.693, -23.759,
-22.157, -20.736, -19.339, -17.965, -26.513, -26.703, -26.427, -25.714, -24.721,
-23.790, -22.185, -20.763, -19.364, -17.984, -26.543, -26.732, -26.457, -25.743,
-24.750, -23.821, -22.213, -20.790, -19.390, -18.003, -26.572, -26.762, -26.486,
-25.773, -24.779, -23.852, -22.241, -20.817, -19.416, -18.022, -26.602, -26.791,
-26.516, -25.802, -24.808, -23.882, -22.269, -20.844, -19.441, -18.041, -26.631,
-26.821, -26.545, -25.831, -24.837, -23.913, -22.297, -20.871, -19.467, -18.059,
 -26.661, -26.851, -26.575, -25.861, -24.866, -23.944, -22.325, -20.898,
-19.492, -18.077, -26.690, -26.880, -26.604, -25.890, -24.895, -23.974, -22.353,
-20.925, -19.517, -18.096, -26.720, -26.910, -26.634, -25.919, -24.923, -24.005,
-22.381, -20.952, -19.543, -18.114, -26.749, -26.939, -26.663, -25.949, -24.952,
-24.036, -22.408, -20.979, -19.568, -18.133, -26.779, -26.969, -26.693, -25.978,
-24.981, -24.067, -22.436, -21.005, -19.594, -18.153, -26.808, -26.998, -26.722,
-26.007, -25.010, -24.097, -22.464, -21.032, -19.619, -18.173, -26.838, -27.028,
-26.752, -26.037, -25.039, -24.128, -22.492, -21.059, -19.645, -18.193, -26.867,
-27.058, -26.781, -26.066, -25.068, -24.159, -22.520, -21.085, -19.670, -18.212,
 -26.897, -27.087, -26.811, -26.095, -25.096, -24.189, -22.548, -21.112,
-19.695, -18.230, -26.926, -27.117, -26.840, -26.124, -25.125, -24.220, -22.575,
-21.138, -19.721, -18.249, -26.956, -27.146, -26.870, -26.154, -25.154, -24.251,
-22.603, -21.165, -19.746, -18.269, -26.985, -27.176, -26.899, -26.183, -25.183,
-24.282, -22.631, -21.191, -19.771, -18.289, -27.015, -27.205, -26.929, -26.212,
-25.212, -24.312, -22.659, -21.218, -19.796, -18.310, -27.044, -27.235, -26.958,
-26.242, -25.240, -24.343, -22.686, -21.244, -19.821, -18.331, -27.074, -27.264,
-26.988, -26.271, -25.269, -24.374, -22.714, -21.270, -19.847, -18.352, -27.103,
-27.294, -27.017, -26.300, -25.298, -24.405, -22.742, -21.296, -19.872, -18.372,
 -27.132, -27.323, -27.046, -26.329, -25.327, -24.435, -22.769, -21.323,
-19.897, -18.392, -27.162, -27.352, -27.076, -26.358, -25.356, -24.466, -22.797,
-21.349, -19.922, -18.412, -27.191, -27.382, -27.105, -26.388, -25.384, -24.497,
-22.824, -21.375, -19.947, -18.432, -27.221, -27.411, -27.135, -26.417, -25.413,
-24.528, -22.852, -21.401, -19.972, -18.453, -27.250, -27.441, -27.164, -26.446,
-25.442, -24.558, -22.880, -21.427, -19.997, -18.473, -27.279, -27.470, -27.193,
-26.475, -25.471, -24.589, -22.907, -21.453, -20.022, -18.494, -27.309, -27.500,
-27.223, -26.505, -25.499, -24.620, -22.935, -21.479, -20.047, -18.515, -27.338,
-27.529, -27.252, -26.534, -25.528, -24.651, -22.962, -21.505, -20.071, -18.535,
 -27.367, -27.558, -27.282, -26.563, -25.557, -24.682, -22.990, -21.531,
-20.096, -18.556, -27.397, -27.588, -27.311, -26.592, -25.586, -24.712, -23.018,
-21.557, -20.121, -18.577, -27.426, -27.617, -27.340, -26.621, -25.614, -24.743,
-23.045, -21.583, -20.146, -18.598, -27.455, -27.647, -27.370, -26.650, -25.643,
-24.774, -23.073, -21.609, -20.170, -18.619, -27.485, -27.676, -27.399, -26.680,
-25.672, -24.805, -23.100, -21.634, -20.195, -18.640, -27.514, -27.705, -27.428,
-26.709, -25.701, -24.835, -23.128, -21.660, -20.220, -18.662, -27.543, -27.735,
-27.458, -26.738, -25.729, -24.866, -23.155, -21.686, -20.244, -18.683, -27.573,
-27.764, -27.487, -26.767, -25.758, -24.897, -23.183, -21.712, -20.269, -18.705,
 -27.602, -27.793, -27.516, -26.796, -25.787, -24.928, -23.210, -21.737,
-20.293, -18.726, -27.631, -27.823, -27.546, -26.825, -25.815, -24.959, -23.238,
-21.763, -20.318, -18.747, -27.660, -27.852, -27.575, -26.855, -25.844, -24.989,
-23.265, -21.788, -20.342, -18.768, -27.690, -27.881, -27.604, -26.884, -25.873,
-25.020, -23.293, -21.814, -20.367, -18.790, -27.719, -27.911, -27.634, -26.913,
-25.902, -25.051, -23.320, -21.840, -20.391, -18.811, -27.748, -27.940, -27.663,
-26.942, -25.930, -25.082, -23.347, -21.865, -20.415, -18.832, -27.778, -27.969,
-27.692, -26.971, -25.959, -25.113, -23.375, -21.891, -20.440, -18.854, -27.807,
-27.999, -27.721, -27.000, -25.988, -25.143, -23.402, -21.916, -20.464, -18.875,
 -27.836, -28.028, -27.751, -27.029, -26.016, -25.174, -23.430, -21.942,
-20.488, -18.897, -27.865, -28.057, -27.780, -27.058, -26.045, -25.205, -23.457,
-21.967, -20.512, -18.919, -27.895, -28.087, -27.809, -27.087, -26.074, -25.236,
-23.485, -21.993, -20.536, -18.940, -27.924, -28.116, -27.839, -27.117, -26.102,
-25.267, -23.512, -22.018, -20.561, -18.962, -27.953, -28.145, -27.868, -27.146,
-26.131, -25.298, -23.539, -22.043, -20.585, -18.984, -27.982, -28.174, -27.897,
-27.175, -26.160, -25.328, -23.567, -22.069, -20.609, -19.005, -28.011, -28.204,
-27.926, -27.204, -26.188, -25.359, -23.594, -22.094, -20.633, -19.027, -28.041,
-28.233, -27.956, -27.233, -26.217, -25.390, -23.622, -22.119, -20.657, -19.049,
 -28.070, -28.262, -27.985, -27.262, -26.246, -25.421, -23.649, -22.145,
-20.681, -19.071, -28.099, -28.291, -28.014, -27.291, -26.274, -25.452, -23.676,
-22.170, -20.705, -19.093, -28.128, -28.321, -28.043, -27.320, -26.303, -25.483,
-23.704, -22.195, -20.729, -19.114, -28.157, -28.350, -28.072, -27.349, -26.332,
-25.513, -23.731, -22.221, -20.752, -19.136, -28.187, -28.379, -28.102, -27.378,
-26.360, -25.544, -23.759, -22.246, -20.776, -19.158, -28.216, -28.408, -28.131,
-27.407, -26.389, -25.575, -23.786, -22.271, -20.800, -19.180, -28.245, -28.438,
-28.160, -27.436, -26.418, -25.606, -23.813, -22.296, -20.824, -19.202, -28.274,
-28.467, -28.189, -27.465, -26.446, -25.637, -23.841, -22.321, -20.848, -19.224,
 -28.303, -28.496, -28.218, -27.494, -26.475, -25.668, -23.868, -22.347,
-20.871, -19.246, -28.332, -28.525, -28.248, -27.524, -26.504, -25.698, -23.896,
-22.372, -20.895, -19.268, -28.361, -28.554, -28.277, -27.553, -26.532, -25.729,
-23.923, -22.397, -20.919, -19.290, -28.391, -28.584, -28.306, -27.582, -26.561,
-25.760, -23.950, -22.422, -20.942, -19.312, -28.420, -28.613, -28.335, -27.611,
-26.590, -25.791, -23.978, -22.447, -20.966, -19.334, -28.449, -28.642, -28.364,
-27.640, -26.618, -25.822, -24.005, -22.472, -20.990, -19.356, -28.478, -28.671,
-28.394, -27.669, -26.647, -25.853, -24.032, -22.497, -21.013, -19.377, -28.507,
-28.700, -28.423, -27.698, -26.675, -25.884, -24.060, -22.523, -21.037, -19.399,
 -28.536, -28.729, -28.452, -27.727, -26.704, -25.915, -24.087, -22.548,
-21.060, -19.421, -28.565, -28.759, -28.481, -27.756, -26.733, -25.945, -24.114,
-22.573, -21.084, -19.443, -28.594, -28.788, -28.510, -27.785, -26.761, -25.976,
-24.142, -22.598, -21.107, -19.465, -28.623, -28.817, -28.539, -27.814, -26.790,
-26.007, -24.169, -22.623, -21.131, -19.487, -28.652, -28.846, -28.569, -27.843,
-26.819, -26.038, -24.197, -22.648, -21.154, -19.509, -28.682, -28.875, -28.598,
-27.872, -26.847, -26.069, -24.224, -22.673, -21.177, -19.531, -28.711, -28.904,
-28.627, -27.901, -26.876, -26.100, -24.251, -22.698, -21.201, -19.553, -28.740,
-28.933, -28.656, -27.930, -26.904, -26.131, -24.279, -22.723, -21.224, -19.575,
 -28.769, -28.963, -28.685, -27.959, -26.933, -26.162, -24.306, -22.748,
-21.248, -19.597, -28.798, -28.992, -28.714, -27.988, -26.962, -26.192, -24.333,
-22.773, -21.271, -19.619, -28.827, -29.021, -28.743, -28.017, -26.990, -26.223,
-24.361, -22.798, -21.294, -19.640, -28.856, -29.050, -28.772, -28.046, -27.019,
-26.254, -24.388, -22.823, -21.318, -19.662, -28.885, -29.079, -28.802, -28.075,
-27.047, -26.285, -24.415, -22.848, -21.341, -19.684, -28.914, -29.108, -28.831,
-28.104, -27.076, -26.316, -24.443, -22.873, -21.364, -19.706, -28.943, -29.137,
-28.860, -28.133, -27.105, -26.347, -24.470, -22.898, -21.387, -19.728, -28.972,
-29.166, -28.889, -28.162, -27.133, -26.378, -24.497, -22.923, -21.411, -19.749,
 -29.001, -29.195, -28.918, -28.191, -27.162, -26.409, -24.525, -22.948,
-21.434, -19.771, -29.030, -29.224, -28.947, -28.220, -27.190, -26.440, -24.552,
-22.973, -21.457, -19.793, -29.059, -29.254, -28.976, -28.249, -27.219, -26.471,
-24.580, -22.997, -21.480, -19.815, -29.088, -29.283, -29.005, -28.277, -27.248,
-26.502, -24.607, -23.022, -21.503, -19.836, -29.117, -29.312, -29.034, -28.306,
-27.276, -26.532, -24.634, -23.047, -21.526, -19.858, -29.146, -29.341, -29.064,
-28.335, -27.305, -26.563, -24.662, -23.072, -21.550, -19.880, -29.175, -29.370,
-29.093, -28.364, -27.333, -26.594, -24.689, -23.097, -21.573, -19.901, -29.204,
-29.399, -29.122, -28.393, -27.362, -26.625, -24.716, -23.122, -21.596, -19.923,
 -29.233, -29.428, -29.151, -28.422, -27.391, -26.656, -24.744, -23.147,
-21.619, -19.945, -29.262, -29.457, -29.180, -28.451, -27.419, -26.687, -24.771,
-23.172, -21.642, -19.966, -29.291, -29.486, -29.209, -28.480, -27.448, -26.718,
-24.798, -23.197, -21.665, -19.988, -29.320, -29.515, -29.238, -28.509, -27.476,
-26.749, -24.826, -23.222, -21.688, -20.009, -29.349, -29.544, -29.267, -28.538,
-27.505, -26.780, -24.853, -23.246, -21.711, -20.031, -29.378, -29.573, -29.296,
-28.567, -27.533, -26.811, -24.880, -23.271, -21.734, -20.052, -29.407, -29.602,
-29.325, -28.596, -27.562, -26.842, -24.908, -23.296, -21.757, -20.074, -29.436,
-29.631, -29.354, -28.625, -27.591, -26.873, -24.935, -23.321, -21.780, -20.095,
 -29.465, -29.660, -29.383, -28.654, -27.619, -26.904, -24.963, -23.346,
-21.803, -20.117, -29.493, -29.689, -29.412, -28.683, -27.648, -26.935, -24.990,
-23.371, -21.826, -20.138, -29.522, -29.718, -29.441, -28.712, -27.676, -26.965,
-25.017, -23.396, -21.849, -20.159, -29.551, -29.747, -29.470, -28.741, -27.705,
-26.996, -25.045, -23.421, -21.872, -20.181, -29.580, -29.776, -29.499, -28.769,
-27.733, -27.027, -25.072, -23.445, -21.895, -20.202, -29.609, -29.805, -29.528,
-28.798, -27.762, -27.058, -25.099, -23.470, -21.918, -20.223, -29.638, -29.834,
-29.558, -28.827, -27.791, -27.089, -25.127, -23.495, -21.941, -20.245, -29.667,
-29.863, -29.587, -28.856, -27.819, -27.120, -25.154, -23.520, -21.964, -20.266,
 -29.696, -29.892, -29.616, -28.885, -27.848, -27.151, -25.182, -23.545,
-21.987, -20.287, -29.725, -29.921, -29.645, -28.914, -27.876, -27.182, -25.209,
-23.570, -22.010, -20.309, -29.754, -29.950, -29.674, -28.943, -27.905, -27.213,
-25.236, -23.594, -22.033, -20.330, -29.783, -29.979, -29.703, -28.972, -27.933,
-27.244, -25.264, -23.619, -22.056, -20.351, -29.811, -30.008, -29.732, -29.001,
-27.962, -27.275, -25.291, -23.644, -22.078, -20.372, -29.840, -30.037, -29.761,
-29.030, -27.990, -27.306, -25.318, -23.669, -22.101, -20.393, -29.869, -30.066,
-29.790, -29.059, -28.019, -27.337, -25.346, -23.694, -22.124, -20.415, -29.898,
-30.095, -29.819, -29.087, -28.048, -27.368, -25.373, -23.719, -22.147, -20.436,
 -29.927, -30.124, -29.848, -29.116, -28.076, -27.399, -25.401, -23.744,
-22.170, -20.457, -29.956, -30.153, -29.877, -29.145, -28.105, -27.430, -25.428,
-23.768, -22.193, -20.478, -29.985, -30.182, -29.906, -29.174, -28.133, -27.461,
-25.455, -23.793, -22.216, -20.499, -30.013, -30.211, -29.935, -29.203, -28.162,
-27.492, -25.483, -23.818, -22.239, -20.520, -30.042, -30.240, -29.964, -29.232,
-28.190, -27.523, -25.510, -23.843, -22.261, -20.541, -30.071, -30.269, -29.993,
-29.261, -28.219, -27.554, -25.537, -23.868, -22.284, -20.562, -30.100, -30.298,
-30.022, -29.290, -28.247, -27.585, -25.565, -23.892, -22.307, -20.583, -30.129,
-30.327, -30.051, -29.319, -28.276, -27.616, -25.592, -23.917, -22.330, -20.604,
 -30.158, -30.355, -30.080, -29.347, -28.304, -27.647, -25.620, -23.942,
-22.353, -20.625, -30.186, -30.384, -30.109, -29.376, -28.333, -27.678, -25.647,
-23.967, -22.376, -20.646, -30.215, -30.413, -30.138, -29.405, -28.362, -27.709,
-25.674, -23.992, -22.399, -20.667, -30.244, -30.442, -30.167, -29.434, -28.390,
-27.740, -25.702, -24.017, -22.421, -20.688, -30.273, -30.471, -30.196, -29.463,
-28.419, -27.771, -25.729, -24.041, -22.444, -20.709, -30.302, -30.500, -30.225,
-29.492, -28.447, -27.802, -25.757, -24.066, -22.467, -20.730, -30.331, -30.529,
-30.254, -29.521, -28.476, -27.833, -25.784, -24.091, -22.490, -20.751, -30.359,
-30.558, -30.283, -29.550, -28.504, -27.863, -25.811, -24.116, -22.513, -20.772,
 -30.388, -30.587, -30.312, -29.578, -28.533, -27.894, -25.839, -24.141,
-22.536, -20.792, -30.417, -30.616, -30.341, -29.607, -28.561, -27.925, -25.866,
-24.166, -22.558, -20.813, -30.446, -30.645, -30.369, -29.636, -28.590, -27.956,
-25.894, -24.190, -22.581, -20.834, -30.474, -30.674, -30.398, -29.665, -28.618,
-27.987, -25.921, -24.215, -22.604, -20.855, -30.503, -30.702, -30.427, -29.694,
-28.647, -28.018, -25.948, -24.240, -22.627, -20.876, -30.532, -30.731, -30.456,
-29.723, -28.675, -28.049, -25.976, -24.265, -22.650, -20.897, -30.561, -30.760,
-30.485, -29.752, -28.704, -28.080, -26.003, -24.290, -22.672, -20.917, -30.590,
-30.789, -30.514, -29.780, -28.732, -28.111, -26.031, -24.315, -22.695, -20.938,
 -30.618, -30.818, -30.543, -29.809, -28.761, -28.142, -26.058, -24.339,
-22.718, -20.959, -30.647, -30.847, -30.572, -29.838, -28.789, -28.173, -26.086,
-24.364, -22.741, -20.980, -30.676, -30.876, -30.601, -29.867, -28.818, -28.204,
-26.113, -24.389, -22.764, -21.001, -30.705, -30.905, -30.630, -29.896, -28.846,
-28.235, -26.140, -24.414, -22.786, -21.021, -30.733, -30.933, -30.659, -29.925,
-28.875, -28.267, -26.168, -24.439, -22.809, -21.042, -30.762, -30.962, -30.688,
-29.954, -28.904, -28.298, -26.195, -24.464, -22.832, -21.063, -30.791, -30.991,
-30.717, -29.982, -28.932, -28.329, -26.223, -24.488, -22.855, -21.083, -30.820,
-31.020, -30.746, -30.011, -28.961, -28.360, -26.250, -24.513, -22.878, -21.104,
 -30.848, -31.049, -30.775, -30.040, -28.989, -28.391, -26.277, -24.538,
-22.900, -21.125, -30.877, -31.078, -30.804, -30.069, -29.018, -28.422, -26.305,
-24.563, -22.923, -21.146, -30.906, -31.107, -30.833, -30.098, -29.046, -28.453,
-26.332, -24.588, -22.946, -21.166, -30.934, -31.136, -30.862, -30.127, -29.075,
-28.484, -26.360, -24.613, -22.969, -21.187, -30.963, -31.164, -30.891, -30.156,
-29.103, -28.515, -26.387, -24.637, -22.992, -21.208, -30.992, -31.193, -30.919,
-30.184, -29.132, -28.546, -26.415, -24.662, -23.014, -21.228, -31.021, -31.222,
-30.948, -30.213, -29.160, -28.577, -26.442, -24.687, -23.037, -21.249, -31.049,
-31.251, -30.977, -30.242, -29.189, -28.608, -26.469, -24.712, -23.060, -21.270,
 -31.078, -31.280, -31.006, -30.271, -29.217, -28.639, -26.497, -24.737,
-23.083, -21.290, -31.107, -31.309, -31.035, -30.300, -29.246, -28.670, -26.524,
-24.762, -23.106, -21.311, -31.135, -31.337, -31.064, -30.329, -29.274, -28.701,
-26.552, -24.786, -23.128, -21.332, -31.164, -31.366, -31.093, -30.357, -29.303,
-28.732, -26.579, -24.811, -23.151, -21.352, -31.193, -31.395, -31.122, -30.386,
-29.331, -28.763, -26.607, -24.836, -23.174, -21.373, -31.222, -31.424, -31.151,
-30.415, -29.360, -28.794, -26.634, -24.861, -23.197, -21.394, -31.250, -31.453,
-31.180, -30.444, -29.388, -28.825, -26.661, -24.886, -23.220, -21.414, -31.279,
-31.482, -31.209, -30.473, -29.417, -28.856, -26.689, -24.911, -23.242, -21.435,
 -31.308, -31.510, -31.238, -30.501, -29.445, -28.887, -26.716, -24.936,
-23.265, -21.456, -31.336, -31.539, -31.266, -30.530, -29.474, -28.918, -26.744,
-24.960, -23.288, -21.476, -31.365, -31.568, -31.295, -30.559, -29.502, -28.949,
-26.771, -24.985, -23.311, -21.497, -31.394, -31.597, -31.324, -30.588, -29.531,
-28.980, -26.799, -25.010, -23.334, -21.518, -31.422, -31.626, -31.353, -30.617,
-29.559, -29.011, -26.826, -25.035, -23.356, -21.538, -31.451, -31.654, -31.382,
-30.646, -29.588, -29.042, -26.854, -25.060, -23.379, -21.559, -31.480, -31.683,
-31.411, -30.674, -29.616, -29.073, -26.881, -25.085, -23.402, -21.579, -31.508,
-31.712, -31.440, -30.703, -29.645, -29.104, -26.908, -25.110, -23.425, -21.600,
 -31.537, -31.741, -31.469, -30.732, -29.673, -29.135, -26.936, -25.134,
-23.448, -21.621, -31.566, -31.770, -31.498, -30.761, -29.702, -29.166, -26.963,
-25.159, -23.470, -21.641, -31.594, -31.798, -31.527, -30.790, -29.730, -29.197,
-26.991, -25.184, -23.493, -21.662, -31.623, -31.827, -31.556, -30.818, -29.759,
-29.228, -27.018, -25.209, -23.516, -21.682, -31.651, -31.856, -31.584, -30.847,
-29.787, -29.259, -27.046, -25.234, -23.539, -21.703, -31.680, -31.885, -31.613,
-30.876, -29.816, -29.291, -27.073, -25.259, -23.562, -21.724, -31.709, -31.914,
-31.642, -30.905, -29.844, -29.322, -27.101, -25.284, -23.585, -21.744, -31.737,
-31.942, -31.671, -30.934, -29.873, -29.353, -27.128, -25.308, -23.607, -21.765,
 -31.766, -31.971, -31.700, -30.963, -29.901, -29.384, -27.156, -25.333,
-23.630, -21.786, -31.795, -32.000, -31.729, -30.991, -29.930, -29.415, -27.183,
-25.358, -23.653, -21.806, -31.823, -32.029, -31.758, -31.020, -29.958, -29.446,
-27.211, -25.383, -23.676, -21.827, -31.852, -32.058, -31.787, -31.049, -29.987,
-29.477, -27.238, -25.408, -23.699, -21.847, -31.880, -32.086, -31.816, -31.078,
-30.015, -29.508, -27.265, -25.433, -23.722, -21.868, -31.909, -32.115, -31.844,
-31.107, -30.044, -29.539, -27.293, -25.458, -23.744, -21.889, -31.938, -32.144,
-31.873, -31.135, -30.072, -29.570, -27.320, -25.483, -23.767, -21.909, -31.966,
-32.173, -31.902, -31.164, -30.101, -29.601, -27.348, -25.508, -23.790, -21.930,
 -31.995, -32.201, -31.931, -31.193, -30.129, -29.632, -27.375, -25.532,
-23.813, -21.951, -32.023, -32.230, -31.960, -31.222, -30.158, -29.663, -27.403,
-25.557, -23.836, -21.971, -32.052, -32.259, -31.989, -31.251, -30.186, -29.694,
-27.430, -25.582, -23.859, -21.992, -32.081, -32.288, -32.018, -31.279, -30.215,
-29.725, -27.458, -25.607, -23.881, -22.013, -32.109, -32.316, -32.047, -31.308,
-30.243, -29.756, -27.485, -25.632, -23.904, -22.033, -32.138, -32.345, -32.075,
-31.337, -30.272, -29.788, -27.513, -25.657, -23.927, -22.054, -32.166, -32.374,
-32.104, -31.366, -30.300, -29.819, -27.540, -25.682, -23.950, -22.074, -32.195,
-32.403, -32.133, -31.395, -30.329, -29.850, -27.568, -25.707, -23.973, -22.095,
 -32.224, -32.431, -32.162, -31.423, -30.357, -29.881, -27.595, -25.732,
-23.996, -22.116, -32.252, -32.460, -32.191, -31.452, -30.385, -29.912, -27.623,
-25.756, -24.019, -22.136, -32.281, -32.489, -32.220, -31.481, -30.414, -29.943,
-27.650, -25.781, -24.041, -22.157, -32.309, -32.518, -32.249, -31.510, -30.442,
-29.974, -27.678, -25.806, -24.064, -22.178, -32.338, -32.546, -32.277, -31.538,
-30.471, -30.005, -27.705, -25.831, -24.087, -22.198, -32.366, -32.575, -32.306,
-31.567, -30.499, -30.036, -27.733, -25.856, -24.110, -22.219, -32.395, -32.604,
-32.335, -31.596, -30.528, -30.067, -27.760, -25.881, -24.133, -22.240, -32.424,
-32.633, -32.364, -31.625, -30.556, -30.098, -27.788, -25.906, -24.156, -22.260,
 -32.452, -32.661, -32.393, -31.654, -30.585, -30.129, -27.815, -25.931,
-24.179, -22.281, -32.481, -32.690, -32.422, -31.682, -30.613, -30.160, -27.843,
-25.956, -24.202, -22.302, -32.509, -32.719, -32.451, -31.711, -30.642, -30.192,
-27.870, -25.981, -24.224, -22.322, -32.538, -32.748, -32.479, -31.740, -30.670,
-30.223, -27.898, -26.005, -24.247, -22.343, -32.566, -32.776, -32.508, -31.769,
-30.699, -30.254, -27.925, -26.030, -24.270, -22.364, -32.595, -32.805, -32.537,
-31.798, -30.727, -30.285, -27.953, -26.055, -24.293, -22.384, -32.623, -32.834,
-32.566, -31.826, -30.756, -30.316, -27.980, -26.080, -24.316, -22.405, -32.652,
-32.863, -32.595, -31.855, -30.784, -30.347, -28.008, -26.105, -24.339, -22.426,
 -32.680, -32.891, -32.624, -31.884, -30.813, -30.378, -28.035, -26.130,
-24.362, -22.447, -32.709, -32.920, -32.653, -31.913, -30.841, -30.409, -28.063,
-26.155, -24.385, -22.467, -32.737, -32.949, -32.681, -31.941, -30.870, -30.440,
-28.090, -26.180, -24.408, -22.488, -32.766, -32.977, -32.710, -31.970, -30.898,
-30.471, -28.118, -26.205, -24.430, -22.509, -32.795, -33.006, -32.739, -31.999,
-30.927, -30.502, -28.145, -26.230, -24.453, -22.529, -32.823, -33.035, -32.768,
-32.028, -30.955, -30.533, -28.173, -26.255, -24.476, -22.550, -32.852, -33.064,
-32.797, -32.057, -30.984, -30.565, -28.200, -26.280, -24.499, -22.571, -32.880,
-33.092, -32.826, -32.085, -31.012, -30.596, -28.228, -26.305, -24.522, -22.592,
 -32.909, -33.121, -32.855, -32.114, -31.040, -30.627, -28.255, -26.329,
-24.545, -22.612, -32.937, -33.150, -32.883, -32.143, -31.069, -30.658, -28.283,
-26.354, -24.568, -22.633, -32.966, -33.178, -32.912, -32.172, -31.097, -30.689,
-28.310, -26.379, -24.591, -22.654, -32.994, -33.207, -32.941, -32.200, -31.126,
-30.720, -28.338, -26.404, -24.614, -22.675, -33.023, -33.236, -32.970, -32.229,
-31.154, -30.751, -28.365, -26.429, -24.637, -22.695, -33.051, -33.265, -32.999,
-32.258, -31.183, -30.782, -28.393, -26.454, -24.660, -22.716, -33.080, -33.293,
-33.028, -32.287, -31.211, -30.813, -28.420, -26.479, -24.683, -22.737, -33.108,
-33.322, -33.056, -32.316, -31.240, -30.844, -28.448, -26.504, -24.706, -22.758,
 -33.137, -33.351, -33.085, -32.344, -31.268, -30.876, -28.475, -26.529,
-24.729, -22.778, -33.165, -33.379, -33.114, -32.373, -31.297, -30.907, -28.503,
-26.554, -24.751, -22.799, -33.194, -33.408, -33.143, -32.402, -31.325, -30.938,
-28.530, -26.579, -24.774, -22.820, -33.222, -33.437, -33.172, -32.431, -31.354,
-30.969, -28.558, -26.604, -24.797, -22.841, -33.251, -33.466, -33.201, -32.459,
-31.382, -31.000, -28.585, -26.629, -24.820, -22.862, -33.279, -33.494, -33.229,
-32.488, -31.411, -31.031, -28.613, -26.654, -24.843, -22.882, -33.308, -33.523,
-33.258, -32.517, -31.439, -31.062, -28.640, -26.679, -24.866, -22.903, -33.336,
-33.552, -33.287, -32.546, -31.468, -31.093, -28.668, -26.704, -24.889, -22.924,
 -33.364, -33.580, -33.316, -32.574, -31.496, -31.124, -28.695, -26.729,
-24.912, -22.945, -33.393, -33.609, -33.345, -32.603, -31.525, -31.156, -28.723,
-26.754, -24.935, -22.966, -33.421, -33.638, -33.374, -32.632, -31.553, -31.187,
-28.750, -26.779, -24.958, -22.986, -33.450, -33.666, -33.402, -32.661, -31.581,
-31.218, -28.778, -26.803, -24.981, -23.007, -33.478, -33.695, -33.431, -32.690,
-31.610, -31.249, -28.805, -26.828, -25.004, -23.028, -33.507, -33.724, -33.460,
-32.718, -31.638, -31.280, -28.833, -26.853, -25.027, -23.049, -33.535, -33.752,
-33.489, -32.747, -31.667, -31.311, -28.860, -26.878, -25.050, -23.070, -33.564,
-33.781, -33.518, -32.776, -31.695, -31.342, -28.888, -26.903, -25.073, -23.091,
 -33.592, -33.810, -33.547, -32.805, -31.724, -31.373, -28.915, -26.928,
-25.096, -23.112, -33.621, -33.838, -33.575, -32.833, -31.752, -31.404, -28.943,
-26.953, -25.119, -23.132, -33.649, -33.867, -33.604, -32.862, -31.781, -31.436,
-28.970, -26.978, -25.142, -23.153, -33.678, -33.896, -33.633, -32.891, -31.809,
-31.467, -28.998, -27.003, -25.165, -23.174, -33.706, -33.925, -33.662, -32.920,
-31.838, -31.498, -29.026, -27.028, -25.188, -23.195, -33.734, -33.953, -33.691,
-32.948, -31.866, -31.529, -29.053, -27.053, -25.211, -23.216, -33.763, -33.982,
-33.719, -32.977, -31.895, -31.560, -29.081, -27.078, -25.234, -23.237, -33.791,
-34.011, -33.748, -33.006, -31.923, -31.591, -29.108, -27.103, -25.257, -23.258,
 -33.820, -34.039, -33.777, -33.035, -31.952, -31.622, -29.136, -27.128,
-25.280, -23.279, -33.848, -34.068, -33.806, -33.063, -31.980, -31.653, -29.163,
-27.153, -25.303, -23.299, -33.877, -34.097, -33.835, -33.092, -32.008, -31.685,
-29.191, -27.178, -25.326, -23.320, -33.905, -34.125, -33.864, -33.121, -32.037,
-31.716, -29.218, -27.203, -25.349, -23.341, -33.934, -34.154, -33.892, -33.150,
-32.065, -31.747, -29.246, -27.228, -25.372, -23.362, -33.962, -34.183, -33.921,
-33.178, -32.094, -31.778, -29.273, -27.253, -25.395, -23.383, -33.990, -34.211,
-33.950, -33.207, -32.122, -31.809, -29.301, -27.278, -25.418, -23.404, -34.019,
-34.240, -33.979, -33.236, -32.151, -31.840, -29.328, -27.303, -25.441, -23.425,
 -34.047, -34.269, -34.008, -33.265, -32.179, -31.871, -29.356, -27.328,
-25.464, -23.446, -34.076, -34.297, -34.036, -33.294, -32.208, -31.902, -29.384,
-27.353, -25.487, -23.467, -34.104, -34.326, -34.065, -33.322, -32.236, -31.934,
-29.411, -27.378, -25.510, -23.488, -34.132, -34.355, -34.094, -33.351, -32.265,
-31.965, -29.439, -27.403, -25.533, -23.509, -34.161, -34.383, -34.123, -33.380,
-32.293, -31.996, -29.466, -27.428, -25.556, -23.530, -34.189, -34.412, -34.152,
-33.409, -32.322, -32.027, -29.494, -27.453, -25.579, -23.551, -34.218, -34.441,
-34.181, -33.437, -32.350, -32.058, -29.521, -27.478, -25.602, -23.572, -34.246,
-34.469, -34.209, -33.466, -32.378, -32.089, -29.549, -27.503, -25.625, -23.593,
 -34.275, -34.498, -34.238, -33.495, -32.407, -32.120, -29.576, -27.528,
-25.648, -23.614, -34.303, -34.527, -34.267, -33.524, -32.435, -32.152, -29.604,
-27.553, -25.671, -23.635, -34.331, -34.555, -34.296, -33.552, -32.464, -32.183,
-29.631, -27.578, -25.694, -23.656, -34.360, -34.584, -34.325, -33.581, -32.492,
-32.214, -29.659, -27.603, -25.717, -23.677, -34.388, -34.612, -34.353, -33.610,
-32.521, -32.245, -29.687, -27.628, -25.740, -23.698, -34.417, -34.641, -34.382,
-33.639, -32.549, -32.276, -29.714, -27.653, -25.763, -23.719, -34.445, -34.670,
-34.411, -33.667, -32.578, -32.307, -29.742, -27.678, -25.786, -23.740, -34.473,
-34.698, -34.440, -33.696, -32.606, -32.338, -29.769, -27.703, -25.809, -23.761,
 -34.502, -34.727, -34.469, -33.725, -32.635, -32.369, -29.797, -27.728,
-25.833, -23.782, -34.530, -34.756, -34.497, -33.754, -32.663, -32.401, -29.824,
-27.753, -25.856, -23.803, -34.558, -34.784, -34.526, -33.782, -32.692, -32.432,
-29.852, -27.778, -25.879, -23.824, -34.587, -34.813, -34.555, -33.811, -32.720,
-32.463, -29.879, -27.803, -25.902, -23.845, -34.615, -34.842, -34.584, -33.840,
-32.748, -32.494, -29.907, -27.828, -25.925, -23.866, -34.644, -34.870, -34.613,
-33.869, -32.777, -32.525, -29.935, -27.853, -25.948, -23.887, -34.672, -34.899,
-34.642, -33.897, -32.805, -32.556, -29.962, -27.878, -25.971, -23.908, -34.700,
-34.928, -34.670, -33.926, -32.834, -32.587, -29.990, -27.903, -25.994, -23.929,
 -34.729, -34.956, -34.699, -33.955, -32.862, -32.619, -30.017, -27.928,
-26.017, -23.950, -34.757, -34.985, -34.728, -33.984, -32.891, -32.650, -30.045,
-27.953, -26.040, -23.971, -34.786, -35.014, -34.757, -34.012, -32.919, -32.681,
-30.072, -27.978, -26.063, -23.992, -34.814, -35.042, -34.786, -34.041, -32.948,
-32.712, -30.100, -28.003, -26.086, -24.013, -34.842, -35.071, -34.814, -34.070,
-32.976, -32.743, -30.127, -28.028, -26.109, -24.034, -34.871, -35.099, -34.843,
-34.099, -33.005, -32.774, -30.155, -28.053, -26.133, -24.055, -34.899, -35.128,
-34.872, -34.127, -33.033, -32.806, -30.183, -28.078, -26.156, -24.076, -34.927,
-35.157, -34.901, -34.156, -33.062, -32.837, -30.210, -28.103, -26.179, -24.098,
 -34.956, -35.185, -34.930, -34.185, -33.090, -32.868, -30.238, -28.128,
-26.202, -24.119, -34.984, -35.214, -34.958, -34.214, -33.118, -32.899, -30.265,
-28.153, -26.225, -24.140, -35.012, -35.243, -34.987, -34.242, -33.147, -32.930,
-30.293, -28.178, -26.248, -24.161, -35.041, -35.271, -35.016, -34.271, -33.175,
-32.961, -30.320, -28.203, -26.271, -24.182, -35.069, -35.300, -35.045, -34.300,
-33.204, -32.992, -30.348, -28.228, -26.294, -24.203, -35.098, -35.329, -35.074,
-34.329, -33.232, -33.024, -30.376, -28.253, -26.317, -24.224, -35.126, -35.357,
-35.102, -34.357, -33.261, -33.055, -30.403, -28.278, -26.340, -24.245, -35.154,
-35.386, -35.131, -34.386, -33.289, -33.086, -30.431, -28.303, -26.364, -24.266,
 -35.183, -35.414, -35.160, -34.415, -33.318, -33.117, -30.458, -28.328,
-26.387, -24.288, -35.211, -35.443, -35.189, -34.444, -33.346, -33.148, -30.486,
-28.353, -26.410, -24.309, -35.239, -35.472, -35.218, -34.472, -33.375, -33.179,
-30.513, -28.378, -26.433, -24.330, -35.268, -35.500, -35.246, -34.501, -33.403,
-33.211, -30.541, -28.404, -26.456, -24.351, -35.296, -35.529, -35.275, -34.530,
-33.432, -33.242, -30.569, -28.429, -26.479, -24.372, -35.324, -35.558, -35.304,
-34.559, -33.460, -33.273, -30.596, -28.454, -26.502, -24.393, -35.353, -35.586,
-35.333, -34.587, -33.488, -33.304, -30.624, -28.479, -26.525, -24.415, -35.381,
-35.615, -35.362, -34.616, -33.517, -33.335, -30.651, -28.504, -26.548, -24.436,
 -35.409, -35.643, -35.390, -34.645, -33.545, -33.366, -30.679, -28.529,
-26.572, -24.457, -35.438, -35.672, -35.419, -34.674, -33.574, -33.398, -30.706,
-28.554, -26.595, -24.478, -35.466, -35.701, -35.448, -34.702, -33.602, -33.429,
-30.734, -28.579, -26.618, -24.499, -35.494, -35.729, -35.477, -34.731, -33.631,
-33.460, -30.762, -28.604, -26.641, -24.520, -35.523, -35.758, -35.506, -34.760,
-33.659, -33.491, -30.789, -28.629, -26.664, -24.542, -35.551, -35.787, -35.534,
-34.789, -33.688, -33.522, -30.817, -28.654, -26.687, -24.563, -35.579, -35.815,
-35.563, -34.817, -33.716, -33.553, -30.844, -28.679, -26.710, -24.584, -35.608,
-35.844, -35.592, -34.846, -33.745, -33.585, -30.872, -28.704, -26.734, -24.605,
 -35.636, -35.872, -35.621, -34.875, -33.773, -33.616, -30.899, -28.729,
-26.757, -24.626, -35.664, -35.901, -35.650, -34.904, -33.802, -33.647, -30.927,
-28.754, -26.780, -24.648, -35.693, -35.930, -35.678, -34.932, -33.830, -33.678,
-30.955, -28.779, -26.803, -24.669, -35.721, -35.958, -35.707, -34.961, -33.858,
-33.709, -30.982, -28.804, -26.826, -24.690, -35.749, -35.987, -35.736, -34.990,
-33.887, -33.740, -31.010, -28.829, -26.849, -24.711, -35.778, -36.016, -35.765,
-35.019, -33.915, -33.772, -31.037, -28.854, -26.872, -24.733, -35.806, -36.044,
-35.793, -35.047, -33.944, -33.803, -31.065, -28.879, -26.896, -24.754, -35.834,
-36.073, -35.822, -35.076, -33.972, -33.834, -31.093, -28.905, -26.919, -24.775,
 -35.862, -36.101, -35.851, -35.105, -34.001, -33.865, -31.120, -28.930,
-26.942, -24.796, -35.891, -36.130, -35.880, -35.133, -34.029, -33.896, -31.148,
-28.955, -26.965, -24.817, -35.919, -36.159, -35.909, -35.162, -34.058, -33.927,
-31.175, -28.980, -26.988, -24.839, -35.947, -36.187, -35.937, -35.191, -34.086,
-33.959, -31.203, -29.005, -27.011, -24.860, -35.976, -36.216, -35.966, -35.220,
-34.115, -33.990, -31.231, -29.030, -27.035, -24.881, -36.004, -36.244, -35.995,
-35.248, -34.143, -34.021, -31.258, -29.055, -27.058, -24.902, -36.032, -36.273,
-36.024, -35.277, -34.171, -34.052, -31.286, -29.080, -27.081, -24.924, -36.061,
-36.302, -36.053, -35.306, -34.200, -34.083, -31.313, -29.105, -27.104, -24.945,
 -36.089, -36.330, -36.081, -35.335, -34.228, -34.115, -31.341, -29.130,
-27.127, -24.966, -36.117, -36.359, -36.110, -35.363, -34.257, -34.146, -31.368,
-29.155, -27.150, -24.988, -36.146, -36.388, -36.139, -35.392, -34.285, -34.177,
-31.396, -29.180, -27.174, -25.009, -36.174, -36.416, -36.168, -35.421, -34.314,
-34.208, -31.424, -29.205, -27.197, -25.030, -36.202, -36.445, -36.197, -35.450,
-34.342, -34.239, -31.451, -29.230, -27.220, -25.051, -36.230, -36.473, -36.225,
-35.478, -34.371, -34.270, -31.479, -29.255, -27.243, -25.073, -36.259, -36.502,
-36.254, -35.507, -34.399, -34.302, -31.506, -29.280, -27.266, -25.094, -36.287,
-36.531, -36.283, -35.536, -34.428, -34.333, -31.534, -29.306, -27.290, -25.115,
 -36.315, -36.559, -36.312, -35.565, -34.456, -34.364, -31.562, -29.331,
-27.313, -25.137, -36.344, -36.588, -36.341, -35.593, -34.485, -34.395, -31.589,
-29.356, -27.336, -25.158, -36.372, -36.616, -36.369, -35.622, -34.513, -34.426,
-31.617, -29.381, -27.359, -25.179, -36.400, -36.645, -36.398, -35.651, -34.541,
-34.458, -31.644, -29.406, -27.382, -25.201, -36.428, -36.674, -36.427, -35.680,
-34.570, -34.489, -31.672, -29.431, -27.406, -25.222, -36.457, -36.702, -36.456,
-35.708, -34.598, -34.520, -31.700, -29.456, -27.429, -25.243, -36.485, -36.731,
-36.485, -35.737, -34.627, -34.551, -31.727, -29.481, -27.452, -25.264, -36.513,
-36.759, -36.513, -35.766, -34.655, -34.582, -31.755, -29.506, -27.475, -25.286,
 -36.542, -36.788, -36.542, -35.795, -34.684, -34.613, -31.782, -29.531,
-27.498, -25.307, -36.570, -36.817, -36.571, -35.823, -34.712, -34.645, -31.810,
-29.556, -27.522, -25.328, -36.598, -36.845, -36.600, -35.852, -34.741, -34.676,
-31.838, -29.581, -27.545, -25.350, -36.626, -36.874, -36.628, -35.881, -34.769,
-34.707, -31.865, -29.607, -27.568, -25.371, -36.655, -36.902, -36.657, -35.910,
-34.798, -34.738, -31.893, -29.632, -27.591, -25.393, -36.683, -36.931, -36.686,
-35.938, -34.826, -34.769, -31.920, -29.657, -27.614, -25.414, -36.711, -36.960,
-36.715, -35.967, -34.855, -34.801, -31.948, -29.682, -27.638, -25.435, -36.740,
-36.988, -36.744, -35.996, -34.883, -34.832, -31.976, -29.707, -27.661, -25.457,
 -36.768, -37.017, -36.772, -36.025, -34.911, -34.863, -32.003, -29.732,
-27.684, -25.478, -36.796, -37.045, -36.801, -36.053, -34.940, -34.894, -32.031,
-29.757, -27.707, -25.499, -36.824, -37.074, -36.830, -36.082, -34.968, -34.925,
-32.058, -29.782, -27.730, -25.521, -36.853, -37.103, -36.859, -36.111, -34.997,
-34.957, -32.086, -29.807, -27.754, -25.542, -36.881, -37.131, -36.888, -36.140,
-35.025, -34.988, -32.114, -29.832, -27.777, -25.563, -36.909, -37.160, -36.916,
-36.168, -35.054, -35.019, -32.141, -29.857, -27.800, -25.585, -36.937, -37.188,
-36.945, -36.197, -35.082, -35.050, -32.169, -29.882, -27.823, -25.606, -36.966,
-37.217, -36.974, -36.226, -35.111, -35.081, -32.196, -29.908, -27.847, -25.628,
 -36.994, -37.246, -37.003, -36.254, -35.139, -35.113, -32.224, -29.933,
-27.870, -25.649, -37.022, -37.274, -37.032, -36.283, -35.168, -35.144, -32.252,
-29.958, -27.893, -25.670, -37.050, -37.303, -37.060, -36.312, -35.196, -35.175,
-32.279, -29.983, -27.916, -25.692, -37.079, -37.331, -37.089, -36.341, -35.224,
-35.206, -32.307, -30.008, -27.940, -25.713, -37.107, -37.360, -37.118, -36.369,
-35.253, -35.237, -32.334, -30.033, -27.963, -25.734, -37.135, -37.389, -37.147,
-36.398, -35.281, -35.269, -32.362, -30.058, -27.986, -25.756, -37.163, -37.417,
-37.175, -36.427, -35.310, -35.300, -32.390, -30.083, -28.009, -25.777, -37.192,
-37.446, -37.204, -36.456, -35.338, -35.331, -32.417, -30.108, -28.032, -25.799,
 -37.220, -37.474, -37.233, -36.484, -35.367, -35.362, -32.445, -30.133,
-28.056, -25.820, -37.248, -37.503, -37.262, -36.513, -35.395, -35.393, -32.473,
-30.159, -28.079, -25.841, -37.277, -37.532, -37.291, -36.542, -35.424, -35.425,
-32.500, -30.184, -28.102, -25.863, -37.305, -37.560, -37.319, -36.571, -35.452,
-35.456, -32.528, -30.209, -28.125, -25.884, -37.333, -37.589, -37.348, -36.599,
-35.481, -35.487, -32.555, -30.234, -28.149, -25.906, -37.361, -37.617, -37.377,
-36.628, -35.509, -35.518, -32.583, -30.259, -28.172, -25.927, -37.390, -37.646,
-37.406, -36.657, -35.538, -35.549, -32.611, -30.284, -28.195, -25.949, -37.418,
-37.675, -37.435, -36.686, -35.566, -35.581, -32.638, -30.309, -28.218, -25.970,
 -37.446, -37.703, -37.463, -36.714, -35.594, -35.612, -32.666, -30.334,
-28.242, -25.991, -37.474, -37.732, -37.492, -36.743, -35.623, -35.643, -32.693,
-30.359, -28.265, -26.013, -37.502, -37.760, -37.521, -36.772, -35.651, -35.674,
-32.721, -30.385, -28.288, -26.034, -37.531, -37.789, -37.550, -36.801, -35.680,
-35.705, -32.749, -30.410, -28.311, -26.056, -37.559, -37.818, -37.578, -36.829,
-35.708, -35.737, -32.776, -30.435, -28.335, -26.077, -37.587, -37.846, -37.607,
-36.858, -35.737, -35.768, -32.804, -30.460, -28.358, -26.099, -37.615, -37.875,
-37.636, -36.887, -35.765, -35.799, -32.832, -30.485, -28.381, -26.120, -37.644,
-37.903, -37.665, -36.916, -35.794, -35.830, -32.859, -30.510, -28.405, -26.142,
 -37.672, -37.932, -37.694, -36.944, -35.822, -35.861, -32.887, -30.535,
-28.428, -26.163, -37.700, -37.961, -37.722, -36.973, -35.851, -35.893, -32.914,
-30.560, -28.451, -26.184, -37.728, -37.989, -37.751, -37.002, -35.879, -35.924,
-32.942, -30.585, -28.474, -26.206, -37.757, -38.018, -37.780, -37.031, -35.908,
-35.955, -32.970, -30.611, -28.498, -26.227, -37.785, -38.046, -37.809, -37.059,
-35.936, -35.986, -32.997, -30.636, -28.521, -26.249, -37.813, -38.075, -37.838,
-37.088, -35.965, -36.017, -33.025, -30.661, -28.544, -26.270, -37.841, -38.103,
-37.866, -37.117, -35.993, -36.049, -33.053, -30.686, -28.567, -26.292, -37.870,
-38.132, -37.895, -37.146, -36.021, -36.080, -33.080, -30.711, -28.591, -26.313,
 -37.898, -38.161, -37.924, -37.174, -36.050, -36.111, -33.108, -30.736,
-28.614, -26.335, -37.926, -38.189, -37.953, -37.203, -36.078, -36.142, -33.135,
-30.761, -28.637, -26.356, -37.954, -38.218, -37.982, -37.232, -36.107, -36.174,
-33.163, -30.786, -28.661, -26.378, -37.982, -38.246, -38.010, -37.261, -36.135,
-36.205, -33.191, -30.811, -28.684, -26.399, -38.011, -38.275, -38.039, -37.289,
-36.164, -36.236, -33.218, -30.837, -28.707, -26.421, -38.039, -38.304, -38.068,
-37.318, -36.192, -36.267, -33.246, -30.862, -28.730, -26.442, -38.067, -38.332,
-38.097, -37.347, -36.221, -36.298, -33.274, -30.887, -28.754, -26.464, -38.095,
-38.361, -38.125, -37.376, -36.249, -36.330, -33.301, -30.912, -28.777, -26.485,
 -38.124, -38.389, -38.154, -37.404, -36.278, -36.361, -33.329, -30.937,
-28.800, -26.507, -38.152, -38.418, -38.183, -37.433, -36.306, -36.392, -33.356,
-30.962, -28.824, -26.528, -38.180, -38.447, -38.212, -37.462, -36.335, -36.423,
-33.384, -30.987, -28.847, -26.550, -38.208, -38.475, -38.241, -37.490, -36.363,
-36.455, -33.412, -31.012, -28.870, -26.571, -38.237, -38.504, -38.269, -37.519,
-36.391, -36.486, -33.439, -31.038, -28.893, -26.593, -38.265, -38.532, -38.298,
-37.548, -36.420, -36.517, -33.467, -31.063, -28.917, -26.614, -38.293, -38.561,
-38.327, -37.577, -36.448, -36.548, -33.495, -31.088, -28.940, -26.636, -38.321,
-38.589, -38.356, -37.605, -36.477, -36.579, -33.522, -31.113, -28.963, -26.657,
 -38.349, -38.618, -38.385, -37.634, -36.505, -36.611, -33.550, -31.138,
-28.987, -26.679, -38.378, -38.647, -38.413, -37.663, -36.534, -36.642, -33.577,
-31.163, -29.010, -26.700, -38.406, -38.675, -38.442, -37.692, -36.562, -36.673,
-33.605, -31.188, -29.033, -26.722, -38.434, -38.704, -38.471, -37.720, -36.591,
-36.704, -33.633, -31.214, -29.057, -26.743, -38.462, -38.732, -38.500, -37.749,
-36.619, -36.736, -33.660, -31.239, -29.080, -26.765, -38.490, -38.761, -38.529,
-37.778, -36.648, -36.767, -33.688, -31.264, -29.103, -26.786, -38.519, -38.790,
-38.557, -37.807, -36.676, -36.798, -33.716, -31.289, -29.126, -26.808, -38.547,
-38.818, -38.586, -37.835, -36.705, -36.829, -33.743, -31.314, -29.150, -26.829,
 -38.575, -38.847, -38.615, -37.864, -36.733, -36.860, -33.771, -31.339,
-29.173, -26.851, -38.603, -38.875, -38.644, -37.893, -36.762, -36.892, -33.799,
-31.364, -29.196, -26.872, -38.631, -38.904, -38.672, -37.922, -36.790, -36.923,
-33.826, -31.389, -29.220, -26.894, -38.660, -38.933, -38.701, -37.950, -36.818,
-36.954, -33.854, -31.415, -29.243, -26.915, -38.688, -38.961, -38.730, -37.979,
-36.847, -36.985, -33.881, -31.440, -29.266, -26.937, -38.716, -38.990, -38.759,
-38.008, -36.875, -37.017, -33.909, -31.465, -29.290, -26.958, -38.744, -39.018,
-38.788, -38.037, -36.904, -37.048, -33.937, -31.490, -29.313, -26.980, -38.773,
-39.047, -38.816, -38.065, -36.932, -37.079, -33.964, -31.515, -29.336, -27.001,
 -38.801, -39.075, -38.845, -38.094, -36.961, -37.110, -33.992, -31.540,
-29.360, -27.023, -38.829, -39.104, -38.874, -38.123, -36.989, -37.141, -34.020,
-31.565, -29.383, -27.045, -38.857, -39.133, -38.903, -38.152, -37.018, -37.173,
-34.047, -31.591, -29.406, -27.066, -38.885, -39.161, -38.932, -38.180, -37.046,
-37.204, -34.075, -31.616, -29.430, -27.088, -38.914, -39.190, -38.960, -38.209,
-37.075, -37.235, -34.103, -31.641, -29.453, -27.109, -38.942, -39.218, -38.989,
-38.238, -37.103, -37.266, -34.130, -31.666, -29.476, -27.131, -38.970, -39.247,
-39.018, -38.267, -37.132, -37.298, -34.158, -31.691, -29.499, -27.152, -38.998,
-39.276, -39.047, -38.295, -37.160, -37.329, -34.185, -31.716, -29.523, -27.174,
 -39.026, -39.304, -39.076, -38.324, -37.189, -37.360, -34.213, -31.741,
-29.546, -27.195, -39.055, -39.333, -39.104, -38.353, -37.217, -37.391, -34.241,
-31.767, -29.569, -27.217, -39.083, -39.361, -39.133, -38.382, -37.245, -37.423,
-34.268, -31.792, -29.593, -27.239, -39.111, -39.390, -39.162, -38.410, -37.274,
-37.454, -34.296, -31.817, -29.616, -27.260, -39.139, -39.418, -39.191, -38.439,
-37.302, -37.485, -34.324, -31.842, -29.639, -27.282, -39.167, -39.447, -39.219,
-38.468, -37.331, -37.516, -34.351, -31.867, -29.663, -27.303, -39.196, -39.476,
-39.248, -38.497, -37.359, -37.547, -34.379, -31.892, -29.686, -27.325, -39.224,
-39.504, -39.277, -38.525, -37.388, -37.579, -34.407, -31.917, -29.709, -27.346,
 -39.252, -39.533, -39.306, -38.554, -37.416, -37.610, -34.434, -31.943,
-29.733, -27.368, -39.280, -39.561, -39.335, -38.583, -37.445, -37.641, -34.462,
-31.968, -29.756, -27.389, -39.308, -39.590, -39.363, -38.612, -37.473, -37.672,
-34.490, -31.993, -29.779, -27.411, -39.336, -39.619, -39.392, -38.640, -37.502,
-37.704, -34.517, -32.018, -29.803, -27.433, -39.365, -39.647, -39.421, -38.669,
-37.530, -37.735, -34.545, -32.043, -29.826, -27.454, -39.393, -39.676, -39.450,
-38.698, -37.559, -37.766, -34.572, -32.068, -29.849, -27.476, -39.421, -39.704,
-39.479, -38.727, -37.587, -37.797, -34.600, -32.094, -29.873, -27.497, -39.449,
-39.733, -39.507, -38.755, -37.616, -37.829, -34.628, -32.119, -29.896, -27.519,
 -39.477, -39.761, -39.536, -38.784, -37.644, -37.860, -34.655, -32.144,
-29.920, -27.540, -39.506, -39.790, -39.565, -38.813, -37.673, -37.891, -34.683,
-32.169, -29.943, -27.562, -39.534, -39.819, -39.594, -38.842, -37.701, -37.922,
-34.711, -32.194, -29.966, -27.584, -39.562, -39.847, -39.623, -38.870, -37.729,
-37.954, -34.738, -32.219, -29.990, -27.605, -39.590, -39.876, -39.651, -38.899,
-37.758, -37.985, -34.766, -32.244, -30.013, -27.627, -39.618, -39.904, -39.680,
-38.928, -37.786, -38.016, -34.794, -32.270, -30.036, -27.648, -39.647, -39.933,
-39.709, -38.957, -37.815, -38.047, -34.821, -32.295, -30.060, -27.670, -39.675,
-39.962, -39.738, -38.985, -37.843, -38.079, -34.849, -32.320, -30.083, -27.692,
 -39.703, -39.990, -39.767, -39.014, -37.872, -38.110, -34.877, -32.345,
-30.106, -27.713, -39.731, -40.019, -39.795, -39.043, -37.900, -38.141, -34.904,
-32.370, -30.130, -27.735, -39.759, -40.047, -39.824, -39.072, -37.929, -38.172,
-34.932, -32.395, -30.153, -27.756, -39.787, -40.076, -39.853, -39.100, -37.957,
-38.204, -34.960, -32.421, -30.176, -27.778, -39.816, -40.104, -39.882, -39.129,
-37.986, -38.235, -34.987, -32.446, -30.200, -27.800, -39.844, -40.133, -39.910,
-39.158, -38.014, -38.266, -35.015, -32.471, -30.223, -27.821, -39.872, -40.162,
-39.939, -39.187, -38.043, -38.297, -35.043, -32.496, -30.246, -27.843, -39.900,
-40.190, -39.968, -39.215, -38.071, -38.329, -35.070, -32.521, -30.270, -27.864,
 -39.928, -40.219, -39.997, -39.244, -38.100, -38.360, -35.098, -32.546,
-30.293, -27.886, -39.957, -40.247, -40.026, -39.273, -38.128, -38.391, -35.125,
-32.572, -30.317, -27.908, -39.985, -40.276, -40.054, -39.302, -38.157, -38.422,
-35.153, -32.597, -30.340, -27.929, -40.013, -40.305, -40.083, -39.330, -38.185,
-38.454, -35.181, -32.622, -30.363, -27.951, -40.041, -40.333, -40.112, -39.359,
-38.214, -38.485, -35.208, -32.647, -30.387, -27.972, -40.069, -40.362, -40.141,
-39.388, -38.242, -38.516, -35.236, -32.672, -30.410, -27.994, -40.097, -40.390,
-40.170, -39.417, -38.270, -38.547, -35.264, -32.697, -30.433, -28.016, -40.126,
-40.419, -40.198, -39.445, -38.299, -38.579, -35.291, -32.723, -30.457, -28.037,
 -40.154, -40.447, -40.227, -39.474, -38.327, -38.610, -35.319, -32.748,
-30.480, -28.059, -40.182, -40.476, -40.256, -39.503, -38.356, -38.641, -35.347,
-32.773, -30.503, -28.080, -40.210, -40.505, -40.285, -39.532, -38.384, -38.672,
-35.374, -32.798, -30.527, -28.102, -40.238, -40.533, -40.314, -39.560, -38.413,
-38.704, -35.402, -32.823, -30.550, -28.124, -40.266, -40.562, -40.342, -39.589,
-38.441, -38.735, -35.430, -32.848, -30.574, -28.145, -40.295, -40.590, -40.371,
-39.618, -38.470, -38.766, -35.457, -32.874, -30.597, -28.167, -40.323, -40.619,
-40.400, -39.647, -38.498, -38.797, -35.485, -32.899, -30.620, -28.188, -40.351,
-40.648, -40.429, -39.676, -38.527, -38.829, -35.513, -32.924, -30.644, -28.210,
 -40.379, -40.676, -40.458, -39.704, -38.555, -38.860, -35.540, -32.949,
-30.667, -28.232, -40.407, -40.705, -40.486, -39.733, -38.584, -38.891, -35.568,
-32.974, -30.690, -28.253, -40.435, -40.733, -40.515, -39.762, -38.612, -38.922,
-35.596, -32.999, -30.714, -28.275, -40.464, -40.762, -40.544, -39.791, -38.641,
-38.954, -35.623, -33.025, -30.737, -28.297, -40.492, -40.790, -40.573, -39.819,
-38.669, -38.985, -35.651, -33.050, -30.761, -28.318, -40.520, -40.819, -40.602,
-39.848, -38.698, -39.016, -35.679, -33.075, -30.784, -28.340, -40.548, -40.848,
-40.630, -39.877, -38.726, -39.047, -35.706, -33.100, -30.807, -28.361, -40.576,
-40.876, -40.659, -39.906, -38.755, -39.079, -35.734, -33.125, -30.831, -28.383,
 -40.604, -40.905, -40.688, -39.934, -38.783, -39.110, -35.762, -33.151,
-30.854, -28.405, -40.633, -40.933, -40.717, -39.963, -38.812, -39.141, -35.789,
-33.176, -30.877, -28.426, -40.661, -40.962, -40.746, -39.992, -38.840, -39.172,
-35.817, -33.201, -30.901, -28.448, -40.689, -40.991, -40.774, -40.021, -38.868,
-39.204, -35.845, -33.226, -30.924, -28.469, -40.717, -41.019, -40.803, -40.049,
-38.897, -39.235, -35.872, -33.251, -30.948, -28.491, -40.745, -41.048, -40.832,
-40.078, -38.925, -39.266, -35.900, -33.276, -30.971, -28.513, -40.773, -41.076,
-40.861, -40.107, -38.954, -39.297, -35.928, -33.302, -30.994, -28.534, -40.802,
-41.105, -40.890, -40.136, -38.982, -39.329, -35.955, -33.327, -31.018, -28.556,
 -40.830, -41.133, -40.918, -40.164, -39.011, -39.360, -35.983, -33.352,
-31.041, -28.578, -40.858, -41.162, -40.947, -40.193, -39.039, -39.391, -36.011,
-33.377, -31.065, -28.599, -40.886, -41.191, -40.976, -40.222, -39.068, -39.422,
-36.038, -33.402, -31.088, -28.621, -40.914, -41.219, -41.005, -40.251, -39.096,
-39.454, -36.066, -33.427, -31.111, -28.642, -40.942, -41.248, -41.034, -40.279,
-39.125, -39.485, -36.094, -33.453, -31.135, -28.664, -40.971, -41.276, -41.062,
-40.308, -39.153, -39.516, -36.121, -33.478, -31.158, -28.686, -40.999, -41.305,
-41.091, -40.337, -39.182, -39.547, -36.149, -33.503, -31.181, -28.707, -41.027,
-41.334, -41.120, -40.366, -39.210, -39.579, -36.177, -33.528, -31.205, -28.729,
 -41.055, -41.362, -41.149, -40.394, -39.239, -39.610, -36.204, -33.553,
-31.228, -28.751, -41.083, -41.391, -41.178, -40.423, -39.267, -39.641, -36.232,
-33.579, -31.252, -28.772, -41.111, -41.419, -41.206, -40.452, -39.296, -39.673,
-36.260, -33.604, -31.275, -28.794, -41.139, -41.448, -41.235, -40.481, -39.324,
-39.704, -36.287, -33.629, -31.298, -28.816, -41.168, -41.476, -41.264, -40.510,
-39.353, -39.735, -36.315, -33.654, -31.322, -28.837, -41.196, -41.505, -41.293,
-40.538, -39.381, -39.766, -36.343, -33.679, -31.345, -28.859, -41.224, -41.534,
-41.322, -40.567, -39.410, -39.798, -36.370, -33.705, -31.369, -28.880, -41.252,
-41.562, -41.350, -40.596, -39.438, -39.829, -36.398, -33.730, -31.392, -28.902,
 -41.280, -41.591, -41.379, -40.625, -39.467, -39.860, -36.426, -33.755,
-31.415, -28.924, -41.308, -41.619, -41.408, -40.653, -39.495, -39.891, -36.453,
-33.780, -31.439, -28.945, -41.337, -41.648, -41.437, -40.682, -39.524, -39.923,
-36.481, -33.805, -31.462, -28.967, -41.365, -41.677, -41.466, -40.711, -39.552,
-39.954, -36.509, -33.830, -31.486, -28.989, -41.393, -41.705, -41.494, -40.740,
-39.580, -39.985, -36.536, -33.856, -31.509, -29.010, -41.421, -41.734, -41.523,
-40.768, -39.609, -40.016, -36.564, -33.881, -31.533, -29.032, -41.449, -41.762,
-41.552, -40.797, -39.637, -40.048, -36.592, -33.906, -31.556, -29.054, -41.477,
-41.791, -41.581, -40.826, -39.666, -40.079, -36.619, -33.931, -31.579, -29.075,
 -41.505, -41.819, -41.610, -40.855, -39.694, -40.110, -36.647, -33.956,
-31.603, -29.097, -41.534, -41.848, -41.638, -40.883, -39.723, -40.142, -36.675,
-33.982, -31.626, -29.118, -41.562, -41.877, -41.667, -40.912, -39.751, -40.173,
-36.702, -34.007, -31.650, -29.140, -41.590, -41.905, -41.696, -40.941, -39.780,
-40.204, -36.730, -34.032, -31.673, -29.162, -41.618, -41.934, -41.725, -40.970,
-39.808, -40.235, -36.758, -34.057, -31.696, -29.183, -41.646, -41.962, -41.754,
-40.998, -39.837, -40.267, -36.785, -34.082, -31.720, -29.205, -41.674, -41.991,
-41.782, -41.027, -39.865, -40.298, -36.813, -34.108, -31.743, -29.227, -41.703,
-42.020, -41.811, -41.056, -39.894, -40.329, -36.841, -34.133, -31.767, -29.248,
 -41.731, -42.048, -41.840, -41.085, -39.922, -40.360, -36.868, -34.158,
-31.790, -29.270, -41.759, -42.077, -41.869, -41.114, -39.951, -40.392, -36.896,
-34.183, -31.813, -29.292, -41.787, -42.105, -41.898, -41.142, -39.979, -40.423,
-36.924, -34.208, -31.837, -29.313, -41.815, -42.134, -41.926, -41.171, -40.008,
-40.454, -36.951, -34.234, -31.860, -29.335, -41.843, -42.163, -41.955, -41.200,
-40.036, -40.486, -36.979, -34.259, -31.884, -29.356, -41.871, -42.191, -41.984,
-41.229, -40.065, -40.517, -37.007, -34.284, -31.907, -29.378, -41.900, -42.220,
-42.013, -41.257, -40.093, -40.548, -37.034, -34.309, -31.931, -29.400, -41.928,
-42.248, -42.042, -41.286, -40.122, -40.579, -37.062, -34.334, -31.954, -29.421,
 -41.956, -42.277, -42.070, -41.315, -40.150, -40.611, -37.090, -34.360,
-31.977, -29.443, -41.984, -42.305, -42.099, -41.344, -40.179, -40.642, -37.118,
-34.385, -32.001, -29.465, -42.012, -42.334, -42.128, -41.372, -40.207, -40.673,
-37.145, -34.410, -32.024, -29.486, -42.040, -42.363, -42.157, -41.401, -40.236,
-40.704, -37.173, -34.435, -32.048, -29.508, -42.068, -42.391, -42.186, -41.430,
-40.264, -40.736, -37.201, -34.460, -32.071, -29.530, -42.097, -42.420, -42.215,
-41.459, -40.293, -40.767, -37.228, -34.486, -32.094, -29.551, -42.125, -42.448,
-42.243, -41.487, -40.321, -40.798, -37.256, -34.511, -32.118, -29.573, -42.153,
-42.477, -42.272, -41.516, -40.350, -40.830, -37.284, -34.536, -32.141, -29.595,
 -42.181, -42.506, -42.301, -41.545, -40.378, -40.861, -37.311, -34.561,
-32.165, -29.616, -42.209, -42.534, -42.330, -41.574, -40.407, -40.892, -37.339,
-34.586, -32.188, -29.638, -42.237, -42.563, -42.359, -41.603, -40.435, -40.923,
-37.367, -34.612, -32.212, -29.659, -42.265, -42.591, -42.387, -41.631, -40.464,
-40.955, -37.394, -34.637, -32.235, -29.681, -42.294, -42.620, -42.416, -41.660,
-40.492, -40.986, -37.422, -34.662, -32.258, -29.703, -42.322, -42.648, -42.445,
-41.689, -40.521, -41.017, -37.450, -34.687, -32.282, -29.724, -42.350, -42.677,
-42.474, -41.718, -40.549, -41.048, -37.477, -34.712, -32.305, -29.746, -42.378,
-42.706, -42.503, -41.746, -40.578, -41.080, -37.505, -34.738, -32.329, -29.768,
 -42.406, -42.734, -42.531, -41.775, -40.606, -41.111, -37.533, -34.763,
-32.352, -29.789, -42.434, -42.763, -42.560, -41.804, -40.635, -41.142, -37.560,
-34.788, -32.376, -29.811, -42.462, -42.791, -42.589, -41.833, -40.663, -41.174,
-37.588, -34.813, -32.399, -29.833, -42.491, -42.820, -42.618, -41.861, -40.691,
-41.205, -37.616, -34.838, -32.422, -29.854, -42.519, -42.849, -42.647, -41.890,
-40.720, -41.236, -37.643, -34.864, -32.446, -29.876, -42.547, -42.877, -42.675,
-41.919, -40.748, -41.267, -37.671, -34.889, -32.469, -29.898, -42.575, -42.906,
-42.704, -41.948, -40.777, -41.299, -37.699, -34.914, -32.493, -29.919, -42.603,
-42.934, -42.733, -41.977, -40.805, -41.330, -37.727, -34.939, -32.516, -29.941,
 -42.631, -42.963, -42.762, -42.005, -40.834, -41.361, -37.754, -34.964,
-32.540, -29.962, -42.659, -42.992, -42.791, -42.034, -40.862, -41.393, -37.782,
-34.990, -32.563, -29.984, -42.688, -43.020, -42.819, -42.063, -40.891, -41.424,
-37.810, -35.015, -32.586, -30.006, -42.716, -43.049, -42.848, -42.092, -40.919,
-41.455, -37.837, -35.040, -32.610, -30.027, -42.744, -43.077, -42.877, -42.120,
-40.948, -41.486, -37.865, -35.065, -32.633, -30.049, -42.772, -43.106, -42.906,
-42.149, -40.976, -41.518, -37.893, -35.091, -32.657, -30.071, -42.800, -43.135,
-42.935, -42.178, -41.005, -41.549, -37.920, -35.116, -32.680, -30.092, -42.828,
-43.163, -42.964, -42.207, -41.033, -41.580, -37.948, -35.141, -32.704, -30.114,
 -42.856, -43.192, -42.992, -42.235, -41.062, -41.612, -37.976, -35.166,
-32.727, -30.136, -42.885, -43.220, -43.021, -42.264, -41.090, -41.643, -38.003,
-35.191, -32.751, -30.157, -42.913, -43.249, -43.050, -42.293, -41.119, -41.674,
-38.031, -35.217, -32.774, -30.179, -42.941, -43.277, -43.079, -42.322, -41.147,
-41.705, -38.059, -35.242, -32.797, -30.201, -42.969, -43.306, -43.108, -42.351,
-41.176, -41.737, -38.086, -35.267, -32.821, -30.222, -42.997, -43.335, -43.136,
-42.379, -41.204, -41.768, -38.114, -35.292, -32.844, -30.244, -43.025, -43.363,
-43.165, -42.408, -41.233, -41.799, -38.142, -35.317, -32.868, -30.265, -43.053,
-43.392, -43.194, -42.437, -41.261, -41.831, -38.170, -35.343, -32.891, -30.287,
 -43.081, -43.420, -43.223, -42.466, -41.290, -41.862, -38.197, -35.368,
-32.915, -30.309, -43.110, -43.449, -43.252, -42.494, -41.318, -41.893, -38.225,
-35.393, -32.938, -30.330, -43.138, -43.478, -43.281, -42.523, -41.347, -41.924,
-38.253, -35.418, -32.962, -30.352, -43.166, -43.506, -43.309, -42.552, -41.375,
-41.956, -38.280, -35.443, -32.985, -30.374, -43.194, -43.535, -43.338, -42.581,
-41.404, -41.987, -38.308, -35.469, -33.008, -30.395, -43.222, -43.563, -43.367,
-42.610, -41.432, -42.018, -38.336, -35.494, -33.032, -30.417, -43.250, -43.592,
-43.396, -42.638, -41.461, -42.050, -38.363, -35.519, -33.055, -30.439, -43.278,
-43.621, -43.425, -42.667, -41.489, -42.081, -38.391, -35.544, -33.079, -30.460,
 -43.307, -43.649, -43.453, -42.696, -41.518, -42.112, -38.419, -35.570,
-33.102, -30.482, -43.335, -43.678, -43.482, -42.725, -41.546, -42.143, -38.446,
-35.595, -33.126, -30.503, -43.363, -43.706, -43.511, -42.753, -41.575, -42.175,
-38.474, -35.620, -33.149, -30.525, -43.391, -43.735, -43.540, -42.782, -41.603,
-42.206, -38.502, -35.645, -33.173, -30.547, -43.419, -43.764, -43.569, -42.811,
-41.632, -42.237, -38.530, -35.670, -33.196, -30.568, -43.447, -43.792, -43.597,
-42.840, -41.660, -42.269, -38.557, -35.696, -33.219, -30.590, -43.475, -43.821,
-43.626, -42.869, -41.689, -42.300, -38.585, -35.721, -33.243, -30.612, -43.503,
-43.849, -43.655, -42.897, -41.717, -42.331, -38.613, -35.746, -33.266, -30.633,
 -43.532, -43.878, -43.684, -42.926, -41.746, -42.363, -38.640, -35.771,
-33.290, -30.655, -43.560, -43.906, -43.713, -42.955, -41.774, -42.394, -38.668,
-35.797, -33.313, -30.677, -43.588, -43.935, -43.742, -42.984, -41.803, -42.425,
-38.696, -35.822, -33.337, -30.698, -43.616, -43.964, -43.770, -43.012, -41.831,
-42.456, -38.723, -35.847, -33.360, -30.720, -43.644, -43.992, -43.799, -43.041,
-41.860, -42.488, -38.751, -35.872, -33.384, -30.741, -43.672, -44.021, -43.828,
-43.070, -41.888, -42.519, -38.779, -35.897, -33.407, -30.763, -43.700, -44.049,
-43.857, -43.099, -41.917, -42.550, -38.807, -35.923, -33.431, -30.785, -43.728,
-44.078, -43.886, -43.128, -41.945, -42.582, -38.834, -35.948, -33.454, -30.806,
 -43.757, -44.107, -43.914, -43.156, -41.974, -42.613, -38.862, -35.973,
-33.477, -30.828, -43.785, -44.135, -43.943, -43.185, -42.002, -42.644, -38.890,
-35.998, -33.501, -30.850, -43.813, -44.164, -43.972, -43.214, -42.031, -42.676,
-38.917, -36.024, -33.524, -30.871, -43.841, -44.192, -44.001, -43.243, -42.059,
-42.707, -38.945, -36.049, -33.548, -30.893, -43.869, -44.221, -44.030, -43.271,
-42.088, -42.738, -38.973, -36.074, -33.571, -30.914, -43.897, -44.250, -44.059,
-43.300, -42.116, -42.769, -39.000, -36.099, -33.595, -30.936, -43.925, -44.278,
-44.087, -43.329, -42.145, -42.801, -39.028, -36.124, -33.618, -30.958, -43.954,
-44.307, -44.116, -43.358, -42.173, -42.832, -39.056, -36.150, -33.642, -30.979,
 -43.982, -44.335, -44.145, -43.387, -42.202, -42.863, -39.084, -36.175,
-33.665, -31.001, -44.010, -44.364, -44.174, -43.415, -42.230, -42.895, -39.111,
-36.200, -33.689, -31.023, -44.038, -44.393, -44.203, -43.444, -42.259, -42.926,
-39.139, -36.225, -33.712, -31.044, -44.066, -44.421, -44.231, -43.473, -42.287,
-42.957, -39.167, -36.251, -33.736, -31.066, -44.094, -44.450, -44.260, -43.502,
-42.316, -42.988, -39.194, -36.276, -33.759, -31.087, -44.122, -44.478, -44.289,
-43.530, -42.344, -43.020, -39.222, -36.301, -33.782, -31.109, -44.150, -44.507,
-44.318, -43.559, -42.373, -43.051, -39.250, -36.326, -33.806, -31.131, -44.179,
-44.536, -44.347, -43.588, -42.401, -43.082, -39.277, -36.352, -33.829, -31.152,
 -44.207, -44.564, -44.376, -43.617, -42.430, -43.114, -39.305, -36.377,
-33.853, -31.174, -44.235, -44.593, -44.404, -43.646, -42.458, -43.145, -39.333,
-36.402, -33.876, -31.196, -44.263, -44.621, -44.433, -43.674, -42.487, -43.176,
-39.361, -36.427, -33.900, -31.217, -44.291, -44.650, -44.462, -43.703, -42.515,
-43.208, -39.388, -36.452, -33.923, -31.239, -44.319, -44.679, -44.491, -43.732,
-42.544, -43.239, -39.416, -36.478, -33.947, -31.260, -44.347, -44.707, -44.520,
-43.761, -42.572, -43.270, -39.444, -36.503, -33.970, -31.282, -44.375, -44.736,
-44.549, -43.789, -42.601, -43.302, -39.471, -36.528, -33.994, -31.304, -44.404,
-44.764, -44.577, -43.818, -42.629, -43.333, -39.499, -36.553, -34.017, -31.325,
 -44.432, -44.793, -44.606, -43.847, -42.658, -43.364, -39.527, -36.579,
-34.041, -31.347, -44.460, -44.822, -44.635, -43.876, -42.686, -43.395, -39.554,
-36.604, -34.064, -31.368, -44.488, -44.850, -44.664, -43.905, -42.715, -43.427,
-39.582, -36.629, -34.088, -31.390, -44.516, -44.879, -44.693, -43.933, -42.743,
-43.458, -39.610, -36.654, -34.111, -31.412, -44.544, -44.907, -44.721, -43.962,
-42.772, -43.489, -39.638, -36.680, -34.135, -31.433, -44.572, -44.936, -44.750,
-43.991, -42.800, -43.521, -39.665, -36.705, -34.158, -31.455, -44.600, -44.965,
-44.779, -44.020, -42.829, -43.552, -39.693, -36.730, -34.181, -31.476, -44.629,
-44.993, -44.808, -44.049, -42.857, -43.583, -39.721, -36.755, -34.205, -31.498,
 -44.657, -45.022, -44.837, -44.077, -42.886, -43.615, -39.748, -36.780,
-34.228, -31.520, -44.685, -45.050, -44.866, -44.106, -42.914, -43.646, -39.776,
-36.806, -34.252, -31.541, -44.713, -45.079, -44.894, -44.135, -42.943, -43.677,
-39.804, -36.831, -34.275, -31.563, -44.741, -45.108, -44.923, -44.164, -42.971,
-43.708, -39.832, -36.856, -34.299, -31.584, -44.769, -45.136, -44.952, -44.192,
-43.000, -43.740, -39.859, -36.881, -34.322, -31.606, -44.797, -45.165, -44.981,
-44.221, -43.028, -43.771, -39.887, -36.907, -34.346, -31.628, -44.825, -45.193,
-45.010, -44.250, -43.057, -43.802, -39.915, -36.932, -34.369, -31.649, -44.854,
-45.222, -45.039, -44.279, -43.085, -43.834, -39.942, -36.957, -34.393, -31.671,
 -44.882, -45.251, -45.067, -44.308, -43.114, -43.865, -39.970, -36.982,
-34.416, -31.692, -44.910, -45.279, -45.096, -44.336, -43.142, -43.896, -39.998,
-37.008, -34.440, -31.714, -44.938, -45.308, -45.125, -44.365, -43.171, -43.928,
-40.026, -37.033, -34.463, -31.736, -44.966, -45.336, -45.154, -44.394, -43.199,
-43.959, -40.053, -37.058, -34.487, -31.757, -44.994, -45.365, -45.183, -44.423,
-43.228, -43.990, -40.081, -37.083, -34.510, -31.779, -45.022, -45.394, -45.212,
-44.452, -43.256, -44.022, -40.109, -37.109, -34.534, -31.800, -45.050, -45.422,
-45.240, -44.480, -43.285, -44.053, -40.136, -37.134, -34.557, -31.822, -45.079,
-45.451, -45.269, -44.509, -43.314, -44.084, -40.164, -37.159, -34.581, -31.844,
 -45.107, -45.479, -45.298, -44.538, -43.342, -44.116, -40.192, -37.184,
-34.604, -31.865, -45.135, -45.508, -45.327, -44.567, -43.371, -44.147, -40.219,
-37.210, -34.628, -31.887, -45.163, -45.537, -45.356, -44.596, -43.399, -44.178,
-40.247, -37.235, -34.651, -31.908, -45.191, -45.565, -45.385, -44.624, -43.428,
-44.209, -40.275, -37.260, -34.674, -31.930, -45.219, -45.594, -45.413, -44.653,
-43.456, -44.241, -40.303, -37.285, -34.698, -31.952, -45.247, -45.622, -45.442,
-44.682, -43.485, -44.272, -40.330, -37.311, -34.721, -31.973, -45.275, -45.651,
-45.471, -44.711, -43.513, -44.303, -40.358, -37.336, -34.745, -31.995, -45.304,
-45.680, -45.500, -44.739, -43.542, -44.335, -40.386, -37.361, -34.768, -32.016,
 -45.332, -45.708, -45.529, -44.768, -43.570, -44.366, -40.413, -37.386,
-34.792, -32.038, -45.360, -45.737, -45.558, -44.797, -43.599, -44.397, -40.441,
-37.411, -34.815, -32.059, -45.388, -45.765, -45.586, -44.826, -43.627, -44.429,
-40.469, -37.437, -34.839, -32.081, -45.416, -45.794, -45.615, -44.855, -43.656,
-44.460, -40.497, -37.462, -34.862, -32.103, -45.444, -45.823, -45.644, -44.883,
-43.684, -44.491, -40.524, -37.487, -34.886, -32.124, -45.472, -45.851, -45.673,
-44.912, -43.713, -44.523, -40.552, -37.512, -34.909, -32.146, -45.500, -45.880,
-45.702, -44.941, -43.741, -44.554, -40.580, -37.538, -34.933, -32.167, -45.528,
-45.908, -45.730, -44.970, -43.770, -44.585, -40.607, -37.563, -34.956, -32.189,
 -45.557, -45.937, -45.759, -44.999, -43.798, -44.617, -40.635, -37.588,
-34.980, -32.210, -45.585, -45.966, -45.788, -45.027, -43.827, -44.648, -40.663,
-37.613, -35.003, -32.232, -45.613, -45.994, -45.817, -45.056, -43.855, -44.679,
-40.691, -37.639, -35.027, -32.254, -45.641, -46.023, -45.846, -45.085, -43.884,
-44.710, -40.718, -37.664, -35.050, -32.275, -45.669, -46.051, -45.875, -45.114,
-43.912, -44.742, -40.746, -37.689, -35.074, -32.297, -45.697, -46.080, -45.903,
-45.143, -43.941, -44.773, -40.774, -37.714, -35.097, -32.318, -45.725, -46.109,
-45.932, -45.171, -43.969, -44.804, -40.801, -37.740, -35.121, -32.340, -45.753,
-46.137, -45.961, -45.200, -43.998, -44.836, -40.829, -37.765, -35.144, -32.361,
 -45.782, -46.166, -45.990, -45.229, -44.026, -44.867, -40.857, -37.790,
-35.168, -32.383, -45.810, -46.194, -46.019, -45.258, -44.055, -44.898, -40.885,
-37.815, -35.191, -32.405, -45.838, -46.223, -46.048, -45.287, -44.083, -44.930,
-40.912, -37.841, -35.215, -32.426, -45.866, -46.252, -46.076, -45.315, -44.112,
-44.961, -40.940, -37.866, -35.238, -32.448, -45.894, -46.280, -46.105, -45.344,
-44.140, -44.992, -40.968, -37.891, -35.262, -32.469, -45.922, -46.309, -46.134,
-45.373, -44.169, -45.024, -40.996, -37.916, -35.285, -32.491, -45.950, -46.337,
-46.163, -45.402, -44.197, -45.055, -41.023, -37.942, -35.309, -32.512, -45.978,
-46.366, -46.192, -45.430, -44.226, -45.086, -41.051, -37.967, -35.332, -32.534,
 -46.007, -46.395, -46.221, -45.459, -44.254, -45.118, -41.079, -37.992,
-35.356, -32.555, -46.035, -46.423, -46.250, -45.488, -44.283, -45.149, -41.106,
-38.017, -35.379, -32.577, -46.063, -46.452, -46.278, -45.517, -44.311, -45.180,
-41.134, -38.043, -35.403, -32.599, -46.091, -46.480, -46.307, -45.546, -44.340,
-45.212, -41.162, -38.068, -35.426, -32.620, -46.119, -46.509, -46.336, -45.574,
-44.369, -45.243, -41.190, -38.093, -35.450, -32.642, -46.147, -46.538, -46.365,
-45.603, -44.397, -45.274, -41.217, -38.118, -35.473, -32.663, -46.175, -46.566,
-46.394, -45.632, -44.426, -45.306, -41.245, -38.144, -35.497, -32.685, -46.203,
-46.595, -46.423, -45.661, -44.454, -45.337, -41.273, -38.169, -35.520, -32.706,
 -46.231, -46.624, -46.451, -45.690, -44.483, -45.368, -41.300, -38.194,
-35.544, -32.728, -46.260, -46.652, -46.480, -45.718, -44.511, -45.400, -41.328,
-38.219, -35.567, -32.749, -46.288, -46.681, -46.509, -45.747, -44.540, -45.431,
-41.356, -38.245, -35.591, -32.771, -46.316, -46.709, -46.538, -45.776, -44.568,
-45.462, -41.384, -38.270, -35.614, -32.792, -46.344, -46.738, -46.567, -45.805,
-44.597, -45.494, -41.411, -38.295, -35.638, -32.814, -46.372, -46.767, -46.596,
-45.834, -44.625, -45.525, -41.439, -38.320, -35.661, -32.836, -46.400, -46.795,
-46.624, -45.862, -44.654, -45.556, -41.467, -38.346, -35.685, -32.857, -46.428,
-46.824, -46.653, -45.891, -44.682, -45.587, -41.494, -38.371, -35.708, -32.879,
 -46.456, -46.852, -46.682, -45.920, -44.711, -45.619, -41.522, -38.396,
-35.732, -32.900, -46.485, -46.881, -46.711, -45.949, -44.739, -45.650, -41.550,
-38.421, -35.755, -32.922, -46.513, -46.910, -46.740, -45.978, -44.768, -45.681,
-41.578, -38.447, -35.779, -32.943, -46.541, -46.938, -46.769, -46.006, -44.796,
-45.713, -41.605, -38.472, -35.802, -32.965, -46.569, -46.967, -46.797, -46.035,
-44.825, -45.744, -41.633, -38.497, -35.826, -32.986, -46.597, -46.995, -46.826,
-46.064, -44.853, -45.775, -41.661, -38.522, -35.849, -33.008, -46.625, -47.024,
-46.855, -46.093, -44.882, -45.807, -41.689, -38.548, -35.873, -33.029, -46.653,
-47.053, -46.884, -46.122, -44.910, -45.838, -41.716, -38.573, -35.896, -33.051,
 -46.681, -47.081, -46.913, -46.150, -44.939, -45.869, -41.744, -38.598,
-35.920, -33.072, -46.710, -47.110, -46.942, -46.179, -44.967, -45.901, -41.772,
-38.624, -35.943, -33.094, -46.738, -47.139, -46.970, -46.208, -44.996, -45.932,
-41.799, -38.649, -35.967, -33.115, -46.766, -47.167, -46.999, -46.237, -45.024,
-45.963, -41.827, -38.674, -35.990, -33.137, -46.794, -47.196, -47.028, -46.266,
-45.053, -45.995, -41.855, -38.699, -36.014, -33.158, -46.822, -47.224, -47.057,
-46.294, -45.081, -46.026, -41.883, -38.725, -36.037, -33.180, -46.850, -47.253,
-47.086, -46.323, -45.110, -46.057, -41.910, -38.750, -36.061, -33.201, -46.878,
-47.282, -47.115, -46.352, -45.139, -46.089, -41.938, -38.775, -36.084, -33.223,
 -46.906, -47.310, -47.144, -46.381, -45.167, -46.120, -41.966, -38.800,
-36.108, -33.244, -46.934, -47.339, -47.172, -46.410, -45.196, -46.151, -41.994,
-38.826, -36.131, -33.266, -46.963, -47.367, -47.201, -46.438, -45.224, -46.183,
-42.021, -38.851, -36.155, -33.287, -46.991, -47.396, -47.230, -46.467, -45.253,
-46.214, -42.049, -38.876, -36.178, -33.309, -47.019, -47.425, -47.259, -46.496,
-45.281, -46.245, -42.077, -38.901, -36.202, -33.330, -47.047, -47.453, -47.288,
-46.525, -45.310, -46.277, -42.104, -38.927, -36.225, -33.352, -47.075, -47.482,
-47.317, -46.554, -45.338, -46.308, -42.132, -38.952, -36.249, -33.373, -47.103,
-47.510, -47.345, -46.582, -45.367, -46.339, -42.160, -38.977, -36.272, -33.395,
 -47.131, -47.539, -47.374, -46.611, -45.395, -46.371, -42.188, -39.002,
-36.296, -33.416, -47.159, -47.568, -47.403, -46.640, -45.424, -46.402, -42.215,
-39.028, -36.319, -33.438, -47.188, -47.596, -47.432, -46.669, -45.452, -46.433,
-42.243, -39.053, -36.343, -33.459, -47.216, -47.625, -47.461, -46.698, -45.481,
-46.465, -42.271, -39.078, -36.366, -33.481, -47.244, -47.654, -47.490, -46.726,
-45.509, -46.496, -42.299, -39.103, -36.390, -33.502, -47.272, -47.682, -47.519,
-46.755, -45.538, -46.527, -42.326, -39.129, -36.413, -33.524, -47.300, -47.711,
-47.547, -46.784, -45.566, -46.559, -42.354, -39.154, -36.437, -33.545, -47.328,
-47.739, -47.576, -46.813, -45.595, -46.590, -42.382, -39.179, -36.460, -33.567,
 -47.356, -47.768, -47.605, -46.842, -45.623, -46.621, -42.409, -39.205,
-36.484, -33.588, -47.384, -47.797, -47.634, -46.871, -45.652, -46.653, -42.437,
-39.230, -36.507, -33.610, -47.412, -47.825, -47.663, -46.899, -45.681, -46.684,
-42.465, -39.255, -36.531, -33.631, -47.441, -47.854, -47.692, -46.928, -45.709,
-46.715, -42.493, -39.280, -36.554, -33.653, -47.469, -47.882, -47.720, -46.957,
-45.738, -46.747, -42.520, -39.306, -36.578, -33.674, -47.497, -47.911, -47.749,
-46.986, -45.766, -46.778, -42.548, -39.331, -36.601, -33.696, -47.525, -47.940,
-47.778, -47.015, -45.795, -46.809, -42.576, -39.356, -36.625, -33.717, -47.553,
-47.968, -47.807, -47.043, -45.823, -46.841, -42.604, -39.381, -36.648, -33.739,
 -47.581, -47.997, -47.836, -47.072, -45.852, -46.872, -42.631, -39.407,
-36.672, -33.760, -47.609, -48.026, -47.865, -47.101, -45.880, -46.903, -42.659,
-39.432, -36.695, -33.782, -47.637, -48.054, -47.894, -47.130, -45.909, -46.935,
-42.687, -39.457, -36.719, -33.803, -47.666, -48.083, -47.922, -47.159, -45.937,
-46.966, -42.715, -39.482, -36.742, -33.824, -47.694, -48.111, -47.951, -47.187,
-45.966, -46.997, -42.742, -39.508, -36.766, -33.846, -47.722, -48.140, -47.980,
-47.216, -45.994, -47.029, -42.770, -39.533, -36.789, -33.867, -47.750, -48.169,
-48.009, -47.245, -46.023, -47.060, -42.798, -39.558, -36.813, -33.889, -47.778,
-48.197, -48.038, -47.274, -46.051, -47.091, -42.825, -39.584, -36.836, -33.910,
 -47.806, -48.226, -48.067, -47.303, -46.080, -47.123, -42.853, -39.609,
-36.860, -33.932, -47.834, -48.255, -48.095, -47.331, -46.108, -47.154, -42.881,
-39.634, -36.883, -33.953, -47.862, -48.283, -48.124, -47.360, -46.137, -47.185,
-42.909, -39.659, -36.907, -33.975, -47.890, -48.312, -48.153, -47.389, -46.165,
-47.217, -42.936, -39.685, -36.930, -33.996, -47.919, -48.340, -48.182, -47.418,
-46.194, -47.248, -42.964, -39.710, -36.954, -34.018, -47.947, -48.369, -48.211,
-47.447, -46.223, -47.279, -42.992, -39.735, -36.977, -34.039, -47.975, -48.398,
-48.240, -47.475, -46.251, -47.311, -43.020, -39.760, -37.001, -34.060, -48.003,
-48.426, -48.269, -47.504, -46.280, -47.342, -43.047, -39.786, -37.025, -34.082,
 -48.031, -48.455, -48.297, -47.533, -46.308, -47.373, -43.075, -39.811,
-37.048, -34.103, -48.059, -48.483, -48.326, -47.562, -46.337, -47.405, -43.103,
-39.836, -37.072, -34.125, -48.087, -48.512, -48.355, -47.591, -46.365, -47.436,
-43.131, -39.861, -37.095, -34.146, -48.115, -48.541, -48.384, -47.620, -46.394,
-47.467, -43.158, -39.887, -37.119, -34.168, -48.144, -48.569, -48.413, -47.648,
-46.422, -47.499, -43.186, -39.912, -37.142, -34.189, -48.172, -48.598, -48.442,
-47.677, -46.451, -47.530, -43.214, -39.937, -37.166, -34.211, -48.200, -48.627,
-48.471, -47.706, -46.479, -47.561, -43.241, -39.963, -37.189, -34.232, -48.228,
-48.655, -48.499, -47.735, -46.508, -47.593, -43.269, -39.988, -37.213, -34.253,
 -48.256, -48.684, -48.528, -47.764, -46.536, -47.624, -43.297, -40.013,
-37.236, -34.275, -48.284, -48.712, -48.557, -47.792, -46.565, -47.655, -43.325,
-40.038, -37.260, -34.296, -48.312, -48.741, -48.586, -47.821, -46.593, -47.687,
-43.352, -40.064, -37.283, -34.318, -48.340, -48.770, -48.615, -47.850, -46.622,
-47.718, -43.380, -40.089, -37.307, -34.339, -48.368, -48.798, -48.644, -47.879,
-46.651, -47.750, -43.408, -40.114, -37.330, -34.361, -48.397, -48.827, -48.673,
-47.908, -46.679, -47.781, -43.436, -40.139, -37.354, -34.382, -48.425, -48.856,
-48.701, -47.936, -46.708, -47.812, -43.463, -40.165, -37.377, -34.403, -48.453,
-48.884, -48.730, -47.965, -46.736, -47.844, -43.491, -40.190, -37.401, -34.425,
 -48.481, -48.913, -48.759, -47.994, -46.765, -47.875, -43.519, -40.215,
-37.424, -34.446, -48.509, -48.941, -48.788, -48.023, -46.793, -47.906, -43.547,
-40.241, -37.448, -34.468, -48.537, -48.970, -48.817, -48.052, -46.822, -47.938,
-43.574, -40.266, -37.471, -34.489, -48.565, -48.999, -48.846, -48.081, -46.850,
-47.969, -43.602, -40.291, -37.495, -34.510, -48.593, -49.027, -48.874, -48.109,
-46.879, -48.000, -43.630, -40.316, -37.518, -34.532, -48.622, -49.056, -48.903,
-48.138, -46.907, -48.032, -43.658, -40.342, -37.542, -34.553, -48.650, -49.085,
-48.932, -48.167, -46.936, -48.063, -43.685, -40.367, -37.565, -34.575, -48.678,
-49.113, -48.961, -48.196, -46.964, -48.094, -43.713, -40.392, -37.589, -34.596,
 -48.706, -49.142, -48.990, -48.225, -46.993, -48.126, -43.741, -40.418,
-37.612, -34.617, -48.734, -49.170, -49.019, -48.253, -47.021, -48.157, -43.769,
-40.443, -37.636, -34.639, -48.762, -49.199, -49.048, -48.282, -47.050, -48.188,
-43.796, -40.468, -37.660, -34.660, -48.790, -49.228, -49.076, -48.311, -47.079,
-48.220, -43.824, -40.493, -37.683, -34.682, -48.818, -49.256, -49.105, -48.340,
-47.107, -48.251, -43.852, -40.519, -37.707, -34.703, -48.846, -49.285, -49.134,
-48.369, -47.136, -48.282, -43.880, -40.544, -37.730, -34.724, -48.875, -49.314,
-49.163, -48.398, -47.164, -48.314, -43.907, -40.569, -37.754, -34.746, -48.903,
-49.342, -49.192, -48.426, -47.193, -48.345, -43.935, -40.594, -37.777, -34.767,
 -48.931, -49.371, -49.221, -48.455, -47.221, -48.376, -43.963, -40.620,
-37.801, -34.788, -48.959, -49.399, -49.250, -48.484, -47.250, -48.408, -43.990,
-40.645, -37.824, -34.810, -48.987, -49.428, -49.278, -48.513, -47.278, -48.439,
-44.018, -40.670, -37.848, -34.831, -49.015, -49.457, -49.307, -48.542, -47.307,
-48.470, -44.046, -40.696, -37.871, -34.853, -49.043, -49.485, -49.336, -48.570,
-47.335, -48.502, -44.074, -40.721, -37.895, -34.874, -49.071, -49.514, -49.365,
-48.599, -47.364, -48.533, -44.101, -40.746, -37.918, -34.895, -49.100, -49.543,
-49.394, -48.628, -47.392, -48.565, -44.129, -40.771, -37.942, -34.917, -49.128,
-49.571, -49.423, -48.657, -47.421, -48.596, -44.157, -40.797, -37.965, -34.938,
 -49.156, -49.600, -49.452, -48.686, -47.450, -48.627, -44.185, -40.822,
-37.989, -34.959, -49.184, -49.628, -49.481, -48.715, -47.478, -48.659, -44.212,
-40.847, -38.012, -34.981, -49.212, -49.657, -49.509, -48.743, -47.507, -48.690,
-44.240, -40.873, -38.036, -35.002, -49.240, -49.686, -49.538, -48.772, -47.535,
-48.721, -44.268, -40.898, -38.059, -35.024, -49.268, -49.714, -49.567, -48.801,
-47.564, -48.753, -44.296, -40.923, -38.083, -35.045, -49.296, -49.743, -49.596,
-48.830, -47.592, -48.784, -44.323, -40.948, -38.106, -35.066, -49.325, -49.772,
-49.625, -48.859, -47.621, -48.815, -44.351, -40.974, -38.130, -35.088, -49.353,
-49.800, -49.654, -48.887, -47.649, -48.847, -44.379, -40.999, -38.154, -35.109,
 -49.381, -49.829, -49.683, -48.916, -47.678, -48.878, -44.407, -41.024,
-38.177, -35.130, -49.409, -49.858, -49.711, -48.945, -47.706, -48.909, -44.434,
-41.050, -38.201, -35.152, -49.437, -49.886, -49.740, -48.974, -47.735, -48.941,
-44.462, -41.075, -38.224, -35.173, -49.465, -49.915, -49.769, -49.003, -47.763,
-48.972, -44.490, -41.100, -38.248, -35.194, -49.493, -49.943, -49.798, -49.032,
-47.792, -49.003, -44.518, -41.125, -38.271, -35.216, -49.521, -49.972, -49.827,
-49.060, -47.821, -49.035, -44.545, -41.151, -38.295, -35.237, -49.549, -50.001,
-49.856, -49.089, -47.849, -49.066, -44.573, -41.176, -38.318, -35.258, -49.578,
-50.029, -49.885, -49.118, -47.878, -49.097, -44.601, -41.201, -38.342, -35.280,
 -49.606, -50.058, -49.913, -49.147, -47.906, -49.129, -44.629, -41.227,
-38.365, -35.301, -49.634, -50.087, -49.942, -49.176, -47.935, -49.160, -44.656,
-41.252, -38.389, -35.322, -49.662, -50.115, -49.971, -49.204, -47.963, -49.192,
-44.684, -41.277, -38.412, -35.344, -49.690, -50.144, -50.000, -49.233, -47.992,
-49.223, -44.712, -41.302, -38.436, -35.365, -49.718, -50.172, -50.029, -49.262,
-48.020, -49.254, -44.740, -41.328, -38.459, -35.386, -49.746, -50.201, -50.058,
-49.291, -48.049, -49.286, -44.767, -41.353, -38.483, -35.408, -49.774, -50.230,
-50.087, -49.320, -48.077, -49.317, -44.795, -41.378, -38.506, -35.429, -49.803,
-50.258, -50.115, -49.349, -48.106, -49.348, -44.823, -41.404, -38.530, -35.450,
 -49.831, -50.287, -50.144, -49.377, -48.135, -49.380, -44.851, -41.429,
-38.553, -35.472, -49.859, -50.316, -50.173, -49.406, -48.163, -49.411, -44.878,
-41.454, -38.577, -35.493, -49.887, -50.344, -50.202, -49.435, -48.192, -49.442,
-44.906, -41.479, -38.601, -35.514, -49.915, -50.373, -50.231, -49.464, -48.220,
-49.474, -44.934, -41.505, -38.624, -35.535, -49.943, -50.402, -50.260, -49.493,
-48.249, -49.505, -44.962, -41.530, -38.648, -35.557, -49.971, -50.430, -50.289,
-49.522, -48.277, -49.536, -44.989, -41.555, -38.671, -35.578, -49.999, -50.459,
-50.318, -49.550, -48.306, -49.568, -45.017, -41.581, -38.695, -35.599, -50.027,
-50.487, -50.346, -49.579, -48.334, -49.599, -45.045, -41.606, -38.718, -35.621,
 -50.056, -50.516, -50.375, -49.608, -48.363, -49.631, -45.073, -41.631,
-38.742, -35.642, -50.084, -50.545, -50.404, -49.637, -48.391, -49.662, -45.100,
-41.656, -38.765, -35.663, -50.112, -50.573, -50.433, -49.666, -48.420, -49.693,
-45.128, -41.682, -38.789, -35.685, -50.140, -50.602, -50.462, -49.694, -48.449,
-49.725, -45.156, -41.707, -38.812, -35.706, -50.168, -50.631, -50.491, -49.723,
-48.477, -49.756, -45.184, -41.732, -38.836, -35.727, -50.196, -50.659, -50.520,
-49.752, -48.506, -49.787, -45.211, -41.758, -38.859, -35.748, -50.224, -50.688,
-50.548, -49.781, -48.534, -49.819, -45.239, -41.783, -38.883, -35.770, -50.252,
-50.716, -50.577, -49.810, -48.563, -49.850, -45.267, -41.808, -38.906, -35.791,
 -50.281, -50.745, -50.606, -49.839, -48.591, -49.881, -45.295, -41.833,
-38.930, -35.812, -50.309, -50.774, -50.635, -49.867, -48.620, -49.913, -45.322,
-41.859, -38.954, -35.834, -50.337, -50.802, -50.664, -49.896, -48.648, -49.944,
-45.350, -41.884, -38.977, -35.855, -50.365, -50.831, -50.693, -49.925, -48.677,
-49.975, -45.378, -41.909, -39.001, -35.876, -50.393, -50.860, -50.722, -49.954,
-48.705, -50.007, -45.406, -41.935, -39.024, -35.897, -50.421, -50.888, -50.751,
-49.983, -48.734, -50.038, -45.433, -41.960, -39.048, -35.919, -50.449, -50.917,
-50.779, -50.012, -48.763, -50.070, -45.461, -41.985, -39.071, -35.940, -50.477,
-50.946, -50.808, -50.040, -48.791, -50.101, -45.489, -42.010, -39.095, -35.961,
 -50.506, -50.974, -50.837, -50.069, -48.820, -50.132, -45.517, -42.036,
-39.118, -35.982, -50.534, -51.003, -50.866, -50.098, -48.848, -50.164, -45.544,
-42.061, -39.142, -36.004, -50.562, -51.032, -50.895, -50.127, -48.877, -50.195,
-45.572, -42.086, -39.165, -36.025, -50.590, -51.060, -50.924, -50.156, -48.905,
-50.226, -45.600, -42.112, -39.189, -36.046, -50.618, -51.089, -50.953, -50.185,
-48.934, -50.258, -45.628, -42.137, -39.212, -36.067, -50.646, -51.117, -50.982,
-50.213, -48.962, -50.289, -45.655, -42.162, -39.236, -36.089, -50.674, -51.146,
-51.010, -50.242, -48.991, -50.320, -45.683, -42.187, -39.259, -36.110, -50.702,
-51.175, -51.039, -50.271, -49.019, -50.352, -45.711, -42.213, -39.283, -36.131,
 -50.731, -51.203, -51.068, -50.300, -49.048, -50.383, -45.739, -42.238,
-39.307, -36.152, -50.759, -51.232, -51.097, -50.329, -49.077, -50.415, -45.766,
-42.263, -39.330, -36.174, -50.787, -51.261, -51.126, -50.357, -49.105, -50.446,
-45.794, -42.289, -39.354, -36.195, -50.815, -51.289, -51.155, -50.386, -49.134,
-50.477, -45.822, -42.314, -39.377, -36.216, -50.843, -51.318, -51.184, -50.415,
-49.162, -50.509, -45.850, -42.339, -39.401, -36.237, -50.871, -51.347, -51.212,
-50.444, -49.191, -50.540, -45.877, -42.365, -39.424, -36.259, -50.899, -51.375,
-51.241, -50.473, -49.219, -50.571, -45.905, -42.390, -39.448, -36.280, -50.927,
-51.404, -51.270, -50.502, -49.248, -50.603, -45.933, -42.415, -39.471, -36.301,
 -50.955, -51.432, -51.299, -50.530, -49.276, -50.634, -45.961, -42.440,
-39.495, -36.322, -50.984, -51.461, -51.328, -50.559, -49.305, -50.665, -45.989,
-42.466, -39.518, -36.344, -51.012, -51.490, -51.357, -50.588, -49.334, -50.697,
-46.016, -42.491, -39.542, -36.365, -51.040, -51.518, -51.386, -50.617, -49.362,
-50.728, -46.044, -42.516, -39.565, -36.386, -51.068, -51.547, -51.415, -50.646,
-49.391, -50.760, -46.072, -42.542, -39.589, -36.407, -51.096, -51.576, -51.443,
-50.675, -49.419, -50.791, -46.100, -42.567, -39.613, -36.428, -51.124, -51.604,
-51.472, -50.703, -49.448, -50.822, -46.127, -42.592, -39.636, -36.450, -51.152,
-51.633, -51.501, -50.732, -49.476, -50.854, -46.155, -42.617, -39.660, -36.471,
 -51.180, -51.662, -51.530, -50.761, -49.505, -50.885, -46.183, -42.643,
-39.683, -36.492, -51.209, -51.690, -51.559, -50.790, -49.533, -50.916, -46.211,
-42.668, -39.707, -36.513, -51.237, -51.719, -51.588, -50.819, -49.562, -50.948,
-46.238, -42.693, -39.730, -36.534, -51.265, -51.748, -51.617, -50.848, -49.590,
-50.979, -46.266, -42.719, -39.754, -36.556, -51.293, -51.776, -51.646, -50.876,
-49.619, -51.010, -46.294, -42.744, -39.777, -36.577, -51.321, -51.805, -51.674,
-50.905, -49.648, -51.042, -46.322, -42.769, -39.801, -36.598, -51.349, -51.833,
-51.703, -50.934, -49.676, -51.073, -46.349, -42.795, -39.824, -36.619, -51.377,
-51.862, -51.732, -50.963, -49.705, -51.105, -46.377, -42.820, -39.848, -36.640,
 -51.405, -51.891, -51.761, -50.992, -49.733, -51.136, -46.405, -42.845,
-39.871, -36.662, -51.434, -51.919, -51.790, -51.021, -49.762, -51.167, -46.433,
-42.870, -39.895, -36.683, -51.462, -51.948, -51.819, -51.049, -49.790, -51.199,
-46.460, -42.896, -39.918, -36.704, -51.490, -51.977, -51.848, -51.078, -49.819,
-51.230, -46.488, -42.921, -39.942, -36.725, -51.518, -52.005, -51.877, -51.107,
-49.847, -51.261, -46.516, -42.946, -39.966, -36.746, -51.546, -52.034, -51.905,
-51.136, -49.876, -51.293, -46.544, -42.972, -39.989, -36.767, -51.574, -52.063,
-51.934, -51.165, -49.905, -51.324, -46.571, -42.997, -40.013, -36.789, -51.602,
-52.091, -51.963, -51.194, -49.933, -51.355, -46.599, -43.022, -40.036, -36.810,
 -51.630, -52.120, -51.992, -51.222, -49.962, -51.387, -46.627, -43.048,
-40.060, -36.831, -51.659, -52.149, -52.021, -51.251, -49.990, -51.418, -46.655,
-43.073, -40.083, -36.852, -51.687, -52.177, -52.050, -51.280, -50.019, -51.450,
-46.683, -43.098, -40.107, -36.873, -51.715, -52.206, -52.079, -51.309, -50.047,
-51.481, -46.710, -43.123, -40.130, -36.894, -51.743, -52.234, -52.108, -51.338,
-50.076, -51.512, -46.738, -43.149, -40.154, -36.916, -51.771, -52.263, -52.136,
-51.367, -50.104, -51.544, -46.766, -43.174, -40.177, -36.937, -51.799, -52.292,
-52.165, -51.395, -50.133, -51.575, -46.794, -43.199, -40.201, -36.958, -51.827,
-52.320, -52.194, -51.424, -50.162, -51.606, -46.821, -43.225, -40.225, -36.979,
 -51.855, -52.349, -52.223, -51.453, -50.190, -51.638, -46.849, -43.250,
-40.248, -37.000, -51.884, -52.378, -52.252, -51.482, -50.219, -51.669, -46.877,
-43.275, -40.272, -37.021, -51.912, -52.406, -52.281, -51.511, -50.247, -51.701,
-46.905, -43.301, -40.295, -37.043, -51.940, -52.435, -52.310, -51.540, -50.276,
-51.732, -46.932, -43.326, -40.319, -37.064, -51.968, -52.464, -52.339, -51.569,
-50.304, -51.763, -46.960, -43.351, -40.342, -37.085, -51.996, -52.492, -52.368,
-51.597, -50.333, -51.795, -46.988, -43.376, -40.366, -37.106, -52.024, -52.521,
-52.396, -51.626, -50.361, -51.826, -47.016, -43.402, -40.389, -37.127, -52.052,
-52.550, -52.425, -51.655, -50.390, -51.857, -47.043, -43.427, -40.413, -37.148,
 -52.080, -52.578, -52.454, -51.684, -50.419, -51.889, -47.071, -43.452,
-40.436, -37.169, -52.109, -52.607, -52.483, -51.713, -50.447, -51.920, -47.099,
-43.478, -40.460, -37.190, -52.137, -52.636, -52.512, -51.742, -50.476, -51.952,
-47.127, -43.503, -40.483, -37.212, -52.165, -52.664, -52.541, -51.770, -50.504,
-51.983, -47.155, -43.528, -40.507, -37.233, -52.193, -52.693, -52.570, -51.799,
-50.533, -52.014, -47.182, -43.554, -40.531, -37.254, -52.221, -52.721, -52.599,
-51.828, -50.561, -52.046, -47.210, -43.579, -40.554, -37.275, -52.249, -52.750,
-52.627, -51.857, -50.590, -52.077, -47.238, -43.604, -40.578, -37.296, -52.277,
-52.779, -52.656, -51.886, -50.619, -52.108, -47.266, -43.630, -40.601, -37.317,
 -52.305, -52.807, -52.685, -51.915, -50.647, -52.140, -47.293, -43.655,
-40.625, -37.338, -52.334, -52.836, -52.714, -51.943, -50.676, -52.171, -47.321,
-43.680, -40.648, -37.359, -52.362, -52.865, -52.743, -51.972, -50.704, -52.203,
-47.349, -43.705, -40.672, -37.380, -52.390, -52.893, -52.772, -52.001, -50.733,
-52.234, -47.377, -43.731, -40.695, -37.402, -52.418, -52.922, -52.801, -52.030,
-50.761, -52.265, -47.404, -43.756, -40.719, -37.423, -52.446, -52.951, -52.830,
-52.059, -50.790, -52.297, -47.432, -43.781, -40.742, -37.444, -52.474, -52.979,
-52.859, -52.088, -50.818, -52.328, -47.460, -43.807, -40.766, -37.465, -52.502,
-53.008, -52.887, -52.116, -50.847, -52.359, -47.488, -43.832, -40.790, -37.486,
 -52.530, -53.037, -52.916, -52.145, -50.876, -52.391, -47.515, -43.857,
-40.813, -37.507, -52.559, -53.065, -52.945, -52.174, -50.904, -52.422, -47.543,
-43.883, -40.837, -37.528, -52.587, -53.094, -52.974, -52.203, -50.933, -52.454,
-47.571, -43.908, -40.860, -37.549, -52.615, -53.123, -53.003, -52.232, -50.961,
-52.485, -47.599, -43.933, -40.884, -37.570, -52.643, -53.151, -53.032, -52.261,
-50.990, -52.516, -47.627, -43.959, -40.907, -37.591, -52.671, -53.180, -53.061,
-52.289, -51.018, -52.548, -47.654, -43.984, -40.931, -37.612, -52.699, -53.209,
-53.090, -52.318, -51.047, -52.579, -47.682, -44.009, -40.954, -37.633, -52.727,
-53.237, -53.118, -52.347, -51.075, -52.610, -47.710, -44.034, -40.978, -37.655,
 -52.755, -53.266, -53.147, -52.376, -51.104, -52.642, -47.738, -44.060,
-41.001, -37.676, -52.784, -53.294, -53.176, -52.405, -51.133, -52.673, -47.765,
-44.085, -41.025, -37.697, -52.812, -53.323, -53.205, -52.434, -51.161, -52.705,
-47.793, -44.110, -41.048, -37.718, -52.840, -53.352, -53.234, -52.463, -51.190,
-52.736, -47.821, -44.136, -41.072, -37.739, -52.868, -53.380, -53.263, -52.491,
-51.218, -52.767, -47.849, -44.161, -41.096, -37.760, -52.896, -53.409, -53.292,
-52.520, -51.247, -52.799, -47.876, -44.186, -41.119, -37.781, -52.924, -53.438,
-53.321, -52.549, -51.275, -52.830, -47.904, -44.212, -41.143, -37.802, -52.952,
-53.466, -53.350, -52.578, -51.304, -52.861, -47.932, -44.237, -41.166, -37.823,
 -52.981, -53.495, -53.378, -52.607, -51.333, -52.893, -47.960, -44.262,
-41.190, -37.844, -53.009, -53.524, -53.407, -52.636, -51.361, -52.924, -47.988,
-44.288, -41.213, -37.865, -53.037, -53.552, -53.436, -52.664, -51.390, -52.956,
-48.015, -44.313, -41.237, -37.886, -53.065, -53.581, -53.465, -52.693, -51.418,
-52.987, -48.043, -44.338, -41.260, -37.907, -53.093, -53.610, -53.494, -52.722,
-51.447, -53.018, -48.071, -44.363, -41.284, -37.928, -53.121, -53.638, -53.523,
-52.751, -51.475, -53.050, -48.099, -44.389, -41.307, -37.949, -53.149, -53.667,
-53.552, -52.780, -51.504, -53.081, -48.126, -44.414, -41.331, -37.970, -53.177,
-53.696, -53.581, -52.809, -51.533, -53.112, -48.154, -44.439, -41.355, -37.991,
 -53.206, -53.724, -53.610, -52.838, -51.561, -53.144, -48.182, -44.465,
-41.378, -38.012, -53.234, -53.753, -53.638, -52.866, -51.590, -53.175, -48.210,
-44.490, -41.402, -38.033, -53.262, -53.782, -53.667, -52.895, -51.618, -53.207,
-48.237, -44.515, -41.425, -38.054, -53.290, -53.810, -53.696, -52.924, -51.647,
-53.238, -48.265, -44.541, -41.449, -38.075, -53.318, -53.839, -53.725, -52.953,
-51.675, -53.269, -48.293, -44.566, -41.472, -38.096, -53.346, -53.868, -53.754,
-52.982, -51.704, -53.301, -48.321, -44.591, -41.496, -38.117, -53.374, -53.896,
-53.783, -53.011, -51.732, -53.332, -48.349, -44.617, -41.519, -38.138, -53.402,
-53.925, -53.812, -53.039, -51.761, -53.364, -48.376, -44.642, -41.543, -38.159,
 -53.431, -53.954, -53.841, -53.068, -51.790, -53.395, -48.404, -44.667,
-41.566, -38.180, -53.459, -53.982, -53.870, -53.097, -51.818, -53.426, -48.432,
-44.693, -41.590, -38.201, -53.487, -54.011, -53.898, -53.126, -51.847, -53.458,
-48.460, -44.718, -41.614, -38.222, -53.515, -54.039, -53.927, -53.155, -51.875,
-53.489, -48.487, -44.743, -41.637, -38.243, -53.543, -54.068, -53.956, -53.184,
-51.904, -53.520, -48.515, -44.768, -41.661, -38.264, -53.571, -54.097, -53.985,
-53.213, -51.932, -53.552, -48.543, -44.794, -41.684, -38.285, -53.599, -54.125,
-54.014, -53.241, -51.961, -53.583, -48.571, -44.819, -41.708, -38.306, -53.627,
-54.154, -54.043, -53.270, -51.990, -53.615, -48.599, -44.844, -41.731, -38.327,
 -53.656, -54.183, -54.072, -53.299, -52.018, -53.646, -48.626, -44.870,
-41.755, -38.348, -53.684, -54.211, -54.101, -53.328, -52.047, -53.677, -48.654,
-44.895, -41.778, -38.369, -53.712, -54.240, -54.130, -53.357, -52.075, -53.709,
-48.682, -44.920, -41.802, -38.390, -53.740, -54.269, -54.158, -53.386, -52.104,
-53.740, -48.710, -44.946, -41.825, -38.411, -53.768, -54.297, -54.187, -53.414,
-52.132, -53.771, -48.737, -44.971, -41.849, -38.432, -53.796, -54.326, -54.216,
-53.443, -52.161, -53.803, -48.765, -44.996, -41.873, -38.453, -53.824, -54.355,
-54.245, -53.472, -52.190, -53.834, -48.793, -45.022, -41.896, -38.474, -53.853,
-54.383, -54.274, -53.501, -52.218, -53.866, -48.821, -45.047, -41.920, -38.495,
 -53.881, -54.412, -54.303, -53.530, -52.247, -53.897, -48.849, -45.072,
-41.943, -38.516, -53.909, -54.441, -54.332, -53.559, -52.275, -53.928, -48.876,
-45.098, -41.967, -38.537, -53.937, -54.469, -54.361, -53.588, -52.304, -53.960,
-48.904, -45.123, -41.990, -38.558, -53.965, -54.498, -54.390, -53.616, -52.332,
-53.991, -48.932, -45.148, -42.014, -38.579, -53.993, -54.527, -54.418, -53.645,
-52.361, -54.023, -48.960, -45.174, -42.037, -38.600, -54.021, -54.555, -54.447,
-53.674, -52.390, -54.054, -48.987, -45.199, -42.061, -38.621, -54.049, -54.584,
-54.476, -53.703, -52.418, -54.085, -49.015, -45.224, -42.084, -38.642, -54.078,
-54.613, -54.505, -53.732, -52.447, -54.117, -49.043, -45.249, -42.108, -38.663,
 -54.106, -54.641, -54.534, -53.761, -52.475, -54.148, -49.071, -45.275,
-42.132, -38.684, -54.134, -54.670, -54.563, -53.790, -52.504, -54.180, -49.099,
-45.300, -42.155, -38.705, -54.162, -54.699, -54.592, -53.818, -52.532, -54.211,
-49.126, -45.325, -42.179, -38.726, -54.190, -54.727, -54.621, -53.847, -52.561,
-54.242, -49.154, -45.351, -42.202, -38.747, -54.218, -54.756, -54.650, -53.876,
-52.590, -54.274, -49.182, -45.376, -42.226, -38.768, -54.246, -54.785, -54.679,
-53.905, -52.618, -54.305, -49.210, -45.401, -42.249, -38.789, -54.275, -54.813,
-54.707, -53.934, -52.647, -54.336, -49.237, -45.427, -42.273, -38.809, -54.303,
-54.842, -54.736, -53.963, -52.675, -54.368, -49.265, -45.452, -42.296, -38.830,
 -54.331, -54.871, -54.765, -53.991, -52.704, -54.399, -49.293, -45.477,
-42.320, -38.851, -54.359, -54.899, -54.794, -54.020, -52.732, -54.431, -49.321,
-45.503, -42.343, -38.872, -54.387, -54.928, -54.823, -54.049, -52.761, -54.462,
-49.349, -45.528, -42.367, -38.893, -54.415, -54.957, -54.852, -54.078, -52.790,
-54.493, -49.376, -45.553, -42.391, -38.914, -54.443, -54.985, -54.881, -54.107,
-52.818, -54.525, -49.404, -45.579, -42.414, -38.935, -54.471, -55.014, -54.910,
-54.136, -52.847, -54.556, -49.432, -45.604, -42.438, -38.956, -54.500, -55.043,
-54.939, -54.165, -52.875, -54.588, -49.460, -45.629, -42.461, -38.977, -54.528,
-55.071, -54.967, -54.193, -52.904, -54.619, -49.487, -45.655, -42.485, -38.998,
 -54.556, -55.100, -54.996, -54.222, -52.932, -54.650, -49.515, -45.680,
-42.508, -39.019, -54.584, -55.129, -55.025, -54.251, -52.961, -54.682, -49.543,
-45.705, -42.532, -39.039, -54.612, -55.157, -55.054, -54.280, -52.990, -54.713,
-49.571, -45.731, -42.555, -39.060, -54.640, -55.186, -55.083, -54.309, -53.018,
-54.744, -49.599, -45.756, -42.579, -39.081, -54.668, -55.215, -55.112, -54.338,
-53.047, -54.776, -49.626, -45.781, -42.603, -39.102, -54.697, -55.243, -55.141,
-54.367, -53.075, -54.807, -49.654, -45.806, -42.626, -39.123, -54.725, -55.272,
-55.170, -54.395, -53.104, -54.839, -49.682, -45.832, -42.650, -39.144, -54.753,
-55.301, -55.199, -54.424, -53.132, -54.870, -49.710, -45.857, -42.673, -39.165,
 -54.781, -55.329, -55.228, -54.453, -53.161, -54.901, -49.737, -45.882,
-42.697, -39.186, -54.809, -55.358, -55.256, -54.482, -53.190, -54.933, -49.765,
-45.908, -42.720, -39.207, -54.837, -55.387, -55.285, -54.511, -53.218, -54.964,
-49.793, -45.933, -42.744, -39.227, -54.865, -55.415, -55.314, -54.540, -53.247,
-54.996, -49.821, -45.958, -42.767, -39.248, -54.893, -55.444, -55.343, -54.569,
-53.275, -55.027, -49.849, -45.984, -42.791, -39.269, -54.922, -55.473, -55.372,
-54.597, -53.304, -55.058, -49.876, -46.009, -42.814, -39.290, -54.950, -55.501,
-55.401, -54.626, -53.332, -55.090, -49.904, -46.034, -42.838, -39.311, -54.978,
-55.530, -55.430, -54.655, -53.361, -55.121, -49.932, -46.060, -42.862, -39.332,
 -55.006, -55.559, -55.459, -54.684, -53.390, -55.153, -49.960, -46.085,
-42.885, -39.353, -55.034, -55.587, -55.488, -54.713, -53.418, -55.184, -49.987,
-46.110, -42.909, -39.373, -55.062, -55.616, -55.517, -54.742, -53.447, -55.215,
-50.015, -46.136, -42.932, -39.394, -55.090, -55.645, -55.545, -54.771, -53.475,
-55.247, -50.043, -46.161, -42.956, -39.415, -55.119, -55.673, -55.574, -54.799,
-53.504, -55.278, -50.071, -46.186, -42.979, -39.436, -55.147, -55.702, -55.603,
-54.828, -53.532, -55.310, -50.099, -46.212, -43.003, -39.457, -55.175, -55.731,
-55.632, -54.857, -53.561, -55.341, -50.126, -46.237, -43.026, -39.478, -55.203,
-55.759, -55.661, -54.886, -53.590, -55.372, -50.154, -46.262, -43.050, -39.499,
 -55.231, -55.788, -55.690, -54.915, -53.618, -55.404, -50.182, -46.288,
-43.073, -39.519, -55.259, -55.817, -55.719, -54.944, -53.647, -55.435, -50.210,
-46.313, -43.097, -39.540, -55.287, -55.845, -55.748, -54.973, -53.675, -55.467,
-50.238, -46.338, -43.121, -39.561, -55.316, -55.874, -55.777, -55.001, -53.704,
-55.498, -50.265, -46.364, -43.144, -39.582, -55.344, -55.903, -55.806, -55.030,
-53.733, -55.529, -50.293, -46.389, -43.168, -39.603, -55.372, -55.931, -55.834,
-55.059, -53.761, -55.561, -50.321, -46.414, -43.191, -39.624, -55.400, -55.960,
-55.863, -55.088, -53.790, -55.592, -50.349, -46.440, -43.215, -39.644, -55.428,
-55.989, -55.892, -55.117, -53.818, -55.623, -50.376, -46.465, -43.238, -39.665,
 -55.456, -56.017, -55.921, -55.146, -53.847, -55.655, -50.404, -46.490,
-43.262, -39.686, -55.484, -56.046, -55.950, -55.175, -53.875, -55.686, -50.432,
-46.516, -43.285, -39.707, -55.512, -56.075, -55.979, -55.203, -53.904, -55.718,
-50.460, -46.541, -43.309, -39.728, -55.541, -56.103, -56.008, -55.232, -53.933,
-55.749, -50.488, -46.566, -43.333, -39.748, -55.569, -56.132, -56.037, -55.261,
-53.961, -55.780, -50.515, -46.592, -43.356, -39.769, -55.597, -56.161, -56.066,
-55.290, -53.990, -55.812, -50.543, -46.617, -43.380, -39.790, -55.625, -56.189,
-56.095, -55.319, -54.018, -55.843, -50.571, -46.642, -43.403, -39.811, -55.653,
-56.218, -56.123, -55.348, -54.047, -55.875, -50.599, -46.668, -43.427, -39.832,
 -55.681, -56.247, -56.152, -55.377, -54.075, -55.906, -50.627, -46.693,
-43.450, -39.852, -55.709, -56.275, -56.181, -55.405, -54.104, -55.937, -50.654,
-46.718, -43.474, -39.873, -55.738, -56.304, -56.210, -55.434, -54.133, -55.969,
-50.682, -46.743, -43.497, -39.894, -55.766, -56.333, -56.239, -55.463, -54.161,
-56.000, -50.710, -46.769, -43.521, -39.915, -55.794, -56.361, -56.268, -55.492,
-54.190, -56.032, -50.738, -46.794, -43.544, -39.936, -55.822, -56.390, -56.297,
-55.521, -54.218, -56.063, -50.765, -46.819, -43.568, -39.956, -55.850, -56.419,
-56.326, -55.550, -54.247, -56.094, -50.793, -46.845, -43.592, -39.977, -55.878,
-56.447, -56.355, -55.579, -54.276, -56.126, -50.821, -46.870, -43.615, -39.998,
 -55.906, -56.476, -56.384, -55.607, -54.304, -56.157, -50.849, -46.895,
-43.639, -40.019, -55.935, -56.505, -56.412, -55.636, -54.333, -56.189, -50.877,
-46.921, -43.662, -40.039, -55.963, -56.533, -56.441, -55.665, -54.361, -56.220,
-50.904, -46.946, -43.686, -40.060, -55.991, -56.562, -56.470, -55.694, -54.390,
-56.251, -50.932, -46.971, -43.709, -40.081, -56.019, -56.591, -56.499, -55.723,
-54.418, -56.283, -50.960, -46.997, -43.733, -40.102, -56.047, -56.619, -56.528,
-55.752, -54.447, -56.314, -50.988, -47.022, -43.756, -40.123, -56.075, -56.648,
-56.557, -55.781, -54.476, -56.346, -51.016, -47.047, -43.780, -40.143, -56.103,
-56.677, -56.586, -55.810, -54.504, -56.377, -51.043, -47.073, -43.803, -40.164,
 -56.132, -56.705, -56.615, -55.838, -54.533, -56.408, -51.071, -47.098,
-43.827, -40.185, -56.160, -56.734, -56.644, -55.867, -54.561, -56.440, -51.099,
-47.123, -43.851, -40.206, -56.188, -56.763, -56.673, -55.896, -54.590, -56.471,
-51.127, -47.149, -43.874, -40.226, -56.216, -56.791, -56.702, -55.925, -54.619,
-56.503, -51.155, -47.174, -43.898, -40.247, -56.244, -56.820, -56.730, -55.954,
-54.647, -56.534, -51.182, -47.199, -43.921, -40.268, -56.272, -56.849, -56.759,
-55.983, -54.676, -56.565, -51.210, -47.225, -43.945, -40.288, -56.300, -56.877,
-56.788, -56.012, -54.704, -56.597, -51.238, -47.250, -43.968, -40.309, -56.329,
-56.906, -56.817, -56.040, -54.733, -56.628, -51.266, -47.275, -43.992, -40.330,
 -56.357, -56.935, -56.846, -56.069, -54.761, -56.660, -51.293, -47.301,
-44.015, -40.351, -56.385, -56.963, -56.875, -56.098, -54.790, -56.691, -51.321,
-47.326, -44.039, -40.371, -56.413, -56.992, -56.904, -56.127, -54.819, -56.722,
-51.349, -47.351, -44.063, -40.392, -56.441, -57.021, -56.933, -56.156, -54.847,
-56.754, -51.377, -47.377, -44.086, -40.413, -56.469, -57.049, -56.962, -56.185,
-54.876, -56.785, -51.405, -47.402, -44.110, -40.434, -56.497, -57.078, -56.991,
-56.214, -54.904, -56.817, -51.432, -47.427, -44.133, -40.454, -56.526, -57.107,
-57.020, -56.242, -54.933, -56.848, -51.460, -47.453, -44.157, -40.475, -56.554,
-57.135, -57.048, -56.271, -54.962, -56.879, -51.488, -47.478, -44.180, -40.496,
 -56.582, -57.164, -57.077, -56.300, -54.990, -56.911, -51.516, -47.503,
-44.204, -40.516, -56.610, -57.193, -57.106, -56.329, -55.019, -56.942, -51.544,
-47.529, -44.227, -40.537, -56.638, -57.222, -57.135, -56.358, -55.047, -56.974,
-51.571, -47.554, -44.251, -40.558, -56.666, -57.250, -57.164, -56.387, -55.076,
-57.005, -51.599, -47.579, -44.274, -40.579, -56.694, -57.279, -57.193, -56.416,
-55.104, -57.036, -51.627, -47.605, -44.298, -40.599, -56.723, -57.308, -57.222,
-56.445, -55.133, -57.068, -51.655, -47.630, -44.322, -40.620, -56.751, -57.336,
-57.251, -56.473, -55.162, -57.099, -51.683, -47.655, -44.345, -40.641, -56.779,
-57.365, -57.280, -56.502, -55.190, -57.131, -51.710, -47.681, -44.369, -40.661,
 -56.807, -57.394, -57.309, -56.531, -55.219, -57.162, -51.738, -47.706,
-44.392, -40.682, -56.835, -57.422, -57.338, -56.560, -55.247, -57.193, -51.766,
-47.731, -44.416, -40.703, -56.863, -57.451, -57.366, -56.589, -55.276, -57.225,
-51.794, -47.757, -44.439, -40.723, -56.891, -57.480, -57.395, -56.618, -55.305,
-57.256, -51.822, -47.782, -44.463, -40.744, -56.920, -57.508, -57.424, -56.647,
-55.333, -57.288, -51.849, -47.807, -44.486, -40.765, -56.948, -57.537, -57.453,
-56.675, -55.362, -57.319, -51.877, -47.833, -44.510, -40.785, -56.976, -57.566,
-57.482, -56.704, -55.390, -57.350, -51.905, -47.858, -44.533, -40.806, -57.004,
-57.594, -57.511, -56.733, -55.419, -57.382, -51.933, -47.883, -44.557, -40.827,
 -57.032, -57.623, -57.540, -56.762, -55.447, -57.413, -51.960, -47.909,
-44.581, -40.847, -57.060, -57.652, -57.569, -56.791, -55.476, -57.445, -51.988,
-47.934, -44.604, -40.868, -57.088, -57.680, -57.598, -56.820, -55.505, -57.476,
-52.016, -47.959, -44.628, -40.889, -57.117, -57.709, -57.627, -56.849, -55.533,
-57.507, -52.044, -47.985, -44.651, -40.909, -57.145, -57.738, -57.656, -56.878,
-55.562, -57.539, -52.072, -48.010, -44.675, -40.930, -57.173, -57.766, -57.684,
-56.906, -55.590, -57.570, -52.099, -48.035, -44.698, -40.951, -57.201, -57.795,
-57.713, -56.935, -55.619, -57.602, -52.127, -48.061, -44.722, -40.971, -57.229,
-57.824, -57.742, -56.964, -55.648, -57.633, -52.155, -48.086, -44.745, -40.992,
 -57.257, -57.852, -57.771, -56.993, -55.676, -57.665, -52.183, -48.111,
-44.769, -41.013, -57.285, -57.881, -57.800, -57.022, -55.705, -57.696, -52.211,
-48.137, -44.793, -41.033, -57.314, -57.910, -57.829, -57.051, -55.733, -57.727,
-52.238, -48.162, -44.816, -41.054, -57.342, -57.938, -57.858, -57.080, -55.762,
-57.759, -52.266, -48.187, -44.840, -41.075, -57.370, -57.967, -57.887, -57.108,
-55.791, -57.790, -52.294, -48.213, -44.863, -41.095, -57.398, -57.996, -57.916,
-57.137, -55.819, -57.822, -52.322, -48.238, -44.887, -41.116, -57.426, -58.024,
-57.945, -57.166, -55.848, -57.853, -52.350, -48.263, -44.910, -41.136, -57.454,
-58.053, -57.974, -57.195, -55.876, -57.884, -52.377, -48.289, -44.934, -41.157,
 -57.482, -58.082, -58.003, -57.224, -55.905, -57.916, -52.405, -48.314,
-44.957, -41.178, -57.511, -58.111, -58.031, -57.253, -55.934, -57.947, -52.433,
-48.339, -44.981, -41.198, -57.539, -58.139, -58.060, -57.282, -55.962, -57.979,
-52.461, -48.365, -45.004, -41.219, -57.567, -58.168, -58.089, -57.311, -55.991,
-58.010, -52.489, -48.390, -45.028, -41.239, -57.595, -58.197, -58.118, -57.339,
-56.019, -58.041, -52.516, -48.415, -45.052, -41.260, -57.623, -58.225, -58.147,
-57.368, -56.048, -58.073, -52.544, -48.441, -45.075, -41.281, -57.651, -58.254,
-58.176, -57.397, -56.076, -58.104, -52.572, -48.466, -45.099, -41.301, -57.679,
-58.283, -58.205, -57.426, -56.105, -58.136, -52.600, -48.491, -45.122, -41.322,
 -57.708, -58.311, -58.234, -57.455, -56.134, -58.167, -52.628, -48.517,
-45.146, -41.343, -57.736, -58.340, -58.263, -57.484, -56.162, -58.198, -52.655,
-48.542, -45.169, -41.363, -57.764, -58.369, -58.292, -57.513, -56.191, -58.230,
-52.683, -48.568, -45.193, -41.384, -57.792, -58.397, -58.321, -57.542, -56.219,
-58.261, -52.711, -48.593, -45.216, -41.404, -57.820, -58.426, -58.349, -57.570,
-56.248, -58.293, -52.739, -48.618, -45.240, -41.425, -57.848, -58.455, -58.378,
-57.599, -56.277, -58.324, -52.767, -48.644, -45.263, -41.445, -57.876, -58.483,
-58.407, -57.628, -56.305, -58.355, -52.794, -48.669, -45.287, -41.466, -57.905,
-58.512, -58.436, -57.657, -56.334, -58.387, -52.822, -48.694, -45.311, -41.487,
 -57.933, -58.541, -58.465, -57.686, -56.362, -58.418, -52.850, -48.720,
-45.334, -41.507, -57.961, -58.569, -58.494, -57.715, -56.391, -58.450, -52.878,
-48.745, -45.358, -41.528, -57.989, -58.598, -58.523, -57.744, -56.420, -58.481,
-52.906, -48.770, -45.381, -41.548, -58.017, -58.627, -58.552, -57.773, -56.448,
-58.513, -52.933, -48.796, -45.405, -41.569, -58.045, -58.656, -58.581, -57.801,
-56.477, -58.544, -52.961, -48.821, -45.428, -41.590, -58.073, -58.684, -58.610,
-57.830, -56.505, -58.575, -52.989, -48.846, -45.452, -41.610, -58.102, -58.713,
-58.639, -57.859, -56.534, -58.607, -53.017, -48.872, -45.475, -41.631, -58.130,
-58.742, -58.668, -57.888, -56.563, -58.638, -53.045, -48.897, -45.499, -41.651,
 -58.158, -58.770, -58.696, -57.917, -56.591, -58.670, -53.072, -48.922,
-45.522, -41.672, -58.186, -58.799, -58.725, -57.946, -56.620, -58.701, -53.100,
-48.948, -45.546, -41.692, -58.214, -58.828, -58.754, -57.975, -56.648, -58.732,
-53.128, -48.973, -45.570, -41.713, -58.242, -58.856, -58.783, -58.004, -56.677,
-58.764, -53.156, -48.998, -45.593, -41.733, -58.271, -58.885, -58.812, -58.032,
-56.705, -58.795, -53.184, -49.024, -45.617, -41.754, -58.299, -58.914, -58.841,
-58.061, -56.734, -58.827, -53.211, -49.049, -45.640, -41.775, -58.327, -58.942,
-58.870, -58.090, -56.763, -58.858, -53.239, -49.074, -45.664, -41.795, -58.355,
-58.971, -58.899, -58.119, -56.791, -58.889, -53.267, -49.100, -45.687, -41.816,
 -58.383, -59.000, -58.928, -58.148, -56.820, -58.921, -53.295, -49.125,
-45.711, -41.836, -58.411, -59.028, -58.957, -58.177, -56.848, -58.952, -53.323,
-49.150, -45.734, -41.857, -58.439, -59.057, -58.986, -58.206, -56.877, -58.984,
-53.350, -49.176, -45.758, -41.877, -58.468, -59.086, -59.015, -58.235, -56.906,
-59.015, -53.378, -49.201, -45.781, -41.898, -58.496, -59.114, -59.044, -58.263,
-56.934, -59.047, -53.406, -49.226, -45.805, -41.918, -58.524, -59.143, -59.072,
-58.292, -56.963, -59.078, -53.434, -49.252, -45.829, -41.939, -58.552, -59.172,
-59.101, -58.321, -56.991, -59.109, -53.462, -49.277, -45.852, -41.959, -58.580,
-59.201, -59.130, -58.350, -57.020, -59.141, -53.489, -49.302, -45.876, -41.980,
 -58.608, -59.229, -59.159, -58.379, -57.049, -59.172, -53.517, -49.328,
-45.899, -42.000, -58.636, -59.258, -59.188, -58.408, -57.077, -59.204, -53.545,
-49.353, -45.923, -42.021, -58.665, -59.287, -59.217, -58.437, -57.106, -59.235,
-53.573, -49.378, -45.946, -42.041, -58.693, -59.315, -59.246, -58.466, -57.134,
-59.266, -53.601, -49.404, -45.970, -42.062, -58.721, -59.344, -59.275, -58.494,
-57.163, -59.298, -53.628, -49.429, -45.993, -42.082, -58.749, -59.373, -59.304,
-58.523, -57.192, -59.329, -53.656, -49.454, -46.017, -42.103, -58.777, -59.401,
-59.333, -58.552, -57.220, -59.361, -53.684, -49.480, -46.040, -42.123, -58.805,
-59.430, -59.362, -58.581, -57.249, -59.392, -53.712, -49.505, -46.064, -42.144,
 -58.834, -59.459, -59.391, -58.610, -57.277, -59.424, -53.740, -49.531,
-46.088, -42.164, -58.862, -59.487, -59.419, -58.639, -57.306, -59.455, -53.767,
-49.556, -46.111, -42.185, -58.890, -59.516, -59.448, -58.668, -57.335, -59.486,
-53.795, -49.581, -46.135, -42.205, -58.918, -59.545, -59.477, -58.697, -57.363,
-59.518, -53.823, -49.607, -46.158, -42.226, -58.946, -59.573, -59.506, -58.725,
-57.392, -59.549, -53.851, -49.632, -46.182, -42.246, -58.974, -59.602, -59.535,
-58.754, -57.420, -59.581, -53.879, -49.657, -46.205, -42.267, -59.002, -59.631,
-59.564, -58.783, -57.449, -59.612, -53.906, -49.683, -46.229, -42.287, -59.031,
-59.660, -59.593, -58.812, -57.478, -59.643, -53.934, -49.708, -46.252, -42.308,
 -59.059, -59.688, -59.622, -58.841, -57.506, -59.675, -53.962, -49.733,
-46.276, -42.328, -59.087, -59.717, -59.651, -58.870, -57.535, -59.706, -53.990,
-49.759, -46.299, -42.349, -59.115, -59.746, -59.680, -58.899, -57.563, -59.738,
-54.018, -49.784, -46.323, -42.369, -59.143, -59.774, -59.709, -58.928, -57.592,
-59.769, -54.045, -49.809, -46.347, -42.390, -59.171, -59.803, -59.738, -58.956,
-57.621, -59.801, -54.073, -49.835, -46.370, -42.410, -59.199, -59.832, -59.767,
-58.985, -57.649, -59.832, -54.101, -49.860, -46.394, -42.431, -59.228, -59.860,
-59.795, -59.014, -57.678, -59.863, -54.129, -49.885, -46.417, -42.451, -59.256,
-59.889, -59.824, -59.043, -57.706, -59.895, -54.157, -49.911, -46.441, -42.472,
 -59.284, -59.918, -59.853, -59.072, -57.735, -59.926, -54.184, -49.936,
-46.464, -42.492, -59.312, -59.946, -59.882, -59.101, -57.764, -59.958, -54.212,
-49.961, -46.488, -42.512, -59.340, -59.975, -59.911, -59.130, -57.792, -59.989,
-54.240, -49.987, -46.511, -42.533, -59.368, -60.004, -59.940, -59.159, -57.821,
-60.020, -54.268, -50.012, -46.535, -42.553, -59.397, -60.033, -59.969, -59.188,
-57.849, -60.052, -54.296, -50.037, -46.558, -42.574, -59.425, -60.061, -59.998,
-59.216, -57.878, -60.083, -54.324, -50.063, -46.582, -42.594, -59.453, -60.090,
-60.027, -59.245, -57.907, -60.115, -54.351, -50.088, -46.606, -42.615, -59.481,
-60.119, -60.056, -59.274, -57.935, -60.146, -54.379, -50.113, -46.629, -42.635,
 -59.509, -60.147, -60.085, -59.303, -57.964, -60.178, -54.407, -50.139,
-46.653, -42.656, -59.537, -60.176, -60.114, -59.332, -57.992, -60.209, -54.435,
-50.164, -46.676, -42.676, -59.565, -60.205, -60.143, -59.361, -58.021, -60.240,
-54.463, -50.190, -46.700, -42.696, -59.594, -60.233, -60.171, -59.390, -58.050,
-60.272, -54.490, -50.215, -46.723, -42.717, -59.622, -60.262, -60.200, -59.419,
-58.078, -60.303, -54.518, -50.240, -46.747, -42.737, -59.650, -60.291, -60.229,
-59.447, -58.107, -60.335, -54.546, -50.266, -46.770, -42.758, -59.678, -60.319,
-60.258, -59.476, -58.135, -60.366, -54.574, -50.291, -46.794, -42.778, -59.706,
-60.348, -60.287, -59.505, -58.164, -60.397, -54.602, -50.316, -46.817, -42.798,
 -59.734, -60.377, -60.316, -59.534, -58.193, -60.429, -54.629, -50.342,
-46.841, -42.819, -59.763, -60.406, -60.345, -59.563, -58.221, -60.460, -54.657,
-50.367, -46.865, -42.839, -59.791, -60.434, -60.374, -59.592, -58.250, -60.492,
-54.685, -50.392, -46.888, -42.860, -59.819, -60.463, -60.403, -59.621, -58.278,
-60.523, -54.713, -50.418, -46.912, -42.880, -59.847, -60.492, -60.432, -59.650,
-58.307, -60.555, -54.741, -50.443, -46.935, -42.901, -59.875, -60.520, -60.461,
-59.678, -58.336, -60.586, -54.768, -50.468, -46.959, -42.921, -59.903, -60.549,
-60.490, -59.707, -58.364, -60.617, -54.796, -50.494, -46.982, -42.941, -59.932,
-60.578, -60.519, -59.736, -58.393, -60.649, -54.824, -50.519, -47.006, -42.962,
 -59.960, -60.606, -60.548, -59.765, -58.421, -60.680, -54.852, -50.544,
-47.029, -42.982, -59.988, -60.635, -60.576, -59.794, -58.450, -60.712, -54.880,
-50.570, -47.053, -43.003, -60.016, -60.664, -60.605, -59.823, -58.479, -60.743,
-54.907, -50.595, -47.076, -43.023, -60.044, -60.692, -60.634, -59.852, -58.507,
-60.774, -54.935, -50.620, -47.100, -43.043, -60.072, -60.721, -60.663, -59.881,
-58.536, -60.806, -54.963, -50.646, -47.124, -43.064, -60.100, -60.750, -60.692,
-59.910, -58.564, -60.837, -54.991, -50.671, -47.147, -43.084, -60.129, -60.779,
-60.721, -59.938, -58.593, -60.869, -55.019, -50.697, -47.171, -43.104, -60.157,
-60.807, -60.750, -59.967, -58.622, -60.900, -55.047, -50.722, -47.194, -43.125,
 -60.185, -60.836, -60.779, -59.996, -58.650, -60.932, -55.074, -50.747,
-47.218, -43.145, -60.213, -60.865, -60.808, -60.025, -58.679, -60.963, -55.102,
-50.773, -47.241, -43.166, -60.241, -60.893, -60.837, -60.054, -58.707, -60.994,
-55.130, -50.798, -47.265, -43.186, -60.269, -60.922, -60.866, -60.083, -58.736,
-61.026, -55.158, -50.823, -47.288, -43.206, -60.298, -60.951, -60.895, -60.112,
-58.765, -61.057, -55.186, -50.849, -47.312, -43.227, -60.326, -60.979, -60.924,
-60.141, -58.793, -61.089, -55.213, -50.874, -47.335, -43.247, -60.354, -61.008,
-60.953, -60.170, -58.822, -61.120, -55.241, -50.899, -47.359, -43.267, -60.382,
-61.037, -60.981, -60.198, -58.850, -61.152, -55.269, -50.925, -47.382, -43.288,
 -60.410, -61.066, -61.010, -60.227, -58.879, -61.183, -55.297, -50.950,
-47.406, -43.308, -60.438, -61.094, -61.039, -60.256, -58.908, -61.214, -55.325,
-50.975, -47.430, -43.328, -60.466, -61.123, -61.068, -60.285, -58.936, -61.246,
-55.352, -51.001, -47.453, -43.349, -60.495, -61.152, -61.097, -60.314, -58.965,
-61.277, -55.380, -51.026, -47.477, -43.369, -60.523, -61.180, -61.126, -60.343,
-58.993, -61.309, -55.408, -51.051, -47.500, -43.389, -60.551, -61.209, -61.155,
-60.372, -59.022, -61.340, -55.436, -51.077, -47.524, -43.410, -60.579, -61.238,
-61.184, -60.401, -59.051, -61.372, -55.464, -51.102, -47.547, -43.430, -60.607,
-61.266, -61.213, -60.429, -59.079, -61.403, -55.491, -51.128, -47.571, -43.450,
 -60.635, -61.295, -61.242, -60.458, -59.108, -61.434, -55.519, -51.153,
-47.594, -43.471, -60.664, -61.324, -61.271, -60.487, -59.137, -61.466, -55.547,
-51.178, -47.618, -43.491, -60.692, -61.352, -61.300, -60.516, -59.165, -61.497,
-55.575, -51.204, -47.641, -43.511, -60.720, -61.381, -61.329, -60.545, -59.194,
-61.529, -55.603, -51.229, -47.665, -43.532, -60.748, -61.410, -61.358, -60.574,
-59.222, -61.560, -55.631, -51.254, -47.688, -43.552, -60.776, -61.439, -61.386,
-60.603, -59.251, -61.591, -55.658, -51.280, -47.712, -43.572, -60.804, -61.467,
-61.415, -60.632, -59.280, -61.623, -55.686, -51.305, -47.736, -43.593, -60.833,
-61.496, -61.444, -60.661, -59.308, -61.654, -55.714, -51.330, -47.759, -43.613,
 -60.861, -61.525, -61.473, -60.689, -59.337, -61.686, -55.742, -51.356,
-47.783, -43.633, -60.889, -61.553, -61.502, -60.718, -59.365, -61.717, -55.770,
-51.381, -47.806, -43.654, -60.917, -61.582, -61.531, -60.747, -59.394, -61.749,
-55.797, -51.406, -47.830, -43.674, -60.945, -61.611, -61.560, -60.776, -59.423,
-61.780, -55.825, -51.432, -47.853, -43.694, -60.973, -61.639, -61.589, -60.805,
-59.451, -61.811, -55.853, -51.457, -47.877, -43.715, -61.002, -61.668, -61.618,
-60.834, -59.480, -61.843, -55.881, -51.483, -47.900, -43.735, -61.030, -61.697,
-61.647, -60.863, -59.508, -61.874, -55.909, -51.508, -47.924, -43.755, -61.058,
-61.726, -61.676, -60.892, -59.537, -61.906, -55.937, -51.533, -47.947, -43.776,
 -61.086, -61.754, -61.705, -60.921, -59.566, -61.937, -55.964, -51.559,
-47.971, -43.796, -61.114, -61.783, -61.734, -60.949, -59.594, -61.969, -55.992,
-51.584, -47.995, -43.816, -61.142, -61.812, -61.763, -60.978, -59.623, -62.000,
-56.020, -51.609, -48.018, -43.836, -61.170, -61.840, -61.792, -61.007, -59.651,
-62.031, -56.048, -51.635, -48.042, -43.857, -61.199, -61.869, -61.820, -61.036,
-59.680, -62.063, -56.076, -51.660, -48.065, -43.877, -61.227, -61.898, -61.849,
-61.065, -59.709, -62.094, -56.103, -51.685, -48.089, -43.897, -61.255, -61.926,
-61.878, -61.094, -59.737, -62.126, -56.131, -51.711, -48.112, -43.918, -61.283,
-61.955, -61.907, -61.123, -59.766, -62.157, -56.159, -51.736, -48.136, -43.938,
 -61.311, -61.984, -61.936, -61.152, -59.795, -62.189, -56.187, -51.761,
-48.159, -43.958, -61.339, -62.013, -61.965, -61.181, -59.823, -62.220, -56.215,
-51.787, -48.183, -43.978, -61.368, -62.041, -61.994, -61.209, -59.852, -62.251,
-56.242, -51.812, -48.206, -43.999, -61.396, -62.070, -62.023, -61.238, -59.880,
-62.283, -56.270, -51.838, -48.230, -44.019, -61.424, -62.099, -62.052, -61.267,
-59.909, -62.314, -56.298, -51.863, -48.253, -44.039, -61.452, -62.127, -62.081,
-61.296, -59.938, -62.346, -56.326, -51.888, -48.277, -44.059, -61.480, -62.156,
-62.110, -61.325, -59.966, -62.377, -56.354, -51.914, -48.301, -44.080, -61.508,
-62.185, -62.139, -61.354, -59.995, -62.409, -56.382, -51.939, -48.324, -44.100,
 -61.537, -62.214, -62.168, -61.383, -60.023, -62.440, -56.409, -51.964,
-48.348, -44.120, -61.565, -62.242, -62.197, -61.412, -60.052, -62.471, -56.437,
-51.990, -48.371, -44.140, -61.593, -62.271, -62.226, -61.441, -60.081, -62.503,
-56.465, -52.015, -48.395, -44.161, -61.621, -62.300, -62.254, -61.469, -60.109,
-62.534, -56.493, -52.040, -48.418, -44.181, -61.649, -62.328, -62.283, -61.498,
-60.138, -62.566, -56.521, -52.066, -48.442, -44.201, -61.677, -62.357, -62.312,
-61.527, -60.166, -62.597, -56.548, -52.091, -48.465, -44.221, -61.706, -62.386,
-62.341, -61.556, -60.195, -62.629, -56.576, -52.116, -48.489, -44.242, -61.734,
-62.414, -62.370, -61.585, -60.224, -62.660, -56.604, -52.142, -48.512, -44.262,
 -61.762, -62.443, -62.399, -61.614, -60.252, -62.691, -56.632, -52.167,
-48.536, -44.282, -61.790, -62.472, -62.428, -61.643, -60.281, -62.723, -56.660,
-52.193, -48.559, -44.302, -61.818, -62.501, -62.457, -61.672, -60.309, -62.754,
-56.688, -52.218, -48.583, -44.323, -61.846, -62.529, -62.486, -61.701, -60.338,
-62.786, -56.715, -52.243, -48.606, -44.343, -61.875, -62.558, -62.515, -61.730,
-60.367, -62.817, -56.743, -52.269, -48.630, -44.363, -61.903, -62.587, -62.544,
-61.758, -60.395, -62.849, -56.771, -52.294, -48.654, -44.383, -61.931, -62.615,
-62.573, -61.787, -60.424, -62.880, -56.799, -52.319, -48.677, -44.404, -61.959,
-62.644, -62.602, -61.816, -60.453, -62.911, -56.827, -52.345, -48.701, -44.424,
 -61.987, -62.673, -62.631, -61.845, -60.481, -62.943, -56.854, -52.370,
-48.724, -44.444, -62.015, -62.701, -62.660, -61.874, -60.510, -62.974, -56.882,
-52.395, -48.748, -44.464, -62.044, -62.730, -62.689, -61.903, -60.538, -63.006,
-56.910, -52.421, -48.771, -44.484, -62.072, -62.759, -62.717, -61.932, -60.567,
-63.037, -56.938, -52.446, -48.795, -44.505, -62.100, -62.788, -62.746, -61.961,
-60.596, -63.069, -56.966, -52.471, -48.818, -44.525, -62.128, -62.816, -62.775,
-61.990, -60.624, -63.100, -56.994, -52.497, -48.842, -44.545, -62.156, -62.845,
-62.804, -62.018, -60.653, -63.131, -57.021, -52.522, -48.865, -44.565, -62.184,
-62.874, -62.833, -62.047, -60.681, -63.163, -57.049, -52.548, -48.889, -44.585,
 -62.213, -62.902, -62.862, -62.076, -60.710, -63.194, -57.077, -52.573,
-48.912, -44.606, -62.241, -62.931, -62.891, -62.105, -60.739, -63.226, -57.105,
-52.598, -48.936, -44.626, -62.269, -62.960, -62.920, -62.134, -60.767, -63.257,
-57.133, -52.624, -48.959, -44.646, -62.297, -62.989, -62.949, -62.163, -60.796,
-63.289, -57.160, -52.649, -48.983, -44.666, -62.325, -63.017, -62.978, -62.192,
-60.825, -63.320, -57.188, -52.674, -49.007, -44.686, -62.353, -63.046, -63.007,
-62.221, -60.853, -63.352, -57.216, -52.700, -49.030, -44.707, -62.382, -63.075,
-63.036, -62.250, -60.882, -63.383, -57.244, -52.725, -49.054, -44.727, -62.410,
-63.103, -63.065, -62.279, -60.910, -63.414, -57.272, -52.750, -49.077, -44.747,
 -62.438, -63.132, -63.094, -62.307, -60.939, -63.446, -57.300, -52.776,
-49.101, -44.767, -62.466, -63.161, -63.123, -62.336, -60.968, -63.477, -57.327,
-52.801, -49.124, -44.787, -62.494, -63.189, -63.152, -62.365, -60.996, -63.509,
-57.355, -52.827, -49.148, -44.808, -62.522, -63.218, -63.181, -62.394, -61.025,
-63.540, -57.383, -52.852, -49.171, -44.828, -62.551, -63.247, -63.209, -62.423,
-61.053, -63.572, -57.411, -52.877, -49.195, -44.848, -62.579, -63.276, -63.238,
-62.452, -61.082, -63.603, -57.439, -52.903, -49.218, -44.868, -62.607, -63.304,
-63.267, -62.481, -61.111, -63.634, -57.466, -52.928, -49.242, -44.888, -62.635,
-63.333, -63.296, -62.510, -61.139, -63.666, -57.494, -52.953, -49.265, -44.908,
 -62.663, -63.362, -63.325, -62.539, -61.168, -63.697, -57.522, -52.979,
-49.289, -44.929, -62.691, -63.390, -63.354, -62.567, -61.197, -63.729, -57.550,
-53.004, -49.312, -44.949, -62.720, -63.419, -63.383, -62.596, -61.225, -63.760,
-57.578, -53.029, -49.336, -44.969, -62.748, -63.448, -63.412, -62.625, -61.254,
-63.792, -57.606, -53.055, -49.360, -44.989, -62.776, -63.477, -63.441, -62.654,
-61.282, -63.823, -57.633, -53.080, -49.383, -45.009, -62.804, -63.505, -63.470,
-62.683, -61.311, -63.854, -57.661, -53.106, -49.407, -45.029, -62.832, -63.534,
-63.499, -62.712, -61.340, -63.886, -57.689, -53.131, -49.430, -45.050, -62.860,
-63.563, -63.528, -62.741, -61.368, -63.917, -57.717, -53.156, -49.454, -45.070,
 -62.889, -63.591, -63.557, -62.770, -61.397, -63.949, -57.745, -53.182,
-49.477, -45.090, -62.917, -63.620, -63.586, -62.799, -61.425, -63.980, -57.773,
-53.207, -49.501, -45.110, -62.945, -63.649, -63.615, -62.828, -61.454, -64.012,
-57.800, -53.232, -49.524, -45.130, -62.973, -63.678, -63.644, -62.856, -61.483,
-64.043, -57.828, -53.258, -49.548, -45.150, -63.001, -63.706, -63.673, -62.885,
-61.511, -64.074, -57.856, -53.283, -49.571, -45.170, -63.029, -63.735, -63.701,
-62.914, -61.540, -64.106, -57.884, -53.308, -49.595, -45.191, -63.058, -63.764,
-63.730, -62.943, -61.569, -64.137, -57.912, -53.334, -49.618, -45.211, -63.086,
-63.792, -63.759, -62.972, -61.597, -64.169, -57.939, -53.359, -49.642, -45.231,
 -63.114, -63.821, -63.788, -63.001, -61.626, -64.200, -57.967, -53.385,
-49.665, -45.251, -63.142, -63.850, -63.817, -63.030, -61.654, -64.232, -57.995,
-53.410, -49.689, -45.271, -63.170, -63.878, -63.846, -63.059, -61.683, -64.263,
-58.023, -53.435, -49.713, -45.291, -63.198, -63.907, -63.875, -63.088, -61.712,
-64.295, -58.051, -53.461, -49.736, -45.311, -63.227, -63.936, -63.904, -63.117,
-61.740, -64.326, -58.079, -53.486, -49.760, -45.331, -63.255, -63.965, -63.933,
-63.145, -61.769, -64.357, -58.106, -53.511, -49.783, -45.352, -63.283, -63.993,
-63.962, -63.174, -61.797, -64.389, -58.134, -53.537, -49.807, -45.372, -63.311,
-64.022, -63.991, -63.203, -61.826, -64.420, -58.162, -53.562, -49.830, -45.392,
 -63.339, -64.051, -64.020, -63.232, -61.855, -64.452, -58.190, -53.588,
-49.854, -45.412, -63.367, -64.079, -64.049, -63.261, -61.883, -64.483, -58.218,
-53.613, -49.877, -45.432, -63.396, -64.108, -64.078, -63.290, -61.912, -64.515,
-58.246, -53.638, -49.901, -45.452, -63.424, -64.137, -64.107, -63.319, -61.941,
-64.546, -58.273, -53.664, -49.924, -45.472, -63.452, -64.166, -64.136, -63.348,
-61.969, -64.577, -58.301, -53.689, -49.948, -45.492, -63.480, -64.194, -64.165,
-63.377, -61.998, -64.609, -58.329, -53.714, -49.971, -45.512, -63.508, -64.223,
-64.193, -63.406, -62.026, -64.640, -58.357, -53.740, -49.995, -45.533, -63.536,
-64.252, -64.222, -63.434, -62.055, -64.672, -58.385, -53.765, -50.018, -45.553,
 -63.565, -64.280, -64.251, -63.463, -62.084, -64.703, -58.412, -53.790,
-50.042, -45.573, -63.593, -64.309, -64.280, -63.492, -62.112, -64.735, -58.440,
-53.816, -50.065, -45.593, -63.621, -64.338, -64.309, -63.521, -62.141, -64.766,
-58.468, -53.841, -50.089, -45.613, -63.649, -64.367, -64.338, -63.550, -62.170,
-64.798, -58.496, -53.867, -50.112, -45.633, -63.677, -64.395, -64.367, -63.579,
-62.198, -64.829, -58.524, -53.892, -50.136, -45.653, -63.705, -64.424, -64.396,
-63.608, -62.227, -64.860, -58.552, -53.917, -50.160, -45.673, -63.734, -64.453,
-64.425, -63.637, -62.255, -64.892, -58.579, -53.943, -50.183, -45.693, -63.762,
-64.481, -64.454, -63.666, -62.284, -64.923, -58.607, -53.968, -50.207, -45.713,
 -63.790, -64.510, -64.483, -63.695, -62.313, -64.955, -58.635, -53.993,
-50.230, -45.733, -63.818, -64.539, -64.512, -63.723, -62.341, -64.986, -58.663,
-54.019, -50.254, -45.754, -63.846, -64.568, -64.541, -63.752, -62.370, -65.018,
-58.691, -54.044, -50.277, -45.774, -63.874, -64.596, -64.570, -63.781, -62.398,
-65.049, -58.719, -54.070, -50.301, -45.794, -63.903, -64.625, -64.599, -63.810,
-62.427, -65.081, -58.746, -54.095, -50.324, -45.814, -63.931, -64.654, -64.628,
-63.839, -62.456, -65.112, -58.774, -54.120, -50.348, -45.834, -63.959, -64.682,
-64.657, -63.868, -62.484, -65.143, -58.802, -54.146, -50.371, -45.854, -63.987,
-64.711, -64.686, -63.897, -62.513, -65.175, -58.830, -54.171, -50.395, -45.874,
 -64.015, -64.740, -64.715, -63.926, -62.542, -65.206, -58.858, -54.196,
-50.418, -45.894, -64.044, -64.769, -64.743, -63.955, -62.570, -65.238, -58.885,
-54.222, -50.442, -45.914, -64.072, -64.797, -64.772, -63.984, -62.599, -65.269,
-58.913, -54.247, -50.465, -45.934, -64.100, -64.826, -64.801, -64.012, -62.627,
-65.301, -58.941, -54.272, -50.489, -45.954, -64.128, -64.855, -64.830, -64.041,
-62.656, -65.332, -58.969, -54.298, -50.512, -45.974, -64.156, -64.883, -64.859,
-64.070, -62.685, -65.363, -58.997, -54.323, -50.536, -45.994, -64.184, -64.912,
-64.888, -64.099, -62.713, -65.395, -59.025, -54.349, -50.559, -46.014, -64.213,
-64.941, -64.917, -64.128, -62.742, -65.426, -59.052, -54.374, -50.583, -46.034,
 -64.241, -64.970, -64.946, -64.157, -62.771, -65.458, -59.080, -54.399,
-50.607, -46.054, -64.269, -64.998, -64.975, -64.186, -62.799, -65.489, -59.108,
-54.425, -50.630, -46.074, -64.297, -65.027, -65.004, -64.215, -62.828, -65.521,
-59.136, -54.450, -50.654, -46.094, -64.325, -65.056, -65.033, -64.244, -62.856,
-65.552, -59.164, -54.475, -50.677, -46.115, -64.353, -65.084, -65.062, -64.273,
-62.885, -65.584, -59.192, -54.501, -50.701, -46.135, -64.382, -65.113, -65.091,
-64.301, -62.914, -65.615, -59.219, -54.526, -50.724, -46.155, -64.410, -65.142,
-65.120, -64.330, -62.942, -65.646, -59.247, -54.552, -50.748, -46.175, -64.438,
-65.171, -65.149, -64.359, -62.971, -65.678, -59.275, -54.577, -50.771, -46.195,
 -64.466, -65.199, -65.178, -64.388, -63.000, -65.709, -59.303, -54.602,
-50.795, -46.215, -64.494, -65.228, -65.207, -64.417, -63.028, -65.741, -59.331,
-54.628, -50.818, -46.235, -64.522, -65.257, -65.236, -64.446, -63.057, -65.772,
-59.359, -54.653, -50.842, -46.255, -64.551, -65.285, -65.265, -64.475, -63.085,
-65.804, -59.386, -54.678, -50.865, -46.275, -64.579, -65.314, -65.293, -64.504,
-63.114, -65.835, -59.414, -54.704, -50.889, -46.295, -64.607, -65.343, -65.322,
-64.533, -63.143, -65.867, -59.442, -54.729, -50.912, -46.315, -64.635, -65.372,
-65.351, -64.562, -63.171, -65.898, -59.470, -54.755, -50.936, -46.335, -64.663,
-65.400, -65.380, -64.591, -63.200, -65.929, -59.498, -54.780, -50.959, -46.355,
 -64.692, -65.429, -65.409, -64.619, -63.229, -65.961, -59.526, -54.805,
-50.983, -46.375, -64.720, -65.458, -65.438, -64.648, -63.257, -65.992, -59.553,
-54.831, -51.006, -46.395, -64.748, -65.486, -65.467, -64.677, -63.286, -66.024,
-59.581, -54.856, -51.030, -46.415, -64.776, -65.515, -65.496, -64.706, -63.314,
-66.055, -59.609, -54.881, -51.053, -46.435, -64.804, -65.544, -65.525, -64.735,
-63.343, -66.087, -59.637, -54.907, -51.077, -46.455, -64.832, -65.573, -65.554,
-64.764, -63.372, -66.118, -59.665, -54.932, -51.100, -46.475, -64.861, -65.601,
-65.583, -64.793, -63.400, -66.150, -59.693, -54.957, -51.124, -46.495, -64.889,
-65.630, -65.612, -64.822, -63.429, -66.181, -59.720, -54.983, -51.148, -46.515,
 -64.917, -65.659, -65.641, -64.851, -63.458, -66.212, -59.748, -55.008,
-51.171, -46.535, -64.945, -65.688, -65.670, -64.880, -63.486, -66.244, -59.776,
-55.034, -51.195, -46.555, -64.973, -65.716, -65.699, -64.908, -63.515, -66.275,
-59.804, -55.059, -51.218, -46.575, -65.001, -65.745, -65.728, -64.937, -63.543,
-66.307, -59.832, -55.084, -51.242, -46.595, -65.030, -65.774, -65.757, -64.966,
-63.572, -66.338, -59.860, -55.110, -51.265, -46.615, -65.058, -65.802, -65.786,
-64.995, -63.601, -66.370, -59.887, -55.135, -51.289, -46.635, -65.086, -65.831,
-65.815, -65.024, -63.629, -66.401, -59.915, -55.160, -51.312, -46.655, -65.114,
-65.860, -65.844, -65.053, -63.658, -66.433, -59.943, -55.186, -51.336, -46.675,
 -65.142, -65.889, -65.872, -65.082, -63.687, -66.464, -59.971, -55.211,
-51.359, -46.695, -65.171, -65.917, -65.901, -65.111, -63.715, -66.495, -59.999,
-55.237, -51.383, -46.715, -65.199, -65.946, -65.930, -65.140, -63.744, -66.527,
-60.026, -55.262, -51.406, -46.735, -65.227, -65.975, -65.959, -65.169, -63.772,
-66.558, -60.054, -55.287, -51.430, -46.755, -65.255, -66.003, -65.988, -65.198,
-63.801, -66.590, -60.082, -55.313, -51.453, -46.775, -65.283, -66.032, -66.017,
-65.226, -63.830, -66.621, -60.110, -55.338, -51.477, -46.795, -65.311, -66.061,
-66.046, -65.255, -63.858, -66.653, -60.138, -55.363, -51.500, -46.815, -65.340,
-66.090, -66.075, -65.284, -63.887, -66.684, -60.166, -55.389, -51.524, -46.835,
 -65.368, -66.118, -66.104, -65.313, -63.916, -66.716, -60.193, -55.414,
-51.547, -46.855, -65.396, -66.147, -66.133, -65.342, -63.944, -66.747, -60.221,
-55.440, -51.571, -46.875, -65.424, -66.176, -66.162, -65.371, -63.973, -66.778,
-60.249, -55.465, -51.594, -46.895, -65.452, -66.204, -66.191, -65.400, -64.001,
-66.810, -60.277, -55.490, -51.618, -46.914, -65.480, -66.233, -66.220, -65.429,
-64.030, -66.841, -60.305, -55.516, -51.641, -46.934, -65.509, -66.262, -66.249,
-65.458, -64.059, -66.873, -60.333, -55.541, -51.665, -46.954, -65.537, -66.291,
-66.278, -65.487, -64.087, -66.904, -60.360, -55.566, -51.688, -46.974, -65.565,
-66.319, -66.307, -65.516, -64.116, -66.936, -60.388, -55.592, -51.712, -46.994,
 -65.593, -66.348, -66.336, -65.544, -64.145, -66.967, -60.416, -55.617,
-51.735, -47.014, -65.621, -66.377, -66.365, -65.573, -64.173, -66.999, -60.444,
-55.643, -51.759, -47.034, -65.650, -66.406, -66.394, -65.602, -64.202, -67.030,
-60.472, -55.668, -51.783, -47.054, -65.678, -66.434, -66.423, -65.631, -64.230,
-67.061, -60.500, -55.693, -51.806, -47.074, -65.706, -66.463, -66.452, -65.660,
-64.259, -67.093, -60.527, -55.719, -51.830, -47.094, -65.734, -66.492, -66.480,
-65.689, -64.288, -67.124, -60.555, -55.744, -51.853, -47.114, -65.762, -66.520,
-66.509, -65.718, -64.316, -67.156, -60.583, -55.769, -51.877, -47.134, -65.790,
-66.549, -66.538, -65.747, -64.345, -67.187, -60.611, -55.795, -51.900, -47.154,
 -65.819, -66.578, -66.567, -65.776, -64.374, -67.219, -60.639, -55.820,
-51.924, -47.174, -65.847, -66.607, -66.596, -65.805, -64.402, -67.250, -60.667,
-55.846, -51.947, -47.194, -65.875, -66.635, -66.625, -65.834, -64.431, -67.282,
-60.694, -55.871, -51.971, -47.214, -65.903, -66.664, -66.654, -65.862, -64.460,
-67.313, -60.722, -55.896, -51.994, -47.234, -65.931, -66.693, -66.683, -65.891,
-64.488, -67.344, -60.750, -55.922, -52.018, -47.254, -65.960, -66.721, -66.712,
-65.920, -64.517, -67.376, -60.778, -55.947, -52.041, -47.273, -65.988, -66.750,
-66.741, -65.949, -64.545, -67.407, -60.806, -55.973, -52.065, -47.293, -66.016,
-66.779, -66.770, -65.978, -64.574, -67.439, -60.834, -55.998, -52.088, -47.313,
 -66.044, -66.808, -66.799, -66.007, -64.603, -67.470, -60.861, -56.023,
-52.112, -47.333, -66.072, -66.836, -66.828, -66.036, -64.631, -67.502, -60.889,
-56.049, -52.135, -47.353, -66.100, -66.865, -66.857, -66.065, -64.660, -67.533,
-60.917, -56.074, -52.159, -47.373, -66.129, -66.894, -66.886, -66.094, -64.689,
-67.565, -60.945, -56.099, -52.182, -47.393, -66.157, -66.923, -66.915, -66.123,
-64.717, -67.596, -60.973, -56.125, -52.206, -47.413, -66.185, -66.951, -66.944,
-66.152, -64.746, -67.628, -61.001, -56.150, -52.229, -47.433, -66.213, -66.980,
-66.973, -66.180, -64.774, -67.659, -61.029, -56.176, -52.253, -47.453, -66.241,
-67.009, -67.002, -66.209, -64.803, -67.690, -61.056, -56.201, -52.276, -47.473,
 -66.270, -67.037, -67.031, -66.238, -64.832, -67.722, -61.084, -56.226,
-52.300, -47.493, -66.298, -67.066, -67.060, -66.267, -64.860, -67.753, -61.112,
-56.252, -52.323, -47.512, -66.326, -67.095, -67.089, -66.296, -64.889, -67.785,
-61.140, -56.277, -52.347, -47.532, -66.354, -67.124, -67.118, -66.325, -64.918,
-67.816, -61.168, -56.302, -52.370, -47.552, -66.382, -67.152, -67.146, -66.354,
-64.946, -67.848, -61.196, -56.328, -52.394, -47.572, -66.410, -67.181, -67.175,
-66.383, -64.975, -67.879, -61.223, -56.353, -52.417, -47.592, -66.439, -67.210,
-67.204, -66.412, -65.003, -67.911, -61.251, -56.379, -52.441, -47.612, -66.467,
-67.239, -67.233, -66.441, -65.032, -67.942, -61.279, -56.404, -52.464, -47.632,
 -66.495, -67.267, -67.262, -66.470, -65.061, -67.973, -61.307, -56.429,
-52.488, -47.652, -66.523, -67.296, -67.291, -66.499, -65.089, -68.005, -61.335,
-56.455, -52.511, -47.672, -66.551, -67.325, -67.320, -66.527, -65.118, -68.036,
-61.363, -56.480, -52.535, -47.692, -66.580, -67.353, -67.349, -66.556, -65.147,
-68.068, -61.390, -56.505, -52.558, -47.711, -66.608, -67.382, -67.378, -66.585,
-65.175, -68.099, -61.418, -56.531, -52.582, -47.731, -66.636, -67.411, -67.407,
-66.614, -65.204, -68.131, -61.446, -56.556, -52.605, -47.751, -66.664, -67.440,
-67.436, -66.643, -65.233, -68.162, -61.474, -56.582, -52.629, -47.771, -66.692,
-67.468, -67.465, -66.672, -65.261, -68.194, -61.502, -56.607, -52.652, -47.791,
 -66.720, -67.497, -67.494, -66.701, -65.290, -68.225, -61.530, -56.632,
-52.676, -47.811, -66.749, -67.526, -67.523, -66.730, -65.318, -68.257, -61.557,
-56.658, -52.699, -47.831, -66.777, -67.554, -67.552, -66.759, -65.347, -68.288,
-61.585, -56.683, -52.723, -47.851, -66.805, -67.583, -67.581, -66.788, -65.376,
-68.319, -61.613, -56.709, -52.746, -47.870, -66.833, -67.612, -67.610, -66.817,
-65.404, -68.351, -61.641, -56.734, -52.770, -47.890, -66.861, -67.641, -67.639,
-66.845, -65.433, -68.382, -61.669, -56.759, -52.793, -47.910, -66.890, -67.669,
-67.668, -66.874, -65.462, -68.414, -61.697, -56.785, -52.817, -47.930, -66.918,
-67.698, -67.697, -66.903, -65.490, -68.445, -61.724, -56.810, -52.840, -47.950,
 -66.946, -67.727, -67.726, -66.932, -65.519, -68.477, -61.752, -56.835,
-52.864, -47.970, -66.974, -67.756, -67.755, -66.961, -65.548, -68.508, -61.780,
-56.861, -52.887, -47.990, -67.002, -67.784, -67.784, -66.990, -65.576, -68.540,
-61.808, -56.886, -52.911, -48.010, -67.030, -67.813, -67.813, -67.019, -65.605,
-68.571, -61.836, -56.912, -52.934, -48.029, -67.059, -67.842, -67.841, -67.048,
-65.633, -68.603, -61.864, -56.937, -52.958, -48.049, -67.087, -67.870, -67.870,
-67.077, -65.662, -68.634, -61.891, -56.962, -52.982, -48.069, -67.115, -67.899,
-67.899, -67.106, -65.691, -68.665, -61.919, -56.988, -53.005, -48.089, -67.143,
-67.928, -67.928, -67.135, -65.719, -68.697, -61.947, -57.013, -53.029, -48.109,
 -67.171, -67.957, -67.957, -67.164, -65.748, -68.728, -61.975, -57.038,
-53.052, -48.129, -67.200, -67.985, -67.986, -67.192, -65.777, -68.760, -62.003,
-57.064, -53.076, -48.149, -67.228, -68.014, -68.015, -67.221, -65.805, -68.791,
-62.031, -57.089, -53.099, -48.168, -67.256, -68.043, -68.044, -67.250, -65.834,
-68.823, -62.059, -57.115, -53.123, -48.188, -67.284, -68.072, -68.073, -67.279,
-65.863, -68.854, -62.086, -57.140, -53.146, -48.208, -67.312, -68.100, -68.102,
-67.308, -65.891, -68.886, -62.114, -57.165, -53.170, -48.228, -67.340, -68.129,
-68.131, -67.337, -65.920, -68.917, -62.142, -57.191, -53.193, -48.248, -67.369,
-68.158, -68.160, -67.366, -65.948, -68.949, -62.170, -57.216, -53.217, -48.268,
 -67.397, -68.187, -68.189, -67.395, -65.977, -68.980, -62.198, -57.242,
-53.240, -48.287, -67.425, -68.215, -68.218, -67.424, -66.006, -69.011, -62.226,
-57.267, -53.264, -48.307, -67.453, -68.244, -68.247, -67.453, -66.034, -69.043,
-62.253, -57.292, -53.287, -48.327, -67.481, -68.273, -68.276, -67.482, -66.063,
-69.074, -62.281, -57.318, -53.311, -48.347, -67.510, -68.301, -68.305, -67.511,
-66.092, -69.106, -62.309, -57.343, -53.334, -48.367, -67.538, -68.330, -68.334,
-67.539, -66.120, -69.137, -62.337, -57.368, -53.358, -48.387, -67.566, -68.359,
-68.363, -67.568, -66.149, -69.169, -62.365, -57.394, -53.381, -48.406, -67.594,
-68.388, -68.392, -67.597, -66.177, -69.200, -62.393, -57.419, -53.405, -48.426,
 -67.622, -68.416, -68.421, -67.626, -66.206, -69.232, -62.420, -57.445,
-53.428, -48.446, -67.651, -68.445, -68.450, -67.655, -66.235, -69.263, -62.448,
-57.470, -53.452, -48.466, -67.679, -68.474, -68.479, -67.684, -66.263, -69.295,
-62.476, -57.495, -53.475, -48.486, -67.707, -68.503, -68.508, -67.713, -66.292,
-69.326, -62.504, -57.521, -53.499, -48.506, -67.735, -68.531, -68.537, -67.742,
-66.321, -69.357, -62.532, -57.546, -53.522, -48.525, -67.763, -68.560, -68.566,
-67.771, -66.349, -69.389, -62.560, -57.572, -53.546, -48.545, -67.791, -68.589,
-68.594, -67.800, -66.378, -69.420, -62.587, -57.597, -53.569, -48.565, -67.820,
-68.617, -68.623, -67.829, -66.407, -69.452, -62.615, -57.622, -53.593, -48.585,
 -67.848, -68.646, -68.652, -67.858, -66.435, -69.483, -62.643, -57.648,
-53.616, -48.605, -67.876, -68.675, -68.681, -67.886, -66.464, -69.515, -62.671,
-57.673, -53.640, -48.624, -67.904, -68.704, -68.710, -67.915, -66.493, -69.546,
-62.699, -57.698, -53.663, -48.644, -67.932, -68.732, -68.739, -67.944, -66.521,
-69.578, -62.727, -57.724, -53.687, -48.664, -67.961, -68.761, -68.768, -67.973,
-66.550, -69.609, -62.755, -57.749, -53.710, -48.684, -67.989, -68.790, -68.797,
-68.002, -66.578, -69.641, -62.782, -57.775, -53.734, -48.704, -68.017, -68.819,
-68.826, -68.031, -66.607, -69.672, -62.810, -57.800, -53.757, -48.724, -68.045,
-68.847, -68.855, -68.060, -66.636, -69.703, -62.838, -57.825, -53.781, -48.743,
 -68.073, -68.876, -68.884, -68.089, -66.664, -69.735, -62.866, -57.851,
-53.804, -48.763, -68.102, -68.905, -68.913, -68.118, -66.693, -69.766, -62.894,
-57.876, -53.828, -48.783, -68.130, -68.934, -68.942, -68.147, -66.722, -69.798,
-62.922, -57.902, -53.851, -48.803, -68.158, -68.962, -68.971, -68.176, -66.750,
-69.829, -62.949, -57.927, -53.875, -48.823, -68.186, -68.991, -69.000, -68.205,
-66.779, -69.861, -62.977, -57.952, -53.898, -48.842, -68.214, -69.020, -69.029,
-68.233, -66.808, -69.892, -63.005, -57.978, -53.922, -48.862, -68.242, -69.048,
-69.058, -68.262, -66.836, -69.924, -63.033, -58.003, -53.945, -48.882, -68.271,
-69.077, -69.087, -68.291, -66.865, -69.955, -63.061, -58.028, -53.969, -48.902,
 -68.299, -69.106, -69.116, -68.320, -66.893, -69.987, -63.089, -58.054,
-53.992, -48.921, -68.327, -69.135, -69.145, -68.349, -66.922, -70.018, -63.116,
-58.079, -54.016, -48.941, -68.355, -69.163, -69.174, -68.378, -66.951, -70.049,
-63.144, -58.105, -54.039, -48.961, -68.383, -69.192, -69.203, -68.407, -66.979,
-70.081, -63.172, -58.130, -54.062, -48.981, -68.412, -69.221, -69.232, -68.436,
-67.008, -70.112, -63.200, -58.155, -54.086, -49.001, -68.440, -69.250, -69.261,
-68.465, -67.037, -70.144, -63.228, -58.181, -54.109, -49.020, -68.468, -69.278,
-69.290, -68.494, -67.065, -70.175, -63.256, -58.206, -54.133, -49.040, -68.496,
-69.307, -69.319, -68.523, -67.094, -70.207, -63.284, -58.232, -54.156, -49.060,
 -68.524, -69.336, -69.348, -68.552, -67.123, -70.238, -63.311, -58.257,
-54.180, -49.080, -68.553, -69.364, -69.377, -68.581, -67.151, -70.270, -63.339,
-58.282, -54.203, -49.100, -68.581, -69.393, -69.405, -68.609, -67.180, -70.301,
-63.367, -58.308, -54.227, -49.119, -68.609, -69.422, -69.434, -68.638, -67.208,
-70.333, -63.395, -58.333, -54.250, -49.139, -68.637, -69.451, -69.463, -68.667,
-67.237, -70.364, -63.423, -58.359, -54.274, -49.159, -68.665, -69.479, -69.492,
-68.696, -67.266, -70.396, -63.451, -58.384, -54.297, -49.179, -68.694, -69.508,
-69.521, -68.725, -67.294, -70.427, -63.478, -58.409, -54.321, -49.198, -68.722,
-69.537, -69.550, -68.754, -67.323, -70.458, -63.506, -58.435, -54.344, -49.218,
 -68.750, -69.566, -69.579, -68.783, -67.352, -70.490, -63.534, -58.460,
-54.368, -49.238, -68.778, -69.594, -69.608, -68.812, -67.380, -70.521, -63.562,
-58.485, -54.391, -49.258, -68.806, -69.623, -69.637, -68.841, -67.409, -70.553,
-63.590, -58.511, -54.415, -49.277, -68.834, -69.652, -69.666, -68.870, -67.438,
-70.584, -63.618, -58.536, -54.438, -49.297, -68.863, -69.681, -69.695, -68.899,
-67.466, -70.616, -63.646, -58.562, -54.462, -49.317, -68.891, -69.709, -69.724,
-68.928, -67.495, -70.647, -63.673, -58.587, -54.485, -49.337, -68.919, -69.738,
-69.753, -68.956, -67.524, -70.679, -63.701, -58.612, -54.509, -49.356, -68.947,
-69.767, -69.782, -68.985, -67.552, -70.710, -63.729, -58.638, -54.532, -49.376,
 -68.975, -69.796, -69.811, -69.014, -67.581, -70.742, -63.757, -58.663,
-54.556, -49.396, -69.004, -69.824, -69.840, -69.043, -67.609, -70.773, -63.785,
-58.689, -54.579, -49.416, -69.032, -69.853, -69.869, -69.072, -67.638, -70.805,
-63.813, -58.714, -54.603, -49.435, -69.060, -69.882, -69.898, -69.101, -67.667,
-70.836, -63.840, -58.739, -54.626, -49.455, -69.088, -69.910, -69.927, -69.130,
-67.695, -70.867, -63.868, -58.765, -54.650, -49.475, -69.116, -69.939, -69.956,
-69.159, -67.724, -70.899, -63.896, -58.790, -54.673, -49.495, -69.145, -69.968,
-69.985, -69.188, -67.753, -70.930, -63.924, -58.816, -54.697, -49.514, -69.173,
-69.997, -70.014, -69.217, -67.781, -70.962, -63.952, -58.841, -54.720, -49.534,
 -69.201, -70.025, -70.043, -69.246, -67.810, -70.993, -63.980, -58.866,
-54.744, -49.554, -69.229, -70.054, -70.072, -69.275, -67.839, -71.025, -64.008,
-58.892, -54.767, -49.574, -69.257, -70.083, -70.101, -69.304, -67.867, -71.056,
-64.035, -58.917, -54.791, -49.593, -69.286, -70.112, -70.130, -69.332, -67.896,
-71.088, -64.063, -58.942, -54.814, -49.613, -69.314, -70.140, -70.159, -69.361,
-67.925, -71.119, -64.091, -58.968, -54.838, -49.633, -69.342, -70.169, -70.188,
-69.390, -67.953, -71.151, -64.119, -58.993, -54.861, -49.653, -69.370, -70.198,
-70.217, -69.419, -67.982, -71.182, -64.147, -59.019, -54.885, -49.672, -69.398,
-70.227, -70.246, -69.448, -68.010, -71.214, -64.175, -59.044, -54.908, -49.692,
 -69.427, -70.255, -70.275, -69.477, -68.039, -71.245, -64.202, -59.069,
-54.932, -49.712, -69.455, -70.284, -70.303, -69.506, -68.068, -71.276, -64.230,
-59.095, -54.955, -49.732, -69.483, -70.313, -70.332, -69.535, -68.096, -71.308,
-64.258, -59.120, -54.979, -49.751, -69.511, -70.342, -70.361, -69.564, -68.125,
-71.339, -64.286, -59.146, -55.002, -49.771, -69.539, -70.370, -70.390, -69.593,
-68.154, -71.371, -64.314, -59.171, -55.026, -49.791, -69.567, -70.399, -70.419,
-69.622, -68.182, -71.402, -64.342, -59.196, -55.049, -49.810, -69.596, -70.428,
-70.448, -69.651, -68.211, -71.434, -64.370, -59.222, -55.073, -49.830, -69.624,
-70.456, -70.477, -69.680, -68.240, -71.465, -64.397, -59.247, -55.096, -49.850,
 -69.652, -70.485, -70.506, -69.708, -68.268, -71.497, -64.425, -59.273,
-55.120, -49.870, -69.680, -70.514, -70.535, -69.737, -68.297, -71.528, -64.453,
-59.298, -55.143, -49.889, -69.708, -70.543, -70.564, -69.766, -68.326, -71.560,
-64.481, -59.323, -55.166, -49.909, -69.737, -70.571, -70.593, -69.795, -68.354,
-71.591, -64.509, -59.349, -55.190, -49.929, -69.765, -70.600, -70.622, -69.824,
-68.383, -71.623, -64.537, -59.374, -55.213, -49.949, -69.793, -70.629, -70.651,
-69.853, -68.411, -71.654, -64.564, -59.399, -55.237, -49.968, -69.821, -70.658,
-70.680, -69.882, -68.440, -71.685, -64.592, -59.425, -55.260, -49.988, -69.849,
-70.686, -70.709, -69.911, -68.469, -71.717, -64.620, -59.450, -55.284, -50.008,
 -69.878, -70.715, -70.738, -69.940, -68.497, -71.748, -64.648, -59.476,
-55.307, -50.027, -69.906, -70.744, -70.767, -69.969, -68.526, -71.780, -64.676,
-59.501, -55.331, -50.047, -69.934, -70.773, -70.796, -69.998, -68.555, -71.811,
-64.704, -59.526, -55.354, -50.067, -69.962, -70.801, -70.825, -70.027, -68.583,
-71.843, -64.732, -59.552, -55.378, -50.086, -69.990, -70.830, -70.854, -70.056,
-68.612, -71.874, -64.759, -59.577, -55.401, -50.106, -70.019, -70.859, -70.883,
-70.085, -68.641, -71.906, -64.787, -59.603, -55.425, -50.126, -70.047, -70.888,
-70.912, -70.113, -68.669, -71.937, -64.815, -59.628, -55.448, -50.146, -70.075,
-70.916, -70.941, -70.142, -68.698, -71.969, -64.843, -59.653, -55.472, -50.165,
 -70.103, -70.945, -70.970, -70.171, -68.727, -72.000, -64.871, -59.679,
-55.495, -50.185, -70.131, -70.974, -70.999, -70.200, -68.755, -72.032, -64.899,
-59.704, -55.519, -50.205, -70.160, -71.003, -71.028, -70.229, -68.784, -72.063,
-64.927, -59.730, -55.542, -50.224, -70.188, -71.031, -71.057, -70.258, -68.812,
-72.094, -64.954, -59.755, -55.566, -50.244, -70.216, -71.060, -71.086, -70.287,
-68.841, -72.126, -64.982, -59.780, -55.589, -50.264, -70.244, -71.089, -71.115,
-70.316, -68.870, -72.157, -65.010, -59.806, -55.613, -50.283, -70.272, -71.117,
-71.144, -70.345, -68.898, -72.189, -65.038, -59.831, -55.636, -50.303, -70.301,
-71.146, -71.173, -70.374, -68.927, -72.220, -65.066, -59.857, -55.660, -50.323,
 -70.329, -71.175, -71.202, -70.403, -68.956, -72.252, -65.094, -59.882,
-55.683, -50.343, -70.357, -71.204, -71.231, -70.432, -68.984, -72.283, -65.121,
-59.907, -55.707, -50.362, -70.385, -71.232, -71.260, -70.461, -69.013, -72.315,
-65.149, -59.933, -55.730, -50.382, -70.413, -71.261, -71.288, -70.489, -69.042,
-72.346, -65.177, -59.958, -55.753, -50.402, -70.442, -71.290, -71.317, -70.518,
-69.070, -72.378, -65.205, -59.984, -55.777, -50.421, -70.470, -71.319, -71.346,
-70.547, -69.099, -72.409, -65.233, -60.009, -55.800, -50.441, -70.498, -71.347,
-71.375, -70.576, -69.128, -72.441, -65.261, -60.034, -55.824, -50.461, -70.526,
-71.376, -71.404, -70.605, -69.156, -72.472, -65.289, -60.060, -55.847, -50.480,
 -70.554, -71.405, -71.433, -70.634, -69.185, -72.504, -65.316, -60.085,
-55.871, -50.500, -70.583, -71.434, -71.462, -70.663, -69.214, -72.535, -65.344,
-60.111, -55.894, -50.520, -70.611, -71.462, -71.491, -70.692, -69.242, -72.566,
-65.372, -60.136, -55.918, -50.539, -70.639, -71.491, -71.520, -70.721, -69.271,
-72.598, -65.400, -60.161, -55.941, -50.559, -70.667, -71.520, -71.549, -70.750,
-69.299, -72.629, -65.428, -60.187, -55.965, -50.579, -70.695, -71.549, -71.578,
-70.779, -69.328, -72.661, -65.456, -60.212, -55.988, -50.598, -70.724, -71.577,
-71.607, -70.808, -69.357, -72.692, -65.484, -60.237, -56.012, -50.618, -70.752,
-71.606, -71.636, -70.837, -69.385, -72.724, -65.511, -60.263, -56.035, -50.638,
 -70.780, -71.635, -71.665, -70.866, -69.414, -72.755, -65.539, -60.288,
-56.059, -50.657, -70.808, -71.664, -71.694, -70.894, -69.443, -72.787, -65.567,
-60.314, -56.082, -50.677, -70.836, -71.692, -71.723, -70.923, -69.471, -72.818,
-65.595, -60.339, -56.106, -50.697, -70.865, -71.721, -71.752, -70.952, -69.500,
-72.850, -65.623, -60.364, -56.129, -50.716, -70.893, -71.750, -71.781, -70.981,
-69.529, -72.881, -65.651, -60.390, -56.153, -50.736, -70.921, -71.779, -71.810,
-71.010, -69.557, -72.913, -65.679, -60.415, -56.176, -50.756, -70.949, -71.807,
-71.839, -71.039, -69.586, -72.944, -65.706, -60.441, -56.199, -50.775, -70.977,
-71.836, -71.868, -71.068, -69.615, -72.976, -65.734, -60.466, -56.223, -50.795,
 -71.006, -71.865, -71.897, -71.097, -69.643, -73.007, -65.762, -60.491,
-56.246, -50.815, -71.034, -71.894, -71.926, -71.126, -69.672, -73.038, -65.790,
-60.517, -56.270, -50.834, -71.062, -71.922, -71.955, -71.155, -69.701, -73.070,
-65.818, -60.542, -56.293, -50.854, -71.090, -71.951, -71.984, -71.184, -69.729,
-73.101, -65.846, -60.568, -56.317, -50.874, -71.118, -71.980, -72.013, -71.213,
-69.758, -73.133, -65.873, -60.593, -56.340, -50.893, -71.147, -72.008, -72.042,
-71.242, -69.787, -73.164, -65.901, -60.618, -56.364, -50.913, -71.175, -72.037,
-72.071, -71.271, -69.815, -73.196, -65.929, -60.644, -56.387, -50.933, -71.203,
-72.066, -72.100, -71.299, -69.844, -73.227, -65.957, -60.669, -56.411, -50.952,
 -71.231, -72.095, -72.129, -71.328, -69.872, -73.259, -65.985, -60.695,
-56.434, -50.972, -71.259, -72.123, -72.158, -71.357, -69.901, -73.290, -66.013,
-60.720, -56.458, -50.992, -71.288, -72.152, -72.187, -71.386, -69.930, -73.322,
-66.041, -60.745, -56.481, -51.011, -71.316, -72.181, -72.216, -71.415, -69.958,
-73.353, -66.068, -60.771, -56.505, -51.031, -71.344, -72.210, -72.245, -71.444,
-69.987, -73.385, -66.096, -60.796, -56.528, -51.050, -71.372, -72.238, -72.274,
-71.473, -70.016, -73.416, -66.124, -60.822, -56.552, -51.070, -71.400, -72.267,
-72.303, -71.502, -70.044, -73.448, -66.152, -60.847, -56.575, -51.090, -71.429,
-72.296, -72.332, -71.531, -70.073, -73.479, -66.180, -60.872, -56.598, -51.109,
 -71.457, -72.325, -72.361, -71.560, -70.102, -73.511, -66.208, -60.898,
-56.622, -51.129, -71.485, -72.353, -72.390, -71.589, -70.130, -73.542, -66.236,
-60.923, -56.645, -51.149, -71.513, -72.382, -72.419, -71.618, -70.159, -73.573,
-66.263, -60.949, -56.669, -51.168, -71.541, -72.411, -72.447, -71.647, -70.188,
-73.605, -66.291, -60.974, -56.692, -51.188, -71.570, -72.440, -72.476, -71.676,
-70.216, -73.636, -66.319, -60.999, -56.716, -51.208, -71.598, -72.468, -72.505,
-71.705, -70.245, -73.668, -66.347, -61.025, -56.739, -51.227, -71.626, -72.497,
-72.534, -71.733, -70.274, -73.699, -66.375, -61.050, -56.763, -51.247, -71.654,
-72.526, -72.563, -71.762, -70.302, -73.731, -66.403, -61.076, -56.786, -51.266,
 -71.682, -72.555, -72.592, -71.791, -70.331, -73.762, -66.431, -61.101,
-56.810, -51.286, -71.711, -72.583, -72.621, -71.820, -70.360, -73.794, -66.458,
-61.126, -56.833, -51.306, -71.739, -72.612, -72.650, -71.849, -70.388, -73.825,
-66.486, -61.152, -56.857, -51.325, -71.767, -72.641, -72.679, -71.878, -70.417,
-73.857, -66.514, -61.177, -56.880, -51.345, -71.795, -72.670, -72.708, -71.907,
-70.445, -73.888, -66.542, -61.203, -56.903, -51.365, -71.823, -72.698, -72.737,
-71.936, -70.474, -73.920, -66.570, -61.228, -56.927, -51.384, -71.852, -72.727,
-72.766, -71.965, -70.503, -73.951, -66.598, -61.253, -56.950, -51.404, -71.880,
-72.756, -72.795, -71.994, -70.531, -73.983, -66.626, -61.279, -56.974, -51.423,
 -71.908, -72.785, -72.824, -72.023, -70.560, -74.014, -66.653, -61.304,
-56.997, -51.443, -71.936, -72.813, -72.853, -72.052, -70.589, -74.045, -66.681,
-61.330, -57.021, -51.463, -71.964, -72.842, -72.882, -72.081, -70.617, -74.077,
-66.709, -61.355, -57.044, -51.482, -71.993, -72.871, -72.911, -72.110, -70.646,
-74.108, -66.737, -61.380, -57.068, -51.502, -72.021, -72.900, -72.940, -72.139,
-70.675, -74.140, -66.765, -61.406, -57.091, -51.522, -72.049, -72.928, -72.969,
-72.167, -70.703, -74.171, -66.793, -61.431, -57.115, -51.541, -72.077, -72.957,
-72.998, -72.196, -70.732, -74.203, -66.821, -61.457, -57.138, -51.561, -72.105,
-72.986, -73.027, -72.225, -70.761, -74.234, -66.848, -61.482, -57.162, -51.580,
 -72.134, -73.015, -73.056, -72.254, -70.789, -74.266, -66.876, -61.507,
-57.185, -51.600, -72.162, -73.043, -73.085, -72.283, -70.818, -74.297, -66.904,
-61.533, -57.208, -51.620, -72.190, -73.072, -73.114, -72.312, -70.847, -74.329,
-66.932, -61.558, -57.232, -51.639, -72.218, -73.101, -73.143, -72.341, -70.875,
-74.360, -66.960, -61.584, -57.255, -51.659, -72.246, -73.130, -73.172, -72.370,
-70.904, -74.392, -66.988, -61.609, -57.279, -51.678, -72.275, -73.158, -73.201,
-72.399, -70.933, -74.423, -67.016, -61.634, -57.302, -51.698, -72.303, -73.187,
-73.230, -72.428, -70.961, -74.455, -67.043, -61.660, -57.326, -51.718, -72.331,
-73.216, -73.259, -72.457, -70.990, -74.486, -67.071, -61.685, -57.349, -51.737,
 -72.359, -73.245, -73.288, -72.486, -71.019, -74.518, -67.099, -61.711,
-57.373, -51.757, -72.387, -73.273, -73.317, -72.515, -71.047, -74.549, -67.127,
-61.736, -57.396, -51.776, -72.416, -73.302, -73.346, -72.544, -71.076, -74.581,
-67.155, -61.761, -57.420, -51.796, -72.444, -73.331, -73.375, -72.573, -71.104,
-74.612, -67.183, -61.787, -57.443, -51.816, -72.472, -73.360, -73.404, -72.601,
-71.133, -74.643, -67.211, -61.812, -57.467, -51.835, -72.500, -73.388, -73.433,
-72.630, -71.162, -74.675, -67.238, -61.838, -57.490, -51.855, -72.528, -73.417,
-73.462, -72.659, -71.190, -74.706, -67.266, -61.863, -57.513, -51.874, -72.557,
-73.446, -73.491, -72.688, -71.219, -74.738, -67.294, -61.888, -57.537, -51.894,
 -72.585, -73.475, -73.520, -72.717, -71.248, -74.769, -67.322, -61.914,
-57.560, -51.914, -72.613, -73.503, -73.549, -72.746, -71.276, -74.801, -67.350,
-61.939, -57.584, -51.933, -72.641, -73.532, -73.578, -72.775, -71.305, -74.832,
-67.378, -61.965, -57.607, -51.953, -72.669, -73.561, -73.607, -72.804, -71.334,
-74.864, -67.406, -61.990, -57.631, -51.972, -72.698, -73.590, -73.636, -72.833,
-71.362, -74.895, -67.433, -62.015, -57.654, -51.992, -72.726, -73.618, -73.665,
-72.862, -71.391, -74.927, -67.461, -62.041, -57.678, -52.012, -72.754, -73.647,
-73.694, -72.891, -71.420, -74.958, -67.489, -62.066, -57.701, -52.031, -72.782,
-73.676, -73.723, -72.920, -71.448, -74.990, -67.517, -62.092, -57.725, -52.051,
 -72.810, -73.705, -73.752, -72.949, -71.477, -75.021, -67.545, -62.117,
-57.748, -52.070, -72.839, -73.733, -73.781, -72.978, -71.506, -75.053, -67.573,
-62.142, -57.771, -52.090, -72.867, -73.762, -73.810, -73.007, -71.534, -75.084,
-67.601, -62.168, -57.795, -52.110, -72.895, -73.791, -73.838, -73.035, -71.563,
-75.116, -67.628, -62.193, -57.818, -52.129, -72.923, -73.820, -73.867, -73.064,
-71.592, -75.147, -67.656, -62.219, -57.842, -52.149, -72.951, -73.848, -73.896,
-73.093, -71.620, -75.179, -67.684, -62.244, -57.865, -52.168, -72.980, -73.877,
-73.925, -73.122, -71.649, -75.210, -67.712, -62.269, -57.889, -52.188, -73.008,
-73.906, -73.954, -73.151, -71.678, -75.241, -67.740, -62.295, -57.912, -52.207,
 -73.036, -73.935, -73.983, -73.180, -71.706, -75.273, -67.768, -62.320,
-57.936, -52.227, -73.064, -73.963, -74.012, -73.209, -71.735, -75.304, -67.796,
-62.346, -57.959, -52.247, -73.093, -73.992, -74.041, -73.238, -71.764, -75.336,
-67.823, -62.371, -57.982, -52.266, -73.121, -74.021, -74.070, -73.267, -71.792,
-75.367, -67.851, -62.396, -58.006, -52.286, -73.149, -74.050, -74.099, -73.296,
-71.821, -75.399, -67.879, -62.422, -58.029, -52.305, -73.177, -74.078, -74.128,
-73.325, -71.850, -75.430, -67.907, -62.447, -58.053, -52.325, -73.205, -74.107,
-74.157, -73.354, -71.878, -75.462, -67.935, -62.473, -58.076, -52.345, -73.234,
-74.136, -74.186, -73.383, -71.907, -75.493, -67.963, -62.498, -58.100, -52.364,
 -73.262, -74.165, -74.215, -73.412, -71.935, -75.525, -67.991, -62.523,
-58.123, -52.384, -73.290, -74.193, -74.244, -73.441, -71.964, -75.556, -68.018,
-62.549, -58.147, -52.403, -73.318, -74.222, -74.273, -73.470, -71.993, -75.588,
-68.046, -62.574, -58.170, -52.423, -73.346, -74.251, -74.302, -73.498, -72.021,
-75.619, -68.074, -62.600, -58.193, -52.442, -73.375, -74.280, -74.331, -73.527,
-72.050, -75.651, -68.102, -62.625, -58.217, -52.462, -73.403, -74.308, -74.360,
-73.556, -72.079, -75.682, -68.130, -62.650, -58.240, -52.481, -73.431, -74.337,
-74.389, -73.585, -72.107, -75.714, -68.158, -62.676, -58.264, -52.501, -73.459,
-74.366, -74.418, -73.614, -72.136, -75.745, -68.186, -62.701, -58.287, -52.521,
 -73.487, -74.395, -74.447, -73.643, -72.165, -75.777, -68.213, -62.727,
-58.311, -52.540, -73.516, -74.423, -74.476, -73.672, -72.193, -75.808, -68.241,
-62.752, -58.334, -52.560, -73.544, -74.452, -74.505, -73.701, -72.222, -75.839,
-68.269, -62.777, -58.358, -52.579, -73.572, -74.481, -74.534, -73.730, -72.251,
-75.871, -68.297, -62.803, -58.381, -52.599, -73.600, -74.510, -74.563, -73.759,
-72.279, -75.902, -68.325, -62.828, -58.405, -52.618, -73.628, -74.538, -74.592,
-73.788, -72.308, -75.934, -68.353, -62.854, -58.428, -52.638, -73.657, -74.567,
-74.621, -73.817, -72.337, -75.965, -68.381, -62.879, -58.451, -52.658, -73.685,
-74.596, -74.650, -73.846, -72.365, -75.997, -68.408, -62.905, -58.475, -52.677,
 -73.713, -74.625, -74.679, -73.875, -72.394, -76.028, -68.436, -62.930,
-58.498, -52.697, -73.741, -74.653, -74.708, -73.904, -72.423, -76.060, -68.464,
-62.955, -58.522, -52.716, -73.769, -74.682, -74.737, -73.933, -72.451, -76.091,
-68.492, -62.981, -58.545, -52.736, -73.798, -74.711, -74.766, -73.961, -72.480,
-76.123, -68.520, -63.006, -58.569, -52.755, -73.826, -74.740, -74.795, -73.990,
-72.509, -76.154, -68.548, -63.032, -58.592, -52.775, -73.854, -74.768, -74.824,
-74.019, -72.537, -76.186, -68.576, -63.057, -58.615, -52.794, -73.882, -74.797,
-74.853, -74.048, -72.566, -76.217, -68.604, -63.082, -58.639, -52.814, -73.911,
-74.826, -74.882, -74.077, -72.595, -76.249, -68.631, -63.108, -58.662, -52.834,
 -73.939, -74.855, -74.911, -74.106, -72.623, -76.280, -68.659, -63.133,
-58.686, -52.853, -73.967, -74.883, -74.940, -74.135, -72.652, -76.312, -68.687,
-63.159, -58.709, -52.873, -73.995, -74.912, -74.969, -74.164, -72.681, -76.343,
-68.715, -63.184, -58.733, -52.892, -74.023, -74.941, -74.998, -74.193, -72.709,
-76.375, -68.743, -63.209, -58.756, -52.912, -74.052, -74.970, -75.027, -74.222,
-72.738, -76.406, -68.771, -63.235, -58.780, -52.931, -74.080, -74.998, -75.056,
-74.251, -72.767, -76.438, -68.799, -63.260, -58.803, -52.951, -74.108, -75.027,
-75.085, -74.280, -72.795, -76.469, -68.826, -63.286, -58.826, -52.970, -74.136,
-75.056, -75.114, -74.309, -72.824, -76.501, -68.854, -63.311, -58.850, -52.990,
 -74.164, -75.085, -75.143, -74.338, -72.853, -76.532, -68.882, -63.336,
-58.873, -53.009, -74.193, -75.113, -75.172, -74.367, -72.881, -76.563, -68.910,
-63.362, -58.897, -53.029, -74.221, -75.142, -75.201, -74.396, -72.910, -76.595,
-68.938, -63.387, -58.920, -53.048, -74.249, -75.171, -75.230, -74.425, -72.939,
-76.626, -68.966, -63.413, -58.944, -53.068, -74.277, -75.200, -75.259, -74.453,
-72.967, -76.658, -68.994, -63.438, -58.967, -53.088, -74.305, -75.228, -75.288,
-74.482, -72.996, -76.689, -69.021, -63.463, -58.990, -53.107, -74.334, -75.257,
-75.317, -74.511, -73.024, -76.721, -69.049, -63.489, -59.014, -53.127, -74.362,
-75.286, -75.346, -74.540, -73.053, -76.752, -69.077, -63.514, -59.037, -53.146,
 -74.390, -75.315, -75.375, -74.569, -73.082, -76.784, -69.105, -63.540,
-59.061, -53.166, -74.418, -75.343, -75.404, -74.598, -73.110, -76.815, -69.133,
-63.565, -59.084, -53.185, -74.447, -75.372, -75.433, -74.627, -73.139, -76.847,
-69.161, -63.590, -59.108, -53.205, -74.475, -75.401, -75.462, -74.656, -73.168,
-76.878, -69.189, -63.616, -59.131, -53.224, -74.503, -75.430, -75.491, -74.685,
-73.196, -76.910, -69.217, -63.641, -59.155, -53.244, -74.531, -75.458, -75.520,
-74.714, -73.225, -76.941, -69.244, -63.667, -59.178, -53.263, -74.559, -75.487,
-75.549, -74.743, -73.254, -76.973, -69.272, -63.692, -59.201, -53.283, -74.588,
-75.516, -75.578, -74.772, -73.282, -77.004, -69.300, -63.718, -59.225, -53.302,
 -74.616, -75.545, -75.607, -74.801, -73.311, -77.036, -69.328, -63.743,
-59.248, -53.322, -74.644, -75.573, -75.636, -74.830, -73.340, -77.067, -69.356,
-63.768, -59.272, -53.341, -74.672, -75.602, -75.665, -74.859, -73.368, -77.099,
-69.384, -63.794, -59.295, -53.361, -74.700, -75.631, -75.694, -74.888, -73.397,
-77.130, -69.412, -63.819, -59.319, -53.381, -74.729, -75.660, -75.722, -74.917,
-73.426, -77.162, -69.439, -63.845, -59.342, -53.400, -74.757, -75.688, -75.751,
-74.945, -73.454, -77.193, -69.467, -63.870, -59.365, -53.420, -74.785, -75.717,
-75.780, -74.974, -73.483, -77.225, -69.495, -63.895, -59.389, -53.439, -74.813,
-75.746, -75.809, -75.003, -73.512, -77.256, -69.523, -63.921, -59.412, -53.459,
 -74.841, -75.775, -75.838, -75.032, -73.540, -77.287, -69.551, -63.946,
-59.436, -53.478, -74.870, -75.804, -75.867, -75.061, -73.569, -77.319, -69.579,
-63.972, -59.459, -53.498, -74.898, -75.832, -75.896, -75.090, -73.598, -77.350,
-69.607, -63.997, -59.483, -53.517, -74.926, -75.861, -75.925, -75.119, -73.626,
-77.382, -69.634, -64.022, -59.506, -53.537, -74.954, -75.890, -75.954, -75.148,
-73.655, -77.413, -69.662, -64.048, -59.529, -53.556, -74.983, -75.919, -75.983,
-75.177, -73.684, -77.445, -69.690, -64.073, -59.553, -53.576, -75.011, -75.947,
-76.012, -75.206, -73.712, -77.476, -69.718, -64.099, -59.576, -53.595, -75.039,
-75.976, -76.041, -75.235, -73.741, -77.508, -69.746, -64.124, -59.600, -53.615,
 -75.067, -76.005, -76.070, -75.264, -73.770, -77.539, -69.774, -64.149,
-59.623, -53.634, -75.095, -76.034, -76.099, -75.293, -73.798, -77.571, -69.802,
-64.175, -59.647, -53.654, -75.124, -76.062, -76.128, -75.322, -73.827, -77.602,
-69.830, -64.200, -59.670, -53.673, -75.152, -76.091, -76.157, -75.351, -73.856,
-77.634, -69.857, -64.226, -59.693, -53.693, -75.180, -76.120, -76.186, -75.380,
-73.884, -77.665, -69.885, -64.251, -59.717, -53.712, -75.208, -76.149, -76.215,
-75.409, -73.913, -77.697, -69.913, -64.277, -59.740, -53.732, -75.236, -76.177,
-76.244, -75.437, -73.942, -77.728, -69.941, -64.302, -59.764, -53.751, -75.265,
-76.206, -76.273, -75.466, -73.970, -77.760, -69.969, -64.327, -59.787, -53.771,
 -75.293, -76.235, -76.302, -75.495, -73.999, -77.791, -69.997, -64.353,
-59.811, -53.790, -75.321, -76.264, -76.331, -75.524, -74.028, -77.823, -70.025,
-64.378, -59.834, -53.810, -75.349, -76.292, -76.360, -75.553, -74.056, -77.854,
-70.052, -64.404, -59.857, -53.829, -75.378, -76.321, -76.389, -75.582, -74.085,
-77.886, -70.080, -64.429, -59.881, -53.849, -75.406, -76.350, -76.418, -75.611,
-74.114, -77.917, -70.108, -64.454, -59.904, -53.868, -75.434, -76.379, -76.447,
-75.640, -74.142, -77.949, -70.136, -64.480, -59.928, -53.888, -75.462, -76.407,
-76.476, -75.669, -74.171, -77.980, -70.164, -64.505, -59.951, -53.907, -75.490,
-76.436, -76.505, -75.698, -74.200, -78.012, -70.192, -64.531, -59.974, -53.927,
 -75.519, -76.465, -76.534, -75.727, -74.228, -78.043, -70.220, -64.556,
-59.998, -53.946, -75.547, -76.494, -76.563, -75.756, -74.257, -78.075, -70.248,
-64.581, -60.021, -53.966, -75.575, -76.522, -76.592, -75.785, -74.286, -78.106,
-70.275, -64.607, -60.045, -53.985, -75.603, -76.551, -76.621, -75.814, -74.314,
-78.137, -70.303, -64.632, -60.068, -54.005, -75.631, -76.580, -76.650, -75.843,
-74.343, -78.169, -70.331, -64.658, -60.092, -54.024, -75.660, -76.609, -76.679,
-75.872, -74.372, -78.200, -70.359, -64.683, -60.115, -54.044, -75.688, -76.637,
-76.708, -75.901, -74.400, -78.232, -70.387, -64.708, -60.138, -54.063, -75.716,
-76.666, -76.737, -75.930, -74.429, -78.263, -70.415, -64.734, -60.162, -54.083,
 -75.744, -76.695, -76.766, -75.958, -74.458, -78.295, -70.443, -64.759,
-60.185, -54.102, -75.773, -76.724, -76.795, -75.987, -74.486, -78.326, -70.470,
-64.785, -60.209, -54.122, -75.801, -76.753, -76.824, -76.016, -74.515, -78.358,
-70.498, -64.810, -60.232, -54.141, -75.829, -76.781, -76.853, -76.045, -74.544,
-78.389, -70.526, -64.836, -60.256, -54.161, -75.857, -76.810, -76.882, -76.074,
-74.572, -78.421, -70.554, -64.861, -60.279, -54.180, -75.885, -76.839, -76.911,
-76.103, -74.601, -78.452, -70.582, -64.886, -60.302, -54.200, -75.914, -76.868,
-76.940, -76.132, -74.630, -78.484, -70.610, -64.912, -60.326, -54.219, -75.942,
-76.896, -76.969, -76.161, -74.658, -78.515, -70.638, -64.937, -60.349, -54.239,
 -75.970, -76.925, -76.998, -76.190, -74.687, -78.547, -70.666, -64.963,
-60.373, -54.258, -75.998, -76.954, -77.027, -76.219, -74.716, -78.578, -70.693,
-64.988, -60.396, -54.278, -76.026, -76.983, -77.056, -76.248, -74.744, -78.610,
-70.721, -65.013, -60.419, -54.297, -76.055, -77.011, -77.085, -76.277, -74.773,
-78.641, -70.749, -65.039, -60.443, -54.317, -76.083, -77.040, -77.114, -76.306,
-74.802, -78.673, -70.777, -65.064, -60.466, -54.336, -76.111, -77.069, -77.143,
-76.335, -74.830, -78.704, -70.805, -65.090, -60.490, -54.356, -76.139, -77.098,
-77.172, -76.364, -74.859, -78.736, -70.833, -65.115, -60.513, -54.375, -76.168,
-77.126, -77.201, -76.393, -74.888, -78.767, -70.861, -65.140, -60.537, -54.395,
 -76.196, -77.155, -77.230, -76.422, -74.916, -78.799, -70.888, -65.166,
-60.560, -54.414, -76.224, -77.184, -77.259, -76.451, -74.945, -78.830, -70.916,
-65.191, -60.583, -54.434, -76.252, -77.213, -77.288, -76.479, -74.974, -78.862,
-70.944, -65.217, -60.607, -54.453, -76.280, -77.241, -77.317, -76.508, -75.002,
-78.893, -70.972, -65.242, -60.630, -54.473, -76.309, -77.270, -77.346, -76.537,
-75.031, -78.925, -71.000, -65.268, -60.654, -54.492, -76.337, -77.299, -77.375,
-76.566, -75.060, -78.956, -71.028, -65.293, -60.677, -54.511, -76.365, -77.328,
-77.404, -76.595, -75.088, -78.988, -71.056, -65.318, -60.700, -54.531, -76.393,
-77.357, -77.433, -76.624, -75.117, -79.019, -71.084, -65.344, -60.724, -54.550,
 -76.421, -77.385, -77.462, -76.653, -75.146, -79.051, -71.111, -65.369,
-60.747, -54.570, -76.450, -77.414, -77.491, -76.682, -75.174, -79.082, -71.139,
-65.395, -60.771, -54.589, -76.478, -77.443, -77.520, -76.711, -75.203, -79.114,
-71.167, -65.420, -60.794, -54.609, -76.506, -77.472, -77.549, -76.740, -75.232,
-79.145, -71.195, -65.445, -60.817, -54.628, -76.534, -77.500, -77.578, -76.769,
-75.260, -79.176, -71.223, -65.471, -60.841, -54.648, -76.563, -77.529, -77.607,
-76.798, -75.289, -79.208, -71.251, -65.496, -60.864, -54.667, -76.591, -77.558,
-77.636, -76.827, -75.318, -79.239, -71.279, -65.522, -60.888, -54.687, -76.619,
-77.587, -77.665, -76.856, -75.346, -79.271, -71.307, -65.547, -60.911, -54.706,
 -76.647, -77.615, -77.694, -76.885, -75.375, -79.302, -71.334, -65.573,
-60.934, -54.726, -76.675, -77.644, -77.723, -76.914, -75.404, -79.334, -71.362,
-65.598, -60.958, -54.745, -76.704, -77.673, -77.752, -76.943, -75.432, -79.365,
-71.390, -65.623, -60.981, -54.765, -76.732, -77.702, -77.781, -76.972, -75.461,
-79.397, -71.418, -65.649, -61.005, -54.784, -76.760, -77.730, -77.810, -77.001,
-75.490, -79.428, -71.446, -65.674, -61.028, -54.804, -76.788, -77.759, -77.839,
-77.029, -75.518, -79.460, -71.474, -65.700, -61.052, -54.823, -76.817, -77.788,
-77.868, -77.058, -75.547, -79.491, -71.502, -65.725, -61.075, -54.842, -76.845,
-77.817, -77.897, -77.087, -75.576, -79.523, -71.530, -65.750, -61.098, -54.862,
 -76.873, -77.845, -77.926, -77.116, -75.604, -79.554, -71.557, -65.776,
-61.122, -54.881, -76.901, -77.874, -77.955, -77.145, -75.633, -79.586, -71.585,
-65.801, -61.145, -54.901, -76.929, -77.903, -77.984, -77.174, -75.662, -79.617,
-71.613, -65.827, -61.169, -54.920, -76.958, -77.932, -78.013, -77.203, -75.690,
-79.649, -71.641, -65.852, -61.192, -54.940, -76.986, -77.961, -78.042, -77.232,
-75.719, -79.680, -71.669, -65.877, -61.215, -54.959, -77.014, -77.989, -78.071,
-77.261, -75.748, -79.712, -71.697, -65.903, -61.239, -54.979, -77.042, -78.018,
-78.100, -77.290, -75.776, -79.743, -71.725, -65.928, -61.262, -54.998, -77.070,
-78.047, -78.129, -77.319, -75.805, -79.775, -71.752, -65.954, -61.286, -55.018,
 -77.099, -78.076, -78.158, -77.348, -75.834, -79.806, -71.780, -65.979,
-61.309, -55.037, -77.127, -78.104, -78.187, -77.377, -75.862, -79.838, -71.808,
-66.005, -61.332, -55.057, -77.155, -78.133, -78.216, -77.406, -75.891, -79.869,
-71.836, -66.030, -61.356, -55.076, -77.183, -78.162, -78.245, -77.435, -75.920,
-79.901, -71.864, -66.055, -61.379, -55.095, -77.212, -78.191, -78.274, -77.464,
-75.948, -79.932, -71.892, -66.081, -61.403, -55.115, -77.240, -78.219, -78.303,
-77.493, -75.977, -79.964, -71.920, -66.106, -61.426, -55.134, -77.268, -78.248,
-78.332, -77.522, -76.006, -79.995, -71.948, -66.132, -61.449, -55.154, -77.296,
-78.277, -78.361, -77.551, -76.034, -80.027, -71.975, -66.157, -61.473, -55.173,
 -77.324, -78.306, -78.390, -77.580, -76.063, -80.058, -72.003, -66.182,
-61.496, -55.193, -77.353, -78.334, -78.419, -77.608, -76.092, -80.090, -72.031,
-66.208, -61.520, -55.212, -77.381, -78.363, -78.448, -77.637, -76.120, -80.121,
-72.059, -66.233, -61.543, -55.232, -77.409, -78.392, -78.477, -77.666, -76.149,
-80.153, -72.087, -66.259, -61.566, -55.251, -77.437, -78.421, -78.506, -77.695,
-76.178, -80.184, -72.115, -66.284, -61.590, -55.271, -77.466, -78.450, -78.535,
-77.724, -76.206, -80.216, -72.143, -66.310, -61.613, -55.290, -77.494, -78.478,
-78.564, -77.753, -76.235, -80.247, -72.171, -66.335, -61.637, -55.309, -77.522,
-78.507, -78.593, -77.782, -76.264, -80.279, -72.198, -66.360, -61.660, -55.329,
 -77.550, -78.536, -78.622, -77.811, -76.292, -80.310, -72.226, -66.386,
-61.683, -55.348, -77.578, -78.565, -78.651, -77.840, -76.321, -80.341, -72.254,
-66.411, -61.707, -55.368, -77.607, -78.593, -78.680, -77.869, -76.350, -80.373,
-72.282, -66.437, -61.730, -55.387, -77.635, -78.622, -78.709, -77.898, -76.378,
-80.404, -72.310, -66.462, -61.754, -55.407, -77.663, -78.651, -78.738, -77.927,
-76.407, -80.436, -72.338, -66.487, -61.777, -55.426, -77.691, -78.680, -78.767,
-77.956, -76.436, -80.467, -72.366, -66.513, -61.800, -55.446, -77.720, -78.708,
-78.796, -77.985, -76.464, -80.499, -72.394, -66.538, -61.824, -55.465, -77.748,
-78.737, -78.825, -78.014, -76.493, -80.530, -72.421, -66.564, -61.847, -55.484,
 -77.776, -78.766, -78.854, -78.043, -76.522, -80.562, -72.449, -66.589,
-61.870, -55.504, -77.804, -78.795, -78.883, -78.072, -76.550, -80.593, -72.477,
-66.615, -61.894, -55.523, -77.832, -78.823, -78.912, -78.101, -76.579, -80.625,
-72.505, -66.640, -61.917, -55.543, -77.861, -78.852, -78.941, -78.130, -76.608,
-80.656, -72.533, -66.665, -61.941, -55.562, -77.889, -78.881, -78.970, -78.159,
-76.636, -80.688, -72.561, -66.691, -61.964, -55.582, -77.917, -78.910, -78.999,
-78.187, -76.665, -80.719, -72.589, -66.716, -61.987, -55.601, -77.945, -78.939,
-79.028, -78.216, -76.694, -80.751, -72.617, -66.742, -62.011, -55.621, -77.974,
-78.967, -79.057, -78.245, -76.722, -80.782, -72.644, -66.767, -62.034, -55.640,
 -78.002, -78.996, -79.086, -78.274, -76.751, -80.814, -72.672, -66.792,
-62.058, -55.659, -78.030, -79.025, -79.115, -78.303, -76.780, -80.845, -72.700,
-66.818, -62.081, -55.679, -78.058, -79.054, -79.144, -78.332, -76.808, -80.877,
-72.728, -66.843, -62.104, -55.698, -78.086, -79.082, -79.173, -78.361, -76.837,
-80.908, -72.756, -66.869, -62.128, -55.718, -78.115, -79.111, -79.202, -78.390,
-76.866, -80.940, -72.784, -66.894, -62.151, -55.737, -78.143, -79.140, -79.231,
-78.419, -76.894, -80.971, -72.812, -66.920, -62.175, -55.757, -78.171, -79.169,
-79.260, -78.448, -76.923, -81.003, -72.840, -66.945, -62.198, -55.776, -78.199,
-79.197, -79.289, -78.477, -76.952, -81.034, -72.867, -66.970, -62.221, -55.795,
 -78.228, -79.226, -79.318, -78.506, -76.980, -81.066, -72.895, -66.996,
-62.245, -55.815, -78.256, -79.255, -79.347, -78.535, -77.009, -81.097, -72.923,
-67.021, -62.268, -55.834, -78.284, -79.284, -79.376, -78.564, -77.038, -81.129,
-72.951, -67.047, -62.291, -55.854, -78.312, -79.312, -79.405, -78.593, -77.066,
-81.160, -72.979, -67.072, -62.315, -55.873, -78.340, -79.341, -79.434, -78.622,
-77.095, -81.192, -73.007, -67.097, -62.338, -55.893, -78.369, -79.370, -79.463,
-78.651, -77.124, -81.223, -73.035, -67.123, -62.362, -55.912, -78.397, -79.399,
-79.492, -78.680, -77.152, -81.255, -73.063, -67.148, -62.385, -55.931, -78.425,
-79.428, -79.521, -78.709, -77.181, -81.286, -73.090, -67.174, -62.408, -55.951,
 -78.453, -79.456, -79.550, -78.738, -77.210, -81.318, -73.118, -67.199,
-62.432, -55.970, -78.482, -79.485, -79.579, -78.767, -77.238, -81.349, -73.146,
-67.225, -62.455, -55.990, -78.510, -79.514, -79.608, -78.795, -77.267, -81.381,
-73.174, -67.250, -62.479, -56.009, -78.538, -79.543, -79.637, -78.824, -77.296,
-81.412, -73.202, -67.275, -62.502, -56.029, -78.566, -79.571, -79.666, -78.853,
-77.324, -81.444, -73.230, -67.301, -62.525, -56.048, -78.594, -79.600, -79.695,
-78.882, -77.353, -81.475, -73.258, -67.326, -62.549, -56.067, -78.623, -79.629,
-79.724, -78.911, -77.382, -81.507, -73.286, -67.352, -62.572, -56.087, -78.651,
-79.658, -79.753, -78.940, -77.410, -81.538, -73.313, -67.377, -62.595, -56.106,
 -78.679, -79.686, -79.782, -78.969, -77.439, -81.570, -73.341, -67.403,
-62.619, -56.126, -78.707, -79.715, -79.811, -78.998, -77.468, -81.601, -73.369,
-67.428, -62.642, -56.145, -78.736, -79.744, -79.840, -79.027, -77.496, -81.633,
-73.397, -67.453, -62.666, -56.165, -78.764, -79.773, -79.869, -79.056, -77.525,
-81.664, -73.425, -67.479, -62.689, -56.184, -78.792, -79.802, -79.898, -79.085,
-77.554, -81.696, -73.453, -67.504, -62.712, -56.203, -78.820, -79.830, -79.927,
-79.114, -77.583, -81.727, -73.481, -67.530, -62.736, -56.223, -78.848, -79.859,
-79.956, -79.143, -77.611, -81.759, -73.509, -67.555, -62.759, -56.242, -78.877,
-79.888, -79.985, -79.172, -77.640, -81.790, -73.536, -67.580, -62.783, -56.262,
 -78.905, -79.917, -80.014, -79.201, -77.669, -81.822, -73.564, -67.606,
-62.806, -56.281, -78.933, -79.945, -80.043, -79.230, -77.697, -81.853, -73.592,
-67.631, -62.829, -56.301, -78.961, -79.974, -80.072, -79.259, -77.726, -81.885,
-73.620, -67.657, -62.853, -56.320, -78.990, -80.003, -80.101, -79.288, -77.755,
-81.916, -73.648, -67.682, -62.876, -56.339, -79.018, -80.032, -80.130, -79.317,
-77.783, -81.947, -73.676, -67.708, -62.899, -56.359, -79.046, -80.060, -80.159,
-79.346, -77.812, -81.979, -73.704, -67.733, -62.923, -56.378, -79.074, -80.089,
-80.188, -79.375, -77.841, -82.010, -73.732, -67.758, -62.946, -56.398, -79.102,
-80.118, -80.217, -79.404, -77.869, -82.042, -73.759, -67.784, -62.970, -56.417,
 -79.131, -80.147, -80.246, -79.432, -77.898, -82.073, -73.787, -67.809,
-62.993, -56.436, -79.159, -80.176, -80.275, -79.461, -77.927, -82.105, -73.815,
-67.835, -63.016, -56.456, -79.187, -80.204, -80.304, -79.490, -77.955, -82.136,
-73.843, -67.860, -63.040, -56.475, -79.215, -80.233, -80.333, -79.519, -77.984,
-82.168, -73.871, -67.886, -63.063, -56.495, -79.244, -80.262, -80.362, -79.548,
-78.013, -82.199, -73.899, -67.911, -63.086, -56.514, -79.272, -80.291, -80.391,
-79.577, -78.041, -82.231, -73.927, -67.936, -63.110, -56.534, -79.300, -80.319,
-80.420, -79.606, -78.070, -82.262, -73.955, -67.962, -63.133, -56.553, -79.328,
-80.348, -80.449, -79.635, -78.099, -82.294, -73.982, -67.987, -63.157, -56.572,
 -79.356, -80.377, -80.478, -79.664, -78.127, -82.325, -74.010, -68.013,
-63.180, -56.592, -79.385, -80.406, -80.507, -79.693, -78.156, -82.357, -74.038,
-68.038, -63.203, -56.611, -79.413, -80.434, -80.536, -79.722, -78.185, -82.388,
-74.066, -68.063, -63.227, -56.631, -79.441, -80.463, -80.565, -79.751, -78.213,
-82.420, -74.094, -68.089, -63.250, -56.650, -79.469, -80.492, -80.594, -79.780,
-78.242, -82.451, -74.122, -68.114, -63.273, -56.669, -79.498, -80.521, -80.623,
-79.809, -78.271, -82.483, -74.150, -68.140, -63.297, -56.689, -79.526, -80.550,
-80.652, -79.838, -78.299, -82.514, -74.178, -68.165, -63.320, -56.708, -79.554,
-80.578, -80.681, -79.867, -78.328, -82.546, -74.205, -68.191, -63.343, -56.728,
 -79.582, -80.607, -80.710, -79.896, -78.357, -82.577, -74.233, -68.216,
-63.367, -56.747, -79.611, -80.636, -80.739, -79.925, -78.385, -82.609, -74.261,
-68.241, -63.390, -56.766, -79.639, -80.665, -80.768, -79.954, -78.414, -82.640,
-74.289, -68.267, -63.414, -56.786, -79.667, -80.693, -80.797, -79.983, -78.443,
-82.672, -74.317, -68.292, -63.437, -56.805, -79.695, -80.722, -80.826, -80.012,
-78.471, -82.703, -74.345, -68.318, -63.460, -56.825, -79.723, -80.751, -80.855,
-80.041, -78.500, -82.735, -74.373, -68.343, -63.484, -56.844, -79.752, -80.780,
-80.884, -80.069, -78.529, -82.766, -74.401, -68.369, -63.507, -56.863, -79.780,
-80.808, -80.913, -80.098, -78.557, -82.798, -74.428, -68.394, -63.530, -56.883,
 -79.808, -80.837, -80.942, -80.127, -78.586, -82.829, -74.456, -68.419,
-63.554, -56.902, -79.836, -80.866, -80.971, -80.156, -78.615, -82.861, -74.484,
-68.445, -63.577, -56.922, -79.865, -80.895, -81.000, -80.185, -78.643, -82.892,
-74.512, -68.470, -63.600, -56.941, -79.893, -80.924, -81.029, -80.214, -78.672,
-82.924, -74.540, -68.496, -63.624, -56.960, -79.921, -80.952, -81.058, -80.243,
-78.701, -82.955, -74.568, -68.521, -63.647, -56.980, -79.949, -80.981, -81.087,
-80.272, -78.729, -82.987, -74.596, -68.546, -63.671, -56.999, -79.977, -81.010,
-81.116, -80.301, -78.758, -83.018, -74.624, -68.572, -63.694, -57.019, -80.006,
-81.039, -81.145, -80.330, -78.787, -83.050, -74.652, -68.597, -63.717, -57.038,
 -80.034, -81.067, -81.174, -80.359, -78.816, -83.081, -74.679, -68.623,
-63.741, -57.057, -80.062, -81.096, -81.203, -80.388, -78.844, -83.113, -74.707,
-68.648, -63.764, -57.077, -80.090, -81.125, -81.232, -80.417, -78.873, -83.144,
-74.735, -68.674, -63.787, -57.096, -80.119, -81.154, -81.261, -80.446, -78.902,
-83.176, -74.763, -68.699, -63.811, -57.116, -80.147, -81.183, -81.290, -80.475,
-78.930, -83.207, -74.791, -68.724, -63.834, -57.135, -80.175, -81.211, -81.319,
-80.504, -78.959, -83.239, -74.819, -68.750, -63.857, -57.154, -80.203, -81.240,
-81.348, -80.533, -78.988, -83.270, -74.847, -68.775, -63.881, -57.174, -80.232,
-81.269, -81.377, -80.562, -79.016, -83.302, -74.875, -68.801, -63.904, -57.193,
 -80.260, -81.298, -81.406, -80.591, -79.045, -83.333, -74.902, -68.826,
-63.928, -57.213, -80.288, -81.326, -81.435, -80.620, -79.074, -83.365, -74.930,
-68.852, -63.951, -57.232, -80.316, -81.355, -81.464, -80.649, -79.102, -83.396,
-74.958, -68.877, -63.974, -57.251, -80.344, -81.384, -81.493, -80.678, -79.131,
-83.428, -74.986, -68.902, -63.998, -57.271, -80.373, -81.413, -81.522, -80.707,
-79.160, -83.459, -75.014, -68.928, -64.021, -57.290, -80.401, -81.441, -81.551,
-80.735, -79.188, -83.491, -75.042, -68.953, -64.044, -57.310, -80.429, -81.470,
-81.580, -80.764, -79.217, -83.522, -75.070, -68.979, -64.068, -57.329, -80.457,
-81.499, -81.609, -80.793, -79.246, -83.554, -75.098, -69.004, -64.091, -57.348,
 -80.486, -81.528, -81.638, -80.822, -79.274, -83.585, -75.125, -69.030,
-64.114, -57.368, -80.514, -81.557, -81.667, -80.851, -79.303, -83.617, -75.153,
-69.055, -64.138, -57.387, -80.542, -81.585, -81.696, -80.880, -79.332, -83.648,
-75.181, -69.080, -64.161, -57.406, -80.570, -81.614, -81.725, -80.909, -79.360,
-83.680, -75.209, -69.106, -64.184, -57.426,
};

const double CIA_Orton::N2H2[][10] = {
 -26.602, -27.013, -27.381, -27.736, -28.088, -28.438, -28.768, -29.127,
-29.485, -29.842, -24.770, -25.180, -25.549, -25.903, -26.255, -26.606, -26.935,
-27.294, -27.653, -28.009, -23.383, -23.794, -24.162, -24.517, -24.869, -25.219,
-25.549, -25.908, -26.266, -26.623, -21.995, -22.406, -22.775, -23.130, -23.482,
-23.832, -24.162, -24.521, -24.880, -25.236, -21.181, -21.592, -21.962, -22.317,
-22.669, -23.020, -23.350, -23.710, -24.069, -24.425, -20.603, -21.015, -21.384,
-21.740, -22.093, -22.444, -22.774, -23.134, -23.493, -23.849, -20.154, -20.566,
-20.936, -21.292, -21.645, -21.997, -22.327, -22.687, -23.046, -23.403, -19.788,
-20.200, -20.570, -20.926, -21.279, -21.631, -21.962, -22.322, -22.681, -23.038,
 -19.477, -19.889, -20.259, -20.616, -20.969, -21.322, -21.653, -22.013,
-22.372, -22.729, -19.207, -19.619, -19.989, -20.346, -20.700, -21.053, -21.385,
-21.745, -22.105, -22.462, -18.966, -19.379, -19.750, -20.107, -20.462, -20.815,
-21.148, -21.508, -21.868, -22.225, -18.751, -19.164, -19.535, -19.893, -20.249,
-20.603, -20.936, -21.297, -21.657, -22.014, -18.555, -18.968, -19.340, -19.699,
-20.056, -20.410, -20.744, -21.105, -21.465, -21.823, -18.373, -18.788, -19.161,
-19.521, -19.878, -20.233, -20.568, -20.930, -21.290, -21.648, -18.204, -18.619,
-18.994, -19.355, -19.713, -20.069, -20.405, -20.768, -21.129, -21.487, -18.046,
-18.462, -18.838, -19.200, -19.560, -19.917, -20.254, -20.617, -20.979, -21.338,
 -17.898, -18.316, -18.692, -19.056, -19.417, -19.776, -20.113, -20.477,
-20.839, -21.199, -17.759, -18.178, -18.556, -18.921, -19.283, -19.643, -19.982,
-20.346, -20.709, -21.068, -17.627, -18.047, -18.427, -18.793, -19.157, -19.518,
-19.858, -20.223, -20.586, -20.946, -17.502, -17.924, -18.304, -18.673, -19.037,
-19.399, -19.741, -20.106, -20.470, -20.830, -17.383, -17.805, -18.188, -18.557,
-18.924, -19.287, -19.629, -19.996, -20.360, -20.721, -17.269, -17.693, -18.077,
-18.448, -18.815, -19.180, -19.524, -19.891, -20.256, -20.617, -16.764, -17.195,
-17.586, -17.965, -18.340, -18.710, -19.061, -19.432, -19.800, -20.164, -16.353,
-16.788, -17.183, -17.567, -17.948, -18.324, -18.681, -19.056, -19.427, -19.793,
 -16.017, -16.452, -16.849, -17.236, -17.620, -18.000, -18.362, -18.740,
-19.112, -19.480, -15.728, -16.162, -16.561, -16.950, -17.337, -17.720, -18.087,
-18.466, -18.840, -19.209, -15.482, -15.916, -16.314, -16.705, -17.093, -17.477,
-17.847, -18.227, -18.602, -18.971, -15.266, -15.699, -16.097, -16.488, -16.877,
-17.262, -17.634, -18.015, -18.391, -18.760, -15.081, -15.513, -15.910, -16.300,
-16.688, -17.072, -17.446, -17.826, -18.201, -18.570, -14.925, -15.354, -15.749,
-16.136, -16.522, -16.904, -17.278, -17.658, -18.031, -18.399, -14.795, -15.221,
-15.613, -15.996, -16.379, -16.757, -17.131, -17.507, -17.879, -18.245, -14.686,
-15.108, -15.495, -15.874, -16.252, -16.626, -16.997, -17.371, -17.739, -18.103,
 -14.584, -15.003, -15.386, -15.761, -16.134, -16.504, -16.874, -17.244,
-17.610, -17.972, -14.486, -14.902, -15.282, -15.652, -16.022, -16.388, -16.757,
-17.125, -17.489, -17.849, -14.384, -14.797, -15.174, -15.543, -15.910, -16.275,
-16.644, -17.010, -17.372, -17.731, -14.282, -14.692, -15.067, -15.434, -15.800,
-16.164, -16.534, -16.899, -17.261, -17.619, -14.178, -14.587, -14.961, -15.328,
-15.694, -16.058, -16.429, -16.794, -17.155, -17.513, -14.095, -14.499, -14.871,
-15.236, -15.600, -15.962, -16.334, -16.697, -17.057, -17.414, -13.901, -14.301,
-14.672, -15.037, -15.402, -15.766, -16.143, -16.506, -16.867, -17.224, -13.687,
-14.091, -14.466, -14.836, -15.207, -15.576, -15.959, -16.325, -16.688, -17.046,
 -13.492, -13.903, -14.285, -14.660, -15.035, -15.406, -15.794, -16.162,
-16.526, -16.885, -13.317, -13.736, -14.122, -14.501, -14.878, -15.252, -15.643,
-16.012, -16.377, -16.736, -13.153, -13.580, -13.972, -14.354, -14.733, -15.109,
-15.503, -15.873, -16.238, -16.598, -13.050, -13.477, -13.867, -14.245, -14.621,
-14.993, -15.386, -15.754, -16.116, -16.474, -12.989, -13.406, -13.787, -14.158,
-14.526, -14.892, -15.282, -15.646, -16.005, -16.361, -12.955, -13.360, -13.730,
-14.091, -14.450, -14.808, -15.192, -15.551, -15.906, -16.258, -12.937, -13.326,
-13.683, -14.032, -14.381, -14.730, -15.110, -15.462, -15.813, -16.161, -12.884,
-13.264, -13.614, -13.958, -14.301, -14.646, -15.022, -15.371, -15.719, -16.065,
 -12.788, -13.177, -13.533, -13.878, -14.220, -14.563, -14.937, -15.284,
-15.630, -15.974, -12.668, -13.069, -13.434, -13.784, -14.129, -14.473, -14.847,
-15.194, -15.540, -15.884, -12.556, -12.970, -13.343, -13.697, -14.044, -14.388,
-14.763, -15.109, -15.454, -15.797, -12.464, -12.885, -13.262, -13.618, -13.966,
-14.309, -14.683, -15.028, -15.373, -15.715, -12.416, -12.831, -13.202, -13.553,
-13.897, -14.237, -14.609, -14.953, -15.296, -15.637, -12.384, -12.787, -13.149,
-13.493, -13.832, -14.169, -14.538, -14.880, -15.222, -15.562, -12.359, -12.747,
-13.098, -13.436, -13.770, -14.103, -14.470, -14.810, -15.151, -15.490, -12.326,
-12.702, -13.046, -13.379, -13.709, -14.040, -14.404, -14.743, -15.083, -15.421,
 -12.250, -12.630, -12.977, -13.312, -13.645, -13.976, -14.339, -14.678,
-15.016, -15.354, -12.170, -12.557, -12.910, -13.250, -13.583, -13.915, -14.277,
-14.615, -14.953, -15.290, -12.085, -12.481, -12.842, -13.187, -13.523, -13.855,
-14.217, -14.554, -14.892, -15.228, -12.018, -12.419, -12.785, -13.132, -13.468,
-13.800, -14.161, -14.497, -14.833, -15.169, -11.992, -12.388, -12.749, -13.091,
-13.424, -13.752, -14.109, -14.443, -14.777, -15.112, -11.970, -12.359, -12.714,
-13.051, -13.380, -13.705, -14.059, -14.391, -14.723, -15.056, -11.971, -12.345,
-12.689, -13.018, -13.340, -13.661, -14.011, -14.341, -14.672, -15.003, -11.965,
-12.326, -12.660, -12.982, -13.300, -13.617, -13.964, -14.292, -14.621, -14.951,
 -11.916, -12.275, -12.610, -12.933, -13.252, -13.569, -13.916, -14.242,
-14.571, -14.901, -11.871, -12.228, -12.563, -12.888, -13.207, -13.524, -13.869,
-14.195, -14.523, -14.852, -11.816, -12.176, -12.515, -12.843, -13.164, -13.481,
-13.825, -14.149, -14.476, -14.804, -11.763, -12.129, -12.472, -12.801, -13.123,
-13.440, -13.782, -14.106, -14.431, -14.758, -11.750, -12.111, -12.450, -12.774,
-13.092, -13.405, -13.744, -14.065, -14.389, -14.714, -11.740, -12.097, -12.430,
-12.749, -13.062, -13.372, -13.707, -14.026, -14.348, -14.671, -11.743, -12.091,
-12.415, -12.727, -13.034, -13.339, -13.671, -13.987, -14.307, -14.630, -11.749,
-12.084, -12.399, -12.704, -13.006, -13.307, -13.635, -13.950, -14.268, -14.589,
 -11.733, -12.058, -12.367, -12.670, -12.971, -13.272, -13.598, -13.911,
-14.229, -14.549, -11.716, -12.032, -12.337, -12.638, -12.938, -13.238, -13.563,
-13.875, -14.191, -14.510, -11.690, -11.999, -12.302, -12.603, -12.903, -13.203,
-13.527, -13.839, -14.154, -14.472, -11.664, -11.970, -12.272, -12.573, -12.873,
-13.172, -13.495, -13.805, -14.119, -14.436, -11.665, -11.964, -12.260, -12.556,
-12.852, -13.147, -13.466, -13.774, -14.086, -14.401, -11.661, -11.955, -12.247,
-12.538, -12.830, -13.121, -13.438, -13.743, -14.053, -14.366, -11.677, -11.963,
-12.246, -12.529, -12.813, -13.100, -13.412, -13.714, -14.021, -14.333, -11.690,
-11.966, -12.240, -12.515, -12.794, -13.077, -13.385, -13.684, -13.990, -14.300,
 -11.680, -11.948, -12.217, -12.489, -12.766, -13.048, -13.355, -13.653,
-13.958, -14.267, -11.677, -11.936, -12.197, -12.466, -12.741, -13.021, -13.326,
-13.624, -13.927, -14.236, -11.660, -11.913, -12.172, -12.440, -12.714, -12.994,
-13.298, -13.595, -13.897, -14.204, -11.639, -11.890, -12.149, -12.416, -12.690,
-12.969, -13.272, -13.567, -13.868, -14.174, -11.642, -11.888, -12.142, -12.406,
-12.676, -12.950, -13.250, -13.542, -13.841, -14.146, -11.640, -11.884, -12.135,
-12.395, -12.661, -12.932, -13.228, -13.518, -13.815, -14.118, -11.651, -11.890,
-12.136, -12.389, -12.649, -12.915, -13.207, -13.494, -13.789, -14.090, -11.668,
-11.898, -12.135, -12.381, -12.636, -12.898, -13.186, -13.471, -13.764, -14.063,
 -11.679, -11.899, -12.127, -12.367, -12.617, -12.876, -13.163, -13.446,
-13.738, -14.036, -11.695, -11.903, -12.121, -12.354, -12.600, -12.856, -13.141,
-13.422, -13.713, -14.010, -11.704, -11.900, -12.110, -12.337, -12.580, -12.835,
-13.118, -13.398, -13.688, -13.984, -11.707, -11.895, -12.099, -12.323, -12.564,
-12.816, -13.098, -13.376, -13.664, -13.959, -11.720, -11.901, -12.099, -12.318,
-12.554, -12.803, -13.081, -13.357, -13.642, -13.935, -11.725, -11.903, -12.097,
-12.312, -12.544, -12.789, -13.064, -13.337, -13.620, -13.911, -11.742, -11.916,
-12.106, -12.314, -12.540, -12.779, -13.050, -13.319, -13.600, -13.889, -11.758,
-11.927, -12.110, -12.312, -12.532, -12.767, -13.033, -13.301, -13.579, -13.866,
 -11.769, -11.931, -12.106, -12.302, -12.517, -12.749, -13.014, -13.280,
-13.557, -13.843, -11.787, -11.940, -12.107, -12.296, -12.506, -12.734, -12.996,
-13.260, -13.536, -13.821, -11.795, -11.939, -12.099, -12.284, -12.491, -12.718,
-12.978, -13.240, -13.515, -13.799, -11.799, -11.937, -12.093, -12.274, -12.479,
-12.703, -12.961, -13.222, -13.495, -13.778, -11.810, -11.944, -12.095, -12.272,
-12.473, -12.693, -12.948, -13.206, -13.477, -13.758, -11.816, -11.947, -12.096,
-12.269, -12.467, -12.683, -12.935, -13.190, -13.459, -13.738, -11.828, -11.958,
-12.103, -12.272, -12.464, -12.676, -12.924, -13.176, -13.442, -13.719, -11.845,
-11.971, -12.110, -12.273, -12.460, -12.667, -12.911, -13.160, -13.424, -13.700,
 -11.864, -11.984, -12.116, -12.272, -12.453, -12.656, -12.897, -13.144,
-13.406, -13.680, -11.888, -12.000, -12.124, -12.272, -12.447, -12.646, -12.883,
-13.128, -13.389, -13.661, -11.909, -12.013, -12.129, -12.270, -12.439, -12.634,
-12.869, -13.112, -13.371, -13.642, -11.925, -12.024, -12.133, -12.268, -12.433,
-12.625, -12.857, -13.098, -13.355, -13.625, -11.941, -12.036, -12.140, -12.271,
-12.432, -12.619, -12.848, -13.085, -13.340, -13.608, -11.956, -12.047, -12.147,
-12.273, -12.429, -12.613, -12.838, -13.073, -13.325, -13.591, -11.972, -12.061,
-12.158, -12.280, -12.432, -12.610, -12.831, -13.062, -13.312, -13.575, -11.988,
-12.075, -12.168, -12.285, -12.431, -12.605, -12.822, -13.050, -13.297, -13.559,
 -12.005, -12.088, -12.175, -12.286, -12.427, -12.596, -12.810, -13.036,
-13.282, -13.542, -12.023, -12.103, -12.185, -12.290, -12.425, -12.590, -12.800,
-13.023, -13.267, -13.526, -12.039, -12.114, -12.189, -12.289, -12.419, -12.581,
-12.789, -13.010, -13.252, -13.510, -12.055, -12.125, -12.195, -12.290, -12.416,
-12.574, -12.779, -12.998, -13.238, -13.494, -12.072, -12.138, -12.204, -12.295,
-12.417, -12.571, -12.773, -12.988, -13.226, -13.480, -12.087, -12.150, -12.212,
-12.299, -12.417, -12.567, -12.766, -12.978, -13.214, -13.466, -12.103, -12.165,
-12.224, -12.307, -12.421, -12.566, -12.760, -12.970, -13.202, -13.452, -12.121,
-12.180, -12.236, -12.314, -12.423, -12.564, -12.754, -12.960, -13.190, -13.438,
 -12.141, -12.198, -12.249, -12.321, -12.425, -12.560, -12.747, -12.950,
-13.178, -13.424, -12.160, -12.216, -12.263, -12.329, -12.427, -12.558, -12.740,
-12.940, -13.166, -13.410, -12.181, -12.233, -12.275, -12.335, -12.427, -12.553,
-12.732, -12.930, -13.153, -13.396, -12.200, -12.249, -12.286, -12.342, -12.429,
-12.551, -12.726, -12.921, -13.142, -13.384, -12.217, -12.264, -12.298, -12.350,
-12.433, -12.550, -12.722, -12.914, -13.132, -13.371, -12.234, -12.278, -12.309,
-12.357, -12.436, -12.550, -12.718, -12.906, -13.122, -13.360, -12.252, -12.295,
-12.323, -12.368, -12.443, -12.552, -12.716, -12.901, -13.114, -13.348, -12.271,
-12.312, -12.336, -12.377, -12.447, -12.552, -12.712, -12.893, -13.104, -13.336,
 -12.292, -12.330, -12.350, -12.386, -12.451, -12.551, -12.707, -12.885,
-13.093, -13.324, -12.312, -12.348, -12.365, -12.396, -12.456, -12.551, -12.703,
-12.878, -13.083, -13.312, -12.332, -12.365, -12.378, -12.403, -12.458, -12.549,
-12.697, -12.869, -13.073, -13.300, -12.350, -12.381, -12.390, -12.411, -12.462,
-12.548, -12.693, -12.862, -13.064, -13.289, -12.365, -12.395, -12.402, -12.420,
-12.467, -12.550, -12.691, -12.857, -13.056, -13.279, -12.377, -12.407, -12.413,
-12.428, -12.472, -12.551, -12.688, -12.852, -13.047, -13.269, -12.391, -12.421,
-12.425, -12.438, -12.479, -12.554, -12.687, -12.847, -13.040, -13.259, -12.407,
-12.436, -12.439, -12.448, -12.485, -12.556, -12.685, -12.842, -13.032, -13.249,
 -12.426, -12.454, -12.454, -12.460, -12.492, -12.558, -12.683, -12.836,
-13.024, -13.239, -12.446, -12.472, -12.470, -12.472, -12.500, -12.561, -12.682,
-12.831, -13.016, -13.229, -12.467, -12.491, -12.485, -12.483, -12.506, -12.562,
-12.679, -12.826, -13.008, -13.219, -12.485, -12.508, -12.500, -12.495, -12.513,
-12.565, -12.678, -12.821, -13.001, -13.210, -12.501, -12.523, -12.513, -12.505,
-12.520, -12.569, -12.678, -12.818, -12.995, -13.201, -12.519, -12.540, -12.528,
-12.516, -12.528, -12.573, -12.678, -12.815, -12.989, -13.193, -12.534, -12.555,
-12.541, -12.528, -12.537, -12.578, -12.680, -12.813, -12.984, -13.185, -12.550,
-12.570, -12.555, -12.540, -12.545, -12.582, -12.679, -12.809, -12.977, -13.176,
 -12.567, -12.587, -12.570, -12.551, -12.553, -12.585, -12.679, -12.805,
-12.970, -13.167, -12.582, -12.602, -12.584, -12.563, -12.561, -12.589, -12.679,
-12.802, -12.964, -13.159, -12.598, -12.618, -12.598, -12.574, -12.568, -12.592,
-12.678, -12.797, -12.958, -13.150, -12.616, -12.634, -12.613, -12.586, -12.576,
-12.596, -12.678, -12.795, -12.952, -13.143, -12.632, -12.650, -12.627, -12.597,
-12.585, -12.602, -12.680, -12.793, -12.947, -13.136, -12.648, -12.665, -12.640,
-12.609, -12.593, -12.606, -12.681, -12.791, -12.943, -13.129, -12.663, -12.680,
-12.655, -12.621, -12.603, -12.613, -12.683, -12.790, -12.939, -13.122, -12.679,
-12.695, -12.669, -12.634, -12.613, -12.618, -12.685, -12.788, -12.934, -13.115,
 -12.695, -12.711, -12.684, -12.647, -12.623, -12.624, -12.687, -12.786,
-12.929, -13.108, -12.711, -12.726, -12.699, -12.659, -12.632, -12.630, -12.689,
-12.785, -12.925, -13.101, -12.727, -12.742, -12.714, -12.672, -12.642, -12.636,
-12.690, -12.783, -12.920, -13.094, -12.744, -12.758, -12.729, -12.685, -12.652,
-12.642, -12.693, -12.782, -12.916, -13.088, -12.760, -12.773, -12.742, -12.697,
-12.661, -12.648, -12.695, -12.782, -12.913, -13.082, -12.778, -12.789, -12.757,
-12.709, -12.671, -12.655, -12.699, -12.781, -12.910, -13.077, -12.794, -12.805,
-12.771, -12.722, -12.681, -12.662, -12.702, -12.782, -12.908, -13.072, -12.811,
-12.821, -12.786, -12.735, -12.692, -12.669, -12.705, -12.782, -12.904, -13.066,
 -12.828, -12.838, -12.802, -12.749, -12.703, -12.677, -12.709, -12.781,
-12.901, -13.060, -12.844, -12.854, -12.817, -12.762, -12.714, -12.684, -12.712,
-12.781, -12.897, -13.054, -12.860, -12.869, -12.832, -12.775, -12.724, -12.690,
-12.715, -12.780, -12.894, -13.048, -12.874, -12.883, -12.846, -12.788, -12.734,
-12.697, -12.718, -12.781, -12.892, -13.044, -12.887, -12.896, -12.858, -12.799,
-12.744, -12.705, -12.722, -12.781, -12.889, -13.039, -12.899, -12.908, -12.870,
-12.810, -12.753, -12.711, -12.725, -12.782, -12.887, -13.034, -12.912, -12.920,
-12.882, -12.822, -12.763, -12.719, -12.730, -12.783, -12.886, -13.030, -12.925,
-12.934, -12.896, -12.834, -12.774, -12.727, -12.734, -12.784, -12.883, -13.026,
 -12.940, -12.949, -12.910, -12.847, -12.785, -12.735, -12.739, -12.785,
-12.882, -13.021, -12.956, -12.965, -12.924, -12.861, -12.797, -12.743, -12.743,
-12.787, -12.880, -13.017, -12.972, -12.980, -12.939, -12.874, -12.808, -12.751,
-12.748, -12.788, -12.878, -13.012, -12.987, -12.995, -12.953, -12.887, -12.819,
-12.760, -12.753, -12.790, -12.877, -13.009, -13.001, -13.008, -12.967, -12.899,
-12.830, -12.768, -12.757, -12.791, -12.876, -13.005, -13.016, -13.023, -12.980,
-12.912, -12.840, -12.776, -12.763, -12.794, -12.875, -13.002, -13.029, -13.036,
-12.993, -12.924, -12.851, -12.785, -12.768, -12.796, -12.875, -12.999, -13.043,
-13.049, -13.006, -12.936, -12.862, -12.793, -12.773, -12.798, -12.874, -12.995,
 -13.057, -13.063, -13.020, -12.949, -12.873, -12.802, -12.778, -12.800,
-12.873, -12.992, -13.069, -13.075, -13.032, -12.961, -12.884, -12.810, -12.783,
-12.802, -12.872, -12.989, -13.083, -13.089, -13.045, -12.974, -12.895, -12.818,
-12.788, -12.804, -12.871, -12.985, -13.097, -13.103, -13.058, -12.986, -12.906,
-12.827, -12.794, -12.807, -12.871, -12.982, -13.111, -13.116, -13.072, -12.998,
-12.917, -12.835, -12.800, -12.809, -12.871, -12.980, -13.124, -13.129, -13.084,
-13.010, -12.927, -12.844, -12.805, -12.812, -12.871, -12.977, -13.137, -13.142,
-13.097, -13.022, -12.938, -12.853, -12.811, -12.815, -12.871, -12.975, -13.150,
-13.155, -13.110, -13.034, -12.949, -12.862, -12.817, -12.818, -12.871, -12.973,
 -13.164, -13.169, -13.123, -13.047, -12.960, -12.871, -12.823, -12.822,
-12.872, -12.970, -13.177, -13.182, -13.136, -13.059, -12.971, -12.880, -12.829,
-12.825, -12.872, -12.968, -13.190, -13.195, -13.149, -13.071, -12.982, -12.889,
-12.835, -12.828, -12.872, -12.966, -13.205, -13.209, -13.162, -13.083, -12.993,
-12.898, -12.842, -12.831, -12.873, -12.964, -13.219, -13.222, -13.174, -13.095,
-13.003, -12.907, -12.848, -12.835, -12.874, -12.962, -13.233, -13.236, -13.187,
-13.107, -13.014, -12.916, -12.854, -12.839, -12.875, -12.961, -13.247, -13.250,
-13.200, -13.119, -13.025, -12.925, -12.861, -12.843, -12.876, -12.960, -13.260,
-13.263, -13.214, -13.132, -13.036, -12.934, -12.867, -12.846, -12.877, -12.958,
 -13.274, -13.277, -13.227, -13.145, -13.048, -12.944, -12.874, -12.850,
-12.878, -12.956, -13.287, -13.290, -13.240, -13.157, -13.059, -12.953, -12.881,
-12.854, -12.879, -12.955, -13.300, -13.303, -13.253, -13.169, -13.070, -12.962,
-12.887, -12.858, -12.880, -12.953, -13.312, -13.315, -13.264, -13.180, -13.080,
-12.971, -12.894, -12.862, -12.882, -12.952, -13.324, -13.326, -13.275, -13.191,
-13.090, -12.980, -12.900, -12.866, -12.883, -12.951, -13.336, -13.338, -13.286,
-13.201, -13.100, -12.988, -12.906, -12.870, -12.885, -12.951, -13.348, -13.349,
-13.297, -13.212, -13.110, -12.997, -12.913, -12.874, -12.886, -12.950, -13.361,
-13.362, -13.309, -13.223, -13.120, -13.006, -12.919, -12.878, -12.888, -12.949,
 -13.374, -13.374, -13.322, -13.235, -13.130, -13.015, -12.926, -12.883,
-12.890, -12.948, -13.387, -13.387, -13.334, -13.247, -13.141, -13.024, -12.933,
-12.887, -12.891, -12.947, -13.399, -13.400, -13.347, -13.259, -13.152, -13.033,
-12.940, -12.891, -12.893, -12.947, -13.412, -13.413, -13.359, -13.270, -13.163,
-13.042, -12.947, -12.896, -12.895, -12.947, -13.425, -13.425, -13.371, -13.281,
-13.173, -13.051, -12.954, -12.901, -12.898, -12.946, -13.437, -13.438, -13.383,
-13.292, -13.183, -13.060, -12.960, -12.905, -12.900, -12.946, -13.450, -13.449,
-13.394, -13.303, -13.193, -13.068, -12.967, -12.910, -12.902, -12.946, -13.462,
-13.461, -13.406, -13.314, -13.203, -13.077, -12.974, -12.915, -12.904, -12.946,
 -13.474, -13.473, -13.417, -13.325, -13.213, -13.086, -12.981, -12.919,
-12.907, -12.946, -13.486, -13.485, -13.428, -13.336, -13.222, -13.094, -12.988,
-12.924, -12.909, -12.946, -13.498, -13.497, -13.440, -13.346, -13.232, -13.103,
-12.994, -12.928, -12.911, -12.946, -13.510, -13.509, -13.451, -13.357, -13.242,
-13.111, -13.001, -12.933, -12.914, -12.946, -13.523, -13.521, -13.463, -13.368,
-13.252, -13.120, -13.008, -12.938, -12.916, -12.946, -13.535, -13.533, -13.474,
-13.378, -13.261, -13.128, -13.014, -12.943, -12.919, -12.947, -13.547, -13.544,
-13.485, -13.389, -13.271, -13.137, -13.021, -12.948, -12.922, -12.947, -13.559,
-13.556, -13.496, -13.399, -13.281, -13.145, -13.028, -12.952, -12.924, -12.947,
 -13.571, -13.568, -13.507, -13.410, -13.290, -13.153, -13.035, -12.957,
-12.927, -12.948, -13.583, -13.580, -13.519, -13.420, -13.299, -13.162, -13.041,
-12.962, -12.930, -12.948, -13.595, -13.591, -13.530, -13.430, -13.309, -13.170,
-13.048, -12.967, -12.932, -12.949, -13.607, -13.603, -13.541, -13.441, -13.318,
-13.178, -13.055, -12.972, -12.935, -12.950, -13.619, -13.615, -13.552, -13.451,
-13.327, -13.186, -13.061, -12.976, -12.938, -12.950, -13.631, -13.627, -13.564,
-13.461, -13.336, -13.194, -13.067, -12.981, -12.941, -12.951, -13.643, -13.639,
-13.575, -13.472, -13.345, -13.202, -13.074, -12.986, -12.944, -12.952, -13.655,
-13.651, -13.586, -13.482, -13.355, -13.210, -13.081, -12.991, -12.947, -12.953,
 -13.667, -13.662, -13.597, -13.492, -13.364, -13.218, -13.087, -12.996,
-12.950, -12.954, -13.678, -13.674, -13.608, -13.502, -13.373, -13.226, -13.094,
-13.001, -12.953, -12.955, -13.690, -13.685, -13.618, -13.512, -13.382, -13.234,
-13.100, -13.006, -12.956, -12.955, -13.702, -13.696, -13.628, -13.521, -13.390,
-13.241, -13.106, -13.010, -12.959, -12.957, -13.713, -13.707, -13.638, -13.530,
-13.397, -13.248, -13.112, -13.015, -12.962, -12.957, -13.724, -13.718, -13.648,
-13.538, -13.405, -13.254, -13.117, -13.019, -12.965, -12.959, -13.736, -13.728,
-13.658, -13.547, -13.412, -13.261, -13.123, -13.024, -12.967, -12.960, -13.747,
-13.739, -13.668, -13.556, -13.421, -13.268, -13.129, -13.028, -12.970, -12.961,
 -13.758, -13.750, -13.678, -13.565, -13.428, -13.275, -13.135, -13.033,
-12.973, -12.962, -13.770, -13.761, -13.688, -13.574, -13.437, -13.282, -13.141,
-13.038, -12.976, -12.963, -13.781, -13.772, -13.698, -13.583, -13.444, -13.289,
-13.146, -13.042, -12.979, -12.964, -13.792, -13.783, -13.708, -13.591, -13.452,
-13.295, -13.152, -13.047, -12.982, -12.966, -13.804, -13.793, -13.717, -13.600,
-13.459, -13.301, -13.157, -13.051, -12.985, -12.967, -13.815, -13.804, -13.727,
-13.608, -13.466, -13.307, -13.162, -13.055, -12.988, -12.968, -13.826, -13.814,
-13.736, -13.616, -13.473, -13.313, -13.167, -13.059, -12.991, -12.970, -13.837,
-13.824, -13.745, -13.624, -13.480, -13.319, -13.172, -13.063, -12.994, -12.971,
 -13.847, -13.834, -13.753, -13.631, -13.486, -13.325, -13.177, -13.067,
-12.997, -12.972, -13.858, -13.844, -13.762, -13.638, -13.492, -13.330, -13.182,
-13.071, -13.000, -12.973, -13.869, -13.854, -13.771, -13.646, -13.499, -13.336,
-13.187, -13.075, -13.003, -12.975, -13.880, -13.864, -13.779, -13.653, -13.505,
-13.341, -13.191, -13.079, -13.005, -12.976, -13.890, -13.873, -13.787, -13.660,
-13.510, -13.346, -13.196, -13.083, -13.008, -12.977, -13.901, -13.883, -13.795,
-13.666, -13.516, -13.351, -13.200, -13.086, -13.011, -12.979, -13.911, -13.892,
-13.803, -13.673, -13.521, -13.355, -13.204, -13.090, -13.014, -12.980, -13.921,
-13.901, -13.811, -13.679, -13.526, -13.360, -13.208, -13.093, -13.016, -12.982,
 -13.932, -13.910, -13.818, -13.685, -13.531, -13.364, -13.212, -13.097,
-13.019, -12.983, -13.942, -13.919, -13.825, -13.690, -13.536, -13.368, -13.216,
-13.100, -13.021, -12.984, -13.952, -13.927, -13.832, -13.696, -13.540, -13.372,
-13.219, -13.103, -13.024, -12.986, -13.962, -13.936, -13.838, -13.701, -13.544,
-13.375, -13.222, -13.106, -13.026, -12.987, -13.971, -13.944, -13.845, -13.706,
-13.548, -13.379, -13.226, -13.109, -13.028, -12.988, -13.981, -13.952, -13.851,
-13.710, -13.552, -13.382, -13.229, -13.112, -13.031, -12.989, -13.990, -13.959,
-13.857, -13.715, -13.555, -13.385, -13.232, -13.114, -13.033, -12.991, -14.000,
-13.967, -13.862, -13.719, -13.558, -13.388, -13.234, -13.117, -13.035, -12.992,
 -14.009, -13.974, -13.867, -13.722, -13.561, -13.390, -13.237, -13.120,
-13.037, -12.993, -14.017, -13.980, -13.871, -13.725, -13.563, -13.393, -13.239,
-13.122, -13.039, -12.994, -14.026, -13.987, -13.876, -13.728, -13.565, -13.394,
-13.241, -13.124, -13.041, -12.996, -14.035, -13.993, -13.879, -13.730, -13.567,
-13.396, -13.243, -13.126, -13.043, -12.997, -14.043, -13.998, -13.882, -13.731,
-13.568, -13.397, -13.244, -13.128, -13.045, -12.998, -14.051, -14.004, -13.885,
-13.733, -13.568, -13.397, -13.245, -13.129, -13.046, -12.999, -14.058, -14.008,
-13.887, -13.733, -13.568, -13.398, -13.246, -13.130, -13.048, -13.000, -14.066,
-14.013, -13.889, -13.734, -13.568, -13.398, -13.247, -13.132, -13.049, -13.001,
 -14.073, -14.017, -13.891, -13.734, -13.568, -13.398, -13.248, -13.133,
-13.051, -13.002, -14.080, -14.020, -13.891, -13.733, -13.568, -13.398, -13.249,
-13.134, -13.052, -13.003, -14.086, -14.023, -13.892, -13.732, -13.566, -13.397,
-13.249, -13.135, -13.053, -13.004, -14.092, -14.025, -13.891, -13.731, -13.565,
-13.396, -13.249, -13.136, -13.054, -13.004, -14.098, -14.027, -13.890, -13.729,
-13.562, -13.395, -13.248, -13.136, -13.055, -13.005, -14.103, -14.028, -13.889,
-13.726, -13.560, -13.393, -13.248, -13.136, -13.056, -13.006, -14.108, -14.029,
-13.886, -13.722, -13.557, -13.391, -13.247, -13.136, -13.056, -13.007, -14.112,
-14.029, -13.883, -13.718, -13.553, -13.388, -13.245, -13.136, -13.057, -13.007,
 -14.115, -14.028, -13.879, -13.714, -13.549, -13.385, -13.244, -13.136,
-13.057, -13.008, -14.119, -14.026, -13.874, -13.708, -13.544, -13.382, -13.242,
-13.135, -13.058, -13.008, -14.121, -14.024, -13.869, -13.702, -13.539, -13.378,
-13.240, -13.135, -13.058, -13.009, -14.123, -14.021, -13.863, -13.695, -13.533,
-13.374, -13.238, -13.134, -13.058, -13.009, -14.124, -14.017, -13.856, -13.688,
-13.527, -13.369, -13.235, -13.133, -13.058, -13.009, -14.124, -14.012, -13.848,
-13.680, -13.520, -13.364, -13.232, -13.131, -13.058, -13.010, -14.124, -14.006,
-13.840, -13.671, -13.512, -13.358, -13.228, -13.130, -13.057, -13.010, -14.123,
-13.999, -13.830, -13.662, -13.504, -13.353, -13.225, -13.128, -13.057, -13.010,
 -14.121, -13.991, -13.820, -13.652, -13.496, -13.346, -13.221, -13.126,
-13.056, -13.010, -14.118, -13.982, -13.808, -13.640, -13.487, -13.340, -13.217,
-13.124, -13.055, -13.010, -14.114, -13.972, -13.796, -13.629, -13.477, -13.332,
-13.212, -13.121, -13.054, -13.010, -14.109, -13.960, -13.782, -13.616, -13.466,
-13.325, -13.207, -13.118, -13.053, -13.009, -14.103, -13.948, -13.768, -13.602,
-13.455, -13.316, -13.202, -13.115, -13.052, -13.009, -14.095, -13.935, -13.752,
-13.588, -13.443, -13.308, -13.196, -13.112, -13.050, -13.008, -14.087, -13.920,
-13.736, -13.573, -13.431, -13.298, -13.190, -13.108, -13.049, -13.008, -14.077,
-13.904, -13.718, -13.557, -13.418, -13.289, -13.184, -13.105, -13.047, -13.007,
 -14.065, -13.886, -13.699, -13.540, -13.405, -13.279, -13.177, -13.101,
-13.045, -13.007, -14.053, -13.867, -13.679, -13.523, -13.391, -13.269, -13.170,
-13.097, -13.043, -13.006, -14.039, -13.847, -13.659, -13.504, -13.376, -13.257,
-13.163, -13.092, -13.041, -13.005, -14.023, -13.826, -13.636, -13.485, -13.360,
-13.245, -13.155, -13.087, -13.038, -13.004, -14.006, -13.803, -13.614, -13.464,
-13.343, -13.233, -13.146, -13.082, -13.036, -13.003, -13.988, -13.779, -13.590,
-13.444, -13.326, -13.220, -13.138, -13.076, -13.033, -13.002, -13.968, -13.754,
-13.565, -13.422, -13.309, -13.207, -13.129, -13.071, -13.030, -13.000, -13.946,
-13.728, -13.539, -13.400, -13.291, -13.194, -13.120, -13.065, -13.027, -12.999,
 -13.922, -13.699, -13.512, -13.377, -13.273, -13.180, -13.110, -13.059,
-13.023, -12.997, -13.897, -13.670, -13.484, -13.353, -13.254, -13.166, -13.100,
-13.052, -13.020, -12.996, -13.870, -13.639, -13.455, -13.328, -13.234, -13.151,
-13.090, -13.046, -13.016, -12.994, -13.840, -13.606, -13.424, -13.301, -13.213,
-13.135, -13.079, -13.039, -13.012, -12.992, -13.810, -13.572, -13.393, -13.275,
-13.191, -13.118, -13.068, -13.031, -13.008, -12.990, -13.778, -13.538, -13.361,
-13.247, -13.169, -13.102, -13.056, -13.024, -13.004, -12.988, -13.743, -13.501,
-13.327, -13.219, -13.146, -13.084, -13.044, -13.016, -12.999, -12.986, -13.707,
-13.463, -13.293, -13.191, -13.124, -13.067, -13.032, -13.008, -12.995, -12.984,
 -13.668, -13.424, -13.258, -13.161, -13.100, -13.049, -13.019, -13.000,
-12.990, -12.982, -13.628, -13.383, -13.222, -13.131, -13.076, -13.031, -13.007,
-12.991, -12.985, -12.979, -13.586, -13.341, -13.185, -13.100, -13.051, -13.012,
-12.993, -12.982, -12.980, -12.976, -13.542, -13.298, -13.146, -13.067, -13.025,
-12.992, -12.979, -12.973, -12.974, -12.974, -13.498, -13.255, -13.108, -13.035,
-12.999, -12.972, -12.965, -12.964, -12.969, -12.971, -13.452, -13.210, -13.068,
-13.002, -12.973, -12.952, -12.951, -12.954, -12.963, -12.968, -13.405, -13.165,
-13.029, -12.969, -12.946, -12.931, -12.936, -12.944, -12.957, -12.965, -13.357,
-13.120, -12.989, -12.936, -12.919, -12.910, -12.921, -12.934, -12.951, -12.962,
 -13.305, -13.072, -12.948, -12.902, -12.891, -12.889, -12.906, -12.923,
-12.945, -12.958, -13.253, -13.024, -12.906, -12.867, -12.863, -12.867, -12.891,
-12.913, -12.938, -12.955, -13.199, -12.974, -12.863, -12.831, -12.835, -12.845,
-12.875, -12.902, -12.931, -12.952, -13.142, -12.922, -12.818, -12.793, -12.804,
-12.821, -12.858, -12.891, -12.925, -12.948, -13.086, -12.870, -12.773, -12.756,
-12.774, -12.798, -12.841, -12.879, -12.918, -12.944, -13.029, -12.818, -12.728,
-12.718, -12.743, -12.774, -12.824, -12.868, -12.910, -12.940, -12.970, -12.766,
-12.683, -12.681, -12.713, -12.750, -12.807, -12.856, -12.903, -12.937, -12.911,
-12.713, -12.638, -12.643, -12.682, -12.727, -12.790, -12.844, -12.896, -12.933,
 -12.849, -12.659, -12.592, -12.605, -12.651, -12.702, -12.772, -12.832,
-12.888, -12.929, -12.786, -12.604, -12.545, -12.566, -12.620, -12.678, -12.755,
-12.820, -12.881, -12.924, -12.722, -12.548, -12.498, -12.527, -12.588, -12.652,
-12.736, -12.807, -12.873, -12.920, -12.656, -12.489, -12.448, -12.486, -12.555,
-12.627, -12.718, -12.794, -12.864, -12.916, -12.593, -12.433, -12.400, -12.445,
-12.522, -12.601, -12.699, -12.781, -12.856, -12.911, -12.528, -12.376, -12.351,
-12.404, -12.488, -12.574, -12.680, -12.768, -12.848, -12.906, -12.464, -12.320,
-12.303, -12.364, -12.455, -12.548, -12.661, -12.754, -12.839, -12.902, -12.402,
-12.266, -12.256, -12.324, -12.422, -12.522, -12.642, -12.741, -12.831, -12.897,
 -12.335, -12.208, -12.207, -12.283, -12.389, -12.496, -12.623, -12.727,
-12.822, -12.892, -12.270, -12.151, -12.159, -12.243, -12.356, -12.469, -12.603,
-12.714, -12.814, -12.887, -12.201, -12.092, -12.109, -12.201, -12.321, -12.442,
-12.584, -12.700, -12.805, -12.882, -12.129, -12.030, -12.057, -12.158, -12.286,
-12.415, -12.564, -12.686, -12.796, -12.877, -12.060, -11.970, -12.005, -12.115,
-12.251, -12.387, -12.544, -12.672, -12.787, -12.872, -11.990, -11.909, -11.954,
-12.072, -12.216, -12.359, -12.523, -12.658, -12.777, -12.867, -11.922, -11.851,
-11.904, -12.030, -12.181, -12.332, -12.503, -12.643, -12.768, -12.862, -11.856,
-11.794, -11.855, -11.989, -12.147, -12.305, -12.483, -12.629, -12.759, -12.856,
 -11.787, -11.735, -11.805, -11.947, -12.113, -12.277, -12.463, -12.615,
-12.750, -12.851, -11.719, -11.677, -11.756, -11.905, -12.079, -12.250, -12.443,
-12.601, -12.740, -12.846, -11.648, -11.616, -11.705, -11.863, -12.044, -12.222,
-12.423, -12.586, -12.731, -12.840, -11.574, -11.554, -11.652, -11.819, -12.009,
-12.194, -12.403, -12.572, -12.722, -12.835, -11.503, -11.493, -11.601, -11.776,
-11.974, -12.167, -12.383, -12.558, -12.712, -12.829, -11.432, -11.431, -11.549,
-11.733, -11.938, -12.139, -12.362, -12.543, -12.703, -12.823, -11.365, -11.373,
-11.499, -11.691, -11.904, -12.111, -12.342, -12.529, -12.693, -12.818, -11.302,
-11.317, -11.451, -11.651, -11.871, -12.085, -12.322, -12.515, -12.684, -12.812,
 -11.234, -11.260, -11.402, -11.610, -11.837, -12.058, -12.303, -12.500,
-12.674, -12.807, -11.170, -11.204, -11.355, -11.570, -11.804, -12.032, -12.283,
-12.486, -12.665, -12.801, -11.101, -11.145, -11.305, -11.529, -11.771, -12.005,
-12.264, -12.472, -12.656, -12.795, -11.028, -11.084, -11.254, -11.487, -11.737,
-11.978, -12.244, -12.458, -12.646, -12.790, -10.958, -11.025, -11.205, -11.446,
-11.704, -11.952, -12.225, -12.444, -12.637, -12.784, -10.889, -10.967, -11.156,
-11.406, -11.671, -11.926, -12.206, -12.431, -12.627, -12.779, -10.825, -10.911,
-11.109, -11.367, -11.639, -11.901, -12.187, -12.417, -12.618, -12.773, -10.764,
-10.859, -11.064, -11.329, -11.608, -11.877, -12.168, -12.404, -12.609, -12.767,
 -10.703, -10.807, -11.020, -11.292, -11.578, -11.853, -12.150, -12.390,
-12.600, -12.762, -10.643, -10.755, -10.977, -11.256, -11.549, -11.829, -12.133,
-12.378, -12.591, -12.756, -10.580, -10.703, -10.933, -11.220, -11.519, -11.806,
-12.115, -12.365, -12.582, -12.751, -10.515, -10.649, -10.889, -11.184, -11.490,
-11.783, -12.098, -12.352, -12.574, -12.745, -10.453, -10.597, -10.846, -11.150,
-11.463, -11.761, -12.081, -12.339, -12.565, -12.740, -10.392, -10.547, -10.805,
-11.116, -11.435, -11.740, -12.064, -12.327, -12.556, -12.734, -10.336, -10.500,
-10.766, -11.084, -11.410, -11.720, -12.048, -12.315, -12.548, -12.729, -10.285,
-10.456, -10.729, -11.054, -11.385, -11.700, -12.032, -12.304, -12.540, -12.724,
 -10.235, -10.414, -10.695, -11.025, -11.362, -11.682, -12.017, -12.292,
-12.532, -12.719, -10.139, -10.336, -10.632, -10.975, -11.323, -11.651, -11.993,
-12.274, -12.519, -12.711, -10.012, -10.235, -10.551, -10.911, -11.273, -11.613,
-11.962, -12.252, -12.504, -12.702, -9.847, -10.102, -10.445, -10.828, -11.208,
-11.564, -11.923, -12.224, -12.485, -12.690, -9.672,  -9.961, -10.331, -10.737,
-11.136, -11.509, -11.880, -12.193, -12.464, -12.677, -9.531,  -9.845, -10.238,
-10.662, -11.077, -11.463, -11.842, -12.165, -12.445, -12.665, -9.431,  -9.767,
-10.175, -10.612, -11.038, -11.433, -11.817, -12.147, -12.432, -12.657, -9.369,
-9.720, -10.139, -10.584, -11.016, -11.416, -11.800, -12.134, -12.423, -12.651,
  -9.367,  -9.723, -10.146, -10.591, -11.022, -11.420, -11.800, -12.133,
-12.421, -12.649, -9.337,  -9.706, -10.135, -10.585, -11.018, -11.417, -11.793,
-12.127, -12.417, -12.646, -9.375,  -9.741, -10.166, -10.610, -11.038, -11.431,
-11.799, -12.130, -12.418, -12.646, -9.410,  -9.776, -10.196, -10.635, -11.057,
-11.445, -11.804, -12.132, -12.418, -12.645, -9.335,  -9.715, -10.147, -10.595,
-11.025, -11.420, -11.779, -12.113, -12.405, -12.637, -9.312,  -9.692, -10.125,
-10.575, -11.007, -11.405, -11.761, -12.098, -12.394, -12.629, -9.200,  -9.596,
-10.045, -10.509, -10.953, -11.362, -11.722, -12.070, -12.374, -12.617, -9.039,
-9.461,  -9.932, -10.416, -10.879, -11.304, -11.673, -12.032, -12.348, -12.600,
  -8.949,  -9.387,  -9.871, -10.367, -10.839, -11.273, -11.644, -12.010,
-12.332, -12.590, -8.917,  -9.368,  -9.861, -10.362, -10.837, -11.272, -11.637,
-12.005, -12.328, -12.587, -8.952,  -9.408,  -9.901, -10.398, -10.868, -11.297,
-11.650, -12.014, -12.333, -12.589, -9.074,  -9.521, -10.000, -10.481, -10.935,
-11.349, -11.683, -12.036, -12.347, -12.598, -9.243,  -9.669, -10.123, -10.579,
-11.011, -11.406, -11.719, -12.061, -12.363, -12.606, -9.339,  -9.748, -10.185,
-10.627, -11.047, -11.432, -11.734, -12.071, -12.369, -12.609, -9.303,  -9.719,
-10.162, -10.609, -11.032, -11.420, -11.725, -12.064, -12.363, -12.605, -9.135,
-9.580, -10.051, -10.521, -10.965, -11.369, -11.691, -12.039, -12.346, -12.594,
  -8.930,  -9.402,  -9.902, -10.400, -10.869, -11.295, -11.640, -12.002,
-12.320, -12.578, -8.789,  -9.274,  -9.790, -10.307, -10.793, -11.235, -11.600,
-11.972, -12.299, -12.564, -8.729,  -9.215,  -9.735, -10.258, -10.751, -11.200,
-11.578, -11.955, -12.287, -12.555, -8.728,  -9.209,  -9.726, -10.248, -10.741,
-11.190, -11.575, -11.952, -12.284, -12.553, -8.799,  -9.269,  -9.778, -10.290,
-10.774, -11.216, -11.599, -11.969, -12.295, -12.559, -8.913,  -9.368,  -9.860,
-10.358, -10.829, -11.258, -11.635, -11.995, -12.312, -12.569, -8.960,  -9.410,
-9.898, -10.390, -10.854, -11.278, -11.655, -12.010, -12.321, -12.574, -8.949,
-9.405,  -9.897, -10.391, -10.856, -11.279, -11.662, -12.015, -12.324, -12.575,
  -8.967,  -9.418,  -9.906, -10.398, -10.861, -11.283, -11.669, -12.020,
-12.327, -12.575, -8.882,  -9.344,  -9.844, -10.347, -10.821, -11.251, -11.650,
-12.006, -12.317, -12.568, -8.822,  -9.290,  -9.797, -10.308, -10.789, -11.226,
-11.635, -11.994, -12.308, -12.563, -8.809,  -9.274,  -9.781, -10.292, -10.775,
-11.214, -11.629, -11.989, -12.304, -12.559, -8.760,  -9.228,  -9.740, -10.258,
-10.747, -11.191, -11.614, -11.978, -12.296, -12.553, -8.806,  -9.264,  -9.768,
-10.279, -10.763, -11.203, -11.624, -11.985, -12.300, -12.555, -8.866,  -9.311,
-9.804, -10.307, -10.784, -11.218, -11.638, -11.994, -12.305, -12.557, -8.982,
-9.407,  -9.883, -10.370, -10.833, -11.256, -11.668, -12.015, -12.318, -12.564,
  -9.120,  -9.519,  -9.973, -10.441, -10.888, -11.297, -11.701, -12.038,
-12.332, -12.572, -9.230,  -9.606, -10.042, -10.496, -10.930, -11.328, -11.725,
-12.054, -12.343, -12.577, -9.317,  -9.676, -10.098, -10.539, -10.963, -11.353,
-11.744, -12.067, -12.350, -12.581, -9.352,  -9.706, -10.122, -10.559, -10.978,
-11.363, -11.753, -12.073, -12.353, -12.582, -9.361,  -9.715, -10.130, -10.566,
-10.984, -11.368, -11.756, -12.074, -12.353, -12.581, -9.372,  -9.726, -10.141,
-10.575, -10.991, -11.373, -11.760, -12.076, -12.354, -12.580, -9.388,  -9.741,
-10.154, -10.586, -11.000, -11.379, -11.765, -12.079, -12.354, -12.579, -9.409,
-9.759, -10.169, -10.598, -11.010, -11.386, -11.770, -12.081, -12.355, -12.579,
  -9.431,  -9.779, -10.186, -10.612, -11.020, -11.394, -11.775, -12.085,
-12.356, -12.579, -9.455,  -9.800, -10.203, -10.626, -11.032, -11.403, -11.781,
-12.088, -12.358, -12.578, -9.476,  -9.819, -10.221, -10.641, -11.044, -11.412,
-11.788, -12.091, -12.359, -12.578, -9.496,  -9.838, -10.238, -10.656, -11.056,
-11.422, -11.794, -12.095, -12.360, -12.578, -9.516,  -9.858, -10.256, -10.672,
-11.070, -11.432, -11.801, -12.099, -12.362, -12.578, -9.536,  -9.878, -10.275,
-10.689, -11.083, -11.443, -11.808, -12.104, -12.364, -12.578, -9.562,  -9.902,
-10.296, -10.707, -11.099, -11.455, -11.816, -12.109, -12.366, -12.578, -9.590,
-9.928, -10.319, -10.727, -11.115, -11.467, -11.825, -12.114, -12.368, -12.578,
  -9.620,  -9.955, -10.343, -10.747, -11.131, -11.480, -11.833, -12.119,
-12.371, -12.578, -9.653,  -9.984, -10.368, -10.768, -11.149, -11.493, -11.843,
-12.124, -12.373, -12.579, -9.680, -10.010, -10.391, -10.788, -11.165, -11.507,
-11.852, -12.130, -12.376, -12.579, -9.706, -10.034, -10.414, -10.808, -11.182,
-11.520, -11.861, -12.136, -12.379, -12.579, -9.733, -10.059, -10.437, -10.829,
-11.200, -11.534, -11.870, -12.141, -12.381, -12.580, -9.757, -10.083, -10.459,
-10.849, -11.217, -11.548, -11.880, -12.148, -12.384, -12.580, -9.786, -10.111,
-10.485, -10.871, -11.236, -11.563, -11.890, -12.154, -12.387, -12.581, -9.818,
-10.141, -10.512, -10.895, -11.256, -11.579, -11.901, -12.161, -12.390, -12.582,
  -9.850, -10.171, -10.538, -10.918, -11.275, -11.594, -11.912, -12.167,
-12.394, -12.582, -9.886, -10.203, -10.566, -10.942, -11.295, -11.610, -11.923,
-12.174, -12.397, -12.583, -9.917, -10.231, -10.592, -10.964, -11.314, -11.626,
-11.933, -12.181, -12.400, -12.583, -9.948, -10.259, -10.617, -10.987, -11.334,
-11.641, -11.944, -12.188, -12.404, -12.584, -9.978, -10.287, -10.642, -11.009,
-11.353, -11.657, -11.955, -12.195, -12.407, -12.585, -10.006, -10.313, -10.667,
-11.031, -11.372, -11.673, -11.966, -12.202, -12.410, -12.585, -10.038, -10.343,
-10.694, -11.055, -11.393, -11.689, -11.978, -12.209, -12.414, -12.586, -10.070,
-10.373, -10.721, -11.080, -11.414, -11.706, -11.990, -12.217, -12.418, -12.587,
 -10.104, -10.405, -10.750, -11.104, -11.435, -11.723, -12.002, -12.224,
-12.421, -12.587, -10.140, -10.438, -10.779, -11.130, -11.456, -11.740, -12.014,
-12.232, -12.425, -12.588, -10.172, -10.467, -10.805, -11.153, -11.476, -11.757,
-12.025, -12.239, -12.428, -12.589, -10.203, -10.496, -10.831, -11.176, -11.496,
-11.773, -12.037, -12.246, -12.432, -12.589, -10.234, -10.524, -10.857, -11.199,
-11.516, -11.790, -12.048, -12.254, -12.435, -12.590, -10.262, -10.550, -10.882,
-11.222, -11.536, -11.806, -12.060, -12.261, -12.439, -12.590, -10.293, -10.580,
-10.909, -11.246, -11.557, -11.823, -12.072, -12.269, -12.442, -12.591, -10.325,
-10.610, -10.936, -11.271, -11.578, -11.840, -12.084, -12.276, -12.446, -12.591,
 -10.357, -10.640, -10.964, -11.295, -11.600, -11.858, -12.096, -12.284,
-12.449, -12.592, -10.392, -10.672, -10.993, -11.321, -11.621, -11.875, -12.108,
-12.291, -12.452, -12.592, -10.425, -10.703, -11.020, -11.345, -11.642, -11.892,
-12.120, -12.298, -12.456, -12.592, -10.459, -10.733, -11.047, -11.369, -11.662,
-11.908, -12.132, -12.305, -12.459, -12.592, -10.493, -10.763, -11.074, -11.392,
-11.682, -11.925, -12.143, -12.313, -12.462, -12.592, -10.523, -10.791, -11.099,
-11.415, -11.702, -11.941, -12.154, -12.320, -12.465, -12.592, -10.556, -10.822,
-11.126, -11.439, -11.723, -11.958, -12.166, -12.327, -12.468, -12.592, -10.588,
-10.851, -11.154, -11.463, -11.744, -11.975, -12.178, -12.334, -12.471, -12.592,
 -10.620, -10.881, -11.182, -11.488, -11.765, -11.992, -12.190, -12.341,
-12.473, -12.592, -10.654, -10.913, -11.210, -11.514, -11.787, -12.009, -12.201,
-12.347, -12.476, -12.592, -10.685, -10.942, -11.237, -11.537, -11.807, -12.025,
-12.212, -12.354, -12.478, -12.591, -10.717, -10.972, -11.263, -11.560, -11.827,
-12.041, -12.223, -12.360, -12.481, -12.591, -10.749, -11.001, -11.289, -11.583,
-11.846, -12.057, -12.234, -12.366, -12.483, -12.590, -10.778, -11.028, -11.314,
-11.605, -11.866, -12.072, -12.244, -12.372, -12.485, -12.589, -10.809, -11.056,
-11.340, -11.629, -11.886, -12.088, -12.255, -12.378, -12.487, -12.588, -10.840,
-11.085, -11.366, -11.653, -11.906, -12.105, -12.266, -12.384, -12.489, -12.587,
 -10.870, -11.114, -11.393, -11.676, -11.927, -12.120, -12.277, -12.390,
-12.490, -12.586, -10.902, -11.144, -11.420, -11.701, -11.947, -12.137, -12.287,
-12.396, -12.492, -12.585, -10.934, -11.173, -11.447, -11.724, -11.967, -12.152,
-12.297, -12.401, -12.493, -12.584, -10.966, -11.203, -11.474, -11.747, -11.986,
-12.167, -12.307, -12.406, -12.494, -12.582, -10.998, -11.233, -11.500, -11.770,
-12.006, -12.182, -12.316, -12.411, -12.495, -12.580, -11.028, -11.261, -11.525,
-11.793, -12.024, -12.196, -12.325, -12.415, -12.496, -12.578, -11.059, -11.290,
-11.552, -11.816, -12.044, -12.211, -12.335, -12.420, -12.496, -12.576, -11.089,
-11.318, -11.577, -11.839, -12.063, -12.226, -12.344, -12.424, -12.497, -12.574,
 -11.119, -11.346, -11.604, -11.862, -12.083, -12.240, -12.353, -12.428,
-12.497, -12.572, -11.149, -11.375, -11.630, -11.886, -12.102, -12.255, -12.362,
-12.432, -12.497, -12.570, -11.179, -11.403, -11.656, -11.908, -12.121, -12.269,
-12.370, -12.435, -12.497, -12.567, -11.209, -11.431, -11.681, -11.931, -12.139,
-12.282, -12.378, -12.438, -12.496, -12.564, -11.239, -11.458, -11.706, -11.952,
-12.157, -12.295, -12.385, -12.441, -12.495, -12.562, -11.267, -11.485, -11.731,
-11.974, -12.174, -12.308, -12.392, -12.444, -12.494, -12.558, -11.297, -11.513,
-11.756, -11.996, -12.193, -12.321, -12.400, -12.446, -12.493, -12.555, -11.325,
-11.540, -11.780, -12.018, -12.210, -12.333, -12.407, -12.448, -12.492, -12.552,
 -11.354, -11.567, -11.805, -12.040, -12.228, -12.346, -12.413, -12.450,
-12.490, -12.548, -11.383, -11.594, -11.830, -12.062, -12.246, -12.358, -12.420,
-12.452, -12.489, -12.545, -11.411, -11.621, -11.855, -12.083, -12.263, -12.370,
-12.426, -12.453, -12.487, -12.541, -11.440, -11.648, -11.881, -12.105, -12.280,
-12.382, -12.431, -12.454, -12.484, -12.536, -11.468, -11.675, -11.905, -12.126,
-12.297, -12.392, -12.436, -12.455, -12.482, -12.532, -11.496, -11.702, -11.929,
-12.147, -12.313, -12.403, -12.441, -12.455, -12.479, -12.528, -11.524, -11.728,
-11.954, -12.168, -12.330, -12.414, -12.446, -12.455, -12.476, -12.523, -11.552,
-11.754, -11.978, -12.189, -12.346, -12.424, -12.450, -12.455, -12.472, -12.518,
 -11.579, -11.780, -12.002, -12.210, -12.362, -12.434, -12.454, -12.454,
-12.469, -12.513, -11.606, -11.806, -12.026, -12.231, -12.378, -12.444, -12.458,
-12.454, -12.465, -12.508, -11.634, -11.832, -12.049, -12.251, -12.393, -12.453,
-12.461, -12.452, -12.461, -12.503, -11.660, -11.858, -12.073, -12.271, -12.407,
-12.461, -12.463, -12.450, -12.457, -12.497, -11.687, -11.883, -12.096, -12.290,
-12.421, -12.469, -12.465, -12.448, -12.452, -12.491, -11.714, -11.908, -12.118,
-12.309, -12.435, -12.476, -12.467, -12.446, -12.447, -12.485, -11.740, -11.933,
-12.141, -12.328, -12.449, -12.484, -12.469, -12.443, -12.442, -12.479, -11.766,
-11.957, -12.164, -12.347, -12.462, -12.491, -12.470, -12.440, -12.437, -12.473,
 -11.792, -11.982, -12.186, -12.366, -12.475, -12.497, -12.470, -12.437,
-12.431, -12.466, -11.818, -12.007, -12.209, -12.384, -12.488, -12.504, -12.471,
-12.433, -12.425, -12.460, -11.843, -12.031, -12.231, -12.402, -12.500, -12.509,
-12.470, -12.429, -12.419, -12.453, -11.869, -12.055, -12.253, -12.420, -12.512,
-12.514, -12.470, -12.425, -12.413, -12.446, -11.894, -12.079, -12.275, -12.438,
-12.523, -12.518, -12.468, -12.420, -12.406, -12.438, -11.919, -12.103, -12.296,
-12.455, -12.534, -12.522, -12.467, -12.414, -12.399, -12.431, -11.944, -12.127,
-12.317, -12.472, -12.545, -12.526, -12.465, -12.409, -12.392, -12.423, -11.968,
-12.150, -12.338, -12.488, -12.555, -12.529, -12.462, -12.403, -12.384, -12.415,
 -11.992, -12.173, -12.359, -12.505, -12.564, -12.531, -12.459, -12.397,
-12.376, -12.408, -12.017, -12.196, -12.379, -12.520, -12.573, -12.533, -12.456,
-12.391, -12.369, -12.399, -12.041, -12.218, -12.400, -12.535, -12.581, -12.534,
-12.452, -12.384, -12.360, -12.391, -12.064, -12.241, -12.420, -12.550, -12.589,
-12.535, -12.448, -12.376, -12.352, -12.383, -12.088, -12.263, -12.439, -12.564,
-12.596, -12.535, -12.442, -12.368, -12.343, -12.374, -12.111, -12.285, -12.458,
-12.578, -12.602, -12.534, -12.437, -12.360, -12.334, -12.365, -12.135, -12.307,
-12.477, -12.592, -12.608, -12.533, -12.431, -12.352, -12.324, -12.356, -12.158,
-12.329, -12.496, -12.604, -12.613, -12.531, -12.425, -12.343, -12.315, -12.347,
 -12.181, -12.350, -12.514, -12.617, -12.618, -12.529, -12.418, -12.334,
-12.305, -12.337, -12.203, -12.371, -12.532, -12.628, -12.622, -12.526, -12.411,
-12.325, -12.295, -12.328, -12.226, -12.392, -12.550, -12.640, -12.625, -12.523,
-12.403, -12.315, -12.285, -12.318, -12.248, -12.413, -12.567, -12.650, -12.627,
-12.518, -12.395, -12.305, -12.274, -12.308, -12.270, -12.433, -12.583, -12.660,
-12.629, -12.513, -12.386, -12.294, -12.264, -12.298, -12.292, -12.454, -12.600,
-12.669, -12.630, -12.508, -12.376, -12.283, -12.253, -12.288, -12.314, -12.473,
-12.616, -12.678, -12.630, -12.502, -12.367, -12.272, -12.241, -12.277, -12.335,
-12.493, -12.631, -12.686, -12.630, -12.495, -12.357, -12.261, -12.230, -12.267,
 -12.356, -12.512, -12.646, -12.693, -12.629, -12.488, -12.346, -12.249,
-12.218, -12.256, -12.378, -12.532, -12.660, -12.700, -12.627, -12.480, -12.335,
-12.237, -12.207, -12.245, -12.399, -12.550, -12.674, -12.705, -12.624, -12.471,
-12.324, -12.225, -12.195, -12.235, -12.419, -12.569, -12.687, -12.710, -12.621,
-12.462, -12.312, -12.212, -12.182, -12.223, -12.440, -12.587, -12.700, -12.714,
-12.616, -12.451, -12.299, -12.198, -12.170, -12.212, -12.460, -12.605, -12.712,
-12.717, -12.610, -12.440, -12.286, -12.185, -12.157, -12.201, -12.480, -12.622,
-12.724, -12.720, -12.604, -12.429, -12.273, -12.171, -12.144, -12.189, -12.500,
-12.639, -12.735, -12.721, -12.597, -12.417, -12.259, -12.157, -12.131, -12.177,
 -12.520, -12.656, -12.745, -12.722, -12.590, -12.405, -12.245, -12.143,
-12.118, -12.166, -12.539, -12.672, -12.754, -12.721, -12.581, -12.392, -12.230,
-12.129, -12.104, -12.154, -12.558, -12.688, -12.763, -12.720, -12.572, -12.378,
-12.216, -12.114, -12.091, -12.142, -12.578, -12.703, -12.770, -12.718, -12.561,
-12.363, -12.200, -12.099, -12.077, -12.130, -12.596, -12.719, -12.777, -12.714,
-12.550, -12.348, -12.184, -12.083, -12.063, -12.117, -12.615, -12.733, -12.784,
-12.710, -12.538, -12.332, -12.168, -12.068, -12.048, -12.105, -12.633, -12.747,
-12.789, -12.705, -12.525, -12.316, -12.151, -12.052, -12.034, -12.092, -12.651,
-12.761, -12.794, -12.699, -12.512, -12.299, -12.134, -12.036, -12.020, -12.080,
 -12.669, -12.774, -12.797, -12.692, -12.498, -12.283, -12.117, -12.020,
-12.005, -12.067, -12.687, -12.786, -12.800, -12.684, -12.482, -12.265, -12.100,
-12.003, -11.990, -12.055, -12.704, -12.798, -12.802, -12.675, -12.467, -12.247,
-12.082, -11.987, -11.976, -12.042, -12.721, -12.809, -12.802, -12.664, -12.449,
-12.228, -12.064, -11.970, -11.960, -12.029, -12.738, -12.820, -12.801, -12.652,
-12.431, -12.207, -12.045, -11.952, -11.945, -12.016, -12.754, -12.829, -12.800,
-12.639, -12.413, -12.187, -12.025, -11.935, -11.930, -12.003, -12.770, -12.839,
-12.797, -12.626, -12.393, -12.167, -12.006, -11.917, -11.914, -11.989, -12.786,
-12.847, -12.793, -12.611, -12.374, -12.146, -11.986, -11.899, -11.899, -11.976,
 -12.801, -12.854, -12.788, -12.596, -12.353, -12.125, -11.967, -11.882,
-11.883, -11.963, -12.816, -12.861, -12.782, -12.579, -12.332, -12.103, -11.947,
-11.864, -11.867, -11.950, -12.831, -12.867, -12.774, -12.561, -12.310, -12.081,
-11.927, -11.846, -11.852, -11.936, -12.845, -12.871, -12.765, -12.542, -12.287,
-12.058, -11.906, -11.827, -11.836, -11.923, -12.859, -12.875, -12.755, -12.522,
-12.263, -12.034, -11.885, -11.808, -11.819, -11.909, -12.872, -12.878, -12.744,
-12.501, -12.239, -12.010, -11.863, -11.789, -11.803, -11.895, -12.885, -12.880,
-12.731, -12.479, -12.214, -11.986, -11.842, -11.770, -11.787, -11.882, -12.897,
-12.881, -12.718, -12.457, -12.189, -11.962, -11.820, -11.752, -11.771, -11.868,
 -12.909, -12.881, -12.704, -12.434, -12.164, -11.938, -11.799, -11.733,
-11.755, -11.855, -12.920, -12.879, -12.688, -12.410, -12.138, -11.913, -11.777,
-11.714, -11.738, -11.841, -12.930, -12.876, -12.671, -12.385, -12.111, -11.888,
-11.755, -11.695, -11.722, -11.827, -12.940, -12.872, -12.651, -12.358, -12.083,
-11.862, -11.733, -11.675, -11.705, -11.814, -12.949, -12.866, -12.630, -12.330,
-12.054, -11.835, -11.710, -11.655, -11.689, -11.800, -12.957, -12.859, -12.609,
-12.302, -12.025, -11.809, -11.687, -11.636, -11.672, -11.786, -12.965, -12.851,
-12.586, -12.272, -11.996, -11.782, -11.664, -11.616, -11.656, -11.772, -12.971,
-12.842, -12.562, -12.243, -11.967, -11.755, -11.641, -11.596, -11.639, -11.759,
 -12.977, -12.831, -12.538, -12.214, -11.938, -11.728, -11.618, -11.577,
-11.623, -11.745, -12.982, -12.818, -12.511, -12.183, -11.908, -11.701, -11.595,
-11.557, -11.606, -11.731, -12.985, -12.804, -12.484, -12.152, -11.877, -11.674,
-11.572, -11.538, -11.590, -11.718, -12.988, -12.787, -12.454, -12.118, -11.846,
-11.646, -11.549, -11.518, -11.573, -11.704, -12.989, -12.770, -12.424, -12.084,
-11.814, -11.618, -11.525, -11.498, -11.556, -11.691, -12.989, -12.751, -12.393,
-12.050, -11.782, -11.589, -11.501, -11.478, -11.540, -11.677, -12.988, -12.731,
-12.360, -12.016, -11.749, -11.561, -11.478, -11.458, -11.523, -11.663, -12.987,
-12.710, -12.329, -11.982, -11.717, -11.532, -11.454, -11.438, -11.507, -11.650,
 -12.984, -12.688, -12.296, -11.948, -11.686, -11.504, -11.431, -11.419,
-11.491, -11.637, -12.979, -12.664, -12.262, -11.913, -11.653, -11.476, -11.408,
-11.399, -11.475, -11.623, -12.973, -12.638, -12.227, -11.877, -11.621, -11.448,
-11.384, -11.380, -11.458, -11.610, -12.964, -12.609, -12.189, -11.840, -11.587,
-11.419, -11.360, -11.360, -11.442, -11.597, -12.954, -12.578, -12.150, -11.801,
-11.553, -11.389, -11.337, -11.340, -11.426, -11.584, -12.943, -12.547, -12.112,
-11.764, -11.519, -11.360, -11.313, -11.321, -11.410, -11.571, -12.929, -12.514,
-12.072, -11.725, -11.485, -11.331, -11.289, -11.301, -11.394, -11.558, -12.916,
-12.482, -12.034, -11.688, -11.452, -11.303, -11.266, -11.282, -11.378, -11.545,
 -12.900, -12.449, -11.995, -11.651, -11.419, -11.274, -11.243, -11.263,
-11.363, -11.532, -12.882, -12.413, -11.955, -11.613, -11.385, -11.246, -11.220,
-11.244, -11.347, -11.520, -12.863, -12.377, -11.915, -11.575, -11.352, -11.218,
-11.197, -11.225, -11.332, -11.507, -12.840, -12.337, -11.872, -11.536, -11.318,
-11.189, -11.174, -11.207, -11.317, -11.495, -12.815, -12.296, -11.828, -11.496,
-11.284, -11.160, -11.151, -11.188, -11.302, -11.483, -12.789, -12.255, -11.785,
-11.457, -11.249, -11.132, -11.129, -11.170, -11.287, -11.470, -12.762, -12.213,
-11.741, -11.417, -11.216, -11.103, -11.106, -11.151, -11.272, -11.458, -12.736,
-12.173, -11.700, -11.380, -11.183, -11.076, -11.084, -11.133, -11.257, -11.447,
 -12.708, -12.132, -11.658, -11.342, -11.150, -11.048, -11.062, -11.116,
-11.243, -11.435, -12.678, -12.090, -11.617, -11.304, -11.118, -11.021, -11.041,
-11.098, -11.229, -11.423, -12.646, -12.048, -11.575, -11.267, -11.086, -10.995,
-11.020, -11.081, -11.215, -11.412, -12.610, -12.002, -11.531, -11.228, -11.053,
-10.968, -10.998, -11.064, -11.201, -11.401, -12.571, -11.955, -11.486, -11.189,
-11.021, -10.941, -10.977, -11.047, -11.187, -11.390, -12.533, -11.908, -11.442,
-11.152, -10.989, -10.915, -10.957, -11.030, -11.174, -11.379, -12.494, -11.862,
-11.399, -11.114, -10.957, -10.889, -10.936, -11.014, -11.160, -11.368, -12.457,
-11.819, -11.358, -11.078, -10.927, -10.864, -10.916, -10.998, -11.147, -11.357,
 -12.421, -11.776, -11.318, -11.043, -10.898, -10.840, -10.897, -10.983,
-11.135, -11.347, -12.383, -11.734, -11.278, -11.009, -10.869, -10.816, -10.878,
-10.967, -11.122, -11.337, -12.344, -11.691, -11.239, -10.975, -10.841, -10.794,
-10.860, -10.952, -11.110, -11.327, -12.301, -11.646, -11.199, -10.941, -10.813,
-10.771, -10.841, -10.938, -11.098, -11.317, -12.257, -11.601, -11.159, -10.908,
-10.786, -10.749, -10.824, -10.923, -11.086, -11.307, -12.213, -11.557, -11.121,
-10.876, -10.759, -10.727, -10.806, -10.909, -11.075, -11.298, -12.171, -11.514,
-11.084, -10.844, -10.733, -10.707, -10.789, -10.896, -11.064, -11.289, -12.133,
-11.475, -11.049, -10.815, -10.710, -10.687, -10.773, -10.882, -11.053, -11.280,
 -12.096, -11.439, -11.016, -10.788, -10.687, -10.669, -10.757, -10.870,
-11.042, -11.271, -12.061, -11.403, -10.986, -10.762, -10.666, -10.652, -10.742,
-10.857, -11.032, -11.263, -11.955, -11.308, -10.906, -10.698, -10.615, -10.612,
-10.709, -10.832, -11.012, -11.247, -11.855, -11.218, -10.832, -10.638, -10.567,
-10.574, -10.678, -10.807, -10.992, -11.232, -11.691, -11.075, -10.713, -10.543,
-10.492, -10.516, -10.630, -10.769, -10.963, -11.210, -11.546, -10.947, -10.607,
-10.456, -10.423, -10.462, -10.584, -10.734, -10.936, -11.188, -11.443, -10.860,
-10.535, -10.399, -10.377, -10.426, -10.553, -10.709, -10.916, -11.173, -11.358,
-10.788, -10.478, -10.353, -10.341, -10.397, -10.528, -10.689, -10.900, -11.160,
 -11.347, -10.783, -10.476, -10.353, -10.341, -10.398, -10.525, -10.685,
-10.897, -11.157, -11.329, -10.774, -10.472, -10.351, -10.341, -10.398, -10.521,
-10.682, -10.893, -11.153, -11.327, -10.778, -10.479, -10.358, -10.348, -10.403,
-10.520, -10.680, -10.891, -11.151, -11.370, -10.819, -10.515, -10.389, -10.372,
-10.421, -10.529, -10.686, -10.894, -11.152, -11.358, -10.812, -10.511, -10.386,
-10.369, -10.419, -10.521, -10.679, -10.888, -11.147, -11.312, -10.771, -10.475,
-10.356, -10.344, -10.398, -10.499, -10.660, -10.873, -11.135, -11.265, -10.727,
-10.436, -10.322, -10.316, -10.375, -10.473, -10.639, -10.856, -11.121, -11.129,
-10.607, -10.335, -10.238, -10.248, -10.321, -10.425, -10.601, -10.826, -11.098,
 -11.002, -10.497, -10.241, -10.161, -10.185, -10.271, -10.380, -10.565,
-10.797, -11.076, -10.943, -10.450, -10.205, -10.133, -10.164, -10.254, -10.362,
-10.550, -10.786, -11.066, -10.941, -10.458, -10.218, -10.147, -10.178, -10.267,
-10.366, -10.553, -10.787, -11.067, -11.009, -10.528, -10.283, -10.205, -10.226,
-10.306, -10.390, -10.571, -10.801, -11.077, -11.157, -10.664, -10.402, -10.303,
-10.305, -10.368, -10.432, -10.603, -10.824, -11.094, -11.304, -10.793, -10.508,
-10.387, -10.370, -10.418, -10.464, -10.627, -10.842, -11.106, -11.349, -10.831,
-10.537, -10.409, -10.386, -10.429, -10.471, -10.631, -10.844, -11.107, -11.259,
-10.756, -10.477, -10.363, -10.350, -10.402, -10.451, -10.616, -10.833, -11.098,
 -11.070, -10.591, -10.341, -10.254, -10.265, -10.336, -10.406, -10.581,
-10.806, -11.077, -10.887, -10.426, -10.200, -10.137, -10.170, -10.261, -10.354,
-10.540, -10.774, -11.053, -10.788, -10.331, -10.115, -10.064, -10.109, -10.211,
-10.320, -10.514, -10.753, -11.037, -10.764, -10.303, -10.086, -10.036, -10.085,
-10.190, -10.308, -10.504, -10.745, -11.030, -10.789, -10.322, -10.100, -10.046,
-10.092, -10.194, -10.317, -10.511, -10.750, -11.033, -10.888, -10.409, -10.175,
-10.109, -10.143, -10.235, -10.354, -10.540, -10.772, -11.049, -10.987, -10.496,
-10.249, -10.171, -10.193, -10.274, -10.390, -10.568, -10.793, -11.065, -10.992,
-10.506, -10.260, -10.182, -10.203, -10.282, -10.403, -10.579, -10.801, -11.071,
 -11.008, -10.521, -10.274, -10.194, -10.213, -10.291, -10.415, -10.589,
-10.809, -11.076, -10.975, -10.489, -10.246, -10.171, -10.195, -10.276, -10.409,
-10.584, -10.805, -11.073, -10.894, -10.418, -10.186, -10.121, -10.155, -10.245,
-10.390, -10.569, -10.794, -11.064, -10.866, -10.388, -10.159, -10.098, -10.135,
-10.229, -10.381, -10.563, -10.789, -11.059, -10.844, -10.365, -10.136, -10.077,
-10.118, -10.214, -10.373, -10.556, -10.784, -11.055, -10.829, -10.349, -10.121,
-10.064, -10.106, -10.204, -10.368, -10.552, -10.780, -11.052, -10.884, -10.393,
-10.156, -10.091, -10.127, -10.220, -10.383, -10.564, -10.789, -11.058, -10.964,
-10.458, -10.208, -10.133, -10.160, -10.246, -10.405, -10.581, -10.802, -11.067,
 -11.087, -10.562, -10.293, -10.202, -10.214, -10.288, -10.441, -10.608,
-10.822, -11.082, -11.227, -10.676, -10.386, -10.275, -10.272, -10.332, -10.477,
-10.635, -10.842, -11.096, -11.308, -10.743, -10.440, -10.318, -10.305, -10.358,
-10.498, -10.652, -10.854, -11.105, -11.388, -10.809, -10.493, -10.360, -10.338,
-10.383, -10.519, -10.667, -10.866, -11.113, -11.397, -10.818, -10.501, -10.367,
-10.344, -10.388, -10.524, -10.671, -10.868, -11.114, -11.407, -10.829, -10.512,
-10.376, -10.351, -10.393, -10.529, -10.675, -10.871, -11.116, -11.421, -10.843,
-10.524, -10.387, -10.360, -10.400, -10.535, -10.679, -10.874, -11.118, -11.441,
-10.860, -10.539, -10.400, -10.371, -10.409, -10.542, -10.684, -10.877, -11.120,
 -11.462, -10.880, -10.556, -10.414, -10.382, -10.418, -10.549, -10.689,
-10.881, -11.123, -11.486, -10.901, -10.574, -10.429, -10.395, -10.427, -10.556,
-10.695, -10.885, -11.125, -11.509, -10.922, -10.593, -10.445, -10.408, -10.438,
-10.564, -10.701, -10.890, -11.128, -11.530, -10.942, -10.611, -10.461, -10.421,
-10.448, -10.573, -10.707, -10.894, -11.131, -11.550, -10.962, -10.629, -10.477,
-10.435, -10.460, -10.581, -10.714, -10.899, -11.135, -11.570, -10.982, -10.648,
-10.494, -10.450, -10.472, -10.591, -10.721, -10.904, -11.138, -11.593, -11.005,
-10.669, -10.513, -10.465, -10.485, -10.600, -10.728, -10.909, -11.142, -11.621,
-11.031, -10.693, -10.533, -10.482, -10.498, -10.611, -10.736, -10.915, -11.146,
 -11.650, -11.058, -10.717, -10.554, -10.500, -10.513, -10.621, -10.744,
-10.921, -11.150, -11.681, -11.087, -10.742, -10.575, -10.518, -10.527, -10.633,
-10.752, -10.927, -11.154, -11.712, -11.115, -10.767, -10.597, -10.536, -10.543,
-10.644, -10.761, -10.934, -11.159, -11.738, -11.140, -10.790, -10.618, -10.554,
-10.557, -10.655, -10.770, -10.940, -11.164, -11.765, -11.166, -10.814, -10.639,
-10.572, -10.573, -10.667, -10.779, -10.947, -11.168, -11.790, -11.191, -10.838,
-10.660, -10.591, -10.588, -10.679, -10.788, -10.954, -11.173, -11.815, -11.216,
-10.862, -10.682, -10.610, -10.605, -10.692, -10.798, -10.962, -11.179, -11.846,
-11.246, -10.889, -10.706, -10.631, -10.622, -10.705, -10.808, -10.969, -11.184,
 -11.878, -11.276, -10.916, -10.730, -10.652, -10.639, -10.718, -10.818,
-10.977, -11.190, -11.911, -11.308, -10.945, -10.755, -10.673, -10.657, -10.732,
-10.829, -10.985, -11.196, -11.945, -11.339, -10.973, -10.780, -10.694, -10.674,
-10.745, -10.840, -10.993, -11.202, -11.975, -11.367, -10.999, -10.803, -10.714,
-10.692, -10.759, -10.850, -11.002, -11.208, -12.005, -11.395, -11.025, -10.826,
-10.735, -10.709, -10.773, -10.861, -11.010, -11.214, -12.034, -11.423, -11.050,
-10.849, -10.755, -10.727, -10.787, -10.873, -11.019, -11.220, -12.063, -11.450,
-11.076, -10.873, -10.777, -10.745, -10.801, -10.884, -11.028, -11.227, -12.095,
-11.481, -11.104, -10.899, -10.799, -10.764, -10.816, -10.896, -11.037, -11.233,
 -12.127, -11.512, -11.133, -10.924, -10.821, -10.783, -10.831, -10.908,
-11.046, -11.240, -12.162, -11.545, -11.163, -10.950, -10.844, -10.802, -10.846,
-10.920, -11.055, -11.247, -12.195, -11.577, -11.192, -10.976, -10.866, -10.822,
-10.862, -10.932, -11.065, -11.254, -12.225, -11.606, -11.218, -11.000, -10.887,
-10.840, -10.876, -10.944, -11.074, -11.261, -12.257, -11.635, -11.245, -11.024,
-10.909, -10.859, -10.892, -10.957, -11.084, -11.268, -12.285, -11.662, -11.270,
-11.048, -10.930, -10.877, -10.907, -10.969, -11.094, -11.276, -12.313, -11.690,
-11.296, -11.072, -10.952, -10.896, -10.922, -10.982, -11.104, -11.283, -12.345,
-11.720, -11.325, -11.098, -10.975, -10.916, -10.938, -10.995, -11.114, -11.291,
 -12.375, -11.750, -11.353, -11.123, -10.997, -10.936, -10.954, -11.008,
-11.124, -11.298, -12.408, -11.782, -11.382, -11.149, -11.021, -10.956, -10.970,
-11.021, -11.135, -11.306, -12.441, -11.814, -11.411, -11.175, -11.043, -10.975,
-10.986, -11.034, -11.145, -11.314, -12.473, -11.844, -11.439, -11.200, -11.065,
-10.995, -11.002, -11.047, -11.155, -11.322, -12.506, -11.875, -11.466, -11.225,
-11.087, -11.014, -11.018, -11.060, -11.166, -11.330, -12.537, -11.904, -11.493,
-11.249, -11.109, -11.033, -11.034, -11.074, -11.176, -11.338, -12.567, -11.932,
-11.519, -11.273, -11.131, -11.052, -11.050, -11.087, -11.187, -11.346, -12.599,
-11.963, -11.548, -11.299, -11.154, -11.072, -11.067, -11.101, -11.198, -11.355,
 -12.629, -11.993, -11.575, -11.324, -11.176, -11.092, -11.083, -11.115,
-11.209, -11.363, -12.662, -12.024, -11.605, -11.351, -11.200, -11.113, -11.100,
-11.128, -11.220, -11.371, -12.693, -12.055, -11.633, -11.376, -11.223, -11.133,
-11.117, -11.142, -11.231, -11.380, -12.723, -12.084, -11.660, -11.400, -11.244,
-11.152, -11.133, -11.156, -11.242, -11.388, -12.754, -12.114, -11.687, -11.425,
-11.266, -11.171, -11.149, -11.169, -11.253, -11.397, -12.783, -12.141, -11.713,
-11.448, -11.288, -11.190, -11.165, -11.183, -11.264, -11.405, -12.812, -12.169,
-11.739, -11.472, -11.309, -11.209, -11.181, -11.197, -11.275, -11.414, -12.842,
-12.198, -11.766, -11.497, -11.332, -11.229, -11.198, -11.211, -11.287, -11.422,
 -12.871, -12.226, -11.793, -11.522, -11.354, -11.249, -11.215, -11.225,
-11.298, -11.431, -12.901, -12.256, -11.821, -11.548, -11.377, -11.269, -11.232,
-11.239, -11.309, -11.440, -12.931, -12.286, -11.848, -11.573, -11.400, -11.289,
-11.248, -11.253, -11.321, -11.449, -12.961, -12.315, -11.875, -11.597, -11.422,
-11.308, -11.265, -11.267, -11.332, -11.457, -12.992, -12.345, -11.903, -11.622,
-11.444, -11.328, -11.281, -11.281, -11.343, -11.466, -13.021, -12.374, -11.929,
-11.646, -11.465, -11.347, -11.297, -11.295, -11.354, -11.475, -13.050, -12.402,
-11.956, -11.670, -11.487, -11.366, -11.314, -11.309, -11.366, -11.484, -13.079,
-12.430, -11.982, -11.695, -11.509, -11.386, -11.331, -11.323, -11.377, -11.492,
 -13.108, -12.458, -12.009, -11.719, -11.531, -11.406, -11.347, -11.337,
-11.389, -11.501, -13.136, -12.487, -12.036, -11.744, -11.554, -11.426, -11.364,
-11.351, -11.400, -11.510, -13.165, -12.515, -12.063, -11.769, -11.576, -11.445,
-11.381, -11.365, -11.412, -11.519, -13.193, -12.543, -12.089, -11.793, -11.597,
-11.464, -11.397, -11.379, -11.423, -11.528, -13.221, -12.571, -12.115, -11.817,
-11.619, -11.483, -11.413, -11.393, -11.434, -11.537, -13.248, -12.598, -12.141,
-11.840, -11.640, -11.502, -11.429, -11.407, -11.446, -11.545, -13.276, -12.625,
-12.166, -11.863, -11.661, -11.521, -11.446, -11.420, -11.457, -11.554, -13.303,
-12.652, -12.192, -11.887, -11.683, -11.540, -11.462, -11.435, -11.468, -11.563,
 -13.330, -12.679, -12.217, -11.911, -11.704, -11.560, -11.479, -11.449,
-11.480, -11.572, -13.357, -12.706, -12.243, -11.935, -11.727, -11.579, -11.496,
-11.463, -11.491, -11.581, -13.384, -12.733, -12.269, -11.959, -11.748, -11.598,
-11.512, -11.477, -11.503, -11.589, -13.411, -12.760, -12.295, -11.983, -11.770,
-11.617, -11.528, -11.490, -11.514, -11.598, -13.438, -12.787, -12.321, -12.007,
-11.791, -11.636, -11.544, -11.504, -11.525, -11.607, -13.464, -12.814, -12.346,
-12.030, -11.812, -11.655, -11.560, -11.517, -11.536, -11.615, -13.490, -12.840,
-12.371, -12.053, -11.833, -11.674, -11.576, -11.531, -11.547, -11.624, -13.516,
-12.867, -12.396, -12.077, -11.855, -11.693, -11.593, -11.545, -11.558, -11.632,
 -13.542, -12.892, -12.421, -12.100, -11.876, -11.712, -11.609, -11.559,
-11.569, -11.641, -13.568, -12.918, -12.446, -12.124, -11.898, -11.731, -11.626,
-11.573, -11.581, -11.650, -13.593, -12.944, -12.471, -12.147, -11.919, -11.750,
-11.642, -11.586, -11.592, -11.658, -13.618, -12.970, -12.496, -12.170, -11.940,
-11.768, -11.657, -11.600, -11.602, -11.666, -13.643, -12.995, -12.520, -12.193,
-11.961, -11.787, -11.673, -11.613, -11.613, -11.675, -13.668, -13.020, -12.545,
-12.215, -11.981, -11.805, -11.689, -11.626, -11.624, -11.683, -13.693, -13.045,
-12.569, -12.238, -12.002, -11.823, -11.705, -11.639, -11.634, -11.691, -13.717,
-13.070, -12.593, -12.260, -12.023, -11.842, -11.720, -11.653, -11.645, -11.699,
 -13.741, -13.094, -12.617, -12.283, -12.043, -11.860, -11.736, -11.666,
-11.656, -11.707, -13.765, -13.119, -12.641, -12.306, -12.064, -11.879, -11.752,
-11.679, -11.667, -11.716, -13.789, -13.143, -12.664, -12.328, -12.085, -11.897,
-11.768, -11.692, -11.677, -11.724, -13.813, -13.168, -12.688, -12.351, -12.106,
-11.915, -11.783, -11.705, -11.687, -11.731, -13.837, -13.192, -12.712, -12.374,
-12.126, -11.934, -11.799, -11.718, -11.698, -11.739, -13.860, -13.216, -12.736,
-12.396, -12.146, -11.951, -11.814, -11.731, -11.708, -11.747, -13.883, -13.240,
-12.759, -12.418, -12.167, -11.970, -11.829, -11.744, -11.718, -11.755, -13.906,
-13.263, -12.782, -12.440, -12.187, -11.988, -11.845, -11.756, -11.728, -11.762,
 -13.929, -13.287, -12.805, -12.462, -12.207, -12.006, -11.860, -11.769,
-11.738, -11.770, -13.952, -13.310, -12.828, -12.484, -12.228, -12.024, -11.876,
-11.782, -11.748, -11.778, -13.974, -13.333, -12.851, -12.506, -12.248, -12.042,
-11.891, -11.794, -11.758, -11.785, -13.997, -13.356, -12.874, -12.528, -12.268,
-12.059, -11.906, -11.807, -11.768, -11.792, -14.019, -13.379, -12.896, -12.549,
-12.288, -12.077, -11.920, -11.819, -11.777, -11.799, -14.041, -13.401, -12.919,
-12.570, -12.308, -12.094, -11.935, -11.831, -11.787, -11.806, -14.063, -13.424,
-12.941, -12.592, -12.327, -12.111, -11.950, -11.843, -11.796, -11.813, -14.084,
-13.446, -12.963, -12.613, -12.347, -12.129, -11.964, -11.855, -11.805, -11.820,
 -14.106, -13.468, -12.985, -12.634, -12.366, -12.146, -11.979, -11.866,
-11.815, -11.827, -14.127, -13.490, -13.007, -12.655, -12.386, -12.163, -11.993,
-11.878, -11.824, -11.834, -14.148, -13.512, -13.029, -12.676, -12.406, -12.181,
-12.008, -11.890, -11.833, -11.840, -14.169, -13.534, -13.051, -12.698, -12.425,
-12.198, -12.022, -11.901, -11.841, -11.847, -14.190, -13.555, -13.072, -12.718,
-12.444, -12.215, -12.036, -11.913, -11.850, -11.853, -14.211, -13.577, -13.094,
-12.739, -12.464, -12.231, -12.050, -11.924, -11.859, -11.859, -14.232, -13.598,
-13.115, -12.760, -12.483, -12.248, -12.064, -11.935, -11.867, -11.866, -14.252,
-13.619, -13.136, -12.780, -12.502, -12.265, -12.078, -11.946, -11.875, -11.872,
 -14.272, -13.640, -13.157, -12.801, -12.521, -12.281, -12.091, -11.957,
-11.884, -11.878, -14.292, -13.661, -13.178, -12.821, -12.539, -12.298, -12.105,
-11.968, -11.892, -11.883, -14.312, -13.682, -13.199, -12.841, -12.558, -12.314,
-12.118, -11.978, -11.900, -11.889, -14.332, -13.702, -13.220, -12.861, -12.577,
-12.330, -12.132, -11.988, -11.907, -11.894, -14.352, -13.723, -13.240, -12.881,
-12.595, -12.346, -12.145, -11.999, -11.915, -11.900, -14.372, -13.743, -13.261,
-12.901, -12.613, -12.362, -12.157, -12.009, -11.922, -11.905, -14.391, -13.763,
-13.281, -12.921, -12.632, -12.378, -12.170, -12.018, -11.930, -11.910, -14.410,
-13.783, -13.301, -12.940, -12.650, -12.393, -12.183, -12.028, -11.937, -11.915,
 -14.430, -13.803, -13.321, -12.960, -12.668, -12.409, -12.195, -12.038,
-11.944, -11.920, -14.449, -13.823, -13.341, -12.979, -12.685, -12.424, -12.207,
-12.047, -11.951, -11.925, -14.468, -13.842, -13.361, -12.999, -12.703, -12.439,
-12.220, -12.056, -11.957, -11.930, -14.486, -13.862, -13.380, -13.018, -12.721,
-12.454, -12.232, -12.065, -11.964, -11.934, -14.505, -13.881, -13.400, -13.037,
-12.738, -12.469, -12.243, -12.074, -11.970, -11.938, -14.524, -13.900, -13.419,
-13.056, -12.756, -12.484, -12.255, -12.083, -11.976, -11.943, -14.542, -13.919,
-13.439, -13.075, -12.773, -12.498, -12.266, -12.091, -11.982, -11.947, -14.561,
-13.938, -13.458, -13.093, -12.790, -12.513, -12.277, -12.100, -11.988, -11.951,
 -14.579, -13.957, -13.477, -13.112, -12.807, -12.527, -12.288, -12.108,
-11.994, -11.954, -14.597, -13.976, -13.496, -13.130, -12.823, -12.541, -12.299,
-12.116, -11.999, -11.958, -14.615, -13.994, -13.515, -13.148, -12.840, -12.554,
-12.310, -12.123, -12.005, -11.962, -14.633, -14.013, -13.534, -13.167, -12.856,
-12.568, -12.320, -12.130, -12.010, -11.965, -14.651, -14.031, -13.552, -13.185,
-12.873, -12.581, -12.330, -12.138, -12.014, -11.968, -14.668, -14.050, -13.571,
-13.202, -12.889, -12.594, -12.340, -12.144, -12.019, -11.971, -14.686, -14.068,
-13.589, -13.220, -12.904, -12.607, -12.349, -12.151, -12.024, -11.974, -14.703,
-14.086, -13.607, -13.238, -12.920, -12.619, -12.359, -12.158, -12.028, -11.977,
 -14.721, -14.104, -13.625, -13.255, -12.935, -12.632, -12.368, -12.164,
-12.032, -11.979, -14.738, -14.122, -13.643, -13.272, -12.950, -12.644, -12.376,
-12.170, -12.036, -11.982, -14.755, -14.139, -13.661, -13.290, -12.965, -12.655,
-12.385, -12.176, -12.039, -11.984, -14.772, -14.157, -13.679, -13.306, -12.980,
-12.667, -12.393, -12.181, -12.043, -11.986, -14.789, -14.174, -13.697, -13.323,
-12.995, -12.678, -12.401, -12.186, -12.046, -11.988, -14.806, -14.192, -13.714,
-13.340, -13.009, -12.689, -12.409, -12.191, -12.049, -11.990, -14.823, -14.209,
-13.732, -13.356, -13.023, -12.700, -12.417, -12.196, -12.052, -11.992, -14.840,
-14.226, -13.749, -13.373, -13.037, -12.710, -12.424, -12.201, -12.055, -11.993,
 -14.857, -14.243, -13.766, -13.389, -13.050, -12.720, -12.430, -12.205,
-12.058, -11.995, -14.873, -14.260, -13.783, -13.405, -13.063, -12.730, -12.437,
-12.209, -12.060, -11.996, -14.890, -14.277, -13.800, -13.420, -13.076, -12.739,
-12.443, -12.212, -12.062, -11.997, -14.906, -14.294, -13.817, -13.436, -13.089,
-12.748, -12.449, -12.216, -12.064, -11.998, -14.923, -14.311, -13.833, -13.451,
-13.102, -12.757, -12.454, -12.219, -12.066, -11.999, -14.939, -14.328, -13.850,
-13.466, -13.114, -12.765, -12.460, -12.222, -12.067, -12.000, -14.955, -14.344,
-13.866, -13.481, -13.125, -12.773, -12.464, -12.225, -12.068, -12.000, -14.971,
-14.361, -13.882, -13.496, -13.137, -12.780, -12.469, -12.227, -12.069, -12.001,
 -14.987, -14.377, -13.898, -13.510, -13.148, -12.787, -12.473, -12.229,
-12.070, -12.001, -15.003, -14.393, -13.914, -13.524, -13.158, -12.794, -12.477,
-12.231, -12.071, -12.001, -15.019, -14.409, -13.930, -13.538, -13.169, -12.800,
-12.481, -12.232, -12.072, -12.002, -15.035, -14.425, -13.946, -13.552, -13.179,
-12.806, -12.484, -12.234, -12.072, -12.002, -15.051, -14.441, -13.962, -13.565,
-13.188, -12.812, -12.487, -12.235, -12.072, -12.001, -15.066, -14.457, -13.977,
-13.579, -13.198, -12.817, -12.489, -12.236, -12.073, -12.001, -15.082, -14.473,
-13.992, -13.591, -13.207, -12.822, -12.492, -12.236, -12.072, -12.001, -15.098,
-14.489, -14.007, -13.604, -13.215, -12.826, -12.493, -12.237, -12.072, -12.000,
 -15.113, -14.504, -14.022, -13.616, -13.223, -12.830, -12.495, -12.237,
-12.072, -12.000, -15.129, -14.520, -14.037, -13.628, -13.230, -12.833, -12.496,
-12.237, -12.071, -11.999, -15.144, -14.535, -14.051, -13.640, -13.237, -12.836,
-12.497, -12.236, -12.070, -11.998, -15.160, -14.551, -14.066, -13.651, -13.244,
-12.839, -12.497, -12.236, -12.069, -11.998, -15.175, -14.566, -14.080, -13.662,
-13.251, -12.841, -12.498, -12.235, -12.068, -11.997, -15.190, -14.581, -14.094,
-13.673, -13.257, -12.843, -12.498, -12.234, -12.067, -11.996, -15.205, -14.596,
-14.108, -13.683, -13.262, -12.844, -12.497, -12.233, -12.066, -11.995, -15.220,
-14.611, -14.121, -13.693, -13.267, -12.845, -12.497, -12.231, -12.065, -11.994,
 -15.235, -14.626, -14.135, -13.702, -13.271, -12.845, -12.496, -12.230,
-12.063, -11.992, -15.250, -14.641, -14.148, -13.712, -13.275, -12.845, -12.494,
-12.228, -12.062, -11.991, -15.265, -14.656, -14.161, -13.720, -13.278, -12.845,
-12.493, -12.226, -12.060, -11.990, -15.280, -14.670, -14.174, -13.729, -13.282,
-12.844, -12.491, -12.224, -12.058, -11.988, -15.295, -14.685, -14.186, -13.737,
-13.284, -12.843, -12.490, -12.222, -12.057, -11.987, -15.310, -14.699, -14.198,
-13.745, -13.287, -12.842, -12.488, -12.220, -12.055, -11.986, -15.325, -14.713,
-14.211, -13.752, -13.289, -12.840, -12.485, -12.218, -12.053, -11.984, -15.339,
-14.728, -14.222, -13.759, -13.290, -12.838, -12.483, -12.215, -12.051, -11.983,
 -15.354, -14.742, -14.234, -13.765, -13.291, -12.836, -12.480, -12.213,
-12.049, -11.981, -15.369, -14.756, -14.245, -13.771, -13.291, -12.833, -12.477,
-12.210, -12.047, -11.980, -15.383, -14.770, -14.256, -13.777, -13.292, -12.831,
-12.474, -12.207, -12.045, -11.978, -15.398, -14.783, -14.267, -13.783, -13.292,
-12.828, -12.471, -12.205, -12.043, -11.977, -15.412, -14.797, -14.278, -13.788,
-13.292, -12.825, -12.468, -12.202, -12.040, -11.975, -15.427, -14.811, -14.288,
-13.793, -13.292, -12.822, -12.465, -12.199, -12.038, -11.974, -15.441, -14.824,
-14.299, -13.798, -13.291, -12.819, -12.462, -12.196, -12.036, -11.972, -15.455,
-14.837, -14.309, -13.802, -13.290, -12.816, -12.459, -12.194, -12.034, -11.971,
 -15.470, -14.851, -14.318, -13.806, -13.289, -12.813, -12.456, -12.191,
-12.032, -11.969, -15.484, -14.864, -14.328, -13.809, -13.288, -12.810, -12.452,
-12.188, -12.030, -11.968, -15.498, -14.877, -14.337, -13.813, -13.288, -12.807,
-12.449, -12.186, -12.028, -11.967, -15.512, -14.890, -14.347, -13.817, -13.287,
-12.804, -12.446, -12.183, -12.026, -11.965, -15.527, -14.903, -14.356, -13.821,
-13.287, -12.802, -12.443, -12.181, -12.025, -11.964, -15.540, -14.914, -14.360,
-13.815, -13.273, -12.788, -12.430, -12.170, -12.017, -11.958, -15.554, -14.925,
-14.362, -13.807, -13.260, -12.773, -12.417, -12.160, -12.009, -11.953, -15.567,
-14.933, -14.358, -13.788, -13.233, -12.747, -12.393, -12.141, -11.995, -11.943,
 -15.580, -14.940, -14.352, -13.768, -13.206, -12.721, -12.370, -12.122,
-11.981, -11.933, -15.594, -14.949, -14.351, -13.756, -13.189, -12.704, -12.354,
-12.110, -11.971, -11.926, -15.607, -14.958, -14.352, -13.748, -13.177, -12.692,
-12.342, -12.100, -11.965, -11.921, -15.621, -14.971, -14.362, -13.755, -13.183,
-12.697, -12.344, -12.102, -11.966, -11.922, -15.635, -14.984, -14.374, -13.765,
-13.190, -12.703, -12.347, -12.104, -11.968, -11.923, -15.648, -14.997, -14.386,
-13.776, -13.200, -12.711, -12.351, -12.107, -11.970, -11.925, -15.663, -15.013,
-14.405, -13.796, -13.220, -12.728, -12.362, -12.115, -11.976, -11.930, -15.677,
-15.026, -14.416, -13.806, -13.227, -12.734, -12.362, -12.116, -11.976, -11.930,
 -15.690, -15.037, -14.420, -13.802, -13.219, -12.726, -12.351, -12.107,
-11.970, -11.925, -15.703, -15.047, -14.424, -13.798, -13.211, -12.717, -12.339,
-12.097, -11.962, -11.920, -15.716, -15.052, -14.412, -13.771, -13.179, -12.687,
-12.310, -12.074, -11.945, -11.907, -15.727, -15.055, -14.398, -13.743, -13.146,
-12.657, -12.281, -12.051, -11.928, -11.895, -15.740, -15.063, -14.399, -13.738,
-13.138, -12.650, -12.272, -12.043, -11.922, -11.891, -15.754, -15.076, -14.412,
-13.750, -13.151, -12.662, -12.278, -12.048, -11.926, -11.894, -15.769, -15.095,
-14.439, -13.784, -13.185, -12.693, -12.299, -12.065, -11.939, -11.903, -15.784,
-15.119, -14.479, -13.836, -13.238, -12.739, -12.333, -12.092, -11.959, -11.918,
 -15.800, -15.142, -14.516, -13.882, -13.284, -12.779, -12.361, -12.114,
-11.976, -11.930, -15.814, -15.158, -14.534, -13.901, -13.301, -12.793, -12.371,
-12.121, -11.982, -11.934, -15.827, -15.166, -14.532, -13.891, -13.288, -12.780,
-12.362, -12.115, -11.977, -11.931, -15.838, -15.165, -14.510, -13.851, -13.244,
-12.742, -12.334, -12.093, -11.961, -11.920, -15.848, -15.162, -14.481, -13.802,
-13.190, -12.694, -12.299, -12.066, -11.942, -11.906, -15.860, -15.163, -14.464,
-13.771, -13.155, -12.661, -12.276, -12.049, -11.929, -11.897, -15.873, -15.171,
-14.463, -13.761, -13.142, -12.648, -12.269, -12.044, -11.925, -11.894, -15.887,
-15.186, -14.476, -13.773, -13.151, -12.655, -12.279, -12.052, -11.931, -11.899,
 -15.902, -15.208, -14.509, -13.811, -13.188, -12.687, -12.310, -12.077,
-11.950, -11.912, -15.918, -15.230, -14.542, -13.851, -13.227, -12.720, -12.341,
-12.102, -11.969, -11.926, -15.931, -15.243, -14.555, -13.864, -13.239, -12.732,
-12.355, -12.114, -11.979, -11.933, -15.945, -15.257, -14.569, -13.877, -13.251,
-12.741, -12.368, -12.124, -11.987, -11.939, -15.958, -15.266, -14.572, -13.874,
-13.246, -12.737, -12.368, -12.125, -11.988, -11.940, -15.970, -15.270, -14.562,
-13.855, -13.224, -12.718, -12.357, -12.117, -11.983, -11.937, -15.982, -15.278,
-14.562, -13.848, -13.215, -12.709, -12.353, -12.115, -11.981, -11.936, -15.995,
-15.288, -14.564, -13.844, -13.208, -12.702, -12.351, -12.113, -11.981, -11.935,
 -16.008, -15.297, -14.566, -13.841, -13.203, -12.697, -12.349, -12.112,
-11.980, -11.935, -16.023, -15.314, -14.587, -13.862, -13.222, -12.713, -12.364,
-12.124, -11.990, -11.942, -16.038, -15.334, -14.613, -13.890, -13.248, -12.735,
-12.384, -12.140, -12.002, -11.950, -16.053, -15.358, -14.650, -13.934, -13.289,
-12.770, -12.414, -12.164, -12.020, -11.963, -16.069, -15.382, -14.689, -13.980,
-13.334, -12.807, -12.446, -12.188, -12.038, -11.976, -16.084, -15.401, -14.714,
-14.009, -13.360, -12.829, -12.465, -12.204, -12.050, -11.985, -16.099, -15.420,
-14.740, -14.038, -13.388, -12.852, -12.485, -12.219, -12.061, -11.993, -16.112,
-15.433, -14.752, -14.049, -13.397, -12.860, -12.492, -12.225, -12.066, -11.996,
 -16.126, -15.446, -14.765, -14.060, -13.406, -12.868, -12.499, -12.231,
-12.071, -12.000, -16.139, -15.460, -14.778, -14.072, -13.417, -12.877, -12.507,
-12.237, -12.076, -12.003, -16.153, -15.473, -14.792, -14.086, -13.429, -12.886,
-12.515, -12.244, -12.081, -12.007, -16.167, -15.487, -14.806, -14.099, -13.441,
-12.897, -12.524, -12.251, -12.086, -12.011, -16.180, -15.502, -14.821, -14.114,
-13.454, -12.908, -12.533, -12.258, -12.092, -12.015, -16.194, -15.516, -14.836,
-14.129, -13.467, -12.919, -12.543, -12.266, -12.098, -12.019, -16.207, -15.530,
-14.851, -14.143, -13.481, -12.931, -12.552, -12.273, -12.104, -12.023, -16.221,
-15.544, -14.866, -14.159, -13.495, -12.943, -12.562, -12.281, -12.110, -12.027,
 -16.235, -15.558, -14.881, -14.174, -13.509, -12.955, -12.572, -12.290,
-12.116, -12.032, -16.248, -15.572, -14.896, -14.190, -13.524, -12.968, -12.583,
-12.298, -12.123, -12.036, -16.262, -15.586, -14.912, -14.207, -13.540, -12.982,
-12.594, -12.307, -12.129, -12.041, -16.276, -15.601, -14.928, -14.224, -13.557,
-12.996, -12.605, -12.316, -12.136, -12.046, -16.289, -15.615, -14.945, -14.242,
-13.573, -13.011, -12.617, -12.325, -12.143, -12.051, -16.303, -15.630, -14.961,
-14.259, -13.590, -13.026, -12.629, -12.335, -12.151, -12.056, -16.317, -15.644,
-14.977, -14.276, -13.607, -13.040, -12.641, -12.344, -12.158, -12.061, -16.330,
-15.658, -14.993, -14.294, -13.623, -13.055, -12.653, -12.354, -12.165, -12.066,
 -16.344, -15.673, -15.009, -14.311, -13.640, -13.071, -12.665, -12.364,
-12.173, -12.071, -16.357, -15.687, -15.025, -14.329, -13.658, -13.086, -12.678,
-12.374, -12.181, -12.076, -16.371, -15.701, -15.042, -14.347, -13.676, -13.102,
-12.691, -12.384, -12.189, -12.082, -16.385, -15.716, -15.059, -14.366, -13.694,
-13.119, -12.704, -12.395, -12.197, -12.087, -16.398, -15.731, -15.076, -14.384,
-13.712, -13.135, -12.718, -12.406, -12.205, -12.093, -16.412, -15.745, -15.093,
-14.403, -13.731, -13.152, -12.731, -12.417, -12.213, -12.099, -16.426, -15.760,
-15.109, -14.421, -13.749, -13.168, -12.745, -12.428, -12.221, -12.104, -16.439,
-15.774, -15.126, -14.439, -13.767, -13.185, -12.758, -12.439, -12.230, -12.110,
 -16.453, -15.788, -15.142, -14.457, -13.785, -13.201, -12.772, -12.450,
-12.238, -12.116, -16.466, -15.802, -15.158, -14.475, -13.803, -13.218, -12.786,
-12.461, -12.247, -12.122, -16.480, -15.817, -15.174, -14.494, -13.822, -13.235,
-12.800, -12.473, -12.256, -12.128, -16.493, -15.831, -15.191, -14.513, -13.841,
-13.253, -12.815, -12.485, -12.265, -12.134, -16.507, -15.846, -15.208, -14.532,
-13.860, -13.270, -12.829, -12.496, -12.274, -12.140, -16.521, -15.860, -15.225,
-14.551, -13.879, -13.288, -12.844, -12.508, -12.283, -12.146, -16.534, -15.874,
-15.241, -14.569, -13.897, -13.305, -12.858, -12.520, -12.291, -12.152, -16.548,
-15.888, -15.257, -14.587, -13.916, -13.322, -12.873, -12.532, -12.300, -12.158,
 -16.561, -15.902, -15.273, -14.604, -13.934, -13.339, -12.887, -12.544,
-12.310, -12.164, -16.575, -15.916, -15.289, -14.622, -13.952, -13.357, -12.902,
-12.556, -12.319, -12.170, -16.588, -15.931, -15.305, -14.641, -13.971, -13.374,
-12.917, -12.568, -12.328, -12.177, -16.602, -15.945, -15.321, -14.659, -13.990,
-13.392, -12.932, -12.581, -12.337, -12.183, -16.615, -15.959, -15.337, -14.678,
-14.010, -13.410, -12.947, -12.593, -12.347, -12.189, -16.629, -15.973, -15.354,
-14.696, -14.029, -13.428, -12.962, -12.605, -12.356, -12.195, -16.642, -15.987,
-15.370, -14.714, -14.047, -13.445, -12.977, -12.618, -12.365, -12.201, -16.656,
-16.001, -15.385, -14.732, -14.065, -13.463, -12.992, -12.630, -12.374, -12.208,
 -16.669, -16.015, -15.401, -14.750, -14.084, -13.480, -13.007, -12.642,
-12.384, -12.214, -16.683, -16.029, -15.417, -14.767, -14.102, -13.497, -13.022,
-12.655, -12.393, -12.220, -16.696, -16.043, -15.433, -14.786, -14.121, -13.515,
-13.037, -12.667, -12.402, -12.226, -16.710, -16.057, -15.448, -14.803, -14.139,
-13.533, -13.052, -12.680, -12.412, -12.232, -16.723, -16.071, -15.464, -14.822,
-14.159, -13.551, -13.068, -12.692, -12.421, -12.238, -16.737, -16.085, -15.480,
-14.840, -14.177, -13.569, -13.083, -12.705, -12.431, -12.244, -16.750, -16.099,
-15.496, -14.857, -14.195, -13.586, -13.098, -12.717, -12.440, -12.250, -16.764,
-16.113, -15.511, -14.875, -14.213, -13.603, -13.113, -12.729, -12.449, -12.256,
 -16.777, -16.127, -15.526, -14.892, -14.231, -13.620, -13.128, -12.742,
-12.458, -12.262, -16.790, -16.141, -15.541, -14.909, -14.249, -13.637, -13.143,
-12.754, -12.467, -12.267, -16.804, -16.154, -15.557, -14.926, -14.267, -13.655,
-13.158, -12.767, -12.476, -12.273, -16.817, -16.168, -15.572, -14.943, -14.286,
-13.672, -13.173, -12.779, -12.485, -12.279, -16.831, -16.182, -15.587, -14.961,
-14.304, -13.690, -13.188, -12.791, -12.494, -12.284, -16.844, -16.196, -15.603,
-14.979, -14.322, -13.707, -13.203, -12.804, -12.503, -12.290, -16.858, -16.210,
-15.618, -14.996, -14.340, -13.724, -13.218, -12.816, -12.512, -12.295, -16.871,
-16.223, -15.633, -15.013, -14.358, -13.742, -13.233, -12.828, -12.521, -12.301,
 -16.884, -16.237, -15.648, -15.030, -14.376, -13.758, -13.247, -12.840,
-12.529, -12.306, -16.898, -16.251, -15.663, -15.046, -14.393, -13.775, -13.262,
-12.852, -12.538, -12.311, -16.911, -16.264, -15.678, -15.063, -14.411, -13.793,
-13.277, -12.864, -12.547, -12.316, -16.925, -16.278, -15.693, -15.080, -14.429,
-13.810, -13.292, -12.876, -12.555, -12.321, -16.938, -16.292, -15.708, -15.097,
-14.447, -13.827, -13.307, -12.888, -12.564, -12.326, -16.951, -16.305, -15.723,
-15.114, -14.465, -13.844, -13.322, -12.900, -12.572, -12.331, -16.965, -16.319,
-15.738, -15.131, -14.482, -13.861, -13.336, -12.911, -12.580, -12.335, -16.978,
-16.332, -15.753, -15.147, -14.500, -13.877, -13.351, -12.923, -12.588, -12.340,
 -16.991, -16.346, -15.767, -15.163, -14.517, -13.894, -13.365, -12.934,
-12.596, -12.344, -17.005, -16.360, -15.782, -15.180, -14.534, -13.910, -13.379,
-12.946, -12.604, -12.349, -17.018, -16.373, -15.796, -15.196, -14.551, -13.927,
-13.393, -12.957, -12.612, -12.353, -17.032, -16.387, -15.811, -15.212, -14.568,
-13.943, -13.408, -12.968, -12.619, -12.357, -17.045, -16.400, -15.825, -15.229,
-14.586, -13.960, -13.422, -12.980, -12.627, -12.361, -17.058, -16.414, -15.840,
-15.245, -14.603, -13.977, -13.436, -12.991, -12.634, -12.365, -17.072, -16.427,
-15.855, -15.261, -14.620, -13.993, -13.450, -13.001, -12.641, -12.368, -17.085,
-16.441, -15.869, -15.277, -14.637, -14.009, -13.464, -13.012, -12.648, -12.372,
 -17.098, -16.454, -15.883, -15.293, -14.654, -14.025, -13.478, -13.023,
-12.655, -12.375, -17.112, -16.467, -15.898, -15.309, -14.671, -14.041, -13.491,
-13.033, -12.662, -12.378, -17.125, -16.481, -15.912, -15.325, -14.688, -14.058,
-13.505, -13.044, -12.668, -12.381, -17.139, -16.494, -15.926, -15.341, -14.704,
-14.074, -13.519, -13.054, -12.675, -12.384, -17.152, -16.508, -15.941, -15.357,
-14.721, -14.090, -13.533, -13.064, -12.681, -12.387, -17.165, -16.521, -15.955,
-15.373, -14.738, -14.106, -13.546, -13.074, -12.687, -12.390, -17.179, -16.534,
-15.969, -15.388, -14.754, -14.122, -13.559, -13.084, -12.693, -12.392, -17.192,
-16.548, -15.983, -15.404, -14.771, -14.137, -13.572, -13.094, -12.699, -12.394,
 -17.205, -16.561, -15.997, -15.419, -14.787, -14.153, -13.585, -13.103,
-12.704, -12.396, -17.219, -16.575, -16.011, -15.435, -14.803, -14.168, -13.598,
-13.112, -12.710, -12.398, -17.232, -16.588, -16.025, -15.450, -14.820, -14.184,
-13.611, -13.122, -12.715, -12.400, -17.245, -16.601, -16.039, -15.466, -14.836,
-14.199, -13.624, -13.131, -12.720, -12.401, -17.259, -16.615, -16.053, -15.481,
-14.852, -14.215, -13.636, -13.140, -12.725, -12.403, -17.272, -16.628, -16.067,
-15.496, -14.868, -14.230, -13.649, -13.148, -12.729, -12.404, -17.285, -16.641,
-16.081, -15.512, -14.885, -14.245, -13.661, -13.157, -12.734, -12.405, -17.299,
-16.654, -16.095, -15.527, -14.901, -14.260, -13.673, -13.165, -12.738, -12.406,
 -17.312, -16.668, -16.109, -15.542, -14.917, -14.275, -13.685, -13.173,
-12.742, -12.407, -17.325, -16.681, -16.122, -15.557, -14.932, -14.290, -13.697,
-13.181, -12.745, -12.407, -17.339, -16.694, -16.136, -15.572, -14.948, -14.305,
-13.709, -13.189, -12.749, -12.407, -17.352, -16.707, -16.150, -15.587, -14.964,
-14.319, -13.721, -13.196, -12.752, -12.407, -17.365, -16.721, -16.164, -15.602,
-14.980, -14.334, -13.732, -13.203, -12.756, -12.407, -17.379, -16.734, -16.177,
-15.617, -14.995, -14.349, -13.743, -13.210, -12.758, -12.407, -17.392, -16.747,
-16.191, -15.632, -15.011, -14.363, -13.754, -13.217, -12.761, -12.407, -17.405,
-16.760, -16.205, -15.646, -15.026, -14.377, -13.765, -13.224, -12.763, -12.406,
 -17.419, -16.774, -16.218, -15.661, -15.042, -14.391, -13.776, -13.230,
-12.766, -12.405, -17.432, -16.787, -16.232, -15.676, -15.057, -14.405, -13.786,
-13.236, -12.768, -12.404, -17.445, -16.800, -16.245, -15.690, -15.072, -14.419,
-13.797, -13.242, -12.769, -12.403, -17.459, -16.813, -16.259, -15.705, -15.087,
-14.433, -13.807, -13.248, -12.771, -12.401, -17.472, -16.826, -16.272, -15.719,
-15.102, -14.447, -13.817, -13.253, -12.772, -12.400, -17.485, -16.840, -16.286,
-15.734, -15.117, -14.460, -13.827, -13.258, -12.773, -12.398, -17.499, -16.853,
-16.299, -15.748, -15.132, -14.474, -13.836, -13.263, -12.774, -12.396, -17.512,
-16.866, -16.313, -15.763, -15.147, -14.487, -13.846, -13.268, -12.774, -12.394,
 -17.525, -16.879, -16.326, -15.777, -15.162, -14.500, -13.855, -13.272,
-12.774, -12.391, -17.539, -16.892, -16.340, -15.791, -15.177, -14.513, -13.864,
-13.276, -12.774, -12.389, -17.552, -16.905, -16.353, -15.805, -15.191, -14.526,
-13.872, -13.280, -12.774, -12.386, -17.565, -16.919, -16.366, -15.819, -15.206,
-14.538, -13.881, -13.283, -12.774, -12.383, -17.579, -16.932, -16.380, -15.834,
-15.220, -14.551, -13.889, -13.287, -12.773, -12.380, -17.592, -16.945, -16.393,
-15.848, -15.234, -14.563, -13.897, -13.289, -12.772, -12.377, -17.605, -16.958,
-16.406, -15.862, -15.249, -14.576, -13.904, -13.292, -12.771, -12.373, -17.619,
-16.971, -16.419, -15.876, -15.263, -14.588, -13.912, -13.294, -12.769, -12.369,
 -17.632, -16.984, -16.433, -15.890, -15.277, -14.599, -13.919, -13.296,
-12.767, -12.366, -17.645, -16.997, -16.446, -15.903, -15.291, -14.611, -13.926,
-13.298, -12.765, -12.362, -17.659, -17.010, -16.459, -15.917, -15.305, -14.622,
-13.932, -13.300, -12.763, -12.357, -17.672, -17.024, -16.472, -15.931, -15.318,
-14.633, -13.938, -13.301, -12.761, -12.353, -17.685, -17.037, -16.485, -15.945,
-15.332, -14.644, -13.944, -13.301, -12.758, -12.348, -17.699, -17.050, -16.498,
-15.958, -15.346, -14.655, -13.950, -13.302, -12.755, -12.344, -17.712, -17.063,
-16.512, -15.972, -15.359, -14.666, -13.955, -13.302, -12.752, -12.339, -17.725,
-17.076, -16.525, -15.986, -15.373, -14.676, -13.960, -13.302, -12.748, -12.334,
 -17.739, -17.089, -16.538, -15.999, -15.386, -14.686, -13.965, -13.302,
-12.744, -12.329, -17.752, -17.102, -16.551, -16.013, -15.399, -14.696, -13.969,
-13.301, -12.741, -12.323, -17.765, -17.115, -16.564, -16.026, -15.412, -14.706,
-13.973, -13.300, -12.736, -12.318, -17.779, -17.128, -16.577, -16.039, -15.425,
-14.715, -13.977, -13.298, -12.732, -12.312, -17.792, -17.141, -16.590, -16.053,
-15.437, -14.724, -13.980, -13.297, -12.727, -12.306, -17.805, -17.154, -16.603,
-16.066, -15.450, -14.733, -13.983, -13.294, -12.722, -12.300, -17.819, -17.167,
-16.616, -16.079, -15.462, -14.741, -13.986, -13.292, -12.717, -12.294, -17.832,
-17.180, -16.629, -16.093, -15.475, -14.750, -13.988, -13.290, -12.712, -12.288,
 -17.845, -17.194, -16.642, -16.106, -15.487, -14.758, -13.990, -13.287,
-12.707, -12.282, -17.859, -17.207, -16.655, -16.119, -15.499, -14.765, -13.992,
-13.284, -12.701, -12.276, -17.872, -17.220, -16.668, -16.132, -15.511, -14.772,
-13.993, -13.280, -12.695, -12.269, -17.885, -17.233, -16.681, -16.145, -15.523,
-14.779, -13.994, -13.276, -12.689, -12.262, -17.898, -17.246, -16.693, -16.158,
-15.534, -14.786, -13.994, -13.272, -12.682, -12.256, -17.912, -17.259, -16.706,
-16.171, -15.545, -14.792, -13.994, -13.267, -12.676, -12.249, -17.925, -17.272,
-16.719, -16.183, -15.557, -14.798, -13.994, -13.263, -12.669, -12.242, -17.938,
-17.285, -16.732, -16.196, -15.568, -14.804, -13.994, -13.258, -12.662, -12.235,
 -17.952, -17.298, -16.745, -16.209, -15.579, -14.809, -13.993, -13.252,
-12.656, -12.228, -17.965, -17.311, -16.758, -16.222, -15.589, -14.814, -13.992,
-13.247, -12.648, -12.220, -17.978, -17.324, -16.770, -16.234, -15.600, -14.819,
-13.990, -13.241, -12.641, -12.213, -17.992, -17.337, -16.783, -16.247, -15.610,
-14.823, -13.988, -13.235, -12.634, -12.206, -18.005, -17.350, -16.796, -16.259,
-15.620, -14.827, -13.985, -13.229, -12.626, -12.198, -18.018, -17.363, -16.809,
-16.272, -15.630, -14.830, -13.982, -13.222, -12.618, -12.191, -18.032, -17.376,
-16.821, -16.284, -15.640, -14.833, -13.979, -13.215, -12.611, -12.184, -18.045,
-17.389, -16.834, -16.296, -15.649, -14.836, -13.976, -13.208, -12.603, -12.176,
 -18.058, -17.402, -16.847, -16.309, -15.658, -14.838, -13.972, -13.201,
-12.595, -12.169, -18.072, -17.415, -16.860, -16.321, -15.667, -14.840, -13.968,
-13.194, -12.587, -12.161, -18.085, -17.428, -16.872, -16.333, -15.676, -14.841,
-13.963, -13.186, -12.578, -12.153, -18.098, -17.441, -16.885, -16.345, -15.684,
-14.842, -13.958, -13.178, -12.570, -12.146, -18.111, -17.454, -16.898, -16.357,
-15.693, -14.843, -13.953, -13.170, -12.562, -12.138, -18.125, -17.467, -16.910,
-16.369, -15.700, -14.843, -13.948, -13.162, -12.553, -12.130, -18.138, -17.480,
-16.923, -16.380, -15.708, -14.843, -13.942, -13.154, -12.545, -12.123, -18.151,
-17.493, -16.935, -16.392, -15.716, -14.843, -13.936, -13.146, -12.537, -12.115,
 -18.165, -17.506, -16.948, -16.404, -15.723, -14.842, -13.930, -13.138,
-12.528, -12.108, -18.178, -17.519, -16.961, -16.415, -15.730, -14.841, -13.924,
-13.129, -12.520, -12.100, -18.191, -17.532, -16.973, -16.427, -15.736, -14.839,
-13.917, -13.120, -12.511, -12.093, -18.204, -17.545, -16.986, -16.438, -15.743,
-14.837, -13.910, -13.111, -12.503, -12.085, -18.218, -17.557, -16.998, -16.449,
-15.749, -14.834, -13.903, -13.103, -12.494, -12.078, -18.231, -17.570, -17.011,
-16.461, -15.754, -14.831, -13.895, -13.094, -12.486, -12.070, -18.244, -17.583,
-17.023, -16.472, -15.760, -14.829, -13.888, -13.085, -12.477, -12.063, -18.257,
-17.596, -17.036, -16.483, -15.765, -14.825, -13.880, -13.076, -12.469, -12.056,
 -18.271, -17.609, -17.048, -16.494, -15.770, -14.822, -13.873, -13.067,
-12.461, -12.048, -18.284, -17.622, -17.060, -16.505, -15.775, -14.818, -13.865,
-13.058, -12.452, -12.041, -18.297, -17.635, -17.073, -16.515, -15.779, -14.814,
-13.857, -13.049, -12.444, -12.034, -18.310, -17.648, -17.085, -16.526, -15.783,
-14.809, -13.849, -13.041, -12.436, -12.027, -18.324, -17.661, -17.098, -16.536,
-15.787, -14.805, -13.841, -13.032, -12.428, -12.021, -18.337, -17.674, -17.110,
-16.547, -15.791, -14.800, -13.833, -13.023, -12.420, -12.014, -18.350, -17.686,
-17.122, -16.557, -15.794, -14.795, -13.825, -13.015, -12.412, -12.007, -18.363,
-17.699, -17.135, -16.567, -15.798, -14.791, -13.817, -13.006, -12.405, -12.001,
 -18.376, -17.712, -17.147, -16.578, -15.801, -14.786, -13.809, -12.998,
-12.397, -11.994, -18.390, -17.725, -17.159, -16.588, -15.804, -14.781, -13.801,
-12.990, -12.390, -11.988, -18.403, -17.738, -17.171, -16.598, -15.807, -14.776,
-13.793, -12.982, -12.382, -11.982, -18.416, -17.751, -17.184, -16.608, -15.809,
-14.771, -13.785, -12.974, -12.375, -11.976, -18.429, -17.763, -17.196, -16.617,
-15.812, -14.766, -13.778, -12.966, -12.368, -11.970, -18.442, -17.776, -17.208,
-16.627, -15.815, -14.762, -13.770, -12.958, -12.362, -11.964, -18.455, -17.789,
-17.220, -16.637, -15.817, -14.758, -13.763, -12.951, -12.355, -11.959, -18.469,
-17.802, -17.232, -16.647, -15.820, -14.754, -13.757, -12.944, -12.349, -11.953,
 -18.482, -17.815, -17.244, -16.655, -15.818, -14.742, -13.742, -12.930,
-12.337, -11.944, -18.495, -17.827, -17.256, -16.662, -15.811, -14.726, -13.722,
-12.913, -12.323, -11.933, -18.508, -17.840, -17.267, -16.667, -15.799, -14.701,
-13.695, -12.889, -12.303, -11.917, -18.521, -17.853, -17.279, -16.670, -15.784,
-14.673, -13.664, -12.862, -12.281, -11.900, -18.534, -17.865, -17.290, -16.675,
-15.772, -14.650, -13.638, -12.839, -12.263, -11.885, -18.547, -17.878, -17.302,
-16.681, -15.766, -14.637, -13.622, -12.824, -12.250, -11.876, -18.560, -17.891,
-17.314, -16.690, -15.767, -14.632, -13.614, -12.817, -12.244, -11.870, -18.573,
-17.903, -17.326, -16.701, -15.777, -14.640, -13.618, -12.819, -12.245, -11.871,
 -18.586, -17.916, -17.338, -16.712, -15.784, -14.644, -13.616, -12.817,
-12.243, -11.869, -18.599, -17.928, -17.350, -16.725, -15.800, -14.660, -13.626,
-12.824, -12.248, -11.872, -18.612, -17.941, -17.362, -16.739, -15.815, -14.674,
-13.634, -12.829, -12.251, -11.874, -18.624, -17.953, -17.374, -16.746, -15.811,
-14.663, -13.617, -12.814, -12.239, -11.865, -18.637, -17.966, -17.386, -16.754,
-15.812, -14.658, -13.605, -12.803, -12.229, -11.857, -18.650, -17.978, -17.397,
-16.757, -15.797, -14.633, -13.576, -12.777, -12.208, -11.840, -18.663, -17.991,
-17.407, -16.756, -15.773, -14.598, -13.537, -12.744, -12.181, -11.819, -18.675,
-18.003, -17.418, -16.761, -15.764, -14.583, -13.517, -12.727, -12.167, -11.808,
 -18.688, -18.015, -17.430, -16.770, -15.771, -14.588, -13.516, -12.725,
-12.166, -11.807, -18.701, -18.028, -17.442, -16.786, -15.793, -14.611, -13.532,
-12.738, -12.176, -11.815, -18.713, -18.040, -17.456, -16.807, -15.831, -14.654,
-13.565, -12.765, -12.197, -11.831, -18.726, -18.052, -17.469, -16.830, -15.871,
-14.699, -13.600, -12.793, -12.219, -11.847, -18.738, -18.064, -17.481, -16.846,
-15.892, -14.721, -13.616, -12.805, -12.228, -11.854, -18.751, -18.076, -17.492,
-16.854, -15.893, -14.717, -13.612, -12.801, -12.225, -11.851, -18.763, -18.089,
-17.503, -16.853, -15.871, -14.685, -13.585, -12.779, -12.207, -11.838, -18.775,
-18.101, -17.512, -16.847, -15.835, -14.637, -13.544, -12.746, -12.181, -11.818,
 -18.787, -18.112, -17.522, -16.844, -15.807, -14.598, -13.513, -12.719,
-12.160, -11.801, -18.800, -18.124, -17.533, -16.846, -15.794, -14.578, -13.498,
-12.707, -12.150, -11.793, -18.812, -18.136, -17.544, -16.855, -15.797, -14.576,
-13.500, -12.708, -12.151, -11.794, -18.823, -18.148, -17.556, -16.872, -15.821,
-14.601, -13.526, -12.729, -12.167, -11.807, -18.835, -18.160, -17.569, -16.892,
-15.854, -14.636, -13.561, -12.759, -12.191, -11.824, -18.847, -18.171, -17.581,
-16.907, -15.873, -14.655, -13.582, -12.776, -12.205, -11.835, -18.859, -18.183,
-17.592, -16.917, -15.881, -14.662, -13.593, -12.785, -12.212, -11.841, -18.870,
-18.194, -17.603, -16.928, -15.889, -14.667, -13.602, -12.793, -12.218, -11.845,
 -18.882, -18.205, -17.613, -16.929, -15.876, -14.648, -13.588, -12.782,
-12.209, -11.838, -18.893, -18.217, -17.623, -16.933, -15.867, -14.634, -13.578,
-12.774, -12.203, -11.834, -18.904, -18.228, -17.634, -16.940, -15.867, -14.630,
-13.576, -12.773, -12.202, -11.833, -18.915, -18.239, -17.644, -16.943, -15.858,
-14.616, -13.566, -12.764, -12.196, -11.828, -18.926, -18.250, -17.655, -16.957,
-15.874, -14.630, -13.580, -12.776, -12.205, -11.835, -18.937, -18.260, -17.667,
-16.971, -15.891, -14.647, -13.596, -12.789, -12.215, -11.843, -18.947, -18.271,
-17.679, -16.991, -15.924, -14.680, -13.627, -12.814, -12.235, -11.858, -18.957,
-18.282, -17.691, -17.012, -15.958, -14.716, -13.660, -12.840, -12.255, -11.873,
 -18.967, -18.292, -17.702, -17.029, -15.984, -14.743, -13.685, -12.860,
-12.271, -11.885, -18.977, -18.302, -17.713, -17.047, -16.012, -14.772, -13.711,
-12.881, -12.287, -11.897, -18.987, -18.312, -17.724, -17.057, -16.020, -14.778,
-13.717, -12.886, -12.291, -11.900, -18.996, -18.322, -17.734, -17.067, -16.030,
-14.786, -13.724, -12.891, -12.295, -11.903, -19.006, -18.332, -17.744, -17.078,
-16.040, -14.794, -13.731, -12.897, -12.300, -11.907, -19.015, -18.341, -17.754,
-17.089, -16.050, -14.803, -13.739, -12.903, -12.304, -11.911, -19.023, -18.351,
-17.764, -17.100, -16.062, -14.813, -13.747, -12.910, -12.310, -11.915, -19.031,
-18.360, -17.774, -17.111, -16.073, -14.824, -13.755, -12.916, -12.315, -11.919,
 -19.039, -18.369, -17.784, -17.122, -16.086, -14.835, -13.764, -12.924,
-12.321, -11.923, -19.047, -18.377, -17.793, -17.133, -16.098, -14.846, -13.774,
-12.931, -12.327, -11.928, -19.054, -18.386, -17.803, -17.144, -16.110, -14.858,
-13.783, -12.939, -12.333, -11.932, -19.061, -18.394, -17.812, -17.155, -16.123,
-14.870, -13.793, -12.947, -12.339, -11.937, -19.068, -18.402, -17.821, -17.166,
-16.136, -14.882, -13.804, -12.955, -12.346, -11.943, -19.074, -18.409, -17.830,
-17.177, -16.150, -14.896, -13.814, -12.964, -12.353, -11.948, -19.080, -18.417,
-17.839, -17.188, -16.164, -14.909, -13.826, -12.973, -12.360, -11.953, -19.085,
-18.424, -17.848, -17.200, -16.178, -14.923, -13.837, -12.982, -12.367, -11.959,
 -19.089, -18.430, -17.856, -17.211, -16.192, -14.938, -13.849, -12.992,
-12.375, -11.965, -19.094, -18.437, -17.865, -17.222, -16.206, -14.952, -13.861,
-13.001, -12.383, -11.971, -19.097, -18.443, -17.873, -17.233, -16.221, -14.967,
-13.873, -13.011, -12.391, -11.977, -19.100, -18.448, -17.880, -17.243, -16.235,
-14.981, -13.886, -13.021, -12.399, -11.984, -19.103, -18.453, -17.888, -17.254,
-16.249, -14.996, -13.898, -13.032, -12.407, -11.990, -19.105, -18.458, -17.895,
-17.265, -16.264, -15.012, -13.911, -13.043, -12.416, -11.997, -19.106, -18.462,
-17.903, -17.276, -16.279, -15.028, -13.925, -13.054, -12.425, -12.004, -19.106,
-18.466, -17.910, -17.286, -16.294, -15.044, -13.938, -13.065, -12.434, -12.011,
 -19.106, -18.470, -17.916, -17.297, -16.310, -15.060, -13.952, -13.077,
-12.443, -12.019, -19.105, -18.473, -17.923, -17.307, -16.325, -15.076, -13.966,
-13.088, -12.452, -12.026, -19.104, -18.475, -17.929, -17.317, -16.339, -15.092,
-13.980, -13.100, -12.462, -12.033, -19.101, -18.477, -17.934, -17.327, -16.354,
-15.108, -13.994, -13.111, -12.471, -12.041, -19.098, -18.479, -17.940, -17.337,
-16.369, -15.124, -14.008, -13.123, -12.481, -12.049, -19.094, -18.479, -17.945,
-17.346, -16.384, -15.141, -14.022, -13.136, -12.491, -12.057, -19.089, -18.480,
-17.950, -17.356, -16.399, -15.158, -14.037, -13.148, -12.502, -12.065, -19.083,
-18.479, -17.954, -17.365, -16.414, -15.175, -14.052, -13.161, -12.512, -12.073,
 -19.076, -18.478, -17.958, -17.375, -16.429, -15.192, -14.067, -13.174,
-12.523, -12.082, -19.068, -18.477, -17.962, -17.384, -16.444, -15.209, -14.082,
-13.187, -12.533, -12.090, -19.059, -18.475, -17.965, -17.392, -16.458, -15.225,
-14.097, -13.199, -12.544, -12.098, -19.049, -18.472, -17.968, -17.401, -16.473,
-15.242, -14.112, -13.212, -12.554, -12.107, -19.039, -18.468, -17.971, -17.409,
-16.487, -15.259, -14.127, -13.225, -12.565, -12.116, -19.027, -18.464, -17.973,
-17.417, -16.501, -15.276, -14.142, -13.239, -12.576, -12.125, -19.014, -18.459,
-17.975, -17.425, -16.516, -15.293, -14.158, -13.252, -12.588, -12.134, -19.000,
-18.453, -17.976, -17.433, -16.530, -15.310, -14.173, -13.266, -12.599, -12.143,
 -18.985, -18.447, -17.977, -17.440, -16.545, -15.327, -14.189, -13.280,
-12.610, -12.152, -18.969, -18.440, -17.977, -17.447, -16.559, -15.344, -14.204,
-13.293, -12.622, -12.161, -18.952, -18.432, -17.977, -17.454, -16.573, -15.361,
-14.220, -13.307, -12.633, -12.171, -18.934, -18.423, -17.977, -17.461, -16.586,
-15.377, -14.235, -13.321, -12.645, -12.180, -18.915, -18.414, -17.976, -17.467,
-16.600, -15.394, -14.251, -13.334, -12.656, -12.189, -18.895, -18.404, -17.975,
-17.474, -16.613, -15.411, -14.267, -13.348, -12.668, -12.199, -18.874, -18.393,
-17.973, -17.479, -16.627, -15.428, -14.283, -13.363, -12.680, -12.209, -18.852,
-18.381, -17.970, -17.485, -16.640, -15.445, -14.299, -13.377, -12.692, -12.218,
 -18.828, -18.369, -17.968, -17.491, -16.654, -15.462, -14.315, -13.391,
-12.704, -12.228, -18.804, -18.356, -17.964, -17.496, -16.667, -15.479, -14.330,
-13.405, -12.716, -12.238, -18.779, -18.342, -17.960, -17.500, -16.680, -15.495,
-14.346, -13.419, -12.728, -12.248, -18.753, -18.327, -17.956, -17.505, -16.692,
-15.511, -14.362, -13.433, -12.740, -12.258, -18.726, -18.312, -17.951, -17.509,
-16.704, -15.528, -14.377, -13.447, -12.752, -12.268, -18.698, -18.295, -17.946,
-17.513, -16.717, -15.544, -14.393, -13.462, -12.765, -12.278, -18.669, -18.279,
-17.940, -17.517, -16.729, -15.561, -14.409, -13.476, -12.777, -12.288, -18.639,
-18.261, -17.934, -17.520, -16.742, -15.577, -14.425, -13.491, -12.789, -12.298,
 -18.609, -18.243, -17.927, -17.523, -16.754, -15.594, -14.441, -13.505,
-12.802, -12.308, -18.578, -18.224, -17.920, -17.526, -16.766, -15.610, -14.457,
-13.519, -12.814, -12.318, -18.546, -18.205, -17.912, -17.528, -16.777, -15.626,
-14.473, -13.534, -12.827, -12.329, -18.513, -18.185, -17.904, -17.531, -16.789,
-15.642, -14.488, -13.548, -12.839, -12.339, -18.480, -18.164, -17.896, -17.533,
-16.800, -15.658, -14.504, -13.562, -12.851, -12.349, -18.447, -18.143, -17.887,
-17.534, -16.811, -15.674, -14.520, -13.577, -12.864, -12.359, -18.412, -18.121,
-17.877, -17.535, -16.822, -15.690, -14.536, -13.592, -12.876, -12.370, -18.378,
-18.099, -17.867, -17.537, -16.833, -15.706, -14.552, -13.606, -12.889, -12.380,
 -18.343, -18.077, -17.857, -17.537, -16.844, -15.722, -14.568, -13.621,
-12.902, -12.391, -18.308, -18.054, -17.847, -17.538, -16.854, -15.738, -14.583,
-13.635, -12.914, -12.401, -18.272, -18.031, -17.836, -17.538, -16.865, -15.753,
-14.599, -13.650, -12.927, -12.411, -18.236, -18.007, -17.825, -17.538, -16.875,
-15.769, -14.614, -13.664, -12.939, -12.422, -18.200, -17.984, -17.814, -17.538,
-16.885, -15.784, -14.629, -13.678, -12.951, -12.432, -18.164, -17.960, -17.802,
-17.538, -16.894, -15.799, -14.645, -13.693, -12.964, -12.443, -18.128, -17.936,
-17.790, -17.537, -16.904, -15.815, -14.661, -13.707, -12.977, -12.453, -18.091,
-17.912, -17.778, -17.536, -16.914, -15.830, -14.677, -13.722, -12.989, -12.463,
 -18.055, -17.888, -17.766, -17.535, -16.923, -15.846, -14.692, -13.736,
-13.002, -12.474, -18.019, -17.864, -17.754, -17.534, -16.933, -15.861, -14.708,
-13.751, -13.014, -12.484, -17.984, -17.840, -17.742, -17.533, -16.942, -15.876,
-14.723, -13.765, -13.027, -12.495, -17.948, -17.816, -17.729, -17.531, -16.951,
-15.891, -14.738, -13.779, -13.039, -12.505, -17.913, -17.792, -17.717, -17.530,
-16.960, -15.905, -14.754, -13.793, -13.052, -12.515, -17.878, -17.769, -17.705,
-17.528, -16.968, -15.920, -14.769, -13.808, -13.064, -12.526, -17.844, -17.745,
-17.693, -17.527, -16.977, -15.935, -14.785, -13.822, -13.077, -12.536, -17.810,
-17.723, -17.681, -17.525, -16.986, -15.950, -14.800, -13.837, -13.090, -12.547,
 -17.777, -17.701, -17.669, -17.524, -16.994, -15.965, -14.816, -13.851,
-13.102, -12.557, -17.745, -17.679, -17.658, -17.522, -17.003, -15.979, -14.831,
-13.865, -13.114, -12.567, -17.714, -17.658, -17.646, -17.520, -17.011, -15.994,
-14.846, -13.879, -13.127, -12.578, -17.683, -17.637, -17.636, -17.519, -17.019,
-16.008, -14.861, -13.893, -13.139, -12.588, -17.653, -17.617, -17.625, -17.518,
-17.027, -16.022, -14.876, -13.907, -13.151, -12.598, -17.625, -17.598, -17.615,
-17.517, -17.035, -16.036, -14.891, -13.922, -13.164, -12.608, -17.598, -17.580,
-17.606, -17.516, -17.043, -16.050, -14.906, -13.936, -13.176, -12.619, -17.572,
-17.563, -17.597, -17.515, -17.052, -16.065, -14.921, -13.950, -13.188, -12.629,
 -17.474, -17.500, -17.560, -17.500, -17.055, -16.078, -14.936, -13.964,
-13.201, -12.639, -17.375, -17.434, -17.520, -17.483, -17.057, -16.091, -14.951,
-13.978, -13.213, -12.649, -17.280, -17.370, -17.482, -17.466, -17.059, -16.103,
-14.965, -13.992, -13.225, -12.659, -17.190, -17.309, -17.444, -17.449, -17.061,
-16.116, -14.980, -14.006, -13.237, -12.669, -17.125, -17.264, -17.417, -17.439,
-17.065, -16.129, -14.994, -14.020, -13.249, -12.679, -17.106, -17.254, -17.413,
-17.440, -17.073, -16.143, -15.009, -14.034, -13.261, -12.689, -17.089, -17.245,
-17.410, -17.442, -17.082, -16.157, -15.024, -14.047, -13.273, -12.699, -17.053,
-17.221, -17.396, -17.438, -17.088, -16.170, -15.039, -14.061, -13.286, -12.709,
 -17.037, -17.210, -17.391, -17.438, -17.095, -16.184, -15.053, -14.075,
-13.298, -12.719, -16.971, -17.164, -17.361, -17.424, -17.097, -16.196, -15.068,
-14.089, -13.309, -12.729, -16.891, -17.105, -17.323, -17.406, -17.097, -16.208,
-15.082, -14.102, -13.321, -12.739, -16.819, -17.054, -17.290, -17.390, -17.098,
-16.220, -15.096, -14.116, -13.333, -12.748, -16.759, -17.013, -17.264, -17.379,
-17.101, -16.232, -15.110, -14.129, -13.344, -12.758, -16.753, -17.012, -17.267,
-17.384, -17.110, -16.245, -15.125, -14.143, -13.356, -12.767, -16.781, -17.038,
-17.289, -17.400, -17.124, -16.260, -15.139, -14.156, -13.368, -12.777, -16.774,
-17.038, -17.292, -17.406, -17.133, -16.274, -15.154, -14.170, -13.380, -12.787,
 -16.833, -17.086, -17.327, -17.429, -17.149, -16.289, -15.168, -14.184,
-13.391, -12.796, -16.855, -17.107, -17.344, -17.442, -17.161, -16.303, -15.183,
-14.197, -13.403, -12.805, -16.801, -17.068, -17.319, -17.430, -17.163, -16.315,
-15.197, -14.210, -13.414, -12.815, -16.832, -17.084, -17.327, -17.436, -17.171,
-16.328, -15.211, -14.223, -13.425, -12.824, -16.752, -17.019, -17.281, -17.411,
-17.168, -16.338, -15.224, -14.236, -13.437, -12.833, -16.572, -16.882, -17.185,
-17.356, -17.151, -16.345, -15.237, -14.249, -13.448, -12.842, -16.450, -16.789,
-17.121, -17.319, -17.143, -16.354, -15.251, -14.262, -13.459, -12.851, -16.381,
-16.743, -17.092, -17.307, -17.144, -16.366, -15.265, -14.276, -13.470, -12.860,
 -16.394, -16.762, -17.112, -17.324, -17.158, -16.380, -15.279, -14.289,
-13.482, -12.869, -16.559, -16.897, -17.213, -17.388, -17.192, -16.399, -15.294,
-14.302, -13.492, -12.878, -16.805, -17.082, -17.341, -17.465, -17.230, -16.418,
-15.309, -14.315, -13.503, -12.887, -16.962, -17.190, -17.410, -17.506, -17.252,
-16.434, -15.323, -14.328, -13.514, -12.896, -16.923, -17.163, -17.394, -17.500,
-17.256, -16.446, -15.336, -14.340, -13.525, -12.904, -16.688, -16.997, -17.287,
-17.442, -17.239, -16.453, -15.349, -14.353, -13.536, -12.913, -16.416, -16.789,
-17.141, -17.358, -17.210, -16.457, -15.361, -14.365, -13.546, -12.921, -16.253,
-16.656, -17.042, -17.297, -17.189, -16.462, -15.373, -14.377, -13.557, -12.930,
 -16.223, -16.625, -17.015, -17.280, -17.187, -16.472, -15.386, -14.390,
-13.567, -12.938, -16.263, -16.651, -17.032, -17.291, -17.197, -16.485, -15.400,
-14.402, -13.578, -12.947, -16.367, -16.733, -17.093, -17.332, -17.221, -16.501,
-15.414, -14.415, -13.588, -12.955, -16.497, -16.835, -17.169, -17.382, -17.250,
-16.519, -15.428, -14.427, -13.598, -12.963, -16.512, -16.853, -17.185, -17.397,
-17.263, -16.532, -15.442, -14.439, -13.608, -12.971, -16.455, -16.816, -17.164,
-17.388, -17.265, -16.544, -15.455, -14.451, -13.618, -12.979, -16.467, -16.824,
-17.170, -17.394, -17.274, -16.556, -15.468, -14.463, -13.628, -12.987, -16.395,
-16.766, -17.128, -17.370, -17.269, -16.565, -15.480, -14.475, -13.638, -12.995,
 -16.315, -16.702, -17.082, -17.343, -17.263, -16.574, -15.493, -14.487,
-13.648, -13.003, -16.293, -16.680, -17.063, -17.332, -17.264, -16.584, -15.505,
-14.499, -13.658, -13.010, -16.207, -16.610, -17.011, -17.301, -17.254, -16.591,
-15.517, -14.510, -13.667, -13.018, -16.198, -16.601, -17.004, -17.298, -17.258,
-16.602, -15.530, -14.522, -13.677, -13.025, -16.229, -16.621, -17.018, -17.308,
-17.269, -16.615, -15.543, -14.533, -13.686, -13.032, -16.267, -16.651, -17.040,
-17.325, -17.282, -16.628, -15.556, -14.545, -13.696, -13.040, -16.314, -16.686,
-17.066, -17.344, -17.297, -16.642, -15.569, -14.556, -13.705, -13.047, -16.335,
-16.701, -17.077, -17.354, -17.307, -16.654, -15.581, -14.567, -13.714, -13.054,
 -16.317, -16.687, -17.068, -17.350, -17.311, -16.665, -15.594, -14.578,
-13.723, -13.061, -16.311, -16.680, -17.062, -17.348, -17.316, -16.676, -15.606,
-14.590, -13.732, -13.068, -16.292, -16.663, -17.050, -17.342, -17.318, -16.686,
-15.618, -14.600, -13.741, -13.075, -16.275, -16.648, -17.038, -17.336, -17.321,
-16.696, -15.630, -14.611, -13.750, -13.082, -16.322, -16.681, -17.062, -17.353,
-17.334, -16.709, -15.642, -14.622, -13.759, -13.089, -16.381, -16.722, -17.091,
-17.374, -17.350, -16.723, -15.655, -14.633, -13.767, -13.096, -16.445, -16.767,
-17.123, -17.396, -17.366, -16.736, -15.667, -14.644, -13.776, -13.102, -16.515,
-16.815, -17.157, -17.420, -17.382, -16.750, -15.679, -14.654, -13.784, -13.109,
 -16.564, -16.848, -17.180, -17.436, -17.395, -16.763, -15.691, -14.665,
-13.793, -13.116, -16.556, -16.842, -17.176, -17.436, -17.401, -16.774, -15.703,
-14.675, -13.801, -13.122, -16.550, -16.837, -17.173, -17.437, -17.407, -16.784,
-15.715, -14.685, -13.809, -13.128, -16.545, -16.833, -17.172, -17.438, -17.413,
-16.795, -15.726, -14.695, -13.817, -13.134, -16.541, -16.831, -17.171, -17.440,
-17.419, -16.806, -15.737, -14.705, -13.825, -13.141, -16.539, -16.830, -17.171,
-17.443, -17.426, -16.817, -15.749, -14.715, -13.833, -13.147, -16.538, -16.829,
-17.172, -17.446, -17.432, -16.827, -15.760, -14.725, -13.841, -13.153, -16.538,
-16.830, -17.173, -17.449, -17.439, -16.838, -15.771, -14.735, -13.849, -13.159,
 -16.539, -16.831, -17.176, -17.453, -17.447, -16.849, -15.782, -14.744,
-13.856, -13.165, -16.541, -16.834, -17.179, -17.458, -17.454, -16.860, -15.793,
-14.754, -13.864, -13.171, -16.543, -16.837, -17.182, -17.463, -17.462, -16.870,
-15.804, -14.764, -13.872, -13.177, -16.547, -16.841, -17.187, -17.468, -17.470,
-16.881, -15.815, -14.773, -13.879, -13.182, -16.552, -16.846, -17.192, -17.474,
-17.478, -16.892, -15.826, -14.782, -13.886, -13.188, -16.557, -16.852, -17.197,
-17.480, -17.486, -16.903, -15.837, -14.792, -13.894, -13.194, -16.563, -16.858,
-17.204, -17.487, -17.494, -16.913, -15.848, -14.801, -13.901, -13.199, -16.570,
-16.865, -17.210, -17.494, -17.503, -16.924, -15.858, -14.810, -13.908, -13.205,
 -16.577, -16.872, -17.218, -17.502, -17.512, -16.935, -15.869, -14.819,
-13.915, -13.211, -16.585, -16.881, -17.225, -17.509, -17.521, -16.946, -15.880,
-14.828, -13.922, -13.216, -16.594, -16.889, -17.234, -17.517, -17.530, -16.957,
-15.890, -14.837, -13.929, -13.222, -16.603, -16.898, -17.242, -17.526, -17.539,
-16.967, -15.901, -14.846, -13.936, -13.227, -16.612, -16.908, -17.251, -17.534,
-17.548, -16.978, -15.911, -14.854, -13.943, -13.233, -16.622, -16.918, -17.261,
-17.543, -17.558, -16.989, -15.921, -14.863, -13.950, -13.238, -16.633, -16.929,
-17.271, -17.553, -17.567, -17.000, -15.932, -14.872, -13.957, -13.243, -16.644,
-16.940, -17.281, -17.562, -17.577, -17.010, -15.942, -14.880, -13.964, -13.249,
 -16.655, -16.951, -17.291, -17.572, -17.587, -17.021, -15.952, -14.889,
-13.971, -13.254, -16.666, -16.963, -17.302, -17.582, -17.597, -17.032, -15.962,
-14.898, -13.978, -13.259, -16.678, -16.975, -17.313, -17.592, -17.607, -17.042,
-15.970, -14.903, -13.981, -13.262, -16.691, -16.987, -17.325, -17.602, -17.617,
-17.052, -15.978, -14.909, -13.985, -13.264, -16.703, -17.000, -17.337, -17.613,
-17.627, -17.061, -15.985, -14.912, -13.986, -13.265, -16.716, -17.013, -17.349,
-17.623, -17.637, -17.071, -15.991, -14.915, -13.987, -13.265, -16.729, -17.026,
-17.361, -17.634, -17.647, -17.081, -15.999, -14.921, -13.991, -13.268, -16.742,
-17.040, -17.373, -17.645, -17.658, -17.091, -16.007, -14.927, -13.995, -13.271,
 -16.756, -17.053, -17.386, -17.656, -17.668, -17.102, -16.019, -14.937,
-14.004, -13.279, -16.770, -17.067, -17.399, -17.668, -17.679, -17.113, -16.029,
-14.946, -14.012, -13.285, -16.783, -17.081, -17.412, -17.679, -17.689, -17.125,
-16.040, -14.956, -14.020, -13.292, -16.798, -17.096, -17.425, -17.691, -17.700,
-17.136, -16.052, -14.968, -14.030, -13.300, -16.812, -17.110, -17.438, -17.703,
-17.711, -17.147, -16.061, -14.975, -14.035, -13.304, -16.826, -17.125, -17.452,
-17.715, -17.722, -17.157, -16.069, -14.981, -14.039, -13.307, -16.841, -17.140,
-17.465, -17.727, -17.732, -17.167, -16.076, -14.984, -14.042, -13.308, -16.856,
-17.155, -17.479, -17.739, -17.743, -17.176, -16.080, -14.985, -14.040, -13.307,
 -16.870, -17.170, -17.493, -17.751, -17.754, -17.185, -16.085, -14.986,
-14.040, -13.306, -16.885, -17.185, -17.507, -17.763, -17.765, -17.195, -16.093,
-14.993, -14.045, -13.310, -16.900, -17.201, -17.521, -17.775, -17.776, -17.207,
-16.105, -15.004, -14.055, -13.318, -16.915, -17.216, -17.536, -17.788, -17.787,
-17.220, -16.120, -15.019, -14.069, -13.330, -16.931, -17.232, -17.550, -17.800,
-17.798, -17.233, -16.136, -15.036, -14.085, -13.344, -16.946, -17.247, -17.565,
-17.813, -17.809, -17.246, -16.150, -15.051, -14.098, -13.354, -16.961, -17.263,
-17.579, -17.826, -17.820, -17.257, -16.161, -15.060, -14.105, -13.361, -16.977,
-17.279, -17.594, -17.838, -17.831, -17.266, -16.167, -15.065, -14.109, -13.363,
 -16.992, -17.295, -17.609, -17.851, -17.843, -17.274, -16.170, -15.064,
-14.106, -13.361, -17.008, -17.311, -17.623, -17.864, -17.854, -17.281, -16.172,
-15.062, -14.103, -13.358, -17.024, -17.327, -17.638, -17.877, -17.865, -17.290,
-16.178, -15.065, -14.104, -13.359, -17.039, -17.343, -17.653, -17.890, -17.876,
-17.300, -16.187, -15.072, -14.110, -13.364, -17.055, -17.359, -17.668, -17.903,
-17.888, -17.311, -16.199, -15.084, -14.121, -13.373, -17.071, -17.375, -17.683,
-17.916, -17.899, -17.324, -16.215, -15.102, -14.138, -13.387, -17.086, -17.391,
-17.698, -17.929, -17.910, -17.336, -16.231, -15.118, -14.152, -13.399, -17.102,
-17.408, -17.713, -17.942, -17.922, -17.347, -16.242, -15.129, -14.162, -13.407,
 -17.118, -17.424, -17.729, -17.955, -17.933, -17.358, -16.254, -15.141,
-14.173, -13.416, -17.134, -17.440, -17.744, -17.969, -17.944, -17.368, -16.262,
-15.147, -14.178, -13.420, -17.150, -17.457, -17.759, -17.982, -17.956, -17.377,
-16.268, -15.151, -14.181, -13.422, -17.166, -17.473, -17.774, -17.995, -17.967,
-17.387, -16.277, -15.159, -14.187, -13.428, -17.182, -17.489, -17.790, -18.009,
-17.979, -17.396, -16.285, -15.165, -14.192, -13.432, -17.198, -17.506, -17.805,
-18.022, -17.990, -17.407, -16.295, -15.174, -14.200, -13.438, -17.213, -17.522,
-17.820, -18.035, -18.001, -17.418, -16.307, -15.186, -14.210, -13.447, -17.229,
-17.538, -17.836, -18.049, -18.013, -17.430, -16.321, -15.201, -14.224, -13.458,
 -17.245, -17.555, -17.851, -18.062, -18.024, -17.443, -16.337, -15.217,
-14.239, -13.471, -17.261, -17.571, -17.867, -18.076, -18.036, -17.456, -16.352,
-15.233, -14.253, -13.482, -17.277, -17.588, -17.882, -18.089, -18.047, -17.468,
-16.365, -15.247, -14.265, -13.492, -17.293, -17.604, -17.897, -18.102, -18.059,
-17.480, -16.378, -15.259, -14.276, -13.501, -17.309, -17.620, -17.913, -18.116,
-18.070, -17.490, -16.388, -15.268, -14.284, -13.508, -17.325, -17.637, -17.928,
-18.129, -18.082, -17.501, -16.399, -15.278, -14.293, -13.515, -17.341, -17.653,
-17.944, -18.143, -18.093, -17.512, -16.409, -15.288, -14.301, -13.522, -17.356,
-17.669, -17.959, -18.157, -18.105, -17.523, -16.420, -15.298, -14.310, -13.529,
 -17.372, -17.686, -17.975, -18.170, -18.117, -17.534, -16.430, -15.308,
-14.319, -13.536, -17.388, -17.702, -17.990, -18.184, -18.128, -17.545, -16.441,
-15.318, -14.327, -13.544, -17.404, -17.718, -18.006, -18.197, -18.140, -17.556,
-16.452, -15.328, -14.336, -13.551, -17.420, -17.735, -18.021, -18.211, -18.151,
-17.567, -16.463, -15.338, -14.345, -13.558, -17.435, -17.751, -18.036, -18.224,
-18.163, -17.578, -16.474, -15.348, -14.354, -13.566, -17.451, -17.767, -18.052,
-18.238, -18.174, -17.589, -16.484, -15.359, -14.363, -13.573, -17.467, -17.783,
-18.067, -18.251, -18.186, -17.600, -16.496, -15.369, -14.372, -13.581, -17.483,
-17.799, -18.083, -18.265, -18.198, -17.611, -16.507, -15.380, -14.382, -13.588,
 -17.498, -17.816, -18.098, -18.279, -18.209, -17.622, -16.518, -15.391,
-14.391, -13.596, -17.514, -17.832, -18.114, -18.292, -18.221, -17.634, -16.529,
-15.401, -14.400, -13.604, -17.530, -17.848, -18.129, -18.306, -18.232, -17.645,
-16.540, -15.412, -14.410, -13.611, -17.545, -17.864, -18.144, -18.319, -18.244,
-17.656, -16.551, -15.423, -14.419, -13.619, -17.561, -17.880, -18.160, -18.333,
-18.256, -17.667, -16.562, -15.434, -14.429, -13.627, -17.576, -17.896, -18.175,
-18.346, -18.267, -17.678, -16.574, -15.445, -14.439, -13.635, -17.592, -17.912,
-18.190, -18.360, -18.279, -17.689, -16.585, -15.456, -14.449, -13.643, -17.607,
-17.928, -18.206, -18.374, -18.290, -17.701, -16.596, -15.467, -14.458, -13.651,
 -17.623, -17.944, -18.221, -18.387, -18.302, -17.712, -16.608, -15.478,
-14.468, -13.659, -17.638, -17.960, -18.236, -18.401, -18.314, -17.723, -16.619,
-15.489, -14.478, -13.668, -17.654, -17.976, -18.251, -18.414, -18.325, -17.734,
-16.631, -15.501, -14.488, -13.676, -17.669, -17.992, -18.266, -18.428, -18.337,
-17.745, -16.642, -15.512, -14.499, -13.684, -17.685, -18.007, -18.282, -18.441,
-18.348, -17.757, -16.653, -15.523, -14.509, -13.692, -17.700, -18.023, -18.297,
-18.455, -18.360, -17.768, -16.665, -15.535, -14.519, -13.701, -17.715, -18.039,
-18.312, -18.468, -18.372, -17.779, -16.676, -15.546, -14.529, -13.709, -17.731,
-18.055, -18.327, -18.482, -18.383, -17.790, -16.688, -15.558, -14.539, -13.718,
 -17.746, -18.070, -18.342, -18.495, -18.395, -17.801, -16.699, -15.569,
-14.550, -13.726, -17.761, -18.086, -18.357, -18.509, -18.406, -17.813, -16.711,
-15.581, -14.560, -13.735, -17.776, -18.102, -18.372, -18.522, -18.418, -17.824,
-16.722, -15.592, -14.571, -13.743, -17.791, -18.117, -18.387, -18.536, -18.430,
-17.835, -16.734, -15.604, -14.581, -13.752, -17.807, -18.133, -18.402, -18.549,
-18.441, -17.846, -16.745, -15.615, -14.591, -13.760, -17.822, -18.148, -18.417,
-18.563, -18.453, -17.857, -16.757, -15.627, -14.602, -13.769, -17.837, -18.164,
-18.432, -18.576, -18.464, -17.868, -16.768, -15.639, -14.612, -13.777, -17.852,
-18.179, -18.447, -18.590, -18.476, -17.880, -16.780, -15.650, -14.623, -13.786,
 -17.867, -18.195, -18.462, -18.603, -18.488, -17.891, -16.791, -15.662,
-14.634, -13.795, -17.882, -18.210, -18.477, -18.616, -18.499, -17.902, -16.802,
-15.674, -14.644, -13.803, -17.897, -18.225, -18.492, -18.630, -18.511, -17.913,
-16.814, -15.686, -14.655, -13.812, -17.912, -18.240, -18.507, -18.643, -18.522,
-17.924, -16.825, -15.697, -14.665, -13.821, -17.926, -18.256, -18.521, -18.656,
-18.534, -17.935, -16.837, -15.709, -14.676, -13.829, -17.941, -18.271, -18.536,
-18.670, -18.545, -17.946, -16.848, -15.721, -14.687, -13.838, -17.956, -18.286,
-18.551, -18.683, -18.557, -17.957, -16.859, -15.732, -14.697, -13.847, -17.971,
-18.301, -18.566, -18.696, -18.568, -17.968, -16.871, -15.744, -14.708, -13.855,
 -17.986, -18.316, -18.580, -18.710, -18.580, -17.979, -16.882, -15.756,
-14.719, -13.864, -18.000, -18.331, -18.595, -18.723, -18.592, -17.991, -16.894,
-15.768, -14.730, -13.873, -18.015, -18.347, -18.610, -18.736, -18.603, -18.002,
-16.905, -15.779, -14.740, -13.881, -18.030, -18.362, -18.624, -18.750, -18.615,
-18.013, -16.916, -15.791, -14.751, -13.890, -18.044, -18.376, -18.639, -18.763,
-18.626, -18.024, -16.928, -15.802, -14.761, -13.899, -18.059, -18.391, -18.653,
-18.776, -18.638, -18.035, -16.939, -15.814, -14.772, -13.907, -18.074, -18.406,
-18.668, -18.789, -18.649, -18.045, -16.950, -15.826, -14.783, -13.916, -18.088,
-18.421, -18.682, -18.803, -18.661, -18.056, -16.961, -15.837, -14.793, -13.924,
 -18.103, -18.436, -18.697, -18.816, -18.672, -18.067, -16.973, -15.849,
-14.804, -13.933, -18.117, -18.451, -18.711, -18.829, -18.684, -18.078, -16.984,
-15.861, -14.815, -13.941, -18.132, -18.466, -18.725, -18.842, -18.695, -18.089,
-16.995, -15.872, -14.825, -13.950, -18.146, -18.480, -18.740, -18.855, -18.707,
-18.100, -17.006, -15.884, -14.836, -13.958, -18.160, -18.495, -18.754, -18.868,
-18.718, -18.111, -17.017, -15.895, -14.846, -13.967, -18.175, -18.510, -18.768,
-18.881, -18.730, -18.122, -17.028, -15.907, -14.857, -13.975, -18.189, -18.524,
-18.783, -18.894, -18.741, -18.133, -17.040, -15.918, -14.867, -13.983, -18.203,
-18.539, -18.797, -18.908, -18.752, -18.144, -17.051, -15.930, -14.877, -13.992,
 -18.218, -18.553, -18.811, -18.921, -18.764, -18.154, -17.062, -15.941,
-14.888, -14.000, -18.232, -18.568, -18.825, -18.934, -18.775, -18.165, -17.073,
-15.953, -14.898, -14.008, -18.246, -18.582, -18.839, -18.947, -18.787, -18.176,
-17.084, -15.964, -14.909, -14.016, -18.260, -18.597, -18.854, -18.960, -18.798,
-18.187, -17.095, -15.976, -14.919, -14.024, -18.275, -18.611, -18.868, -18.973,
-18.809, -18.198, -17.106, -15.987, -14.929, -14.032, -18.289, -18.626, -18.882,
-18.986, -18.821, -18.208, -17.117, -15.998, -14.939, -14.040, -18.303, -18.640,
-18.896, -18.998, -18.832, -18.219, -17.128, -16.009, -14.949, -14.048, -18.317,
-18.654, -18.910, -19.011, -18.844, -18.230, -17.139, -16.021, -14.960, -14.056,
 -18.331, -18.669, -18.924, -19.024, -18.855, -18.241, -17.149, -16.032,
-14.970, -14.064, -18.345, -18.683, -18.938, -19.037, -18.866, -18.251, -17.160,
-16.043, -14.980, -14.071, -18.359, -18.697, -18.952, -19.050, -18.878, -18.262,
-17.171, -16.054, -14.990, -14.079, -18.373, -18.711, -18.965, -19.063, -18.889,
-18.273, -17.182, -16.065, -15.000, -14.087, -18.387, -18.725, -18.979, -19.076,
-18.900, -18.283, -17.193, -16.076, -15.010, -14.094, -18.401, -18.739, -18.993,
-19.088, -18.912, -18.294, -17.203, -16.087, -15.019, -14.101, -18.415, -18.754,
-19.007, -19.101, -18.923, -18.305, -17.214, -16.098, -15.029, -14.109, -18.428,
-18.768, -19.021, -19.114, -18.934, -18.315, -17.225, -16.109, -15.039, -14.116,
 -18.442, -18.782, -19.035, -19.127, -18.945, -18.326, -17.236, -16.120,
-15.049, -14.123, -18.456, -18.796, -19.048, -19.140, -18.957, -18.336, -17.246,
-16.131, -15.058, -14.130, -18.470, -18.810, -19.062, -19.152, -18.968, -18.347,
-17.257, -16.142, -15.068, -14.137, -18.484, -18.824, -19.076, -19.165, -18.979,
-18.357, -17.268, -16.153, -15.077, -14.144, -18.497, -18.837, -19.089, -19.178,
-18.990, -18.368, -17.278, -16.164, -15.087, -14.151, -18.511, -18.851, -19.103,
-19.190, -19.001, -18.378, -17.289, -16.174, -15.096, -14.158, -18.525, -18.865,
-19.116, -19.203, -19.013, -18.389, -17.299, -16.185, -15.105, -14.165, -18.538,
-18.879, -19.130, -19.215, -19.024, -18.399, -17.310, -16.196, -15.114, -14.171,
 -18.552, -18.893, -19.143, -19.228, -19.035, -18.410, -17.320, -16.206,
-15.123, -14.178, -18.565, -18.907, -19.157, -19.241, -19.046, -18.420, -17.331,
-16.217, -15.133, -14.184, -18.579, -18.920, -19.170, -19.253, -19.057, -18.431,
-17.341, -16.227, -15.142, -14.190, -18.592, -18.934, -19.184, -19.266, -19.068,
-18.441, -17.352, -16.238, -15.150, -14.196, -18.606, -18.948, -19.197, -19.278,
-19.079, -18.451, -17.362, -16.248, -15.159, -14.203, -18.619, -18.961, -19.211,
-19.291, -19.090, -18.462, -17.372, -16.259, -15.168, -14.209, -18.633, -18.975,
-19.224, -19.303, -19.102, -18.472, -17.383, -16.269, -15.177, -14.214, -18.646,
-18.988, -19.237, -19.316, -19.113, -18.482, -17.393, -16.279, -15.185, -14.220,
 -18.660, -19.002, -19.251, -19.328, -19.124, -18.493, -17.403, -16.289,
-15.194, -14.226, -18.673, -19.016, -19.264, -19.340, -19.135, -18.503, -17.414,
-16.300, -15.202, -14.232, -18.686, -19.029, -19.277, -19.353, -19.146, -18.513,
-17.424, -16.310, -15.210, -14.237, -18.700, -19.043, -19.290, -19.365, -19.157,
-18.523, -17.434, -16.320, -15.219, -14.242, -18.713, -19.056, -19.304, -19.377,
-19.168, -18.534, -17.444, -16.330, -15.227, -14.247, -18.726, -19.069, -19.317,
-19.390, -19.178, -18.544, -17.454, -16.339, -15.235, -14.253, -18.740, -19.083,
-19.330, -19.402, -19.189, -18.554, -17.465, -16.349, -15.243, -14.258, -18.753,
-19.096, -19.343, -19.414, -19.200, -18.564, -17.475, -16.359, -15.250, -14.262,
 -18.766, -19.109, -19.356, -19.427, -19.211, -18.574, -17.485, -16.369,
-15.258, -14.267, -18.779, -19.123, -19.369, -19.439, -19.222, -18.584, -17.495,
-16.379, -15.266, -14.272, -18.792, -19.136, -19.382, -19.451, -19.233, -18.594,
-17.505, -16.388, -15.273, -14.276, -18.805, -19.149, -19.395, -19.463, -19.244,
-18.604, -17.515, -16.398, -15.281, -14.281, -18.819, -19.163, -19.408, -19.475,
-19.254, -18.615, -17.525, -16.407, -15.288, -14.285, -18.832, -19.176, -19.421,
-19.487, -19.265, -18.625, -17.535, -16.417, -15.295, -14.289, -18.845, -19.189,
-19.434, -19.499, -19.276, -18.634, -17.545, -16.426, -15.302, -14.293, -18.858,
-19.202, -19.447, -19.512, -19.287, -18.644, -17.555, -16.435, -15.309, -14.297,
 -18.871, -19.215, -19.460, -19.524, -19.297, -18.654, -17.565, -16.445,
-15.316, -14.301, -18.884, -19.228, -19.473, -19.536, -19.308, -18.664, -17.574,
-16.454, -15.323, -14.305, -18.897, -19.241, -19.486, -19.548, -19.318, -18.674,
-17.584, -16.463, -15.330, -14.309, -18.910, -19.254, -19.498, -19.560, -19.329,
-18.684, -17.594, -16.472, -15.336, -14.312, -18.923, -19.267, -19.511, -19.571,
-19.340, -18.694, -17.604, -16.481, -15.343, -14.316, -18.935, -19.280, -19.524,
-19.583, -19.350, -18.703, -17.613, -16.490, -15.349, -14.319, -18.948, -19.293,
-19.537, -19.595, -19.361, -18.713, -17.623, -16.498, -15.355, -14.322, -18.961,
-19.306, -19.549, -19.607, -19.371, -18.723, -17.633, -16.507, -15.362, -14.325,
 -18.974, -19.319, -19.562, -19.619, -19.382, -18.733, -17.642, -16.516,
-15.368, -14.328, -18.987, -19.332, -19.575, -19.631, -19.392, -18.742, -17.652,
-16.524, -15.373, -14.331, -19.000, -19.345, -19.587, -19.643, -19.402, -18.752,
-17.662, -16.533, -15.379, -14.334, -19.012, -19.358, -19.600, -19.654, -19.413,
-18.761, -17.671, -16.541, -15.385, -14.336, -19.025, -19.371, -19.613, -19.666,
-19.423, -18.771, -17.681, -16.549, -15.391, -14.339, -19.038, -19.384, -19.625,
-19.678, -19.433, -18.780, -17.690, -16.558, -15.396, -14.341, -19.051, -19.396,
-19.638, -19.689, -19.444, -18.790, -17.700, -16.566, -15.401, -14.344, -19.063,
-19.409, -19.650, -19.701, -19.454, -18.799, -17.709, -16.574, -15.407, -14.346,
 -19.076, -19.422, -19.663, -19.713, -19.464, -18.809, -17.718, -16.582,
-15.412, -14.348, -19.089, -19.435, -19.675, -19.724, -19.474, -18.818, -17.728,
-16.590, -15.417, -14.350, -19.101, -19.447, -19.687, -19.736, -19.484, -18.827,
-17.737, -16.598, -15.422, -14.352, -19.114, -19.460, -19.700, -19.747, -19.494,
-18.837, -17.746, -16.606, -15.427, -14.354, -19.126, -19.473, -19.712, -19.759,
-19.504, -18.846, -17.756, -16.613, -15.431, -14.356, -19.139, -19.485, -19.725,
-19.770, -19.514, -18.855, -17.765, -16.621, -15.436, -14.358, -19.152, -19.498,
-19.737, -19.781, -19.524, -18.864, -17.774, -16.628, -15.441, -14.359, -19.164,
-19.510, -19.749, -19.793, -19.534, -18.873, -17.783, -16.636, -15.445, -14.361,
 -19.177, -19.523, -19.761, -19.804, -19.544, -18.882, -17.792, -16.643,
-15.449, -14.362, -19.189, -19.535, -19.774, -19.815, -19.554, -18.891, -17.801,
-16.651, -15.454, -14.364, -19.202, -19.548, -19.786, -19.827, -19.563, -18.900,
-17.810, -16.658, -15.458, -14.365, -19.214, -19.560, -19.798, -19.838, -19.573,
-18.909, -17.819, -16.665, -15.462, -14.367, -19.226, -19.573, -19.810, -19.849,
-19.583, -18.918, -17.828, -16.672, -15.466, -14.368, -19.239, -19.585, -19.822,
-19.860, -19.592, -18.927, -17.837, -16.679, -15.470, -14.369, -19.251, -19.598,
-19.834, -19.871, -19.602, -18.936, -17.846, -16.686, -15.474, -14.370, -19.264,
-19.610, -19.846, -19.882, -19.611, -18.944, -17.855, -16.693, -15.478, -14.372,
 -19.276, -19.623, -19.858, -19.893, -19.621, -18.953, -17.864, -16.700,
-15.481, -14.373, -19.288, -19.635, -19.870, -19.904, -19.630, -18.961, -17.872,
-16.706, -15.485, -14.374, -19.301, -19.647, -19.882, -19.915, -19.639, -18.970,
-17.881, -16.713, -15.489, -14.375, -19.313, -19.660, -19.894, -19.926, -19.648,
-18.978, -17.890, -16.720, -15.492, -14.376, -19.325, -19.672, -19.906, -19.937,
-19.657, -18.987, -17.899, -16.726, -15.496, -14.377, -19.337, -19.684, -19.918,
-19.947, -19.667, -18.995, -17.907, -16.733, -15.499, -14.378, -19.350, -19.696,
-19.930, -19.958, -19.676, -19.003, -17.916, -16.739, -15.503, -14.379, -19.362,
-19.709, -19.942, -19.969, -19.684, -19.012, -17.924, -16.745, -15.506, -14.380,
 -19.374, -19.721, -19.954, -19.979, -19.693, -19.020, -17.933, -16.752,
-15.510, -14.381, -19.386, -19.733, -19.965, -19.990, -19.702, -19.028, -17.941,
-16.758, -15.513, -14.382, -19.399, -19.745, -19.977, -20.000, -19.711, -19.036,
-17.949, -16.764, -15.516, -14.383, -19.411, -19.757, -19.989, -20.011, -19.719,
-19.044, -17.958, -16.771, -15.520, -14.385, -19.423, -19.769, -20.000, -20.021,
-19.728, -19.051, -17.966, -16.777, -15.523, -14.386, -19.435, -19.782, -20.012,
-20.031, -19.736, -19.059, -17.974, -16.783, -15.526, -14.387, -19.447, -19.794,
-20.024, -20.041, -19.744, -19.067, -17.982, -16.789, -15.530, -14.388, -19.459,
-19.806, -20.035, -20.052, -19.753, -19.074, -17.991, -16.793, -15.529, -14.385,
 -19.471, -19.818, -20.047, -20.062, -19.761, -19.082, -17.999, -16.797,
-15.529, -14.383, -19.483, -19.830, -20.058, -20.072, -19.769, -19.089, -18.007,
-16.799, -15.525, -14.377, -19.495, -19.842, -20.069, -20.081, -19.777, -19.097,
-18.015, -16.800, -15.522, -14.370, -19.507, -19.854, -20.081, -20.091, -19.785,
-19.104, -18.023, -16.803, -15.521, -14.367, -19.519, -19.866, -20.092, -20.101,
-19.792, -19.111, -18.030, -16.807, -15.521, -14.365, -19.531, -19.877, -20.104,
-20.111, -19.800, -19.118, -18.038, -16.814, -15.526, -14.368, -19.543, -19.889,
-20.115, -20.120, -19.807, -19.125, -18.046, -16.821, -15.531, -14.372, -19.555,
-19.901, -20.126, -20.130, -19.815, -19.132, -18.054, -16.828, -15.536, -14.375,
 -19.567, -19.913, -20.137, -20.139, -19.822, -19.139, -18.061, -16.837,
-15.544, -14.381, -19.579, -19.925, -20.148, -20.148, -19.829, -19.145, -18.069,
-16.843, -15.548, -14.383, -19.591, -19.937, -20.159, -20.158, -19.836, -19.152,
-18.076, -16.847, -15.548, -14.382, -19.603, -19.948, -20.170, -20.167, -19.843,
-19.158, -18.084, -16.850, -15.547, -14.379, -19.615, -19.960, -20.181, -20.176,
-19.850, -19.165, -18.091, -16.849, -15.540, -14.370, -19.626, -19.972, -20.192,
-20.185, -19.856, -19.171, -18.099, -16.848, -15.534, -14.362, -19.638, -19.984,
-20.203, -20.193, -19.863, -19.177, -18.106, -16.852, -15.535, -14.361, -19.650,
-19.995, -20.214, -20.202, -19.869, -19.183, -18.113, -16.859, -15.541, -14.366,
 -19.662, -20.007, -20.224, -20.210, -19.875, -19.189, -18.120, -16.870,
-15.553, -14.376, -19.674, -20.019, -20.235, -20.219, -19.881, -19.195, -18.127,
-16.884, -15.569, -14.391, -19.685, -20.030, -20.246, -20.227, -19.887, -19.200,
-18.134, -16.897, -15.582, -14.403, -19.697, -20.042, -20.256, -20.235, -19.893,
-19.206, -18.141, -16.905, -15.590, -14.409, -19.709, -20.053, -20.267, -20.243,
-19.898, -19.211, -18.148, -16.908, -15.590, -14.408, -19.721, -20.065, -20.277,
-20.251, -19.903, -19.217, -18.155, -16.908, -15.584, -14.401, -19.732, -20.076,
-20.287, -20.259, -19.909, -19.222, -18.162, -16.905, -15.576, -14.391, -19.744,
-20.088, -20.298, -20.266, -19.913, -19.227, -18.168, -16.905, -15.572, -14.385,
 -19.756, -20.099, -20.308, -20.274, -19.918, -19.231, -18.175, -16.909,
-15.573, -14.385, -19.767, -20.111, -20.318, -20.281, -19.923, -19.236, -18.181,
-16.917, -15.581, -14.391, -19.779, -20.122, -20.328, -20.288, -19.927, -19.241,
-18.188, -16.931, -15.596, -14.405, -19.791, -20.133, -20.338, -20.295, -19.931,
-19.245, -18.194, -16.944, -15.612, -14.420, -19.802, -20.145, -20.348, -20.301,
-19.935, -19.249, -18.200, -16.953, -15.621, -14.428, -19.814, -20.156, -20.357,
-20.308, -19.939, -19.253, -18.207, -16.962, -15.630, -14.435, -19.825, -20.167,
-20.367, -20.314, -19.943, -19.257, -18.213, -16.967, -15.634, -14.438, -19.837,
-20.178, -20.376, -20.320, -19.946, -19.261, -18.219, -16.970, -15.634, -14.437,
 -19.848, -20.189, -20.386, -20.326, -19.949, -19.265, -18.225, -16.974,
-15.637, -14.438, -19.860, -20.200, -20.395, -20.332, -19.952, -19.268, -18.230,
-16.979, -15.640, -14.440, -19.871, -20.212, -20.404, -20.338, -19.955, -19.271,
-18.236, -16.984, -15.643, -14.442, -19.883, -20.223, -20.413, -20.343, -19.957,
-19.274, -18.242, -16.993, -15.653, -14.450, -19.894, -20.234, -20.422, -20.348,
-19.959, -19.277, -18.247, -17.003, -15.664, -14.460, -19.906, -20.244, -20.431,
-20.353, -19.961, -19.280, -18.253, -17.016, -15.679, -14.474, -19.917, -20.255,
-20.440, -20.357, -19.963, -19.282, -18.258, -17.029, -15.694, -14.487, -19.929,
-20.266, -20.448, -20.361, -19.964, -19.285, -18.264, -17.039, -15.705, -14.497,
 -19.940, -20.277, -20.457, -20.366, -19.966, -19.287, -18.269, -17.048,
-15.715, -14.507, -19.951, -20.288, -20.465, -20.369, -19.966, -19.289, -18.274,
-17.055, -15.722, -14.512, -19.963, -20.298, -20.473, -20.373, -19.967, -19.291,
-18.279, -17.062, -15.728, -14.517, -19.974, -20.309, -20.481, -20.376, -19.967,
-19.292, -18.284, -17.069, -15.735, -14.523, -19.985, -20.320, -20.489, -20.379,
-19.968, -19.294, -18.288, -17.076, -15.742, -14.528, -19.997, -20.330, -20.496,
-20.382, -19.967, -19.295, -18.293, -17.083, -15.749, -14.534, -20.008, -20.341,
-20.504, -20.384, -19.967, -19.296, -18.298, -17.090, -15.756, -14.540, -20.019,
-20.351, -20.511, -20.386, -19.966, -19.297, -18.302, -17.097, -15.763, -14.546,
 -20.030, -20.361, -20.518, -20.388, -19.965, -19.297, -18.307, -17.104,
-15.770, -14.552, -20.042, -20.372, -20.524, -20.389, -19.964, -19.298, -18.311,
-17.111, -15.777, -14.558, -20.053, -20.382, -20.531, -20.390, -19.962, -19.298,
-18.315, -17.118, -15.785, -14.564, -20.064, -20.392, -20.537, -20.391, -19.960,
-19.298, -18.319, -17.126, -15.792, -14.571, -20.075, -20.402, -20.543, -20.391,
-19.958, -19.297, -18.323, -17.133, -15.800, -14.578, -20.086, -20.412, -20.549,
-20.392, -19.955, -19.297, -18.327, -17.140, -15.808, -14.584, -20.097, -20.422,
-20.555, -20.391, -19.952, -19.296, -18.330, -17.147, -15.816, -14.591, -20.108,
-20.431, -20.560, -20.390, -19.949, -19.295, -18.334, -17.155, -15.824, -14.598,
 -20.120, -20.441, -20.565, -20.389, -19.945, -19.294, -18.337, -17.162,
-15.832, -14.605, -20.131, -20.450, -20.570, -20.388, -19.941, -19.293, -18.341,
-17.169, -15.840, -14.612, -20.142, -20.460, -20.575, -20.386, -19.937, -19.291,
-18.344, -17.177, -15.848, -14.620, -20.152, -20.469, -20.579, -20.384, -19.932,
-19.289, -18.347, -17.184, -15.856, -14.627, -20.163, -20.478, -20.583, -20.381,
-19.927, -19.287, -18.350, -17.191, -15.865, -14.635, -20.174, -20.487, -20.586,
-20.378, -19.922, -19.285, -18.353, -17.199, -15.873, -14.642, -20.185, -20.496,
-20.590, -20.375, -19.917, -19.282, -18.356, -17.206, -15.882, -14.650, -20.196,
-20.505, -20.593, -20.371, -19.911, -19.279, -18.358, -17.214, -15.890, -14.658,
 -20.207, -20.514, -20.595, -20.366, -19.904, -19.276, -18.361, -17.221,
-15.899, -14.666, -20.218, -20.522, -20.597, -20.362, -19.898, -19.273, -18.363,
-17.228, -15.908, -14.674, -20.228, -20.531, -20.599, -20.357, -19.891, -19.270,
-18.365, -17.235, -15.917, -14.682, -20.239, -20.539, -20.601, -20.351, -19.884,
-19.266, -18.367, -17.243, -15.926, -14.690, -20.250, -20.547, -20.602, -20.345,
-19.876, -19.262, -18.369, -17.250, -15.935, -14.698, -20.260, -20.555, -20.602,
-20.338, -19.868, -19.258, -18.371, -17.257, -15.944, -14.707, -20.271, -20.562,
-20.603, -20.331, -19.860, -19.253, -18.373, -17.264, -15.953, -14.715, -20.281,
-20.570, -20.602, -20.324, -19.851, -19.248, -18.374, -17.271, -15.962, -14.724,
 -20.292, -20.577, -20.602, -20.316, -19.842, -19.244, -18.376, -17.279,
-15.971, -14.732, -20.302, -20.584, -20.601, -20.308, -19.833, -19.238, -18.377,
-17.286, -15.980, -14.741, -20.312, -20.591, -20.599, -20.299, -19.823, -19.233,
-18.378, -17.293, -15.989, -14.750, -20.323, -20.598, -20.597, -20.289, -19.813,
-19.227, -18.379, -17.299, -15.998, -14.759, -20.333, -20.604, -20.594, -20.280,
-19.803, -19.221, -18.380, -17.306, -16.008, -14.767, -20.343, -20.610, -20.591,
-20.270, -19.792, -19.215, -18.380, -17.313, -16.017, -14.776, -20.353, -20.616,
-20.588, -20.259, -19.781, -19.209, -18.381, -17.320, -16.026, -14.785, -20.363,
-20.621, -20.584, -20.248, -19.770, -19.202, -18.381, -17.327, -16.036, -14.794,
 -20.373, -20.626, -20.579, -20.236, -19.758, -19.196, -18.382, -17.334,
-16.045, -14.803, -20.383, -20.631, -20.574, -20.224, -19.746, -19.189, -18.382,
-17.340, -16.055, -14.813, -20.393, -20.636, -20.568, -20.211, -19.734, -19.181,
-18.382, -17.347, -16.064, -14.822, -20.403, -20.640, -20.562, -20.198, -19.721,
-19.174, -18.382, -17.353, -16.073, -14.831, -20.412, -20.644, -20.555, -20.185,
-19.709, -19.166, -18.381, -17.360, -16.083, -14.840, -20.422, -20.648, -20.548,
-20.171, -19.695, -19.158, -18.381, -17.366, -16.092, -14.850, -20.431, -20.651,
-20.540, -20.157, -19.682, -19.150, -18.380, -17.372, -16.102, -14.859, -20.441,
-20.653, -20.531, -20.142, -19.668, -19.141, -18.380, -17.379, -16.111, -14.868,
 -20.450, -20.656, -20.522, -20.126, -19.654, -19.133, -18.379, -17.385,
-16.120, -14.878, -20.459, -20.658, -20.512, -20.111, -19.640, -19.124, -18.378,
-17.391, -16.130, -14.887, -20.468, -20.659, -20.501, -20.095, -19.625, -19.115,
-18.376, -17.397, -16.139, -14.897, -20.477, -20.660, -20.490, -20.078, -19.610,
-19.106, -18.375, -17.402, -16.148, -14.906, -20.485, -20.661, -20.479, -20.061,
-19.595, -19.096, -18.374, -17.408, -16.158, -14.916, -20.494, -20.661, -20.467,
-20.043, -19.579, -19.086, -18.372, -17.414, -16.167, -14.925, -20.502, -20.660,
-20.454, -20.026, -19.563, -19.077, -18.370, -17.420, -16.177, -14.935, -20.510,
-20.659, -20.440, -20.007, -19.547, -19.066, -18.368, -17.425, -16.186, -14.944,
 -20.518, -20.657, -20.426, -19.989, -19.531, -19.056, -18.366, -17.431,
-16.195, -14.954, -20.526, -20.655, -20.411, -19.970, -19.514, -19.046, -18.364,
-17.436, -16.205, -14.964, -20.534, -20.652, -20.396, -19.950, -19.498, -19.035,
-18.362, -17.441, -16.214, -14.973, -20.541, -20.649, -20.380, -19.931, -19.481,
-19.024, -18.359, -17.446, -16.223, -14.983, -20.549, -20.645, -20.363, -19.910,
-19.463, -19.013, -18.356, -17.452, -16.232, -14.992, -20.556, -20.641, -20.346,
-19.890, -19.446, -19.002, -18.354, -17.457, -16.241, -15.002, -20.562, -20.635,
-20.329, -19.869, -19.428, -18.990, -18.351, -17.461, -16.251, -15.012, -20.569,
-20.629, -20.310, -19.848, -19.410, -18.979, -18.348, -17.466, -16.260, -15.022,
 -20.575, -20.623, -20.291, -19.826, -19.392, -18.967, -18.344, -17.471,
-16.269, -15.031, -20.581, -20.615, -20.272, -19.804, -19.374, -18.955, -18.341,
-17.476, -16.278, -15.041, -20.586, -20.607, -20.252, -19.782, -19.355, -18.943,
-18.337, -17.480, -16.287, -15.050, -20.592, -20.599, -20.231, -19.760, -19.336,
-18.930, -18.334, -17.484, -16.296, -15.060, -20.597, -20.589, -20.210, -19.737,
-19.318, -18.918, -18.330, -17.489, -16.305, -15.070, -20.601, -20.579, -20.188,
-19.714, -19.298, -18.905, -18.326, -17.493, -16.314, -15.079, -20.606, -20.568,
-20.166, -19.691, -19.279, -18.892, -18.322, -17.497, -16.323, -15.089, -20.609,
-20.557, -20.143, -19.667, -19.260, -18.879, -18.318, -17.501, -16.331, -15.099,
 -20.613, -20.544, -20.119, -19.643, -19.240, -18.866, -18.313, -17.505,
-16.340, -15.108, -20.616, -20.531, -20.096, -19.619, -19.220, -18.853, -18.309,
-17.509, -16.349, -15.118, -20.618, -20.517, -20.071, -19.595, -19.200, -18.840,
-18.304, -17.512, -16.357, -15.127, -20.621, -20.502, -20.047, -19.570, -19.180,
-18.826, -18.299, -17.516, -16.366, -15.137, -20.622, -20.486, -20.021, -19.546,
-19.160, -18.812, -18.294, -17.520, -16.375, -15.147, -20.623, -20.470, -19.996,
-19.521, -19.140, -18.799, -18.289, -17.523, -16.383, -15.156, -20.624, -20.453,
-19.970, -19.496, -19.120, -18.785, -18.284, -17.526, -16.392, -15.166, -20.624,
-20.435, -19.943, -19.470, -19.099, -18.771, -18.279, -17.529, -16.400, -15.175,
 -20.623, -20.416, -19.916, -19.445, -19.078, -18.757, -18.273, -17.532,
-16.409, -15.185, -20.622, -20.397, -19.889, -19.419, -19.058, -18.742, -18.268,
-17.535, -16.417, -15.194, -20.620, -20.377, -19.862, -19.394, -19.037, -18.728,
-18.262, -17.538, -16.425, -15.204, -20.618, -20.356, -19.834, -19.368, -19.016,
-18.713, -18.256, -17.541, -16.434, -15.213, -20.615, -20.334, -19.806, -19.342,
-18.995, -18.699, -18.250, -17.544, -16.442, -15.223, -20.611, -20.312, -19.777,
-19.315, -18.974, -18.684, -18.244, -17.546, -16.450, -15.232, -20.606, -20.289,
-19.748, -19.289, -18.952, -18.669, -18.238, -17.549, -16.458, -15.242, -20.601,
-20.265, -19.719, -19.263, -18.931, -18.654, -18.232, -17.551, -16.466, -15.251,
 -20.595, -20.241, -19.690, -19.236, -18.910, -18.639, -18.225, -17.553,
-16.474, -15.260, -20.589, -20.216, -19.660, -19.210, -18.889, -18.624, -18.219,
-17.555, -16.482, -15.270, -20.581, -20.190, -19.631, -19.183, -18.867, -18.609,
-18.212, -17.557, -16.490, -15.279, -20.573, -20.164, -19.601, -19.157, -18.846,
-18.594, -18.205, -17.559, -16.498, -15.288, -20.564, -20.137, -19.570, -19.130,
-18.824, -18.579, -18.198, -17.561, -16.506, -15.298, -20.554, -20.110, -19.540,
-19.104, -18.803, -18.563, -18.191, -17.563, -16.514, -15.307, -20.543, -20.082,
-19.510, -19.077, -18.782, -18.548, -18.184, -17.565, -16.522, -15.316, -20.531,
-20.053, -19.479, -19.050, -18.760, -18.532, -18.177, -17.566, -16.530, -15.325,
 -20.519, -20.024, -19.449, -19.024, -18.739, -18.517, -18.169, -17.568,
-16.537, -15.335, -20.506, -19.995, -19.418, -18.997, -18.717, -18.501, -18.162,
-17.569, -16.545, -15.344, -20.492, -19.965, -19.387, -18.970, -18.696, -18.486,
-18.154, -17.570, -16.552, -15.353, -20.477, -19.935, -19.357, -18.944, -18.674,
-18.470, -18.146, -17.571, -16.560, -15.362, -20.461, -19.905, -19.326, -18.917,
-18.653, -18.454, -18.139, -17.573, -16.567, -15.371, -20.444, -19.874, -19.295,
-18.891, -18.632, -18.439, -18.131, -17.573, -16.575, -15.380, -20.427, -19.843,
-19.264, -18.865, -18.611, -18.423, -18.123, -17.574, -16.582, -15.389, -20.409,
-19.812, -19.234, -18.839, -18.590, -18.408, -18.115, -17.575, -16.590, -15.398,
 -20.390, -19.780, -19.203, -18.813, -18.569, -18.392, -18.107, -17.576,
-16.597, -15.407, -20.370, -19.748, -19.173, -18.787, -18.548, -18.376, -18.099,
-17.577, -16.604, -15.416, -20.350, -19.717, -19.142, -18.761, -18.527, -18.361,
-18.091, -17.577, -16.612, -15.425, -20.328, -19.685, -19.112, -18.736, -18.506,
-18.345, -18.082, -17.578, -16.619, -15.434, -20.306, -19.653, -19.082, -18.710,
-18.486, -18.330, -18.074, -17.578, -16.626, -15.443, -20.284, -19.621, -19.053,
-18.685, -18.466, -18.314, -18.066, -17.579, -16.633, -15.452, -20.261, -19.589,
-19.023, -18.661, -18.445, -18.299, -18.057, -17.579, -16.640, -15.461, -20.237,
-19.557, -18.994, -18.636, -18.425, -18.284, -18.049, -17.579, -16.648, -15.470,
 -20.213, -19.525, -18.965, -18.612, -18.406, -18.269, -18.041, -17.580,
-16.655, -15.479, -20.189, -19.493, -18.936, -18.588, -18.386, -18.254, -18.032,
-17.580, -16.662, -15.487, -20.163, -19.462, -18.908, -18.564, -18.367, -18.239,
-18.024, -17.580, -16.669, -15.496, -20.138, -19.431, -18.880, -18.541, -18.348,
-18.224, -18.015, -17.580, -16.676, -15.505, -20.112, -19.400, -18.853, -18.518,
-18.329, -18.209, -18.007, -17.580, -16.683, -15.514, -20.087, -19.369, -18.826,
-18.496, -18.311, -18.195, -17.999, -17.580, -16.690, -15.523, -20.060, -19.339,
-18.799, -18.473, -18.293, -18.181, -17.990, -17.580, -16.697, -15.531, -20.034,
-19.310, -18.773, -18.452, -18.275, -18.166, -17.982, -17.580, -16.704, -15.540,
 -20.008, -19.280, -18.748, -18.431, -18.257, -18.153, -17.974, -17.581,
-16.711, -15.549, -19.982, -19.252, -18.723, -18.410, -18.240, -18.139, -17.966,
-17.581, -16.718, -15.558, -19.956, -19.224, -18.699, -18.390, -18.224, -18.126,
-17.958, -17.581, -16.725, -15.566, -19.930, -19.196, -18.675, -18.371, -18.207,
-18.113, -17.950, -17.581, -16.732, -15.575, -19.905, -19.170, -18.653, -18.352,
-18.192, -18.100, -17.943, -17.581, -16.739, -15.584, -19.880, -19.144, -18.631,
-18.334, -18.177, -18.087, -17.935, -17.581, -16.746, -15.592, -19.855, -19.119,
-18.609, -18.316, -18.162, -18.075, -17.928, -17.582, -16.753, -15.601, -19.831,
-19.095, -18.589, -18.299, -18.148, -18.064, -17.921, -17.582, -16.760, -15.610,
 -19.807, -19.071, -18.570, -18.283, -18.134, -18.052, -17.914, -17.582,
-16.767, -15.618, -19.784, -19.049, -18.551, -18.268, -18.121, -18.041, -17.907,
-17.583, -16.774, -15.627, -19.763, -19.028, -18.534, -18.253, -18.108, -18.031,
-17.901, -17.583, -16.781, -15.635, -19.742, -19.008, -18.517, -18.240, -18.096,
-18.021, -17.894, -17.584, -16.788, -15.644, -19.652, -18.933, -18.459, -18.195,
-18.062, -17.994, -17.877, -17.579, -16.794, -15.652, -19.566, -18.860, -18.403,
-18.151, -18.028, -17.968, -17.861, -17.575, -16.799, -15.661, -19.483, -18.791,
-18.349, -18.110, -17.995, -17.943, -17.845, -17.570, -16.805, -15.669, -19.402,
-18.724, -18.297, -18.069, -17.964, -17.919, -17.829, -17.566, -16.810, -15.677,
 -19.351, -18.682, -18.264, -18.044, -17.944, -17.903, -17.819, -17.564,
-16.816, -15.686, -19.335, -18.671, -18.256, -18.037, -17.938, -17.897, -17.815,
-17.566, -16.824, -15.694, -19.321, -18.660, -18.249, -18.032, -17.932, -17.892,
-17.812, -17.567, -16.831, -15.703, -19.285, -18.631, -18.227, -18.014, -17.918,
-17.880, -17.804, -17.567, -16.838, -15.711, -19.272, -18.620, -18.218, -18.006,
-17.910, -17.872, -17.799, -17.568, -16.844, -15.720, -19.209, -18.568, -18.176,
-17.973, -17.884, -17.852, -17.785, -17.564, -16.850, -15.728, -19.135, -18.506,
-18.127, -17.935, -17.853, -17.828, -17.770, -17.558, -16.854, -15.735, -19.071,
-18.454, -18.086, -17.903, -17.828, -17.808, -17.756, -17.555, -16.860, -15.743,
 -19.017, -18.412, -18.055, -17.879, -17.810, -17.793, -17.747, -17.553,
-16.865, -15.750, -19.020, -18.418, -18.061, -17.884, -17.813, -17.794, -17.747,
-17.556, -16.872, -15.757, -19.043, -18.441, -18.081, -17.900, -17.824, -17.802,
-17.753, -17.562, -16.880, -15.765, -19.038, -18.442, -18.085, -17.904, -17.827,
-17.803, -17.753, -17.566, -16.887, -15.773, -19.099, -18.494, -18.127, -17.936,
-17.849, -17.818, -17.763, -17.574, -16.896, -15.782, -19.111, -18.507, -18.139,
-17.945, -17.855, -17.820, -17.765, -17.578, -16.904, -15.790, -19.071, -18.474,
-18.111, -17.923, -17.837, -17.806, -17.755, -17.576, -16.909, -15.799, -19.090,
-18.481, -18.112, -17.920, -17.832, -17.800, -17.751, -17.576, -16.916, -15.807,
 -18.998, -18.401, -18.046, -17.867, -17.791, -17.769, -17.730, -17.567,
-16.919, -15.815, -18.834, -18.263, -17.935, -17.781, -17.725, -17.720, -17.697,
-17.551, -16.919, -15.821, -18.728, -18.176, -17.867, -17.728, -17.685, -17.690,
-17.677, -17.542, -16.921, -15.827, -18.661, -18.128, -17.834, -17.705, -17.668,
-17.678, -17.670, -17.541, -16.926, -15.833, -18.682, -18.155, -17.860, -17.729,
-17.688, -17.693, -17.681, -17.550, -16.934, -15.840, -18.860, -18.308, -17.985,
-17.825, -17.760, -17.745, -17.716, -17.573, -16.947, -15.849, -19.099, -18.501,
-18.132, -17.934, -17.838, -17.800, -17.753, -17.596, -16.961, -15.859, -19.231,
-18.599, -18.202, -17.982, -17.871, -17.822, -17.768, -17.606, -16.970, -15.869,
 -19.173, -18.554, -18.169, -17.959, -17.854, -17.810, -17.760, -17.605,
-16.976, -15.879, -18.939, -18.372, -18.033, -17.860, -17.784, -17.761, -17.729,
-17.589, -16.977, -15.887, -18.682, -18.159, -17.866, -17.733, -17.690, -17.694,
-17.684, -17.566, -16.974, -15.893, -18.541, -18.036, -17.764, -17.651, -17.627,
-17.647, -17.653, -17.549, -16.973, -15.899, -18.521, -18.011, -17.739, -17.629,
-17.608, -17.633, -17.643, -17.545, -16.976, -15.904, -18.566, -18.043, -17.761,
-17.644, -17.619, -17.640, -17.649, -17.550, -16.982, -15.910, -18.673, -18.131,
-17.831, -17.699, -17.660, -17.671, -17.671, -17.565, -16.991, -15.917, -18.792,
-18.230, -17.910, -17.760, -17.707, -17.706, -17.696, -17.582, -17.001, -15.925,
 -18.795, -18.239, -17.921, -17.771, -17.717, -17.715, -17.703, -17.588,
-17.009, -15.933, -18.754, -18.211, -17.903, -17.759, -17.710, -17.712, -17.703,
-17.590, -17.015, -15.943, -18.753, -18.208, -17.900, -17.757, -17.709, -17.712,
-17.704, -17.592, -17.021, -15.952, -18.679, -18.145, -17.849, -17.718, -17.680,
-17.692, -17.691, -17.587, -17.024, -15.960, -18.610, -18.086, -17.801, -17.680,
-17.652, -17.672, -17.679, -17.581, -17.026, -15.967, -18.590, -18.064, -17.780,
-17.663, -17.639, -17.664, -17.674, -17.579, -17.030, -15.974, -18.504, -17.991,
-17.720, -17.616, -17.604, -17.639, -17.658, -17.571, -17.031, -15.979, -18.506,
-17.990, -17.718, -17.614, -17.603, -17.639, -17.659, -17.573, -17.035, -15.986,
 -18.531, -18.006, -17.729, -17.621, -17.609, -17.644, -17.664, -17.577,
-17.040, -15.993, -18.573, -18.039, -17.755, -17.642, -17.625, -17.658, -17.674,
-17.585, -17.047, -16.000, -18.619, -18.075, -17.782, -17.663, -17.643, -17.672,
-17.686, -17.592, -17.054, -16.008, -18.636, -18.087, -17.792, -17.671, -17.649,
-17.678, -17.691, -17.597, -17.060, -16.017, -18.620, -18.073, -17.781, -17.663,
-17.644, -17.676, -17.691, -17.598, -17.065, -16.025, -18.611, -18.063, -17.772,
-17.656, -17.640, -17.674, -17.691, -17.599, -17.069, -16.034, -18.593, -18.046,
-17.757, -17.644, -17.632, -17.669, -17.689, -17.599, -17.073, -16.042, -18.577,
-18.031, -17.744, -17.634, -17.625, -17.665, -17.687, -17.599, -17.077, -16.049,
 -18.630, -18.070, -17.773, -17.656, -17.642, -17.679, -17.698, -17.606,
-17.083, -16.057, -18.688, -18.112, -17.804, -17.680, -17.661, -17.694, -17.710,
-17.614, -17.090, -16.064, -18.751, -18.158, -17.838, -17.705, -17.681, -17.710,
-17.722, -17.622, -17.096, -16.072, -18.820, -18.207, -17.874, -17.732, -17.702,
-17.727, -17.734, -17.630, -17.102, -16.079, -18.861, -18.235, -17.894, -17.748,
-17.714, -17.738, -17.743, -17.636, -17.108, -16.086, -18.854, -18.229, -17.889,
-17.744, -17.713, -17.738, -17.744, -17.637, -17.112, -16.093, -18.848, -18.224,
-17.885, -17.742, -17.712, -17.739, -17.746, -17.639, -17.116, -16.100, -18.844,
-18.221, -17.883, -17.741, -17.712, -17.741, -17.748, -17.642, -17.120, -16.107,
 -18.841, -18.218, -17.881, -17.740, -17.713, -17.743, -17.751, -17.644,
-17.125, -16.114, -18.840, -18.217, -17.880, -17.740, -17.714, -17.745, -17.754,
-17.647, -17.129, -16.121, -18.839, -18.216, -17.880, -17.741, -17.716, -17.748,
-17.757, -17.649, -17.133, -16.128, -18.840, -18.217, -17.881, -17.743, -17.718,
-17.751, -17.760, -17.652, -17.138, -16.135, -18.841, -18.219, -17.883, -17.745,
-17.721, -17.754, -17.764, -17.655, -17.142, -16.142, -18.844, -18.221, -17.886,
-17.748, -17.725, -17.758, -17.768, -17.658, -17.147, -16.149, -18.847, -18.225,
-17.889, -17.751, -17.728, -17.762, -17.772, -17.662, -17.151, -16.156, -18.851,
-18.229, -17.893, -17.755, -17.733, -17.767, -17.776, -17.665, -17.156, -16.163,
 -18.856, -18.234, -17.898, -17.760, -17.737, -17.772, -17.780, -17.669,
-17.161, -16.170, -18.862, -18.240, -17.903, -17.765, -17.742, -17.777, -17.785,
-17.672, -17.166, -16.178, -18.868, -18.246, -17.909, -17.770, -17.748, -17.782,
-17.790, -17.676, -17.170, -16.185, -18.875, -18.253, -17.916, -17.776, -17.754,
-17.788, -17.794, -17.680, -17.175, -16.192, -18.883, -18.261, -17.923, -17.783,
-17.760, -17.793, -17.800, -17.684, -17.180, -16.199, -18.891, -18.269, -17.930,
-17.790, -17.766, -17.799, -17.805, -17.688, -17.185, -16.206, -18.900, -18.278,
-17.939, -17.797, -17.773, -17.806, -17.810, -17.692, -17.190, -16.214, -18.909,
-18.287, -17.947, -17.805, -17.780, -17.812, -17.816, -17.697, -17.196, -16.221,
 -18.919, -18.297, -17.956, -17.813, -17.787, -17.819, -17.821, -17.701,
-17.201, -16.228, -18.929, -18.307, -17.966, -17.822, -17.795, -17.825, -17.827,
-17.706, -17.206, -16.236, -18.940, -18.318, -17.976, -17.830, -17.803, -17.832,
-17.833, -17.710, -17.212, -16.243, -18.951, -18.329, -17.986, -17.840, -17.811,
-17.839, -17.838, -17.715, -17.217, -16.251, -18.962, -18.341, -17.997, -17.849,
-17.819, -17.847, -17.844, -17.720, -17.222, -16.258, -18.974, -18.353, -18.008,
-17.859, -17.828, -17.854, -17.850, -17.725, -17.228, -16.266, -18.986, -18.365,
-18.019, -17.869, -17.836, -17.861, -17.856, -17.729, -17.234, -16.273, -18.998,
-18.378, -18.030, -17.879, -17.845, -17.869, -17.863, -17.734, -17.239, -16.281,
 -19.011, -18.391, -18.042, -17.889, -17.854, -17.877, -17.869, -17.739,
-17.245, -16.288, -19.024, -18.404, -18.054, -17.900, -17.864, -17.885, -17.875,
-17.744, -17.251, -16.296, -19.037, -18.417, -18.067, -17.911, -17.873, -17.892,
-17.881, -17.750, -17.256, -16.303, -19.051, -18.431, -18.079, -17.922, -17.882,
-17.900, -17.888, -17.755, -17.262, -16.311, -19.064, -18.445, -18.092, -17.933,
-17.892, -17.908, -17.894, -17.760, -17.268, -16.319, -19.078, -18.459, -18.105,
-17.945, -17.902, -17.917, -17.901, -17.765, -17.274, -16.327, -19.092, -18.473,
-18.119, -17.957, -17.912, -17.925, -17.907, -17.770, -17.280, -16.334, -19.106,
-18.488, -18.132, -17.968, -17.922, -17.933, -17.914, -17.776, -17.286, -16.342,
 -19.120, -18.502, -18.146, -17.980, -17.932, -17.941, -17.920, -17.781,
-17.292, -16.350, -19.135, -18.517, -18.160, -17.993, -17.942, -17.950, -17.927,
-17.786, -17.298, -16.358, -19.149, -18.532, -18.174, -18.005, -17.953, -17.958,
-17.934, -17.792, -17.304, -16.365, -19.164, -18.548, -18.188, -18.017, -17.963,
-17.967, -17.940, -17.797, -17.310, -16.373, -19.179, -18.563, -18.202, -18.030,
-17.974, -17.975, -17.947, -17.803, -17.317, -16.381, -19.194, -18.578, -18.216,
-18.042, -17.985, -17.984, -17.954, -17.808, -17.323, -16.389, -19.209, -18.594,
-18.231, -18.055, -17.995, -17.993, -17.960, -17.814, -17.329, -16.397, -19.224,
-18.610, -18.246, -18.068, -18.006, -18.002, -17.967, -17.820, -17.335, -16.405,
 -19.239, -18.625, -18.260, -18.081, -18.017, -18.010, -17.974, -17.825,
-17.342, -16.413, -19.255, -18.641, -18.275, -18.094, -18.028, -18.019, -17.981,
-17.831, -17.348, -16.421, -19.270, -18.657, -18.290, -18.107, -18.039, -18.028,
-17.988, -17.837, -17.355, -16.429, -19.285, -18.673, -18.305, -18.120, -18.050,
-18.037, -17.995, -17.842, -17.361, -16.437, -19.301, -18.689, -18.320, -18.134,
-18.061, -18.046, -18.001, -17.848, -17.367, -16.445, -19.317, -18.706, -18.336,
-18.147, -18.073, -18.055, -18.008, -17.854, -17.374, -16.453, -19.332, -18.722,
-18.351, -18.161, -18.084, -18.064, -18.015, -17.860, -17.380, -16.461, -19.348,
-18.738, -18.366, -18.174, -18.095, -18.073, -18.022, -17.866, -17.387, -16.469,
 -19.363, -18.754, -18.382, -18.188, -18.107, -18.082, -18.029, -17.872,
-17.394, -16.477, -19.379, -18.771, -18.397, -18.201, -18.118, -18.091, -18.036,
-17.877, -17.400, -16.485, -19.395, -18.787, -18.413, -18.215, -18.129, -18.100,
-18.044, -17.883, -17.407, -16.493, -19.411, -18.804, -18.428, -18.229, -18.141,
-18.110, -18.051, -17.889, -17.413, -16.501, -19.426, -18.820, -18.444, -18.243,
-18.153, -18.119, -18.058, -17.895, -17.420, -16.510, -19.442, -18.837, -18.459,
-18.256, -18.164, -18.128, -18.065, -17.901, -17.427, -16.518, -19.458, -18.853,
-18.475, -18.270, -18.176, -18.138, -18.072, -17.908, -17.434, -16.526, -19.474,
-18.870, -18.491, -18.284, -18.188, -18.147, -18.079, -17.914, -17.440, -16.534,
 -19.489, -18.886, -18.507, -18.298, -18.199, -18.156, -18.086, -17.920,
-17.447, -16.542, -19.505, -18.903, -18.522, -18.312, -18.211, -18.166, -18.094,
-17.926, -17.454, -16.551, -19.521, -18.919, -18.538, -18.326, -18.223, -18.175,
-18.101, -17.932, -17.461, -16.559, -19.537, -18.936, -18.554, -18.340, -18.235,
-18.184, -18.108, -17.938, -17.468, -16.567, -19.553, -18.952, -18.570, -18.354,
-18.246, -18.194, -18.116, -17.945, -17.475, -16.575, -19.569, -18.969, -18.586,
-18.368, -18.258, -18.203, -18.123, -17.951, -17.482, -16.583, -19.584, -18.986,
-18.601, -18.383, -18.270, -18.213, -18.130, -17.957, -17.488, -16.592, -19.600,
-19.002, -18.617, -18.397, -18.282, -18.222, -18.138, -17.964, -17.495, -16.600,
 -19.616, -19.019, -18.633, -18.411, -18.294, -18.232, -18.145, -17.970,
-17.502, -16.608, -19.632, -19.035, -18.649, -18.425, -18.306, -18.242, -18.153,
-17.976, -17.509, -16.616, -19.648, -19.052, -18.665, -18.439, -18.318, -18.251,
-18.160, -17.983, -17.516, -16.625, -19.663, -19.068, -18.681, -18.454, -18.330,
-18.261, -18.168, -17.989, -17.523, -16.633, -19.679, -19.085, -18.697, -18.468,
-18.342, -18.271, -18.175, -17.996, -17.530, -16.641, -19.695, -19.101, -18.713,
-18.482, -18.354, -18.280, -18.183, -18.002, -17.538, -16.650, -19.710, -19.118,
-18.729, -18.496, -18.366, -18.290, -18.191, -18.009, -17.545, -16.658, -19.726,
-19.134, -18.744, -18.511, -18.379, -18.300, -18.198, -18.016, -17.552, -16.666,
 -19.742, -19.151, -18.760, -18.525, -18.391, -18.310, -18.206, -18.022,
-17.559, -16.674, -19.757, -19.167, -18.776, -18.539, -18.403, -18.319, -18.214,
-18.029, -17.566, -16.683, -19.773, -19.183, -18.792, -18.554, -18.415, -18.329,
-18.221, -18.036, -17.573, -16.691, -19.789, -19.200, -18.808, -18.568, -18.427,
-18.339, -18.229, -18.042, -17.580, -16.699, -19.804, -19.216, -18.824, -18.582,
-18.439, -18.349, -18.237, -18.049, -17.588, -16.708, -19.820, -19.232, -18.839,
-18.596, -18.452, -18.359, -18.245, -18.056, -17.595, -16.716, -19.835, -19.249,
-18.855, -18.611, -18.464, -18.369, -18.253, -18.063, -17.602, -16.724, -19.851,
-19.265, -18.871, -18.625, -18.476, -18.379, -18.260, -18.069, -17.609, -16.733,
 -19.866, -19.281, -18.887, -18.639, -18.488, -18.389, -18.268, -18.076,
-17.617, -16.741, -19.882, -19.297, -18.903, -18.654, -18.500, -18.399, -18.276,
-18.083, -17.624, -16.749, -19.897, -19.313, -18.918, -18.668, -18.513, -18.409,
-18.284, -18.090, -17.631, -16.758, -19.912, -19.330, -18.934, -18.682, -18.525,
-18.419, -18.292, -18.097, -17.639, -16.766, -19.928, -19.346, -18.950, -18.697,
-18.537, -18.429, -18.300, -18.104, -17.646, -16.774, -19.943, -19.362, -18.965,
-18.711, -18.550, -18.439, -18.308, -18.111, -17.653, -16.783, -19.958, -19.378,
-18.981, -18.725, -18.562, -18.449, -18.316, -18.118, -17.661, -16.791, -19.974,
-19.394, -18.997, -18.740, -18.574, -18.459, -18.324, -18.125, -17.668, -16.799,
 -19.989, -19.410, -19.012, -18.754, -18.586, -18.469, -18.332, -18.132,
-17.675, -16.808, -20.004, -19.426, -19.028, -18.768, -18.599, -18.479, -18.341,
-18.139, -17.683, -16.816, -20.019, -19.441, -19.044, -18.782, -18.611, -18.489,
-18.349, -18.147, -17.690, -16.824, -20.034, -19.457, -19.059, -18.797, -18.623,
-18.499, -18.357, -18.154, -17.698, -16.833, -20.050, -19.473, -19.075, -18.811,
-18.636, -18.509, -18.365, -18.161, -17.705, -16.841, -20.065, -19.489, -19.090,
-18.825, -18.648, -18.520, -18.373, -18.168, -17.713, -16.849, -20.080, -19.505,
-19.106, -18.839, -18.660, -18.530, -18.382, -18.175, -17.720, -16.858, -20.095,
-19.520, -19.121, -18.853, -18.673, -18.540, -18.390, -18.183, -17.728, -16.866,
 -20.110, -19.536, -19.136, -18.868, -18.685, -18.550, -18.398, -18.190,
-17.735, -16.874, -20.125, -19.552, -19.152, -18.882, -18.697, -18.560, -18.407,
-18.197, -17.743, -16.883, -20.140, -19.567, -19.167, -18.896, -18.710, -18.571,
-18.415, -18.205, -17.750, -16.891, -20.154, -19.583, -19.182, -18.910, -18.722,
-18.581, -18.423, -18.212, -17.758, -16.899, -20.169, -19.598, -19.198, -18.924,
-18.734, -18.591, -18.432, -18.219, -17.765, -16.908, -20.184, -19.614, -19.213,
-18.938, -18.747, -18.602, -18.440, -18.227, -17.773, -16.916, -20.199, -19.629,
-19.228, -18.953, -18.759, -18.612, -18.449, -18.234, -17.780, -16.924, -20.214,
-19.645, -19.243, -18.967, -18.771, -18.622, -18.457, -18.242, -17.788, -16.932,
 -20.228, -19.660, -19.259, -18.981, -18.784, -18.633, -18.465, -18.249,
-17.796, -16.941, -20.243, -19.675, -19.274, -18.995, -18.796, -18.643, -18.474,
-18.257, -17.803, -16.949, -20.258, -19.691, -19.289, -19.009, -18.808, -18.653,
-18.483, -18.264, -17.811, -16.957, -20.272, -19.706, -19.304, -19.023, -18.821,
-18.664, -18.491, -18.272, -17.818, -16.966, -20.287, -19.721, -19.319, -19.037,
-18.833, -18.674, -18.500, -18.279, -17.826, -16.974, -20.302, -19.736, -19.334,
-19.051, -18.846, -18.684, -18.508, -18.287, -17.834, -16.982, -20.316, -19.752,
-19.349, -19.065, -18.858, -18.695, -18.517, -18.294, -17.841, -16.991, -20.331,
-19.767, -19.364, -19.079, -18.870, -18.705, -18.525, -18.302, -17.849, -16.999,
 -20.345, -19.782, -19.379, -19.093, -18.883, -18.716, -18.534, -18.310,
-17.857, -17.007, -20.360, -19.797, -19.394, -19.107, -18.895, -18.726, -18.543,
-18.317, -17.864, -17.015, -20.374, -19.812, -19.409, -19.121, -18.907, -18.736,
-18.551, -18.325, -17.872, -17.024, -20.388, -19.827, -19.424, -19.135, -18.919,
-18.747, -18.560, -18.333, -17.880, -17.032, -20.403, -19.842, -19.438, -19.149,
-18.932, -18.757, -18.569, -18.340, -17.887, -17.040, -20.417, -19.857, -19.453,
-19.162, -18.944, -18.768, -18.578, -18.348, -17.895, -17.048, -20.431, -19.871,
-19.468, -19.176, -18.956, -18.778, -18.586, -18.356, -17.903, -17.057, -20.446,
-19.886, -19.483, -19.190, -18.969, -18.789, -18.595, -18.363, -17.910, -17.065,
 -20.460, -19.901, -19.497, -19.204, -18.981, -18.799, -18.604, -18.371,
-17.918, -17.073, -20.474, -19.916, -19.512, -19.218, -18.993, -18.810, -18.613,
-18.379, -17.926, -17.081, -20.488, -19.931, -19.526, -19.231, -19.006, -18.820,
-18.621, -18.387, -17.934, -17.090, -20.502, -19.945, -19.541, -19.245, -19.018,
-18.831, -18.630, -18.395, -17.941, -17.098, -20.517, -19.960, -19.556, -19.259,
-19.030, -18.841, -18.639, -18.402, -17.949, -17.106, -20.531, -19.975, -19.570,
-19.273, -19.042, -18.852, -18.648, -18.410, -17.957, -17.114, -20.545, -19.989,
-19.585, -19.286, -19.055, -18.862, -18.657, -18.418, -17.965, -17.122, -20.559,
-20.004, -19.599, -19.300, -19.067, -18.873, -18.666, -18.426, -17.972, -17.131,
 -20.573, -20.018, -19.614, -19.314, -19.079, -18.883, -18.674, -18.434,
-17.980, -17.139, -20.587, -20.033, -19.628, -19.327, -19.091, -18.894, -18.683,
-18.442, -17.988, -17.147, -20.601, -20.047, -19.642, -19.341, -19.104, -18.904,
-18.692, -18.450, -17.995, -17.155, -20.615, -20.061, -19.657, -19.354, -19.116,
-18.915, -18.701, -18.458, -18.003, -17.163, -20.628, -20.076, -19.671, -19.368,
-19.128, -18.926, -18.710, -18.465, -18.011, -17.171, -20.642, -20.090, -19.685,
-19.382, -19.140, -18.936, -18.719, -18.473, -18.019, -17.180, -20.656, -20.104,
-19.699, -19.395, -19.152, -18.947, -18.728, -18.481, -18.026, -17.188, -20.670,
-20.119, -19.714, -19.409, -19.165, -18.957, -18.737, -18.489, -18.034, -17.196,
 -20.684, -20.133, -19.728, -19.422, -19.177, -18.968, -18.746, -18.497,
-18.042, -17.204, -20.697, -20.147, -19.742, -19.435, -19.189, -18.978, -18.755,
-18.505, -18.050, -17.212, -20.711, -20.161, -19.756, -19.449, -19.201, -18.989,
-18.764, -18.513, -18.058, -17.220, -20.725, -20.176, -19.770, -19.462, -19.213,
-18.999, -18.773, -18.521, -18.065, -17.228, -20.738, -20.190, -19.784, -19.476,
-19.225, -19.010, -18.782, -18.529, -18.073, -17.236, -20.752, -20.204, -19.798,
-19.489, -19.237, -19.020, -18.791, -18.537, -18.081, -17.244, -20.766, -20.218,
-19.812, -19.502, -19.250, -19.031, -18.800, -18.545, -18.089, -17.253, -20.779,
-20.232, -19.826, -19.516, -19.262, -19.042, -18.809, -18.553, -18.096, -17.261,
 -20.793, -20.246, -19.840, -19.529, -19.274, -19.052, -18.818, -18.561,
-18.104, -17.269, -20.806, -20.260, -19.854, -19.542, -19.286, -19.063, -18.827,
-18.569, -18.112, -17.277, -20.820, -20.274, -19.868, -19.556, -19.298, -19.073,
-18.836, -18.577, -18.120, -17.285, -20.833, -20.287, -19.882, -19.569, -19.310,
-19.084, -18.845, -18.585, -18.127, -17.293, -20.847, -20.301, -19.895, -19.582,
-19.322, -19.094, -18.854, -18.593, -18.135, -17.301, -20.860, -20.315, -19.909,
-19.595, -19.334, -19.105, -18.863, -18.601, -18.143, -17.309, -20.873, -20.329,
-19.923, -19.608, -19.346, -19.116, -18.872, -18.609, -18.150, -17.317, -20.887,
-20.343, -19.937, -19.621, -19.358, -19.126, -18.881, -18.617, -18.158, -17.325,
 -20.900, -20.356, -19.950, -19.635, -19.370, -19.137, -18.890, -18.626,
-18.166, -17.333, -20.913, -20.370, -19.964, -19.648, -19.382, -19.147, -18.899,
-18.634, -18.174, -17.341, -20.927, -20.384, -19.978, -19.661, -19.394, -19.158,
-18.908, -18.642, -18.181, -17.349, -20.940, -20.397, -19.991, -19.674, -19.406,
-19.168, -18.918, -18.650, -18.189, -17.357, -20.953, -20.411, -20.005, -19.687,
-19.418, -19.179, -18.927, -18.658, -18.197, -17.365, -20.966, -20.425, -20.018,
-19.700, -19.430, -19.189, -18.936, -18.666, -18.204, -17.373, -20.980, -20.438,
-20.032, -19.713, -19.442, -19.200, -18.945, -18.674, -18.212, -17.380, -20.993,
-20.452, -20.045, -19.726, -19.454, -19.210, -18.954, -18.682, -18.220, -17.388,
 -21.006, -20.465, -20.059, -19.739, -19.466, -19.221, -18.963, -18.690,
-18.228, -17.396, -21.019, -20.479, -20.072, -19.752, -19.477, -19.231, -18.972,
-18.698, -18.235, -17.404, -21.032, -20.492, -20.086, -19.765, -19.489, -19.242,
-18.981, -18.706, -18.243, -17.412, -21.045, -20.506, -20.099, -19.777, -19.501,
-19.252, -18.990, -18.714, -18.251, -17.420, -21.058, -20.519, -20.112, -19.790,
-19.513, -19.263, -18.999, -18.722, -18.258, -17.428, -21.071, -20.532, -20.126,
-19.803, -19.525, -19.273, -19.008, -18.730, -18.266, -17.436, -21.084, -20.546,
-20.139, -19.816, -19.537, -19.284, -19.017, -18.738, -18.274, -17.443, -21.097,
-20.559, -20.152, -19.829, -19.549, -19.294, -19.027, -18.746, -18.281, -17.451,
 -21.110, -20.572, -20.165, -19.841, -19.560, -19.305, -19.036, -18.754,
-18.289, -17.459, -21.123, -20.586, -20.179, -19.854, -19.572, -19.315, -19.045,
-18.762, -18.296, -17.467, -21.136, -20.599, -20.192, -19.867, -19.584, -19.326,
-19.054, -18.770, -18.304, -17.475, -21.149, -20.612, -20.205, -19.880, -19.596,
-19.336, -19.063, -18.779, -18.312, -17.482, -21.162, -20.625, -20.218, -19.892,
-19.607, -19.347, -19.072, -18.787, -18.319, -17.490, -21.174, -20.638, -20.231,
-19.905, -19.619, -19.357, -19.081, -18.795, -18.327, -17.498, -21.187, -20.651,
-20.244, -19.918, -19.631, -19.368, -19.090, -18.803, -18.334, -17.506, -21.200,
-20.665, -20.257, -19.930, -19.642, -19.378, -19.099, -18.811, -18.342, -17.513,
 -21.213, -20.678, -20.270, -19.943, -19.654, -19.389, -19.108, -18.819,
-18.349, -17.521, -21.225, -20.691, -20.283, -19.955, -19.666, -19.399, -19.117,
-18.827, -18.357, -17.529, -21.238, -20.704, -20.296, -19.968, -19.677, -19.409,
-19.126, -18.835, -18.364, -17.537, -21.251, -20.717, -20.309, -19.980, -19.689,
-19.420, -19.135, -18.843, -18.372, -17.544, -21.264, -20.730, -20.322, -19.993,
-19.701, -19.430, -19.144, -18.851, -18.379, -17.552, -21.276, -20.743, -20.335,
-20.005, -19.712, -19.441, -19.153, -18.859, -18.387, -17.559, -21.289, -20.755,
-20.348, -20.018, -19.724, -19.451, -19.162, -18.866, -18.394, -17.567, -21.301,
-20.768, -20.361, -20.030, -19.735, -19.461, -19.171, -18.874, -18.402, -17.575,
 -21.314, -20.781, -20.374, -20.043, -19.747, -19.472, -19.180, -18.882,
-18.409, -17.582, -21.327, -20.794, -20.386, -20.055, -19.759, -19.482, -19.189,
-18.890, -18.417, -17.590, -21.339, -20.807, -20.399, -20.067, -19.770, -19.492,
-19.198, -18.898, -18.424, -17.597, -21.352, -20.820, -20.412, -20.080, -19.782,
-19.503, -19.207, -18.906, -18.432, -17.605, -21.364, -20.833, -20.425, -20.092,
-19.793, -19.513, -19.216, -18.914, -18.439, -17.613, -21.377, -20.845, -20.437,
-20.104, -19.805, -19.523, -19.225, -18.922, -18.446, -17.620, -21.389, -20.858,
-20.450, -20.117, -19.816, -19.534, -19.234, -18.930, -18.454, -17.628, -21.402,
-20.871, -20.463, -20.129, -19.828, -19.544, -19.243, -18.938, -18.461, -17.635,
 -21.414, -20.883, -20.475, -20.141, -19.839, -19.554, -19.252, -18.945,
-18.468, -17.643, -21.426, -20.896, -20.488, -20.153, -19.850, -19.564, -19.261,
-18.953, -18.476, -17.650, -21.439, -20.909, -20.501, -20.166, -19.862, -19.575,
-19.270, -18.961, -18.483, -17.657, -21.451, -20.921, -20.513, -20.178, -19.873,
-19.585, -19.279, -18.969, -18.490, -17.665, -21.463, -20.934, -20.526, -20.190,
-19.885, -19.595, -19.287, -18.977, -18.497, -17.672, -21.476, -20.947, -20.538,
-20.202, -19.896, -19.605, -19.296, -18.984, -18.505, -17.680, -21.488, -20.959,
-20.551, -20.214, -19.907, -19.615, -19.305, -18.992, -18.512, -17.687, -21.500,
-20.972, -20.563, -20.226, -19.919, -19.626, -19.314, -19.000, -18.519, -17.694,
 -21.513, -20.984, -20.576, -20.238, -19.930, -19.636, -19.323, -19.008,
-18.526, -17.702, -21.525, -20.997, -20.588, -20.250, -19.941, -19.646, -19.332,
-19.015, -18.533, -17.709, -21.537, -21.009, -20.600, -20.262, -19.952, -19.656,
-19.340, -19.023, -18.540, -17.716, -21.549, -21.022, -20.613, -20.274, -19.964,
-19.666, -19.349, -19.031, -18.548, -17.724, -21.561, -21.034, -20.625, -20.286,
-19.975, -19.676, -19.358, -19.038, -18.555, -17.731, -21.574, -21.046, -20.637,
-20.298, -19.986, -19.686, -19.367, -19.046, -18.562, -17.738, -21.586, -21.059,
-20.650, -20.310, -19.997, -19.696, -19.375, -19.053, -18.569, -17.745, -21.598,
-21.071, -20.662, -20.322, -20.008, -19.706, -19.384, -19.061, -18.576, -17.753,
 -21.610, -21.084, -20.674, -20.334, -20.020, -19.716, -19.393, -19.069,
-18.583, -17.760, -21.622, -21.096, -20.687, -20.346, -20.031, -19.726, -19.401,
-19.076, -18.590, -17.767, -21.634, -21.108, -20.699, -20.358, -20.042, -19.736,
-19.410, -19.084, -18.597, -17.774, -21.646, -21.120, -20.711, -20.370, -20.053,
-19.746, -19.419, -19.091, -18.604, -17.781, -21.658, -21.133, -20.723, -20.382,
-20.064, -19.756, -19.427, -19.099, -18.611, -17.788, -21.670, -21.145, -20.735,
-20.393, -20.075, -19.766, -19.436, -19.106, -18.617, -17.796, -21.682, -21.157,
-20.748, -20.405, -20.086, -19.776, -19.444, -19.113, -18.624, -17.803, -21.694,
-21.169, -20.760, -20.417, -20.097, -19.786, -19.453, -19.121, -18.631, -17.810,
 -21.706, -21.182, -20.772, -20.429, -20.108, -19.796, -19.461, -19.128,
-18.638, -17.817, -21.718, -21.194, -20.784, -20.440, -20.119, -19.806, -19.470,
-19.136, -18.645, -17.824, -21.730, -21.206, -20.796, -20.452, -20.130, -19.816,
-19.478, -19.143, -18.651, -17.831, -21.742, -21.218, -20.808, -20.464, -20.141,
-19.826, -19.487, -19.150, -18.658, -17.838, -21.754, -21.230, -20.820, -20.476,
-20.152, -19.835, -19.495, -19.157, -18.665, -17.845, -21.766, -21.242, -20.832,
-20.487, -20.163, -19.845, -19.504, -19.165, -18.672, -17.852, -21.778, -21.254,
-20.844, -20.499, -20.174, -19.855, -19.512, -19.172, -18.678, -17.858, -21.790,
-21.266, -20.856, -20.510, -20.185, -19.865, -19.520, -19.179, -18.685, -17.865,
 -21.802, -21.278, -20.868, -20.522, -20.196, -19.874, -19.529, -19.186,
-18.691, -17.872, -21.813, -21.290, -20.880, -20.534, -20.206, -19.884, -19.537,
-19.193, -18.698, -17.879, -21.825, -21.302, -20.892, -20.545, -20.217, -19.894,
-19.545, -19.200, -18.704, -17.886, -21.837, -21.314, -20.904, -20.557, -20.228,
-19.903, -19.553, -19.207, -18.711, -17.893, -21.849, -21.326, -20.915, -20.568,
-20.239, -19.913, -19.561, -19.214, -18.717, -17.899, -21.861, -21.338, -20.927,
-20.580, -20.249, -19.922, -19.570, -19.221, -18.724, -17.906, -21.872, -21.350,
-20.939, -20.591, -20.260, -19.932, -19.578, -19.228, -18.730, -17.913, -21.884,
-21.362, -20.951, -20.603, -20.271, -19.941, -19.586, -19.235, -18.737, -17.920,
 -21.896, -21.374, -20.963, -20.614, -20.282, -19.951, -19.594, -19.242,
-18.743, -17.926, -21.907, -21.386, -20.974, -20.625, -20.292, -19.960, -19.602,
-19.249, -18.749, -17.933, -21.919, -21.398, -20.986, -20.637, -20.303, -19.970,
-19.610, -19.256, -18.755, -17.939, -21.931, -21.410, -20.998, -20.648, -20.313,
-19.979, -19.618, -19.262, -18.762, -17.946, -21.942, -21.421, -21.010, -20.659,
-20.324, -19.989, -19.626, -19.269, -18.768, -17.953, -21.954, -21.433, -21.021,
-20.671, -20.335, -19.998, -19.634, -19.276, -18.774, -17.959, -21.966, -21.445,
-21.033, -20.682, -20.345, -20.007, -19.642, -19.282, -18.780, -17.966, -21.977,
-21.457, -21.045, -20.693, -20.356, -20.017, -19.649, -19.289, -18.786, -17.972,
 -21.989, -21.469, -21.056, -20.705, -20.366, -20.026, -19.657, -19.295,
-18.792, -17.979, -22.000, -21.480, -21.068, -20.716, -20.376, -20.035, -19.665,
-19.302, -18.798, -17.985, -22.012, -21.492, -21.080, -20.727, -20.387, -20.044,
-19.673, -19.308, -18.804, -17.991, -22.024, -21.504, -21.091, -20.738, -20.397,
-20.053, -19.680, -19.315, -18.810, -17.998, -22.035, -21.516, -21.103, -20.749,
-20.408, -20.062, -19.688, -19.321, -18.816, -18.004, -22.047, -21.527, -21.114,
-20.761, -20.418, -20.071, -19.695, -19.327, -18.822, -18.010, -22.058, -21.539,
-21.126, -20.772, -20.428, -20.080, -19.703, -19.334, -18.827, -18.017, -22.070,
-21.551, -21.137, -20.783, -20.439, -20.089, -19.710, -19.340, -18.833, -18.023,
 -22.081, -21.562, -21.149, -20.794, -20.449, -20.098, -19.718, -19.346,
-18.839, -18.029, -22.093, -21.574, -21.160, -20.805, -20.459, -20.107, -19.725,
-19.352, -18.845, -18.035, -22.104, -21.585, -21.172, -20.816, -20.469, -20.116,
-19.732, -19.358, -18.850, -18.042, -22.115, -21.597, -21.183, -20.827, -20.479,
-20.125, -19.740, -19.364, -18.856, -18.048, -22.127, -21.609, -21.195, -20.838,
-20.490, -20.134, -19.747, -19.370, -18.861, -18.054, -22.138, -21.620, -21.206,
-20.849, -20.500, -20.143, -19.754, -19.376, -18.867, -18.060, -22.150, -21.632,
-21.218, -20.860, -20.510, -20.151, -19.761, -19.382, -18.872, -18.066, -22.161,
-21.643, -21.229, -20.871, -20.520, -20.160, -19.768, -19.388, -18.878, -18.072,
 -22.172, -21.655, -21.240, -20.882, -20.530, -20.168, -19.775, -19.394,
-18.883, -18.078, -22.184, -21.666, -21.252, -20.893, -20.540, -20.177, -19.782,
-19.400, -18.888, -18.084, -22.195, -21.678, -21.263, -20.904, -20.550, -20.185,
-19.789, -19.405, -18.894, -18.090, -22.207, -21.689, -21.274, -20.915, -20.559,
-20.194, -19.796, -19.411, -18.899, -18.096, -22.218, -21.701, -21.286, -20.925,
-20.569, -20.202, -19.803, -19.416, -18.904, -18.102, -22.229, -21.712, -21.297,
-20.936, -20.579, -20.211, -19.810, -19.422, -18.909, -18.107, -22.240, -21.723,
-21.308, -20.947, -20.589, -20.219, -19.816, -19.427, -18.914, -18.113, -22.252,
-21.735, -21.319, -20.958, -20.599, -20.227, -19.823, -19.433, -18.919, -18.119,
 -22.263, -21.746, -21.331, -20.968, -20.608, -20.235, -19.830, -19.438,
-18.924, -18.125, -22.274, -21.758, -21.342, -20.979, -20.618, -20.243, -19.836,
-19.443, -18.929, -18.130, -22.285, -21.769, -21.353, -20.990, -20.627, -20.251,
-19.842, -19.449, -18.934, -18.136, -22.297, -21.780, -21.364, -21.000, -20.637,
-20.259, -19.849, -19.454, -18.939, -18.142, -22.308, -21.792, -21.375, -21.011,
-20.646, -20.267, -19.855, -19.459, -18.944, -18.147, -22.319, -21.803, -21.386,
-21.021, -20.656, -20.275, -19.861, -19.464, -18.948, -18.153, -22.330, -21.814,
-21.398, -21.032, -20.665, -20.283, -19.868, -19.469, -18.953, -18.158, -22.341,
-21.826, -21.409, -21.043, -20.675, -20.291, -19.874, -19.474, -18.958, -18.164,
 -22.353, -21.837, -21.420, -21.053, -20.684, -20.298, -19.880, -19.478,
-18.962, -18.169, -22.364, -21.848, -21.431, -21.064, -20.693, -20.306, -19.886,
-19.483, -18.967, -18.175, -22.375, -21.859, -21.442, -21.074, -20.702, -20.313,
-19.891, -19.488, -18.971, -18.180, -22.386, -21.871, -21.453, -21.084, -20.712,
-20.321, -19.897, -19.493, -18.976, -18.185, -22.397, -21.882, -21.464, -21.095,
-20.721, -20.328, -19.903, -19.497, -18.980, -18.191, -22.408, -21.893, -21.475,
-21.105, -20.730, -20.336, -19.909, -19.502, -18.984, -18.196, -22.419, -21.904,
-21.486, -21.115, -20.739, -20.343, -19.914, -19.506, -18.988, -18.201, -22.430,
-21.915, -21.497, -21.126, -20.748, -20.350, -19.920, -19.510, -18.993, -18.206,
 -22.442, -21.927, -21.508, -21.136, -20.756, -20.357, -19.925, -19.515,
-18.997, -18.211, -22.453, -21.938, -21.519, -21.146, -20.765, -20.364, -19.930,
-19.519, -19.001, -18.217, -22.464, -21.949, -21.530, -21.156, -20.774, -20.371,
-19.936, -19.523, -19.005, -18.222, -22.475, -21.960, -21.540, -21.166, -20.783,
-20.378, -19.941, -19.527, -19.009, -18.227, -22.486, -21.971, -21.551, -21.176,
-20.791, -20.384, -19.946, -19.531, -19.013, -18.232, -22.497, -21.982, -21.562,
-21.186, -20.800, -20.391, -19.951, -19.535, -19.017, -18.237, -22.508, -21.993,
-21.573, -21.196, -20.808, -20.397, -19.956, -19.539, -19.020, -18.242, -22.519,
-22.004, -21.584, -21.206, -20.817, -20.404, -19.961, -19.543, -19.024, -18.246,
 -22.530, -22.015, -21.594, -21.216, -20.825, -20.410, -19.965, -19.546,
-19.028, -18.251, -22.541, -22.026, -21.605, -21.226, -20.833, -20.416, -19.970,
-19.550, -19.031, -18.256, -22.552, -22.037, -21.616, -21.236, -20.841, -20.423,
-19.975, -19.553, -19.035, -18.261, -22.563, -22.048, -21.627, -21.246, -20.849,
-20.429, -19.979, -19.557, -19.038, -18.266, -22.573, -22.059, -21.637, -21.255,
-20.857, -20.435, -19.983, -19.560, -19.042, -18.270, -22.584, -22.070, -21.648,
-21.265, -20.865, -20.440, -19.988, -19.564, -19.045, -18.275, -22.595, -22.081,
-21.659, -21.275, -20.873, -20.446, -19.992, -19.567, -19.049, -18.280, -22.606,
-22.092, -21.669, -21.284, -20.881, -20.452, -19.996, -19.570, -19.052, -18.284,
 -22.617, -22.103, -21.680, -21.294, -20.888, -20.457, -20.000, -19.573,
-19.055, -18.289, -22.628, -22.114, -21.690, -21.303, -20.896, -20.463, -20.004,
-19.576, -19.058, -18.294, -22.639, -22.125, -21.701, -21.313, -20.903, -20.468,
-20.008, -19.579, -19.061, -18.298, -22.650, -22.136, -21.711, -21.322, -20.911,
-20.473, -20.011, -19.582, -19.064, -18.302, -22.661, -22.147, -21.722, -21.331,
-20.918, -20.478, -20.015, -19.585, -19.067, -18.307, -22.671, -22.158, -21.732,
-21.340, -20.925, -20.483, -20.018, -19.587, -19.070, -18.311, -22.682, -22.169,
-21.743, -21.349, -20.932, -20.488, -20.022, -19.590, -19.073, -18.316, -22.693,
-22.180, -21.753, -21.359, -20.939, -20.493, -20.025, -19.593, -19.076, -18.320,
 -22.704, -22.190, -21.763, -21.368, -20.946, -20.497, -20.028, -19.595,
-19.079, -18.324, -22.715, -22.201, -21.774, -21.376, -20.952, -20.502, -20.031,
-19.597, -19.082, -18.329, -22.725, -22.212, -21.784, -21.385, -20.959, -20.506,
-20.034, -19.600, -19.084, -18.333, -22.736, -22.223, -21.794, -21.394, -20.965,
-20.510, -20.037, -19.602, -19.087, -18.337, -22.747, -22.234, -21.805, -21.403,
-20.972, -20.514, -20.040, -19.604, -19.089, -18.341, -22.758, -22.244, -21.815,
-21.411, -20.978, -20.518, -20.042, -19.606, -19.092, -18.345, -22.768, -22.255,
-21.825, -21.420, -20.984, -20.522, -20.045, -19.608, -19.094, -18.349, -22.779,
-22.266, -21.835, -21.428, -20.990, -20.525, -20.047, -19.610, -19.097, -18.354,
 -22.790, -22.277, -21.845, -21.437, -20.995, -20.529, -20.049, -19.612,
-19.099, -18.358, -22.801, -22.287, -21.855, -21.445, -21.001, -20.532, -20.051,
-19.613, -19.101, -18.362, -22.811, -22.298, -21.865, -21.453, -21.007, -20.535,
-20.053, -19.615, -19.104, -18.366, -22.822, -22.309, -21.875, -21.461, -21.012,
-20.538, -20.055, -19.617, -19.106, -18.370, -22.833, -22.319, -21.885, -21.469,
-21.017, -20.541, -20.057, -19.618, -19.108, -18.374, -22.843, -22.330, -21.895,
-21.477, -21.022, -20.544, -20.059, -19.619, -19.110, -18.377, -22.854, -22.341,
-21.905, -21.485, -21.027, -20.546, -20.060, -19.621, -19.112, -18.381, -22.865,
-22.351, -21.915, -21.493, -21.032, -20.549, -20.062, -19.622, -19.114, -18.385,
 -22.875, -22.362, -21.924, -21.500, -21.036, -20.551, -20.063, -19.623,
-19.116, -18.389, -22.886, -22.372, -21.934, -21.508, -21.041, -20.553, -20.064,
-19.624, -19.118, -18.393, -22.897, -22.383, -21.944, -21.515, -21.045, -20.555,
-20.065, -19.625, -19.120, -18.397, -22.907, -22.393, -21.953, -21.522, -21.049,
-20.557, -20.066, -19.626, -19.122, -18.401, -22.918, -22.404, -21.963, -21.529,
-21.053, -20.559, -20.067, -19.627, -19.123, -18.404, -22.928, -22.414, -21.972,
-21.536, -21.057, -20.560, -20.068, -19.627, -19.125, -18.408, -22.939, -22.425,
-21.982, -21.543, -21.060, -20.561, -20.068, -19.628, -19.127, -18.412, -22.950,
-22.435, -21.991, -21.550, -21.064, -20.562, -20.069, -19.629, -19.128, -18.416,
 -22.960, -22.446, -22.000, -21.556, -21.067, -20.563, -20.069, -19.629,
-19.130, -18.419, -22.971, -22.456, -22.009, -21.563, -21.070, -20.564, -20.069,
-19.630, -19.132, -18.423, -22.981, -22.467, -22.019, -21.569, -21.073, -20.565,
-20.069, -19.630, -19.133, -18.427, -22.992, -22.477, -22.028, -21.575, -21.075,
-20.565, -20.069, -19.630, -19.135, -18.431, -23.002, -22.487, -22.037, -21.581,
-21.078, -20.565, -20.069, -19.630, -19.136, -18.434, -23.013, -22.498, -22.046,
-21.587, -21.080, -20.565, -20.069, -19.630, -19.137, -18.438, -23.023, -22.508,
-22.054, -21.592, -21.082, -20.565, -20.068, -19.630, -19.139, -18.442, -23.034,
-22.518, -22.063, -21.598, -21.084, -20.565, -20.068, -19.630, -19.140, -18.446,
 -23.044, -22.528, -22.072, -21.603, -21.086, -20.565, -20.067, -19.630,
-19.142, -18.449, -23.055, -22.539, -22.080, -21.608, -21.087, -20.564, -20.066,
-19.630, -19.143, -18.453, -23.065, -22.549, -22.089, -21.613, -21.088, -20.563,
-20.065, -19.630, -19.144, -18.457, -23.076, -22.559, -22.097, -21.618, -21.089,
-20.562, -20.064, -19.629, -19.146, -18.461, -23.086, -22.569, -22.106, -21.623,
-21.090, -20.561, -20.063, -19.629, -19.147, -18.464, -23.097, -22.579, -22.114,
-21.627, -21.091, -20.560, -20.061, -19.628, -19.148, -18.468, -23.107, -22.589,
-22.122, -21.631, -21.091, -20.558, -20.060, -19.628, -19.149, -18.472, -23.117,
-22.599, -22.130, -21.636, -21.092, -20.557, -20.058, -19.627, -19.151, -18.476,
 -23.128, -22.609, -22.138, -21.639, -21.092, -20.555, -20.057, -19.626,
-19.152, -18.480, -23.138, -22.619, -22.146, -21.643, -21.092, -20.553, -20.055,
-19.626, -19.153, -18.484, -23.148, -22.629, -22.154, -21.647, -21.091, -20.551,
-20.053, -19.625, -19.155, -18.487, -23.159, -22.639, -22.161, -21.650, -21.091,
-20.549, -20.051, -19.624, -19.156, -18.491, -23.169, -22.649, -22.169, -21.653,
-21.090, -20.546, -20.049, -19.623, -19.157, -18.495, -23.180, -22.659, -22.176,
-21.656, -21.089, -20.544, -20.046, -19.622, -19.158, -18.499, -23.190, -22.669,
-22.184, -21.659, -21.088, -20.541, -20.044, -19.621, -19.160, -18.504, -23.200,
-22.678, -22.191, -21.662, -21.087, -20.538, -20.042, -19.620, -19.161, -18.508,
 -23.211, -22.688, -22.198, -21.664, -21.086, -20.535, -20.039, -19.619,
-19.162, -18.512, -23.221, -22.698, -22.205, -21.667, -21.084, -20.532, -20.036,
-19.617, -19.164, -18.516, -23.231, -22.707, -22.212, -21.669, -21.082, -20.529,
-20.034, -19.616, -19.165, -18.520, -23.241, -22.717, -22.219, -21.671, -21.081,
-20.526, -20.031, -19.615, -19.166, -18.525, -23.252, -22.727, -22.225, -21.673,
-21.079, -20.522, -20.028, -19.614, -19.168, -18.529, -23.262, -22.736, -22.232,
-21.674, -21.077, -20.519, -20.025, -19.612, -19.169, -18.533, -23.272, -22.746,
-22.238, -21.676, -21.074, -20.515, -20.022, -19.611, -19.171, -18.538, -23.282,
-22.755, -22.245, -21.677, -21.072, -20.511, -20.019, -19.610, -19.173, -18.542,
 -23.293, -22.764, -22.251, -21.679, -21.070, -20.508, -20.016, -19.609,
-19.174, -18.547, -23.303, -22.774, -22.257, -21.680, -21.067, -20.504, -20.013,
-19.607, -19.176, -18.552, -23.313, -22.783, -22.263, -21.681, -21.065, -20.500,
-20.009, -19.606, -19.178, -18.556, -23.323, -22.792, -22.269, -21.682, -21.062,
-20.496, -20.006, -19.605, -19.180, -18.561, -23.333, -22.802, -22.275, -21.683,
-21.060, -20.492, -20.003, -19.604, -19.182, -18.566, -23.343, -22.811, -22.281,
-21.684, -21.057, -20.489, -20.000, -19.602, -19.184, -18.571, -23.354, -22.820,
-22.286, -21.685, -21.055, -20.485, -19.997, -19.601, -19.186, -18.576, -23.364,
-22.829, -22.292, -21.686, -21.052, -20.481, -19.993, -19.600, -19.188, -18.581,
 -23.374, -22.838, -22.298, -21.686, -21.050, -20.477, -19.990, -19.599,
-19.190, -18.586, -23.384, -22.848, -22.304, -21.687, -21.047, -20.474, -19.987,
-19.598, -19.192, -18.591, -23.394, -22.857, -22.309, -21.688, -21.045, -20.470,
-19.984, -19.597, -19.195, -18.597, -23.404, -22.866, -22.315, -21.689, -21.042,
-20.466, -19.981, -19.596, -19.197, -18.602, -23.414, -22.875, -22.320, -21.690,
-21.040, -20.463, -19.978, -19.596, -19.200, -18.608, -23.424, -22.884, -22.326,
-21.691, -21.038, -20.460, -19.976, -19.595, -19.203, -18.613, -23.434, -22.893,
-22.332, -21.693, -21.037, -20.457, -19.973, -19.594, -19.206, -18.619, -23.444,
-22.902, -22.338, -21.694, -21.035, -20.454, -19.971, -19.594, -19.208, -18.624,
 -23.454, -22.910, -22.339, -21.689, -21.027, -20.445, -19.964, -19.591,
-19.210, -18.629, -23.464, -22.916, -22.338, -21.679, -21.013, -20.432, -19.954,
-19.585, -19.209, -18.633, -23.473, -22.923, -22.336, -21.669, -21.000, -20.419,
-19.943, -19.579, -19.209, -18.638, -23.483, -22.929, -22.334, -21.658, -20.986,
-20.406, -19.933, -19.573, -19.209, -18.642, -23.493, -22.935, -22.331, -21.648,
-20.973, -20.394, -19.924, -19.568, -19.208, -18.646, -23.503, -22.944, -22.337,
-21.651, -20.973, -20.393, -19.923, -19.568, -19.212, -18.652, -23.513, -22.953,
-22.344, -21.654, -20.974, -20.392, -19.922, -19.569, -19.216, -18.658, -23.522,
-22.961, -22.349, -21.656, -20.973, -20.390, -19.920, -19.570, -19.219, -18.664,
 -23.532, -22.970, -22.354, -21.657, -20.972, -20.388, -19.918, -19.569,
-19.222, -18.670, -23.542, -22.977, -22.355, -21.652, -20.965, -20.380, -19.912,
-19.566, -19.223, -18.674, -23.551, -22.982, -22.351, -21.640, -20.950, -20.367,
-19.902, -19.560, -19.222, -18.678, -23.561, -22.987, -22.347, -21.628, -20.936,
-20.353, -19.891, -19.554, -19.222, -18.682, -23.570, -22.993, -22.346, -21.622,
-20.928, -20.345, -19.885, -19.551, -19.223, -18.686, -23.580, -23.001, -22.348,
-21.619, -20.923, -20.341, -19.882, -19.549, -19.225, -18.691, -23.590, -23.013,
-22.365, -21.638, -20.940, -20.353, -19.891, -19.557, -19.233, -18.699, -23.600,
-23.022, -22.372, -21.643, -20.943, -20.355, -19.892, -19.560, -19.237, -18.705,
 -23.610, -23.034, -22.386, -21.657, -20.955, -20.364, -19.899, -19.565,
-19.243, -18.712, -23.621, -23.046, -22.401, -21.673, -20.968, -20.373, -19.905,
-19.571, -19.250, -18.719, -23.630, -23.054, -22.404, -21.672, -20.965, -20.370,
-19.903, -19.570, -19.252, -18.723, -23.640, -23.063, -22.410, -21.673, -20.964,
-20.367, -19.900, -19.569, -19.253, -18.727, -23.650, -23.069, -22.409, -21.666,
-20.953, -20.357, -19.892, -19.564, -19.252, -18.731, -23.658, -23.067, -22.388,
-21.632, -20.918, -20.327, -19.870, -19.549, -19.245, -18.731, -23.666, -23.065,
-22.367, -21.598, -20.885, -20.300, -19.849, -19.535, -19.238, -18.731, -23.674,
-23.067, -22.359, -21.585, -20.872, -20.289, -19.841, -19.530, -19.237, -18.733,
 -23.684, -23.076, -22.366, -21.591, -20.878, -20.293, -19.845, -19.534,
-19.241, -18.739, -23.695, -23.095, -22.397, -21.629, -20.912, -20.321, -19.866,
-19.550, -19.254, -18.748, -23.709, -23.121, -22.447, -21.690, -20.968, -20.365,
-19.899, -19.575, -19.271, -18.759, -23.721, -23.144, -22.486, -21.736, -21.009,
-20.397, -19.922, -19.592, -19.283, -18.767, -23.731, -23.155, -22.497, -21.746,
-21.015, -20.401, -19.926, -19.595, -19.287, -18.772, -23.739, -23.154, -22.481,
-21.719, -20.990, -20.382, -19.911, -19.586, -19.282, -18.772, -23.745, -23.142,
-22.436, -21.658, -20.933, -20.336, -19.879, -19.563, -19.268, -18.769, -23.751,
-23.128, -22.392, -21.598, -20.876, -20.291, -19.845, -19.539, -19.254, -18.764,
 -23.759, -23.130, -22.380, -21.577, -20.855, -20.273, -19.832, -19.530,
-19.249, -18.764, -23.770, -23.141, -22.390, -21.584, -20.859, -20.276, -19.835,
-19.532, -19.251, -18.767, -23.781, -23.157, -22.413, -21.608, -20.880, -20.293,
-19.847, -19.541, -19.258, -18.772, -23.793, -23.179, -22.450, -21.651, -20.919,
-20.324, -19.871, -19.559, -19.270, -18.780, -23.804, -23.195, -22.472, -21.676,
-20.942, -20.343, -19.886, -19.571, -19.278, -18.786, -23.813, -23.200, -22.472,
-21.673, -20.940, -20.343, -19.887, -19.572, -19.279, -18.788, -23.823, -23.209,
-22.479, -21.679, -20.945, -20.348, -19.892, -19.575, -19.281, -18.791, -23.832,
-23.213, -22.475, -21.669, -20.935, -20.341, -19.888, -19.572, -19.279, -18.792,
 -23.840, -23.211, -22.458, -21.645, -20.913, -20.324, -19.876, -19.564,
-19.274, -18.791, -23.849, -23.216, -22.454, -21.636, -20.904, -20.317, -19.871,
-19.561, -19.272, -18.791, -23.857, -23.217, -22.444, -21.620, -20.889, -20.306,
-19.864, -19.555, -19.268, -18.790, -23.866, -23.220, -22.438, -21.609, -20.879,
-20.298, -19.859, -19.552, -19.265, -18.790, -23.876, -23.231, -22.447, -21.617,
-20.885, -20.304, -19.864, -19.555, -19.267, -18.792, -23.886, -23.243, -22.460,
-21.629, -20.895, -20.313, -19.872, -19.561, -19.269, -18.794, -23.897, -23.257,
-22.477, -21.646, -20.910, -20.326, -19.882, -19.568, -19.274, -18.797, -23.908,
-23.269, -22.491, -21.659, -20.923, -20.337, -19.891, -19.574, -19.277, -18.799,
 -23.917, -23.275, -22.492, -21.657, -20.921, -20.336, -19.892, -19.575,
-19.276, -18.800, -23.926, -23.284, -22.498, -21.662, -20.924, -20.340, -19.896,
-19.577, -19.277, -18.801, -23.936, -23.290, -22.498, -21.658, -20.921, -20.338,
-19.895, -19.576, -19.275, -18.801, -23.945, -23.296, -22.498, -21.655, -20.918,
-20.337, -19.895, -19.576, -19.274, -18.800, -23.955, -23.306, -22.507, -21.662,
-20.924, -20.343, -19.900, -19.579, -19.275, -18.801, -23.966, -23.321, -22.526,
-21.681, -20.941, -20.357, -19.912, -19.587, -19.279, -18.804, -23.977, -23.336,
-22.545, -21.700, -20.958, -20.371, -19.923, -19.595, -19.283, -18.806, -23.988,
-23.351, -22.565, -21.720, -20.975, -20.386, -19.935, -19.602, -19.286, -18.808,
 -23.999, -23.367, -22.586, -21.741, -20.994, -20.402, -19.947, -19.611,
-19.290, -18.810, -24.009, -23.374, -22.590, -21.743, -20.995, -20.404, -19.950,
-19.612, -19.290, -18.810, -24.018, -23.382, -22.594, -21.745, -20.997, -20.407,
-19.953, -19.614, -19.290, -18.810, -24.028, -23.390, -22.599, -21.748, -21.000,
-20.410, -19.956, -19.616, -19.290, -18.811, -24.037, -23.398, -22.604, -21.751,
-21.002, -20.413, -19.959, -19.618, -19.290, -18.811, -24.047, -23.406, -22.609,
-21.754, -21.006, -20.417, -19.963, -19.620, -19.290, -18.811, -24.057, -23.414,
-22.615, -21.758, -21.009, -20.420, -19.967, -19.622, -19.291, -18.812, -24.066,
-23.423, -22.621, -21.762, -21.013, -20.425, -19.971, -19.625, -19.291, -18.812,
 -24.076, -23.431, -22.627, -21.767, -21.018, -20.429, -19.975, -19.627,
-19.292, -18.812, -24.085, -23.440, -22.634, -21.772, -21.022, -20.434, -19.979,
-19.630, -19.293, -18.813, -24.095, -23.449, -22.641, -21.778, -21.027, -20.439,
-19.984, -19.633, -19.293, -18.814, -24.105, -23.458, -22.648, -21.784, -21.033,
-20.444, -19.989, -19.636, -19.294, -18.815, -24.114, -23.467, -22.655, -21.790,
-21.038, -20.450, -19.994, -19.640, -19.296, -18.815, -24.124, -23.476, -22.663,
-21.796, -21.044, -20.456, -19.999, -19.643, -19.297, -18.816, -24.134, -23.485,
-22.671, -21.803, -21.051, -20.461, -20.004, -19.647, -19.298, -18.817, -24.143,
-23.494, -22.679, -21.810, -21.057, -20.468, -20.009, -19.650, -19.300, -18.819,
 -24.153, -23.504, -22.688, -21.818, -21.064, -20.474, -20.015, -19.654,
-19.302, -18.820, -24.163, -23.513, -22.696, -21.826, -21.071, -20.480, -20.021,
-19.658, -19.303, -18.821, -24.173, -23.523, -22.705, -21.834, -21.078, -20.487,
-20.026, -19.662, -19.305, -18.823, -24.182, -23.533, -22.714, -21.842, -21.086,
-20.494, -20.032, -19.666, -19.307, -18.824, -24.192, -23.542, -22.724, -21.850,
-21.094, -20.501, -20.038, -19.670, -19.310, -18.826, -24.202, -23.552, -22.733,
-21.859, -21.101, -20.508, -20.044, -19.674, -19.312, -18.827, -24.211, -23.562,
-22.743, -21.868, -21.110, -20.515, -20.050, -19.679, -19.314, -18.829, -24.221,
-23.572, -22.753, -21.877, -21.118, -20.523, -20.056, -19.683, -19.317, -18.831,
 -24.231, -23.582, -22.763, -21.887, -21.126, -20.530, -20.063, -19.688,
-19.319, -18.833, -24.241, -23.592, -22.773, -21.896, -21.135, -20.538, -20.069,
-19.692, -19.322, -18.835, -24.250, -23.602, -22.783, -21.906, -21.144, -20.545,
-20.075, -19.697, -19.325, -18.838, -24.260, -23.612, -22.793, -21.916, -21.152,
-20.553, -20.082, -19.701, -19.328, -18.840, -24.270, -23.622, -22.804, -21.926,
-21.161, -20.561, -20.088, -19.706, -19.331, -18.842, -24.280, -23.632, -22.814,
-21.936, -21.171, -20.569, -20.095, -19.711, -19.334, -18.845, -24.290, -23.643,
-22.825, -21.946, -21.180, -20.577, -20.101, -19.716, -19.337, -18.848, -24.299,
-23.653, -22.836, -21.956, -21.189, -20.585, -20.108, -19.721, -19.341, -18.850,
 -24.309, -23.663, -22.847, -21.967, -21.199, -20.593, -20.115, -19.726,
-19.344, -18.853, -24.319, -23.673, -22.858, -21.978, -21.208, -20.601, -20.121,
-19.731, -19.347, -18.856, -24.329, -23.684, -22.869, -21.988, -21.218, -20.610,
-20.128, -19.736, -19.351, -18.859, -24.338, -23.694, -22.880, -21.999, -21.228,
-20.618, -20.135, -19.741, -19.355, -18.862, -24.348, -23.705, -22.891, -22.010,
-21.238, -20.626, -20.142, -19.746, -19.358, -18.865, -24.358, -23.715, -22.902,
-22.021, -21.247, -20.635, -20.148, -19.751, -19.362, -18.869, -24.368, -23.725,
-22.913, -22.032, -21.258, -20.643, -20.155, -19.756, -19.366, -18.872, -24.377,
-23.736, -22.925, -22.044, -21.268, -20.652, -20.162, -19.762, -19.370, -18.875,
 -24.387, -23.746, -22.936, -22.055, -21.278, -20.660, -20.169, -19.767,
-19.374, -18.879, -24.397, -23.757, -22.948, -22.066, -21.288, -20.669, -20.176,
-19.772, -19.378, -18.883, -24.407, -23.767, -22.959, -22.078, -21.298, -20.678,
-20.183, -19.778, -19.382, -18.886, -24.416, -23.778, -22.971, -22.089, -21.309,
-20.686, -20.190, -19.783, -19.386, -18.890, -24.426, -23.788, -22.982, -22.101,
-21.319, -20.695, -20.197, -19.788, -19.391, -18.894, -24.436, -23.799, -22.994,
-22.112, -21.329, -20.704, -20.204, -19.794, -19.395, -18.898, -24.446, -23.809,
-23.005, -22.124, -21.340, -20.713, -20.211, -19.799, -19.399, -18.902, -24.455,
-23.820, -23.017, -22.136, -21.350, -20.722, -20.218, -19.805, -19.404, -18.906,
 -24.465, -23.830, -23.029, -22.147, -21.361, -20.730, -20.225, -19.810,
-19.408, -18.910, -24.475, -23.840, -23.040, -22.159, -21.372, -20.739, -20.232,
-19.816, -19.413, -18.914, -24.484, -23.851, -23.052, -22.171, -21.382, -20.748,
-20.239, -19.822, -19.417, -18.918, -24.494, -23.861, -23.064, -22.183, -21.393,
-20.757, -20.246, -19.827, -19.422, -18.922, -24.504, -23.872, -23.076, -22.195,
-21.404, -20.766, -20.253, -19.833, -19.427, -18.927, -24.514, -23.882, -23.087,
-22.207, -21.414, -20.775, -20.261, -19.839, -19.432, -18.931, -24.523, -23.893,
-23.099, -22.219, -21.425, -20.784, -20.268, -19.844, -19.436, -18.936, -24.533,
-23.903, -23.111, -22.231, -21.436, -20.793, -20.275, -19.850, -19.441, -18.940,
 -24.543, -23.914, -23.123, -22.242, -21.447, -20.802, -20.282, -19.856,
-19.446, -18.945, -24.552, -23.924, -23.134, -22.255, -21.458, -20.811, -20.290,
-19.862, -19.451, -18.949, -24.562, -23.935, -23.146, -22.267, -21.469, -20.820,
-20.297, -19.868, -19.456, -18.954, -24.572, -23.945, -23.158, -22.279, -21.480,
-20.830, -20.304, -19.873, -19.461, -18.959, -24.581, -23.956, -23.170, -22.291,
-21.490, -20.839, -20.311, -19.879, -19.466, -18.964, -24.591, -23.966, -23.181,
-22.303, -21.501, -20.848, -20.319, -19.885, -19.471, -18.968, -24.601, -23.976,
-23.193, -22.315, -21.512, -20.857, -20.326, -19.891, -19.477, -18.973, -24.610,
-23.987, -23.205, -22.327, -21.523, -20.866, -20.333, -19.897, -19.482, -18.978,
 -24.620, -23.997, -23.217, -22.339, -21.534, -20.876, -20.341, -19.903,
-19.487, -18.983, -24.630, -24.008, -23.228, -22.351, -21.546, -20.885, -20.348,
-19.909, -19.492, -18.988, -24.639, -24.018, -23.240, -22.363, -21.557, -20.894,
-20.356, -19.915, -19.498, -18.993, -24.649, -24.028, -23.252, -22.375, -21.568,
-20.904, -20.363, -19.921, -19.503, -18.998, -24.659, -24.039, -23.264, -22.388,
-21.579, -20.913, -20.371, -19.928, -19.509, -19.003, -24.668, -24.049, -23.275,
-22.400, -21.590, -20.922, -20.378, -19.934, -19.514, -19.008, -24.678, -24.059,
-23.287, -22.412, -21.601, -20.932, -20.386, -19.940, -19.519, -19.014, -24.687,
-24.070, -23.299, -22.424, -21.612, -20.941, -20.393, -19.946, -19.525, -19.019,
 -24.697, -24.080, -23.310, -22.436, -21.623, -20.950, -20.401, -19.952,
-19.530, -19.024, -24.707, -24.090, -23.322, -22.448, -21.634, -20.960, -20.408,
-19.959, -19.536, -19.029, -24.716, -24.101, -23.334, -22.460, -21.646, -20.969,
-20.416, -19.965, -19.542, -19.035, -24.726, -24.111, -23.345, -22.472, -21.657,
-20.979, -20.424, -19.971, -19.547, -19.040, -24.735, -24.121, -23.357, -22.485,
-21.668, -20.988, -20.431, -19.978, -19.553, -19.045, -24.745, -24.131, -23.369,
-22.497, -21.679, -20.998, -20.439, -19.984, -19.559, -19.051, -24.755, -24.142,
-23.380, -22.509, -21.690, -21.007, -20.447, -19.990, -19.564, -19.056, -24.764,
-24.152, -23.392, -22.521, -21.702, -21.017, -20.455, -19.997, -19.570, -19.062,
 -24.774, -24.162, -23.403, -22.533, -21.713, -21.026, -20.462, -20.003,
-19.576, -19.067, -24.783, -24.172, -23.415, -22.545, -21.724, -21.036, -20.470,
-20.010, -19.582, -19.073, -24.793, -24.182, -23.426, -22.557, -21.735, -21.045,
-20.478, -20.016, -19.588, -19.078, -24.802, -24.193, -23.438, -22.569, -21.746,
-21.055, -20.486, -20.023, -19.593, -19.084, -24.812, -24.203, -23.449, -22.581,
-21.758, -21.064, -20.493, -20.029, -19.599, -19.089, -24.821, -24.213, -23.461,
-22.593, -21.769, -21.074, -20.501, -20.036, -19.605, -19.095, -24.831, -24.223,
-23.472, -22.605, -21.780, -21.084, -20.509, -20.043, -19.611, -19.100, -24.840,
-24.233, -23.483, -22.617, -21.791, -21.093, -20.517, -20.049, -19.617, -19.106,
 -24.850, -24.243, -23.495, -22.629, -21.803, -21.103, -20.525, -20.056,
-19.623, -19.111, -24.860, -24.253, -23.506, -22.641, -21.814, -21.112, -20.533,
-20.063, -19.629, -19.117, -24.869, -24.264, -23.518, -22.653, -21.825, -21.122,
-20.541, -20.069, -19.635, -19.123, -24.879, -24.274, -23.529, -22.665, -21.836,
-21.132, -20.549, -20.076, -19.641, -19.128, -24.888, -24.284, -23.540, -22.677,
-21.847, -21.141, -20.557, -20.083, -19.647, -19.134, -24.898, -24.294, -23.552,
-22.689, -21.859, -21.151, -20.565, -20.090, -19.653, -19.140, -24.907, -24.304,
-23.563, -22.701, -21.870, -21.161, -20.573, -20.096, -19.659, -19.145, -24.917,
-24.314, -23.574, -22.713, -21.881, -21.171, -20.581, -20.103, -19.665, -19.151,
 -24.926, -24.324, -23.585, -22.725, -21.892, -21.180, -20.589, -20.110,
-19.672, -19.157, -24.936, -24.334, -23.596, -22.737, -21.904, -21.190, -20.597,
-20.117, -19.678, -19.162, -24.945, -24.344, -23.608, -22.749, -21.915, -21.200,
-20.605, -20.124, -19.684, -19.168, -24.954, -24.354, -23.619, -22.761, -21.926,
-21.210, -20.613, -20.131, -19.690, -19.174, -24.964, -24.364, -23.630, -22.773,
-21.937, -21.219, -20.621, -20.137, -19.696, -19.180, -24.973, -24.374, -23.641,
-22.784, -21.949, -21.229, -20.630, -20.144, -19.702, -19.185, -24.983, -24.384,
-23.652, -22.796, -21.960, -21.239, -20.638, -20.151, -19.709, -19.191, -24.992,
-24.394, -23.663, -22.808, -21.971, -21.249, -20.646, -20.158, -19.715, -19.197,
 -25.002, -24.404, -23.674, -22.820, -21.982, -21.258, -20.654, -20.165,
-19.721, -19.202, -25.011, -24.414, -23.685, -22.832, -21.993, -21.268, -20.662,
-20.172, -19.727, -19.208, -25.021, -24.424, -23.696, -22.843, -22.005, -21.278,
-20.671, -20.179, -19.733, -19.214, -25.030, -24.433, -23.707, -22.855, -22.016,
-21.288, -20.679, -20.186, -19.740, -19.220, -25.040, -24.443, -23.718, -22.867,
-22.027, -21.298, -20.687, -20.193, -19.746, -19.225, -25.049, -24.453, -23.729,
-22.878, -22.038, -21.308, -20.695, -20.200, -19.752, -19.231, -25.058, -24.463,
-23.740, -22.890, -22.049, -21.317, -20.704, -20.207, -19.758, -19.237, -25.068,
-24.473, -23.751, -22.902, -22.061, -21.327, -20.712, -20.214, -19.765, -19.242,
 -25.077, -24.483, -23.762, -22.913, -22.072, -21.337, -20.720, -20.222,
-19.771, -19.248, -25.087, -24.493, -23.773, -22.925, -22.083, -21.347, -20.729,
-20.229, -19.777, -19.254, -25.096, -24.502, -23.784, -22.937, -22.094, -21.357,
-20.737, -20.236, -19.783, -19.259, -25.105, -24.512, -23.795, -22.948, -22.105,
-21.367, -20.745, -20.243, -19.790, -19.265, -25.115, -24.522, -23.806, -22.960,
-22.116, -21.376, -20.754, -20.250, -19.796, -19.271, -25.124, -24.532, -23.816,
-22.971, -22.127, -21.386, -20.762, -20.257, -19.802, -19.276, -25.134, -24.542,
-23.827, -22.983, -22.139, -21.396, -20.770, -20.264, -19.808, -19.282, -25.143,
-24.552, -23.838, -22.994, -22.150, -21.406, -20.779, -20.271, -19.814, -19.288,
 -25.152, -24.561, -23.849, -23.006, -22.161, -21.416, -20.787, -20.278,
-19.821, -19.293, -25.162, -24.571, -23.859, -23.017, -22.172, -21.426, -20.796,
-20.286, -19.827, -19.299, -25.171, -24.581, -23.870, -23.029, -22.183, -21.436,
-20.804, -20.293, -19.833, -19.304, -25.181, -24.591, -23.881, -23.040, -22.194,
-21.446, -20.813, -20.300, -19.839, -19.310, -25.190, -24.600, -23.891, -23.052,
-22.205, -21.456, -20.821, -20.307, -19.846, -19.315, -25.199, -24.610, -23.902,
-23.063, -22.216, -21.465, -20.829, -20.314, -19.852, -19.321, -25.209, -24.620,
-23.913, -23.074, -22.227, -21.475, -20.838, -20.321, -19.858, -19.326, -25.218,
-24.629, -23.923, -23.086, -22.238, -21.485, -20.846, -20.329, -19.864, -19.332,
 -25.227, -24.639, -23.934, -23.097, -22.249, -21.495, -20.855, -20.336,
-19.870, -19.337, -25.237, -24.649, -23.945, -23.108, -22.260, -21.505, -20.863,
-20.343, -19.876, -19.343, -25.246, -24.659, -23.955, -23.120, -22.271, -21.515,
-20.872, -20.350, -19.883, -19.348, -25.255, -24.668, -23.966, -23.131, -22.282,
-21.525, -20.880, -20.357, -19.889, -19.353, -25.265, -24.678, -23.976, -23.142,
-22.293, -21.535, -20.889, -20.364, -19.895, -19.359, -25.274, -24.688, -23.987,
-23.154, -22.304, -21.545, -20.897, -20.372, -19.901, -19.364, -25.283, -24.697,
-23.997, -23.165, -22.315, -21.554, -20.906, -20.379, -19.907, -19.369, -25.293,
-24.707, -24.008, -23.176, -22.326, -21.564, -20.914, -20.386, -19.913, -19.375,
 -25.302, -24.716, -24.018, -23.187, -22.337, -21.574, -20.922, -20.393,
-19.919, -19.380, -25.311, -24.726, -24.029, -23.198, -22.348, -21.584, -20.931,
-20.400, -19.925, -19.385, -25.321, -24.736, -24.039, -23.209, -22.359, -21.594,
-20.939, -20.407, -19.931, -19.390, -25.330, -24.745, -24.049, -23.221, -22.370,
-21.604, -20.948, -20.415, -19.937, -19.396, -25.339, -24.755, -24.060, -23.232,
-22.381, -21.614, -20.956, -20.422, -19.943, -19.401, -25.348, -24.764, -24.070,
-23.243, -22.392, -21.624, -20.965, -20.429, -19.949, -19.406, -25.358, -24.774,
-24.080, -23.254, -22.403, -21.633, -20.973, -20.436, -19.955, -19.411, -25.367,
-24.784, -24.091, -23.265, -22.414, -21.643, -20.982, -20.443, -19.961, -19.416,
 -25.376, -24.793, -24.101, -23.276, -22.424, -21.653, -20.990, -20.450,
-19.967, -19.421, -25.386, -24.803, -24.111, -23.287, -22.435, -21.663, -20.999,
-20.457, -19.973, -19.426, -25.395, -24.812, -24.122, -23.298, -22.446, -21.673,
-21.007, -20.464, -19.978, -19.431, -25.404, -24.822, -24.132, -23.309, -22.457,
-21.683, -21.016, -20.471, -19.984, -19.436, -25.413, -24.831, -24.142, -23.320,
-22.468, -21.693, -21.024, -20.478, -19.990, -19.440, -25.423, -24.841, -24.152,
-23.331, -22.479, -21.702, -21.033, -20.485, -19.996, -19.445, -25.432, -24.850,
-24.163, -23.342, -22.489, -21.712, -21.041, -20.492, -20.001, -19.450, -25.441,
-24.860, -24.173, -23.353, -22.500, -21.722, -21.050, -20.499, -20.007, -19.455,
 -25.450, -24.869, -24.183, -23.364, -22.511, -21.732, -21.058, -20.506,
-20.013, -19.459, -25.460, -24.879, -24.193, -23.374, -22.522, -21.742, -21.066,
-20.513, -20.018, -19.464, -25.469, -24.888, -24.203, -23.385, -22.532, -21.752,
-21.075, -20.520, -20.024, -19.469, -25.478, -24.898, -24.214, -23.396, -22.543,
-21.761, -21.083, -20.527, -20.029, -19.473, -25.487, -24.907, -24.224, -23.407,
-22.554, -21.771, -21.092, -20.534, -20.035, -19.478, -25.497, -24.917, -24.234,
-23.418, -22.564, -21.781, -21.100, -20.541, -20.040, -19.482, -25.506, -24.926,
-24.244, -23.428, -22.575, -21.791, -21.108, -20.548, -20.046, -19.486, -25.515,
-24.936, -24.254, -23.439, -22.586, -21.800, -21.117, -20.555, -20.051, -19.491,
 -25.524, -24.945, -24.264, -23.450, -22.596, -21.810, -21.125, -20.562,
-20.057, -19.495, -25.534, -24.955, -24.274, -23.461, -22.607, -21.820, -21.134,
-20.568, -20.062, -19.499, -25.543, -24.964, -24.284, -23.471, -22.618, -21.830,
-21.142, -20.575, -20.067, -19.504, -25.552, -24.974, -24.294, -23.482, -22.628,
-21.839, -21.150, -20.582, -20.072, -19.508, -25.561, -24.983, -24.304, -23.493,
-22.639, -21.849, -21.159, -20.589, -20.077, -19.512, -25.570, -24.992, -24.314,
-23.503, -22.649, -21.859, -21.167, -20.595, -20.083, -19.516, -25.580, -25.002,
-24.324, -23.514, -22.660, -21.869, -21.175, -20.602, -20.088, -19.520, -25.589,
-25.011, -24.334, -23.525, -22.671, -21.878, -21.183, -20.609, -20.093, -19.524,
 -25.598, -25.021, -24.344, -23.535, -22.681, -21.888, -21.192, -20.615,
-20.098, -19.528, -25.607, -25.030, -24.354, -23.546, -22.692, -21.898, -21.200,
-20.622, -20.103, -19.531, -25.616, -25.039, -24.364, -23.556, -22.702, -21.907,
-21.208, -20.628, -20.107, -19.535, -25.626, -25.049, -24.374, -23.567, -22.713,
-21.917, -21.216, -20.635, -20.112, -19.539, -25.635, -25.058, -24.384, -23.577,
-22.723, -21.926, -21.225, -20.641, -20.117, -19.543, -25.644, -25.068, -24.394,
-23.588, -22.734, -21.936, -21.233, -20.648, -20.122, -19.546, -25.653, -25.077,
-24.404, -23.598, -22.744, -21.946, -21.241, -20.654, -20.126, -19.550, -25.662,
-25.086, -24.414, -23.609, -22.755, -21.955, -21.249, -20.660, -20.131, -19.553,
 -25.671, -25.096, -24.424, -23.619, -22.765, -21.965, -21.257, -20.667,
-20.136, -19.556, -25.681, -25.105, -24.433, -23.630, -22.775, -21.974, -21.265,
-20.673, -20.140, -19.560, -25.690, -25.114, -24.443, -23.640, -22.786, -21.984,
-21.273, -20.679, -20.145, -19.563, -25.699, -25.124, -24.453, -23.651, -22.796,
-21.993, -21.281, -20.685, -20.149, -19.566, -25.708, -25.133, -24.463, -23.661,
-22.806, -22.003, -21.289, -20.692, -20.153, -19.569, -25.717, -25.142, -24.473,
-23.671, -22.817, -22.012, -21.297, -20.698, -20.157, -19.572, -25.726, -25.152,
-24.482, -23.682, -22.827, -22.022, -21.305, -20.704, -20.162, -19.575, -25.736,
-25.161, -24.492, -23.692, -22.837, -22.031, -21.313, -20.710, -20.166, -19.578,
 -25.745, -25.170, -24.502, -23.703, -22.848, -22.041, -21.321, -20.716,
-20.170, -19.581, -25.754, -25.180, -24.512, -23.713, -22.858, -22.050, -21.329,
-20.722, -20.174, -19.584, -25.763, -25.189, -24.522, -23.723, -22.868, -22.060,
-21.337, -20.727, -20.178, -19.587, -25.772, -25.198, -24.531, -23.733, -22.879,
-22.069, -21.345, -20.733, -20.182, -19.589, -25.781, -25.207, -24.541, -23.744,
-22.889, -22.078, -21.352, -20.739, -20.185, -19.592, -25.790, -25.217, -24.551,
-23.754, -22.899, -22.088, -21.360, -20.745, -20.189, -19.594, -25.799, -25.226,
-24.560, -23.764, -22.909, -22.097, -21.368, -20.750, -20.193, -19.597, -25.809,
-25.235, -24.570, -23.774, -22.919, -22.107, -21.375, -20.756, -20.196, -19.599,
 -25.818, -25.245, -24.580, -23.785, -22.930, -22.116, -21.383, -20.762,
-20.200, -19.601, -25.827, -25.254, -24.589, -23.795, -22.940, -22.125, -21.391,
-20.767, -20.203, -19.604, -25.836, -25.263, -24.599, -23.805, -22.950, -22.134,
-21.398, -20.773, -20.207, -19.606, -25.845, -25.272, -24.609, -23.815, -22.960,
-22.144, -21.406, -20.778, -20.210, -19.608, -25.854, -25.282, -24.618, -23.825,
-22.970, -22.153, -21.413, -20.783, -20.213, -19.610, -25.863, -25.291, -24.628,
-23.835, -22.980, -22.162, -21.421, -20.789, -20.216, -19.612, -25.872, -25.300,
-24.638, -23.845, -22.990, -22.171, -21.428, -20.794, -20.219, -19.613, -25.881,
-25.309, -24.647, -23.856, -23.000, -22.180, -21.436, -20.799, -20.222, -19.615,
 -25.891, -25.318, -24.657, -23.866, -23.010, -22.189, -21.443, -20.804,
-20.225, -19.617, -25.900, -25.328, -24.666, -23.876, -23.020, -22.198, -21.450,
-20.809, -20.228, -19.618, -25.909, -25.337, -24.676, -23.886, -23.030, -22.208,
-21.458, -20.814, -20.231, -19.620, -25.918, -25.346, -24.686, -23.896, -23.040,
-22.217, -21.465, -20.819, -20.233, -19.621, -25.927, -25.355, -24.695, -23.906,
-23.050, -22.226, -21.472, -20.824, -20.236, -19.623, -25.936, -25.364, -24.705,
-23.916, -23.060, -22.235, -21.479, -20.828, -20.238, -19.624, -25.945, -25.374,
-24.714, -23.926, -23.070, -22.244, -21.486, -20.833, -20.241, -19.625, -25.954,
-25.383, -24.724, -23.936, -23.080, -22.252, -21.493, -20.838, -20.243, -19.626,
 -25.963, -25.392, -24.733, -23.946, -23.090, -22.261, -21.500, -20.842,
-20.245, -19.627, -25.972, -25.401, -24.743, -23.956, -23.100, -22.270, -21.507,
-20.846, -20.247, -19.628, -25.981, -25.410, -24.752, -23.966, -23.110, -22.279,
-21.514, -20.851, -20.249, -19.629, -25.990, -25.420, -24.762, -23.976, -23.120,
-22.288, -21.521, -20.855, -20.251, -19.629, -25.999, -25.429, -24.771, -23.986,
-23.130, -22.297, -21.527, -20.859, -20.253, -19.630, -26.008, -25.438, -24.781,
-23.995, -23.139, -22.305, -21.534, -20.863, -20.255, -19.631, -26.018, -25.447,
-24.790, -24.005, -23.149, -22.314, -21.541, -20.867, -20.257, -19.631, -26.027,
-25.456, -24.800, -24.015, -23.159, -22.323, -21.547, -20.871, -20.258, -19.631,
 -26.036, -25.465, -24.809, -24.025, -23.169, -22.331, -21.554, -20.875,
-20.260, -19.632, -26.045, -25.474, -24.819, -24.035, -23.178, -22.340, -21.560,
-20.879, -20.261, -19.632, -26.054, -25.484, -24.828, -24.045, -23.188, -22.349,
-21.567, -20.883, -20.262, -19.632, -26.063, -25.493, -24.837, -24.054, -23.198,
-22.357, -21.573, -20.886, -20.264, -19.632, -26.072, -25.502, -24.847, -24.064,
-23.207, -22.366, -21.579, -20.890, -20.265, -19.632, -26.081, -25.511, -24.856,
-24.074, -23.217, -22.374, -21.585, -20.893, -20.266, -19.632, -26.090, -25.520,
-24.866, -24.084, -23.227, -22.382, -21.592, -20.897, -20.267, -19.632, -26.099,
-25.529, -24.875, -24.094, -23.236, -22.391, -21.598, -20.900, -20.267, -19.631,
 -26.108, -25.538, -24.884, -24.103, -23.246, -22.399, -21.604, -20.903,
-20.268, -19.631, -26.117, -25.547, -24.894, -24.113, -23.255, -22.407, -21.610,
-20.906, -20.269, -19.630, -26.126, -25.557, -24.903, -24.123, -23.265, -22.416,
-21.615, -20.909, -20.269, -19.630, -26.135, -25.566, -24.912, -24.132, -23.274,
-22.424, -21.621, -20.912, -20.270, -19.629, -26.144, -25.575, -24.922, -24.142,
-23.284, -22.432, -21.627, -20.915, -20.270, -19.628, -26.153, -25.584, -24.931,
-24.152, -23.293, -22.440, -21.633, -20.918, -20.270, -19.627, -26.162, -25.593,
-24.941, -24.161, -23.303, -22.448, -21.638, -20.920, -20.271, -19.627, -26.171,
-25.602, -24.950, -24.171, -23.312, -22.456, -21.644, -20.923, -20.271, -19.626,
 -26.180, -25.611, -24.959, -24.181, -23.321, -22.464, -21.649, -20.925,
-20.271, -19.624, -26.189, -25.620, -24.968, -24.190, -23.331, -22.472, -21.654,
-20.928, -20.270, -19.623, -26.198, -25.629, -24.978, -24.200, -23.340, -22.480,
-21.659, -20.930, -20.270, -19.622, -26.207, -25.638, -24.987, -24.210, -23.349,
-22.487, -21.665, -20.932, -20.270, -19.621, -26.216, -25.647, -24.996, -24.219,
-23.359, -22.495, -21.670, -20.934, -20.269, -19.619, -26.225, -25.656, -25.006,
-24.229, -23.368, -22.503, -21.675, -20.936, -20.269, -19.618, -26.234, -25.665,
-25.015, -24.238, -23.377, -22.510, -21.679, -20.938, -20.268, -19.616, -26.243,
-25.674, -25.024, -24.248, -23.386, -22.518, -21.684, -20.940, -20.268, -19.614,
 -26.252, -25.683, -25.033, -24.257, -23.396, -22.526, -21.689, -20.941,
-20.267, -19.612, -26.261, -25.693, -25.043, -24.267, -23.405, -22.533, -21.694,
-20.943, -20.266, -19.611, -26.270, -25.702, -25.052, -24.276, -23.414, -22.540,
-21.698, -20.944, -20.265, -19.609, -26.279, -25.711, -25.061, -24.286, -23.423,
-22.548, -21.702, -20.945, -20.264, -19.607, -26.288, -25.720, -25.070, -24.295,
-23.432, -22.555, -21.707, -20.947, -20.263, -19.605, -26.297, -25.729, -25.080,
-24.305, -23.441, -22.562, -21.711, -20.948, -20.261, -19.602, -26.306, -25.738,
-25.089, -24.314, -23.450, -22.569, -21.715, -20.949, -20.260, -19.600, -26.315,
-25.747, -25.098, -24.324, -23.459, -22.576, -21.719, -20.950, -20.258, -19.598,
 -26.324, -25.756, -25.107, -24.333, -23.468, -22.583, -21.723, -20.950,
-20.257, -19.595, -26.333, -25.765, -25.116, -24.342, -23.477, -22.590, -21.727,
-20.951, -20.255, -19.593, -26.342, -25.774, -25.126, -24.352, -23.485, -22.597,
-21.731, -20.952, -20.253, -19.590, -26.351, -25.783, -25.135, -24.361, -23.494,
-22.604, -21.734, -20.952, -20.252, -19.588, -26.360, -25.792, -25.144, -24.370,
-23.503, -22.610, -21.738, -20.952, -20.250, -19.585, -26.369, -25.801, -25.153,
-24.380, -23.512, -22.617, -21.741, -20.953, -20.248, -19.582, -26.377, -25.810,
-25.162, -24.389, -23.520, -22.623, -21.744, -20.953, -20.245, -19.579, -26.386,
-25.819, -25.171, -24.398, -23.529, -22.630, -21.747, -20.953, -20.243, -19.576,
 -26.395, -25.828, -25.180, -24.408, -23.538, -22.636, -21.750, -20.953,
-20.241, -19.573, -26.404, -25.837, -25.190, -24.417, -23.546, -22.642, -21.753,
-20.952, -20.238, -19.570, -26.413, -25.846, -25.199, -24.426, -23.555, -22.648,
-21.756, -20.952, -20.236, -19.567, -26.422, -25.855, -25.208, -24.436, -23.563,
-22.654, -21.759, -20.952, -20.233, -19.564, -26.431, -25.864, -25.217, -24.445,
-23.572, -22.660, -21.761, -20.951, -20.231, -19.560, -26.440, -25.873, -25.226,
-24.454, -23.580, -22.666, -21.764, -20.951, -20.228, -19.557, -26.449, -25.882,
-25.235, -24.463, -23.588, -22.672, -21.766, -20.950, -20.225, -19.554, -26.458,
-25.890, -25.244, -24.472, -23.597, -22.678, -21.768, -20.949, -20.222, -19.550,
 -26.467, -25.899, -25.253, -24.481, -23.605, -22.683, -21.771, -20.948,
-20.219, -19.547, -26.476, -25.908, -25.262, -24.491, -23.613, -22.689, -21.773,
-20.947, -20.216, -19.543, -26.485, -25.917, -25.272, -24.500, -23.621, -22.694,
-21.774, -20.946, -20.213, -19.540, -26.494, -25.926, -25.281, -24.509, -23.629,
-22.699, -21.776, -20.944, -20.210, -19.536, -26.502, -25.935, -25.290, -24.518,
-23.637, -22.704, -21.778, -20.943, -20.207, -19.532, -26.511, -25.944, -25.299,
-24.527, -23.645, -22.709, -21.779, -20.942, -20.203, -19.528, -26.520, -25.953,
-25.308, -24.536, -23.653, -22.714, -21.780, -20.940, -20.200, -19.524, -26.529,
-25.962, -25.317, -24.545, -23.661, -22.719, -21.782, -20.938, -20.196, -19.521,
 -26.538, -25.971, -25.326, -24.554, -23.669, -22.724, -21.783, -20.937,
-20.193, -19.517, -26.547, -25.980, -25.335, -24.563, -23.676, -22.729, -21.784,
-20.935, -20.189, -19.513, -26.556, -25.989, -25.344, -24.572, -23.684, -22.733,
-21.785, -20.933, -20.185, -19.509, -26.565, -25.998, -25.353, -24.581, -23.692,
-22.737, -21.785, -20.930, -20.181, -19.505, -26.574, -26.007, -25.362, -24.590,
-23.699, -22.742, -21.786, -20.928, -20.178, -19.500, -26.583, -26.016, -25.371,
-24.599, -23.707, -22.746, -21.786, -20.926, -20.174, -19.496, -26.591, -26.024,
-25.380, -24.608, -23.714, -22.750, -21.787, -20.924, -20.170, -19.492, -26.600,
-26.033, -25.389, -24.617, -23.721, -22.753, -21.787, -20.921, -20.166, -19.488,
 -26.609, -26.042, -25.398, -24.625, -23.729, -22.757, -21.787, -20.918,
-20.162, -19.484, -26.618, -26.051, -25.407, -24.634, -23.736, -22.761, -21.787,
-20.916, -20.157, -19.479, -26.627, -26.060, -25.416, -24.643, -23.743, -22.764,
-21.786, -20.913, -20.153, -19.475, -26.636, -26.069, -25.425, -24.652, -23.750,
-22.767, -21.786, -20.910, -20.149, -19.471, -26.645, -26.078, -25.434, -24.661,
-23.757, -22.770, -21.785, -20.907, -20.145, -19.467, -26.654, -26.087, -25.443,
-24.669, -23.764, -22.773, -21.785, -20.904, -20.140, -19.462, -26.663, -26.096,
-25.452, -24.678, -23.770, -22.776, -21.784, -20.901, -20.136, -19.458, -26.671,
-26.104, -25.461, -24.687, -23.777, -22.779, -21.783, -20.897, -20.131, -19.453,
 -26.680, -26.113, -25.469, -24.695, -23.784, -22.782, -21.782, -20.894,
-20.127, -19.449, -26.689, -26.122, -25.478, -24.704, -23.790, -22.784, -21.781,
-20.891, -20.122, -19.445, -26.698, -26.131, -25.487, -24.712, -23.797, -22.786,
-21.780, -20.887, -20.118, -19.440, -26.707, -26.140, -25.496, -24.721, -23.803,
-22.788, -21.778, -20.883, -20.113, -19.436, -26.716, -26.149, -25.505, -24.730,
-23.809, -22.790, -21.776, -20.880, -20.108, -19.431, -26.725, -26.158, -25.514,
-24.738, -23.815, -22.792, -21.775, -20.876, -20.104, -19.427, -26.733, -26.167,
-25.523, -24.747, -23.821, -22.794, -21.773, -20.872, -20.099, -19.423, -26.742,
-26.175, -25.532, -24.755, -23.827, -22.795, -21.771, -20.868, -20.094, -19.418,
 -26.751, -26.184, -25.541, -24.763, -23.833, -22.797, -21.769, -20.864,
-20.089, -19.414, -26.760, -26.193, -25.549, -24.772, -23.838, -22.798, -21.766,
-20.860, -20.084, -19.409, -26.769, -26.202, -25.558, -24.780, -23.844, -22.799,
-21.764, -20.856, -20.080, -19.405, -26.778, -26.211, -25.567, -24.788, -23.849,
-22.799, -21.761, -20.851, -20.075, -19.401, -26.786, -26.220, -25.576, -24.797,
-23.855, -22.800, -21.759, -20.847, -20.070, -19.396, -26.795, -26.228, -25.585,
-24.805, -23.860, -22.801, -21.756, -20.843, -20.065, -19.392, -26.804, -26.237,
-25.594, -24.813, -23.865, -22.801, -21.753, -20.838, -20.060, -19.388, -26.813,
-26.246, -25.602, -24.821, -23.870, -22.801, -21.750, -20.833, -20.055, -19.384,
 -26.822, -26.255, -25.611, -24.829, -23.875, -22.801, -21.746, -20.829,
-20.050, -19.379, -26.831, -26.264, -25.620, -24.837, -23.879, -22.801, -21.743,
-20.824, -20.045, -19.375, -26.839, -26.273, -25.629, -24.845, -23.884, -22.800,
-21.739, -20.819, -20.040, -19.371, -26.848, -26.281, -25.638, -24.853, -23.888,
-22.800, -21.736, -20.814, -20.035, -19.367, -26.857, -26.290, -25.646, -24.861,
-23.892, -22.799, -21.732, -20.809, -20.030, -19.363, -26.866, -26.299, -25.655,
-24.869, -23.897, -22.798, -21.728, -20.804, -20.025, -19.359, -26.875, -26.308,
-25.664, -24.877, -23.900, -22.797, -21.724, -20.799, -20.020, -19.355, -26.884,
-26.317, -25.673, -24.884, -23.904, -22.796, -21.720, -20.794, -20.015, -19.351,
 -26.892, -26.325, -25.681, -24.892, -23.908, -22.794, -21.716, -20.789,
-20.010, -19.347, -26.901, -26.334, -25.690, -24.900, -23.911, -22.793, -21.711,
-20.784, -20.005, -19.344, -26.910, -26.343, -25.699, -24.907, -23.915, -22.791,
-21.707, -20.778, -20.000, -19.340, -26.919, -26.352, -25.708, -24.915, -23.918,
-22.789, -21.702, -20.773, -19.995, -19.336,
};

const double CIA_Orton::N2CH4[][10] = {
 -22.850, -23.681, -24.411, -25.086, -25.706, -26.257, -26.758, -27.207,
-27.615, -27.985, -21.017, -21.849, -22.578, -23.253, -23.873, -24.425, -24.925,
-25.374, -25.782, -26.153, -19.631, -20.463, -21.192, -21.867, -22.487, -23.038,
-23.539, -23.988, -24.396, -24.766, -18.246, -19.077, -19.806, -20.481, -21.101,
-21.652, -22.153, -22.602, -23.010, -23.380, -17.436, -18.266, -18.995, -19.670,
-20.290, -20.841, -21.342, -21.791, -22.199, -22.569, -16.862, -17.692, -18.421,
-19.095, -19.715, -20.266, -20.767, -21.216, -21.624, -21.994, -16.418, -17.247,
-17.975, -18.650, -19.269, -19.820, -20.321, -20.769, -21.177, -21.548, -16.055,
-16.884, -17.612, -18.286, -18.905, -19.456, -19.956, -20.405, -20.813, -21.183,
 -15.750, -16.577, -17.305, -17.978, -18.597, -19.148, -19.648, -20.097,
-20.505, -20.875, -15.486, -16.312, -17.039, -17.712, -18.331, -18.881, -19.382,
-19.830, -20.238, -20.608, -15.253, -16.079, -16.805, -17.478, -18.096, -18.646,
-19.147, -19.595, -20.002, -20.373, -15.047, -15.871, -16.596, -17.268, -17.886,
-18.436, -18.936, -19.384, -19.792, -20.162, -14.860, -15.683, -16.407, -17.079,
-17.696, -18.246, -18.746, -19.194, -19.602, -19.972, -14.691, -15.512, -16.235,
-16.906, -17.523, -18.073, -18.573, -19.020, -19.428, -19.798, -14.536, -15.355,
-16.077, -16.748, -17.364, -17.914, -18.413, -18.861, -19.268, -19.638, -14.393,
-15.210, -15.931, -16.601, -17.217, -17.766, -18.266, -18.713, -19.120, -19.490,
 -14.260, -15.076, -15.796, -16.465, -17.081, -17.629, -18.128, -18.576,
-18.983, -19.352, -14.137, -14.951, -15.670, -16.338, -16.953, -17.501, -18.000,
-18.447, -18.854, -19.224, -14.022, -14.834, -15.551, -16.218, -16.833, -17.381,
-17.879, -18.326, -18.733, -19.103, -13.915, -14.724, -15.440, -16.106, -16.720,
-17.268, -17.766, -18.213, -18.619, -18.989, -13.813, -14.620, -15.335, -16.000,
-16.614, -17.161, -17.658, -18.105, -18.511, -18.881, -13.718, -14.523, -15.236,
-15.900, -16.513, -17.059, -17.557, -18.003, -18.409, -18.779, -13.311, -14.103,
-14.808, -15.467, -16.076, -16.620, -17.115, -17.560, -17.966, -18.334, -12.991,
-13.769, -14.465, -15.117, -15.722, -16.263, -16.757, -17.200, -17.604, -17.972,
 -12.730, -13.494, -14.181, -14.827, -15.427, -15.964, -16.455, -16.897,
-17.300, -17.666, -12.513, -13.263, -13.940, -14.579, -15.174, -15.708, -16.196,
-16.636, -17.037, -17.403, -12.326, -13.064, -13.732, -14.364, -14.954, -15.484,
-15.969, -16.407, -16.807, -17.171, -12.163, -12.890, -13.550, -14.176, -14.760,
-15.286, -15.768, -16.203, -16.601, -16.964, -12.018, -12.736, -13.389, -14.008,
-14.587, -15.109, -15.588, -16.021, -16.416, -16.778, -11.885, -12.597, -13.244,
-13.857, -14.431, -14.949, -15.425, -15.855, -16.249, -16.608, -11.761, -12.469,
-13.112, -13.721, -14.290, -14.804, -15.276, -15.704, -16.095, -16.453, -11.645,
-12.351, -12.992, -13.596, -14.161, -14.671, -15.140, -15.565, -15.954, -16.311,
 -11.533, -12.241, -12.880, -13.481, -14.042, -14.549, -15.015, -15.437,
-15.824, -16.178, -11.425, -12.137, -12.777, -13.375, -13.933, -14.436, -14.899,
-15.318, -15.703, -16.055, -11.319, -12.038, -12.679, -13.277, -13.831, -14.331,
-14.791, -15.207, -15.590, -15.940, -11.216, -11.943, -12.588, -13.184, -13.736,
-14.234, -14.690, -15.104, -15.484, -15.833, -11.116, -11.852, -12.501, -13.097,
-13.647, -14.142, -14.596, -15.007, -15.385, -15.731, -11.019, -11.765, -12.418,
-13.015, -13.563, -14.056, -14.507, -14.916, -15.291, -15.636, -10.839, -11.602,
-12.264, -12.863, -13.409, -13.898, -14.344, -14.748, -15.119, -15.460, -10.681,
-11.454, -12.123, -12.725, -13.269, -13.755, -14.198, -14.598, -14.964, -15.301,
 -10.546, -11.320, -11.993, -12.597, -13.141, -13.625, -14.065, -14.461,
-14.824, -15.157, -10.432, -11.199, -11.873, -12.477, -13.022, -13.505, -13.942,
-14.336, -14.695, -15.025, -10.328, -11.085, -11.758, -12.364, -12.909, -13.393,
-13.829, -14.221, -14.577, -14.903, -10.225, -10.976, -11.649, -12.256, -12.803,
-13.288, -13.724, -14.113, -14.468, -14.791, -10.119, -10.868, -11.542, -12.152,
-12.701, -13.188, -13.624, -14.013, -14.365, -14.686, -10.009, -10.761, -11.438,
-12.051, -12.603, -13.092, -13.530, -13.919, -14.270, -14.588, -9.900, -10.655,
-11.335, -11.952, -12.508, -13.001, -13.440, -13.830, -14.180, -14.496, -9.796,
-10.552, -11.235, -11.855, -12.416, -12.912, -13.354, -13.745, -14.094, -14.409,
  -9.699, -10.451, -11.136, -11.761, -12.325, -12.826, -13.271, -13.664,
-14.013, -14.327, -9.606, -10.352, -11.038, -11.668, -12.237, -12.743, -13.191,
-13.586, -13.936, -14.249, -9.510, -10.253, -10.942, -11.577, -12.152, -12.662,
-13.114, -13.511, -13.862, -14.175, -9.408, -10.153, -10.848, -11.488, -12.068,
-12.584, -13.039, -13.439, -13.791, -14.104, -9.301, -10.054, -10.757, -11.403,
-11.988, -12.507, -12.967, -13.369, -13.723, -14.036, -9.195,  -9.961, -10.670,
-11.321, -11.910, -12.434, -12.896, -13.301, -13.657, -13.970, -9.102,  -9.876,
-10.590, -11.244, -11.836, -12.362, -12.828, -13.236, -13.593, -13.907, -9.031,
-9.806, -10.519, -11.172, -11.764, -12.293, -12.761, -13.172, -13.531, -13.846,
  -8.986,  -9.750, -10.456, -11.106, -11.697, -12.226, -12.697, -13.109,
-13.470, -13.786, -8.964,  -9.707, -10.401, -11.044, -11.632, -12.162, -12.634,
-13.049, -13.412, -13.729, -8.955,  -9.673, -10.352, -10.986, -11.571, -12.100,
-12.573, -12.990, -13.354, -13.673, -8.948,  -9.643, -10.306, -10.932, -11.512,
-12.039, -12.513, -12.932, -13.298, -13.619, -8.937,  -9.613, -10.263, -10.880,
-11.455, -11.981, -12.455, -12.875, -13.244, -13.566, -8.920,  -9.583, -10.221,
-10.829, -11.399, -11.924, -12.398, -12.820, -13.191, -13.514, -8.900,  -9.551,
-10.179, -10.779, -11.345, -11.868, -12.343, -12.766, -13.138, -13.463, -8.880,
-9.519, -10.137, -10.731, -11.293, -11.814, -12.289, -12.713, -13.087, -13.414,
  -8.864,  -9.488, -10.095, -10.683, -11.241, -11.761, -12.236, -12.662,
-13.037, -13.365, -8.849,  -9.456, -10.054, -10.636, -11.191, -11.709, -12.184,
-12.611, -12.988, -13.318, -8.831,  -9.422, -10.012, -10.589, -11.142, -11.659,
-12.134, -12.562, -12.940, -13.271, -8.805,  -9.386,  -9.970, -10.544, -11.094,
-11.610, -12.085, -12.514, -12.893, -13.226, -8.770,  -9.348,  -9.929, -10.501,
-11.049, -11.563, -12.037, -12.466, -12.847, -13.181, -8.732,  -9.311,  -9.891,
-10.460, -11.006, -11.518, -11.991, -12.420, -12.802, -13.137, -8.701,  -9.280,
-9.857, -10.423, -10.965, -11.475, -11.947, -12.375, -12.758, -13.094, -8.684,
-9.259,  -9.830, -10.390, -10.927, -11.434, -11.903, -12.332, -12.714, -13.052,
  -8.686,  -9.249,  -9.810, -10.361, -10.893, -11.395, -11.862, -12.289,
-12.672, -13.010, -8.706,  -9.250,  -9.796, -10.337, -10.861, -11.358, -11.822,
-12.248, -12.630, -12.969, -8.736,  -9.258,  -9.788, -10.317, -10.831, -11.322,
-11.783, -12.207, -12.589, -12.929, -8.768,  -9.270,  -9.783, -10.299, -10.804,
-11.288, -11.745, -12.168, -12.550, -12.890, -8.797,  -9.283,  -9.780, -10.282,
-10.778, -11.256, -11.709, -12.129, -12.511, -12.851, -8.819,  -9.294,  -9.777,
-10.268, -10.754, -11.225, -11.673, -12.092, -12.472, -12.813, -8.837,  -9.303,
-9.774, -10.253, -10.730, -11.195, -11.639, -12.055, -12.435, -12.776, -8.854,
-9.312,  -9.771, -10.239, -10.708, -11.166, -11.606, -12.019, -12.398, -12.739,
  -8.872,  -9.320,  -9.767, -10.225, -10.686, -11.138, -11.574, -11.985,
-12.362, -12.703, -8.890,  -9.327,  -9.763, -10.211, -10.664, -11.111, -11.542,
-11.951, -12.327, -12.668, -8.905,  -9.333,  -9.757, -10.197, -10.643, -11.084,
-11.512, -11.918, -12.293, -12.633, -8.913,  -9.335,  -9.751, -10.184, -10.623,
-11.059, -11.483, -11.886, -12.259, -12.599, -8.915,  -9.334,  -9.744, -10.171,
-10.604, -11.035, -11.454, -11.855, -12.226, -12.566, -8.911,  -9.333,  -9.738,
-10.159, -10.587, -11.012, -11.427, -11.824, -12.194, -12.533, -8.908,  -9.334,
-9.735, -10.150, -10.571, -10.991, -11.401, -11.795, -12.163, -12.501, -8.911,
-9.340,  -9.735, -10.144, -10.558, -10.971, -11.376, -11.766, -12.132, -12.469,
  -8.923,  -9.354,  -9.741, -10.140, -10.546, -10.952, -11.352, -11.739,
-12.103, -12.438, -8.945,  -9.375,  -9.751, -10.140, -10.537, -10.936, -11.330,
-11.712, -12.073, -12.408, -8.974,  -9.403,  -9.765, -10.143, -10.530, -10.920,
-11.308, -11.687, -12.045, -12.378, -9.007,  -9.434,  -9.782, -10.148, -10.524,
-10.906, -11.288, -11.662, -12.017, -12.349, -9.042,  -9.467,  -9.801, -10.155,
-10.520, -10.893, -11.268, -11.638, -11.990, -12.320, -9.076,  -9.502,  -9.820,
-10.162, -10.517, -10.881, -11.250, -11.614, -11.964, -12.292, -9.111,  -9.537,
-9.839, -10.170, -10.514, -10.870, -11.232, -11.592, -11.938, -12.265, -9.148,
-9.574,  -9.858, -10.178, -10.512, -10.859, -11.215, -11.570, -11.913, -12.238,
  -9.186,  -9.613,  -9.877, -10.186, -10.510, -10.849, -11.198, -11.549,
-11.889, -12.211, -9.225,  -9.654,  -9.896, -10.194, -10.509, -10.840, -11.182,
-11.528, -11.865, -12.185, -9.265,  -9.697,  -9.913, -10.202, -10.508, -10.831,
-11.167, -11.508, -11.842, -12.160, -9.304,  -9.741,  -9.930, -10.210, -10.507,
-10.822, -11.153, -11.489, -11.819, -12.135, -9.344,  -9.784,  -9.947, -10.218,
-10.507, -10.815, -11.139, -11.471, -11.797, -12.111, -9.382,  -9.826,  -9.964,
-10.226, -10.508, -10.808, -11.127, -11.453, -11.776, -12.087, -9.420,  -9.867,
-9.982, -10.236, -10.510, -10.803, -11.115, -11.436, -11.755, -12.064, -9.457,
-9.906, -10.003, -10.248, -10.513, -10.798, -11.104, -11.420, -11.735, -12.042,
  -9.494,  -9.943, -10.026, -10.262, -10.519, -10.795, -11.094, -11.405,
-11.716, -12.019, -9.528,  -9.978, -10.052, -10.278, -10.525, -10.793, -11.085,
-11.390, -11.697, -11.998, -9.560, -10.010, -10.081, -10.296, -10.533, -10.792,
-11.076, -11.376, -11.679, -11.977, -9.589, -10.039, -10.111, -10.314, -10.543,
-10.792, -11.069, -11.363, -11.661, -11.956, -9.614, -10.064, -10.141, -10.334,
-10.553, -10.793, -11.062, -11.350, -11.644, -11.936, -9.633, -10.085, -10.171,
-10.354, -10.564, -10.794, -11.056, -11.338, -11.628, -11.916, -9.645, -10.101,
-10.200, -10.373, -10.575, -10.796, -11.050, -11.326, -11.611, -11.897, -9.651,
-10.112, -10.228, -10.392, -10.586, -10.798, -11.045, -11.315, -11.596, -11.878,
  -9.649, -10.119, -10.254, -10.411, -10.597, -10.801, -11.040, -11.305,
-11.581, -11.860, -9.640, -10.120, -10.279, -10.428, -10.608, -10.804, -11.036,
-11.295, -11.566, -11.842, -9.625, -10.118, -10.303, -10.445, -10.619, -10.807,
-11.033, -11.285, -11.552, -11.825, -9.603, -10.113, -10.324, -10.461, -10.631,
-10.811, -11.030, -11.276, -11.539, -11.808, -9.577, -10.103, -10.344, -10.477,
-10.642, -10.814, -11.027, -11.268, -11.526, -11.791, -9.548, -10.091, -10.361,
-10.493, -10.654, -10.819, -11.025, -11.260, -11.513, -11.775, -9.518, -10.078,
-10.376, -10.510, -10.667, -10.824, -11.023, -11.253, -11.501, -11.760, -9.490,
-10.066, -10.390, -10.527, -10.681, -10.830, -11.022, -11.246, -11.490, -11.744,
  -9.473, -10.055, -10.402, -10.546, -10.696, -10.836, -11.022, -11.240,
-11.479, -11.730, -9.454, -10.047, -10.414, -10.567, -10.712, -10.844, -11.023,
-11.234, -11.468, -11.715, -9.443, -10.043, -10.425, -10.588, -10.730, -10.852,
-11.024, -11.229, -11.458, -11.701, -9.440, -10.042, -10.437, -10.611, -10.750,
-10.861, -11.026, -11.225, -11.449, -11.688, -9.444, -10.045, -10.449, -10.634,
-10.770, -10.870, -11.028, -11.221, -11.440, -11.675, -9.455, -10.052, -10.461,
-10.658, -10.790, -10.880, -11.031, -11.217, -11.431, -11.662, -9.471, -10.061,
-10.474, -10.681, -10.812, -10.890, -11.034, -11.214, -11.423, -11.650, -9.490,
-10.080, -10.486, -10.704, -10.834, -10.901, -11.037, -11.211, -11.415, -11.638,
  -9.511, -10.093, -10.498, -10.725, -10.857, -10.911, -11.041, -11.209,
-11.407, -11.626, -9.532, -10.105, -10.509, -10.745, -10.879, -10.922, -11.045,
-11.207, -11.400, -11.615, -9.553, -10.118, -10.519, -10.764, -10.902, -10.933,
-11.049, -11.205, -11.394, -11.604, -9.573, -10.129, -10.528, -10.780, -10.924,
-10.944, -11.054, -11.204, -11.387, -11.594, -9.595, -10.139, -10.535, -10.795,
-10.946, -10.956, -11.059, -11.203, -11.381, -11.584, -9.612, -10.147, -10.540,
-10.808, -10.966, -10.967, -11.064, -11.202, -11.376, -11.574, -9.626, -10.152,
-10.544, -10.819, -10.986, -10.979, -11.070, -11.202, -11.371, -11.565, -9.638,
-10.155, -10.545, -10.828, -11.004, -10.992, -11.076, -11.203, -11.366, -11.556,
  -9.647, -10.156, -10.553, -10.835, -11.021, -11.005, -11.082, -11.204,
-11.362, -11.547, -9.654, -10.154, -10.551, -10.841, -11.036, -11.019, -11.089,
-11.205, -11.358, -11.539, -9.658, -10.151, -10.547, -10.845, -11.051, -11.034,
-11.097, -11.206, -11.354, -11.531, -9.659, -10.148, -10.542, -10.847, -11.064,
-11.049, -11.105, -11.208, -11.351, -11.524, -9.661, -10.141, -10.537, -10.849,
-11.075, -11.064, -11.113, -11.211, -11.348, -11.517, -9.657, -10.133, -10.531,
-10.850, -11.086, -11.081, -11.122, -11.213, -11.346, -11.510, -9.652, -10.124,
-10.524, -10.851, -11.096, -11.097, -11.131, -11.216, -11.344, -11.503, -9.644,
-10.115, -10.518, -10.851, -11.105, -11.114, -11.140, -11.220, -11.342, -11.497,
  -9.635, -10.106, -10.513, -10.851, -11.113, -11.131, -11.150, -11.223,
-11.340, -11.491, -9.624, -10.097, -10.507, -10.852, -11.121, -11.149, -11.159,
-11.227, -11.339, -11.485, -9.611, -10.088, -10.503, -10.859, -11.128, -11.167,
-11.170, -11.231, -11.338, -11.480, -9.595, -10.081, -10.499, -10.860, -11.135,
-11.184, -11.181, -11.235, -11.338, -11.475, -9.585, -10.072, -10.497, -10.861,
-11.142, -11.201, -11.191, -11.240, -11.338, -11.470, -9.568, -10.065, -10.495,
-10.862, -11.148, -11.218, -11.203, -11.245, -11.338, -11.466, -9.552, -10.059,
-10.494, -10.864, -11.154, -11.235, -11.214, -11.250, -11.338, -11.462, -9.539,
-10.055, -10.494, -10.867, -11.160, -11.251, -11.226, -11.256, -11.339, -11.458,
  -9.531, -10.054, -10.496, -10.870, -11.166, -11.266, -11.239, -11.262,
-11.340, -11.454, -9.531, -10.056, -10.499, -10.874, -11.171, -11.282, -11.252,
-11.268, -11.341, -11.451, -9.538, -10.063, -10.504, -10.878, -11.177, -11.296,
-11.265, -11.274, -11.342, -11.448, -9.553, -10.078, -10.511, -10.883, -11.182,
-11.309, -11.279, -11.280, -11.344, -11.446, -9.578, -10.092, -10.521, -10.888,
-11.187, -11.322, -11.294, -11.287, -11.346, -11.443, -9.606, -10.109, -10.530,
-10.893, -11.192, -11.335, -11.308, -11.294, -11.348, -11.441, -9.638, -10.127,
-10.541, -10.900, -11.204, -11.346, -11.323, -11.301, -11.351, -11.439, -9.672,
-10.148, -10.552, -10.906, -11.208, -11.357, -11.338, -11.309, -11.354, -11.437,
  -9.706, -10.169, -10.563, -10.912, -11.212, -11.366, -11.353, -11.316,
-11.356, -11.436, -9.739, -10.189, -10.574, -10.917, -11.216, -11.376, -11.368,
-11.324, -11.360, -11.435, -9.771, -10.209, -10.585, -10.922, -11.219, -11.384,
-11.383, -11.331, -11.363, -11.434, -9.800, -10.231, -10.600, -10.927, -11.221,
-11.392, -11.398, -11.339, -11.366, -11.433, -9.831, -10.247, -10.608, -10.931,
-11.223, -11.398, -11.412, -11.347, -11.370, -11.432, -9.853, -10.262, -10.616,
-10.934, -11.224, -11.404, -11.426, -11.355, -11.374, -11.432, -9.871, -10.274,
-10.621, -10.937, -11.225, -11.409, -11.439, -11.363, -11.378, -11.432, -9.885,
-10.283, -10.625, -10.938, -11.225, -11.413, -11.452, -11.372, -11.382, -11.432,
  -9.894, -10.288, -10.627, -10.937, -11.224, -11.417, -11.464, -11.380,
-11.387, -11.433, -9.898, -10.290, -10.626, -10.935, -11.222, -11.419, -11.476,
-11.389, -11.392, -11.433, -9.898, -10.288, -10.623, -10.931, -11.220, -11.421,
-11.487, -11.397, -11.396, -11.434, -9.892, -10.287, -10.619, -10.926, -11.216,
-11.422, -11.497, -11.406, -11.401, -11.435, -9.884, -10.277, -10.610, -10.919,
-11.211, -11.423, -11.507, -11.415, -11.406, -11.437, -9.869, -10.262, -10.599,
-10.911, -11.206, -11.429, -11.516, -11.423, -11.412, -11.438, -9.849, -10.244,
-10.585, -10.905, -11.200, -11.428, -11.524, -11.432, -11.417, -11.440, -9.824,
-10.222, -10.569, -10.894, -11.193, -11.427, -11.531, -11.442, -11.423, -11.441,
  -9.796, -10.198, -10.551, -10.882, -11.185, -11.425, -11.538, -11.451,
-11.429, -11.443, -9.764, -10.171, -10.532, -10.870, -11.178, -11.423, -11.544,
-11.460, -11.434, -11.446, -9.729, -10.142, -10.513, -10.857, -11.171, -11.420,
-11.550, -11.469, -11.440, -11.448, -9.693, -10.115, -10.498, -10.845, -11.164,
-11.418, -11.555, -11.479, -11.446, -11.451, -9.660, -10.088, -10.479, -10.833,
-11.157, -11.416, -11.560, -11.488, -11.453, -11.453, -9.627, -10.063, -10.462,
-10.823, -11.151, -11.413, -11.565, -11.498, -11.459, -11.456, -9.600, -10.043,
-10.448, -10.817, -11.146, -11.412, -11.569, -11.508, -11.465, -11.459, -9.580,
-10.028, -10.438, -10.810, -11.142, -11.410, -11.573, -11.518, -11.472, -11.463,
  -9.570, -10.019, -10.432, -10.806, -11.139, -11.409, -11.577, -11.528,
-11.479, -11.466, -9.570, -10.018, -10.430, -10.804, -11.137, -11.409, -11.580,
-11.538, -11.486, -11.470, -9.581, -10.024, -10.433, -10.805, -11.139, -11.409,
-11.584, -11.549, -11.493, -11.473, -9.603, -10.040, -10.442, -10.809, -11.141,
-11.410, -11.588, -11.559, -11.500, -11.477, -9.634, -10.059, -10.454, -10.815,
-11.143, -11.411, -11.591, -11.569, -11.507, -11.482, -9.673, -10.085, -10.469,
-10.824, -11.147, -11.413, -11.595, -11.579, -11.515, -11.486, -9.717, -10.115,
-10.488, -10.838, -11.152, -11.415, -11.605, -11.590, -11.523, -11.490, -9.765,
-10.149, -10.510, -10.851, -11.158, -11.418, -11.608, -11.600, -11.531, -11.495,
  -9.815, -10.184, -10.533, -10.865, -11.165, -11.421, -11.612, -11.610,
-11.538, -11.499, -9.865, -10.222, -10.557, -10.879, -11.172, -11.425, -11.615,
-11.619, -11.546, -11.504, -9.914, -10.259, -10.582, -10.894, -11.181, -11.428,
-11.618, -11.629, -11.554, -11.509, -9.963, -10.296, -10.610, -10.909, -11.188,
-11.432, -11.621, -11.638, -11.563, -11.515, -10.008, -10.332, -10.634, -10.923,
-11.196, -11.435, -11.624, -11.647, -11.571, -11.520, -10.051, -10.365, -10.656,
-10.937, -11.203, -11.438, -11.627, -11.655, -11.579, -11.525, -10.091, -10.396,
-10.677, -10.952, -11.210, -11.442, -11.629, -11.664, -11.588, -11.531, -10.128,
-10.424, -10.697, -10.964, -11.216, -11.444, -11.632, -11.672, -11.597, -11.536,
 -10.159, -10.448, -10.714, -10.974, -11.221, -11.447, -11.634, -11.679,
-11.605, -11.542, -10.186, -10.469, -10.728, -10.982, -11.226, -11.451, -11.636,
-11.687, -11.614, -11.548, -10.208, -10.486, -10.739, -10.989, -11.234, -11.453,
-11.638, -11.693, -11.623, -11.555, -10.224, -10.498, -10.749, -10.995, -11.237,
-11.455, -11.640, -11.700, -11.632, -11.561, -10.234, -10.506, -10.755, -10.999,
-11.239, -11.456, -11.641, -11.706, -11.641, -11.567, -10.239, -10.510, -10.758,
-11.001, -11.240, -11.457, -11.642, -11.712, -11.650, -11.574, -10.238, -10.510,
-10.758, -11.005, -11.241, -11.458, -11.644, -11.717, -11.659, -11.581, -10.232,
-10.506, -10.756, -11.004, -11.241, -11.458, -11.645, -11.722, -11.668, -11.587,
 -10.222, -10.499, -10.752, -11.003, -11.242, -11.459, -11.645, -11.727,
-11.677, -11.594, -10.207, -10.489, -10.747, -11.001, -11.242, -11.461, -11.646,
-11.732, -11.686, -11.601, -10.190, -10.478, -10.741, -10.999, -11.244, -11.461,
-11.648, -11.736, -11.695, -11.609, -10.171, -10.465, -10.736, -10.997, -11.244,
-11.462, -11.649, -11.747, -11.704, -11.616, -10.152, -10.453, -10.730, -10.995,
-11.244, -11.462, -11.650, -11.751, -11.712, -11.623, -10.134, -10.443, -10.726,
-10.994, -11.245, -11.463, -11.651, -11.754, -11.721, -11.631, -10.121, -10.436,
-10.723, -10.996, -11.246, -11.465, -11.652, -11.758, -11.729, -11.639, -10.114,
-10.433, -10.724, -10.998, -11.249, -11.467, -11.653, -11.761, -11.737, -11.646,
 -10.113, -10.434, -10.727, -11.002, -11.252, -11.469, -11.654, -11.764,
-11.745, -11.654, -10.120, -10.441, -10.734, -11.008, -11.256, -11.475, -11.656,
-11.768, -11.753, -11.662, -10.133, -10.453, -10.744, -11.016, -11.265, -11.478,
-11.659, -11.771, -11.761, -11.671, -10.154, -10.471, -10.758, -11.025, -11.272,
-11.481, -11.661, -11.773, -11.768, -11.679, -10.183, -10.493, -10.775, -11.037,
-11.279, -11.486, -11.663, -11.776, -11.775, -11.687, -10.215, -10.519, -10.795,
-11.051, -11.287, -11.490, -11.666, -11.779, -11.782, -11.695, -10.250, -10.548,
-10.817, -11.068, -11.297, -11.495, -11.668, -11.782, -11.789, -11.704, -10.288,
-10.580, -10.841, -11.084, -11.307, -11.501, -11.671, -11.785, -11.795, -11.712,
 -10.328, -10.613, -10.866, -11.101, -11.317, -11.506, -11.673, -11.787,
-11.801, -11.720, -10.369, -10.647, -10.892, -11.119, -11.328, -11.513, -11.676,
-11.790, -11.807, -11.729, -10.410, -10.682, -10.918, -11.137, -11.340, -11.519,
-11.680, -11.792, -11.813, -11.737, -10.451, -10.717, -10.945, -11.155, -11.351,
-11.525, -11.683, -11.795, -11.818, -11.745, -10.491, -10.751, -10.971, -11.173,
-11.362, -11.531, -11.685, -11.797, -11.823, -11.753, -10.531, -10.785, -10.996,
-11.191, -11.373, -11.537, -11.688, -11.799, -11.828, -11.762, -10.569, -10.817,
-11.021, -11.207, -11.383, -11.543, -11.691, -11.801, -11.833, -11.769, -10.608,
-10.847, -11.043, -11.223, -11.393, -11.549, -11.693, -11.803, -11.837, -11.777,
 -10.643, -10.875, -11.064, -11.237, -11.402, -11.554, -11.696, -11.805,
-11.841, -11.785, -10.676, -10.901, -11.083, -11.249, -11.410, -11.562, -11.698,
-11.807, -11.845, -11.792, -10.706, -10.924, -11.099, -11.262, -11.420, -11.566,
-11.703, -11.808, -11.855, -11.800, -10.733, -10.944, -11.113, -11.271, -11.426,
-11.570, -11.704, -11.810, -11.859, -11.807, -10.758, -10.962, -11.125, -11.280,
-11.432, -11.573, -11.706, -11.811, -11.862, -11.814, -10.778, -10.976, -11.135,
-11.287, -11.436, -11.576, -11.708, -11.812, -11.865, -11.820, -10.795, -10.987,
-11.142, -11.292, -11.440, -11.578, -11.709, -11.813, -11.868, -11.827, -10.809,
-10.995, -11.147, -11.296, -11.443, -11.581, -11.710, -11.815, -11.870, -11.833,
 -10.818, -11.001, -11.150, -11.298, -11.446, -11.583, -11.711, -11.816,
-11.873, -11.839, -10.824, -11.002, -11.152, -11.300, -11.448, -11.586, -11.712,
-11.817, -11.875, -11.845, -10.826, -11.002, -11.152, -11.301, -11.451, -11.587,
-11.714, -11.817, -11.877, -11.850, -10.825, -11.000, -11.151, -11.302, -11.452,
-11.589, -11.715, -11.818, -11.879, -11.855, -10.821, -10.997, -11.150, -11.303,
-11.454, -11.590, -11.716, -11.818, -11.880, -11.860, -10.816, -10.993, -11.149,
-11.304, -11.456, -11.592, -11.717, -11.818, -11.882, -11.865, -10.810, -10.989,
-11.148, -11.305, -11.458, -11.593, -11.717, -11.819, -11.883, -11.870, -10.804,
-10.986, -11.148, -11.307, -11.460, -11.595, -11.718, -11.820, -11.884, -11.874,
 -10.798, -10.985, -11.150, -11.310, -11.463, -11.597, -11.719, -11.820,
-11.886, -11.878, -10.792, -10.985, -11.153, -11.314, -11.467, -11.601, -11.719,
-11.820, -11.886, -11.882, -10.786, -10.986, -11.158, -11.319, -11.472, -11.604,
-11.723, -11.820, -11.887, -11.886, -10.781, -10.988, -11.165, -11.326, -11.477,
-11.606, -11.724, -11.820, -11.888, -11.889, -10.780, -10.992, -11.172, -11.333,
-11.483, -11.609, -11.724, -11.820, -11.888, -11.892, -10.776, -10.998, -11.181,
-11.341, -11.488, -11.612, -11.725, -11.820, -11.889, -11.895, -10.773, -11.004,
-11.191, -11.350, -11.494, -11.615, -11.727, -11.819, -11.889, -11.898, -10.771,
-11.011, -11.201, -11.360, -11.500, -11.618, -11.727, -11.821, -11.890, -11.901,
 -10.772, -11.020, -11.213, -11.370, -11.507, -11.620, -11.728, -11.821,
-11.890, -11.903, -10.774, -11.030, -11.225, -11.380, -11.513, -11.624, -11.728,
-11.820, -11.890, -11.905, -10.780, -11.041, -11.237, -11.390, -11.520, -11.627,
-11.730, -11.820, -11.889, -11.907, -10.788, -11.054, -11.250, -11.400, -11.525,
-11.629, -11.730, -11.819, -11.889, -11.909, -10.800, -11.067, -11.263, -11.410,
-11.531, -11.632, -11.730, -11.818, -11.889, -11.911, -10.815, -11.085, -11.276,
-11.420, -11.536, -11.634, -11.730, -11.817, -11.888, -11.913, -10.833, -11.101,
-11.289, -11.429, -11.541, -11.635, -11.730, -11.818, -11.889, -11.920, -10.855,
-11.118, -11.303, -11.437, -11.545, -11.636, -11.729, -11.817, -11.888, -11.922,
 -10.879, -11.137, -11.316, -11.444, -11.548, -11.637, -11.729, -11.815,
-11.888, -11.923, -10.906, -11.156, -11.328, -11.452, -11.550, -11.638, -11.728,
-11.814, -11.887, -11.924, -10.934, -11.176, -11.341, -11.457, -11.552, -11.638,
-11.729, -11.813, -11.887, -11.925, -10.965, -11.197, -11.352, -11.462, -11.553,
-11.638, -11.728, -11.812, -11.886, -11.925, -10.997, -11.217, -11.363, -11.467,
-11.554, -11.637, -11.727, -11.811, -11.885, -11.926, -11.029, -11.238, -11.374,
-11.470, -11.554, -11.636, -11.726, -11.810, -11.884, -11.927, -11.061, -11.258,
-11.383, -11.473, -11.554, -11.635, -11.724, -11.811, -11.884, -11.927, -11.093,
-11.277, -11.392, -11.475, -11.553, -11.634, -11.723, -11.810, -11.884, -11.928,
 -11.125, -11.296, -11.402, -11.477, -11.553, -11.634, -11.722, -11.809,
-11.883, -11.928, -11.157, -11.314, -11.410, -11.479, -11.553, -11.633, -11.721,
-11.808, -11.882, -11.929, -11.186, -11.332, -11.418, -11.481, -11.553, -11.632,
-11.722, -11.807, -11.882, -11.929, -11.215, -11.348, -11.425, -11.483, -11.553,
-11.632, -11.721, -11.807, -11.881, -11.930, -11.242, -11.364, -11.432, -11.485,
-11.553, -11.632, -11.721, -11.806, -11.881, -11.930, -11.269, -11.380, -11.440,
-11.489, -11.555, -11.633, -11.721, -11.806, -11.881, -11.930, -11.294, -11.396,
-11.449, -11.493, -11.557, -11.634, -11.722, -11.807, -11.882, -11.931, -11.318,
-11.413, -11.459, -11.499, -11.561, -11.637, -11.723, -11.807, -11.882, -11.931,
 -11.342, -11.429, -11.469, -11.506, -11.565, -11.640, -11.724, -11.808,
-11.882, -11.931, -11.363, -11.446, -11.481, -11.514, -11.570, -11.644, -11.726,
-11.809, -11.882, -11.931, -11.383, -11.463, -11.493, -11.523, -11.577, -11.648,
-11.730, -11.810, -11.882, -11.932, -11.407, -11.479, -11.506, -11.533, -11.585,
-11.653, -11.733, -11.811, -11.882, -11.932, -11.423, -11.494, -11.520, -11.545,
-11.594, -11.659, -11.736, -11.813, -11.882, -11.932, -11.437, -11.508, -11.534,
-11.557, -11.603, -11.665, -11.740, -11.814, -11.883, -11.933, -11.448, -11.521,
-11.547, -11.571, -11.613, -11.672, -11.744, -11.818, -11.884, -11.933, -11.456,
-11.533, -11.561, -11.584, -11.623, -11.679, -11.748, -11.820, -11.885, -11.933,
 -11.462, -11.543, -11.574, -11.597, -11.634, -11.687, -11.752, -11.822,
-11.885, -11.934, -11.466, -11.551, -11.587, -11.610, -11.645, -11.694, -11.757,
-11.824, -11.886, -11.934, -11.467, -11.558, -11.598, -11.623, -11.656, -11.702,
-11.762, -11.827, -11.887, -11.935, -11.466, -11.564, -11.609, -11.636, -11.667,
-11.710, -11.767, -11.829, -11.888, -11.935, -11.463, -11.571, -11.620, -11.649,
-11.679, -11.718, -11.772, -11.832, -11.889, -11.936, -11.459, -11.575, -11.631,
-11.661, -11.690, -11.727, -11.777, -11.834, -11.890, -11.936, -11.454, -11.577,
-11.641, -11.673, -11.700, -11.735, -11.782, -11.838, -11.893, -11.936, -11.449,
-11.580, -11.649, -11.684, -11.711, -11.743, -11.787, -11.841, -11.894, -11.937,
 -11.444, -11.582, -11.658, -11.695, -11.721, -11.750, -11.792, -11.843,
-11.895, -11.937, -11.440, -11.585, -11.666, -11.706, -11.730, -11.758, -11.797,
-11.846, -11.896, -11.938, -11.438, -11.588, -11.675, -11.716, -11.740, -11.765,
-11.803, -11.849, -11.898, -11.938, -11.437, -11.593, -11.683, -11.726, -11.748,
-11.772, -11.808, -11.852, -11.899, -11.939, -11.439, -11.599, -11.692, -11.736,
-11.757, -11.779, -11.813, -11.855, -11.900, -11.939, -11.443, -11.607, -11.702,
-11.745, -11.766, -11.785, -11.818, -11.858, -11.902, -11.940, -11.450, -11.615,
-11.712, -11.755, -11.774, -11.792, -11.823, -11.862, -11.904, -11.940, -11.460,
-11.626, -11.722, -11.765, -11.782, -11.799, -11.827, -11.865, -11.905, -11.941,
 -11.472, -11.639, -11.736, -11.774, -11.791, -11.806, -11.832, -11.868,
-11.907, -11.941, -11.487, -11.653, -11.748, -11.785, -11.800, -11.812, -11.837,
-11.871, -11.908, -11.942, -11.505, -11.669, -11.762, -11.796, -11.808, -11.819,
-11.843, -11.874, -11.910, -11.943, -11.525, -11.687, -11.776, -11.808, -11.817,
-11.826, -11.848, -11.878, -11.912, -11.944, -11.546, -11.706, -11.792, -11.820,
-11.827, -11.834, -11.853, -11.881, -11.913, -11.945, -11.570, -11.727, -11.808,
-11.833, -11.837, -11.841, -11.859, -11.885, -11.915, -11.945, -11.597, -11.750,
-11.826, -11.847, -11.848, -11.850, -11.865, -11.889, -11.919, -11.946, -11.624,
-11.774, -11.846, -11.862, -11.859, -11.858, -11.871, -11.893, -11.921, -11.946,
 -11.651, -11.799, -11.866, -11.878, -11.871, -11.867, -11.877, -11.896,
-11.923, -11.947, -11.679, -11.825, -11.887, -11.895, -11.884, -11.877, -11.883,
-11.900, -11.924, -11.948, -11.708, -11.852, -11.910, -11.912, -11.897, -11.887,
-11.891, -11.904, -11.926, -11.949, -11.745, -11.879, -11.933, -11.931, -11.911,
-11.897, -11.898, -11.909, -11.929, -11.950, -11.774, -11.907, -11.957, -11.949,
-11.926, -11.908, -11.905, -11.913, -11.931, -11.950, -11.804, -11.935, -11.981,
-11.969, -11.941, -11.919, -11.913, -11.917, -11.934, -11.951, -11.833, -11.963,
-12.005, -11.990, -11.956, -11.930, -11.920, -11.923, -11.937, -11.952, -11.863,
-11.990, -12.029, -12.010, -11.972, -11.941, -11.928, -11.928, -11.939, -11.953,
 -11.892, -12.018, -12.052, -12.030, -11.987, -11.953, -11.935, -11.932,
-11.941, -11.953, -11.926, -12.046, -12.076, -12.050, -12.003, -11.964, -11.943,
-11.937, -11.943, -11.954, -11.954, -12.072, -12.099, -12.069, -12.019, -11.976,
-11.951, -11.941, -11.946, -11.956, -11.981, -12.098, -12.121, -12.088, -12.034,
-11.987, -11.959, -11.946, -11.948, -11.957, -12.008, -12.129, -12.142, -12.107,
-12.049, -11.999, -11.966, -11.950, -11.950, -11.958, -12.034, -12.152, -12.163,
-12.125, -12.065, -12.010, -11.974, -11.955, -11.952, -11.959, -12.059, -12.174,
-12.183, -12.143, -12.080, -12.021, -11.981, -11.960, -11.957, -11.959, -12.082,
-12.195, -12.202, -12.159, -12.094, -12.031, -11.988, -11.964, -11.959, -11.960,
 -12.105, -12.215, -12.220, -12.176, -12.107, -12.041, -11.996, -11.969,
-11.961, -11.961, -12.126, -12.233, -12.237, -12.191, -12.120, -12.051, -12.002,
-11.973, -11.963, -11.961, -12.147, -12.251, -12.253, -12.206, -12.132, -12.061,
-12.009, -11.977, -11.965, -11.963, -12.166, -12.267, -12.267, -12.219, -12.144,
-12.070, -12.015, -11.981, -11.967, -11.964, -12.184, -12.285, -12.282, -12.232,
-12.155, -12.078, -12.022, -11.985, -11.969, -11.964, -12.200, -12.299, -12.295,
-12.245, -12.166, -12.087, -12.028, -11.989, -11.971, -11.965, -12.215, -12.311,
-12.307, -12.257, -12.176, -12.095, -12.034, -11.994, -11.974, -11.966, -12.230,
-12.323, -12.325, -12.269, -12.187, -12.103, -12.040, -11.998, -11.976, -11.966,
 -12.243, -12.335, -12.338, -12.280, -12.197, -12.111, -12.046, -12.001,
-11.978, -11.967, -12.255, -12.346, -12.349, -12.291, -12.207, -12.119, -12.051,
-12.005, -11.980, -11.968, -12.268, -12.357, -12.360, -12.302, -12.216, -12.127,
-12.057, -12.009, -11.982, -11.970, -12.280, -12.368, -12.371, -12.313, -12.226,
-12.135, -12.063, -12.012, -11.984, -11.971, -12.292, -12.379, -12.383, -12.324,
-12.236, -12.143, -12.069, -12.016, -11.986, -11.971, -12.304, -12.390, -12.394,
-12.335, -12.246, -12.151, -12.075, -12.020, -11.988, -11.972, -12.320, -12.403,
-12.406, -12.347, -12.257, -12.159, -12.081, -12.024, -11.992, -11.973, -12.333,
-12.416, -12.419, -12.359, -12.267, -12.168, -12.087, -12.028, -11.994, -11.973,
 -12.347, -12.429, -12.432, -12.372, -12.279, -12.177, -12.093, -12.032,
-11.996, -11.974, -12.361, -12.443, -12.446, -12.385, -12.290, -12.186, -12.100,
-12.035, -11.998, -11.974, -12.380, -12.458, -12.461, -12.399, -12.303, -12.196,
-12.106, -12.040, -11.999, -11.976, -12.395, -12.473, -12.478, -12.413, -12.315,
-12.205, -12.112, -12.043, -12.001, -11.977, -12.411, -12.489, -12.494, -12.428,
-12.327, -12.214, -12.119, -12.047, -12.003, -11.977, -12.426, -12.506, -12.510,
-12.448, -12.340, -12.224, -12.126, -12.051, -12.005, -11.978, -12.442, -12.523,
-12.527, -12.463, -12.353, -12.233, -12.132, -12.056, -12.008, -11.978, -12.458,
-12.541, -12.544, -12.479, -12.366, -12.243, -12.139, -12.060, -12.010, -11.979,
 -12.474, -12.558, -12.561, -12.494, -12.379, -12.253, -12.145, -12.063,
-12.012, -11.980, -12.489, -12.579, -12.579, -12.510, -12.391, -12.262, -12.151,
-12.067, -12.014, -11.980, -12.503, -12.596, -12.597, -12.526, -12.404, -12.272,
-12.158, -12.071, -12.016, -11.983, -12.517, -12.618, -12.614, -12.542, -12.417,
-12.281, -12.164, -12.074, -12.017, -11.984, -12.529, -12.635, -12.631, -12.557,
-12.429, -12.290, -12.170, -12.078, -12.019, -11.985, -12.540, -12.652, -12.650,
-12.572, -12.441, -12.299, -12.175, -12.082, -12.021, -11.985, -12.551, -12.668,
-12.666, -12.587, -12.453, -12.307, -12.181, -12.086, -12.024, -11.986, -12.560,
-12.683, -12.683, -12.601, -12.464, -12.315, -12.187, -12.089, -12.026, -11.987,
 -12.567, -12.698, -12.698, -12.614, -12.475, -12.323, -12.192, -12.092,
-12.028, -11.987, -12.574, -12.712, -12.714, -12.627, -12.485, -12.330, -12.197,
-12.095, -12.030, -11.989, -12.580, -12.725, -12.728, -12.640, -12.494, -12.337,
-12.202, -12.099, -12.031, -11.989, -12.590, -12.737, -12.742, -12.651, -12.504,
-12.344, -12.206, -12.102, -12.033, -11.990, -12.594, -12.749, -12.755, -12.662,
-12.512, -12.351, -12.211, -12.105, -12.035, -11.991, -12.596, -12.759, -12.767,
-12.674, -12.521, -12.357, -12.216, -12.108, -12.037, -11.992, -12.598, -12.769,
-12.778, -12.684, -12.532, -12.364, -12.220, -12.111, -12.039, -11.992, -12.599,
-12.778, -12.794, -12.693, -12.539, -12.369, -12.224, -12.115, -12.041, -11.993,
 -12.600, -12.786, -12.805, -12.702, -12.546, -12.375, -12.229, -12.118,
-12.043, -11.994, -12.600, -12.794, -12.814, -12.710, -12.553, -12.381, -12.234,
-12.121, -12.045, -11.996, -12.601, -12.801, -12.822, -12.718, -12.560, -12.387,
-12.239, -12.124, -12.047, -11.997, -12.602, -12.807, -12.830, -12.725, -12.567,
-12.393, -12.243, -12.128, -12.049, -11.998, -12.603, -12.813, -12.837, -12.733,
-12.574, -12.399, -12.248, -12.131, -12.051, -11.999, -12.610, -12.819, -12.844,
-12.740, -12.581, -12.405, -12.253, -12.135, -12.054, -12.000, -12.612, -12.828,
-12.850, -12.746, -12.588, -12.411, -12.259, -12.139, -12.057, -12.001, -12.615,
-12.834, -12.856, -12.753, -12.595, -12.418, -12.264, -12.143, -12.059, -12.003,
 -12.618, -12.839, -12.862, -12.761, -12.602, -12.425, -12.270, -12.147,
-12.062, -12.004, -12.623, -12.844, -12.868, -12.768, -12.610, -12.433, -12.276,
-12.152, -12.064, -12.006, -12.630, -12.850, -12.875, -12.776, -12.619, -12.440,
-12.282, -12.156, -12.067, -12.007, -12.637, -12.856, -12.881, -12.783, -12.627,
-12.448, -12.289, -12.161, -12.070, -12.009, -12.645, -12.863, -12.888, -12.791,
-12.636, -12.456, -12.296, -12.166, -12.073, -12.010, -12.654, -12.870, -12.895,
-12.803, -12.645, -12.465, -12.302, -12.171, -12.076, -12.012, -12.665, -12.879,
-12.903, -12.811, -12.654, -12.474, -12.310, -12.176, -12.080, -12.013, -12.677,
-12.888, -12.911, -12.820, -12.663, -12.483, -12.317, -12.181, -12.083, -12.015,
 -12.691, -12.903, -12.920, -12.829, -12.673, -12.492, -12.324, -12.186,
-12.086, -12.016, -12.705, -12.914, -12.930, -12.839, -12.684, -12.502, -12.332,
-12.191, -12.089, -12.019, -12.721, -12.926, -12.941, -12.851, -12.694, -12.511,
-12.339, -12.197, -12.093, -12.021, -12.739, -12.941, -12.952, -12.862, -12.705,
-12.523, -12.347, -12.202, -12.096, -12.023, -12.757, -12.955, -12.964, -12.873,
-12.717, -12.533, -12.355, -12.208, -12.100, -12.025, -12.776, -12.970, -12.979,
-12.885, -12.728, -12.543, -12.362, -12.213, -12.104, -12.027, -12.797, -12.986,
-12.993, -12.898, -12.740, -12.553, -12.371, -12.219, -12.107, -12.029, -12.818,
-13.003, -13.008, -12.911, -12.751, -12.562, -12.378, -12.225, -12.111, -12.031,
 -12.840, -13.020, -13.024, -12.924, -12.763, -12.572, -12.386, -12.230,
-12.115, -12.033, -12.863, -13.039, -13.040, -12.938, -12.775, -12.582, -12.394,
-12.236, -12.118, -12.035, -12.886, -13.058, -13.056, -12.953, -12.787, -12.592,
-12.402, -12.242, -12.122, -12.037, -12.910, -13.078, -13.074, -12.967, -12.800,
-12.602, -12.410, -12.247, -12.126, -12.040, -12.934, -13.098, -13.091, -12.982,
-12.812, -12.612, -12.417, -12.253, -12.130, -12.042, -12.958, -13.118, -13.109,
-12.998, -12.824, -12.622, -12.425, -12.259, -12.134, -12.044, -12.983, -13.138,
-13.130, -13.013, -12.840, -12.633, -12.433, -12.265, -12.138, -12.046, -13.008,
-13.159, -13.153, -13.029, -12.853, -12.643, -12.441, -12.271, -12.142, -12.049,
 -13.032, -13.180, -13.171, -13.045, -12.866, -12.653, -12.448, -12.277,
-12.146, -12.051, -13.057, -13.200, -13.190, -13.060, -12.879, -12.663, -12.456,
-12.283, -12.150, -12.055, -13.081, -13.221, -13.208, -13.076, -12.891, -12.673,
-12.464, -12.289, -12.154, -12.057, -13.106, -13.241, -13.226, -13.092, -12.904,
-12.683, -12.473, -12.295, -12.158, -12.059, -13.130, -13.260, -13.244, -13.107,
-12.917, -12.694, -12.481, -12.301, -12.163, -12.062, -13.166, -13.280, -13.262,
-13.123, -12.930, -12.705, -12.489, -12.307, -12.167, -12.064, -13.189, -13.299,
-13.279, -13.138, -12.942, -12.715, -12.498, -12.314, -12.172, -12.067, -13.210,
-13.317, -13.296, -13.154, -12.955, -12.726, -12.506, -12.320, -12.176, -12.070,
 -13.231, -13.334, -13.312, -13.168, -12.968, -12.737, -12.515, -12.327,
-12.181, -12.072, -13.252, -13.351, -13.327, -13.183, -12.981, -12.748, -12.524,
-12.334, -12.185, -12.076, -13.270, -13.367, -13.342, -13.197, -12.993, -12.759,
-12.533, -12.341, -12.190, -12.078, -13.288, -13.382, -13.357, -13.210, -13.006,
-12.770, -12.544, -12.348, -12.195, -12.081, -13.304, -13.396, -13.371, -13.226,
-13.018, -12.781, -12.553, -12.354, -12.199, -12.084, -13.319, -13.409, -13.384,
-13.239, -13.030, -12.792, -12.562, -12.361, -12.204, -12.087, -13.333, -13.422,
-13.396, -13.251, -13.042, -12.803, -12.571, -12.369, -12.209, -12.090, -13.344,
-13.433, -13.408, -13.263, -13.054, -12.814, -12.580, -12.376, -12.214, -12.093,
 -13.354, -13.453, -13.419, -13.275, -13.065, -12.825, -12.590, -12.383,
-12.219, -12.096, -13.362, -13.462, -13.429, -13.286, -13.077, -12.836, -12.599,
-12.390, -12.224, -12.099, -13.369, -13.471, -13.439, -13.301, -13.088, -12.849,
-12.609, -12.397, -12.229, -12.102, -13.374, -13.478, -13.448, -13.312, -13.099,
-12.859, -12.618, -12.405, -12.234, -12.105, -13.377, -13.484, -13.457, -13.322,
-13.111, -12.870, -12.627, -12.412, -12.239, -12.108, -13.378, -13.489, -13.465,
-13.332, -13.122, -12.881, -12.636, -12.419, -12.244, -12.111, -13.378, -13.493,
-13.473, -13.343, -13.133, -12.891, -12.646, -12.426, -12.249, -12.114, -13.377,
-13.497, -13.481, -13.353, -13.144, -12.902, -12.655, -12.433, -12.254, -12.117,
 -13.374, -13.500, -13.488, -13.363, -13.155, -12.912, -12.664, -12.440,
-12.259, -12.120, -13.371, -13.502, -13.496, -13.373, -13.166, -12.923, -12.673,
-12.447, -12.264, -12.124, -13.366, -13.505, -13.503, -13.383, -13.176, -12.934,
-12.682, -12.454, -12.269, -12.127, -13.361, -13.507, -13.511, -13.394, -13.187,
-12.944, -12.691, -12.461, -12.274, -12.130, -13.355, -13.508, -13.518, -13.404,
-13.199, -12.955, -12.700, -12.468, -12.279, -12.133, -13.349, -13.510, -13.528,
-13.415, -13.212, -12.965, -12.708, -12.475, -12.284, -12.136, -13.344, -13.511,
-13.535, -13.425, -13.224, -12.976, -12.717, -12.482, -12.289, -12.139, -13.338,
-13.513, -13.550, -13.436, -13.235, -12.986, -12.726, -12.489, -12.294, -12.142,
 -13.333, -13.515, -13.558, -13.448, -13.247, -12.997, -12.735, -12.496,
-12.299, -12.145, -13.328, -13.518, -13.567, -13.459, -13.259, -13.007, -12.744,
-12.503, -12.304, -12.149, -13.325, -13.521, -13.575, -13.471, -13.270, -13.018,
-12.753, -12.509, -12.309, -12.152, -13.322, -13.524, -13.584, -13.483, -13.282,
-13.029, -12.762, -12.516, -12.314, -12.155, -13.322, -13.528, -13.594, -13.495,
-13.295, -13.040, -12.771, -12.523, -12.319, -12.158, -13.322, -13.533, -13.603,
-13.508, -13.307, -13.051, -12.781, -12.530, -12.323, -12.161, -13.323, -13.539,
-13.613, -13.520, -13.319, -13.062, -12.790, -12.537, -12.328, -12.164, -13.326,
-13.545, -13.624, -13.533, -13.332, -13.073, -12.799, -12.544, -12.333, -12.168,
 -13.331, -13.552, -13.634, -13.546, -13.345, -13.084, -12.808, -12.553,
-12.338, -12.171, -13.338, -13.560, -13.645, -13.559, -13.360, -13.095, -12.817,
-12.560, -12.343, -12.174, -13.346, -13.569, -13.657, -13.573, -13.373, -13.107,
-12.827, -12.567, -12.348, -12.177, -13.356, -13.579, -13.669, -13.586, -13.386,
-13.118, -12.837, -12.574, -12.353, -12.180, -13.368, -13.590, -13.681, -13.602,
-13.399, -13.129, -12.846, -12.581, -12.358, -12.184, -13.382, -13.601, -13.694,
-13.616, -13.412, -13.140, -12.856, -12.588, -12.363, -12.187, -13.410, -13.614,
-13.707, -13.629, -13.424, -13.151, -12.865, -12.595, -12.368, -12.190, -13.426,
-13.628, -13.720, -13.642, -13.437, -13.163, -12.874, -12.602, -12.373, -12.193,
 -13.444, -13.643, -13.733, -13.656, -13.450, -13.174, -12.884, -12.609,
-12.378, -12.196, -13.464, -13.658, -13.747, -13.674, -13.463, -13.185, -12.893,
-12.616, -12.383, -12.200, -13.484, -13.674, -13.761, -13.687, -13.475, -13.198,
-12.902, -12.623, -12.388, -12.203, -13.506, -13.691, -13.775, -13.700, -13.488,
-13.209, -12.911, -12.630, -12.393, -12.206, -13.530, -13.708, -13.790, -13.714,
-13.500, -13.219, -12.920, -12.637, -12.398, -12.209, -13.554, -13.727, -13.805,
-13.727, -13.512, -13.230, -12.929, -12.644, -12.403, -12.212, -13.579, -13.745,
-13.819, -13.740, -13.524, -13.240, -12.938, -12.651, -12.408, -12.216, -13.605,
-13.764, -13.835, -13.753, -13.535, -13.251, -12.947, -12.658, -12.413, -12.219,
 -13.632, -13.784, -13.850, -13.765, -13.547, -13.261, -12.955, -12.665,
-12.418, -12.222, -13.659, -13.804, -13.865, -13.778, -13.559, -13.271, -12.964,
-12.671, -12.423, -12.225, -13.687, -13.839, -13.880, -13.790, -13.570, -13.281,
-12.973, -12.678, -12.427, -12.228, -13.715, -13.859, -13.896, -13.803, -13.581,
-13.291, -12.981, -12.685, -12.432, -12.232, -13.743, -13.880, -13.911, -13.815,
-13.592, -13.301, -12.990, -12.691, -12.437, -12.235, -13.772, -13.901, -13.927,
-13.827, -13.607, -13.311, -12.998, -12.698, -12.442, -12.238, -13.801, -13.922,
-13.944, -13.840, -13.618, -13.321, -13.006, -12.705, -12.447, -12.242, -13.830,
-13.943, -13.959, -13.852, -13.629, -13.330, -13.015, -12.712, -12.452, -12.245,
 -13.859, -13.964, -13.974, -13.864, -13.640, -13.340, -13.023, -12.719,
-12.457, -12.248, -13.887, -13.985, -13.989, -13.876, -13.651, -13.350, -13.031,
-12.725, -12.462, -12.252, -13.917, -14.005, -14.004, -13.888, -13.662, -13.360,
-13.040, -12.732, -12.467, -12.255, -13.945, -14.025, -14.018, -13.900, -13.673,
-13.370, -13.048, -12.739, -12.472, -12.259, -13.973, -14.045, -14.033, -13.912,
-13.684, -13.382, -13.057, -12.746, -12.477, -12.262, -13.999, -14.064, -14.047,
-13.924, -13.695, -13.392, -13.066, -12.753, -12.483, -12.266, -14.025, -14.082,
-14.061, -13.935, -13.706, -13.402, -13.075, -12.760, -12.488, -12.270, -14.050,
-14.100, -14.075, -13.947, -13.717, -13.412, -13.084, -12.768, -12.493, -12.273,
 -14.074, -14.117, -14.088, -13.959, -13.731, -13.423, -13.093, -12.775,
-12.499, -12.277, -14.097, -14.134, -14.112, -13.971, -13.742, -13.433, -13.102,
-12.783, -12.505, -12.281, -14.118, -14.150, -14.125, -13.982, -13.753, -13.444,
-13.112, -12.790, -12.510, -12.285, -14.139, -14.164, -14.137, -13.995, -13.765,
-13.455, -13.122, -12.798, -12.516, -12.289, -14.159, -14.178, -14.148, -14.007,
-13.776, -13.466, -13.131, -12.805, -12.522, -12.292, -14.177, -14.191, -14.159,
-14.018, -13.787, -13.476, -13.140, -12.813, -12.527, -12.296, -14.193, -14.204,
-14.170, -14.029, -13.799, -13.487, -13.150, -12.821, -12.533, -12.300, -14.208,
-14.215, -14.180, -14.040, -13.810, -13.498, -13.160, -12.828, -12.539, -12.304,
 -14.221, -14.225, -14.189, -14.051, -13.821, -13.510, -13.169, -12.836,
-12.545, -12.308, -14.232, -14.234, -14.198, -14.063, -13.833, -13.524, -13.179,
-12.844, -12.550, -12.312, -14.242, -14.241, -14.207, -14.074, -13.844, -13.535,
-13.188, -12.852, -12.556, -12.316, -14.250, -14.248, -14.215, -14.084, -13.855,
-13.546, -13.198, -12.860, -12.562, -12.320, -14.256, -14.254, -14.222, -14.094,
-13.867, -13.557, -13.207, -12.868, -12.568, -12.324, -14.262, -14.259, -14.229,
-14.104, -13.878, -13.568, -13.217, -12.876, -12.574, -12.329, -14.265, -14.263,
-14.236, -14.114, -13.889, -13.579, -13.226, -12.884, -12.580, -12.333, -14.269,
-14.266, -14.242, -14.123, -13.900, -13.590, -13.236, -12.891, -12.586, -12.337,
 -14.282, -14.269, -14.248, -14.132, -13.911, -13.601, -13.245, -12.899,
-12.592, -12.341, -14.282, -14.272, -14.253, -14.142, -13.921, -13.613, -13.254,
-12.907, -12.598, -12.345, -14.281, -14.273, -14.259, -14.150, -13.932, -13.623,
-13.264, -12.915, -12.604, -12.350, -14.280, -14.275, -14.264, -14.167, -13.943,
-13.634, -13.273, -12.922, -12.610, -12.354, -14.279, -14.276, -14.269, -14.175,
-13.955, -13.645, -13.282, -12.930, -12.616, -12.358, -14.277, -14.278, -14.274,
-14.184, -13.965, -13.655, -13.291, -12.938, -12.622, -12.363, -14.275, -14.279,
-14.281, -14.192, -13.975, -13.666, -13.301, -12.945, -12.628, -12.367, -14.274,
-14.281, -14.286, -14.201, -13.985, -13.676, -13.310, -12.953, -12.634, -12.371,
 -14.274, -14.283, -14.292, -14.210, -13.995, -13.686, -13.319, -12.961,
-12.640, -12.376, -14.274, -14.286, -14.297, -14.218, -14.005, -13.697, -13.328,
-12.969, -12.647, -12.380, -14.275, -14.289, -14.303, -14.226, -14.015, -13.707,
-13.337, -12.977, -12.653, -12.385, -14.277, -14.294, -14.309, -14.235, -14.025,
-13.719, -13.346, -12.985, -12.659, -12.389, -14.281, -14.310, -14.315, -14.243,
-14.034, -13.729, -13.357, -12.992, -12.665, -12.394, -14.286, -14.315, -14.322,
-14.252, -14.044, -13.739, -13.366, -13.000, -12.671, -12.398, -14.292, -14.322,
-14.330, -14.261, -14.054, -13.749, -13.375, -13.008, -12.678, -12.403, -14.300,
-14.329, -14.338, -14.270, -14.064, -13.759, -13.384, -13.017, -12.684, -12.407,
 -14.310, -14.337, -14.346, -14.279, -14.075, -13.769, -13.393, -13.025,
-12.690, -12.412, -14.321, -14.347, -14.355, -14.288, -14.085, -13.778, -13.402,
-13.033, -12.697, -12.417, -14.334, -14.357, -14.365, -14.298, -14.095, -13.788,
-13.414, -13.041, -12.703, -12.421, -14.349, -14.369, -14.376, -14.308, -14.105,
-13.798, -13.423, -13.049, -12.710, -12.426, -14.365, -14.382, -14.386, -14.318,
-14.115, -13.808, -13.433, -13.057, -12.716, -12.430, -14.382, -14.396, -14.398,
-14.329, -14.125, -13.818, -13.442, -13.065, -12.723, -12.435, -14.402, -14.411,
-14.410, -14.340, -14.135, -13.828, -13.451, -13.073, -12.729, -12.440, -14.422,
-14.426, -14.423, -14.351, -14.146, -13.838, -13.460, -13.081, -12.735, -12.444,
 -14.444, -14.443, -14.436, -14.364, -14.156, -13.848, -13.470, -13.089,
-12.742, -12.450, -14.467, -14.461, -14.450, -14.376, -14.167, -13.861, -13.479,
-13.097, -12.748, -12.455, -14.492, -14.480, -14.475, -14.388, -14.183, -13.871,
-13.488, -13.106, -12.755, -12.459, -14.517, -14.499, -14.490, -14.400, -14.194,
-13.881, -13.498, -13.114, -12.762, -12.464, -14.544, -14.519, -14.506, -14.412,
-14.205, -13.891, -13.507, -13.122, -12.768, -12.469, -14.571, -14.540, -14.521,
-14.425, -14.216, -13.901, -13.517, -13.130, -12.775, -12.474, -14.603, -14.561,
-14.538, -14.438, -14.228, -13.912, -13.526, -13.138, -12.781, -12.478, -14.632,
-14.583, -14.554, -14.451, -14.239, -13.922, -13.536, -13.146, -12.788, -12.483,
 -14.661, -14.606, -14.571, -14.464, -14.251, -13.933, -13.545, -13.154,
-12.794, -12.488, -14.691, -14.629, -14.588, -14.478, -14.262, -13.943, -13.555,
-13.162, -12.801, -12.492, -14.721, -14.652, -14.605, -14.491, -14.274, -13.954,
-13.564, -13.170, -12.807, -12.497, -14.752, -14.676, -14.622, -14.504, -14.286,
-13.964, -13.574, -13.178, -12.813, -12.502, -14.782, -14.699, -14.639, -14.518,
-14.299, -13.975, -13.583, -13.186, -12.820, -12.507, -14.813, -14.723, -14.657,
-14.531, -14.310, -13.985, -13.592, -13.194, -12.826, -12.512, -14.844, -14.747,
-14.674, -14.545, -14.322, -13.996, -13.602, -13.202, -12.833, -12.516, -14.875,
-14.770, -14.692, -14.558, -14.333, -14.006, -13.611, -13.210, -12.839, -12.521,
 -14.906, -14.794, -14.709, -14.571, -14.345, -14.016, -13.621, -13.218,
-12.846, -12.526, -14.937, -14.817, -14.726, -14.584, -14.356, -14.027, -13.630,
-13.226, -12.852, -12.531, -14.971, -14.840, -14.742, -14.597, -14.367, -14.037,
-13.639, -13.234, -12.859, -12.536, -15.001, -14.866, -14.759, -14.610, -14.379,
-14.049, -13.650, -13.242, -12.865, -12.540, -15.031, -14.888, -14.775, -14.622,
-14.390, -14.059, -13.659, -13.250, -12.871, -12.545, -15.060, -14.910, -14.790,
-14.645, -14.400, -14.069, -13.669, -13.258, -12.878, -12.550, -15.088, -14.931,
-14.806, -14.656, -14.411, -14.079, -13.678, -13.268, -12.884, -12.555, -15.116,
-14.952, -14.820, -14.668, -14.424, -14.089, -13.687, -13.276, -12.891, -12.560,
 -15.143, -14.971, -14.835, -14.679, -14.435, -14.098, -13.696, -13.284,
-12.898, -12.565, -15.169, -14.990, -14.849, -14.690, -14.445, -14.108, -13.708,
-13.292, -12.905, -12.570, -15.194, -15.009, -14.862, -14.702, -14.455, -14.118,
-13.717, -13.301, -12.911, -12.575, -15.230, -15.026, -14.874, -14.712, -14.464,
-14.127, -13.726, -13.309, -12.918, -12.579, -15.253, -15.043, -14.887, -14.722,
-14.474, -14.137, -13.735, -13.317, -12.925, -12.584, -15.275, -15.060, -14.898,
-14.731, -14.483, -14.149, -13.744, -13.325, -12.931, -12.589, -15.296, -15.080,
-14.909, -14.741, -14.493, -14.158, -13.754, -13.333, -12.938, -12.594, -15.316,
-15.093, -14.919, -14.749, -14.502, -14.168, -13.763, -13.341, -12.944, -12.599,
 -15.334, -15.106, -14.929, -14.759, -14.511, -14.178, -13.772, -13.348,
-12.951, -12.604, -15.352, -15.118, -14.938, -14.767, -14.520, -14.188, -13.781,
-13.357, -12.958, -12.609, -15.368, -15.129, -14.947, -14.775, -14.529, -14.197,
-13.790, -13.365, -12.964, -12.614, -15.382, -15.140, -14.956, -14.783, -14.537,
-14.206, -13.799, -13.373, -12.971, -12.619, -15.396, -15.149, -14.965, -14.791,
-14.546, -14.215, -13.808, -13.380, -12.977, -12.624, -15.408, -15.158, -14.973,
-14.799, -14.555, -14.224, -13.818, -13.389, -12.984, -12.629, -15.420, -15.166,
-14.980, -14.806, -14.563, -14.233, -13.827, -13.396, -12.990, -12.634, -15.430,
-15.173, -14.987, -14.814, -14.572, -14.242, -13.836, -13.404, -12.997, -12.639,
 -15.439, -15.185, -14.993, -14.821, -14.580, -14.251, -13.845, -13.412,
-13.004, -12.644, -15.448, -15.191, -15.000, -14.828, -14.589, -14.260, -13.854,
-13.420, -13.010, -12.649, -15.456, -15.197, -15.006, -14.835, -14.597, -14.269,
-13.864, -13.428, -13.017, -12.655, -15.464, -15.203, -15.012, -14.842, -14.607,
-14.278, -13.873, -13.436, -13.023, -12.660, -15.471, -15.209, -15.024, -14.850,
-14.623, -14.288, -13.882, -13.444, -13.030, -12.665, -15.478, -15.215, -15.031,
-14.857, -14.632, -14.297, -13.891, -13.452, -13.037, -12.670, -15.485, -15.221,
-15.038, -14.865, -14.640, -14.306, -13.901, -13.461, -13.044, -12.675, -15.493,
-15.227, -15.045, -14.873, -14.649, -14.316, -13.910, -13.469, -13.051, -12.680,
 -15.501, -15.234, -15.051, -14.880, -14.658, -14.325, -13.920, -13.477,
-13.058, -12.685, -15.509, -15.241, -15.059, -14.888, -14.667, -14.335, -13.929,
-13.486, -13.064, -12.690, -15.518, -15.249, -15.066, -14.897, -14.676, -14.347,
-13.939, -13.494, -13.071, -12.695, -15.528, -15.258, -15.074, -14.905, -14.685,
-14.357, -13.949, -13.502, -13.078, -12.701, -15.539, -15.267, -15.083, -14.914,
-14.695, -14.367, -13.959, -13.510, -13.085, -12.706, -15.551, -15.277, -15.092,
-14.923, -14.705, -14.376, -13.968, -13.519, -13.092, -12.711, -15.565, -15.287,
-15.101, -14.932, -14.715, -14.386, -13.978, -13.529, -13.099, -12.716, -15.579,
-15.299, -15.111, -14.942, -14.725, -14.396, -13.987, -13.537, -13.106, -12.722,
 -15.595, -15.312, -15.122, -14.952, -14.735, -14.406, -13.997, -13.546,
-13.114, -12.727, -15.612, -15.325, -15.135, -14.963, -14.745, -14.416, -14.008,
-13.555, -13.121, -12.733, -15.630, -15.339, -15.147, -14.974, -14.755, -14.426,
-14.018, -13.564, -13.128, -12.738, -15.657, -15.355, -15.159, -14.985, -14.765,
-14.436, -14.028, -13.572, -13.135, -12.744, -15.679, -15.371, -15.172, -14.996,
-14.776, -14.446, -14.037, -13.581, -13.142, -12.749, -15.701, -15.388, -15.186,
-15.008, -14.786, -14.457, -14.049, -13.589, -13.149, -12.754, -15.724, -15.406,
-15.200, -15.019, -14.797, -14.467, -14.058, -13.598, -13.156, -12.760, -15.748,
-15.425, -15.215, -15.038, -14.807, -14.477, -14.068, -13.606, -13.163, -12.766,
 -15.773, -15.445, -15.230, -15.050, -14.818, -14.489, -14.077, -13.615,
-13.171, -12.772, -15.799, -15.466, -15.246, -15.062, -14.829, -14.499, -14.087,
-13.623, -13.178, -12.777, -15.826, -15.487, -15.262, -15.075, -14.840, -14.509,
-14.096, -13.632, -13.185, -12.783, -15.853, -15.509, -15.279, -15.088, -14.850,
-14.519, -14.106, -13.641, -13.192, -12.788, -15.882, -15.532, -15.296, -15.102,
-14.861, -14.529, -14.115, -13.649, -13.199, -12.794, -15.911, -15.555, -15.314,
-15.115, -14.872, -14.539, -14.125, -13.658, -13.207, -12.800, -15.940, -15.578,
-15.331, -15.129, -14.883, -14.549, -14.134, -13.667, -13.215, -12.805, -15.970,
-15.611, -15.350, -15.143, -14.894, -14.564, -14.144, -13.675, -13.222, -12.811,
 -16.001, -15.636, -15.368, -15.156, -14.905, -14.574, -14.153, -13.684,
-13.229, -12.816, -16.032, -15.661, -15.387, -15.170, -14.917, -14.584, -14.163,
-13.692, -13.236, -12.822, -16.063, -15.686, -15.405, -15.184, -14.928, -14.594,
-14.172, -13.701, -13.243, -12.828, -16.094, -15.711, -15.424, -15.198, -14.939,
-14.604, -14.181, -13.709, -13.251, -12.833, -16.126, -15.737, -15.443, -15.212,
-14.950, -14.614, -14.191, -13.718, -13.258, -12.839, -16.157, -15.763, -15.462,
-15.226, -14.961, -14.624, -14.200, -13.726, -13.265, -12.845, -16.189, -15.788,
-15.482, -15.241, -14.971, -14.634, -14.209, -13.735, -13.272, -12.851, -16.221,
-15.814, -15.500, -15.255, -14.982, -14.643, -14.218, -13.744, -13.279, -12.857,
 -16.252, -15.840, -15.519, -15.268, -14.993, -14.653, -14.228, -13.752,
-13.287, -12.862, -16.284, -15.865, -15.538, -15.282, -15.004, -14.663, -14.237,
-13.761, -13.294, -12.868, -16.315, -15.890, -15.557, -15.295, -15.014, -14.674,
-14.248, -13.769, -13.301, -12.874, -16.346, -15.915, -15.575, -15.308, -15.025,
-14.683, -14.257, -13.778, -13.308, -12.880, -16.377, -15.940, -15.593, -15.321,
-15.036, -14.693, -14.267, -13.786, -13.316, -12.886, -16.407, -15.964, -15.611,
-15.334, -15.046, -14.702, -14.276, -13.796, -13.323, -12.892, -16.437, -15.988,
-15.634, -15.347, -15.063, -14.712, -14.285, -13.805, -13.330, -12.897, -16.467,
-16.012, -15.651, -15.359, -15.073, -14.721, -14.294, -13.814, -13.339, -12.903,
 -16.496, -16.035, -15.672, -15.371, -15.083, -14.731, -14.305, -13.823,
-13.346, -12.909, -16.524, -16.057, -15.688, -15.383, -15.093, -14.740, -14.314,
-13.832, -13.353, -12.915, -16.595, -16.079, -15.703, -15.395, -15.103, -14.750,
-14.323, -13.840, -13.361, -12.921, -16.623, -16.100, -15.719, -15.406, -15.112,
-14.759, -14.332, -13.849, -13.368, -12.927, -16.649, -16.120, -15.733, -15.417,
-15.122, -14.768, -14.341, -13.857, -13.376, -12.933, -16.675, -16.140, -15.747,
-15.428, -15.131, -14.777, -14.350, -13.866, -13.383, -12.939, -16.701, -16.159,
-15.761, -15.438, -15.141, -14.787, -14.359, -13.875, -13.390, -12.945, -16.725,
-16.177, -15.774, -15.449, -15.151, -14.797, -14.368, -13.883, -13.397, -12.951,
 -16.748, -16.195, -15.787, -15.459, -15.160, -14.806, -14.377, -13.892,
-13.405, -12.957, -16.771, -16.212, -15.799, -15.469, -15.169, -14.815, -14.386,
-13.900, -13.412, -12.963, -16.793, -16.228, -15.811, -15.478, -15.178, -14.824,
-14.395, -13.909, -13.419, -12.969, -16.814, -16.243, -15.823, -15.488, -15.187,
-14.833, -14.404, -13.918, -13.427, -12.975, -16.834, -16.258, -15.834, -15.497,
-15.196, -14.842, -14.412, -13.927, -13.434, -12.981, -16.854, -16.272, -15.844,
-15.506, -15.205, -14.851, -14.421, -13.935, -13.442, -12.987, -16.873, -16.286,
-15.855, -15.515, -15.214, -14.860, -14.430, -13.944, -13.449, -12.993, -16.890,
-16.316, -15.865, -15.524, -15.223, -14.869, -14.442, -13.952, -13.456, -12.999,
 -16.908, -16.329, -15.875, -15.532, -15.232, -14.878, -14.451, -13.961,
-13.463, -13.005, -16.924, -16.341, -15.885, -15.549, -15.241, -14.887, -14.460,
-13.969, -13.470, -13.011, -16.940, -16.353, -15.895, -15.557, -15.250, -14.895,
-14.468, -13.977, -13.478, -13.017, -16.956, -16.365, -15.904, -15.566, -15.259,
-14.904, -14.477, -13.986, -13.485, -13.023, -16.972, -16.376, -15.914, -15.575,
-15.268, -14.913, -14.486, -13.994, -13.492, -13.028, -16.987, -16.388, -15.924,
-15.583, -15.276, -14.922, -14.494, -14.003, -13.499, -13.034, -17.003, -16.400,
-15.934, -15.595, -15.285, -14.931, -14.503, -14.011, -13.507, -13.040, -17.028,
-16.412, -15.944, -15.604, -15.294, -14.939, -14.512, -14.019, -13.514, -13.046,
 -17.044, -16.424, -15.954, -15.613, -15.303, -14.948, -14.520, -14.028,
-13.521, -13.052, -17.060, -16.438, -15.965, -15.622, -15.312, -14.962, -14.529,
-14.036, -13.528, -13.058, -17.077, -16.451, -15.976, -15.632, -15.321, -14.971,
-14.539, -14.044, -13.536, -13.064, -17.094, -16.465, -15.987, -15.641, -15.331,
-14.980, -14.547, -14.052, -13.543, -13.070, -17.112, -16.479, -15.998, -15.651,
-15.340, -14.989, -14.556, -14.060, -13.550, -13.076, -17.130, -16.493, -16.011,
-15.662, -15.350, -14.998, -14.564, -14.071, -13.557, -13.082, -17.149, -16.509,
-16.023, -15.672, -15.360, -15.007, -14.573, -14.079, -13.564, -13.088, -17.170,
-16.525, -16.036, -15.683, -15.369, -15.016, -14.582, -14.087, -13.573, -13.094,
 -17.191, -16.542, -16.057, -15.694, -15.379, -15.025, -14.591, -14.097,
-13.580, -13.100, -17.212, -16.559, -16.071, -15.705, -15.389, -15.034, -14.599,
-14.105, -13.587, -13.106, -17.235, -16.577, -16.085, -15.717, -15.399, -15.043,
-14.608, -14.113, -13.594, -13.112, -17.258, -16.596, -16.101, -15.729, -15.409,
-15.052, -14.616, -14.121, -13.602, -13.118, -17.283, -16.622, -16.116, -15.741,
-15.420, -15.061, -14.625, -14.129, -13.609, -13.124, -17.308, -16.643, -16.132,
-15.754, -15.430, -15.071, -14.634, -14.137, -13.616, -13.130, -17.334, -16.664,
-16.149, -15.767, -15.441, -15.080, -14.642, -14.145, -13.623, -13.137, -17.360,
-16.686, -16.166, -15.780, -15.451, -15.090, -14.651, -14.154, -13.630, -13.143,
 -17.388, -16.708, -16.186, -15.793, -15.470, -15.099, -14.659, -14.162,
-13.638, -13.149, -17.415, -16.731, -16.204, -15.807, -15.481, -15.109, -14.668,
-14.170, -13.645, -13.155, -17.444, -16.755, -16.223, -15.821, -15.492, -15.118,
-14.677, -14.178, -13.652, -13.161, -17.473, -16.779, -16.242, -15.835, -15.503,
-15.128, -14.685, -14.186, -13.660, -13.167, -17.502, -16.804, -16.261, -15.850,
-15.514, -15.138, -14.694, -14.194, -13.668, -13.173, -17.532, -16.829, -16.281,
-15.864, -15.525, -15.147, -14.703, -14.203, -13.675, -13.179, -17.562, -16.855,
-16.301, -15.879, -15.536, -15.157, -14.711, -14.211, -13.682, -13.185, -17.592,
-16.881, -16.321, -15.894, -15.548, -15.167, -14.720, -14.219, -13.690, -13.191,
 -17.623, -16.907, -16.341, -15.909, -15.561, -15.177, -14.729, -14.227,
-13.697, -13.197, -17.654, -16.934, -16.362, -15.924, -15.572, -15.187, -14.738,
-14.235, -13.704, -13.203, -17.685, -16.961, -16.383, -15.940, -15.584, -15.197,
-14.746, -14.243, -13.711, -13.209, -17.716, -16.988, -16.404, -15.955, -15.595,
-15.207, -14.755, -14.252, -13.719, -13.215, -17.747, -17.015, -16.425, -15.970,
-15.606, -15.216, -14.764, -14.260, -13.726, -13.221, -17.779, -17.042, -16.451,
-15.990, -15.618, -15.226, -14.773, -14.270, -13.734, -13.227, -17.810, -17.070,
-16.472, -16.005, -15.629, -15.236, -14.782, -14.278, -13.742, -13.233, -17.841,
-17.097, -16.493, -16.020, -15.640, -15.246, -14.791, -14.287, -13.749, -13.239,
 -17.872, -17.124, -16.514, -16.035, -15.652, -15.256, -14.799, -14.295,
-13.757, -13.246, -17.904, -17.152, -16.535, -16.051, -15.663, -15.267, -14.812,
-14.303, -13.764, -13.252, -17.935, -17.179, -16.556, -16.066, -15.674, -15.276,
-14.821, -14.311, -13.771, -13.258, -17.965, -17.206, -16.577, -16.085, -15.685,
-15.286, -14.830, -14.319, -13.779, -13.264, -17.996, -17.232, -16.597, -16.100,
-15.696, -15.296, -14.838, -14.329, -13.786, -13.270, -18.027, -17.259, -16.617,
-16.114, -15.707, -15.306, -14.847, -14.337, -13.794, -13.276, -18.057, -17.285,
-16.637, -16.129, -15.718, -15.315, -14.856, -14.345, -13.803, -13.283, -18.087,
-17.311, -16.657, -16.143, -15.728, -15.325, -14.865, -14.355, -13.811, -13.289,
 -18.116, -17.336, -16.676, -16.158, -15.739, -15.334, -14.874, -14.363,
-13.818, -13.296, -18.268, -17.361, -16.695, -16.172, -15.749, -15.344, -14.882,
-14.371, -13.826, -13.302, -18.297, -17.386, -16.714, -16.185, -15.760, -15.353,
-14.891, -14.379, -13.834, -13.308, -18.326, -17.410, -16.732, -16.198, -15.770,
-15.370, -14.899, -14.387, -13.841, -13.314, -18.355, -17.434, -16.750, -16.212,
-15.780, -15.380, -14.908, -14.395, -13.849, -13.320, -18.383, -17.457, -16.768,
-16.225, -15.790, -15.389, -14.916, -14.403, -13.856, -13.328, -18.411, -17.480,
-16.785, -16.237, -15.800, -15.398, -14.925, -14.411, -13.864, -13.334, -18.438,
-17.502, -16.801, -16.250, -15.810, -15.407, -14.933, -14.419, -13.871, -13.340,
 -18.465, -17.524, -16.818, -16.262, -15.820, -15.416, -14.942, -14.427,
-13.879, -13.346, -18.491, -17.545, -16.834, -16.274, -15.829, -15.425, -14.950,
-14.435, -13.886, -13.353, -18.517, -17.565, -16.849, -16.289, -15.839, -15.434,
-14.959, -14.443, -13.894, -13.359, -18.542, -17.585, -16.864, -16.301, -15.848,
-15.443, -14.967, -14.451, -13.903, -13.366, -18.577, -17.605, -16.879, -16.312,
-15.858, -15.452, -14.975, -14.459, -13.910, -13.372, -18.601, -17.624, -16.894,
-16.323, -15.867, -15.461, -14.984, -14.467, -13.918, -13.378, -18.624, -17.690,
-16.908, -16.335, -15.876, -15.470, -14.992, -14.475, -13.926, -13.385, -18.648,
-17.708, -16.922, -16.346, -15.888, -15.480, -15.000, -14.483, -13.933, -13.391,
 -18.670, -17.726, -16.936, -16.357, -15.897, -15.489, -15.009, -14.491,
-13.941, -13.398, -18.693, -17.744, -16.950, -16.368, -15.906, -15.498, -15.017,
-14.499, -13.948, -13.404, -18.715, -17.761, -16.965, -16.379, -15.921, -15.507,
-15.026, -14.507, -13.956, -13.410, -18.737, -17.778, -16.978, -16.390, -15.930,
-15.515, -15.034, -14.515, -13.963, -13.417, -18.759, -17.795, -16.992, -16.401,
-15.939, -15.524, -15.042, -14.524, -13.971, -13.423, -18.781, -17.812, -17.006,
-16.412, -15.949, -15.533, -15.051, -14.532, -13.979, -13.429, -18.802, -17.829,
-17.020, -16.423, -15.958, -15.542, -15.060, -14.540, -13.987, -13.436, -18.824,
-17.846, -17.034, -16.434, -15.968, -15.550, -15.068, -14.548, -13.994, -13.442,
 -18.846, -17.871, -17.048, -16.446, -15.977, -15.560, -15.076, -14.556,
-14.002, -13.449, -18.868, -17.889, -17.062, -16.457, -15.987, -15.568, -15.086,
-14.564, -14.009, -13.455, -18.891, -17.906, -17.076, -16.469, -15.997, -15.577,
-15.095, -14.572, -14.017, -13.462, -18.913, -17.925, -17.091, -16.481, -16.007,
-15.586, -15.103, -14.581, -14.024, -13.468, -18.936, -17.943, -17.106, -16.493,
-16.017, -15.595, -15.112, -14.589, -14.032, -13.474, -18.960, -17.962, -17.122,
-16.506, -16.027, -15.604, -15.120, -14.597, -14.041, -13.481, -18.984, -17.982,
-17.138, -16.518, -16.037, -15.613, -15.129, -14.605, -14.048, -13.488, -19.008,
-18.002, -17.173, -16.531, -16.048, -15.622, -15.138, -14.616, -14.056, -13.495,
 -19.033, -18.023, -17.190, -16.545, -16.058, -15.631, -15.146, -14.624,
-14.065, -13.501, -19.059, -18.044, -17.207, -16.558, -16.069, -15.640, -15.155,
-14.632, -14.072, -13.508, -19.084, -18.065, -17.224, -16.572, -16.080, -15.649,
-15.163, -14.640, -14.080, -13.514, -19.111, -18.088, -17.242, -16.586, -16.090,
-15.659, -15.178, -14.648, -14.087, -13.521, -19.138, -18.110, -17.261, -16.600,
-16.104, -15.668, -15.186, -14.656, -14.095, -13.527, -19.165, -18.134, -17.279,
-16.617, -16.115, -15.677, -15.195, -14.664, -14.102, -13.535, -19.192, -18.158,
-17.299, -16.632, -16.126, -15.686, -15.203, -14.672, -14.109, -13.542, -19.220,
-18.182, -17.318, -16.648, -16.138, -15.696, -15.212, -14.680, -14.117, -13.548,
 -19.249, -18.207, -17.338, -16.663, -16.150, -15.705, -15.220, -14.688,
-14.124, -13.555, -19.277, -18.232, -17.364, -16.679, -16.161, -15.715, -15.229,
-14.696, -14.132, -13.562, -19.306, -18.257, -17.385, -16.695, -16.173, -15.724,
-15.237, -14.704, -14.139, -13.568, -19.336, -18.283, -17.406, -16.711, -16.185,
-15.734, -15.246, -14.712, -14.148, -13.575, -19.365, -18.310, -17.428, -16.728,
-16.198, -15.743, -15.255, -14.720, -14.155, -13.581, -19.395, -18.337, -17.449,
-16.744, -16.210, -15.759, -15.263, -14.728, -14.162, -13.588, -19.425, -18.364,
-17.471, -16.761, -16.222, -15.769, -15.272, -14.736, -14.170, -13.594, -19.455,
-18.391, -17.494, -16.778, -16.235, -15.780, -15.280, -14.744, -14.177, -13.601,
 -19.485, -18.418, -17.516, -16.795, -16.247, -15.789, -15.289, -14.752,
-14.184, -13.608, -19.515, -18.446, -17.539, -16.812, -16.260, -15.799, -15.297,
-14.759, -14.192, -13.614, -19.545, -18.474, -17.562, -16.830, -16.272, -15.808,
-15.306, -14.767, -14.199, -13.621, -19.576, -18.502, -17.585, -16.847, -16.285,
-15.818, -15.314, -14.775, -14.206, -13.628, -19.606, -18.530, -17.608, -16.874,
-16.298, -15.828, -15.324, -14.783, -14.214, -13.634, -19.637, -18.558, -17.631,
-16.891, -16.310, -15.837, -15.333, -14.791, -14.221, -13.641, -19.667, -18.586,
-17.654, -16.909, -16.323, -15.847, -15.341, -14.799, -14.228, -13.647, -19.698,
-18.615, -17.677, -16.926, -16.336, -15.857, -15.350, -14.807, -14.236, -13.654,
 -19.728, -18.643, -17.700, -16.943, -16.348, -15.866, -15.359, -14.815,
-14.243, -13.661, -19.759, -18.671, -17.723, -16.961, -16.361, -15.877, -15.368,
-14.823, -14.250, -13.667, -19.789, -18.699, -17.746, -16.978, -16.373, -15.886,
-15.376, -14.830, -14.258, -13.674, -19.820, -18.727, -17.769, -16.995, -16.386,
-15.896, -15.385, -14.839, -14.265, -13.681, -19.850, -18.755, -17.792, -17.012,
-16.402, -15.905, -15.393, -14.847, -14.272, -13.687, -19.880, -18.782, -17.815,
-17.029, -16.415, -15.915, -15.402, -14.855, -14.281, -13.695, -19.911, -18.810,
-17.837, -17.050, -16.427, -15.924, -15.411, -14.865, -14.288, -13.702, -19.941,
-18.837, -17.859, -17.066, -16.439, -15.933, -15.419, -14.872, -14.295, -13.708,
 -20.261, -18.864, -17.881, -17.083, -16.451, -15.943, -15.428, -14.881,
-14.303, -13.715, -20.292, -18.890, -17.903, -17.099, -16.463, -15.952, -15.436,
-14.888, -14.310, -13.722, -20.324, -18.917, -17.924, -17.115, -16.475, -15.961,
-15.445, -14.896, -14.317, -13.728, -20.356, -18.943, -17.945, -17.131, -16.487,
-15.970, -15.453, -14.904, -14.324, -13.735, -20.388, -18.969, -17.965, -17.146,
-16.499, -15.981, -15.462, -14.912, -14.332, -13.743, -20.419, -18.994, -17.986,
-17.161, -16.510, -15.990, -15.470, -14.920, -14.339, -13.750, -20.450, -19.019,
-18.006, -17.177, -16.522, -15.999, -15.479, -14.927, -14.346, -13.756, -20.482,
-19.044, -18.025, -17.192, -16.533, -16.008, -15.487, -14.935, -14.353, -13.763,
 -20.513, -19.069, -18.045, -17.206, -16.544, -16.017, -15.496, -14.943,
-14.361, -13.770, -20.544, -19.093, -18.064, -17.221, -16.556, -16.026, -15.504,
-14.955, -14.368, -13.777, -20.575, -19.117, -18.083, -17.236, -16.567, -16.035,
-15.513, -14.963, -14.377, -13.784, -20.606, -19.140, -18.101, -17.250, -16.578,
-16.044, -15.521, -14.971, -14.385, -13.790, -20.637, -19.283, -18.120, -17.264,
-16.589, -16.053, -15.530, -14.979, -14.392, -13.797, -20.668, -19.305, -18.138,
-17.278, -16.600, -16.062, -15.544, -14.986, -14.399, -13.804, -20.699, -19.328,
-18.155, -17.292, -16.616, -16.071, -15.552, -14.994, -14.406, -13.810, -20.729,
-19.350, -18.173, -17.307, -16.627, -16.080, -15.561, -15.002, -14.413, -13.817,
 -20.760, -19.372, -18.191, -17.321, -16.637, -16.088, -15.569, -15.009,
-14.421, -13.824, -20.790, -19.393, -18.208, -17.335, -16.648, -16.097, -15.577,
-15.017, -14.428, -13.831, -20.820, -19.415, -18.225, -17.348, -16.659, -16.106,
-15.586, -15.025, -14.435, -13.838, -20.851, -19.436, -18.243, -17.362, -16.670,
-16.115, -15.594, -15.033, -14.442, -13.845, -20.881, -19.458, -18.260, -17.376,
-16.681, -16.124, -15.604, -15.040, -14.449, -13.851, -20.911, -19.479, -18.277,
-17.390, -16.692, -16.133, -15.612, -15.048, -14.457, -13.858, -20.941, -19.500,
-18.294, -17.404, -16.703, -16.142, -15.620, -15.056, -14.464, -13.865, -20.970,
-19.522, -18.312, -17.418, -16.714, -16.157, -15.629, -15.063, -14.471, -13.871,
 -21.000, -19.543, -18.330, -17.432, -16.725, -16.166, -15.638, -15.071,
-14.478, -13.878, -21.030, -19.565, -18.347, -17.446, -16.740, -16.175, -15.646,
-15.079, -14.485, -13.885, -21.059, -19.587, -18.365, -17.460, -16.751, -16.184,
-15.655, -15.087, -14.493, -13.891, -21.089, -19.610, -18.383, -17.475, -16.763,
-16.193, -15.663, -15.095, -14.500, -13.898, -21.118, -19.632, -18.402, -17.490,
-16.774, -16.202, -15.672, -15.103, -14.508, -13.905, -21.147, -19.655, -18.471,
-17.505, -16.786, -16.212, -15.680, -15.110, -14.516, -13.912, -21.177, -19.678,
-18.489, -17.520, -16.798, -16.221, -15.690, -15.119, -14.523, -13.919, -21.206,
-19.702, -18.509, -17.535, -16.810, -16.231, -15.698, -15.127, -14.530, -13.925,
 -21.235, -19.725, -18.528, -17.551, -16.822, -16.240, -15.707, -15.135,
-14.537, -13.932, -21.264, -19.750, -18.548, -17.567, -16.835, -16.250, -15.715,
-15.143, -14.544, -13.939, -21.293, -19.774, -18.568, -17.583, -16.847, -16.260,
-15.724, -15.150, -14.551, -13.945, -21.321, -19.799, -18.588, -17.600, -16.860,
-16.270, -15.732, -15.158, -14.558, -13.952, -21.350, -19.824, -18.609, -17.616,
-16.873, -16.280, -15.741, -15.166, -14.565, -13.960, -21.379, -19.850, -18.630,
-17.633, -16.885, -16.290, -15.750, -15.174, -14.573, -13.966, -21.407, -19.876,
-18.652, -17.650, -16.901, -16.300, -15.758, -15.181, -14.580, -13.973, -21.436,
-19.902, -18.674, -17.668, -16.914, -16.310, -15.767, -15.189, -14.587, -13.979,
 -21.464, -19.929, -18.696, -17.685, -16.928, -16.320, -15.776, -15.197,
-14.594, -13.986, -21.493, -19.956, -18.718, -17.703, -16.941, -16.330, -15.785,
-15.204, -14.601, -13.992, -21.521, -19.983, -18.741, -17.721, -16.955, -16.340,
-15.794, -15.212, -14.610, -13.999, -21.549, -20.011, -18.764, -17.740, -16.969,
-16.351, -15.802, -15.220, -14.617, -14.005, -21.577, -20.038, -18.787, -17.758,
-16.983, -16.361, -15.811, -15.228, -14.624, -14.012, -21.605, -20.066, -18.811,
-17.777, -16.997, -16.371, -15.820, -15.235, -14.631, -14.018, -21.633, -20.095,
-18.834, -17.796, -17.011, -16.384, -15.828, -15.243, -14.639, -14.025, -21.661,
-20.123, -18.858, -17.814, -17.025, -16.395, -15.837, -15.251, -14.646, -14.031,
 -21.689, -20.152, -18.882, -17.833, -17.039, -16.405, -15.846, -15.259,
-14.653, -14.038, -21.717, -20.180, -18.907, -17.853, -17.054, -16.416, -15.855,
-15.266, -14.660, -14.044, -21.745, -20.209, -18.931, -17.872, -17.068, -16.426,
-15.863, -15.274, -14.667, -14.051, -21.773, -20.238, -18.956, -17.916, -17.082,
-16.437, -15.872, -15.286, -14.674, -14.058, -21.804, -20.268, -18.980, -17.936,
-17.097, -16.447, -15.881, -15.294, -14.684, -14.064, -21.832, -20.297, -19.005,
-17.955, -17.111, -16.458, -15.889, -15.302, -14.691, -14.071, -21.859, -20.326,
-19.029, -17.974, -17.126, -16.469, -15.898, -15.309, -14.698, -14.077, -21.886,
-20.356, -19.054, -17.993, -17.140, -16.479, -15.913, -15.317, -14.706, -14.083,
 -21.914, -20.385, -19.078, -18.013, -17.154, -16.492, -15.921, -15.325,
-14.713, -14.090, -21.941, -20.415, -19.103, -18.033, -17.169, -16.502, -15.930,
-15.333, -14.720, -14.096, -21.969, -20.444, -19.127, -18.052, -17.183, -16.513,
-15.938, -15.341, -14.727, -14.103, -21.996, -20.474, -19.152, -18.071, -17.197,
-16.523, -15.947, -15.348, -14.734, -14.109, -22.023, -20.503, -19.176, -18.090,
-17.212, -16.534, -15.955, -15.356, -14.742, -14.117, -22.050, -20.533, -19.200,
-18.109, -17.226, -16.544, -15.964, -15.364, -14.749, -14.124, -22.077, -20.563,
-19.224, -18.127, -17.240, -16.555, -15.972, -15.372, -14.756, -14.130, -22.715,
-20.592, -19.248, -18.146, -17.254, -16.565, -15.981, -15.380, -14.763, -14.137,
 -22.744, -20.622, -19.272, -18.164, -17.268, -16.575, -15.989, -15.388,
-14.770, -14.143, -22.773, -20.657, -19.295, -18.183, -17.282, -16.590, -15.997,
-15.396, -14.777, -14.149, -22.801, -20.687, -19.319, -18.201, -17.295, -16.600,
-16.006, -15.403, -14.784, -14.156, -22.830, -20.716, -19.342, -18.219, -17.309,
-16.610, -16.014, -15.411, -14.791, -14.164, -22.858, -20.746, -19.365, -18.237,
-17.323, -16.620, -16.023, -15.419, -14.798, -14.170, -22.887, -20.775, -19.387,
-18.254, -17.336, -16.631, -16.031, -15.426, -14.805, -14.176, -22.915, -20.805,
-19.410, -18.272, -17.350, -16.641, -16.039, -15.434, -14.812, -14.183, -22.943,
-20.834, -19.432, -18.289, -17.363, -16.651, -16.048, -15.442, -14.819, -14.189,
 -22.971, -20.864, -19.454, -18.306, -17.376, -16.661, -16.056, -15.449,
-14.826, -14.196, -22.999, -20.893, -19.476, -18.323, -17.390, -16.671, -16.065,
-15.457, -14.834, -14.202, -23.028, -20.922, -19.497, -18.340, -17.403, -16.681,
-16.073, -15.465, -14.841, -14.209, -23.056, -21.192, -19.519, -18.357, -17.416,
-16.691, -16.081, -15.473, -14.848, -14.215, -23.083, -21.222, -19.540, -18.373,
-17.429, -16.701, -16.090, -15.480, -14.855, -14.221, -23.111, -21.253, -19.561,
-18.390, -17.456, -16.712, -16.098, -15.488, -14.862, -14.228, -23.139, -21.283,
-19.582, -18.406, -17.469, -16.722, -16.106, -15.496, -14.869, -14.234, -23.167,
-21.313, -19.603, -18.422, -17.483, -16.732, -16.114, -15.504, -14.876, -14.241,
 -23.195, -21.344, -19.623, -18.439, -17.496, -16.742, -16.123, -15.511,
-14.883, -14.247, -23.222, -21.374, -19.644, -18.455, -17.509, -16.752, -16.131,
-15.519, -14.890, -14.253, -23.250, -21.404, -19.670, -18.471, -17.522, -16.762,
-16.139, -15.527, -14.898, -14.260, -23.278, -21.434, -19.690, -18.487, -17.535,
-16.772, -16.149, -15.535, -14.905, -14.266, -23.305, -21.464, -19.711, -18.504,
-17.548, -16.782, -16.157, -15.543, -14.912, -14.272, -23.332, -21.494, -19.732,
-18.520, -17.561, -16.793, -16.166, -15.551, -14.919, -14.279, -23.360, -21.524,
-19.752, -18.536, -17.574, -16.803, -16.174, -15.558, -14.926, -14.285, -23.387,
-21.553, -19.773, -18.553, -17.587, -16.813, -16.183, -15.566, -14.933, -14.292,
 -23.415, -21.583, -19.794, -18.569, -17.600, -16.824, -16.191, -15.574,
-14.940, -14.298, -23.442, -21.613, -19.815, -18.586, -17.613, -16.834, -16.199,
-15.582, -14.947, -14.305, -23.469, -21.642, -19.836, -18.603, -17.627, -16.845,
-16.208, -15.590, -14.954, -14.311, -23.496, -21.672, -19.857, -18.620, -17.640,
-16.855, -16.216, -15.598, -14.962, -14.317, -23.523, -21.701, -19.969, -18.637,
-17.654, -16.866, -16.225, -15.606, -14.969, -14.325, -23.550, -21.731, -19.991,
-18.654, -17.668, -16.876, -16.235, -15.619, -14.976, -14.332, -23.577, -21.760,
-20.013, -18.672, -17.682, -16.887, -16.243, -15.627, -14.986, -14.338, -23.604,
-21.789, -20.034, -18.690, -17.695, -16.898, -16.252, -15.635, -14.993, -14.344,
 -23.631, -21.818, -20.057, -18.707, -17.710, -16.909, -16.260, -15.643,
-15.000, -14.351, -23.658, -21.847, -20.079, -18.726, -17.724, -16.920, -16.274,
-15.651, -15.007, -14.357, -23.685, -21.876, -20.102, -18.744, -17.738, -16.931,
-16.283, -15.659, -15.014, -14.364, -23.712, -21.905, -20.125, -18.762, -17.753,
-16.942, -16.291, -15.666, -15.021, -14.371, -23.739, -21.934, -20.148, -18.781,
-17.767, -16.954, -16.300, -15.674, -15.028, -14.377, -23.765, -21.963, -20.172,
-18.800, -17.782, -16.965, -16.308, -15.682, -15.035, -14.384, -23.792, -21.992,
-20.196, -18.819, -17.797, -16.976, -16.317, -15.690, -15.042, -14.392, -23.818,
-22.021, -20.220, -18.843, -17.812, -16.988, -16.326, -15.698, -15.049, -14.399,
 -23.845, -22.049, -20.245, -18.863, -17.827, -16.999, -16.335, -15.706,
-15.056, -14.405, -23.872, -22.078, -20.269, -18.882, -17.843, -17.011, -16.343,
-15.714, -15.064, -14.411, -23.898, -22.106, -20.294, -18.902, -17.858, -17.023,
-16.352, -15.722, -15.071, -14.418, -23.925, -22.135, -20.320, -18.922, -17.873,
-17.037, -16.361, -15.730, -15.078, -14.424, -23.951, -22.163, -20.345, -18.943,
-17.889, -17.057, -16.370, -15.738, -15.086, -14.431, -23.977, -22.191, -20.371,
-18.963, -17.905, -17.069, -16.379, -15.746, -15.093, -14.437, -24.004, -22.220,
-20.397, -18.983, -17.920, -17.081, -16.388, -15.753, -15.100, -14.443, -24.030,
-22.248, -20.423, -19.004, -17.936, -17.092, -16.397, -15.761, -15.107, -14.450,
 -24.056, -22.276, -20.449, -19.024, -17.952, -17.104, -16.406, -15.769,
-15.114, -14.456, -24.083, -22.304, -20.476, -19.045, -17.968, -17.116, -16.415,
-15.777, -15.121, -14.462, -24.109, -22.332, -20.503, -19.106, -17.984, -17.128,
-16.424, -15.785, -15.128, -14.469, -24.135, -22.360, -20.529, -19.127, -18.000,
-17.140, -16.433, -15.794, -15.135, -14.475, -24.161, -22.388, -20.556, -19.148,
-18.016, -17.152, -16.442, -15.801, -15.142, -14.481, -24.187, -22.420, -20.584,
-19.169, -18.032, -17.164, -16.451, -15.809, -15.149, -14.488, -24.213, -22.448,
-20.611, -19.190, -18.048, -17.176, -16.460, -15.817, -15.156, -14.494, -24.239,
-22.475, -20.638, -19.211, -18.064, -17.188, -16.469, -15.825, -15.163, -14.500,
 -24.265, -22.503, -20.666, -19.232, -18.080, -17.200, -16.478, -15.833,
-15.170, -14.507, -24.291, -22.531, -20.694, -19.253, -18.097, -17.212, -16.487,
-15.841, -15.177, -14.513, -24.317, -22.558, -20.721, -19.273, -18.113, -17.224,
-16.496, -15.849, -15.184, -14.520, -24.343, -22.586, -20.749, -19.294, -18.129,
-17.236, -16.505, -15.857, -15.192, -14.527, -24.369, -22.613, -20.777, -19.315,
-18.145, -17.248, -16.514, -15.865, -15.199, -14.533, -24.394, -22.641, -20.805,
-19.335, -18.160, -17.260, -16.523, -15.873, -15.206, -14.540, -25.527, -22.668,
-20.833, -19.356, -18.176, -17.272, -16.532, -15.881, -15.213, -14.546, -25.554,
-22.696, -20.861, -19.376, -18.192, -17.284, -16.541, -15.889, -15.220, -14.552,
 -25.581, -22.723, -20.889, -19.397, -18.208, -17.296, -16.550, -15.897,
-15.227, -14.559, -25.608, -22.750, -20.917, -19.417, -18.223, -17.308, -16.558,
-15.905, -15.234, -14.565, -25.635, -22.777, -20.946, -19.437, -18.243, -17.319,
-16.567, -15.913, -15.241, -14.572, -25.662, -22.805, -20.974, -19.457, -18.258,
-17.331, -16.576, -15.920, -15.248, -14.579, -25.689, -22.832, -21.002, -19.477,
-18.274, -17.343, -16.585, -15.928, -15.255, -14.585, -25.716, -22.859, -21.030,
-19.497, -18.289, -17.355, -16.594, -15.936, -15.262, -14.591, -25.743, -22.886,
-21.063, -19.516, -18.304, -17.366, -16.603, -15.949, -15.270, -14.598, -25.770,
-22.913, -21.091, -19.536, -18.319, -17.378, -16.612, -15.956, -15.277, -14.604,
 -25.797, -22.940, -21.120, -19.555, -18.334, -17.390, -16.621, -15.965,
-15.289, -14.610, -25.824, -22.967, -21.148, -19.574, -18.349, -17.401, -16.630,
-15.973, -15.296, -14.616, -25.850, -23.496, -21.176, -19.594, -18.364, -17.413,
-16.639, -15.980, -15.303, -14.623, -25.877, -23.525, -21.205, -19.613, -18.379,
-17.425, -16.652, -15.988, -15.310, -14.629, -25.904, -23.553, -21.233, -19.632,
-18.413, -17.436, -16.661, -15.996, -15.317, -14.636, -25.930, -23.581, -21.261,
-19.651, -18.428, -17.448, -16.670, -16.003, -15.324, -14.642, -25.957, -23.609,
-21.289, -19.670, -18.442, -17.459, -16.679, -16.011, -15.331, -14.648, -25.983,
-23.636, -21.317, -19.688, -18.457, -17.471, -16.688, -16.018, -15.338, -14.655,
 -26.010, -23.664, -21.345, -19.707, -18.472, -17.483, -16.697, -16.026,
-15.345, -14.663, -26.036, -23.692, -21.374, -19.726, -18.486, -17.494, -16.706,
-16.034, -15.352, -14.669, -26.063, -23.720, -21.402, -19.745, -18.501, -17.506,
-16.719, -16.041, -15.359, -14.676, -26.089, -23.747, -21.430, -19.764, -18.516,
-17.518, -16.728, -16.049, -15.366, -14.682, -26.116, -23.775, -21.458, -19.783,
-18.530, -17.529, -16.737, -16.057, -15.373, -14.688, -26.142, -23.803, -21.486,
-19.802, -18.545, -17.543, -16.746, -16.064, -15.380, -14.695, -26.168, -23.830,
-21.514, -19.821, -18.560, -17.555, -16.755, -16.072, -15.387, -14.701, -26.194,
-23.858, -21.542, -19.840, -18.575, -17.566, -16.765, -16.080, -15.394, -14.707,
 -26.221, -23.885, -21.570, -19.859, -18.590, -17.578, -16.774, -16.087,
-15.401, -14.713, -26.247, -23.912, -21.597, -19.878, -18.605, -17.590, -16.783,
-16.095, -15.408, -14.720, -26.273, -23.940, -21.625, -19.898, -18.620, -17.602,
-16.792, -16.103, -15.415, -14.726, -26.299, -23.967, -21.855, -19.917, -18.635,
-17.614, -16.801, -16.111, -15.423, -14.733, -26.325, -23.994, -21.883, -19.942,
-18.651, -17.626, -16.811, -16.118, -15.430, -14.740, -26.351, -24.022, -21.912,
-19.962, -18.666, -17.639, -16.820, -16.126, -15.437, -14.746, -26.377, -24.049,
-21.940, -19.982, -18.682, -17.651, -16.829, -16.134, -15.444, -14.752, -26.403,
-24.076, -21.969, -20.002, -18.697, -17.663, -16.839, -16.142, -15.451, -14.758,
 -26.429, -24.103, -21.997, -20.022, -18.713, -17.676, -16.848, -16.150,
-15.458, -14.765, -26.455, -24.130, -22.025, -20.043, -18.729, -17.691, -16.858,
-16.157, -15.465, -14.771, -26.481, -24.157, -22.054, -20.064, -18.745, -17.704,
-16.867, -16.165, -15.472, -14.778, -26.507, -24.184, -22.082, -20.085, -18.761,
-17.716, -16.877, -16.173, -15.479, -14.784, -26.533, -24.211, -22.110, -20.106,
-18.778, -17.729, -16.886, -16.180, -15.486, -14.791, -26.558, -24.238, -22.138,
-20.128, -18.794, -17.742, -16.896, -16.188, -15.493, -14.797, -26.584, -24.265,
-22.166, -20.150, -18.811, -17.755, -16.905, -16.196, -15.501, -14.803, -26.610,
-24.291, -22.194, -20.171, -18.827, -17.767, -16.915, -16.204, -15.509, -14.810,
 -26.636, -24.318, -22.222, -20.194, -18.844, -17.780, -16.925, -16.211,
-15.516, -14.816, -26.661, -24.345, -22.250, -20.216, -18.861, -17.793, -16.934,
-16.219, -15.523, -14.822, -26.687, -24.371, -22.278, -20.239, -18.878, -17.806,
-16.944, -16.227, -15.530, -14.828, -26.713, -24.398, -22.306, -20.261, -18.895,
-17.829, -16.954, -16.235, -15.537, -14.835, -26.738, -24.425, -22.334, -20.284,
-18.912, -17.842, -16.963, -16.242, -15.545, -14.841, -26.764, -24.451, -22.362,
-20.307, -18.929, -17.855, -16.973, -16.250, -15.552, -14.848, -26.789, -24.478,
-22.389, -20.331, -18.947, -17.868, -16.983, -16.258, -15.559, -14.854, -26.815,
-24.504, -22.417, -20.354, -18.964, -17.881, -16.993, -16.266, -15.566, -14.860,
 -26.840, -24.531, -22.445, -20.378, -18.981, -17.894, -17.003, -16.279,
-15.574, -14.866, -26.866, -24.557, -22.472, -20.477, -18.999, -17.908, -17.013,
-16.287, -15.581, -14.873, -26.891, -24.584, -22.500, -20.501, -19.016, -17.921,
-17.023, -16.294, -15.592, -14.879, -26.916, -24.610, -22.527, -20.526, -19.034,
-17.934, -17.032, -16.302, -15.600, -14.885, -26.942, -24.636, -22.555, -20.550,
-19.051, -17.947, -17.042, -16.310, -15.607, -14.892, -26.967, -24.662, -22.582,
-20.575, -19.069, -17.960, -17.055, -16.318, -15.614, -14.898, -26.993, -24.689,
-22.609, -20.599, -19.091, -17.974, -17.065, -16.325, -15.621, -14.905, -27.018,
-24.715, -22.637, -20.624, -19.108, -17.987, -17.075, -16.333, -15.628, -14.911,
 -27.043, -24.741, -22.664, -20.649, -19.126, -18.000, -17.085, -16.341,
-15.635, -14.917, -27.068, -24.767, -22.691, -20.675, -19.144, -18.013, -17.095,
-16.349, -15.643, -14.923, -27.093, -24.793, -22.718, -20.700, -19.161, -18.027,
-17.105, -16.356, -15.650, -14.934, -27.119, -24.819, -22.750, -20.725, -19.179,
-18.040, -17.115, -16.364, -15.658, -14.940, -27.144, -24.846, -22.777, -20.751,
-19.196, -18.053, -17.125, -16.372, -15.665, -14.947, -33.865, -24.872, -22.804,
-20.776, -19.214, -18.066, -17.135, -16.383, -15.672, -14.953, -33.891, -24.897,
-22.831, -20.802, -19.231, -18.079, -17.145, -16.391, -15.679, -14.959, -33.918,
-24.923, -22.858, -20.828, -19.249, -18.094, -17.154, -16.398, -15.686, -14.965,
 -33.944, -24.949, -22.885, -20.854, -19.266, -18.107, -17.164, -16.406,
-15.694, -14.972, -33.971, -24.975, -22.911, -20.879, -19.283, -18.120, -17.174,
-16.414, -15.701, -14.979, -33.997, -25.001, -22.938, -20.905, -19.301, -18.133,
-17.184, -16.422, -15.708, -14.985, -34.024, -25.027, -22.965, -20.931, -19.318,
-18.146, -17.194, -16.429, -15.715, -14.991, -34.050, -25.053, -22.992, -20.958,
-19.335, -18.159, -17.204, -16.437, -15.722, -14.998, -34.076, -25.079, -23.019,
-20.984, -19.352, -18.172, -17.214, -16.445, -15.729, -15.004, -34.103, -25.104,
-23.045, -21.010, -19.369, -18.185, -17.227, -16.453, -15.737, -15.010, -34.129,
-25.130, -23.072, -21.036, -19.386, -18.198, -17.237, -16.461, -15.744, -15.017,
 -34.156, -25.156, -23.098, -21.062, -19.403, -18.211, -17.247, -16.469,
-15.751, -15.023, -34.182, -26.124, -23.125, -21.089, -19.420, -18.224, -17.257,
-16.477, -15.758, -15.029, -34.209, -26.150, -23.151, -21.115, -19.437, -18.237,
-17.267, -16.484, -15.766, -15.036, -34.235, -26.177, -23.178, -21.141, -19.489,
-18.250, -17.277, -16.492, -15.773, -15.042, -34.262, -26.204, -23.204, -21.168,
-19.506, -18.263, -17.287, -16.500, -15.780, -15.048, -34.288, -26.230, -23.231,
-21.194, -19.522, -18.276, -17.297, -16.508, -15.787, -15.055, -34.315, -26.257,
-23.257, -21.220, -19.539, -18.289, -17.307, -16.516, -15.794, -15.061, -34.341,
-26.284, -23.283, -21.247, -19.556, -18.302, -17.317, -16.523, -15.801, -15.068,
 -34.368, -26.310, -23.310, -21.278, -19.573, -18.315, -17.332, -16.531,
-15.809, -15.074, -34.394, -26.337, -23.336, -21.304, -19.590, -18.328, -17.342,
-16.539, -15.816, -15.080, -34.421, -26.363, -23.362, -21.331, -19.607, -18.341,
-17.352, -16.547, -15.823, -15.087, -34.447, -26.389, -23.388, -21.357, -19.624,
-18.354, -17.363, -16.555, -15.830, -15.093, -34.473, -26.416, -23.414, -21.384,
-19.641, -18.367, -17.373, -16.563, -15.837, -15.099, -34.500, -26.442, -23.440,
-21.410, -19.658, -18.383, -17.383, -16.571, -15.844, -15.106, -34.526, -26.469,
-23.466, -21.437, -19.675, -18.397, -17.393, -16.578, -15.851, -15.112, -34.553,
-26.495, -23.492, -21.463, -19.692, -18.410, -17.403, -16.586, -15.859, -15.118,
 -34.579, -26.521, -23.518, -21.489, -19.710, -18.423, -17.414, -16.594,
-15.866, -15.125, -34.606, -26.547, -23.544, -21.516, -19.727, -18.436, -17.424,
-16.602, -15.873, -15.131, -34.632, -26.574, -23.987, -21.542, -19.745, -18.449,
-17.434, -16.614, -15.880, -15.138, -34.659, -26.600, -24.014, -21.569, -19.762,
-18.463, -17.445, -16.622, -15.887, -15.145, -34.685, -26.626, -24.041, -21.595,
-19.780, -18.476, -17.455, -16.630, -15.899, -15.151, -34.711, -26.652, -24.067,
-21.621, -19.798, -18.490, -17.465, -16.638, -15.906, -15.158, -34.738, -26.678,
-24.094, -21.648, -19.816, -18.503, -17.476, -16.646, -15.913, -15.164, -34.764,
-26.704, -24.121, -21.674, -19.835, -18.517, -17.489, -16.654, -15.920, -15.170,
 -34.791, -26.730, -24.147, -21.700, -19.853, -18.531, -17.499, -16.662,
-15.927, -15.178, -34.817, -26.756, -24.174, -21.726, -19.872, -18.544, -17.510,
-16.670, -15.934, -15.184, -34.844, -26.782, -24.200, -21.753, -19.890, -18.558,
-17.520, -16.678, -15.941, -15.190, -34.870, -26.808, -24.227, -21.779, -19.909,
-18.572, -17.531, -16.686, -15.948, -15.197, -34.896, -26.834, -24.253, -21.805,
-19.929, -18.586, -17.542, -16.694, -15.955, -15.206, -34.923, -26.860, -24.279,
-21.831, -19.948, -18.600, -17.553, -16.702, -15.962, -15.213, -34.949, -26.886,
-24.306, -21.857, -19.967, -18.614, -17.563, -16.711, -15.970, -15.219, -34.976,
-26.911, -24.332, -21.883, -19.987, -18.628, -17.574, -16.719, -15.977, -15.226,
 -35.002, -26.937, -24.358, -21.909, -20.007, -18.643, -17.585, -16.727,
-15.984, -15.232, -35.029, -26.963, -24.385, -21.935, -20.027, -18.657, -17.596,
-16.735, -15.991, -15.239, -35.055, -26.989, -24.411, -21.961, -20.051, -18.688,
-17.607, -16.743, -15.998, -15.245, -35.081, -27.014, -24.437, -21.987, -20.071,
-18.702, -17.618, -16.752, -16.005, -15.252, -35.108, -27.040, -24.463, -22.013,
-20.092, -18.717, -17.629, -16.760, -16.012, -15.258, -35.134, -27.066, -24.489,
-22.039, -20.112, -18.731, -17.639, -16.768, -16.019, -15.265, -35.161, -27.091,
-24.515, -22.065, -20.133, -18.746, -17.650, -16.776, -16.025, -15.271, -35.187,
-27.117, -24.541, -22.091, -20.154, -18.760, -17.661, -16.784, -16.034, -15.278,
 -35.213, -27.143, -24.567, -22.289, -20.175, -18.775, -17.673, -16.793,
-16.041, -15.284, -35.240, -27.168, -24.593, -22.315, -20.196, -18.789, -17.684,
-16.803, -16.048, -15.290, -35.266, -27.194, -24.619, -22.342, -20.218, -18.804,
-17.695, -16.811, -16.055, -15.297, -35.293, -27.219, -24.645, -22.368, -20.239,
-18.818, -17.706, -16.820, -16.062, -15.303, -35.319, -27.245, -24.671, -22.394,
-20.261, -18.833, -17.717, -16.828, -16.069, -15.310, -35.345, -27.270, -24.697,
-22.420, -20.283, -18.848, -17.728, -16.836, -16.076, -15.317, -35.372, -27.295,
-24.723, -22.446, -20.305, -18.862, -17.739, -16.845, -16.083, -15.323, -35.398,
-27.321, -24.749, -22.473, -20.327, -18.877, -17.750, -16.853, -16.090, -15.330,
 -35.425, -27.346, -24.774, -22.499, -20.349, -18.892, -17.761, -16.861,
-16.097, -15.336, -35.451, -27.372, -24.800, -22.525, -20.372, -18.907, -17.773,
-16.870, -16.104, -15.343, -35.477, -27.397, -24.826, -22.551, -20.394, -18.921,
-17.787, -16.878, -16.111, -15.350, -35.504, -27.422, -24.851, -22.577, -20.417,
-18.936, -17.798, -16.886, -16.118, -15.356, -35.530, -27.447, -24.877, -22.603,
-20.439, -18.951, -17.809, -16.897, -16.125, -15.363, -35.557, -27.473, -24.903,
-22.629, -20.462, -18.966, -17.820, -16.906, -16.132, -15.369, -35.583, -27.498,
-24.928, -22.655, -20.485, -18.980, -17.831, -16.914, -16.139, -15.376, -35.609,
-27.523, -24.954, -22.681, -20.508, -18.995, -17.842, -16.923, -16.146, -15.383,
 -35.636, -27.548, -24.979, -22.706, -20.531, -19.010, -17.853, -16.931,
-16.152, -15.389, -35.662, -27.573, -25.005, -22.732, -20.554, -19.024, -17.864,
-16.939, -16.159, -15.396, -35.689, -27.599, -25.030, -22.758, -20.578, -19.039,
-17.875, -16.948, -16.166, -15.403, -35.715, -27.624, -25.056, -22.784, -20.601,
-19.054, -17.886, -16.956, -16.173, -15.409, -35.741, -27.649, -25.081, -22.814,
-20.624, -19.069, -17.898, -16.968, -16.180, -15.416, -35.768, -27.674, -25.106,
-22.840, -20.648, -19.083, -17.909, -16.976, -16.187, -15.422, -35.794, -27.699,
-25.132, -22.866, -20.671, -19.098, -17.920, -16.985, -16.198, -15.429, -35.820,
-27.724, -25.157, -22.891, -20.695, -19.113, -17.931, -16.993, -16.205, -15.435,
 -35.847, -35.922, -25.182, -22.917, -20.718, -19.128, -17.942, -17.002,
-16.212, -15.442, -35.873, -35.949, -25.208, -22.942, -20.742, -19.143, -17.955,
-17.010, -16.219, -15.449, -35.900, -35.976, -25.233, -22.968, -20.833, -19.161,
-17.966, -17.019, -16.226, -15.455, -35.926, -36.002, -25.258, -22.993, -20.857,
-19.176, -17.977, -17.027, -16.233, -15.462, -35.952, -36.029, -25.283, -23.019,
-20.881, -19.191, -17.988, -17.035, -16.239, -15.469, -35.979, -36.056, -25.308,
-23.044, -20.905, -19.206, -17.999, -17.044, -16.246, -15.475, -36.005, -36.082,
-25.334, -23.070, -20.929, -19.221, -18.010, -17.052, -16.253, -15.486, -36.031,
-36.109, -25.359, -23.095, -20.953, -19.236, -18.030, -17.061, -16.260, -15.492,
 -36.058, -36.136, -25.384, -23.120, -20.977, -19.251, -18.041, -17.069,
-16.267, -15.499, -36.084, -36.163, -25.409, -23.146, -21.001, -19.266, -18.052,
-17.078, -16.274, -15.506, -36.110, -36.189, -25.434, -23.171, -21.026, -19.281,
-18.063, -17.087, -16.281, -15.512, -36.137, -36.216, -25.459, -23.196, -21.050,
-19.296, -18.075, -17.095, -16.288, -15.519, -36.163, -36.243, -25.484, -23.221,
-21.074, -19.311, -18.086, -17.104, -16.295, -15.525, -36.190, -36.269, -25.509,
-23.247, -21.098, -19.327, -18.097, -17.112, -16.301, -15.532, -36.216, -36.296,
-25.534, -23.272, -21.123, -19.342, -18.108, -17.121, -16.308, -15.538, -36.242,
-36.323, -25.559, -23.297, -21.147, -19.358, -18.120, -17.130, -16.315, -15.545,
 -36.269, -36.350, -25.583, -23.322, -21.171, -19.373, -18.131, -17.138,
-16.322, -15.552, -36.295, -36.376, -26.422, -23.347, -21.196, -19.389, -18.142,
-17.147, -16.329, -15.559, -36.321, -36.403, -26.447, -23.372, -21.220, -19.405,
-18.154, -17.156, -16.336, -15.565, -36.348, -36.430, -26.473, -23.397, -21.244,
-19.421, -18.165, -17.165, -16.343, -15.572, -36.374, -36.456, -26.499, -23.422,
-21.269, -19.437, -18.177, -17.174, -16.350, -15.578, -36.400, -36.483, -26.524,
-23.447, -21.293, -19.453, -18.188, -17.182, -16.357, -15.585, -36.427, -36.510,
-26.550, -23.472, -21.317, -19.469, -18.200, -17.191, -16.364, -15.591, -36.453,
-36.536, -26.575, -23.497, -21.347, -19.486, -18.211, -17.200, -16.371, -15.598,
 -36.479, -36.563, -26.601, -23.521, -21.371, -19.502, -18.223, -17.209,
-16.378, -15.605, -36.506, -36.590, -26.626, -23.546, -21.395, -19.519, -18.234,
-17.218, -16.385, -15.611, -36.532, -36.617, -26.652, -23.571, -21.420, -19.535,
-18.246, -17.226, -16.392, -15.618, -36.558, -36.643, -26.677, -23.596, -21.444,
-19.552, -18.258, -17.235, -16.400, -15.624, -36.585, -36.670, -26.703, -23.621,
-21.468, -19.569, -18.270, -17.244, -16.407, -15.631, -36.611, -36.697, -26.728,
-23.645, -21.493, -19.587, -18.281, -17.253, -16.414, -15.637, -36.637, -36.723,
-26.753, -23.670, -21.517, -19.604, -18.293, -17.264, -16.421, -15.644, -36.664,
-36.750, -26.779, -23.695, -21.541, -19.621, -18.305, -17.273, -16.428, -15.650,
 -36.690, -36.777, -26.804, -23.719, -21.566, -19.639, -18.317, -17.282,
-16.435, -15.657, -36.716, -36.803, -26.829, -23.744, -21.590, -19.685, -18.328,
-17.291, -16.442, -15.664, -36.743, -36.830, -26.854, -23.768, -21.614, -19.703,
-18.340, -17.300, -16.449, -15.671, -36.769, -36.857, -26.880, -23.793, -21.639,
-19.721, -18.352, -17.309, -16.456, -15.677, -36.795, -36.883, -26.905, -23.817,
-21.663, -19.739, -18.364, -17.318, -16.463, -15.684, -36.822, -36.910, -26.930,
-23.842, -21.687, -19.757, -18.376, -17.327, -16.470, -15.690, -36.848, -36.937,
-26.955, -24.219, -21.711, -19.776, -18.388, -17.339, -16.479, -15.697, -36.874,
-36.963, -26.980, -24.244, -21.736, -19.794, -18.403, -17.348, -16.486, -15.703,
 -36.901, -36.990, -27.005, -24.269, -21.760, -19.813, -18.415, -17.357,
-16.497, -15.709, -36.927, -37.017, -27.030, -24.294, -21.784, -19.831, -18.427,
-17.366, -16.504, -15.716, -36.953, -37.043, -27.055, -24.319, -21.808, -19.850,
-18.440, -17.375, -16.511, -15.722, -36.980, -37.070, -27.080, -24.344, -21.832,
-19.869, -18.452, -17.384, -16.518, -15.729, -37.006, -37.097, -27.105, -24.369,
-21.857, -19.888, -18.464, -17.393, -16.525, -15.735, -37.032, -37.123, -27.130,
-24.394, -21.881, -19.908, -18.476, -17.402, -16.532, -15.742, -37.059, -37.150,
-27.155, -24.419, -21.905, -19.927, -18.488, -17.412, -16.539, -15.748, -37.085,
-37.177, -27.180, -24.444, -21.929, -19.946, -18.500, -17.421, -16.547, -15.755,
 -37.111, -37.203, -27.205, -24.468, -21.953, -19.966, -18.512, -17.430,
-16.554, -15.765, -37.137, -37.230, -27.230, -24.493, -21.977, -19.986, -18.524,
-17.439, -16.561, -15.772, -37.164, -37.257, -27.255, -24.518, -22.001, -20.006,
-18.537, -17.448, -16.568, -15.778, -37.190, -37.283, -27.280, -24.543, -22.025,
-20.026, -18.549, -17.462, -16.575, -15.784, -37.216, -37.310, -27.305, -24.567,
-22.049, -20.046, -18.561, -17.471, -16.582, -15.791, -37.243, -37.337, -27.329,
-24.592, -22.073, -20.066, -18.573, -17.481, -16.589, -15.797, -37.269, -37.363,
-27.354, -24.617, -22.097, -20.090, -18.585, -17.490, -16.596, -15.803, -37.295,
-37.390, -27.379, -24.641, -22.121, -20.110, -18.597, -17.499, -16.603, -15.810,
 -37.322, -37.417, -27.404, -24.666, -22.145, -20.131, -18.609, -17.508,
-16.610, -15.816, -37.348, -37.443, -27.428, -24.690, -22.169, -20.151, -18.621,
-17.518, -16.617, -15.822, -37.374, -37.470, -27.453, -24.715, -22.193, -20.172,
-18.633, -17.527, -16.624, -15.829, -37.401, -37.496, -27.478, -24.740, -22.216,
-20.193, -18.645, -17.536, -16.631, -15.835, -37.427, -37.523, -27.502, -24.764,
-22.240, -20.213, -18.658, -17.545, -16.639, -15.841, -37.453, -37.550, -27.527,
-24.789, -22.264, -20.234, -18.670, -17.555, -16.646, -15.847, -37.479, -37.576,
-27.551, -24.813, -22.288, -20.255, -18.682, -17.564, -16.653, -15.854, -37.506,
-37.603, -27.576, -24.837, -22.312, -20.276, -18.694, -17.574, -16.660, -15.860,
 -37.532, -37.630, -27.601, -24.862, -22.335, -20.297, -18.706, -17.583,
-16.667, -15.866, -37.558, -37.656, -27.625, -24.886, -22.509, -20.318, -18.718,
-17.592, -16.674, -15.873, -37.585, -37.683, -27.650, -24.910, -22.533, -20.340,
-18.731, -17.602, -16.681, -15.879, -37.611, -37.710, -27.674, -24.935, -22.557,
-20.361, -18.743, -17.611, -16.689, -15.885, -37.637, -37.736, -27.699, -24.959,
-22.581, -20.382, -18.755, -17.621, -16.696, -15.891, -37.663, -37.763, -27.723,
-24.983, -22.605, -20.403, -18.767, -17.630, -16.704, -15.898, -37.690, -37.789,
-27.747, -25.008, -22.629, -20.425, -18.793, -17.639, -16.711, -15.904, -37.716,
-37.816, -27.772, -25.032, -22.653, -20.446, -18.805, -17.649, -16.718, -15.910,
 -37.742, -37.843, -27.796, -25.056, -22.677, -20.468, -18.818, -17.658,
-16.725, -15.916, -37.769, -37.869, -27.821, -25.080, -22.701, -20.489, -18.830,
-17.668, -16.732, -15.923, -37.795, -37.896, -27.845, -25.104, -22.725, -20.511,
-18.842, -17.677, -16.740, -15.929, -37.821, -37.923, -27.869, -25.129, -22.754,
-20.533, -18.854, -17.687, -16.747, -15.935, -37.847, -37.949, -27.894, -25.153,
-22.778, -20.554, -18.866, -17.696, -16.754, -15.941, -37.874, -37.976, -27.918,
-25.177, -22.802, -20.576, -18.879, -17.706, -16.761, -15.947, -37.900, -38.002,
-27.942, -25.201, -22.826, -20.598, -18.891, -17.715, -16.769, -15.954, -37.926,
-38.029, -27.967, -25.225, -22.849, -20.619, -18.903, -17.725, -16.776, -15.960,
 -37.952, -38.056, -37.289, -25.249, -22.873, -20.641, -18.916, -17.736,
-16.785, -15.966, -37.979, -38.082, -37.315, -25.273, -22.897, -20.663, -18.928,
-17.746, -16.792, -15.973, -38.005, -38.109, -37.341, -25.297, -22.921, -20.685,
-18.941, -17.758, -16.803, -15.979, -38.031, -38.136, -37.367, -25.321, -22.944,
-20.707, -18.953, -17.768, -16.810, -15.985, -38.058, -38.162, -37.393, -25.345,
-22.968, -20.729, -18.966, -17.777, -16.817, -15.991, -38.084, -38.189, -37.420,
-25.369, -22.991, -20.751, -18.978, -17.787, -16.824, -15.998, -38.110, -38.215,
-37.446, -25.393, -23.015, -20.773, -18.991, -17.797, -16.832, -16.005, -38.136,
-38.242, -37.472, -25.416, -23.039, -20.794, -19.004, -17.806, -16.839, -16.011,
 -38.163, -38.269, -37.498, -25.440, -23.062, -20.816, -19.016, -17.816,
-16.846, -16.017, -38.189, -38.295, -37.525, -25.464, -23.086, -20.838, -19.029,
-17.826, -16.854, -16.023, -38.215, -38.322, -37.551, -25.488, -23.109, -20.860,
-19.042, -17.835, -16.861, -16.034, -38.241, -38.348, -37.577, -25.512, -23.133,
-20.882, -19.058, -17.845, -16.869, -16.040, -38.268, -38.375, -37.603, -25.535,
-23.156, -20.904, -19.071, -17.855, -16.876, -16.046, -38.294, -38.402, -37.629,
-25.559, -23.180, -20.926, -19.083, -17.865, -16.883, -16.052, -38.320, -38.428,
-37.656, -25.583, -23.203, -20.948, -19.096, -17.875, -16.891, -16.058, -38.346,
-38.455, -37.682, -25.607, -23.226, -20.971, -19.109, -17.885, -16.898, -16.064,
 -38.373, -38.481, -37.708, -25.630, -23.250, -21.055, -19.122, -17.894,
-16.906, -16.071, -38.399, -38.508, -37.734, -25.654, -23.273, -21.077, -19.135,
-17.904, -16.913, -16.078, -38.425, -38.535, -37.760, -25.678, -23.296, -21.099,
-19.148, -17.914, -16.920, -16.084, -38.451, -38.561, -37.787, -25.701, -23.320,
-21.121, -19.161, -17.924, -16.928, -16.090, -38.478, -38.588, -37.813, -25.725,
-23.343, -21.144, -19.174, -17.934, -16.935, -16.096, -38.504, -38.614, -37.839,
-26.465, -23.366, -21.166, -19.188, -17.944, -16.945, -16.102, -38.530, -38.641,
-37.865, -26.489, -23.389, -21.188, -19.201, -17.954, -16.953, -16.108, -38.556,
-38.667, -37.891, -26.514, -23.413, -21.210, -19.214, -17.964, -16.960, -16.114,
 -38.583, -38.694, -37.918, -26.538, -23.436, -21.232, -19.227, -17.974,
-16.968, -16.120, -38.609, -38.721, -37.944, -26.562, -23.459, -21.255, -19.240,
-17.984, -16.975, -16.127, -38.635, -38.747, -37.970, -26.586, -23.482, -21.282,
-19.254, -17.994, -16.983, -16.133, -38.661, -38.774, -37.996, -26.610, -23.505,
-21.304, -19.267, -18.004, -16.990, -16.139, -38.688, -38.800, -38.022, -26.634,
-23.528, -21.327, -19.280, -18.014, -16.998, -16.145, -38.714, -38.827, -38.049,
-26.658, -23.551, -21.349, -19.294, -18.024, -17.006, -16.151, -38.740, -38.854,
-38.075, -26.682, -23.574, -21.371, -19.307, -18.034, -17.013, -16.157, -38.766,
-38.880, -38.101, -26.706, -23.597, -21.393, -19.320, -18.044, -17.021, -16.164,
 -38.793, -38.907, -38.127, -26.730, -23.620, -21.415, -19.334, -18.054,
-17.029, -16.170, -38.819, -38.933, -38.153, -26.754, -23.643, -21.438, -19.347,
-18.072, -17.036, -16.176, -38.845, -38.960, -38.180, -26.778, -23.666, -21.460,
-19.361, -18.082, -17.044, -16.182, -38.871, -38.986, -38.206, -26.802, -23.689,
-21.482, -19.374, -18.092, -17.051, -16.188, -38.898, -39.013, -38.232, -26.826,
-23.712, -21.504, -19.388, -18.102, -17.059, -16.194, -38.924, -39.040, -38.258,
-26.850, -23.735, -21.526, -19.401, -18.112, -17.067, -16.200, -38.950, -39.066,
-38.284, -26.874, -23.758, -21.548, -19.415, -18.122, -17.074, -16.207, -38.976,
-39.093, -38.310, -26.897, -23.781, -21.571, -19.428, -18.132, -17.082, -16.213,
 -39.003, -39.119, -38.337, -26.921, -23.804, -21.593, -19.442, -18.142,
-17.090, -16.219, -39.029, -39.146, -38.363, -26.945, -23.826, -21.615, -19.456,
-18.152, -17.097, -16.225, -39.055, -39.172, -38.389, -26.969, -23.849, -21.637,
-19.469, -18.163, -17.105, -16.231, -39.081, -39.199, -38.415, -26.992, -23.872,
-21.659, -19.483, -18.174, -17.113, -16.237, -39.107, -39.226, -38.441, -27.016,
-23.895, -21.681, -19.497, -18.184, -17.123, -16.243, -39.134, -39.252, -38.467,
-27.040, -23.917, -21.703, -19.510, -18.194, -17.131, -16.249, -39.160, -39.279,
-38.494, -27.064, -23.940, -21.725, -19.524, -18.204, -17.139, -16.256, -39.186,
-39.305, -38.520, -27.087, -23.963, -21.747, -19.538, -18.214, -17.146, -16.262,
 -39.212, -39.332, -38.546, -27.111, -24.293, -21.769, -19.552, -18.224,
-17.154, -16.268, -39.239, -39.358, -38.572, -27.135, -24.316, -21.791, -19.566,
-18.234, -17.162, -16.275, -39.265, -39.385, -38.598, -27.158, -24.340, -21.813,
-19.580, -18.244, -17.170, -16.281, -39.291, -39.411, -38.624, -27.182, -24.363,
-21.835, -19.594, -18.255, -17.177, -16.287, -39.317, -39.438, -38.651, -27.205,
-24.386, -21.857, -19.608, -18.265, -17.187, -16.297, -39.343, -39.465, -38.677,
-27.229, -24.409, -21.879, -19.643, -18.275, -17.195, -16.303, -39.370, -39.491,
-38.703, -27.253, -24.432, -21.901, -19.657, -18.288, -17.203, -16.309, -39.396,
-39.518, -38.729, -27.276, -24.455, -21.923, -19.671, -18.298, -17.210, -16.315,
 -39.422, -39.544, -38.755, -27.300, -24.478, -21.945, -19.686, -18.308,
-17.218, -16.321, -39.448, -39.571, -38.781, -27.323, -24.501, -21.967, -19.700,
-18.319, -17.226, -16.327, -39.475, -39.597, -38.808, -27.347, -24.523, -21.988,
-19.714, -18.329, -17.234, -16.335, -39.501, -39.624, -38.834, -27.370, -24.546,
-22.010, -19.728, -18.339, -17.242, -16.341, -39.527, -39.650, -38.860, -27.393,
-24.569, -22.032, -19.743, -18.349, -17.250, -16.347, -39.553, -39.677, -38.886,
-27.417, -24.592, -22.054, -19.757, -18.360, -17.257, -16.353, -39.579, -39.703,
-38.912, -27.440, -24.615, -22.076, -19.771, -18.370, -17.265, -16.359, -39.606,
-39.730, -38.938, -27.464, -24.638, -22.098, -19.786, -18.380, -17.273, -16.366,
 -39.632, -39.757, -38.964, -27.487, -24.660, -22.119, -19.804, -18.391,
-17.281, -16.372, -39.658, -39.783, -38.991, -27.510, -24.683, -22.141, -19.819,
-18.401, -17.289, -16.378, -39.684, -39.810, -39.017, -27.534, -24.706, -22.163,
-19.834, -18.412, -17.297, -16.384, -39.710, -39.836, -39.043, -27.557, -24.729,
-22.185, -19.849, -18.422, -17.305, -16.390, -39.737, -39.863, -39.069, -27.580,
-24.751, -22.206, -19.863, -18.432, -17.313, -16.396, -39.763, -39.889, -39.095,
-27.604, -24.774, -22.228, -19.879, -18.443, -17.321, -16.403, -39.789, -39.916,
-39.121, -27.627, -24.797, -22.250, -19.894, -18.453, -17.329, -16.409, -39.815,
-39.942, -39.147, -27.650, -24.819, -22.271, -19.909, -18.464, -17.337, -16.415,
 -39.841, -39.969, -39.174, -27.673, -24.842, -22.293, -19.924, -18.474,
-17.345, -16.421, -39.868, -39.995, -39.200, -27.697, -24.865, -22.315, -19.940,
-18.485, -17.353, -16.427, -39.894, -40.022, -39.226, -27.720, -24.887, -22.336,
-19.955, -18.495, -17.361, -16.434, -39.920, -40.048, -39.252, -27.743, -24.910,
-22.358, -19.971, -18.506, -17.370, -16.440, -39.946, -40.075, -39.278, -27.766,
-24.932, -22.379, -19.986, -18.516, -17.378, -16.446, -39.972, -40.101, -39.304,
-27.789, -24.955, -22.401, -20.002, -18.527, -17.386, -16.452, -39.999, -40.128,
-39.330, -27.812, -24.978, -22.422, -20.018, -18.538, -17.394, -16.459, -40.025,
-40.155, -39.356, -27.836, -25.000, -22.586, -20.034, -18.548, -17.402, -16.466,
 -40.051, -40.181, -39.383, -27.859, -25.023, -22.608, -20.050, -18.559,
-17.410, -16.472, -40.077, -40.208, -39.409, -27.882, -25.045, -22.630, -20.066,
-18.569, -17.418, -16.479, -40.103, -40.234, -39.435, -27.905, -25.068, -22.651,
-20.082, -18.580, -17.426, -16.485, -40.130, -40.261, -39.461, -27.928, -25.090,
-22.673, -20.099, -18.591, -17.435, -16.491, -40.156, -40.287, -39.487, -38.070,
-25.112, -22.695, -20.115, -18.602, -17.447, -16.497, -40.182, -40.314, -39.513,
-38.095, -25.135, -22.716, -20.132, -18.613, -17.457, -16.504, -40.208, -40.340,
-39.539, -38.120, -25.157, -22.738, -20.148, -18.624, -17.465, -16.510, -40.234,
-40.367, -39.565, -38.145, -25.180, -22.760, -20.165, -18.635, -17.473, -16.516,
 -40.260, -40.393, -39.592, -38.170, -25.202, -22.781, -20.182, -18.645,
-17.482, -16.523, -40.287, -40.420, -39.618, -38.195, -25.224, -22.803, -20.199,
-18.656, -17.490, -16.529, -40.313, -40.446, -39.644, -38.220, -25.247, -22.825,
-20.216, -18.667, -17.498, -16.536, -40.339, -40.473, -39.670, -38.246, -25.269,
-22.846, -20.233, -18.678, -17.506, -16.542, -40.365, -40.499, -39.696, -38.271,
-25.291, -22.868, -20.250, -18.688, -17.515, -16.548, -40.391, -40.526, -39.722,
-38.296, -25.313, -22.889, -20.267, -18.699, -17.523, -16.555, -40.418, -40.552,
-39.748, -38.321, -25.336, -22.911, -20.285, -18.710, -17.531, -16.565, -40.444,
-40.579, -39.774, -38.346, -25.358, -22.932, -20.302, -18.721, -17.539, -16.571,
 -40.470, -40.605, -39.800, -38.371, -25.380, -22.954, -20.320, -18.743,
-17.548, -16.577, -40.496, -40.632, -39.827, -38.396, -25.402, -22.975, -20.337,
-18.754, -17.556, -16.583, -40.522, -40.658, -39.853, -38.422, -25.425, -22.997,
-20.355, -18.764, -17.564, -16.590, -40.548, -40.685, -39.879, -38.447, -25.447,
-23.018, -20.373, -18.775, -17.573, -16.596, -40.575, -40.711, -39.905, -38.472,
-25.469, -23.040, -20.391, -18.786, -17.581, -16.602, -40.601, -40.738, -39.931,
-38.497, -25.491, -23.061, -20.409, -18.797, -17.589, -16.609, -40.627, -40.764,
-39.957, -38.522, -25.513, -23.082, -20.427, -18.808, -17.598, -16.615, -40.653,
-40.791, -39.983, -38.547, -25.535, -23.104, -20.445, -18.818, -17.606, -16.621,
 -40.679, -40.817, -40.009, -38.572, -25.557, -23.125, -20.463, -18.829,
-17.614, -16.628, -40.706, -40.844, -40.035, -38.598, -25.579, -23.147, -20.481,
-18.840, -17.625, -16.634, -40.732, -40.870, -40.061, -38.623, -25.602, -23.168,
-20.500, -18.851, -17.633, -16.640, -40.758, -40.897, -40.088, -38.648, -25.624,
-23.189, -20.518, -18.865, -17.641, -16.647, -40.784, -40.923, -40.114, -38.673,
-25.646, -23.211, -20.537, -18.876, -17.650, -16.653, -40.810, -40.950, -40.140,
-38.698, -25.668, -23.232, -20.555, -18.887, -17.658, -16.660, -40.836, -40.976,
-40.166, -38.723, -26.337, -23.253, -20.574, -18.898, -17.667, -16.667, -40.863,
-41.003, -40.192, -38.748, -26.359, -23.274, -20.592, -18.909, -17.675, -16.674,
 -40.889, -41.029, -40.218, -38.773, -26.382, -23.296, -20.611, -18.920,
-17.684, -16.680, -40.915, -41.056, -40.244, -38.798, -26.404, -23.317, -20.630,
-18.931, -17.692, -16.686, -40.941, -41.082, -40.270, -38.824, -26.426, -23.338,
-20.648, -18.942, -17.700, -16.693, -40.967, -41.109, -40.296, -38.849, -26.449,
-23.359, -20.667, -18.953, -17.709, -16.699, -40.993, -41.135, -40.322, -38.874,
-26.471, -23.380, -20.729, -18.964, -17.718, -16.706, -41.019, -41.162, -40.348,
-38.899, -26.493, -23.402, -20.748, -18.975, -17.726, -16.712, -41.046, -41.188,
-40.374, -38.924, -26.516, -23.423, -20.767, -18.986, -17.735, -16.719, -41.072,
-41.215, -40.401, -38.949, -26.538, -23.444, -20.786, -18.997, -17.743, -16.725,
 -41.098, -41.241, -40.427, -38.974, -26.560, -23.465, -20.805, -19.008,
-17.752, -16.732, -41.124, -41.268, -40.453, -38.999, -26.583, -23.486, -20.824,
-19.019, -17.760, -16.738, -41.150, -41.294, -40.479, -39.024, -26.605, -23.507,
-20.844, -19.030, -17.769, -16.745, -41.176, -41.321, -40.505, -39.050, -26.627,
-23.528, -20.863, -19.041, -17.777, -16.751, -41.203, -41.347, -40.531, -39.075,
-26.649, -23.549, -20.882, -19.053, -17.786, -16.757, -41.229, -41.374, -40.557,
-39.100, -26.672, -23.570, -20.902, -19.064, -17.795, -16.764, -41.255, -41.400,
-40.583, -39.125, -26.694, -23.591, -20.921, -19.075, -17.803, -16.771, -41.281,
-41.427, -40.609, -39.150, -26.716, -23.612, -20.940, -19.086, -17.814, -16.777,
 -41.307, -41.453, -40.635, -39.175, -26.738, -23.633, -20.960, -19.097,
-17.822, -16.784, -41.333, -41.480, -40.661, -39.200, -26.760, -23.654, -20.979,
-19.109, -17.831, -16.790, -41.360, -41.506, -40.687, -39.225, -26.782, -23.675,
-20.999, -19.120, -17.839, -16.797, -41.386, -41.532, -40.713, -39.250, -26.805,
-23.696, -21.018, -19.131, -17.848, -16.804, -41.412, -41.559, -40.740, -39.275,
-26.827, -23.717, -21.038, -19.143, -17.857, -16.810, -41.438, -41.585, -40.766,
-39.301, -26.849, -23.738, -21.058, -19.154, -17.865, -16.817, -41.464, -41.612,
-40.792, -39.326, -26.871, -23.759, -21.077, -19.166, -17.874, -16.823, -41.490,
-41.638, -40.818, -39.351, -26.893, -23.780, -21.097, -19.177, -17.883, -16.830,
 -41.516, -41.665, -40.844, -39.376, -26.915, -23.800, -21.117, -19.188,
-17.891, -16.840, -41.543, -41.691, -40.870, -39.401, -26.937, -23.821, -21.136,
-19.200, -17.900, -16.846, -41.569, -41.718, -40.896, -39.426, -26.959, -23.842,
-21.156, -19.211, -17.909, -16.853, -41.595, -41.744, -40.922, -39.451, -26.981,
-23.863, -21.176, -19.223, -17.918, -16.859, -41.621, -41.771, -40.948, -39.476,
-27.003, -23.884, -21.195, -19.235, -17.926, -16.866, -41.647, -41.797, -40.974,
-39.501, -27.025, -23.905, -21.215, -19.246, -17.935, -16.873, -41.673, -41.824,
-41.000, -39.526, -27.047, -24.202, -21.235, -19.258, -17.944, -16.881, -41.699,
-41.850, -41.026, -39.551, -27.069, -24.223, -21.255, -19.269, -17.953, -16.888,
 -41.726, -41.877, -41.052, -39.576, -27.091, -24.244, -21.275, -19.281,
-17.961, -16.895, -41.752, -41.903, -41.078, -39.602, -27.113, -24.265, -21.294,
-19.293, -17.970, -16.901, -41.778, -41.930, -41.104, -39.627, -27.135, -24.286,
-21.314, -19.305, -17.979, -16.908, -41.804, -41.956, -41.130, -39.652, -27.157,
-24.307, -21.334, -19.316, -17.994, -16.915, -41.830, -41.982, -41.156, -39.677,
-27.178, -24.328, -21.354, -19.328, -18.002, -16.921, -41.856, -42.009, -41.183,
-39.702, -27.200, -24.349, -21.374, -19.340, -18.011, -16.928, -41.882, -42.035,
-41.209, -39.727, -27.222, -24.370, -21.394, -19.352, -18.020, -16.935, -41.908,
-42.062, -41.235, -39.752, -27.244, -24.391, -21.414, -19.364, -18.029, -16.941,
 -41.935, -42.088, -41.261, -39.777, -27.266, -24.412, -21.433, -19.376,
-18.038, -16.948, -41.961, -42.115, -41.287, -39.802, -27.288, -24.433, -21.453,
-19.387, -18.046, -16.955, -41.987, -42.141, -41.313, -39.827, -27.309, -24.454,
-21.473, -19.399, -18.055, -16.962, -42.013, -42.168, -41.339, -39.852, -27.331,
-24.475, -21.493, -19.411, -18.064, -16.968, -42.039, -42.194, -41.365, -39.877,
-27.353, -24.496, -21.513, -19.423, -18.073, -16.975, -42.065, -42.221, -41.391,
-39.902, -27.375, -24.516, -21.533, -19.435, -18.084, -16.982, -42.091, -42.247,
-41.417, -39.927, -27.396, -24.537, -21.553, -19.447, -18.093, -16.989, -42.118,
-42.273, -41.443, -39.952, -27.418, -24.558, -21.573, -19.479, -18.102, -16.995,
 -42.144, -42.300, -41.469, -39.978, -27.440, -24.579, -21.593, -19.491,
-18.111, -17.002, -42.170, -42.326, -41.495, -40.003, -27.462, -24.600, -21.613,
-19.503, -18.120, -17.009, -42.196, -42.353, -41.521, -40.028, -27.483, -24.621,
-21.633, -19.515, -18.129, -17.017, -42.222, -42.379, -41.547, -40.053, -27.505,
-24.641, -21.653, -19.527, -18.138, -17.024, -42.248, -42.406, -41.573, -40.078,
-27.527, -24.662, -21.672, -19.540, -18.147, -17.031, -42.274, -42.432, -41.599,
-40.103, -27.548, -24.683, -21.692, -19.552, -18.156, -17.038, -42.300, -42.459,
-41.625, -40.128, -27.570, -24.704, -21.712, -19.564, -18.165, -17.044, -42.327,
-42.485, -41.651, -40.153, -27.592, -24.724, -21.732, -19.576, -18.174, -17.051,
 -42.353, -42.511, -41.677, -40.178, -27.613, -24.745, -21.757, -19.588,
-18.183, -17.058, -42.379, -42.538, -41.703, -40.203, -27.635, -24.766, -21.777,
-19.601, -18.193, -17.065, -42.405, -42.564, -41.729, -40.228, -27.656, -24.786,
-21.797, -19.613, -18.202, -17.072, -42.431, -42.591, -41.755, -40.253, -27.678,
-24.807, -21.817, -19.625, -18.211, -17.079, -42.457, -42.617, -41.781, -40.278,
-27.699, -24.828, -21.836, -19.638, -18.220, -17.086, -42.483, -42.644, -41.807,
-40.303, -38.355, -24.848, -21.856, -19.650, -18.229, -17.093, -42.509, -42.670,
-41.833, -40.328, -38.379, -24.869, -21.876, -19.663, -18.238, -17.100, -42.535,
-42.697, -41.860, -40.353, -38.403, -24.890, -21.896, -19.675, -18.247, -17.107,
 -42.562, -42.723, -41.886, -40.378, -38.426, -24.910, -21.916, -19.688,
-18.256, -17.114, -42.588, -42.749, -41.912, -40.403, -38.450, -24.931, -21.936,
-19.701, -18.265, -17.123, -42.614, -42.776, -41.938, -40.428, -38.474, -24.951,
-21.956, -19.713, -18.274, -17.130, -42.640, -42.802, -41.964, -40.453, -38.497,
-24.972, -22.054, -19.726, -18.283, -17.137, -42.666, -42.829, -41.990, -40.478,
-38.521, -24.992, -22.074, -19.739, -18.292, -17.144, -42.692, -42.855, -42.016,
-40.504, -38.544, -25.013, -22.094, -19.752, -18.301, -17.151, -42.718, -42.882,
-42.042, -40.529, -38.568, -25.033, -22.114, -19.764, -18.310, -17.158, -42.744,
-42.908, -42.068, -40.554, -38.592, -25.054, -22.134, -19.777, -18.319, -17.165,
 -42.770, -42.934, -42.094, -40.579, -38.615, -25.074, -22.154, -19.790,
-18.328, -17.172, -42.797, -42.961, -42.120, -40.604, -38.639, -25.095, -22.174,
-19.803, -18.337, -17.179, -42.823, -42.987, -42.146, -40.629, -38.663, -25.115,
-22.194, -19.816, -18.346, -17.186, -42.849, -43.014, -42.172, -40.654, -38.686,
-25.136, -22.214, -19.829, -18.356, -17.193, -42.875, -43.040, -42.198, -40.679,
-38.710, -25.156, -22.234, -19.843, -18.365, -17.200, -42.901, -43.067, -42.224,
-40.704, -38.734, -25.177, -22.254, -19.856, -18.374, -17.207, -42.927, -43.093,
-42.250, -40.729, -38.757, -25.197, -22.274, -19.869, -18.383, -17.214, -42.953,
-43.119, -42.276, -40.754, -38.781, -25.218, -22.294, -19.883, -18.392, -17.221,
 -42.979, -43.146, -42.302, -40.779, -38.805, -25.238, -22.314, -19.896,
-18.401, -17.228, -43.005, -43.172, -42.328, -40.804, -38.828, -25.258, -22.334,
-19.909, -18.411, -17.235, -43.032, -43.199, -42.354, -40.829, -38.852, -25.279,
-22.354, -19.923, -18.420, -17.243, -43.058, -43.225, -42.380, -40.854, -38.876,
-25.299, -22.374, -19.937, -18.429, -17.250, -43.084, -43.252, -42.406, -40.879,
-38.899, -25.319, -22.394, -19.950, -18.438, -17.257, -43.110, -43.278, -42.432,
-40.904, -38.923, -25.340, -22.413, -19.964, -18.448, -17.264, -43.136, -43.304,
-42.458, -40.929, -38.946, -25.360, -22.433, -19.978, -18.457, -17.271, -43.162,
-43.331, -42.484, -40.954, -38.970, -25.380, -22.453, -19.992, -18.466, -17.278,
 -43.188, -43.357, -42.510, -40.979, -38.994, -25.401, -22.473, -20.006,
-18.476, -17.285, -43.214, -43.384, -42.536, -41.004, -39.017, -25.421, -22.493,
-20.020, -18.485, -17.292, -43.240, -43.410, -42.562, -41.029, -39.041, -25.441,
-22.513, -20.034, -18.494, -17.300, -43.266, -43.436, -42.588, -41.054, -39.065,
-25.462, -22.532, -20.048, -18.504, -17.307, -43.293, -43.463, -42.614, -41.079,
-39.088, -25.482, -22.552, -20.062, -18.513, -17.317, -43.319, -43.489, -42.640,
-41.104, -39.112, -26.100, -22.572, -20.076, -18.522, -17.324, -43.345, -43.516,
-42.666, -41.129, -39.135, -26.121, -22.592, -20.091, -18.532, -17.332, -43.371,
-43.542, -42.692, -41.154, -39.159, -26.141, -22.612, -20.105, -18.541, -17.339,
 -43.397, -43.569, -42.718, -41.179, -39.183, -26.162, -22.631, -20.120,
-18.551, -17.346, -43.423, -43.595, -42.744, -41.204, -39.206, -26.182, -22.651,
-20.134, -18.571, -17.353, -43.449, -43.621, -42.770, -41.229, -39.230, -26.203,
-22.671, -20.149, -18.580, -17.360, -43.475, -43.648, -42.796, -41.254, -39.254,
-26.223, -22.690, -20.163, -18.591, -17.368, -43.501, -43.674, -42.822, -41.279,
-39.277, -26.244, -22.710, -20.181, -18.600, -17.375, -43.527, -43.701, -42.848,
-41.304, -39.301, -26.264, -22.730, -20.196, -18.610, -17.382, -43.553, -43.727,
-42.874, -41.329, -39.324, -26.285, -22.750, -20.211, -18.619, -17.389, -43.580,
-43.753, -42.900, -41.354, -39.348, -26.305, -22.769, -20.226, -18.628, -17.399,
 -43.606, -43.780, -42.926, -41.379, -39.372, -26.326, -22.789, -20.241,
-18.638, -17.406, -43.632, -43.806, -42.952, -41.404, -39.395, -26.346, -22.809,
-20.256, -18.647, -17.413, -43.658, -43.833, -42.978, -41.429, -39.419, -26.367,
-22.828, -20.272, -18.657, -17.420, -43.684, -43.859, -43.004, -41.454, -39.442,
-26.387, -22.848, -20.287, -18.666, -17.430, -43.710, -43.885, -43.030, -41.479,
-39.466, -26.407, -22.867, -20.302, -18.676, -17.437, -43.736, -43.912, -43.056,
-41.504, -39.490, -26.428, -22.892, -20.318, -18.685, -17.444, -43.762, -43.938,
-43.082, -41.529, -39.513, -26.448, -22.912, -20.359, -18.695, -17.452, -43.788,
-43.965, -43.108, -41.554, -39.537, -26.469, -22.932, -20.375, -18.704, -17.459,
 -43.814, -43.991, -43.134, -41.579, -39.561, -26.489, -22.951, -20.390,
-18.714, -17.466, -43.840, -44.017, -43.160, -41.604, -39.584, -26.509, -22.971,
-20.406, -18.723, -17.474, -43.866, -44.044, -43.186, -41.629, -39.608, -26.529,
-22.990, -20.422, -18.733, -17.481, -43.893, -44.070, -43.212, -41.654, -39.631,
-26.550, -23.010, -20.437, -18.742, -17.488, -43.919, -44.097, -43.238, -41.679,
-39.655, -26.570, -23.029, -20.453, -18.752, -17.496, -43.945, -44.123, -43.264,
-41.704, -39.679, -26.590, -23.049, -20.469, -18.761, -17.503, -43.971, -44.149,
-43.290, -41.729, -39.702, -26.611, -23.068, -20.485, -18.771, -17.510, -43.997,
-44.176, -43.316, -41.754, -39.726, -26.631, -23.088, -20.501, -18.781, -17.518,
 -44.023, -44.202, -43.342, -41.779, -39.749, -26.651, -23.107, -20.517,
-18.790, -17.525, -44.049, -44.229, -43.368, -41.804, -39.773, -26.671, -23.127,
-20.533, -18.800, -17.532, -44.075, -44.255, -43.393, -41.829, -39.796, -26.692,
-23.146, -20.549, -18.810, -17.540, -44.101, -44.281, -43.419, -41.854, -39.820,
-26.712, -23.166, -20.566, -18.819, -17.547, -44.127, -44.308, -43.445, -41.879,
-39.844, -26.732, -23.185, -20.582, -18.829, -17.555, -44.153, -44.334, -43.471,
-41.904, -39.867, -26.752, -23.205, -20.598, -18.839, -17.562, -44.179, -44.361,
-43.497, -41.929, -39.891, -26.772, -23.224, -20.615, -18.848, -17.569, -44.205,
-44.387, -43.523, -41.954, -39.914, -26.793, -23.243, -20.631, -18.858, -17.577,
 -44.232, -44.413, -43.549, -41.979, -39.938, -26.813, -23.263, -20.648,
-18.868, -17.584, -44.258, -44.440, -43.575, -42.004, -39.962, -26.833, -23.282,
-20.664, -18.878, -17.592, -44.284, -44.466, -43.601, -42.029, -39.985, -26.853,
-23.455, -20.681, -18.888, -17.599, -44.310, -44.492, -43.627, -42.054, -40.009,
-26.873, -23.474, -20.698, -18.897, -17.607, -44.336, -44.519, -43.653, -42.079,
-40.032, -26.893, -23.494, -20.714, -18.907, -17.614, -44.362, -44.545, -43.679,
-42.104, -40.056, -26.913, -23.513, -20.731, -18.917, -17.622, -44.388, -44.572,
-43.705, -42.129, -40.080, -26.934, -23.533, -20.748, -18.927, -17.629, -44.414,
-44.598, -43.731, -42.154, -40.103, -26.954, -23.552, -20.765, -18.937, -17.637,
 -44.440, -44.624, -43.757, -42.179, -40.127, -26.974, -23.572, -20.782,
-18.947, -17.644, -44.466, -44.651, -43.783, -42.204, -40.150, -26.994, -23.591,
-20.799, -18.956, -17.652, -44.492, -44.677, -43.809, -42.229, -40.174, -27.014,
-23.611, -20.816, -18.966, -17.659, -44.518, -44.704, -43.835, -42.254, -40.197,
-27.034, -23.630, -20.833, -18.976, -17.667, -44.544, -44.730, -43.861, -42.279,
-40.221, -27.054, -23.650, -20.850, -18.986, -17.674, -44.570, -44.756, -43.887,
-42.304, -40.245, -27.074, -23.669, -20.867, -18.996, -17.682, -44.596, -44.783,
-43.913, -42.329, -40.268, -27.094, -23.689, -20.884, -19.006, -17.689, -44.623,
-44.809, -43.939, -42.354, -40.292, -27.114, -23.708, -20.901, -19.016, -17.697,
 -44.649, -44.835, -43.965, -42.378, -40.315, -27.134, -23.727, -20.918,
-19.026, -17.704, -44.675, -44.862, -43.991, -42.403, -40.339, -27.154, -23.747,
-20.935, -19.036, -17.712, -44.701, -44.888, -44.017, -42.428, -40.362, -27.174,
-23.766, -20.953, -19.046, -17.720, -44.727, -44.915, -44.043, -42.453, -40.386,
-27.194, -23.785, -20.970, -19.057, -17.727, -44.753, -44.941, -44.069, -42.478,
-40.410, -27.214, -23.805, -20.987, -19.067, -17.735, -44.779, -44.967, -44.095,
-42.503, -40.433, -27.234, -23.824, -21.005, -19.077, -17.744, -44.805, -44.994,
-44.120, -42.528, -40.457, -27.254, -23.843, -21.022, -19.090, -17.752, -44.831,
-45.020, -44.146, -42.553, -40.480, -27.274, -23.863, -21.039, -19.100, -17.760,
 -44.857, -45.046, -44.172, -42.578, -40.504, -27.294, -23.882, -21.061,
-19.110, -17.767, -44.883, -45.073, -44.198, -42.603, -40.527, -27.313, -23.901,
-21.078, -19.120, -17.775, -44.909, -45.099, -44.224, -42.628, -40.551, -27.333,
-23.921, -21.095, -19.130, -17.783, -44.935, -45.126, -44.250, -42.653, -40.574,
-27.353, -23.940, -21.113, -19.141, -17.797, -44.961, -45.152, -44.276, -42.678,
-40.598, -38.169, -23.959, -21.130, -19.151, -17.804, -44.987, -45.178, -44.302,
-42.703, -40.622, -38.191, -23.978, -21.148, -19.161, -17.812, -45.013, -45.205,
-44.328, -42.728, -40.645, -38.213, -23.998, -21.165, -19.172, -17.820, -45.039,
-45.231, -44.354, -42.753, -40.669, -38.235, -24.017, -21.183, -19.182, -17.827,
 -45.065, -45.257, -44.380, -42.778, -40.692, -38.257, -24.036, -21.201,
-19.205, -17.835, -45.092, -45.284, -44.406, -42.803, -40.716, -38.279, -24.055,
-21.218, -19.215, -17.843, -45.118, -45.310, -44.432, -42.828, -40.739, -38.301,
-24.074, -21.236, -19.225, -17.850, -45.144, -45.337, -44.458, -42.853, -40.763,
-38.323, -24.094, -21.253, -19.236, -17.858, -45.170, -45.363, -44.484, -42.878,
-40.786, -38.345, -24.113, -21.271, -19.246, -17.866, -45.196, -45.389, -44.510,
-42.903, -40.810, -38.367, -24.132, -21.289, -19.257, -17.873, -45.222, -45.416,
-44.536, -42.927, -40.834, -38.388, -24.151, -21.306, -19.267, -17.881, -45.248,
-45.442, -44.562, -42.952, -40.857, -38.410, -24.170, -21.324, -19.278, -17.889,
 -45.274, -45.468, -44.587, -42.977, -40.881, -38.432, -24.189, -21.342,
-19.288, -17.896, -45.300, -45.495, -44.613, -43.002, -40.904, -38.454, -24.208,
-21.360, -19.299, -17.904, -45.326, -45.521, -44.639, -43.027, -40.928, -38.476,
-24.227, -21.377, -19.309, -17.912, -45.352, -45.547, -44.665, -43.052, -40.951,
-38.498, -24.246, -21.395, -19.320, -17.919, -45.378, -45.574, -44.691, -43.077,
-40.975, -38.520, -24.266, -21.413, -19.331, -17.927, -45.404, -45.600, -44.717,
-43.102, -40.998, -38.542, -24.285, -21.479, -19.341, -17.935, -45.430, -45.626,
-44.743, -43.127, -41.022, -38.564, -24.304, -21.496, -19.352, -17.943, -45.456,
-45.653, -44.769, -43.152, -41.045, -38.586, -24.323, -21.514, -19.363, -17.950,
 -45.482, -45.679, -44.795, -43.177, -41.069, -38.608, -24.342, -21.532,
-19.373, -17.958, -45.508, -45.706, -44.821, -43.202, -41.093, -38.629, -24.361,
-21.550, -19.384, -17.966, -45.534, -45.732, -44.847, -43.227, -41.116, -38.651,
-24.380, -21.568, -19.395, -17.974, -45.560, -45.758, -44.873, -43.252, -41.140,
-38.673, -24.399, -21.586, -19.406, -17.981, -45.586, -45.785, -44.899, -43.277,
-41.163, -38.695, -24.418, -21.604, -19.417, -17.989, -45.612, -45.811, -44.925,
-43.302, -41.187, -38.717, -24.437, -21.621, -19.428, -17.997, -45.639, -45.837,
-44.951, -43.327, -41.210, -38.739, -24.456, -21.639, -19.439, -18.005, -45.665,
-45.864, -44.977, -43.351, -41.234, -38.761, -24.474, -21.657, -19.450, -18.013,
 -45.691, -45.890, -45.002, -43.376, -41.257, -38.783, -24.493, -21.675,
-19.461, -18.021, -45.717, -45.916, -45.028, -43.401, -41.281, -38.805, -24.512,
-21.693, -19.472, -18.028, -45.743, -45.943, -45.054, -43.426, -41.304, -38.827,
-24.531, -21.711, -19.483, -18.036, -45.769, -45.969, -45.080, -43.451, -41.328,
-38.848, -24.550, -21.729, -19.494, -18.044, -45.795, -45.995, -45.106, -43.476,
-41.351, -38.870, -24.569, -21.747, -19.505, -18.052, -45.821, -46.022, -45.132,
-43.501, -41.375, -38.892, -24.588, -21.765, -19.516, -18.060, -45.847, -46.048,
-45.158, -43.526, -41.398, -38.914, -24.607, -21.783, -19.528, -18.068, -45.873,
-46.074, -45.184, -43.551, -41.422, -38.936, -24.626, -21.801, -19.539, -18.077,
 -45.899, -46.101, -45.210, -43.576, -41.446, -38.958, -24.644, -21.818,
-19.550, -18.085, -45.925, -46.127, -45.236, -43.601, -41.469, -38.980, -24.953,
-21.836, -19.562, -18.093, -45.951, -46.154, -45.262, -43.626, -41.493, -39.002,
-24.972, -21.854, -19.573, -18.101, -45.977, -46.180, -45.288, -43.651, -41.516,
-39.024, -24.991, -21.872, -19.585, -18.109, -46.003, -46.206, -45.314, -43.676,
-41.540, -39.045, -25.011, -21.890, -19.596, -18.117, -46.029, -46.233, -45.339,
-43.700, -41.563, -39.067, -25.030, -21.908, -19.608, -18.125, -46.055, -46.259,
-45.365, -43.725, -41.587, -39.089, -25.049, -21.926, -19.619, -18.132, -46.081,
-46.285, -45.391, -43.750, -41.610, -39.111, -25.068, -21.944, -19.631, -18.140,
 -46.107, -46.312, -45.417, -43.775, -41.634, -39.133, -25.087, -21.962,
-19.642, -18.148, -46.133, -46.338, -45.443, -43.800, -41.657, -39.155, -25.106,
-21.980, -19.654, -18.156, -46.159, -46.364, -45.469, -43.825, -41.681, -39.177,
-25.125, -21.998, -19.669, -18.164, -46.185, -46.391, -45.495, -43.850, -41.704,
-39.199, -25.144, -22.015, -19.681, -18.172, -46.211, -46.417, -45.521, -43.875,
-41.728, -39.221, -25.162, -22.038, -19.693, -18.180, -46.237, -46.443, -45.547,
-43.900, -41.751, -39.242, -25.181, -22.056, -19.704, -18.188, -46.263, -46.470,
-45.573, -43.925, -41.775, -39.264, -25.200, -22.074, -19.716, -18.196, -46.289,
-46.496, -45.599, -43.950, -41.798, -39.286, -25.219, -22.092, -19.728, -18.204,
 -46.315, -46.522, -45.625, -43.975, -41.822, -39.308, -25.238, -22.110,
-19.740, -18.215, -46.341, -46.549, -45.650, -43.999, -41.845, -39.330, -25.257,
-22.128, -19.753, -18.223, -46.367, -46.575, -45.676, -44.024, -41.869, -39.352,
-25.276, -22.145, -19.765, -18.231, -46.393, -46.601, -45.702, -44.049, -41.892,
-39.374, -25.295, -22.163, -19.777, -18.239, -46.419, -46.628, -45.728, -44.074,
-41.916, -39.396, -25.314, -22.181, -19.789, -18.247, -46.445, -46.654, -45.754,
-44.099, -41.939, -39.417, -25.333, -22.199, -19.801, -18.255, -46.471, -46.680,
-45.780, -44.124, -41.963, -39.439, -25.351, -22.217, -19.814, -18.263, -46.498,
-46.707, -45.806, -44.149, -41.986, -39.461, -25.370, -22.235, -19.826, -18.271,
 -46.524, -46.733, -45.832, -44.174, -42.010, -39.483, -25.389, -22.253,
-19.839, -18.279, -46.550, -46.759, -45.858, -44.199, -42.033, -39.505, -25.408,
-22.271, -19.851, -18.287, -46.576, -46.786, -45.884, -44.224, -42.057, -39.527,
-25.427, -22.288, -19.864, -18.302, -46.602, -46.812, -45.909, -44.248, -42.080,
-39.549, -25.446, -22.306, -19.876, -18.310, -46.628, -46.838, -45.935, -44.273,
-42.104, -39.570, -25.464, -22.324, -19.889, -18.318, -46.654, -46.865, -45.961,
-44.298, -42.127, -39.592, -25.483, -22.342, -19.902, -18.326, -46.680, -46.891,
-45.987, -44.323, -42.151, -39.614, -25.502, -22.360, -19.915, -18.334, -46.706,
-46.917, -46.013, -44.348, -42.174, -39.636, -25.521, -22.378, -19.946, -18.342,
 -46.732, -46.944, -46.039, -44.373, -42.198, -39.658, -25.539, -22.395,
-19.959, -18.350, -46.758, -46.970, -46.065, -44.398, -42.221, -39.680, -25.558,
-22.413, -19.971, -18.359, -46.784, -46.996, -46.091, -44.423, -42.245, -39.702,
-25.577, -22.431, -19.984, -18.367, -46.810, -47.023, -46.117, -44.448, -42.268,
-39.723, -25.596, -22.449, -19.998, -18.375, -46.836, -47.049, -46.143, -44.473,
-42.292, -39.745, -25.614, -22.467, -20.011, -18.383, -46.862, -47.075, -46.168,
-44.497, -42.315, -39.767, -25.633, -22.484, -20.024, -18.392, -46.888, -47.102,
-46.194, -44.522, -42.339, -39.789, -25.652, -22.502, -20.037, -18.400, -46.914,
-47.128, -46.220, -44.547, -42.362, -39.811, -25.670, -22.520, -20.050, -18.408,
 -46.940, -47.154, -46.246, -44.572, -42.386, -39.833, -25.689, -22.538,
-20.064, -18.416, -46.966, -47.181, -46.272, -44.597, -42.409, -39.855, -25.708,
-22.556, -20.077, -18.426, -46.992, -47.207, -46.298, -44.622, -42.433, -39.876,
-25.726, -22.573, -20.090, -18.434, -47.018, -47.233, -46.324, -44.647, -42.456,
-39.898, -25.745, -22.591, -20.104, -18.442, -47.044, -47.260, -46.350, -44.672,
-42.480, -39.920, -25.764, -22.699, -20.117, -18.450, -47.070, -47.286, -46.376,
-44.697, -42.503, -39.942, -25.782, -22.717, -20.131, -18.459, -47.096, -47.312,
-46.402, -44.722, -42.527, -39.964, -25.801, -22.735, -20.145, -18.467, -47.122,
-47.338, -46.427, -44.746, -42.550, -39.986, -25.819, -22.752, -20.158, -18.475,
 -47.148, -47.365, -46.453, -44.771, -42.574, -40.007, -25.838, -22.770,
-20.172, -18.484, -47.174, -47.391, -46.479, -44.796, -42.597, -40.029, -25.857,
-22.788, -20.186, -18.492, -47.200, -47.417, -46.505, -44.821, -42.621, -40.051,
-25.875, -22.806, -20.200, -18.500, -47.226, -47.444, -46.531, -44.846, -42.644,
-40.073, -25.894, -22.824, -20.214, -18.509, -47.252, -47.470, -46.557, -44.871,
-42.668, -40.095, -25.912, -22.841, -20.228, -18.517, -47.278, -47.496, -46.583,
-44.896, -42.691, -40.117, -25.931, -22.859, -20.242, -18.525, -47.304, -47.523,
-46.609, -44.921, -42.715, -40.138, -25.949, -22.877, -20.256, -18.534, -47.330,
-47.549, -46.635, -44.946, -42.738, -40.160, -25.968, -22.895, -20.270, -18.542,
 -47.356, -47.575, -46.660, -44.970, -42.762, -40.182, -25.987, -22.913,
-20.284, -18.551, -47.382, -47.602, -46.686, -44.995, -42.785, -40.204, -26.005,
-22.930, -20.298, -18.559, -47.408, -47.628, -46.712, -45.020, -42.809, -40.226,
-26.024, -22.948, -20.312, -18.568, -47.434, -47.654, -46.738, -45.045, -42.832,
-40.248, -26.042, -22.966, -20.327, -18.576, -47.460, -47.681, -46.764, -45.070,
-42.856, -40.269, -26.061, -22.984, -20.341, -18.585, -47.486, -47.707, -46.790,
-45.095, -42.879, -40.291, -26.079, -23.001, -20.355, -18.593, -47.512, -47.733,
-46.816, -45.120, -42.902, -40.313, -26.098, -23.019, -20.370, -18.602, -47.538,
-47.760, -46.842, -45.145, -42.926, -40.335, -26.720, -23.037, -20.387, -18.610,
 -47.564, -47.786, -46.868, -45.169, -42.949, -40.357, -26.739, -23.054,
-20.402, -18.619, -47.590, -47.812, -46.893, -45.194, -42.973, -40.379, -26.757,
-23.078, -20.416, -18.627, -47.616, -47.838, -46.919, -45.219, -42.996, -40.400,
-26.776, -23.095, -20.431, -18.636, -47.642, -47.865, -46.945, -45.244, -43.020,
-40.422, -26.795, -23.113, -20.446, -18.644, -47.668, -47.891, -46.971, -45.269,
-43.043, -40.444, -26.813, -23.131, -20.460, -18.655, -47.694, -47.917, -46.997,
-45.294, -43.067, -40.466, -26.832, -23.148, -20.475, -18.664, -47.720, -47.944,
-47.023, -45.319, -43.090, -40.488, -26.851, -23.166, -20.490, -18.673, -47.746,
-47.970, -47.049, -45.344, -43.114, -40.510, -26.869, -23.184, -20.505, -18.681,
 -47.772, -47.996, -47.075, -45.368, -43.137, -40.531, -26.888, -23.201,
-20.519, -18.690, -47.798, -48.023, -47.100, -45.393, -43.161, -40.553, -26.906,
-23.219, -20.534, -18.699, -47.824, -48.049, -47.126, -45.418, -43.184, -40.575,
-26.925, -23.237, -20.549, -18.707, -47.850, -48.075, -47.152, -45.443, -43.208,
-40.597, -26.944, -23.254, -20.564, -18.716, -47.876, -48.101, -47.178, -45.468,
-43.231, -40.619, -26.962, -23.272, -20.579, -18.725, -47.902, -48.128, -47.204,
-45.493, -43.254, -40.640, -26.981, -23.289, -20.594, -18.734, -47.928, -48.154,
-47.230, -45.518, -43.278, -40.662, -26.999, -23.307, -20.609, -18.742, -47.954,
-48.180, -47.256, -45.543, -43.301, -40.684, -27.018, -23.325, -20.624, -18.751,
 -47.980, -48.207, -47.282, -45.567, -43.325, -40.706, -27.037, -23.342,
-20.639, -18.760, -48.006, -48.233, -47.307, -45.592, -43.348, -40.728, -27.055,
-23.360, -20.654, -18.769, -48.032, -48.259, -47.333, -45.617, -43.372, -40.750,
-27.074, -23.377, -20.670, -18.778, -48.058, -48.286, -47.359, -45.642, -43.395,
-40.771, -27.092, -23.395, -20.685, -18.787, -48.084, -48.312, -47.385, -45.667,
-43.419, -40.793, -27.111, -23.412, -20.700, -18.796, -48.110, -48.338, -47.411,
-45.692, -43.442, -40.815, -27.129, -23.430, -20.715, -18.804, -48.136, -48.365,
-47.437, -45.717, -43.466, -40.837, -27.148, -23.447, -20.730, -18.813, -48.162,
-48.391, -47.463, -45.741, -43.489, -40.859, -27.166, -23.465, -20.746, -18.822,
 -48.188, -48.417, -47.488, -45.766, -43.512, -40.880, -27.185, -23.482,
-20.761, -18.831, -48.214, -48.443, -47.514, -45.791, -43.536, -40.902, -27.203,
-23.500, -20.776, -18.849, -48.240, -48.470, -47.540, -45.816, -43.559, -40.924,
-27.222, -23.517, -20.792, -18.858, -48.266, -48.496, -47.566, -45.841, -43.583,
-40.946, -27.240, -23.535, -20.807, -18.867, -48.292, -48.522, -47.592, -45.866,
-43.606, -40.968, -27.259, -23.552, -20.823, -18.876, -48.318, -48.549, -47.618,
-45.891, -43.630, -40.989, -27.277, -23.570, -20.838, -18.885, -48.344, -48.575,
-47.644, -45.915, -43.653, -41.011, -27.296, -23.587, -20.884, -18.894, -48.370,
-48.601, -47.670, -45.940, -43.677, -41.033, -27.314, -23.605, -20.900, -18.904,
 -48.396, -48.627, -47.695, -45.965, -43.700, -41.055, -27.332, -23.622,
-20.915, -18.913, -48.422, -48.654, -47.721, -45.990, -43.724, -41.077, -27.351,
-23.640, -20.931, -18.922, -48.448, -48.680, -47.747, -46.015, -43.747, -41.098,
-27.369, -23.657, -20.947, -18.931, -48.474, -48.706, -47.773, -46.040, -43.770,
-41.120, -27.388, -23.674, -20.962, -18.940, -48.500, -48.733, -47.799, -46.065,
-43.794, -41.142, -27.406, -23.692, -20.978, -18.949, -48.526, -48.759, -47.825,
-46.089, -43.817, -41.164, -27.424, -23.709, -20.993, -18.959, -48.552, -48.785,
-47.851, -46.114, -43.841, -41.186, -27.443, -23.727, -21.009, -18.968, -48.578,
-48.812, -47.876, -46.139, -43.864, -41.207, -27.461, -23.744, -21.025, -18.977,
 -48.604, -48.838, -47.902, -46.164, -43.888, -41.229, -27.479, -23.761,
-21.040, -18.987, -48.630, -48.864, -47.928, -46.189, -43.911, -41.251, -27.498,
-23.779, -21.056, -18.996, -48.656, -48.890, -47.954, -46.214, -43.935, -41.273,
-27.516, -23.796, -21.072, -19.005, -48.682, -48.917, -47.980, -46.239, -43.958,
-41.295, -27.534, -23.977, -21.087, -19.015, -48.708, -48.943, -48.006, -46.263,
-43.981, -41.316, -27.553, -23.994, -21.103, -19.024, -48.734, -48.969, -48.032,
-46.288, -44.005, -41.338, -27.571, -24.012, -21.119, -19.034, -48.760, -48.996,
-48.057, -46.313, -44.028, -41.360, -27.589, -24.029, -21.135, -19.043, -48.785,
-49.022, -48.083, -46.338, -44.052, -41.382, -27.608, -24.046, -21.150, -19.053,
 -48.811, -49.048, -48.109, -46.363, -44.075, -41.404, -27.626, -24.064,
-21.166, -19.062, -48.837, -49.074, -48.135, -46.388, -44.099, -41.425, -27.644,
-24.081, -21.182, -19.072, -48.863, -49.101, -48.161, -46.413, -44.122, -41.447,
-27.663, -24.099, -21.198, -19.081, -48.889, -49.127, -48.187, -46.437, -44.145,
-41.469, -27.681, -24.116, -21.218, -19.091, -48.915, -49.153, -48.213, -46.462,
-44.169, -41.491, -27.699, -24.134, -21.233, -19.101, -48.941, -49.180, -48.238,
-46.487, -44.192, -41.512, -27.717, -24.151, -21.249, -19.110, -48.967, -49.206,
-48.264, -46.512, -44.216, -41.534, -27.736, -24.168, -21.265, -19.120, -48.993,
-49.232, -48.290, -46.537, -44.239, -41.556, -27.754, -24.186, -21.281, -19.130,
 -49.019, -49.258, -48.316, -46.562, -44.263, -41.578, -27.772, -24.203,
-21.297, -19.142, -49.045, -49.285, -48.342, -46.586, -44.286, -41.600, -27.790,
-24.221, -21.313, -19.152, -49.071, -49.311, -48.368, -46.611, -44.310, -41.621,
-27.809, -24.238, -21.329, -19.162, -49.097, -49.337, -48.393, -46.636, -44.333,
-41.643, -27.827, -24.255, -21.345, -19.172, -49.123, -49.363, -48.419, -46.661,
-44.356, -41.665, -27.845, -24.273, -21.360, -19.182, -49.149, -49.390, -48.445,
-46.686, -44.380, -41.687, -27.863, -24.290, -21.376, -19.192, -49.175, -49.416,
-48.471, -46.711, -44.403, -41.708, -38.832, -24.307, -21.392, -19.202, -49.201,
-49.442, -48.497, -46.735, -44.427, -41.730, -38.852, -24.325, -21.408, -19.212,
 -49.227, -49.469, -48.523, -46.760, -44.450, -41.752, -38.872, -24.342,
-21.424, -19.222, -49.253, -49.495, -48.549, -46.785, -44.473, -41.774, -38.892,
-24.359, -21.440, -19.232, -49.279, -49.521, -48.574, -46.810, -44.497, -41.796,
-38.913, -24.377, -21.456, -19.242, -49.305, -49.547, -48.600, -46.835, -44.520,
-41.817, -38.933, -24.394, -21.472, -19.252, -49.331, -49.574, -48.626, -46.860,
-44.544, -41.839, -38.953, -24.411, -21.488, -19.263, -49.357, -49.600, -48.652,
-46.884, -44.567, -41.861, -38.973, -24.428, -21.504, -19.273, -49.383, -49.626,
-48.678, -46.909, -44.591, -41.883, -38.993, -24.446, -21.520, -19.283, -49.409,
-49.653, -48.704, -46.934, -44.614, -41.904, -39.013, -24.463, -21.536, -19.293,
 -49.435, -49.679, -48.729, -46.959, -44.637, -41.926, -39.033, -24.480,
-21.552, -19.304, -49.461, -49.705, -48.755, -46.984, -44.661, -41.948, -39.053,
-24.497, -21.568, -19.314, -49.487, -49.731, -48.781, -47.009, -44.684, -41.970,
-39.073, -24.515, -21.584, -19.325, -49.513, -49.758, -48.807, -47.033, -44.708,
-41.991, -39.093, -24.532, -21.600, -19.335, -49.539, -49.784, -48.833, -47.058,
-44.731, -42.013, -39.113, -24.549, -21.615, -19.346, -49.565, -49.810, -48.859,
-47.083, -44.755, -42.035, -39.133, -24.566, -21.631, -19.356, -49.591, -49.836,
-48.884, -47.108, -44.778, -42.057, -39.153, -24.583, -21.647, -19.367, -49.617,
-49.863, -48.910, -47.133, -44.801, -42.079, -39.173, -24.601, -21.663, -19.377,
 -49.643, -49.889, -48.936, -47.158, -44.825, -42.100, -39.193, -24.618,
-21.679, -19.388, -49.669, -49.915, -48.962, -47.182, -44.848, -42.122, -39.213,
-24.635, -21.695, -19.399, -49.694, -49.942, -48.988, -47.207, -44.872, -42.144,
-39.233, -24.652, -21.711, -19.410, -49.720, -49.968, -49.014, -47.232, -44.895,
-42.166, -39.253, -24.669, -21.727, -19.420, -49.746, -49.994, -49.040, -47.257,
-44.918, -42.187, -39.273, -24.686, -21.743, -19.431, -49.772, -50.020, -49.065,
-47.282, -44.942, -42.209, -39.293, -24.704, -21.759, -19.442, -49.798, -50.047,
-49.091, -47.307, -44.965, -42.231, -39.313, -24.721, -21.775, -19.453, -49.824,
-50.073, -49.117, -47.331, -44.989, -42.253, -39.333, -24.738, -21.791, -19.464,
 -49.850, -50.099, -49.143, -47.356, -45.012, -42.274, -39.353, -24.755,
-21.807, -19.488, -49.876, -50.125, -49.169, -47.381, -45.035, -42.296, -39.373,
-24.772, -21.823, -19.499, -49.902, -50.152, -49.195, -47.406, -45.059, -42.318,
-39.393, -24.789, -21.839, -19.510, -49.928, -50.178, -49.220, -47.431, -45.082,
-42.340, -39.413, -24.806, -21.855, -19.521, -49.954, -50.204, -49.246, -47.455,
-45.106, -42.361, -39.433, -24.823, -21.871, -19.533, -49.980, -50.230, -49.272,
-47.480, -45.129, -42.383, -39.453, -24.840, -21.943, -19.544, -50.006, -50.257,
-49.298, -47.505, -45.153, -42.405, -39.473, -24.857, -21.959, -19.555, -50.032,
-50.283, -49.324, -47.530, -45.176, -42.427, -39.493, -24.874, -21.975, -19.566,
 -50.058, -50.309, -49.349, -47.555, -45.199, -42.448, -39.513, -24.892,
-21.991, -19.578, -50.084, -50.336, -49.375, -47.580, -45.223, -42.470, -39.533,
-24.909, -22.007, -19.589, -50.110, -50.362, -49.401, -47.604, -45.246, -42.492,
-39.553, -24.926, -22.023, -19.600, -50.136, -50.388, -49.427, -47.629, -45.270,
-42.514, -39.573, -24.943, -22.039, -19.612, -50.162, -50.414, -49.453, -47.654,
-45.293, -42.535, -39.593, -24.960, -22.055, -19.624, -50.188, -50.441, -49.479,
-47.679, -45.316, -42.557, -39.613, -24.977, -22.071, -19.635, -50.214, -50.467,
-49.504, -47.704, -45.340, -42.579, -39.633, -24.994, -22.087, -19.647, -50.240,
-50.493, -49.530, -47.728, -45.363, -42.601, -39.653, -25.011, -22.108, -19.658,
 -50.266, -50.519, -49.556, -47.753, -45.387, -42.622, -39.673, -25.028,
-22.124, -19.670, -50.292, -50.546, -49.582, -47.778, -45.410, -42.644, -39.693,
-25.045, -22.140, -19.682, -50.317, -50.572, -49.608, -47.803, -45.433, -42.666,
-39.713, -25.357, -22.156, -19.694, -50.343, -50.598, -49.634, -47.828, -45.457,
-42.688, -39.733, -25.374, -22.172, -19.705, -50.369, -50.624, -49.659, -47.853,
-45.480, -42.709, -39.753, -25.391, -22.188, -19.717, -50.395, -50.651, -49.685,
-47.877, -45.504, -42.731, -39.773, -25.409, -22.204, -19.732, -50.421, -50.677,
-49.711, -47.902, -45.527, -42.753, -39.793, -25.426, -22.220, -19.744, -50.447,
-50.703, -49.737, -47.927, -45.550, -42.775, -39.812, -25.443, -22.236, -19.756,
 -50.473, -50.729, -49.763, -47.952, -45.574, -42.796, -39.832, -25.460,
-22.252, -19.768, -50.499, -50.756, -49.788, -47.977, -45.597, -42.818, -39.852,
-25.477, -22.268, -19.780, -50.525, -50.782, -49.814, -48.001, -45.621, -42.840,
-39.872, -25.494, -22.284, -19.792, -50.551, -50.808, -49.840, -48.026, -45.644,
-42.862, -39.892, -25.511, -22.300, -19.804, -50.577, -50.834, -49.866, -48.051,
-45.667, -42.883, -39.912, -25.528, -22.316, -19.817, -50.603, -50.861, -49.892,
-48.076, -45.691, -42.905, -39.932, -25.545, -22.332, -19.829, -50.629, -50.887,
-49.918, -48.101, -45.714, -42.927, -39.952, -25.562, -22.348, -19.841, -50.655,
-50.913, -49.943, -48.125, -45.737, -42.948, -39.972, -25.579, -22.364, -19.854,
 -50.681, -50.939, -49.969, -48.150, -45.761, -42.970, -39.992, -25.596,
-22.380, -19.866, -50.707, -50.966, -49.995, -48.175, -45.784, -42.992, -40.012,
-25.613, -22.396, -19.878, -50.733, -50.992, -50.021, -48.200, -45.808, -43.014,
-40.032, -25.631, -22.412, -19.891, -50.759, -51.018, -50.047, -48.225, -45.831,
-43.035, -40.052, -25.648, -22.428, -19.903, -50.785, -51.044, -50.072, -48.249,
-45.854, -43.057, -40.072, -25.665, -22.443, -19.916, -50.811, -51.071, -50.098,
-48.274, -45.878, -43.079, -40.092, -25.682, -22.459, -19.928, -50.836, -51.097,
-50.124, -48.299, -45.901, -43.101, -40.112, -25.699, -22.475, -19.941, -50.862,
-51.123, -50.150, -48.324, -45.925, -43.122, -40.132, -25.716, -22.491, -19.953,
 -50.888, -51.149, -50.176, -48.349, -45.948, -43.144, -40.152, -25.733,
-22.507, -19.966, -50.914, -51.176, -50.202, -48.373, -45.971, -43.166, -40.172,
-25.749, -22.523, -19.979, -50.940, -51.202, -50.227, -48.398, -45.995, -43.187,
-40.192, -25.766, -22.539, -19.992, -50.966, -51.228, -50.253, -48.423, -46.018,
-43.209, -40.212, -25.783, -22.555, -20.004, -50.992, -51.254, -50.279, -48.448,
-46.042, -43.231, -40.232, -25.800, -22.571, -20.017, -51.018, -51.281, -50.305,
-48.473, -46.065, -43.253, -40.252, -25.817, -22.587, -20.030, -51.044, -51.307,
-50.331, -48.497, -46.088, -43.274, -40.272, -25.834, -22.603, -20.043, -51.070,
-51.333, -50.356, -48.522, -46.112, -43.296, -40.292, -25.851, -22.619, -20.056,
 -51.096, -51.359, -50.382, -48.547, -46.135, -43.318, -40.312, -25.868,
-22.634, -20.069, -51.122, -51.386, -50.408, -48.572, -46.158, -43.340, -40.332,
-25.885, -22.650, -20.082, -51.148, -51.412, -50.434, -48.597, -46.182, -43.361,
-40.352, -25.902, -22.666, -20.095, -51.174, -51.438, -50.460, -48.621, -46.205,
-43.383, -40.372, -25.919, -22.682, -20.108, -51.200, -51.464, -50.485, -48.646,
-46.229, -43.405, -40.392, -25.936, -22.698, -20.121, -51.226, -51.491, -50.511,
-48.671, -46.252, -43.426, -40.412, -25.953, -22.714, -20.134, -51.252, -51.517,
-50.537, -48.696, -46.275, -43.448, -40.432, -25.969, -22.730, -20.147, -51.278,
-51.543, -50.563, -48.721, -46.299, -43.470, -40.451, -25.986, -22.745, -20.160,
 -51.303, -51.569, -50.589, -48.745, -46.322, -43.492, -40.471, -26.003,
-22.761, -20.173, -51.329, -51.596, -50.614, -48.770, -46.345, -43.513, -40.491,
-26.020, -22.777, -20.186, -51.355, -51.622, -50.640, -48.795, -46.369, -43.535,
-40.511, -26.037, -22.793, -20.200, -51.381, -51.648, -50.666, -48.820, -46.392,
-43.557, -40.531, -26.054, -22.809, -20.213, -51.407, -51.674, -50.692, -48.845,
-46.416, -43.578, -40.551, -26.071, -22.825, -20.226, -51.433, -51.701, -50.718,
-48.869, -46.439, -43.600, -40.571, -26.087, -22.840, -20.240, -51.459, -51.727,
-50.743, -48.894, -46.462, -43.622, -40.591, -26.104, -22.856, -20.253, -51.485,
-51.753, -50.769, -48.919, -46.486, -43.644, -40.611, -26.121, -22.872, -20.287,
 -51.511, -51.779, -50.795, -48.944, -46.509, -43.665, -40.631, -26.138,
-22.888, -20.300, -51.537, -51.806, -50.821, -48.969, -46.532, -43.687, -40.651,
-26.155, -22.904, -20.314, -51.563, -51.832, -50.847, -48.993, -46.556, -43.709,
-40.671, -26.171, -22.919, -20.327, -51.589, -51.858, -50.873, -49.018, -46.579,
-43.730, -40.691, -26.188, -22.935, -20.341, -51.615, -51.884, -50.898, -49.043,
-46.603, -43.752, -40.711, -26.205, -23.054, -20.354, -51.641, -51.910, -50.924,
-49.068, -46.626, -43.774, -40.731, -26.222, -23.070, -20.368, -51.667, -51.937,
-50.950, -49.092, -46.649, -43.796, -40.751, -26.239, -23.086, -20.381, -51.692,
-51.963, -50.976, -49.117, -46.673, -43.817, -40.771, -26.255, -23.102, -20.395,
 -51.718, -51.989, -51.001, -49.142, -46.696, -43.839, -40.791, -26.272,
-23.118, -20.408, -51.744, -52.015, -51.027, -49.167, -46.719, -43.861, -40.811,
-26.289, -23.134, -20.426, -51.770, -52.042, -51.053, -49.192, -46.743, -43.882,
-40.830, -26.306, -23.150, -20.439, -51.796, -52.068, -51.079, -49.216, -46.766,
-43.904, -40.850, -26.322, -23.165, -20.453, -51.822, -52.094, -51.105, -49.241,
-46.789, -43.926, -40.870, -26.339, -23.181, -20.467, -51.848, -52.120, -51.130,
-49.266, -46.813, -43.948, -40.890, -26.356, -23.197, -20.480, -51.874, -52.147,
-51.156, -49.291, -46.836, -43.969, -40.910, -26.372, -23.213, -20.494, -51.900,
-52.173, -51.182, -49.316, -46.860, -43.991, -40.930, -26.389, -23.229, -20.508,
 -51.926, -52.199, -51.208, -49.340, -46.883, -44.013, -40.950, -27.009,
-23.244, -20.521, -51.952, -52.225, -51.234, -49.365, -46.906, -44.034, -40.970,
-27.025, -23.260, -20.535, -51.978, -52.252, -51.259, -49.390, -46.930, -44.056,
-40.990, -27.042, -23.276, -20.549, -52.004, -52.278, -51.285, -49.415, -46.953,
-44.078, -41.010, -27.059, -23.292, -20.563, -52.030, -52.304, -51.311, -49.439,
-46.976, -44.099, -41.030, -27.076, -23.308, -20.576, -52.056, -52.330, -51.337,
-49.464, -47.000, -44.121, -41.050, -27.093, -23.323, -20.590, -52.081, -52.356,
-51.363, -49.489, -47.023, -44.143, -41.070, -27.110, -23.339, -20.604, -52.107,
-52.383, -51.388, -49.514, -47.046, -44.165, -41.090, -27.127, -23.355, -20.618,
 -52.133, -52.409, -51.414, -49.539, -47.070, -44.186, -41.110, -27.143,
-23.371, -20.632, -52.159, -52.435, -51.440, -49.563, -47.093, -44.208, -41.130,
-27.160, -23.386, -20.646, -52.185, -52.461, -51.466, -49.588, -47.116, -44.230,
-41.149, -27.177, -23.402, -20.660, -52.211, -52.488, -51.492, -49.613, -47.140,
-44.251, -41.169, -27.194, -23.418, -20.673, -52.237, -52.514, -51.517, -49.638,
-47.163, -44.273, -41.189, -27.211, -23.433, -20.687, -52.263, -52.540, -51.543,
-49.662, -47.187, -44.295, -41.209, -27.228, -23.449, -20.701, -52.289, -52.566,
-51.569, -49.687, -47.210, -44.316, -41.229, -27.244, -23.465, -20.715, -52.315,
-52.593, -51.595, -49.712, -47.233, -44.338, -41.249, -27.261, -23.481, -20.729,
 -52.341, -52.619, -51.621, -49.737, -47.257, -44.360, -41.269, -27.278,
-23.496, -20.743, -52.367, -52.645, -51.646, -49.761, -47.280, -44.381, -41.289,
-27.295, -23.512, -20.757, -52.393, -52.671, -51.672, -49.786, -47.303, -44.403,
-41.309, -27.311, -23.528, -20.771, -52.418, -52.697, -51.698, -49.811, -47.327,
-44.425, -41.329, -27.328, -23.543, -20.785, -52.444, -52.724, -51.724, -49.836,
-47.350, -44.447, -41.349, -27.345, -23.559, -20.799, -52.470, -52.750, -51.749,
-49.861, -47.373, -44.468, -41.369, -27.362, -23.575, -20.813, -52.496, -52.776,
-51.775, -49.885, -47.397, -44.490, -41.389, -27.378, -23.590, -20.827, -52.522,
-52.802, -51.801, -49.910, -47.420, -44.512, -41.408, -27.395, -23.606, -20.841,
 -52.548, -52.829, -51.827, -49.935, -47.443, -44.533, -41.428, -27.412,
-23.622, -20.855, -52.574, -52.855, -51.853, -49.960, -47.467, -44.555, -41.448,
-27.429, -23.637, -20.869, -52.600, -52.881, -51.878, -49.984, -47.490, -44.577,
-41.468, -27.445, -23.653, -20.883, -52.626, -52.907, -51.904, -50.009, -47.513,
-44.598, -41.488, -27.462, -23.668, -20.897, -52.652, -52.933, -51.930, -50.034,
-47.537, -44.620, -41.508, -27.479, -23.684, -20.911, -52.678, -52.960, -51.956,
-50.059, -47.560, -44.642, -41.528, -27.495, -23.700, -20.926, -52.704, -52.986,
-51.981, -50.083, -47.584, -44.663, -41.548, -27.512, -23.715, -20.940, -52.730,
-53.012, -52.007, -50.108, -47.607, -44.685, -41.568, -27.529, -23.731, -20.954,
 -52.755, -53.038, -52.033, -50.133, -47.630, -44.707, -41.588, -27.546,
-23.747, -20.968, -52.781, -53.065, -52.059, -50.158, -47.654, -44.728, -41.608,
-27.562, -23.762, -20.982, -52.807, -53.091, -52.085, -50.183, -47.677, -44.750,
-41.628, -27.579, -23.778, -20.996, -52.833, -53.117, -52.110, -50.207, -47.700,
-44.772, -41.647, -27.596, -23.793, -21.010, -52.859, -53.143, -52.136, -50.232,
-47.724, -44.793, -41.667, -27.612, -23.809, -21.024, -52.885, -53.169, -52.162,
-50.257, -47.747, -44.815, -41.687, -27.629, -23.824, -21.038, -52.911, -53.196,
-52.188, -50.282, -47.770, -44.837, -41.707, -27.645, -23.840, -21.053, -52.937,
-53.222, -52.214, -50.306, -47.794, -44.859, -41.727, -27.662, -23.855, -21.067,
 -52.963, -53.248, -52.239, -50.331, -47.817, -44.880, -41.747, -27.679,
-23.871, -21.081, -52.989, -53.274, -52.265, -50.356, -47.840, -44.902, -41.767,
-27.695, -23.887, -21.095, -53.015, -53.300, -52.291, -50.381, -47.864, -44.924,
-41.787, -27.712, -23.902, -21.109, -53.041, -53.327, -52.317, -50.405, -47.887,
-44.945, -41.807, -27.729, -23.918, -21.123, -53.066, -53.353, -52.342, -50.430,
-47.910, -44.967, -41.827, -27.745, -23.933, -21.138, -53.092, -53.379, -52.368,
-50.455, -47.934, -44.989, -41.847, -27.762, -23.949, -21.152, -53.118, -53.405,
-52.394, -50.480, -47.957, -45.010, -41.866, -27.778, -23.964, -21.206, -53.144,
-53.432, -52.420, -50.504, -47.980, -45.032, -41.886, -27.795, -23.980, -21.220,
 -53.170, -53.458, -52.445, -50.529, -48.004, -45.054, -41.906, -27.812,
-23.995, -21.234, -53.196, -53.484, -52.471, -50.554, -48.027, -45.075, -41.926,
-27.828, -24.011, -21.249, -53.222, -53.510, -52.497, -50.579, -48.050, -45.097,
-41.946, -27.845, -24.026, -21.263, -53.248, -53.536, -52.523, -50.603, -48.074,
-45.119, -41.966, -27.861, -24.210, -21.277, -53.274, -53.563, -52.549, -50.628,
-48.097, -45.140, -41.986, -27.878, -24.226, -21.291, -53.300, -53.589, -52.574,
-50.653, -48.120, -45.162, -42.006, -27.894, -24.241, -21.306, -53.326, -53.615,
-52.600, -50.678, -48.144, -45.184, -42.026, -27.911, -24.257, -21.320, -53.351,
-53.641, -52.626, -50.702, -48.167, -45.205, -42.046, -27.927, -24.272, -21.334,
 -53.377, -53.667, -52.652, -50.727, -48.190, -45.227, -42.065, -27.944,
-24.288, -21.348, -53.403, -53.694, -52.677, -50.752, -48.214, -45.249, -42.085,
-27.960, -24.304, -21.363, -53.429, -53.720, -52.703, -50.777, -48.237, -45.270,
-42.105, -27.977, -24.319, -21.377, -53.455, -53.746, -52.729, -50.801, -48.260,
-45.292, -42.125, -27.993, -24.335, -21.391, -53.481, -53.772, -52.755, -50.826,
-48.284, -45.314, -42.145, -28.010, -24.350, -21.405, -53.507, -53.799, -52.781,
-50.851, -48.307, -45.335, -42.165, -28.026, -24.366, -21.420, -53.533, -53.825,
-52.806, -50.876, -48.330, -45.357, -42.185, -28.043, -24.381, -21.434, -53.559,
-53.851, -52.832, -50.900, -48.354, -45.379, -42.205, -38.981, -24.397, -21.448,
 -53.585, -53.877, -52.858, -50.925, -48.377, -45.400, -42.225, -38.999,
-24.412, -21.462, -53.611, -53.903, -52.884, -50.950, -48.400, -45.422, -42.245,
-39.017, -24.428, -21.477, -53.636, -53.930, -52.909, -50.975, -48.424, -45.444,
-42.264, -39.035, -24.443, -21.491, -53.662, -53.956, -52.935, -50.999, -48.447,
-45.465, -42.284, -39.053, -24.459, -21.505, -53.688, -53.982, -52.961, -51.024,
-48.470, -45.487, -42.304, -39.071, -24.474, -21.519, -53.714, -54.008, -52.987,
-51.049, -48.494, -45.509, -42.324, -39.089, -24.490, -21.534, -53.740, -54.034,
-53.012, -51.074, -48.517, -45.530, -42.344, -39.107, -24.505, -21.548, -53.766,
-54.061, -53.038, -51.098, -48.540, -45.552, -42.364, -39.126, -24.521, -21.562,
 -53.792, -54.087, -53.064, -51.123, -48.564, -45.574, -42.384, -39.144,
-24.536, -21.576, -53.818, -54.113, -53.090, -51.148, -48.587, -45.595, -42.404,
-39.162, -24.552, -21.591, -53.844, -54.139, -53.115, -51.173, -48.610, -45.617,
-42.424, -39.180, -24.567, -21.605, -53.870, -54.165, -53.141, -51.197, -48.634,
-45.639, -42.443, -39.198, -24.582, -21.619, -53.896, -54.192, -53.167, -51.222,
-48.657, -45.660, -42.463, -39.216, -24.598, -21.633, -53.921, -54.218, -53.193,
-51.247, -48.680, -45.682, -42.483, -39.234, -24.613, -21.648, -53.947, -54.244,
-53.219, -51.272, -48.704, -45.704, -42.503, -39.252, -24.629, -21.662, -53.973,
-54.270, -53.244, -51.296, -48.727, -45.725, -42.523, -39.270, -24.644, -21.676,
 -53.999, -54.296, -53.270, -51.321, -48.750, -45.747, -42.543, -39.289,
-24.660, -21.690, -54.025, -54.323, -53.296, -51.346, -48.773, -45.769, -42.563,
-39.307, -24.675, -21.705, -54.051, -54.349, -53.322, -51.371, -48.797, -45.790,
-42.583, -39.325, -24.690, -21.719, -54.077, -54.375, -53.347, -51.395, -48.820,
-45.812, -42.603, -39.343, -24.706, -21.733, -54.103, -54.401, -53.373, -51.420,
-48.843, -45.833, -42.622, -39.361, -24.721, -21.747, -54.129, -54.427, -53.399,
-51.445, -48.867, -45.855, -42.642, -39.379, -24.737, -21.762, -54.155, -54.454,
-53.425, -51.470, -48.890, -45.877, -42.662, -39.397, -24.752, -21.776, -54.180,
-54.480, -53.450, -51.494, -48.913, -45.898, -42.682, -39.415, -24.767, -21.790,
 -54.206, -54.506, -53.476, -51.519, -48.937, -45.920, -42.702, -39.433,
-24.783, -21.804, -54.232, -54.532, -53.502, -51.544, -48.960, -45.942, -42.722,
-39.451, -24.798, -21.819, -54.258, -54.558, -53.528, -51.569, -48.983, -45.963,
-42.742, -39.470, -24.813, -21.833, -54.284, -54.585, -53.553, -51.593, -49.007,
-45.985, -42.762, -39.488, -24.829, -21.847, -54.310, -54.611, -53.579, -51.618,
-49.030, -46.007, -42.781, -39.506, -24.844, -21.861, -54.336, -54.637, -53.605,
-51.643, -49.053, -46.028, -42.801, -39.524, -24.859, -21.875, -54.362, -54.663,
-53.631, -51.668, -49.077, -46.050, -42.821, -39.542, -24.875, -21.890, -54.388,
-54.689, -53.656, -51.692, -49.100, -46.072, -42.841, -39.560, -24.890, -21.904,
 -54.414, -54.716, -53.682, -51.717, -49.123, -46.093, -42.861, -39.578,
-24.905, -21.918, -54.439, -54.742, -53.708, -51.742, -49.147, -46.115, -42.881,
-39.596, -24.921, -21.932, -54.465, -54.768, -53.734, -51.767, -49.170, -46.137,
-42.901, -39.614, -24.936, -21.952, -54.491, -54.794, -53.759, -51.791, -49.193,
-46.158, -42.921, -39.632, -24.951, -21.966, -54.517, -54.820, -53.785, -51.816,
-49.216, -46.180, -42.940, -39.651, -24.967, -21.980, -54.543, -54.847, -53.811,
-51.841, -49.240, -46.202, -42.960, -39.669, -24.982, -21.994, -54.569, -54.873,
-53.837, -51.865, -49.263, -46.223, -42.980, -39.687, -24.997, -22.009, -54.595,
-54.899, -53.862, -51.890, -49.286, -46.245, -43.000, -39.705, -25.012, -22.023,
 -54.621, -54.925, -53.888, -51.915, -49.310, -46.266, -43.020, -39.723,
-25.028, -22.037, -54.647, -54.951, -53.914, -51.940, -49.333, -46.288, -43.040,
-39.741, -25.043, -22.051, -54.673, -54.978, -53.940, -51.964, -49.356, -46.310,
-43.060, -39.759, -25.058, -22.065, -54.698, -55.004, -53.965, -51.989, -49.380,
-46.331, -43.079, -39.777, -25.073, -22.079, -54.724, -55.030, -53.991, -52.014,
-49.403, -46.353, -43.099, -39.795, -25.089, -22.094, -54.750, -55.056, -54.017,
-52.039, -49.426, -46.375, -43.119, -39.813, -25.104, -22.169, -54.776, -55.082,
-54.043, -52.063, -49.450, -46.396, -43.139, -39.831, -25.119, -22.183, -54.802,
-55.109, -54.068, -52.088, -49.473, -46.418, -43.159, -39.850, -25.134, -22.198,
 -54.828, -55.135, -54.094, -52.113, -49.496, -46.440, -43.179, -39.868,
-25.150, -22.212, -54.854, -55.161, -54.120, -52.138, -49.520, -46.461, -43.199,
-39.886, -25.165, -22.226, -54.880, -55.187, -54.146, -52.162, -49.543, -46.483,
-43.219, -39.904, -25.477, -22.240, -54.906, -55.213, -54.171, -52.187, -49.566,
-46.504, -43.238, -39.922, -25.492, -22.255, -54.931, -55.240, -54.197, -52.212,
-49.589, -46.526, -43.258, -39.940, -25.508, -22.269, -54.957, -55.266, -54.223,
-52.236, -49.613, -46.548, -43.278, -39.958, -25.523, -22.283, -54.983, -55.292,
-54.249, -52.261, -49.636, -46.569, -43.298, -39.976, -25.538, -22.297, -55.009,
-55.318, -54.274, -52.286, -49.659, -46.591, -43.318, -39.994, -25.553, -22.311,
 -55.035, -55.344, -54.300, -52.311, -49.683, -46.613, -43.338, -40.012,
-25.569, -22.326, -55.061, -55.370, -54.326, -52.335, -49.706, -46.634, -43.358,
-40.030, -25.584, -22.340, -55.087, -55.397, -54.352, -52.360, -49.729, -46.656,
-43.377, -40.048, -25.599, -22.354, -55.113, -55.423, -54.377, -52.385, -49.753,
-46.678, -43.397, -40.067, -25.615, -22.368, -55.139, -55.449, -54.403, -52.410,
-49.776, -46.699, -43.417, -40.085, -25.630, -22.382, -55.164, -55.475, -54.429,
-52.434, -49.799, -46.721, -43.437, -40.103, -25.645, -22.397, -55.190, -55.501,
-54.455, -52.459, -49.822, -46.742, -43.457, -40.121, -25.661, -22.411, -55.216,
-55.528, -54.480, -52.484, -49.846, -46.764, -43.477, -40.139, -25.676, -22.425,
 -55.242, -55.554, -54.506, -52.508, -49.869, -46.786, -43.497, -40.157,
-25.691, -22.439, -55.268, -55.580, -54.532, -52.533, -49.892, -46.807, -43.516,
-40.175, -25.706, -22.453, -55.294, -55.606, -54.558, -52.558, -49.916, -46.829,
-43.536, -40.193, -25.722, -22.467, -55.320, -55.632, -54.583, -52.583, -49.939,
-46.851, -43.556, -40.211, -25.737, -22.482, -55.346, -55.659, -54.609, -52.607,
-49.962, -46.872, -43.576, -40.229, -25.752, -22.496, -55.372, -55.685, -54.635,
-52.632, -49.986, -46.894, -43.596, -40.247, -25.767, -22.510, -55.397, -55.711,
-54.661, -52.657, -50.009, -46.915, -43.616, -40.265, -25.783, -22.524, -55.423,
-55.737, -54.686, -52.681, -50.032, -46.937, -43.636, -40.283, -25.798, -22.538,
 -55.449, -55.763, -54.712, -52.706, -50.055, -46.959, -43.655, -40.302,
-25.813, -22.552, -55.475, -55.789, -54.738, -52.731, -50.079, -46.980, -43.675,
-40.320, -25.828, -22.567, -55.501, -55.816, -54.764, -52.756, -50.102, -47.002,
-43.695, -40.338, -25.844, -22.581, -55.527, -55.842, -54.789, -52.780, -50.125,
-47.024, -43.715, -40.356, -25.859, -22.595, -55.553, -55.868, -54.815, -52.805,
-50.149, -47.045, -43.735, -40.374, -25.874, -22.609, -55.579, -55.894, -54.841,
-52.830, -50.172, -47.067, -43.755, -40.392, -25.889, -22.623, -55.605, -55.920,
-54.867, -52.854, -50.195, -47.089, -43.774, -40.410, -25.904, -22.637, -55.630,
-55.947, -54.892, -52.879, -50.218, -47.110, -43.794, -40.428, -25.919, -22.651,
 -55.656, -55.973, -54.918, -52.904, -50.242, -47.132, -43.814, -40.446,
-25.935, -22.665, -55.682, -55.999, -54.944, -52.929, -50.265, -47.153, -43.834,
-40.464, -25.950, -22.679, -55.708, -56.025, -54.969, -52.953, -50.288, -47.175,
-43.854, -40.482, -25.965, -22.694, -55.734, -56.051, -54.995, -52.978, -50.312,
-47.197, -43.874, -40.500, -25.980, -22.708, -55.760, -56.077, -55.021, -53.003,
-50.335, -47.218, -43.894, -40.518, -25.995, -22.722, -55.786, -56.104, -55.047,
-53.028, -50.358, -47.240, -43.913, -40.536, -26.010, -22.736, -55.812, -56.130,
-55.072, -53.052, -50.382, -47.261, -43.933, -40.554, -26.026, -22.756, -55.837,
-56.156, -55.098, -53.077, -50.405, -47.283, -43.953, -40.573, -26.041, -22.770,
 -55.863, -56.182, -55.124, -53.102, -50.428, -47.305, -43.973, -40.591,
-26.056, -22.784, -55.889, -56.208, -55.150, -53.126, -50.451, -47.326, -43.993,
-40.609, -26.071, -22.798, -55.915, -56.235, -55.175, -53.151, -50.475, -47.348,
-44.013, -40.627, -26.086, -22.812, -55.941, -56.261, -55.201, -53.176, -50.498,
-47.370, -44.032, -40.645, -26.101, -22.826, -55.967, -56.287, -55.227, -53.201,
-50.521, -47.391, -44.052, -40.663, -26.116, -22.841, -55.993, -56.313, -55.253,
-53.225, -50.545, -47.413, -44.072, -40.681, -26.132, -22.855, -56.019, -56.339,
-55.278, -53.250, -50.568, -47.434, -44.092, -40.699, -26.147, -22.869, -56.045,
-56.365, -55.304, -53.275, -50.591, -47.456, -44.112, -40.717, -26.162, -22.883,
 -56.070, -56.392, -55.330, -53.299, -50.614, -47.478, -44.132, -40.735,
-26.177, -22.897, -56.096, -56.418, -55.356, -53.324, -50.638, -47.499, -44.151,
-40.753, -26.192, -22.911, -56.122, -56.444, -55.381, -53.349, -50.661, -47.521,
-44.171, -40.771, -26.207, -22.925, -56.148, -56.470, -55.407, -53.373, -50.684,
-47.543, -44.191, -40.789, -26.222, -22.939, -56.174, -56.496, -55.433, -53.398,
-50.708, -47.564, -44.211, -40.807, -26.237, -22.953, -56.200, -56.522, -55.458,
-53.423, -50.731, -47.586, -44.231, -40.825, -26.252, -22.967, -56.226, -56.549,
-55.484, -53.448, -50.754, -47.607, -44.251, -40.843, -26.267, -22.981, -56.252,
-56.575, -55.510, -53.472, -50.777, -47.629, -44.270, -40.861, -26.282, -22.995,
 -56.277, -56.601, -55.536, -53.497, -50.801, -47.651, -44.290, -40.879,
-26.297, -23.009, -56.303, -56.627, -55.561, -53.522, -50.824, -47.672, -44.310,
-40.897, -26.312, -23.023, -56.329, -56.653, -55.587, -53.546, -50.847, -47.694,
-44.330, -40.916, -26.327, -23.037, -56.355, -56.680, -55.613, -53.571, -50.871,
-47.715, -44.350, -40.934, -26.343, -23.154, -56.381, -56.706, -55.639, -53.596,
-50.894, -47.737, -44.370, -40.952, -26.358, -23.168, -56.407, -56.732, -55.664,
-53.621, -50.917, -47.759, -44.389, -40.970, -26.373, -23.182, -56.433, -56.758,
-55.690, -53.645, -50.940, -47.780, -44.409, -40.988, -26.388, -23.196, -56.459,
-56.784, -55.716, -53.670, -50.964, -47.802, -44.429, -41.006, -26.403, -23.210,
 -56.484, -56.810, -55.741, -53.695, -50.987, -47.823, -44.449, -41.024,
-27.016, -23.224, -56.510, -56.837, -55.767, -53.719, -51.010, -47.845, -44.469,
-41.042, -27.031, -23.238, -56.536, -56.863, -55.793, -53.744, -51.033, -47.867,
-44.489, -41.060, -27.046, -23.252, -56.562, -56.889, -55.819, -53.769, -51.057,
-47.888, -44.508, -41.078, -27.062, -23.266, -56.588, -56.915, -55.844, -53.794,
-51.080, -47.910, -44.528, -41.096, -27.077, -23.280, -56.614, -56.941, -55.870,
-53.818, -51.103, -47.931, -44.548, -41.114, -27.092, -23.294, -56.640, -56.967,
-55.896, -53.843, -51.127, -47.953, -44.568, -41.132, -27.107, -23.308, -56.666,
-56.994, -55.922, -53.868, -51.150, -47.975, -44.588, -41.150, -27.122, -23.322,
 -56.691, -57.020, -55.947, -53.892, -51.173, -47.996, -44.608, -41.168,
-27.137, -23.336, -56.717, -57.046, -55.973, -53.917, -51.196, -48.018, -44.627,
-41.186, -27.152, -23.350, -56.743, -57.072, -55.999, -53.942, -51.220, -48.040,
-44.647, -41.204, -27.167, -23.364, -56.769, -57.098, -56.024, -53.966, -51.243,
-48.061, -44.667, -41.222, -27.183, -23.378, -56.795, -57.124, -56.050, -53.991,
-51.266, -48.083, -44.687, -41.240, -27.198, -23.392, -56.821, -57.151, -56.076,
-54.016, -51.289, -48.104, -44.707, -41.258, -27.213, -23.406, -56.847, -57.177,
-56.102, -54.041, -51.313, -48.126, -44.726, -41.276, -27.228, -23.420, -56.872,
-57.203, -56.127, -54.065, -51.336, -48.148, -44.746, -41.294, -27.243, -23.434,
 -56.898, -57.229, -56.153, -54.090, -51.359, -48.169, -44.766, -41.312,
-27.258, -23.448, -56.924, -57.255, -56.179, -54.115, -51.383, -48.191, -44.786,
-41.330, -27.273, -23.462, -56.950, -57.281, -56.204, -54.139, -51.406, -48.212,
-44.806, -41.349, -27.288, -23.476, -56.976, -57.308, -56.230, -54.164, -51.429,
-48.234, -44.826, -41.367, -27.303, -23.490, -57.002, -57.334, -56.256, -54.189,
-51.452, -48.256, -44.845, -41.385, -27.318, -23.504, -57.028, -57.360, -56.282,
-54.213, -51.476, -48.277, -44.865, -41.403, -27.333, -23.518, -57.054, -57.386,
-56.307, -54.238, -51.499, -48.299, -44.885, -41.421, -27.348, -23.532, -57.079,
-57.412, -56.333, -54.263, -51.522, -48.320, -44.905, -41.439, -27.363, -23.546,
 -57.105, -57.438, -56.359, -54.288, -51.545, -48.342, -44.925, -41.457,
-27.378, -23.560, -57.131, -57.465, -56.385, -54.312, -51.569, -48.364, -44.944,
-41.475, -27.393, -23.574, -57.157, -57.491, -56.410, -54.337, -51.592, -48.385,
-44.964, -41.493, -27.408, -23.588, -57.183, -57.517, -56.436, -54.362, -51.615,
-48.407, -44.984, -41.511, -27.423, -23.601, -57.209, -57.543, -56.462, -54.386,
-51.639, -48.428, -45.004, -41.529, -27.438, -23.615, -57.235, -57.569, -56.487,
-54.411, -51.662, -48.450, -45.024, -41.547, -27.453, -23.629, -57.260, -57.595,
-56.513, -54.436, -51.685, -48.471, -45.044, -41.565, -27.468, -23.643, -57.286,
-57.622, -56.539, -54.460, -51.708, -48.493, -45.063, -41.583, -27.483, -23.657,
 -57.312, -57.648, -56.565, -54.485, -51.732, -48.515, -45.083, -41.601,
-27.498, -23.671, -57.338, -57.674, -56.590, -54.510, -51.755, -48.536, -45.103,
-41.619, -27.513, -23.685, -57.364, -57.700, -56.616, -54.534, -51.778, -48.558,
-45.123, -41.637, -27.528, -23.699, -57.390, -57.726, -56.642, -54.559, -51.801,
-48.579, -45.143, -41.655, -27.543, -23.713, -57.416, -57.752, -56.667, -54.584,
-51.825, -48.601, -45.162, -41.673, -27.558, -23.726, -57.442, -57.779, -56.693,
-54.609, -51.848, -48.623, -45.182, -41.691, -27.573, -23.740, -57.467, -57.805,
-56.719, -54.633, -51.871, -48.644, -45.202, -41.709, -27.588, -23.754, -57.493,
-57.831, -56.745, -54.658, -51.894, -48.666, -45.222, -41.727, -27.603, -23.768,
 -57.519, -57.857, -56.770, -54.683, -51.918, -48.687, -45.242, -41.745,
-27.618, -23.782, -57.545, -57.883, -56.796, -54.707, -51.941, -48.709, -45.261,
-41.763, -27.633, -23.796, -57.571, -57.909, -56.822, -54.732, -51.964, -48.731,
-45.281, -41.781, -27.648, -23.810, -57.597, -57.935, -56.847, -54.757, -51.988,
-48.752, -45.301, -41.799, -27.663, -23.823, -57.623, -57.962, -56.873, -54.781,
-52.011, -48.774, -45.321, -41.817, -27.678, -23.837, -57.648, -57.988, -56.899,
-54.806, -52.034, -48.795, -45.341, -41.835, -27.693, -23.851, -57.674, -58.014,
-56.925, -54.831, -52.057, -48.817, -45.361, -41.853, -27.708, -23.865, -57.700,
-58.040, -56.950, -54.855, -52.081, -48.839, -45.380, -41.871, -27.723, -23.879,
 -57.726, -58.066, -56.976, -54.880, -52.104, -48.860, -45.400, -41.889,
-27.738, -23.893, -57.752, -58.092, -57.002, -54.905, -52.127, -48.882, -45.420,
-41.907, -27.753, -23.906, -57.778, -58.119, -57.027, -54.930, -52.150, -48.903,
-45.440, -41.925, -27.768, -23.920, -57.804, -58.145, -57.053, -54.954, -52.174,
-48.925, -45.460, -41.943, -27.783, -23.934, -57.829, -58.171, -57.079, -54.979,
-52.197, -48.946, -45.479, -41.961, -27.797, -23.948, -57.855, -58.197, -57.105,
-55.004, -52.220, -48.968, -45.499, -41.979, -27.812, -23.962, -57.881, -58.223,
-57.130, -55.028, -52.243, -48.990, -45.519, -41.997, -27.827, -23.975, -57.907,
-58.249, -57.156, -55.053, -52.267, -49.011, -45.539, -42.015, -27.842, -23.989,
 -57.933, -58.276, -57.182, -55.078, -52.290, -49.033, -45.559, -42.033,
-27.857, -24.003, -57.959, -58.302, -57.207, -55.102, -52.313, -49.054, -45.578,
-42.051, -27.872, -24.017, -57.985, -58.328, -57.233, -55.127, -52.336, -49.076,
-45.598, -42.069, -27.887, -24.202, -58.010, -58.354, -57.259, -55.152, -52.360,
-49.098, -45.618, -42.087, -27.902, -24.216, -58.036, -58.380, -57.284, -55.176,
-52.383, -49.119, -45.638, -42.105, -27.917, -24.229, -58.062, -58.406, -57.310,
-55.201, -52.406, -49.141, -45.658, -42.123, -27.931, -24.243, -58.088, -58.432,
-57.336, -55.226, -52.429, -49.162, -45.677, -42.141, -27.946, -24.257, -58.114,
-58.459, -57.362, -55.250, -52.453, -49.184, -45.697, -42.159, -38.753, -24.271,
 -58.140, -58.485, -57.387, -55.275, -52.476, -49.205, -45.717, -42.177,
-38.769, -24.285, -58.166, -58.511, -57.413, -55.300, -52.499, -49.227, -45.737,
-42.195, -38.786, -24.299, -58.191, -58.537, -57.439, -55.324, -52.522, -49.249,
-45.757, -42.213, -38.802, -24.312, -58.217, -58.563, -57.464, -55.349, -52.546,
-49.270, -45.776, -42.232, -38.818, -24.326, -58.243, -58.589, -57.490, -55.374,
-52.569, -49.292, -45.796, -42.250, -38.834, -24.340, -58.269, -58.616, -57.516,
-55.399, -52.592, -49.313, -45.816, -42.268, -38.851, -24.354, -58.295, -58.642,
-57.542, -55.423, -52.615, -49.335, -45.836, -42.286, -38.867, -24.368, -58.321,
-58.668, -57.567, -55.448, -52.639, -49.357, -45.856, -42.304, -38.883, -24.382,
 -58.347, -58.694, -57.593, -55.473, -52.662, -49.378, -45.875, -42.322,
-38.900, -24.395, -58.372, -58.720, -57.619, -55.497, -52.685, -49.400, -45.895,
-42.340, -38.916, -24.409, -58.398, -58.746, -57.644, -55.522, -52.708, -49.421,
-45.915, -42.358, -38.932, -24.423, -58.424, -58.772, -57.670, -55.547, -52.732,
-49.443, -45.935, -42.376, -38.948, -24.437, -58.450, -58.799, -57.696, -55.571,
-52.755, -49.464, -45.954, -42.394, -38.965, -24.451, -58.476, -58.825, -57.721,
-55.596, -52.778, -49.486, -45.974, -42.412, -38.981, -24.464, -58.502, -58.851,
-57.747, -55.621, -52.801, -49.508, -45.994, -42.430, -38.997, -24.478, -58.528,
-58.877, -57.773, -55.645, -52.825, -49.529, -46.014, -42.448, -39.013, -24.492,
 -58.553, -58.903, -57.799, -55.670, -52.848, -49.551, -46.034, -42.466,
-39.030, -24.506, -58.579, -58.929, -57.824, -55.695, -52.871, -49.572, -46.053,
-42.484, -39.046, -24.519, -58.605, -58.955, -57.850, -55.719, -52.894, -49.594,
-46.073, -42.502, -39.062, -24.533, -58.631, -58.982, -57.876, -55.744, -52.918,
-49.615, -46.093, -42.520, -39.079, -24.547, -58.657, -59.008, -57.901, -55.769,
-52.941, -49.637, -46.113, -42.538, -39.095, -24.561, -58.683, -59.034, -57.927,
-55.793, -52.964, -49.659, -46.133, -42.556, -39.111, -24.574, -58.709, -59.060,
-57.953, -55.818, -52.987, -49.680, -46.152, -42.574, -39.127, -24.588, -58.734,
-59.086, -57.979, -55.843, -53.011, -49.702, -46.172, -42.592, -39.144, -24.602,
 -58.760, -59.112, -58.004, -55.867, -53.034, -49.723, -46.192, -42.610,
-39.160, -24.616, -58.786, -59.139, -58.030, -55.892, -53.057, -49.745, -46.212,
-42.628, -39.176, -24.629, -58.812, -59.165, -58.056, -55.917, -53.080, -49.766,
-46.232, -42.646, -39.192, -24.643, -58.838, -59.191, -58.081, -55.941, -53.104,
-49.788, -46.251, -42.664, -39.209, -24.657, -58.864, -59.217, -58.107, -55.966,
-53.127, -49.810, -46.271, -42.681, -39.225, -24.670, -58.890, -59.243, -58.133,
-55.991, -53.150, -49.831, -46.291, -42.699, -39.241, -24.684, -58.915, -59.269,
-58.158, -56.015, -53.173, -49.853, -46.311, -42.717, -39.258, -24.698, -58.941,
-59.295, -58.184, -56.040, -53.197, -49.874, -46.330, -42.735, -39.274, -24.712,
 -58.967, -59.322, -58.210, -56.065, -53.220, -49.896, -46.350, -42.753,
-39.290, -24.725, -58.993, -59.348, -58.235, -56.089, -53.243, -49.917, -46.370,
-42.771, -39.306, -24.739, -59.019, -59.374, -58.261, -56.114, -53.266, -49.939,
-46.390, -42.789, -39.323, -24.753, -59.045, -59.400, -58.287, -56.139, -53.290,
-49.961, -46.410, -42.807, -39.339, -24.766, -59.070, -59.426, -58.313, -56.163,
-53.313, -49.982, -46.429, -42.825, -39.355, -24.780, -59.096, -59.452, -58.338,
-56.188, -53.336, -50.004, -46.449, -42.843, -39.371, -24.794, -59.122, -59.478,
-58.364, -56.213, -53.359, -50.025, -46.469, -42.861, -39.388, -24.807, -59.148,
-59.505, -58.390, -56.238, -53.383, -50.047, -46.489, -42.879, -39.404, -24.821,
 -59.174, -59.531, -58.415, -56.262, -53.406, -50.068, -46.508, -42.897,
-39.420, -24.835, -59.200, -59.557, -58.441, -56.287, -53.429, -50.090, -46.528,
-42.915, -39.436, -24.848, -59.226, -59.583, -58.467, -56.312, -53.452, -50.111,
-46.548, -42.933, -39.453, -24.862, -59.251, -59.609, -58.492, -56.336, -53.476,
-50.133, -46.568, -42.951, -39.469, -24.876, -59.277, -59.635, -58.518, -56.361,
-53.499, -50.155, -46.588, -42.969, -39.485, -24.889, -59.303, -59.661, -58.544,
-56.386, -53.522, -50.176, -46.607, -42.987, -39.501, -24.903, -59.329, -59.688,
-58.570, -56.410, -53.545, -50.198, -46.627, -43.005, -39.518, -24.917, -59.355,
-59.714, -58.595, -56.435, -53.568, -50.219, -46.647, -43.023, -39.534, -24.930,
 -59.381, -59.740, -58.621, -56.460, -53.592, -50.241, -46.667, -43.041,
-39.550, -24.944, -59.406, -59.766, -58.647, -56.484, -53.615, -50.262, -46.686,
-43.059, -39.566, -24.957, -59.432, -59.792, -58.672, -56.509, -53.638, -50.284,
-46.706, -43.077, -39.583, -24.971, -59.458, -59.818, -58.698, -56.534, -53.661,
-50.306, -46.726, -43.095, -39.599, -24.985, -59.484, -59.844, -58.724, -56.558,
-53.685, -50.327, -46.746, -43.113, -39.615, -24.998, -59.510, -59.871, -58.749,
-56.583, -53.708, -50.349, -46.766, -43.131, -39.631, -25.012, -59.536, -59.897,
-58.775, -56.608, -53.731, -50.370, -46.785, -43.149, -39.648, -25.026, -59.562,
-59.923, -58.801, -56.632, -53.754, -50.392, -46.805, -43.167, -39.664, -25.039,
 -59.587, -59.949, -58.826, -56.657, -53.778, -50.413, -46.825, -43.185,
-39.680, -25.053, -59.613, -59.975, -58.852, -56.682, -53.801, -50.435, -46.845,
-43.203, -39.696, -25.362, -59.639, -60.001, -58.878, -56.706, -53.824, -50.456,
-46.864, -43.221, -39.713, -25.376, -59.665, -60.027, -58.904, -56.731, -53.847,
-50.478, -46.884, -43.239, -39.729, -25.389, -59.691, -60.053, -58.929, -56.756,
-53.871, -50.500, -46.904, -43.257, -39.745, -25.403, -59.717, -60.080, -58.955,
-56.780, -53.894, -50.521, -46.924, -43.275, -39.761, -25.417, -59.742, -60.106,
-58.981, -56.805, -53.917, -50.543, -46.944, -43.293, -39.778, -25.430, -59.768,
-60.132, -59.006, -56.830, -53.940, -50.564, -46.963, -43.311, -39.794, -25.444,
 -59.794, -60.158, -59.032, -56.854, -53.963, -50.586, -46.983, -43.329,
-39.810, -25.458, -59.820, -60.184, -59.058, -56.879, -53.987, -50.607, -47.003,
-43.347, -39.826, -25.471, -59.846, -60.210, -59.083, -56.904, -54.010, -50.629,
-47.023, -43.365, -39.843, -25.485, -59.872, -60.236, -59.109, -56.928, -54.033,
-50.650, -47.042, -43.383, -39.859, -25.499, -59.897, -60.263, -59.135, -56.953,
-54.056, -50.672, -47.062, -43.401, -39.875, -25.512, -59.923, -60.289, -59.160,
-56.977, -54.080, -50.694, -47.082, -43.419, -39.891, -25.526, -59.949, -60.315,
-59.186, -57.002, -54.103, -50.715, -47.102, -43.437, -39.908, -25.540, -59.975,
-60.341, -59.212, -57.027, -54.126, -50.737, -47.121, -43.455, -39.924, -25.553,
 -60.001, -60.367, -59.237, -57.051, -54.149, -50.758, -47.141, -43.473,
-39.940, -25.567, -60.027, -60.393, -59.263, -57.076, -54.173, -50.780, -47.161,
-43.491, -39.956, -25.581, -60.053, -60.419, -59.289, -57.101, -54.196, -50.801,
-47.181, -43.509, -39.973, -25.594, -60.078, -60.446, -59.315, -57.125, -54.219,
-50.823, -47.200, -43.527, -39.989, -25.608, -60.104, -60.472, -59.340, -57.150,
-54.242, -50.844, -47.220, -43.545, -40.005, -25.621, -60.130, -60.498, -59.366,
-57.175, -54.265, -50.866, -47.240, -43.563, -40.021, -25.635, -60.156, -60.524,
-59.392, -57.199, -54.289, -50.888, -47.260, -43.581, -40.038, -25.649, -60.182,
-60.550, -59.417, -57.224, -54.312, -50.909, -47.280, -43.599, -40.054, -25.662,
 -60.208, -60.576, -59.443, -57.249, -54.335, -50.931, -47.299, -43.616,
-40.070, -25.676, -60.233, -60.602, -59.469, -57.273, -54.358, -50.952, -47.319,
-43.634, -40.086, -25.689, -60.259, -60.628, -59.494, -57.298, -54.382, -50.974,
-47.339, -43.652, -40.103, -25.703, -60.285, -60.655, -59.520, -57.323, -54.405,
-50.995, -47.359, -43.670, -40.119, -25.717, -60.311, -60.681, -59.546, -57.347,
-54.428, -51.017, -47.378, -43.688, -40.135, -25.730, -60.337, -60.707, -59.571,
-57.372, -54.451, -51.038, -47.398, -43.706, -40.151, -25.744, -60.363, -60.733,
-59.597, -57.397, -54.475, -51.060, -47.418, -43.724, -40.168, -25.757, -60.388,
-60.759, -59.623, -57.421, -54.498, -51.081, -47.438, -43.742, -40.184, -25.771,
 -60.414, -60.785, -59.648, -57.446, -54.521, -51.103, -47.457, -43.760,
-40.200, -25.784, -60.440, -60.811, -59.674, -57.471, -54.544, -51.125, -47.477,
-43.778, -40.216, -25.798, -60.466, -60.837, -59.700, -57.495, -54.567, -51.146,
-47.497, -43.796, -40.232, -25.812, -60.492, -60.864, -59.725, -57.520, -54.591,
-51.168, -47.517, -43.814, -40.249, -25.825, -60.518, -60.890, -59.751, -57.545,
-54.614, -51.189, -47.536, -43.832, -40.265, -25.839, -60.543, -60.916, -59.777,
-57.569, -54.637, -51.211, -47.556, -43.850, -40.281, -25.852, -60.569, -60.942,
-59.802, -57.594, -54.660, -51.232, -47.576, -43.868, -40.297, -25.866, -60.595,
-60.968, -59.828, -57.619, -54.684, -51.254, -47.596, -43.886, -40.314, -25.879,
 -60.621, -60.994, -59.854, -57.643, -54.707, -51.275, -47.615, -43.904,
-40.330, -25.893, -60.647, -61.020, -59.880, -57.668, -54.730, -51.297, -47.635,
-43.922, -40.346, -25.906, -60.673, -61.047, -59.905, -57.693, -54.753, -51.318,
-47.655, -43.940, -40.362, -25.920, -60.698, -61.073, -59.931, -57.717, -54.776,
-51.340, -47.675, -43.958, -40.379, -25.933, -60.724, -61.099, -59.957, -57.742,
-54.800, -51.362, -47.694, -43.976, -40.395, -25.947, -60.750, -61.125, -59.982,
-57.766, -54.823, -51.383, -47.714, -43.994, -40.411, -25.960, -60.776, -61.151,
-60.008, -57.791, -54.846, -51.405, -47.734, -44.012, -40.427, -25.974, -60.802,
-61.177, -60.034, -57.816, -54.869, -51.426, -47.754, -44.030, -40.443, -25.987,
 -60.828, -61.203, -60.059, -57.840, -54.892, -51.448, -47.773, -44.048,
-40.460, -26.001, -60.853, -61.229, -60.085, -57.865, -54.916, -51.469, -47.793,
-44.066, -40.476, -26.014, -60.879, -61.256, -60.111, -57.890, -54.939, -51.491,
-47.813, -44.083, -40.492, -26.028, -60.905, -61.282, -60.136, -57.914, -54.962,
-51.512, -47.833, -44.101, -40.508, -26.041, -60.931, -61.308, -60.162, -57.939,
-54.985, -51.534, -47.852, -44.119, -40.525, -26.055, -60.957, -61.334, -60.188,
-57.964, -55.009, -51.555, -47.872, -44.137, -40.541, -26.068, -60.983, -61.360,
-60.213, -57.988, -55.032, -51.577, -47.892, -44.155, -40.557, -26.082, -61.008,
-61.386, -60.239, -58.013, -55.055, -51.599, -47.912, -44.173, -40.573, -26.095,
 -61.034, -61.412, -60.265, -58.038, -55.078, -51.620, -47.931, -44.191,
-40.589, -26.109, -61.060, -61.438, -60.290, -58.062, -55.101, -51.642, -47.951,
-44.209, -40.606, -26.122, -61.086, -61.465, -60.316, -58.087, -55.125, -51.663,
-47.971, -44.227, -40.622, -26.136, -61.112, -61.491, -60.342, -58.112, -55.148,
-51.685, -47.991, -44.245, -40.638, -26.149, -61.138, -61.517, -60.367, -58.136,
-55.171, -51.706, -48.010, -44.263, -40.654, -26.163, -61.163, -61.543, -60.393,
-58.161, -55.194, -51.728, -48.030, -44.281, -40.671, -26.176, -61.189, -61.569,
-60.419, -58.186, -55.218, -51.749, -48.050, -44.299, -40.687, -26.190, -61.215,
-61.595, -60.444, -58.210, -55.241, -51.771, -48.070, -44.317, -40.703, -26.203,
 -61.241, -61.621, -60.470, -58.235, -55.264, -51.792, -48.089, -44.335,
-40.719, -26.810, -61.267, -61.647, -60.496, -58.259, -55.287, -51.814, -48.109,
-44.353, -40.735, -26.823, -61.293, -61.674, -60.521, -58.284, -55.310, -51.835,
-48.129, -44.371, -40.752, -26.837, -61.318, -61.700, -60.547, -58.309, -55.334,
-51.857, -48.149, -44.389, -40.768, -26.850, -61.344, -61.726, -60.573, -58.333,
-55.357, -51.879, -48.168, -44.407, -40.784, -26.864, -61.370, -61.752, -60.598,
-58.358, -55.380, -51.900, -48.188, -44.425, -40.800, -26.877, -61.396, -61.778,
-60.624, -58.383, -55.403, -51.922, -48.208, -44.443, -40.816, -26.891, -61.422,
-61.804, -60.650, -58.407, -55.426, -51.943, -48.228, -44.460, -40.833, -26.904,
 -61.447, -61.830, -60.675, -58.432, -55.450, -51.965, -48.247, -44.478,
-40.849, -26.918, -61.473, -61.856, -60.701, -58.457, -55.473, -51.986, -48.267,
-44.496, -40.865, -26.931, -61.499, -61.882, -60.727, -58.481, -55.496, -52.008,
-48.287, -44.514, -40.881, -26.945, -61.525, -61.909, -60.752, -58.506, -55.519,
-52.029, -48.307, -44.532, -40.898, -26.958, -61.551, -61.935, -60.778, -58.531,
-55.543, -52.051, -48.326, -44.550, -40.914, -26.972, -61.577, -61.961, -60.804,
-58.555, -55.566, -52.072, -48.346, -44.568, -40.930, -26.985, -61.602, -61.987,
-60.829, -58.580, -55.589, -52.094, -48.366, -44.586, -40.946, -26.999, -61.628,
-62.013, -60.855, -58.605, -55.612, -52.115, -48.386, -44.604, -40.962, -27.012,
 -61.654, -62.039, -60.881, -58.629, -55.635, -52.137, -48.405, -44.622,
-40.979, -27.026, -61.680, -62.065, -60.907, -58.654, -55.659, -52.158, -48.425,
-44.640, -40.995, -27.039, -61.706, -62.091, -60.932, -58.678, -55.682, -52.180,
-48.445, -44.658, -41.011, -27.053, -61.732, -62.118, -60.958, -58.703, -55.705,
-52.201, -48.465, -44.676, -41.027, -27.066, -61.757, -62.144, -60.984, -58.728,
-55.728, -52.223, -48.484, -44.694, -41.043, -27.080, -61.783, -62.170, -61.009,
-58.752, -55.751, -52.245, -48.504, -44.712, -41.060, -27.093, -61.809, -62.196,
-61.035, -58.777, -55.775, -52.266, -48.524, -44.730, -41.076, -27.107, -61.835,
-62.222, -61.061, -58.802, -55.798, -52.288, -48.544, -44.748, -41.092, -27.120,
 -61.861, -62.248, -61.086, -58.826, -55.821, -52.309, -48.563, -44.765,
-41.108, -27.134, -61.887, -62.274, -61.112, -58.851, -55.844, -52.331, -48.583,
-44.783, -41.124, -27.147, -61.912, -62.300, -61.138, -58.876, -55.867, -52.352,
-48.603, -44.801, -41.141, -27.161, -61.938, -62.327, -61.163, -58.900, -55.891,
-52.374, -48.622, -44.819, -41.157, -27.174, -61.964, -62.353, -61.189, -58.925,
-55.914, -52.395, -48.642, -44.837, -41.173, -27.187, -61.990, -62.379, -61.215,
-58.949, -55.937, -52.417, -48.662, -44.855, -41.189, -27.201, -62.016, -62.405,
-61.240, -58.974, -55.960, -52.438, -48.682, -44.873, -41.205, -27.214, -62.041,
-62.431, -61.266, -58.999, -55.983, -52.460, -48.701, -44.891, -41.222, -27.228,
 -62.067, -62.457, -61.292, -59.023, -56.007, -52.481, -48.721, -44.909,
-41.238, -27.241, -62.093, -62.483, -61.317, -59.048, -56.030, -52.503, -48.741,
-44.927, -41.254, -27.255, -62.119, -62.509, -61.343, -59.073, -56.053, -52.524,
-48.761, -44.945, -41.270, -27.268, -62.145, -62.535, -61.369, -59.097, -56.076,
-52.546, -48.780, -44.963, -41.286, -27.281, -62.171, -62.562, -61.394, -59.122,
-56.099, -52.567, -48.800, -44.981, -41.303, -27.295, -62.196, -62.588, -61.420,
-59.147, -56.123, -52.589, -48.820, -44.999, -41.319, -27.308, -62.222, -62.614,
-61.446, -59.171, -56.146, -52.610, -48.840, -45.017, -41.335, -27.322, -62.248,
-62.640, -61.471, -59.196, -56.169, -52.632, -48.859, -45.035, -41.351, -27.335,
 -62.274, -62.666, -61.497, -59.220, -56.192, -52.654, -48.879, -45.052,
-41.367, -27.348, -62.300, -62.692, -61.522, -59.245, -56.215, -52.675, -48.899,
-45.070, -41.384, -27.362, -62.325, -62.718, -61.548, -59.270, -56.239, -52.697,
-48.918, -45.088, -41.400, -27.375, -62.351, -62.744, -61.574, -59.294, -56.262,
-52.718, -48.938, -45.106, -41.416, -27.389, -62.377, -62.770, -61.599, -59.319,
-56.285, -52.740, -48.958, -45.124, -41.432, -27.402, -62.403, -62.797, -61.625,
-59.344, -56.308, -52.761, -48.978, -45.142, -41.448, -27.415, -62.429, -62.823,
-61.651, -59.368, -56.331, -52.783, -48.997, -45.160, -41.465, -27.429, -62.455,
-62.849, -61.676, -59.393, -56.355, -52.804, -49.017, -45.178, -41.481, -27.442,
 -62.480, -62.875, -61.702, -59.418, -56.378, -52.826, -49.037, -45.196,
-41.497, -27.456, -62.506, -62.901, -61.728, -59.442, -56.401, -52.847, -49.057,
-45.214, -41.513, -27.469, -62.532, -62.927, -61.753, -59.467, -56.424, -52.869,
-49.076, -45.232, -41.529, -27.482, -62.558, -62.953, -61.779, -59.491, -56.447,
-52.890, -49.096, -45.250, -41.546, -27.496, -62.584, -62.979, -61.805, -59.516,
-56.471, -52.912, -49.116, -45.268, -41.562, -27.509, -62.609, -63.005, -61.830,
-59.541, -56.494, -52.933, -49.136, -45.286, -41.578, -27.522, -62.635, -63.032,
-61.856, -59.565, -56.517, -52.955, -49.155, -45.303, -41.594, -27.536, -62.661,
-63.058, -61.882, -59.590, -56.540, -52.976, -49.175, -45.321, -41.610, -27.549,
 -62.687, -63.084, -61.907, -59.615, -56.563, -52.998, -49.195, -45.339,
-41.627, -27.562, -62.713, -63.110, -61.933, -59.639, -56.587, -53.019, -49.214,
-45.357, -41.643, -27.576, -62.739, -63.136, -61.959, -59.664, -56.610, -53.041,
-49.234, -45.375, -41.659, -27.589, -62.764, -63.162, -61.984, -59.689, -56.633,
-53.062, -49.254, -45.393, -41.675, -27.602, -62.790, -63.188, -62.010, -59.713,
-56.656, -53.084, -49.274, -45.411, -41.691, -27.616, -62.816, -63.214, -62.036,
-59.738, -56.679, -53.105, -49.293, -45.429, -41.707, -27.629, -62.842, -63.240,
-62.061, -59.762, -56.703, -53.127, -49.313, -45.447, -41.724, -27.642, -62.868,
-63.267, -62.087, -59.787, -56.726, -53.148, -49.333, -45.465, -41.740, -38.241,
 -62.893, -63.293, -62.113, -59.812, -56.749, -53.170, -49.352, -45.483,
-41.756, -38.256, -62.919, -63.319, -62.138, -59.836, -56.772, -53.192, -49.372,
-45.501, -41.772, -38.271, -62.945, -63.345, -62.164, -59.861, -56.795, -53.213,
-49.392, -45.519, -41.788, -38.285, -62.971, -63.371, -62.190, -59.886, -56.819,
-53.235, -49.412, -45.536, -41.805, -38.300, -62.997, -63.397, -62.215, -59.910,
-56.842, -53.256, -49.431, -45.554, -41.821, -38.314, -63.023, -63.423, -62.241,
-59.935, -56.865, -53.278, -49.451, -45.572, -41.837, -38.329, -63.048, -63.449,
-62.267, -59.959, -56.888, -53.299, -49.471, -45.590, -41.853, -38.343, -63.074,
-63.475, -62.292, -59.984, -56.911, -53.321, -49.491, -45.608, -41.869, -38.358,
 -63.100, -63.502, -62.318, -60.009, -56.935, -53.342, -49.510, -45.626,
-41.886, -38.372, -63.126, -63.528, -62.344, -60.033, -56.958, -53.364, -49.530,
-45.644, -41.902, -38.387, -63.152, -63.554, -62.369, -60.058, -56.981, -53.385,
-49.550, -45.662, -41.918, -38.402, -63.177, -63.580, -62.395, -60.083, -57.004,
-53.407, -49.569, -45.680, -41.934, -38.416, -63.203, -63.606, -62.421, -60.107,
-57.027, -53.428, -49.589, -45.698, -41.950, -38.431, -63.229, -63.632, -62.446,
-60.132, -57.051, -53.450, -49.609, -45.716, -41.966, -38.445, -63.255, -63.658,
-62.472, -60.156, -57.074, -53.471, -49.629, -45.734, -41.983, -38.460, -63.281,
-63.684, -62.498, -60.181, -57.097, -53.493, -49.648, -45.752, -41.999, -38.474,
 -63.306, -63.710, -62.523, -60.206, -57.120, -53.514, -49.668, -45.769,
-42.015, -38.489, -63.332, -63.736, -62.549, -60.230, -57.143, -53.536, -49.688,
-45.787, -42.031, -38.503, -63.358, -63.763, -62.574, -60.255, -57.167, -53.557,
-49.707, -45.805, -42.047, -38.518, -63.384, -63.789, -62.600, -60.280, -57.190,
-53.579, -49.727, -45.823, -42.064, -38.533, -63.410, -63.815, -62.626, -60.304,
-57.213, -53.600, -49.747, -45.841, -42.080, -38.547, -63.436, -63.841, -62.651,
-60.329, -57.236, -53.622, -49.767, -45.859, -42.096, -38.562, -63.461, -63.867,
-62.677, -60.353, -57.259, -53.643, -49.786, -45.877, -42.112, -38.576, -63.487,
-63.893, -62.703, -60.378, -57.282, -53.665, -49.806, -45.895, -42.128, -38.591,
 -63.513, -63.919, -62.728, -60.403, -57.306, -53.686, -49.826, -45.913,
-42.144, -38.605, -63.539, -63.945, -62.754, -60.427, -57.329, -53.708, -49.845,
-45.931, -42.161, -38.620, -63.565, -63.971, -62.780, -60.452, -57.352, -53.729,
-49.865, -45.949, -42.177, -38.634, -63.590, -63.998, -62.805, -60.477, -57.375,
-53.751, -49.885, -45.966, -42.193, -38.649, -63.616, -64.024, -62.831, -60.501,
-57.398, -53.772, -49.905, -45.984, -42.209, -38.663, -63.642, -64.050, -62.857,
-60.526, -57.422, -53.794, -49.924, -46.002, -42.225, -38.678, -63.668, -64.076,
-62.882, -60.550, -57.445, -53.815, -49.944, -46.020, -42.241, -38.693, -63.694,
-64.102, -62.908, -60.575, -57.468, -53.837, -49.964, -46.038, -42.258, -38.707,
 -63.719, -64.128, -62.934, -60.600, -57.491, -53.858, -49.983, -46.056,
-42.274, -38.722, -63.745, -64.154, -62.959, -60.624, -57.514, -53.880, -50.003,
-46.074, -42.290, -38.736, -63.771, -64.180, -62.985, -60.649, -57.538, -53.901,
-50.023, -46.092, -42.306, -38.751, -63.797, -64.206, -63.011, -60.674, -57.561,
-53.923, -50.043, -46.110, -42.322, -38.765, -63.823, -64.232, -63.036, -60.698,
-57.584, -53.944, -50.062, -46.128, -42.338, -38.780, -63.849, -64.259, -63.062,
-60.723, -57.607, -53.966, -50.082, -46.146, -42.355, -38.794, -63.874, -64.285,
-63.088, -60.747, -57.630, -53.987, -50.102, -46.164, -42.371, -38.809, -63.900,
-64.311, -63.113, -60.772, -57.653, -54.009, -50.121, -46.181, -42.387, -38.823,
 -63.926, -64.337, -63.139, -60.797, -57.677, -54.030, -50.141, -46.199,
-42.403, -38.838, -63.952, -64.363, -63.164, -60.821, -57.700, -54.052, -50.161,
-46.217, -42.419, -38.853, -63.978, -64.389, -63.190, -60.846, -57.723, -54.073,
-50.181, -46.235, -42.435, -38.867, -64.003, -64.415, -63.216, -60.870, -57.746,
-54.095, -50.200, -46.253, -42.452, -38.882, -64.029, -64.441, -63.241, -60.895,
-57.769, -54.116, -50.220, -46.271, -42.468, -38.896, -64.055, -64.467, -63.267,
-60.920, -57.793, -54.138, -50.240, -46.289, -42.484, -38.911, -64.081, -64.493,
-63.293, -60.944, -57.816, -54.159, -50.259, -46.307, -42.500, -38.925, -64.107,
-64.520, -63.318, -60.969, -57.839, -54.181, -50.279, -46.325, -42.516, -38.940,
 -64.132, -64.546, -63.344, -60.994, -57.862, -54.202, -50.299, -46.343,
-42.532, -38.954, -64.158, -64.572, -63.370, -61.018, -57.885, -54.224, -50.319,
-46.360, -42.549, -38.969, -64.184, -64.598, -63.395, -61.043, -57.908, -54.245,
-50.338, -46.378, -42.565, -38.983, -64.210, -64.624, -63.421, -61.067, -57.932,
-54.267, -50.358, -46.396, -42.581, -38.998, -64.236, -64.650, -63.447, -61.092,
-57.955, -54.288, -50.378, -46.414, -42.597, -39.013, -64.261, -64.676, -63.472,
-61.117, -57.978, -54.310, -50.397, -46.432, -42.613, -39.027, -64.287, -64.702,
-63.498, -61.141, -58.001, -54.331, -50.417, -46.450, -42.629, -39.042, -64.313,
-64.728, -63.524, -61.166, -58.024, -54.353, -50.437, -46.468, -42.646, -39.056,
 -64.339, -64.754, -63.549, -61.191, -58.048, -54.374, -50.456, -46.486,
-42.662, -39.071, -64.365, -64.781, -63.575, -61.215, -58.071, -54.396, -50.476,
-46.504, -42.678, -39.085, -64.390, -64.807, -63.600, -61.240, -58.094, -54.417,
-50.496, -46.522, -42.694, -39.100, -64.416, -64.833, -63.626, -61.264, -58.117,
-54.439, -50.516, -46.539, -42.710, -39.114, -64.442, -64.859, -63.652, -61.289,
-58.140, -54.460, -50.535, -46.557, -42.726, -39.129, -64.468, -64.885, -63.677,
-61.314, -58.163, -54.482, -50.555, -46.575, -42.743, -39.143, -64.494, -64.911,
-63.703, -61.338, -58.187, -54.503, -50.575, -46.593, -42.759, -39.158, -64.519,
-64.937, -63.729, -61.363, -58.210, -54.525, -50.594, -46.611, -42.775, -39.172,
 -64.545, -64.963, -63.754, -61.387, -58.233, -54.546, -50.614, -46.629,
-42.791, -39.187, -64.571, -64.989, -63.780, -61.412, -58.256, -54.568, -50.634,
-46.647, -42.807, -39.201, -64.597, -65.015, -63.806, -61.437, -58.279, -54.589,
-50.653, -46.665, -42.823, -39.216, -64.623, -65.041, -63.831, -61.461, -58.303,
-54.611, -50.673, -46.683, -42.840, -39.231, -64.648, -65.068, -63.857, -61.486,
-58.326, -54.632, -50.693, -46.701, -42.856, -39.245, -64.674, -65.094, -63.883,
-61.511, -58.349, -54.654, -50.713, -46.718, -42.872, -39.260, -64.700, -65.120,
-63.908, -61.535, -58.372, -54.675, -50.732, -46.736, -42.888, -39.274, -64.726,
-65.146, -63.934, -61.560, -58.395, -54.697, -50.752, -46.754, -42.904, -39.289,
 -64.752, -65.172, -63.959, -61.584, -58.418, -54.718, -50.772, -46.772,
-42.920, -39.303, -64.777, -65.198, -63.985, -61.609, -58.442, -54.740, -50.791,
-46.790, -42.936, -39.318, -64.803, -65.224, -64.011, -61.634, -58.465, -54.761,
-50.811, -46.808, -42.953, -39.332, -64.829, -65.250, -64.036, -61.658, -58.488,
-54.783, -50.831, -46.826, -42.969, -39.347, -64.855, -65.276, -64.062, -61.683,
-58.511, -54.804, -50.850, -46.844, -42.985, -39.361, -64.881, -65.302, -64.088,
-61.707, -58.534, -54.826, -50.870, -46.862, -43.001, -39.376, -64.906, -65.328,
-64.113, -61.732, -58.557, -54.847, -50.890, -46.880, -43.017, -39.390, -64.932,
-65.355, -64.139, -61.757, -58.581, -54.869, -50.910, -46.897, -43.033, -39.405,
 -64.958, -65.381, -64.165, -61.781, -58.604, -54.890, -50.929, -46.915,
-43.050, -39.419, -64.984, -65.407, -64.190, -61.806, -58.627, -54.912, -50.949,
-46.933, -43.066, -39.434, -65.010, -65.433, -64.216, -61.830, -58.650, -54.933,
-50.969, -46.951, -43.082, -39.448, -65.035, -65.459, -64.242, -61.855, -58.673,
-54.955, -50.988, -46.969, -43.098, -39.463, -65.061, -65.485, -64.267, -61.880,
-58.697, -54.976, -51.008, -46.987, -43.114, -39.477, -65.087, -65.511, -64.293,
-61.904, -58.720, -54.998, -51.028, -47.005, -43.130, -39.492, -65.113, -65.537,
-64.318, -61.929, -58.743, -55.019, -51.047, -47.023, -43.146, -39.506, -65.139,
-65.563, -64.344, -61.953, -58.766, -55.041, -51.067, -47.041, -43.163, -39.521,
 -65.164, -65.589, -64.370, -61.978, -58.789, -55.062, -51.087, -47.058,
-43.179, -39.536, -65.190, -65.615, -64.395, -62.003, -58.812, -55.084, -51.107,
-47.076, -43.195, -39.550, -65.216, -65.642, -64.421, -62.027, -58.836, -55.105,
-51.126, -47.094, -43.211, -39.565, -65.242, -65.668, -64.447, -62.052, -58.859,
-55.127, -51.146, -47.112, -43.227, -39.579, -65.268, -65.694, -64.472, -62.077,
-58.882, -55.148, -51.166, -47.130, -43.243, -39.594, -65.293, -65.720, -64.498,
-62.101, -58.905, -55.170, -51.185, -47.148, -43.259, -39.608, -65.319, -65.746,
-64.524, -62.126, -58.928, -55.191, -51.205, -47.166, -43.276, -39.623, -65.345,
-65.772, -64.549, -62.150, -58.951, -55.213, -51.225, -47.184, -43.292, -39.637,
 -65.371, -65.798, -64.575, -62.175, -58.975, -55.234, -51.244, -47.202,
-43.308, -39.652, -65.397, -65.824, -64.600, -62.200, -58.998, -55.256, -51.264,
-47.219, -43.324, -39.666, -65.422, -65.850, -64.626, -62.224, -59.021, -55.277,
-51.284, -47.237, -43.340, -39.681, -65.448, -65.876, -64.652, -62.249, -59.044,
-55.299, -51.303, -47.255, -43.356, -39.695, -65.474, -65.902, -64.677, -62.273,
-59.067, -55.320, -51.323, -47.273, -43.372, -39.710, -65.500, -65.929, -64.703,
-62.298, -59.090, -55.342, -51.343, -47.291, -43.389, -39.724, -65.526, -65.955,
-64.729, -62.323, -59.114, -55.363, -51.363, -47.309, -43.405, -39.739, -65.551,
-65.981, -64.754, -62.347, -59.137, -55.385, -51.382, -47.327, -43.421, -39.753,
 -65.577, -66.007, -64.780, -62.372, -59.160, -55.406, -51.402, -47.345,
-43.437, -39.768, -65.603, -66.033, -64.806, -62.396, -59.183, -55.427, -51.422,
-47.363, -43.453, -39.782, -65.629, -66.059, -64.831, -62.421, -59.206, -55.449,
-51.441, -47.380, -43.469, -39.797, -65.655, -66.085, -64.857, -62.446, -59.229,
-55.470, -51.461, -47.398, -43.485, -39.811, -65.680, -66.111, -64.882, -62.470,
-59.253, -55.492, -51.481, -47.416, -43.502, -39.826, -65.706, -66.137, -64.908,
-62.495, -59.276, -55.513, -51.500, -47.434, -43.518, -39.840, -65.732, -66.163,
-64.934, -62.519, -59.299, -55.535, -51.520, -47.452, -43.534, -39.855, -65.758,
-66.189, -64.959, -62.544, -59.322, -55.556, -51.540, -47.470, -43.550, -39.869,
 -65.784, -66.216, -64.985, -62.569, -59.345, -55.578, -51.559, -47.488,
-43.566, -39.884, -65.809, -66.242, -65.011, -62.593, -59.368, -55.599, -51.579,
-47.506, -43.582, -39.898, -65.835, -66.268, -65.036, -62.618, -59.392, -55.621,
-51.599, -47.523, -43.598, -39.913, -65.861, -66.294, -65.062, -62.642, -59.415,
-55.642, -51.618, -47.541, -43.615, -39.927, -65.887, -66.320, -65.088, -62.667,
-59.438, -55.664, -51.638, -47.559, -43.631, -39.942, -65.913, -66.346, -65.113,
-62.692, -59.461, -55.685, -51.658, -47.577, -43.647, -39.956, -65.938, -66.372,
-65.139, -62.716, -59.484, -55.707, -51.678, -47.595, -43.663, -39.971, -65.964,
-66.398, -65.164, -62.741, -59.507, -55.728, -51.697, -47.613, -43.679, -39.985,
 -65.990, -66.424, -65.190, -62.765, -59.531, -55.750, -51.717, -47.631,
-43.695, -40.000, -66.016, -66.450, -65.216, -62.790, -59.554, -55.771, -51.737,
-47.649, -43.711, -40.014, -66.042, -66.476, -65.241, -62.815, -59.577, -55.793,
-51.756, -47.667, -43.728, -40.029, -66.067, -66.502, -65.267, -62.839, -59.600,
-55.814, -51.776, -47.684, -43.744, -40.043, -66.093, -66.529, -65.293, -62.864,
-59.623, -55.836, -51.796, -47.702, -43.760, -40.058, -66.119, -66.555, -65.318,
-62.888, -59.646, -55.857, -51.815, -47.720, -43.776, -40.072, -66.145, -66.581,
-65.344, -62.913, -59.670, -55.879, -51.835, -47.738, -43.792, -40.087, -66.170,
-66.607, -65.369, -62.938, -59.693, -55.900, -51.855, -47.756, -43.808, -40.101,
 -66.196, -66.633, -65.395, -62.962, -59.716, -55.922, -51.874, -47.774,
-43.824, -40.116, -66.222, -66.659, -65.421, -62.987, -59.739, -55.943, -51.894,
-47.792, -43.841, -40.130, -66.248, -66.685, -65.446, -63.011, -59.762, -55.965,
-51.914, -47.810, -43.857, -40.145, -66.274, -66.711, -65.472, -63.036, -59.785,
-55.986, -51.933, -47.827, -43.873, -40.159, -66.299, -66.737, -65.498, -63.061,
-59.809, -56.007, -51.953, -47.845, -43.889, -40.174, -66.325, -66.763, -65.523,
-63.085, -59.832, -56.029, -51.973, -47.863, -43.905, -40.188, -66.351, -66.789,
-65.549, -63.110, -59.855, -56.050, -51.992, -47.881, -43.921, -40.203, -66.377,
-66.815, -65.574, -63.134, -59.878, -56.072, -52.012, -47.899, -43.937, -40.217,
 -66.403, -66.842, -65.600, -63.159, -59.901, -56.093, -52.032, -47.917,
-43.953, -40.232, -66.428, -66.868, -65.626, -63.184, -59.924, -56.115, -52.052,
-47.935, -43.970, -40.246, -66.454, -66.894, -65.651, -63.208, -59.947, -56.136,
-52.071, -47.953, -43.986, -40.261, -66.480, -66.920, -65.677, -63.233, -59.971,
-56.158, -52.091, -47.970, -44.002, -40.275, -66.506, -66.946, -65.703, -63.257,
-59.994, -56.179, -52.111, -47.988, -44.018, -40.290, -66.532, -66.972, -65.728,
-63.282, -60.017, -56.201, -52.130, -48.006, -44.034, -40.304, -66.557, -66.998,
-65.754, -63.307, -60.040, -56.222, -52.150, -48.024, -44.050, -40.319, -66.583,
-67.024, -65.780, -63.331, -60.063, -56.244, -52.170, -48.042, -44.066, -40.333,
 -66.609, -67.050, -65.805, -63.356, -60.086, -56.265, -52.189, -48.060,
-44.082, -40.348, -66.635, -67.076, -65.831, -63.380, -60.110, -56.287, -52.209,
-48.078, -44.099, -40.362, -66.660, -67.102, -65.856, -63.405, -60.133, -56.308,
-52.229, -48.096, -44.115, -40.377, -66.686, -67.128, -65.882, -63.430, -60.156,
-56.330, -52.248, -48.113, -44.131, -40.391, -66.712, -67.154, -65.908, -63.454,
-60.179, -56.351, -52.268, -48.131, -44.147, -40.406, -66.738, -67.181, -65.933,
-63.479, -60.202, -56.373, -52.288, -48.149, -44.163, -40.420, -66.764, -67.207,
-65.959, -63.503, -60.225, -56.394, -52.307, -48.167, -44.179, -40.435, -66.789,
-67.233, -65.985, -63.528, -60.249, -56.416, -52.327, -48.185, -44.195, -40.449,
 -66.815, -67.259, -66.010, -63.553, -60.272, -56.437, -52.347, -48.203,
-44.211, -40.464, -66.841, -67.285, -66.036, -63.577, -60.295, -56.458, -52.366,
-48.221, -44.228, -40.478, -66.867, -67.311, -66.061, -63.602, -60.318, -56.480,
-52.386, -48.238, -44.244, -40.493, -66.893, -67.337, -66.087, -63.626, -60.341,
-56.501, -52.406, -48.256, -44.260, -40.507, -66.918, -67.363, -66.113, -63.651,
-60.364, -56.523, -52.425, -48.274, -44.276, -40.522, -66.944, -67.389, -66.138,
-63.676, -60.387, -56.544, -52.445, -48.292, -44.292, -40.536, -66.970, -67.415,
-66.164, -63.700, -60.411, -56.566, -52.465, -48.310, -44.308, -40.551, -66.996,
-67.441, -66.190, -63.725, -60.434, -56.587, -52.484, -48.328, -44.324, -40.565,
 -67.022, -67.467, -66.215, -63.749, -60.457, -56.609, -52.504, -48.346,
-44.340, -40.580, -67.047, -67.494, -66.241, -63.774, -60.480, -56.630, -52.524,
-48.364, -44.357, -40.594, -67.073, -67.520, -66.266, -63.799, -60.503, -56.652,
-52.543, -48.381, -44.373, -40.609, -67.099, -67.546, -66.292, -63.823, -60.526,
-56.673, -52.563, -48.399, -44.389, -40.623, -67.125, -67.572, -66.318, -63.848,
-60.550, -56.695, -52.583, -48.417, -44.405, -40.638, -67.150, -67.598, -66.343,
-63.872, -60.573, -56.716, -52.602, -48.435, -44.421, -40.652, -67.176, -67.624,
-66.369, -63.897, -60.596, -56.738, -52.622, -48.453, -44.437, -40.667, -67.202,
-67.650, -66.395, -63.921, -60.619, -56.759, -52.642, -48.471, -44.453, -40.681,
 -67.228, -67.676, -66.420, -63.946, -60.642, -56.781, -52.661, -48.489,
-44.469, -40.696, -67.254, -67.702, -66.446, -63.971, -60.665, -56.802, -52.681,
-48.506, -44.486, -40.710, -67.279, -67.728, -66.471, -63.995, -60.689, -56.823,
-52.701, -48.524, -44.502, -40.725, -67.305, -67.754, -66.497, -64.020, -60.712,
-56.845, -52.720, -48.542, -44.518, -40.739, -67.331, -67.780, -66.523, -64.044,
-60.735, -56.866, -52.740, -48.560, -44.534, -40.754, -67.357, -67.806, -66.548,
-64.069, -60.758, -56.888, -52.760, -48.578, -44.550, -40.768, -67.383, -67.832,
-66.574, -64.094, -60.781, -56.909, -52.779, -48.596, -44.566, -40.783, -67.408,
-67.859, -66.599, -64.118, -60.804, -56.931, -52.799, -48.614, -44.582, -40.797,
 -67.434, -67.885, -66.625, -64.143, -60.827, -56.952, -52.819, -48.631,
-44.598, -40.812, -67.460, -67.911, -66.651, -64.167, -60.851, -56.974, -52.838,
-48.649, -44.614, -40.826, -67.486, -67.937, -66.676, -64.192, -60.874, -56.995,
-52.858, -48.667, -44.631, -40.841, -67.511, -67.963, -66.702, -64.217, -60.897,
-57.017, -52.878, -48.685, -44.647, -40.855, -67.537, -67.989, -66.728, -64.241,
-60.920, -57.038, -52.898, -48.703, -44.663, -40.870, -67.563, -68.015, -66.753,
-64.266, -60.943, -57.060, -52.917, -48.721, -44.679, -40.884, -67.589, -68.041,
-66.779, -64.290, -60.966, -57.081, -52.937, -48.739, -44.695, -40.899, -67.615,
-68.067, -66.804, -64.315, -60.989, -57.103, -52.957, -48.757, -44.711, -40.913,
 -67.640, -68.093, -66.830, -64.339, -61.013, -57.124, -52.976, -48.774,
-44.727, -40.927, -67.666, -68.119, -66.856, -64.364, -61.036, -57.145, -52.996,
-48.792, -44.743, -40.942, -67.692, -68.145, -66.881, -64.389, -61.059, -57.167,
-53.016, -48.810, -44.759, -40.956, -67.718, -68.171, -66.907, -64.413, -61.082,
-57.188, -53.035, -48.828, -44.776, -40.971, -67.743, -68.198, -66.933, -64.438,
-61.105, -57.210, -53.055, -48.846, -44.792, -40.985, -67.769, -68.224, -66.958,
-64.462, -61.128, -57.231, -53.075, -48.864, -44.808, -41.000, -67.795, -68.250,
-66.984, -64.487, -61.152, -57.253, -53.094, -48.882, -44.824, -41.014, -67.821,
-68.276, -67.009, -64.512, -61.175, -57.274, -53.114, -48.899, -44.840, -41.029,
 -67.847, -68.302, -67.035, -64.536, -61.198, -57.296, -53.134, -48.917,
-44.856, -41.043, -67.872, -68.328, -67.061, -64.561, -61.221, -57.317, -53.153,
-48.935, -44.872, -41.058, -67.898, -68.354, -67.086, -64.585, -61.244, -57.339,
-53.173, -48.953, -44.888, -41.072, -67.924, -68.380, -67.112, -64.610, -61.267,
-57.360, -53.192, -48.971, -44.904, -41.087, -67.950, -68.406, -67.137, -64.635,
-61.290, -57.382, -53.212, -48.989, -44.921, -41.101, -67.975, -68.432, -67.163,
-64.659, -61.314, -57.403, -53.232, -49.007, -44.937, -41.116, -68.001, -68.458,
-67.189, -64.684, -61.337, -57.424, -53.251, -49.024, -44.953, -41.130, -68.027,
-68.484, -67.214, -64.708, -61.360, -57.446, -53.271, -49.042, -44.969, -41.145,
 -68.053, -68.510, -67.240, -64.733, -61.383, -57.467, -53.291, -49.060,
-44.985, -41.159, -68.079, -68.536, -67.266, -64.757, -61.406, -57.489, -53.310,
-49.078, -45.001, -41.174, -68.104, -68.563, -67.291, -64.782, -61.429, -57.510,
-53.330, -49.096, -45.017, -41.188, -68.130, -68.589, -67.317, -64.807, -61.452,
-57.532, -53.350, -49.114, -45.033, -41.203, -68.156, -68.615, -67.342, -64.831,
-61.476, -57.553, -53.369, -49.131, -45.049, -41.217, -68.182, -68.641, -67.368,
-64.856, -61.499, -57.575, -53.389, -49.149, -45.066, -41.231, -68.208, -68.667,
-67.394, -64.880, -61.522, -57.596, -53.409, -49.167, -45.082, -41.246, -68.233,
-68.693, -67.419, -64.905, -61.545, -57.618, -53.428, -49.185, -45.098, -41.260,
 -68.259, -68.719, -67.445, -64.930, -61.568, -57.639, -53.448, -49.203,
-45.114, -41.275, -68.285, -68.745, -67.470, -64.954, -61.591, -57.661, -53.468,
-49.221, -45.130, -41.289, -68.311, -68.771, -67.496, -64.979, -61.614, -57.682,
-53.487, -49.239, -45.146, -41.304, -68.336, -68.797, -67.522, -65.003, -61.638,
-57.703, -53.507, -49.256, -45.162, -41.318, -68.362, -68.823, -67.547, -65.028,
-61.661, -57.725, -53.527, -49.274, -45.178, -41.333, -68.388, -68.849, -67.573,
-65.052, -61.684, -57.746, -53.546, -49.292, -45.194, -41.347, -68.414, -68.875,
-67.599, -65.077, -61.707, -57.768, -53.566, -49.310, -45.211, -41.362, -68.440,
-68.901, -67.624, -65.102, -61.730, -57.789, -53.586, -49.328, -45.227, -41.376,
 -68.465, -68.927, -67.650, -65.126, -61.753, -57.811, -53.605, -49.346,
-45.243, -41.391, -68.491, -68.954, -67.675, -65.151, -61.776, -57.832, -53.625,
-49.364, -45.259, -41.405, -68.517, -68.980, -67.701, -65.175, -61.800, -57.854,
-53.645, -49.381, -45.275, -41.420, -68.543, -69.006, -67.727, -65.200, -61.823,
-57.875, -53.664, -49.399, -45.291, -41.434, -68.568, -69.032, -67.752, -65.225,
-61.846, -57.897, -53.684, -49.417, -45.307, -41.448, -68.594, -69.058, -67.778,
-65.249, -61.869, -57.918, -53.704, -49.435, -45.323, -41.463, -68.620, -69.084,
-67.803, -65.274, -61.892, -57.939, -53.723, -49.453, -45.339, -41.477, -68.646,
-69.110, -67.829, -65.298, -61.915, -57.961, -53.743, -49.471, -45.355, -41.492,
 -68.671, -69.136, -67.855, -65.323, -61.938, -57.982, -53.763, -49.488,
-45.372, -41.506, -68.697, -69.162, -67.880, -65.347, -61.962, -58.004, -53.782,
-49.506, -45.388, -41.521, -68.723, -69.188, -67.906, -65.372, -61.985, -58.025,
-53.802, -49.524, -45.404, -41.535, -68.749, -69.214, -67.932, -65.397, -62.008,
-58.047, -53.822, -49.542, -45.420, -41.550, -68.775, -69.240, -67.957, -65.421,
-62.031, -58.068, -53.841, -49.560, -45.436, -41.564, -68.800, -69.266, -67.983,
-65.446, -62.054, -58.090, -53.861, -49.578, -45.452, -41.579, -68.826, -69.292,
-68.008, -65.470, -62.077, -58.111, -53.881, -49.596, -45.468, -41.593, -68.852,
-69.318, -68.034, -65.495, -62.100, -58.133, -53.900, -49.613, -45.484, -41.608,
 -68.878, -69.345, -68.060, -65.519, -62.124, -58.154, -53.920, -49.631,
-45.500, -41.622, -68.903, -69.371, -68.085, -65.544, -62.147, -58.175, -53.940,
-49.649, -45.516, -41.637, -68.929, -69.397, -68.111, -65.569, -62.170, -58.197,
-53.959, -49.667, -45.532, -41.651, -68.955, -69.423, -68.136, -65.593, -62.193,
-58.218, -53.979, -49.685, -45.549, -41.665, -68.981, -69.449, -68.162, -65.618,
-62.216, -58.240, -53.999, -49.703, -45.565, -41.680, -69.007, -69.475, -68.188,
-65.642, -62.239, -58.261, -54.018, -49.720, -45.581, -41.694, -69.032, -69.501,
-68.213, -65.667, -62.262, -58.283, -54.038, -49.738, -45.597, -41.709, -69.058,
-69.527, -68.239, -65.691, -62.286, -58.304, -54.057, -49.756, -45.613, -41.723,
 -69.084, -69.553, -68.264, -65.716, -62.309, -58.326, -54.077, -49.774,
-45.629, -41.738, -69.110, -69.579, -68.290, -65.741, -62.332, -58.347, -54.097,
-49.792, -45.645, -41.752, -69.135, -69.605, -68.316, -65.765, -62.355, -58.369,
-54.116, -49.810, -45.661, -41.767, -69.161, -69.631, -68.341, -65.790, -62.378,
-58.390, -54.136, -49.828, -45.677, -41.781, -69.187, -69.657, -68.367, -65.814,
-62.401, -58.411, -54.156, -49.845, -45.693, -41.796, -69.213, -69.683, -68.393,
-65.839, -62.424, -58.433, -54.175, -49.863, -45.710, -41.810, -69.239, -69.709,
-68.418, -65.864, -62.447, -58.454, -54.195, -49.881, -45.726, -41.824, -69.264,
-69.735, -68.444, -65.888, -62.471, -58.476, -54.215, -49.899, -45.742, -41.839,
 -69.290, -69.762, -68.469, -65.913, -62.494, -58.497, -54.234, -49.917,
-45.758, -41.853, -69.316, -69.788, -68.495, -65.937, -62.517, -58.519, -54.254,
-49.935, -45.774, -41.868, -69.342, -69.814, -68.521, -65.962, -62.540, -58.540,
-54.274, -49.952, -45.790, -41.882, -69.367, -69.840, -68.546, -65.986, -62.563,
-58.562, -54.293, -49.970, -45.806, -41.897, -69.393, -69.866, -68.572, -66.011,
-62.586, -58.583, -54.313, -49.988, -45.822, -41.911, -69.419, -69.892, -68.597,
-66.036, -62.609, -58.605, -54.333, -50.006, -45.838, -41.926, -69.445, -69.918,
-68.623, -66.060, -62.633, -58.626, -54.352, -50.024, -45.854, -41.940, -69.470,
-69.944, -68.649, -66.085, -62.656, -58.647, -54.372, -50.042, -45.870, -41.955,
 -69.496, -69.970, -68.674, -66.109, -62.679, -58.669, -54.392, -50.059,
-45.887, -41.969, -69.522, -69.996, -68.700, -66.134, -62.702, -58.690, -54.411,
-50.077, -45.903, -41.983, -69.548, -70.022, -68.725, -66.158, -62.725, -58.712,
-54.431, -50.095, -45.919, -41.998, -69.574, -70.048, -68.751, -66.183, -62.748,
-58.733, -54.450, -50.113, -45.935, -42.012, -69.599, -70.074, -68.777, -66.208,
-62.771, -58.755, -54.470, -50.131, -45.951, -42.027, -69.625, -70.100, -68.802,
-66.232, -62.795, -58.776, -54.490, -50.149, -45.967, -42.041, -69.651, -70.126,
-68.828, -66.257, -62.818, -58.798, -54.509, -50.166, -45.983, -42.056, -69.677,
-70.152, -68.853, -66.281, -62.841, -58.819, -54.529, -50.184, -45.999, -42.070,
 -69.702, -70.179, -68.879, -66.306, -62.864, -58.840, -54.549, -50.202,
-46.015, -42.085, -69.728, -70.205, -68.905, -66.330, -62.887, -58.862, -54.568,
-50.220, -46.031, -42.099, -69.754, -70.231, -68.930, -66.355, -62.910, -58.883,
-54.588, -50.238, -46.047, -42.114, -69.780, -70.257, -68.956, -66.380, -62.933,
-58.905, -54.608, -50.256, -46.064, -42.128, -69.805, -70.283, -68.981, -66.404,
-62.956, -58.926, -54.627, -50.273, -46.080, -42.142, -69.831, -70.309, -69.007,
-66.429, -62.980, -58.948, -54.647, -50.291, -46.096, -42.157, -69.857, -70.335,
-69.033, -66.453, -63.003, -58.969, -54.667, -50.309, -46.112, -42.171, -69.883,
-70.361, -69.058, -66.478, -63.026, -58.991, -54.686, -50.327, -46.128, -42.186,
 -69.909, -70.387, -69.084, -66.502, -63.049, -59.012, -54.706, -50.345,
-46.144, -42.200, -69.934, -70.413, -69.109, -66.527, -63.072, -59.033, -54.726,
-50.363, -46.160, -42.215, -69.960, -70.439, -69.135, -66.552, -63.095, -59.055,
-54.745, -50.380, -46.176, -42.229, -69.986, -70.465, -69.161, -66.576, -63.118,
-59.076, -54.765, -50.398, -46.192, -42.244, -70.012, -70.491, -69.186, -66.601,
-63.141, -59.098, -54.784, -50.416, -46.208, -42.258, -70.037, -70.517, -69.212,
-66.625, -63.165, -59.119, -54.804, -50.434, -46.224, -42.272, -70.063, -70.543,
-69.237, -66.650, -63.188, -59.141, -54.824, -50.452, -46.240, -42.287, -70.089,
-70.569, -69.263, -66.674, -63.211, -59.162, -54.843, -50.470, -46.257, -42.301,
 -70.115, -70.595, -69.289, -66.699, -63.234, -59.184, -54.863, -50.487,
-46.273, -42.316, -70.140, -70.622, -69.314, -66.724, -63.257, -59.205, -54.883,
-50.505, -46.289, -42.330, -70.166, -70.648, -69.340, -66.748, -63.280, -59.226,
-54.902, -50.523, -46.305, -42.345, -70.192, -70.674, -69.366, -66.773, -63.303,
-59.248, -54.922, -50.541, -46.321, -42.359, -70.218, -70.700, -69.391, -66.797,
-63.327, -59.269, -54.942, -50.559, -46.337, -42.374, -70.244, -70.726, -69.417,
-66.822, -63.350, -59.291, -54.961, -50.577, -46.353, -42.388, -70.269, -70.752,
-69.442, -66.846, -63.373, -59.312, -54.981, -50.594, -46.369, -42.402, -70.295,
-70.778, -69.468, -66.871, -63.396, -59.334, -55.001, -50.612, -46.385, -42.417,
 -70.321, -70.804, -69.494, -66.896, -63.419, -59.355, -55.020, -50.630,
-46.401, -42.431, -70.347, -70.830, -69.519, -66.920, -63.442, -59.377, -55.040,
-50.648, -46.417, -42.446, -70.372, -70.856, -69.545, -66.945, -63.465, -59.398,
-55.059, -50.666, -46.433, -42.460, -70.398, -70.882, -69.570, -66.969, -63.488,
-59.419, -55.079, -50.684, -46.449, -42.475, -70.424, -70.908, -69.596, -66.994,
-63.512, -59.441, -55.099, -50.701, -46.466, -42.489, -70.450, -70.934, -69.622,
-67.018, -63.535, -59.462, -55.118, -50.719, -46.482, -42.504, -70.475, -70.960,
-69.647, -67.043, -63.558, -59.484, -55.138, -50.737, -46.498, -42.518, -70.501,
-70.986, -69.673, -67.068, -63.581, -59.505, -55.158, -50.755, -46.514, -42.532,
 -70.527, -71.012, -69.698, -67.092, -63.604, -59.527, -55.177, -50.773,
-46.530, -42.547, -70.553, -71.038, -69.724, -67.117, -63.627, -59.548, -55.197,
-50.791, -46.546, -42.561, -70.578, -71.064, -69.750, -67.141, -63.650, -59.570,
-55.217, -50.808, -46.562, -42.576, -70.604, -71.091, -69.775, -67.166, -63.673,
-59.591, -55.236, -50.826, -46.578, -42.590, -70.630, -71.117, -69.801, -67.190,
-63.697, -59.612, -55.256, -50.844, -46.594, -42.605, -70.656, -71.143, -69.826,
-67.215, -63.720, -59.634, -55.275, -50.862, -46.610, -42.619, -70.682, -71.169,
-69.852, -67.239, -63.743, -59.655, -55.295, -50.880, -46.626, -42.634, -70.707,
-71.195, -69.878, -67.264, -63.766, -59.677, -55.315, -50.898, -46.642, -42.648,
 -70.733, -71.221, -69.903, -67.289, -63.789, -59.698, -55.334, -50.915,
-46.658, -42.662, -70.759, -71.247, -69.929, -67.313, -63.812, -59.720, -55.354,
-50.933, -46.675, -42.677, -70.785, -71.273, -69.954, -67.338, -63.835, -59.741,
-55.374, -50.951, -46.691, -42.691, -70.810, -71.299, -69.980, -67.362, -63.858,
-59.762, -55.393, -50.969, -46.707, -42.706, -70.836, -71.325, -70.006, -67.387,
-63.882, -59.784, -55.413, -50.987, -46.723, -42.720, -70.862, -71.351, -70.031,
-67.411, -63.905, -59.805, -55.433, -51.004, -46.739, -42.735, -70.888, -71.377,
-70.057, -67.436, -63.928, -59.827, -55.452, -51.022, -46.755, -42.749, -70.913,
-71.403, -70.082, -67.461, -63.951, -59.848, -55.472, -51.040, -46.771, -42.763,
 -70.939, -71.429, -70.108, -67.485, -63.974, -59.870, -55.491, -51.058,
-46.787, -42.778, -70.965, -71.455, -70.133, -67.510, -63.997, -59.891, -55.511,
-51.076, -46.803, -42.792, -70.991, -71.481, -70.159, -67.534, -64.020, -59.913,
-55.531, -51.094, -46.819, -42.807, -71.016, -71.507, -70.185, -67.559, -64.043,
-59.934, -55.550, -51.111, -46.835, -42.821, -71.042, -71.533, -70.210, -67.583,
-64.067, -59.955, -55.570, -51.129, -46.851, -42.836, -71.068, -71.559, -70.236,
-67.608, -64.090, -59.977, -55.590, -51.147, -46.867, -42.850, -71.094, -71.586,
-70.261, -67.632, -64.113, -59.998, -55.609, -51.165, -46.884, -42.864, -71.120,
-71.612, -70.287, -67.657, -64.136, -60.020, -55.629, -51.183, -46.900, -42.879,
 -71.145, -71.638, -70.313, -67.682, -64.159, -60.041, -55.649, -51.201,
-46.916, -42.893, -71.171, -71.664, -70.338, -67.706, -64.182, -60.063, -55.668,
-51.218, -46.932, -42.908, -71.197, -71.690, -70.364, -67.731, -64.205, -60.084,
-55.688, -51.236, -46.948, -42.922, -71.223, -71.716, -70.389, -67.755, -64.228,
-60.105, -55.707, -51.254, -46.964, -42.937, -71.248, -71.742, -70.415, -67.780,
-64.252, -60.127, -55.727, -51.272, -46.980, -42.951, -71.274, -71.768, -70.441,
-67.804, -64.275, -60.148, -55.747, -51.290, -46.996, -42.966, -71.300, -71.794,
-70.466, -67.829, -64.298, -60.170, -55.766, -51.307, -47.012, -42.980, -71.326,
-71.820, -70.492, -67.854, -64.321, -60.191, -55.786, -51.325, -47.028, -42.994,
 -71.351, -71.846, -70.517, -67.878, -64.344, -60.213, -55.806, -51.343,
-47.044, -43.009, -71.377, -71.872, -70.543, -67.903, -64.367, -60.234, -55.825,
-51.361, -47.060, -43.023, -71.403, -71.898, -70.569, -67.927, -64.390, -60.255,
-55.845, -51.379, -47.076, -43.038, -71.429, -71.924, -70.594, -67.952, -64.413,
-60.277, -55.864, -51.397, -47.092, -43.052, -71.454, -71.950, -70.620, -67.976,
-64.436, -60.298, -55.884, -51.414, -47.108, -43.067, -71.480, -71.976, -70.645,
-68.001, -64.460, -60.320, -55.904, -51.432, -47.125, -43.081, -71.506, -72.002,
-70.671, -68.025, -64.483, -60.341, -55.923, -51.450, -47.141, -43.095, -71.532,
-72.028, -70.697, -68.050, -64.506, -60.363, -55.943, -51.468, -47.157, -43.110,
 -71.557, -72.054, -70.722, -68.075, -64.529, -60.384, -55.963, -51.486,
-47.173, -43.124, -71.583, -72.080, -70.748, -68.099, -64.552, -60.405, -55.982,
-51.503, -47.189, -43.139, -71.609, -72.107, -70.773, -68.124, -64.575, -60.427,
-56.002, -51.521, -47.205, -43.153, -71.635, -72.133, -70.799, -68.148, -64.598,
-60.448, -56.022, -51.539, -47.221, -43.167, -71.660, -72.159, -70.825, -68.173,
-64.621, -60.470, -56.041, -51.557, -47.237, -43.182, -71.686, -72.185, -70.850,
-68.197, -64.645, -60.491, -56.061, -51.575, -47.253, -43.196, -71.712, -72.211,
-70.876, -68.222, -64.668, -60.513, -56.080, -51.593, -47.269, -43.211, -71.738,
-72.237, -70.901, -68.246, -64.691, -60.534, -56.100, -51.610, -47.285, -43.225,
 -71.764, -72.263, -70.927, -68.271, -64.714, -60.555, -56.120, -51.628,
-47.301, -43.240, -71.789, -72.289, -70.953, -68.296, -64.737, -60.577, -56.139,
-51.646, -47.317, -43.254, -71.815, -72.315, -70.978, -68.320, -64.760, -60.598,
-56.159, -51.664, -47.333, -43.268, -71.841, -72.341, -71.004, -68.345, -64.783,
-60.620, -56.179, -51.682, -47.349, -43.283, -71.867, -72.367, -71.029, -68.369,
-64.806, -60.641, -56.198, -51.699, -47.366, -43.297, -71.892, -72.393, -71.055,
-68.394, -64.830, -60.663, -56.218, -51.717, -47.382, -43.312, -71.918, -72.419,
-71.080, -68.418, -64.853, -60.684, -56.237, -51.735, -47.398, -43.326, -71.944,
-72.445, -71.106, -68.443, -64.876, -60.706, -56.257, -51.753, -47.414, -43.341,
 -71.970, -72.471, -71.132, -68.468, -64.899, -60.727, -56.277, -51.771,
-47.430, -43.355, -71.995, -72.497, -71.157, -68.492, -64.922, -60.748, -56.296,
-51.788, -47.446, -43.369, -72.021, -72.523, -71.183, -68.517, -64.945, -60.770,
-56.316, -51.806, -47.462, -43.384, -72.047, -72.549, -71.208, -68.541, -64.968,
-60.791, -56.336, -51.824, -47.478, -43.398, -72.073, -72.575, -71.234, -68.566,
-64.991, -60.813, -56.355, -51.842, -47.494, -43.413, -72.098, -72.601, -71.260,
-68.590, -65.014, -60.834, -56.375, -51.860, -47.510, -43.427, -72.124, -72.627,
-71.285, -68.615, -65.038, -60.855, -56.394, -51.878, -47.526, -43.442, -72.150,
-72.654, -71.311, -68.639, -65.061, -60.877, -56.414, -51.895, -47.542, -43.456,
 -72.176, -72.680, -71.336, -68.664, -65.084, -60.898, -56.434, -51.913,
-47.558, -43.470, -72.201, -72.706, -71.362, -68.689, -65.107, -60.920, -56.453,
-51.931, -47.574, -43.485, -72.227, -72.732, -71.388, -68.713, -65.130, -60.941,
-56.473, -51.949, -47.590, -43.499, -72.253, -72.758, -71.413, -68.738, -65.153,
-60.963, -56.493, -51.967, -47.606, -43.514, -72.279, -72.784, -71.439, -68.762,
-65.176, -60.984, -56.512, -51.984, -47.622, -43.528, -72.304, -72.810, -71.464,
-68.787, -65.199, -61.005, -56.532, -52.002, -47.639, -43.542, -72.330, -72.836,
-71.490, -68.811, -65.222, -61.027, -56.551, -52.020, -47.655, -43.557, -72.356,
-72.862, -71.515, -68.836, -65.246, -61.048, -56.571, -52.038, -47.671, -43.571,
 -72.382, -72.888, -71.541, -68.860, -65.269, -61.070, -56.591, -52.056,
-47.687, -43.586, -72.407, -72.914, -71.567, -68.885, -65.292, -61.091, -56.610,
-52.073, -47.703, -43.600, -72.433, -72.940, -71.592, -68.909, -65.315, -61.113,
-56.630, -52.091, -47.719, -43.615, -72.459, -72.966, -71.618, -68.934, -65.338,
-61.134, -56.650, -52.109, -47.735, -43.629, -72.485, -72.992, -71.643, -68.959,
-65.361, -61.155, -56.669, -52.127, -47.751, -43.643, -72.510, -73.018, -71.669,
-68.983, -65.384, -61.177, -56.689, -52.145, -47.767, -43.658, -72.536, -73.044,
-71.695, -69.008, -65.407, -61.198, -56.708, -52.163, -47.783, -43.672, -72.562,
-73.070, -71.720, -69.032, -65.430, -61.220, -56.728, -52.180, -47.799, -43.687,
 -72.588, -73.096, -71.746, -69.057, -65.454, -61.241, -56.748, -52.198,
-47.815, -43.701, -72.613, -73.122, -71.771, -69.081, -65.477, -61.263, -56.767,
-52.216, -47.831, -43.715, -72.639, -73.148, -71.797, -69.106, -65.500, -61.284,
-56.787, -52.234, -47.847, -43.730, -72.665, -73.174, -71.823, -69.130, -65.523,
-61.305, -56.807, -52.252, -47.863, -43.744, -72.691, -73.200, -71.848, -69.155,
-65.546, -61.327, -56.826, -52.269, -47.879, -43.759, -72.716, -73.226, -71.874,
-69.180, -65.569, -61.348, -56.846, -52.287, -47.895, -43.773, -72.742, -73.253,
-71.899, -69.204, -65.592, -61.370, -56.865, -52.305, -47.911, -43.788, -72.768,
-73.279, -71.925, -69.229, -65.615, -61.391, -56.885, -52.323, -47.928, -43.802,
 -72.794, -73.305, -71.950, -69.253, -65.638, -61.413, -56.905, -52.341,
-47.944, -43.816, -72.820, -73.331, -71.976, -69.278, -65.662, -61.434, -56.924,
-52.358, -47.960, -43.831, -72.845, -73.357, -72.002, -69.302, -65.685, -61.455,
-56.944, -52.376, -47.976, -43.845, -72.871, -73.383, -72.027, -69.327, -65.708,
-61.477, -56.963, -52.394, -47.992, -43.860, -72.897, -73.409, -72.053, -69.351,
-65.731, -61.498, -56.983, -52.412, -48.008, -43.874, -72.923, -73.435, -72.078,
-69.376, -65.754, -61.520, -57.003, -52.430, -48.024, -43.888, -72.948, -73.461,
-72.104, -69.401, -65.777, -61.541, -57.022, -52.447, -48.040, -43.903, -72.974,
-73.487, -72.130, -69.425, -65.800, -61.562, -57.042, -52.465, -48.056, -43.917,
 -73.000, -73.513, -72.155, -69.450, -65.823, -61.584, -57.062, -52.483,
-48.072, -43.932, -73.026, -73.539, -72.181, -69.474, -65.846, -61.605, -57.081,
-52.501, -48.088, -43.946, -73.051, -73.565, -72.206, -69.499, -65.870, -61.627,
-57.101, -52.519, -48.104, -43.960, -73.077, -73.591, -72.232, -69.523, -65.893,
-61.648, -57.120, -52.536, -48.120, -43.975,
};

const double CIA_Orton::CH4CH4[][10] = {
 -24.226, -25.228, -26.130, -26.936, -27.647, -28.264, -28.786, -29.211,
-29.534, -29.750, -22.393, -23.395, -24.298, -25.103, -25.814, -26.432, -26.954,
-27.378, -27.701, -27.918, -21.007, -22.009, -22.911, -23.717, -24.428, -25.045,
-25.567, -25.992, -26.315, -26.531, -19.621, -20.623, -21.525, -22.331, -23.042,
-23.659, -24.181, -24.606, -24.929, -25.145, -18.809, -19.812, -20.714, -21.519,
-22.230, -22.848, -23.370, -23.795, -24.118, -24.334, -18.233, -19.236, -20.138,
-20.944, -21.655, -22.272, -22.794, -23.219, -23.542, -23.759, -17.786, -18.789,
-19.691, -20.497, -21.208, -21.826, -22.348, -22.773, -23.096, -23.312, -17.420,
-18.423, -19.326, -20.132, -20.843, -21.461, -21.983, -22.408, -22.731, -22.948,
 -17.110, -18.114, -19.017, -19.823, -20.535, -21.152, -21.675, -22.100,
-22.423, -22.639, -16.841, -17.845, -18.749, -19.555, -20.267, -20.885, -21.407,
-21.833, -22.156, -22.372, -16.604, -17.608, -18.512, -19.319, -20.031, -20.649,
-21.172, -21.597, -21.920, -22.137, -16.391, -17.396, -18.300, -19.107, -19.820,
-20.438, -20.961, -21.386, -21.709, -21.926, -16.198, -17.203, -18.108, -18.916,
-19.628, -20.247, -20.770, -21.195, -21.519, -21.735, -16.021, -17.027, -17.933,
-18.741, -19.454, -20.072, -20.595, -21.021, -21.345, -21.561, -15.858, -16.865,
-17.771, -18.580, -19.293, -19.912, -20.435, -20.860, -21.184, -21.401, -15.706,
-16.714, -17.621, -18.430, -19.144, -19.763, -20.286, -20.712, -21.036, -21.253,
 -15.565, -16.574, -17.481, -18.291, -19.005, -19.624, -20.148, -20.574,
-20.898, -21.115, -15.432, -16.442, -17.350, -18.160, -18.875, -19.494, -20.018,
-20.444, -20.769, -20.985, -15.307, -16.318, -17.227, -18.038, -18.752, -19.372,
-19.897, -20.323, -20.647, -20.864, -15.189, -16.201, -17.110, -17.922, -18.637,
-19.257, -19.782, -20.208, -20.533, -20.750, -15.077, -16.089, -17.000, -17.812,
-18.528, -19.148, -19.673, -20.100, -20.424, -20.641, -14.970, -15.984, -16.895,
-17.708, -18.424, -19.045, -19.570, -19.997, -20.321, -20.539, -14.502, -15.520,
-16.436, -17.252, -17.971, -18.594, -19.120, -19.548, -19.874, -20.092, -14.116,
-15.137, -16.057, -16.877, -17.598, -18.224, -18.752, -19.181, -19.508, -19.726,
 -13.786, -14.811, -15.734, -16.557, -17.282, -17.910, -18.440, -18.870,
-19.198, -19.417, -13.501, -14.527, -15.452, -16.278, -17.006, -17.636, -18.168,
-18.600, -18.929, -19.148, -13.250, -14.276, -15.203, -16.031, -16.761, -17.394,
-17.928, -18.362, -18.691, -18.912, -13.027, -14.052, -14.979, -15.809, -16.542,
-17.176, -17.712, -18.148, -18.478, -18.699, -12.828, -13.850, -14.777, -15.608,
-16.342, -16.979, -17.517, -17.953, -18.285, -18.507, -12.650, -13.668, -14.594,
-15.425, -16.160, -16.798, -17.338, -17.776, -18.109, -18.332, -12.489, -13.502,
-14.425, -15.256, -15.992, -16.632, -17.173, -17.612, -17.946, -18.170, -12.342,
-13.350, -14.271, -15.101, -15.837, -16.477, -17.020, -17.461, -17.796, -18.020,
 -12.210, -13.211, -14.128, -14.956, -15.692, -16.334, -16.877, -17.319,
-17.655, -17.880, -12.088, -13.083, -13.996, -14.822, -15.558, -16.199, -16.744,
-17.187, -17.524, -17.750, -11.976, -12.965, -13.874, -14.697, -15.432, -16.073,
-16.618, -17.062, -17.400, -17.627, -11.873, -12.855, -13.760, -14.580, -15.313,
-15.955, -16.500, -16.945, -17.283, -17.511, -11.777, -12.753, -13.653, -14.471,
-15.202, -15.843, -16.388, -16.834, -17.173, -17.401, -11.688, -12.658, -13.553,
-14.367, -15.097, -15.737, -16.283, -16.728, -17.068, -17.297, -11.524, -12.484,
-13.371, -14.178, -14.904, -15.542, -16.087, -16.533, -16.874, -17.104, -11.378,
-12.330, -13.208, -14.009, -14.730, -15.366, -15.910, -16.356, -16.697, -16.927,
 -11.244, -12.191, -13.062, -13.857, -14.573, -15.205, -15.748, -16.193,
-16.534, -16.765, -11.123, -12.065, -12.930, -13.719, -14.429, -15.059, -15.599,
-16.043, -16.385, -16.615, -11.012, -11.951, -12.809, -13.592, -14.298, -14.924,
-15.462, -15.905, -16.246, -16.477, -10.912, -11.846, -12.699, -13.475, -14.177,
-14.799, -15.335, -15.777, -16.117, -16.347, -10.822, -11.750, -12.596, -13.367,
-14.064, -14.684, -15.217, -15.657, -15.996, -16.226, -10.738, -11.660, -12.500,
-13.266, -13.959, -14.576, -15.107, -15.546, -15.883, -16.113, -10.659, -11.573,
-12.408, -13.171, -13.861, -14.475, -15.004, -15.441, -15.777, -16.006, -10.579,
-11.488, -12.320, -13.080, -13.768, -14.380, -14.907, -15.342, -15.677, -15.905,
 -10.494, -11.402, -12.233, -12.992, -13.679, -14.290, -14.816, -15.250,
-15.583, -15.810, -10.403, -11.315, -12.148, -12.908, -13.595, -14.205, -14.730,
-15.162, -15.494, -15.720, -10.308, -11.227, -12.064, -12.826, -13.514, -14.124,
-14.648, -15.079, -15.410, -15.634, -10.213, -11.140, -11.982, -12.747, -13.437,
-14.047, -14.571, -15.000, -15.330, -15.553, -10.123, -11.056, -11.903, -12.671,
-13.363, -13.974, -14.497, -14.925, -15.253, -15.475, -10.043, -10.978, -11.827,
-12.597, -13.291, -13.903, -14.426, -14.854, -15.181, -15.401, -9.972, -10.905,
-11.754, -12.527, -13.223, -13.836, -14.359, -14.786, -15.111, -15.331, -9.912,
-10.838, -11.685, -12.459, -13.156, -13.771, -14.294, -14.720, -15.045, -15.263,
  -9.858, -10.774, -11.618, -12.393, -13.093, -13.709, -14.232, -14.658,
-14.981, -15.198, -9.806, -10.711, -11.552, -12.328, -13.031, -13.648, -14.172,
-14.598, -14.920, -15.136, -9.748, -10.646, -11.487, -12.265, -12.970, -13.590,
-14.115, -14.540, -14.861, -15.076, -9.682, -10.578, -11.421, -12.203, -12.912,
-13.533, -14.059, -14.484, -14.805, -15.019, -9.607, -10.506, -11.355, -12.142,
-12.855, -13.479, -14.005, -14.430, -14.750, -14.963, -9.529, -10.434, -11.290,
-12.083, -12.800, -13.426, -13.953, -14.378, -14.697, -14.910, -9.453, -10.365,
-11.228, -12.027, -12.747, -13.374, -13.902, -14.327, -14.646, -14.858, -9.384,
-10.302, -11.170, -11.973, -12.696, -13.325, -13.853, -14.278, -14.596, -14.807,
  -9.326, -10.247, -11.118, -11.923, -12.646, -13.276, -13.805, -14.229,
-14.548, -14.759, -9.280, -10.200, -11.071, -11.876, -12.599, -13.229, -13.758,
-14.183, -14.501, -14.711, -9.242, -10.159, -11.027, -11.830, -12.553, -13.183,
-13.712, -14.137, -14.455, -14.665, -9.209, -10.121, -10.985, -11.786, -12.507,
-13.137, -13.666, -14.091, -14.410, -14.620, -9.175, -10.081, -10.941, -11.740,
-12.461, -13.091, -13.621, -14.047, -14.366, -14.577, -9.133, -10.035, -10.894,
-11.692, -12.414, -13.045, -13.576, -14.003, -14.323, -14.534, -9.081,  -9.983,
-10.842, -11.643, -12.366, -12.999, -13.532, -13.960, -14.281, -14.492, -9.021,
-9.925, -10.787, -11.591, -12.317, -12.953, -13.488, -13.918, -14.240, -14.451,
  -8.958,  -9.865, -10.732, -11.539, -12.269, -12.907, -13.444, -13.876,
-14.199, -14.411, -8.899,  -9.808, -10.677, -11.487, -12.220, -12.861, -13.401,
-13.834, -14.159, -14.372, -8.848,  -9.757, -10.627, -11.439, -12.174, -12.817,
-13.358, -13.793, -14.119, -14.333, -8.804,  -9.712, -10.581, -11.393, -12.129,
-12.773, -13.316, -13.753, -14.080, -14.295, -8.769,  -9.675, -10.540, -11.350,
-12.086, -12.731, -13.275, -13.714, -14.042, -14.258, -8.739,  -9.642, -10.504,
-11.311, -12.045, -12.690, -13.235, -13.675, -14.005, -14.222, -8.711,  -9.612,
-10.471, -11.274, -12.006, -12.650, -13.195, -13.636, -13.968, -14.186, -8.679,
-9.583, -10.439, -11.239, -11.968, -12.611, -13.157, -13.599, -13.931, -14.150,
  -8.641,  -9.551, -10.408, -11.206, -11.932, -12.574, -13.119, -13.562,
-13.895, -14.116, -8.597,  -9.517, -10.377, -11.173, -11.897, -12.537, -13.082,
-13.526, -13.860, -14.081, -8.551,  -9.484, -10.347, -11.142, -11.864, -12.502,
-13.047, -13.490, -13.825, -14.048, -8.509,  -9.453, -10.319, -11.113, -11.832,
-12.468, -13.012, -13.455, -13.791, -14.015, -8.475,  -9.428, -10.295, -11.087,
-11.802, -12.436, -12.978, -13.421, -13.758, -13.982, -8.454,  -9.410, -10.275,
-11.063, -11.775, -12.405, -12.945, -13.388, -13.725, -13.950, -8.445,  -9.399,
-10.260, -11.042, -11.749, -12.375, -12.913, -13.355, -13.692, -13.918, -8.447,
-9.394, -10.248, -11.024, -11.724, -12.346, -12.882, -13.323, -13.660, -13.887,
  -8.454,  -9.392, -10.238, -11.006, -11.701, -12.319, -12.851, -13.291,
-13.629, -13.856, -8.457,  -9.388, -10.227, -10.989, -11.678, -12.292, -12.822,
-13.260, -13.598, -13.825, -8.448,  -9.377, -10.213, -10.971, -11.656, -12.265,
-12.793, -13.230, -13.567, -13.795, -8.422,  -9.357, -10.196, -10.952, -11.633,
-12.239, -12.764, -13.200, -13.537, -13.765, -8.383,  -9.331, -10.175, -10.932,
-11.611, -12.214, -12.737, -13.171, -13.507, -13.736, -8.339,  -9.302, -10.154,
-10.912, -11.589, -12.189, -12.709, -13.142, -13.478, -13.707, -8.300,  -9.275,
-10.133, -10.893, -11.568, -12.165, -12.683, -13.114, -13.449, -13.678, -8.270,
-9.252, -10.115, -10.874, -11.547, -12.142, -12.656, -13.086, -13.421, -13.650,
  -8.254,  -9.237, -10.100, -10.858, -11.528, -12.119, -12.631, -13.059,
-13.392, -13.622, -8.251,  -9.230, -10.090, -10.844, -11.509, -12.096, -12.606,
-13.032, -13.365, -13.594, -8.258,  -9.228, -10.081, -10.830, -11.491, -12.074,
-12.581, -13.005, -13.337, -13.567, -8.271,  -9.229, -10.074, -10.817, -11.473,
-12.053, -12.556, -12.978, -13.310, -13.539, -8.284,  -9.230, -10.066, -10.803,
-11.455, -12.031, -12.532, -12.952, -13.283, -13.512, -8.293,  -9.227, -10.056,
-10.788, -11.436, -12.009, -12.507, -12.926, -13.256, -13.486, -8.295,  -9.219,
-10.042, -10.771, -11.416, -11.987, -12.483, -12.901, -13.230, -13.459, -8.292,
-9.207, -10.026, -10.752, -11.396, -11.965, -12.459, -12.875, -13.204, -13.433,
  -8.285,  -9.192, -10.008, -10.733, -11.375, -11.942, -12.435, -12.850,
-13.178, -13.407, -8.278,  -9.177,  -9.990, -10.713, -11.355, -11.921, -12.412,
-12.826, -13.152, -13.381, -8.271,  -9.163,  -9.972, -10.695, -11.335, -11.899,
-12.389, -12.801, -13.127, -13.356, -8.267,  -9.150,  -9.956, -10.677, -11.316,
-11.879, -12.367, -12.777, -13.102, -13.331, -8.263,  -9.139,  -9.942, -10.662,
-11.299, -11.860, -12.345, -12.754, -13.077, -13.306, -8.261,  -9.130,  -9.930,
-10.648, -11.283, -11.841, -12.324, -12.731, -13.053, -13.281, -8.258,  -9.123,
-9.920, -10.636, -11.269, -11.824, -12.304, -12.708, -13.029, -13.257, -8.255,
-9.117,  -9.912, -10.626, -11.256, -11.808, -12.285, -12.686, -13.006, -13.233,
  -8.252,  -9.114,  -9.908, -10.619, -11.245, -11.792, -12.266, -12.665,
-12.983, -13.209, -8.251,  -9.113,  -9.906, -10.613, -11.235, -11.778, -12.247,
-12.644, -12.960, -13.185, -8.252,  -9.117,  -9.907, -10.611, -11.227, -11.764,
-12.230, -12.623, -12.938, -13.162, -8.258,  -9.124,  -9.912, -10.610, -11.219,
-11.751, -12.213, -12.603, -12.916, -13.139, -8.270,  -9.138,  -9.921, -10.611,
-11.213, -11.739, -12.196, -12.583, -12.894, -13.117, -8.291,  -9.157,  -9.932,
-10.614, -11.208, -11.728, -12.180, -12.564, -12.873, -13.095, -8.320,  -9.180,
-9.946, -10.617, -11.203, -11.716, -12.164, -12.545, -12.852, -13.073, -8.353,
-9.204,  -9.958, -10.619, -11.197, -11.704, -12.148, -12.526, -12.831, -13.051,
  -8.383,  -9.224,  -9.967, -10.618, -11.189, -11.691, -12.132, -12.508,
-12.811, -13.030, -8.399,  -9.232,  -9.968, -10.612, -11.179, -11.678, -12.116,
-12.489, -12.791, -13.008, -8.393,  -9.226,  -9.958, -10.601, -11.166, -11.664,
-12.099, -12.471, -12.771, -12.988, -8.364,  -9.203,  -9.939, -10.584, -11.150,
-11.648, -12.083, -12.453, -12.752, -12.967, -8.322,  -9.170,  -9.913, -10.563,
-11.132, -11.631, -12.066, -12.436, -12.733, -12.947, -8.277,  -9.132,  -9.881,
-10.538, -11.112, -11.614, -12.050, -12.418, -12.714, -12.927, -8.238,  -9.095,
-9.849, -10.512, -11.091, -11.597, -12.033, -12.401, -12.695, -12.907, -8.208,
-9.062,  -9.818, -10.486, -11.071, -11.580, -12.017, -12.384, -12.677, -12.888,
  -8.189,  -9.036,  -9.791, -10.461, -11.050, -11.563, -12.001, -12.368,
-12.659, -12.868, -8.180,  -9.016,  -9.767, -10.438, -11.032, -11.547, -11.986,
-12.352, -12.642, -12.850, -8.178,  -9.002,  -9.747, -10.418, -11.014, -11.532,
-11.971, -12.336, -12.625, -12.831, -8.183,  -8.993,  -9.731, -10.401, -10.999,
-11.518, -11.957, -12.321, -12.608, -12.813, -8.192,  -8.988,  -9.718, -10.387,
-10.985, -11.505, -11.944, -12.307, -12.591, -12.794, -8.203,  -8.986,  -9.709,
-10.376, -10.974, -11.494, -11.932, -12.293, -12.575, -12.777, -8.216,  -8.987,
-9.703, -10.367, -10.965, -11.484, -11.921, -12.279, -12.560, -12.759, -8.229,
-8.990,  -9.701, -10.361, -10.957, -11.475, -11.910, -12.266, -12.544, -12.742,
  -8.242,  -8.995,  -9.701, -10.358, -10.952, -11.467, -11.900, -12.253,
-12.529, -12.725, -8.255,  -9.001,  -9.703, -10.358, -10.948, -11.461, -11.890,
-12.241, -12.514, -12.708, -8.267,  -9.010,  -9.708, -10.359, -10.946, -11.454,
-11.881, -12.228, -12.499, -12.691, -8.278,  -9.018,  -9.714, -10.361, -10.944,
-11.449, -11.872, -12.216, -12.485, -12.675, -8.287,  -9.026,  -9.720, -10.363,
-10.941, -11.443, -11.862, -12.204, -12.470, -12.658, -8.292,  -9.032,  -9.723,
-10.363, -10.938, -11.436, -11.853, -12.192, -12.456, -12.642, -8.291,  -9.033,
-9.724, -10.361, -10.933, -11.429, -11.843, -12.180, -12.442, -12.626, -8.284,
-9.028,  -9.720, -10.356, -10.926, -11.420, -11.833, -12.168, -12.427, -12.610,
  -8.272,  -9.019,  -9.712, -10.348, -10.917, -11.410, -11.822, -12.155,
-12.413, -12.594, -8.258,  -9.009,  -9.702, -10.338, -10.907, -11.399, -11.810,
-12.143, -12.399, -12.579, -8.248,  -9.000,  -9.693, -10.327, -10.895, -11.387,
-11.798, -12.130, -12.385, -12.563, -8.244,  -8.996,  -9.686, -10.316, -10.883,
-11.375, -11.786, -12.117, -12.371, -12.548, -8.251,  -8.999,  -9.682, -10.307,
-10.871, -11.363, -11.774, -12.105, -12.358, -12.533, -8.268,  -9.008,  -9.682,
-10.300, -10.860, -11.351, -11.762, -12.092, -12.344, -12.518, -8.295,  -9.023,
-9.685, -10.294, -10.850, -11.339, -11.749, -12.079, -12.330, -12.503, -8.325,
-9.040,  -9.689, -10.289, -10.840, -11.327, -11.737, -12.067, -12.317, -12.488,
  -8.350,  -9.054,  -9.691, -10.283, -10.830, -11.315, -11.725, -12.055,
-12.304, -12.474, -8.360,  -9.059,  -9.690, -10.277, -10.820, -11.304, -11.714,
-12.042, -12.291, -12.460, -8.351,  -9.054,  -9.684, -10.268, -10.809, -11.293,
-11.702, -12.031, -12.278, -12.446, -8.328,  -9.040,  -9.674, -10.259, -10.799,
-11.282, -11.691, -12.019, -12.266, -12.432, -8.298,  -9.022,  -9.662, -10.248,
-10.788, -11.271, -11.680, -12.007, -12.253, -12.418, -8.270,  -9.003,  -9.649,
-10.238, -10.778, -11.260, -11.669, -11.996, -12.241, -12.405, -8.250,  -8.987,
-9.638, -10.228, -10.768, -11.250, -11.658, -11.985, -12.229, -12.392, -8.241,
-8.977,  -9.628, -10.219, -10.758, -11.239, -11.648, -11.974, -12.218, -12.379,
  -8.240,  -8.972,  -9.622, -10.211, -10.749, -11.229, -11.637, -11.964,
-12.206, -12.366, -8.248,  -8.971,  -9.618, -10.204, -10.740, -11.219, -11.627,
-11.953, -12.195, -12.354, -8.263,  -8.975,  -9.615, -10.198, -10.731, -11.209,
-11.617, -11.943, -12.184, -12.342, -8.281,  -8.980,  -9.612, -10.191, -10.722,
-11.199, -11.606, -11.933, -12.173, -12.330, -8.301,  -8.986,  -9.610, -10.183,
-10.712, -11.189, -11.596, -11.922, -12.163, -12.318, -8.320,  -8.992,  -9.606,
-10.175, -10.702, -11.178, -11.586, -11.912, -12.152, -12.306, -8.338,  -8.996,
-9.601, -10.166, -10.691, -11.167, -11.576, -11.903, -12.142, -12.295, -8.354,
-8.998,  -9.595, -10.156, -10.680, -11.156, -11.566, -11.893, -12.132, -12.284,
  -8.367,  -8.999,  -9.589, -10.146, -10.669, -11.146, -11.556, -11.884,
-12.122, -12.273, -8.376,  -8.998,  -9.582, -10.137, -10.659, -11.136, -11.547,
-11.875, -12.113, -12.262, -8.381,  -8.995,  -9.575, -10.128, -10.650, -11.127,
-11.538, -11.866, -12.104, -12.252, -8.382,  -8.990,  -9.568, -10.121, -10.643,
-11.120, -11.531, -11.859, -12.095, -12.242, -8.379,  -8.985,  -9.562, -10.115,
-10.637, -11.113, -11.524, -11.851, -12.086, -12.232, -8.373,  -8.978,  -9.558,
-10.111, -10.633, -11.108, -11.518, -11.844, -12.078, -12.222, -8.364,  -8.973,
-9.555, -10.110, -10.631, -11.105, -11.513, -11.838, -12.071, -12.213, -8.355,
-8.968,  -9.556, -10.112, -10.632, -11.104, -11.510, -11.832, -12.063, -12.203,
  -8.346,  -8.967,  -9.560, -10.117, -10.635, -11.104, -11.507, -11.827,
-12.056, -12.194, -8.342,  -8.971,  -9.568, -10.126, -10.641, -11.106, -11.505,
-11.822, -12.049, -12.185, -8.344,  -8.982,  -9.582, -10.139, -10.649, -11.109,
-11.503, -11.817, -12.042, -12.176, -8.356,  -9.000,  -9.602, -10.155, -10.659,
-11.113, -11.502, -11.813, -12.035, -12.168, -8.379,  -9.026,  -9.626, -10.173,
-10.671, -11.117, -11.501, -11.808, -12.028, -12.159, -8.412,  -9.059,  -9.653,
-10.193, -10.682, -11.121, -11.499, -11.803, -12.021, -12.150, -8.451,  -9.094,
-9.681, -10.211, -10.691, -11.123, -11.497, -11.798, -12.014, -12.142, -8.487,
-9.123,  -9.702, -10.225, -10.698, -11.124, -11.493, -11.792, -12.007, -12.133,
  -8.506,  -9.139,  -9.714, -10.231, -10.699, -11.121, -11.488, -11.786,
-11.999, -12.124, -8.500,  -9.135,  -9.711, -10.229, -10.696, -11.116, -11.482,
-11.778, -11.991, -12.116, -8.470,  -9.113,  -9.696, -10.218, -10.687, -11.107,
-11.473, -11.770, -11.983, -12.107, -8.428,  -9.079,  -9.671, -10.201, -10.674,
-11.096, -11.464, -11.761, -11.975, -12.098, -8.385,  -9.041,  -9.642, -10.179,
-10.657, -11.083, -11.453, -11.752, -11.966, -12.089, -8.352,  -9.008,  -9.614,
-10.156, -10.639, -11.069, -11.441, -11.742, -11.957, -12.080, -8.331,  -8.982,
-9.588, -10.134, -10.621, -11.054, -11.429, -11.732, -11.948, -12.071, -8.324,
-8.966,  -9.569, -10.115, -10.604, -11.039, -11.417, -11.722, -11.940, -12.062,
  -8.330,  -8.960,  -9.556, -10.099, -10.588, -11.026, -11.405, -11.712,
-11.931, -12.054, -8.346,  -8.962,  -9.549, -10.088, -10.575, -11.013, -11.394,
-11.703, -11.922, -12.045, -8.371,  -8.972,  -9.547, -10.080, -10.565, -11.002,
-11.384, -11.694, -11.914, -12.037, -8.402,  -8.987,  -9.550, -10.075, -10.557,
-10.993, -11.375, -11.685, -11.906, -12.029, -8.438,  -9.007,  -9.558, -10.074,
-10.552, -10.986, -11.367, -11.678, -11.898, -12.021, -8.476,  -9.030,  -9.568,
-10.076, -10.549, -10.981, -11.361, -11.670, -11.891, -12.013, -8.516,  -9.055,
-9.580, -10.080, -10.548, -10.977, -11.355, -11.664, -11.884, -12.005, -8.555,
-9.081,  -9.594, -10.086, -10.549, -10.975, -11.351, -11.658, -11.877, -11.998,
  -8.593,  -9.106,  -9.609, -10.094, -10.552, -10.974, -11.347, -11.653,
-11.871, -11.991, -8.630,  -9.131,  -9.625, -10.103, -10.557, -10.975, -11.345,
-11.648, -11.865, -11.984, -8.663,  -9.154,  -9.640, -10.113, -10.563, -10.977,
-11.343, -11.644, -11.859, -11.977, -8.693,  -9.176,  -9.656, -10.125, -10.570,
-10.980, -11.343, -11.640, -11.854, -11.970, -8.720,  -9.197,  -9.671, -10.137,
-10.578, -10.984, -11.342, -11.637, -11.849, -11.964, -8.744,  -9.216,  -9.687,
-10.150, -10.587, -10.989, -11.343, -11.634, -11.844, -11.958, -8.765,  -9.234,
-9.704, -10.164, -10.598, -10.994, -11.343, -11.631, -11.839, -11.951, -8.785,
-9.253,  -9.722, -10.179, -10.609, -11.000, -11.344, -11.629, -11.834, -11.945,
  -8.804,  -9.273,  -9.742, -10.197, -10.621, -11.007, -11.346, -11.626,
-11.829, -11.939, -8.824,  -9.296,  -9.764, -10.216, -10.635, -11.014, -11.347,
-11.624, -11.825, -11.933, -8.848,  -9.322,  -9.790, -10.237, -10.650, -11.021,
-11.349, -11.621, -11.820, -11.927, -8.876,  -9.354,  -9.820, -10.261, -10.665,
-11.029, -11.350, -11.619, -11.815, -11.921, -8.910,  -9.389,  -9.852, -10.286,
-10.681, -11.037, -11.352, -11.617, -11.811, -11.916, -8.948,  -9.427,  -9.885,
-10.310, -10.696, -11.044, -11.353, -11.614, -11.806, -11.910, -8.985,  -9.464,
-9.916, -10.333, -10.710, -11.050, -11.354, -11.612, -11.802, -11.904, -9.016,
-9.493,  -9.940, -10.351, -10.721, -11.055, -11.355, -11.609, -11.797, -11.898,
  -9.034,  -9.511,  -9.956, -10.363, -10.729, -11.059, -11.354, -11.606,
-11.792, -11.892, -9.037,  -9.515,  -9.961, -10.368, -10.733, -11.061, -11.354,
-11.603, -11.788, -11.887, -9.028,  -9.507,  -9.957, -10.368, -10.734, -11.061,
-11.352, -11.599, -11.783, -11.881, -9.015,  -9.494,  -9.948, -10.363, -10.732,
-11.060, -11.350, -11.596, -11.778, -11.875, -9.003,  -9.480,  -9.936, -10.356,
-10.728, -11.058, -11.347, -11.592, -11.773, -11.869, -8.998,  -9.470,  -9.926,
-10.348, -10.724, -11.055, -11.345, -11.588, -11.768, -11.863, -9.002,  -9.466,
-9.919, -10.342, -10.720, -11.052, -11.342, -11.584, -11.763, -11.857, -9.014,
-9.469,  -9.917, -10.338, -10.717, -11.050, -11.339, -11.580, -11.758, -11.852,
  -9.033,  -9.478,  -9.919, -10.337, -10.715, -11.047, -11.335, -11.576,
-11.752, -11.846, -9.060,  -9.494,  -9.926, -10.338, -10.714, -11.046, -11.332,
-11.571, -11.747, -11.840, -9.091,  -9.515,  -9.937, -10.342, -10.714, -11.044,
-11.329, -11.567, -11.742, -11.834, -9.126,  -9.540,  -9.951, -10.349, -10.716,
-11.043, -11.326, -11.562, -11.736, -11.828, -9.164,  -9.568,  -9.968, -10.357,
-10.719, -11.042, -11.324, -11.558, -11.731, -11.822, -9.203,  -9.598,  -9.987,
-10.367, -10.723, -11.042, -11.321, -11.554, -11.726, -11.816, -9.244,  -9.630,
-10.008, -10.378, -10.727, -11.043, -11.319, -11.550, -11.720, -11.810, -9.284,
-9.661, -10.029, -10.391, -10.733, -11.044, -11.317, -11.546, -11.715, -11.804,
  -9.324,  -9.693, -10.051, -10.404, -10.740, -11.047, -11.316, -11.542,
-11.710, -11.799, -9.362,  -9.725, -10.074, -10.419, -10.748, -11.050, -11.315,
-11.539, -11.705, -11.793, -9.399,  -9.755, -10.096, -10.434, -10.757, -11.054,
-11.316, -11.536, -11.701, -11.788, -9.435,  -9.784, -10.118, -10.450, -10.767,
-11.059, -11.317, -11.534, -11.697, -11.782, -9.468,  -9.813, -10.141, -10.466,
-10.779, -11.065, -11.318, -11.532, -11.693, -11.777, -9.500,  -9.840, -10.163,
-10.484, -10.792, -11.073, -11.321, -11.531, -11.689, -11.772, -9.530,  -9.866,
-10.186, -10.503, -10.806, -11.081, -11.324, -11.530, -11.685, -11.767, -9.559,
-9.893, -10.210, -10.524, -10.821, -11.091, -11.329, -11.530, -11.682, -11.763,
  -9.586,  -9.920, -10.235, -10.546, -10.838, -11.102, -11.333, -11.530,
-11.679, -11.758, -9.614,  -9.948, -10.262, -10.569, -10.856, -11.113, -11.339,
-11.530, -11.676, -11.753, -9.642,  -9.977, -10.291, -10.594, -10.875, -11.125,
-11.344, -11.531, -11.673, -11.749, -9.671, -10.008, -10.321, -10.620, -10.894,
-11.137, -11.349, -11.531, -11.670, -11.745, -9.701, -10.041, -10.352, -10.645,
-10.912, -11.148, -11.354, -11.531, -11.668, -11.740, -9.731, -10.072, -10.382,
-10.669, -10.929, -11.158, -11.358, -11.531, -11.664, -11.736, -9.758, -10.100,
-10.408, -10.690, -10.944, -11.167, -11.362, -11.530, -11.661, -11.731, -9.779,
-10.122, -10.427, -10.706, -10.955, -11.173, -11.364, -11.529, -11.658, -11.726,
  -9.794, -10.135, -10.440, -10.716, -10.962, -11.177, -11.364, -11.527,
-11.654, -11.722, -9.801, -10.141, -10.445, -10.721, -10.965, -11.178, -11.364,
-11.525, -11.650, -11.717, -9.804, -10.141, -10.445, -10.721, -10.965, -11.177,
-11.362, -11.521, -11.646, -11.712, -9.806, -10.139, -10.441, -10.717, -10.962,
-11.175, -11.359, -11.518, -11.642, -11.707, -9.809, -10.137, -10.437, -10.713,
-10.958, -11.171, -11.355, -11.514, -11.637, -11.702, -9.815, -10.138, -10.434,
-10.708, -10.953, -11.166, -11.351, -11.510, -11.632, -11.698, -9.824, -10.142,
-10.434, -10.704, -10.948, -11.162, -11.347, -11.505, -11.628, -11.693, -9.836,
-10.150, -10.436, -10.702, -10.943, -11.157, -11.342, -11.501, -11.623, -11.688,
  -9.851, -10.161, -10.441, -10.701, -10.940, -11.153, -11.338, -11.497,
-11.619, -11.683, -9.868, -10.176, -10.449, -10.702, -10.937, -11.149, -11.335,
-11.493, -11.615, -11.679, -9.888, -10.193, -10.459, -10.705, -10.936, -11.146,
-11.332, -11.490, -11.611, -11.674, -9.909, -10.211, -10.471, -10.710, -10.936,
-11.144, -11.329, -11.487, -11.608, -11.670, -9.930, -10.231, -10.484, -10.716,
-10.937, -11.143, -11.328, -11.485, -11.604, -11.666, -9.953, -10.252, -10.499,
-10.723, -10.939, -11.143, -11.327, -11.483, -11.601, -11.662, -9.977, -10.274,
-10.514, -10.731, -10.942, -11.145, -11.327, -11.482, -11.599, -11.658, -10.001,
-10.296, -10.529, -10.740, -10.947, -11.147, -11.328, -11.482, -11.597, -11.654,
 -10.025, -10.318, -10.545, -10.749, -10.952, -11.150, -11.330, -11.482,
-11.595, -11.651, -10.049, -10.340, -10.561, -10.759, -10.959, -11.155, -11.332,
-11.482, -11.593, -11.647, -10.074, -10.361, -10.577, -10.770, -10.966, -11.160,
-11.336, -11.483, -11.591, -11.644, -10.099, -10.383, -10.593, -10.782, -10.975,
-11.167, -11.340, -11.484, -11.590, -11.641, -10.124, -10.404, -10.610, -10.795,
-10.985, -11.174, -11.345, -11.486, -11.589, -11.638, -10.149, -10.426, -10.628,
-10.810, -10.997, -11.183, -11.350, -11.488, -11.588, -11.635, -10.175, -10.448,
-10.647, -10.826, -11.011, -11.193, -11.356, -11.490, -11.587, -11.632, -10.202,
-10.472, -10.667, -10.844, -11.026, -11.204, -11.363, -11.493, -11.586, -11.629,
 -10.229, -10.497, -10.690, -10.865, -11.043, -11.216, -11.370, -11.495,
-11.585, -11.626, -10.258, -10.524, -10.716, -10.888, -11.061, -11.229, -11.377,
-11.498, -11.585, -11.623, -10.288, -10.553, -10.744, -10.913, -11.081, -11.242,
-11.384, -11.500, -11.584, -11.620, -10.319, -10.584, -10.774, -10.939, -11.100,
-11.255, -11.391, -11.502, -11.582, -11.617, -10.352, -10.616, -10.804, -10.964,
-11.119, -11.267, -11.397, -11.504, -11.581, -11.613, -10.384, -10.647, -10.832,
-10.987, -11.135, -11.277, -11.402, -11.505, -11.579, -11.610, -10.415, -10.675,
-10.856, -11.005, -11.148, -11.285, -11.406, -11.506, -11.578, -11.607, -10.443,
-10.697, -10.873, -11.018, -11.157, -11.290, -11.408, -11.506, -11.576, -11.603,
 -10.467, -10.713, -10.882, -11.023, -11.160, -11.292, -11.409, -11.505,
-11.573, -11.600, -10.488, -10.723, -10.885, -11.023, -11.160, -11.291, -11.408,
-11.503, -11.571, -11.596, -10.508, -10.730, -10.884, -11.020, -11.156, -11.289,
-11.406, -11.501, -11.568, -11.592, -10.527, -10.737, -10.882, -11.014, -11.150,
-11.284, -11.403, -11.499, -11.565, -11.588, -10.547, -10.745, -10.881, -11.008,
-11.143, -11.279, -11.400, -11.496, -11.562, -11.585, -10.568, -10.755, -10.881,
-11.003, -11.137, -11.274, -11.396, -11.493, -11.559, -11.581, -10.590, -10.767,
-10.885, -11.001, -11.132, -11.269, -11.392, -11.490, -11.556, -11.577, -10.613,
-10.783, -10.892, -11.001, -11.129, -11.265, -11.389, -11.488, -11.553, -11.573,
 -10.637, -10.801, -10.902, -11.004, -11.128, -11.262, -11.386, -11.485,
-11.550, -11.570, -10.662, -10.820, -10.916, -11.010, -11.128, -11.261, -11.384,
-11.483, -11.547, -11.566, -10.686, -10.842, -10.931, -11.019, -11.131, -11.260,
-11.383, -11.481, -11.545, -11.563, -10.710, -10.864, -10.949, -11.030, -11.136,
-11.262, -11.383, -11.480, -11.543, -11.560, -10.734, -10.887, -10.968, -11.043,
-11.143, -11.264, -11.383, -11.479, -11.541, -11.557, -10.758, -10.911, -10.989,
-11.058, -11.152, -11.269, -11.385, -11.479, -11.539, -11.554, -10.781, -10.934,
-11.011, -11.074, -11.162, -11.274, -11.388, -11.480, -11.538, -11.551, -10.803,
-10.958, -11.033, -11.091, -11.173, -11.281, -11.391, -11.481, -11.537, -11.549,
 -10.825, -10.981, -11.055, -11.109, -11.186, -11.290, -11.396, -11.484,
-11.537, -11.547, -10.846, -11.004, -11.078, -11.128, -11.201, -11.299, -11.402,
-11.486, -11.537, -11.545, -10.867, -11.027, -11.101, -11.148, -11.216, -11.311,
-11.410, -11.490, -11.538, -11.543, -10.888, -11.050, -11.124, -11.169, -11.233,
-11.323, -11.418, -11.495, -11.539, -11.542, -10.909, -11.072, -11.148, -11.191,
-11.251, -11.337, -11.427, -11.500, -11.541, -11.541, -10.930, -11.095, -11.172,
-11.214, -11.271, -11.352, -11.437, -11.505, -11.543, -11.539, -10.952, -11.118,
-11.197, -11.239, -11.292, -11.368, -11.449, -11.512, -11.545, -11.539, -10.974,
-11.142, -11.223, -11.264, -11.315, -11.386, -11.461, -11.518, -11.547, -11.538,
 -10.997, -11.167, -11.250, -11.292, -11.339, -11.404, -11.473, -11.525,
-11.550, -11.537, -11.020, -11.192, -11.279, -11.320, -11.363, -11.423, -11.486,
-11.533, -11.552, -11.536, -11.045, -11.219, -11.308, -11.349, -11.389, -11.443,
-11.499, -11.540, -11.555, -11.535, -11.071, -11.247, -11.338, -11.379, -11.414,
-11.461, -11.511, -11.547, -11.557, -11.535, -11.097, -11.275, -11.368, -11.408,
-11.438, -11.479, -11.523, -11.553, -11.560, -11.534, -11.125, -11.303, -11.397,
-11.434, -11.460, -11.496, -11.534, -11.559, -11.562, -11.533, -11.153, -11.330,
-11.424, -11.458, -11.480, -11.510, -11.543, -11.564, -11.563, -11.532, -11.181,
-11.356, -11.447, -11.479, -11.496, -11.522, -11.551, -11.568, -11.564, -11.530,
 -11.210, -11.381, -11.468, -11.496, -11.510, -11.532, -11.557, -11.572,
-11.565, -11.529, -11.239, -11.405, -11.487, -11.510, -11.520, -11.539, -11.562,
-11.575, -11.566, -11.527, -11.268, -11.430, -11.505, -11.522, -11.529, -11.545,
-11.566, -11.577, -11.566, -11.525, -11.299, -11.455, -11.522, -11.534, -11.536,
-11.550, -11.569, -11.578, -11.566, -11.523, -11.330, -11.481, -11.541, -11.545,
-11.543, -11.554, -11.572, -11.579, -11.565, -11.521, -11.362, -11.508, -11.560,
-11.558, -11.550, -11.558, -11.574, -11.580, -11.565, -11.520, -11.396, -11.536,
-11.581, -11.571, -11.558, -11.562, -11.576, -11.581, -11.564, -11.518, -11.429,
-11.566, -11.604, -11.587, -11.567, -11.567, -11.579, -11.582, -11.564, -11.516,
 -11.464, -11.596, -11.629, -11.603, -11.577, -11.573, -11.581, -11.583,
-11.564, -11.514, -11.499, -11.628, -11.654, -11.622, -11.588, -11.579, -11.585,
-11.585, -11.564, -11.512, -11.534, -11.660, -11.681, -11.641, -11.601, -11.586,
-11.589, -11.587, -11.564, -11.511, -11.569, -11.692, -11.709, -11.662, -11.615,
-11.595, -11.594, -11.590, -11.565, -11.510, -11.604, -11.725, -11.737, -11.684,
-11.630, -11.604, -11.600, -11.593, -11.566, -11.509, -11.639, -11.758, -11.766,
-11.707, -11.646, -11.615, -11.606, -11.597, -11.567, -11.508, -11.674, -11.790,
-11.795, -11.731, -11.663, -11.626, -11.614, -11.601, -11.569, -11.508, -11.708,
-11.823, -11.824, -11.754, -11.681, -11.639, -11.622, -11.606, -11.571, -11.507,
 -11.743, -11.855, -11.853, -11.779, -11.699, -11.652, -11.631, -11.611,
-11.574, -11.507, -11.777, -11.886, -11.882, -11.803, -11.719, -11.666, -11.641,
-11.617, -11.576, -11.507, -11.810, -11.917, -11.910, -11.828, -11.739, -11.682,
-11.651, -11.624, -11.579, -11.507, -11.843, -11.948, -11.939, -11.854, -11.760,
-11.698, -11.663, -11.631, -11.583, -11.508, -11.875, -11.977, -11.967, -11.880,
-11.782, -11.715, -11.675, -11.638, -11.586, -11.508, -11.907, -12.007, -11.995,
-11.906, -11.805, -11.733, -11.687, -11.646, -11.590, -11.509, -11.938, -12.036,
-12.023, -11.933, -11.829, -11.752, -11.701, -11.654, -11.594, -11.509, -11.969,
-12.065, -12.051, -11.960, -11.853, -11.771, -11.714, -11.662, -11.598, -11.510,
 -12.000, -12.093, -12.080, -11.988, -11.878, -11.791, -11.728, -11.671,
-11.601, -11.511, -12.030, -12.121, -12.108, -12.017, -11.904, -11.811, -11.742,
-11.679, -11.605, -11.511, -12.060, -12.150, -12.137, -12.045, -11.930, -11.831,
-11.755, -11.686, -11.609, -11.512, -12.089, -12.177, -12.165, -12.074, -11.954,
-11.850, -11.768, -11.694, -11.612, -11.512, -12.119, -12.205, -12.193, -12.101,
-11.978, -11.868, -11.779, -11.701, -11.615, -11.513, -12.148, -12.232, -12.219,
-12.126, -11.999, -11.884, -11.790, -11.707, -11.618, -11.513, -12.177, -12.258,
-12.244, -12.149, -12.018, -11.898, -11.799, -11.712, -11.620, -11.513, -12.205,
-12.284, -12.267, -12.169, -12.034, -11.909, -11.807, -11.717, -11.623, -11.513,
 -12.234, -12.310, -12.289, -12.187, -12.048, -11.919, -11.813, -11.720,
-11.624, -11.513, -12.263, -12.335, -12.310, -12.203, -12.060, -11.927, -11.818,
-11.724, -11.626, -11.512, -12.292, -12.361, -12.331, -12.218, -12.070, -11.934,
-11.822, -11.726, -11.627, -11.512, -12.322, -12.387, -12.351, -12.233, -12.080,
-11.939, -11.826, -11.728, -11.628, -11.511, -12.352, -12.413, -12.373, -12.248,
-12.089, -11.945, -11.829, -11.730, -11.628, -11.511, -12.382, -12.440, -12.395,
-12.264, -12.099, -11.950, -11.832, -11.731, -11.629, -11.510, -12.413, -12.468,
-12.418, -12.280, -12.109, -11.955, -11.834, -11.733, -11.630, -11.509, -12.444,
-12.496, -12.442, -12.298, -12.120, -11.961, -11.837, -11.735, -11.630, -11.509,
 -12.475, -12.525, -12.467, -12.317, -12.132, -11.968, -11.841, -11.736,
-11.631, -11.508, -12.507, -12.554, -12.492, -12.337, -12.145, -11.975, -11.845,
-11.739, -11.632, -11.508, -12.539, -12.584, -12.518, -12.357, -12.159, -11.984,
-11.849, -11.741, -11.633, -11.508, -12.571, -12.614, -12.545, -12.378, -12.174,
-11.992, -11.855, -11.744, -11.635, -11.507, -12.604, -12.645, -12.572, -12.400,
-12.189, -12.002, -11.860, -11.748, -11.636, -11.507, -12.637, -12.676, -12.599,
-12.422, -12.205, -12.013, -11.867, -11.752, -11.638, -11.508, -12.670, -12.706,
-12.626, -12.445, -12.222, -12.024, -11.875, -11.756, -11.641, -11.508, -12.703,
-12.737, -12.654, -12.468, -12.239, -12.037, -11.883, -11.762, -11.644, -11.509,
 -12.737, -12.768, -12.681, -12.491, -12.257, -12.050, -11.892, -11.768,
-11.647, -11.510, -12.770, -12.799, -12.708, -12.514, -12.276, -12.064, -11.902,
-11.774, -11.651, -11.511, -12.803, -12.830, -12.735, -12.537, -12.295, -12.079,
-11.913, -11.782, -11.655, -11.513, -12.837, -12.860, -12.762, -12.560, -12.315,
-12.095, -11.925, -11.790, -11.659, -11.514, -12.870, -12.890, -12.788, -12.584,
-12.336, -12.112, -11.938, -11.798, -11.664, -11.516, -12.903, -12.920, -12.814,
-12.608, -12.358, -12.130, -11.951, -11.807, -11.669, -11.518, -12.935, -12.950,
-12.840, -12.632, -12.380, -12.149, -11.965, -11.816, -11.675, -11.520, -12.968,
-12.979, -12.866, -12.657, -12.403, -12.168, -11.980, -11.826, -11.680, -11.522,
 -13.000, -13.008, -12.892, -12.681, -12.426, -12.188, -11.994, -11.835,
-11.685, -11.525, -13.032, -13.036, -12.917, -12.706, -12.450, -12.208, -12.009,
-11.845, -11.691, -11.527, -13.063, -13.064, -12.942, -12.730, -12.473, -12.227,
-12.022, -11.853, -11.696, -11.529, -13.095, -13.091, -12.966, -12.754, -12.495,
-12.245, -12.035, -11.862, -11.700, -11.531, -13.125, -13.118, -12.989, -12.776,
-12.515, -12.261, -12.047, -11.869, -11.705, -11.533, -13.155, -13.144, -13.012,
-12.796, -12.533, -12.276, -12.058, -11.876, -11.709, -11.534, -13.185, -13.169,
-13.032, -12.814, -12.549, -12.288, -12.066, -11.882, -11.712, -11.535, -13.215,
-13.194, -13.052, -12.830, -12.562, -12.299, -12.074, -11.887, -11.715, -11.537,
 -13.244, -13.217, -13.071, -12.845, -12.573, -12.307, -12.079, -11.891,
-11.717, -11.538, -13.272, -13.241, -13.089, -12.858, -12.583, -12.314, -12.084,
-11.894, -11.719, -11.538, -13.301, -13.264, -13.106, -12.871, -12.592, -12.320,
-12.088, -11.896, -11.721, -11.539, -13.329, -13.287, -13.124, -12.884, -12.601,
-12.325, -12.091, -11.898, -11.722, -11.539, -13.356, -13.310, -13.142, -12.898,
-12.610, -12.330, -12.094, -11.900, -11.724, -11.540, -13.384, -13.332, -13.160,
-12.912, -12.620, -12.336, -12.097, -11.902, -11.725, -11.540, -13.411, -13.355,
-13.179, -12.927, -12.630, -12.342, -12.101, -11.904, -11.726, -11.541, -13.439,
-13.378, -13.199, -12.943, -12.642, -12.350, -12.105, -11.907, -11.727, -11.541,
 -13.466, -13.401, -13.218, -12.960, -12.655, -12.358, -12.110, -11.909,
-11.729, -11.542, -13.494, -13.425, -13.239, -12.978, -12.669, -12.367, -12.115,
-11.913, -11.731, -11.542, -13.521, -13.449, -13.260, -12.997, -12.684, -12.377,
-12.122, -11.917, -11.733, -11.543, -13.549, -13.473, -13.281, -13.016, -12.700,
-12.389, -12.129, -11.921, -11.736, -11.544, -13.577, -13.497, -13.304, -13.037,
-12.717, -12.401, -12.137, -11.926, -11.739, -11.546, -13.605, -13.522, -13.326,
-13.058, -12.735, -12.415, -12.147, -11.932, -11.743, -11.547, -13.634, -13.548,
-13.349, -13.079, -12.753, -12.429, -12.157, -11.939, -11.747, -11.549, -13.663,
-13.574, -13.373, -13.102, -12.773, -12.445, -12.168, -11.947, -11.751, -11.551,
 -13.692, -13.601, -13.397, -13.124, -12.793, -12.461, -12.180, -11.955,
-11.756, -11.553, -13.721, -13.628, -13.422, -13.147, -12.814, -12.479, -12.193,
-11.964, -11.762, -11.555, -13.751, -13.656, -13.447, -13.171, -12.836, -12.497,
-12.207, -11.973, -11.767, -11.558, -13.781, -13.684, -13.473, -13.196, -12.859,
-12.516, -12.222, -11.983, -11.774, -11.561, -13.812, -13.713, -13.500, -13.220,
-12.882, -12.537, -12.238, -11.994, -11.780, -11.564, -13.842, -13.742, -13.526,
-13.246, -12.906, -12.558, -12.254, -12.005, -11.787, -11.567, -13.873, -13.772,
-13.554, -13.272, -12.931, -12.579, -12.271, -12.017, -11.794, -11.571, -13.904,
-13.802, -13.582, -13.298, -12.956, -12.601, -12.288, -12.029, -11.801, -11.574,
 -13.936, -13.833, -13.610, -13.325, -12.982, -12.624, -12.305, -12.040,
-11.808, -11.577, -13.968, -13.864, -13.638, -13.352, -13.007, -12.646, -12.323,
-12.052, -11.814, -11.580, -13.999, -13.895, -13.667, -13.378, -13.033, -12.668,
-12.339, -12.063, -11.821, -11.584, -14.031, -13.927, -13.696, -13.405, -13.057,
-12.689, -12.355, -12.074, -11.828, -11.587, -14.063, -13.958, -13.725, -13.431,
-13.081, -12.709, -12.370, -12.084, -11.834, -11.590, -14.096, -13.990, -13.754,
-13.457, -13.103, -12.728, -12.384, -12.094, -11.839, -11.593, -14.128, -14.022,
-13.783, -13.481, -13.124, -12.745, -12.397, -12.102, -11.845, -11.596, -14.160,
-14.054, -13.811, -13.505, -13.144, -12.761, -12.409, -12.110, -11.850, -11.598,
 -14.192, -14.086, -13.839, -13.528, -13.163, -12.775, -12.419, -12.117,
-11.854, -11.601, -14.224, -14.118, -13.868, -13.551, -13.180, -12.788, -12.429,
-12.124, -11.859, -11.603, -14.256, -14.150, -13.896, -13.574, -13.198, -12.801,
-12.438, -12.130, -11.863, -11.605, -14.288, -14.182, -13.924, -13.597, -13.215,
-12.814, -12.446, -12.136, -11.867, -11.608, -14.320, -14.214, -13.952, -13.620,
-13.233, -12.826, -12.455, -12.142, -11.871, -11.610, -14.352, -14.245, -13.980,
-13.643, -13.250, -12.839, -12.464, -12.148, -11.875, -11.612, -14.384, -14.277,
-14.008, -13.666, -13.269, -12.852, -12.473, -12.154, -11.878, -11.614, -14.415,
-14.308, -14.035, -13.689, -13.287, -12.866, -12.482, -12.160, -11.883, -11.617,
 -14.446, -14.339, -14.063, -13.713, -13.306, -12.880, -12.492, -12.167,
-11.887, -11.619, -14.478, -14.369, -14.091, -13.737, -13.326, -12.895, -12.503,
-12.174, -11.892, -11.622, -14.509, -14.400, -14.118, -13.760, -13.346, -12.911,
-12.514, -12.182, -11.897, -11.625, -14.540, -14.430, -14.145, -13.784, -13.367,
-12.927, -12.526, -12.190, -11.902, -11.628, -14.570, -14.460, -14.172, -13.808,
-13.388, -12.944, -12.539, -12.199, -11.907, -11.631, -14.601, -14.490, -14.199,
-13.833, -13.409, -12.962, -12.552, -12.208, -11.913, -11.635, -14.631, -14.520,
-14.226, -13.857, -13.431, -12.980, -12.566, -12.218, -11.920, -11.638, -14.662,
-14.550, -14.253, -13.881, -13.453, -12.999, -12.581, -12.228, -11.926, -11.642,
 -14.692, -14.579, -14.280, -13.905, -13.476, -13.019, -12.596, -12.239,
-11.934, -11.646, -14.722, -14.608, -14.306, -13.930, -13.499, -13.039, -12.612,
-12.250, -11.941, -11.650, -14.752, -14.638, -14.333, -13.954, -13.522, -13.060,
-12.629, -12.262, -11.949, -11.655, -14.782, -14.667, -14.360, -13.979, -13.545,
-13.081, -12.646, -12.275, -11.957, -11.659, -14.812, -14.697, -14.387, -14.004,
-13.569, -13.103, -12.664, -12.288, -11.965, -11.663, -14.842, -14.726, -14.414,
-14.029, -13.594, -13.125, -12.682, -12.301, -11.973, -11.668, -14.872, -14.755,
-14.441, -14.054, -13.618, -13.148, -12.701, -12.314, -11.981, -11.672, -14.902,
-14.785, -14.468, -14.080, -13.643, -13.171, -12.719, -12.327, -11.989, -11.677,
 -14.932, -14.815, -14.496, -14.106, -13.668, -13.194, -12.738, -12.340,
-11.997, -11.681, -14.962, -14.845, -14.524, -14.131, -13.693, -13.217, -12.756,
-12.353, -12.005, -11.685, -14.992, -14.875, -14.552, -14.157, -13.717, -13.239,
-12.774, -12.365, -12.013, -11.689, -15.022, -14.905, -14.580, -14.183, -13.742,
-13.261, -12.791, -12.377, -12.020, -11.693, -15.052, -14.935, -14.609, -14.209,
-13.765, -13.281, -12.807, -12.388, -12.027, -11.697, -15.082, -14.966, -14.638,
-14.235, -13.788, -13.301, -12.822, -12.398, -12.033, -11.700, -15.112, -14.997,
-14.666, -14.260, -13.811, -13.319, -12.836, -12.408, -12.039, -11.704, -15.142,
-15.028, -14.695, -14.286, -13.832, -13.337, -12.849, -12.417, -12.045, -11.707,
 -15.172, -15.059, -14.725, -14.311, -13.853, -13.354, -12.861, -12.425,
-12.050, -11.710, -15.202, -15.090, -14.754, -14.337, -13.874, -13.370, -12.873,
-12.433, -12.055, -11.713, -15.232, -15.121, -14.784, -14.363, -13.895, -13.386,
-12.884, -12.441, -12.060, -11.715, -15.262, -15.152, -14.814, -14.388, -13.916,
-13.402, -12.895, -12.448, -12.065, -11.718, -15.292, -15.184, -14.844, -14.414,
-13.937, -13.417, -12.906, -12.455, -12.069, -11.721, -15.322, -15.215, -14.874,
-14.441, -13.958, -13.433, -12.917, -12.462, -12.074, -11.724, -15.352, -15.247,
-14.904, -14.467, -13.980, -13.449, -12.928, -12.470, -12.079, -11.727, -15.382,
-15.278, -14.934, -14.493, -14.001, -13.466, -12.940, -12.477, -12.084, -11.730,
 -15.412, -15.310, -14.964, -14.520, -14.023, -13.483, -12.952, -12.485,
-12.089, -11.733, -15.442, -15.341, -14.994, -14.546, -14.046, -13.500, -12.964,
-12.494, -12.095, -11.737, -15.472, -15.373, -15.025, -14.573, -14.068, -13.518,
-12.977, -12.503, -12.100, -11.740, -15.502, -15.404, -15.055, -14.599, -14.090,
-13.536, -12.991, -12.512, -12.107, -11.744, -15.531, -15.435, -15.085, -14.625,
-14.113, -13.554, -13.005, -12.522, -12.114, -11.748, -15.561, -15.467, -15.115,
-14.652, -14.136, -13.573, -13.020, -12.533, -12.121, -11.753, -15.591, -15.498,
-15.144, -14.678, -14.159, -13.593, -13.035, -12.544, -12.128, -11.758, -15.620,
-15.528, -15.174, -14.704, -14.182, -13.612, -13.051, -12.556, -12.136, -11.763,
 -15.650, -15.559, -15.203, -14.730, -14.205, -13.633, -13.067, -12.568,
-12.145, -11.768, -15.679, -15.590, -15.232, -14.756, -14.228, -13.653, -13.084,
-12.581, -12.153, -11.773, -15.708, -15.620, -15.261, -14.782, -14.251, -13.674,
-13.102, -12.594, -12.163, -11.779, -15.738, -15.651, -15.290, -14.807, -14.274,
-13.696, -13.120, -12.608, -12.172, -11.785, -15.767, -15.681, -15.319, -14.833,
-14.298, -13.718, -13.139, -12.622, -12.181, -11.791, -15.796, -15.711, -15.347,
-14.858, -14.321, -13.740, -13.158, -12.636, -12.191, -11.797, -15.825, -15.741,
-15.375, -14.883, -14.345, -13.762, -13.177, -12.651, -12.201, -11.802, -15.854,
-15.770, -15.404, -14.909, -14.368, -13.785, -13.196, -12.665, -12.210, -11.808,
 -15.883, -15.800, -15.432, -14.934, -14.392, -13.807, -13.215, -12.679,
-12.220, -11.814, -15.912, -15.830, -15.460, -14.959, -14.415, -13.829, -13.234,
-12.693, -12.229, -11.819, -15.941, -15.859, -15.487, -14.984, -14.439, -13.850,
-13.251, -12.706, -12.237, -11.825, -15.970, -15.888, -15.515, -15.009, -14.461,
-13.871, -13.268, -12.718, -12.246, -11.829, -15.998, -15.918, -15.543, -15.033,
-14.483, -13.891, -13.284, -12.730, -12.253, -11.834, -16.027, -15.947, -15.571,
-15.058, -14.505, -13.909, -13.299, -12.741, -12.260, -11.838, -16.056, -15.976,
-15.598, -15.083, -14.526, -13.927, -13.313, -12.751, -12.267, -11.842, -16.084,
-16.005, -15.626, -15.107, -14.547, -13.944, -13.327, -12.760, -12.273, -11.846,
 -16.113, -16.034, -15.654, -15.132, -14.568, -13.961, -13.339, -12.769,
-12.279, -11.850, -16.142, -16.063, -15.682, -15.157, -14.589, -13.978, -13.351,
-12.777, -12.284, -11.853, -16.170, -16.093, -15.710, -15.182, -14.610, -13.994,
-13.363, -12.785, -12.289, -11.856, -16.199, -16.122, -15.738, -15.207, -14.631,
-14.010, -13.374, -12.793, -12.294, -11.859, -16.227, -16.151, -15.767, -15.232,
-14.652, -14.027, -13.386, -12.800, -12.299, -11.862, -16.256, -16.180, -15.795,
-15.258, -14.674, -14.044, -13.398, -12.808, -12.304, -11.866, -16.284, -16.209,
-15.824, -15.284, -14.696, -14.061, -13.410, -12.817, -12.310, -11.869, -16.313,
-16.239, -15.853, -15.310, -14.719, -14.079, -13.423, -12.825, -12.315, -11.872,
 -16.341, -16.268, -15.882, -15.336, -14.741, -14.097, -13.437, -12.834,
-12.321, -11.876, -16.369, -16.298, -15.911, -15.363, -14.764, -14.116, -13.451,
-12.844, -12.327, -11.880, -16.398, -16.327, -15.940, -15.389, -14.788, -14.135,
-13.465, -12.854, -12.334, -11.884, -16.426, -16.356, -15.969, -15.416, -14.811,
-14.155, -13.480, -12.865, -12.341, -11.888, -16.455, -16.386, -15.999, -15.443,
-14.835, -14.175, -13.496, -12.876, -12.349, -11.893, -16.483, -16.415, -16.028,
-15.470, -14.859, -14.196, -13.513, -12.888, -12.357, -11.898, -16.511, -16.445,
-16.058, -15.498, -14.883, -14.217, -13.530, -12.901, -12.365, -11.903, -16.540,
-16.474, -16.088, -15.525, -14.908, -14.238, -13.547, -12.914, -12.374, -11.909,
 -16.568, -16.504, -16.117, -15.553, -14.932, -14.260, -13.565, -12.928,
-12.384, -11.914, -16.596, -16.533, -16.147, -15.580, -14.957, -14.282, -13.584,
-12.942, -12.394, -11.921, -16.625, -16.563, -16.177, -15.608, -14.981, -14.305,
-13.604, -12.957, -12.404, -11.927, -16.653, -16.592, -16.207, -15.635, -15.006,
-14.327, -13.623, -12.972, -12.414, -11.934, -16.681, -16.621, -16.237, -15.663,
-15.031, -14.350, -13.643, -12.988, -12.425, -11.940, -16.709, -16.651, -16.266,
-15.690, -15.056, -14.374, -13.664, -13.004, -12.436, -11.947, -16.738, -16.680,
-16.296, -15.718, -15.081, -14.397, -13.685, -13.020, -12.447, -11.954, -16.766,
-16.709, -16.325, -15.745, -15.106, -14.420, -13.705, -13.036, -12.458, -11.961,
 -16.794, -16.738, -16.355, -15.773, -15.131, -14.443, -13.726, -13.052,
-12.469, -11.968, -16.822, -16.767, -16.384, -15.800, -15.155, -14.466, -13.746,
-13.068, -12.480, -11.974, -16.850, -16.796, -16.414, -15.827, -15.180, -14.489,
-13.765, -13.083, -12.491, -11.981, -16.878, -16.825, -16.443, -15.854, -15.204,
-14.511, -13.784, -13.097, -12.501, -11.987, -16.906, -16.854, -16.472, -15.881,
-15.227, -14.532, -13.802, -13.111, -12.511, -11.994, -16.934, -16.883, -16.501,
-15.907, -15.251, -14.553, -13.820, -13.125, -12.520, -12.000, -16.962, -16.912,
-16.530, -15.934, -15.274, -14.573, -13.837, -13.138, -12.529, -12.006, -16.990,
-16.940, -16.558, -15.960, -15.297, -14.593, -13.853, -13.150, -12.538, -12.011,
 -17.018, -16.969, -16.587, -15.986, -15.320, -14.612, -13.869, -13.162,
-12.546, -12.017, -17.046, -16.997, -16.615, -16.013, -15.343, -14.632, -13.885,
-13.174, -12.555, -12.023, -17.074, -17.026, -16.644, -16.039, -15.365, -14.651,
-13.900, -13.185, -12.563, -12.028, -17.102, -17.054, -16.672, -16.065, -15.388,
-14.670, -13.916, -13.197, -12.571, -12.033, -17.130, -17.083, -16.700, -16.091,
-15.411, -14.690, -13.931, -13.208, -12.579, -12.039, -17.158, -17.111, -16.729,
-16.117, -15.434, -14.709, -13.947, -13.220, -12.587, -12.044, -17.186, -17.139,
-16.757, -16.143, -15.457, -14.729, -13.963, -13.232, -12.595, -12.049, -17.214,
-17.167, -16.785, -16.169, -15.480, -14.749, -13.979, -13.244, -12.603, -12.055,
 -17.242, -17.196, -16.813, -16.195, -15.503, -14.770, -13.996, -13.256,
-12.612, -12.060, -17.270, -17.224, -16.841, -16.221, -15.527, -14.790, -14.013,
-13.269, -12.621, -12.066, -17.297, -17.252, -16.869, -16.248, -15.550, -14.811,
-14.030, -13.282, -12.630, -12.072, -17.325, -17.280, -16.897, -16.274, -15.574,
-14.832, -14.048, -13.296, -12.639, -12.078, -17.353, -17.308, -16.925, -16.301,
-15.598, -14.854, -14.066, -13.310, -12.649, -12.084, -17.381, -17.336, -16.954,
-16.327, -15.622, -14.875, -14.085, -13.325, -12.659, -12.090, -17.409, -17.364,
-16.982, -16.354, -15.646, -14.897, -14.104, -13.339, -12.670, -12.097, -17.436,
-17.392, -17.010, -16.381, -15.670, -14.919, -14.123, -13.355, -12.681, -12.104,
 -17.464, -17.420, -17.038, -16.407, -15.694, -14.942, -14.143, -13.371,
-12.692, -12.111, -17.492, -17.448, -17.067, -16.434, -15.719, -14.964, -14.163,
-13.387, -12.703, -12.118, -17.520, -17.476, -17.095, -16.461, -15.744, -14.987,
-14.184, -13.403, -12.715, -12.125, -17.547, -17.504, -17.123, -16.489, -15.768,
-15.010, -14.205, -13.420, -12.727, -12.133, -17.575, -17.532, -17.152, -16.516,
-15.793, -15.033, -14.226, -13.437, -12.739, -12.140, -17.603, -17.560, -17.180,
-16.543, -15.818, -15.056, -14.247, -13.454, -12.751, -12.148, -17.631, -17.588,
-17.209, -16.571, -15.843, -15.080, -14.268, -13.471, -12.763, -12.156, -17.658,
-17.616, -17.237, -16.598, -15.868, -15.103, -14.289, -13.488, -12.775, -12.163,
 -17.686, -17.644, -17.265, -16.626, -15.893, -15.126, -14.310, -13.505,
-12.787, -12.170, -17.714, -17.672, -17.294, -16.653, -15.918, -15.149, -14.330,
-13.521, -12.798, -12.178, -17.742, -17.700, -17.322, -16.681, -15.943, -15.172,
-14.351, -13.537, -12.809, -12.185, -17.769, -17.728, -17.351, -16.708, -15.968,
-15.194, -14.370, -13.552, -12.820, -12.192, -17.797, -17.756, -17.379, -16.736,
-15.993, -15.216, -14.389, -13.567, -12.831, -12.199, -17.825, -17.784, -17.408,
-16.763, -16.018, -15.238, -14.408, -13.582, -12.841, -12.205, -17.852, -17.812,
-17.436, -16.791, -16.043, -15.259, -14.426, -13.596, -12.851, -12.212, -17.880,
-17.839, -17.464, -16.818, -16.067, -15.281, -14.444, -13.610, -12.861, -12.218,
 -17.908, -17.867, -17.493, -16.846, -16.092, -15.302, -14.461, -13.623,
-12.870, -12.225, -17.935, -17.895, -17.521, -16.873, -16.116, -15.323, -14.478,
-13.636, -12.879, -12.231, -17.963, -17.923, -17.549, -16.900, -16.140, -15.343,
-14.495, -13.649, -12.889, -12.237, -17.991, -17.950, -17.577, -16.928, -16.165,
-15.364, -14.513, -13.662, -12.898, -12.243, -18.018, -17.978, -17.605, -16.955,
-16.189, -15.385, -14.530, -13.675, -12.907, -12.250, -18.046, -18.006, -17.633,
-16.982, -16.213, -15.406, -14.547, -13.689, -12.917, -12.256, -18.074, -18.033,
-17.661, -17.009, -16.238, -15.427, -14.565, -13.702, -12.927, -12.263, -18.101,
-18.061, -17.689, -17.036, -16.262, -15.449, -14.583, -13.716, -12.936, -12.270,
 -18.129, -18.088, -17.717, -17.062, -16.286, -15.470, -14.601, -13.730,
-12.947, -12.276, -18.157, -18.116, -17.745, -17.089, -16.310, -15.491, -14.619,
-13.744, -12.957, -12.284, -18.184, -18.144, -17.772, -17.116, -16.335, -15.513,
-14.638, -13.759, -12.968, -12.291, -18.212, -18.171, -17.800, -17.143, -16.359,
-15.535, -14.657, -13.774, -12.979, -12.298, -18.239, -18.199, -17.828, -17.169,
-16.383, -15.556, -14.676, -13.790, -12.990, -12.306, -18.267, -18.226, -17.855,
-17.196, -16.407, -15.578, -14.695, -13.806, -13.002, -12.314, -18.295, -18.254,
-17.883, -17.222, -16.431, -15.600, -14.715, -13.822, -13.014, -12.322, -18.322,
-18.281, -17.910, -17.249, -16.456, -15.622, -14.735, -13.839, -13.026, -12.330,
 -18.350, -18.308, -17.937, -17.275, -16.480, -15.645, -14.756, -13.856,
-13.039, -12.339, -18.377, -18.336, -17.965, -17.302, -16.504, -15.667, -14.776,
-13.873, -13.052, -12.347, -18.405, -18.363, -17.992, -17.328, -16.529, -15.689,
-14.797, -13.891, -13.065, -12.356, -18.433, -18.391, -18.019, -17.354, -16.553,
-15.712, -14.818, -13.909, -13.078, -12.364, -18.460, -18.418, -18.047, -17.381,
-16.577, -15.735, -14.840, -13.926, -13.091, -12.373, -18.488, -18.445, -18.074,
-17.407, -16.602, -15.758, -14.861, -13.944, -13.104, -12.381, -18.515, -18.473,
-18.101, -17.434, -16.627, -15.781, -14.882, -13.962, -13.117, -12.390, -18.543,
-18.500, -18.128, -17.461, -16.651, -15.804, -14.904, -13.980, -13.130, -12.398,
 -18.571, -18.527, -18.155, -17.487, -16.676, -15.826, -14.925, -13.997,
-13.142, -12.406, -18.598, -18.555, -18.183, -17.514, -16.701, -15.849, -14.945,
-14.014, -13.155, -12.414, -18.626, -18.582, -18.210, -17.541, -16.726, -15.872,
-14.966, -14.031, -13.166, -12.422, -18.653, -18.609, -18.237, -17.567, -16.751,
-15.894, -14.985, -14.047, -13.178, -12.429, -18.681, -18.637, -18.264, -17.594,
-16.776, -15.917, -15.005, -14.062, -13.189, -12.436, -18.708, -18.664, -18.292,
-17.621, -16.801, -15.939, -15.024, -14.077, -13.200, -12.444, -18.736, -18.691,
-18.319, -17.648, -16.826, -15.961, -15.043, -14.092, -13.210, -12.450, -18.764,
-18.718, -18.346, -17.675, -16.851, -15.983, -15.061, -14.106, -13.220, -12.457,
 -18.791, -18.746, -18.373, -17.702, -16.876, -16.005, -15.080, -14.120,
-13.230, -12.464, -18.819, -18.773, -18.400, -17.729, -16.902, -16.027, -15.098,
-14.134, -13.240, -12.470, -18.846, -18.800, -18.428, -17.756, -16.927, -16.049,
-15.116, -14.148, -13.250, -12.477, -18.874, -18.828, -18.455, -17.783, -16.953,
-16.072, -15.135, -14.162, -13.260, -12.484, -18.901, -18.855, -18.482, -17.810,
-16.978, -16.094, -15.153, -14.176, -13.270, -12.490, -18.929, -18.882, -18.509,
-17.837, -17.004, -16.117, -15.172, -14.190, -13.280, -12.497, -18.957, -18.909,
-18.536, -17.864, -17.029, -16.139, -15.191, -14.205, -13.290, -12.504, -18.984,
-18.937, -18.564, -17.892, -17.055, -16.162, -15.210, -14.220, -13.301, -12.512,
 -19.012, -18.964, -18.591, -17.919, -17.080, -16.185, -15.230, -14.235,
-13.312, -12.519, -19.039, -18.991, -18.618, -17.946, -17.106, -16.208, -15.249,
-14.251, -13.323, -12.527, -19.067, -19.018, -18.645, -17.973, -17.132, -16.230,
-15.269, -14.267, -13.335, -12.535, -19.094, -19.045, -18.672, -18.000, -17.157,
-16.254, -15.290, -14.283, -13.347, -12.543, -19.122, -19.073, -18.699, -18.027,
-17.183, -16.277, -15.310, -14.300, -13.360, -12.552, -19.149, -19.100, -18.726,
-18.054, -17.208, -16.300, -15.331, -14.317, -13.373, -12.561, -19.177, -19.127,
-18.753, -18.081, -17.234, -16.323, -15.352, -14.335, -13.386, -12.570, -19.205,
-19.154, -18.780, -18.108, -17.259, -16.347, -15.373, -14.353, -13.400, -12.579,
 -19.232, -19.181, -18.807, -18.135, -17.284, -16.370, -15.395, -14.372,
-13.414, -12.589, -19.260, -19.209, -18.834, -18.161, -17.310, -16.394, -15.417,
-14.390, -13.428, -12.599, -19.287, -19.236, -18.861, -18.188, -17.335, -16.417,
-15.439, -14.409, -13.443, -12.609, -19.315, -19.263, -18.887, -18.215, -17.361,
-16.441, -15.461, -14.428, -13.457, -12.619, -19.342, -19.290, -18.914, -18.242,
-17.386, -16.464, -15.483, -14.448, -13.472, -12.629, -19.370, -19.317, -18.941,
-18.268, -17.411, -16.488, -15.505, -14.467, -13.487, -12.639, -19.397, -19.344,
-18.968, -18.295, -17.436, -16.511, -15.527, -14.486, -13.501, -12.649, -19.425,
-19.372, -18.994, -18.321, -17.461, -16.535, -15.549, -14.505, -13.516, -12.659,
 -19.452, -19.399, -19.021, -18.348, -17.487, -16.558, -15.571, -14.524,
-13.530, -12.668, -19.480, -19.426, -19.048, -18.374, -17.512, -16.582, -15.593,
-14.543, -13.544, -12.678, -19.507, -19.453, -19.074, -18.401, -17.537, -16.605,
-15.614, -14.561, -13.558, -12.687, -19.535, -19.480, -19.101, -18.427, -17.562,
-16.628, -15.635, -14.579, -13.572, -12.696, -19.563, -19.507, -19.128, -18.454,
-17.587, -16.651, -15.656, -14.596, -13.585, -12.705, -19.590, -19.535, -19.154,
-18.480, -17.612, -16.674, -15.677, -14.614, -13.597, -12.714, -19.618, -19.562,
-19.181, -18.506, -17.637, -16.697, -15.697, -14.630, -13.610, -12.723, -19.645,
-19.589, -19.207, -18.533, -17.663, -16.720, -15.717, -14.647, -13.622, -12.731,
 -19.673, -19.616, -19.234, -18.559, -17.688, -16.743, -15.737, -14.663,
-13.634, -12.739, -19.700, -19.643, -19.261, -18.585, -17.713, -16.766, -15.758,
-14.680, -13.647, -12.748, -19.728, -19.670, -19.287, -18.612, -17.738, -16.789,
-15.778, -14.696, -13.659, -12.756, -19.755, -19.697, -19.314, -18.638, -17.763,
-16.813, -15.798, -14.712, -13.671, -12.764, -19.783, -19.724, -19.340, -18.664,
-17.788, -16.836, -15.818, -14.729, -13.683, -12.772, -19.810, -19.752, -19.367,
-18.690, -17.814, -16.859, -15.839, -14.746, -13.695, -12.781, -19.838, -19.779,
-19.393, -18.717, -17.839, -16.882, -15.860, -14.762, -13.708, -12.789, -19.865,
-19.806, -19.420, -18.743, -17.864, -16.906, -15.880, -14.779, -13.720, -12.798,
 -19.893, -19.833, -19.446, -18.769, -17.889, -16.929, -15.901, -14.797,
-13.733, -12.807, -19.920, -19.860, -19.473, -18.795, -17.914, -16.952, -15.922,
-14.814, -13.747, -12.816, -19.948, -19.887, -19.499, -18.821, -17.940, -16.976,
-15.943, -14.832, -13.760, -12.826, -19.975, -19.914, -19.525, -18.848, -17.965,
-16.999, -15.965, -14.850, -13.774, -12.835, -20.003, -19.941, -19.552, -18.874,
-17.990, -17.023, -15.986, -14.869, -13.788, -12.845, -20.030, -19.969, -19.578,
-18.900, -18.015, -17.046, -16.008, -14.887, -13.803, -12.855, -20.058, -19.996,
-19.605, -18.926, -18.041, -17.070, -16.030, -14.906, -13.818, -12.865, -20.085,
-20.023, -19.631, -18.952, -18.066, -17.094, -16.052, -14.926, -13.833, -12.876,
 -20.113, -20.050, -19.657, -18.978, -18.091, -17.117, -16.074, -14.945,
-13.848, -12.886, -20.140, -20.077, -19.684, -19.004, -18.116, -17.141, -16.096,
-14.965, -13.863, -12.897, -20.168, -20.104, -19.710, -19.030, -18.141, -17.164,
-16.119, -14.984, -13.879, -12.908, -20.195, -20.131, -19.737, -19.056, -18.166,
-17.188, -16.141, -15.004, -13.895, -12.919, -20.223, -20.158, -19.763, -19.082,
-18.192, -17.212, -16.163, -15.024, -13.910, -12.930, -20.250, -20.186, -19.789,
-19.108, -18.217, -17.235, -16.186, -15.044, -13.926, -12.941, -20.278, -20.213,
-19.816, -19.134, -18.242, -17.259, -16.208, -15.064, -13.942, -12.952, -20.305,
-20.240, -19.842, -19.160, -18.267, -17.282, -16.231, -15.084, -13.957, -12.963,
 -20.333, -20.267, -19.868, -19.186, -18.292, -17.305, -16.253, -15.104,
-13.973, -12.973, -20.360, -20.294, -19.894, -19.212, -18.317, -17.329, -16.275,
-15.123, -13.988, -12.984, -20.388, -20.321, -19.921, -19.238, -18.341, -17.352,
-16.297, -15.142, -14.003, -12.994, -20.415, -20.348, -19.947, -19.263, -18.366,
-17.376, -16.319, -15.161, -14.018, -13.005, -20.443, -20.375, -19.973, -19.289,
-18.391, -17.399, -16.340, -15.179, -14.032, -13.015, -20.470, -20.403, -20.000,
-19.315, -18.416, -17.422, -16.362, -15.198, -14.046, -13.025, -20.498, -20.430,
-20.026, -19.341, -18.441, -17.446, -16.383, -15.216, -14.060, -13.034, -20.525,
-20.457, -20.052, -19.366, -18.465, -17.469, -16.405, -15.234, -14.074, -13.044,
 -20.553, -20.484, -20.078, -19.392, -18.490, -17.492, -16.426, -15.252,
-14.088, -13.054, -20.580, -20.511, -20.104, -19.417, -18.515, -17.515, -16.447,
-15.269, -14.101, -13.063, -20.608, -20.538, -20.131, -19.443, -18.539, -17.538,
-16.469, -15.287, -14.115, -13.073, -20.635, -20.565, -20.157, -19.469, -18.564,
-17.562, -16.490, -15.305, -14.129, -13.083, -20.663, -20.592, -20.183, -19.494,
-18.589, -17.585, -16.512, -15.323, -14.142, -13.092, -20.690, -20.620, -20.209,
-19.520, -18.613, -17.608, -16.533, -15.341, -14.156, -13.102, -20.718, -20.647,
-20.236, -19.545, -18.638, -17.631, -16.555, -15.359, -14.170, -13.112, -20.745,
-20.674, -20.262, -19.570, -18.662, -17.654, -16.576, -15.377, -14.184, -13.122,
 -20.773, -20.701, -20.288, -19.596, -18.687, -17.677, -16.598, -15.396,
-14.199, -13.132, -20.800, -20.728, -20.314, -19.621, -18.712, -17.700, -16.620,
-15.414, -14.213, -13.142, -20.828, -20.755, -20.340, -19.647, -18.736, -17.723,
-16.642, -15.433, -14.228, -13.153, -20.855, -20.782, -20.366, -19.672, -18.761,
-17.746, -16.664, -15.452, -14.243, -13.164, -20.883, -20.809, -20.393, -19.698,
-18.785, -17.770, -16.686, -15.471, -14.258, -13.174, -20.910, -20.837, -20.419,
-19.723, -18.810, -17.793, -16.708, -15.491, -14.274, -13.185, -20.938, -20.864,
-20.445, -19.748, -18.834, -17.816, -16.730, -15.510, -14.290, -13.197, -20.965,
-20.891, -20.471, -19.774, -18.859, -17.839, -16.752, -15.530, -14.306, -13.208,
 -20.992, -20.918, -20.497, -19.799, -18.883, -17.862, -16.774, -15.550,
-14.322, -13.220, -21.020, -20.945, -20.524, -19.824, -18.908, -17.885, -16.797,
-15.570, -14.338, -13.231, -21.047, -20.972, -20.550, -19.850, -18.932, -17.908,
-16.819, -15.591, -14.354, -13.243, -21.075, -20.999, -20.576, -19.875, -18.957,
-17.932, -16.841, -15.611, -14.371, -13.255, -21.102, -21.027, -20.602, -19.900,
-18.982, -17.955, -16.864, -15.631, -14.387, -13.266, -21.130, -21.054, -20.628,
-19.926, -19.006, -17.978, -16.886, -15.651, -14.404, -13.278, -21.157, -21.081,
-20.654, -19.951, -19.031, -18.001, -16.908, -15.671, -14.420, -13.289, -21.185,
-21.108, -20.680, -19.976, -19.055, -18.024, -16.931, -15.691, -14.436, -13.301,
 -21.212, -21.135, -20.707, -20.002, -19.080, -18.048, -16.953, -15.711,
-14.452, -13.312, -21.240, -21.162, -20.733, -20.027, -19.104, -18.071, -16.976,
-15.731, -14.467, -13.323, -21.267, -21.189, -20.759, -20.052, -19.129, -18.094,
-16.998, -15.750, -14.482, -13.334, -21.294, -21.217, -20.785, -20.077, -19.153,
-18.117, -17.020, -15.769, -14.497, -13.345, -21.322, -21.244, -20.811, -20.103,
-19.178, -18.140, -17.043, -15.788, -14.512, -13.355, -21.349, -21.271, -20.837,
-20.128, -19.202, -18.164, -17.065, -15.807, -14.527, -13.366, -21.377, -21.298,
-20.864, -20.153, -19.227, -18.187, -17.088, -15.825, -14.541, -13.376, -21.404,
-21.325, -20.890, -20.178, -19.251, -18.210, -17.110, -15.844, -14.556, -13.386,
 -21.432, -21.352, -20.916, -20.204, -19.276, -18.233, -17.132, -15.862,
-14.570, -13.397, -21.459, -21.379, -20.942, -20.229, -19.300, -18.256, -17.155,
-15.881, -14.585, -13.407, -21.487, -21.407, -20.968, -20.254, -19.325, -18.280,
-17.177, -15.900, -14.599, -13.417, -21.514, -21.434, -20.994, -20.279, -19.349,
-18.303, -17.199, -15.918, -14.613, -13.427, -21.541, -21.461, -21.020, -20.304,
-19.374, -18.326, -17.222, -15.937, -14.628, -13.438, -21.569, -21.488, -21.047,
-20.329, -19.398, -18.349, -17.244, -15.956, -14.643, -13.449, -21.596, -21.515,
-21.073, -20.354, -19.422, -18.372, -17.266, -15.975, -14.658, -13.459, -21.624,
-21.542, -21.099, -20.380, -19.447, -18.395, -17.288, -15.994, -14.673, -13.470,
 -21.651, -21.569, -21.125, -20.405, -19.471, -18.418, -17.311, -16.014,
-14.688, -13.482, -21.679, -21.597, -21.151, -20.430, -19.495, -18.441, -17.333,
-16.033, -14.704, -13.493, -21.706, -21.624, -21.177, -20.455, -19.519, -18.464,
-17.355, -16.053, -14.720, -13.505, -21.733, -21.651, -21.203, -20.480, -19.544,
-18.487, -17.377, -16.073, -14.736, -13.516, -21.761, -21.678, -21.230, -20.505,
-19.568, -18.510, -17.399, -16.094, -14.753, -13.528, -21.788, -21.705, -21.256,
-20.530, -19.592, -18.533, -17.421, -16.114, -14.769, -13.540, -21.816, -21.732,
-21.282, -20.555, -19.616, -18.556, -17.443, -16.135, -14.786, -13.553, -21.843,
-21.759, -21.308, -20.580, -19.640, -18.579, -17.465, -16.155, -14.803, -13.565,
 -21.870, -21.787, -21.334, -20.605, -19.665, -18.602, -17.487, -16.176,
-14.820, -13.578, -21.898, -21.814, -21.360, -20.630, -19.689, -18.625, -17.509,
-16.197, -14.838, -13.591, -21.925, -21.841, -21.386, -20.655, -19.713, -18.648,
-17.531, -16.218, -14.855, -13.603, -21.953, -21.868, -21.413, -20.680, -19.737,
-18.671, -17.553, -16.240, -14.873, -13.616, -21.980, -21.895, -21.439, -20.705,
-19.761, -18.694, -17.574, -16.261, -14.890, -13.629, -22.007, -21.922, -21.465,
-20.730, -19.785, -18.717, -17.596, -16.282, -14.907, -13.641, -22.035, -21.950,
-21.491, -20.755, -19.809, -18.740, -17.618, -16.303, -14.925, -13.654, -22.062,
-21.977, -21.517, -20.780, -19.833, -18.762, -17.640, -16.324, -14.942, -13.667,
 -22.090, -22.004, -21.543, -20.805, -19.857, -18.785, -17.662, -16.345,
-14.959, -13.679, -22.117, -22.031, -21.569, -20.830, -19.881, -18.808, -17.683,
-16.366, -14.975, -13.691, -22.144, -22.058, -21.596, -20.855, -19.905, -18.831,
-17.705, -16.386, -14.992, -13.703, -22.172, -22.085, -21.622, -20.880, -19.929,
-18.854, -17.727, -16.407, -15.008, -13.715, -22.199, -22.113, -21.648, -20.905,
-19.953, -18.877, -17.748, -16.427, -15.024, -13.727, -22.226, -22.140, -21.674,
-20.930, -19.977, -18.900, -17.770, -16.447, -15.041, -13.739, -22.254, -22.167,
-21.700, -20.955, -20.001, -18.922, -17.792, -16.468, -15.057, -13.751, -22.281,
-22.194, -21.726, -20.980, -20.025, -18.945, -17.813, -16.488, -15.072, -13.762,
 -22.309, -22.221, -21.753, -21.004, -20.049, -18.968, -17.835, -16.508,
-15.088, -13.774, -22.336, -22.248, -21.779, -21.029, -20.073, -18.991, -17.857,
-16.528, -15.104, -13.786, -22.363, -22.276, -21.805, -21.054, -20.097, -19.014,
-17.878, -16.549, -15.120, -13.797, -22.391, -22.303, -21.831, -21.079, -20.121,
-19.036, -17.900, -16.569, -15.136, -13.809, -22.418, -22.330, -21.857, -21.104,
-20.145, -19.059, -17.921, -16.589, -15.152, -13.821, -22.445, -22.357, -21.883,
-21.129, -20.169, -19.082, -17.943, -16.610, -15.169, -13.833, -22.473, -22.384,
-21.910, -21.154, -20.193, -19.105, -17.964, -16.630, -15.185, -13.845, -22.500,
-22.411, -21.936, -21.179, -20.217, -19.127, -17.986, -16.651, -15.202, -13.858,
 -22.527, -22.438, -21.962, -21.204, -20.240, -19.150, -18.007, -16.672,
-15.219, -13.870, -22.555, -22.466, -21.988, -21.228, -20.264, -19.173, -18.029,
-16.693, -15.236, -13.883, -22.582, -22.493, -22.014, -21.253, -20.288, -19.196,
-18.050, -16.714, -15.253, -13.896, -22.610, -22.520, -22.040, -21.278, -20.312,
-19.218, -18.072, -16.735, -15.271, -13.909, -22.637, -22.547, -22.067, -21.303,
-20.336, -19.241, -18.093, -16.756, -15.288, -13.922, -22.664, -22.574, -22.093,
-21.328, -20.359, -19.264, -18.114, -16.777, -15.306, -13.936, -22.692, -22.601,
-22.119, -21.353, -20.383, -19.286, -18.136, -16.799, -15.324, -13.949, -22.719,
-22.629, -22.145, -21.378, -20.407, -19.309, -18.157, -16.820, -15.342, -13.963,
 -22.746, -22.656, -22.171, -21.403, -20.431, -19.332, -18.179, -16.842,
-15.361, -13.977, -22.774, -22.683, -22.197, -21.427, -20.454, -19.354, -18.200,
-16.863, -15.379, -13.990, -22.801, -22.710, -22.224, -21.452, -20.478, -19.377,
-18.221, -16.885, -15.397, -14.004, -22.828, -22.737, -22.250, -21.477, -20.502,
-19.399, -18.242, -16.906, -15.416, -14.018, -22.856, -22.764, -22.276, -21.502,
-20.525, -19.422, -18.264, -16.928, -15.434, -14.032, -22.883, -22.792, -22.302,
-21.527, -20.549, -19.444, -18.285, -16.950, -15.452, -14.046, -22.910, -22.819,
-22.328, -21.552, -20.573, -19.467, -18.306, -16.971, -15.471, -14.059, -22.938,
-22.846, -22.354, -21.576, -20.596, -19.489, -18.327, -16.993, -15.489, -14.073,
 -22.965, -22.873, -22.381, -21.601, -20.620, -19.512, -18.349, -17.015,
-15.507, -14.086, -22.992, -22.900, -22.407, -21.626, -20.644, -19.534, -18.370,
-17.036, -15.524, -14.099, -23.019, -22.927, -22.433, -21.651, -20.667, -19.557,
-18.391, -17.058, -15.542, -14.113, -23.047, -22.954, -22.459, -21.676, -20.691,
-19.579, -18.412, -17.080, -15.559, -14.126, -23.074, -22.982, -22.485, -21.701,
-20.715, -19.602, -18.433, -17.101, -15.577, -14.138, -23.101, -23.009, -22.512,
-21.725, -20.738, -19.624, -18.454, -17.123, -15.594, -14.151, -23.129, -23.036,
-22.538, -21.750, -20.762, -19.646, -18.475, -17.144, -15.611, -14.164, -23.156,
-23.063, -22.564, -21.775, -20.785, -19.669, -18.496, -17.166, -15.628, -14.177,
 -23.183, -23.090, -22.590, -21.800, -20.809, -19.691, -18.517, -17.187,
-15.645, -14.189, -23.211, -23.117, -22.616, -21.825, -20.832, -19.713, -18.538,
-17.209, -15.663, -14.202, -23.238, -23.145, -22.643, -21.849, -20.856, -19.736,
-18.559, -17.230, -15.680, -14.215, -23.265, -23.172, -22.669, -21.874, -20.879,
-19.758, -18.580, -17.252, -15.697, -14.228, -23.292, -23.199, -22.695, -21.899,
-20.903, -19.780, -18.601, -17.273, -15.715, -14.241, -23.320, -23.226, -22.721,
-21.924, -20.926, -19.803, -18.622, -17.294, -15.732, -14.254, -23.347, -23.253,
-22.747, -21.949, -20.950, -19.825, -18.643, -17.316, -15.750, -14.267, -23.374,
-23.280, -22.774, -21.973, -20.973, -19.847, -18.664, -17.337, -15.768, -14.280,
 -23.402, -23.307, -22.800, -21.998, -20.997, -19.870, -18.685, -17.358,
-15.786, -14.294, -23.429, -23.335, -22.826, -22.023, -21.020, -19.892, -18.706,
-17.379, -15.804, -14.307, -23.456, -23.362, -22.852, -22.048, -21.044, -19.914,
-18.727, -17.400, -15.822, -14.321, -23.483, -23.389, -22.878, -22.073, -21.067,
-19.936, -18.748, -17.421, -15.841, -14.335, -23.511, -23.416, -22.905, -22.097,
-21.091, -19.958, -18.769, -17.442, -15.859, -14.349, -23.538, -23.443, -22.931,
-22.122, -21.114, -19.981, -18.790, -17.463, -15.878, -14.363, -23.565, -23.470,
-22.957, -22.147, -21.138, -20.003, -18.810, -17.484, -15.897, -14.378, -23.592,
-23.497, -22.983, -22.172, -21.161, -20.025, -18.831, -17.505, -15.916, -14.392,
 -23.620, -23.525, -23.010, -22.197, -21.184, -20.047, -18.852, -17.526,
-15.936, -14.407, -23.647, -23.552, -23.036, -22.221, -21.208, -20.069, -18.873,
-17.547, -15.955, -14.421, -23.674, -23.579, -23.062, -22.246, -21.231, -20.091,
-18.894, -17.567, -15.974, -14.436, -23.701, -23.606, -23.088, -22.271, -21.255,
-20.114, -18.914, -17.588, -15.994, -14.450, -23.729, -23.633, -23.114, -22.296,
-21.278, -20.136, -18.935, -17.609, -16.013, -14.465, -23.756, -23.660, -23.141,
-22.321, -21.301, -20.158, -18.956, -17.629, -16.032, -14.479, -23.783, -23.687,
-23.167, -22.345, -21.325, -20.180, -18.977, -17.650, -16.052, -14.493, -23.810,
-23.715, -23.193, -22.370, -21.348, -20.202, -18.997, -17.671, -16.071, -14.508,
 -23.838, -23.742, -23.219, -22.395, -21.371, -20.224, -19.018, -17.691,
-16.090, -14.522, -23.865, -23.769, -23.245, -22.420, -21.395, -20.246, -19.039,
-17.711, -16.109, -14.535, -23.892, -23.796, -23.272, -22.445, -21.418, -20.268,
-19.059, -17.732, -16.127, -14.549, -23.919, -23.823, -23.298, -22.469, -21.441,
-20.290, -19.080, -17.752, -16.146, -14.563, -23.947, -23.850, -23.324, -22.494,
-21.465, -20.312, -19.100, -17.773, -16.165, -14.576, -23.974, -23.877, -23.350,
-22.519, -21.488, -20.334, -19.121, -17.793, -16.183, -14.590, -24.001, -23.904,
-23.377, -22.544, -21.511, -20.356, -19.142, -17.813, -16.202, -14.603, -24.028,
-23.932, -23.403, -22.569, -21.535, -20.378, -19.162, -17.833, -16.220, -14.617,
 -24.055, -23.959, -23.429, -22.593, -21.558, -20.400, -19.183, -17.853,
-16.238, -14.630, -24.083, -23.986, -23.455, -22.618, -21.581, -20.422, -19.203,
-17.873, -16.257, -14.643, -24.110, -24.013, -23.482, -22.643, -21.605, -20.444,
-19.224, -17.893, -16.275, -14.657, -24.137, -24.040, -23.508, -22.668, -21.628,
-20.466, -19.244, -17.913, -16.294, -14.671, -24.164, -24.067, -23.534, -22.693,
-21.651, -20.488, -19.265, -17.933, -16.313, -14.684, -24.191, -24.094, -23.560,
-22.717, -21.675, -20.510, -19.285, -17.953, -16.331, -14.698, -24.219, -24.121,
-23.586, -22.742, -21.698, -20.532, -19.306, -17.973, -16.350, -14.712, -24.246,
-24.149, -23.613, -22.767, -21.721, -20.553, -19.326, -17.993, -16.369, -14.726,
 -24.273, -24.176, -23.639, -22.792, -21.744, -20.575, -19.347, -18.013,
-16.388, -14.741, -24.300, -24.203, -23.665, -22.817, -21.768, -20.597, -19.367,
-18.033, -16.408, -14.755, -24.327, -24.230, -23.691, -22.841, -21.791, -20.619,
-19.387, -18.052, -16.427, -14.770, -24.355, -24.257, -23.718, -22.866, -21.814,
-20.641, -19.408, -18.072, -16.446, -14.784, -24.382, -24.284, -23.744, -22.891,
-21.838, -20.663, -19.428, -18.092, -16.466, -14.799, -24.409, -24.311, -23.770,
-22.916, -21.861, -20.684, -19.448, -18.111, -16.485, -14.814, -24.436, -24.338,
-23.796, -22.941, -21.884, -20.706, -19.469, -18.131, -16.505, -14.830, -24.463,
-24.365, -23.823, -22.965, -21.907, -20.728, -19.489, -18.150, -16.525, -14.845,
 -24.491, -24.392, -23.849, -22.990, -21.931, -20.750, -19.509, -18.170,
-16.545, -14.860, -24.518, -24.420, -23.875, -23.015, -21.954, -20.772, -19.530,
-18.189, -16.565, -14.876, -24.545, -24.447, -23.901, -23.040, -21.977, -20.793,
-19.550, -18.209, -16.585, -14.891, -24.572, -24.474, -23.928, -23.065, -22.000,
-20.815, -19.570, -18.228, -16.605, -14.907, -24.599, -24.501, -23.954, -23.089,
-22.023, -20.837, -19.590, -18.247, -16.625, -14.922, -24.626, -24.528, -23.980,
-23.114, -22.047, -20.859, -19.610, -18.267, -16.645, -14.938, -24.653, -24.555,
-24.006, -23.139, -22.070, -20.880, -19.631, -18.286, -16.665, -14.953, -24.681,
-24.582, -24.033, -23.164, -22.093, -20.902, -19.651, -18.305, -16.685, -14.968,
 -24.708, -24.609, -24.059, -23.189, -22.116, -20.924, -19.671, -18.324,
-16.705, -14.983, -24.735, -24.636, -24.085, -23.213, -22.140, -20.945, -19.691,
-18.343, -16.726, -14.998, -24.762, -24.663, -24.111, -23.238, -22.163, -20.967,
-19.711, -18.363, -16.746, -15.013, -24.789, -24.691, -24.138, -23.263, -22.186,
-20.989, -19.731, -18.382, -16.766, -15.028, -24.816, -24.718, -24.164, -23.288,
-22.209, -21.010, -19.751, -18.401, -16.786, -15.043, -24.844, -24.745, -24.190,
-23.313, -22.232, -21.032, -19.771, -18.420, -16.807, -15.058, -24.871, -24.772,
-24.216, -23.338, -22.256, -21.054, -19.792, -18.439, -16.827, -15.073, -24.898,
-24.799, -24.243, -23.362, -22.279, -21.075, -19.812, -18.458, -16.847, -15.088,
 -24.925, -24.826, -24.269, -23.387, -22.302, -21.097, -19.832, -18.477,
-16.867, -15.103, -24.952, -24.853, -24.295, -23.412, -22.325, -21.119, -19.852,
-18.496, -16.887, -15.118, -24.979, -24.880, -24.321, -23.437, -22.348, -21.140,
-19.872, -18.514, -16.907, -15.133, -25.006, -24.907, -24.347, -23.462, -22.372,
-21.162, -19.892, -18.533, -16.928, -15.148, -25.033, -24.934, -24.374, -23.486,
-22.395, -21.183, -19.912, -18.552, -16.948, -15.163, -25.061, -24.961, -24.400,
-23.511, -22.418, -21.205, -19.932, -18.571, -16.968, -15.179, -25.088, -24.988,
-24.426, -23.536, -22.441, -21.227, -19.951, -18.589, -16.988, -15.195, -25.115,
-25.015, -24.452, -23.561, -22.464, -21.248, -19.971, -18.608, -17.008, -15.210,
 -25.142, -25.043, -24.479, -23.586, -22.488, -21.270, -19.991, -18.627,
-17.027, -15.226, -25.169, -25.070, -24.505, -23.611, -22.511, -21.291, -20.011,
-18.645, -17.047, -15.242, -25.196, -25.097, -24.531, -23.635, -22.534, -21.313,
-20.031, -18.664, -17.067, -15.259, -25.223, -25.124, -24.557, -23.660, -22.557,
-21.334, -20.051, -18.683, -17.087, -15.275, -25.250, -25.151, -24.584, -23.685,
-22.580, -21.356, -20.071, -18.701, -17.106, -15.292, -25.277, -25.178, -24.610,
-23.710, -22.603, -21.378, -20.091, -18.720, -17.126, -15.308, -25.304, -25.205,
-24.636, -23.735, -22.627, -21.399, -20.111, -18.738, -17.145, -15.325, -25.332,
-25.232, -24.662, -23.760, -22.650, -21.421, -20.130, -18.757, -17.165, -15.342,
 -25.359, -25.259, -24.689, -23.784, -22.673, -21.442, -20.150, -18.775,
-17.184, -15.359, -25.386, -25.286, -24.715, -23.809, -22.696, -21.464, -20.170,
-18.793, -17.204, -15.376, -25.413, -25.313, -24.741, -23.834, -22.719, -21.485,
-20.190, -18.812, -17.223, -15.393, -25.440, -25.340, -24.767, -23.859, -22.742,
-21.507, -20.210, -18.830, -17.242, -15.410, -25.467, -25.367, -24.794, -23.884,
-22.766, -21.528, -20.229, -18.848, -17.261, -15.427, -25.494, -25.394, -24.820,
-23.908, -22.789, -21.550, -20.249, -18.866, -17.280, -15.444, -25.521, -25.421,
-24.846, -23.933, -22.812, -21.571, -20.269, -18.885, -17.299, -15.461, -25.548,
-25.448, -24.872, -23.958, -22.835, -21.593, -20.289, -18.903, -17.318, -15.478,
 -25.575, -25.475, -24.899, -23.983, -22.858, -21.614, -20.308, -18.921,
-17.337, -15.494, -25.602, -25.502, -24.925, -24.008, -22.881, -21.636, -20.328,
-18.939, -17.356, -15.511, -25.629, -25.529, -24.951, -24.033, -22.905, -21.657,
-20.348, -18.957, -17.374, -15.528, -25.656, -25.556, -24.977, -24.057, -22.928,
-21.679, -20.367, -18.975, -17.393, -15.545, -25.684, -25.583, -25.004, -24.082,
-22.951, -21.700, -20.387, -18.993, -17.412, -15.561, -25.711, -25.611, -25.030,
-24.107, -22.974, -21.722, -20.407, -19.011, -17.430, -15.578, -25.738, -25.638,
-25.056, -24.132, -22.997, -21.743, -20.426, -19.029, -17.449, -15.595, -25.765,
-25.665, -25.082, -24.157, -23.020, -21.765, -20.446, -19.047, -17.467, -15.611,
 -25.792, -25.692, -25.109, -24.182, -23.044, -21.786, -20.466, -19.065,
-17.486, -15.628, -25.819, -25.719, -25.135, -24.206, -23.067, -21.808, -20.485,
-19.083, -17.504, -15.645, -25.846, -25.746, -25.161, -24.231, -23.090, -21.829,
-20.505, -19.101, -17.522, -15.662, -25.873, -25.773, -25.187, -24.256, -23.113,
-21.851, -20.524, -19.119, -17.540, -15.678, -25.900, -25.800, -25.214, -24.281,
-23.136, -21.872, -20.544, -19.137, -17.559, -15.695, -25.927, -25.827, -25.240,
-24.306, -23.159, -21.893, -20.564, -19.155, -17.577, -15.712, -25.954, -25.854,
-25.266, -24.331, -23.182, -21.915, -20.583, -19.173, -17.595, -15.730, -25.981,
-25.881, -25.292, -24.356, -23.206, -21.936, -20.603, -19.190, -17.613, -15.747,
 -26.008, -25.908, -25.318, -24.380, -23.229, -21.958, -20.622, -19.208,
-17.631, -15.764, -26.035, -25.935, -25.345, -24.405, -23.252, -21.979, -20.642,
-19.226, -17.649, -15.782, -26.062, -25.962, -25.371, -24.430, -23.275, -22.001,
-20.662, -19.244, -17.666, -15.799, -26.089, -25.989, -25.397, -24.455, -23.298,
-22.022, -20.681, -19.261, -17.684, -15.817, -26.116, -26.016, -25.423, -24.480,
-23.321, -22.044, -20.701, -19.279, -17.702, -15.834, -26.143, -26.043, -25.450,
-24.505, -23.344, -22.065, -20.720, -19.297, -17.720, -15.852, -26.170, -26.070,
-25.476, -24.529, -23.368, -22.086, -20.740, -19.314, -17.737, -15.870, -26.197,
-26.097, -25.502, -24.554, -23.391, -22.108, -20.759, -19.332, -17.755, -15.888,
 -26.224, -26.124, -25.528, -24.579, -23.414, -22.129, -20.779, -19.349,
-17.772, -15.906, -26.251, -26.151, -25.555, -24.604, -23.437, -22.151, -20.798,
-19.367, -17.789, -15.924, -26.278, -26.178, -25.581, -24.629, -23.460, -22.172,
-20.818, -19.384, -17.807, -15.942, -26.305, -26.205, -25.607, -24.654, -23.483,
-22.194, -20.837, -19.402, -17.824, -15.960, -26.332, -26.232, -25.633, -24.678,
-23.506, -22.215, -20.857, -19.419, -17.841, -15.978, -26.359, -26.259, -25.660,
-24.703, -23.530, -22.236, -20.876, -19.437, -17.858, -15.996, -26.386, -26.286,
-25.686, -24.728, -23.553, -22.258, -20.896, -19.454, -17.875, -16.015, -26.413,
-26.313, -25.712, -24.753, -23.576, -22.279, -20.915, -19.472, -17.892, -16.033,
 -26.440, -26.340, -25.738, -24.778, -23.599, -22.301, -20.935, -19.489,
-17.909, -16.051, -26.467, -26.367, -25.764, -24.803, -23.622, -22.322, -20.955,
-19.506, -17.926, -16.069, -26.494, -26.394, -25.791, -24.827, -23.645, -22.343,
-20.974, -19.523, -17.942, -16.087, -26.521, -26.421, -25.817, -24.852, -23.668,
-22.365, -20.994, -19.541, -17.959, -16.105, -26.548, -26.448, -25.843, -24.877,
-23.692, -22.386, -21.013, -19.558, -17.975, -16.123, -26.575, -26.474, -25.869,
-24.902, -23.715, -22.408, -21.033, -19.575, -17.992, -16.141, -26.602, -26.501,
-25.896, -24.927, -23.738, -22.429, -21.052, -19.592, -18.008, -16.159, -26.629,
-26.528, -25.922, -24.952, -23.761, -22.450, -21.072, -19.609, -18.024, -16.177,
 -26.656, -26.555, -25.948, -24.977, -23.784, -22.472, -21.091, -19.627,
-18.040, -16.195, -26.683, -26.582, -25.974, -25.001, -23.807, -22.493, -21.111,
-19.644, -18.057, -16.213, -26.710, -26.609, -26.000, -25.026, -23.830, -22.514,
-21.130, -19.661, -18.073, -16.230, -26.737, -26.636, -26.027, -25.051, -23.853,
-22.536, -21.150, -19.678, -18.089, -16.248, -26.764, -26.663, -26.053, -25.076,
-23.877, -22.557, -21.169, -19.695, -18.105, -16.266, -26.791, -26.690, -26.079,
-25.101, -23.900, -22.579, -21.189, -19.712, -18.121, -16.284, -26.818, -26.717,
-26.105, -25.126, -23.923, -22.600, -21.209, -19.729, -18.137, -16.302, -26.845,
-26.744, -26.131, -25.150, -23.946, -22.621, -21.228, -19.746, -18.153, -16.320,
 -26.872, -26.771, -26.158, -25.175, -23.969, -22.643, -21.248, -19.763,
-18.169, -16.337, -26.899, -26.798, -26.184, -25.200, -23.992, -22.664, -21.267,
-19.780, -18.185, -16.355, -26.926, -26.825, -26.210, -25.225, -24.015, -22.685,
-21.287, -19.797, -18.200, -16.373, -26.953, -26.852, -26.236, -25.250, -24.039,
-22.707, -21.306, -19.814, -18.216, -16.391, -26.979, -26.879, -26.263, -25.275,
-24.062, -22.728, -21.326, -19.831, -18.232, -16.409, -27.006, -26.906, -26.289,
-25.300, -24.085, -22.749, -21.346, -19.848, -18.248, -16.426, -27.033, -26.933,
-26.315, -25.324, -24.108, -22.771, -21.365, -19.865, -18.264, -16.444, -27.060,
-26.960, -26.341, -25.349, -24.131, -22.792, -21.385, -19.881, -18.280, -16.462,
 -27.087, -26.987, -26.367, -25.374, -24.154, -22.813, -21.404, -19.898,
-18.296, -16.480, -27.114, -27.013, -26.394, -25.399, -24.177, -22.835, -21.424,
-19.915, -18.311, -16.497, -27.141, -27.040, -26.420, -25.424, -24.200, -22.856,
-21.444, -19.932, -18.327, -16.515, -27.168, -27.067, -26.446, -25.449, -24.224,
-22.878, -21.463, -19.949, -18.343, -16.533, -27.195, -27.094, -26.472, -25.473,
-24.247, -22.899, -21.483, -19.966, -18.359, -16.550, -27.222, -27.121, -26.498,
-25.498, -24.270, -22.920, -21.503, -19.983, -18.375, -16.568, -27.249, -27.148,
-26.525, -25.523, -24.293, -22.942, -21.522, -20.000, -18.390, -16.585, -27.276,
-27.175, -26.551, -25.548, -24.316, -22.963, -21.542, -20.017, -18.406, -16.603,
 -27.303, -27.202, -26.577, -25.573, -24.339, -22.984, -21.561, -20.034,
-18.422, -16.620, -27.329, -27.229, -26.603, -25.598, -24.362, -23.006, -21.581,
-20.051, -18.437, -16.638, -27.356, -27.256, -26.629, -25.622, -24.385, -23.027,
-21.601, -20.067, -18.453, -16.655, -27.383, -27.283, -26.656, -25.647, -24.409,
-23.048, -21.620, -20.084, -18.468, -16.672, -27.410, -27.310, -26.682, -25.672,
-24.432, -23.069, -21.640, -20.101, -18.484, -16.689, -27.437, -27.336, -26.708,
-25.697, -24.455, -23.091, -21.660, -20.118, -18.499, -16.706, -27.464, -27.363,
-26.734, -25.722, -24.478, -23.112, -21.679, -20.134, -18.514, -16.723, -27.491,
-27.390, -26.760, -25.747, -24.501, -23.133, -21.699, -20.151, -18.529, -16.739,
 -27.518, -27.417, -26.786, -25.771, -24.524, -23.155, -21.718, -20.168,
-18.544, -16.755, -27.545, -27.444, -26.813, -25.796, -24.547, -23.176, -21.738,
-20.185, -18.559, -16.771, -27.572, -27.471, -26.839, -25.821, -24.570, -23.197,
-21.758, -20.201, -18.574, -16.787, -27.598, -27.498, -26.865, -25.846, -24.594,
-23.219, -21.777, -20.218, -18.588, -16.803, -27.625, -27.525, -26.891, -25.871,
-24.617, -23.240, -21.797, -20.235, -18.603, -16.818, -27.652, -27.552, -26.917,
-25.896, -24.640, -23.261, -21.817, -20.251, -18.617, -16.833, -27.679, -27.578,
-26.944, -25.921, -24.663, -23.283, -21.836, -20.268, -18.631, -16.849, -27.706,
-27.605, -26.970, -25.945, -24.686, -23.304, -21.856, -20.284, -18.646, -16.864,
 -27.733, -27.632, -26.996, -25.970, -24.709, -23.325, -21.875, -20.301,
-18.660, -16.879, -27.760, -27.659, -27.022, -25.995, -24.732, -23.347, -21.895,
-20.317, -18.674, -16.893, -27.787, -27.686, -27.048, -26.020, -24.755, -23.368,
-21.915, -20.334, -18.688, -16.908, -27.813, -27.713, -27.074, -26.045, -24.778,
-23.389, -21.934, -20.350, -18.702, -16.923, -27.840, -27.740, -27.101, -26.070,
-24.802, -23.410, -21.954, -20.367, -18.716, -16.938, -27.867, -27.767, -27.127,
-26.094, -24.825, -23.432, -21.973, -20.384, -18.730, -16.952, -27.894, -27.793,
-27.153, -26.119, -24.848, -23.453, -21.993, -20.400, -18.744, -16.967, -27.921,
-27.820, -27.179, -26.144, -24.871, -23.474, -22.013, -20.417, -18.758, -16.982,
 -27.948, -27.847, -27.205, -26.169, -24.894, -23.496, -22.032, -20.433,
-18.772, -16.996, -27.975, -27.874, -27.231, -26.194, -24.917, -23.517, -22.052,
-20.450, -18.787, -17.011, -28.001, -27.901, -27.258, -26.218, -24.940, -23.538,
-22.071, -20.467, -18.801, -17.026, -28.028, -27.928, -27.284, -26.243, -24.963,
-23.559, -22.091, -20.483, -18.815, -17.040, -28.055, -27.955, -27.310, -26.268,
-24.987, -23.581, -22.110, -20.500, -18.829, -17.055, -28.082, -27.981, -27.336,
-26.293, -25.010, -23.602, -22.130, -20.517, -18.843, -17.070, -28.109, -28.008,
-27.362, -26.318, -25.033, -23.623, -22.150, -20.533, -18.858, -17.085, -28.136,
-28.035, -27.388, -26.343, -25.056, -23.645, -22.169, -20.550, -18.872, -17.099,
 -28.163, -28.062, -27.415, -26.367, -25.079, -23.666, -22.189, -20.567,
-18.886, -17.114, -28.189, -28.089, -27.441, -26.392, -25.102, -23.687, -22.208,
-20.584, -18.901, -17.129, -28.216, -28.116, -27.467, -26.417, -25.125, -23.708,
-22.228, -20.600, -18.915, -17.144, -28.243, -28.143, -27.493, -26.442, -25.148,
-23.730, -22.247, -20.617, -18.930, -17.159, -28.270, -28.169, -27.519, -26.467,
-25.171, -23.751, -22.267, -20.634, -18.944, -17.175, -28.297, -28.196, -27.545,
-26.492, -25.194, -23.772, -22.286, -20.651, -18.959, -17.190, -28.324, -28.223,
-27.571, -26.516, -25.218, -23.794, -22.306, -20.668, -18.973, -17.205, -28.350,
-28.250, -27.598, -26.541, -25.241, -23.815, -22.325, -20.685, -18.988, -17.220,
 -28.377, -28.277, -27.624, -26.566, -25.264, -23.836, -22.345, -20.702,
-19.003, -17.235, -28.404, -28.304, -27.650, -26.591, -25.287, -23.857, -22.365,
-20.719, -19.017, -17.250, -28.431, -28.330, -27.676, -26.616, -25.310, -23.879,
-22.384, -20.736, -19.032, -17.266, -28.458, -28.357, -27.702, -26.641, -25.333,
-23.900, -22.404, -20.753, -19.046, -17.281, -28.484, -28.384, -27.728, -26.665,
-25.356, -23.921, -22.423, -20.770, -19.061, -17.296, -28.511, -28.411, -27.754,
-26.690, -25.379, -23.942, -22.443, -20.787, -19.075, -17.310, -28.538, -28.438,
-27.781, -26.715, -25.402, -23.964, -22.462, -20.805, -19.090, -17.325, -28.565,
-28.465, -27.807, -26.740, -25.426, -23.985, -22.482, -20.822, -19.104, -17.339,
 -28.592, -28.491, -27.833, -26.765, -25.449, -24.006, -22.501, -20.839,
-19.118, -17.353, -28.619, -28.518, -27.859, -26.789, -25.472, -24.027, -22.521,
-20.856, -19.132, -17.367, -28.645, -28.545, -27.885, -26.814, -25.495, -24.049,
-22.540, -20.873, -19.146, -17.380, -28.672, -28.572, -27.911, -26.839, -25.518,
-24.070, -22.560, -20.890, -19.160, -17.393, -28.699, -28.599, -27.937, -26.864,
-25.541, -24.091, -22.579, -20.907, -19.173, -17.406, -28.726, -28.625, -27.964,
-26.889, -25.564, -24.112, -22.599, -20.924, -19.187, -17.418, -28.752, -28.652,
-27.990, -26.913, -25.587, -24.134, -22.618, -20.942, -19.200, -17.431, -28.779,
-28.679, -28.016, -26.938, -25.610, -24.155, -22.637, -20.959, -19.213, -17.443,
 -28.806, -28.706, -28.042, -26.963, -25.633, -24.176, -22.657, -20.976,
-19.226, -17.454, -28.833, -28.733, -28.068, -26.988, -25.656, -24.197, -22.676,
-20.993, -19.239, -17.466, -28.860, -28.759, -28.094, -27.013, -25.680, -24.219,
-22.696, -21.010, -19.252, -17.478, -28.886, -28.786, -28.120, -27.038, -25.703,
-24.240, -22.715, -21.028, -19.265, -17.489, -28.913, -28.813, -28.146, -27.062,
-25.726, -24.261, -22.735, -21.045, -19.278, -17.501, -28.940, -28.840, -28.172,
-27.087, -25.749, -24.282, -22.754, -21.062, -19.291, -17.512, -28.967, -28.867,
-28.199, -27.112, -25.772, -24.304, -22.774, -21.080, -19.304, -17.524, -28.994,
-28.893, -28.225, -27.137, -25.795, -24.325, -22.793, -21.097, -19.317, -17.536,
 -29.020, -28.920, -28.251, -27.162, -25.818, -24.346, -22.813, -21.114,
-19.330, -17.547, -29.047, -28.947, -28.277, -27.186, -25.841, -24.367, -22.832,
-21.132, -19.343, -17.559, -29.074, -28.974, -28.303, -27.211, -25.864, -24.388,
-22.851, -21.149, -19.356, -17.571, -29.101, -29.000, -28.329, -27.236, -25.887,
-24.410, -22.871, -21.167, -19.370, -17.583, -29.127, -29.027, -28.355, -27.261,
-25.910, -24.431, -22.890, -21.184, -19.383, -17.595, -29.154, -29.054, -28.381,
-27.286, -25.934, -24.452, -22.910, -21.202, -19.396, -17.607, -29.181, -29.081,
-28.407, -27.310, -25.957, -24.473, -22.929, -21.220, -19.410, -17.619, -29.208,
-29.107, -28.434, -27.335, -25.980, -24.495, -22.949, -21.237, -19.423, -17.632,
 -29.234, -29.134, -28.460, -27.360, -26.003, -24.516, -22.968, -21.255,
-19.437, -17.644, -29.261, -29.161, -28.486, -27.385, -26.026, -24.537, -22.987,
-21.273, -19.450, -17.657, -29.288, -29.188, -28.512, -27.410, -26.049, -24.558,
-23.007, -21.291, -19.464, -17.670, -29.315, -29.215, -28.538, -27.434, -26.072,
-24.579, -23.026, -21.309, -19.478, -17.683, -29.341, -29.241, -28.564, -27.459,
-26.095, -24.601, -23.046, -21.327, -19.492, -17.696, -29.368, -29.268, -28.590,
-27.484, -26.118, -24.622, -23.065, -21.345, -19.506, -17.710, -29.395, -29.295,
-28.616, -27.509, -26.141, -24.643, -23.084, -21.363, -19.520, -17.723, -29.422,
-29.322, -28.642, -27.533, -26.164, -24.664, -23.104, -21.381, -19.535, -17.737,
 -29.448, -29.348, -28.668, -27.558, -26.187, -24.686, -23.123, -21.399,
-19.549, -17.751, -29.475, -29.375, -28.694, -27.583, -26.210, -24.707, -23.143,
-21.417, -19.563, -17.764, -29.502, -29.402, -28.721, -27.608, -26.234, -24.728,
-23.162, -21.435, -19.578, -17.778, -29.529, -29.429, -28.747, -27.633, -26.257,
-24.749, -23.181, -21.453, -19.592, -17.792, -29.555, -29.455, -28.773, -27.657,
-26.280, -24.770, -23.201, -21.471, -19.606, -17.806, -29.582, -29.482, -28.799,
-27.682, -26.303, -24.792, -23.220, -21.490, -19.621, -17.820, -29.609, -29.509,
-28.825, -27.707, -26.326, -24.813, -23.240, -21.508, -19.635, -17.834, -29.636,
-29.536, -28.851, -27.732, -26.349, -24.834, -23.259, -21.526, -19.649, -17.847,
 -29.662, -29.562, -28.877, -27.757, -26.372, -24.855, -23.278, -21.545,
-19.664, -17.860, -29.689, -29.589, -28.903, -27.781, -26.395, -24.876, -23.298,
-21.563, -19.678, -17.873, -29.716, -29.616, -28.929, -27.806, -26.418, -24.898,
-23.317, -21.581, -19.692, -17.885, -29.742, -29.642, -28.955, -27.831, -26.441,
-24.919, -23.337, -21.600, -19.705, -17.897, -29.769, -29.669, -28.981, -27.856,
-26.464, -24.940, -23.356, -21.618, -19.719, -17.909, -29.796, -29.696, -29.007,
-27.880, -26.487, -24.961, -23.375, -21.636, -19.733, -17.920, -29.823, -29.723,
-29.033, -27.905, -26.510, -24.982, -23.395, -21.655, -19.746, -17.930, -29.849,
-29.749, -29.059, -27.930, -26.533, -25.004, -23.414, -21.673, -19.759, -17.941,
 -29.876, -29.776, -29.085, -27.955, -26.556, -25.025, -23.433, -21.691,
-19.772, -17.951, -29.903, -29.803, -29.112, -27.979, -26.580, -25.046, -23.453,
-21.710, -19.785, -17.960, -29.929, -29.829, -29.138, -28.004, -26.603, -25.067,
-23.472, -21.728, -19.798, -17.970, -29.956, -29.856, -29.164, -28.029, -26.626,
-25.088, -23.491, -21.747, -19.811, -17.979, -29.983, -29.883, -29.190, -28.054,
-26.649, -25.110, -23.511, -21.765, -19.824, -17.988, -30.010, -29.910, -29.216,
-28.079, -26.672, -25.131, -23.530, -21.784, -19.837, -17.997, -30.036, -29.936,
-29.242, -28.103, -26.695, -25.152, -23.549, -21.802, -19.850, -18.007, -30.063,
-29.963, -29.268, -28.128, -26.718, -25.173, -23.569, -21.820, -19.862, -18.016,
 -30.090, -29.990, -29.294, -28.153, -26.741, -25.194, -23.588, -21.839,
-19.875, -18.025, -30.116, -30.016, -29.320, -28.178, -26.764, -25.215, -23.607,
-21.857, -19.888, -18.035, -30.143, -30.043, -29.346, -28.202, -26.787, -25.237,
-23.627, -21.876, -19.901, -18.045, -30.170, -30.070, -29.372, -28.227, -26.810,
-25.258, -23.646, -21.894, -19.914, -18.054, -30.196, -30.096, -29.398, -28.252,
-26.833, -25.279, -23.665, -21.912, -19.927, -18.064, -30.223, -30.123, -29.424,
-28.277, -26.856, -25.300, -23.685, -21.931, -19.941, -18.074, -30.250, -30.150,
-29.450, -28.301, -26.879, -25.321, -23.704, -21.949, -19.954, -18.085, -30.276,
-30.177, -29.476, -28.326, -26.902, -25.343, -23.723, -21.967, -19.968, -18.095,
 -30.303, -30.203, -29.502, -28.351, -26.925, -25.364, -23.743, -21.986,
-19.982, -18.106, -30.330, -30.230, -29.528, -28.376, -26.948, -25.385, -23.762,
-22.004, -19.996, -18.117, -30.356, -30.257, -29.554, -28.400, -26.971, -25.406,
-23.781, -22.022, -20.010, -18.129, -30.383, -30.283, -29.580, -28.425, -26.994,
-25.427, -23.801, -22.041, -20.024, -18.140, -30.410, -30.310, -29.606, -28.450,
-27.017, -25.448, -23.820, -22.059, -20.038, -18.152, -30.436, -30.337, -29.632,
-28.475, -27.040, -25.470, -23.839, -22.077, -20.053, -18.164, -30.463, -30.363,
-29.658, -28.499, -27.064, -25.491, -23.859, -22.096, -20.068, -18.176, -30.490,
-30.390, -29.684, -28.524, -27.087, -25.512, -23.878, -22.114, -20.083, -18.189,
 -30.516, -30.417, -29.710, -28.549, -27.110, -25.533, -23.897, -22.132,
-20.098, -18.202, -30.543, -30.443, -29.736, -28.574, -27.133, -25.554, -23.917,
-22.150, -20.113, -18.215, -30.570, -30.470, -29.762, -28.598, -27.156, -25.575,
-23.936, -22.169, -20.128, -18.228, -30.596, -30.497, -29.789, -28.623, -27.179,
-25.596, -23.955, -22.187, -20.144, -18.241, -30.623, -30.523, -29.815, -28.648,
-27.202, -25.618, -23.974, -22.205, -20.159, -18.255, -30.650, -30.550, -29.841,
-28.673, -27.225, -25.639, -23.994, -22.223, -20.175, -18.268, -30.676, -30.577,
-29.867, -28.697, -27.248, -25.660, -24.013, -22.242, -20.191, -18.282, -30.703,
-30.603, -29.893, -28.722, -27.271, -25.681, -24.032, -22.260, -20.206, -18.295,
 -30.730, -30.630, -29.919, -28.747, -27.294, -25.702, -24.052, -22.278,
-20.222, -18.308, -30.756, -30.657, -29.945, -28.772, -27.317, -25.723, -24.071,
-22.296, -20.238, -18.321, -30.783, -30.683, -29.971, -28.796, -27.340, -25.745,
-24.090, -22.314, -20.253, -18.334, -30.810, -30.710, -29.997, -28.821, -27.363,
-25.766, -24.110, -22.332, -20.269, -18.346, -30.836, -30.737, -30.023, -28.846,
-27.386, -25.787, -24.129, -22.351, -20.284, -18.358, -30.863, -30.763, -30.049,
-28.870, -27.409, -25.808, -24.148, -22.369, -20.300, -18.369, -30.890, -30.790,
-30.075, -28.895, -27.432, -25.829, -24.167, -22.387, -20.315, -18.380, -30.916,
-30.817, -30.101, -28.920, -27.455, -25.850, -24.187, -22.405, -20.330, -18.390,
 -30.943, -30.843, -30.127, -28.945, -27.478, -25.871, -24.206, -22.423,
-20.346, -18.400, -30.970, -30.870, -30.153, -28.969, -27.501, -25.893, -24.225,
-22.441, -20.361, -18.409, -30.996, -30.896, -30.179, -28.994, -27.524, -25.914,
-24.244, -22.459, -20.375, -18.418, -31.023, -30.923, -30.205, -29.019, -27.547,
-25.935, -24.264, -22.477, -20.390, -18.426, -31.049, -30.950, -30.231, -29.044,
-27.570, -25.956, -24.283, -22.495, -20.405, -18.435, -31.076, -30.976, -30.257,
-29.068, -27.593, -25.977, -24.302, -22.513, -20.420, -18.443, -31.103, -31.003,
-30.283, -29.093, -27.616, -25.998, -24.322, -22.531, -20.435, -18.451, -31.129,
-31.030, -30.309, -29.118, -27.639, -26.019, -24.341, -22.549, -20.449, -18.460,
 -31.156, -31.056, -30.334, -29.142, -27.662, -26.040, -24.360, -22.567,
-20.464, -18.468, -31.183, -31.083, -30.360, -29.167, -27.685, -26.062, -24.379,
-22.585, -20.479, -18.476, -31.209, -31.109, -30.386, -29.192, -27.708, -26.083,
-24.399, -22.603, -20.494, -18.485, -31.236, -31.136, -30.412, -29.217, -27.731,
-26.104, -24.418, -22.621, -20.509, -18.493, -31.262, -31.163, -30.438, -29.241,
-27.754, -26.125, -24.437, -22.639, -20.524, -18.502, -31.289, -31.189, -30.464,
-29.266, -27.777, -26.146, -24.456, -22.657, -20.539, -18.512, -31.316, -31.216,
-30.490, -29.291, -27.800, -26.167, -24.476, -22.675, -20.554, -18.521, -31.342,
-31.243, -30.516, -29.315, -27.823, -26.188, -24.495, -22.693, -20.570, -18.531,
 -31.369, -31.269, -30.542, -29.340, -27.846, -26.209, -24.514, -22.711,
-20.585, -18.540, -31.396, -31.296, -30.568, -29.365, -27.869, -26.231, -24.533,
-22.729, -20.601, -18.551, -31.422, -31.322, -30.594, -29.390, -27.892, -26.252,
-24.553, -22.747, -20.617, -18.561, -31.449, -31.349, -30.620, -29.414, -27.915,
-26.273, -24.572, -22.765, -20.633, -18.572, -31.475, -31.376, -30.646, -29.439,
-27.938, -26.294, -24.591, -22.782, -20.649, -18.583, -31.502, -31.402, -30.672,
-29.464, -27.961, -26.315, -24.610, -22.800, -20.665, -18.595, -31.529, -31.429,
-30.698, -29.488, -27.984, -26.336, -24.630, -22.818, -20.682, -18.607, -31.555,
-31.455, -30.724, -29.513, -28.007, -26.357, -24.649, -22.836, -20.698, -18.619,
 -31.582, -31.482, -30.750, -29.538, -28.030, -26.378, -24.668, -22.854,
-20.715, -18.631, -31.608, -31.509, -30.776, -29.562, -28.053, -26.399, -24.687,
-22.872, -20.731, -18.644, -31.635, -31.535, -30.802, -29.587, -28.076, -26.421,
-24.706, -22.890, -20.748, -18.657, -31.662, -31.562, -30.828, -29.612, -28.099,
-26.442, -24.726, -22.907, -20.765, -18.671, -31.688, -31.588, -30.854, -29.636,
-28.122, -26.463, -24.745, -22.925, -20.782, -18.684, -31.715, -31.615, -30.880,
-29.661, -28.145, -26.484, -24.764, -22.943, -20.799, -18.698, -31.741, -31.642,
-30.906, -29.686, -28.168, -26.505, -24.783, -22.961, -20.817, -18.712, -31.768,
-31.668, -30.932, -29.711, -28.191, -26.526, -24.803, -22.979, -20.834, -18.726,
 -31.794, -31.695, -30.958, -29.735, -28.214, -26.547, -24.822, -22.996,
-20.851, -18.741, -31.821, -31.721, -30.984, -29.760, -28.237, -26.568, -24.841,
-23.014, -20.869, -18.755, -31.848, -31.748, -31.009, -29.785, -28.260, -26.589,
-24.860, -23.032, -20.887, -18.769, -31.874, -31.774, -31.035, -29.809, -28.283,
-26.610, -24.879, -23.050, -20.904, -18.782, -31.901, -31.801, -31.061, -29.834,
-28.306, -26.631, -24.899, -23.068, -20.922, -18.795, -31.927, -31.828, -31.087,
-29.859, -28.329, -26.653, -24.918, -23.085, -20.940, -18.808, -31.954, -31.854,
-31.113, -29.883, -28.352, -26.674, -24.937, -23.103, -20.958, -18.820, -31.981,
-31.881, -31.139, -29.908, -28.375, -26.695, -24.956, -23.121, -20.975, -18.832,
 -32.007, -31.907, -31.165, -29.933, -28.398, -26.716, -24.975, -23.138,
-20.993, -18.843, -32.034, -31.934, -31.191, -29.957, -28.421, -26.737, -24.995,
-23.156, -21.011, -18.854, -32.060, -31.960, -31.217, -29.982, -28.444, -26.758,
-25.014, -23.174, -21.029, -18.864, -32.087, -31.987, -31.243, -30.007, -28.467,
-26.779, -25.033, -23.192, -21.047, -18.873, -32.113, -32.014, -31.269, -30.031,
-28.490, -26.800, -25.052, -23.209, -21.065, -18.882, -32.140, -32.040, -31.295,
-30.056, -28.513, -26.821, -25.071, -23.227, -21.083, -18.891, -32.167, -32.067,
-31.321, -30.081, -28.536, -26.842, -25.091, -23.245, -21.101, -18.900, -32.193,
-32.093, -31.347, -30.105, -28.559, -26.863, -25.110, -23.262, -21.119, -18.908,
 -32.220, -32.120, -31.373, -30.130, -28.582, -26.884, -25.129, -23.280,
-21.138, -18.916, -32.246, -32.146, -31.398, -30.155, -28.605, -26.906, -25.148,
-23.298, -21.156, -18.925, -32.273, -32.173, -31.424, -30.179, -28.628, -26.927,
-25.167, -23.315, -21.174, -18.934, -32.299, -32.200, -31.450, -30.204, -28.651,
-26.948, -25.187, -23.333, -21.192, -18.942, -32.326, -32.226, -31.476, -30.229,
-28.674, -26.969, -25.206, -23.351, -21.210, -18.951, -32.352, -32.253, -31.502,
-30.253, -28.696, -26.990, -25.225, -23.368, -21.228, -18.960, -32.379, -32.279,
-31.528, -30.278, -28.719, -27.011, -25.244, -23.386, -21.246, -18.970, -32.406,
-32.306, -31.554, -30.303, -28.742, -27.032, -25.263, -23.404, -21.264, -18.980,
 -32.432, -32.332, -31.580, -30.327, -28.765, -27.053, -25.283, -23.421,
-21.282, -18.990, -32.459, -32.359, -31.606, -30.352, -28.788, -27.074, -25.302,
-23.439, -21.300, -19.000, -32.485, -32.385, -31.632, -30.377, -28.811, -27.095,
-25.321, -23.457, -21.319, -19.011, -32.512, -32.412, -31.658, -30.401, -28.834,
-27.116, -25.340, -23.474, -21.337, -19.022, -32.538, -32.438, -31.683, -30.426,
-28.857, -27.137, -25.359, -23.492, -21.355, -19.034, -32.565, -32.465, -31.709,
-30.451, -28.880, -27.158, -25.378, -23.509, -21.373, -19.046, -32.591, -32.491,
-31.735, -30.475, -28.903, -27.179, -25.398, -23.527, -21.391, -19.058, -32.618,
-32.518, -31.761, -30.500, -28.926, -27.200, -25.417, -23.545, -21.409, -19.071,
 -32.644, -32.545, -31.787, -30.524, -28.949, -27.222, -25.436, -23.562,
-21.427, -19.084, -32.671, -32.571, -31.813, -30.549, -28.972, -27.243, -25.455,
-23.580, -21.445, -19.098, -32.698, -32.598, -31.839, -30.574, -28.995, -27.264,
-25.474, -23.597, -21.462, -19.112, -32.724, -32.624, -31.865, -30.598, -29.018,
-27.285, -25.493, -23.615, -21.480, -19.127, -32.751, -32.651, -31.891, -30.623,
-29.041, -27.306, -25.513, -23.632, -21.498, -19.141, -32.777, -32.677, -31.916,
-30.648, -29.064, -27.327, -25.532, -23.650, -21.516, -19.157, -32.804, -32.704,
-31.942, -30.672, -29.087, -27.348, -25.551, -23.668, -21.534, -19.172, -32.830,
-32.730, -31.968, -30.697, -29.110, -27.369, -25.570, -23.685, -21.552, -19.188,
 -32.857, -32.757, -31.994, -30.722, -29.132, -27.390, -25.589, -23.703,
-21.570, -19.204, -32.883, -32.783, -32.020, -30.746, -29.155, -27.411, -25.608,
-23.720, -21.587, -19.221, -32.910, -32.810, -32.046, -30.771, -29.178, -27.432,
-25.627, -23.738, -21.605, -19.237, -32.936, -32.836, -32.072, -30.795, -29.201,
-27.453, -25.647, -23.755, -21.623, -19.254, -32.963, -32.863, -32.098, -30.820,
-29.224, -27.474, -25.666, -23.773, -21.640, -19.270, -32.989, -32.889, -32.123,
-30.845, -29.247, -27.495, -25.685, -23.790, -21.658, -19.286, -33.016, -32.916,
-32.149, -30.869, -29.270, -27.516, -25.704, -23.808, -21.676, -19.302, -33.042,
-32.942, -32.175, -30.894, -29.293, -27.537, -25.723, -23.826, -21.693, -19.318,
 -33.069, -32.969, -32.201, -30.919, -29.316, -27.558, -25.742, -23.843,
-21.711, -19.333, -33.095, -32.995, -32.227, -30.943, -29.339, -27.579, -25.762,
-23.861, -21.728, -19.347, -33.122, -33.022, -32.253, -30.968, -29.362, -27.600,
-25.781, -23.878, -21.746, -19.361, -33.148, -33.048, -32.279, -30.992, -29.385,
-27.621, -25.800, -23.896, -21.763, -19.375, -33.175, -33.075, -32.305, -31.017,
-29.408, -27.642, -25.819, -23.913, -21.781, -19.388, -33.202, -33.101, -32.330,
-31.042, -29.431, -27.663, -25.838, -23.931, -21.798, -19.400, -33.228, -33.128,
-32.356, -31.066, -29.453, -27.684, -25.857, -23.948, -21.816, -19.412, -33.255,
-33.154, -32.382, -31.091, -29.476, -27.705, -25.876, -23.966, -21.833, -19.424,
 -33.281, -33.181, -32.408, -31.116, -29.499, -27.726, -25.895, -23.983,
-21.851, -19.436, -33.308, -33.207, -32.434, -31.140, -29.522, -27.747, -25.915,
-24.001, -21.868, -19.447, -33.334, -33.234, -32.460, -31.165, -29.545, -27.768,
-25.934, -24.018, -21.885, -19.459, -33.361, -33.260, -32.486, -31.189, -29.568,
-27.789, -25.953, -24.035, -21.903, -19.470, -33.387, -33.287, -32.511, -31.214,
-29.591, -27.810, -25.972, -24.053, -21.920, -19.482, -33.414, -33.313, -32.537,
-31.239, -29.614, -27.831, -25.991, -24.070, -21.937, -19.494, -33.440, -33.340,
-32.563, -31.263, -29.637, -27.853, -26.010, -24.088, -21.954, -19.506, -33.467,
-33.366, -32.589, -31.288, -29.660, -27.874, -26.029, -24.105, -21.972, -19.518,
 -33.493, -33.393, -32.615, -31.312, -29.683, -27.895, -26.048, -24.123,
-21.989, -19.531, -33.520, -33.419, -32.641, -31.337, -29.705, -27.916, -26.068,
-24.140, -22.006, -19.543, -33.546, -33.446, -32.666, -31.362, -29.728, -27.937,
-26.087, -24.158, -22.023, -19.556, -33.573, -33.472, -32.692, -31.386, -29.751,
-27.958, -26.106, -24.175, -22.040, -19.570, -33.599, -33.499, -32.718, -31.411,
-29.774, -27.979, -26.125, -24.193, -22.057, -19.583, -33.626, -33.525, -32.744,
-31.435, -29.797, -28.000, -26.144, -24.210, -22.074, -19.597, -33.652, -33.552,
-32.770, -31.460, -29.820, -28.021, -26.163, -24.228, -22.091, -19.611, -33.678,
-33.578, -32.796, -31.485, -29.843, -28.042, -26.182, -24.245, -22.108, -19.625,
 -33.705, -33.604, -32.822, -31.509, -29.866, -28.063, -26.201, -24.262,
-22.125, -19.640, -33.731, -33.631, -32.847, -31.534, -29.889, -28.084, -26.220,
-24.280, -22.142, -19.655, -33.758, -33.657, -32.873, -31.558, -29.912, -28.105,
-26.239, -24.297, -22.159, -19.670, -33.784, -33.684, -32.899, -31.583, -29.934,
-28.126, -26.259, -24.315, -22.176, -19.685, -33.811, -33.710, -32.925, -31.608,
-29.957, -28.147, -26.278, -24.332, -22.193, -19.701, -33.837, -33.737, -32.951,
-31.632, -29.980, -28.168, -26.297, -24.350, -22.210, -19.717, -33.864, -33.763,
-32.976, -31.657, -30.003, -28.189, -26.316, -24.367, -22.227, -19.733, -33.890,
-33.790, -33.002, -31.681, -30.026, -28.210, -26.335, -24.384, -22.244, -19.749,
 -33.917, -33.816, -33.028, -31.706, -30.049, -28.231, -26.354, -24.402,
-22.261, -19.765, -33.943, -33.843, -33.054, -31.730, -30.072, -28.252, -26.373,
-24.419, -22.278, -19.782, -33.970, -33.869, -33.080, -31.755, -30.095, -28.273,
-26.392, -24.437, -22.294, -19.798, -33.996, -33.895, -33.106, -31.780, -30.118,
-28.293, -26.411, -24.454, -22.311, -19.815, -34.023, -33.922, -33.131, -31.804,
-30.140, -28.314, -26.430, -24.471, -22.328, -19.832, -34.049, -33.948, -33.157,
-31.829, -30.163, -28.335, -26.450, -24.489, -22.345, -19.849, -34.076, -33.975,
-33.183, -31.853, -30.186, -28.356, -26.469, -24.506, -22.361, -19.866, -34.102,
-34.001, -33.209, -31.878, -30.209, -28.377, -26.488, -24.524, -22.378, -19.884,
 -34.129, -34.028, -33.235, -31.903, -30.232, -28.398, -26.507, -24.541,
-22.395, -19.901, -34.155, -34.054, -33.260, -31.927, -30.255, -28.419, -26.526,
-24.558, -22.412, -19.918, -34.181, -34.081, -33.286, -31.952, -30.278, -28.440,
-26.545, -24.576, -22.428, -19.936, -34.208, -34.107, -33.312, -31.976, -30.301,
-28.461, -26.564, -24.593, -22.445, -19.954, -34.234, -34.133, -33.338, -32.001,
-30.323, -28.482, -26.583, -24.611, -22.461, -19.971, -34.261, -34.160, -33.364,
-32.025, -30.346, -28.503, -26.602, -24.628, -22.478, -19.989, -34.287, -34.186,
-33.389, -32.050, -30.369, -28.524, -26.621, -24.645, -22.495, -20.007, -34.314,
-34.213, -33.415, -32.074, -30.392, -28.545, -26.640, -24.663, -22.511, -20.025,
 -34.340, -34.239, -33.441, -32.099, -30.415, -28.566, -26.659, -24.680,
-22.528, -20.043, -34.367, -34.266, -33.467, -32.124, -30.438, -28.587, -26.678,
-24.697, -22.544, -20.060, -34.393, -34.292, -33.493, -32.148, -30.461, -28.608,
-26.697, -24.715, -22.561, -20.078, -34.420, -34.319, -33.518, -32.173, -30.483,
-28.629, -26.717, -24.732, -22.577, -20.096, -34.446, -34.345, -33.544, -32.197,
-30.506, -28.650, -26.736, -24.750, -22.594, -20.114, -34.473, -34.371, -33.570,
-32.222, -30.529, -28.671, -26.755, -24.767, -22.610, -20.132, -34.499, -34.398,
-33.596, -32.246, -30.552, -28.692, -26.774, -24.784, -22.627, -20.150, -34.525,
-34.424, -33.622, -32.271, -30.575, -28.713, -26.793, -24.802, -22.643, -20.168,
 -34.552, -34.451, -33.647, -32.295, -30.598, -28.734, -26.812, -24.819,
-22.660, -20.186, -34.578, -34.477, -33.673, -32.320, -30.621, -28.755, -26.831,
-24.836, -22.676, -20.204, -34.605, -34.503, -33.699, -32.345, -30.644, -28.776,
-26.850, -24.854, -22.693, -20.223, -34.631, -34.530, -33.725, -32.369, -30.666,
-28.797, -26.869, -24.871, -22.709, -20.241, -34.658, -34.556, -33.751, -32.394,
-30.689, -28.818, -26.888, -24.888, -22.726, -20.259, -34.684, -34.583, -33.776,
-32.418, -30.712, -28.839, -26.907, -24.906, -22.742, -20.276, -34.711, -34.609,
-33.802, -32.443, -30.735, -28.860, -26.926, -24.923, -22.758, -20.294, -34.737,
-34.636, -33.828, -32.467, -30.758, -28.881, -26.945, -24.940, -22.775, -20.312,
 -34.763, -34.662, -33.854, -32.492, -30.781, -28.902, -26.964, -24.958,
-22.791, -20.330, -34.790, -34.688, -33.879, -32.516, -30.803, -28.922, -26.983,
-24.975, -22.807, -20.348, -34.816, -34.715, -33.905, -32.541, -30.826, -28.943,
-27.002, -24.992, -22.824, -20.366, -34.843, -34.741, -33.931, -32.565, -30.849,
-28.964, -27.021, -25.010, -22.840, -20.384, -34.869, -34.768, -33.957, -32.590,
-30.872, -28.985, -27.040, -25.027, -22.856, -20.402, -34.896, -34.794, -33.982,
-32.615, -30.895, -29.006, -27.059, -25.044, -22.873, -20.420, -34.922, -34.820,
-34.008, -32.639, -30.918, -29.027, -27.079, -25.062, -22.889, -20.437, -34.948,
-34.847, -34.034, -32.664, -30.941, -29.048, -27.098, -25.079, -22.905, -20.455,
 -34.975, -34.873, -34.060, -32.688, -30.963, -29.069, -27.117, -25.096,
-22.921, -20.473, -35.001, -34.900, -34.086, -32.713, -30.986, -29.090, -27.136,
-25.114, -22.938, -20.490, -35.028, -34.926, -34.111, -32.737, -31.009, -29.111,
-27.155, -25.131, -22.954, -20.508, -35.054, -34.952, -34.137, -32.762, -31.032,
-29.132, -27.174, -25.148, -22.970, -20.526, -35.081, -34.979, -34.163, -32.786,
-31.055, -29.153, -27.193, -25.165, -22.986, -20.543, -35.107, -35.005, -34.189,
-32.811, -31.078, -29.174, -27.212, -25.183, -23.003, -20.561, -35.133, -35.032,
-34.214, -32.835, -31.100, -29.195, -27.231, -25.200, -23.019, -20.578, -35.160,
-35.058, -34.240, -32.860, -31.123, -29.216, -27.250, -25.217, -23.035, -20.595,
 -35.186, -35.084, -34.266, -32.884, -31.146, -29.237, -27.269, -25.235,
-23.051, -20.613, -35.213, -35.111, -34.292, -32.909, -31.169, -29.257, -27.288,
-25.252, -23.067, -20.630, -35.239, -35.137, -34.317, -32.933, -31.192, -29.278,
-27.307, -25.269, -23.083, -20.648, -35.265, -35.163, -34.343, -32.958, -31.214,
-29.299, -27.326, -25.287, -23.100, -20.665, -35.292, -35.190, -34.369, -32.982,
-31.237, -29.320, -27.345, -25.304, -23.116, -20.682, -35.318, -35.216, -34.395,
-33.007, -31.260, -29.341, -27.364, -25.321, -23.132, -20.699, -35.345, -35.243,
-34.420, -33.031, -31.283, -29.362, -27.383, -25.338, -23.148, -20.716, -35.371,
-35.269, -34.446, -33.056, -31.306, -29.383, -27.402, -25.356, -23.164, -20.734,
 -35.398, -35.295, -34.472, -33.080, -31.329, -29.404, -27.421, -25.373,
-23.180, -20.751, -35.424, -35.322, -34.498, -33.105, -31.351, -29.425, -27.440,
-25.390, -23.196, -20.768, -35.450, -35.348, -34.523, -33.130, -31.374, -29.446,
-27.459, -25.408, -23.212, -20.785, -35.477, -35.375, -34.549, -33.154, -31.397,
-29.467, -27.478, -25.425, -23.228, -20.802, -35.503, -35.401, -34.575, -33.179,
-31.420, -29.488, -27.497, -25.442, -23.244, -20.819, -35.530, -35.427, -34.601,
-33.203, -31.443, -29.508, -27.516, -25.459, -23.261, -20.835, -35.556, -35.454,
-34.626, -33.228, -31.465, -29.529, -27.535, -25.477, -23.277, -20.852, -35.582,
-35.480, -34.652, -33.252, -31.488, -29.550, -27.554, -25.494, -23.293, -20.869,
 -35.609, -35.506, -34.678, -33.277, -31.511, -29.571, -27.573, -25.511,
-23.309, -20.886, -35.635, -35.533, -34.703, -33.301, -31.534, -29.592, -27.592,
-25.528, -23.325, -20.903, -35.662, -35.559, -34.729, -33.326, -31.557, -29.613,
-27.611, -25.546, -23.341, -20.919, -35.688, -35.585, -34.755, -33.350, -31.580,
-29.634, -27.630, -25.563, -23.357, -20.936, -35.714, -35.612, -34.781, -33.375,
-31.602, -29.655, -27.649, -25.580, -23.373, -20.953, -35.741, -35.638, -34.806,
-33.399, -31.625, -29.676, -27.668, -25.598, -23.389, -20.969, -35.767, -35.665,
-34.832, -33.424, -31.648, -29.697, -27.687, -25.615, -23.405, -20.986, -35.794,
-35.691, -34.858, -33.448, -31.671, -29.717, -27.706, -25.632, -23.421, -21.002,
 -35.820, -35.717, -34.884, -33.473, -31.694, -29.738, -27.725, -25.649,
-23.437, -21.019, -35.846, -35.744, -34.909, -33.497, -31.716, -29.759, -27.744,
-25.667, -23.453, -21.035, -35.873, -35.770, -34.935, -33.521, -31.739, -29.780,
-27.763, -25.684, -23.469, -21.052, -35.899, -35.796, -34.961, -33.546, -31.762,
-29.801, -27.782, -25.701, -23.485, -21.068, -35.926, -35.823, -34.986, -33.570,
-31.785, -29.822, -27.801, -25.718, -23.501, -21.084, -35.952, -35.849, -35.012,
-33.595, -31.808, -29.843, -27.820, -25.736, -23.517, -21.101, -35.978, -35.875,
-35.038, -33.619, -31.830, -29.864, -27.839, -25.753, -23.532, -21.117, -36.005,
-35.902, -35.064, -33.644, -31.853, -29.885, -27.858, -25.770, -23.548, -21.133,
 -36.031, -35.928, -35.089, -33.668, -31.876, -29.906, -27.877, -25.787,
-23.564, -21.149, -36.057, -35.954, -35.115, -33.693, -31.899, -29.926, -27.896,
-25.804, -23.580, -21.165, -36.084, -35.981, -35.141, -33.717, -31.921, -29.947,
-27.915, -25.822, -23.596, -21.182, -36.110, -36.007, -35.166, -33.742, -31.944,
-29.968, -27.934, -25.839, -23.612, -21.198, -36.137, -36.033, -35.192, -33.766,
-31.967, -29.989, -27.953, -25.856, -23.628, -21.214, -36.163, -36.060, -35.218,
-33.791, -31.990, -30.010, -27.972, -25.873, -23.644, -21.230, -36.189, -36.086,
-35.243, -33.815, -32.013, -30.031, -27.991, -25.891, -23.660, -21.246, -36.216,
-36.112, -35.269, -33.840, -32.035, -30.052, -28.010, -25.908, -23.676, -21.262,
 -36.242, -36.139, -35.295, -33.864, -32.058, -30.073, -28.029, -25.925,
-23.692, -21.278, -36.268, -36.165, -35.321, -33.889, -32.081, -30.093, -28.048,
-25.942, -23.707, -21.294, -36.295, -36.192, -35.346, -33.913, -32.104, -30.114,
-28.067, -25.960, -23.723, -21.310, -36.321, -36.218, -35.372, -33.938, -32.127,
-30.135, -28.086, -25.977, -23.739, -21.326, -36.348, -36.244, -35.398, -33.962,
-32.149, -30.156, -28.105, -25.994, -23.755, -21.341, -36.374, -36.271, -35.423,
-33.987, -32.172, -30.177, -28.123, -26.011, -23.771, -21.357, -36.400, -36.297,
-35.449, -34.011, -32.195, -30.198, -28.142, -26.028, -23.787, -21.373, -36.427,
-36.323, -35.475, -34.036, -32.218, -30.219, -28.161, -26.046, -23.803, -21.389,
 -36.453, -36.349, -35.500, -34.060, -32.240, -30.240, -28.180, -26.063,
-23.818, -21.404, -36.479, -36.376, -35.526, -34.084, -32.263, -30.260, -28.199,
-26.080, -23.834, -21.420, -36.506, -36.402, -35.552, -34.109, -32.286, -30.281,
-28.218, -26.097, -23.850, -21.436, -36.532, -36.428, -35.578, -34.133, -32.309,
-30.302, -28.237, -26.114, -23.866, -21.451, -36.559, -36.455, -35.603, -34.158,
-32.331, -30.323, -28.256, -26.132, -23.882, -21.467, -36.585, -36.481, -35.629,
-34.182, -32.354, -30.344, -28.275, -26.149, -23.898, -21.483, -36.611, -36.507,
-35.655, -34.207, -32.377, -30.365, -28.294, -26.166, -23.913, -21.498, -36.638,
-36.534, -35.680, -34.231, -32.400, -30.386, -28.313, -26.183, -23.929, -21.514,
 -36.664, -36.560, -35.706, -34.256, -32.423, -30.406, -28.332, -26.200,
-23.945, -21.529, -36.690, -36.586, -35.732, -34.280, -32.445, -30.427, -28.351,
-26.218, -23.961, -21.545, -36.717, -36.613, -35.757, -34.305, -32.468, -30.448,
-28.370, -26.235, -23.977, -21.560, -36.743, -36.639, -35.783, -34.329, -32.491,
-30.469, -28.389, -26.252, -23.992, -21.576, -36.769, -36.665, -35.809, -34.353,
-32.514, -30.490, -28.408, -26.269, -24.008, -21.591, -36.796, -36.692, -35.834,
-34.378, -32.536, -30.511, -28.427, -26.286, -24.024, -21.607, -36.822, -36.718,
-35.860, -34.402, -32.559, -30.532, -28.446, -26.304, -24.040, -21.622, -36.849,
-36.744, -35.886, -34.427, -32.582, -30.552, -28.465, -26.321, -24.055, -21.637,
 -36.875, -36.771, -35.911, -34.451, -32.605, -30.573, -28.484, -26.338,
-24.071, -21.653, -36.901, -36.797, -35.937, -34.476, -32.627, -30.594, -28.502,
-26.355, -24.087, -21.668, -36.928, -36.823, -35.963, -34.500, -32.650, -30.615,
-28.521, -26.372, -24.103, -21.683, -36.954, -36.850, -35.988, -34.525, -32.673,
-30.636, -28.540, -26.390, -24.118, -21.699, -36.980, -36.876, -36.014, -34.549,
-32.696, -30.657, -28.559, -26.407, -24.134, -21.714, -37.007, -36.902, -36.040,
-34.573, -32.718, -30.677, -28.578, -26.424, -24.150, -21.729, -37.033, -36.928,
-36.065, -34.598, -32.741, -30.698, -28.597, -26.441, -24.166, -21.744, -37.059,
-36.955, -36.091, -34.622, -32.764, -30.719, -28.616, -26.458, -24.181, -21.760,
 -37.086, -36.981, -36.117, -34.647, -32.787, -30.740, -28.635, -26.475,
-24.197, -21.775, -37.112, -37.007, -36.142, -34.671, -32.809, -30.761, -28.654,
-26.493, -24.213, -21.790, -37.138, -37.034, -36.168, -34.696, -32.832, -30.782,
-28.673, -26.510, -24.229, -21.805, -37.165, -37.060, -36.194, -34.720, -32.855,
-30.803, -28.692, -26.527, -24.244, -21.820, -37.191, -37.086, -36.219, -34.745,
-32.878, -30.823, -28.711, -26.544, -24.260, -21.835, -37.217, -37.113, -36.245,
-34.769, -32.900, -30.844, -28.730, -26.561, -24.276, -21.850, -37.244, -37.139,
-36.271, -34.793, -32.923, -30.865, -28.749, -26.578, -24.291, -21.866, -37.270,
-37.165, -36.296, -34.818, -32.946, -30.886, -28.767, -26.596, -24.307, -21.881,
 -37.296, -37.191, -36.322, -34.842, -32.969, -30.907, -28.786, -26.613,
-24.323, -21.896, -37.323, -37.218, -36.348, -34.867, -32.991, -30.928, -28.805,
-26.630, -24.339, -21.911, -37.349, -37.244, -36.373, -34.891, -33.014, -30.948,
-28.824, -26.647, -24.354, -21.926, -37.375, -37.270, -36.399, -34.916, -33.037,
-30.969, -28.843, -26.664, -24.370, -21.941, -37.402, -37.297, -36.424, -34.940,
-33.059, -30.990, -28.862, -26.681, -24.386, -21.956, -37.428, -37.323, -36.450,
-34.964, -33.082, -31.011, -28.881, -26.699, -24.401, -21.971, -37.454, -37.349,
-36.476, -34.989, -33.105, -31.032, -28.900, -26.716, -24.417, -21.986, -37.481,
-37.375, -36.501, -35.013, -33.128, -31.052, -28.919, -26.733, -24.433, -22.000,
 -37.507, -37.402, -36.527, -35.038, -33.150, -31.073, -28.938, -26.750,
-24.448, -22.015, -37.533, -37.428, -36.553, -35.062, -33.173, -31.094, -28.957,
-26.767, -24.464, -22.030, -37.560, -37.454, -36.578, -35.086, -33.196, -31.115,
-28.975, -26.784, -24.480, -22.045, -37.586, -37.481, -36.604, -35.111, -33.219,
-31.136, -28.994, -26.801, -24.495, -22.060, -37.612, -37.507, -36.630, -35.135,
-33.241, -31.157, -29.013, -26.819, -24.511, -22.075, -37.639, -37.533, -36.655,
-35.160, -33.264, -31.177, -29.032, -26.836, -24.527, -22.090, -37.665, -37.559,
-36.681, -35.184, -33.287, -31.198, -29.051, -26.853, -24.542, -22.105, -37.691,
-37.586, -36.707, -35.209, -33.309, -31.219, -29.070, -26.870, -24.558, -22.119,
 -37.718, -37.612, -36.732, -35.233, -33.332, -31.240, -29.089, -26.887,
-24.574, -22.134, -37.744, -37.638, -36.758, -35.257, -33.355, -31.261, -29.108,
-26.904, -24.589, -22.149, -37.770, -37.664, -36.783, -35.282, -33.378, -31.281,
-29.127, -26.921, -24.605, -22.164, -37.797, -37.691, -36.809, -35.306, -33.400,
-31.302, -29.146, -26.939, -24.621, -22.178, -37.823, -37.717, -36.835, -35.331,
-33.423, -31.323, -29.164, -26.956, -24.636, -22.193, -37.849, -37.743, -36.860,
-35.355, -33.446, -31.344, -29.183, -26.973, -24.652, -22.208, -37.876, -37.770,
-36.886, -35.379, -33.468, -31.365, -29.202, -26.990, -24.667, -22.223, -37.902,
-37.796, -36.912, -35.404, -33.491, -31.385, -29.221, -27.007, -24.683, -22.237,
 -37.928, -37.822, -36.937, -35.428, -33.514, -31.406, -29.240, -27.024,
-24.699, -22.252, -37.955, -37.848, -36.963, -35.453, -33.537, -31.427, -29.259,
-27.041, -24.714, -22.267, -37.981, -37.875, -36.988, -35.477, -33.559, -31.448,
-29.278, -27.058, -24.730, -22.281, -38.007, -37.901, -37.014, -35.501, -33.582,
-31.469, -29.297, -27.076, -24.746, -22.296, -38.034, -37.927, -37.040, -35.526,
-33.605, -31.489, -29.316, -27.093, -24.761, -22.311, -38.060, -37.953, -37.065,
-35.550, -33.627, -31.510, -29.334, -27.110, -24.777, -22.325, -38.086, -37.980,
-37.091, -35.575, -33.650, -31.531, -29.353, -27.127, -24.792, -22.340, -38.112,
-38.006, -37.116, -35.599, -33.673, -31.552, -29.372, -27.144, -24.808, -22.355,
 -38.139, -38.032, -37.142, -35.623, -33.695, -31.573, -29.391, -27.161,
-24.824, -22.369, -38.165, -38.058, -37.168, -35.648, -33.718, -31.593, -29.410,
-27.178, -24.839, -22.384, -38.191, -38.085, -37.193, -35.672, -33.741, -31.614,
-29.429, -27.195, -24.855, -22.398, -38.218, -38.111, -37.219, -35.697, -33.764,
-31.635, -29.448, -27.213, -24.870, -22.413, -38.244, -38.137, -37.245, -35.721,
-33.786, -31.656, -29.467, -27.230, -24.886, -22.427, -38.270, -38.163, -37.270,
-35.745, -33.809, -31.677, -29.485, -27.247, -24.902, -22.442, -38.297, -38.190,
-37.296, -35.770, -33.832, -31.697, -29.504, -27.264, -24.917, -22.457, -38.323,
-38.216, -37.321, -35.794, -33.854, -31.718, -29.523, -27.281, -24.933, -22.471,
 -38.349, -38.242, -37.347, -35.818, -33.877, -31.739, -29.542, -27.298,
-24.948, -22.486, -38.376, -38.268, -37.373, -35.843, -33.900, -31.760, -29.561,
-27.315, -24.964, -22.500, -38.402, -38.295, -37.398, -35.867, -33.922, -31.781,
-29.580, -27.332, -24.979, -22.515, -38.428, -38.321, -37.424, -35.892, -33.945,
-31.801, -29.599, -27.349, -24.995, -22.529, -38.454, -38.347, -37.449, -35.916,
-33.968, -31.822, -29.618, -27.366, -25.011, -22.544, -38.481, -38.373, -37.475,
-35.940, -33.991, -31.843, -29.636, -27.384, -25.026, -22.558, -38.507, -38.400,
-37.501, -35.965, -34.013, -31.864, -29.655, -27.401, -25.042, -22.572, -38.533,
-38.426, -37.526, -35.989, -34.036, -31.884, -29.674, -27.418, -25.057, -22.587,
 -38.560, -38.452, -37.552, -36.013, -34.059, -31.905, -29.693, -27.435,
-25.073, -22.601, -38.586, -38.478, -37.577, -36.038, -34.081, -31.926, -29.712,
-27.452, -25.088, -22.616, -38.612, -38.505, -37.603, -36.062, -34.104, -31.947,
-29.731, -27.469, -25.104, -22.630, -38.639, -38.531, -37.629, -36.087, -34.127,
-31.967, -29.750, -27.486, -25.120, -22.645, -38.665, -38.557, -37.654, -36.111,
-34.149, -31.988, -29.768, -27.503, -25.135, -22.659, -38.691, -38.583, -37.680,
-36.135, -34.172, -32.009, -29.787, -27.520, -25.151, -22.673, -38.717, -38.610,
-37.705, -36.160, -34.195, -32.030, -29.806, -27.537, -25.166, -22.688, -38.744,
-38.636, -37.731, -36.184, -34.217, -32.051, -29.825, -27.554, -25.182, -22.702,
 -38.770, -38.662, -37.757, -36.208, -34.240, -32.071, -29.844, -27.572,
-25.197, -22.717, -38.796, -38.688, -37.782, -36.233, -34.263, -32.092, -29.863,
-27.589, -25.213, -22.731, -38.823, -38.715, -37.808, -36.257, -34.285, -32.113,
-29.882, -27.606, -25.228, -22.745, -38.849, -38.741, -37.833, -36.282, -34.308,
-32.134, -29.900, -27.623, -25.244, -22.760, -38.875, -38.767, -37.859, -36.306,
-34.331, -32.154, -29.919, -27.640, -25.259, -22.774, -38.901, -38.793, -37.884,
-36.330, -34.353, -32.175, -29.938, -27.657, -25.275, -22.788, -38.928, -38.819,
-37.910, -36.355, -34.376, -32.196, -29.957, -27.674, -25.291, -22.803, -38.954,
-38.846, -37.936, -36.379, -34.399, -32.217, -29.976, -27.691, -25.306, -22.817,
 -38.980, -38.872, -37.961, -36.403, -34.421, -32.237, -29.995, -27.708,
-25.322, -22.831, -39.007, -38.898, -37.987, -36.428, -34.444, -32.258, -30.013,
-27.725, -25.337, -22.846, -39.033, -38.924, -38.012, -36.452, -34.467, -32.279,
-30.032, -27.742, -25.353, -22.860, -39.059, -38.951, -38.038, -36.476, -34.489,
-32.300, -30.051, -27.759, -25.368, -22.874, -39.085, -38.977, -38.063, -36.501,
-34.512, -32.320, -30.070, -27.776, -25.384, -22.888, -39.112, -39.003, -38.089,
-36.525, -34.535, -32.341, -30.089, -27.794, -25.399, -22.903, -39.138, -39.029,
-38.115, -36.550, -34.557, -32.362, -30.108, -27.811, -25.415, -22.917, -39.164,
-39.055, -38.140, -36.574, -34.580, -32.383, -30.126, -27.828, -25.430, -22.931,
 -39.191, -39.082, -38.166, -36.598, -34.603, -32.403, -30.145, -27.845,
-25.446, -22.945, -39.217, -39.108, -38.191, -36.623, -34.625, -32.424, -30.164,
-27.862, -25.461, -22.960, -39.243, -39.134, -38.217, -36.647, -34.648, -32.445,
-30.183, -27.879, -25.477, -22.974, -39.269, -39.160, -38.242, -36.671, -34.671,
-32.466, -30.202, -27.896, -25.492, -22.988, -39.296, -39.187, -38.268, -36.696,
-34.693, -32.486, -30.221, -27.913, -25.508, -23.002, -39.322, -39.213, -38.294,
-36.720, -34.716, -32.507, -30.239, -27.930, -25.523, -23.017, -39.348, -39.239,
-38.319, -36.744, -34.739, -32.528, -30.258, -27.947, -25.539, -23.031, -39.375,
-39.265, -38.345, -36.769, -34.761, -32.549, -30.277, -27.964, -25.554, -23.045,
 -39.401, -39.291, -38.370, -36.793, -34.784, -32.569, -30.296, -27.981,
-25.570, -23.059, -39.427, -39.318, -38.396, -36.817, -34.807, -32.590, -30.315,
-27.998, -25.585, -23.074, -39.453, -39.344, -38.421, -36.842, -34.829, -32.611,
-30.334, -28.015, -25.601, -23.088, -39.480, -39.370, -38.447, -36.866, -34.852,
-32.632, -30.352, -28.032, -25.616, -23.102, -39.506, -39.396, -38.473, -36.890,
-34.875, -32.652, -30.371, -28.049, -25.632, -23.116, -39.532, -39.422, -38.498,
-36.915, -34.897, -32.673, -30.390, -28.066, -25.647, -23.130, -39.558, -39.449,
-38.524, -36.939, -34.920, -32.694, -30.409, -28.083, -25.663, -23.144, -39.585,
-39.475, -38.549, -36.963, -34.943, -32.715, -30.428, -28.101, -25.678, -23.159,
 -39.611, -39.501, -38.575, -36.988, -34.965, -32.735, -30.447, -28.118,
-25.694, -23.173, -39.637, -39.527, -38.600, -37.012, -34.988, -32.756, -30.465,
-28.135, -25.709, -23.187, -39.663, -39.553, -38.626, -37.036, -35.010, -32.777,
-30.484, -28.152, -25.724, -23.201, -39.690, -39.580, -38.651, -37.061, -35.033,
-32.798, -30.503, -28.169, -25.740, -23.215, -39.716, -39.606, -38.677, -37.085,
-35.056, -32.818, -30.522, -28.186, -25.755, -23.229, -39.742, -39.632, -38.703,
-37.109, -35.078, -32.839, -30.541, -28.203, -25.771, -23.243, -39.769, -39.658,
-38.728, -37.134, -35.101, -32.860, -30.559, -28.220, -25.786, -23.258, -39.795,
-39.684, -38.754, -37.158, -35.124, -32.880, -30.578, -28.237, -25.802, -23.272,
 -39.821, -39.711, -38.779, -37.182, -35.146, -32.901, -30.597, -28.254,
-25.817, -23.286, -39.847, -39.737, -38.805, -37.207, -35.169, -32.922, -30.616,
-28.271, -25.833, -23.300, -39.874, -39.763, -38.830, -37.231, -35.192, -32.943,
-30.635, -28.288, -25.848, -23.314, -39.900, -39.789, -38.856, -37.255, -35.214,
-32.963, -30.653, -28.305, -25.864, -23.328, -39.926, -39.815, -38.881, -37.280,
-35.237, -32.984, -30.672, -28.322, -25.879, -23.342, -39.952, -39.842, -38.907,
-37.304, -35.259, -33.005, -30.691, -28.339, -25.895, -23.356, -39.979, -39.868,
-38.932, -37.328, -35.282, -33.025, -30.710, -28.356, -25.910, -23.370, -40.005,
-39.894, -38.958, -37.353, -35.305, -33.046, -30.729, -28.373, -25.925, -23.384,
 -40.031, -39.920, -38.984, -37.377, -35.327, -33.067, -30.747, -28.390,
-25.941, -23.399, -40.057, -39.946, -39.009, -37.401, -35.350, -33.088, -30.766,
-28.407, -25.956, -23.413, -40.084, -39.973, -39.035, -37.426, -35.373, -33.108,
-30.785, -28.424, -25.972, -23.427, -40.110, -39.999, -39.060, -37.450, -35.395,
-33.129, -30.804, -28.441, -25.987, -23.441, -40.136, -40.025, -39.086, -37.474,
-35.418, -33.150, -30.823, -28.458, -26.003, -23.455, -40.162, -40.051, -39.111,
-37.499, -35.441, -33.171, -30.841, -28.475, -26.018, -23.469, -40.189, -40.077,
-39.137, -37.523, -35.463, -33.191, -30.860, -28.492, -26.034, -23.483, -40.215,
-40.104, -39.162, -37.547, -35.486, -33.212, -30.879, -28.509, -26.049, -23.497,
 -40.241, -40.130, -39.188, -37.572, -35.508, -33.233, -30.898, -28.526,
-26.064, -23.511, -40.267, -40.156, -39.213, -37.596, -35.531, -33.253, -30.917,
-28.543, -26.080, -23.525, -40.294, -40.182, -39.239, -37.620, -35.554, -33.274,
-30.935, -28.560, -26.095, -23.539, -40.320, -40.208, -39.264, -37.644, -35.576,
-33.295, -30.954, -28.577, -26.111, -23.553, -40.346, -40.234, -39.290, -37.669,
-35.599, -33.315, -30.973, -28.594, -26.126, -23.567, -40.372, -40.261, -39.315,
-37.693, -35.621, -33.336, -30.992, -28.611, -26.142, -23.581, -40.399, -40.287,
-39.341, -37.717, -35.644, -33.357, -31.010, -28.628, -26.157, -23.595, -40.425,
-40.313, -39.366, -37.742, -35.667, -33.378, -31.029, -28.645, -26.172, -23.609,
 -40.451, -40.339, -39.392, -37.766, -35.689, -33.398, -31.048, -28.662,
-26.188, -23.623, -40.477, -40.365, -39.418, -37.790, -35.712, -33.419, -31.067,
-28.679, -26.203, -23.637, -40.504, -40.392, -39.443, -37.815, -35.735, -33.440,
-31.086, -28.696, -26.219, -23.651, -40.530, -40.418, -39.469, -37.839, -35.757,
-33.460, -31.104, -28.713, -26.234, -23.665, -40.556, -40.444, -39.494, -37.863,
-35.780, -33.481, -31.123, -28.730, -26.249, -23.679, -40.582, -40.470, -39.520,
-37.888, -35.802, -33.502, -31.142, -28.747, -26.265, -23.693, -40.609, -40.496,
-39.545, -37.912, -35.825, -33.522, -31.161, -28.764, -26.280, -23.707, -40.635,
-40.522, -39.571, -37.936, -35.848, -33.543, -31.179, -28.781, -26.296, -23.721,
 -40.661, -40.549, -39.596, -37.960, -35.870, -33.564, -31.198, -28.798,
-26.311, -23.735, -40.687, -40.575, -39.622, -37.985, -35.893, -33.585, -31.217,
-28.815, -26.327, -23.749, -40.714, -40.601, -39.647, -38.009, -35.915, -33.605,
-31.236, -28.832, -26.342, -23.763, -40.740, -40.627, -39.673, -38.033, -35.938,
-33.626, -31.255, -28.849, -26.357, -23.777, -40.766, -40.653, -39.698, -38.058,
-35.961, -33.647, -31.273, -28.866, -26.373, -23.791, -40.792, -40.679, -39.724,
-38.082, -35.983, -33.667, -31.292, -28.883, -26.388, -23.805, -40.818, -40.706,
-39.749, -38.106, -36.006, -33.688, -31.311, -28.900, -26.404, -23.819, -40.845,
-40.732, -39.775, -38.130, -36.028, -33.709, -31.330, -28.917, -26.419, -23.833,
 -40.871, -40.758, -39.800, -38.155, -36.051, -33.729, -31.348, -28.934,
-26.434, -23.847, -40.897, -40.784, -39.826, -38.179, -36.074, -33.750, -31.367,
-28.951, -26.450, -23.861, -40.923, -40.810, -39.851, -38.203, -36.096, -33.771,
-31.386, -28.968, -26.465, -23.875, -40.950, -40.836, -39.877, -38.228, -36.119,
-33.791, -31.405, -28.985, -26.480, -23.888, -40.976, -40.863, -39.902, -38.252,
-36.141, -33.812, -31.423, -29.002, -26.496, -23.902, -41.002, -40.889, -39.928,
-38.276, -36.164, -33.833, -31.442, -29.019, -26.511, -23.916, -41.028, -40.915,
-39.953, -38.301, -36.187, -33.853, -31.461, -29.036, -26.527, -23.930, -41.055,
-40.941, -39.979, -38.325, -36.209, -33.874, -31.480, -29.053, -26.542, -23.944,
 -41.081, -40.967, -40.004, -38.349, -36.232, -33.895, -31.498, -29.070,
-26.557, -23.958, -41.107, -40.993, -40.030, -38.373, -36.254, -33.916, -31.517,
-29.087, -26.573, -23.972, -41.133, -41.019, -40.055, -38.398, -36.277, -33.936,
-31.536, -29.104, -26.588, -23.986, -41.159, -41.046, -40.081, -38.422, -36.300,
-33.957, -31.555, -29.121, -26.604, -24.000, -41.186, -41.072, -40.106, -38.446,
-36.322, -33.978, -31.574, -29.138, -26.619, -24.014, -41.212, -41.098, -40.132,
-38.470, -36.345, -33.998, -31.592, -29.155, -26.634, -24.028, -41.238, -41.124,
-40.157, -38.495, -36.367, -34.019, -31.611, -29.172, -26.650, -24.042, -41.264,
-41.150, -40.183, -38.519, -36.390, -34.040, -31.630, -29.189, -26.665, -24.055,
 -41.291, -41.176, -40.208, -38.543, -36.413, -34.060, -31.649, -29.206,
-26.680, -24.069, -41.317, -41.203, -40.234, -38.568, -36.435, -34.081, -31.667,
-29.223, -26.696, -24.083, -41.343, -41.229, -40.259, -38.592, -36.458, -34.102,
-31.686, -29.240, -26.711, -24.097, -41.369, -41.255, -40.285, -38.616, -36.480,
-34.122, -31.705, -29.257, -26.727, -24.111, -41.395, -41.281, -40.310, -38.640,
-36.503, -34.143, -31.724, -29.274, -26.742, -24.125, -41.422, -41.307, -40.336,
-38.665, -36.525, -34.164, -31.742, -29.291, -26.757, -24.139, -41.448, -41.333,
-40.361, -38.689, -36.548, -34.184, -31.761, -29.308, -26.773, -24.153, -41.474,
-41.359, -40.387, -38.713, -36.571, -34.205, -31.780, -29.325, -26.788, -24.167,
 -41.500, -41.386, -40.412, -38.738, -36.593, -34.226, -31.798, -29.342,
-26.803, -24.180, -41.527, -41.412, -40.438, -38.762, -36.616, -34.246, -31.817,
-29.358, -26.819, -24.194, -41.553, -41.438, -40.463, -38.786, -36.638, -34.267,
-31.836, -29.375, -26.834, -24.208, -41.579, -41.464, -40.489, -38.810, -36.661,
-34.288, -31.855, -29.392, -26.849, -24.222, -41.605, -41.490, -40.514, -38.835,
-36.684, -34.308, -31.873, -29.409, -26.865, -24.236, -41.631, -41.516, -40.540,
-38.859, -36.706, -34.329, -31.892, -29.426, -26.880, -24.250, -41.658, -41.542,
-40.565, -38.883, -36.729, -34.350, -31.911, -29.443, -26.895, -24.264, -41.684,
-41.569, -40.591, -38.907, -36.751, -34.370, -31.930, -29.460, -26.911, -24.277,
 -41.710, -41.595, -40.616, -38.932, -36.774, -34.391, -31.948, -29.477,
-26.926, -24.291, -41.736, -41.621, -40.642, -38.956, -36.796, -34.412, -31.967,
-29.494, -26.941, -24.305, -41.763, -41.647, -40.667, -38.980, -36.819, -34.432,
-31.986, -29.511, -26.957, -24.319, -41.789, -41.673, -40.692, -39.004, -36.842,
-34.453, -32.005, -29.528, -26.972, -24.333, -41.815, -41.699, -40.718, -39.029,
-36.864, -34.473, -32.023, -29.545, -26.987, -24.347, -41.841, -41.725, -40.743,
-39.053, -36.887, -34.494, -32.042, -29.562, -27.003, -24.360, -41.867, -41.751,
-40.769, -39.077, -36.909, -34.515, -32.061, -29.579, -27.018, -24.374, -41.894,
-41.778, -40.794, -39.102, -36.932, -34.535, -32.079, -29.596, -27.034, -24.388,
 -41.920, -41.804, -40.820, -39.126, -36.954, -34.556, -32.098, -29.613,
-27.049, -24.402, -41.946, -41.830, -40.845, -39.150, -36.977, -34.577, -32.117,
-29.630, -27.064, -24.416, -41.972, -41.856, -40.871, -39.174, -37.000, -34.597,
-32.136, -29.647, -27.080, -24.430, -41.998, -41.882, -40.896, -39.199, -37.022,
-34.618, -32.154, -29.664, -27.095, -24.443, -42.025, -41.908, -40.922, -39.223,
-37.045, -34.639, -32.173, -29.680, -27.110, -24.457, -42.051, -41.934, -40.947,
-39.247, -37.067, -34.659, -32.192, -29.697, -27.126, -24.471, -42.077, -41.961,
-40.973, -39.271, -37.090, -34.680, -32.211, -29.714, -27.141, -24.485, -42.103,
-41.987, -40.998, -39.296, -37.112, -34.701, -32.229, -29.731, -27.156, -24.499,
 -42.129, -42.013, -41.024, -39.320, -37.135, -34.721, -32.248, -29.748,
-27.171, -24.512, -42.156, -42.039, -41.049, -39.344, -37.157, -34.742, -32.267,
-29.765, -27.187, -24.526, -42.182, -42.065, -41.075, -39.368, -37.180, -34.763,
-32.285, -29.782, -27.202, -24.540, -42.208, -42.091, -41.100, -39.393, -37.203,
-34.783, -32.304, -29.799, -27.217, -24.554, -42.234, -42.117, -41.125, -39.417,
-37.225, -34.804, -32.323, -29.816, -27.233, -24.568, -42.260, -42.143, -41.151,
-39.441, -37.248, -34.824, -32.342, -29.833, -27.248, -24.581, -42.287, -42.170,
-41.176, -39.465, -37.270, -34.845, -32.360, -29.850, -27.263, -24.595, -42.313,
-42.196, -41.202, -39.490, -37.293, -34.866, -32.379, -29.867, -27.279, -24.609,
 -42.339, -42.222, -41.227, -39.514, -37.315, -34.886, -32.398, -29.884,
-27.294, -24.623, -42.365, -42.248, -41.253, -39.538, -37.338, -34.907, -32.416,
-29.901, -27.309, -24.637, -42.391, -42.274, -41.278, -39.562, -37.360, -34.928,
-32.435, -29.917, -27.325, -24.650, -42.418, -42.300, -41.304, -39.587, -37.383,
-34.948, -32.454, -29.934, -27.340, -24.664, -42.444, -42.326, -41.329, -39.611,
-37.406, -34.969, -32.473, -29.951, -27.355, -24.678, -42.470, -42.352, -41.355,
-39.635, -37.428, -34.990, -32.491, -29.968, -27.371, -24.692, -42.496, -42.378,
-41.380, -39.659, -37.451, -35.010, -32.510, -29.985, -27.386, -24.705, -42.522,
-42.405, -41.405, -39.683, -37.473, -35.031, -32.529, -30.002, -27.401, -24.719,
 -42.549, -42.431, -41.431, -39.708, -37.496, -35.051, -32.547, -30.019,
-27.416, -24.733, -42.575, -42.457, -41.456, -39.732, -37.518, -35.072, -32.566,
-30.036, -27.432, -24.747, -42.601, -42.483, -41.482, -39.756, -37.541, -35.093,
-32.585, -30.053, -27.447, -24.761, -42.627, -42.509, -41.507, -39.780, -37.563,
-35.113, -32.603, -30.070, -27.462, -24.774, -42.653, -42.535, -41.533, -39.805,
-37.586, -35.134, -32.622, -30.087, -27.478, -24.788, -42.680, -42.561, -41.558,
-39.829, -37.608, -35.155, -32.641, -30.104, -27.493, -24.802, -42.706, -42.587,
-41.584, -39.853, -37.631, -35.175, -32.660, -30.120, -27.508, -24.816, -42.732,
-42.614, -41.609, -39.877, -37.654, -35.196, -32.678, -30.137, -27.524, -24.829,
 -42.758, -42.640, -41.635, -39.902, -37.676, -35.216, -32.697, -30.154,
-27.539, -24.843, -42.784, -42.666, -41.660, -39.926, -37.699, -35.237, -32.716,
-30.171, -27.554, -24.857, -42.811, -42.692, -41.685, -39.950, -37.721, -35.258,
-32.734, -30.188, -27.569, -24.871, -42.837, -42.718, -41.711, -39.974, -37.744,
-35.278, -32.753, -30.205, -27.585, -24.884, -42.863, -42.744, -41.736, -39.999,
-37.766, -35.299, -32.772, -30.222, -27.600, -24.898, -42.889, -42.770, -41.762,
-40.023, -37.789, -35.320, -32.790, -30.239, -27.615, -24.912, -42.915, -42.796,
-41.787, -40.047, -37.811, -35.340, -32.809, -30.256, -27.631, -24.926, -42.941,
-42.822, -41.813, -40.071, -37.834, -35.361, -32.828, -30.273, -27.646, -24.939,
 -42.968, -42.848, -41.838, -40.095, -37.856, -35.381, -32.847, -30.290,
-27.661, -24.953, -42.994, -42.875, -41.864, -40.120, -37.879, -35.402, -32.865,
-30.306, -27.676, -24.967, -43.020, -42.901, -41.889, -40.144, -37.901, -35.423,
-32.884, -30.323, -27.692, -24.981, -43.046, -42.927, -41.914, -40.168, -37.924,
-35.443, -32.903, -30.340, -27.707, -24.994, -43.072, -42.953, -41.940, -40.192,
-37.947, -35.464, -32.921, -30.357, -27.722, -25.008, -43.099, -42.979, -41.965,
-40.217, -37.969, -35.484, -32.940, -30.374, -27.738, -25.022, -43.125, -43.005,
-41.991, -40.241, -37.992, -35.505, -32.959, -30.391, -27.753, -25.035, -43.151,
-43.031, -42.016, -40.265, -38.014, -35.526, -32.977, -30.408, -27.768, -25.049,
 -43.177, -43.057, -42.042, -40.289, -38.037, -35.546, -32.996, -30.425,
-27.783, -25.063, -43.203, -43.083, -42.067, -40.313, -38.059, -35.567, -33.015,
-30.442, -27.799, -25.077, -43.229, -43.109, -42.092, -40.338, -38.082, -35.588,
-33.033, -30.458, -27.814, -25.090, -43.256, -43.136, -42.118, -40.362, -38.104,
-35.608, -33.052, -30.475, -27.829, -25.104, -43.282, -43.162, -42.143, -40.386,
-38.127, -35.629, -33.071, -30.492, -27.844, -25.118, -43.308, -43.188, -42.169,
-40.410, -38.149, -35.649, -33.089, -30.509, -27.860, -25.131, -43.334, -43.214,
-42.194, -40.435, -38.172, -35.670, -33.108, -30.526, -27.875, -25.145, -43.360,
-43.240, -42.220, -40.459, -38.194, -35.691, -33.127, -30.543, -27.890, -25.159,
 -43.387, -43.266, -42.245, -40.483, -38.217, -35.711, -33.145, -30.560,
-27.906, -25.173, -43.413, -43.292, -42.270, -40.507, -38.239, -35.732, -33.164,
-30.577, -27.921, -25.186, -43.439, -43.318, -42.296, -40.531, -38.262, -35.752,
-33.183, -30.594, -27.936, -25.200, -43.465, -43.344, -42.321, -40.556, -38.284,
-35.773, -33.201, -30.610, -27.951, -25.214, -43.491, -43.370, -42.347, -40.580,
-38.307, -35.794, -33.220, -30.627, -27.967, -25.227, -43.517, -43.396, -42.372,
-40.604, -38.329, -35.814, -33.239, -30.644, -27.982, -25.241, -43.544, -43.423,
-42.398, -40.628, -38.352, -35.835, -33.258, -30.661, -27.997, -25.255, -43.570,
-43.449, -42.423, -40.652, -38.375, -35.855, -33.276, -30.678, -28.012, -25.268,
 -43.596, -43.475, -42.448, -40.677, -38.397, -35.876, -33.295, -30.695,
-28.028, -25.282, -43.622, -43.501, -42.474, -40.701, -38.420, -35.897, -33.314,
-30.712, -28.043, -25.296, -43.648, -43.527, -42.499, -40.725, -38.442, -35.917,
-33.332, -30.729, -28.058, -25.309, -43.674, -43.553, -42.525, -40.749, -38.465,
-35.938, -33.351, -30.745, -28.073, -25.323, -43.701, -43.579, -42.550, -40.773,
-38.487, -35.958, -33.370, -30.762, -28.089, -25.337, -43.727, -43.605, -42.576,
-40.798, -38.510, -35.979, -33.388, -30.779, -28.104, -25.351, -43.753, -43.631,
-42.601, -40.822, -38.532, -36.000, -33.407, -30.796, -28.119, -25.364, -43.779,
-43.657, -42.626, -40.846, -38.555, -36.020, -33.426, -30.813, -28.134, -25.378,
 -43.805, -43.683, -42.652, -40.870, -38.577, -36.041, -33.444, -30.830,
-28.150, -25.392, -43.831, -43.710, -42.677, -40.894, -38.600, -36.061, -33.463,
-30.847, -28.165, -25.405, -43.858, -43.736, -42.703, -40.919, -38.622, -36.082,
-33.482, -30.864, -28.180, -25.419, -43.884, -43.762, -42.728, -40.943, -38.645,
-36.102, -33.500, -30.880, -28.195, -25.433, -43.910, -43.788, -42.753, -40.967,
-38.667, -36.123, -33.519, -30.897, -28.211, -25.446, -43.936, -43.814, -42.779,
-40.991, -38.690, -36.144, -33.537, -30.914, -28.226, -25.460, -43.962, -43.840,
-42.804, -41.015, -38.712, -36.164, -33.556, -30.931, -28.241, -25.474, -43.988,
-43.866, -42.830, -41.040, -38.735, -36.185, -33.575, -30.948, -28.256, -25.487,
 -44.015, -43.892, -42.855, -41.064, -38.757, -36.205, -33.593, -30.965,
-28.271, -25.501, -44.041, -43.918, -42.880, -41.088, -38.780, -36.226, -33.612,
-30.982, -28.287, -25.515, -44.067, -43.944, -42.906, -41.112, -38.802, -36.247,
-33.631, -30.998, -28.302, -25.528, -44.093, -43.970, -42.931, -41.136, -38.825,
-36.267, -33.649, -31.015, -28.317, -25.542, -44.119, -43.996, -42.957, -41.161,
-38.847, -36.288, -33.668, -31.032, -28.332, -25.556, -44.145, -44.022, -42.982,
-41.185, -38.870, -36.308, -33.687, -31.049, -28.348, -25.569, -44.172, -44.049,
-43.007, -41.209, -38.892, -36.329, -33.705, -31.066, -28.363, -25.583, -44.198,
-44.075, -43.033, -41.233, -38.915, -36.349, -33.724, -31.083, -28.378, -25.596,
 -44.224, -44.101, -43.058, -41.257, -38.937, -36.370, -33.743, -31.100,
-28.393, -25.610, -44.250, -44.127, -43.084, -41.282, -38.960, -36.391, -33.761,
-31.116, -28.409, -25.624, -44.276, -44.153, -43.109, -41.306, -38.982, -36.411,
-33.780, -31.133, -28.424, -25.637, -44.302, -44.179, -43.135, -41.330, -39.005,
-36.432, -33.799, -31.150, -28.439, -25.651, -44.328, -44.205, -43.160, -41.354,
-39.027, -36.452, -33.817, -31.167, -28.454, -25.665, -44.355, -44.231, -43.185,
-41.378, -39.050, -36.473, -33.836, -31.184, -28.469, -25.678, -44.381, -44.257,
-43.211, -41.403, -39.072, -36.494, -33.855, -31.201, -28.485, -25.692, -44.407,
-44.283, -43.236, -41.427, -39.095, -36.514, -33.873, -31.218, -28.500, -25.706,
 -44.433, -44.309, -43.262, -41.451, -39.117, -36.535, -33.892, -31.234,
-28.515, -25.719, -44.459, -44.335, -43.287, -41.475, -39.140, -36.555, -33.911,
-31.251, -28.530, -25.733, -44.485, -44.361, -43.312, -41.499, -39.162, -36.576,
-33.929, -31.268, -28.545, -25.747, -44.512, -44.387, -43.338, -41.523, -39.185,
-36.596, -33.948, -31.285, -28.561, -25.760, -44.538, -44.414, -43.363, -41.548,
-39.207, -36.617, -33.966, -31.302, -28.576, -25.774, -44.564, -44.440, -43.388,
-41.572, -39.230, -36.638, -33.985, -31.319, -28.591, -25.787, -44.590, -44.466,
-43.414, -41.596, -39.252, -36.658, -34.004, -31.335, -28.606, -25.801, -44.616,
-44.492, -43.439, -41.620, -39.275, -36.679, -34.022, -31.352, -28.622, -25.815,
 -44.642, -44.518, -43.465, -41.644, -39.297, -36.699, -34.041, -31.369,
-28.637, -25.828, -44.668, -44.544, -43.490, -41.669, -39.320, -36.720, -34.060,
-31.386, -28.652, -25.842, -44.695, -44.570, -43.515, -41.693, -39.342, -36.740,
-34.078, -31.403, -28.667, -25.855, -44.721, -44.596, -43.541, -41.717, -39.365,
-36.761, -34.097, -31.420, -28.682, -25.869, -44.747, -44.622, -43.566, -41.741,
-39.387, -36.781, -34.116, -31.437, -28.698, -25.883, -44.773, -44.648, -43.592,
-41.765, -39.410, -36.802, -34.134, -31.453, -28.713, -25.896, -44.799, -44.674,
-43.617, -41.789, -39.432, -36.823, -34.153, -31.470, -28.728, -25.910, -44.825,
-44.700, -43.642, -41.814, -39.455, -36.843, -34.171, -31.487, -28.743, -25.924,
 -44.852, -44.726, -43.668, -41.838, -39.477, -36.864, -34.190, -31.504,
-28.758, -25.937, -44.878, -44.752, -43.693, -41.862, -39.500, -36.884, -34.209,
-31.521, -28.774, -25.951, -44.904, -44.778, -43.719, -41.886, -39.522, -36.905,
-34.227, -31.538, -28.789, -25.964, -44.930, -44.804, -43.744, -41.910, -39.544,
-36.925, -34.246, -31.554, -28.804, -25.978, -44.956, -44.830, -43.769, -41.935,
-39.567, -36.946, -34.265, -31.571, -28.819, -25.992, -44.982, -44.857, -43.795,
-41.959, -39.589, -36.967, -34.283, -31.588, -28.834, -26.005, -45.008, -44.883,
-43.820, -41.983, -39.612, -36.987, -34.302, -31.605, -28.850, -26.019, -45.035,
-44.909, -43.845, -42.007, -39.634, -37.008, -34.320, -31.622, -28.865, -26.032,
 -45.061, -44.935, -43.871, -42.031, -39.657, -37.028, -34.339, -31.638,
-28.880, -26.046, -45.087, -44.961, -43.896, -42.055, -39.679, -37.049, -34.358,
-31.655, -28.895, -26.060, -45.113, -44.987, -43.922, -42.080, -39.702, -37.069,
-34.376, -31.672, -28.910, -26.073, -45.139, -45.013, -43.947, -42.104, -39.724,
-37.090, -34.395, -31.689, -28.925, -26.087, -45.165, -45.039, -43.972, -42.128,
-39.747, -37.110, -34.414, -31.706, -28.941, -26.100, -45.191, -45.065, -43.998,
-42.152, -39.769, -37.131, -34.432, -31.723, -28.956, -26.114, -45.218, -45.091,
-44.023, -42.176, -39.792, -37.151, -34.451, -31.739, -28.971, -26.128, -45.244,
-45.117, -44.048, -42.200, -39.814, -37.172, -34.469, -31.756, -28.986, -26.141,
 -45.270, -45.143, -44.074, -42.225, -39.837, -37.193, -34.488, -31.773,
-29.001, -26.155, -45.296, -45.169, -44.099, -42.249, -39.859, -37.213, -34.507,
-31.790, -29.017, -26.168, -45.322, -45.195, -44.125, -42.273, -39.882, -37.234,
-34.525, -31.807, -29.032, -26.182, -45.348, -45.221, -44.150, -42.297, -39.904,
-37.254, -34.544, -31.823, -29.047, -26.195, -45.374, -45.247, -44.175, -42.321,
-39.927, -37.275, -34.563, -31.840, -29.062, -26.209, -45.401, -45.273, -44.201,
-42.345, -39.949, -37.295, -34.581, -31.857, -29.077, -26.223, -45.427, -45.299,
-44.226, -42.369, -39.972, -37.316, -34.600, -31.874, -29.092, -26.236, -45.453,
-45.325, -44.251, -42.394, -39.994, -37.336, -34.618, -31.891, -29.108, -26.250,
 -45.479, -45.351, -44.277, -42.418, -40.016, -37.357, -34.637, -31.908,
-29.123, -26.263, -45.505, -45.378, -44.302, -42.442, -40.039, -37.378, -34.656,
-31.924, -29.138, -26.277, -45.531, -45.404, -44.328, -42.466, -40.061, -37.398,
-34.674, -31.941, -29.153, -26.290, -45.557, -45.430, -44.353, -42.490, -40.084,
-37.419, -34.693, -31.958, -29.168, -26.304, -45.583, -45.456, -44.378, -42.514,
-40.106, -37.439, -34.711, -31.975, -29.183, -26.318, -45.610, -45.482, -44.404,
-42.539, -40.129, -37.460, -34.730, -31.992, -29.199, -26.331, -45.636, -45.508,
-44.429, -42.563, -40.151, -37.480, -34.749, -32.008, -29.214, -26.345, -45.662,
-45.534, -44.454, -42.587, -40.174, -37.501, -34.767, -32.025, -29.229, -26.358,
 -45.688, -45.560, -44.480, -42.611, -40.196, -37.521, -34.786, -32.042,
-29.244, -26.372, -45.714, -45.586, -44.505, -42.635, -40.219, -37.542, -34.804,
-32.059, -29.259, -26.385, -45.740, -45.612, -44.531, -42.659, -40.241, -37.562,
-34.823, -32.076, -29.274, -26.399, -45.766, -45.638, -44.556, -42.684, -40.264,
-37.583, -34.842, -32.092, -29.290, -26.413, -45.793, -45.664, -44.581, -42.708,
-40.286, -37.603, -34.860, -32.109, -29.305, -26.426, -45.819, -45.690, -44.607,
-42.732, -40.308, -37.624, -34.879, -32.126, -29.320, -26.440, -45.845, -45.716,
-44.632, -42.756, -40.331, -37.645, -34.898, -32.143, -29.335, -26.453, -45.871,
-45.742, -44.657, -42.780, -40.353, -37.665, -34.916, -32.160, -29.350, -26.467,
 -45.897, -45.768, -44.683, -42.804, -40.376, -37.686, -34.935, -32.176,
-29.365, -26.480, -45.923, -45.794, -44.708, -42.828, -40.398, -37.706, -34.953,
-32.193, -29.380, -26.494, -45.949, -45.820, -44.733, -42.853, -40.421, -37.727,
-34.972, -32.210, -29.396, -26.507, -45.975, -45.846, -44.759, -42.877, -40.443,
-37.747, -34.991, -32.227, -29.411, -26.521, -46.002, -45.872, -44.784, -42.901,
-40.466, -37.768, -35.009, -32.244, -29.426, -26.535, -46.028, -45.898, -44.809,
-42.925, -40.488, -37.788, -35.028, -32.260, -29.441, -26.548, -46.054, -45.924,
-44.835, -42.949, -40.511, -37.809, -35.046, -32.277, -29.456, -26.562, -46.080,
-45.950, -44.860, -42.973, -40.533, -37.829, -35.065, -32.294, -29.471, -26.575,
 -46.106, -45.976, -44.886, -42.997, -40.555, -37.850, -35.083, -32.311,
-29.487, -26.589, -46.132, -46.002, -44.911, -43.022, -40.578, -37.870, -35.102,
-32.328, -29.502, -26.602, -46.158, -46.028, -44.936, -43.046, -40.600, -37.891,
-35.121, -32.344, -29.517, -26.616, -46.184, -46.054, -44.962, -43.070, -40.623,
-37.911, -35.139, -32.361, -29.532, -26.629, -46.211, -46.080, -44.987, -43.094,
-40.645, -37.932, -35.158, -32.378, -29.547, -26.643, -46.237, -46.107, -45.012,
-43.118, -40.668, -37.952, -35.176, -32.395, -29.562, -26.656, -46.263, -46.133,
-45.038, -43.142, -40.690, -37.973, -35.195, -32.412, -29.577, -26.670, -46.289,
-46.159, -45.063, -43.166, -40.713, -37.994, -35.214, -32.428, -29.593, -26.683,
 -46.315, -46.185, -45.088, -43.191, -40.735, -38.014, -35.232, -32.445,
-29.608, -26.697, -46.341, -46.211, -45.114, -43.215, -40.758, -38.035, -35.251,
-32.462, -29.623, -26.710, -46.367, -46.237, -45.139, -43.239, -40.780, -38.055,
-35.269, -32.479, -29.638, -26.724, -46.393, -46.263, -45.164, -43.263, -40.802,
-38.076, -35.288, -32.495, -29.653, -26.738, -46.420, -46.289, -45.190, -43.287,
-40.825, -38.096, -35.307, -32.512, -29.668, -26.751, -46.446, -46.315, -45.215,
-43.311, -40.847, -38.117, -35.325, -32.529, -29.683, -26.765, -46.472, -46.341,
-45.240, -43.335, -40.870, -38.137, -35.344, -32.546, -29.698, -26.778, -46.498,
-46.367, -45.266, -43.359, -40.892, -38.158, -35.362, -32.563, -29.714, -26.792,
 -46.524, -46.393, -45.291, -43.384, -40.915, -38.178, -35.381, -32.579,
-29.729, -26.805, -46.550, -46.419, -45.316, -43.408, -40.937, -38.199, -35.400,
-32.596, -29.744, -26.819, -46.576, -46.445, -45.342, -43.432, -40.960, -38.219,
-35.418, -32.613, -29.759, -26.832, -46.602, -46.471, -45.367, -43.456, -40.982,
-38.240, -35.437, -32.630, -29.774, -26.846, -46.628, -46.497, -45.392, -43.480,
-41.004, -38.260, -35.455, -32.646, -29.789, -26.859, -46.655, -46.523, -45.418,
-43.504, -41.027, -38.281, -35.474, -32.663, -29.804, -26.873, -46.681, -46.549,
-45.443, -43.528, -41.049, -38.301, -35.492, -32.680, -29.819, -26.886, -46.707,
-46.575, -45.469, -43.553, -41.072, -38.322, -35.511, -32.697, -29.835, -26.900,
 -46.733, -46.601, -45.494, -43.577, -41.094, -38.342, -35.530, -32.714,
-29.850, -26.913, -46.759, -46.627, -45.519, -43.601, -41.117, -38.363, -35.548,
-32.730, -29.865, -26.927, -46.785, -46.653, -45.545, -43.625, -41.139, -38.383,
-35.567, -32.747, -29.880, -26.940, -46.811, -46.679, -45.570, -43.649, -41.161,
-38.404, -35.585, -32.764, -29.895, -26.954, -46.837, -46.705, -45.595, -43.673,
-41.184, -38.424, -35.604, -32.781, -29.910, -26.967, -46.863, -46.731, -45.621,
-43.697, -41.206, -38.445, -35.622, -32.797, -29.925, -26.981, -46.890, -46.757,
-45.646, -43.721, -41.229, -38.465, -35.641, -32.814, -29.940, -26.994, -46.916,
-46.783, -45.671, -43.746, -41.251, -38.486, -35.660, -32.831, -29.956, -27.008,
 -46.942, -46.809, -45.697, -43.770, -41.274, -38.506, -35.678, -32.848,
-29.971, -27.021, -46.968, -46.835, -45.722, -43.794, -41.296, -38.527, -35.697,
-32.864, -29.986, -27.035, -46.994, -46.861, -45.747, -43.818, -41.318, -38.547,
-35.715, -32.881, -30.001, -27.048, -47.020, -46.887, -45.773, -43.842, -41.341,
-38.568, -35.734, -32.898, -30.016, -27.062, -47.046, -46.913, -45.798, -43.866,
-41.363, -38.588, -35.753, -32.915, -30.031, -27.075, -47.072, -46.939, -45.823,
-43.890, -41.386, -38.609, -35.771, -32.931, -30.046, -27.089, -47.098, -46.965,
-45.849, -43.914, -41.408, -38.629, -35.790, -32.948, -30.061, -27.102, -47.125,
-46.991, -45.874, -43.939, -41.431, -38.650, -35.808, -32.965, -30.076, -27.116,
 -47.151, -47.017, -45.899, -43.963, -41.453, -38.670, -35.827, -32.982,
-30.091, -27.129, -47.177, -47.043, -45.925, -43.987, -41.476, -38.691, -35.845,
-32.999, -30.107, -27.143, -47.203, -47.069, -45.950, -44.011, -41.498, -38.711,
-35.864, -33.015, -30.122, -27.156, -47.229, -47.095, -45.975, -44.035, -41.520,
-38.732, -35.882, -33.032, -30.137, -27.170, -47.255, -47.121, -46.001, -44.059,
-41.543, -38.752, -35.901, -33.049, -30.152, -27.183, -47.281, -47.147, -46.026,
-44.083, -41.565, -38.773, -35.920, -33.066, -30.167, -27.197, -47.307, -47.173,
-46.051, -44.107, -41.588, -38.793, -35.938, -33.082, -30.182, -27.210, -47.333,
-47.199, -46.076, -44.131, -41.610, -38.814, -35.957, -33.099, -30.197, -27.224,
 -47.360, -47.225, -46.102, -44.156, -41.632, -38.834, -35.975, -33.116,
-30.212, -27.237, -47.386, -47.251, -46.127, -44.180, -41.655, -38.855, -35.994,
-33.133, -30.227, -27.251, -47.412, -47.277, -46.152, -44.204, -41.677, -38.875,
-36.012, -33.149, -30.242, -27.264, -47.438, -47.303, -46.178, -44.228, -41.700,
-38.896, -36.031, -33.166, -30.258, -27.277, -47.464, -47.329, -46.203, -44.252,
-41.722, -38.916, -36.050, -33.183, -30.273, -27.291, -47.490, -47.355, -46.228,
-44.276, -41.745, -38.937, -36.068, -33.200, -30.288, -27.304, -47.516, -47.381,
-46.254, -44.300, -41.767, -38.957, -36.087, -33.216, -30.303, -27.318, -47.542,
-47.407, -46.279, -44.324, -41.789, -38.978, -36.105, -33.233, -30.318, -27.331,
 -47.568, -47.433, -46.304, -44.348, -41.812, -38.998, -36.124, -33.250,
-30.333, -27.345, -47.594, -47.459, -46.330, -44.373, -41.834, -39.019, -36.142,
-33.267, -30.348, -27.358, -47.621, -47.485, -46.355, -44.397, -41.857, -39.039,
-36.161, -33.283, -30.363, -27.372, -47.647, -47.511, -46.380, -44.421, -41.879,
-39.060, -36.179, -33.300, -30.378, -27.385, -47.673, -47.537, -46.406, -44.445,
-41.902, -39.080, -36.198, -33.317, -30.393, -27.399, -47.699, -47.563, -46.431,
-44.469, -41.924, -39.101, -36.217, -33.334, -30.408, -27.412, -47.725, -47.589,
-46.456, -44.493, -41.946, -39.121, -36.235, -33.350, -30.424, -27.426, -47.751,
-47.615, -46.482, -44.517, -41.969, -39.142, -36.254, -33.367, -30.439, -27.439,
 -47.777, -47.641, -46.507, -44.541, -41.991, -39.162, -36.272, -33.384,
-30.454, -27.452, -47.803, -47.667, -46.532, -44.565, -42.014, -39.183, -36.291,
-33.400, -30.469, -27.466, -47.829, -47.693, -46.558, -44.590, -42.036, -39.203,
-36.309, -33.417, -30.484, -27.479, -47.855, -47.719, -46.583, -44.614, -42.058,
-39.224, -36.328, -33.434, -30.499, -27.493, -47.882, -47.745, -46.608, -44.638,
-42.081, -39.244, -36.346, -33.451, -30.514, -27.506, -47.908, -47.771, -46.634,
-44.662, -42.103, -39.265, -36.365, -33.467, -30.529, -27.520, -47.934, -47.797,
-46.659, -44.686, -42.126, -39.285, -36.384, -33.484, -30.544, -27.533, -47.960,
-47.823, -46.684, -44.710, -42.148, -39.306, -36.402, -33.501, -30.559, -27.547,
 -47.986, -47.849, -46.709, -44.734, -42.171, -39.326, -36.421, -33.518,
-30.574, -27.560, -48.012, -47.875, -46.735, -44.758, -42.193, -39.347, -36.439,
-33.534, -30.589, -27.574, -48.038, -47.901, -46.760, -44.782, -42.215, -39.367,
-36.458, -33.551, -30.604, -27.587, -48.064, -47.927, -46.785, -44.806, -42.238,
-39.388, -36.476, -33.568, -30.620, -27.600, -48.090, -47.953, -46.811, -44.831,
-42.260, -39.408, -36.495, -33.585, -30.635, -27.614, -48.116, -47.979, -46.836,
-44.855, -42.283, -39.429, -36.513, -33.601, -30.650, -27.627, -48.142, -48.005,
-46.861, -44.879, -42.305, -39.449, -36.532, -33.618, -30.665, -27.641, -48.169,
-48.031, -46.887, -44.903, -42.327, -39.470, -36.550, -33.635, -30.680, -27.654,
 -48.195, -48.057, -46.912, -44.927, -42.350, -39.490, -36.569, -33.652,
-30.695, -27.668, -48.221, -48.083, -46.937, -44.951, -42.372, -39.511, -36.588,
-33.668, -30.710, -27.681, -48.247, -48.109, -46.963, -44.975, -42.395, -39.531,
-36.606, -33.685, -30.725, -27.694, -48.273, -48.135, -46.988, -44.999, -42.417,
-39.551, -36.625, -33.702, -30.740, -27.708, -48.299, -48.161, -47.013, -45.023,
-42.439, -39.572, -36.643, -33.718, -30.755, -27.721, -48.325, -48.187, -47.038,
-45.047, -42.462, -39.592, -36.662, -33.735, -30.770, -27.735, -48.351, -48.213,
-47.064, -45.072, -42.484, -39.613, -36.680, -33.752, -30.785, -27.748, -48.377,
-48.239, -47.089, -45.096, -42.507, -39.633, -36.699, -33.769, -30.800, -27.762,
 -48.403, -48.265, -47.114, -45.120, -42.529, -39.654, -36.717, -33.785,
-30.815, -27.775, -48.429, -48.291, -47.140, -45.144, -42.551, -39.674, -36.736,
-33.802, -30.830, -27.788, -48.456, -48.317, -47.165, -45.168, -42.574, -39.695,
-36.754, -33.819, -30.845, -27.802, -48.482, -48.343, -47.190, -45.192, -42.596,
-39.715, -36.773, -33.835, -30.860, -27.815, -48.508, -48.369, -47.216, -45.216,
-42.619, -39.736, -36.792, -33.852, -30.876, -27.829, -48.534, -48.395, -47.241,
-45.240, -42.641, -39.756, -36.810, -33.869, -30.891, -27.842, -48.560, -48.421,
-47.266, -45.264, -42.663, -39.777, -36.829, -33.886, -30.906, -27.856, -48.586,
-48.447, -47.291, -45.288, -42.686, -39.797, -36.847, -33.902, -30.921, -27.869,
 -48.612, -48.473, -47.317, -45.312, -42.708, -39.818, -36.866, -33.919,
-30.936, -27.882, -48.638, -48.499, -47.342, -45.337, -42.731, -39.838, -36.884,
-33.936, -30.951, -27.896, -48.664, -48.525, -47.367, -45.361, -42.753, -39.859,
-36.903, -33.953, -30.966, -27.909, -48.690, -48.551, -47.393, -45.385, -42.775,
-39.879, -36.921, -33.969, -30.981, -27.923, -48.716, -48.577, -47.418, -45.409,
-42.798, -39.900, -36.940, -33.986, -30.996, -27.936, -48.742, -48.603, -47.443,
-45.433, -42.820, -39.920, -36.958, -34.003, -31.011, -27.950, -48.769, -48.629,
-47.469, -45.457, -42.843, -39.940, -36.977, -34.019, -31.026, -27.963, -48.795,
-48.655, -47.494, -45.481, -42.865, -39.961, -36.995, -34.036, -31.041, -27.976,
 -48.821, -48.681, -47.519, -45.505, -42.887, -39.981, -37.014, -34.053,
-31.056, -27.990, -48.847, -48.707, -47.544, -45.529, -42.910, -40.002, -37.032,
-34.070, -31.071, -28.003, -48.873, -48.733, -47.570, -45.553, -42.932, -40.022,
-37.051, -34.086, -31.086, -28.017, -48.899, -48.759, -47.595, -45.577, -42.955,
-40.043, -37.069, -34.103, -31.101, -28.030, -48.925, -48.785, -47.620, -45.601,
-42.977, -40.063, -37.088, -34.120, -31.116, -28.043, -48.951, -48.811, -47.646,
-45.626, -42.999, -40.084, -37.107, -34.136, -31.131, -28.057, -48.977, -48.837,
-47.671, -45.650, -43.022, -40.104, -37.125, -34.153, -31.146, -28.070, -49.003,
-48.863, -47.696, -45.674, -43.044, -40.125, -37.144, -34.170, -31.161, -28.084,
 -49.029, -48.889, -47.721, -45.698, -43.066, -40.145, -37.162, -34.186,
-31.176, -28.097, -49.055, -48.915, -47.747, -45.722, -43.089, -40.166, -37.181,
-34.203, -31.191, -28.110, -49.082, -48.941, -47.772, -45.746, -43.111, -40.186,
-37.199, -34.220, -31.206, -28.124, -49.108, -48.967, -47.797, -45.770, -43.134,
-40.206, -37.218, -34.237, -31.222, -28.137, -49.134, -48.993, -47.823, -45.794,
-43.156, -40.227, -37.236, -34.253, -31.237, -28.151, -49.160, -49.019, -47.848,
-45.818, -43.178, -40.247, -37.255, -34.270, -31.252, -28.164, -49.186, -49.045,
-47.873, -45.842, -43.201, -40.268, -37.273, -34.287, -31.267, -28.177, -49.212,
-49.071, -47.898, -45.866, -43.223, -40.288, -37.292, -34.303, -31.282, -28.191,
 -49.238, -49.097, -47.924, -45.890, -43.246, -40.309, -37.310, -34.320,
-31.297, -28.204, -49.264, -49.122, -47.949, -45.914, -43.268, -40.329, -37.329,
-34.337, -31.312, -28.217, -49.290, -49.148, -47.974, -45.939, -43.290, -40.350,
-37.347, -34.353, -31.327, -28.231, -49.316, -49.174, -48.000, -45.963, -43.313,
-40.370, -37.366, -34.370, -31.342, -28.244, -49.342, -49.200, -48.025, -45.987,
-43.335, -40.391, -37.384, -34.387, -31.357, -28.258, -49.368, -49.226, -48.050,
-46.011, -43.358, -40.411, -37.403, -34.404, -31.372, -28.271, -49.394, -49.252,
-48.075, -46.035, -43.380, -40.431, -37.421, -34.420, -31.387, -28.284, -49.421,
-49.278, -48.101, -46.059, -43.402, -40.452, -37.440, -34.437, -31.402, -28.298,
 -49.447, -49.304, -48.126, -46.083, -43.425, -40.472, -37.458, -34.454,
-31.417, -28.311, -49.473, -49.330, -48.151, -46.107, -43.447, -40.493, -37.477,
-34.470, -31.432, -28.325, -49.499, -49.356, -48.177, -46.131, -43.469, -40.513,
-37.495, -34.487, -31.447, -28.338, -49.525, -49.382, -48.202, -46.155, -43.492,
-40.534, -37.514, -34.504, -31.462, -28.351, -49.551, -49.408, -48.227, -46.179,
-43.514, -40.554, -37.532, -34.520, -31.477, -28.365, -49.577, -49.434, -48.252,
-46.203, -43.537, -40.575, -37.551, -34.537, -31.492, -28.378, -49.603, -49.460,
-48.278, -46.227, -43.559, -40.595, -37.570, -34.554, -31.507, -28.391, -49.629,
-49.486, -48.303, -46.251, -43.581, -40.616, -37.588, -34.571, -31.522, -28.405,
 -49.655, -49.512, -48.328, -46.276, -43.604, -40.636, -37.607, -34.587,
-31.537, -28.418, -49.681, -49.538, -48.354, -46.300, -43.626, -40.656, -37.625,
-34.604, -31.552, -28.432, -49.707, -49.564, -48.379, -46.324, -43.648, -40.677,
-37.644, -34.621, -31.567, -28.445, -49.733, -49.590, -48.404, -46.348, -43.671,
-40.697, -37.662, -34.637, -31.582, -28.458, -49.759, -49.616, -48.429, -46.372,
-43.693, -40.718, -37.681, -34.654, -31.597, -28.472, -49.786, -49.642, -48.455,
-46.396, -43.716, -40.738, -37.699, -34.671, -31.612, -28.485, -49.812, -49.668,
-48.480, -46.420, -43.738, -40.759, -37.718, -34.687, -31.627, -28.498, -49.838,
-49.694, -48.505, -46.444, -43.760, -40.779, -37.736, -34.704, -31.642, -28.512,
 -49.864, -49.720, -48.530, -46.468, -43.783, -40.800, -37.755, -34.721,
-31.657, -28.525, -49.890, -49.746, -48.556, -46.492, -43.805, -40.820, -37.773,
-34.737, -31.672, -28.538, -49.916, -49.772, -48.581, -46.516, -43.827, -40.840,
-37.792, -34.754, -31.687, -28.552, -49.942, -49.798, -48.606, -46.540, -43.850,
-40.861, -37.810, -34.771, -31.702, -28.565, -49.968, -49.824, -48.632, -46.564,
-43.872, -40.881, -37.829, -34.787, -31.717, -28.578, -49.994, -49.850, -48.657,
-46.588, -43.895, -40.902, -37.847, -34.804, -31.732, -28.592, -50.020, -49.876,
-48.682, -46.612, -43.917, -40.922, -37.866, -34.821, -31.747, -28.605, -50.046,
-49.901, -48.707, -46.637, -43.939, -40.943, -37.884, -34.837, -31.762, -28.619,
 -50.072, -49.927, -48.733, -46.661, -43.962, -40.963, -37.903, -34.854,
-31.777, -28.632, -50.098, -49.953, -48.758, -46.685, -43.984, -40.984, -37.921,
-34.871, -31.792, -28.645, -50.124, -49.979, -48.783, -46.709, -44.006, -41.004,
-37.940, -34.888, -31.807, -28.659, -50.150, -50.005, -48.808, -46.733, -44.029,
-41.024, -37.958, -34.904, -31.822, -28.672, -50.177, -50.031, -48.834, -46.757,
-44.051, -41.045, -37.977, -34.921, -31.837, -28.685, -50.203, -50.057, -48.859,
-46.781, -44.074, -41.065, -37.995, -34.938, -31.852, -28.699, -50.229, -50.083,
-48.884, -46.805, -44.096, -41.086, -38.014, -34.954, -31.867, -28.712, -50.255,
-50.109, -48.909, -46.829, -44.118, -41.106, -38.032, -34.971, -31.882, -28.725,
 -50.281, -50.135, -48.935, -46.853, -44.141, -41.127, -38.051, -34.988,
-31.897, -28.739, -50.307, -50.161, -48.960, -46.877, -44.163, -41.147, -38.069,
-35.004, -31.912, -28.752, -50.333, -50.187, -48.985, -46.901, -44.185, -41.167,
-38.088, -35.021, -31.927, -28.765, -50.359, -50.213, -49.011, -46.925, -44.208,
-41.188, -38.106, -35.038, -31.942, -28.779, -50.385, -50.239, -49.036, -46.949,
-44.230, -41.208, -38.125, -35.054, -31.957, -28.792, -50.411, -50.265, -49.061,
-46.973, -44.252, -41.229, -38.143, -35.071, -31.972, -28.805, -50.437, -50.291,
-49.086, -46.997, -44.275, -41.249, -38.162, -35.088, -31.987, -28.819, -50.463,
-50.317, -49.112, -47.021, -44.297, -41.270, -38.180, -35.104, -32.002, -28.832,
 -50.489, -50.343, -49.137, -47.045, -44.320, -41.290, -38.199, -35.121,
-32.017, -28.845, -50.515, -50.369, -49.162, -47.070, -44.342, -41.311, -38.217,
-35.138, -32.032, -28.859, -50.541, -50.395, -49.187, -47.094, -44.364, -41.331,
-38.236, -35.154, -32.047, -28.872, -50.567, -50.421, -49.213, -47.118, -44.387,
-41.351, -38.254, -35.171, -32.062, -28.885, -50.593, -50.447, -49.238, -47.142,
-44.409, -41.372, -38.273, -35.188, -32.077, -28.899, -50.620, -50.473, -49.263,
-47.166, -44.431, -41.392, -38.291, -35.204, -32.092, -28.912, -50.646, -50.498,
-49.288, -47.190, -44.454, -41.413, -38.310, -35.221, -32.107, -28.925, -50.672,
-50.524, -49.314, -47.214, -44.476, -41.433, -38.328, -35.238, -32.122, -28.939,
 -50.698, -50.550, -49.339, -47.238, -44.498, -41.454, -38.347, -35.254,
-32.137, -28.952, -50.724, -50.576, -49.364, -47.262, -44.521, -41.474, -38.365,
-35.271, -32.152, -28.965, -50.750, -50.602, -49.389, -47.286, -44.543, -41.494,
-38.384, -35.288, -32.167, -28.979, -50.776, -50.628, -49.415, -47.310, -44.566,
-41.515, -38.402, -35.304, -32.182, -28.992, -50.802, -50.654, -49.440, -47.334,
-44.588, -41.535, -38.420, -35.321, -32.197, -29.005, -50.828, -50.680, -49.465,
-47.358, -44.610, -41.556, -38.439, -35.338, -32.212, -29.019, -50.854, -50.706,
-49.490, -47.382, -44.633, -41.576, -38.457, -35.354, -32.227, -29.032, -50.880,
-50.732, -49.516, -47.406, -44.655, -41.597, -38.476, -35.371, -32.242, -29.045,
 -50.906, -50.758, -49.541, -47.430, -44.677, -41.617, -38.494, -35.388,
-32.257, -29.059, -50.932, -50.784, -49.566, -47.454, -44.700, -41.637, -38.513,
-35.404, -32.272, -29.072, -50.958, -50.810, -49.591, -47.478, -44.722, -41.658,
-38.531, -35.421, -32.287, -29.085, -50.984, -50.836, -49.617, -47.502, -44.744,
-41.678, -38.550, -35.438, -32.301, -29.098, -51.010, -50.862, -49.642, -47.526,
-44.767, -41.699, -38.568, -35.454, -32.316, -29.112, -51.036, -50.888, -49.667,
-47.550, -44.789, -41.719, -38.587, -35.471, -32.331, -29.125, -51.062, -50.914,
-49.692, -47.574, -44.811, -41.739, -38.605, -35.488, -32.346, -29.138, -51.089,
-50.940, -49.718, -47.599, -44.834, -41.760, -38.624, -35.504, -32.361, -29.152,
 -51.115, -50.966, -49.743, -47.623, -44.856, -41.780, -38.642, -35.521,
-32.376, -29.165, -51.141, -50.991, -49.768, -47.647, -44.878, -41.801, -38.661,
-35.537, -32.391, -29.178, -51.167, -51.017, -49.793, -47.671, -44.901, -41.821,
-38.679, -35.554, -32.406, -29.192, -51.193, -51.043, -49.819, -47.695, -44.923,
-41.842, -38.698, -35.571, -32.421, -29.205, -51.219, -51.069, -49.844, -47.719,
-44.946, -41.862, -38.716, -35.587, -32.436, -29.218, -51.245, -51.095, -49.869,
-47.743, -44.968, -41.882, -38.735, -35.604, -32.451, -29.232, -51.271, -51.121,
-49.894, -47.767, -44.990, -41.903, -38.753, -35.621, -32.466, -29.245, -51.297,
-51.147, -49.920, -47.791, -45.013, -41.923, -38.772, -35.637, -32.481, -29.258,
 -51.323, -51.173, -49.945, -47.815, -45.035, -41.944, -38.790, -35.654,
-32.496, -29.271, -51.349, -51.199, -49.970, -47.839, -45.057, -41.964, -38.809,
-35.671, -32.511, -29.285, -51.375, -51.225, -49.995, -47.863, -45.080, -41.984,
-38.827, -35.687, -32.526, -29.298, -51.401, -51.251, -50.021, -47.887, -45.102,
-42.005, -38.846, -35.704, -32.541, -29.311, -51.427, -51.277, -50.046, -47.911,
-45.124, -42.025, -38.864, -35.721, -32.556, -29.325, -51.453, -51.303, -50.071,
-47.935, -45.147, -42.046, -38.883, -35.737, -32.571, -29.338, -51.479, -51.329,
-50.096, -47.959, -45.169, -42.066, -38.901, -35.754, -32.586, -29.351, -51.505,
-51.355, -50.122, -47.983, -45.191, -42.087, -38.919, -35.771, -32.601, -29.364,
 -51.531, -51.381, -50.147, -48.007, -45.214, -42.107, -38.938, -35.787,
-32.616, -29.378, -51.557, -51.407, -50.172, -48.031, -45.236, -42.127, -38.956,
-35.804, -32.631, -29.391, -51.583, -51.432, -50.197, -48.055, -45.258, -42.148,
-38.975, -35.820, -32.646, -29.404, -51.609, -51.458, -50.223, -48.079, -45.281,
-42.168, -38.993, -35.837, -32.660, -29.418, -51.636, -51.484, -50.248, -48.103,
-45.303, -42.189, -39.012, -35.854, -32.675, -29.431, -51.662, -51.510, -50.273,
-48.127, -45.325, -42.209, -39.030, -35.870, -32.690, -29.444, -51.688, -51.536,
-50.298, -48.151, -45.348, -42.229, -39.049, -35.887, -32.705, -29.457, -51.714,
-51.562, -50.324, -48.175, -45.370, -42.250, -39.067, -35.904, -32.720, -29.471,
 -51.740, -51.588, -50.349, -48.199, -45.392, -42.270, -39.086, -35.920,
-32.735, -29.484, -51.766, -51.614, -50.374, -48.223, -45.415, -42.291, -39.104,
-35.937, -32.750, -29.497, -51.792, -51.640, -50.399, -48.247, -45.437, -42.311,
-39.123, -35.954, -32.765, -29.511, -51.818, -51.666, -50.424, -48.271, -45.459,
-42.331, -39.141, -35.970, -32.780, -29.524, -51.844, -51.692, -50.450, -48.296,
-45.482, -42.352, -39.160, -35.987, -32.795, -29.537, -51.870, -51.718, -50.475,
-48.320, -45.504, -42.372, -39.178, -36.004, -32.810, -29.550, -51.896, -51.744,
-50.500, -48.344, -45.526, -42.393, -39.197, -36.020, -32.825, -29.564, -51.922,
-51.770, -50.525, -48.368, -45.549, -42.413, -39.215, -36.037, -32.840, -29.577,
 -51.948, -51.796, -50.551, -48.392, -45.571, -42.434, -39.233, -36.053,
-32.855, -29.590, -51.974, -51.821, -50.576, -48.416, -45.593, -42.454, -39.252,
-36.070, -32.870, -29.603, -52.000, -51.847, -50.601, -48.440, -45.616, -42.474,
-39.270, -36.087, -32.885, -29.617, -52.026, -51.873, -50.626, -48.464, -45.638,
-42.495, -39.289, -36.103, -32.900, -29.630, -52.052, -51.899, -50.652, -48.488,
-45.660, -42.515, -39.307, -36.120, -32.914, -29.643, -52.078, -51.925, -50.677,
-48.512, -45.683, -42.536, -39.326, -36.137, -32.929, -29.657, -52.104, -51.951,
-50.702, -48.536, -45.705, -42.556, -39.344, -36.153, -32.944, -29.670, -52.130,
-51.977, -50.727, -48.560, -45.727, -42.576, -39.363, -36.170, -32.959, -29.683,
 -52.156, -52.003, -50.753, -48.584, -45.750, -42.597, -39.381, -36.187,
-32.974, -29.696, -52.182, -52.029, -50.778, -48.608, -45.772, -42.617, -39.400,
-36.203, -32.989, -29.710, -52.208, -52.055, -50.803, -48.632, -45.794, -42.638,
-39.418, -36.220, -33.004, -29.723, -52.234, -52.081, -50.828, -48.656, -45.817,
-42.658, -39.437, -36.236, -33.019, -29.736, -52.261, -52.107, -50.853, -48.680,
-45.839, -42.678, -39.455, -36.253, -33.034, -29.749, -52.287, -52.133, -50.879,
-48.704, -45.861, -42.699, -39.473, -36.270, -33.049, -29.763, -52.313, -52.159,
-50.904, -48.728, -45.884, -42.719, -39.492, -36.286, -33.064, -29.776, -52.339,
-52.184, -50.929, -48.752, -45.906, -42.740, -39.510, -36.303, -33.079, -29.789,
 -52.365, -52.210, -50.954, -48.776, -45.928, -42.760, -39.529, -36.320,
-33.094, -29.802, -52.391, -52.236, -50.980, -48.800, -45.951, -42.780, -39.547,
-36.336, -33.109, -29.816, -52.417, -52.262, -51.005, -48.824, -45.973, -42.801,
-39.566, -36.353, -33.124, -29.829, -52.443, -52.288, -51.030, -48.848, -45.995,
-42.821, -39.584, -36.369, -33.138, -29.842, -52.469, -52.314, -51.055, -48.872,
-46.018, -42.842, -39.603, -36.386, -33.153, -29.855, -52.495, -52.340, -51.081,
-48.896, -46.040, -42.862, -39.621, -36.403, -33.168, -29.869, -52.521, -52.366,
-51.106, -48.920, -46.062, -42.882, -39.640, -36.419, -33.183, -29.882, -52.547,
-52.392, -51.131, -48.944, -46.085, -42.903, -39.658, -36.436, -33.198, -29.895,
 -52.573, -52.418, -51.156, -48.968, -46.107, -42.923, -39.676, -36.453,
-33.213, -29.908, -52.599, -52.444, -51.181, -48.992, -46.129, -42.944, -39.695,
-36.469, -33.228, -29.922, -52.625, -52.470, -51.207, -49.016, -46.152, -42.964,
-39.713, -36.486, -33.243, -29.935, -52.651, -52.496, -51.232, -49.040, -46.174,
-42.984, -39.732, -36.502, -33.258, -29.948, -52.677, -52.522, -51.257, -49.064,
-46.196, -43.005, -39.750, -36.519, -33.273, -29.961, -52.703, -52.547, -51.282,
-49.088, -46.219, -43.025, -39.769, -36.536, -33.288, -29.975, -52.729, -52.573,
-51.308, -49.112, -46.241, -43.046, -39.787, -36.552, -33.303, -29.988, -52.755,
-52.599, -51.333, -49.136, -46.263, -43.066, -39.806, -36.569, -33.317, -30.001,
 -52.781, -52.625, -51.358, -49.160, -46.286, -43.086, -39.824, -36.585,
-33.332, -30.014, -52.807, -52.651, -51.383, -49.184, -46.308, -43.107, -39.843,
-36.602, -33.347, -30.027, -52.833, -52.677, -51.408, -49.208, -46.330, -43.127,
-39.861, -36.619, -33.362, -30.041, -52.859, -52.703, -51.434, -49.232, -46.353,
-43.147, -39.879, -36.635, -33.377, -30.054, -52.885, -52.729, -51.459, -49.256,
-46.375, -43.168, -39.898, -36.652, -33.392, -30.067, -52.911, -52.755, -51.484,
-49.280, -46.397, -43.188, -39.916, -36.669, -33.407, -30.080, -52.937, -52.781,
-51.509, -49.304, -46.420, -43.209, -39.935, -36.685, -33.422, -30.094, -52.963,
-52.807, -51.534, -49.328, -46.442, -43.229, -39.953, -36.702, -33.437, -30.107,
 -52.989, -52.833, -51.560, -49.352, -46.464, -43.249, -39.972, -36.718,
-33.452, -30.120, -53.015, -52.858, -51.585, -49.376, -46.487, -43.270, -39.990,
-36.735, -33.467, -30.133, -53.042, -52.884, -51.610, -49.400, -46.509, -43.290,
-40.009, -36.752, -33.481, -30.147, -53.068, -52.910, -51.635, -49.424, -46.531,
-43.311, -40.027, -36.768, -33.496, -30.160, -53.094, -52.936, -51.661, -49.448,
-46.554, -43.331, -40.045, -36.785, -33.511, -30.173, -53.120, -52.962, -51.686,
-49.472, -46.576, -43.351, -40.064, -36.801, -33.526, -30.186, -53.146, -52.988,
-51.711, -49.496, -46.598, -43.372, -40.082, -36.818, -33.541, -30.199, -53.172,
-53.014, -51.736, -49.520, -46.621, -43.392, -40.101, -36.835, -33.556, -30.213,
 -53.198, -53.040, -51.761, -49.544, -46.643, -43.413, -40.119, -36.851,
-33.571, -30.226, -53.224, -53.066, -51.787, -49.568, -46.665, -43.433, -40.138,
-36.868, -33.586, -30.239, -53.250, -53.092, -51.812, -49.592, -46.688, -43.453,
-40.156, -36.884, -33.601, -30.252, -53.276, -53.118, -51.837, -49.616, -46.710,
-43.474, -40.175, -36.901, -33.616, -30.266, -53.302, -53.144, -51.862, -49.640,
-46.732, -43.494, -40.193, -36.918, -33.630, -30.279, -53.328, -53.169, -51.888,
-49.664, -46.754, -43.514, -40.211, -36.934, -33.645, -30.292, -53.354, -53.195,
-51.913, -49.688, -46.777, -43.535, -40.230, -36.951, -33.660, -30.305, -53.380,
-53.221, -51.938, -49.712, -46.799, -43.555, -40.248, -36.968, -33.675, -30.318,
 -53.406, -53.247, -51.963, -49.736, -46.821, -43.576, -40.267, -36.984,
-33.690, -30.332, -53.432, -53.273, -51.988, -49.760, -46.844, -43.596, -40.285,
-37.001, -33.705, -30.345, -53.458, -53.299, -52.014, -49.784, -46.866, -43.616,
-40.304, -37.017, -33.720, -30.358, -53.484, -53.325, -52.039, -49.808, -46.888,
-43.637, -40.322, -37.034, -33.735, -30.371, -53.510, -53.351, -52.064, -49.832,
-46.911, -43.657, -40.341, -37.051, -33.750, -30.384, -53.536, -53.377, -52.089,
-49.856, -46.933, -43.677, -40.359, -37.067, -33.764, -30.398, -53.562, -53.403,
-52.114, -49.880, -46.955, -43.698, -40.377, -37.084, -33.779, -30.411, -53.588,
-53.429, -52.140, -49.904, -46.978, -43.718, -40.396, -37.100, -33.794, -30.424,
 -53.614, -53.454, -52.165, -49.928, -47.000, -43.739, -40.414, -37.117,
-33.809, -30.437, -53.640, -53.480, -52.190, -49.952, -47.022, -43.759, -40.433,
-37.134, -33.824, -30.450, -53.666, -53.506, -52.215, -49.976, -47.045, -43.779,
-40.451, -37.150, -33.839, -30.464, -53.692, -53.532, -52.240, -50.000, -47.067,
-43.800, -40.470, -37.167, -33.854, -30.477, -53.718, -53.558, -52.266, -50.024,
-47.089, -43.820, -40.488, -37.183, -33.869, -30.490, -53.744, -53.584, -52.291,
-50.048, -47.111, -43.841, -40.506, -37.200, -33.884, -30.503, -53.770, -53.610,
-52.316, -50.072, -47.134, -43.861, -40.525, -37.217, -33.898, -30.516, -53.796,
-53.636, -52.341, -50.096, -47.156, -43.881, -40.543, -37.233, -33.913, -30.530,
 -53.822, -53.662, -52.366, -50.120, -47.178, -43.902, -40.562, -37.250,
-33.928, -30.543, -53.848, -53.688, -52.392, -50.144, -47.201, -43.922, -40.580,
-37.266, -33.943, -30.556, -53.874, -53.714, -52.417, -50.168, -47.223, -43.942,
-40.599, -37.283, -33.958, -30.569, -53.900, -53.739, -52.442, -50.192, -47.245,
-43.963, -40.617, -37.299, -33.973, -30.582, -53.926, -53.765, -52.467, -50.216,
-47.268, -43.983, -40.635, -37.316, -33.988, -30.596, -53.952, -53.791, -52.492,
-50.240, -47.290, -44.004, -40.654, -37.333, -34.003, -30.609, -53.978, -53.817,
-52.518, -50.264, -47.312, -44.024, -40.672, -37.349, -34.018, -30.622, -54.004,
-53.843, -52.543, -50.288, -47.335, -44.044, -40.691, -37.366, -34.032, -30.635,
 -54.030, -53.869, -52.568, -50.312, -47.357, -44.065, -40.709, -37.382,
-34.047, -30.648, -54.056, -53.895, -52.593, -50.336, -47.379, -44.085, -40.728,
-37.399, -34.062, -30.661, -54.082, -53.921, -52.618, -50.360, -47.401, -44.105,
-40.746, -37.416, -34.077, -30.675, -54.108, -53.947, -52.644, -50.384, -47.424,
-44.126, -40.764, -37.432, -34.092, -30.688, -54.134, -53.973, -52.669, -50.408,
-47.446, -44.146, -40.783, -37.449, -34.107, -30.701, -54.160, -53.999, -52.694,
-50.432, -47.468, -44.166, -40.801, -37.465, -34.122, -30.714, -54.186, -54.024,
-52.719, -50.456, -47.491, -44.187, -40.820, -37.482, -34.137, -30.727, -54.213,
-54.050, -52.744, -50.480, -47.513, -44.207, -40.838, -37.499, -34.151, -30.741,
 -54.239, -54.076, -52.770, -50.504, -47.535, -44.228, -40.857, -37.515,
-34.166, -30.754, -54.265, -54.102, -52.795, -50.528, -47.558, -44.248, -40.875,
-37.532, -34.181, -30.767, -54.291, -54.128, -52.820, -50.552, -47.580, -44.268,
-40.893, -37.548, -34.196, -30.780, -54.317, -54.154, -52.845, -50.576, -47.602,
-44.289, -40.912, -37.565, -34.211, -30.793, -54.343, -54.180, -52.870, -50.600,
-47.625, -44.309, -40.930, -37.582, -34.226, -30.806, -54.369, -54.206, -52.896,
-50.624, -47.647, -44.329, -40.949, -37.598, -34.241, -30.820, -54.395, -54.232,
-52.921, -50.648, -47.669, -44.350, -40.967, -37.615, -34.255, -30.833, -54.421,
-54.258, -52.946, -50.672, -47.691, -44.370, -40.986, -37.631, -34.270, -30.846,
 -54.447, -54.283, -52.971, -50.696, -47.714, -44.391, -41.004, -37.648,
-34.285, -30.859, -54.473, -54.309, -52.996, -50.720, -47.736, -44.411, -41.022,
-37.664, -34.300, -30.872, -54.499, -54.335, -53.022, -50.744, -47.758, -44.431,
-41.041, -37.681, -34.315, -30.885, -54.525, -54.361, -53.047, -50.768, -47.781,
-44.452, -41.059, -37.698, -34.330, -30.899, -54.551, -54.387, -53.072, -50.792,
-47.803, -44.472, -41.078, -37.714, -34.345, -30.912, -54.577, -54.413, -53.097,
-50.816, -47.825, -44.492, -41.096, -37.731, -34.359, -30.925, -54.603, -54.439,
-53.122, -50.840, -47.848, -44.513, -41.115, -37.747, -34.374, -30.938, -54.629,
-54.465, -53.148, -50.864, -47.870, -44.533, -41.133, -37.764, -34.389, -30.951,
 -54.655, -54.491, -53.173, -50.888, -47.892, -44.553, -41.151, -37.780,
-34.404, -30.964, -54.681, -54.517, -53.198, -50.912, -47.914, -44.574, -41.170,
-37.797, -34.419, -30.978, -54.707, -54.542, -53.223, -50.936, -47.937, -44.594,
-41.188, -37.814, -34.434, -30.991, -54.733, -54.568, -53.248, -50.960, -47.959,
-44.615, -41.207, -37.830, -34.449, -31.004, -54.759, -54.594, -53.274, -50.984,
-47.981, -44.635, -41.225, -37.847, -34.463, -31.017, -54.785, -54.620, -53.299,
-51.008, -48.004, -44.655, -41.243, -37.863, -34.478, -31.030, -54.811, -54.646,
-53.324, -51.032, -48.026, -44.676, -41.262, -37.880, -34.493, -31.043, -54.837,
-54.672, -53.349, -51.056, -48.048, -44.696, -41.280, -37.897, -34.508, -31.057,
 -54.863, -54.698, -53.374, -51.080, -48.070, -44.716, -41.299, -37.913,
-34.523, -31.070, -54.889, -54.724, -53.400, -51.104, -48.093, -44.737, -41.317,
-37.930, -34.538, -31.083, -54.915, -54.750, -53.425, -51.128, -48.115, -44.757,
-41.336, -37.946, -34.553, -31.096, -54.941, -54.775, -53.450, -51.152, -48.137,
-44.777, -41.354, -37.963, -34.567, -31.109, -54.967, -54.801, -53.475, -51.176,
-48.160, -44.798, -41.372, -37.979, -34.582, -31.122, -54.993, -54.827, -53.500,
-51.200, -48.182, -44.818, -41.391, -37.996, -34.597, -31.135, -55.019, -54.853,
-53.525, -51.224, -48.204, -44.839, -41.409, -38.013, -34.612, -31.149, -55.045,
-54.879, -53.551, -51.248, -48.227, -44.859, -41.428, -38.029, -34.627, -31.162,
 -55.071, -54.905, -53.576, -51.272, -48.249, -44.879, -41.446, -38.046,
-34.642, -31.175, -55.097, -54.931, -53.601, -51.296, -48.271, -44.900, -41.464,
-38.062, -34.657, -31.188, -55.123, -54.957, -53.626, -51.320, -48.293, -44.920,
-41.483, -38.079, -34.671, -31.201, -55.149, -54.983, -53.651, -51.344, -48.316,
-44.940, -41.501, -38.095, -34.686, -31.214, -55.175, -55.009, -53.677, -51.368,
-48.338, -44.961, -41.520, -38.112, -34.701, -31.227, -55.201, -55.034, -53.702,
-51.392, -48.360, -44.981, -41.538, -38.129, -34.716, -31.241, -55.227, -55.060,
-53.727, -51.416, -48.383, -45.001, -41.557, -38.145, -34.731, -31.254, -55.253,
-55.086, -53.752, -51.440, -48.405, -45.022, -41.575, -38.162, -34.746, -31.267,
 -55.279, -55.112, -53.777, -51.464, -48.427, -45.042, -41.593, -38.178,
-34.760, -31.280, -55.305, -55.138, -53.802, -51.488, -48.449, -45.062, -41.612,
-38.195, -34.775, -31.293, -55.331, -55.164, -53.828, -51.512, -48.472, -45.083,
-41.630, -38.211, -34.790, -31.306, -55.357, -55.190, -53.853, -51.536, -48.494,
-45.103, -41.649, -38.228, -34.805, -31.319, -55.383, -55.216, -53.878, -51.560,
-48.516, -45.123, -41.667, -38.245, -34.820, -31.332, -55.409, -55.242, -53.903,
-51.583, -48.539, -45.144, -41.685, -38.261, -34.835, -31.346, -55.435, -55.267,
-53.928, -51.607, -48.561, -45.164, -41.704, -38.278, -34.849, -31.359, -55.461,
-55.293, -53.954, -51.631, -48.583, -45.185, -41.722, -38.294, -34.864, -31.372,
 -55.487, -55.319, -53.979, -51.655, -48.605, -45.205, -41.741, -38.311,
-34.879, -31.385, -55.513, -55.345, -54.004, -51.679, -48.628, -45.225, -41.759,
-38.327, -34.894, -31.398, -55.539, -55.371, -54.029, -51.703, -48.650, -45.246,
-41.777, -38.344, -34.909, -31.411, -55.565, -55.397, -54.054, -51.727, -48.672,
-45.266, -41.796, -38.360, -34.924, -31.424, -55.591, -55.423, -54.079, -51.751,
-48.695, -45.286, -41.814, -38.377, -34.939, -31.437, -55.617, -55.449, -54.105,
-51.775, -48.717, -45.307, -41.833, -38.394, -34.953, -31.451, -55.643, -55.475,
-54.130, -51.799, -48.739, -45.327, -41.851, -38.410, -34.968, -31.464, -55.669,
-55.500, -54.155, -51.823, -48.761, -45.347, -41.869, -38.427, -34.983, -31.477,
 -55.695, -55.526, -54.180, -51.847, -48.784, -45.368, -41.888, -38.443,
-34.998, -31.490, -55.721, -55.552, -54.205, -51.871, -48.806, -45.388, -41.906,
-38.460, -35.013, -31.503, -55.747, -55.578, -54.231, -51.895, -48.828, -45.408,
-41.925, -38.476, -35.027, -31.516, -55.773, -55.604, -54.256, -51.919, -48.851,
-45.429, -41.943, -38.493, -35.042, -31.529, -55.799, -55.630, -54.281, -51.943,
-48.873, -45.449, -41.961, -38.509, -35.057, -31.542, -55.825, -55.656, -54.306,
-51.967, -48.895, -45.469, -41.980, -38.526, -35.072, -31.556, -55.851, -55.682,
-54.331, -51.991, -48.917, -45.490, -41.998, -38.543, -35.087, -31.569, -55.877,
-55.708, -54.356, -52.015, -48.940, -45.510, -42.017, -38.559, -35.102, -31.582,
 -55.903, -55.733, -54.382, -52.039, -48.962, -45.530, -42.035, -38.576,
-35.116, -31.595, -55.929, -55.759, -54.407, -52.063, -48.984, -45.551, -42.053,
-38.592, -35.131, -31.608, -55.955, -55.785, -54.432, -52.087, -49.007, -45.571,
-42.072, -38.609, -35.146, -31.621, -55.981, -55.811, -54.457, -52.111, -49.029,
-45.591, -42.090, -38.625, -35.161, -31.634, -56.007, -55.837, -54.482, -52.135,
-49.051, -45.612, -42.109, -38.642, -35.176, -31.647, -56.033, -55.863, -54.507,
-52.159, -49.073, -45.632, -42.127, -38.658, -35.191, -31.660, -56.059, -55.889,
-54.533, -52.183, -49.096, -45.653, -42.145, -38.675, -35.205, -31.674, -56.085,
-55.915, -54.558, -52.207, -49.118, -45.673, -42.164, -38.692, -35.220, -31.687,
 -56.111, -55.940, -54.583, -52.231, -49.140, -45.693, -42.182, -38.708,
-35.235, -31.700, -56.137, -55.966, -54.608, -52.255, -49.163, -45.714, -42.201,
-38.725, -35.250, -31.713, -56.163, -55.992, -54.633, -52.279, -49.185, -45.734,
-42.219, -38.741, -35.265, -31.726, -56.189, -56.018, -54.658, -52.303, -49.207,
-45.754, -42.237, -38.758, -35.279, -31.739, -56.215, -56.044, -54.684, -52.326,
-49.229, -45.775, -42.256, -38.774, -35.294, -31.752, -56.241, -56.070, -54.709,
-52.350, -49.252, -45.795, -42.274, -38.791, -35.309, -31.765, -56.267, -56.096,
-54.734, -52.374, -49.274, -45.815, -42.293, -38.807, -35.324, -31.778, -56.293,
-56.122, -54.759, -52.398, -49.296, -45.836, -42.311, -38.824, -35.339, -31.791,
 -56.319, -56.147, -54.784, -52.422, -49.318, -45.856, -42.329, -38.841,
-35.354, -31.805, -56.345, -56.173, -54.809, -52.446, -49.341, -45.876, -42.348,
-38.857, -35.368, -31.818, -56.371, -56.199, -54.835, -52.470, -49.363, -45.897,
-42.366, -38.874, -35.383, -31.831, -56.397, -56.225, -54.860, -52.494, -49.385,
-45.917, -42.385, -38.890, -35.398, -31.844, -56.423, -56.251, -54.885, -52.518,
-49.408, -45.937, -42.403, -38.907, -35.413, -31.857, -56.449, -56.277, -54.910,
-52.542, -49.430, -45.958, -42.421, -38.923, -35.428, -31.870, -56.475, -56.303,
-54.935, -52.566, -49.452, -45.978, -42.440, -38.940, -35.442, -31.883, -56.501,
-56.329, -54.961, -52.590, -49.474, -45.998, -42.458, -38.956, -35.457, -31.896,
 -56.527, -56.355, -54.986, -52.614, -49.497, -46.019, -42.477, -38.973,
-35.472, -31.909, -56.553, -56.380, -55.011, -52.638, -49.519, -46.039, -42.495,
-38.989, -35.487, -31.922, -56.579, -56.406, -55.036, -52.662, -49.541, -46.059,
-42.513, -39.006, -35.502, -31.935, -56.605, -56.432, -55.061, -52.686, -49.563,
-46.080, -42.532, -39.023, -35.516, -31.949, -56.631, -56.458, -55.086, -52.710,
-49.586, -46.100, -42.550, -39.039, -35.531, -31.962, -56.657, -56.484, -55.111,
-52.734, -49.608, -46.120, -42.569, -39.056, -35.546, -31.975, -56.683, -56.510,
-55.137, -52.758, -49.630, -46.141, -42.587, -39.072, -35.561, -31.988, -56.709,
-56.536, -55.162, -52.782, -49.653, -46.161, -42.605, -39.089, -35.576, -32.001,
 -56.735, -56.562, -55.187, -52.806, -49.675, -46.181, -42.624, -39.105,
-35.590, -32.014, -56.761, -56.587, -55.212, -52.830, -49.697, -46.202, -42.642,
-39.122, -35.605, -32.027, -56.787, -56.613, -55.237, -52.854, -49.719, -46.222,
-42.660, -39.138, -35.620, -32.040, -56.813, -56.639, -55.262, -52.878, -49.742,
-46.242, -42.679, -39.155, -35.635, -32.053, -56.839, -56.665, -55.288, -52.901,
-49.764, -46.263, -42.697, -39.171, -35.650, -32.066, -56.865, -56.691, -55.313,
-52.925, -49.786, -46.283, -42.716, -39.188, -35.664, -32.079, -56.891, -56.717,
-55.338, -52.949, -49.808, -46.303, -42.734, -39.204, -35.679, -32.092, -56.917,
-56.743, -55.363, -52.973, -49.831, -46.324, -42.752, -39.221, -35.694, -32.105,
 -56.943, -56.768, -55.388, -52.997, -49.853, -46.344, -42.771, -39.238,
-35.709, -32.119, -56.969, -56.794, -55.413, -53.021, -49.875, -46.364, -42.789,
-39.254, -35.724, -32.132, -56.995, -56.820, -55.439, -53.045, -49.898, -46.385,
-42.808, -39.271, -35.738, -32.145, -57.021, -56.846, -55.464, -53.069, -49.920,
-46.405, -42.826, -39.287, -35.753, -32.158, -57.047, -56.872, -55.489, -53.093,
-49.942, -46.425, -42.844, -39.304, -35.768, -32.171, -57.073, -56.898, -55.514,
-53.117, -49.964, -46.446, -42.863, -39.320, -35.783, -32.184, -57.099, -56.924,
-55.539, -53.141, -49.987, -46.466, -42.881, -39.337, -35.798, -32.197, -57.125,
-56.950, -55.564, -53.165, -50.009, -46.486, -42.899, -39.353, -35.812, -32.210,
 -57.151, -56.975, -55.590, -53.189, -50.031, -46.507, -42.918, -39.370,
-35.827, -32.223, -57.177, -57.001, -55.615, -53.213, -50.053, -46.527, -42.936,
-39.386, -35.842, -32.236, -57.203, -57.027, -55.640, -53.237, -50.076, -46.547,
-42.955, -39.403, -35.857, -32.249, -57.229, -57.053, -55.665, -53.261, -50.098,
-46.568, -42.973, -39.419, -35.872, -32.262, -57.255, -57.079, -55.690, -53.285,
-50.120, -46.588, -42.991, -39.436, -35.886, -32.275, -57.281, -57.105, -55.715,
-53.309, -50.142, -46.608, -43.010, -39.453, -35.901, -32.288, -57.307, -57.131,
-55.740, -53.333, -50.165, -46.629, -43.028, -39.469, -35.916, -32.301, -57.333,
-57.157, -55.766, -53.357, -50.187, -46.649, -43.047, -39.486, -35.931, -32.315,
 -57.359, -57.182, -55.791, -53.380, -50.209, -46.669, -43.065, -39.502,
-35.946, -32.328, -57.385, -57.208, -55.816, -53.404, -50.231, -46.690, -43.083,
-39.519, -35.960, -32.341, -57.411, -57.234, -55.841, -53.428, -50.254, -46.710,
-43.102, -39.535, -35.975, -32.354, -57.436, -57.260, -55.866, -53.452, -50.276,
-46.730, -43.120, -39.552, -35.990, -32.367, -57.462, -57.286, -55.891, -53.476,
-50.298, -46.751, -43.138, -39.568, -36.005, -32.380, -57.488, -57.312, -55.917,
-53.500, -50.321, -46.771, -43.157, -39.585, -36.019, -32.393, -57.514, -57.338,
-55.942, -53.524, -50.343, -46.791, -43.175, -39.601, -36.034, -32.406, -57.540,
-57.363, -55.967, -53.548, -50.365, -46.812, -43.194, -39.618, -36.049, -32.419,
 -57.566, -57.389, -55.992, -53.572, -50.387, -46.832, -43.212, -39.634,
-36.064, -32.432, -57.592, -57.415, -56.017, -53.596, -50.410, -46.852, -43.230,
-39.651, -36.079, -32.445, -57.618, -57.441, -56.042, -53.620, -50.432, -46.873,
-43.249, -39.667, -36.093, -32.458, -57.644, -57.467, -56.067, -53.644, -50.454,
-46.893, -43.267, -39.684, -36.108, -32.471, -57.670, -57.493, -56.093, -53.668,
-50.476, -46.913, -43.285, -39.700, -36.123, -32.484, -57.696, -57.519, -56.118,
-53.692, -50.499, -46.933, -43.304, -39.717, -36.138, -32.497, -57.722, -57.545,
-56.143, -53.716, -50.521, -46.954, -43.322, -39.733, -36.153, -32.510, -57.748,
-57.570, -56.168, -53.740, -50.543, -46.974, -43.341, -39.750, -36.167, -32.523,
 -57.774, -57.596, -56.193, -53.764, -50.565, -46.994, -43.359, -39.767,
-36.182, -32.536, -57.800, -57.622, -56.218, -53.788, -50.588, -47.015, -43.377,
-39.783, -36.197, -32.549, -57.826, -57.648, -56.244, -53.811, -50.610, -47.035,
-43.396, -39.800, -36.212, -32.562, -57.852, -57.674, -56.269, -53.835, -50.632,
-47.055, -43.414, -39.816, -36.226, -32.575, -57.878, -57.700, -56.294, -53.859,
-50.654, -47.076, -43.432, -39.833, -36.241, -32.589, -57.904, -57.726, -56.319,
-53.883, -50.677, -47.096, -43.451, -39.849, -36.256, -32.602, -57.930, -57.751,
-56.344, -53.907, -50.699, -47.116, -43.469, -39.866, -36.271, -32.615, -57.956,
-57.777, -56.369, -53.931, -50.721, -47.137, -43.488, -39.882, -36.285, -32.628,
 -57.982, -57.803, -56.394, -53.955, -50.743, -47.157, -43.506, -39.899,
-36.300, -32.641, -58.008, -57.829, -56.420, -53.979, -50.766, -47.177, -43.524,
-39.915, -36.315, -32.654, -58.034, -57.855, -56.445, -54.003, -50.788, -47.198,
-43.543, -39.932, -36.330, -32.667, -58.060, -57.881, -56.470, -54.027, -50.810,
-47.218, -43.561, -39.948, -36.345, -32.680, -58.086, -57.907, -56.495, -54.051,
-50.832, -47.238, -43.579, -39.965, -36.359, -32.693, -58.112, -57.932, -56.520,
-54.075, -50.855, -47.259, -43.598, -39.981, -36.374, -32.706, -58.138, -57.958,
-56.545, -54.099, -50.877, -47.279, -43.616, -39.998, -36.389, -32.719, -58.164,
-57.984, -56.570, -54.123, -50.899, -47.299, -43.635, -40.014, -36.404, -32.732,
 -58.190, -58.010, -56.596, -54.147, -50.921, -47.320, -43.653, -40.031,
-36.418, -32.745, -58.216, -58.036, -56.621, -54.171, -50.944, -47.340, -43.671,
-40.047, -36.433, -32.758, -58.242, -58.062, -56.646, -54.194, -50.966, -47.360,
-43.690, -40.064, -36.448, -32.771, -58.268, -58.088, -56.671, -54.218, -50.988,
-47.380, -43.708, -40.080, -36.463, -32.784, -58.294, -58.114, -56.696, -54.242,
-51.010, -47.401, -43.726, -40.097, -36.477, -32.797, -58.320, -58.139, -56.721,
-54.266, -51.033, -47.421, -43.745, -40.113, -36.492, -32.810, -58.346, -58.165,
-56.746, -54.290, -51.055, -47.441, -43.763, -40.130, -36.507, -32.823, -58.372,
-58.191, -56.772, -54.314, -51.077, -47.462, -43.782, -40.146, -36.522, -32.836,
 -58.398, -58.217, -56.797, -54.338, -51.099, -47.482, -43.800, -40.163,
-36.537, -32.849, -58.424, -58.243, -56.822, -54.362, -51.122, -47.502, -43.818,
-40.180, -36.551, -32.862, -58.450, -58.269, -56.847, -54.386, -51.144, -47.523,
-43.837, -40.196, -36.566, -32.875, -58.476, -58.295, -56.872, -54.410, -51.166,
-47.543, -43.855, -40.213, -36.581, -32.888, -58.502, -58.320, -56.897, -54.434,
-51.188, -47.563, -43.873, -40.229, -36.596, -32.901, -58.528, -58.346, -56.922,
-54.458, -51.211, -47.584, -43.892, -40.246, -36.610, -32.914, -58.554, -58.372,
-56.948, -54.482, -51.233, -47.604, -43.910, -40.262, -36.625, -32.927, -58.580,
-58.398, -56.973, -54.506, -51.255, -47.624, -43.928, -40.279, -36.640, -32.940,
 -58.606, -58.424, -56.998, -54.530, -51.277, -47.645, -43.947, -40.295,
-36.655, -32.953, -58.632, -58.450, -57.023, -54.553, -51.300, -47.665, -43.965,
-40.312, -36.669, -32.966, -58.658, -58.476, -57.048, -54.577, -51.322, -47.685,
-43.984, -40.328, -36.684, -32.979, -58.683, -58.501, -57.073, -54.601, -51.344,
-47.705, -44.002, -40.345, -36.699, -32.992, -58.709, -58.527, -57.098, -54.625,
-51.366, -47.726, -44.020, -40.361, -36.714, -33.005, -58.735, -58.553, -57.124,
-54.649, -51.389, -47.746, -44.039, -40.378, -36.728, -33.018, -58.761, -58.579,
-57.149, -54.673, -51.411, -47.766, -44.057, -40.394, -36.743, -33.031, -58.787,
-58.605, -57.174, -54.697, -51.433, -47.787, -44.075, -40.411, -36.758, -33.044,
 -58.813, -58.631, -57.199, -54.721, -51.455, -47.807, -44.094, -40.427,
-36.773, -33.057, -58.839, -58.656, -57.224, -54.745, -51.478, -47.827, -44.112,
-40.444, -36.787, -33.070, -58.865, -58.682, -57.249, -54.769, -51.500, -47.848,
-44.130, -40.460, -36.802, -33.083, -58.891, -58.708, -57.274, -54.793, -51.522,
-47.868, -44.149, -40.477, -36.817, -33.096, -58.917, -58.734, -57.300, -54.817,
-51.544, -47.888, -44.167, -40.493, -36.832, -33.109, -58.943, -58.760, -57.325,
-54.841, -51.567, -47.909, -44.186, -40.510, -36.846, -33.122, -58.969, -58.786,
-57.350, -54.865, -51.589, -47.929, -44.204, -40.526, -36.861, -33.135, -58.995,
-58.812, -57.375, -54.888, -51.611, -47.949, -44.222, -40.543, -36.876, -33.148,
 -59.021, -58.837, -57.400, -54.912, -51.633, -47.970, -44.241, -40.559,
-36.891, -33.161, -59.047, -58.863, -57.425, -54.936, -51.656, -47.990, -44.259,
-40.576, -36.905, -33.174, -59.073, -58.889, -57.450, -54.960, -51.678, -48.010,
-44.277, -40.592, -36.920, -33.187, -59.099, -58.915, -57.475, -54.984, -51.700,
-48.030, -44.296, -40.609, -36.935, -33.200, -59.125, -58.941, -57.501, -55.008,
-51.722, -48.051, -44.314, -40.625, -36.950, -33.213, -59.151, -58.967, -57.526,
-55.032, -51.745, -48.071, -44.332, -40.642, -36.964, -33.226, -59.177, -58.993,
-57.551, -55.056, -51.767, -48.091, -44.351, -40.658, -36.979, -33.239, -59.203,
-59.018, -57.576, -55.080, -51.789, -48.112, -44.369, -40.675, -36.994, -33.252,
 -59.229, -59.044, -57.601, -55.104, -51.811, -48.132, -44.387, -40.691,
-37.009, -33.265, -59.255, -59.070, -57.626, -55.128, -51.834, -48.152, -44.406,
-40.708, -37.023, -33.278, -59.281, -59.096, -57.651, -55.152, -51.856, -48.173,
-44.424, -40.724, -37.038, -33.291, -59.307, -59.122, -57.677, -55.176, -51.878,
-48.193, -44.443, -40.741, -37.053, -33.304, -59.333, -59.148, -57.702, -55.200,
-51.900, -48.213, -44.461, -40.757, -37.068, -33.317, -59.359, -59.174, -57.727,
-55.223, -51.923, -48.233, -44.479, -40.774, -37.082, -33.330, -59.385, -59.199,
-57.752, -55.247, -51.945, -48.254, -44.498, -40.790, -37.097, -33.343, -59.411,
-59.225, -57.777, -55.271, -51.967, -48.274, -44.516, -40.807, -37.112, -33.356,
 -59.437, -59.251, -57.802, -55.295, -51.989, -48.294, -44.534, -40.823,
-37.126, -33.369, -59.463, -59.277, -57.827, -55.319, -52.012, -48.315, -44.553,
-40.840, -37.141, -33.382, -59.489, -59.303, -57.852, -55.343, -52.034, -48.335,
-44.571, -40.856, -37.156, -33.395, -59.515, -59.329, -57.878, -55.367, -52.056,
-48.355, -44.589, -40.873, -37.171, -33.408, -59.541, -59.354, -57.903, -55.391,
-52.078, -48.376, -44.608, -40.889, -37.185, -33.421, -59.566, -59.380, -57.928,
-55.415, -52.100, -48.396, -44.626, -40.906, -37.200, -33.434, -59.592, -59.406,
-57.953, -55.439, -52.123, -48.416, -44.644, -40.922, -37.215, -33.447, -59.618,
-59.432, -57.978, -55.463, -52.145, -48.436, -44.663, -40.939, -37.230, -33.460,
 -59.644, -59.458, -58.003, -55.487, -52.167, -48.457, -44.681, -40.955,
-37.244, -33.473, -59.670, -59.484, -58.028, -55.510, -52.189, -48.477, -44.699,
-40.972, -37.259, -33.486, -59.696, -59.510, -58.053, -55.534, -52.212, -48.497,
-44.718, -40.988, -37.274, -33.499, -59.722, -59.535, -58.079, -55.558, -52.234,
-48.518, -44.736, -41.005, -37.289, -33.512, -59.748, -59.561, -58.104, -55.582,
-52.256, -48.538, -44.755, -41.021, -37.303, -33.525, -59.774, -59.587, -58.129,
-55.606, -52.278, -48.558, -44.773, -41.038, -37.318, -33.538, -59.800, -59.613,
-58.154, -55.630, -52.301, -48.579, -44.791, -41.054, -37.333, -33.551, -59.826,
-59.639, -58.179, -55.654, -52.323, -48.599, -44.810, -41.071, -37.347, -33.564,
 -59.852, -59.665, -58.204, -55.678, -52.345, -48.619, -44.828, -41.087,
-37.362, -33.577, -59.878, -59.690, -58.229, -55.702, -52.367, -48.639, -44.846,
-41.104, -37.377, -33.590, -59.904, -59.716, -58.255, -55.726, -52.390, -48.660,
-44.865, -41.120, -37.392, -33.603, -59.930, -59.742, -58.280, -55.750, -52.412,
-48.680, -44.883, -41.137, -37.406, -33.616, -59.956, -59.768, -58.305, -55.774,
-52.434, -48.700, -44.901, -41.153, -37.421, -33.629, -59.982, -59.794, -58.330,
-55.797, -52.456, -48.721, -44.920, -41.170, -37.436, -33.642, -60.008, -59.820,
-58.355, -55.821, -52.478, -48.741, -44.938, -41.186, -37.451, -33.654, -60.034,
-59.846, -58.380, -55.845, -52.501, -48.761, -44.956, -41.203, -37.465, -33.667,
 -60.060, -59.871, -58.405, -55.869, -52.523, -48.782, -44.975, -41.219,
-37.480, -33.680, -60.086, -59.897, -58.430, -55.893, -52.545, -48.802, -44.993,
-41.236, -37.495, -33.693, -60.112, -59.923, -58.455, -55.917, -52.567, -48.822,
-45.011, -41.252, -37.509, -33.706, -60.138, -59.949, -58.481, -55.941, -52.590,
-48.842, -45.030, -41.269, -37.524, -33.719, -60.164, -59.975, -58.506, -55.965,
-52.612, -48.863, -45.048, -41.285, -37.539, -33.732, -60.190, -60.001, -58.531,
-55.989, -52.634, -48.883, -45.066, -41.301, -37.554, -33.745, -60.216, -60.026,
-58.556, -56.013, -52.656, -48.903, -45.085, -41.318, -37.568, -33.758, -60.242,
-60.052, -58.581, -56.037, -52.679, -48.924, -45.103, -41.334, -37.583, -33.771,
 -60.268, -60.078, -58.606, -56.061, -52.701, -48.944, -45.121, -41.351,
-37.598, -33.784, -60.293, -60.104, -58.631, -56.084, -52.723, -48.964, -45.140,
-41.367, -37.612, -33.797, -60.319, -60.130, -58.656, -56.108, -52.745, -48.984,
-45.158, -41.384, -37.627, -33.810, -60.345, -60.156, -58.682, -56.132, -52.768,
-49.005, -45.176, -41.400, -37.642, -33.823, -60.371, -60.181, -58.707, -56.156,
-52.790, -49.025, -45.195, -41.417, -37.657, -33.836, -60.397, -60.207, -58.732,
-56.180, -52.812, -49.045, -45.213, -41.433, -37.671, -33.849, -60.423, -60.233,
-58.757, -56.204, -52.834, -49.066, -45.232, -41.450, -37.686, -33.862, -60.449,
-60.259, -58.782, -56.228, -52.856, -49.086, -45.250, -41.466, -37.701, -33.875,
 -60.475, -60.285, -58.807, -56.252, -52.879, -49.106, -45.268, -41.483,
-37.715, -33.888, -60.501, -60.311, -58.832, -56.276, -52.901, -49.127, -45.287,
-41.499, -37.730, -33.901, -60.527, -60.336, -58.857, -56.300, -52.923, -49.147,
-45.305, -41.516, -37.745, -33.914, -60.553, -60.362, -58.883, -56.324, -52.945,
-49.167, -45.323, -41.532, -37.760, -33.926, -60.579, -60.388, -58.908, -56.347,
-52.968, -49.187, -45.342, -41.549, -37.774, -33.939, -60.605, -60.414, -58.933,
-56.371, -52.990, -49.208, -45.360, -41.565, -37.789, -33.952, -60.631, -60.440,
-58.958, -56.395, -53.012, -49.228, -45.378, -41.582, -37.804, -33.965, -60.657,
-60.466, -58.983, -56.419, -53.034, -49.248, -45.397, -41.598, -37.818, -33.978,
 -60.683, -60.491, -59.008, -56.443, -53.056, -49.269, -45.415, -41.615,
-37.833, -33.991, -60.709, -60.517, -59.033, -56.467, -53.079, -49.289, -45.433,
-41.631, -37.848, -34.004, -60.735, -60.543, -59.058, -56.491, -53.101, -49.309,
-45.452, -41.648, -37.863, -34.017, -60.761, -60.569, -59.083, -56.515, -53.123,
-49.329, -45.470, -41.664, -37.877, -34.030, -60.787, -60.595, -59.109, -56.539,
-53.145, -49.350, -45.488, -41.681, -37.892, -34.043, -60.813, -60.621, -59.134,
-56.563, -53.168, -49.370, -45.507, -41.697, -37.907, -34.056, -60.839, -60.647,
-59.159, -56.587, -53.190, -49.390, -45.525, -41.714, -37.921, -34.069, -60.865,
-60.672, -59.184, -56.610, -53.212, -49.411, -45.543, -41.730, -37.936, -34.082,
 -60.891, -60.698, -59.209, -56.634, -53.234, -49.431, -45.562, -41.746,
-37.951, -34.095, -60.916, -60.724, -59.234, -56.658, -53.257, -49.451, -45.580,
-41.763, -37.966, -34.108, -60.942, -60.750, -59.259, -56.682, -53.279, -49.471,
-45.598, -41.779, -37.980, -34.121, -60.968, -60.776, -59.284, -56.706, -53.301,
-49.492, -45.617, -41.796, -37.995, -34.133, -60.994, -60.802, -59.310, -56.730,
-53.323, -49.512, -45.635, -41.812, -38.010, -34.146, -61.020, -60.827, -59.335,
-56.754, -53.345, -49.532, -45.653, -41.829, -38.024, -34.159, -61.046, -60.853,
-59.360, -56.778, -53.368, -49.553, -45.672, -41.845, -38.039, -34.172, -61.072,
-60.879, -59.385, -56.802, -53.390, -49.573, -45.690, -41.862, -38.054, -34.185,
 -61.098, -60.905, -59.410, -56.826, -53.412, -49.593, -45.708, -41.878,
-38.068, -34.198, -61.124, -60.931, -59.435, -56.849, -53.434, -49.613, -45.727,
-41.895, -38.083, -34.211, -61.150, -60.957, -59.460, -56.873, -53.457, -49.634,
-45.745, -41.911, -38.098, -34.224, -61.176, -60.982, -59.485, -56.897, -53.479,
-49.654, -45.763, -41.928, -38.113, -34.237, -61.202, -61.008, -59.510, -56.921,
-53.501, -49.674, -45.782, -41.944, -38.127, -34.250, -61.228, -61.034, -59.536,
-56.945, -53.523, -49.695, -45.800, -41.961, -38.142, -34.263, -61.254, -61.060,
-59.561, -56.969, -53.545, -49.715, -45.818, -41.977, -38.157, -34.276, -61.280,
-61.086, -59.586, -56.993, -53.568, -49.735, -45.837, -41.994, -38.171, -34.289,
 -61.306, -61.112, -59.611, -57.017, -53.590, -49.755, -45.855, -42.010,
-38.186, -34.301, -61.332, -61.137, -59.636, -57.041, -53.612, -49.776, -45.873,
-42.027, -38.201, -34.314, -61.358, -61.163, -59.661, -57.065, -53.634, -49.796,
-45.892, -42.043, -38.215, -34.327, -61.384, -61.189, -59.686, -57.089, -53.657,
-49.816, -45.910, -42.059, -38.230, -34.340, -61.410, -61.215, -59.711, -57.112,
-53.679, -49.837, -45.928, -42.076, -38.245, -34.353, -61.436, -61.241, -59.736,
-57.136, -53.701, -49.857, -45.947, -42.092, -38.259, -34.366, -61.462, -61.267,
-59.762, -57.160, -53.723, -49.877, -45.965, -42.109, -38.274, -34.379, -61.487,
-61.292, -59.787, -57.184, -53.745, -49.897, -45.983, -42.125, -38.289, -34.392,
 -61.513, -61.318, -59.812, -57.208, -53.768, -49.918, -46.002, -42.142,
-38.304, -34.405, -61.539, -61.344, -59.837, -57.232, -53.790, -49.938, -46.020,
-42.158, -38.318, -34.418, -61.565, -61.370, -59.862, -57.256, -53.812, -49.958,
-46.038, -42.175, -38.333, -34.431, -61.591, -61.396, -59.887, -57.280, -53.834,
-49.978, -46.057, -42.191, -38.348, -34.444, -61.617, -61.421, -59.912, -57.304,
-53.856, -49.999, -46.075, -42.208, -38.362, -34.456, -61.643, -61.447, -59.937,
-57.328, -53.879, -50.019, -46.093, -42.224, -38.377, -34.469, -61.669, -61.473,
-59.962, -57.351, -53.901, -50.039, -46.112, -42.241, -38.392, -34.482, -61.695,
-61.499, -59.987, -57.375, -53.923, -50.060, -46.130, -42.257, -38.406, -34.495,
 -61.721, -61.525, -60.013, -57.399, -53.945, -50.080, -46.148, -42.274,
-38.421, -34.508, -61.747, -61.551, -60.038, -57.423, -53.968, -50.100, -46.167,
-42.290, -38.436, -34.521, -61.773, -61.576, -60.063, -57.447, -53.990, -50.120,
-46.185, -42.306, -38.450, -34.534, -61.799, -61.602, -60.088, -57.471, -54.012,
-50.141, -46.203, -42.323, -38.465, -34.547, -61.825, -61.628, -60.113, -57.495,
-54.034, -50.161, -46.222, -42.339, -38.480, -34.560, -61.851, -61.654, -60.138,
-57.519, -54.056, -50.181, -46.240, -42.356, -38.495, -34.573, -61.877, -61.680,
-60.163, -57.543, -54.079, -50.202, -46.258, -42.372, -38.509, -34.585, -61.903,
-61.706, -60.188, -57.566, -54.101, -50.222, -46.277, -42.389, -38.524, -34.598,
 -61.929, -61.731, -60.213, -57.590, -54.123, -50.242, -46.295, -42.405,
-38.539, -34.611, -61.955, -61.757, -60.239, -57.614, -54.145, -50.262, -46.313,
-42.422, -38.553, -34.624, -61.981, -61.783, -60.264, -57.638, -54.168, -50.283,
-46.332, -42.438, -38.568, -34.637, -62.006, -61.809, -60.289, -57.662, -54.190,
-50.303, -46.350, -42.455, -38.583, -34.650, -62.032, -61.835, -60.314, -57.686,
-54.212, -50.323, -46.368, -42.471, -38.597, -34.663, -62.058, -61.861, -60.339,
-57.710, -54.234, -50.343, -46.387, -42.488, -38.612, -34.676, -62.084, -61.886,
-60.364, -57.734, -54.256, -50.364, -46.405, -42.504, -38.627, -34.689, -62.110,
-61.912, -60.389, -57.758, -54.279, -50.384, -46.423, -42.520, -38.641, -34.702,
 -62.136, -61.938, -60.414, -57.782, -54.301, -50.404, -46.442, -42.537,
-38.656, -34.714, -62.162, -61.964, -60.439, -57.805, -54.323, -50.425, -46.460,
-42.553, -38.671, -34.727, -62.188, -61.990, -60.464, -57.829, -54.345, -50.445,
-46.478, -42.570, -38.685, -34.740, -62.214, -62.016, -60.490, -57.853, -54.367,
-50.465, -46.496, -42.586, -38.700, -34.753, -62.240, -62.041, -60.515, -57.877,
-54.390, -50.485, -46.515, -42.603, -38.715, -34.766, -62.266, -62.067, -60.540,
-57.901, -54.412, -50.506, -46.533, -42.619, -38.729, -34.779, -62.292, -62.093,
-60.565, -57.925, -54.434, -50.526, -46.551, -42.636, -38.744, -34.792, -62.318,
-62.119, -60.590, -57.949, -54.456, -50.546, -46.570, -42.652, -38.759, -34.805,
 -62.344, -62.145, -60.615, -57.973, -54.478, -50.566, -46.588, -42.669,
-38.773, -34.818, -62.370, -62.170, -60.640, -57.997, -54.501, -50.587, -46.606,
-42.685, -38.788, -34.830, -62.396, -62.196, -60.665, -58.020, -54.523, -50.607,
-46.625, -42.702, -38.803, -34.843, -62.422, -62.222, -60.690, -58.044, -54.545,
-50.627, -46.643, -42.718, -38.817, -34.856, -62.448, -62.248, -60.715, -58.068,
-54.567, -50.648, -46.661, -42.734, -38.832, -34.869, -62.473, -62.274, -60.741,
-58.092, -54.590, -50.668, -46.680, -42.751, -38.847, -34.882, -62.499, -62.300,
-60.766, -58.116, -54.612, -50.688, -46.698, -42.767, -38.861, -34.895, -62.525,
-62.325, -60.791, -58.140, -54.634, -50.708, -46.716, -42.784, -38.876, -34.908,
 -62.551, -62.351, -60.816, -58.164, -54.656, -50.729, -46.735, -42.800,
-38.891, -34.921, -62.577, -62.377, -60.841, -58.188, -54.678, -50.749, -46.753,
-42.817, -38.905, -34.933, -62.603, -62.403, -60.866, -58.212, -54.701, -50.769,
-46.771, -42.833, -38.920, -34.946, -62.629, -62.429, -60.891, -58.235, -54.723,
-50.789, -46.790, -42.850, -38.935, -34.959, -62.655, -62.455, -60.916, -58.259,
-54.745, -50.810, -46.808, -42.866, -38.949, -34.972, -62.681, -62.480, -60.941,
-58.283, -54.767, -50.830, -46.826, -42.882, -38.964, -34.985, -62.707, -62.506,
-60.966, -58.307, -54.789, -50.850, -46.845, -42.899, -38.979, -34.998, -62.733,
-62.532, -60.992, -58.331, -54.812, -50.871, -46.863, -42.915, -38.993, -35.011,
 -62.759, -62.558, -61.017, -58.355, -54.834, -50.891, -46.881, -42.932,
-39.008, -35.024, -62.785, -62.584, -61.042, -58.379, -54.856, -50.911, -46.900,
-42.948, -39.023, -35.036, -62.811, -62.609, -61.067, -58.403, -54.878, -50.931,
-46.918, -42.965, -39.037, -35.049, -62.837, -62.635, -61.092, -58.427, -54.900,
-50.952, -46.936, -42.981, -39.052, -35.062, -62.863, -62.661, -61.117, -58.450,
-54.923, -50.972, -46.954, -42.998, -39.067, -35.075, -62.889, -62.687, -61.142,
-58.474, -54.945, -50.992, -46.973, -43.014, -39.081, -35.088, -62.915, -62.713,
-61.167, -58.498, -54.967, -51.012, -46.991, -43.031, -39.096, -35.101, -62.940,
-62.739, -61.192, -58.522, -54.989, -51.033, -47.009, -43.047, -39.111, -35.114,
 -62.966, -62.764, -61.217, -58.546, -55.011, -51.053, -47.028, -43.063,
-39.125, -35.127, -62.992, -62.790, -61.242, -58.570, -55.034, -51.073, -47.046,
-43.080, -39.140, -35.139, -63.018, -62.816, -61.268, -58.594, -55.056, -51.093,
-47.064, -43.096, -39.155, -35.152, -63.044, -62.842, -61.293, -58.618, -55.078,
-51.114, -47.083, -43.113, -39.169, -35.165, -63.070, -62.868, -61.318, -58.641,
-55.100, -51.134, -47.101, -43.129, -39.184, -35.178, -63.096, -62.893, -61.343,
-58.665, -55.122, -51.154, -47.119, -43.146, -39.199, -35.191, -63.122, -62.919,
-61.368, -58.689, -55.145, -51.174, -47.138, -43.162, -39.213, -35.204, -63.148,
-62.945, -61.393, -58.713, -55.167, -51.195, -47.156, -43.179, -39.228, -35.217,
 -63.174, -62.971, -61.418, -58.737, -55.189, -51.215, -47.174, -43.195,
-39.243, -35.229, -63.200, -62.997, -61.443, -58.761, -55.211, -51.235, -47.193,
-43.211, -39.257, -35.242, -63.226, -63.023, -61.468, -58.785, -55.234, -51.256,
-47.211, -43.228, -39.272, -35.255, -63.252, -63.048, -61.493, -58.809, -55.256,
-51.276, -47.229, -43.244, -39.287, -35.268, -63.278, -63.074, -61.518, -58.833,
-55.278, -51.296, -47.248, -43.261, -39.301, -35.281, -63.304, -63.100, -61.544,
-58.856, -55.300, -51.316, -47.266, -43.277, -39.316, -35.294, -63.330, -63.126,
-61.569, -58.880, -55.322, -51.337, -47.284, -43.294, -39.331, -35.307, -63.356,
-63.152, -61.594, -58.904, -55.345, -51.357, -47.302, -43.310, -39.345, -35.319,
 -63.381, -63.177, -61.619, -58.928, -55.367, -51.377, -47.321, -43.327,
-39.360, -35.332, -63.407, -63.203, -61.644, -58.952, -55.389, -51.397, -47.339,
-43.343, -39.375, -35.345, -63.433, -63.229, -61.669, -58.976, -55.411, -51.418,
-47.357, -43.359, -39.389, -35.358, -63.459, -63.255, -61.694, -59.000, -55.433,
-51.438, -47.376, -43.376, -39.404, -35.371, -63.485, -63.281, -61.719, -59.024,
-55.456, -51.458, -47.394, -43.392, -39.418, -35.384, -63.511, -63.306, -61.744,
-59.047, -55.478, -51.478, -47.412, -43.409, -39.433, -35.396, -63.537, -63.332,
-61.769, -59.071, -55.500, -51.499, -47.431, -43.425, -39.448, -35.409, -63.563,
-63.358, -61.794, -59.095, -55.522, -51.519, -47.449, -43.442, -39.462, -35.422,
 -63.589, -63.384, -61.820, -59.119, -55.544, -51.539, -47.467, -43.458,
-39.477, -35.435, -63.615, -63.410, -61.845, -59.143, -55.567, -51.559, -47.486,
-43.475, -39.492, -35.448, -63.641, -63.436, -61.870, -59.167, -55.589, -51.580,
-47.504, -43.491, -39.506, -35.461, -63.667, -63.461, -61.895, -59.191, -55.611,
-51.600, -47.522, -43.507, -39.521, -35.474, -63.693, -63.487, -61.920, -59.215,
-55.633, -51.620, -47.540, -43.524, -39.536, -35.486, -63.719, -63.513, -61.945,
-59.238, -55.655, -51.641, -47.559, -43.540, -39.550, -35.499, -63.745, -63.539,
-61.970, -59.262, -55.678, -51.661, -47.577, -43.557, -39.565, -35.512, -63.771,
-63.565, -61.995, -59.286, -55.700, -51.681, -47.595, -43.573, -39.580, -35.525,
 -63.796, -63.590, -62.020, -59.310, -55.722, -51.701, -47.614, -43.590,
-39.594, -35.538, -63.822, -63.616, -62.045, -59.334, -55.744, -51.722, -47.632,
-43.606, -39.609, -35.551, -63.848, -63.642, -62.070, -59.358, -55.766, -51.742,
-47.650, -43.623, -39.624, -35.563, -63.874, -63.668, -62.095, -59.382, -55.789,
-51.762, -47.669, -43.639, -39.638, -35.576, -63.900, -63.694, -62.121, -59.406,
-55.811, -51.782, -47.687, -43.655, -39.653, -35.589, -63.926, -63.720, -62.146,
-59.429, -55.833, -51.803, -47.705, -43.672, -39.667, -35.602, -63.952, -63.745,
-62.171, -59.453, -55.855, -51.823, -47.724, -43.688, -39.682, -35.615, -63.978,
-63.771, -62.196, -59.477, -55.877, -51.843, -47.742, -43.705, -39.697, -35.628,
 -64.004, -63.797, -62.221, -59.501, -55.899, -51.863, -47.760, -43.721,
-39.711, -35.640, -64.030, -63.823, -62.246, -59.525, -55.922, -51.884, -47.778,
-43.738, -39.726, -35.653, -64.056, -63.849, -62.271, -59.549, -55.944, -51.904,
-47.797, -43.754, -39.741, -35.666, -64.082, -63.874, -62.296, -59.573, -55.966,
-51.924, -47.815, -43.770, -39.755, -35.679, -64.108, -63.900, -62.321, -59.597,
-55.988, -51.944, -47.833, -43.787, -39.770, -35.692, -64.134, -63.926, -62.346,
-59.620, -56.010, -51.965, -47.852, -43.803, -39.785, -35.705, -64.160, -63.952,
-62.371, -59.644, -56.033, -51.985, -47.870, -43.820, -39.799, -35.717, -64.185,
-63.978, -62.397, -59.668, -56.055, -52.005, -47.888, -43.836, -39.814, -35.730,
 -64.211, -64.003, -62.422, -59.692, -56.077, -52.025, -47.907, -43.853,
-39.828, -35.743, -64.237, -64.029, -62.447, -59.716, -56.099, -52.046, -47.925,
-43.869, -39.843, -35.756, -64.263, -64.055, -62.472, -59.740, -56.121, -52.066,
-47.943, -43.885, -39.858, -35.769, -64.289, -64.081, -62.497, -59.764, -56.144,
-52.086, -47.961, -43.902, -39.872, -35.782, -64.315, -64.107, -62.522, -59.788,
-56.166, -52.106, -47.980, -43.918, -39.887, -35.794, -64.341, -64.132, -62.547,
-59.811, -56.188, -52.127, -47.998, -43.935, -39.902, -35.807, -64.367, -64.158,
-62.572, -59.835, -56.210, -52.147, -48.016, -43.951, -39.916, -35.820, -64.393,
-64.184, -62.597, -59.859, -56.232, -52.167, -48.035, -43.968, -39.931, -35.833,
 -64.419, -64.210, -62.622, -59.883, -56.255, -52.187, -48.053, -43.984,
-39.946, -35.846, -64.445, -64.236, -62.647, -59.907, -56.277, -52.208, -48.071,
-44.000, -39.960, -35.858, -64.471, -64.262, -62.672, -59.931, -56.299, -52.228,
-48.090, -44.017, -39.975, -35.871, -64.497, -64.287, -62.697, -59.955, -56.321,
-52.248, -48.108, -44.033, -39.989, -35.884, -64.523, -64.313, -62.723, -59.979,
-56.343, -52.268, -48.126, -44.050, -40.004, -35.897, -64.549, -64.339, -62.748,
-60.002, -56.366, -52.289, -48.145, -44.066, -40.019, -35.910, -64.574, -64.365,
-62.773, -60.026, -56.388, -52.309, -48.163, -44.083, -40.033, -35.923, -64.600,
-64.391, -62.798, -60.050, -56.410, -52.329, -48.181, -44.099, -40.048, -35.935,
 -64.626, -64.416, -62.823, -60.074, -56.432, -52.349, -48.199, -44.115,
-40.063, -35.948, -64.652, -64.442, -62.848, -60.098, -56.454, -52.370, -48.218,
-44.132, -40.077, -35.961, -64.678, -64.468, -62.873, -60.122, -56.477, -52.390,
-48.236, -44.148, -40.092, -35.974, -64.704, -64.494, -62.898, -60.146, -56.499,
-52.410, -48.254, -44.165, -40.106, -35.987, -64.730, -64.520, -62.923, -60.169,
-56.521, -52.430, -48.273, -44.181, -40.121, -35.999, -64.756, -64.545, -62.948,
-60.193, -56.543, -52.451, -48.291, -44.198, -40.136, -36.012, -64.782, -64.571,
-62.973, -60.217, -56.565, -52.471, -48.309, -44.214, -40.150, -36.025, -64.808,
-64.597, -62.998, -60.241, -56.587, -52.491, -48.327, -44.230, -40.165, -36.038,
 -64.834, -64.623, -63.023, -60.265, -56.610, -52.511, -48.346, -44.247,
-40.180, -36.051, -64.860, -64.649, -63.049, -60.289, -56.632, -52.532, -48.364,
-44.263, -40.194, -36.063, -64.886, -64.674, -63.074, -60.313, -56.654, -52.552,
-48.382, -44.280, -40.209, -36.076, -64.912, -64.700, -63.099, -60.337, -56.676,
-52.572, -48.401, -44.296, -40.223, -36.089, -64.937, -64.726, -63.124, -60.360,
-56.698, -52.592, -48.419, -44.313, -40.238, -36.102, -64.963, -64.752, -63.149,
-60.384, -56.721, -52.613, -48.437, -44.329, -40.253, -36.115, -64.989, -64.778,
-63.174, -60.408, -56.743, -52.633, -48.456, -44.345, -40.267, -36.127, -65.015,
-64.803, -63.199, -60.432, -56.765, -52.653, -48.474, -44.362, -40.282, -36.140,
 -65.041, -64.829, -63.224, -60.456, -56.787, -52.673, -48.492, -44.378,
-40.297, -36.153, -65.067, -64.855, -63.249, -60.480, -56.809, -52.694, -48.510,
-44.395, -40.311, -36.166, -65.093, -64.881, -63.274, -60.504, -56.832, -52.714,
-48.529, -44.411, -40.326, -36.179, -65.119, -64.907, -63.299, -60.527, -56.854,
-52.734, -48.547, -44.428, -40.340, -36.191, -65.145, -64.932, -63.324, -60.551,
-56.876, -52.754, -48.565, -44.444, -40.355, -36.204, -65.171, -64.958, -63.349,
-60.575, -56.898, -52.775, -48.584, -44.460, -40.370, -36.217, -65.197, -64.984,
-63.375, -60.599, -56.920, -52.795, -48.602, -44.477, -40.384, -36.230, -65.223,
-65.010, -63.400, -60.623, -56.942, -52.815, -48.620, -44.493, -40.399, -36.243,
 -65.249, -65.036, -63.425, -60.647, -56.965, -52.835, -48.639, -44.510,
-40.413, -36.255, -65.275, -65.061, -63.450, -60.671, -56.987, -52.856, -48.657,
-44.526, -40.428, -36.268, -65.300, -65.087, -63.475, -60.694, -57.009, -52.876,
-48.675, -44.543, -40.443, -36.281, -65.326, -65.113, -63.500, -60.718, -57.031,
-52.896, -48.693, -44.559, -40.457, -36.294, -65.352, -65.139, -63.525, -60.742,
-57.053, -52.916, -48.712, -44.575, -40.472, -36.307, -65.378, -65.165, -63.550,
-60.766, -57.076, -52.937, -48.730, -44.592, -40.487, -36.319, -65.404, -65.190,
-63.575, -60.790, -57.098, -52.957, -48.748, -44.608, -40.501, -36.332, -65.430,
-65.216, -63.600, -60.814, -57.120, -52.977, -48.767, -44.625, -40.516, -36.345,
 -65.456, -65.242, -63.625, -60.838, -57.142, -52.997, -48.785, -44.641,
-40.530, -36.358, -65.482, -65.268, -63.650, -60.862, -57.164, -53.018, -48.803,
-44.657, -40.545, -36.371, -65.508, -65.294, -63.675, -60.885, -57.187, -53.038,
-48.821, -44.674, -40.560, -36.383, -65.534, -65.319, -63.700, -60.909, -57.209,
-53.058, -48.840, -44.690, -40.574, -36.396, -65.560, -65.345, -63.726, -60.933,
-57.231, -53.078, -48.858, -44.707, -40.589, -36.409, -65.586, -65.371, -63.751,
-60.957, -57.253, -53.099, -48.876, -44.723, -40.603, -36.422, -65.612, -65.397,
-63.776, -60.981, -57.275, -53.119, -48.895, -44.740, -40.618, -36.434, -65.637,
-65.423, -63.801, -61.005, -57.297, -53.139, -48.913, -44.756, -40.633, -36.447,
 -65.663, -65.448, -63.826, -61.029, -57.320, -53.159, -48.931, -44.772,
-40.647, -36.460, -65.689, -65.474, -63.851, -61.052, -57.342, -53.180, -48.949,
-44.789, -40.662, -36.473, -65.715, -65.500, -63.876, -61.076, -57.364, -53.200,
-48.968, -44.805, -40.676, -36.486, -65.741, -65.526, -63.901, -61.100, -57.386,
-53.220, -48.986, -44.822, -40.691, -36.498, -65.767, -65.552, -63.926, -61.124,
-57.408, -53.240, -49.004, -44.838, -40.706, -36.511, -65.793, -65.577, -63.951,
-61.148, -57.431, -53.260, -49.023, -44.854, -40.720, -36.524, -65.819, -65.603,
-63.976, -61.172, -57.453, -53.281, -49.041, -44.871, -40.735, -36.537, -65.845,
-65.629, -64.001, -61.196, -57.475, -53.301, -49.059, -44.887, -40.750, -36.550,
 -65.871, -65.655, -64.026, -61.219, -57.497, -53.321, -49.078, -44.904,
-40.764, -36.562, -65.897, -65.681, -64.051, -61.243, -57.519, -53.341, -49.096,
-44.920, -40.779, -36.575, -65.923, -65.706, -64.077, -61.267, -57.541, -53.362,
-49.114, -44.937, -40.793, -36.588, -65.949, -65.732, -64.102, -61.291, -57.564,
-53.382, -49.132, -44.953, -40.808, -36.601, -65.974, -65.758, -64.127, -61.315,
-57.586, -53.402, -49.151, -44.969, -40.823, -36.613, -66.000, -65.784, -64.152,
-61.339, -57.608, -53.422, -49.169, -44.986, -40.837, -36.626, -66.026, -65.810,
-64.177, -61.363, -57.630, -53.443, -49.187, -45.002, -40.852, -36.639, -66.052,
-65.835, -64.202, -61.386, -57.652, -53.463, -49.206, -45.019, -40.866, -36.652,
 -66.078, -65.861, -64.227, -61.410, -57.675, -53.483, -49.224, -45.035,
-40.881, -36.664, -66.104, -65.887, -64.252, -61.434, -57.697, -53.503, -49.242,
-45.051, -40.896, -36.677, -66.130, -65.913, -64.277, -61.458, -57.719, -53.524,
-49.260, -45.068, -40.910, -36.690, -66.156, -65.939, -64.302, -61.482, -57.741,
-53.544, -49.279, -45.084, -40.925, -36.703, -66.182, -65.964, -64.327, -61.506,
-57.763, -53.564, -49.297, -45.101, -40.939, -36.716, -66.208, -65.990, -64.352,
-61.529, -57.785, -53.584, -49.315, -45.117, -40.954, -36.728, -66.234, -66.016,
-64.377, -61.553, -57.808, -53.605, -49.334, -45.133, -40.969, -36.741, -66.260,
-66.042, -64.402, -61.577, -57.830, -53.625, -49.352, -45.150, -40.983, -36.754,
 -66.286, -66.068, -64.427, -61.601, -57.852, -53.645, -49.370, -45.166,
-40.998, -36.767, -66.311, -66.093, -64.453, -61.625, -57.874, -53.665, -49.388,
-45.183, -41.012, -36.779, -66.337, -66.119, -64.478, -61.649, -57.896, -53.686,
-49.407, -45.199, -41.027, -36.792, -66.363, -66.145, -64.503, -61.673, -57.918,
-53.706, -49.425, -45.215, -41.042, -36.805, -66.389, -66.171, -64.528, -61.696,
-57.941, -53.726, -49.443, -45.232, -41.056, -36.818, -66.415, -66.197, -64.553,
-61.720, -57.963, -53.746, -49.462, -45.248, -41.071, -36.830, -66.441, -66.222,
-64.578, -61.744, -57.985, -53.766, -49.480, -45.265, -41.085, -36.843, -66.467,
-66.248, -64.603, -61.768, -58.007, -53.787, -49.498, -45.281, -41.100, -36.856,
 -66.493, -66.274, -64.628, -61.792, -58.029, -53.807, -49.516, -45.298,
-41.114, -36.869, -66.519, -66.300, -64.653, -61.816, -58.052, -53.827, -49.535,
-45.314, -41.129, -36.881, -66.545, -66.326, -64.678, -61.840, -58.074, -53.847,
-49.553, -45.330, -41.144, -36.894, -66.571, -66.351, -64.703, -61.863, -58.096,
-53.868, -49.571, -45.347, -41.158, -36.907, -66.597, -66.377, -64.728, -61.887,
-58.118, -53.888, -49.590, -45.363, -41.173, -36.920, -66.622, -66.403, -64.753,
-61.911, -58.140, -53.908, -49.608, -45.380, -41.187, -36.932, -66.648, -66.429,
-64.778, -61.935, -58.162, -53.928, -49.626, -45.396, -41.202, -36.945, -66.674,
-66.455, -64.803, -61.959, -58.185, -53.949, -49.644, -45.412, -41.217, -36.958,
 -66.700, -66.480, -64.828, -61.983, -58.207, -53.969, -49.663, -45.429,
-41.231, -36.971, -66.726, -66.506, -64.854, -62.007, -58.229, -53.989, -49.681,
-45.445, -41.246, -36.983, -66.752, -66.532, -64.879, -62.030, -58.251, -54.009,
-49.699, -45.462, -41.260, -36.996, -66.778, -66.558, -64.904, -62.054, -58.273,
-54.029, -49.717, -45.478, -41.275, -37.009, -66.804, -66.583, -64.929, -62.078,
-58.295, -54.050, -49.736, -45.494, -41.290, -37.022, -66.830, -66.609, -64.954,
-62.102, -58.318, -54.070, -49.754, -45.511, -41.304, -37.034, -66.856, -66.635,
-64.979, -62.126, -58.340, -54.090, -49.772, -45.527, -41.319, -37.047, -66.882,
-66.661, -65.004, -62.150, -58.362, -54.110, -49.791, -45.544, -41.333, -37.060,
 -66.908, -66.687, -65.029, -62.173, -58.384, -54.131, -49.809, -45.560,
-41.348, -37.073, -66.933, -66.712, -65.054, -62.197, -58.406, -54.151, -49.827,
-45.576, -41.362, -37.085, -66.959, -66.738, -65.079, -62.221, -58.428, -54.171,
-49.845, -45.593, -41.377, -37.098, -66.985, -66.764, -65.104, -62.245, -58.451,
-54.191, -49.864, -45.609, -41.392, -37.111, -67.011, -66.790, -65.129, -62.269,
-58.473, -54.212, -49.882, -45.626, -41.406, -37.124, -67.037, -66.816, -65.154,
-62.293, -58.495, -54.232, -49.900, -45.642, -41.421, -37.136, -67.063, -66.841,
-65.179, -62.317, -58.517, -54.252, -49.919, -45.658, -41.435, -37.149, -67.089,
-66.867, -65.204, -62.340, -58.539, -54.272, -49.937, -45.675, -41.450, -37.162,
 -67.115, -66.893, -65.229, -62.364, -58.562, -54.293, -49.955, -45.691,
-41.465, -37.175, -67.141, -66.919, -65.254, -62.388, -58.584, -54.313, -49.973,
-45.708, -41.479, -37.187, -67.167, -66.945, -65.280, -62.412, -58.606, -54.333,
-49.992, -45.724, -41.494, -37.200, -67.193, -66.970, -65.305, -62.436, -58.628,
-54.353, -50.010, -45.740, -41.508, -37.213, -67.219, -66.996, -65.330, -62.460,
-58.650, -54.373, -50.028, -45.757, -41.523, -37.226, -67.244, -67.022, -65.355,
-62.483, -58.672, -54.394, -50.046, -45.773, -41.537, -37.238, -67.270, -67.048,
-65.380, -62.507, -58.695, -54.414, -50.065, -45.790, -41.552, -37.251, -67.296,
-67.073, -65.405, -62.531, -58.717, -54.434, -50.083, -45.806, -41.567, -37.264,
 -67.322, -67.099, -65.430, -62.555, -58.739, -54.454, -50.101, -45.822,
-41.581, -37.277, -67.348, -67.125, -65.455, -62.579, -58.761, -54.475, -50.120,
-45.839, -41.596, -37.289, -67.374, -67.151, -65.480, -62.603, -58.783, -54.495,
-50.138, -45.855, -41.610, -37.302, -67.400, -67.177, -65.505, -62.627, -58.805,
-54.515, -50.156, -45.872, -41.625, -37.315, -67.426, -67.202, -65.530, -62.650,
-58.828, -54.535, -50.174, -45.888, -41.639, -37.327, -67.452, -67.228, -65.555,
-62.674, -58.850, -54.555, -50.193, -45.904, -41.654, -37.340, -67.478, -67.254,
-65.580, -62.698, -58.872, -54.576, -50.211, -45.921, -41.669, -37.353, -67.504,
-67.280, -65.605, -62.722, -58.894, -54.596, -50.229, -45.937, -41.683, -37.366,
 -67.530, -67.306, -65.630, -62.746, -58.916, -54.616, -50.248, -45.954,
-41.698, -37.378, -67.555, -67.331, -65.655, -62.770, -58.938, -54.636, -50.266,
-45.970, -41.712, -37.391, -67.581, -67.357, -65.680, -62.793, -58.961, -54.657,
-50.284, -45.986, -41.727, -37.404, -67.607, -67.383, -65.705, -62.817, -58.983,
-54.677, -50.302, -46.003, -41.741, -37.417, -67.633, -67.409, -65.731, -62.841,
-59.005, -54.697, -50.321, -46.019, -41.756, -37.429, -67.659, -67.435, -65.756,
-62.865, -59.027, -54.717, -50.339, -46.036, -41.771, -37.442, -67.685, -67.460,
-65.781, -62.889, -59.049, -54.738, -50.357, -46.052, -41.785, -37.455, -67.711,
-67.486, -65.806, -62.913, -59.071, -54.758, -50.375, -46.068, -41.800, -37.467,
 -67.737, -67.512, -65.831, -62.936, -59.094, -54.778, -50.394, -46.085,
-41.814, -37.480, -67.763, -67.538, -65.856, -62.960, -59.116, -54.798, -50.412,
-46.101, -41.829, -37.493, -67.789, -67.563, -65.881, -62.984, -59.138, -54.818,
-50.430, -46.118, -41.843, -37.506, -67.815, -67.589, -65.906, -63.008, -59.160,
-54.839, -50.449, -46.134, -41.858, -37.518
};
*/

#endif  // H2
